/* Generated by Cython 3.0.11 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "define_macros": [
            [
                "HB_NO_MT",
                "1"
            ],
            [
                "HB_EXPERIMENTAL_API",
                "1"
            ],
            [
                "HAVE_MMAP",
                "1"
            ],
            [
                "HAVE_UNISTD_H",
                "1"
            ],
            [
                "HAVE_SYS_MMAN_H",
                "1"
            ]
        ],
        "depends": [
            "harfbuzz/src/hb-ot.h",
            "harfbuzz/src/hb-subset-repacker.h",
            "harfbuzz/src/hb-subset.h",
            "harfbuzz/src/hb.h"
        ],
        "extra_compile_args": [
            "-std=c++11"
        ],
        "include_dirs": [
            "harfbuzz/src"
        ],
        "language": "c++",
        "name": "uharfbuzz._harfbuzz",
        "sources": [
            "src/uharfbuzz/_harfbuzz.pyx",
            "harfbuzz/src/harfbuzz-subset.cc",
            "harfbuzz/src/hb-coretext.cc",
            "harfbuzz/src/hb-directwrite.cc",
            "harfbuzz/src/hb-uniscribe.cc"
        ]
    },
    "module_name": "uharfbuzz._harfbuzz"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#if defined(CYTHON_LIMITED_API) && 0
  #ifndef Py_LIMITED_API
    #if CYTHON_LIMITED_API+0 > 0x03030000
      #define Py_LIMITED_API CYTHON_LIMITED_API
    #else
      #define Py_LIMITED_API 0x03030000
    #endif
  #endif
#endif

#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02070000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.7+ or Python 3.3+.
#else
#if defined(CYTHON_LIMITED_API) && CYTHON_LIMITED_API
#define __PYX_EXTRA_ABI_MODULE_NAME "limited"
#else
#define __PYX_EXTRA_ABI_MODULE_NAME ""
#endif
#define CYTHON_ABI "3_0_11" __PYX_EXTRA_ABI_MODULE_NAME
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."
#define CYTHON_HEX_VERSION 0x03000BF0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #define HAVE_LONG_LONG
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1 && PYPY_VERSION_NUM >= 0x07030C00)
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_CLINE_IN_TRACEBACK
  #define CYTHON_CLINE_IN_TRACEBACK 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 1
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(Py_GIL_DISABLED) || defined(Py_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL 1
  #endif
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS 0
  #endif
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL (PY_MAJOR_VERSION < 3 || PY_VERSION_HEX >= 0x03060000 && PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL (PY_VERSION_HEX >= 0x030700A1)
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #if PY_VERSION_HEX < 0x030400a1
    #undef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #elif !defined(CYTHON_USE_TP_FINALIZE)
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #if PY_VERSION_HEX < 0x030600B1
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5)
  #endif
  #if PY_VERSION_HEX < 0x030700A3
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if !defined(CYTHON_VECTORCALL)
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL && PY_VERSION_HEX >= 0x030800B1)
#endif
#define CYTHON_BACKPORT_VECTORCALL (CYTHON_METH_FASTCALL && PY_VERSION_HEX < 0x030800B1)
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
            typedef unsigned char     uint8_t;
            typedef unsigned short    uint16_t;
            typedef unsigned int      uint32_t;
        #else
            typedef unsigned __int8   uint8_t;
            typedef unsigned __int16  uint16_t;
            typedef unsigned __int32  uint32_t;
        #endif
    #endif
    #if _MSC_VER < 1300
        #ifdef _WIN64
            typedef unsigned long long  __pyx_uintptr_t;
        #else
            typedef unsigned int        __pyx_uintptr_t;
        #endif
    #else
        #ifdef _WIN64
            typedef unsigned __int64    __pyx_uintptr_t;
        #else
            typedef unsigned __int32    __pyx_uintptr_t;
        #endif
    #endif
#else
    #include <stdint.h>
    typedef uintptr_t  __pyx_uintptr_t;
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(const U& other) const { return *ptr == other; }
    template<typename U> bool operator !=(const U& other) const { return *ptr != other; }
    template<typename U> bool operator==(const __Pyx_FakeReference<U>& other) const { return *ptr == *other.ptr; }
    template<typename U> bool operator!=(const __Pyx_FakeReference<U>& other) const { return *ptr != *other.ptr; }
  private:
    T *ptr;
};

#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_DefaultClassType PyClass_Type
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    static CYTHON_INLINE PyObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *exception_table = NULL;
        PyObject *types_module=NULL, *code_type=NULL, *result=NULL;
        #if __PYX_LIMITED_VERSION_HEX < 0x030B0000
        PyObject *version_info;
        PyObject *py_minor_version = NULL;
        #endif
        long minor_version = 0;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        #if __PYX_LIMITED_VERSION_HEX >= 0x030B0000
        minor_version = 11;
        #else
        if (!(version_info = PySys_GetObject("version_info"))) goto end;
        if (!(py_minor_version = PySequence_GetItem(version_info, 1))) goto end;
        minor_version = PyLong_AsLong(py_minor_version);
        Py_DECREF(py_minor_version);
        if (minor_version == -1 && PyErr_Occurred()) goto end;
        #endif
        if (!(types_module = PyImport_ImportModule("types"))) goto end;
        if (!(code_type = PyObject_GetAttrString(types_module, "CodeType"))) goto end;
        if (minor_version <= 7) {
            (void)p;
            result = PyObject_CallFunction(code_type, "iiiiiOOOOOOiOO", a, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else if (minor_version <= 10) {
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOiOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else {
            if (!(exception_table = PyBytes_FromStringAndSize(NULL, 0))) goto end;
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOOiOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, name, fline, lnos, exception_table, fv, cell);
        }
    end:
        Py_XDECREF(code_type);
        Py_XDECREF(exception_table);
        Py_XDECREF(types_module);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
    #ifndef CO_OPTIMIZED
    #define CO_OPTIMIZED 0x0001
    #endif
    #ifndef CO_NEWLOCALS
    #define CO_NEWLOCALS 0x0002
    #endif
    #ifndef CO_VARARGS
    #define CO_VARARGS 0x0004
    #endif
    #ifndef CO_VARKEYWORDS
    #define CO_VARKEYWORDS 0x0008
    #endif
    #ifndef CO_ASYNC_GENERATOR
    #define CO_ASYNC_GENERATOR 0x0200
    #endif
    #ifndef CO_GENERATOR
    #define CO_GENERATOR 0x0020
    #endif
    #ifndef CO_COROUTINE
    #define CO_COROUTINE 0x0080
    #endif
#elif PY_VERSION_HEX >= 0x030B0000
  static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
    PyCodeObject *result;
    PyObject *empty_bytes = PyBytes_FromStringAndSize("", 0);
    if (!empty_bytes) return NULL;
    result =
      #if PY_VERSION_HEX >= 0x030C0000
        PyUnstable_Code_NewWithPosOnlyArgs
      #else
        PyCode_NewWithPosOnlyArgs
      #endif
        (a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, name, fline, lnos, empty_bytes);
    Py_DECREF(empty_bytes);
    return result;
  }
#elif PY_VERSION_HEX >= 0x030800B2 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
#endif
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef CO_COROUTINE
  #define CO_COROUTINE 0x80
#endif
#ifndef CO_ASYNC_GENERATOR
  #define CO_ASYNC_GENERATOR 0x200
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #if PY_VERSION_HEX >= 0x030d00A4
  #  define __Pyx_PyCFunctionFast PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords PyCFunctionFastWithKeywords
  #else
  #  define __Pyx_PyCFunctionFast _PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
  #endif
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#elif CYTHON_BACKPORT_VECTORCALL
  typedef PyObject *(*__pyx_vectorcallfunc)(PyObject *callable, PyObject *const *args,
                                            size_t nargsf, PyObject *kwnames);
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  ((size_t)1 << (8 * sizeof(size_t) - 1))
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(((size_t)(n)) & ~__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_MAJOR_VERSION >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void *cfunc) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if __PYX_LIMITED_VERSION_HEX < 0x030900B1
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE void *__Pyx_PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE(obj), name, func_ctype)
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if PY_MAJOR_VERSION < 3
    #if CYTHON_COMPILING_IN_PYPY
        #if PYPY_VERSION_NUM < 0x07030600
            #if defined(__cplusplus) && __cplusplus >= 201402L
                [[deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")]]
            #elif defined(__GNUC__) || defined(__clang__)
                __attribute__ ((__deprecated__("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")))
            #elif defined(_MSC_VER)
                __declspec(deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6"))
            #endif
            static CYTHON_INLINE int PyGILState_Check(void) {
                return 0;
            }
        #else  // PYPY_VERSION_NUM < 0x07030600
        #endif  // PYPY_VERSION_NUM < 0x07030600
    #else
        static CYTHON_INLINE int PyGILState_Check(void) {
            PyThreadState * tstate = _PyThreadState_Current;
            return tstate && (tstate == PyGILState_GetThisThreadState());
        }
    #endif
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000 || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030600B4 && PY_VERSION_HEX < 0x030d0000 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif PY_MAJOR_VERSION >= 3 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000)
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  (Py_TYPE(obj)->tp_iternext)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  PyIter_Next
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_SetItemOnTypeDict(tp, k, v) PyObject_GenericSetAttr((PyObject*)tp, k, v)
#else
  #define __Pyx_SetItemOnTypeDict(tp, k, v) PyDict_SetItem(tp->tp_dict, k, v)
#endif
#if CYTHON_USE_TYPE_SPECS && PY_VERSION_HEX >= 0x03080000
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GetLength(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#elif PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535U : 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((int)sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = (Py_UNICODE) ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains) || (PY_MAJOR_VERSION == 2 && PYPY_VERSION_NUM < 0x07030500)
    #undef PyUnicode_Contains
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#if CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static CYTHON_INLINE PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define __Pyx_Py3Int_Check(op)       PyLong_Check(op)
  #define __Pyx_Py3Int_CheckExact(op)  PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#else
  #define __Pyx_Py3Int_Check(op)       (PyLong_Check(op) || PyInt_Check(op))
  #define __Pyx_Py3Int_CheckExact(op)  (PyLong_CheckExact(op) || PyInt_CheckExact(op))
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__;  (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
    #define __PYX_EXTERN_C extern "C++"
#endif

#define __PYX_HAVE__uharfbuzz___harfbuzz
#define __PYX_HAVE_API__uharfbuzz___harfbuzz
/* Early includes */
#include <stdint.h>
#include "hb.h"
#include "hb-ot.h"
#include "hb-subset-repacker.h"
#include "hb-subset.h"
#include <string.h>
#include <stdlib.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
#include <string.h>
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = (char) c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#include <string.h>
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* #### Code section: filename_table ### */

static const char *__pyx_f[] = {
  "src/uharfbuzz/_harfbuzz.pyx",
  "<stringsource>",
};
/* #### Code section: utility_code_proto_before_types ### */
/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* #### Code section: numeric_typedefs ### */
/* #### Code section: complex_type_declarations ### */
/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphInfo;
struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition;
struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer;
struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob;
struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face;
struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font;
struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs;
struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine;
struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs;
struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs;
struct __pyx_obj_9uharfbuzz_9_harfbuzz_HBObject;
struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput;
struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetPlan;
struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set;
struct __pyx_obj_9uharfbuzz_9_harfbuzz_SetIter;
struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map;
struct __pyx_obj_9uharfbuzz_9_harfbuzz_MapIter;
struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr;
struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr;
struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr;
struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr;
struct __pyx_t_9uharfbuzz_9_harfbuzz__pen_methods;
struct __pyx_defaults;
typedef struct __pyx_defaults __pyx_defaults;
struct __pyx_defaults1;
typedef struct __pyx_defaults1 __pyx_defaults1;

/* "uharfbuzz/_harfbuzz.pyx":897
 *         hb_font_draw_glyph(self._hb_font, gid, drawfuncs, <void*>&methods)
 * 
 * cdef struct _pen_methods:             # <<<<<<<<<<<<<<
 *     void *moveTo
 *     void *lineTo
 */
struct __pyx_t_9uharfbuzz_9_harfbuzz__pen_methods {
  void *moveTo;
  void *lineTo;
  void *curveTo;
  void *qCurveTo;
  void *closePath;
};
struct __pyx_defaults {
  PyObject *__pyx_arg_script_index;
  PyObject *__pyx_arg_language_index;
};
struct __pyx_defaults1 {
  PyObject *__pyx_arg_script_index;
};

/* "uharfbuzz/_harfbuzz.pyx":56
 *     SAFE_TO_INSERT_TATWEEL = HB_GLYPH_FLAG_SAFE_TO_INSERT_TATWEEL
 * 
 * cdef class GlyphInfo:             # <<<<<<<<<<<<<<
 * 
 *     cdef hb_glyph_info_t _hb_glyph_info
 */
struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphInfo {
  PyObject_HEAD
  struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_GlyphInfo *__pyx_vtab;
  hb_glyph_info_t _hb_glyph_info;
};


/* "uharfbuzz/_harfbuzz.pyx":77
 * 
 * 
 * cdef class GlyphPosition:             # <<<<<<<<<<<<<<
 *     cdef hb_glyph_position_t _hb_glyph_position
 *     # could maybe store Buffer to prevent GC
 */
struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition {
  PyObject_HEAD
  struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_GlyphPosition *__pyx_vtab;
  hb_glyph_position_t _hb_glyph_position;
};


/* "uharfbuzz/_harfbuzz.pyx":132
 *     GLYPHS = HB_BUFFER_CONTENT_TYPE_GLYPHS
 * 
 * cdef class Buffer:             # <<<<<<<<<<<<<<
 *     cdef hb_buffer_t* _hb_buffer
 *     cdef object _message_callback
 */
struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer {
  PyObject_HEAD
  hb_buffer_t *_hb_buffer;
  PyObject *_message_callback;
};


/* "uharfbuzz/_harfbuzz.pyx":375
 * 
 * 
 * cdef class Blob:             # <<<<<<<<<<<<<<
 *     cdef hb_blob_t* _hb_blob
 *     cdef object _data
 */
struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob {
  PyObject_HEAD
  struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_Blob *__pyx_vtab;
  hb_blob_t *_hb_blob;
  PyObject *_data;
};


/* "uharfbuzz/_harfbuzz.pyx":444
 * 
 * 
 * cdef class Face:             # <<<<<<<<<<<<<<
 *     cdef hb_face_t* _hb_face
 *     cdef object _reference_table_func
 */
struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face {
  PyObject_HEAD
  struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_Face *__pyx_vtab;
  hb_face_t *_hb_face;
  PyObject *_reference_table_func;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *_blob;
};


/* "uharfbuzz/_harfbuzz.pyx":570
 * )
 * 
 * cdef class Font:             # <<<<<<<<<<<<<<
 *     cdef hb_font_t* _hb_font
 *     # GC bookkeeping
 */
struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font {
  PyObject_HEAD
  struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_Font *__pyx_vtab;
  hb_font_t *_hb_font;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *_face;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *_ffuncs;
};


/* "uharfbuzz/_harfbuzz.pyx":1049
 * 
 * 
 * cdef class FontFuncs:             # <<<<<<<<<<<<<<
 *     cdef hb_font_funcs_t* _hb_ffuncs
 *     cdef object _glyph_h_advance_func
 */
struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs {
  PyObject_HEAD
  hb_font_funcs_t *_hb_ffuncs;
  PyObject *_glyph_h_advance_func;
  PyObject *_glyph_v_advance_func;
  PyObject *_glyph_v_origin_func;
  PyObject *_glyph_name_func;
  PyObject *_nominal_glyph_func;
  PyObject *_variation_glyph_func;
  PyObject *_font_h_extents_func;
  PyObject *_font_v_extents_func;
};


/* "uharfbuzz/_harfbuzz.pyx":1684
 * 
 * 
 * cdef class ColorLine:             # <<<<<<<<<<<<<<
 *     cdef hb_color_line_t* _color_line
 * 
 */
struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine {
  PyObject_HEAD
  struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_ColorLine *__pyx_vtab;
  hb_color_line_t *_color_line;
};


/* "uharfbuzz/_harfbuzz.pyx":1896
 * 
 * 
 * cdef class PaintFuncs:             # <<<<<<<<<<<<<<
 *     cdef hb_paint_funcs_t* _hb_paintfuncs
 *     cdef object _push_transform_func
 */
struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs {
  PyObject_HEAD
  hb_paint_funcs_t *_hb_paintfuncs;
  PyObject *_push_transform_func;
  PyObject *_pop_transform_func;
  PyObject *_color_glyph_func;
  PyObject *_push_clip_glyph_func;
  PyObject *_push_clip_rectangle_func;
  PyObject *_pop_clip_func;
  PyObject *_color_func;
  PyObject *_image_func;
  PyObject *_linear_gradient_func;
  PyObject *_radial_gradient_func;
  PyObject *_sweep_gradient_func;
  PyObject *_push_group_func;
  PyObject *_pop_group_func;
  PyObject *_custom_palette_color_func;
};


/* "uharfbuzz/_harfbuzz.pyx":2123
 * 
 * 
 * cdef class DrawFuncs:             # <<<<<<<<<<<<<<
 *     cdef hb_draw_funcs_t* _hb_drawfuncs
 *     cdef object _move_to_func
 */
struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs {
  PyObject_HEAD
  hb_draw_funcs_t *_hb_drawfuncs;
  PyObject *_move_to_func;
  PyObject *_line_to_func;
  PyObject *_cubic_to_func;
  PyObject *_quadratic_to_func;
  PyObject *_close_path_func;
  int _warned;
};


/* "uharfbuzz/_harfbuzz.pyx":2281
 *             self._hb_drawfuncs, func_p, user_data_p, NULL)
 * 
 * cdef class HBObject:             # <<<<<<<<<<<<<<
 *     cdef hb_object_t* _hb_obj_list
 *     cdef unsigned int _num
 */
struct __pyx_obj_9uharfbuzz_9_harfbuzz_HBObject {
  PyObject_HEAD
  struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_HBObject *__pyx_vtab;
  hb_object_t *_hb_obj_list;
  unsigned int _num;
};


/* "uharfbuzz/_harfbuzz.pyx":2470
 * 
 * 
 * cdef class SubsetInput:             # <<<<<<<<<<<<<<
 *     cdef hb_subset_input_t* _hb_input
 * 
 */
struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput {
  PyObject_HEAD
  hb_subset_input_t *_hb_input;
};


/* "uharfbuzz/_harfbuzz.pyx":2545
 * 
 * 
 * cdef class SubsetPlan:             # <<<<<<<<<<<<<<
 *     cdef hb_subset_plan_t* _hb_plan
 * 
 */
struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetPlan {
  PyObject_HEAD
  hb_subset_plan_t *_hb_plan;
};


/* "uharfbuzz/_harfbuzz.pyx":2576
 * 
 * 
 * cdef class Set:             # <<<<<<<<<<<<<<
 *     cdef hb_set_t* _hb_set
 * 
 */
struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set {
  PyObject_HEAD
  struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_Set *__pyx_vtab;
  hb_set_t *_hb_set;
};


/* "uharfbuzz/_harfbuzz.pyx":2744
 *         return ("Set({%s})" % s)
 * 
 * cdef class SetIter:             # <<<<<<<<<<<<<<
 *     cdef Set s
 *     cdef hb_set_t *_hb_set
 */
struct __pyx_obj_9uharfbuzz_9_harfbuzz_SetIter {
  PyObject_HEAD
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *s;
  hb_set_t *_hb_set;
  hb_codepoint_t _c;
};


/* "uharfbuzz/_harfbuzz.pyx":2764
 * 
 * 
 * cdef class Map:             # <<<<<<<<<<<<<<
 *     cdef hb_map_t* _hb_map
 * 
 */
struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map {
  PyObject_HEAD
  struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_Map *__pyx_vtab;
  hb_map_t *_hb_map;
};


/* "uharfbuzz/_harfbuzz.pyx":2872
 *         return ("Map({%s})" % s)
 * 
 * cdef class MapIter:             # <<<<<<<<<<<<<<
 *     cdef Map m
 *     cdef hb_map_t *_hb_map
 */
struct __pyx_obj_9uharfbuzz_9_harfbuzz_MapIter {
  PyObject_HEAD
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *m;
  hb_map_t *_hb_map;
  int _i;
};


/* "uharfbuzz/_harfbuzz.pyx":2741
 *         if self.is_inverted():
 *             return "Set({...})"
 *         s = ', '.join(repr(v) for v in self)             # <<<<<<<<<<<<<<
 *         return ("Set({%s})" % s)
 * 
 */
struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr {
  PyObject_HEAD
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_genexpr_arg_0;
  PyObject *__pyx_v_v;
};


/* "uharfbuzz/_harfbuzz.pyx":2860
 * 
 *     def keys(self):
 *         return (k for k,v in self.items())             # <<<<<<<<<<<<<<
 * 
 *     def values(self):
 */
struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_genexpr_arg_0;
  PyObject *__pyx_v_k;
  PyObject *__pyx_v_v;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
};


/* "uharfbuzz/_harfbuzz.pyx":2863
 * 
 *     def values(self):
 *         return (v for k,v in self.items())             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr {
  PyObject_HEAD
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_genexpr_arg_0;
  PyObject *__pyx_v_k;
  PyObject *__pyx_v_v;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
};


/* "uharfbuzz/_harfbuzz.pyx":2869
 * 
 *     def __repr__(self):
 *         s = ', '.join("%s: %s" % (repr(k), repr(v)) for k,v in sorted(self.items()))             # <<<<<<<<<<<<<<
 *         return ("Map({%s})" % s)
 * 
 */
struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr {
  PyObject_HEAD
  PyObject *__pyx_genexpr_arg_0;
  PyObject *__pyx_v_k;
  PyObject *__pyx_v_v;
};



/* "uharfbuzz/_harfbuzz.pyx":56
 *     SAFE_TO_INSERT_TATWEEL = HB_GLYPH_FLAG_SAFE_TO_INSERT_TATWEEL
 * 
 * cdef class GlyphInfo:             # <<<<<<<<<<<<<<
 * 
 *     cdef hb_glyph_info_t _hb_glyph_info
 */

struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_GlyphInfo {
  PyObject *(*set)(struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphInfo *, hb_glyph_info_t);
};
static struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_GlyphInfo *__pyx_vtabptr_9uharfbuzz_9_harfbuzz_GlyphInfo;


/* "uharfbuzz/_harfbuzz.pyx":77
 * 
 * 
 * cdef class GlyphPosition:             # <<<<<<<<<<<<<<
 *     cdef hb_glyph_position_t _hb_glyph_position
 *     # could maybe store Buffer to prevent GC
 */

struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_GlyphPosition {
  PyObject *(*set)(struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition *, hb_glyph_position_t);
};
static struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_GlyphPosition *__pyx_vtabptr_9uharfbuzz_9_harfbuzz_GlyphPosition;


/* "uharfbuzz/_harfbuzz.pyx":375
 * 
 * 
 * cdef class Blob:             # <<<<<<<<<<<<<<
 *     cdef hb_blob_t* _hb_blob
 *     cdef object _data
 */

struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_Blob {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *(*from_ptr)(hb_blob_t *);
};
static struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_Blob *__pyx_vtabptr_9uharfbuzz_9_harfbuzz_Blob;


/* "uharfbuzz/_harfbuzz.pyx":444
 * 
 * 
 * cdef class Face:             # <<<<<<<<<<<<<<
 *     cdef hb_face_t* _hb_face
 *     cdef object _reference_table_func
 */

struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_Face {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *(*from_ptr)(hb_face_t *);
};
static struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_Face *__pyx_vtabptr_9uharfbuzz_9_harfbuzz_Face;


/* "uharfbuzz/_harfbuzz.pyx":570
 * )
 * 
 * cdef class Font:             # <<<<<<<<<<<<<<
 *     cdef hb_font_t* _hb_font
 *     # GC bookkeeping
 */

struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_Font {
  PyObject *(*_Font__create)(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *);
  PyObject *(*_Font__create_sub_font)(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *);
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *(*from_ptr)(hb_font_t *);
};
static struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_Font *__pyx_vtabptr_9uharfbuzz_9_harfbuzz_Font;


/* "uharfbuzz/_harfbuzz.pyx":1684
 * 
 * 
 * cdef class ColorLine:             # <<<<<<<<<<<<<<
 *     cdef hb_color_line_t* _color_line
 * 
 */

struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_ColorLine {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine *(*from_ptr)(hb_color_line_t *);
};
static struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_ColorLine *__pyx_vtabptr_9uharfbuzz_9_harfbuzz_ColorLine;


/* "uharfbuzz/_harfbuzz.pyx":2281
 *             self._hb_drawfuncs, func_p, user_data_p, NULL)
 * 
 * cdef class HBObject:             # <<<<<<<<<<<<<<
 *     cdef hb_object_t* _hb_obj_list
 *     cdef unsigned int _num
 */

struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_HBObject {
  PyObject *(*update_obj_length)(struct __pyx_obj_9uharfbuzz_9_harfbuzz_HBObject *, unsigned int, char *, char *);
  hb_link_t *(*create_links)(struct __pyx_obj_9uharfbuzz_9_harfbuzz_HBObject *, unsigned int, unsigned int, int);
  PyObject *(*update_links)(struct __pyx_obj_9uharfbuzz_9_harfbuzz_HBObject *, unsigned int, int, PyObject *);
};
static struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_HBObject *__pyx_vtabptr_9uharfbuzz_9_harfbuzz_HBObject;


/* "uharfbuzz/_harfbuzz.pyx":2576
 * 
 * 
 * cdef class Set:             # <<<<<<<<<<<<<<
 *     cdef hb_set_t* _hb_set
 * 
 */

struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_Set {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *(*from_ptr)(hb_set_t *);
};
static struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_Set *__pyx_vtabptr_9uharfbuzz_9_harfbuzz_Set;


/* "uharfbuzz/_harfbuzz.pyx":2764
 * 
 * 
 * cdef class Map:             # <<<<<<<<<<<<<<
 *     cdef hb_map_t* _hb_map
 * 
 */

struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_Map {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *(*from_ptr)(hb_map_t *);
};
static struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_Map *__pyx_vtabptr_9uharfbuzz_9_harfbuzz_Map;
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__))
  #define __Pyx_RefNannyFinishContextNogil() __Pyx_RefNannyFinishContext()
#endif
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* IncludeStringH.proto */
#include <string.h>

/* decode_c_string_utf16.proto */
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 0;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16LE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = -1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16BE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}

/* decode_c_string.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#if !CYTHON_VECTORCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if !CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03080000
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()
  #define __Pyx_PyFrame_GetLocalsplus(frame)  ((frame)->f_localsplus)
#else
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif
#endif
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs);

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* TupleAndListFromArray.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_Arg_VARARGS(args, i) PySequence_GetItem(args, i)
#elif CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_Arg_VARARGS(args, i) PyTuple_GET_ITEM(args, i)
#else
    #define __Pyx_Arg_VARARGS(args, i) PyTuple_GetItem(args, i)
#endif
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_Arg_NewRef_VARARGS(arg) __Pyx_NewRef(arg)
    #define __Pyx_Arg_XDECREF_VARARGS(arg) Py_XDECREF(arg)
#else
    #define __Pyx_Arg_NewRef_VARARGS(arg) arg
    #define __Pyx_Arg_XDECREF_VARARGS(arg)
#endif
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_Arg_FASTCALL(args, i) args[i]
    #define __Pyx_NumKwargs_FASTCALL(kwds) PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
    CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues);
  #else
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
  #endif
    #define __Pyx_Arg_NewRef_FASTCALL(arg) arg  /* no-op, __Pyx_Arg_FASTCALL is direct and this needs
                                                   to have the same reference counting */
    #define __Pyx_Arg_XDECREF_FASTCALL(arg)
#else
    #define __Pyx_Arg_FASTCALL __Pyx_Arg_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
    #define __Pyx_Arg_NewRef_FASTCALL(arg) __Pyx_Arg_NewRef_VARARGS(arg)
    #define __Pyx_Arg_XDECREF_FASTCALL(arg) __Pyx_Arg_XDECREF_VARARGS(arg)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_VARARGS(args, start), stop - start)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_FASTCALL(args, start), stop - start)
#else
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* decode_c_bytes.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_bytes(
         const char* cstring, Py_ssize_t length, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* decode_bytes.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_bytes(
         PyObject* string, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    char* as_c_string;
    Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS
    as_c_string = PyBytes_AS_STRING(string);
    size = PyBytes_GET_SIZE(string);
#else
    if (PyBytes_AsStringAndSize(string, &as_c_string, &size) < 0) {
        return NULL;
    }
#endif
    return __Pyx_decode_c_bytes(
        as_c_string, size,
        start, stop, encoding, errors, decode_func);
}

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kw, const char* function_name, int kw_allowed);

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,
    const char* function_name);

/* PyObjectCallNoArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely(__Pyx_IS_TYPE(obj, type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* AssertionsEnabled.proto */
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define __Pyx_init_assertions_enabled()  (0)
  #define __pyx_assertions_enabled()  (1)
#elif CYTHON_COMPILING_IN_LIMITED_API  ||  (CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030C0000)
  static int __pyx_assertions_enabled_flag;
  #define __pyx_assertions_enabled() (__pyx_assertions_enabled_flag)
  static int __Pyx_init_assertions_enabled(void) {
    PyObject *builtins, *debug, *debug_str;
    int flag;
    builtins = PyEval_GetBuiltins();
    if (!builtins) goto bad;
    debug_str = PyUnicode_FromStringAndSize("__debug__", 9);
    if (!debug_str) goto bad;
    debug = PyObject_GetItem(builtins, debug_str);
    Py_DECREF(debug_str);
    if (!debug) goto bad;
    flag = PyObject_IsTrue(debug);
    Py_DECREF(debug);
    if (flag == -1) goto bad;
    __pyx_assertions_enabled_flag = flag;
    return 0;
  bad:
    __pyx_assertions_enabled_flag = 1;
    return -1;
  }
#else
  #define __Pyx_init_assertions_enabled()  (0)
  #define __pyx_assertions_enabled()  (!Py_OptimizeFlag)
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* RaiseUnexpectedTypeError.proto */
static int __Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj);

/* PyObjectFormatSimple.proto */
#if CYTHON_COMPILING_IN_PYPY
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#elif PY_MAJOR_VERSION < 3
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyString_CheckExact(s)) ? PyUnicode_FromEncodedObject(s, NULL, "strict") :\
        PyObject_Format(s, f))
#elif CYTHON_USE_TYPE_SLOTS
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyLong_CheckExact(s)) ? PyLong_Type.tp_repr(s) :\
        likely(PyFloat_CheckExact(s)) ? PyFloat_Type.tp_repr(s) :\
        PyObject_Format(s, f))
#else
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#endif

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_MultiplyObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_MultiplyObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceMultiply(op1, op2) : PyNumber_Multiply(op1, op2))
#endif

/* SliceTupleAndList.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_GetSlice(PyObject* src, Py_ssize_t start, Py_ssize_t stop);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_GetSlice(PyObject* src, Py_ssize_t start, Py_ssize_t stop);
#else
#define __Pyx_PyList_GetSlice(seq, start, stop)   PySequence_GetSlice(seq, start, stop)
#define __Pyx_PyTuple_GetSlice(seq, start, stop)  PySequence_GetSlice(seq, start, stop)
#endif

/* RaiseUnboundLocalError.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* JoinPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* GCCDiagnostics.proto */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* BuildPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char);

/* CIntToPyUnicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char, char format_char);

/* PyObjectCall2Args.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* StringJoin.proto */
#if PY_MAJOR_VERSION < 3
#define __Pyx_PyString_Join __Pyx_PyBytes_Join
#define __Pyx_PyBaseString_Join(s, v) (PyUnicode_CheckExact(s) ? PyUnicode_Join(s, v) : __Pyx_PyBytes_Join(s, v))
#else
#define __Pyx_PyString_Join PyUnicode_Join
#define __Pyx_PyBaseString_Join PyUnicode_Join
#endif
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* pep479.proto */
static void __Pyx_Generator_Replace_StopIteration(int in_async_gen);

/* PyObjectFormatAndDecref.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f);
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f);

/* IncludeStructmemberH.proto */
#include <structmember.h>

/* FixUpExtensionType.proto */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);
#endif

/* ValidateBasesTuple.proto */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases);
#endif

/* PyType_Ready.proto */
CYTHON_UNUSED static int __Pyx_PyType_Ready(PyTypeObject *t);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyTypeObject* typeptr , void* vtable);

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyTypeObject *type);

/* MergeVTables.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_MergeVtables(PyTypeObject *type);
#endif

/* SetupReduce.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce(PyObject* type_obj);
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportDottedModule.proto */
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple);
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_ImportDottedModule_WalkParts(PyObject *module, PyObject *name, PyObject *parts_tuple);
#endif

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* FetchSharedCythonModule.proto */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* FetchCommonType.proto */
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);
#else
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases);
#endif

/* PyMethodNew.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    PyObject *typesModule=NULL, *methodType=NULL, *result=NULL;
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    typesModule = PyImport_ImportModule("types");
    if (!typesModule) return NULL;
    methodType = PyObject_GetAttrString(typesModule, "MethodType");
    Py_DECREF(typesModule);
    if (!methodType) return NULL;
    result = PyObject_CallFunctionObjArgs(methodType, func, self, NULL);
    Py_DECREF(methodType);
    return result;
}
#elif PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#else
    #define __Pyx_PyMethod_New PyMethod_New
#endif

/* PyVectorcallFastCallDict.proto */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject_HEAD
    PyObject *func;
#elif PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_BACKPORT_VECTORCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_classobj;
#endif
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#undef __Pyx_CyOrPyCFunction_Check
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_CyFunctionType)
#define __Pyx_CyOrPyCFunction_Check(obj)  __Pyx_TypeCheck2(obj, __pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_CyFunctionType)
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc);
#undef __Pyx_IsSameCFunction
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCyOrCFunction(func, cfunc)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_BACKPORT_VECTORCALL
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* Py3UpdateBases.proto */
static PyObject* __Pyx_PEP560_update_bases(PyObject *bases);

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* PyObjectLookupSpecial.proto */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_LookupSpecialNoError(obj, attr_name)  __Pyx__PyObject_LookupSpecial(obj, attr_name, 0)
#define __Pyx_PyObject_LookupSpecial(obj, attr_name)  __Pyx__PyObject_LookupSpecial(obj, attr_name, 1)
static CYTHON_INLINE PyObject* __Pyx__PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name, int with_error);
#else
#define __Pyx_PyObject_LookupSpecialNoError(o,n) __Pyx_PyObject_GetAttrStrNoError(o,n)
#define __Pyx_PyObject_LookupSpecial(o,n) __Pyx_PyObject_GetAttrStr(o,n)
#endif

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* SetNameInClass.proto */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && PY_VERSION_HEX < 0x030d0000
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? _PyDict_SetItem_KnownHash(ns, name, value, ((PyASCIIObject *) name)->hash) : PyObject_SetItem(ns, name, value))
#elif CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? PyDict_SetItem(ns, name, value) : PyObject_SetItem(ns, name, value))
#else
#define __Pyx_SetNameInClass(ns, name, value)  PyObject_SetItem(ns, name, value)
#endif

/* ClassMethod.proto */
#include "descrobject.h"
CYTHON_UNUSED static PyObject* __Pyx_Method_ClassMethod(PyObject *method);

/* GetNameInClass.proto */
#define __Pyx_GetNameInClass(var, nmspace, name)  (var) = __Pyx__GetNameInClass(nmspace, name)
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);
#endif

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* None.proto */
#include <new>

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_glyph_flags_t(hb_glyph_flags_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_buffer_flags_t(hb_buffer_flags_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_buffer_cluster_level_t(hb_buffer_cluster_level_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_buffer_content_type_t(hb_buffer_content_type_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_codepoint_t(hb_codepoint_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE hb_codepoint_t __Pyx_PyInt_As_hb_codepoint_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_ot_layout_glyph_class_t(hb_ot_layout_glyph_class_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_ot_color_palette_flags_t(hb_ot_color_palette_flags_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_ot_math_constant_t(hb_ot_math_constant_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_ot_math_kern_t(hb_ot_math_kern_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_ot_math_glyph_part_flags_t(hb_ot_math_glyph_part_flags_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_paint_composite_mode_t(hb_paint_composite_mode_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_paint_extend_t(hb_paint_extend_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_subset_sets_t(hb_subset_sets_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_subset_flags_t(hb_subset_flags_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_long(unsigned long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_mask_t(hb_mask_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_position_t(hb_position_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_int(unsigned int value);

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned int __Pyx_PyInt_As_unsigned_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE hb_buffer_flags_t __Pyx_PyInt_As_hb_buffer_flags_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE hb_buffer_cluster_level_t __Pyx_PyInt_As_hb_buffer_cluster_level_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE hb_buffer_content_type_t __Pyx_PyInt_As_hb_buffer_content_type_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_bool_t(hb_bool_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE hb_bool_t __Pyx_PyInt_As_hb_bool_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE uint32_t __Pyx_PyInt_As_uint32_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE hb_position_t __Pyx_PyInt_As_hb_position_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned long __Pyx_PyInt_As_unsigned_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint32_t(uint32_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_ot_name_id_t(hb_ot_name_id_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE hb_ot_math_constant_t __Pyx_PyInt_As_hb_ot_math_constant_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE hb_ot_math_kern_t __Pyx_PyInt_As_hb_ot_math_kern_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE uint8_t __Pyx_PyInt_As_uint8_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint8_t(uint8_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE hb_subset_sets_t __Pyx_PyInt_As_hb_subset_sets_t(PyObject *);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
static __Pyx_TypeName __Pyx_PyType_GetName(PyTypeObject* tp);
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#else
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CoroutineBase.proto */
struct __pyx_CoroutineObject;
typedef PyObject *(*__pyx_coroutine_body_t)(struct __pyx_CoroutineObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct __pyx_CoroutineObject {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    PyObject *gi_frame;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
#define __Pyx_Generator_CheckExact(obj) __Pyx_IS_TYPE(obj, __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(PyObject *module);

/* CheckBinaryVersion.proto */
static unsigned long __Pyx_get_runtime_version(void);
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

/* #### Code section: module_declarations ### */
static PyObject *__pyx_f_9uharfbuzz_9_harfbuzz_9GlyphInfo_set(struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphInfo *__pyx_v_self, hb_glyph_info_t __pyx_v_info); /* proto*/
static PyObject *__pyx_f_9uharfbuzz_9_harfbuzz_13GlyphPosition_set(struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition *__pyx_v_self, hb_glyph_position_t __pyx_v_position); /* proto*/
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *__pyx_f_9uharfbuzz_9_harfbuzz_4Blob_from_ptr(hb_blob_t *__pyx_v_hb_blob); /* proto*/
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_f_9uharfbuzz_9_harfbuzz_4Face_from_ptr(hb_face_t *__pyx_v_hb_face); /* proto*/
static PyObject *__pyx_f_9uharfbuzz_9_harfbuzz_4Font__Font__create(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face); /* proto*/
static PyObject *__pyx_f_9uharfbuzz_9_harfbuzz_4Font__Font__create_sub_font(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font); /* proto*/
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_f_9uharfbuzz_9_harfbuzz_4Font_from_ptr(hb_font_t *__pyx_v_hb_font); /* proto*/
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine *__pyx_f_9uharfbuzz_9_harfbuzz_9ColorLine_from_ptr(hb_color_line_t *__pyx_v_color_line); /* proto*/
static PyObject *__pyx_f_9uharfbuzz_9_harfbuzz_8HBObject_update_obj_length(struct __pyx_obj_9uharfbuzz_9_harfbuzz_HBObject *__pyx_v_self, unsigned int __pyx_v_idx, char *__pyx_v_head, char *__pyx_v_tail); /* proto*/
static hb_link_t *__pyx_f_9uharfbuzz_9_harfbuzz_8HBObject_create_links(struct __pyx_obj_9uharfbuzz_9_harfbuzz_HBObject *__pyx_v_self, unsigned int __pyx_v_idx, unsigned int __pyx_v_link_num, int __pyx_v_is_real_link); /* proto*/
static PyObject *__pyx_f_9uharfbuzz_9_harfbuzz_8HBObject_update_links(struct __pyx_obj_9uharfbuzz_9_harfbuzz_HBObject *__pyx_v_self, unsigned int __pyx_v_idx, int __pyx_v_is_real_link, PyObject *__pyx_v_links); /* proto*/
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_f_9uharfbuzz_9_harfbuzz_3Set_from_ptr(hb_set_t *__pyx_v_hb_set); /* proto*/
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_f_9uharfbuzz_9_harfbuzz_3Map_from_ptr(hb_map_t *__pyx_v_hb_map); /* proto*/

/* Module declarations from "cython" */

/* Module declarations from "libc.stdint" */

/* Module declarations from "uharfbuzz.charfbuzz" */

/* Module declarations from "libc.string" */

/* Module declarations from "libc.stdlib" */

/* Module declarations from "cpython.pycapsule" */

/* Module declarations from "uharfbuzz._harfbuzz" */
static hb_user_data_key_t __pyx_v_9uharfbuzz_9_harfbuzz_k;
static hb_draw_funcs_t *__pyx_v_9uharfbuzz_9_harfbuzz_drawfuncs;
static int __pyx_f_9uharfbuzz_9_harfbuzz_msgcallback(hb_buffer_t *, hb_font_t *, char const *, void *); /*proto*/
static hb_blob_t *__pyx_f_9uharfbuzz_9_harfbuzz__reference_table_func(hb_face_t *, hb_tag_t, void *); /*proto*/
static void __pyx_f_9uharfbuzz_9_harfbuzz__pen_move_to_func(hb_draw_funcs_t *, void *, hb_draw_state_t *, float, float, void *); /*proto*/
static void __pyx_f_9uharfbuzz_9_harfbuzz__pen_line_to_func(hb_draw_funcs_t *, void *, hb_draw_state_t *, float, float, void *); /*proto*/
static void __pyx_f_9uharfbuzz_9_harfbuzz__pen_close_path_func(hb_draw_funcs_t *, void *, hb_draw_state_t *, void *); /*proto*/
static void __pyx_f_9uharfbuzz_9_harfbuzz__pen_quadratic_to_func(hb_draw_funcs_t *, void *, hb_draw_state_t *, float, float, float, float, void *); /*proto*/
static void __pyx_f_9uharfbuzz_9_harfbuzz__pen_cubic_to_func(hb_draw_funcs_t *, void *, hb_draw_state_t *, float, float, float, float, float, float, void *); /*proto*/
static hb_position_t __pyx_f_9uharfbuzz_9_harfbuzz__glyph_h_advance_func(hb_font_t *, void *, hb_codepoint_t, void *); /*proto*/
static hb_position_t __pyx_f_9uharfbuzz_9_harfbuzz__glyph_v_advance_func(hb_font_t *, void *, hb_codepoint_t, void *); /*proto*/
static hb_bool_t __pyx_f_9uharfbuzz_9_harfbuzz__glyph_v_origin_func(hb_font_t *, void *, hb_codepoint_t, hb_position_t *, hb_position_t *, void *); /*proto*/
static hb_bool_t __pyx_f_9uharfbuzz_9_harfbuzz__glyph_name_func(hb_font_t *, void *, hb_codepoint_t, char *, unsigned int, void *); /*proto*/
static hb_bool_t __pyx_f_9uharfbuzz_9_harfbuzz__nominal_glyph_func(hb_font_t *, void *, hb_codepoint_t, hb_codepoint_t *, void *); /*proto*/
static hb_bool_t __pyx_f_9uharfbuzz_9_harfbuzz__variation_glyph_func(hb_font_t *, void *, hb_codepoint_t, hb_codepoint_t, hb_codepoint_t *, void *); /*proto*/
static hb_bool_t __pyx_f_9uharfbuzz_9_harfbuzz__font_h_extents_func(hb_font_t *, void *, hb_font_extents_t *, void *); /*proto*/
static hb_bool_t __pyx_f_9uharfbuzz_9_harfbuzz__font_v_extents_func(hb_font_t *, void *, hb_font_extents_t *, void *); /*proto*/
static void __pyx_f_9uharfbuzz_9_harfbuzz__paint_push_transform_func(hb_paint_funcs_t *, void *, float, float, float, float, float, float, void *); /*proto*/
static void __pyx_f_9uharfbuzz_9_harfbuzz__paint_pop_transform_func(hb_paint_funcs_t *, void *, void *); /*proto*/
static hb_bool_t __pyx_f_9uharfbuzz_9_harfbuzz__paint_color_glyph_func(hb_paint_funcs_t *, void *, hb_codepoint_t, hb_font_t *, void *); /*proto*/
static void __pyx_f_9uharfbuzz_9_harfbuzz__paint_push_clip_glyph_func(hb_paint_funcs_t *, void *, hb_codepoint_t, hb_font_t *, void *); /*proto*/
static void __pyx_f_9uharfbuzz_9_harfbuzz__paint_push_clip_rectangle_func(hb_paint_funcs_t *, void *, float, float, float, float, void *); /*proto*/
static void __pyx_f_9uharfbuzz_9_harfbuzz__paint_pop_clip_func(hb_paint_funcs_t *, void *, void *); /*proto*/
static void __pyx_f_9uharfbuzz_9_harfbuzz__paint_color_func(hb_paint_funcs_t *, void *, hb_bool_t, hb_color_t, void *); /*proto*/
static hb_bool_t __pyx_f_9uharfbuzz_9_harfbuzz__paint_image_func(hb_paint_funcs_t *, void *, hb_blob_t *, unsigned int, unsigned int, hb_tag_t, float, hb_glyph_extents_t *, void *); /*proto*/
static void __pyx_f_9uharfbuzz_9_harfbuzz__paint_linear_gradient_func(hb_paint_funcs_t *, void *, hb_color_line_t *, float, float, float, float, float, float, void *); /*proto*/
static void __pyx_f_9uharfbuzz_9_harfbuzz__paint_radial_gradient_func(hb_paint_funcs_t *, void *, hb_color_line_t *, float, float, float, float, float, float, void *); /*proto*/
static void __pyx_f_9uharfbuzz_9_harfbuzz__paint_sweep_gradient_func(hb_paint_funcs_t *, void *, hb_color_line_t *, float, float, float, float, void *); /*proto*/
static void __pyx_f_9uharfbuzz_9_harfbuzz__paint_push_group_func(hb_paint_funcs_t *, void *, void *); /*proto*/
static void __pyx_f_9uharfbuzz_9_harfbuzz__paint_pop_group_func(hb_paint_funcs_t *, void *, hb_paint_composite_mode_t, void *); /*proto*/
static hb_bool_t __pyx_f_9uharfbuzz_9_harfbuzz__paint_custom_palette_color_func(hb_paint_funcs_t *, void *, unsigned int, hb_color_t *, void *); /*proto*/
static void __pyx_f_9uharfbuzz_9_harfbuzz__move_to_func(hb_draw_funcs_t *, void *, hb_draw_state_t *, float, float, void *); /*proto*/
static void __pyx_f_9uharfbuzz_9_harfbuzz__line_to_func(hb_draw_funcs_t *, void *, hb_draw_state_t *, float, float, void *); /*proto*/
static void __pyx_f_9uharfbuzz_9_harfbuzz__close_path_func(hb_draw_funcs_t *, void *, hb_draw_state_t *, void *); /*proto*/
static void __pyx_f_9uharfbuzz_9_harfbuzz__quadratic_to_func(hb_draw_funcs_t *, void *, hb_draw_state_t *, float, float, float, float, void *); /*proto*/
static void __pyx_f_9uharfbuzz_9_harfbuzz__cubic_to_func(hb_draw_funcs_t *, void *, hb_draw_state_t *, float, float, float, float, float, float, void *); /*proto*/
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "uharfbuzz._harfbuzz"
extern int __pyx_module_is_main_uharfbuzz___harfbuzz;
int __pyx_module_is_main_uharfbuzz___harfbuzz = 0;

/* Implementation of "uharfbuzz._harfbuzz" */
/* #### Code section: global_var ### */
static PyObject *__pyx_builtin_staticmethod;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_AssertionError;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_round;
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_DeprecationWarning;
static PyObject *__pyx_builtin_KeyError;
static PyObject *__pyx_builtin_NotImplemented;
static PyObject *__pyx_builtin_StopIteration;
/* #### Code section: string_decls ### */
static const char __pyx_k_a[] = "a";
static const char __pyx_k_b[] = "b";
static const char __pyx_k_c[] = "c";
static const char __pyx_k_g[] = "g";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_k[] = "k";
static const char __pyx_k_m[] = "m";
static const char __pyx_k_p[] = "p";
static const char __pyx_k_r[] = "r";
static const char __pyx_k_s[] = "s";
static const char __pyx_k_v[] = "v";
static const char __pyx_k_x[] = "x";
static const char __pyx_k_y[] = "y";
static const char __pyx_k_gc[] = "gc";
static const char __pyx_k_os[] = "os";
static const char __pyx_k_BOT[] = "BOT";
static const char __pyx_k_EOT[] = "EOT";
static const char __pyx_k_Map[] = "Map";
static const char __pyx_k_PAD[] = "PAD";
static const char __pyx_k_SRC[] = "SRC";
static const char __pyx_k_Set[] = "Set({...})";
static const char __pyx_k_XOR[] = "XOR";
static const char __pyx_k__10[] = "'";
static const char __pyx_k__15[] = "";
static const char __pyx_k__16[] = ")";
static const char __pyx_k__19[] = ", ";
static const char __pyx_k__21[] = ": ";
static const char __pyx_k__22[] = "*";
static const char __pyx_k__23[] = ".";
static const char __pyx_k_add[] = "add";
static const char __pyx_k_cls[] = "cls";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_end[] = "end";
static const char __pyx_k_get[] = "get";
static const char __pyx_k_gid[] = "gid";
static const char __pyx_k_int[] = "int";
static const char __pyx_k_pen[] = "pen";
static const char __pyx_k_red[] = "red";
static const char __pyx_k_ret[] = "ret";
static const char __pyx_k_set[] = "set";
static const char __pyx_k_str[] = "str";
static const char __pyx_k_tag[] = "tag";
static const char __pyx_k_Blob[] = "Blob";
static const char __pyx_k_DEST[] = "DEST";
static const char __pyx_k_Dict[] = "Dict";
static const char __pyx_k_Face[] = "Face";
static const char __pyx_k_Font[] = "Font";
static const char __pyx_k_List[] = "List";
static const char __pyx_k_MARK[] = "MARK";
static const char __pyx_k_None[] = "None";
static const char __pyx_k_PLUS[] = "PLUS";
static const char __pyx_k_Path[] = "Path";
static const char __pyx_k__287[] = "?";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_blob[] = "blob";
static const char __pyx_k_blue[] = "blue";
static const char __pyx_k_bool[] = "bool";
static const char __pyx_k_copy[] = "copy";
static const char __pyx_k_cstr[] = "cstr";
static const char __pyx_k_data[] = "data";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_enum[] = "enum";
static const char __pyx_k_face[] = "face";
static const char __pyx_k_feat[] = "feat";
static const char __pyx_k_font[] = "font";
static const char __pyx_k_func[] = "func";
static const char __pyx_k_hang[] = "hang";
static const char __pyx_k_head[] = "head";
static const char __pyx_k_icfb[] = "icfb";
static const char __pyx_k_icft[] = "icft";
static const char __pyx_k_ideo[] = "ideo";
static const char __pyx_k_idtp[] = "idtp";
static const char __pyx_k_init[] = "init";
static const char __pyx_k_inst[] = "inst";
static const char __pyx_k_join[] = "join";
static const char __pyx_k_kern[] = "kern";
static const char __pyx_k_keys[] = "keys";
static const char __pyx_k_kind[] = "kind";
static const char __pyx_k_last[] = "last";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_math[] = "math";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_node[] = "node";
static const char __pyx_k_romn[] = "romn";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_sets[] = "sets";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_spec[] = "__spec__";
static const char __pyx_k_tags[] = "tags";
static const char __pyx_k_tail[] = "tail";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_text[] = "text";
static const char __pyx_k_warn[] = "warn";
static const char __pyx_k_CLEAR[] = "CLEAR";
static const char __pyx_k_Color[] = "Color";
static const char __pyx_k_Map_s[] = "Map({%s})";
static const char __pyx_k_Set_2[] = "Set";
static const char __pyx_k_Set_s[] = "Set({%s})";
static const char __pyx_k_Tuple[] = "Tuple";
static const char __pyx_k_Union[] = "Union";
static const char __pyx_k_alpha[] = "alpha";
static const char __pyx_k_bytes[] = "bytes";
static const char __pyx_k_clear[] = "clear";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_color[] = "color";
static const char __pyx_k_count[] = "count";
static const char __pyx_k_first[] = "first";
static const char __pyx_k_flags[] = "flags";
static const char __pyx_k_float[] = "float";
static const char __pyx_k_funcs[] = "funcs";
static const char __pyx_k_glyph[] = "glyph";
static const char __pyx_k_green[] = "green";
static const char __pyx_k_index[] = "index";
static const char __pyx_k_input[] = "input";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_kerns[] = "kerns";
static const char __pyx_k_other[] = "other";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_reset[] = "reset";
static const char __pyx_k_round[] = "round";
static const char __pyx_k_set_2[] = "_set";
static const char __pyx_k_shape[] = "shape";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_super[] = "super";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_width[] = "width";
static const char __pyx_k_Buffer[] = "Buffer";
static const char __pyx_k_DARKEN[] = "DARKEN";
static const char __pyx_k_GLYPHS[] = "GLYPHS";
static const char __pyx_k_REPEAT[] = "REPEAT";
static const char __pyx_k_SCREEN[] = "SCREEN";
static const char __pyx_k_SRC_IN[] = "SRC_IN";
static const char __pyx_k_VERIFY[] = "VERIFY";
static const char __pyx_k_buffer[] = "buffer";
static const char __pyx_k_colors[] = "colors";
static const char __pyx_k_coords[] = "coords";
static const char __pyx_k_copy_2[] = "__copy__";
static const char __pyx_k_create[] = "create";
static const char __pyx_k_enable[] = "enable";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_func_p[] = "func_p";
static const char __pyx_k_hb_tag[] = "hb_tag";
static const char __pyx_k_height[] = "height";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_invert[] = "invert";
static const char __pyx_k_layers[] = "layers";
static const char __pyx_k_length[] = "length";
static const char __pyx_k_lineTo[] = "lineTo";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_moveTo[] = "moveTo";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_object[] = "object";
static const char __pyx_k_offset[] = "offset";
static const char __pyx_k_packed[] = "packed";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_remove[] = "remove";
static const char __pyx_k_repack[] = "repack";
static const char __pyx_k_return[] = "return";
static const char __pyx_k_shaper[] = "shaper";
static const char __pyx_k_source[] = "source";
static const char __pyx_k_string[] = "string";
static const char __pyx_k_subset[] = "subset";
static const char __pyx_k_to_int[] = "to_int";
static const char __pyx_k_typing[] = "typing";
static const char __pyx_k_update[] = "_update";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_DEFAULT[] = "DEFAULT";
static const char __pyx_k_DEST_IN[] = "DEST_IN";
static const char __pyx_k_HSL_HUE[] = "HSL_HUE";
static const char __pyx_k_INVALID[] = "INVALID";
static const char __pyx_k_IntEnum[] = "IntEnum";
static const char __pyx_k_IntFlag[] = "IntFlag";
static const char __pyx_k_LIGHTEN[] = "LIGHTEN";
static const char __pyx_k_MapIter[] = "MapIter";
static const char __pyx_k_Map_get[] = "Map.get";
static const char __pyx_k_NAME_ID[] = "NAME_ID";
static const char __pyx_k_OVERLAY[] = "OVERLAY";
static const char __pyx_k_REFLECT[] = "REFLECT";
static const char __pyx_k_SRC_OUT[] = "SRC_OUT";
static const char __pyx_k_SetIter[] = "SetIter";
static const char __pyx_k_Set_add[] = "Set.add";
static const char __pyx_k_Set_set[] = "Set.set";
static const char __pyx_k_UNICODE[] = "UNICODE";
static const char __pyx_k_add_str[] = "add_str";
static const char __pyx_k_advance[] = "advance";
static const char __pyx_k_curveTo[] = "curveTo";
static const char __pyx_k_disable[] = "disable";
static const char __pyx_k_discard[] = "discard";
static const char __pyx_k_execute[] = "execute";
static const char __pyx_k_extents[] = "extents";
static const char __pyx_k_genexpr[] = "genexpr";
static const char __pyx_k_methods[] = "methods";
static const char __pyx_k_name_id[] = "name_id";
static const char __pyx_k_pathlib[] = "pathlib";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_shapers[] = "shapers";
static const char __pyx_k_success[] = "success";
static const char __pyx_k_unicode[] = "unicode";
static const char __pyx_k_Callable[] = "Callable";
static const char __pyx_k_DEST_OUT[] = "DEST_OUT";
static const char __pyx_k_EXTENDER[] = "EXTENDER";
static const char __pyx_k_HBObject[] = "HBObject";
static const char __pyx_k_KeyError[] = "KeyError";
static const char __pyx_k_LIGATURE[] = "LIGATURE";
static const char __pyx_k_List_int[] = "List[int]";
static const char __pyx_k_List_str[] = "List[str]";
static const char __pyx_k_MULTIPLY[] = "MULTIPLY";
static const char __pyx_k_Map_copy[] = "Map.copy";
static const char __pyx_k_Map_keys[] = "Map.keys";
static const char __pyx_k_SRC_ATOP[] = "SRC_ATOP";
static const char __pyx_k_SRC_OVER[] = "SRC_OVER";
static const char __pyx_k_Sequence[] = "Sequence";
static const char __pyx_k_Set__set[] = "Set._set";
static const char __pyx_k_Set_copy[] = "Set.copy";
static const char __pyx_k_TOP_LEFT[] = "TOP_LEFT";
static const char __pyx_k_add_utf8[] = "add_utf8";
static const char __pyx_k_ascender[] = "ascender";
static const char __pyx_k_assembly[] = "assembly";
static const char __pyx_k_c_coords[] = "c_coords";
static const char __pyx_k_callback[] = "callback";
static const char __pyx_k_constant[] = "constant";
static const char __pyx_k_features[] = "features";
static const char __pyx_k_filename[] = "filename";
static const char __pyx_k_from_int[] = "from_int";
static const char __pyx_k_fsencode[] = "fsencode";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_int_None[] = "int | None";
static const char __pyx_k_is_equal[] = "_is_equal";
static const char __pyx_k_issubset[] = "issubset";
static const char __pyx_k_line_gap[] = "line_gap";
static const char __pyx_k_new_face[] = "new_face";
static const char __pyx_k_obj_list[] = "obj_list";
static const char __pyx_k_qCurveTo[] = "qCurveTo";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_set_name[] = "__set_name__";
static const char __pyx_k_set_type[] = "set_type";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_update_2[] = "update";
static const char __pyx_k_variants[] = "variants";
static const char __pyx_k_warnings[] = "warnings";
static const char __pyx_k_COMPONENT[] = "COMPONENT";
static const char __pyx_k_ColorLine[] = "ColorLine";
static const char __pyx_k_ColorStop[] = "ColorStop";
static const char __pyx_k_DEST_ATOP[] = "DEST_ATOP";
static const char __pyx_k_DEST_OVER[] = "DEST_OVER";
static const char __pyx_k_DrawFuncs[] = "DrawFuncs";
static const char __pyx_k_EXCLUSION[] = "EXCLUSION";
static const char __pyx_k_FontFuncs[] = "FontFuncs";
static const char __pyx_k_GlyphInfo[] = "GlyphInfo";
static const char __pyx_k_HSL_COLOR[] = "HSL_COLOR";
static const char __pyx_k_Map_clear[] = "Map.clear";
static const char __pyx_k_Map_items[] = "Map.items";
static const char __pyx_k_Set_clear[] = "Set.clear";
static const char __pyx_k_TOP_RIGHT[] = "TOP_RIGHT";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_add_range[] = "add_range";
static const char __pyx_k_blob_data[] = "blob_data";
static const char __pyx_k_c_shapers[] = "c_shapers";
static const char __pyx_k_closePath[] = "closePath";
static const char __pyx_k_del_range[] = "del_range";
static const char __pyx_k_descender[] = "descender";
static const char __pyx_k_direction[] = "direction";
static const char __pyx_k_draw_data[] = "draw_data";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_hb_script[] = "hb_script";
static const char __pyx_k_isenabled[] = "isenabled";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_num_nodes[] = "num_nodes";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_subtables[] = "subtables";
static const char __pyx_k_user_data[] = "user_data";
static const char __pyx_k_variation[] = "variation";
static const char __pyx_k_x_bearing[] = "x_bearing";
static const char __pyx_k_y_bearing[] = "y_bearing";
static const char __pyx_k_BASE_GLYPH[] = "BASE_GLYPH";
static const char __pyx_k_CHARACTERS[] = "CHARACTERS";
static const char __pyx_k_COLOR_BURN[] = "COLOR_BURN";
static const char __pyx_k_Color_None[] = "Color | None";
static const char __pyx_k_DIFFERENCE[] = "DIFFERENCE";
static const char __pyx_k_GlyphFlags[] = "GlyphFlags";
static const char __pyx_k_HARD_LIGHT[] = "HARD_LIGHT";
static const char __pyx_k_List_Color[] = "List[Color]";
static const char __pyx_k_List_bytes[] = "List[bytes]";
static const char __pyx_k_Map___copy[] = "Map.__copy__";
static const char __pyx_k_Map_update[] = "Map.update";
static const char __pyx_k_Map_values[] = "Map.values";
static const char __pyx_k_NO_HINTING[] = "NO_HINTING";
static const char __pyx_k_OTMathKern[] = "OTMathKern";
static const char __pyx_k_PaintFuncs[] = "PaintFuncs";
static const char __pyx_k_SOFT_LIGHT[] = "SOFT_LIGHT";
static const char __pyx_k_Set___copy[] = "Set.__copy__";
static const char __pyx_k_Set_invert[] = "Set.invert";
static const char __pyx_k_Set_remove[] = "Set.remove";
static const char __pyx_k_Set_update[] = "Set.update";
static const char __pyx_k_SubsetPlan[] = "SubsetPlan";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_alternates[] = "alternates";
static const char __pyx_k_codepoints[] = "codepoints";
static const char __pyx_k_draw_funcs[] = "draw_funcs";
static const char __pyx_k_draw_glyph[] = "draw_glyph";
static const char __pyx_k_draw_state[] = "draw_state";
static const char __pyx_k_foreground[] = "foreground";
static const char __pyx_k_graphnodes[] = "graphnodes";
static const char __pyx_k_issuperset[] = "issuperset";
static const char __pyx_k_kern_value[] = "kern_value";
static const char __pyx_k_larger_set[] = "larger_set";
static const char __pyx_k_namedtuple[] = "namedtuple";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_script_tag[] = "script_tag";
static const char __pyx_k_table_data[] = "table_data";
static const char __pyx_k_tag_packed[] = "tag_packed";
static const char __pyx_k_variations[] = "variations";
static const char __pyx_k_AXIS_HEIGHT[] = "AXIS_HEIGHT";
static const char __pyx_k_BOTTOM_LEFT[] = "BOTTOM_LEFT";
static const char __pyx_k_BufferFlags[] = "BufferFlags";
static const char __pyx_k_COLOR_DODGE[] = "COLOR_DODGE";
static const char __pyx_k_Face_create[] = "Face.create";
static const char __pyx_k_FontExtents[] = "FontExtents";
static const char __pyx_k_Font_create[] = "Font.create";
static const char __pyx_k_GLYPH_INDEX[] = "GLYPH_INDEX";
static const char __pyx_k_GLYPH_NAMES[] = "GLYPH_NAMES";
static const char __pyx_k_Map__update[] = "Map._update";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_NAME_LEGACY[] = "NAME_LEGACY";
static const char __pyx_k_PaintExtend[] = "PaintExtend";
static const char __pyx_k_RETAIN_GIDS[] = "RETAIN_GIDS";
static const char __pyx_k_Set__update[] = "Set._update";
static const char __pyx_k_Set_discard[] = "Set.discard";
static const char __pyx_k_SubsetFlags[] = "SubsetFlags";
static const char __pyx_k_SubsetInput[] = "SubsetInput";
static const char __pyx_k_blob_length[] = "blob_length";
static const char __pyx_k_collections[] = "collections";
static const char __pyx_k_color_count[] = "color_count";
static const char __pyx_k_color_index[] = "color_index";
static const char __pyx_k_hb_features[] = "hb_features";
static const char __pyx_k_hb_language[] = "hb_language";
static const char __pyx_k_hb_position[] = "hb_position";
static const char __pyx_k_is_inverted[] = "is_inverted";
static const char __pyx_k_item_length[] = "item_length";
static const char __pyx_k_item_offset[] = "item_offset";
static const char __pyx_k_kerns_array[] = "kerns_array";
static const char __pyx_k_layer_count[] = "layer_count";
static const char __pyx_k_mro_entries[] = "__mro_entries__";
static const char __pyx_k_packed_blob[] = "packed_blob";
static const char __pyx_k_paint_funcs[] = "paint_funcs";
static const char __pyx_k_paint_glyph[] = "paint_glyph";
static const char __pyx_k_paint_state[] = "paint_state";
static const char __pyx_k_script_tags[] = "script_tags";
static const char __pyx_k_smaller_set[] = "smaller_set";
static const char __pyx_k_table_bytes[] = "table_bytes";
static const char __pyx_k_user_data_p[] = "user_data_p";
static const char __pyx_k_BOTTOM_RIGHT[] = "BOTTOM_RIGHT";
static const char __pyx_k_Buffer_reset[] = "Buffer.reset";
static const char __pyx_k_Color_to_int[] = "Color.to_int";
static const char __pyx_k_GlyphExtents[] = "GlyphExtents";
static const char __pyx_k_MATH_LEADING[] = "MATH_LEADING";
static const char __pyx_k_NAME_LANG_ID[] = "NAME_LANG_ID";
static const char __pyx_k_OTColorLayer[] = "OTColorLayer";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_Set_issubset[] = "Set.issubset";
static const char __pyx_k_UNCLASSIFIED[] = "UNCLASSIFIED";
static const char __pyx_k_baseline_tag[] = "baseline_tag";
static const char __pyx_k_c_foreground[] = "c_foreground";
static const char __pyx_k_draw_state_p[] = "draw_state_p";
static const char __pyx_k_face_or_font[] = "face_or_font";
static const char __pyx_k_feature_tags[] = "feature_tags";
static const char __pyx_k_full_advance[] = "full_advance";
static const char __pyx_k_hb_direction[] = "hb_direction";
static const char __pyx_k_initializing[] = "_initializing";
static const char __pyx_k_invalid_kern[] = "invalid kern";
static const char __pyx_k_is_coroutine[] = "_is_coroutine";
static const char __pyx_k_language_tag[] = "language_tag";
static const char __pyx_k_lookup_index[] = "lookup_index";
static const char __pyx_k_script_count[] = "script_count";
static const char __pyx_k_script_index[] = "script_index";
static const char __pyx_k_start_offset[] = "start_offset";
static const char __pyx_k_staticmethod[] = "staticmethod";
static const char __pyx_k_stringsource[] = "<stringsource>";
static const char __pyx_k_Buffer_create[] = "Buffer.create";
static const char __pyx_k_GlyphPosition[] = "GlyphPosition";
static const char __pyx_k_HarfBuzzError[] = "HarfBuzzError";
static const char __pyx_k_INVALID_VALUE[] = "INVALID_VALUE";
static const char __pyx_k_Map__is_equal[] = "Map._is_equal";
static const char __pyx_k_RepackerError[] = "RepackerError";
static const char __pyx_k_STACK_GAP_MIN[] = "STACK_GAP_MIN";
static const char __pyx_k_Set__is_equal[] = "Set._is_equal";
static const char __pyx_k_Set_add_range[] = "Set.add_range";
static const char __pyx_k_Set_del_range[] = "Set.del_range";
static const char __pyx_k_StopIteration[] = "StopIteration";
static const char __pyx_k_coords_2dot14[] = "coords_2dot14";
static const char __pyx_k_feature_count[] = "feature_count";
static const char __pyx_k_hb_codepoints[] = "hb_codepoints";
static const char __pyx_k_hb_script_tag[] = "hb_script_tag";
static const char __pyx_k_hb_variations[] = "hb_variations";
static const char __pyx_k_init_subclass[] = "__init_subclass__";
static const char __pyx_k_is_foreground[] = "is_foreground";
static const char __pyx_k_language_tags[] = "language_tags";
static const char __pyx_k_paint_state_p[] = "paint_state_p";
static const char __pyx_k_palette_index[] = "palette_index";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_set_variation[] = "set_variation";
static const char __pyx_k_AssertionError[] = "AssertionError";
static const char __pyx_k_Buffer_add_str[] = "Buffer.add_str";
static const char __pyx_k_Color_from_int[] = "Color.from_int";
static const char __pyx_k_DESUBROUTINIZE[] = "DESUBROUTINIZE";
static const char __pyx_k_DROP_TABLE_TAG[] = "DROP_TABLE_TAG";
static const char __pyx_k_Dict_str_float[] = "Dict[str, float]";
static const char __pyx_k_Failed_to_open[] = "Failed to open: ";
static const char __pyx_k_HSL_LUMINOSITY[] = "HSL_LUMINOSITY";
static const char __pyx_k_HSL_SATURATION[] = "HSL_SATURATION";
static const char __pyx_k_NOTDEF_OUTLINE[] = "NOTDEF_OUTLINE";
static const char __pyx_k_NotImplemented[] = "NotImplemented";
static const char __pyx_k_OTMathConstant[] = "OTMathConstant";
static const char __pyx_k_Set_issuperset[] = "Set.issuperset";
static const char __pyx_k_Union_str_Path[] = "Union[str, Path]";
static const char __pyx_k_add_codepoints[] = "add_codepoints";
static const char __pyx_k_assembly_array[] = "assembly_array";
static const char __pyx_k_clear_contents[] = "clear_contents";
static const char __pyx_k_from_file_path[] = "from_file_path";
static const char __pyx_k_get_glyph_name[] = "get_glyph_name";
static const char __pyx_k_hb_codepoint_t[] = "hb_codepoint_t";
static const char __pyx_k_language_count[] = "language_count";
static const char __pyx_k_language_index[] = "language_index";
static const char __pyx_k_set_color_func[] = "set_color_func";
static const char __pyx_k_set_image_func[] = "set_image_func";
static const char __pyx_k_set_variations[] = "set_variations";
static const char __pyx_k_variants_array[] = "variants_array";
static const char __pyx_k_version_string[] = "version_string";
static const char __pyx_k_Buffer_add_utf8[] = "Buffer.add_utf8";
static const char __pyx_k_Font_draw_glyph[] = "Font.draw_glyph";
static const char __pyx_k_OTMathGlyphPart[] = "OTMathGlyphPart";
static const char __pyx_k_OTMathKernEntry[] = "OTMathKernEntry";
static const char __pyx_k_Set_is_inverted[] = "Set.is_inverted";
static const char __pyx_k_SubsetInputSets[] = "SubsetInputSets";
static const char __pyx_k_UNSAFE_TO_BREAK[] = "UNSAFE_TO_BREAK";
static const char __pyx_k_alternate_count[] = "alternate_count";
static const char __pyx_k_get_glyph_shape[] = "get_glyph_shape";
static const char __pyx_k_glyph_to_string[] = "glyph_to_string";
static const char __pyx_k_hb_baseline_tag[] = "hb_baseline_tag";
static const char __pyx_k_hb_language_tag[] = "hb_language_tag";
static const char __pyx_k_keep_everything[] = "keep_everything";
static const char __pyx_k_repack_with_tag[] = "repack_with_tag";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_FontFuncs_create[] = "FontFuncs.create";
static const char __pyx_k_Font_paint_glyph[] = "Font.paint_glyph";
static const char __pyx_k_SubsetInput_sets[] = "SubsetInput.sets";
static const char __pyx_k_UNSAFE_TO_CONCAT[] = "UNSAFE_TO_CONCAT";
static const char __pyx_k_alternate_glyphs[] = "alternate_glyphs";
static const char __pyx_k_get_font_extents[] = "get_font_extents";
static const char __pyx_k_invalid_constant[] = "invalid constant";
static const char __pyx_k_ot_color_has_png[] = "ot_color_has_png";
static const char __pyx_k_ot_color_has_svg[] = "ot_color_has_svg";
static const char __pyx_k_ot_math_has_data[] = "ot_math_has_data";
static const char __pyx_k_ot_tag_to_script[] = "ot_tag_to_script";
static const char __pyx_k_set_line_to_func[] = "set_line_to_func";
static const char __pyx_k_set_message_func[] = "set_message_func";
static const char __pyx_k_set_move_to_func[] = "set_move_to_func";
static const char __pyx_k_BufferContentType[] = "BufferContentType";
static const char __pyx_k_LAYOUT_SCRIPT_TAG[] = "LAYOUT_SCRIPT_TAG";
static const char __pyx_k_List_OTColorLayer[] = "List[OTColorLayer]";
static const char __pyx_k_NO_LAYOUT_CLOSURE[] = "NO_LAYOUT_CLOSURE";
static const char __pyx_k_SET_OVERLAPS_FLAG[] = "SET_OVERLAPS_FLAG";
static const char __pyx_k_SUBSCRIPT_TOP_MAX[] = "SUBSCRIPT_TOP_MAX";
static const char __pyx_k_Subsetting_failed[] = "Subsetting failed";
static const char __pyx_k_correction_height[] = "correction_height";
static const char __pyx_k_create_for_tables[] = "create_for_tables";
static const char __pyx_k_difference_update[] = "difference_update";
static const char __pyx_k_get_glyph_extents[] = "get_glyph_extents";
static const char __pyx_k_get_nominal_glyph[] = "get_nominal_glyph";
static const char __pyx_k_glyph_from_string[] = "glyph_from_string";
static const char __pyx_k_ot_font_set_funcs[] = "ot_font_set_funcs";
static const char __pyx_k_pin_axis_location[] = "pin_axis_location";
static const char __pyx_k_set_cubic_to_func[] = "set_cubic_to_func";
static const char __pyx_k_set_pop_clip_func[] = "set_pop_clip_func";
static const char __pyx_k_subset_preprocess[] = "subset_preprocess";
static const char __pyx_k_ACCENT_BASE_HEIGHT[] = "ACCENT_BASE_HEIGHT";
static const char __pyx_k_All_shapers_failed[] = "All shapers failed";
static const char __pyx_k_BufferClusterLevel[] = "BufferClusterLevel";
static const char __pyx_k_DeprecationWarning[] = "DeprecationWarning";
static const char __pyx_k_Font_set_variation[] = "Font.set_variation";
static const char __pyx_k_LAYOUT_FEATURE_TAG[] = "LAYOUT_FEATURE_TAG";
static const char __pyx_k_MONOTONE_GRAPHEMES[] = "MONOTONE_GRAPHEMES";
static const char __pyx_k_OTLayoutGlyphClass[] = "OTLayoutGlyphClass";
static const char __pyx_k_OTMathGlyphVariant[] = "OTMathGlyphVariant";
static const char __pyx_k_PaintCompositeMode[] = "PaintCompositeMode";
static const char __pyx_k_SPACE_AFTER_SCRIPT[] = "SPACE_AFTER_SCRIPT";
static const char __pyx_k_STACK_TOP_SHIFT_UP[] = "STACK_TOP_SHIFT_UP";
static const char __pyx_k_SubsetInput_subset[] = "SubsetInput.subset";
static const char __pyx_k_SubsetPlan_execute[] = "SubsetPlan.execute";
static const char __pyx_k_asyncio_coroutines[] = "asyncio.coroutines";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_get_glyph_h_origin[] = "get_glyph_h_origin";
static const char __pyx_k_get_glyph_v_origin[] = "get_glyph_v_origin";
static const char __pyx_k_italics_correction[] = "italics_correction";
static const char __pyx_k_num_of_graph_nodes[] = ") != num of graph nodes(";
static const char __pyx_k_ot_color_has_paint[] = "ot_color_has_paint";
static const char __pyx_k_ot_tag_to_language[] = "ot_tag_to_language";
static const char __pyx_k_set_pop_group_func[] = "set_pop_group_func";
static const char __pyx_k_variation_selector[] = "variation_selector";
static const char __pyx_k_variation_unicodes[] = "variation_unicodes";
static const char __pyx_k_Blob_from_file_path[] = "Blob.from_file_path";
static const char __pyx_k_Font_get_glyph_name[] = "Font.get_glyph_name";
static const char __pyx_k_Font_set_variations[] = "Font.set_variations";
static const char __pyx_k_LOWER_LIMIT_GAP_MIN[] = "LOWER_LIMIT_GAP_MIN";
static const char __pyx_k_MONOTONE_CHARACTERS[] = "MONOTONE_CHARACTERS";
static const char __pyx_k_Map___reduce_cython[] = "Map.__reduce_cython__";
static const char __pyx_k_NO_SUBSET_TABLE_TAG[] = "NO_SUBSET_TABLE_TAG";
static const char __pyx_k_OTColorPaletteFlags[] = "OTColorPaletteFlags";
static const char __pyx_k_Set___reduce_cython[] = "Set.__reduce_cython__";
static const char __pyx_k_UPPER_LIMIT_GAP_MIN[] = "UPPER_LIMIT_GAP_MIN";
static const char __pyx_k_draw_glyph_with_pen[] = "draw_glyph_with_pen";
static const char __pyx_k_get_glyph_from_name[] = "get_glyph_from_name";
static const char __pyx_k_get_glyph_h_advance[] = "get_glyph_h_advance";
static const char __pyx_k_get_glyph_v_advance[] = "get_glyph_v_advance";
static const char __pyx_k_get_variation_glyph[] = "get_variation_glyph";
static const char __pyx_k_intersection_update[] = "intersection_update";
static const char __pyx_k_keys_locals_genexpr[] = "keys.<locals>.genexpr";
static const char __pyx_k_ot_color_has_layers[] = "ot_color_has_layers";
static const char __pyx_k_pin_axis_to_default[] = "pin_axis_to_default";
static const char __pyx_k_set_close_path_func[] = "set_close_path_func";
static const char __pyx_k_set_glyph_name_func[] = "set_glyph_name_func";
static const char __pyx_k_set_push_group_func[] = "set_push_group_func";
static const char __pyx_k_uharfbuzz__harfbuzz[] = "uharfbuzz._harfbuzz";
static const char __pyx_k_Blob___reduce_cython[] = "Blob.__reduce_cython__";
static const char __pyx_k_Callable_object_None[] = "Callable[[object], None]";
static const char __pyx_k_DrawFuncs_draw_glyph[] = "DrawFuncs.draw_glyph";
static const char __pyx_k_Face___reduce_cython[] = "Face.__reduce_cython__";
static const char __pyx_k_Font___reduce_cython[] = "Font.__reduce_cython__";
static const char __pyx_k_Font_glyph_to_string[] = "Font.glyph_to_string";
static const char __pyx_k_List_OTMathKernEntry[] = "List[OTMathKernEntry]";
static const char __pyx_k_OTMathGlyphPartFlags[] = "OTMathGlyphPartFlags";
static const char __pyx_k_OVERBAR_VERTICAL_GAP[] = "OVERBAR_VERTICAL_GAP";
static const char __pyx_k_RADICAL_VERTICAL_GAP[] = "RADICAL_VERTICAL_GAP";
static const char __pyx_k_SUBSCRIPT_SHIFT_DOWN[] = "SUBSCRIPT_SHIFT_DOWN";
static const char __pyx_k_SUPERSCRIPT_SHIFT_UP[] = "SUPERSCRIPT_SHIFT_UP";
static const char __pyx_k_end_connector_length[] = "end_connector_length";
static const char __pyx_k_invalid_baseline_tag[] = "invalid baseline tag '";
static const char __pyx_k_ot_math_get_constant[] = "ot_math_get_constant";
static const char __pyx_k_set_color_glyph_func[] = "set_color_glyph_func";
static const char __pyx_k_Buffer_add_codepoints[] = "Buffer.add_codepoints";
static const char __pyx_k_Buffer_clear_contents[] = "Buffer.clear_contents";
static const char __pyx_k_Font_get_font_extents[] = "Font.get_font_extents";
static const char __pyx_k_Map___setstate_cython[] = "Map.__setstate_cython__";
static const char __pyx_k_Set___setstate_cython[] = "Set.__setstate_cython__";
static const char __pyx_k_Set_difference_update[] = "Set.difference_update";
static const char __pyx_k_UNDERBAR_VERTICAL_GAP[] = "UNDERBAR_VERTICAL_GAP";
static const char __pyx_k_get_var_coords_design[] = "get_var_coords_design";
static const char __pyx_k_max_correction_height[] = "max_correction_height";
static const char __pyx_k_ot_color_has_palettes[] = "ot_color_has_palettes";
static const char __pyx_k_repr___locals_genexpr[] = "__repr__.<locals>.genexpr";
static const char __pyx_k_set_quadratic_to_func[] = "set_quadratic_to_func";
static const char __pyx_k_set_var_coords_design[] = "set_var_coords_design";
static const char __pyx_k_values_locals_genexpr[] = "values.<locals>.genexpr";
static const char __pyx_k_Blob___setstate_cython[] = "Blob.__setstate_cython__";
static const char __pyx_k_Buffer___reduce_cython[] = "Buffer.__reduce_cython__";
static const char __pyx_k_Face___setstate_cython[] = "Face.__setstate_cython__";
static const char __pyx_k_Face_create_for_tables[] = "Face.create_for_tables";
static const char __pyx_k_Font___setstate_cython[] = "Font.__setstate_cython__";
static const char __pyx_k_Font_get_glyph_extents[] = "Font.get_glyph_extents";
static const char __pyx_k_Font_get_nominal_glyph[] = "Font.get_nominal_glyph";
static const char __pyx_k_Font_glyph_from_string[] = "Font.glyph_from_string";
static const char __pyx_k_Input_num_of_subtables[] = "Input num of subtables(";
static const char __pyx_k_OVERBAR_EXTRA_ASCENDER[] = "OVERBAR_EXTRA_ASCENDER";
static const char __pyx_k_OVERBAR_RULE_THICKNESS[] = "OVERBAR_RULE_THICKNESS";
static const char __pyx_k_RADICAL_EXTRA_ASCENDER[] = "RADICAL_EXTRA_ASCENDER";
static const char __pyx_k_RADICAL_RULE_THICKNESS[] = "RADICAL_RULE_THICKNESS";
static const char __pyx_k_SAFE_TO_INSERT_TATWEEL[] = "SAFE_TO_INSERT_TATWEEL";
static const char __pyx_k_SUPERSCRIPT_BOTTOM_MIN[] = "SUPERSCRIPT_BOTTOM_MIN";
static const char __pyx_k_ot_color_glyph_get_png[] = "ot_color_glyph_get_png";
static const char __pyx_k_ot_color_glyph_get_svg[] = "ot_color_glyph_get_svg";
static const char __pyx_k_ot_layout_get_baseline[] = "ot_layout_get_baseline";
static const char __pyx_k_set_nominal_glyph_func[] = "set_nominal_glyph_func";
static const char __pyx_k_set_pop_transform_func[] = "set_pop_transform_func";
static const char __pyx_k_set_script_from_ot_tag[] = "set_script_from_ot_tag";
static const char __pyx_k_start_connector_length[] = "start_connector_length";
static const char __pyx_k_Buffer_set_message_func[] = "Buffer.set_message_func";
static const char __pyx_k_FRACTION_RULE_THICKNESS[] = "FRACTION_RULE_THICKNESS";
static const char __pyx_k_Face_variation_unicodes[] = "Face.variation_unicodes";
static const char __pyx_k_Font_get_glyph_h_origin[] = "Font.get_glyph_h_origin";
static const char __pyx_k_Font_get_glyph_v_origin[] = "Font.get_glyph_v_origin";
static const char __pyx_k_List_OTMathGlyphVariant[] = "List[OTMathGlyphVariant]";
static const char __pyx_k_MapIter___reduce_cython[] = "MapIter.__reduce_cython__";
static const char __pyx_k_NO_PRUNE_UNICODE_RANGES[] = "NO_PRUNE_UNICODE_RANGES";
static const char __pyx_k_STACK_BOTTOM_SHIFT_DOWN[] = "STACK_BOTTOM_SHIFT_DOWN";
static const char __pyx_k_SUB_SUPERSCRIPT_GAP_MIN[] = "SUB_SUPERSCRIPT_GAP_MIN";
static const char __pyx_k_SetIter___reduce_cython[] = "SetIter.__reduce_cython__";
static const char __pyx_k_Set_intersection_update[] = "Set.intersection_update";
static const char __pyx_k_UNDERBAR_RULE_THICKNESS[] = "UNDERBAR_RULE_THICKNESS";
static const char __pyx_k_set_font_h_extents_func[] = "set_font_h_extents_func";
static const char __pyx_k_set_font_v_extents_func[] = "set_font_v_extents_func";
static const char __pyx_k_set_glyph_v_origin_func[] = "set_glyph_v_origin_func";
static const char __pyx_k_set_push_transform_func[] = "set_push_transform_func";
static const char __pyx_k_set_sweep_gradient_func[] = "set_sweep_gradient_func";
static const char __pyx_k_Buffer___setstate_cython[] = "Buffer.__setstate_cython__";
static const char __pyx_k_Callable_int_object_None[] = "Callable[[int, object], None]";
static const char __pyx_k_Callable_int_object_bool[] = "Callable[[int, object], bool]";
static const char __pyx_k_Font_draw_glyph_with_pen[] = "Font.draw_glyph_with_pen";
static const char __pyx_k_Font_get_glyph_from_name[] = "Font.get_glyph_from_name";
static const char __pyx_k_Font_get_glyph_h_advance[] = "Font.get_glyph_h_advance";
static const char __pyx_k_Font_get_glyph_v_advance[] = "Font.get_glyph_v_advance";
static const char __pyx_k_Font_get_variation_glyph[] = "Font.get_variation_glyph";
static const char __pyx_k_HBObject___reduce_cython[] = "HBObject.__reduce_cython__";
static const char __pyx_k_PASSTHROUGH_UNRECOGNIZED[] = "PASSTHROUGH_UNRECOGNIZED";
static const char __pyx_k_PRODUCE_UNSAFE_TO_CONCAT[] = "PRODUCE_UNSAFE_TO_CONCAT";
static const char __pyx_k_UNDERBAR_EXTRA_DESCENDER[] = "UNDERBAR_EXTRA_DESCENDER";
static const char __pyx_k_guess_segment_properties[] = "guess_segment_properties";
static const char __pyx_k_ot_color_glyph_has_paint[] = "ot_color_glyph_has_paint";
static const char __pyx_k_set_glyph_h_advance_func[] = "set_glyph_h_advance_func";
static const char __pyx_k_set_glyph_v_advance_func[] = "set_glyph_v_advance_func";
static const char __pyx_k_set_language_from_ot_tag[] = "set_language_from_ot_tag";
static const char __pyx_k_set_linear_gradient_func[] = "set_linear_gradient_func";
static const char __pyx_k_set_push_clip_glyph_func[] = "set_push_clip_glyph_func";
static const char __pyx_k_set_radial_gradient_func[] = "set_radial_gradient_func";
static const char __pyx_k_set_variation_glyph_func[] = "set_variation_glyph_func";
static const char __pyx_k_Callable_int_object_Color[] = "Callable[[int, object], Color]";
static const char __pyx_k_ColorLine___reduce_cython[] = "ColorLine.__reduce_cython__";
static const char __pyx_k_DrawFuncs___reduce_cython[] = "DrawFuncs.__reduce_cython__";
static const char __pyx_k_DrawFuncs_get_glyph_shape[] = "DrawFuncs.get_glyph_shape";
static const char __pyx_k_FontFuncs___reduce_cython[] = "FontFuncs.__reduce_cython__";
static const char __pyx_k_GlyphInfo___reduce_cython[] = "GlyphInfo.__reduce_cython__";
static const char __pyx_k_MapIter___setstate_cython[] = "MapIter.__setstate_cython__";
static const char __pyx_k_PaintFuncs_set_color_func[] = "PaintFuncs.set_color_func";
static const char __pyx_k_PaintFuncs_set_image_func[] = "PaintFuncs.set_image_func";
static const char __pyx_k_RADICAL_KERN_AFTER_DEGREE[] = "RADICAL_KERN_AFTER_DEGREE";
static const char __pyx_k_REMOVE_DEFAULT_IGNORABLES[] = "REMOVE_DEFAULT_IGNORABLES";
static const char __pyx_k_SCRIPT_PERCENT_SCALE_DOWN[] = "SCRIPT_PERCENT_SCALE_DOWN";
static const char __pyx_k_SetIter___setstate_cython[] = "SetIter.__setstate_cython__";
static const char __pyx_k_get_var_coords_normalized[] = "get_var_coords_normalized";
static const char __pyx_k_ot_color_glyph_get_layers[] = "ot_color_glyph_get_layers";
static const char __pyx_k_ot_layout_get_glyph_class[] = "ot_layout_get_glyph_class";
static const char __pyx_k_ot_layout_has_positioning[] = "ot_layout_has_positioning";
static const char __pyx_k_ot_math_get_glyph_kerning[] = "ot_math_get_glyph_kerning";
static const char __pyx_k_set_var_coords_normalized[] = "set_var_coords_normalized";
static const char __pyx_k_DrawFuncs_set_line_to_func[] = "DrawFuncs.set_line_to_func";
static const char __pyx_k_DrawFuncs_set_move_to_func[] = "DrawFuncs.set_move_to_func";
static const char __pyx_k_FRACTION_NUMERATOR_GAP_MIN[] = "FRACTION_NUMERATOR_GAP_MIN";
static const char __pyx_k_Font_get_var_coords_design[] = "Font.get_var_coords_design";
static const char __pyx_k_Font_set_var_coords_design[] = "Font.set_var_coords_design";
static const char __pyx_k_HBObject___setstate_cython[] = "HBObject.__setstate_cython__";
static const char __pyx_k_PaintFuncs___reduce_cython[] = "PaintFuncs.__reduce_cython__";
static const char __pyx_k_RADICAL_KERN_BEFORE_DEGREE[] = "RADICAL_KERN_BEFORE_DEGREE";
static const char __pyx_k_STRETCH_STACK_TOP_SHIFT_UP[] = "STRETCH_STACK_TOP_SHIFT_UP";
static const char __pyx_k_SubsetPlan___reduce_cython[] = "SubsetPlan.__reduce_cython__";
static const char __pyx_k_ot_color_palette_get_count[] = "ot_color_palette_get_count";
static const char __pyx_k_ot_color_palette_get_flags[] = "ot_color_palette_get_flags";
static const char __pyx_k_ot_layout_has_substitution[] = "ot_layout_has_substitution";
static const char __pyx_k_ot_math_get_glyph_assembly[] = "ot_math_get_glyph_assembly";
static const char __pyx_k_ot_math_get_glyph_kernings[] = "ot_math_get_glyph_kernings";
static const char __pyx_k_ot_math_get_glyph_variants[] = "ot_math_get_glyph_variants";
static const char __pyx_k_ColorLine___setstate_cython[] = "ColorLine.__setstate_cython__";
static const char __pyx_k_DISPLAY_OPERATOR_MIN_HEIGHT[] = "DISPLAY_OPERATOR_MIN_HEIGHT";
static const char __pyx_k_DO_NOT_INSERT_DOTTED_CIRCLE[] = "DO_NOT_INSERT_DOTTED_CIRCLE";
static const char __pyx_k_DrawFuncs___setstate_cython[] = "DrawFuncs.__setstate_cython__";
static const char __pyx_k_DrawFuncs_set_cubic_to_func[] = "DrawFuncs.set_cubic_to_func";
static const char __pyx_k_FRACTION_NUMERATOR_SHIFT_UP[] = "FRACTION_NUMERATOR_SHIFT_UP";
static const char __pyx_k_FontFuncs___setstate_cython[] = "FontFuncs.__setstate_cython__";
static const char __pyx_k_GlyphInfo___setstate_cython[] = "GlyphInfo.__setstate_cython__";
static const char __pyx_k_PRESERVE_DEFAULT_IGNORABLES[] = "PRESERVE_DEFAULT_IGNORABLES";
static const char __pyx_k_STACK_DISPLAY_STYLE_GAP_MIN[] = "STACK_DISPLAY_STYLE_GAP_MIN";
static const char __pyx_k_STRETCH_STACK_GAP_ABOVE_MIN[] = "STRETCH_STACK_GAP_ABOVE_MIN";
static const char __pyx_k_STRETCH_STACK_GAP_BELOW_MIN[] = "STRETCH_STACK_GAP_BELOW_MIN";
static const char __pyx_k_SUBSCRIPT_BASELINE_DROP_MIN[] = "SUBSCRIPT_BASELINE_DROP_MIN";
static const char __pyx_k_SubsetInput___reduce_cython[] = "SubsetInput.__reduce_cython__";
static const char __pyx_k_SubsetInput_keep_everything[] = "SubsetInput.keep_everything";
static const char __pyx_k_USABLE_WITH_DARK_BACKGROUND[] = "USABLE_WITH_DARK_BACKGROUND";
static const char __pyx_k_ot_color_palette_get_colors[] = "ot_color_palette_get_colors";
static const char __pyx_k_ot_layout_has_glyph_classes[] = "ot_layout_has_glyph_classes";
static const char __pyx_k_src_uharfbuzz__harfbuzz_pyx[] = "src/uharfbuzz/_harfbuzz.pyx";
static const char __pyx_k_symmetric_difference_update[] = "symmetric_difference_update";
static const char __pyx_k_Callable_Font_int_object_int[] = "Callable[[Font, int, object], int]";
static const char __pyx_k_Callable_Font_int_object_str[] = "Callable[[Font, int, object], str]";
static const char __pyx_k_FLATTENED_ACCENT_BASE_HEIGHT[] = "FLATTENED_ACCENT_BASE_HEIGHT";
static const char __pyx_k_FRACTION_DENOMINATOR_GAP_MIN[] = "FRACTION_DENOMINATOR_GAP_MIN";
static const char __pyx_k_PaintFuncs___setstate_cython[] = "PaintFuncs.__setstate_cython__";
static const char __pyx_k_PaintFuncs_set_pop_clip_func[] = "PaintFuncs.set_pop_clip_func";
static const char __pyx_k_SKEWED_FRACTION_VERTICAL_GAP[] = "SKEWED_FRACTION_VERTICAL_GAP";
static const char __pyx_k_SUPERSCRIPT_SHIFT_UP_CRAMPED[] = "SUPERSCRIPT_SHIFT_UP_CRAMPED";
static const char __pyx_k_SubsetPlan___setstate_cython[] = "SubsetPlan.__setstate_cython__";
static const char __pyx_k_USABLE_WITH_LIGHT_BACKGROUND[] = "USABLE_WITH_LIGHT_BACKGROUND";
static const char __pyx_k_ot_color_palette_get_name_id[] = "ot_color_palette_get_name_id";
static const char __pyx_k_set_push_clip_rectangle_func[] = "set_push_clip_rectangle_func";
static const char __pyx_k_Buffer_set_script_from_ot_tag[] = "Buffer.set_script_from_ot_tag";
static const char __pyx_k_DEFAULT_REPLACEMENT_CODEPOINT[] = "DEFAULT_REPLACEMENT_CODEPOINT";
static const char __pyx_k_DrawFuncs_set_close_path_func[] = "DrawFuncs.set_close_path_func";
static const char __pyx_k_FontFuncs_set_glyph_name_func[] = "FontFuncs.set_glyph_name_func";
static const char __pyx_k_GlyphPosition___reduce_cython[] = "GlyphPosition.__reduce_cython__";
static const char __pyx_k_LOWER_LIMIT_BASELINE_DROP_MIN[] = "LOWER_LIMIT_BASELINE_DROP_MIN";
static const char __pyx_k_PaintFuncs_set_pop_group_func[] = "PaintFuncs.set_pop_group_func";
static const char __pyx_k_SUPERSCRIPT_BASELINE_DROP_MAX[] = "SUPERSCRIPT_BASELINE_DROP_MAX";
static const char __pyx_k_SubsetInput___setstate_cython[] = "SubsetInput.__setstate_cython__";
static const char __pyx_k_SubsetInput_pin_axis_location[] = "SubsetInput.pin_axis_location";
static const char __pyx_k_UPPER_LIMIT_BASELINE_RISE_MIN[] = "UPPER_LIMIT_BASELINE_RISE_MIN";
static const char __pyx_k_set_custom_palette_color_func[] = "set_custom_palette_color_func";
static const char __pyx_k_Callable_Face_str_object_bytes[] = "Callable[[Face, str, object], bytes]";
static const char __pyx_k_Font_get_var_coords_normalized[] = "Font.get_var_coords_normalized";
static const char __pyx_k_Font_set_var_coords_normalized[] = "Font.set_var_coords_normalized";
static const char __pyx_k_PRODUCE_SAFE_TO_INSERT_TATWEEL[] = "PRODUCE_SAFE_TO_INSERT_TATWEEL";
static const char __pyx_k_PaintFuncs_set_push_group_func[] = "PaintFuncs.set_push_group_func";
static const char __pyx_k_SKEWED_FRACTION_HORIZONTAL_GAP[] = "SKEWED_FRACTION_HORIZONTAL_GAP";
static const char __pyx_k_Tuple_List_OTMathGlyphPart_int[] = "Tuple[List[OTMathGlyphPart], int]";
static const char __pyx_k_Buffer_guess_segment_properties[] = "Buffer.guess_segment_properties";
static const char __pyx_k_Buffer_set_language_from_ot_tag[] = "Buffer.set_language_from_ot_tag";
static const char __pyx_k_Callable_Blob_int_int_str_float[] = "Callable[[Blob, int, int, str, float, GlyphExtents, object], bool]";
static const char __pyx_k_Callable_Color_bool_object_None[] = "Callable[[Color, bool, object], None]";
static const char __pyx_k_DrawFuncs_set_quadratic_to_func[] = "DrawFuncs.set_quadratic_to_func";
static const char __pyx_k_FRACTION_DENOMINATOR_SHIFT_DOWN[] = "FRACTION_DENOMINATOR_SHIFT_DOWN";
static const char __pyx_k_GlyphPosition___setstate_cython[] = "GlyphPosition.__setstate_cython__";
static const char __pyx_k_PaintFuncs_set_color_glyph_func[] = "PaintFuncs.set_color_glyph_func";
static const char __pyx_k_Pickling_of_struct_members_such[] = "Pickling of struct members such as self._hb_glyph_info must be explicitly requested with @auto_pickle(True)";
static const char __pyx_k_STRETCH_STACK_BOTTOM_SHIFT_DOWN[] = "STRETCH_STACK_BOTTOM_SHIFT_DOWN";
static const char __pyx_k_Set_symmetric_difference_update[] = "Set.symmetric_difference_update";
static const char __pyx_k_SubsetInput_pin_axis_to_default[] = "SubsetInput.pin_axis_to_default";
static const char __pyx_k_ot_color_palette_color_get_name[] = "ot_color_palette_color_get_name_id";
static const char __pyx_k_ot_layout_table_get_script_tags[] = "ot_layout_table_get_script_tags";
static const char __pyx_k_ot_math_is_glyph_extended_shape[] = "ot_math_is_glyph_extended_shape";
static const char __pyx_k_Callable_ColorLine_float_float_f[] = "Callable[[ColorLine, float, float, float, float, float, float, object], None]";
static const char __pyx_k_Callable_Font_int_int_object_int[] = "Callable[[Font, int, int, object], int]";
static const char __pyx_k_Callable_Font_int_object_int_int[] = "Callable[[Font, int, object], (int, int, int)]";
static const char __pyx_k_Callable_Font_object_FontExtents[] = "Callable[[Font, object], FontExtents]";
static const char __pyx_k_Callable_PaintCompositeMode_obje[] = "Callable[[PaintCompositeMode, object], None]";
static const char __pyx_k_Callable_float_float_float_float[] = "Callable[[float, float, float, float, float, float, object], None]";
static const char __pyx_k_Callable_float_float_object_None[] = "Callable[[float, float, object], None]";
static const char __pyx_k_DELIMITED_SUB_FORMULA_MIN_HEIGHT[] = "DELIMITED_SUB_FORMULA_MIN_HEIGHT";
static const char __pyx_k_Dict_str_Union_int_bool_Sequence[] = "Dict[str, Union[int, bool, Sequence[Tuple[int, int, Union[int, bool]]]]]";
static const char __pyx_k_FRACTION_DENOMINATOR_DISPLAY_STY[] = "FRACTION_DENOMINATOR_DISPLAY_STYLE_SHIFT_DOWN";
static const char __pyx_k_FRACTION_DENOM_DISPLAY_STYLE_GAP[] = "FRACTION_DENOM_DISPLAY_STYLE_GAP_MIN";
static const char __pyx_k_FRACTION_NUMERATOR_DISPLAY_STYLE[] = "FRACTION_NUMERATOR_DISPLAY_STYLE_SHIFT_UP";
static const char __pyx_k_FRACTION_NUM_DISPLAY_STYLE_GAP_M[] = "FRACTION_NUM_DISPLAY_STYLE_GAP_MIN";
static const char __pyx_k_FontFuncs_set_font_h_extents_fun[] = "FontFuncs.set_font_h_extents_func";
static const char __pyx_k_FontFuncs_set_font_v_extents_fun[] = "FontFuncs.set_font_v_extents_func";
static const char __pyx_k_FontFuncs_set_glyph_h_advance_fu[] = "FontFuncs.set_glyph_h_advance_func";
static const char __pyx_k_FontFuncs_set_glyph_v_advance_fu[] = "FontFuncs.set_glyph_v_advance_func";
static const char __pyx_k_FontFuncs_set_glyph_v_origin_fun[] = "FontFuncs.set_glyph_v_origin_func";
static const char __pyx_k_FontFuncs_set_nominal_glyph_func[] = "FontFuncs.set_nominal_glyph_func";
static const char __pyx_k_FontFuncs_set_variation_glyph_fu[] = "FontFuncs.set_variation_glyph_func";
static const char __pyx_k_List_Tuple_List_Tuple_int_int_in[] = "List[Tuple[List[Tuple[int, int, int]], List[Tuple[int, int, int]]]]";
static const char __pyx_k_PaintFuncs_set_custom_palette_co[] = "PaintFuncs.set_custom_palette_color_func";
static const char __pyx_k_PaintFuncs_set_linear_gradient_f[] = "PaintFuncs.set_linear_gradient_func";
static const char __pyx_k_PaintFuncs_set_pop_transform_fun[] = "PaintFuncs.set_pop_transform_func";
static const char __pyx_k_PaintFuncs_set_push_clip_glyph_f[] = "PaintFuncs.set_push_clip_glyph_func";
static const char __pyx_k_PaintFuncs_set_push_clip_rectang[] = "PaintFuncs.set_push_clip_rectangle_func";
static const char __pyx_k_PaintFuncs_set_push_transform_fu[] = "PaintFuncs.set_push_transform_func";
static const char __pyx_k_PaintFuncs_set_radial_gradient_f[] = "PaintFuncs.set_radial_gradient_func";
static const char __pyx_k_PaintFuncs_set_sweep_gradient_fu[] = "PaintFuncs.set_sweep_gradient_func";
static const char __pyx_k_Pass_draw_state_to_Font_draw_gly[] = "Pass draw_state to Font.draw_glyph";
static const char __pyx_k_RADICAL_DEGREE_BOTTOM_RAISE_PERC[] = "RADICAL_DEGREE_BOTTOM_RAISE_PERCENT";
static const char __pyx_k_RADICAL_DISPLAY_STYLE_VERTICAL_G[] = "RADICAL_DISPLAY_STYLE_VERTICAL_GAP";
static const char __pyx_k_SCRIPT_SCRIPT_PERCENT_SCALE_DOWN[] = "SCRIPT_SCRIPT_PERCENT_SCALE_DOWN";
static const char __pyx_k_STACK_BOTTOM_DISPLAY_STYLE_SHIFT[] = "STACK_BOTTOM_DISPLAY_STYLE_SHIFT_DOWN";
static const char __pyx_k_STACK_TOP_DISPLAY_STYLE_SHIFT_UP[] = "STACK_TOP_DISPLAY_STYLE_SHIFT_UP";
static const char __pyx_k_SUPERSCRIPT_BOTTOM_MAX_WITH_SUBS[] = "SUPERSCRIPT_BOTTOM_MAX_WITH_SUBSCRIPT";
static const char __pyx_k_draw_glyph_is_deprecated_use_Fon[] = "draw_glyph() is deprecated, use Font.draw_glyph() instead";
static const char __pyx_k_get_glyph_shape_is_deprecated_us[] = "get_glyph_shape() is deprecated, use Font.draw_glyph() instead";
static const char __pyx_k_index_out_of_range_for_feature_a[] = "index out of range for feature array capacity";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static const char __pyx_k_ot_layout_language_get_feature_t[] = "ot_layout_language_get_feature_tags";
static const char __pyx_k_ot_layout_lookup_get_glyph_alter[] = "ot_layout_lookup_get_glyph_alternates";
static const char __pyx_k_ot_layout_script_get_language_ta[] = "ot_layout_script_get_language_tags";
static const char __pyx_k_ot_math_get_glyph_italics_correc[] = "ot_math_get_glyph_italics_correction";
static const char __pyx_k_ot_math_get_glyph_top_accent_att[] = "ot_math_get_glyph_top_accent_attachment";
static const char __pyx_k_ot_math_get_min_connector_overla[] = "ot_math_get_min_connector_overlap";
static const char __pyx_k_Pickling_of_struct_members_such_2[] = "Pickling of struct members such as self._hb_glyph_position must be explicitly requested with @auto_pickle(True)";
static const char __pyx_k_Callable_ColorLine_float_float_f_2[] = "Callable[[ColorLine, float, float, float, float, object], None]";
static const char __pyx_k_Callable_float_float_float_float_2[] = "Callable[[float, float, float, float, object], None]";
/* #### Code section: decls ### */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_version_string(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9GlyphInfo_9codepoint___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphInfo *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9GlyphInfo_7cluster___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphInfo *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9GlyphInfo_5flags___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphInfo *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9GlyphInfo___reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphInfo *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9GlyphInfo_2__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphInfo *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_13GlyphPosition_8position___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_13GlyphPosition_9x_advance___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_13GlyphPosition_9y_advance___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_13GlyphPosition_8x_offset___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_13GlyphPosition_8y_offset___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_13GlyphPosition___reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_13GlyphPosition_2__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer___cinit__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self); /* proto */
static void __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_2__dealloc__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_4create(PyTypeObject *__pyx_v_cls); /* proto */
static Py_ssize_t __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_6__len__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_8reset(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_10clear_contents(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_9direction___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_9direction_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_11glyph_infos___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_15glyph_positions___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_8language___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_8language_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_6script___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_6script_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_5flags___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_5flags_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_13cluster_level___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_13cluster_level_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_12content_type___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_12content_type_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_21replacement_codepoint___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_21replacement_codepoint_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_15invisible_glyph___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_15invisible_glyph_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_15not_found_glyph___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_15not_found_glyph_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_12set_language_from_ot_tag(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_14set_script_from_ot_tag(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_16add_codepoints(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self, PyObject *__pyx_v_codepoints, PyObject *__pyx_v_item_offset, PyObject *__pyx_v_item_length); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_18add_utf8(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self, PyObject *__pyx_v_text, PyObject *__pyx_v_item_offset, PyObject *__pyx_v_item_length); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_20add_str(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self, PyObject *__pyx_v_text, PyObject *__pyx_v_item_offset, PyObject *__pyx_v_item_length); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_22guess_segment_properties(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_24set_message_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self, PyObject *__pyx_v_callback); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_26__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_28__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_4Blob___cinit__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *__pyx_v_self, PyObject *__pyx_v_data); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Blob_2from_file_path(PyTypeObject *__pyx_v_cls, PyObject *__pyx_v_filename); /* proto */
static void __pyx_pf_9uharfbuzz_9_harfbuzz_4Blob_4__dealloc__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_9uharfbuzz_9_harfbuzz_4Blob_6__len__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *__pyx_v_self); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_4Blob_8__bool__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Blob_4data___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Blob_10__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Blob_12__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_4Face___cinit__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_self, PyObject *__pyx_v_blob, int __pyx_v_index); /* proto */
static void __pyx_pf_9uharfbuzz_9_harfbuzz_4Face_2__dealloc__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Face_4create(PyTypeObject *__pyx_v_cls, PyObject *__pyx_v_blob, int __pyx_v_index); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Face_6create_for_tables(PyTypeObject *__pyx_v_cls, PyObject *__pyx_v_func, PyObject *__pyx_v_user_data); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Face_5index___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_self); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_4Face_5index_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Face_4upem___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_self); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_4Face_4upem_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Face_11glyph_count___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_self); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_4Face_11glyph_count_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Face_4blob___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Face_10table_tags___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Face_8unicodes___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Face_19variation_selectors___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Face_8variation_unicodes(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_self, PyObject *__pyx_v_variation_selector); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Face_10__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Face_12__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_4Font___cinit__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_face_or_font); /* proto */
static void __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_2__dealloc__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_4create(PyTypeObject *__pyx_v_cls, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_4face___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_5funcs___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_5funcs_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *__pyx_v_ffuncs); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_5scale___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_5scale_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_4ppem___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_4ppem_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_4ptem___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_4ptem_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, double __pyx_v_value); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_15synthetic_slant___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_15synthetic_slant_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, double __pyx_v_value); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_14synthetic_bold___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_14synthetic_bold_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_18var_named_instance___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_18var_named_instance_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_6set_variations(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_variations); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_8set_variation(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_name, double __pyx_v_value); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_10get_glyph_name(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_gid); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_12get_glyph_from_name(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_14get_glyph_extents(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_gid); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_16get_glyph_h_advance(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_gid); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_18get_glyph_v_advance(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_gid); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_20get_glyph_h_origin(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_gid); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_22get_glyph_v_origin(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_gid); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_24get_font_extents(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_direction); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_26get_variation_glyph(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_unicode, PyObject *__pyx_v_variation_selector); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_28get_nominal_glyph(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_unicode); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_30get_var_coords_normalized(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_32set_var_coords_normalized(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_coords); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_34get_var_coords_design(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_36set_var_coords_design(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_coords); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_38glyph_to_string(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_gid); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_40glyph_from_string(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_string); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_42draw_glyph(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_gid, struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *__pyx_v_draw_funcs, PyObject *__pyx_v_draw_state); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_44paint_glyph(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_gid, struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_paint_funcs, PyObject *__pyx_v_paint_state, PyObject *__pyx_v_palette_index, PyObject *__pyx_v_foreground); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_46draw_glyph_with_pen(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_gid, PyObject *__pyx_v_pen); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_48__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_50__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs___cinit__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *__pyx_v_self); /* proto */
static void __pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_2__dealloc__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_4create(PyTypeObject *__pyx_v_cls); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_6set_glyph_h_advance_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *__pyx_v_self, PyObject *__pyx_v_func, PyObject *__pyx_v_user_data); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_8set_glyph_v_advance_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *__pyx_v_self, PyObject *__pyx_v_func, PyObject *__pyx_v_user_data); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_10set_glyph_v_origin_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *__pyx_v_self, PyObject *__pyx_v_func, PyObject *__pyx_v_user_data); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_12set_glyph_name_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *__pyx_v_self, PyObject *__pyx_v_func, PyObject *__pyx_v_user_data); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_14set_nominal_glyph_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *__pyx_v_self, PyObject *__pyx_v_func, PyObject *__pyx_v_user_data); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_16set_variation_glyph_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *__pyx_v_self, PyObject *__pyx_v_func, PyObject *__pyx_v_user_data); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_18set_font_h_extents_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *__pyx_v_self, PyObject *__pyx_v_func, PyObject *__pyx_v_user_data); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_20set_font_v_extents_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *__pyx_v_self, PyObject *__pyx_v_func, PyObject *__pyx_v_user_data); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_22__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_24__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_2shape(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_buffer, PyObject *__pyx_v_features, PyObject *__pyx_v_shapers); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4ot_tag_to_script(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_tag); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6ot_tag_to_language(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_tag); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_8ot_layout_lookup_get_glyph_alternates(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, PyObject *__pyx_v_lookup_index, hb_codepoint_t __pyx_v_glyph); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_84__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10ot_layout_language_get_feature_tags(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, PyObject *__pyx_v_tag, PyObject *__pyx_v_script_index, PyObject *__pyx_v_language_index); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_86__defaults__(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_12ot_layout_script_get_language_tags(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, PyObject *__pyx_v_tag, PyObject *__pyx_v_script_index); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_14ot_layout_table_get_script_tags(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, PyObject *__pyx_v_tag); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_16ot_layout_get_baseline(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font, PyObject *__pyx_v_baseline_tag, PyObject *__pyx_v_direction, PyObject *__pyx_v_script_tag, PyObject *__pyx_v_language_tag); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_18ot_layout_has_glyph_classes(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_20ot_layout_has_positioning(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_22ot_layout_has_substitution(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_24ot_layout_get_glyph_class(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, PyObject *__pyx_v_glyph); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_26ot_color_has_palettes(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_28ot_color_palette_get_count(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_30ot_color_palette_get_flags(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, PyObject *__pyx_v_palette_index); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_32ot_color_palette_get_colors(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, PyObject *__pyx_v_palette_index); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_34ot_color_palette_get_name_id(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, PyObject *__pyx_v_palette_index); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_36ot_color_palette_color_get_name_id(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, PyObject *__pyx_v_color_index); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_38ot_color_has_layers(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_40ot_color_glyph_get_layers(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, PyObject *__pyx_v_glyph); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_42ot_color_has_paint(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_44ot_color_glyph_has_paint(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, PyObject *__pyx_v_glyph); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_46ot_color_has_svg(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face); /* proto */
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *__pyx_pf_9uharfbuzz_9_harfbuzz_48ot_color_glyph_get_svg(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, PyObject *__pyx_v_glyph); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_50ot_color_has_png(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face); /* proto */
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *__pyx_pf_9uharfbuzz_9_harfbuzz_52ot_color_glyph_get_png(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font, PyObject *__pyx_v_glyph); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_54ot_math_has_data(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_56ot_math_get_constant(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font, PyObject *__pyx_v_constant); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_58ot_math_get_glyph_italics_correction(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font, PyObject *__pyx_v_glyph); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_60ot_math_get_glyph_top_accent_attachment(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font, PyObject *__pyx_v_glyph); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_62ot_math_is_glyph_extended_shape(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, PyObject *__pyx_v_glyph); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_64ot_math_get_min_connector_overlap(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font, PyObject *__pyx_v_direction); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_66ot_math_get_glyph_kerning(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font, PyObject *__pyx_v_glyph, PyObject *__pyx_v_kern, int __pyx_v_correction_height); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_68ot_math_get_glyph_kernings(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font, PyObject *__pyx_v_glyph, PyObject *__pyx_v_kern); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_70ot_math_get_glyph_variants(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font, PyObject *__pyx_v_glyph, PyObject *__pyx_v_direction); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_72ot_math_get_glyph_assembly(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font, PyObject *__pyx_v_glyph, PyObject *__pyx_v_direction); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_74ot_font_set_funcs(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_5Color_to_int(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_5Color_2from_int(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_value); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_9ColorLine___cinit__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9ColorLine_11color_stops___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9ColorLine_6extend___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9ColorLine_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9ColorLine_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs___cinit__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self); /* proto */
static void __pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_2__dealloc__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_4set_push_transform_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self, PyObject *__pyx_v_func); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_6set_pop_transform_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self, PyObject *__pyx_v_func); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_8set_color_glyph_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self, PyObject *__pyx_v_func); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_10set_push_clip_glyph_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self, PyObject *__pyx_v_func); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_12set_push_clip_rectangle_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self, PyObject *__pyx_v_func); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_14set_pop_clip_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self, PyObject *__pyx_v_func); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_16set_color_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self, PyObject *__pyx_v_func); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_18set_image_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self, PyObject *__pyx_v_func); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_20set_linear_gradient_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self, PyObject *__pyx_v_func); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_22set_radial_gradient_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self, PyObject *__pyx_v_func); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_24set_sweep_gradient_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self, PyObject *__pyx_v_func); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_26set_push_group_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self, PyObject *__pyx_v_func); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_28set_pop_group_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self, PyObject *__pyx_v_func); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_30set_custom_palette_color_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self, PyObject *__pyx_v_func); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_32__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_34__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_9DrawFuncs___cinit__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *__pyx_v_self); /* proto */
static void __pyx_pf_9uharfbuzz_9_harfbuzz_9DrawFuncs_2__dealloc__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9DrawFuncs_4get_glyph_shape(struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font, PyObject *__pyx_v_gid); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9DrawFuncs_6draw_glyph(struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font, PyObject *__pyx_v_gid, PyObject *__pyx_v_draw_data); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9DrawFuncs_8set_move_to_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *__pyx_v_self, PyObject *__pyx_v_func, PyObject *__pyx_v_user_data); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9DrawFuncs_10set_line_to_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *__pyx_v_self, PyObject *__pyx_v_func, PyObject *__pyx_v_user_data); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9DrawFuncs_12set_cubic_to_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *__pyx_v_self, PyObject *__pyx_v_func, PyObject *__pyx_v_user_data); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9DrawFuncs_14set_quadratic_to_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *__pyx_v_self, PyObject *__pyx_v_func, PyObject *__pyx_v_user_data); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9DrawFuncs_16set_close_path_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *__pyx_v_self, PyObject *__pyx_v_func, PyObject *__pyx_v_user_data); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9DrawFuncs_18__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9DrawFuncs_20__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_8HBObject___cinit__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_HBObject *__pyx_v_self, PyObject *__pyx_v_num_nodes); /* proto */
static void __pyx_pf_9uharfbuzz_9_harfbuzz_8HBObject_2__dealloc__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_HBObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_8HBObject_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_HBObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_8HBObject_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_HBObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_76repack(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_subtables, PyObject *__pyx_v_graphnodes); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_78repack_with_tag(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_tag, PyObject *__pyx_v_subtables, PyObject *__pyx_v_graphnodes); /* proto */
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_pf_9uharfbuzz_9_harfbuzz_80subset_preprocess(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face); /* proto */
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_pf_9uharfbuzz_9_harfbuzz_82subset(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_input); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput___cinit__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self); /* proto */
static void __pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_2__dealloc__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self); /* proto */
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_4subset(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_source); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_6keep_everything(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_8pin_axis_to_default(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, PyObject *__pyx_v_tag); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_10pin_axis_location(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, PyObject *__pyx_v_tag, double __pyx_v_value); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_11unicode_set___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_9glyph_set___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self); /* proto */
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_12sets(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self, PyObject *__pyx_v_set_type); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_23no_subset_table_tag_set___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_18drop_table_tag_set___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_11name_id_set___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_16name_lang_id_set___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_22layout_feature_tag_set___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_21layout_script_tag_set___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_5flags___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_5flags_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self, PyObject *__pyx_v_flags); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_14__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_16__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_10SubsetPlan___cinit__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetPlan *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_input); /* proto */
static void __pyx_pf_9uharfbuzz_9_harfbuzz_10SubsetPlan_2__dealloc__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetPlan *__pyx_v_self); /* proto */
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_pf_9uharfbuzz_9_harfbuzz_10SubsetPlan_4execute(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetPlan *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10SubsetPlan_24old_to_new_glyph_mapping___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetPlan *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10SubsetPlan_24new_to_old_glyph_mapping___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetPlan *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10SubsetPlan_28unicode_to_old_glyph_mapping___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetPlan *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10SubsetPlan_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetPlan *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10SubsetPlan_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetPlan *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_3Set___cinit__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, PyObject *__pyx_v_init); /* proto */
static void __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_2__dealloc__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self); /* proto */
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_4copy(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self); /* proto */
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_6__copy__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_8clear(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_10__bool__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_12invert(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_14is_inverted(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_16__contains__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, PyObject *__pyx_v_c); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_18add(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, PyObject *__pyx_v_c); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_20add_range(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, PyObject *__pyx_v_first, PyObject *__pyx_v_last); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_22remove(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, PyObject *__pyx_v_c); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_24discard(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, PyObject *__pyx_v_c); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_26del_range(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, PyObject *__pyx_v_first, PyObject *__pyx_v_last); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_28_is_equal(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_30__eq__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_32issubset(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_larger_set); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_34issuperset(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_smaller_set); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_36_set(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_38set(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_40_update(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_42update(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_44__ior__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_46intersection_update(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_48__iand__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_50difference_update(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_52__isub__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_54symmetric_difference_update(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_56__ixor__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_other); /* proto */
static Py_ssize_t __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_58__len__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_3min___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_3max___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_60__iter__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_8__repr___genexpr(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_genexpr_arg_0); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_62__repr__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_64__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_66__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_7SetIter___cinit__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SetIter *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_s); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_7SetIter_2__iter__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SetIter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_7SetIter_4__next__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SetIter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_7SetIter_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_SetIter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_7SetIter_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_SetIter *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_3Map___cinit__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self, PyObject *__pyx_v_init); /* proto */
static void __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_2__dealloc__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self); /* proto */
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_4copy(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self); /* proto */
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_6__copy__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_8_update(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_10update(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_12clear(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_14__bool__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_16__len__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_18_is_equal(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_other); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_20__eq__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self, PyObject *__pyx_v_other); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_22__setitem__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self, PyObject *__pyx_v_k, PyObject *__pyx_v_v); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_24get(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self, PyObject *__pyx_v_k); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_26__getitem__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self, PyObject *__pyx_v_k); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_28__contains__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self, PyObject *__pyx_v_k); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_30__delitem__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self, PyObject *__pyx_v_c); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_32items(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_4keys_genexpr(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_genexpr_arg_0); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_34keys(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_6values_genexpr(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_genexpr_arg_0); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_36values(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_38__iter__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_8__repr___genexpr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_40__repr__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_42__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_44__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_9uharfbuzz_9_harfbuzz_7MapIter___cinit__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_MapIter *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_m); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_7MapIter_2__iter__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_MapIter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_7MapIter_4__next__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_MapIter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_7MapIter_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_MapIter *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_7MapIter_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_MapIter *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_GlyphInfo(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_GlyphPosition(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_Buffer(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_Blob(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_Face(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_Font(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_FontFuncs(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_ColorLine(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_PaintFuncs(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_DrawFuncs(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_HBObject(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_SubsetInput(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_SubsetPlan(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_Set(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_SetIter(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_Map(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_MapIter(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  #ifdef __Pyx_CyFunction_USED
  PyTypeObject *__pyx_CyFunctionType;
  #endif
  #ifdef __Pyx_FusedFunction_USED
  PyTypeObject *__pyx_FusedFunctionType;
  #endif
  #ifdef __Pyx_Generator_USED
  PyTypeObject *__pyx_GeneratorType;
  #endif
  #ifdef __Pyx_IterableCoroutine_USED
  PyTypeObject *__pyx_IterableCoroutineType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineAwaitType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineType;
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  PyObject *__pyx_type_9uharfbuzz_9_harfbuzz_GlyphInfo;
  PyObject *__pyx_type_9uharfbuzz_9_harfbuzz_GlyphPosition;
  PyObject *__pyx_type_9uharfbuzz_9_harfbuzz_Buffer;
  PyObject *__pyx_type_9uharfbuzz_9_harfbuzz_Blob;
  PyObject *__pyx_type_9uharfbuzz_9_harfbuzz_Face;
  PyObject *__pyx_type_9uharfbuzz_9_harfbuzz_Font;
  PyObject *__pyx_type_9uharfbuzz_9_harfbuzz_FontFuncs;
  PyObject *__pyx_type_9uharfbuzz_9_harfbuzz_ColorLine;
  PyObject *__pyx_type_9uharfbuzz_9_harfbuzz_PaintFuncs;
  PyObject *__pyx_type_9uharfbuzz_9_harfbuzz_DrawFuncs;
  PyObject *__pyx_type_9uharfbuzz_9_harfbuzz_HBObject;
  PyObject *__pyx_type_9uharfbuzz_9_harfbuzz_SubsetInput;
  PyObject *__pyx_type_9uharfbuzz_9_harfbuzz_SubsetPlan;
  PyObject *__pyx_type_9uharfbuzz_9_harfbuzz_Set;
  PyObject *__pyx_type_9uharfbuzz_9_harfbuzz_SetIter;
  PyObject *__pyx_type_9uharfbuzz_9_harfbuzz_Map;
  PyObject *__pyx_type_9uharfbuzz_9_harfbuzz_MapIter;
  PyObject *__pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr;
  PyObject *__pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr;
  PyObject *__pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr;
  PyObject *__pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr;
  #endif
  PyTypeObject *__pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphInfo;
  PyTypeObject *__pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphPosition;
  PyTypeObject *__pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer;
  PyTypeObject *__pyx_ptype_9uharfbuzz_9_harfbuzz_Blob;
  PyTypeObject *__pyx_ptype_9uharfbuzz_9_harfbuzz_Face;
  PyTypeObject *__pyx_ptype_9uharfbuzz_9_harfbuzz_Font;
  PyTypeObject *__pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs;
  PyTypeObject *__pyx_ptype_9uharfbuzz_9_harfbuzz_ColorLine;
  PyTypeObject *__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs;
  PyTypeObject *__pyx_ptype_9uharfbuzz_9_harfbuzz_DrawFuncs;
  PyTypeObject *__pyx_ptype_9uharfbuzz_9_harfbuzz_HBObject;
  PyTypeObject *__pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetInput;
  PyTypeObject *__pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetPlan;
  PyTypeObject *__pyx_ptype_9uharfbuzz_9_harfbuzz_Set;
  PyTypeObject *__pyx_ptype_9uharfbuzz_9_harfbuzz_SetIter;
  PyTypeObject *__pyx_ptype_9uharfbuzz_9_harfbuzz_Map;
  PyTypeObject *__pyx_ptype_9uharfbuzz_9_harfbuzz_MapIter;
  PyTypeObject *__pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr;
  PyTypeObject *__pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr;
  PyTypeObject *__pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr;
  PyTypeObject *__pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr;
  PyObject *__pyx_n_s_ACCENT_BASE_HEIGHT;
  PyObject *__pyx_n_s_AXIS_HEIGHT;
  PyObject *__pyx_kp_u_All_shapers_failed;
  PyObject *__pyx_n_s_AssertionError;
  PyObject *__pyx_n_s_BASE_GLYPH;
  PyObject *__pyx_n_s_BOT;
  PyObject *__pyx_n_s_BOTTOM_LEFT;
  PyObject *__pyx_n_s_BOTTOM_RIGHT;
  PyObject *__pyx_n_s_Blob;
  PyObject *__pyx_n_s_Blob___reduce_cython;
  PyObject *__pyx_n_s_Blob___setstate_cython;
  PyObject *__pyx_n_s_Blob_from_file_path;
  PyObject *__pyx_n_s_Buffer;
  PyObject *__pyx_n_s_BufferClusterLevel;
  PyObject *__pyx_n_s_BufferContentType;
  PyObject *__pyx_n_s_BufferFlags;
  PyObject *__pyx_n_s_Buffer___reduce_cython;
  PyObject *__pyx_n_s_Buffer___setstate_cython;
  PyObject *__pyx_n_s_Buffer_add_codepoints;
  PyObject *__pyx_n_s_Buffer_add_str;
  PyObject *__pyx_n_s_Buffer_add_utf8;
  PyObject *__pyx_n_s_Buffer_clear_contents;
  PyObject *__pyx_n_s_Buffer_create;
  PyObject *__pyx_n_s_Buffer_guess_segment_properties;
  PyObject *__pyx_n_s_Buffer_reset;
  PyObject *__pyx_n_s_Buffer_set_language_from_ot_tag;
  PyObject *__pyx_n_s_Buffer_set_message_func;
  PyObject *__pyx_n_s_Buffer_set_script_from_ot_tag;
  PyObject *__pyx_n_s_CHARACTERS;
  PyObject *__pyx_n_s_CLEAR;
  PyObject *__pyx_n_s_COLOR_BURN;
  PyObject *__pyx_n_s_COLOR_DODGE;
  PyObject *__pyx_n_s_COMPONENT;
  PyObject *__pyx_n_s_Callable;
  PyObject *__pyx_kp_s_Callable_Blob_int_int_str_float;
  PyObject *__pyx_kp_s_Callable_ColorLine_float_float_f;
  PyObject *__pyx_kp_s_Callable_ColorLine_float_float_f_2;
  PyObject *__pyx_kp_s_Callable_Color_bool_object_None;
  PyObject *__pyx_kp_s_Callable_Face_str_object_bytes;
  PyObject *__pyx_kp_s_Callable_Font_int_int_object_int;
  PyObject *__pyx_kp_s_Callable_Font_int_object_int;
  PyObject *__pyx_kp_s_Callable_Font_int_object_int_int;
  PyObject *__pyx_kp_s_Callable_Font_int_object_str;
  PyObject *__pyx_kp_s_Callable_Font_object_FontExtents;
  PyObject *__pyx_kp_s_Callable_PaintCompositeMode_obje;
  PyObject *__pyx_kp_s_Callable_float_float_float_float;
  PyObject *__pyx_kp_s_Callable_float_float_float_float_2;
  PyObject *__pyx_kp_s_Callable_float_float_object_None;
  PyObject *__pyx_kp_s_Callable_int_object_Color;
  PyObject *__pyx_kp_s_Callable_int_object_None;
  PyObject *__pyx_kp_s_Callable_int_object_bool;
  PyObject *__pyx_kp_s_Callable_object_None;
  PyObject *__pyx_n_s_Color;
  PyObject *__pyx_n_u_Color;
  PyObject *__pyx_n_s_ColorLine;
  PyObject *__pyx_n_s_ColorLine___reduce_cython;
  PyObject *__pyx_n_s_ColorLine___setstate_cython;
  PyObject *__pyx_n_s_ColorStop;
  PyObject *__pyx_n_u_ColorStop;
  PyObject *__pyx_kp_s_Color_None;
  PyObject *__pyx_n_s_Color_from_int;
  PyObject *__pyx_n_s_Color_to_int;
  PyObject *__pyx_n_s_DARKEN;
  PyObject *__pyx_n_s_DEFAULT;
  PyObject *__pyx_n_s_DEFAULT_REPLACEMENT_CODEPOINT;
  PyObject *__pyx_n_s_DELIMITED_SUB_FORMULA_MIN_HEIGHT;
  PyObject *__pyx_n_s_DEST;
  PyObject *__pyx_n_s_DEST_ATOP;
  PyObject *__pyx_n_s_DEST_IN;
  PyObject *__pyx_n_s_DEST_OUT;
  PyObject *__pyx_n_s_DEST_OVER;
  PyObject *__pyx_n_s_DESUBROUTINIZE;
  PyObject *__pyx_n_s_DIFFERENCE;
  PyObject *__pyx_n_s_DISPLAY_OPERATOR_MIN_HEIGHT;
  PyObject *__pyx_n_s_DO_NOT_INSERT_DOTTED_CIRCLE;
  PyObject *__pyx_n_s_DROP_TABLE_TAG;
  PyObject *__pyx_n_s_DeprecationWarning;
  PyObject *__pyx_n_s_Dict;
  PyObject *__pyx_kp_s_Dict_str_Union_int_bool_Sequence;
  PyObject *__pyx_kp_s_Dict_str_float;
  PyObject *__pyx_n_s_DrawFuncs;
  PyObject *__pyx_n_s_DrawFuncs___reduce_cython;
  PyObject *__pyx_n_s_DrawFuncs___setstate_cython;
  PyObject *__pyx_n_s_DrawFuncs_draw_glyph;
  PyObject *__pyx_n_s_DrawFuncs_get_glyph_shape;
  PyObject *__pyx_n_s_DrawFuncs_set_close_path_func;
  PyObject *__pyx_n_s_DrawFuncs_set_cubic_to_func;
  PyObject *__pyx_n_s_DrawFuncs_set_line_to_func;
  PyObject *__pyx_n_s_DrawFuncs_set_move_to_func;
  PyObject *__pyx_n_s_DrawFuncs_set_quadratic_to_func;
  PyObject *__pyx_n_s_EOT;
  PyObject *__pyx_n_s_EXCLUSION;
  PyObject *__pyx_n_s_EXTENDER;
  PyObject *__pyx_n_s_FLATTENED_ACCENT_BASE_HEIGHT;
  PyObject *__pyx_n_s_FRACTION_DENOMINATOR_DISPLAY_STY;
  PyObject *__pyx_n_s_FRACTION_DENOMINATOR_GAP_MIN;
  PyObject *__pyx_n_s_FRACTION_DENOMINATOR_SHIFT_DOWN;
  PyObject *__pyx_n_s_FRACTION_DENOM_DISPLAY_STYLE_GAP;
  PyObject *__pyx_n_s_FRACTION_NUMERATOR_DISPLAY_STYLE;
  PyObject *__pyx_n_s_FRACTION_NUMERATOR_GAP_MIN;
  PyObject *__pyx_n_s_FRACTION_NUMERATOR_SHIFT_UP;
  PyObject *__pyx_n_s_FRACTION_NUM_DISPLAY_STYLE_GAP_M;
  PyObject *__pyx_n_s_FRACTION_RULE_THICKNESS;
  PyObject *__pyx_n_s_Face;
  PyObject *__pyx_n_s_Face___reduce_cython;
  PyObject *__pyx_n_s_Face___setstate_cython;
  PyObject *__pyx_n_s_Face_create;
  PyObject *__pyx_n_s_Face_create_for_tables;
  PyObject *__pyx_n_s_Face_variation_unicodes;
  PyObject *__pyx_kp_u_Failed_to_open;
  PyObject *__pyx_n_s_Font;
  PyObject *__pyx_n_s_FontExtents;
  PyObject *__pyx_n_u_FontExtents;
  PyObject *__pyx_n_s_FontFuncs;
  PyObject *__pyx_n_s_FontFuncs___reduce_cython;
  PyObject *__pyx_n_s_FontFuncs___setstate_cython;
  PyObject *__pyx_n_s_FontFuncs_create;
  PyObject *__pyx_n_s_FontFuncs_set_font_h_extents_fun;
  PyObject *__pyx_n_s_FontFuncs_set_font_v_extents_fun;
  PyObject *__pyx_n_s_FontFuncs_set_glyph_h_advance_fu;
  PyObject *__pyx_n_s_FontFuncs_set_glyph_name_func;
  PyObject *__pyx_n_s_FontFuncs_set_glyph_v_advance_fu;
  PyObject *__pyx_n_s_FontFuncs_set_glyph_v_origin_fun;
  PyObject *__pyx_n_s_FontFuncs_set_nominal_glyph_func;
  PyObject *__pyx_n_s_FontFuncs_set_variation_glyph_fu;
  PyObject *__pyx_n_s_Font___reduce_cython;
  PyObject *__pyx_n_s_Font___setstate_cython;
  PyObject *__pyx_n_s_Font_create;
  PyObject *__pyx_n_s_Font_draw_glyph;
  PyObject *__pyx_n_s_Font_draw_glyph_with_pen;
  PyObject *__pyx_n_s_Font_get_font_extents;
  PyObject *__pyx_n_s_Font_get_glyph_extents;
  PyObject *__pyx_n_s_Font_get_glyph_from_name;
  PyObject *__pyx_n_s_Font_get_glyph_h_advance;
  PyObject *__pyx_n_s_Font_get_glyph_h_origin;
  PyObject *__pyx_n_s_Font_get_glyph_name;
  PyObject *__pyx_n_s_Font_get_glyph_v_advance;
  PyObject *__pyx_n_s_Font_get_glyph_v_origin;
  PyObject *__pyx_n_s_Font_get_nominal_glyph;
  PyObject *__pyx_n_s_Font_get_var_coords_design;
  PyObject *__pyx_n_s_Font_get_var_coords_normalized;
  PyObject *__pyx_n_s_Font_get_variation_glyph;
  PyObject *__pyx_n_s_Font_glyph_from_string;
  PyObject *__pyx_n_s_Font_glyph_to_string;
  PyObject *__pyx_n_s_Font_paint_glyph;
  PyObject *__pyx_n_s_Font_set_var_coords_design;
  PyObject *__pyx_n_s_Font_set_var_coords_normalized;
  PyObject *__pyx_n_s_Font_set_variation;
  PyObject *__pyx_n_s_Font_set_variations;
  PyObject *__pyx_n_s_GLYPHS;
  PyObject *__pyx_n_s_GLYPH_INDEX;
  PyObject *__pyx_n_s_GLYPH_NAMES;
  PyObject *__pyx_n_s_GlyphExtents;
  PyObject *__pyx_n_u_GlyphExtents;
  PyObject *__pyx_n_s_GlyphFlags;
  PyObject *__pyx_n_s_GlyphInfo;
  PyObject *__pyx_n_s_GlyphInfo___reduce_cython;
  PyObject *__pyx_n_s_GlyphInfo___setstate_cython;
  PyObject *__pyx_n_s_GlyphPosition;
  PyObject *__pyx_n_s_GlyphPosition___reduce_cython;
  PyObject *__pyx_n_s_GlyphPosition___setstate_cython;
  PyObject *__pyx_n_s_HARD_LIGHT;
  PyObject *__pyx_n_s_HBObject;
  PyObject *__pyx_n_s_HBObject___reduce_cython;
  PyObject *__pyx_n_s_HBObject___setstate_cython;
  PyObject *__pyx_n_s_HSL_COLOR;
  PyObject *__pyx_n_s_HSL_HUE;
  PyObject *__pyx_n_s_HSL_LUMINOSITY;
  PyObject *__pyx_n_s_HSL_SATURATION;
  PyObject *__pyx_n_s_HarfBuzzError;
  PyObject *__pyx_n_s_INVALID;
  PyObject *__pyx_n_s_INVALID_VALUE;
  PyObject *__pyx_kp_u_Input_num_of_subtables;
  PyObject *__pyx_n_s_IntEnum;
  PyObject *__pyx_n_s_IntFlag;
  PyObject *__pyx_n_s_KeyError;
  PyObject *__pyx_n_s_LAYOUT_FEATURE_TAG;
  PyObject *__pyx_n_s_LAYOUT_SCRIPT_TAG;
  PyObject *__pyx_n_s_LIGATURE;
  PyObject *__pyx_n_s_LIGHTEN;
  PyObject *__pyx_n_s_LOWER_LIMIT_BASELINE_DROP_MIN;
  PyObject *__pyx_n_s_LOWER_LIMIT_GAP_MIN;
  PyObject *__pyx_n_s_List;
  PyObject *__pyx_kp_s_List_Color;
  PyObject *__pyx_kp_s_List_OTColorLayer;
  PyObject *__pyx_kp_s_List_OTMathGlyphVariant;
  PyObject *__pyx_kp_s_List_OTMathKernEntry;
  PyObject *__pyx_kp_s_List_Tuple_List_Tuple_int_int_in;
  PyObject *__pyx_kp_s_List_bytes;
  PyObject *__pyx_kp_s_List_int;
  PyObject *__pyx_kp_s_List_str;
  PyObject *__pyx_n_s_MARK;
  PyObject *__pyx_n_s_MATH_LEADING;
  PyObject *__pyx_n_s_MONOTONE_CHARACTERS;
  PyObject *__pyx_n_s_MONOTONE_GRAPHEMES;
  PyObject *__pyx_n_s_MULTIPLY;
  PyObject *__pyx_n_s_Map;
  PyObject *__pyx_n_s_MapIter;
  PyObject *__pyx_n_s_MapIter___reduce_cython;
  PyObject *__pyx_n_s_MapIter___setstate_cython;
  PyObject *__pyx_n_s_Map___copy;
  PyObject *__pyx_n_s_Map___reduce_cython;
  PyObject *__pyx_n_s_Map___setstate_cython;
  PyObject *__pyx_n_s_Map__is_equal;
  PyObject *__pyx_n_s_Map__update;
  PyObject *__pyx_n_s_Map_clear;
  PyObject *__pyx_n_s_Map_copy;
  PyObject *__pyx_n_s_Map_get;
  PyObject *__pyx_n_s_Map_items;
  PyObject *__pyx_n_s_Map_keys;
  PyObject *__pyx_kp_u_Map_s;
  PyObject *__pyx_n_s_Map_update;
  PyObject *__pyx_n_s_Map_values;
  PyObject *__pyx_n_s_MemoryError;
  PyObject *__pyx_n_s_NAME_ID;
  PyObject *__pyx_n_s_NAME_LANG_ID;
  PyObject *__pyx_n_s_NAME_LEGACY;
  PyObject *__pyx_n_s_NOTDEF_OUTLINE;
  PyObject *__pyx_n_s_NO_HINTING;
  PyObject *__pyx_n_s_NO_LAYOUT_CLOSURE;
  PyObject *__pyx_n_s_NO_PRUNE_UNICODE_RANGES;
  PyObject *__pyx_n_s_NO_SUBSET_TABLE_TAG;
  PyObject *__pyx_n_s_None;
  PyObject *__pyx_n_s_NotImplemented;
  PyObject *__pyx_n_s_OTColorLayer;
  PyObject *__pyx_n_u_OTColorLayer;
  PyObject *__pyx_n_s_OTColorPaletteFlags;
  PyObject *__pyx_n_s_OTLayoutGlyphClass;
  PyObject *__pyx_n_s_OTMathConstant;
  PyObject *__pyx_n_s_OTMathGlyphPart;
  PyObject *__pyx_n_u_OTMathGlyphPart;
  PyObject *__pyx_n_s_OTMathGlyphPartFlags;
  PyObject *__pyx_n_s_OTMathGlyphVariant;
  PyObject *__pyx_n_u_OTMathGlyphVariant;
  PyObject *__pyx_n_s_OTMathKern;
  PyObject *__pyx_n_s_OTMathKernEntry;
  PyObject *__pyx_n_u_OTMathKernEntry;
  PyObject *__pyx_n_s_OVERBAR_EXTRA_ASCENDER;
  PyObject *__pyx_n_s_OVERBAR_RULE_THICKNESS;
  PyObject *__pyx_n_s_OVERBAR_VERTICAL_GAP;
  PyObject *__pyx_n_s_OVERLAY;
  PyObject *__pyx_n_s_PAD;
  PyObject *__pyx_n_s_PASSTHROUGH_UNRECOGNIZED;
  PyObject *__pyx_n_s_PLUS;
  PyObject *__pyx_n_s_PRESERVE_DEFAULT_IGNORABLES;
  PyObject *__pyx_n_s_PRODUCE_SAFE_TO_INSERT_TATWEEL;
  PyObject *__pyx_n_s_PRODUCE_UNSAFE_TO_CONCAT;
  PyObject *__pyx_n_s_PaintCompositeMode;
  PyObject *__pyx_n_s_PaintExtend;
  PyObject *__pyx_n_s_PaintFuncs;
  PyObject *__pyx_n_s_PaintFuncs___reduce_cython;
  PyObject *__pyx_n_s_PaintFuncs___setstate_cython;
  PyObject *__pyx_n_s_PaintFuncs_set_color_func;
  PyObject *__pyx_n_s_PaintFuncs_set_color_glyph_func;
  PyObject *__pyx_n_s_PaintFuncs_set_custom_palette_co;
  PyObject *__pyx_n_s_PaintFuncs_set_image_func;
  PyObject *__pyx_n_s_PaintFuncs_set_linear_gradient_f;
  PyObject *__pyx_n_s_PaintFuncs_set_pop_clip_func;
  PyObject *__pyx_n_s_PaintFuncs_set_pop_group_func;
  PyObject *__pyx_n_s_PaintFuncs_set_pop_transform_fun;
  PyObject *__pyx_n_s_PaintFuncs_set_push_clip_glyph_f;
  PyObject *__pyx_n_s_PaintFuncs_set_push_clip_rectang;
  PyObject *__pyx_n_s_PaintFuncs_set_push_group_func;
  PyObject *__pyx_n_s_PaintFuncs_set_push_transform_fu;
  PyObject *__pyx_n_s_PaintFuncs_set_radial_gradient_f;
  PyObject *__pyx_n_s_PaintFuncs_set_sweep_gradient_fu;
  PyObject *__pyx_kp_u_Pass_draw_state_to_Font_draw_gly;
  PyObject *__pyx_n_s_Path;
  PyObject *__pyx_kp_s_Pickling_of_struct_members_such;
  PyObject *__pyx_kp_s_Pickling_of_struct_members_such_2;
  PyObject *__pyx_n_s_RADICAL_DEGREE_BOTTOM_RAISE_PERC;
  PyObject *__pyx_n_s_RADICAL_DISPLAY_STYLE_VERTICAL_G;
  PyObject *__pyx_n_s_RADICAL_EXTRA_ASCENDER;
  PyObject *__pyx_n_s_RADICAL_KERN_AFTER_DEGREE;
  PyObject *__pyx_n_s_RADICAL_KERN_BEFORE_DEGREE;
  PyObject *__pyx_n_s_RADICAL_RULE_THICKNESS;
  PyObject *__pyx_n_s_RADICAL_VERTICAL_GAP;
  PyObject *__pyx_n_s_REFLECT;
  PyObject *__pyx_n_s_REMOVE_DEFAULT_IGNORABLES;
  PyObject *__pyx_n_s_REPEAT;
  PyObject *__pyx_n_s_RETAIN_GIDS;
  PyObject *__pyx_n_s_RepackerError;
  PyObject *__pyx_n_s_RuntimeError;
  PyObject *__pyx_n_s_SAFE_TO_INSERT_TATWEEL;
  PyObject *__pyx_n_s_SCREEN;
  PyObject *__pyx_n_s_SCRIPT_PERCENT_SCALE_DOWN;
  PyObject *__pyx_n_s_SCRIPT_SCRIPT_PERCENT_SCALE_DOWN;
  PyObject *__pyx_n_s_SET_OVERLAPS_FLAG;
  PyObject *__pyx_n_s_SKEWED_FRACTION_HORIZONTAL_GAP;
  PyObject *__pyx_n_s_SKEWED_FRACTION_VERTICAL_GAP;
  PyObject *__pyx_n_s_SOFT_LIGHT;
  PyObject *__pyx_n_s_SPACE_AFTER_SCRIPT;
  PyObject *__pyx_n_s_SRC;
  PyObject *__pyx_n_s_SRC_ATOP;
  PyObject *__pyx_n_s_SRC_IN;
  PyObject *__pyx_n_s_SRC_OUT;
  PyObject *__pyx_n_s_SRC_OVER;
  PyObject *__pyx_n_s_STACK_BOTTOM_DISPLAY_STYLE_SHIFT;
  PyObject *__pyx_n_s_STACK_BOTTOM_SHIFT_DOWN;
  PyObject *__pyx_n_s_STACK_DISPLAY_STYLE_GAP_MIN;
  PyObject *__pyx_n_s_STACK_GAP_MIN;
  PyObject *__pyx_n_s_STACK_TOP_DISPLAY_STYLE_SHIFT_UP;
  PyObject *__pyx_n_s_STACK_TOP_SHIFT_UP;
  PyObject *__pyx_n_s_STRETCH_STACK_BOTTOM_SHIFT_DOWN;
  PyObject *__pyx_n_s_STRETCH_STACK_GAP_ABOVE_MIN;
  PyObject *__pyx_n_s_STRETCH_STACK_GAP_BELOW_MIN;
  PyObject *__pyx_n_s_STRETCH_STACK_TOP_SHIFT_UP;
  PyObject *__pyx_n_s_SUBSCRIPT_BASELINE_DROP_MIN;
  PyObject *__pyx_n_s_SUBSCRIPT_SHIFT_DOWN;
  PyObject *__pyx_n_s_SUBSCRIPT_TOP_MAX;
  PyObject *__pyx_n_s_SUB_SUPERSCRIPT_GAP_MIN;
  PyObject *__pyx_n_s_SUPERSCRIPT_BASELINE_DROP_MAX;
  PyObject *__pyx_n_s_SUPERSCRIPT_BOTTOM_MAX_WITH_SUBS;
  PyObject *__pyx_n_s_SUPERSCRIPT_BOTTOM_MIN;
  PyObject *__pyx_n_s_SUPERSCRIPT_SHIFT_UP;
  PyObject *__pyx_n_s_SUPERSCRIPT_SHIFT_UP_CRAMPED;
  PyObject *__pyx_n_s_Sequence;
  PyObject *__pyx_kp_u_Set;
  PyObject *__pyx_n_s_SetIter;
  PyObject *__pyx_n_s_SetIter___reduce_cython;
  PyObject *__pyx_n_s_SetIter___setstate_cython;
  PyObject *__pyx_n_s_Set_2;
  PyObject *__pyx_n_s_Set___copy;
  PyObject *__pyx_n_s_Set___reduce_cython;
  PyObject *__pyx_n_s_Set___setstate_cython;
  PyObject *__pyx_n_s_Set__is_equal;
  PyObject *__pyx_n_s_Set__set;
  PyObject *__pyx_n_s_Set__update;
  PyObject *__pyx_n_s_Set_add;
  PyObject *__pyx_n_s_Set_add_range;
  PyObject *__pyx_n_s_Set_clear;
  PyObject *__pyx_n_s_Set_copy;
  PyObject *__pyx_n_s_Set_del_range;
  PyObject *__pyx_n_s_Set_difference_update;
  PyObject *__pyx_n_s_Set_discard;
  PyObject *__pyx_n_s_Set_intersection_update;
  PyObject *__pyx_n_s_Set_invert;
  PyObject *__pyx_n_s_Set_is_inverted;
  PyObject *__pyx_n_s_Set_issubset;
  PyObject *__pyx_n_s_Set_issuperset;
  PyObject *__pyx_n_s_Set_remove;
  PyObject *__pyx_kp_u_Set_s;
  PyObject *__pyx_n_s_Set_set;
  PyObject *__pyx_n_s_Set_symmetric_difference_update;
  PyObject *__pyx_n_s_Set_update;
  PyObject *__pyx_n_s_StopIteration;
  PyObject *__pyx_n_s_SubsetFlags;
  PyObject *__pyx_n_s_SubsetInput;
  PyObject *__pyx_n_s_SubsetInputSets;
  PyObject *__pyx_n_s_SubsetInput___reduce_cython;
  PyObject *__pyx_n_s_SubsetInput___setstate_cython;
  PyObject *__pyx_n_s_SubsetInput_keep_everything;
  PyObject *__pyx_n_s_SubsetInput_pin_axis_location;
  PyObject *__pyx_n_s_SubsetInput_pin_axis_to_default;
  PyObject *__pyx_n_s_SubsetInput_sets;
  PyObject *__pyx_n_s_SubsetInput_subset;
  PyObject *__pyx_n_s_SubsetPlan;
  PyObject *__pyx_n_s_SubsetPlan___reduce_cython;
  PyObject *__pyx_n_s_SubsetPlan___setstate_cython;
  PyObject *__pyx_n_s_SubsetPlan_execute;
  PyObject *__pyx_kp_u_Subsetting_failed;
  PyObject *__pyx_n_s_TOP_LEFT;
  PyObject *__pyx_n_s_TOP_RIGHT;
  PyObject *__pyx_n_s_Tuple;
  PyObject *__pyx_kp_s_Tuple_List_OTMathGlyphPart_int;
  PyObject *__pyx_n_s_TypeError;
  PyObject *__pyx_n_s_UNCLASSIFIED;
  PyObject *__pyx_n_s_UNDERBAR_EXTRA_DESCENDER;
  PyObject *__pyx_n_s_UNDERBAR_RULE_THICKNESS;
  PyObject *__pyx_n_s_UNDERBAR_VERTICAL_GAP;
  PyObject *__pyx_n_s_UNICODE;
  PyObject *__pyx_n_s_UNSAFE_TO_BREAK;
  PyObject *__pyx_n_s_UNSAFE_TO_CONCAT;
  PyObject *__pyx_n_s_UPPER_LIMIT_BASELINE_RISE_MIN;
  PyObject *__pyx_n_s_UPPER_LIMIT_GAP_MIN;
  PyObject *__pyx_n_s_USABLE_WITH_DARK_BACKGROUND;
  PyObject *__pyx_n_s_USABLE_WITH_LIGHT_BACKGROUND;
  PyObject *__pyx_n_s_Union;
  PyObject *__pyx_kp_s_Union_str_Path;
  PyObject *__pyx_n_s_VERIFY;
  PyObject *__pyx_n_s_ValueError;
  PyObject *__pyx_n_s_XOR;
  PyObject *__pyx_kp_u__10;
  PyObject *__pyx_kp_b__15;
  PyObject *__pyx_kp_u__15;
  PyObject *__pyx_kp_u__16;
  PyObject *__pyx_kp_u__19;
  PyObject *__pyx_kp_u__21;
  PyObject *__pyx_n_s__22;
  PyObject *__pyx_kp_u__23;
  PyObject *__pyx_n_s__287;
  PyObject *__pyx_n_s_a;
  PyObject *__pyx_n_s_add;
  PyObject *__pyx_n_s_add_codepoints;
  PyObject *__pyx_n_s_add_range;
  PyObject *__pyx_n_s_add_str;
  PyObject *__pyx_n_s_add_utf8;
  PyObject *__pyx_n_u_advance;
  PyObject *__pyx_n_s_alpha;
  PyObject *__pyx_n_u_alpha;
  PyObject *__pyx_n_s_alternate_count;
  PyObject *__pyx_n_s_alternate_glyphs;
  PyObject *__pyx_n_s_alternates;
  PyObject *__pyx_n_s_args;
  PyObject *__pyx_n_s_ascender;
  PyObject *__pyx_n_u_ascender;
  PyObject *__pyx_n_s_assembly;
  PyObject *__pyx_n_s_assembly_array;
  PyObject *__pyx_n_s_asyncio_coroutines;
  PyObject *__pyx_n_s_b;
  PyObject *__pyx_n_s_baseline_tag;
  PyObject *__pyx_n_s_blob;
  PyObject *__pyx_n_s_blob_data;
  PyObject *__pyx_n_s_blob_length;
  PyObject *__pyx_n_s_blue;
  PyObject *__pyx_n_u_blue;
  PyObject *__pyx_n_s_bool;
  PyObject *__pyx_n_s_buffer;
  PyObject *__pyx_n_s_bytes;
  PyObject *__pyx_n_s_c;
  PyObject *__pyx_n_s_c_coords;
  PyObject *__pyx_n_s_c_foreground;
  PyObject *__pyx_n_s_c_shapers;
  PyObject *__pyx_n_s_callback;
  PyObject *__pyx_n_s_clear;
  PyObject *__pyx_n_s_clear_contents;
  PyObject *__pyx_n_s_cline_in_traceback;
  PyObject *__pyx_n_s_close;
  PyObject *__pyx_n_s_closePath;
  PyObject *__pyx_n_s_cls;
  PyObject *__pyx_n_s_codepoints;
  PyObject *__pyx_n_s_collections;
  PyObject *__pyx_n_u_color;
  PyObject *__pyx_n_s_color_count;
  PyObject *__pyx_n_s_color_index;
  PyObject *__pyx_n_u_color_index;
  PyObject *__pyx_n_s_colors;
  PyObject *__pyx_n_s_constant;
  PyObject *__pyx_n_s_coords;
  PyObject *__pyx_n_s_coords_2dot14;
  PyObject *__pyx_n_s_copy;
  PyObject *__pyx_n_s_copy_2;
  PyObject *__pyx_n_s_correction_height;
  PyObject *__pyx_n_s_count;
  PyObject *__pyx_n_s_create;
  PyObject *__pyx_n_s_create_for_tables;
  PyObject *__pyx_n_s_cstr;
  PyObject *__pyx_n_s_curveTo;
  PyObject *__pyx_n_s_data;
  PyObject *__pyx_n_s_del_range;
  PyObject *__pyx_n_s_descender;
  PyObject *__pyx_n_u_descender;
  PyObject *__pyx_n_s_dict;
  PyObject *__pyx_n_s_difference_update;
  PyObject *__pyx_n_s_direction;
  PyObject *__pyx_kp_u_disable;
  PyObject *__pyx_n_s_discard;
  PyObject *__pyx_n_s_doc;
  PyObject *__pyx_n_s_draw_data;
  PyObject *__pyx_n_s_draw_funcs;
  PyObject *__pyx_n_s_draw_glyph;
  PyObject *__pyx_kp_u_draw_glyph_is_deprecated_use_Fon;
  PyObject *__pyx_n_s_draw_glyph_with_pen;
  PyObject *__pyx_n_s_draw_state;
  PyObject *__pyx_n_s_draw_state_p;
  PyObject *__pyx_kp_u_enable;
  PyObject *__pyx_n_s_encode;
  PyObject *__pyx_n_s_end;
  PyObject *__pyx_n_u_end_connector_length;
  PyObject *__pyx_n_s_enum;
  PyObject *__pyx_n_s_enumerate;
  PyObject *__pyx_n_s_execute;
  PyObject *__pyx_n_s_extents;
  PyObject *__pyx_n_s_face;
  PyObject *__pyx_n_s_face_or_font;
  PyObject *__pyx_n_s_feat;
  PyObject *__pyx_n_s_feature_count;
  PyObject *__pyx_n_s_feature_tags;
  PyObject *__pyx_n_s_features;
  PyObject *__pyx_n_s_filename;
  PyObject *__pyx_n_s_first;
  PyObject *__pyx_n_u_flags;
  PyObject *__pyx_n_s_float;
  PyObject *__pyx_n_s_font;
  PyObject *__pyx_n_s_foreground;
  PyObject *__pyx_n_s_from_file_path;
  PyObject *__pyx_n_s_from_int;
  PyObject *__pyx_n_s_fsencode;
  PyObject *__pyx_n_u_full_advance;
  PyObject *__pyx_n_s_func;
  PyObject *__pyx_n_s_func_p;
  PyObject *__pyx_n_s_funcs;
  PyObject *__pyx_n_s_g;
  PyObject *__pyx_kp_u_gc;
  PyObject *__pyx_n_s_genexpr;
  PyObject *__pyx_n_s_get;
  PyObject *__pyx_n_s_get_font_extents;
  PyObject *__pyx_n_s_get_glyph_extents;
  PyObject *__pyx_n_s_get_glyph_from_name;
  PyObject *__pyx_n_s_get_glyph_h_advance;
  PyObject *__pyx_n_s_get_glyph_h_origin;
  PyObject *__pyx_n_s_get_glyph_name;
  PyObject *__pyx_n_s_get_glyph_shape;
  PyObject *__pyx_kp_u_get_glyph_shape_is_deprecated_us;
  PyObject *__pyx_n_s_get_glyph_v_advance;
  PyObject *__pyx_n_s_get_glyph_v_origin;
  PyObject *__pyx_n_s_get_nominal_glyph;
  PyObject *__pyx_n_s_get_var_coords_design;
  PyObject *__pyx_n_s_get_var_coords_normalized;
  PyObject *__pyx_n_s_get_variation_glyph;
  PyObject *__pyx_n_s_getstate;
  PyObject *__pyx_n_s_gid;
  PyObject *__pyx_n_s_glyph;
  PyObject *__pyx_n_u_glyph;
  PyObject *__pyx_n_s_glyph_from_string;
  PyObject *__pyx_n_s_glyph_to_string;
  PyObject *__pyx_n_s_graphnodes;
  PyObject *__pyx_n_s_green;
  PyObject *__pyx_n_u_green;
  PyObject *__pyx_n_s_guess_segment_properties;
  PyObject *__pyx_n_u_hang;
  PyObject *__pyx_n_s_hb_baseline_tag;
  PyObject *__pyx_n_s_hb_codepoint_t;
  PyObject *__pyx_n_s_hb_codepoints;
  PyObject *__pyx_n_s_hb_direction;
  PyObject *__pyx_n_s_hb_features;
  PyObject *__pyx_n_s_hb_language;
  PyObject *__pyx_n_s_hb_language_tag;
  PyObject *__pyx_n_s_hb_position;
  PyObject *__pyx_n_s_hb_script;
  PyObject *__pyx_n_s_hb_script_tag;
  PyObject *__pyx_n_s_hb_tag;
  PyObject *__pyx_n_s_hb_variations;
  PyObject *__pyx_n_s_head;
  PyObject *__pyx_n_u_height;
  PyObject *__pyx_n_s_i;
  PyObject *__pyx_n_u_icfb;
  PyObject *__pyx_n_u_icft;
  PyObject *__pyx_n_u_ideo;
  PyObject *__pyx_n_u_idtp;
  PyObject *__pyx_n_s_import;
  PyObject *__pyx_n_s_index;
  PyObject *__pyx_kp_u_index_out_of_range_for_feature_a;
  PyObject *__pyx_n_s_init;
  PyObject *__pyx_n_s_init_subclass;
  PyObject *__pyx_n_s_initializing;
  PyObject *__pyx_n_s_input;
  PyObject *__pyx_n_s_inst;
  PyObject *__pyx_n_s_int;
  PyObject *__pyx_kp_s_int_None;
  PyObject *__pyx_n_s_intersection_update;
  PyObject *__pyx_kp_u_invalid_baseline_tag;
  PyObject *__pyx_kp_u_invalid_constant;
  PyObject *__pyx_kp_u_invalid_kern;
  PyObject *__pyx_n_s_invert;
  PyObject *__pyx_n_s_is_coroutine;
  PyObject *__pyx_n_s_is_equal;
  PyObject *__pyx_n_u_is_foreground;
  PyObject *__pyx_n_s_is_inverted;
  PyObject *__pyx_kp_u_isenabled;
  PyObject *__pyx_n_s_issubset;
  PyObject *__pyx_n_s_issuperset;
  PyObject *__pyx_n_s_italics_correction;
  PyObject *__pyx_n_s_item_length;
  PyObject *__pyx_n_s_item_offset;
  PyObject *__pyx_n_s_items;
  PyObject *__pyx_n_s_join;
  PyObject *__pyx_n_s_k;
  PyObject *__pyx_n_s_keep_everything;
  PyObject *__pyx_n_s_kern;
  PyObject *__pyx_n_u_kern_value;
  PyObject *__pyx_n_s_kerns;
  PyObject *__pyx_n_s_kerns_array;
  PyObject *__pyx_n_s_keys;
  PyObject *__pyx_n_s_keys_locals_genexpr;
  PyObject *__pyx_n_s_kind;
  PyObject *__pyx_n_s_language_count;
  PyObject *__pyx_n_s_language_index;
  PyObject *__pyx_n_s_language_tag;
  PyObject *__pyx_n_s_language_tags;
  PyObject *__pyx_n_s_larger_set;
  PyObject *__pyx_n_s_last;
  PyObject *__pyx_n_s_layer_count;
  PyObject *__pyx_n_s_layers;
  PyObject *__pyx_n_s_length;
  PyObject *__pyx_n_s_lineTo;
  PyObject *__pyx_n_s_line_gap;
  PyObject *__pyx_n_u_line_gap;
  PyObject *__pyx_n_s_lookup_index;
  PyObject *__pyx_n_s_m;
  PyObject *__pyx_n_s_main;
  PyObject *__pyx_n_u_math;
  PyObject *__pyx_n_u_max_correction_height;
  PyObject *__pyx_n_s_metaclass;
  PyObject *__pyx_n_s_methods;
  PyObject *__pyx_n_s_module;
  PyObject *__pyx_n_s_moveTo;
  PyObject *__pyx_n_s_mro_entries;
  PyObject *__pyx_n_s_name;
  PyObject *__pyx_n_s_name_2;
  PyObject *__pyx_n_s_name_id;
  PyObject *__pyx_n_s_namedtuple;
  PyObject *__pyx_n_s_new_face;
  PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
  PyObject *__pyx_n_s_node;
  PyObject *__pyx_n_s_num_nodes;
  PyObject *__pyx_kp_u_num_of_graph_nodes;
  PyObject *__pyx_n_s_obj_list;
  PyObject *__pyx_n_s_object;
  PyObject *__pyx_n_u_offset;
  PyObject *__pyx_n_s_os;
  PyObject *__pyx_n_s_ot_color_glyph_get_layers;
  PyObject *__pyx_n_s_ot_color_glyph_get_png;
  PyObject *__pyx_n_s_ot_color_glyph_get_svg;
  PyObject *__pyx_n_s_ot_color_glyph_has_paint;
  PyObject *__pyx_n_s_ot_color_has_layers;
  PyObject *__pyx_n_s_ot_color_has_paint;
  PyObject *__pyx_n_s_ot_color_has_palettes;
  PyObject *__pyx_n_s_ot_color_has_png;
  PyObject *__pyx_n_s_ot_color_has_svg;
  PyObject *__pyx_n_s_ot_color_palette_color_get_name;
  PyObject *__pyx_n_s_ot_color_palette_get_colors;
  PyObject *__pyx_n_s_ot_color_palette_get_count;
  PyObject *__pyx_n_s_ot_color_palette_get_flags;
  PyObject *__pyx_n_s_ot_color_palette_get_name_id;
  PyObject *__pyx_n_s_ot_font_set_funcs;
  PyObject *__pyx_n_s_ot_layout_get_baseline;
  PyObject *__pyx_n_s_ot_layout_get_glyph_class;
  PyObject *__pyx_n_s_ot_layout_has_glyph_classes;
  PyObject *__pyx_n_s_ot_layout_has_positioning;
  PyObject *__pyx_n_s_ot_layout_has_substitution;
  PyObject *__pyx_n_s_ot_layout_language_get_feature_t;
  PyObject *__pyx_n_s_ot_layout_lookup_get_glyph_alter;
  PyObject *__pyx_n_s_ot_layout_script_get_language_ta;
  PyObject *__pyx_n_s_ot_layout_table_get_script_tags;
  PyObject *__pyx_n_s_ot_math_get_constant;
  PyObject *__pyx_n_s_ot_math_get_glyph_assembly;
  PyObject *__pyx_n_s_ot_math_get_glyph_italics_correc;
  PyObject *__pyx_n_s_ot_math_get_glyph_kerning;
  PyObject *__pyx_n_s_ot_math_get_glyph_kernings;
  PyObject *__pyx_n_s_ot_math_get_glyph_top_accent_att;
  PyObject *__pyx_n_s_ot_math_get_glyph_variants;
  PyObject *__pyx_n_s_ot_math_get_min_connector_overla;
  PyObject *__pyx_n_s_ot_math_has_data;
  PyObject *__pyx_n_s_ot_math_is_glyph_extended_shape;
  PyObject *__pyx_n_s_ot_tag_to_language;
  PyObject *__pyx_n_s_ot_tag_to_script;
  PyObject *__pyx_n_s_other;
  PyObject *__pyx_n_s_p;
  PyObject *__pyx_n_s_packed;
  PyObject *__pyx_n_s_packed_blob;
  PyObject *__pyx_n_s_paint_funcs;
  PyObject *__pyx_n_s_paint_glyph;
  PyObject *__pyx_n_s_paint_state;
  PyObject *__pyx_n_s_paint_state_p;
  PyObject *__pyx_n_s_palette_index;
  PyObject *__pyx_n_s_pathlib;
  PyObject *__pyx_n_s_pen;
  PyObject *__pyx_n_s_pin_axis_location;
  PyObject *__pyx_n_s_pin_axis_to_default;
  PyObject *__pyx_n_s_prepare;
  PyObject *__pyx_n_s_pyx_state;
  PyObject *__pyx_n_s_pyx_vtable;
  PyObject *__pyx_n_s_qCurveTo;
  PyObject *__pyx_n_s_qualname;
  PyObject *__pyx_n_s_r;
  PyObject *__pyx_n_s_range;
  PyObject *__pyx_n_s_red;
  PyObject *__pyx_n_u_red;
  PyObject *__pyx_n_s_reduce;
  PyObject *__pyx_n_s_reduce_cython;
  PyObject *__pyx_n_s_reduce_ex;
  PyObject *__pyx_n_s_remove;
  PyObject *__pyx_n_s_repack;
  PyObject *__pyx_n_s_repack_with_tag;
  PyObject *__pyx_n_s_repr___locals_genexpr;
  PyObject *__pyx_n_s_reset;
  PyObject *__pyx_n_s_ret;
  PyObject *__pyx_n_s_return;
  PyObject *__pyx_n_u_romn;
  PyObject *__pyx_n_s_round;
  PyObject *__pyx_n_s_s;
  PyObject *__pyx_n_s_script_count;
  PyObject *__pyx_n_s_script_index;
  PyObject *__pyx_n_s_script_tag;
  PyObject *__pyx_n_s_script_tags;
  PyObject *__pyx_n_s_self;
  PyObject *__pyx_n_s_send;
  PyObject *__pyx_n_s_set;
  PyObject *__pyx_n_s_set_2;
  PyObject *__pyx_n_s_set_close_path_func;
  PyObject *__pyx_n_s_set_color_func;
  PyObject *__pyx_n_s_set_color_glyph_func;
  PyObject *__pyx_n_s_set_cubic_to_func;
  PyObject *__pyx_n_s_set_custom_palette_color_func;
  PyObject *__pyx_n_s_set_font_h_extents_func;
  PyObject *__pyx_n_s_set_font_v_extents_func;
  PyObject *__pyx_n_s_set_glyph_h_advance_func;
  PyObject *__pyx_n_s_set_glyph_name_func;
  PyObject *__pyx_n_s_set_glyph_v_advance_func;
  PyObject *__pyx_n_s_set_glyph_v_origin_func;
  PyObject *__pyx_n_s_set_image_func;
  PyObject *__pyx_n_s_set_language_from_ot_tag;
  PyObject *__pyx_n_s_set_line_to_func;
  PyObject *__pyx_n_s_set_linear_gradient_func;
  PyObject *__pyx_n_s_set_message_func;
  PyObject *__pyx_n_s_set_move_to_func;
  PyObject *__pyx_n_s_set_name;
  PyObject *__pyx_n_s_set_nominal_glyph_func;
  PyObject *__pyx_n_s_set_pop_clip_func;
  PyObject *__pyx_n_s_set_pop_group_func;
  PyObject *__pyx_n_s_set_pop_transform_func;
  PyObject *__pyx_n_s_set_push_clip_glyph_func;
  PyObject *__pyx_n_s_set_push_clip_rectangle_func;
  PyObject *__pyx_n_s_set_push_group_func;
  PyObject *__pyx_n_s_set_push_transform_func;
  PyObject *__pyx_n_s_set_quadratic_to_func;
  PyObject *__pyx_n_s_set_radial_gradient_func;
  PyObject *__pyx_n_s_set_script_from_ot_tag;
  PyObject *__pyx_n_s_set_sweep_gradient_func;
  PyObject *__pyx_n_s_set_type;
  PyObject *__pyx_n_s_set_var_coords_design;
  PyObject *__pyx_n_s_set_var_coords_normalized;
  PyObject *__pyx_n_s_set_variation;
  PyObject *__pyx_n_s_set_variation_glyph_func;
  PyObject *__pyx_n_s_set_variations;
  PyObject *__pyx_n_s_sets;
  PyObject *__pyx_n_s_setstate;
  PyObject *__pyx_n_s_setstate_cython;
  PyObject *__pyx_n_s_shape;
  PyObject *__pyx_n_s_shaper;
  PyObject *__pyx_n_s_shapers;
  PyObject *__pyx_n_s_size;
  PyObject *__pyx_n_s_smaller_set;
  PyObject *__pyx_n_s_source;
  PyObject *__pyx_n_s_spec;
  PyObject *__pyx_kp_s_src_uharfbuzz__harfbuzz_pyx;
  PyObject *__pyx_n_s_start;
  PyObject *__pyx_n_u_start_connector_length;
  PyObject *__pyx_n_s_start_offset;
  PyObject *__pyx_n_s_staticmethod;
  PyObject *__pyx_n_s_str;
  PyObject *__pyx_n_s_string;
  PyObject *__pyx_kp_s_stringsource;
  PyObject *__pyx_n_s_subset;
  PyObject *__pyx_n_s_subset_preprocess;
  PyObject *__pyx_n_s_subtables;
  PyObject *__pyx_n_s_success;
  PyObject *__pyx_n_s_super;
  PyObject *__pyx_n_s_symmetric_difference_update;
  PyObject *__pyx_n_s_table_bytes;
  PyObject *__pyx_n_s_table_data;
  PyObject *__pyx_n_s_tag;
  PyObject *__pyx_n_s_tag_packed;
  PyObject *__pyx_n_s_tags;
  PyObject *__pyx_n_s_tail;
  PyObject *__pyx_n_s_test;
  PyObject *__pyx_n_s_text;
  PyObject *__pyx_n_s_throw;
  PyObject *__pyx_n_s_to_int;
  PyObject *__pyx_n_s_typing;
  PyObject *__pyx_n_s_uharfbuzz__harfbuzz;
  PyObject *__pyx_n_s_unicode;
  PyObject *__pyx_n_s_update;
  PyObject *__pyx_n_s_update_2;
  PyObject *__pyx_n_s_user_data;
  PyObject *__pyx_n_s_user_data_p;
  PyObject *__pyx_n_s_v;
  PyObject *__pyx_n_s_value;
  PyObject *__pyx_n_s_values;
  PyObject *__pyx_n_s_values_locals_genexpr;
  PyObject *__pyx_n_s_variants;
  PyObject *__pyx_n_s_variants_array;
  PyObject *__pyx_n_s_variation;
  PyObject *__pyx_n_s_variation_selector;
  PyObject *__pyx_n_s_variation_unicodes;
  PyObject *__pyx_n_s_variations;
  PyObject *__pyx_n_s_version_string;
  PyObject *__pyx_n_s_warn;
  PyObject *__pyx_n_s_warnings;
  PyObject *__pyx_n_u_width;
  PyObject *__pyx_n_s_x;
  PyObject *__pyx_n_u_x_bearing;
  PyObject *__pyx_n_s_y;
  PyObject *__pyx_n_u_y_bearing;
  PyObject *__pyx_int_0;
  PyObject *__pyx_int_1;
  PyObject *__pyx_int_16384;
  PyObject *__pyx_int_65535;
  PyObject *__pyx_int_neg_1;
  PyObject *__pyx_k_;
  PyObject *__pyx_k__2;
  PyObject *__pyx_k__3;
  PyObject *__pyx_k__4;
  PyObject *__pyx_k__5;
  PyObject *__pyx_k__6;
  PyObject *__pyx_k__8;
  PyObject *__pyx_k__18;
  PyObject *__pyx_k__20;
  PyObject *__pyx_tuple__7;
  PyObject *__pyx_tuple__9;
  PyObject *__pyx_tuple__11;
  PyObject *__pyx_tuple__12;
  PyObject *__pyx_tuple__13;
  PyObject *__pyx_tuple__14;
  PyObject *__pyx_tuple__17;
  PyObject *__pyx_tuple__24;
  PyObject *__pyx_tuple__26;
  PyObject *__pyx_tuple__28;
  PyObject *__pyx_tuple__32;
  PyObject *__pyx_tuple__36;
  PyObject *__pyx_tuple__39;
  PyObject *__pyx_tuple__41;
  PyObject *__pyx_tuple__43;
  PyObject *__pyx_tuple__46;
  PyObject *__pyx_tuple__50;
  PyObject *__pyx_tuple__54;
  PyObject *__pyx_tuple__56;
  PyObject *__pyx_tuple__57;
  PyObject *__pyx_tuple__59;
  PyObject *__pyx_tuple__63;
  PyObject *__pyx_tuple__65;
  PyObject *__pyx_tuple__67;
  PyObject *__pyx_tuple__69;
  PyObject *__pyx_tuple__71;
  PyObject *__pyx_tuple__73;
  PyObject *__pyx_tuple__75;
  PyObject *__pyx_tuple__78;
  PyObject *__pyx_tuple__81;
  PyObject *__pyx_tuple__83;
  PyObject *__pyx_tuple__85;
  PyObject *__pyx_tuple__87;
  PyObject *__pyx_tuple__89;
  PyObject *__pyx_tuple__92;
  PyObject *__pyx_tuple__94;
  PyObject *__pyx_tuple__96;
  PyObject *__pyx_tuple__98;
  PyObject *__pyx_tuple__100;
  PyObject *__pyx_tuple__102;
  PyObject *__pyx_tuple__107;
  PyObject *__pyx_tuple__118;
  PyObject *__pyx_tuple__120;
  PyObject *__pyx_tuple__121;
  PyObject *__pyx_tuple__123;
  PyObject *__pyx_tuple__125;
  PyObject *__pyx_tuple__127;
  PyObject *__pyx_tuple__129;
  PyObject *__pyx_tuple__131;
  PyObject *__pyx_tuple__133;
  PyObject *__pyx_tuple__135;
  PyObject *__pyx_tuple__139;
  PyObject *__pyx_tuple__143;
  PyObject *__pyx_tuple__145;
  PyObject *__pyx_tuple__147;
  PyObject *__pyx_tuple__149;
  PyObject *__pyx_tuple__152;
  PyObject *__pyx_tuple__157;
  PyObject *__pyx_tuple__160;
  PyObject *__pyx_tuple__163;
  PyObject *__pyx_tuple__165;
  PyObject *__pyx_tuple__169;
  PyObject *__pyx_tuple__171;
  PyObject *__pyx_tuple__173;
  PyObject *__pyx_tuple__175;
  PyObject *__pyx_tuple__177;
  PyObject *__pyx_tuple__179;
  PyObject *__pyx_tuple__182;
  PyObject *__pyx_tuple__186;
  PyObject *__pyx_tuple__203;
  PyObject *__pyx_tuple__205;
  PyObject *__pyx_tuple__207;
  PyObject *__pyx_tuple__217;
  PyObject *__pyx_tuple__219;
  PyObject *__pyx_tuple__221;
  PyObject *__pyx_tuple__223;
  PyObject *__pyx_tuple__225;
  PyObject *__pyx_tuple__228;
  PyObject *__pyx_tuple__230;
  PyObject *__pyx_tuple__232;
  PyObject *__pyx_tuple__236;
  PyObject *__pyx_tuple__240;
  PyObject *__pyx_tuple__247;
  PyObject *__pyx_tuple__252;
  PyObject *__pyx_tuple__254;
  PyObject *__pyx_tuple__256;
  PyObject *__pyx_tuple__259;
  PyObject *__pyx_tuple__273;
  PyObject *__pyx_tuple__277;
  PyObject *__pyx_tuple__280;
  PyObject *__pyx_codeobj__25;
  PyObject *__pyx_codeobj__27;
  PyObject *__pyx_codeobj__29;
  PyObject *__pyx_codeobj__30;
  PyObject *__pyx_codeobj__31;
  PyObject *__pyx_codeobj__33;
  PyObject *__pyx_codeobj__34;
  PyObject *__pyx_codeobj__35;
  PyObject *__pyx_codeobj__37;
  PyObject *__pyx_codeobj__38;
  PyObject *__pyx_codeobj__40;
  PyObject *__pyx_codeobj__42;
  PyObject *__pyx_codeobj__44;
  PyObject *__pyx_codeobj__45;
  PyObject *__pyx_codeobj__47;
  PyObject *__pyx_codeobj__48;
  PyObject *__pyx_codeobj__49;
  PyObject *__pyx_codeobj__51;
  PyObject *__pyx_codeobj__52;
  PyObject *__pyx_codeobj__53;
  PyObject *__pyx_codeobj__55;
  PyObject *__pyx_codeobj__58;
  PyObject *__pyx_codeobj__60;
  PyObject *__pyx_codeobj__61;
  PyObject *__pyx_codeobj__62;
  PyObject *__pyx_codeobj__64;
  PyObject *__pyx_codeobj__66;
  PyObject *__pyx_codeobj__68;
  PyObject *__pyx_codeobj__70;
  PyObject *__pyx_codeobj__72;
  PyObject *__pyx_codeobj__74;
  PyObject *__pyx_codeobj__76;
  PyObject *__pyx_codeobj__77;
  PyObject *__pyx_codeobj__79;
  PyObject *__pyx_codeobj__80;
  PyObject *__pyx_codeobj__82;
  PyObject *__pyx_codeobj__84;
  PyObject *__pyx_codeobj__86;
  PyObject *__pyx_codeobj__88;
  PyObject *__pyx_codeobj__90;
  PyObject *__pyx_codeobj__91;
  PyObject *__pyx_codeobj__93;
  PyObject *__pyx_codeobj__95;
  PyObject *__pyx_codeobj__97;
  PyObject *__pyx_codeobj__99;
  PyObject *__pyx_codeobj__101;
  PyObject *__pyx_codeobj__103;
  PyObject *__pyx_codeobj__104;
  PyObject *__pyx_codeobj__105;
  PyObject *__pyx_codeobj__106;
  PyObject *__pyx_codeobj__108;
  PyObject *__pyx_codeobj__109;
  PyObject *__pyx_codeobj__110;
  PyObject *__pyx_codeobj__111;
  PyObject *__pyx_codeobj__112;
  PyObject *__pyx_codeobj__113;
  PyObject *__pyx_codeobj__114;
  PyObject *__pyx_codeobj__115;
  PyObject *__pyx_codeobj__116;
  PyObject *__pyx_codeobj__117;
  PyObject *__pyx_codeobj__119;
  PyObject *__pyx_codeobj__122;
  PyObject *__pyx_codeobj__124;
  PyObject *__pyx_codeobj__126;
  PyObject *__pyx_codeobj__128;
  PyObject *__pyx_codeobj__130;
  PyObject *__pyx_codeobj__132;
  PyObject *__pyx_codeobj__134;
  PyObject *__pyx_codeobj__136;
  PyObject *__pyx_codeobj__137;
  PyObject *__pyx_codeobj__138;
  PyObject *__pyx_codeobj__140;
  PyObject *__pyx_codeobj__141;
  PyObject *__pyx_codeobj__142;
  PyObject *__pyx_codeobj__144;
  PyObject *__pyx_codeobj__146;
  PyObject *__pyx_codeobj__148;
  PyObject *__pyx_codeobj__150;
  PyObject *__pyx_codeobj__151;
  PyObject *__pyx_codeobj__153;
  PyObject *__pyx_codeobj__154;
  PyObject *__pyx_codeobj__155;
  PyObject *__pyx_codeobj__156;
  PyObject *__pyx_codeobj__158;
  PyObject *__pyx_codeobj__159;
  PyObject *__pyx_codeobj__161;
  PyObject *__pyx_codeobj__162;
  PyObject *__pyx_codeobj__164;
  PyObject *__pyx_codeobj__166;
  PyObject *__pyx_codeobj__167;
  PyObject *__pyx_codeobj__168;
  PyObject *__pyx_codeobj__170;
  PyObject *__pyx_codeobj__172;
  PyObject *__pyx_codeobj__174;
  PyObject *__pyx_codeobj__176;
  PyObject *__pyx_codeobj__178;
  PyObject *__pyx_codeobj__180;
  PyObject *__pyx_codeobj__181;
  PyObject *__pyx_codeobj__183;
  PyObject *__pyx_codeobj__184;
  PyObject *__pyx_codeobj__185;
  PyObject *__pyx_codeobj__187;
  PyObject *__pyx_codeobj__188;
  PyObject *__pyx_codeobj__189;
  PyObject *__pyx_codeobj__190;
  PyObject *__pyx_codeobj__191;
  PyObject *__pyx_codeobj__192;
  PyObject *__pyx_codeobj__193;
  PyObject *__pyx_codeobj__194;
  PyObject *__pyx_codeobj__195;
  PyObject *__pyx_codeobj__196;
  PyObject *__pyx_codeobj__197;
  PyObject *__pyx_codeobj__198;
  PyObject *__pyx_codeobj__199;
  PyObject *__pyx_codeobj__200;
  PyObject *__pyx_codeobj__201;
  PyObject *__pyx_codeobj__202;
  PyObject *__pyx_codeobj__204;
  PyObject *__pyx_codeobj__206;
  PyObject *__pyx_codeobj__208;
  PyObject *__pyx_codeobj__209;
  PyObject *__pyx_codeobj__210;
  PyObject *__pyx_codeobj__211;
  PyObject *__pyx_codeobj__212;
  PyObject *__pyx_codeobj__213;
  PyObject *__pyx_codeobj__214;
  PyObject *__pyx_codeobj__215;
  PyObject *__pyx_codeobj__216;
  PyObject *__pyx_codeobj__218;
  PyObject *__pyx_codeobj__220;
  PyObject *__pyx_codeobj__222;
  PyObject *__pyx_codeobj__224;
  PyObject *__pyx_codeobj__226;
  PyObject *__pyx_codeobj__227;
  PyObject *__pyx_codeobj__229;
  PyObject *__pyx_codeobj__231;
  PyObject *__pyx_codeobj__233;
  PyObject *__pyx_codeobj__234;
  PyObject *__pyx_codeobj__235;
  PyObject *__pyx_codeobj__237;
  PyObject *__pyx_codeobj__238;
  PyObject *__pyx_codeobj__239;
  PyObject *__pyx_codeobj__241;
  PyObject *__pyx_codeobj__242;
  PyObject *__pyx_codeobj__243;
  PyObject *__pyx_codeobj__244;
  PyObject *__pyx_codeobj__245;
  PyObject *__pyx_codeobj__246;
  PyObject *__pyx_codeobj__248;
  PyObject *__pyx_codeobj__249;
  PyObject *__pyx_codeobj__250;
  PyObject *__pyx_codeobj__251;
  PyObject *__pyx_codeobj__253;
  PyObject *__pyx_codeobj__255;
  PyObject *__pyx_codeobj__257;
  PyObject *__pyx_codeobj__258;
  PyObject *__pyx_codeobj__260;
  PyObject *__pyx_codeobj__261;
  PyObject *__pyx_codeobj__262;
  PyObject *__pyx_codeobj__263;
  PyObject *__pyx_codeobj__264;
  PyObject *__pyx_codeobj__265;
  PyObject *__pyx_codeobj__266;
  PyObject *__pyx_codeobj__267;
  PyObject *__pyx_codeobj__268;
  PyObject *__pyx_codeobj__269;
  PyObject *__pyx_codeobj__270;
  PyObject *__pyx_codeobj__271;
  PyObject *__pyx_codeobj__272;
  PyObject *__pyx_codeobj__274;
  PyObject *__pyx_codeobj__275;
  PyObject *__pyx_codeobj__276;
  PyObject *__pyx_codeobj__278;
  PyObject *__pyx_codeobj__279;
  PyObject *__pyx_codeobj__281;
  PyObject *__pyx_codeobj__282;
  PyObject *__pyx_codeobj__283;
  PyObject *__pyx_codeobj__284;
  PyObject *__pyx_codeobj__285;
  PyObject *__pyx_codeobj__286;
} __pyx_mstate;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
  extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate(o) ((__pyx_mstate *)__Pyx_PyModule_GetState(o))

#define __pyx_mstate_global (__pyx_mstate(PyState_FindModule(&__pyx_moduledef)))

#define __pyx_m (PyState_FindModule(&__pyx_moduledef))
#else
static __pyx_mstate __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstate *__pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_clear(PyObject *m) {
  __pyx_mstate *clear_module_state = __pyx_mstate(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_CLEAR(clear_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_CLEAR(clear_module_state->__pyx_FusedFunctionType);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphInfo);
  Py_CLEAR(clear_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_GlyphInfo);
  Py_CLEAR(clear_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphPosition);
  Py_CLEAR(clear_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_GlyphPosition);
  Py_CLEAR(clear_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer);
  Py_CLEAR(clear_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_Buffer);
  Py_CLEAR(clear_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_Blob);
  Py_CLEAR(clear_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_Blob);
  Py_CLEAR(clear_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_Face);
  Py_CLEAR(clear_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_Face);
  Py_CLEAR(clear_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_Font);
  Py_CLEAR(clear_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_Font);
  Py_CLEAR(clear_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs);
  Py_CLEAR(clear_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_FontFuncs);
  Py_CLEAR(clear_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_ColorLine);
  Py_CLEAR(clear_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_ColorLine);
  Py_CLEAR(clear_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs);
  Py_CLEAR(clear_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_PaintFuncs);
  Py_CLEAR(clear_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_DrawFuncs);
  Py_CLEAR(clear_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_DrawFuncs);
  Py_CLEAR(clear_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_HBObject);
  Py_CLEAR(clear_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_HBObject);
  Py_CLEAR(clear_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetInput);
  Py_CLEAR(clear_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_SubsetInput);
  Py_CLEAR(clear_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetPlan);
  Py_CLEAR(clear_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_SubsetPlan);
  Py_CLEAR(clear_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_Set);
  Py_CLEAR(clear_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_Set);
  Py_CLEAR(clear_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_SetIter);
  Py_CLEAR(clear_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_SetIter);
  Py_CLEAR(clear_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_Map);
  Py_CLEAR(clear_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_Map);
  Py_CLEAR(clear_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_MapIter);
  Py_CLEAR(clear_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_MapIter);
  Py_CLEAR(clear_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr);
  Py_CLEAR(clear_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr);
  Py_CLEAR(clear_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr);
  Py_CLEAR(clear_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr);
  Py_CLEAR(clear_module_state->__pyx_n_s_ACCENT_BASE_HEIGHT);
  Py_CLEAR(clear_module_state->__pyx_n_s_AXIS_HEIGHT);
  Py_CLEAR(clear_module_state->__pyx_kp_u_All_shapers_failed);
  Py_CLEAR(clear_module_state->__pyx_n_s_AssertionError);
  Py_CLEAR(clear_module_state->__pyx_n_s_BASE_GLYPH);
  Py_CLEAR(clear_module_state->__pyx_n_s_BOT);
  Py_CLEAR(clear_module_state->__pyx_n_s_BOTTOM_LEFT);
  Py_CLEAR(clear_module_state->__pyx_n_s_BOTTOM_RIGHT);
  Py_CLEAR(clear_module_state->__pyx_n_s_Blob);
  Py_CLEAR(clear_module_state->__pyx_n_s_Blob___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Blob___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Blob_from_file_path);
  Py_CLEAR(clear_module_state->__pyx_n_s_Buffer);
  Py_CLEAR(clear_module_state->__pyx_n_s_BufferClusterLevel);
  Py_CLEAR(clear_module_state->__pyx_n_s_BufferContentType);
  Py_CLEAR(clear_module_state->__pyx_n_s_BufferFlags);
  Py_CLEAR(clear_module_state->__pyx_n_s_Buffer___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Buffer___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Buffer_add_codepoints);
  Py_CLEAR(clear_module_state->__pyx_n_s_Buffer_add_str);
  Py_CLEAR(clear_module_state->__pyx_n_s_Buffer_add_utf8);
  Py_CLEAR(clear_module_state->__pyx_n_s_Buffer_clear_contents);
  Py_CLEAR(clear_module_state->__pyx_n_s_Buffer_create);
  Py_CLEAR(clear_module_state->__pyx_n_s_Buffer_guess_segment_properties);
  Py_CLEAR(clear_module_state->__pyx_n_s_Buffer_reset);
  Py_CLEAR(clear_module_state->__pyx_n_s_Buffer_set_language_from_ot_tag);
  Py_CLEAR(clear_module_state->__pyx_n_s_Buffer_set_message_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_Buffer_set_script_from_ot_tag);
  Py_CLEAR(clear_module_state->__pyx_n_s_CHARACTERS);
  Py_CLEAR(clear_module_state->__pyx_n_s_CLEAR);
  Py_CLEAR(clear_module_state->__pyx_n_s_COLOR_BURN);
  Py_CLEAR(clear_module_state->__pyx_n_s_COLOR_DODGE);
  Py_CLEAR(clear_module_state->__pyx_n_s_COMPONENT);
  Py_CLEAR(clear_module_state->__pyx_n_s_Callable);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Callable_Blob_int_int_str_float);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Callable_ColorLine_float_float_f);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Callable_ColorLine_float_float_f_2);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Callable_Color_bool_object_None);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Callable_Face_str_object_bytes);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Callable_Font_int_int_object_int);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Callable_Font_int_object_int);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Callable_Font_int_object_int_int);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Callable_Font_int_object_str);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Callable_Font_object_FontExtents);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Callable_PaintCompositeMode_obje);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Callable_float_float_float_float);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Callable_float_float_float_float_2);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Callable_float_float_object_None);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Callable_int_object_Color);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Callable_int_object_None);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Callable_int_object_bool);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Callable_object_None);
  Py_CLEAR(clear_module_state->__pyx_n_s_Color);
  Py_CLEAR(clear_module_state->__pyx_n_u_Color);
  Py_CLEAR(clear_module_state->__pyx_n_s_ColorLine);
  Py_CLEAR(clear_module_state->__pyx_n_s_ColorLine___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_ColorLine___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_ColorStop);
  Py_CLEAR(clear_module_state->__pyx_n_u_ColorStop);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Color_None);
  Py_CLEAR(clear_module_state->__pyx_n_s_Color_from_int);
  Py_CLEAR(clear_module_state->__pyx_n_s_Color_to_int);
  Py_CLEAR(clear_module_state->__pyx_n_s_DARKEN);
  Py_CLEAR(clear_module_state->__pyx_n_s_DEFAULT);
  Py_CLEAR(clear_module_state->__pyx_n_s_DEFAULT_REPLACEMENT_CODEPOINT);
  Py_CLEAR(clear_module_state->__pyx_n_s_DELIMITED_SUB_FORMULA_MIN_HEIGHT);
  Py_CLEAR(clear_module_state->__pyx_n_s_DEST);
  Py_CLEAR(clear_module_state->__pyx_n_s_DEST_ATOP);
  Py_CLEAR(clear_module_state->__pyx_n_s_DEST_IN);
  Py_CLEAR(clear_module_state->__pyx_n_s_DEST_OUT);
  Py_CLEAR(clear_module_state->__pyx_n_s_DEST_OVER);
  Py_CLEAR(clear_module_state->__pyx_n_s_DESUBROUTINIZE);
  Py_CLEAR(clear_module_state->__pyx_n_s_DIFFERENCE);
  Py_CLEAR(clear_module_state->__pyx_n_s_DISPLAY_OPERATOR_MIN_HEIGHT);
  Py_CLEAR(clear_module_state->__pyx_n_s_DO_NOT_INSERT_DOTTED_CIRCLE);
  Py_CLEAR(clear_module_state->__pyx_n_s_DROP_TABLE_TAG);
  Py_CLEAR(clear_module_state->__pyx_n_s_DeprecationWarning);
  Py_CLEAR(clear_module_state->__pyx_n_s_Dict);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Dict_str_Union_int_bool_Sequence);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Dict_str_float);
  Py_CLEAR(clear_module_state->__pyx_n_s_DrawFuncs);
  Py_CLEAR(clear_module_state->__pyx_n_s_DrawFuncs___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_DrawFuncs___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_DrawFuncs_draw_glyph);
  Py_CLEAR(clear_module_state->__pyx_n_s_DrawFuncs_get_glyph_shape);
  Py_CLEAR(clear_module_state->__pyx_n_s_DrawFuncs_set_close_path_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_DrawFuncs_set_cubic_to_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_DrawFuncs_set_line_to_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_DrawFuncs_set_move_to_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_DrawFuncs_set_quadratic_to_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_EOT);
  Py_CLEAR(clear_module_state->__pyx_n_s_EXCLUSION);
  Py_CLEAR(clear_module_state->__pyx_n_s_EXTENDER);
  Py_CLEAR(clear_module_state->__pyx_n_s_FLATTENED_ACCENT_BASE_HEIGHT);
  Py_CLEAR(clear_module_state->__pyx_n_s_FRACTION_DENOMINATOR_DISPLAY_STY);
  Py_CLEAR(clear_module_state->__pyx_n_s_FRACTION_DENOMINATOR_GAP_MIN);
  Py_CLEAR(clear_module_state->__pyx_n_s_FRACTION_DENOMINATOR_SHIFT_DOWN);
  Py_CLEAR(clear_module_state->__pyx_n_s_FRACTION_DENOM_DISPLAY_STYLE_GAP);
  Py_CLEAR(clear_module_state->__pyx_n_s_FRACTION_NUMERATOR_DISPLAY_STYLE);
  Py_CLEAR(clear_module_state->__pyx_n_s_FRACTION_NUMERATOR_GAP_MIN);
  Py_CLEAR(clear_module_state->__pyx_n_s_FRACTION_NUMERATOR_SHIFT_UP);
  Py_CLEAR(clear_module_state->__pyx_n_s_FRACTION_NUM_DISPLAY_STYLE_GAP_M);
  Py_CLEAR(clear_module_state->__pyx_n_s_FRACTION_RULE_THICKNESS);
  Py_CLEAR(clear_module_state->__pyx_n_s_Face);
  Py_CLEAR(clear_module_state->__pyx_n_s_Face___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Face___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Face_create);
  Py_CLEAR(clear_module_state->__pyx_n_s_Face_create_for_tables);
  Py_CLEAR(clear_module_state->__pyx_n_s_Face_variation_unicodes);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Failed_to_open);
  Py_CLEAR(clear_module_state->__pyx_n_s_Font);
  Py_CLEAR(clear_module_state->__pyx_n_s_FontExtents);
  Py_CLEAR(clear_module_state->__pyx_n_u_FontExtents);
  Py_CLEAR(clear_module_state->__pyx_n_s_FontFuncs);
  Py_CLEAR(clear_module_state->__pyx_n_s_FontFuncs___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_FontFuncs___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_FontFuncs_create);
  Py_CLEAR(clear_module_state->__pyx_n_s_FontFuncs_set_font_h_extents_fun);
  Py_CLEAR(clear_module_state->__pyx_n_s_FontFuncs_set_font_v_extents_fun);
  Py_CLEAR(clear_module_state->__pyx_n_s_FontFuncs_set_glyph_h_advance_fu);
  Py_CLEAR(clear_module_state->__pyx_n_s_FontFuncs_set_glyph_name_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_FontFuncs_set_glyph_v_advance_fu);
  Py_CLEAR(clear_module_state->__pyx_n_s_FontFuncs_set_glyph_v_origin_fun);
  Py_CLEAR(clear_module_state->__pyx_n_s_FontFuncs_set_nominal_glyph_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_FontFuncs_set_variation_glyph_fu);
  Py_CLEAR(clear_module_state->__pyx_n_s_Font___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Font___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Font_create);
  Py_CLEAR(clear_module_state->__pyx_n_s_Font_draw_glyph);
  Py_CLEAR(clear_module_state->__pyx_n_s_Font_draw_glyph_with_pen);
  Py_CLEAR(clear_module_state->__pyx_n_s_Font_get_font_extents);
  Py_CLEAR(clear_module_state->__pyx_n_s_Font_get_glyph_extents);
  Py_CLEAR(clear_module_state->__pyx_n_s_Font_get_glyph_from_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_Font_get_glyph_h_advance);
  Py_CLEAR(clear_module_state->__pyx_n_s_Font_get_glyph_h_origin);
  Py_CLEAR(clear_module_state->__pyx_n_s_Font_get_glyph_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_Font_get_glyph_v_advance);
  Py_CLEAR(clear_module_state->__pyx_n_s_Font_get_glyph_v_origin);
  Py_CLEAR(clear_module_state->__pyx_n_s_Font_get_nominal_glyph);
  Py_CLEAR(clear_module_state->__pyx_n_s_Font_get_var_coords_design);
  Py_CLEAR(clear_module_state->__pyx_n_s_Font_get_var_coords_normalized);
  Py_CLEAR(clear_module_state->__pyx_n_s_Font_get_variation_glyph);
  Py_CLEAR(clear_module_state->__pyx_n_s_Font_glyph_from_string);
  Py_CLEAR(clear_module_state->__pyx_n_s_Font_glyph_to_string);
  Py_CLEAR(clear_module_state->__pyx_n_s_Font_paint_glyph);
  Py_CLEAR(clear_module_state->__pyx_n_s_Font_set_var_coords_design);
  Py_CLEAR(clear_module_state->__pyx_n_s_Font_set_var_coords_normalized);
  Py_CLEAR(clear_module_state->__pyx_n_s_Font_set_variation);
  Py_CLEAR(clear_module_state->__pyx_n_s_Font_set_variations);
  Py_CLEAR(clear_module_state->__pyx_n_s_GLYPHS);
  Py_CLEAR(clear_module_state->__pyx_n_s_GLYPH_INDEX);
  Py_CLEAR(clear_module_state->__pyx_n_s_GLYPH_NAMES);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlyphExtents);
  Py_CLEAR(clear_module_state->__pyx_n_u_GlyphExtents);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlyphFlags);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlyphInfo);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlyphInfo___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlyphInfo___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlyphPosition);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlyphPosition___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_GlyphPosition___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_HARD_LIGHT);
  Py_CLEAR(clear_module_state->__pyx_n_s_HBObject);
  Py_CLEAR(clear_module_state->__pyx_n_s_HBObject___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_HBObject___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_HSL_COLOR);
  Py_CLEAR(clear_module_state->__pyx_n_s_HSL_HUE);
  Py_CLEAR(clear_module_state->__pyx_n_s_HSL_LUMINOSITY);
  Py_CLEAR(clear_module_state->__pyx_n_s_HSL_SATURATION);
  Py_CLEAR(clear_module_state->__pyx_n_s_HarfBuzzError);
  Py_CLEAR(clear_module_state->__pyx_n_s_INVALID);
  Py_CLEAR(clear_module_state->__pyx_n_s_INVALID_VALUE);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Input_num_of_subtables);
  Py_CLEAR(clear_module_state->__pyx_n_s_IntEnum);
  Py_CLEAR(clear_module_state->__pyx_n_s_IntFlag);
  Py_CLEAR(clear_module_state->__pyx_n_s_KeyError);
  Py_CLEAR(clear_module_state->__pyx_n_s_LAYOUT_FEATURE_TAG);
  Py_CLEAR(clear_module_state->__pyx_n_s_LAYOUT_SCRIPT_TAG);
  Py_CLEAR(clear_module_state->__pyx_n_s_LIGATURE);
  Py_CLEAR(clear_module_state->__pyx_n_s_LIGHTEN);
  Py_CLEAR(clear_module_state->__pyx_n_s_LOWER_LIMIT_BASELINE_DROP_MIN);
  Py_CLEAR(clear_module_state->__pyx_n_s_LOWER_LIMIT_GAP_MIN);
  Py_CLEAR(clear_module_state->__pyx_n_s_List);
  Py_CLEAR(clear_module_state->__pyx_kp_s_List_Color);
  Py_CLEAR(clear_module_state->__pyx_kp_s_List_OTColorLayer);
  Py_CLEAR(clear_module_state->__pyx_kp_s_List_OTMathGlyphVariant);
  Py_CLEAR(clear_module_state->__pyx_kp_s_List_OTMathKernEntry);
  Py_CLEAR(clear_module_state->__pyx_kp_s_List_Tuple_List_Tuple_int_int_in);
  Py_CLEAR(clear_module_state->__pyx_kp_s_List_bytes);
  Py_CLEAR(clear_module_state->__pyx_kp_s_List_int);
  Py_CLEAR(clear_module_state->__pyx_kp_s_List_str);
  Py_CLEAR(clear_module_state->__pyx_n_s_MARK);
  Py_CLEAR(clear_module_state->__pyx_n_s_MATH_LEADING);
  Py_CLEAR(clear_module_state->__pyx_n_s_MONOTONE_CHARACTERS);
  Py_CLEAR(clear_module_state->__pyx_n_s_MONOTONE_GRAPHEMES);
  Py_CLEAR(clear_module_state->__pyx_n_s_MULTIPLY);
  Py_CLEAR(clear_module_state->__pyx_n_s_Map);
  Py_CLEAR(clear_module_state->__pyx_n_s_MapIter);
  Py_CLEAR(clear_module_state->__pyx_n_s_MapIter___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_MapIter___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Map___copy);
  Py_CLEAR(clear_module_state->__pyx_n_s_Map___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Map___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Map__is_equal);
  Py_CLEAR(clear_module_state->__pyx_n_s_Map__update);
  Py_CLEAR(clear_module_state->__pyx_n_s_Map_clear);
  Py_CLEAR(clear_module_state->__pyx_n_s_Map_copy);
  Py_CLEAR(clear_module_state->__pyx_n_s_Map_get);
  Py_CLEAR(clear_module_state->__pyx_n_s_Map_items);
  Py_CLEAR(clear_module_state->__pyx_n_s_Map_keys);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Map_s);
  Py_CLEAR(clear_module_state->__pyx_n_s_Map_update);
  Py_CLEAR(clear_module_state->__pyx_n_s_Map_values);
  Py_CLEAR(clear_module_state->__pyx_n_s_MemoryError);
  Py_CLEAR(clear_module_state->__pyx_n_s_NAME_ID);
  Py_CLEAR(clear_module_state->__pyx_n_s_NAME_LANG_ID);
  Py_CLEAR(clear_module_state->__pyx_n_s_NAME_LEGACY);
  Py_CLEAR(clear_module_state->__pyx_n_s_NOTDEF_OUTLINE);
  Py_CLEAR(clear_module_state->__pyx_n_s_NO_HINTING);
  Py_CLEAR(clear_module_state->__pyx_n_s_NO_LAYOUT_CLOSURE);
  Py_CLEAR(clear_module_state->__pyx_n_s_NO_PRUNE_UNICODE_RANGES);
  Py_CLEAR(clear_module_state->__pyx_n_s_NO_SUBSET_TABLE_TAG);
  Py_CLEAR(clear_module_state->__pyx_n_s_None);
  Py_CLEAR(clear_module_state->__pyx_n_s_NotImplemented);
  Py_CLEAR(clear_module_state->__pyx_n_s_OTColorLayer);
  Py_CLEAR(clear_module_state->__pyx_n_u_OTColorLayer);
  Py_CLEAR(clear_module_state->__pyx_n_s_OTColorPaletteFlags);
  Py_CLEAR(clear_module_state->__pyx_n_s_OTLayoutGlyphClass);
  Py_CLEAR(clear_module_state->__pyx_n_s_OTMathConstant);
  Py_CLEAR(clear_module_state->__pyx_n_s_OTMathGlyphPart);
  Py_CLEAR(clear_module_state->__pyx_n_u_OTMathGlyphPart);
  Py_CLEAR(clear_module_state->__pyx_n_s_OTMathGlyphPartFlags);
  Py_CLEAR(clear_module_state->__pyx_n_s_OTMathGlyphVariant);
  Py_CLEAR(clear_module_state->__pyx_n_u_OTMathGlyphVariant);
  Py_CLEAR(clear_module_state->__pyx_n_s_OTMathKern);
  Py_CLEAR(clear_module_state->__pyx_n_s_OTMathKernEntry);
  Py_CLEAR(clear_module_state->__pyx_n_u_OTMathKernEntry);
  Py_CLEAR(clear_module_state->__pyx_n_s_OVERBAR_EXTRA_ASCENDER);
  Py_CLEAR(clear_module_state->__pyx_n_s_OVERBAR_RULE_THICKNESS);
  Py_CLEAR(clear_module_state->__pyx_n_s_OVERBAR_VERTICAL_GAP);
  Py_CLEAR(clear_module_state->__pyx_n_s_OVERLAY);
  Py_CLEAR(clear_module_state->__pyx_n_s_PAD);
  Py_CLEAR(clear_module_state->__pyx_n_s_PASSTHROUGH_UNRECOGNIZED);
  Py_CLEAR(clear_module_state->__pyx_n_s_PLUS);
  Py_CLEAR(clear_module_state->__pyx_n_s_PRESERVE_DEFAULT_IGNORABLES);
  Py_CLEAR(clear_module_state->__pyx_n_s_PRODUCE_SAFE_TO_INSERT_TATWEEL);
  Py_CLEAR(clear_module_state->__pyx_n_s_PRODUCE_UNSAFE_TO_CONCAT);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaintCompositeMode);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaintExtend);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaintFuncs);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaintFuncs___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaintFuncs___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaintFuncs_set_color_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaintFuncs_set_color_glyph_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaintFuncs_set_custom_palette_co);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaintFuncs_set_image_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaintFuncs_set_linear_gradient_f);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaintFuncs_set_pop_clip_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaintFuncs_set_pop_group_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaintFuncs_set_pop_transform_fun);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaintFuncs_set_push_clip_glyph_f);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaintFuncs_set_push_clip_rectang);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaintFuncs_set_push_group_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaintFuncs_set_push_transform_fu);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaintFuncs_set_radial_gradient_f);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaintFuncs_set_sweep_gradient_fu);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Pass_draw_state_to_Font_draw_gly);
  Py_CLEAR(clear_module_state->__pyx_n_s_Path);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Pickling_of_struct_members_such);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Pickling_of_struct_members_such_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_RADICAL_DEGREE_BOTTOM_RAISE_PERC);
  Py_CLEAR(clear_module_state->__pyx_n_s_RADICAL_DISPLAY_STYLE_VERTICAL_G);
  Py_CLEAR(clear_module_state->__pyx_n_s_RADICAL_EXTRA_ASCENDER);
  Py_CLEAR(clear_module_state->__pyx_n_s_RADICAL_KERN_AFTER_DEGREE);
  Py_CLEAR(clear_module_state->__pyx_n_s_RADICAL_KERN_BEFORE_DEGREE);
  Py_CLEAR(clear_module_state->__pyx_n_s_RADICAL_RULE_THICKNESS);
  Py_CLEAR(clear_module_state->__pyx_n_s_RADICAL_VERTICAL_GAP);
  Py_CLEAR(clear_module_state->__pyx_n_s_REFLECT);
  Py_CLEAR(clear_module_state->__pyx_n_s_REMOVE_DEFAULT_IGNORABLES);
  Py_CLEAR(clear_module_state->__pyx_n_s_REPEAT);
  Py_CLEAR(clear_module_state->__pyx_n_s_RETAIN_GIDS);
  Py_CLEAR(clear_module_state->__pyx_n_s_RepackerError);
  Py_CLEAR(clear_module_state->__pyx_n_s_RuntimeError);
  Py_CLEAR(clear_module_state->__pyx_n_s_SAFE_TO_INSERT_TATWEEL);
  Py_CLEAR(clear_module_state->__pyx_n_s_SCREEN);
  Py_CLEAR(clear_module_state->__pyx_n_s_SCRIPT_PERCENT_SCALE_DOWN);
  Py_CLEAR(clear_module_state->__pyx_n_s_SCRIPT_SCRIPT_PERCENT_SCALE_DOWN);
  Py_CLEAR(clear_module_state->__pyx_n_s_SET_OVERLAPS_FLAG);
  Py_CLEAR(clear_module_state->__pyx_n_s_SKEWED_FRACTION_HORIZONTAL_GAP);
  Py_CLEAR(clear_module_state->__pyx_n_s_SKEWED_FRACTION_VERTICAL_GAP);
  Py_CLEAR(clear_module_state->__pyx_n_s_SOFT_LIGHT);
  Py_CLEAR(clear_module_state->__pyx_n_s_SPACE_AFTER_SCRIPT);
  Py_CLEAR(clear_module_state->__pyx_n_s_SRC);
  Py_CLEAR(clear_module_state->__pyx_n_s_SRC_ATOP);
  Py_CLEAR(clear_module_state->__pyx_n_s_SRC_IN);
  Py_CLEAR(clear_module_state->__pyx_n_s_SRC_OUT);
  Py_CLEAR(clear_module_state->__pyx_n_s_SRC_OVER);
  Py_CLEAR(clear_module_state->__pyx_n_s_STACK_BOTTOM_DISPLAY_STYLE_SHIFT);
  Py_CLEAR(clear_module_state->__pyx_n_s_STACK_BOTTOM_SHIFT_DOWN);
  Py_CLEAR(clear_module_state->__pyx_n_s_STACK_DISPLAY_STYLE_GAP_MIN);
  Py_CLEAR(clear_module_state->__pyx_n_s_STACK_GAP_MIN);
  Py_CLEAR(clear_module_state->__pyx_n_s_STACK_TOP_DISPLAY_STYLE_SHIFT_UP);
  Py_CLEAR(clear_module_state->__pyx_n_s_STACK_TOP_SHIFT_UP);
  Py_CLEAR(clear_module_state->__pyx_n_s_STRETCH_STACK_BOTTOM_SHIFT_DOWN);
  Py_CLEAR(clear_module_state->__pyx_n_s_STRETCH_STACK_GAP_ABOVE_MIN);
  Py_CLEAR(clear_module_state->__pyx_n_s_STRETCH_STACK_GAP_BELOW_MIN);
  Py_CLEAR(clear_module_state->__pyx_n_s_STRETCH_STACK_TOP_SHIFT_UP);
  Py_CLEAR(clear_module_state->__pyx_n_s_SUBSCRIPT_BASELINE_DROP_MIN);
  Py_CLEAR(clear_module_state->__pyx_n_s_SUBSCRIPT_SHIFT_DOWN);
  Py_CLEAR(clear_module_state->__pyx_n_s_SUBSCRIPT_TOP_MAX);
  Py_CLEAR(clear_module_state->__pyx_n_s_SUB_SUPERSCRIPT_GAP_MIN);
  Py_CLEAR(clear_module_state->__pyx_n_s_SUPERSCRIPT_BASELINE_DROP_MAX);
  Py_CLEAR(clear_module_state->__pyx_n_s_SUPERSCRIPT_BOTTOM_MAX_WITH_SUBS);
  Py_CLEAR(clear_module_state->__pyx_n_s_SUPERSCRIPT_BOTTOM_MIN);
  Py_CLEAR(clear_module_state->__pyx_n_s_SUPERSCRIPT_SHIFT_UP);
  Py_CLEAR(clear_module_state->__pyx_n_s_SUPERSCRIPT_SHIFT_UP_CRAMPED);
  Py_CLEAR(clear_module_state->__pyx_n_s_Sequence);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Set);
  Py_CLEAR(clear_module_state->__pyx_n_s_SetIter);
  Py_CLEAR(clear_module_state->__pyx_n_s_SetIter___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_SetIter___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Set_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_Set___copy);
  Py_CLEAR(clear_module_state->__pyx_n_s_Set___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Set___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_Set__is_equal);
  Py_CLEAR(clear_module_state->__pyx_n_s_Set__set);
  Py_CLEAR(clear_module_state->__pyx_n_s_Set__update);
  Py_CLEAR(clear_module_state->__pyx_n_s_Set_add);
  Py_CLEAR(clear_module_state->__pyx_n_s_Set_add_range);
  Py_CLEAR(clear_module_state->__pyx_n_s_Set_clear);
  Py_CLEAR(clear_module_state->__pyx_n_s_Set_copy);
  Py_CLEAR(clear_module_state->__pyx_n_s_Set_del_range);
  Py_CLEAR(clear_module_state->__pyx_n_s_Set_difference_update);
  Py_CLEAR(clear_module_state->__pyx_n_s_Set_discard);
  Py_CLEAR(clear_module_state->__pyx_n_s_Set_intersection_update);
  Py_CLEAR(clear_module_state->__pyx_n_s_Set_invert);
  Py_CLEAR(clear_module_state->__pyx_n_s_Set_is_inverted);
  Py_CLEAR(clear_module_state->__pyx_n_s_Set_issubset);
  Py_CLEAR(clear_module_state->__pyx_n_s_Set_issuperset);
  Py_CLEAR(clear_module_state->__pyx_n_s_Set_remove);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Set_s);
  Py_CLEAR(clear_module_state->__pyx_n_s_Set_set);
  Py_CLEAR(clear_module_state->__pyx_n_s_Set_symmetric_difference_update);
  Py_CLEAR(clear_module_state->__pyx_n_s_Set_update);
  Py_CLEAR(clear_module_state->__pyx_n_s_StopIteration);
  Py_CLEAR(clear_module_state->__pyx_n_s_SubsetFlags);
  Py_CLEAR(clear_module_state->__pyx_n_s_SubsetInput);
  Py_CLEAR(clear_module_state->__pyx_n_s_SubsetInputSets);
  Py_CLEAR(clear_module_state->__pyx_n_s_SubsetInput___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_SubsetInput___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_SubsetInput_keep_everything);
  Py_CLEAR(clear_module_state->__pyx_n_s_SubsetInput_pin_axis_location);
  Py_CLEAR(clear_module_state->__pyx_n_s_SubsetInput_pin_axis_to_default);
  Py_CLEAR(clear_module_state->__pyx_n_s_SubsetInput_sets);
  Py_CLEAR(clear_module_state->__pyx_n_s_SubsetInput_subset);
  Py_CLEAR(clear_module_state->__pyx_n_s_SubsetPlan);
  Py_CLEAR(clear_module_state->__pyx_n_s_SubsetPlan___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_SubsetPlan___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_SubsetPlan_execute);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Subsetting_failed);
  Py_CLEAR(clear_module_state->__pyx_n_s_TOP_LEFT);
  Py_CLEAR(clear_module_state->__pyx_n_s_TOP_RIGHT);
  Py_CLEAR(clear_module_state->__pyx_n_s_Tuple);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Tuple_List_OTMathGlyphPart_int);
  Py_CLEAR(clear_module_state->__pyx_n_s_TypeError);
  Py_CLEAR(clear_module_state->__pyx_n_s_UNCLASSIFIED);
  Py_CLEAR(clear_module_state->__pyx_n_s_UNDERBAR_EXTRA_DESCENDER);
  Py_CLEAR(clear_module_state->__pyx_n_s_UNDERBAR_RULE_THICKNESS);
  Py_CLEAR(clear_module_state->__pyx_n_s_UNDERBAR_VERTICAL_GAP);
  Py_CLEAR(clear_module_state->__pyx_n_s_UNICODE);
  Py_CLEAR(clear_module_state->__pyx_n_s_UNSAFE_TO_BREAK);
  Py_CLEAR(clear_module_state->__pyx_n_s_UNSAFE_TO_CONCAT);
  Py_CLEAR(clear_module_state->__pyx_n_s_UPPER_LIMIT_BASELINE_RISE_MIN);
  Py_CLEAR(clear_module_state->__pyx_n_s_UPPER_LIMIT_GAP_MIN);
  Py_CLEAR(clear_module_state->__pyx_n_s_USABLE_WITH_DARK_BACKGROUND);
  Py_CLEAR(clear_module_state->__pyx_n_s_USABLE_WITH_LIGHT_BACKGROUND);
  Py_CLEAR(clear_module_state->__pyx_n_s_Union);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Union_str_Path);
  Py_CLEAR(clear_module_state->__pyx_n_s_VERIFY);
  Py_CLEAR(clear_module_state->__pyx_n_s_ValueError);
  Py_CLEAR(clear_module_state->__pyx_n_s_XOR);
  Py_CLEAR(clear_module_state->__pyx_kp_u__10);
  Py_CLEAR(clear_module_state->__pyx_kp_b__15);
  Py_CLEAR(clear_module_state->__pyx_kp_u__15);
  Py_CLEAR(clear_module_state->__pyx_kp_u__16);
  Py_CLEAR(clear_module_state->__pyx_kp_u__19);
  Py_CLEAR(clear_module_state->__pyx_kp_u__21);
  Py_CLEAR(clear_module_state->__pyx_n_s__22);
  Py_CLEAR(clear_module_state->__pyx_kp_u__23);
  Py_CLEAR(clear_module_state->__pyx_n_s__287);
  Py_CLEAR(clear_module_state->__pyx_n_s_a);
  Py_CLEAR(clear_module_state->__pyx_n_s_add);
  Py_CLEAR(clear_module_state->__pyx_n_s_add_codepoints);
  Py_CLEAR(clear_module_state->__pyx_n_s_add_range);
  Py_CLEAR(clear_module_state->__pyx_n_s_add_str);
  Py_CLEAR(clear_module_state->__pyx_n_s_add_utf8);
  Py_CLEAR(clear_module_state->__pyx_n_u_advance);
  Py_CLEAR(clear_module_state->__pyx_n_s_alpha);
  Py_CLEAR(clear_module_state->__pyx_n_u_alpha);
  Py_CLEAR(clear_module_state->__pyx_n_s_alternate_count);
  Py_CLEAR(clear_module_state->__pyx_n_s_alternate_glyphs);
  Py_CLEAR(clear_module_state->__pyx_n_s_alternates);
  Py_CLEAR(clear_module_state->__pyx_n_s_args);
  Py_CLEAR(clear_module_state->__pyx_n_s_ascender);
  Py_CLEAR(clear_module_state->__pyx_n_u_ascender);
  Py_CLEAR(clear_module_state->__pyx_n_s_assembly);
  Py_CLEAR(clear_module_state->__pyx_n_s_assembly_array);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncio_coroutines);
  Py_CLEAR(clear_module_state->__pyx_n_s_b);
  Py_CLEAR(clear_module_state->__pyx_n_s_baseline_tag);
  Py_CLEAR(clear_module_state->__pyx_n_s_blob);
  Py_CLEAR(clear_module_state->__pyx_n_s_blob_data);
  Py_CLEAR(clear_module_state->__pyx_n_s_blob_length);
  Py_CLEAR(clear_module_state->__pyx_n_s_blue);
  Py_CLEAR(clear_module_state->__pyx_n_u_blue);
  Py_CLEAR(clear_module_state->__pyx_n_s_bool);
  Py_CLEAR(clear_module_state->__pyx_n_s_buffer);
  Py_CLEAR(clear_module_state->__pyx_n_s_bytes);
  Py_CLEAR(clear_module_state->__pyx_n_s_c);
  Py_CLEAR(clear_module_state->__pyx_n_s_c_coords);
  Py_CLEAR(clear_module_state->__pyx_n_s_c_foreground);
  Py_CLEAR(clear_module_state->__pyx_n_s_c_shapers);
  Py_CLEAR(clear_module_state->__pyx_n_s_callback);
  Py_CLEAR(clear_module_state->__pyx_n_s_clear);
  Py_CLEAR(clear_module_state->__pyx_n_s_clear_contents);
  Py_CLEAR(clear_module_state->__pyx_n_s_cline_in_traceback);
  Py_CLEAR(clear_module_state->__pyx_n_s_close);
  Py_CLEAR(clear_module_state->__pyx_n_s_closePath);
  Py_CLEAR(clear_module_state->__pyx_n_s_cls);
  Py_CLEAR(clear_module_state->__pyx_n_s_codepoints);
  Py_CLEAR(clear_module_state->__pyx_n_s_collections);
  Py_CLEAR(clear_module_state->__pyx_n_u_color);
  Py_CLEAR(clear_module_state->__pyx_n_s_color_count);
  Py_CLEAR(clear_module_state->__pyx_n_s_color_index);
  Py_CLEAR(clear_module_state->__pyx_n_u_color_index);
  Py_CLEAR(clear_module_state->__pyx_n_s_colors);
  Py_CLEAR(clear_module_state->__pyx_n_s_constant);
  Py_CLEAR(clear_module_state->__pyx_n_s_coords);
  Py_CLEAR(clear_module_state->__pyx_n_s_coords_2dot14);
  Py_CLEAR(clear_module_state->__pyx_n_s_copy);
  Py_CLEAR(clear_module_state->__pyx_n_s_copy_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_correction_height);
  Py_CLEAR(clear_module_state->__pyx_n_s_count);
  Py_CLEAR(clear_module_state->__pyx_n_s_create);
  Py_CLEAR(clear_module_state->__pyx_n_s_create_for_tables);
  Py_CLEAR(clear_module_state->__pyx_n_s_cstr);
  Py_CLEAR(clear_module_state->__pyx_n_s_curveTo);
  Py_CLEAR(clear_module_state->__pyx_n_s_data);
  Py_CLEAR(clear_module_state->__pyx_n_s_del_range);
  Py_CLEAR(clear_module_state->__pyx_n_s_descender);
  Py_CLEAR(clear_module_state->__pyx_n_u_descender);
  Py_CLEAR(clear_module_state->__pyx_n_s_dict);
  Py_CLEAR(clear_module_state->__pyx_n_s_difference_update);
  Py_CLEAR(clear_module_state->__pyx_n_s_direction);
  Py_CLEAR(clear_module_state->__pyx_kp_u_disable);
  Py_CLEAR(clear_module_state->__pyx_n_s_discard);
  Py_CLEAR(clear_module_state->__pyx_n_s_doc);
  Py_CLEAR(clear_module_state->__pyx_n_s_draw_data);
  Py_CLEAR(clear_module_state->__pyx_n_s_draw_funcs);
  Py_CLEAR(clear_module_state->__pyx_n_s_draw_glyph);
  Py_CLEAR(clear_module_state->__pyx_kp_u_draw_glyph_is_deprecated_use_Fon);
  Py_CLEAR(clear_module_state->__pyx_n_s_draw_glyph_with_pen);
  Py_CLEAR(clear_module_state->__pyx_n_s_draw_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_draw_state_p);
  Py_CLEAR(clear_module_state->__pyx_kp_u_enable);
  Py_CLEAR(clear_module_state->__pyx_n_s_encode);
  Py_CLEAR(clear_module_state->__pyx_n_s_end);
  Py_CLEAR(clear_module_state->__pyx_n_u_end_connector_length);
  Py_CLEAR(clear_module_state->__pyx_n_s_enum);
  Py_CLEAR(clear_module_state->__pyx_n_s_enumerate);
  Py_CLEAR(clear_module_state->__pyx_n_s_execute);
  Py_CLEAR(clear_module_state->__pyx_n_s_extents);
  Py_CLEAR(clear_module_state->__pyx_n_s_face);
  Py_CLEAR(clear_module_state->__pyx_n_s_face_or_font);
  Py_CLEAR(clear_module_state->__pyx_n_s_feat);
  Py_CLEAR(clear_module_state->__pyx_n_s_feature_count);
  Py_CLEAR(clear_module_state->__pyx_n_s_feature_tags);
  Py_CLEAR(clear_module_state->__pyx_n_s_features);
  Py_CLEAR(clear_module_state->__pyx_n_s_filename);
  Py_CLEAR(clear_module_state->__pyx_n_s_first);
  Py_CLEAR(clear_module_state->__pyx_n_u_flags);
  Py_CLEAR(clear_module_state->__pyx_n_s_float);
  Py_CLEAR(clear_module_state->__pyx_n_s_font);
  Py_CLEAR(clear_module_state->__pyx_n_s_foreground);
  Py_CLEAR(clear_module_state->__pyx_n_s_from_file_path);
  Py_CLEAR(clear_module_state->__pyx_n_s_from_int);
  Py_CLEAR(clear_module_state->__pyx_n_s_fsencode);
  Py_CLEAR(clear_module_state->__pyx_n_u_full_advance);
  Py_CLEAR(clear_module_state->__pyx_n_s_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_func_p);
  Py_CLEAR(clear_module_state->__pyx_n_s_funcs);
  Py_CLEAR(clear_module_state->__pyx_n_s_g);
  Py_CLEAR(clear_module_state->__pyx_kp_u_gc);
  Py_CLEAR(clear_module_state->__pyx_n_s_genexpr);
  Py_CLEAR(clear_module_state->__pyx_n_s_get);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_font_extents);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_glyph_extents);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_glyph_from_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_glyph_h_advance);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_glyph_h_origin);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_glyph_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_glyph_shape);
  Py_CLEAR(clear_module_state->__pyx_kp_u_get_glyph_shape_is_deprecated_us);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_glyph_v_advance);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_glyph_v_origin);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_nominal_glyph);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_var_coords_design);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_var_coords_normalized);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_variation_glyph);
  Py_CLEAR(clear_module_state->__pyx_n_s_getstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_gid);
  Py_CLEAR(clear_module_state->__pyx_n_s_glyph);
  Py_CLEAR(clear_module_state->__pyx_n_u_glyph);
  Py_CLEAR(clear_module_state->__pyx_n_s_glyph_from_string);
  Py_CLEAR(clear_module_state->__pyx_n_s_glyph_to_string);
  Py_CLEAR(clear_module_state->__pyx_n_s_graphnodes);
  Py_CLEAR(clear_module_state->__pyx_n_s_green);
  Py_CLEAR(clear_module_state->__pyx_n_u_green);
  Py_CLEAR(clear_module_state->__pyx_n_s_guess_segment_properties);
  Py_CLEAR(clear_module_state->__pyx_n_u_hang);
  Py_CLEAR(clear_module_state->__pyx_n_s_hb_baseline_tag);
  Py_CLEAR(clear_module_state->__pyx_n_s_hb_codepoint_t);
  Py_CLEAR(clear_module_state->__pyx_n_s_hb_codepoints);
  Py_CLEAR(clear_module_state->__pyx_n_s_hb_direction);
  Py_CLEAR(clear_module_state->__pyx_n_s_hb_features);
  Py_CLEAR(clear_module_state->__pyx_n_s_hb_language);
  Py_CLEAR(clear_module_state->__pyx_n_s_hb_language_tag);
  Py_CLEAR(clear_module_state->__pyx_n_s_hb_position);
  Py_CLEAR(clear_module_state->__pyx_n_s_hb_script);
  Py_CLEAR(clear_module_state->__pyx_n_s_hb_script_tag);
  Py_CLEAR(clear_module_state->__pyx_n_s_hb_tag);
  Py_CLEAR(clear_module_state->__pyx_n_s_hb_variations);
  Py_CLEAR(clear_module_state->__pyx_n_s_head);
  Py_CLEAR(clear_module_state->__pyx_n_u_height);
  Py_CLEAR(clear_module_state->__pyx_n_s_i);
  Py_CLEAR(clear_module_state->__pyx_n_u_icfb);
  Py_CLEAR(clear_module_state->__pyx_n_u_icft);
  Py_CLEAR(clear_module_state->__pyx_n_u_ideo);
  Py_CLEAR(clear_module_state->__pyx_n_u_idtp);
  Py_CLEAR(clear_module_state->__pyx_n_s_import);
  Py_CLEAR(clear_module_state->__pyx_n_s_index);
  Py_CLEAR(clear_module_state->__pyx_kp_u_index_out_of_range_for_feature_a);
  Py_CLEAR(clear_module_state->__pyx_n_s_init);
  Py_CLEAR(clear_module_state->__pyx_n_s_init_subclass);
  Py_CLEAR(clear_module_state->__pyx_n_s_initializing);
  Py_CLEAR(clear_module_state->__pyx_n_s_input);
  Py_CLEAR(clear_module_state->__pyx_n_s_inst);
  Py_CLEAR(clear_module_state->__pyx_n_s_int);
  Py_CLEAR(clear_module_state->__pyx_kp_s_int_None);
  Py_CLEAR(clear_module_state->__pyx_n_s_intersection_update);
  Py_CLEAR(clear_module_state->__pyx_kp_u_invalid_baseline_tag);
  Py_CLEAR(clear_module_state->__pyx_kp_u_invalid_constant);
  Py_CLEAR(clear_module_state->__pyx_kp_u_invalid_kern);
  Py_CLEAR(clear_module_state->__pyx_n_s_invert);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_coroutine);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_equal);
  Py_CLEAR(clear_module_state->__pyx_n_u_is_foreground);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_inverted);
  Py_CLEAR(clear_module_state->__pyx_kp_u_isenabled);
  Py_CLEAR(clear_module_state->__pyx_n_s_issubset);
  Py_CLEAR(clear_module_state->__pyx_n_s_issuperset);
  Py_CLEAR(clear_module_state->__pyx_n_s_italics_correction);
  Py_CLEAR(clear_module_state->__pyx_n_s_item_length);
  Py_CLEAR(clear_module_state->__pyx_n_s_item_offset);
  Py_CLEAR(clear_module_state->__pyx_n_s_items);
  Py_CLEAR(clear_module_state->__pyx_n_s_join);
  Py_CLEAR(clear_module_state->__pyx_n_s_k);
  Py_CLEAR(clear_module_state->__pyx_n_s_keep_everything);
  Py_CLEAR(clear_module_state->__pyx_n_s_kern);
  Py_CLEAR(clear_module_state->__pyx_n_u_kern_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_kerns);
  Py_CLEAR(clear_module_state->__pyx_n_s_kerns_array);
  Py_CLEAR(clear_module_state->__pyx_n_s_keys);
  Py_CLEAR(clear_module_state->__pyx_n_s_keys_locals_genexpr);
  Py_CLEAR(clear_module_state->__pyx_n_s_kind);
  Py_CLEAR(clear_module_state->__pyx_n_s_language_count);
  Py_CLEAR(clear_module_state->__pyx_n_s_language_index);
  Py_CLEAR(clear_module_state->__pyx_n_s_language_tag);
  Py_CLEAR(clear_module_state->__pyx_n_s_language_tags);
  Py_CLEAR(clear_module_state->__pyx_n_s_larger_set);
  Py_CLEAR(clear_module_state->__pyx_n_s_last);
  Py_CLEAR(clear_module_state->__pyx_n_s_layer_count);
  Py_CLEAR(clear_module_state->__pyx_n_s_layers);
  Py_CLEAR(clear_module_state->__pyx_n_s_length);
  Py_CLEAR(clear_module_state->__pyx_n_s_lineTo);
  Py_CLEAR(clear_module_state->__pyx_n_s_line_gap);
  Py_CLEAR(clear_module_state->__pyx_n_u_line_gap);
  Py_CLEAR(clear_module_state->__pyx_n_s_lookup_index);
  Py_CLEAR(clear_module_state->__pyx_n_s_m);
  Py_CLEAR(clear_module_state->__pyx_n_s_main);
  Py_CLEAR(clear_module_state->__pyx_n_u_math);
  Py_CLEAR(clear_module_state->__pyx_n_u_max_correction_height);
  Py_CLEAR(clear_module_state->__pyx_n_s_metaclass);
  Py_CLEAR(clear_module_state->__pyx_n_s_methods);
  Py_CLEAR(clear_module_state->__pyx_n_s_module);
  Py_CLEAR(clear_module_state->__pyx_n_s_moveTo);
  Py_CLEAR(clear_module_state->__pyx_n_s_mro_entries);
  Py_CLEAR(clear_module_state->__pyx_n_s_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_name_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_name_id);
  Py_CLEAR(clear_module_state->__pyx_n_s_namedtuple);
  Py_CLEAR(clear_module_state->__pyx_n_s_new_face);
  Py_CLEAR(clear_module_state->__pyx_kp_s_no_default___reduce___due_to_non);
  Py_CLEAR(clear_module_state->__pyx_n_s_node);
  Py_CLEAR(clear_module_state->__pyx_n_s_num_nodes);
  Py_CLEAR(clear_module_state->__pyx_kp_u_num_of_graph_nodes);
  Py_CLEAR(clear_module_state->__pyx_n_s_obj_list);
  Py_CLEAR(clear_module_state->__pyx_n_s_object);
  Py_CLEAR(clear_module_state->__pyx_n_u_offset);
  Py_CLEAR(clear_module_state->__pyx_n_s_os);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_color_glyph_get_layers);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_color_glyph_get_png);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_color_glyph_get_svg);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_color_glyph_has_paint);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_color_has_layers);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_color_has_paint);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_color_has_palettes);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_color_has_png);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_color_has_svg);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_color_palette_color_get_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_color_palette_get_colors);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_color_palette_get_count);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_color_palette_get_flags);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_color_palette_get_name_id);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_font_set_funcs);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_layout_get_baseline);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_layout_get_glyph_class);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_layout_has_glyph_classes);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_layout_has_positioning);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_layout_has_substitution);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_layout_language_get_feature_t);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_layout_lookup_get_glyph_alter);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_layout_script_get_language_ta);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_layout_table_get_script_tags);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_math_get_constant);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_math_get_glyph_assembly);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_math_get_glyph_italics_correc);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_math_get_glyph_kerning);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_math_get_glyph_kernings);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_math_get_glyph_top_accent_att);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_math_get_glyph_variants);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_math_get_min_connector_overla);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_math_has_data);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_math_is_glyph_extended_shape);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_tag_to_language);
  Py_CLEAR(clear_module_state->__pyx_n_s_ot_tag_to_script);
  Py_CLEAR(clear_module_state->__pyx_n_s_other);
  Py_CLEAR(clear_module_state->__pyx_n_s_p);
  Py_CLEAR(clear_module_state->__pyx_n_s_packed);
  Py_CLEAR(clear_module_state->__pyx_n_s_packed_blob);
  Py_CLEAR(clear_module_state->__pyx_n_s_paint_funcs);
  Py_CLEAR(clear_module_state->__pyx_n_s_paint_glyph);
  Py_CLEAR(clear_module_state->__pyx_n_s_paint_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_paint_state_p);
  Py_CLEAR(clear_module_state->__pyx_n_s_palette_index);
  Py_CLEAR(clear_module_state->__pyx_n_s_pathlib);
  Py_CLEAR(clear_module_state->__pyx_n_s_pen);
  Py_CLEAR(clear_module_state->__pyx_n_s_pin_axis_location);
  Py_CLEAR(clear_module_state->__pyx_n_s_pin_axis_to_default);
  Py_CLEAR(clear_module_state->__pyx_n_s_prepare);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_vtable);
  Py_CLEAR(clear_module_state->__pyx_n_s_qCurveTo);
  Py_CLEAR(clear_module_state->__pyx_n_s_qualname);
  Py_CLEAR(clear_module_state->__pyx_n_s_r);
  Py_CLEAR(clear_module_state->__pyx_n_s_range);
  Py_CLEAR(clear_module_state->__pyx_n_s_red);
  Py_CLEAR(clear_module_state->__pyx_n_u_red);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_ex);
  Py_CLEAR(clear_module_state->__pyx_n_s_remove);
  Py_CLEAR(clear_module_state->__pyx_n_s_repack);
  Py_CLEAR(clear_module_state->__pyx_n_s_repack_with_tag);
  Py_CLEAR(clear_module_state->__pyx_n_s_repr___locals_genexpr);
  Py_CLEAR(clear_module_state->__pyx_n_s_reset);
  Py_CLEAR(clear_module_state->__pyx_n_s_ret);
  Py_CLEAR(clear_module_state->__pyx_n_s_return);
  Py_CLEAR(clear_module_state->__pyx_n_u_romn);
  Py_CLEAR(clear_module_state->__pyx_n_s_round);
  Py_CLEAR(clear_module_state->__pyx_n_s_s);
  Py_CLEAR(clear_module_state->__pyx_n_s_script_count);
  Py_CLEAR(clear_module_state->__pyx_n_s_script_index);
  Py_CLEAR(clear_module_state->__pyx_n_s_script_tag);
  Py_CLEAR(clear_module_state->__pyx_n_s_script_tags);
  Py_CLEAR(clear_module_state->__pyx_n_s_self);
  Py_CLEAR(clear_module_state->__pyx_n_s_send);
  Py_CLEAR(clear_module_state->__pyx_n_s_set);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_close_path_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_color_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_color_glyph_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_cubic_to_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_custom_palette_color_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_font_h_extents_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_font_v_extents_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_glyph_h_advance_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_glyph_name_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_glyph_v_advance_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_glyph_v_origin_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_image_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_language_from_ot_tag);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_line_to_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_linear_gradient_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_message_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_move_to_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_nominal_glyph_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_pop_clip_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_pop_group_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_pop_transform_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_push_clip_glyph_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_push_clip_rectangle_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_push_group_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_push_transform_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_quadratic_to_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_radial_gradient_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_script_from_ot_tag);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_sweep_gradient_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_var_coords_design);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_var_coords_normalized);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_variation);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_variation_glyph_func);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_variations);
  Py_CLEAR(clear_module_state->__pyx_n_s_sets);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_shape);
  Py_CLEAR(clear_module_state->__pyx_n_s_shaper);
  Py_CLEAR(clear_module_state->__pyx_n_s_shapers);
  Py_CLEAR(clear_module_state->__pyx_n_s_size);
  Py_CLEAR(clear_module_state->__pyx_n_s_smaller_set);
  Py_CLEAR(clear_module_state->__pyx_n_s_source);
  Py_CLEAR(clear_module_state->__pyx_n_s_spec);
  Py_CLEAR(clear_module_state->__pyx_kp_s_src_uharfbuzz__harfbuzz_pyx);
  Py_CLEAR(clear_module_state->__pyx_n_s_start);
  Py_CLEAR(clear_module_state->__pyx_n_u_start_connector_length);
  Py_CLEAR(clear_module_state->__pyx_n_s_start_offset);
  Py_CLEAR(clear_module_state->__pyx_n_s_staticmethod);
  Py_CLEAR(clear_module_state->__pyx_n_s_str);
  Py_CLEAR(clear_module_state->__pyx_n_s_string);
  Py_CLEAR(clear_module_state->__pyx_kp_s_stringsource);
  Py_CLEAR(clear_module_state->__pyx_n_s_subset);
  Py_CLEAR(clear_module_state->__pyx_n_s_subset_preprocess);
  Py_CLEAR(clear_module_state->__pyx_n_s_subtables);
  Py_CLEAR(clear_module_state->__pyx_n_s_success);
  Py_CLEAR(clear_module_state->__pyx_n_s_super);
  Py_CLEAR(clear_module_state->__pyx_n_s_symmetric_difference_update);
  Py_CLEAR(clear_module_state->__pyx_n_s_table_bytes);
  Py_CLEAR(clear_module_state->__pyx_n_s_table_data);
  Py_CLEAR(clear_module_state->__pyx_n_s_tag);
  Py_CLEAR(clear_module_state->__pyx_n_s_tag_packed);
  Py_CLEAR(clear_module_state->__pyx_n_s_tags);
  Py_CLEAR(clear_module_state->__pyx_n_s_tail);
  Py_CLEAR(clear_module_state->__pyx_n_s_test);
  Py_CLEAR(clear_module_state->__pyx_n_s_text);
  Py_CLEAR(clear_module_state->__pyx_n_s_throw);
  Py_CLEAR(clear_module_state->__pyx_n_s_to_int);
  Py_CLEAR(clear_module_state->__pyx_n_s_typing);
  Py_CLEAR(clear_module_state->__pyx_n_s_uharfbuzz__harfbuzz);
  Py_CLEAR(clear_module_state->__pyx_n_s_unicode);
  Py_CLEAR(clear_module_state->__pyx_n_s_update);
  Py_CLEAR(clear_module_state->__pyx_n_s_update_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_user_data);
  Py_CLEAR(clear_module_state->__pyx_n_s_user_data_p);
  Py_CLEAR(clear_module_state->__pyx_n_s_v);
  Py_CLEAR(clear_module_state->__pyx_n_s_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_values);
  Py_CLEAR(clear_module_state->__pyx_n_s_values_locals_genexpr);
  Py_CLEAR(clear_module_state->__pyx_n_s_variants);
  Py_CLEAR(clear_module_state->__pyx_n_s_variants_array);
  Py_CLEAR(clear_module_state->__pyx_n_s_variation);
  Py_CLEAR(clear_module_state->__pyx_n_s_variation_selector);
  Py_CLEAR(clear_module_state->__pyx_n_s_variation_unicodes);
  Py_CLEAR(clear_module_state->__pyx_n_s_variations);
  Py_CLEAR(clear_module_state->__pyx_n_s_version_string);
  Py_CLEAR(clear_module_state->__pyx_n_s_warn);
  Py_CLEAR(clear_module_state->__pyx_n_s_warnings);
  Py_CLEAR(clear_module_state->__pyx_n_u_width);
  Py_CLEAR(clear_module_state->__pyx_n_s_x);
  Py_CLEAR(clear_module_state->__pyx_n_u_x_bearing);
  Py_CLEAR(clear_module_state->__pyx_n_s_y);
  Py_CLEAR(clear_module_state->__pyx_n_u_y_bearing);
  Py_CLEAR(clear_module_state->__pyx_int_0);
  Py_CLEAR(clear_module_state->__pyx_int_1);
  Py_CLEAR(clear_module_state->__pyx_int_16384);
  Py_CLEAR(clear_module_state->__pyx_int_65535);
  Py_CLEAR(clear_module_state->__pyx_int_neg_1);
  Py_CLEAR(clear_module_state->__pyx_k_);
  Py_CLEAR(clear_module_state->__pyx_k__2);
  Py_CLEAR(clear_module_state->__pyx_k__3);
  Py_CLEAR(clear_module_state->__pyx_k__4);
  Py_CLEAR(clear_module_state->__pyx_k__5);
  Py_CLEAR(clear_module_state->__pyx_k__6);
  Py_CLEAR(clear_module_state->__pyx_k__8);
  Py_CLEAR(clear_module_state->__pyx_k__18);
  Py_CLEAR(clear_module_state->__pyx_k__20);
  Py_CLEAR(clear_module_state->__pyx_tuple__7);
  Py_CLEAR(clear_module_state->__pyx_tuple__9);
  Py_CLEAR(clear_module_state->__pyx_tuple__11);
  Py_CLEAR(clear_module_state->__pyx_tuple__12);
  Py_CLEAR(clear_module_state->__pyx_tuple__13);
  Py_CLEAR(clear_module_state->__pyx_tuple__14);
  Py_CLEAR(clear_module_state->__pyx_tuple__17);
  Py_CLEAR(clear_module_state->__pyx_tuple__24);
  Py_CLEAR(clear_module_state->__pyx_tuple__26);
  Py_CLEAR(clear_module_state->__pyx_tuple__28);
  Py_CLEAR(clear_module_state->__pyx_tuple__32);
  Py_CLEAR(clear_module_state->__pyx_tuple__36);
  Py_CLEAR(clear_module_state->__pyx_tuple__39);
  Py_CLEAR(clear_module_state->__pyx_tuple__41);
  Py_CLEAR(clear_module_state->__pyx_tuple__43);
  Py_CLEAR(clear_module_state->__pyx_tuple__46);
  Py_CLEAR(clear_module_state->__pyx_tuple__50);
  Py_CLEAR(clear_module_state->__pyx_tuple__54);
  Py_CLEAR(clear_module_state->__pyx_tuple__56);
  Py_CLEAR(clear_module_state->__pyx_tuple__57);
  Py_CLEAR(clear_module_state->__pyx_tuple__59);
  Py_CLEAR(clear_module_state->__pyx_tuple__63);
  Py_CLEAR(clear_module_state->__pyx_tuple__65);
  Py_CLEAR(clear_module_state->__pyx_tuple__67);
  Py_CLEAR(clear_module_state->__pyx_tuple__69);
  Py_CLEAR(clear_module_state->__pyx_tuple__71);
  Py_CLEAR(clear_module_state->__pyx_tuple__73);
  Py_CLEAR(clear_module_state->__pyx_tuple__75);
  Py_CLEAR(clear_module_state->__pyx_tuple__78);
  Py_CLEAR(clear_module_state->__pyx_tuple__81);
  Py_CLEAR(clear_module_state->__pyx_tuple__83);
  Py_CLEAR(clear_module_state->__pyx_tuple__85);
  Py_CLEAR(clear_module_state->__pyx_tuple__87);
  Py_CLEAR(clear_module_state->__pyx_tuple__89);
  Py_CLEAR(clear_module_state->__pyx_tuple__92);
  Py_CLEAR(clear_module_state->__pyx_tuple__94);
  Py_CLEAR(clear_module_state->__pyx_tuple__96);
  Py_CLEAR(clear_module_state->__pyx_tuple__98);
  Py_CLEAR(clear_module_state->__pyx_tuple__100);
  Py_CLEAR(clear_module_state->__pyx_tuple__102);
  Py_CLEAR(clear_module_state->__pyx_tuple__107);
  Py_CLEAR(clear_module_state->__pyx_tuple__118);
  Py_CLEAR(clear_module_state->__pyx_tuple__120);
  Py_CLEAR(clear_module_state->__pyx_tuple__121);
  Py_CLEAR(clear_module_state->__pyx_tuple__123);
  Py_CLEAR(clear_module_state->__pyx_tuple__125);
  Py_CLEAR(clear_module_state->__pyx_tuple__127);
  Py_CLEAR(clear_module_state->__pyx_tuple__129);
  Py_CLEAR(clear_module_state->__pyx_tuple__131);
  Py_CLEAR(clear_module_state->__pyx_tuple__133);
  Py_CLEAR(clear_module_state->__pyx_tuple__135);
  Py_CLEAR(clear_module_state->__pyx_tuple__139);
  Py_CLEAR(clear_module_state->__pyx_tuple__143);
  Py_CLEAR(clear_module_state->__pyx_tuple__145);
  Py_CLEAR(clear_module_state->__pyx_tuple__147);
  Py_CLEAR(clear_module_state->__pyx_tuple__149);
  Py_CLEAR(clear_module_state->__pyx_tuple__152);
  Py_CLEAR(clear_module_state->__pyx_tuple__157);
  Py_CLEAR(clear_module_state->__pyx_tuple__160);
  Py_CLEAR(clear_module_state->__pyx_tuple__163);
  Py_CLEAR(clear_module_state->__pyx_tuple__165);
  Py_CLEAR(clear_module_state->__pyx_tuple__169);
  Py_CLEAR(clear_module_state->__pyx_tuple__171);
  Py_CLEAR(clear_module_state->__pyx_tuple__173);
  Py_CLEAR(clear_module_state->__pyx_tuple__175);
  Py_CLEAR(clear_module_state->__pyx_tuple__177);
  Py_CLEAR(clear_module_state->__pyx_tuple__179);
  Py_CLEAR(clear_module_state->__pyx_tuple__182);
  Py_CLEAR(clear_module_state->__pyx_tuple__186);
  Py_CLEAR(clear_module_state->__pyx_tuple__203);
  Py_CLEAR(clear_module_state->__pyx_tuple__205);
  Py_CLEAR(clear_module_state->__pyx_tuple__207);
  Py_CLEAR(clear_module_state->__pyx_tuple__217);
  Py_CLEAR(clear_module_state->__pyx_tuple__219);
  Py_CLEAR(clear_module_state->__pyx_tuple__221);
  Py_CLEAR(clear_module_state->__pyx_tuple__223);
  Py_CLEAR(clear_module_state->__pyx_tuple__225);
  Py_CLEAR(clear_module_state->__pyx_tuple__228);
  Py_CLEAR(clear_module_state->__pyx_tuple__230);
  Py_CLEAR(clear_module_state->__pyx_tuple__232);
  Py_CLEAR(clear_module_state->__pyx_tuple__236);
  Py_CLEAR(clear_module_state->__pyx_tuple__240);
  Py_CLEAR(clear_module_state->__pyx_tuple__247);
  Py_CLEAR(clear_module_state->__pyx_tuple__252);
  Py_CLEAR(clear_module_state->__pyx_tuple__254);
  Py_CLEAR(clear_module_state->__pyx_tuple__256);
  Py_CLEAR(clear_module_state->__pyx_tuple__259);
  Py_CLEAR(clear_module_state->__pyx_tuple__273);
  Py_CLEAR(clear_module_state->__pyx_tuple__277);
  Py_CLEAR(clear_module_state->__pyx_tuple__280);
  Py_CLEAR(clear_module_state->__pyx_codeobj__25);
  Py_CLEAR(clear_module_state->__pyx_codeobj__27);
  Py_CLEAR(clear_module_state->__pyx_codeobj__29);
  Py_CLEAR(clear_module_state->__pyx_codeobj__30);
  Py_CLEAR(clear_module_state->__pyx_codeobj__31);
  Py_CLEAR(clear_module_state->__pyx_codeobj__33);
  Py_CLEAR(clear_module_state->__pyx_codeobj__34);
  Py_CLEAR(clear_module_state->__pyx_codeobj__35);
  Py_CLEAR(clear_module_state->__pyx_codeobj__37);
  Py_CLEAR(clear_module_state->__pyx_codeobj__38);
  Py_CLEAR(clear_module_state->__pyx_codeobj__40);
  Py_CLEAR(clear_module_state->__pyx_codeobj__42);
  Py_CLEAR(clear_module_state->__pyx_codeobj__44);
  Py_CLEAR(clear_module_state->__pyx_codeobj__45);
  Py_CLEAR(clear_module_state->__pyx_codeobj__47);
  Py_CLEAR(clear_module_state->__pyx_codeobj__48);
  Py_CLEAR(clear_module_state->__pyx_codeobj__49);
  Py_CLEAR(clear_module_state->__pyx_codeobj__51);
  Py_CLEAR(clear_module_state->__pyx_codeobj__52);
  Py_CLEAR(clear_module_state->__pyx_codeobj__53);
  Py_CLEAR(clear_module_state->__pyx_codeobj__55);
  Py_CLEAR(clear_module_state->__pyx_codeobj__58);
  Py_CLEAR(clear_module_state->__pyx_codeobj__60);
  Py_CLEAR(clear_module_state->__pyx_codeobj__61);
  Py_CLEAR(clear_module_state->__pyx_codeobj__62);
  Py_CLEAR(clear_module_state->__pyx_codeobj__64);
  Py_CLEAR(clear_module_state->__pyx_codeobj__66);
  Py_CLEAR(clear_module_state->__pyx_codeobj__68);
  Py_CLEAR(clear_module_state->__pyx_codeobj__70);
  Py_CLEAR(clear_module_state->__pyx_codeobj__72);
  Py_CLEAR(clear_module_state->__pyx_codeobj__74);
  Py_CLEAR(clear_module_state->__pyx_codeobj__76);
  Py_CLEAR(clear_module_state->__pyx_codeobj__77);
  Py_CLEAR(clear_module_state->__pyx_codeobj__79);
  Py_CLEAR(clear_module_state->__pyx_codeobj__80);
  Py_CLEAR(clear_module_state->__pyx_codeobj__82);
  Py_CLEAR(clear_module_state->__pyx_codeobj__84);
  Py_CLEAR(clear_module_state->__pyx_codeobj__86);
  Py_CLEAR(clear_module_state->__pyx_codeobj__88);
  Py_CLEAR(clear_module_state->__pyx_codeobj__90);
  Py_CLEAR(clear_module_state->__pyx_codeobj__91);
  Py_CLEAR(clear_module_state->__pyx_codeobj__93);
  Py_CLEAR(clear_module_state->__pyx_codeobj__95);
  Py_CLEAR(clear_module_state->__pyx_codeobj__97);
  Py_CLEAR(clear_module_state->__pyx_codeobj__99);
  Py_CLEAR(clear_module_state->__pyx_codeobj__101);
  Py_CLEAR(clear_module_state->__pyx_codeobj__103);
  Py_CLEAR(clear_module_state->__pyx_codeobj__104);
  Py_CLEAR(clear_module_state->__pyx_codeobj__105);
  Py_CLEAR(clear_module_state->__pyx_codeobj__106);
  Py_CLEAR(clear_module_state->__pyx_codeobj__108);
  Py_CLEAR(clear_module_state->__pyx_codeobj__109);
  Py_CLEAR(clear_module_state->__pyx_codeobj__110);
  Py_CLEAR(clear_module_state->__pyx_codeobj__111);
  Py_CLEAR(clear_module_state->__pyx_codeobj__112);
  Py_CLEAR(clear_module_state->__pyx_codeobj__113);
  Py_CLEAR(clear_module_state->__pyx_codeobj__114);
  Py_CLEAR(clear_module_state->__pyx_codeobj__115);
  Py_CLEAR(clear_module_state->__pyx_codeobj__116);
  Py_CLEAR(clear_module_state->__pyx_codeobj__117);
  Py_CLEAR(clear_module_state->__pyx_codeobj__119);
  Py_CLEAR(clear_module_state->__pyx_codeobj__122);
  Py_CLEAR(clear_module_state->__pyx_codeobj__124);
  Py_CLEAR(clear_module_state->__pyx_codeobj__126);
  Py_CLEAR(clear_module_state->__pyx_codeobj__128);
  Py_CLEAR(clear_module_state->__pyx_codeobj__130);
  Py_CLEAR(clear_module_state->__pyx_codeobj__132);
  Py_CLEAR(clear_module_state->__pyx_codeobj__134);
  Py_CLEAR(clear_module_state->__pyx_codeobj__136);
  Py_CLEAR(clear_module_state->__pyx_codeobj__137);
  Py_CLEAR(clear_module_state->__pyx_codeobj__138);
  Py_CLEAR(clear_module_state->__pyx_codeobj__140);
  Py_CLEAR(clear_module_state->__pyx_codeobj__141);
  Py_CLEAR(clear_module_state->__pyx_codeobj__142);
  Py_CLEAR(clear_module_state->__pyx_codeobj__144);
  Py_CLEAR(clear_module_state->__pyx_codeobj__146);
  Py_CLEAR(clear_module_state->__pyx_codeobj__148);
  Py_CLEAR(clear_module_state->__pyx_codeobj__150);
  Py_CLEAR(clear_module_state->__pyx_codeobj__151);
  Py_CLEAR(clear_module_state->__pyx_codeobj__153);
  Py_CLEAR(clear_module_state->__pyx_codeobj__154);
  Py_CLEAR(clear_module_state->__pyx_codeobj__155);
  Py_CLEAR(clear_module_state->__pyx_codeobj__156);
  Py_CLEAR(clear_module_state->__pyx_codeobj__158);
  Py_CLEAR(clear_module_state->__pyx_codeobj__159);
  Py_CLEAR(clear_module_state->__pyx_codeobj__161);
  Py_CLEAR(clear_module_state->__pyx_codeobj__162);
  Py_CLEAR(clear_module_state->__pyx_codeobj__164);
  Py_CLEAR(clear_module_state->__pyx_codeobj__166);
  Py_CLEAR(clear_module_state->__pyx_codeobj__167);
  Py_CLEAR(clear_module_state->__pyx_codeobj__168);
  Py_CLEAR(clear_module_state->__pyx_codeobj__170);
  Py_CLEAR(clear_module_state->__pyx_codeobj__172);
  Py_CLEAR(clear_module_state->__pyx_codeobj__174);
  Py_CLEAR(clear_module_state->__pyx_codeobj__176);
  Py_CLEAR(clear_module_state->__pyx_codeobj__178);
  Py_CLEAR(clear_module_state->__pyx_codeobj__180);
  Py_CLEAR(clear_module_state->__pyx_codeobj__181);
  Py_CLEAR(clear_module_state->__pyx_codeobj__183);
  Py_CLEAR(clear_module_state->__pyx_codeobj__184);
  Py_CLEAR(clear_module_state->__pyx_codeobj__185);
  Py_CLEAR(clear_module_state->__pyx_codeobj__187);
  Py_CLEAR(clear_module_state->__pyx_codeobj__188);
  Py_CLEAR(clear_module_state->__pyx_codeobj__189);
  Py_CLEAR(clear_module_state->__pyx_codeobj__190);
  Py_CLEAR(clear_module_state->__pyx_codeobj__191);
  Py_CLEAR(clear_module_state->__pyx_codeobj__192);
  Py_CLEAR(clear_module_state->__pyx_codeobj__193);
  Py_CLEAR(clear_module_state->__pyx_codeobj__194);
  Py_CLEAR(clear_module_state->__pyx_codeobj__195);
  Py_CLEAR(clear_module_state->__pyx_codeobj__196);
  Py_CLEAR(clear_module_state->__pyx_codeobj__197);
  Py_CLEAR(clear_module_state->__pyx_codeobj__198);
  Py_CLEAR(clear_module_state->__pyx_codeobj__199);
  Py_CLEAR(clear_module_state->__pyx_codeobj__200);
  Py_CLEAR(clear_module_state->__pyx_codeobj__201);
  Py_CLEAR(clear_module_state->__pyx_codeobj__202);
  Py_CLEAR(clear_module_state->__pyx_codeobj__204);
  Py_CLEAR(clear_module_state->__pyx_codeobj__206);
  Py_CLEAR(clear_module_state->__pyx_codeobj__208);
  Py_CLEAR(clear_module_state->__pyx_codeobj__209);
  Py_CLEAR(clear_module_state->__pyx_codeobj__210);
  Py_CLEAR(clear_module_state->__pyx_codeobj__211);
  Py_CLEAR(clear_module_state->__pyx_codeobj__212);
  Py_CLEAR(clear_module_state->__pyx_codeobj__213);
  Py_CLEAR(clear_module_state->__pyx_codeobj__214);
  Py_CLEAR(clear_module_state->__pyx_codeobj__215);
  Py_CLEAR(clear_module_state->__pyx_codeobj__216);
  Py_CLEAR(clear_module_state->__pyx_codeobj__218);
  Py_CLEAR(clear_module_state->__pyx_codeobj__220);
  Py_CLEAR(clear_module_state->__pyx_codeobj__222);
  Py_CLEAR(clear_module_state->__pyx_codeobj__224);
  Py_CLEAR(clear_module_state->__pyx_codeobj__226);
  Py_CLEAR(clear_module_state->__pyx_codeobj__227);
  Py_CLEAR(clear_module_state->__pyx_codeobj__229);
  Py_CLEAR(clear_module_state->__pyx_codeobj__231);
  Py_CLEAR(clear_module_state->__pyx_codeobj__233);
  Py_CLEAR(clear_module_state->__pyx_codeobj__234);
  Py_CLEAR(clear_module_state->__pyx_codeobj__235);
  Py_CLEAR(clear_module_state->__pyx_codeobj__237);
  Py_CLEAR(clear_module_state->__pyx_codeobj__238);
  Py_CLEAR(clear_module_state->__pyx_codeobj__239);
  Py_CLEAR(clear_module_state->__pyx_codeobj__241);
  Py_CLEAR(clear_module_state->__pyx_codeobj__242);
  Py_CLEAR(clear_module_state->__pyx_codeobj__243);
  Py_CLEAR(clear_module_state->__pyx_codeobj__244);
  Py_CLEAR(clear_module_state->__pyx_codeobj__245);
  Py_CLEAR(clear_module_state->__pyx_codeobj__246);
  Py_CLEAR(clear_module_state->__pyx_codeobj__248);
  Py_CLEAR(clear_module_state->__pyx_codeobj__249);
  Py_CLEAR(clear_module_state->__pyx_codeobj__250);
  Py_CLEAR(clear_module_state->__pyx_codeobj__251);
  Py_CLEAR(clear_module_state->__pyx_codeobj__253);
  Py_CLEAR(clear_module_state->__pyx_codeobj__255);
  Py_CLEAR(clear_module_state->__pyx_codeobj__257);
  Py_CLEAR(clear_module_state->__pyx_codeobj__258);
  Py_CLEAR(clear_module_state->__pyx_codeobj__260);
  Py_CLEAR(clear_module_state->__pyx_codeobj__261);
  Py_CLEAR(clear_module_state->__pyx_codeobj__262);
  Py_CLEAR(clear_module_state->__pyx_codeobj__263);
  Py_CLEAR(clear_module_state->__pyx_codeobj__264);
  Py_CLEAR(clear_module_state->__pyx_codeobj__265);
  Py_CLEAR(clear_module_state->__pyx_codeobj__266);
  Py_CLEAR(clear_module_state->__pyx_codeobj__267);
  Py_CLEAR(clear_module_state->__pyx_codeobj__268);
  Py_CLEAR(clear_module_state->__pyx_codeobj__269);
  Py_CLEAR(clear_module_state->__pyx_codeobj__270);
  Py_CLEAR(clear_module_state->__pyx_codeobj__271);
  Py_CLEAR(clear_module_state->__pyx_codeobj__272);
  Py_CLEAR(clear_module_state->__pyx_codeobj__274);
  Py_CLEAR(clear_module_state->__pyx_codeobj__275);
  Py_CLEAR(clear_module_state->__pyx_codeobj__276);
  Py_CLEAR(clear_module_state->__pyx_codeobj__278);
  Py_CLEAR(clear_module_state->__pyx_codeobj__279);
  Py_CLEAR(clear_module_state->__pyx_codeobj__281);
  Py_CLEAR(clear_module_state->__pyx_codeobj__282);
  Py_CLEAR(clear_module_state->__pyx_codeobj__283);
  Py_CLEAR(clear_module_state->__pyx_codeobj__284);
  Py_CLEAR(clear_module_state->__pyx_codeobj__285);
  Py_CLEAR(clear_module_state->__pyx_codeobj__286);
  return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstate *traverse_module_state = __pyx_mstate(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  Py_VISIT(traverse_module_state->__pyx_empty_tuple);
  Py_VISIT(traverse_module_state->__pyx_empty_bytes);
  Py_VISIT(traverse_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_VISIT(traverse_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_VISIT(traverse_module_state->__pyx_FusedFunctionType);
  #endif
  Py_VISIT(traverse_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphInfo);
  Py_VISIT(traverse_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_GlyphInfo);
  Py_VISIT(traverse_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphPosition);
  Py_VISIT(traverse_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_GlyphPosition);
  Py_VISIT(traverse_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer);
  Py_VISIT(traverse_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_Buffer);
  Py_VISIT(traverse_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_Blob);
  Py_VISIT(traverse_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_Blob);
  Py_VISIT(traverse_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_Face);
  Py_VISIT(traverse_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_Face);
  Py_VISIT(traverse_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_Font);
  Py_VISIT(traverse_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_Font);
  Py_VISIT(traverse_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs);
  Py_VISIT(traverse_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_FontFuncs);
  Py_VISIT(traverse_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_ColorLine);
  Py_VISIT(traverse_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_ColorLine);
  Py_VISIT(traverse_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs);
  Py_VISIT(traverse_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_PaintFuncs);
  Py_VISIT(traverse_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_DrawFuncs);
  Py_VISIT(traverse_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_DrawFuncs);
  Py_VISIT(traverse_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_HBObject);
  Py_VISIT(traverse_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_HBObject);
  Py_VISIT(traverse_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetInput);
  Py_VISIT(traverse_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_SubsetInput);
  Py_VISIT(traverse_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetPlan);
  Py_VISIT(traverse_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_SubsetPlan);
  Py_VISIT(traverse_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_Set);
  Py_VISIT(traverse_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_Set);
  Py_VISIT(traverse_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_SetIter);
  Py_VISIT(traverse_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_SetIter);
  Py_VISIT(traverse_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_Map);
  Py_VISIT(traverse_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_Map);
  Py_VISIT(traverse_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz_MapIter);
  Py_VISIT(traverse_module_state->__pyx_type_9uharfbuzz_9_harfbuzz_MapIter);
  Py_VISIT(traverse_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr);
  Py_VISIT(traverse_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr);
  Py_VISIT(traverse_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr);
  Py_VISIT(traverse_module_state->__pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr);
  Py_VISIT(traverse_module_state->__pyx_n_s_ACCENT_BASE_HEIGHT);
  Py_VISIT(traverse_module_state->__pyx_n_s_AXIS_HEIGHT);
  Py_VISIT(traverse_module_state->__pyx_kp_u_All_shapers_failed);
  Py_VISIT(traverse_module_state->__pyx_n_s_AssertionError);
  Py_VISIT(traverse_module_state->__pyx_n_s_BASE_GLYPH);
  Py_VISIT(traverse_module_state->__pyx_n_s_BOT);
  Py_VISIT(traverse_module_state->__pyx_n_s_BOTTOM_LEFT);
  Py_VISIT(traverse_module_state->__pyx_n_s_BOTTOM_RIGHT);
  Py_VISIT(traverse_module_state->__pyx_n_s_Blob);
  Py_VISIT(traverse_module_state->__pyx_n_s_Blob___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Blob___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Blob_from_file_path);
  Py_VISIT(traverse_module_state->__pyx_n_s_Buffer);
  Py_VISIT(traverse_module_state->__pyx_n_s_BufferClusterLevel);
  Py_VISIT(traverse_module_state->__pyx_n_s_BufferContentType);
  Py_VISIT(traverse_module_state->__pyx_n_s_BufferFlags);
  Py_VISIT(traverse_module_state->__pyx_n_s_Buffer___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Buffer___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Buffer_add_codepoints);
  Py_VISIT(traverse_module_state->__pyx_n_s_Buffer_add_str);
  Py_VISIT(traverse_module_state->__pyx_n_s_Buffer_add_utf8);
  Py_VISIT(traverse_module_state->__pyx_n_s_Buffer_clear_contents);
  Py_VISIT(traverse_module_state->__pyx_n_s_Buffer_create);
  Py_VISIT(traverse_module_state->__pyx_n_s_Buffer_guess_segment_properties);
  Py_VISIT(traverse_module_state->__pyx_n_s_Buffer_reset);
  Py_VISIT(traverse_module_state->__pyx_n_s_Buffer_set_language_from_ot_tag);
  Py_VISIT(traverse_module_state->__pyx_n_s_Buffer_set_message_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_Buffer_set_script_from_ot_tag);
  Py_VISIT(traverse_module_state->__pyx_n_s_CHARACTERS);
  Py_VISIT(traverse_module_state->__pyx_n_s_CLEAR);
  Py_VISIT(traverse_module_state->__pyx_n_s_COLOR_BURN);
  Py_VISIT(traverse_module_state->__pyx_n_s_COLOR_DODGE);
  Py_VISIT(traverse_module_state->__pyx_n_s_COMPONENT);
  Py_VISIT(traverse_module_state->__pyx_n_s_Callable);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Callable_Blob_int_int_str_float);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Callable_ColorLine_float_float_f);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Callable_ColorLine_float_float_f_2);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Callable_Color_bool_object_None);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Callable_Face_str_object_bytes);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Callable_Font_int_int_object_int);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Callable_Font_int_object_int);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Callable_Font_int_object_int_int);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Callable_Font_int_object_str);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Callable_Font_object_FontExtents);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Callable_PaintCompositeMode_obje);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Callable_float_float_float_float);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Callable_float_float_float_float_2);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Callable_float_float_object_None);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Callable_int_object_Color);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Callable_int_object_None);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Callable_int_object_bool);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Callable_object_None);
  Py_VISIT(traverse_module_state->__pyx_n_s_Color);
  Py_VISIT(traverse_module_state->__pyx_n_u_Color);
  Py_VISIT(traverse_module_state->__pyx_n_s_ColorLine);
  Py_VISIT(traverse_module_state->__pyx_n_s_ColorLine___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_ColorLine___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_ColorStop);
  Py_VISIT(traverse_module_state->__pyx_n_u_ColorStop);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Color_None);
  Py_VISIT(traverse_module_state->__pyx_n_s_Color_from_int);
  Py_VISIT(traverse_module_state->__pyx_n_s_Color_to_int);
  Py_VISIT(traverse_module_state->__pyx_n_s_DARKEN);
  Py_VISIT(traverse_module_state->__pyx_n_s_DEFAULT);
  Py_VISIT(traverse_module_state->__pyx_n_s_DEFAULT_REPLACEMENT_CODEPOINT);
  Py_VISIT(traverse_module_state->__pyx_n_s_DELIMITED_SUB_FORMULA_MIN_HEIGHT);
  Py_VISIT(traverse_module_state->__pyx_n_s_DEST);
  Py_VISIT(traverse_module_state->__pyx_n_s_DEST_ATOP);
  Py_VISIT(traverse_module_state->__pyx_n_s_DEST_IN);
  Py_VISIT(traverse_module_state->__pyx_n_s_DEST_OUT);
  Py_VISIT(traverse_module_state->__pyx_n_s_DEST_OVER);
  Py_VISIT(traverse_module_state->__pyx_n_s_DESUBROUTINIZE);
  Py_VISIT(traverse_module_state->__pyx_n_s_DIFFERENCE);
  Py_VISIT(traverse_module_state->__pyx_n_s_DISPLAY_OPERATOR_MIN_HEIGHT);
  Py_VISIT(traverse_module_state->__pyx_n_s_DO_NOT_INSERT_DOTTED_CIRCLE);
  Py_VISIT(traverse_module_state->__pyx_n_s_DROP_TABLE_TAG);
  Py_VISIT(traverse_module_state->__pyx_n_s_DeprecationWarning);
  Py_VISIT(traverse_module_state->__pyx_n_s_Dict);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Dict_str_Union_int_bool_Sequence);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Dict_str_float);
  Py_VISIT(traverse_module_state->__pyx_n_s_DrawFuncs);
  Py_VISIT(traverse_module_state->__pyx_n_s_DrawFuncs___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_DrawFuncs___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_DrawFuncs_draw_glyph);
  Py_VISIT(traverse_module_state->__pyx_n_s_DrawFuncs_get_glyph_shape);
  Py_VISIT(traverse_module_state->__pyx_n_s_DrawFuncs_set_close_path_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_DrawFuncs_set_cubic_to_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_DrawFuncs_set_line_to_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_DrawFuncs_set_move_to_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_DrawFuncs_set_quadratic_to_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_EOT);
  Py_VISIT(traverse_module_state->__pyx_n_s_EXCLUSION);
  Py_VISIT(traverse_module_state->__pyx_n_s_EXTENDER);
  Py_VISIT(traverse_module_state->__pyx_n_s_FLATTENED_ACCENT_BASE_HEIGHT);
  Py_VISIT(traverse_module_state->__pyx_n_s_FRACTION_DENOMINATOR_DISPLAY_STY);
  Py_VISIT(traverse_module_state->__pyx_n_s_FRACTION_DENOMINATOR_GAP_MIN);
  Py_VISIT(traverse_module_state->__pyx_n_s_FRACTION_DENOMINATOR_SHIFT_DOWN);
  Py_VISIT(traverse_module_state->__pyx_n_s_FRACTION_DENOM_DISPLAY_STYLE_GAP);
  Py_VISIT(traverse_module_state->__pyx_n_s_FRACTION_NUMERATOR_DISPLAY_STYLE);
  Py_VISIT(traverse_module_state->__pyx_n_s_FRACTION_NUMERATOR_GAP_MIN);
  Py_VISIT(traverse_module_state->__pyx_n_s_FRACTION_NUMERATOR_SHIFT_UP);
  Py_VISIT(traverse_module_state->__pyx_n_s_FRACTION_NUM_DISPLAY_STYLE_GAP_M);
  Py_VISIT(traverse_module_state->__pyx_n_s_FRACTION_RULE_THICKNESS);
  Py_VISIT(traverse_module_state->__pyx_n_s_Face);
  Py_VISIT(traverse_module_state->__pyx_n_s_Face___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Face___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Face_create);
  Py_VISIT(traverse_module_state->__pyx_n_s_Face_create_for_tables);
  Py_VISIT(traverse_module_state->__pyx_n_s_Face_variation_unicodes);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Failed_to_open);
  Py_VISIT(traverse_module_state->__pyx_n_s_Font);
  Py_VISIT(traverse_module_state->__pyx_n_s_FontExtents);
  Py_VISIT(traverse_module_state->__pyx_n_u_FontExtents);
  Py_VISIT(traverse_module_state->__pyx_n_s_FontFuncs);
  Py_VISIT(traverse_module_state->__pyx_n_s_FontFuncs___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_FontFuncs___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_FontFuncs_create);
  Py_VISIT(traverse_module_state->__pyx_n_s_FontFuncs_set_font_h_extents_fun);
  Py_VISIT(traverse_module_state->__pyx_n_s_FontFuncs_set_font_v_extents_fun);
  Py_VISIT(traverse_module_state->__pyx_n_s_FontFuncs_set_glyph_h_advance_fu);
  Py_VISIT(traverse_module_state->__pyx_n_s_FontFuncs_set_glyph_name_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_FontFuncs_set_glyph_v_advance_fu);
  Py_VISIT(traverse_module_state->__pyx_n_s_FontFuncs_set_glyph_v_origin_fun);
  Py_VISIT(traverse_module_state->__pyx_n_s_FontFuncs_set_nominal_glyph_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_FontFuncs_set_variation_glyph_fu);
  Py_VISIT(traverse_module_state->__pyx_n_s_Font___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Font___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Font_create);
  Py_VISIT(traverse_module_state->__pyx_n_s_Font_draw_glyph);
  Py_VISIT(traverse_module_state->__pyx_n_s_Font_draw_glyph_with_pen);
  Py_VISIT(traverse_module_state->__pyx_n_s_Font_get_font_extents);
  Py_VISIT(traverse_module_state->__pyx_n_s_Font_get_glyph_extents);
  Py_VISIT(traverse_module_state->__pyx_n_s_Font_get_glyph_from_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_Font_get_glyph_h_advance);
  Py_VISIT(traverse_module_state->__pyx_n_s_Font_get_glyph_h_origin);
  Py_VISIT(traverse_module_state->__pyx_n_s_Font_get_glyph_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_Font_get_glyph_v_advance);
  Py_VISIT(traverse_module_state->__pyx_n_s_Font_get_glyph_v_origin);
  Py_VISIT(traverse_module_state->__pyx_n_s_Font_get_nominal_glyph);
  Py_VISIT(traverse_module_state->__pyx_n_s_Font_get_var_coords_design);
  Py_VISIT(traverse_module_state->__pyx_n_s_Font_get_var_coords_normalized);
  Py_VISIT(traverse_module_state->__pyx_n_s_Font_get_variation_glyph);
  Py_VISIT(traverse_module_state->__pyx_n_s_Font_glyph_from_string);
  Py_VISIT(traverse_module_state->__pyx_n_s_Font_glyph_to_string);
  Py_VISIT(traverse_module_state->__pyx_n_s_Font_paint_glyph);
  Py_VISIT(traverse_module_state->__pyx_n_s_Font_set_var_coords_design);
  Py_VISIT(traverse_module_state->__pyx_n_s_Font_set_var_coords_normalized);
  Py_VISIT(traverse_module_state->__pyx_n_s_Font_set_variation);
  Py_VISIT(traverse_module_state->__pyx_n_s_Font_set_variations);
  Py_VISIT(traverse_module_state->__pyx_n_s_GLYPHS);
  Py_VISIT(traverse_module_state->__pyx_n_s_GLYPH_INDEX);
  Py_VISIT(traverse_module_state->__pyx_n_s_GLYPH_NAMES);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlyphExtents);
  Py_VISIT(traverse_module_state->__pyx_n_u_GlyphExtents);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlyphFlags);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlyphInfo);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlyphInfo___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlyphInfo___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlyphPosition);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlyphPosition___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_GlyphPosition___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_HARD_LIGHT);
  Py_VISIT(traverse_module_state->__pyx_n_s_HBObject);
  Py_VISIT(traverse_module_state->__pyx_n_s_HBObject___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_HBObject___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_HSL_COLOR);
  Py_VISIT(traverse_module_state->__pyx_n_s_HSL_HUE);
  Py_VISIT(traverse_module_state->__pyx_n_s_HSL_LUMINOSITY);
  Py_VISIT(traverse_module_state->__pyx_n_s_HSL_SATURATION);
  Py_VISIT(traverse_module_state->__pyx_n_s_HarfBuzzError);
  Py_VISIT(traverse_module_state->__pyx_n_s_INVALID);
  Py_VISIT(traverse_module_state->__pyx_n_s_INVALID_VALUE);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Input_num_of_subtables);
  Py_VISIT(traverse_module_state->__pyx_n_s_IntEnum);
  Py_VISIT(traverse_module_state->__pyx_n_s_IntFlag);
  Py_VISIT(traverse_module_state->__pyx_n_s_KeyError);
  Py_VISIT(traverse_module_state->__pyx_n_s_LAYOUT_FEATURE_TAG);
  Py_VISIT(traverse_module_state->__pyx_n_s_LAYOUT_SCRIPT_TAG);
  Py_VISIT(traverse_module_state->__pyx_n_s_LIGATURE);
  Py_VISIT(traverse_module_state->__pyx_n_s_LIGHTEN);
  Py_VISIT(traverse_module_state->__pyx_n_s_LOWER_LIMIT_BASELINE_DROP_MIN);
  Py_VISIT(traverse_module_state->__pyx_n_s_LOWER_LIMIT_GAP_MIN);
  Py_VISIT(traverse_module_state->__pyx_n_s_List);
  Py_VISIT(traverse_module_state->__pyx_kp_s_List_Color);
  Py_VISIT(traverse_module_state->__pyx_kp_s_List_OTColorLayer);
  Py_VISIT(traverse_module_state->__pyx_kp_s_List_OTMathGlyphVariant);
  Py_VISIT(traverse_module_state->__pyx_kp_s_List_OTMathKernEntry);
  Py_VISIT(traverse_module_state->__pyx_kp_s_List_Tuple_List_Tuple_int_int_in);
  Py_VISIT(traverse_module_state->__pyx_kp_s_List_bytes);
  Py_VISIT(traverse_module_state->__pyx_kp_s_List_int);
  Py_VISIT(traverse_module_state->__pyx_kp_s_List_str);
  Py_VISIT(traverse_module_state->__pyx_n_s_MARK);
  Py_VISIT(traverse_module_state->__pyx_n_s_MATH_LEADING);
  Py_VISIT(traverse_module_state->__pyx_n_s_MONOTONE_CHARACTERS);
  Py_VISIT(traverse_module_state->__pyx_n_s_MONOTONE_GRAPHEMES);
  Py_VISIT(traverse_module_state->__pyx_n_s_MULTIPLY);
  Py_VISIT(traverse_module_state->__pyx_n_s_Map);
  Py_VISIT(traverse_module_state->__pyx_n_s_MapIter);
  Py_VISIT(traverse_module_state->__pyx_n_s_MapIter___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_MapIter___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Map___copy);
  Py_VISIT(traverse_module_state->__pyx_n_s_Map___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Map___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Map__is_equal);
  Py_VISIT(traverse_module_state->__pyx_n_s_Map__update);
  Py_VISIT(traverse_module_state->__pyx_n_s_Map_clear);
  Py_VISIT(traverse_module_state->__pyx_n_s_Map_copy);
  Py_VISIT(traverse_module_state->__pyx_n_s_Map_get);
  Py_VISIT(traverse_module_state->__pyx_n_s_Map_items);
  Py_VISIT(traverse_module_state->__pyx_n_s_Map_keys);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Map_s);
  Py_VISIT(traverse_module_state->__pyx_n_s_Map_update);
  Py_VISIT(traverse_module_state->__pyx_n_s_Map_values);
  Py_VISIT(traverse_module_state->__pyx_n_s_MemoryError);
  Py_VISIT(traverse_module_state->__pyx_n_s_NAME_ID);
  Py_VISIT(traverse_module_state->__pyx_n_s_NAME_LANG_ID);
  Py_VISIT(traverse_module_state->__pyx_n_s_NAME_LEGACY);
  Py_VISIT(traverse_module_state->__pyx_n_s_NOTDEF_OUTLINE);
  Py_VISIT(traverse_module_state->__pyx_n_s_NO_HINTING);
  Py_VISIT(traverse_module_state->__pyx_n_s_NO_LAYOUT_CLOSURE);
  Py_VISIT(traverse_module_state->__pyx_n_s_NO_PRUNE_UNICODE_RANGES);
  Py_VISIT(traverse_module_state->__pyx_n_s_NO_SUBSET_TABLE_TAG);
  Py_VISIT(traverse_module_state->__pyx_n_s_None);
  Py_VISIT(traverse_module_state->__pyx_n_s_NotImplemented);
  Py_VISIT(traverse_module_state->__pyx_n_s_OTColorLayer);
  Py_VISIT(traverse_module_state->__pyx_n_u_OTColorLayer);
  Py_VISIT(traverse_module_state->__pyx_n_s_OTColorPaletteFlags);
  Py_VISIT(traverse_module_state->__pyx_n_s_OTLayoutGlyphClass);
  Py_VISIT(traverse_module_state->__pyx_n_s_OTMathConstant);
  Py_VISIT(traverse_module_state->__pyx_n_s_OTMathGlyphPart);
  Py_VISIT(traverse_module_state->__pyx_n_u_OTMathGlyphPart);
  Py_VISIT(traverse_module_state->__pyx_n_s_OTMathGlyphPartFlags);
  Py_VISIT(traverse_module_state->__pyx_n_s_OTMathGlyphVariant);
  Py_VISIT(traverse_module_state->__pyx_n_u_OTMathGlyphVariant);
  Py_VISIT(traverse_module_state->__pyx_n_s_OTMathKern);
  Py_VISIT(traverse_module_state->__pyx_n_s_OTMathKernEntry);
  Py_VISIT(traverse_module_state->__pyx_n_u_OTMathKernEntry);
  Py_VISIT(traverse_module_state->__pyx_n_s_OVERBAR_EXTRA_ASCENDER);
  Py_VISIT(traverse_module_state->__pyx_n_s_OVERBAR_RULE_THICKNESS);
  Py_VISIT(traverse_module_state->__pyx_n_s_OVERBAR_VERTICAL_GAP);
  Py_VISIT(traverse_module_state->__pyx_n_s_OVERLAY);
  Py_VISIT(traverse_module_state->__pyx_n_s_PAD);
  Py_VISIT(traverse_module_state->__pyx_n_s_PASSTHROUGH_UNRECOGNIZED);
  Py_VISIT(traverse_module_state->__pyx_n_s_PLUS);
  Py_VISIT(traverse_module_state->__pyx_n_s_PRESERVE_DEFAULT_IGNORABLES);
  Py_VISIT(traverse_module_state->__pyx_n_s_PRODUCE_SAFE_TO_INSERT_TATWEEL);
  Py_VISIT(traverse_module_state->__pyx_n_s_PRODUCE_UNSAFE_TO_CONCAT);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaintCompositeMode);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaintExtend);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaintFuncs);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaintFuncs___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaintFuncs___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaintFuncs_set_color_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaintFuncs_set_color_glyph_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaintFuncs_set_custom_palette_co);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaintFuncs_set_image_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaintFuncs_set_linear_gradient_f);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaintFuncs_set_pop_clip_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaintFuncs_set_pop_group_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaintFuncs_set_pop_transform_fun);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaintFuncs_set_push_clip_glyph_f);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaintFuncs_set_push_clip_rectang);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaintFuncs_set_push_group_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaintFuncs_set_push_transform_fu);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaintFuncs_set_radial_gradient_f);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaintFuncs_set_sweep_gradient_fu);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Pass_draw_state_to_Font_draw_gly);
  Py_VISIT(traverse_module_state->__pyx_n_s_Path);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Pickling_of_struct_members_such);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Pickling_of_struct_members_such_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_RADICAL_DEGREE_BOTTOM_RAISE_PERC);
  Py_VISIT(traverse_module_state->__pyx_n_s_RADICAL_DISPLAY_STYLE_VERTICAL_G);
  Py_VISIT(traverse_module_state->__pyx_n_s_RADICAL_EXTRA_ASCENDER);
  Py_VISIT(traverse_module_state->__pyx_n_s_RADICAL_KERN_AFTER_DEGREE);
  Py_VISIT(traverse_module_state->__pyx_n_s_RADICAL_KERN_BEFORE_DEGREE);
  Py_VISIT(traverse_module_state->__pyx_n_s_RADICAL_RULE_THICKNESS);
  Py_VISIT(traverse_module_state->__pyx_n_s_RADICAL_VERTICAL_GAP);
  Py_VISIT(traverse_module_state->__pyx_n_s_REFLECT);
  Py_VISIT(traverse_module_state->__pyx_n_s_REMOVE_DEFAULT_IGNORABLES);
  Py_VISIT(traverse_module_state->__pyx_n_s_REPEAT);
  Py_VISIT(traverse_module_state->__pyx_n_s_RETAIN_GIDS);
  Py_VISIT(traverse_module_state->__pyx_n_s_RepackerError);
  Py_VISIT(traverse_module_state->__pyx_n_s_RuntimeError);
  Py_VISIT(traverse_module_state->__pyx_n_s_SAFE_TO_INSERT_TATWEEL);
  Py_VISIT(traverse_module_state->__pyx_n_s_SCREEN);
  Py_VISIT(traverse_module_state->__pyx_n_s_SCRIPT_PERCENT_SCALE_DOWN);
  Py_VISIT(traverse_module_state->__pyx_n_s_SCRIPT_SCRIPT_PERCENT_SCALE_DOWN);
  Py_VISIT(traverse_module_state->__pyx_n_s_SET_OVERLAPS_FLAG);
  Py_VISIT(traverse_module_state->__pyx_n_s_SKEWED_FRACTION_HORIZONTAL_GAP);
  Py_VISIT(traverse_module_state->__pyx_n_s_SKEWED_FRACTION_VERTICAL_GAP);
  Py_VISIT(traverse_module_state->__pyx_n_s_SOFT_LIGHT);
  Py_VISIT(traverse_module_state->__pyx_n_s_SPACE_AFTER_SCRIPT);
  Py_VISIT(traverse_module_state->__pyx_n_s_SRC);
  Py_VISIT(traverse_module_state->__pyx_n_s_SRC_ATOP);
  Py_VISIT(traverse_module_state->__pyx_n_s_SRC_IN);
  Py_VISIT(traverse_module_state->__pyx_n_s_SRC_OUT);
  Py_VISIT(traverse_module_state->__pyx_n_s_SRC_OVER);
  Py_VISIT(traverse_module_state->__pyx_n_s_STACK_BOTTOM_DISPLAY_STYLE_SHIFT);
  Py_VISIT(traverse_module_state->__pyx_n_s_STACK_BOTTOM_SHIFT_DOWN);
  Py_VISIT(traverse_module_state->__pyx_n_s_STACK_DISPLAY_STYLE_GAP_MIN);
  Py_VISIT(traverse_module_state->__pyx_n_s_STACK_GAP_MIN);
  Py_VISIT(traverse_module_state->__pyx_n_s_STACK_TOP_DISPLAY_STYLE_SHIFT_UP);
  Py_VISIT(traverse_module_state->__pyx_n_s_STACK_TOP_SHIFT_UP);
  Py_VISIT(traverse_module_state->__pyx_n_s_STRETCH_STACK_BOTTOM_SHIFT_DOWN);
  Py_VISIT(traverse_module_state->__pyx_n_s_STRETCH_STACK_GAP_ABOVE_MIN);
  Py_VISIT(traverse_module_state->__pyx_n_s_STRETCH_STACK_GAP_BELOW_MIN);
  Py_VISIT(traverse_module_state->__pyx_n_s_STRETCH_STACK_TOP_SHIFT_UP);
  Py_VISIT(traverse_module_state->__pyx_n_s_SUBSCRIPT_BASELINE_DROP_MIN);
  Py_VISIT(traverse_module_state->__pyx_n_s_SUBSCRIPT_SHIFT_DOWN);
  Py_VISIT(traverse_module_state->__pyx_n_s_SUBSCRIPT_TOP_MAX);
  Py_VISIT(traverse_module_state->__pyx_n_s_SUB_SUPERSCRIPT_GAP_MIN);
  Py_VISIT(traverse_module_state->__pyx_n_s_SUPERSCRIPT_BASELINE_DROP_MAX);
  Py_VISIT(traverse_module_state->__pyx_n_s_SUPERSCRIPT_BOTTOM_MAX_WITH_SUBS);
  Py_VISIT(traverse_module_state->__pyx_n_s_SUPERSCRIPT_BOTTOM_MIN);
  Py_VISIT(traverse_module_state->__pyx_n_s_SUPERSCRIPT_SHIFT_UP);
  Py_VISIT(traverse_module_state->__pyx_n_s_SUPERSCRIPT_SHIFT_UP_CRAMPED);
  Py_VISIT(traverse_module_state->__pyx_n_s_Sequence);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Set);
  Py_VISIT(traverse_module_state->__pyx_n_s_SetIter);
  Py_VISIT(traverse_module_state->__pyx_n_s_SetIter___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_SetIter___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Set_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_Set___copy);
  Py_VISIT(traverse_module_state->__pyx_n_s_Set___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Set___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_Set__is_equal);
  Py_VISIT(traverse_module_state->__pyx_n_s_Set__set);
  Py_VISIT(traverse_module_state->__pyx_n_s_Set__update);
  Py_VISIT(traverse_module_state->__pyx_n_s_Set_add);
  Py_VISIT(traverse_module_state->__pyx_n_s_Set_add_range);
  Py_VISIT(traverse_module_state->__pyx_n_s_Set_clear);
  Py_VISIT(traverse_module_state->__pyx_n_s_Set_copy);
  Py_VISIT(traverse_module_state->__pyx_n_s_Set_del_range);
  Py_VISIT(traverse_module_state->__pyx_n_s_Set_difference_update);
  Py_VISIT(traverse_module_state->__pyx_n_s_Set_discard);
  Py_VISIT(traverse_module_state->__pyx_n_s_Set_intersection_update);
  Py_VISIT(traverse_module_state->__pyx_n_s_Set_invert);
  Py_VISIT(traverse_module_state->__pyx_n_s_Set_is_inverted);
  Py_VISIT(traverse_module_state->__pyx_n_s_Set_issubset);
  Py_VISIT(traverse_module_state->__pyx_n_s_Set_issuperset);
  Py_VISIT(traverse_module_state->__pyx_n_s_Set_remove);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Set_s);
  Py_VISIT(traverse_module_state->__pyx_n_s_Set_set);
  Py_VISIT(traverse_module_state->__pyx_n_s_Set_symmetric_difference_update);
  Py_VISIT(traverse_module_state->__pyx_n_s_Set_update);
  Py_VISIT(traverse_module_state->__pyx_n_s_StopIteration);
  Py_VISIT(traverse_module_state->__pyx_n_s_SubsetFlags);
  Py_VISIT(traverse_module_state->__pyx_n_s_SubsetInput);
  Py_VISIT(traverse_module_state->__pyx_n_s_SubsetInputSets);
  Py_VISIT(traverse_module_state->__pyx_n_s_SubsetInput___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_SubsetInput___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_SubsetInput_keep_everything);
  Py_VISIT(traverse_module_state->__pyx_n_s_SubsetInput_pin_axis_location);
  Py_VISIT(traverse_module_state->__pyx_n_s_SubsetInput_pin_axis_to_default);
  Py_VISIT(traverse_module_state->__pyx_n_s_SubsetInput_sets);
  Py_VISIT(traverse_module_state->__pyx_n_s_SubsetInput_subset);
  Py_VISIT(traverse_module_state->__pyx_n_s_SubsetPlan);
  Py_VISIT(traverse_module_state->__pyx_n_s_SubsetPlan___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_SubsetPlan___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_SubsetPlan_execute);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Subsetting_failed);
  Py_VISIT(traverse_module_state->__pyx_n_s_TOP_LEFT);
  Py_VISIT(traverse_module_state->__pyx_n_s_TOP_RIGHT);
  Py_VISIT(traverse_module_state->__pyx_n_s_Tuple);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Tuple_List_OTMathGlyphPart_int);
  Py_VISIT(traverse_module_state->__pyx_n_s_TypeError);
  Py_VISIT(traverse_module_state->__pyx_n_s_UNCLASSIFIED);
  Py_VISIT(traverse_module_state->__pyx_n_s_UNDERBAR_EXTRA_DESCENDER);
  Py_VISIT(traverse_module_state->__pyx_n_s_UNDERBAR_RULE_THICKNESS);
  Py_VISIT(traverse_module_state->__pyx_n_s_UNDERBAR_VERTICAL_GAP);
  Py_VISIT(traverse_module_state->__pyx_n_s_UNICODE);
  Py_VISIT(traverse_module_state->__pyx_n_s_UNSAFE_TO_BREAK);
  Py_VISIT(traverse_module_state->__pyx_n_s_UNSAFE_TO_CONCAT);
  Py_VISIT(traverse_module_state->__pyx_n_s_UPPER_LIMIT_BASELINE_RISE_MIN);
  Py_VISIT(traverse_module_state->__pyx_n_s_UPPER_LIMIT_GAP_MIN);
  Py_VISIT(traverse_module_state->__pyx_n_s_USABLE_WITH_DARK_BACKGROUND);
  Py_VISIT(traverse_module_state->__pyx_n_s_USABLE_WITH_LIGHT_BACKGROUND);
  Py_VISIT(traverse_module_state->__pyx_n_s_Union);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Union_str_Path);
  Py_VISIT(traverse_module_state->__pyx_n_s_VERIFY);
  Py_VISIT(traverse_module_state->__pyx_n_s_ValueError);
  Py_VISIT(traverse_module_state->__pyx_n_s_XOR);
  Py_VISIT(traverse_module_state->__pyx_kp_u__10);
  Py_VISIT(traverse_module_state->__pyx_kp_b__15);
  Py_VISIT(traverse_module_state->__pyx_kp_u__15);
  Py_VISIT(traverse_module_state->__pyx_kp_u__16);
  Py_VISIT(traverse_module_state->__pyx_kp_u__19);
  Py_VISIT(traverse_module_state->__pyx_kp_u__21);
  Py_VISIT(traverse_module_state->__pyx_n_s__22);
  Py_VISIT(traverse_module_state->__pyx_kp_u__23);
  Py_VISIT(traverse_module_state->__pyx_n_s__287);
  Py_VISIT(traverse_module_state->__pyx_n_s_a);
  Py_VISIT(traverse_module_state->__pyx_n_s_add);
  Py_VISIT(traverse_module_state->__pyx_n_s_add_codepoints);
  Py_VISIT(traverse_module_state->__pyx_n_s_add_range);
  Py_VISIT(traverse_module_state->__pyx_n_s_add_str);
  Py_VISIT(traverse_module_state->__pyx_n_s_add_utf8);
  Py_VISIT(traverse_module_state->__pyx_n_u_advance);
  Py_VISIT(traverse_module_state->__pyx_n_s_alpha);
  Py_VISIT(traverse_module_state->__pyx_n_u_alpha);
  Py_VISIT(traverse_module_state->__pyx_n_s_alternate_count);
  Py_VISIT(traverse_module_state->__pyx_n_s_alternate_glyphs);
  Py_VISIT(traverse_module_state->__pyx_n_s_alternates);
  Py_VISIT(traverse_module_state->__pyx_n_s_args);
  Py_VISIT(traverse_module_state->__pyx_n_s_ascender);
  Py_VISIT(traverse_module_state->__pyx_n_u_ascender);
  Py_VISIT(traverse_module_state->__pyx_n_s_assembly);
  Py_VISIT(traverse_module_state->__pyx_n_s_assembly_array);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncio_coroutines);
  Py_VISIT(traverse_module_state->__pyx_n_s_b);
  Py_VISIT(traverse_module_state->__pyx_n_s_baseline_tag);
  Py_VISIT(traverse_module_state->__pyx_n_s_blob);
  Py_VISIT(traverse_module_state->__pyx_n_s_blob_data);
  Py_VISIT(traverse_module_state->__pyx_n_s_blob_length);
  Py_VISIT(traverse_module_state->__pyx_n_s_blue);
  Py_VISIT(traverse_module_state->__pyx_n_u_blue);
  Py_VISIT(traverse_module_state->__pyx_n_s_bool);
  Py_VISIT(traverse_module_state->__pyx_n_s_buffer);
  Py_VISIT(traverse_module_state->__pyx_n_s_bytes);
  Py_VISIT(traverse_module_state->__pyx_n_s_c);
  Py_VISIT(traverse_module_state->__pyx_n_s_c_coords);
  Py_VISIT(traverse_module_state->__pyx_n_s_c_foreground);
  Py_VISIT(traverse_module_state->__pyx_n_s_c_shapers);
  Py_VISIT(traverse_module_state->__pyx_n_s_callback);
  Py_VISIT(traverse_module_state->__pyx_n_s_clear);
  Py_VISIT(traverse_module_state->__pyx_n_s_clear_contents);
  Py_VISIT(traverse_module_state->__pyx_n_s_cline_in_traceback);
  Py_VISIT(traverse_module_state->__pyx_n_s_close);
  Py_VISIT(traverse_module_state->__pyx_n_s_closePath);
  Py_VISIT(traverse_module_state->__pyx_n_s_cls);
  Py_VISIT(traverse_module_state->__pyx_n_s_codepoints);
  Py_VISIT(traverse_module_state->__pyx_n_s_collections);
  Py_VISIT(traverse_module_state->__pyx_n_u_color);
  Py_VISIT(traverse_module_state->__pyx_n_s_color_count);
  Py_VISIT(traverse_module_state->__pyx_n_s_color_index);
  Py_VISIT(traverse_module_state->__pyx_n_u_color_index);
  Py_VISIT(traverse_module_state->__pyx_n_s_colors);
  Py_VISIT(traverse_module_state->__pyx_n_s_constant);
  Py_VISIT(traverse_module_state->__pyx_n_s_coords);
  Py_VISIT(traverse_module_state->__pyx_n_s_coords_2dot14);
  Py_VISIT(traverse_module_state->__pyx_n_s_copy);
  Py_VISIT(traverse_module_state->__pyx_n_s_copy_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_correction_height);
  Py_VISIT(traverse_module_state->__pyx_n_s_count);
  Py_VISIT(traverse_module_state->__pyx_n_s_create);
  Py_VISIT(traverse_module_state->__pyx_n_s_create_for_tables);
  Py_VISIT(traverse_module_state->__pyx_n_s_cstr);
  Py_VISIT(traverse_module_state->__pyx_n_s_curveTo);
  Py_VISIT(traverse_module_state->__pyx_n_s_data);
  Py_VISIT(traverse_module_state->__pyx_n_s_del_range);
  Py_VISIT(traverse_module_state->__pyx_n_s_descender);
  Py_VISIT(traverse_module_state->__pyx_n_u_descender);
  Py_VISIT(traverse_module_state->__pyx_n_s_dict);
  Py_VISIT(traverse_module_state->__pyx_n_s_difference_update);
  Py_VISIT(traverse_module_state->__pyx_n_s_direction);
  Py_VISIT(traverse_module_state->__pyx_kp_u_disable);
  Py_VISIT(traverse_module_state->__pyx_n_s_discard);
  Py_VISIT(traverse_module_state->__pyx_n_s_doc);
  Py_VISIT(traverse_module_state->__pyx_n_s_draw_data);
  Py_VISIT(traverse_module_state->__pyx_n_s_draw_funcs);
  Py_VISIT(traverse_module_state->__pyx_n_s_draw_glyph);
  Py_VISIT(traverse_module_state->__pyx_kp_u_draw_glyph_is_deprecated_use_Fon);
  Py_VISIT(traverse_module_state->__pyx_n_s_draw_glyph_with_pen);
  Py_VISIT(traverse_module_state->__pyx_n_s_draw_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_draw_state_p);
  Py_VISIT(traverse_module_state->__pyx_kp_u_enable);
  Py_VISIT(traverse_module_state->__pyx_n_s_encode);
  Py_VISIT(traverse_module_state->__pyx_n_s_end);
  Py_VISIT(traverse_module_state->__pyx_n_u_end_connector_length);
  Py_VISIT(traverse_module_state->__pyx_n_s_enum);
  Py_VISIT(traverse_module_state->__pyx_n_s_enumerate);
  Py_VISIT(traverse_module_state->__pyx_n_s_execute);
  Py_VISIT(traverse_module_state->__pyx_n_s_extents);
  Py_VISIT(traverse_module_state->__pyx_n_s_face);
  Py_VISIT(traverse_module_state->__pyx_n_s_face_or_font);
  Py_VISIT(traverse_module_state->__pyx_n_s_feat);
  Py_VISIT(traverse_module_state->__pyx_n_s_feature_count);
  Py_VISIT(traverse_module_state->__pyx_n_s_feature_tags);
  Py_VISIT(traverse_module_state->__pyx_n_s_features);
  Py_VISIT(traverse_module_state->__pyx_n_s_filename);
  Py_VISIT(traverse_module_state->__pyx_n_s_first);
  Py_VISIT(traverse_module_state->__pyx_n_u_flags);
  Py_VISIT(traverse_module_state->__pyx_n_s_float);
  Py_VISIT(traverse_module_state->__pyx_n_s_font);
  Py_VISIT(traverse_module_state->__pyx_n_s_foreground);
  Py_VISIT(traverse_module_state->__pyx_n_s_from_file_path);
  Py_VISIT(traverse_module_state->__pyx_n_s_from_int);
  Py_VISIT(traverse_module_state->__pyx_n_s_fsencode);
  Py_VISIT(traverse_module_state->__pyx_n_u_full_advance);
  Py_VISIT(traverse_module_state->__pyx_n_s_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_func_p);
  Py_VISIT(traverse_module_state->__pyx_n_s_funcs);
  Py_VISIT(traverse_module_state->__pyx_n_s_g);
  Py_VISIT(traverse_module_state->__pyx_kp_u_gc);
  Py_VISIT(traverse_module_state->__pyx_n_s_genexpr);
  Py_VISIT(traverse_module_state->__pyx_n_s_get);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_font_extents);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_glyph_extents);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_glyph_from_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_glyph_h_advance);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_glyph_h_origin);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_glyph_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_glyph_shape);
  Py_VISIT(traverse_module_state->__pyx_kp_u_get_glyph_shape_is_deprecated_us);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_glyph_v_advance);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_glyph_v_origin);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_nominal_glyph);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_var_coords_design);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_var_coords_normalized);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_variation_glyph);
  Py_VISIT(traverse_module_state->__pyx_n_s_getstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_gid);
  Py_VISIT(traverse_module_state->__pyx_n_s_glyph);
  Py_VISIT(traverse_module_state->__pyx_n_u_glyph);
  Py_VISIT(traverse_module_state->__pyx_n_s_glyph_from_string);
  Py_VISIT(traverse_module_state->__pyx_n_s_glyph_to_string);
  Py_VISIT(traverse_module_state->__pyx_n_s_graphnodes);
  Py_VISIT(traverse_module_state->__pyx_n_s_green);
  Py_VISIT(traverse_module_state->__pyx_n_u_green);
  Py_VISIT(traverse_module_state->__pyx_n_s_guess_segment_properties);
  Py_VISIT(traverse_module_state->__pyx_n_u_hang);
  Py_VISIT(traverse_module_state->__pyx_n_s_hb_baseline_tag);
  Py_VISIT(traverse_module_state->__pyx_n_s_hb_codepoint_t);
  Py_VISIT(traverse_module_state->__pyx_n_s_hb_codepoints);
  Py_VISIT(traverse_module_state->__pyx_n_s_hb_direction);
  Py_VISIT(traverse_module_state->__pyx_n_s_hb_features);
  Py_VISIT(traverse_module_state->__pyx_n_s_hb_language);
  Py_VISIT(traverse_module_state->__pyx_n_s_hb_language_tag);
  Py_VISIT(traverse_module_state->__pyx_n_s_hb_position);
  Py_VISIT(traverse_module_state->__pyx_n_s_hb_script);
  Py_VISIT(traverse_module_state->__pyx_n_s_hb_script_tag);
  Py_VISIT(traverse_module_state->__pyx_n_s_hb_tag);
  Py_VISIT(traverse_module_state->__pyx_n_s_hb_variations);
  Py_VISIT(traverse_module_state->__pyx_n_s_head);
  Py_VISIT(traverse_module_state->__pyx_n_u_height);
  Py_VISIT(traverse_module_state->__pyx_n_s_i);
  Py_VISIT(traverse_module_state->__pyx_n_u_icfb);
  Py_VISIT(traverse_module_state->__pyx_n_u_icft);
  Py_VISIT(traverse_module_state->__pyx_n_u_ideo);
  Py_VISIT(traverse_module_state->__pyx_n_u_idtp);
  Py_VISIT(traverse_module_state->__pyx_n_s_import);
  Py_VISIT(traverse_module_state->__pyx_n_s_index);
  Py_VISIT(traverse_module_state->__pyx_kp_u_index_out_of_range_for_feature_a);
  Py_VISIT(traverse_module_state->__pyx_n_s_init);
  Py_VISIT(traverse_module_state->__pyx_n_s_init_subclass);
  Py_VISIT(traverse_module_state->__pyx_n_s_initializing);
  Py_VISIT(traverse_module_state->__pyx_n_s_input);
  Py_VISIT(traverse_module_state->__pyx_n_s_inst);
  Py_VISIT(traverse_module_state->__pyx_n_s_int);
  Py_VISIT(traverse_module_state->__pyx_kp_s_int_None);
  Py_VISIT(traverse_module_state->__pyx_n_s_intersection_update);
  Py_VISIT(traverse_module_state->__pyx_kp_u_invalid_baseline_tag);
  Py_VISIT(traverse_module_state->__pyx_kp_u_invalid_constant);
  Py_VISIT(traverse_module_state->__pyx_kp_u_invalid_kern);
  Py_VISIT(traverse_module_state->__pyx_n_s_invert);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_coroutine);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_equal);
  Py_VISIT(traverse_module_state->__pyx_n_u_is_foreground);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_inverted);
  Py_VISIT(traverse_module_state->__pyx_kp_u_isenabled);
  Py_VISIT(traverse_module_state->__pyx_n_s_issubset);
  Py_VISIT(traverse_module_state->__pyx_n_s_issuperset);
  Py_VISIT(traverse_module_state->__pyx_n_s_italics_correction);
  Py_VISIT(traverse_module_state->__pyx_n_s_item_length);
  Py_VISIT(traverse_module_state->__pyx_n_s_item_offset);
  Py_VISIT(traverse_module_state->__pyx_n_s_items);
  Py_VISIT(traverse_module_state->__pyx_n_s_join);
  Py_VISIT(traverse_module_state->__pyx_n_s_k);
  Py_VISIT(traverse_module_state->__pyx_n_s_keep_everything);
  Py_VISIT(traverse_module_state->__pyx_n_s_kern);
  Py_VISIT(traverse_module_state->__pyx_n_u_kern_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_kerns);
  Py_VISIT(traverse_module_state->__pyx_n_s_kerns_array);
  Py_VISIT(traverse_module_state->__pyx_n_s_keys);
  Py_VISIT(traverse_module_state->__pyx_n_s_keys_locals_genexpr);
  Py_VISIT(traverse_module_state->__pyx_n_s_kind);
  Py_VISIT(traverse_module_state->__pyx_n_s_language_count);
  Py_VISIT(traverse_module_state->__pyx_n_s_language_index);
  Py_VISIT(traverse_module_state->__pyx_n_s_language_tag);
  Py_VISIT(traverse_module_state->__pyx_n_s_language_tags);
  Py_VISIT(traverse_module_state->__pyx_n_s_larger_set);
  Py_VISIT(traverse_module_state->__pyx_n_s_last);
  Py_VISIT(traverse_module_state->__pyx_n_s_layer_count);
  Py_VISIT(traverse_module_state->__pyx_n_s_layers);
  Py_VISIT(traverse_module_state->__pyx_n_s_length);
  Py_VISIT(traverse_module_state->__pyx_n_s_lineTo);
  Py_VISIT(traverse_module_state->__pyx_n_s_line_gap);
  Py_VISIT(traverse_module_state->__pyx_n_u_line_gap);
  Py_VISIT(traverse_module_state->__pyx_n_s_lookup_index);
  Py_VISIT(traverse_module_state->__pyx_n_s_m);
  Py_VISIT(traverse_module_state->__pyx_n_s_main);
  Py_VISIT(traverse_module_state->__pyx_n_u_math);
  Py_VISIT(traverse_module_state->__pyx_n_u_max_correction_height);
  Py_VISIT(traverse_module_state->__pyx_n_s_metaclass);
  Py_VISIT(traverse_module_state->__pyx_n_s_methods);
  Py_VISIT(traverse_module_state->__pyx_n_s_module);
  Py_VISIT(traverse_module_state->__pyx_n_s_moveTo);
  Py_VISIT(traverse_module_state->__pyx_n_s_mro_entries);
  Py_VISIT(traverse_module_state->__pyx_n_s_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_name_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_name_id);
  Py_VISIT(traverse_module_state->__pyx_n_s_namedtuple);
  Py_VISIT(traverse_module_state->__pyx_n_s_new_face);
  Py_VISIT(traverse_module_state->__pyx_kp_s_no_default___reduce___due_to_non);
  Py_VISIT(traverse_module_state->__pyx_n_s_node);
  Py_VISIT(traverse_module_state->__pyx_n_s_num_nodes);
  Py_VISIT(traverse_module_state->__pyx_kp_u_num_of_graph_nodes);
  Py_VISIT(traverse_module_state->__pyx_n_s_obj_list);
  Py_VISIT(traverse_module_state->__pyx_n_s_object);
  Py_VISIT(traverse_module_state->__pyx_n_u_offset);
  Py_VISIT(traverse_module_state->__pyx_n_s_os);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_color_glyph_get_layers);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_color_glyph_get_png);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_color_glyph_get_svg);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_color_glyph_has_paint);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_color_has_layers);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_color_has_paint);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_color_has_palettes);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_color_has_png);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_color_has_svg);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_color_palette_color_get_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_color_palette_get_colors);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_color_palette_get_count);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_color_palette_get_flags);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_color_palette_get_name_id);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_font_set_funcs);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_layout_get_baseline);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_layout_get_glyph_class);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_layout_has_glyph_classes);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_layout_has_positioning);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_layout_has_substitution);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_layout_language_get_feature_t);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_layout_lookup_get_glyph_alter);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_layout_script_get_language_ta);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_layout_table_get_script_tags);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_math_get_constant);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_math_get_glyph_assembly);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_math_get_glyph_italics_correc);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_math_get_glyph_kerning);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_math_get_glyph_kernings);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_math_get_glyph_top_accent_att);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_math_get_glyph_variants);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_math_get_min_connector_overla);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_math_has_data);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_math_is_glyph_extended_shape);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_tag_to_language);
  Py_VISIT(traverse_module_state->__pyx_n_s_ot_tag_to_script);
  Py_VISIT(traverse_module_state->__pyx_n_s_other);
  Py_VISIT(traverse_module_state->__pyx_n_s_p);
  Py_VISIT(traverse_module_state->__pyx_n_s_packed);
  Py_VISIT(traverse_module_state->__pyx_n_s_packed_blob);
  Py_VISIT(traverse_module_state->__pyx_n_s_paint_funcs);
  Py_VISIT(traverse_module_state->__pyx_n_s_paint_glyph);
  Py_VISIT(traverse_module_state->__pyx_n_s_paint_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_paint_state_p);
  Py_VISIT(traverse_module_state->__pyx_n_s_palette_index);
  Py_VISIT(traverse_module_state->__pyx_n_s_pathlib);
  Py_VISIT(traverse_module_state->__pyx_n_s_pen);
  Py_VISIT(traverse_module_state->__pyx_n_s_pin_axis_location);
  Py_VISIT(traverse_module_state->__pyx_n_s_pin_axis_to_default);
  Py_VISIT(traverse_module_state->__pyx_n_s_prepare);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_vtable);
  Py_VISIT(traverse_module_state->__pyx_n_s_qCurveTo);
  Py_VISIT(traverse_module_state->__pyx_n_s_qualname);
  Py_VISIT(traverse_module_state->__pyx_n_s_r);
  Py_VISIT(traverse_module_state->__pyx_n_s_range);
  Py_VISIT(traverse_module_state->__pyx_n_s_red);
  Py_VISIT(traverse_module_state->__pyx_n_u_red);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_ex);
  Py_VISIT(traverse_module_state->__pyx_n_s_remove);
  Py_VISIT(traverse_module_state->__pyx_n_s_repack);
  Py_VISIT(traverse_module_state->__pyx_n_s_repack_with_tag);
  Py_VISIT(traverse_module_state->__pyx_n_s_repr___locals_genexpr);
  Py_VISIT(traverse_module_state->__pyx_n_s_reset);
  Py_VISIT(traverse_module_state->__pyx_n_s_ret);
  Py_VISIT(traverse_module_state->__pyx_n_s_return);
  Py_VISIT(traverse_module_state->__pyx_n_u_romn);
  Py_VISIT(traverse_module_state->__pyx_n_s_round);
  Py_VISIT(traverse_module_state->__pyx_n_s_s);
  Py_VISIT(traverse_module_state->__pyx_n_s_script_count);
  Py_VISIT(traverse_module_state->__pyx_n_s_script_index);
  Py_VISIT(traverse_module_state->__pyx_n_s_script_tag);
  Py_VISIT(traverse_module_state->__pyx_n_s_script_tags);
  Py_VISIT(traverse_module_state->__pyx_n_s_self);
  Py_VISIT(traverse_module_state->__pyx_n_s_send);
  Py_VISIT(traverse_module_state->__pyx_n_s_set);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_close_path_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_color_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_color_glyph_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_cubic_to_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_custom_palette_color_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_font_h_extents_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_font_v_extents_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_glyph_h_advance_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_glyph_name_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_glyph_v_advance_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_glyph_v_origin_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_image_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_language_from_ot_tag);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_line_to_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_linear_gradient_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_message_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_move_to_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_nominal_glyph_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_pop_clip_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_pop_group_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_pop_transform_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_push_clip_glyph_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_push_clip_rectangle_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_push_group_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_push_transform_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_quadratic_to_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_radial_gradient_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_script_from_ot_tag);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_sweep_gradient_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_var_coords_design);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_var_coords_normalized);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_variation);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_variation_glyph_func);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_variations);
  Py_VISIT(traverse_module_state->__pyx_n_s_sets);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_shape);
  Py_VISIT(traverse_module_state->__pyx_n_s_shaper);
  Py_VISIT(traverse_module_state->__pyx_n_s_shapers);
  Py_VISIT(traverse_module_state->__pyx_n_s_size);
  Py_VISIT(traverse_module_state->__pyx_n_s_smaller_set);
  Py_VISIT(traverse_module_state->__pyx_n_s_source);
  Py_VISIT(traverse_module_state->__pyx_n_s_spec);
  Py_VISIT(traverse_module_state->__pyx_kp_s_src_uharfbuzz__harfbuzz_pyx);
  Py_VISIT(traverse_module_state->__pyx_n_s_start);
  Py_VISIT(traverse_module_state->__pyx_n_u_start_connector_length);
  Py_VISIT(traverse_module_state->__pyx_n_s_start_offset);
  Py_VISIT(traverse_module_state->__pyx_n_s_staticmethod);
  Py_VISIT(traverse_module_state->__pyx_n_s_str);
  Py_VISIT(traverse_module_state->__pyx_n_s_string);
  Py_VISIT(traverse_module_state->__pyx_kp_s_stringsource);
  Py_VISIT(traverse_module_state->__pyx_n_s_subset);
  Py_VISIT(traverse_module_state->__pyx_n_s_subset_preprocess);
  Py_VISIT(traverse_module_state->__pyx_n_s_subtables);
  Py_VISIT(traverse_module_state->__pyx_n_s_success);
  Py_VISIT(traverse_module_state->__pyx_n_s_super);
  Py_VISIT(traverse_module_state->__pyx_n_s_symmetric_difference_update);
  Py_VISIT(traverse_module_state->__pyx_n_s_table_bytes);
  Py_VISIT(traverse_module_state->__pyx_n_s_table_data);
  Py_VISIT(traverse_module_state->__pyx_n_s_tag);
  Py_VISIT(traverse_module_state->__pyx_n_s_tag_packed);
  Py_VISIT(traverse_module_state->__pyx_n_s_tags);
  Py_VISIT(traverse_module_state->__pyx_n_s_tail);
  Py_VISIT(traverse_module_state->__pyx_n_s_test);
  Py_VISIT(traverse_module_state->__pyx_n_s_text);
  Py_VISIT(traverse_module_state->__pyx_n_s_throw);
  Py_VISIT(traverse_module_state->__pyx_n_s_to_int);
  Py_VISIT(traverse_module_state->__pyx_n_s_typing);
  Py_VISIT(traverse_module_state->__pyx_n_s_uharfbuzz__harfbuzz);
  Py_VISIT(traverse_module_state->__pyx_n_s_unicode);
  Py_VISIT(traverse_module_state->__pyx_n_s_update);
  Py_VISIT(traverse_module_state->__pyx_n_s_update_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_user_data);
  Py_VISIT(traverse_module_state->__pyx_n_s_user_data_p);
  Py_VISIT(traverse_module_state->__pyx_n_s_v);
  Py_VISIT(traverse_module_state->__pyx_n_s_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_values);
  Py_VISIT(traverse_module_state->__pyx_n_s_values_locals_genexpr);
  Py_VISIT(traverse_module_state->__pyx_n_s_variants);
  Py_VISIT(traverse_module_state->__pyx_n_s_variants_array);
  Py_VISIT(traverse_module_state->__pyx_n_s_variation);
  Py_VISIT(traverse_module_state->__pyx_n_s_variation_selector);
  Py_VISIT(traverse_module_state->__pyx_n_s_variation_unicodes);
  Py_VISIT(traverse_module_state->__pyx_n_s_variations);
  Py_VISIT(traverse_module_state->__pyx_n_s_version_string);
  Py_VISIT(traverse_module_state->__pyx_n_s_warn);
  Py_VISIT(traverse_module_state->__pyx_n_s_warnings);
  Py_VISIT(traverse_module_state->__pyx_n_u_width);
  Py_VISIT(traverse_module_state->__pyx_n_s_x);
  Py_VISIT(traverse_module_state->__pyx_n_u_x_bearing);
  Py_VISIT(traverse_module_state->__pyx_n_s_y);
  Py_VISIT(traverse_module_state->__pyx_n_u_y_bearing);
  Py_VISIT(traverse_module_state->__pyx_int_0);
  Py_VISIT(traverse_module_state->__pyx_int_1);
  Py_VISIT(traverse_module_state->__pyx_int_16384);
  Py_VISIT(traverse_module_state->__pyx_int_65535);
  Py_VISIT(traverse_module_state->__pyx_int_neg_1);
  Py_VISIT(traverse_module_state->__pyx_k_);
  Py_VISIT(traverse_module_state->__pyx_k__2);
  Py_VISIT(traverse_module_state->__pyx_k__3);
  Py_VISIT(traverse_module_state->__pyx_k__4);
  Py_VISIT(traverse_module_state->__pyx_k__5);
  Py_VISIT(traverse_module_state->__pyx_k__6);
  Py_VISIT(traverse_module_state->__pyx_k__8);
  Py_VISIT(traverse_module_state->__pyx_k__18);
  Py_VISIT(traverse_module_state->__pyx_k__20);
  Py_VISIT(traverse_module_state->__pyx_tuple__7);
  Py_VISIT(traverse_module_state->__pyx_tuple__9);
  Py_VISIT(traverse_module_state->__pyx_tuple__11);
  Py_VISIT(traverse_module_state->__pyx_tuple__12);
  Py_VISIT(traverse_module_state->__pyx_tuple__13);
  Py_VISIT(traverse_module_state->__pyx_tuple__14);
  Py_VISIT(traverse_module_state->__pyx_tuple__17);
  Py_VISIT(traverse_module_state->__pyx_tuple__24);
  Py_VISIT(traverse_module_state->__pyx_tuple__26);
  Py_VISIT(traverse_module_state->__pyx_tuple__28);
  Py_VISIT(traverse_module_state->__pyx_tuple__32);
  Py_VISIT(traverse_module_state->__pyx_tuple__36);
  Py_VISIT(traverse_module_state->__pyx_tuple__39);
  Py_VISIT(traverse_module_state->__pyx_tuple__41);
  Py_VISIT(traverse_module_state->__pyx_tuple__43);
  Py_VISIT(traverse_module_state->__pyx_tuple__46);
  Py_VISIT(traverse_module_state->__pyx_tuple__50);
  Py_VISIT(traverse_module_state->__pyx_tuple__54);
  Py_VISIT(traverse_module_state->__pyx_tuple__56);
  Py_VISIT(traverse_module_state->__pyx_tuple__57);
  Py_VISIT(traverse_module_state->__pyx_tuple__59);
  Py_VISIT(traverse_module_state->__pyx_tuple__63);
  Py_VISIT(traverse_module_state->__pyx_tuple__65);
  Py_VISIT(traverse_module_state->__pyx_tuple__67);
  Py_VISIT(traverse_module_state->__pyx_tuple__69);
  Py_VISIT(traverse_module_state->__pyx_tuple__71);
  Py_VISIT(traverse_module_state->__pyx_tuple__73);
  Py_VISIT(traverse_module_state->__pyx_tuple__75);
  Py_VISIT(traverse_module_state->__pyx_tuple__78);
  Py_VISIT(traverse_module_state->__pyx_tuple__81);
  Py_VISIT(traverse_module_state->__pyx_tuple__83);
  Py_VISIT(traverse_module_state->__pyx_tuple__85);
  Py_VISIT(traverse_module_state->__pyx_tuple__87);
  Py_VISIT(traverse_module_state->__pyx_tuple__89);
  Py_VISIT(traverse_module_state->__pyx_tuple__92);
  Py_VISIT(traverse_module_state->__pyx_tuple__94);
  Py_VISIT(traverse_module_state->__pyx_tuple__96);
  Py_VISIT(traverse_module_state->__pyx_tuple__98);
  Py_VISIT(traverse_module_state->__pyx_tuple__100);
  Py_VISIT(traverse_module_state->__pyx_tuple__102);
  Py_VISIT(traverse_module_state->__pyx_tuple__107);
  Py_VISIT(traverse_module_state->__pyx_tuple__118);
  Py_VISIT(traverse_module_state->__pyx_tuple__120);
  Py_VISIT(traverse_module_state->__pyx_tuple__121);
  Py_VISIT(traverse_module_state->__pyx_tuple__123);
  Py_VISIT(traverse_module_state->__pyx_tuple__125);
  Py_VISIT(traverse_module_state->__pyx_tuple__127);
  Py_VISIT(traverse_module_state->__pyx_tuple__129);
  Py_VISIT(traverse_module_state->__pyx_tuple__131);
  Py_VISIT(traverse_module_state->__pyx_tuple__133);
  Py_VISIT(traverse_module_state->__pyx_tuple__135);
  Py_VISIT(traverse_module_state->__pyx_tuple__139);
  Py_VISIT(traverse_module_state->__pyx_tuple__143);
  Py_VISIT(traverse_module_state->__pyx_tuple__145);
  Py_VISIT(traverse_module_state->__pyx_tuple__147);
  Py_VISIT(traverse_module_state->__pyx_tuple__149);
  Py_VISIT(traverse_module_state->__pyx_tuple__152);
  Py_VISIT(traverse_module_state->__pyx_tuple__157);
  Py_VISIT(traverse_module_state->__pyx_tuple__160);
  Py_VISIT(traverse_module_state->__pyx_tuple__163);
  Py_VISIT(traverse_module_state->__pyx_tuple__165);
  Py_VISIT(traverse_module_state->__pyx_tuple__169);
  Py_VISIT(traverse_module_state->__pyx_tuple__171);
  Py_VISIT(traverse_module_state->__pyx_tuple__173);
  Py_VISIT(traverse_module_state->__pyx_tuple__175);
  Py_VISIT(traverse_module_state->__pyx_tuple__177);
  Py_VISIT(traverse_module_state->__pyx_tuple__179);
  Py_VISIT(traverse_module_state->__pyx_tuple__182);
  Py_VISIT(traverse_module_state->__pyx_tuple__186);
  Py_VISIT(traverse_module_state->__pyx_tuple__203);
  Py_VISIT(traverse_module_state->__pyx_tuple__205);
  Py_VISIT(traverse_module_state->__pyx_tuple__207);
  Py_VISIT(traverse_module_state->__pyx_tuple__217);
  Py_VISIT(traverse_module_state->__pyx_tuple__219);
  Py_VISIT(traverse_module_state->__pyx_tuple__221);
  Py_VISIT(traverse_module_state->__pyx_tuple__223);
  Py_VISIT(traverse_module_state->__pyx_tuple__225);
  Py_VISIT(traverse_module_state->__pyx_tuple__228);
  Py_VISIT(traverse_module_state->__pyx_tuple__230);
  Py_VISIT(traverse_module_state->__pyx_tuple__232);
  Py_VISIT(traverse_module_state->__pyx_tuple__236);
  Py_VISIT(traverse_module_state->__pyx_tuple__240);
  Py_VISIT(traverse_module_state->__pyx_tuple__247);
  Py_VISIT(traverse_module_state->__pyx_tuple__252);
  Py_VISIT(traverse_module_state->__pyx_tuple__254);
  Py_VISIT(traverse_module_state->__pyx_tuple__256);
  Py_VISIT(traverse_module_state->__pyx_tuple__259);
  Py_VISIT(traverse_module_state->__pyx_tuple__273);
  Py_VISIT(traverse_module_state->__pyx_tuple__277);
  Py_VISIT(traverse_module_state->__pyx_tuple__280);
  Py_VISIT(traverse_module_state->__pyx_codeobj__25);
  Py_VISIT(traverse_module_state->__pyx_codeobj__27);
  Py_VISIT(traverse_module_state->__pyx_codeobj__29);
  Py_VISIT(traverse_module_state->__pyx_codeobj__30);
  Py_VISIT(traverse_module_state->__pyx_codeobj__31);
  Py_VISIT(traverse_module_state->__pyx_codeobj__33);
  Py_VISIT(traverse_module_state->__pyx_codeobj__34);
  Py_VISIT(traverse_module_state->__pyx_codeobj__35);
  Py_VISIT(traverse_module_state->__pyx_codeobj__37);
  Py_VISIT(traverse_module_state->__pyx_codeobj__38);
  Py_VISIT(traverse_module_state->__pyx_codeobj__40);
  Py_VISIT(traverse_module_state->__pyx_codeobj__42);
  Py_VISIT(traverse_module_state->__pyx_codeobj__44);
  Py_VISIT(traverse_module_state->__pyx_codeobj__45);
  Py_VISIT(traverse_module_state->__pyx_codeobj__47);
  Py_VISIT(traverse_module_state->__pyx_codeobj__48);
  Py_VISIT(traverse_module_state->__pyx_codeobj__49);
  Py_VISIT(traverse_module_state->__pyx_codeobj__51);
  Py_VISIT(traverse_module_state->__pyx_codeobj__52);
  Py_VISIT(traverse_module_state->__pyx_codeobj__53);
  Py_VISIT(traverse_module_state->__pyx_codeobj__55);
  Py_VISIT(traverse_module_state->__pyx_codeobj__58);
  Py_VISIT(traverse_module_state->__pyx_codeobj__60);
  Py_VISIT(traverse_module_state->__pyx_codeobj__61);
  Py_VISIT(traverse_module_state->__pyx_codeobj__62);
  Py_VISIT(traverse_module_state->__pyx_codeobj__64);
  Py_VISIT(traverse_module_state->__pyx_codeobj__66);
  Py_VISIT(traverse_module_state->__pyx_codeobj__68);
  Py_VISIT(traverse_module_state->__pyx_codeobj__70);
  Py_VISIT(traverse_module_state->__pyx_codeobj__72);
  Py_VISIT(traverse_module_state->__pyx_codeobj__74);
  Py_VISIT(traverse_module_state->__pyx_codeobj__76);
  Py_VISIT(traverse_module_state->__pyx_codeobj__77);
  Py_VISIT(traverse_module_state->__pyx_codeobj__79);
  Py_VISIT(traverse_module_state->__pyx_codeobj__80);
  Py_VISIT(traverse_module_state->__pyx_codeobj__82);
  Py_VISIT(traverse_module_state->__pyx_codeobj__84);
  Py_VISIT(traverse_module_state->__pyx_codeobj__86);
  Py_VISIT(traverse_module_state->__pyx_codeobj__88);
  Py_VISIT(traverse_module_state->__pyx_codeobj__90);
  Py_VISIT(traverse_module_state->__pyx_codeobj__91);
  Py_VISIT(traverse_module_state->__pyx_codeobj__93);
  Py_VISIT(traverse_module_state->__pyx_codeobj__95);
  Py_VISIT(traverse_module_state->__pyx_codeobj__97);
  Py_VISIT(traverse_module_state->__pyx_codeobj__99);
  Py_VISIT(traverse_module_state->__pyx_codeobj__101);
  Py_VISIT(traverse_module_state->__pyx_codeobj__103);
  Py_VISIT(traverse_module_state->__pyx_codeobj__104);
  Py_VISIT(traverse_module_state->__pyx_codeobj__105);
  Py_VISIT(traverse_module_state->__pyx_codeobj__106);
  Py_VISIT(traverse_module_state->__pyx_codeobj__108);
  Py_VISIT(traverse_module_state->__pyx_codeobj__109);
  Py_VISIT(traverse_module_state->__pyx_codeobj__110);
  Py_VISIT(traverse_module_state->__pyx_codeobj__111);
  Py_VISIT(traverse_module_state->__pyx_codeobj__112);
  Py_VISIT(traverse_module_state->__pyx_codeobj__113);
  Py_VISIT(traverse_module_state->__pyx_codeobj__114);
  Py_VISIT(traverse_module_state->__pyx_codeobj__115);
  Py_VISIT(traverse_module_state->__pyx_codeobj__116);
  Py_VISIT(traverse_module_state->__pyx_codeobj__117);
  Py_VISIT(traverse_module_state->__pyx_codeobj__119);
  Py_VISIT(traverse_module_state->__pyx_codeobj__122);
  Py_VISIT(traverse_module_state->__pyx_codeobj__124);
  Py_VISIT(traverse_module_state->__pyx_codeobj__126);
  Py_VISIT(traverse_module_state->__pyx_codeobj__128);
  Py_VISIT(traverse_module_state->__pyx_codeobj__130);
  Py_VISIT(traverse_module_state->__pyx_codeobj__132);
  Py_VISIT(traverse_module_state->__pyx_codeobj__134);
  Py_VISIT(traverse_module_state->__pyx_codeobj__136);
  Py_VISIT(traverse_module_state->__pyx_codeobj__137);
  Py_VISIT(traverse_module_state->__pyx_codeobj__138);
  Py_VISIT(traverse_module_state->__pyx_codeobj__140);
  Py_VISIT(traverse_module_state->__pyx_codeobj__141);
  Py_VISIT(traverse_module_state->__pyx_codeobj__142);
  Py_VISIT(traverse_module_state->__pyx_codeobj__144);
  Py_VISIT(traverse_module_state->__pyx_codeobj__146);
  Py_VISIT(traverse_module_state->__pyx_codeobj__148);
  Py_VISIT(traverse_module_state->__pyx_codeobj__150);
  Py_VISIT(traverse_module_state->__pyx_codeobj__151);
  Py_VISIT(traverse_module_state->__pyx_codeobj__153);
  Py_VISIT(traverse_module_state->__pyx_codeobj__154);
  Py_VISIT(traverse_module_state->__pyx_codeobj__155);
  Py_VISIT(traverse_module_state->__pyx_codeobj__156);
  Py_VISIT(traverse_module_state->__pyx_codeobj__158);
  Py_VISIT(traverse_module_state->__pyx_codeobj__159);
  Py_VISIT(traverse_module_state->__pyx_codeobj__161);
  Py_VISIT(traverse_module_state->__pyx_codeobj__162);
  Py_VISIT(traverse_module_state->__pyx_codeobj__164);
  Py_VISIT(traverse_module_state->__pyx_codeobj__166);
  Py_VISIT(traverse_module_state->__pyx_codeobj__167);
  Py_VISIT(traverse_module_state->__pyx_codeobj__168);
  Py_VISIT(traverse_module_state->__pyx_codeobj__170);
  Py_VISIT(traverse_module_state->__pyx_codeobj__172);
  Py_VISIT(traverse_module_state->__pyx_codeobj__174);
  Py_VISIT(traverse_module_state->__pyx_codeobj__176);
  Py_VISIT(traverse_module_state->__pyx_codeobj__178);
  Py_VISIT(traverse_module_state->__pyx_codeobj__180);
  Py_VISIT(traverse_module_state->__pyx_codeobj__181);
  Py_VISIT(traverse_module_state->__pyx_codeobj__183);
  Py_VISIT(traverse_module_state->__pyx_codeobj__184);
  Py_VISIT(traverse_module_state->__pyx_codeobj__185);
  Py_VISIT(traverse_module_state->__pyx_codeobj__187);
  Py_VISIT(traverse_module_state->__pyx_codeobj__188);
  Py_VISIT(traverse_module_state->__pyx_codeobj__189);
  Py_VISIT(traverse_module_state->__pyx_codeobj__190);
  Py_VISIT(traverse_module_state->__pyx_codeobj__191);
  Py_VISIT(traverse_module_state->__pyx_codeobj__192);
  Py_VISIT(traverse_module_state->__pyx_codeobj__193);
  Py_VISIT(traverse_module_state->__pyx_codeobj__194);
  Py_VISIT(traverse_module_state->__pyx_codeobj__195);
  Py_VISIT(traverse_module_state->__pyx_codeobj__196);
  Py_VISIT(traverse_module_state->__pyx_codeobj__197);
  Py_VISIT(traverse_module_state->__pyx_codeobj__198);
  Py_VISIT(traverse_module_state->__pyx_codeobj__199);
  Py_VISIT(traverse_module_state->__pyx_codeobj__200);
  Py_VISIT(traverse_module_state->__pyx_codeobj__201);
  Py_VISIT(traverse_module_state->__pyx_codeobj__202);
  Py_VISIT(traverse_module_state->__pyx_codeobj__204);
  Py_VISIT(traverse_module_state->__pyx_codeobj__206);
  Py_VISIT(traverse_module_state->__pyx_codeobj__208);
  Py_VISIT(traverse_module_state->__pyx_codeobj__209);
  Py_VISIT(traverse_module_state->__pyx_codeobj__210);
  Py_VISIT(traverse_module_state->__pyx_codeobj__211);
  Py_VISIT(traverse_module_state->__pyx_codeobj__212);
  Py_VISIT(traverse_module_state->__pyx_codeobj__213);
  Py_VISIT(traverse_module_state->__pyx_codeobj__214);
  Py_VISIT(traverse_module_state->__pyx_codeobj__215);
  Py_VISIT(traverse_module_state->__pyx_codeobj__216);
  Py_VISIT(traverse_module_state->__pyx_codeobj__218);
  Py_VISIT(traverse_module_state->__pyx_codeobj__220);
  Py_VISIT(traverse_module_state->__pyx_codeobj__222);
  Py_VISIT(traverse_module_state->__pyx_codeobj__224);
  Py_VISIT(traverse_module_state->__pyx_codeobj__226);
  Py_VISIT(traverse_module_state->__pyx_codeobj__227);
  Py_VISIT(traverse_module_state->__pyx_codeobj__229);
  Py_VISIT(traverse_module_state->__pyx_codeobj__231);
  Py_VISIT(traverse_module_state->__pyx_codeobj__233);
  Py_VISIT(traverse_module_state->__pyx_codeobj__234);
  Py_VISIT(traverse_module_state->__pyx_codeobj__235);
  Py_VISIT(traverse_module_state->__pyx_codeobj__237);
  Py_VISIT(traverse_module_state->__pyx_codeobj__238);
  Py_VISIT(traverse_module_state->__pyx_codeobj__239);
  Py_VISIT(traverse_module_state->__pyx_codeobj__241);
  Py_VISIT(traverse_module_state->__pyx_codeobj__242);
  Py_VISIT(traverse_module_state->__pyx_codeobj__243);
  Py_VISIT(traverse_module_state->__pyx_codeobj__244);
  Py_VISIT(traverse_module_state->__pyx_codeobj__245);
  Py_VISIT(traverse_module_state->__pyx_codeobj__246);
  Py_VISIT(traverse_module_state->__pyx_codeobj__248);
  Py_VISIT(traverse_module_state->__pyx_codeobj__249);
  Py_VISIT(traverse_module_state->__pyx_codeobj__250);
  Py_VISIT(traverse_module_state->__pyx_codeobj__251);
  Py_VISIT(traverse_module_state->__pyx_codeobj__253);
  Py_VISIT(traverse_module_state->__pyx_codeobj__255);
  Py_VISIT(traverse_module_state->__pyx_codeobj__257);
  Py_VISIT(traverse_module_state->__pyx_codeobj__258);
  Py_VISIT(traverse_module_state->__pyx_codeobj__260);
  Py_VISIT(traverse_module_state->__pyx_codeobj__261);
  Py_VISIT(traverse_module_state->__pyx_codeobj__262);
  Py_VISIT(traverse_module_state->__pyx_codeobj__263);
  Py_VISIT(traverse_module_state->__pyx_codeobj__264);
  Py_VISIT(traverse_module_state->__pyx_codeobj__265);
  Py_VISIT(traverse_module_state->__pyx_codeobj__266);
  Py_VISIT(traverse_module_state->__pyx_codeobj__267);
  Py_VISIT(traverse_module_state->__pyx_codeobj__268);
  Py_VISIT(traverse_module_state->__pyx_codeobj__269);
  Py_VISIT(traverse_module_state->__pyx_codeobj__270);
  Py_VISIT(traverse_module_state->__pyx_codeobj__271);
  Py_VISIT(traverse_module_state->__pyx_codeobj__272);
  Py_VISIT(traverse_module_state->__pyx_codeobj__274);
  Py_VISIT(traverse_module_state->__pyx_codeobj__275);
  Py_VISIT(traverse_module_state->__pyx_codeobj__276);
  Py_VISIT(traverse_module_state->__pyx_codeobj__278);
  Py_VISIT(traverse_module_state->__pyx_codeobj__279);
  Py_VISIT(traverse_module_state->__pyx_codeobj__281);
  Py_VISIT(traverse_module_state->__pyx_codeobj__282);
  Py_VISIT(traverse_module_state->__pyx_codeobj__283);
  Py_VISIT(traverse_module_state->__pyx_codeobj__284);
  Py_VISIT(traverse_module_state->__pyx_codeobj__285);
  Py_VISIT(traverse_module_state->__pyx_codeobj__286);
  return 0;
}
#endif
/* #### Code section: module_state_defines ### */
#define __pyx_d __pyx_mstate_global->__pyx_d
#define __pyx_b __pyx_mstate_global->__pyx_b
#define __pyx_cython_runtime __pyx_mstate_global->__pyx_cython_runtime
#define __pyx_empty_tuple __pyx_mstate_global->__pyx_empty_tuple
#define __pyx_empty_bytes __pyx_mstate_global->__pyx_empty_bytes
#define __pyx_empty_unicode __pyx_mstate_global->__pyx_empty_unicode
#ifdef __Pyx_CyFunction_USED
#define __pyx_CyFunctionType __pyx_mstate_global->__pyx_CyFunctionType
#endif
#ifdef __Pyx_FusedFunction_USED
#define __pyx_FusedFunctionType __pyx_mstate_global->__pyx_FusedFunctionType
#endif
#ifdef __Pyx_Generator_USED
#define __pyx_GeneratorType __pyx_mstate_global->__pyx_GeneratorType
#endif
#ifdef __Pyx_IterableCoroutine_USED
#define __pyx_IterableCoroutineType __pyx_mstate_global->__pyx_IterableCoroutineType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineAwaitType __pyx_mstate_global->__pyx_CoroutineAwaitType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineType __pyx_mstate_global->__pyx_CoroutineType
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#define __pyx_type_9uharfbuzz_9_harfbuzz_GlyphInfo __pyx_mstate_global->__pyx_type_9uharfbuzz_9_harfbuzz_GlyphInfo
#define __pyx_type_9uharfbuzz_9_harfbuzz_GlyphPosition __pyx_mstate_global->__pyx_type_9uharfbuzz_9_harfbuzz_GlyphPosition
#define __pyx_type_9uharfbuzz_9_harfbuzz_Buffer __pyx_mstate_global->__pyx_type_9uharfbuzz_9_harfbuzz_Buffer
#define __pyx_type_9uharfbuzz_9_harfbuzz_Blob __pyx_mstate_global->__pyx_type_9uharfbuzz_9_harfbuzz_Blob
#define __pyx_type_9uharfbuzz_9_harfbuzz_Face __pyx_mstate_global->__pyx_type_9uharfbuzz_9_harfbuzz_Face
#define __pyx_type_9uharfbuzz_9_harfbuzz_Font __pyx_mstate_global->__pyx_type_9uharfbuzz_9_harfbuzz_Font
#define __pyx_type_9uharfbuzz_9_harfbuzz_FontFuncs __pyx_mstate_global->__pyx_type_9uharfbuzz_9_harfbuzz_FontFuncs
#define __pyx_type_9uharfbuzz_9_harfbuzz_ColorLine __pyx_mstate_global->__pyx_type_9uharfbuzz_9_harfbuzz_ColorLine
#define __pyx_type_9uharfbuzz_9_harfbuzz_PaintFuncs __pyx_mstate_global->__pyx_type_9uharfbuzz_9_harfbuzz_PaintFuncs
#define __pyx_type_9uharfbuzz_9_harfbuzz_DrawFuncs __pyx_mstate_global->__pyx_type_9uharfbuzz_9_harfbuzz_DrawFuncs
#define __pyx_type_9uharfbuzz_9_harfbuzz_HBObject __pyx_mstate_global->__pyx_type_9uharfbuzz_9_harfbuzz_HBObject
#define __pyx_type_9uharfbuzz_9_harfbuzz_SubsetInput __pyx_mstate_global->__pyx_type_9uharfbuzz_9_harfbuzz_SubsetInput
#define __pyx_type_9uharfbuzz_9_harfbuzz_SubsetPlan __pyx_mstate_global->__pyx_type_9uharfbuzz_9_harfbuzz_SubsetPlan
#define __pyx_type_9uharfbuzz_9_harfbuzz_Set __pyx_mstate_global->__pyx_type_9uharfbuzz_9_harfbuzz_Set
#define __pyx_type_9uharfbuzz_9_harfbuzz_SetIter __pyx_mstate_global->__pyx_type_9uharfbuzz_9_harfbuzz_SetIter
#define __pyx_type_9uharfbuzz_9_harfbuzz_Map __pyx_mstate_global->__pyx_type_9uharfbuzz_9_harfbuzz_Map
#define __pyx_type_9uharfbuzz_9_harfbuzz_MapIter __pyx_mstate_global->__pyx_type_9uharfbuzz_9_harfbuzz_MapIter
#define __pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr __pyx_mstate_global->__pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr
#define __pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr __pyx_mstate_global->__pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr
#define __pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr __pyx_mstate_global->__pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr
#define __pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr __pyx_mstate_global->__pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr
#endif
#define __pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphInfo __pyx_mstate_global->__pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphInfo
#define __pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphPosition __pyx_mstate_global->__pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphPosition
#define __pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer __pyx_mstate_global->__pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer
#define __pyx_ptype_9uharfbuzz_9_harfbuzz_Blob __pyx_mstate_global->__pyx_ptype_9uharfbuzz_9_harfbuzz_Blob
#define __pyx_ptype_9uharfbuzz_9_harfbuzz_Face __pyx_mstate_global->__pyx_ptype_9uharfbuzz_9_harfbuzz_Face
#define __pyx_ptype_9uharfbuzz_9_harfbuzz_Font __pyx_mstate_global->__pyx_ptype_9uharfbuzz_9_harfbuzz_Font
#define __pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs __pyx_mstate_global->__pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs
#define __pyx_ptype_9uharfbuzz_9_harfbuzz_ColorLine __pyx_mstate_global->__pyx_ptype_9uharfbuzz_9_harfbuzz_ColorLine
#define __pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs __pyx_mstate_global->__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs
#define __pyx_ptype_9uharfbuzz_9_harfbuzz_DrawFuncs __pyx_mstate_global->__pyx_ptype_9uharfbuzz_9_harfbuzz_DrawFuncs
#define __pyx_ptype_9uharfbuzz_9_harfbuzz_HBObject __pyx_mstate_global->__pyx_ptype_9uharfbuzz_9_harfbuzz_HBObject
#define __pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetInput __pyx_mstate_global->__pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetInput
#define __pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetPlan __pyx_mstate_global->__pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetPlan
#define __pyx_ptype_9uharfbuzz_9_harfbuzz_Set __pyx_mstate_global->__pyx_ptype_9uharfbuzz_9_harfbuzz_Set
#define __pyx_ptype_9uharfbuzz_9_harfbuzz_SetIter __pyx_mstate_global->__pyx_ptype_9uharfbuzz_9_harfbuzz_SetIter
#define __pyx_ptype_9uharfbuzz_9_harfbuzz_Map __pyx_mstate_global->__pyx_ptype_9uharfbuzz_9_harfbuzz_Map
#define __pyx_ptype_9uharfbuzz_9_harfbuzz_MapIter __pyx_mstate_global->__pyx_ptype_9uharfbuzz_9_harfbuzz_MapIter
#define __pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr __pyx_mstate_global->__pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr
#define __pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr __pyx_mstate_global->__pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr
#define __pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr __pyx_mstate_global->__pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr
#define __pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr __pyx_mstate_global->__pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr
#define __pyx_n_s_ACCENT_BASE_HEIGHT __pyx_mstate_global->__pyx_n_s_ACCENT_BASE_HEIGHT
#define __pyx_n_s_AXIS_HEIGHT __pyx_mstate_global->__pyx_n_s_AXIS_HEIGHT
#define __pyx_kp_u_All_shapers_failed __pyx_mstate_global->__pyx_kp_u_All_shapers_failed
#define __pyx_n_s_AssertionError __pyx_mstate_global->__pyx_n_s_AssertionError
#define __pyx_n_s_BASE_GLYPH __pyx_mstate_global->__pyx_n_s_BASE_GLYPH
#define __pyx_n_s_BOT __pyx_mstate_global->__pyx_n_s_BOT
#define __pyx_n_s_BOTTOM_LEFT __pyx_mstate_global->__pyx_n_s_BOTTOM_LEFT
#define __pyx_n_s_BOTTOM_RIGHT __pyx_mstate_global->__pyx_n_s_BOTTOM_RIGHT
#define __pyx_n_s_Blob __pyx_mstate_global->__pyx_n_s_Blob
#define __pyx_n_s_Blob___reduce_cython __pyx_mstate_global->__pyx_n_s_Blob___reduce_cython
#define __pyx_n_s_Blob___setstate_cython __pyx_mstate_global->__pyx_n_s_Blob___setstate_cython
#define __pyx_n_s_Blob_from_file_path __pyx_mstate_global->__pyx_n_s_Blob_from_file_path
#define __pyx_n_s_Buffer __pyx_mstate_global->__pyx_n_s_Buffer
#define __pyx_n_s_BufferClusterLevel __pyx_mstate_global->__pyx_n_s_BufferClusterLevel
#define __pyx_n_s_BufferContentType __pyx_mstate_global->__pyx_n_s_BufferContentType
#define __pyx_n_s_BufferFlags __pyx_mstate_global->__pyx_n_s_BufferFlags
#define __pyx_n_s_Buffer___reduce_cython __pyx_mstate_global->__pyx_n_s_Buffer___reduce_cython
#define __pyx_n_s_Buffer___setstate_cython __pyx_mstate_global->__pyx_n_s_Buffer___setstate_cython
#define __pyx_n_s_Buffer_add_codepoints __pyx_mstate_global->__pyx_n_s_Buffer_add_codepoints
#define __pyx_n_s_Buffer_add_str __pyx_mstate_global->__pyx_n_s_Buffer_add_str
#define __pyx_n_s_Buffer_add_utf8 __pyx_mstate_global->__pyx_n_s_Buffer_add_utf8
#define __pyx_n_s_Buffer_clear_contents __pyx_mstate_global->__pyx_n_s_Buffer_clear_contents
#define __pyx_n_s_Buffer_create __pyx_mstate_global->__pyx_n_s_Buffer_create
#define __pyx_n_s_Buffer_guess_segment_properties __pyx_mstate_global->__pyx_n_s_Buffer_guess_segment_properties
#define __pyx_n_s_Buffer_reset __pyx_mstate_global->__pyx_n_s_Buffer_reset
#define __pyx_n_s_Buffer_set_language_from_ot_tag __pyx_mstate_global->__pyx_n_s_Buffer_set_language_from_ot_tag
#define __pyx_n_s_Buffer_set_message_func __pyx_mstate_global->__pyx_n_s_Buffer_set_message_func
#define __pyx_n_s_Buffer_set_script_from_ot_tag __pyx_mstate_global->__pyx_n_s_Buffer_set_script_from_ot_tag
#define __pyx_n_s_CHARACTERS __pyx_mstate_global->__pyx_n_s_CHARACTERS
#define __pyx_n_s_CLEAR __pyx_mstate_global->__pyx_n_s_CLEAR
#define __pyx_n_s_COLOR_BURN __pyx_mstate_global->__pyx_n_s_COLOR_BURN
#define __pyx_n_s_COLOR_DODGE __pyx_mstate_global->__pyx_n_s_COLOR_DODGE
#define __pyx_n_s_COMPONENT __pyx_mstate_global->__pyx_n_s_COMPONENT
#define __pyx_n_s_Callable __pyx_mstate_global->__pyx_n_s_Callable
#define __pyx_kp_s_Callable_Blob_int_int_str_float __pyx_mstate_global->__pyx_kp_s_Callable_Blob_int_int_str_float
#define __pyx_kp_s_Callable_ColorLine_float_float_f __pyx_mstate_global->__pyx_kp_s_Callable_ColorLine_float_float_f
#define __pyx_kp_s_Callable_ColorLine_float_float_f_2 __pyx_mstate_global->__pyx_kp_s_Callable_ColorLine_float_float_f_2
#define __pyx_kp_s_Callable_Color_bool_object_None __pyx_mstate_global->__pyx_kp_s_Callable_Color_bool_object_None
#define __pyx_kp_s_Callable_Face_str_object_bytes __pyx_mstate_global->__pyx_kp_s_Callable_Face_str_object_bytes
#define __pyx_kp_s_Callable_Font_int_int_object_int __pyx_mstate_global->__pyx_kp_s_Callable_Font_int_int_object_int
#define __pyx_kp_s_Callable_Font_int_object_int __pyx_mstate_global->__pyx_kp_s_Callable_Font_int_object_int
#define __pyx_kp_s_Callable_Font_int_object_int_int __pyx_mstate_global->__pyx_kp_s_Callable_Font_int_object_int_int
#define __pyx_kp_s_Callable_Font_int_object_str __pyx_mstate_global->__pyx_kp_s_Callable_Font_int_object_str
#define __pyx_kp_s_Callable_Font_object_FontExtents __pyx_mstate_global->__pyx_kp_s_Callable_Font_object_FontExtents
#define __pyx_kp_s_Callable_PaintCompositeMode_obje __pyx_mstate_global->__pyx_kp_s_Callable_PaintCompositeMode_obje
#define __pyx_kp_s_Callable_float_float_float_float __pyx_mstate_global->__pyx_kp_s_Callable_float_float_float_float
#define __pyx_kp_s_Callable_float_float_float_float_2 __pyx_mstate_global->__pyx_kp_s_Callable_float_float_float_float_2
#define __pyx_kp_s_Callable_float_float_object_None __pyx_mstate_global->__pyx_kp_s_Callable_float_float_object_None
#define __pyx_kp_s_Callable_int_object_Color __pyx_mstate_global->__pyx_kp_s_Callable_int_object_Color
#define __pyx_kp_s_Callable_int_object_None __pyx_mstate_global->__pyx_kp_s_Callable_int_object_None
#define __pyx_kp_s_Callable_int_object_bool __pyx_mstate_global->__pyx_kp_s_Callable_int_object_bool
#define __pyx_kp_s_Callable_object_None __pyx_mstate_global->__pyx_kp_s_Callable_object_None
#define __pyx_n_s_Color __pyx_mstate_global->__pyx_n_s_Color
#define __pyx_n_u_Color __pyx_mstate_global->__pyx_n_u_Color
#define __pyx_n_s_ColorLine __pyx_mstate_global->__pyx_n_s_ColorLine
#define __pyx_n_s_ColorLine___reduce_cython __pyx_mstate_global->__pyx_n_s_ColorLine___reduce_cython
#define __pyx_n_s_ColorLine___setstate_cython __pyx_mstate_global->__pyx_n_s_ColorLine___setstate_cython
#define __pyx_n_s_ColorStop __pyx_mstate_global->__pyx_n_s_ColorStop
#define __pyx_n_u_ColorStop __pyx_mstate_global->__pyx_n_u_ColorStop
#define __pyx_kp_s_Color_None __pyx_mstate_global->__pyx_kp_s_Color_None
#define __pyx_n_s_Color_from_int __pyx_mstate_global->__pyx_n_s_Color_from_int
#define __pyx_n_s_Color_to_int __pyx_mstate_global->__pyx_n_s_Color_to_int
#define __pyx_n_s_DARKEN __pyx_mstate_global->__pyx_n_s_DARKEN
#define __pyx_n_s_DEFAULT __pyx_mstate_global->__pyx_n_s_DEFAULT
#define __pyx_n_s_DEFAULT_REPLACEMENT_CODEPOINT __pyx_mstate_global->__pyx_n_s_DEFAULT_REPLACEMENT_CODEPOINT
#define __pyx_n_s_DELIMITED_SUB_FORMULA_MIN_HEIGHT __pyx_mstate_global->__pyx_n_s_DELIMITED_SUB_FORMULA_MIN_HEIGHT
#define __pyx_n_s_DEST __pyx_mstate_global->__pyx_n_s_DEST
#define __pyx_n_s_DEST_ATOP __pyx_mstate_global->__pyx_n_s_DEST_ATOP
#define __pyx_n_s_DEST_IN __pyx_mstate_global->__pyx_n_s_DEST_IN
#define __pyx_n_s_DEST_OUT __pyx_mstate_global->__pyx_n_s_DEST_OUT
#define __pyx_n_s_DEST_OVER __pyx_mstate_global->__pyx_n_s_DEST_OVER
#define __pyx_n_s_DESUBROUTINIZE __pyx_mstate_global->__pyx_n_s_DESUBROUTINIZE
#define __pyx_n_s_DIFFERENCE __pyx_mstate_global->__pyx_n_s_DIFFERENCE
#define __pyx_n_s_DISPLAY_OPERATOR_MIN_HEIGHT __pyx_mstate_global->__pyx_n_s_DISPLAY_OPERATOR_MIN_HEIGHT
#define __pyx_n_s_DO_NOT_INSERT_DOTTED_CIRCLE __pyx_mstate_global->__pyx_n_s_DO_NOT_INSERT_DOTTED_CIRCLE
#define __pyx_n_s_DROP_TABLE_TAG __pyx_mstate_global->__pyx_n_s_DROP_TABLE_TAG
#define __pyx_n_s_DeprecationWarning __pyx_mstate_global->__pyx_n_s_DeprecationWarning
#define __pyx_n_s_Dict __pyx_mstate_global->__pyx_n_s_Dict
#define __pyx_kp_s_Dict_str_Union_int_bool_Sequence __pyx_mstate_global->__pyx_kp_s_Dict_str_Union_int_bool_Sequence
#define __pyx_kp_s_Dict_str_float __pyx_mstate_global->__pyx_kp_s_Dict_str_float
#define __pyx_n_s_DrawFuncs __pyx_mstate_global->__pyx_n_s_DrawFuncs
#define __pyx_n_s_DrawFuncs___reduce_cython __pyx_mstate_global->__pyx_n_s_DrawFuncs___reduce_cython
#define __pyx_n_s_DrawFuncs___setstate_cython __pyx_mstate_global->__pyx_n_s_DrawFuncs___setstate_cython
#define __pyx_n_s_DrawFuncs_draw_glyph __pyx_mstate_global->__pyx_n_s_DrawFuncs_draw_glyph
#define __pyx_n_s_DrawFuncs_get_glyph_shape __pyx_mstate_global->__pyx_n_s_DrawFuncs_get_glyph_shape
#define __pyx_n_s_DrawFuncs_set_close_path_func __pyx_mstate_global->__pyx_n_s_DrawFuncs_set_close_path_func
#define __pyx_n_s_DrawFuncs_set_cubic_to_func __pyx_mstate_global->__pyx_n_s_DrawFuncs_set_cubic_to_func
#define __pyx_n_s_DrawFuncs_set_line_to_func __pyx_mstate_global->__pyx_n_s_DrawFuncs_set_line_to_func
#define __pyx_n_s_DrawFuncs_set_move_to_func __pyx_mstate_global->__pyx_n_s_DrawFuncs_set_move_to_func
#define __pyx_n_s_DrawFuncs_set_quadratic_to_func __pyx_mstate_global->__pyx_n_s_DrawFuncs_set_quadratic_to_func
#define __pyx_n_s_EOT __pyx_mstate_global->__pyx_n_s_EOT
#define __pyx_n_s_EXCLUSION __pyx_mstate_global->__pyx_n_s_EXCLUSION
#define __pyx_n_s_EXTENDER __pyx_mstate_global->__pyx_n_s_EXTENDER
#define __pyx_n_s_FLATTENED_ACCENT_BASE_HEIGHT __pyx_mstate_global->__pyx_n_s_FLATTENED_ACCENT_BASE_HEIGHT
#define __pyx_n_s_FRACTION_DENOMINATOR_DISPLAY_STY __pyx_mstate_global->__pyx_n_s_FRACTION_DENOMINATOR_DISPLAY_STY
#define __pyx_n_s_FRACTION_DENOMINATOR_GAP_MIN __pyx_mstate_global->__pyx_n_s_FRACTION_DENOMINATOR_GAP_MIN
#define __pyx_n_s_FRACTION_DENOMINATOR_SHIFT_DOWN __pyx_mstate_global->__pyx_n_s_FRACTION_DENOMINATOR_SHIFT_DOWN
#define __pyx_n_s_FRACTION_DENOM_DISPLAY_STYLE_GAP __pyx_mstate_global->__pyx_n_s_FRACTION_DENOM_DISPLAY_STYLE_GAP
#define __pyx_n_s_FRACTION_NUMERATOR_DISPLAY_STYLE __pyx_mstate_global->__pyx_n_s_FRACTION_NUMERATOR_DISPLAY_STYLE
#define __pyx_n_s_FRACTION_NUMERATOR_GAP_MIN __pyx_mstate_global->__pyx_n_s_FRACTION_NUMERATOR_GAP_MIN
#define __pyx_n_s_FRACTION_NUMERATOR_SHIFT_UP __pyx_mstate_global->__pyx_n_s_FRACTION_NUMERATOR_SHIFT_UP
#define __pyx_n_s_FRACTION_NUM_DISPLAY_STYLE_GAP_M __pyx_mstate_global->__pyx_n_s_FRACTION_NUM_DISPLAY_STYLE_GAP_M
#define __pyx_n_s_FRACTION_RULE_THICKNESS __pyx_mstate_global->__pyx_n_s_FRACTION_RULE_THICKNESS
#define __pyx_n_s_Face __pyx_mstate_global->__pyx_n_s_Face
#define __pyx_n_s_Face___reduce_cython __pyx_mstate_global->__pyx_n_s_Face___reduce_cython
#define __pyx_n_s_Face___setstate_cython __pyx_mstate_global->__pyx_n_s_Face___setstate_cython
#define __pyx_n_s_Face_create __pyx_mstate_global->__pyx_n_s_Face_create
#define __pyx_n_s_Face_create_for_tables __pyx_mstate_global->__pyx_n_s_Face_create_for_tables
#define __pyx_n_s_Face_variation_unicodes __pyx_mstate_global->__pyx_n_s_Face_variation_unicodes
#define __pyx_kp_u_Failed_to_open __pyx_mstate_global->__pyx_kp_u_Failed_to_open
#define __pyx_n_s_Font __pyx_mstate_global->__pyx_n_s_Font
#define __pyx_n_s_FontExtents __pyx_mstate_global->__pyx_n_s_FontExtents
#define __pyx_n_u_FontExtents __pyx_mstate_global->__pyx_n_u_FontExtents
#define __pyx_n_s_FontFuncs __pyx_mstate_global->__pyx_n_s_FontFuncs
#define __pyx_n_s_FontFuncs___reduce_cython __pyx_mstate_global->__pyx_n_s_FontFuncs___reduce_cython
#define __pyx_n_s_FontFuncs___setstate_cython __pyx_mstate_global->__pyx_n_s_FontFuncs___setstate_cython
#define __pyx_n_s_FontFuncs_create __pyx_mstate_global->__pyx_n_s_FontFuncs_create
#define __pyx_n_s_FontFuncs_set_font_h_extents_fun __pyx_mstate_global->__pyx_n_s_FontFuncs_set_font_h_extents_fun
#define __pyx_n_s_FontFuncs_set_font_v_extents_fun __pyx_mstate_global->__pyx_n_s_FontFuncs_set_font_v_extents_fun
#define __pyx_n_s_FontFuncs_set_glyph_h_advance_fu __pyx_mstate_global->__pyx_n_s_FontFuncs_set_glyph_h_advance_fu
#define __pyx_n_s_FontFuncs_set_glyph_name_func __pyx_mstate_global->__pyx_n_s_FontFuncs_set_glyph_name_func
#define __pyx_n_s_FontFuncs_set_glyph_v_advance_fu __pyx_mstate_global->__pyx_n_s_FontFuncs_set_glyph_v_advance_fu
#define __pyx_n_s_FontFuncs_set_glyph_v_origin_fun __pyx_mstate_global->__pyx_n_s_FontFuncs_set_glyph_v_origin_fun
#define __pyx_n_s_FontFuncs_set_nominal_glyph_func __pyx_mstate_global->__pyx_n_s_FontFuncs_set_nominal_glyph_func
#define __pyx_n_s_FontFuncs_set_variation_glyph_fu __pyx_mstate_global->__pyx_n_s_FontFuncs_set_variation_glyph_fu
#define __pyx_n_s_Font___reduce_cython __pyx_mstate_global->__pyx_n_s_Font___reduce_cython
#define __pyx_n_s_Font___setstate_cython __pyx_mstate_global->__pyx_n_s_Font___setstate_cython
#define __pyx_n_s_Font_create __pyx_mstate_global->__pyx_n_s_Font_create
#define __pyx_n_s_Font_draw_glyph __pyx_mstate_global->__pyx_n_s_Font_draw_glyph
#define __pyx_n_s_Font_draw_glyph_with_pen __pyx_mstate_global->__pyx_n_s_Font_draw_glyph_with_pen
#define __pyx_n_s_Font_get_font_extents __pyx_mstate_global->__pyx_n_s_Font_get_font_extents
#define __pyx_n_s_Font_get_glyph_extents __pyx_mstate_global->__pyx_n_s_Font_get_glyph_extents
#define __pyx_n_s_Font_get_glyph_from_name __pyx_mstate_global->__pyx_n_s_Font_get_glyph_from_name
#define __pyx_n_s_Font_get_glyph_h_advance __pyx_mstate_global->__pyx_n_s_Font_get_glyph_h_advance
#define __pyx_n_s_Font_get_glyph_h_origin __pyx_mstate_global->__pyx_n_s_Font_get_glyph_h_origin
#define __pyx_n_s_Font_get_glyph_name __pyx_mstate_global->__pyx_n_s_Font_get_glyph_name
#define __pyx_n_s_Font_get_glyph_v_advance __pyx_mstate_global->__pyx_n_s_Font_get_glyph_v_advance
#define __pyx_n_s_Font_get_glyph_v_origin __pyx_mstate_global->__pyx_n_s_Font_get_glyph_v_origin
#define __pyx_n_s_Font_get_nominal_glyph __pyx_mstate_global->__pyx_n_s_Font_get_nominal_glyph
#define __pyx_n_s_Font_get_var_coords_design __pyx_mstate_global->__pyx_n_s_Font_get_var_coords_design
#define __pyx_n_s_Font_get_var_coords_normalized __pyx_mstate_global->__pyx_n_s_Font_get_var_coords_normalized
#define __pyx_n_s_Font_get_variation_glyph __pyx_mstate_global->__pyx_n_s_Font_get_variation_glyph
#define __pyx_n_s_Font_glyph_from_string __pyx_mstate_global->__pyx_n_s_Font_glyph_from_string
#define __pyx_n_s_Font_glyph_to_string __pyx_mstate_global->__pyx_n_s_Font_glyph_to_string
#define __pyx_n_s_Font_paint_glyph __pyx_mstate_global->__pyx_n_s_Font_paint_glyph
#define __pyx_n_s_Font_set_var_coords_design __pyx_mstate_global->__pyx_n_s_Font_set_var_coords_design
#define __pyx_n_s_Font_set_var_coords_normalized __pyx_mstate_global->__pyx_n_s_Font_set_var_coords_normalized
#define __pyx_n_s_Font_set_variation __pyx_mstate_global->__pyx_n_s_Font_set_variation
#define __pyx_n_s_Font_set_variations __pyx_mstate_global->__pyx_n_s_Font_set_variations
#define __pyx_n_s_GLYPHS __pyx_mstate_global->__pyx_n_s_GLYPHS
#define __pyx_n_s_GLYPH_INDEX __pyx_mstate_global->__pyx_n_s_GLYPH_INDEX
#define __pyx_n_s_GLYPH_NAMES __pyx_mstate_global->__pyx_n_s_GLYPH_NAMES
#define __pyx_n_s_GlyphExtents __pyx_mstate_global->__pyx_n_s_GlyphExtents
#define __pyx_n_u_GlyphExtents __pyx_mstate_global->__pyx_n_u_GlyphExtents
#define __pyx_n_s_GlyphFlags __pyx_mstate_global->__pyx_n_s_GlyphFlags
#define __pyx_n_s_GlyphInfo __pyx_mstate_global->__pyx_n_s_GlyphInfo
#define __pyx_n_s_GlyphInfo___reduce_cython __pyx_mstate_global->__pyx_n_s_GlyphInfo___reduce_cython
#define __pyx_n_s_GlyphInfo___setstate_cython __pyx_mstate_global->__pyx_n_s_GlyphInfo___setstate_cython
#define __pyx_n_s_GlyphPosition __pyx_mstate_global->__pyx_n_s_GlyphPosition
#define __pyx_n_s_GlyphPosition___reduce_cython __pyx_mstate_global->__pyx_n_s_GlyphPosition___reduce_cython
#define __pyx_n_s_GlyphPosition___setstate_cython __pyx_mstate_global->__pyx_n_s_GlyphPosition___setstate_cython
#define __pyx_n_s_HARD_LIGHT __pyx_mstate_global->__pyx_n_s_HARD_LIGHT
#define __pyx_n_s_HBObject __pyx_mstate_global->__pyx_n_s_HBObject
#define __pyx_n_s_HBObject___reduce_cython __pyx_mstate_global->__pyx_n_s_HBObject___reduce_cython
#define __pyx_n_s_HBObject___setstate_cython __pyx_mstate_global->__pyx_n_s_HBObject___setstate_cython
#define __pyx_n_s_HSL_COLOR __pyx_mstate_global->__pyx_n_s_HSL_COLOR
#define __pyx_n_s_HSL_HUE __pyx_mstate_global->__pyx_n_s_HSL_HUE
#define __pyx_n_s_HSL_LUMINOSITY __pyx_mstate_global->__pyx_n_s_HSL_LUMINOSITY
#define __pyx_n_s_HSL_SATURATION __pyx_mstate_global->__pyx_n_s_HSL_SATURATION
#define __pyx_n_s_HarfBuzzError __pyx_mstate_global->__pyx_n_s_HarfBuzzError
#define __pyx_n_s_INVALID __pyx_mstate_global->__pyx_n_s_INVALID
#define __pyx_n_s_INVALID_VALUE __pyx_mstate_global->__pyx_n_s_INVALID_VALUE
#define __pyx_kp_u_Input_num_of_subtables __pyx_mstate_global->__pyx_kp_u_Input_num_of_subtables
#define __pyx_n_s_IntEnum __pyx_mstate_global->__pyx_n_s_IntEnum
#define __pyx_n_s_IntFlag __pyx_mstate_global->__pyx_n_s_IntFlag
#define __pyx_n_s_KeyError __pyx_mstate_global->__pyx_n_s_KeyError
#define __pyx_n_s_LAYOUT_FEATURE_TAG __pyx_mstate_global->__pyx_n_s_LAYOUT_FEATURE_TAG
#define __pyx_n_s_LAYOUT_SCRIPT_TAG __pyx_mstate_global->__pyx_n_s_LAYOUT_SCRIPT_TAG
#define __pyx_n_s_LIGATURE __pyx_mstate_global->__pyx_n_s_LIGATURE
#define __pyx_n_s_LIGHTEN __pyx_mstate_global->__pyx_n_s_LIGHTEN
#define __pyx_n_s_LOWER_LIMIT_BASELINE_DROP_MIN __pyx_mstate_global->__pyx_n_s_LOWER_LIMIT_BASELINE_DROP_MIN
#define __pyx_n_s_LOWER_LIMIT_GAP_MIN __pyx_mstate_global->__pyx_n_s_LOWER_LIMIT_GAP_MIN
#define __pyx_n_s_List __pyx_mstate_global->__pyx_n_s_List
#define __pyx_kp_s_List_Color __pyx_mstate_global->__pyx_kp_s_List_Color
#define __pyx_kp_s_List_OTColorLayer __pyx_mstate_global->__pyx_kp_s_List_OTColorLayer
#define __pyx_kp_s_List_OTMathGlyphVariant __pyx_mstate_global->__pyx_kp_s_List_OTMathGlyphVariant
#define __pyx_kp_s_List_OTMathKernEntry __pyx_mstate_global->__pyx_kp_s_List_OTMathKernEntry
#define __pyx_kp_s_List_Tuple_List_Tuple_int_int_in __pyx_mstate_global->__pyx_kp_s_List_Tuple_List_Tuple_int_int_in
#define __pyx_kp_s_List_bytes __pyx_mstate_global->__pyx_kp_s_List_bytes
#define __pyx_kp_s_List_int __pyx_mstate_global->__pyx_kp_s_List_int
#define __pyx_kp_s_List_str __pyx_mstate_global->__pyx_kp_s_List_str
#define __pyx_n_s_MARK __pyx_mstate_global->__pyx_n_s_MARK
#define __pyx_n_s_MATH_LEADING __pyx_mstate_global->__pyx_n_s_MATH_LEADING
#define __pyx_n_s_MONOTONE_CHARACTERS __pyx_mstate_global->__pyx_n_s_MONOTONE_CHARACTERS
#define __pyx_n_s_MONOTONE_GRAPHEMES __pyx_mstate_global->__pyx_n_s_MONOTONE_GRAPHEMES
#define __pyx_n_s_MULTIPLY __pyx_mstate_global->__pyx_n_s_MULTIPLY
#define __pyx_n_s_Map __pyx_mstate_global->__pyx_n_s_Map
#define __pyx_n_s_MapIter __pyx_mstate_global->__pyx_n_s_MapIter
#define __pyx_n_s_MapIter___reduce_cython __pyx_mstate_global->__pyx_n_s_MapIter___reduce_cython
#define __pyx_n_s_MapIter___setstate_cython __pyx_mstate_global->__pyx_n_s_MapIter___setstate_cython
#define __pyx_n_s_Map___copy __pyx_mstate_global->__pyx_n_s_Map___copy
#define __pyx_n_s_Map___reduce_cython __pyx_mstate_global->__pyx_n_s_Map___reduce_cython
#define __pyx_n_s_Map___setstate_cython __pyx_mstate_global->__pyx_n_s_Map___setstate_cython
#define __pyx_n_s_Map__is_equal __pyx_mstate_global->__pyx_n_s_Map__is_equal
#define __pyx_n_s_Map__update __pyx_mstate_global->__pyx_n_s_Map__update
#define __pyx_n_s_Map_clear __pyx_mstate_global->__pyx_n_s_Map_clear
#define __pyx_n_s_Map_copy __pyx_mstate_global->__pyx_n_s_Map_copy
#define __pyx_n_s_Map_get __pyx_mstate_global->__pyx_n_s_Map_get
#define __pyx_n_s_Map_items __pyx_mstate_global->__pyx_n_s_Map_items
#define __pyx_n_s_Map_keys __pyx_mstate_global->__pyx_n_s_Map_keys
#define __pyx_kp_u_Map_s __pyx_mstate_global->__pyx_kp_u_Map_s
#define __pyx_n_s_Map_update __pyx_mstate_global->__pyx_n_s_Map_update
#define __pyx_n_s_Map_values __pyx_mstate_global->__pyx_n_s_Map_values
#define __pyx_n_s_MemoryError __pyx_mstate_global->__pyx_n_s_MemoryError
#define __pyx_n_s_NAME_ID __pyx_mstate_global->__pyx_n_s_NAME_ID
#define __pyx_n_s_NAME_LANG_ID __pyx_mstate_global->__pyx_n_s_NAME_LANG_ID
#define __pyx_n_s_NAME_LEGACY __pyx_mstate_global->__pyx_n_s_NAME_LEGACY
#define __pyx_n_s_NOTDEF_OUTLINE __pyx_mstate_global->__pyx_n_s_NOTDEF_OUTLINE
#define __pyx_n_s_NO_HINTING __pyx_mstate_global->__pyx_n_s_NO_HINTING
#define __pyx_n_s_NO_LAYOUT_CLOSURE __pyx_mstate_global->__pyx_n_s_NO_LAYOUT_CLOSURE
#define __pyx_n_s_NO_PRUNE_UNICODE_RANGES __pyx_mstate_global->__pyx_n_s_NO_PRUNE_UNICODE_RANGES
#define __pyx_n_s_NO_SUBSET_TABLE_TAG __pyx_mstate_global->__pyx_n_s_NO_SUBSET_TABLE_TAG
#define __pyx_n_s_None __pyx_mstate_global->__pyx_n_s_None
#define __pyx_n_s_NotImplemented __pyx_mstate_global->__pyx_n_s_NotImplemented
#define __pyx_n_s_OTColorLayer __pyx_mstate_global->__pyx_n_s_OTColorLayer
#define __pyx_n_u_OTColorLayer __pyx_mstate_global->__pyx_n_u_OTColorLayer
#define __pyx_n_s_OTColorPaletteFlags __pyx_mstate_global->__pyx_n_s_OTColorPaletteFlags
#define __pyx_n_s_OTLayoutGlyphClass __pyx_mstate_global->__pyx_n_s_OTLayoutGlyphClass
#define __pyx_n_s_OTMathConstant __pyx_mstate_global->__pyx_n_s_OTMathConstant
#define __pyx_n_s_OTMathGlyphPart __pyx_mstate_global->__pyx_n_s_OTMathGlyphPart
#define __pyx_n_u_OTMathGlyphPart __pyx_mstate_global->__pyx_n_u_OTMathGlyphPart
#define __pyx_n_s_OTMathGlyphPartFlags __pyx_mstate_global->__pyx_n_s_OTMathGlyphPartFlags
#define __pyx_n_s_OTMathGlyphVariant __pyx_mstate_global->__pyx_n_s_OTMathGlyphVariant
#define __pyx_n_u_OTMathGlyphVariant __pyx_mstate_global->__pyx_n_u_OTMathGlyphVariant
#define __pyx_n_s_OTMathKern __pyx_mstate_global->__pyx_n_s_OTMathKern
#define __pyx_n_s_OTMathKernEntry __pyx_mstate_global->__pyx_n_s_OTMathKernEntry
#define __pyx_n_u_OTMathKernEntry __pyx_mstate_global->__pyx_n_u_OTMathKernEntry
#define __pyx_n_s_OVERBAR_EXTRA_ASCENDER __pyx_mstate_global->__pyx_n_s_OVERBAR_EXTRA_ASCENDER
#define __pyx_n_s_OVERBAR_RULE_THICKNESS __pyx_mstate_global->__pyx_n_s_OVERBAR_RULE_THICKNESS
#define __pyx_n_s_OVERBAR_VERTICAL_GAP __pyx_mstate_global->__pyx_n_s_OVERBAR_VERTICAL_GAP
#define __pyx_n_s_OVERLAY __pyx_mstate_global->__pyx_n_s_OVERLAY
#define __pyx_n_s_PAD __pyx_mstate_global->__pyx_n_s_PAD
#define __pyx_n_s_PASSTHROUGH_UNRECOGNIZED __pyx_mstate_global->__pyx_n_s_PASSTHROUGH_UNRECOGNIZED
#define __pyx_n_s_PLUS __pyx_mstate_global->__pyx_n_s_PLUS
#define __pyx_n_s_PRESERVE_DEFAULT_IGNORABLES __pyx_mstate_global->__pyx_n_s_PRESERVE_DEFAULT_IGNORABLES
#define __pyx_n_s_PRODUCE_SAFE_TO_INSERT_TATWEEL __pyx_mstate_global->__pyx_n_s_PRODUCE_SAFE_TO_INSERT_TATWEEL
#define __pyx_n_s_PRODUCE_UNSAFE_TO_CONCAT __pyx_mstate_global->__pyx_n_s_PRODUCE_UNSAFE_TO_CONCAT
#define __pyx_n_s_PaintCompositeMode __pyx_mstate_global->__pyx_n_s_PaintCompositeMode
#define __pyx_n_s_PaintExtend __pyx_mstate_global->__pyx_n_s_PaintExtend
#define __pyx_n_s_PaintFuncs __pyx_mstate_global->__pyx_n_s_PaintFuncs
#define __pyx_n_s_PaintFuncs___reduce_cython __pyx_mstate_global->__pyx_n_s_PaintFuncs___reduce_cython
#define __pyx_n_s_PaintFuncs___setstate_cython __pyx_mstate_global->__pyx_n_s_PaintFuncs___setstate_cython
#define __pyx_n_s_PaintFuncs_set_color_func __pyx_mstate_global->__pyx_n_s_PaintFuncs_set_color_func
#define __pyx_n_s_PaintFuncs_set_color_glyph_func __pyx_mstate_global->__pyx_n_s_PaintFuncs_set_color_glyph_func
#define __pyx_n_s_PaintFuncs_set_custom_palette_co __pyx_mstate_global->__pyx_n_s_PaintFuncs_set_custom_palette_co
#define __pyx_n_s_PaintFuncs_set_image_func __pyx_mstate_global->__pyx_n_s_PaintFuncs_set_image_func
#define __pyx_n_s_PaintFuncs_set_linear_gradient_f __pyx_mstate_global->__pyx_n_s_PaintFuncs_set_linear_gradient_f
#define __pyx_n_s_PaintFuncs_set_pop_clip_func __pyx_mstate_global->__pyx_n_s_PaintFuncs_set_pop_clip_func
#define __pyx_n_s_PaintFuncs_set_pop_group_func __pyx_mstate_global->__pyx_n_s_PaintFuncs_set_pop_group_func
#define __pyx_n_s_PaintFuncs_set_pop_transform_fun __pyx_mstate_global->__pyx_n_s_PaintFuncs_set_pop_transform_fun
#define __pyx_n_s_PaintFuncs_set_push_clip_glyph_f __pyx_mstate_global->__pyx_n_s_PaintFuncs_set_push_clip_glyph_f
#define __pyx_n_s_PaintFuncs_set_push_clip_rectang __pyx_mstate_global->__pyx_n_s_PaintFuncs_set_push_clip_rectang
#define __pyx_n_s_PaintFuncs_set_push_group_func __pyx_mstate_global->__pyx_n_s_PaintFuncs_set_push_group_func
#define __pyx_n_s_PaintFuncs_set_push_transform_fu __pyx_mstate_global->__pyx_n_s_PaintFuncs_set_push_transform_fu
#define __pyx_n_s_PaintFuncs_set_radial_gradient_f __pyx_mstate_global->__pyx_n_s_PaintFuncs_set_radial_gradient_f
#define __pyx_n_s_PaintFuncs_set_sweep_gradient_fu __pyx_mstate_global->__pyx_n_s_PaintFuncs_set_sweep_gradient_fu
#define __pyx_kp_u_Pass_draw_state_to_Font_draw_gly __pyx_mstate_global->__pyx_kp_u_Pass_draw_state_to_Font_draw_gly
#define __pyx_n_s_Path __pyx_mstate_global->__pyx_n_s_Path
#define __pyx_kp_s_Pickling_of_struct_members_such __pyx_mstate_global->__pyx_kp_s_Pickling_of_struct_members_such
#define __pyx_kp_s_Pickling_of_struct_members_such_2 __pyx_mstate_global->__pyx_kp_s_Pickling_of_struct_members_such_2
#define __pyx_n_s_RADICAL_DEGREE_BOTTOM_RAISE_PERC __pyx_mstate_global->__pyx_n_s_RADICAL_DEGREE_BOTTOM_RAISE_PERC
#define __pyx_n_s_RADICAL_DISPLAY_STYLE_VERTICAL_G __pyx_mstate_global->__pyx_n_s_RADICAL_DISPLAY_STYLE_VERTICAL_G
#define __pyx_n_s_RADICAL_EXTRA_ASCENDER __pyx_mstate_global->__pyx_n_s_RADICAL_EXTRA_ASCENDER
#define __pyx_n_s_RADICAL_KERN_AFTER_DEGREE __pyx_mstate_global->__pyx_n_s_RADICAL_KERN_AFTER_DEGREE
#define __pyx_n_s_RADICAL_KERN_BEFORE_DEGREE __pyx_mstate_global->__pyx_n_s_RADICAL_KERN_BEFORE_DEGREE
#define __pyx_n_s_RADICAL_RULE_THICKNESS __pyx_mstate_global->__pyx_n_s_RADICAL_RULE_THICKNESS
#define __pyx_n_s_RADICAL_VERTICAL_GAP __pyx_mstate_global->__pyx_n_s_RADICAL_VERTICAL_GAP
#define __pyx_n_s_REFLECT __pyx_mstate_global->__pyx_n_s_REFLECT
#define __pyx_n_s_REMOVE_DEFAULT_IGNORABLES __pyx_mstate_global->__pyx_n_s_REMOVE_DEFAULT_IGNORABLES
#define __pyx_n_s_REPEAT __pyx_mstate_global->__pyx_n_s_REPEAT
#define __pyx_n_s_RETAIN_GIDS __pyx_mstate_global->__pyx_n_s_RETAIN_GIDS
#define __pyx_n_s_RepackerError __pyx_mstate_global->__pyx_n_s_RepackerError
#define __pyx_n_s_RuntimeError __pyx_mstate_global->__pyx_n_s_RuntimeError
#define __pyx_n_s_SAFE_TO_INSERT_TATWEEL __pyx_mstate_global->__pyx_n_s_SAFE_TO_INSERT_TATWEEL
#define __pyx_n_s_SCREEN __pyx_mstate_global->__pyx_n_s_SCREEN
#define __pyx_n_s_SCRIPT_PERCENT_SCALE_DOWN __pyx_mstate_global->__pyx_n_s_SCRIPT_PERCENT_SCALE_DOWN
#define __pyx_n_s_SCRIPT_SCRIPT_PERCENT_SCALE_DOWN __pyx_mstate_global->__pyx_n_s_SCRIPT_SCRIPT_PERCENT_SCALE_DOWN
#define __pyx_n_s_SET_OVERLAPS_FLAG __pyx_mstate_global->__pyx_n_s_SET_OVERLAPS_FLAG
#define __pyx_n_s_SKEWED_FRACTION_HORIZONTAL_GAP __pyx_mstate_global->__pyx_n_s_SKEWED_FRACTION_HORIZONTAL_GAP
#define __pyx_n_s_SKEWED_FRACTION_VERTICAL_GAP __pyx_mstate_global->__pyx_n_s_SKEWED_FRACTION_VERTICAL_GAP
#define __pyx_n_s_SOFT_LIGHT __pyx_mstate_global->__pyx_n_s_SOFT_LIGHT
#define __pyx_n_s_SPACE_AFTER_SCRIPT __pyx_mstate_global->__pyx_n_s_SPACE_AFTER_SCRIPT
#define __pyx_n_s_SRC __pyx_mstate_global->__pyx_n_s_SRC
#define __pyx_n_s_SRC_ATOP __pyx_mstate_global->__pyx_n_s_SRC_ATOP
#define __pyx_n_s_SRC_IN __pyx_mstate_global->__pyx_n_s_SRC_IN
#define __pyx_n_s_SRC_OUT __pyx_mstate_global->__pyx_n_s_SRC_OUT
#define __pyx_n_s_SRC_OVER __pyx_mstate_global->__pyx_n_s_SRC_OVER
#define __pyx_n_s_STACK_BOTTOM_DISPLAY_STYLE_SHIFT __pyx_mstate_global->__pyx_n_s_STACK_BOTTOM_DISPLAY_STYLE_SHIFT
#define __pyx_n_s_STACK_BOTTOM_SHIFT_DOWN __pyx_mstate_global->__pyx_n_s_STACK_BOTTOM_SHIFT_DOWN
#define __pyx_n_s_STACK_DISPLAY_STYLE_GAP_MIN __pyx_mstate_global->__pyx_n_s_STACK_DISPLAY_STYLE_GAP_MIN
#define __pyx_n_s_STACK_GAP_MIN __pyx_mstate_global->__pyx_n_s_STACK_GAP_MIN
#define __pyx_n_s_STACK_TOP_DISPLAY_STYLE_SHIFT_UP __pyx_mstate_global->__pyx_n_s_STACK_TOP_DISPLAY_STYLE_SHIFT_UP
#define __pyx_n_s_STACK_TOP_SHIFT_UP __pyx_mstate_global->__pyx_n_s_STACK_TOP_SHIFT_UP
#define __pyx_n_s_STRETCH_STACK_BOTTOM_SHIFT_DOWN __pyx_mstate_global->__pyx_n_s_STRETCH_STACK_BOTTOM_SHIFT_DOWN
#define __pyx_n_s_STRETCH_STACK_GAP_ABOVE_MIN __pyx_mstate_global->__pyx_n_s_STRETCH_STACK_GAP_ABOVE_MIN
#define __pyx_n_s_STRETCH_STACK_GAP_BELOW_MIN __pyx_mstate_global->__pyx_n_s_STRETCH_STACK_GAP_BELOW_MIN
#define __pyx_n_s_STRETCH_STACK_TOP_SHIFT_UP __pyx_mstate_global->__pyx_n_s_STRETCH_STACK_TOP_SHIFT_UP
#define __pyx_n_s_SUBSCRIPT_BASELINE_DROP_MIN __pyx_mstate_global->__pyx_n_s_SUBSCRIPT_BASELINE_DROP_MIN
#define __pyx_n_s_SUBSCRIPT_SHIFT_DOWN __pyx_mstate_global->__pyx_n_s_SUBSCRIPT_SHIFT_DOWN
#define __pyx_n_s_SUBSCRIPT_TOP_MAX __pyx_mstate_global->__pyx_n_s_SUBSCRIPT_TOP_MAX
#define __pyx_n_s_SUB_SUPERSCRIPT_GAP_MIN __pyx_mstate_global->__pyx_n_s_SUB_SUPERSCRIPT_GAP_MIN
#define __pyx_n_s_SUPERSCRIPT_BASELINE_DROP_MAX __pyx_mstate_global->__pyx_n_s_SUPERSCRIPT_BASELINE_DROP_MAX
#define __pyx_n_s_SUPERSCRIPT_BOTTOM_MAX_WITH_SUBS __pyx_mstate_global->__pyx_n_s_SUPERSCRIPT_BOTTOM_MAX_WITH_SUBS
#define __pyx_n_s_SUPERSCRIPT_BOTTOM_MIN __pyx_mstate_global->__pyx_n_s_SUPERSCRIPT_BOTTOM_MIN
#define __pyx_n_s_SUPERSCRIPT_SHIFT_UP __pyx_mstate_global->__pyx_n_s_SUPERSCRIPT_SHIFT_UP
#define __pyx_n_s_SUPERSCRIPT_SHIFT_UP_CRAMPED __pyx_mstate_global->__pyx_n_s_SUPERSCRIPT_SHIFT_UP_CRAMPED
#define __pyx_n_s_Sequence __pyx_mstate_global->__pyx_n_s_Sequence
#define __pyx_kp_u_Set __pyx_mstate_global->__pyx_kp_u_Set
#define __pyx_n_s_SetIter __pyx_mstate_global->__pyx_n_s_SetIter
#define __pyx_n_s_SetIter___reduce_cython __pyx_mstate_global->__pyx_n_s_SetIter___reduce_cython
#define __pyx_n_s_SetIter___setstate_cython __pyx_mstate_global->__pyx_n_s_SetIter___setstate_cython
#define __pyx_n_s_Set_2 __pyx_mstate_global->__pyx_n_s_Set_2
#define __pyx_n_s_Set___copy __pyx_mstate_global->__pyx_n_s_Set___copy
#define __pyx_n_s_Set___reduce_cython __pyx_mstate_global->__pyx_n_s_Set___reduce_cython
#define __pyx_n_s_Set___setstate_cython __pyx_mstate_global->__pyx_n_s_Set___setstate_cython
#define __pyx_n_s_Set__is_equal __pyx_mstate_global->__pyx_n_s_Set__is_equal
#define __pyx_n_s_Set__set __pyx_mstate_global->__pyx_n_s_Set__set
#define __pyx_n_s_Set__update __pyx_mstate_global->__pyx_n_s_Set__update
#define __pyx_n_s_Set_add __pyx_mstate_global->__pyx_n_s_Set_add
#define __pyx_n_s_Set_add_range __pyx_mstate_global->__pyx_n_s_Set_add_range
#define __pyx_n_s_Set_clear __pyx_mstate_global->__pyx_n_s_Set_clear
#define __pyx_n_s_Set_copy __pyx_mstate_global->__pyx_n_s_Set_copy
#define __pyx_n_s_Set_del_range __pyx_mstate_global->__pyx_n_s_Set_del_range
#define __pyx_n_s_Set_difference_update __pyx_mstate_global->__pyx_n_s_Set_difference_update
#define __pyx_n_s_Set_discard __pyx_mstate_global->__pyx_n_s_Set_discard
#define __pyx_n_s_Set_intersection_update __pyx_mstate_global->__pyx_n_s_Set_intersection_update
#define __pyx_n_s_Set_invert __pyx_mstate_global->__pyx_n_s_Set_invert
#define __pyx_n_s_Set_is_inverted __pyx_mstate_global->__pyx_n_s_Set_is_inverted
#define __pyx_n_s_Set_issubset __pyx_mstate_global->__pyx_n_s_Set_issubset
#define __pyx_n_s_Set_issuperset __pyx_mstate_global->__pyx_n_s_Set_issuperset
#define __pyx_n_s_Set_remove __pyx_mstate_global->__pyx_n_s_Set_remove
#define __pyx_kp_u_Set_s __pyx_mstate_global->__pyx_kp_u_Set_s
#define __pyx_n_s_Set_set __pyx_mstate_global->__pyx_n_s_Set_set
#define __pyx_n_s_Set_symmetric_difference_update __pyx_mstate_global->__pyx_n_s_Set_symmetric_difference_update
#define __pyx_n_s_Set_update __pyx_mstate_global->__pyx_n_s_Set_update
#define __pyx_n_s_StopIteration __pyx_mstate_global->__pyx_n_s_StopIteration
#define __pyx_n_s_SubsetFlags __pyx_mstate_global->__pyx_n_s_SubsetFlags
#define __pyx_n_s_SubsetInput __pyx_mstate_global->__pyx_n_s_SubsetInput
#define __pyx_n_s_SubsetInputSets __pyx_mstate_global->__pyx_n_s_SubsetInputSets
#define __pyx_n_s_SubsetInput___reduce_cython __pyx_mstate_global->__pyx_n_s_SubsetInput___reduce_cython
#define __pyx_n_s_SubsetInput___setstate_cython __pyx_mstate_global->__pyx_n_s_SubsetInput___setstate_cython
#define __pyx_n_s_SubsetInput_keep_everything __pyx_mstate_global->__pyx_n_s_SubsetInput_keep_everything
#define __pyx_n_s_SubsetInput_pin_axis_location __pyx_mstate_global->__pyx_n_s_SubsetInput_pin_axis_location
#define __pyx_n_s_SubsetInput_pin_axis_to_default __pyx_mstate_global->__pyx_n_s_SubsetInput_pin_axis_to_default
#define __pyx_n_s_SubsetInput_sets __pyx_mstate_global->__pyx_n_s_SubsetInput_sets
#define __pyx_n_s_SubsetInput_subset __pyx_mstate_global->__pyx_n_s_SubsetInput_subset
#define __pyx_n_s_SubsetPlan __pyx_mstate_global->__pyx_n_s_SubsetPlan
#define __pyx_n_s_SubsetPlan___reduce_cython __pyx_mstate_global->__pyx_n_s_SubsetPlan___reduce_cython
#define __pyx_n_s_SubsetPlan___setstate_cython __pyx_mstate_global->__pyx_n_s_SubsetPlan___setstate_cython
#define __pyx_n_s_SubsetPlan_execute __pyx_mstate_global->__pyx_n_s_SubsetPlan_execute
#define __pyx_kp_u_Subsetting_failed __pyx_mstate_global->__pyx_kp_u_Subsetting_failed
#define __pyx_n_s_TOP_LEFT __pyx_mstate_global->__pyx_n_s_TOP_LEFT
#define __pyx_n_s_TOP_RIGHT __pyx_mstate_global->__pyx_n_s_TOP_RIGHT
#define __pyx_n_s_Tuple __pyx_mstate_global->__pyx_n_s_Tuple
#define __pyx_kp_s_Tuple_List_OTMathGlyphPart_int __pyx_mstate_global->__pyx_kp_s_Tuple_List_OTMathGlyphPart_int
#define __pyx_n_s_TypeError __pyx_mstate_global->__pyx_n_s_TypeError
#define __pyx_n_s_UNCLASSIFIED __pyx_mstate_global->__pyx_n_s_UNCLASSIFIED
#define __pyx_n_s_UNDERBAR_EXTRA_DESCENDER __pyx_mstate_global->__pyx_n_s_UNDERBAR_EXTRA_DESCENDER
#define __pyx_n_s_UNDERBAR_RULE_THICKNESS __pyx_mstate_global->__pyx_n_s_UNDERBAR_RULE_THICKNESS
#define __pyx_n_s_UNDERBAR_VERTICAL_GAP __pyx_mstate_global->__pyx_n_s_UNDERBAR_VERTICAL_GAP
#define __pyx_n_s_UNICODE __pyx_mstate_global->__pyx_n_s_UNICODE
#define __pyx_n_s_UNSAFE_TO_BREAK __pyx_mstate_global->__pyx_n_s_UNSAFE_TO_BREAK
#define __pyx_n_s_UNSAFE_TO_CONCAT __pyx_mstate_global->__pyx_n_s_UNSAFE_TO_CONCAT
#define __pyx_n_s_UPPER_LIMIT_BASELINE_RISE_MIN __pyx_mstate_global->__pyx_n_s_UPPER_LIMIT_BASELINE_RISE_MIN
#define __pyx_n_s_UPPER_LIMIT_GAP_MIN __pyx_mstate_global->__pyx_n_s_UPPER_LIMIT_GAP_MIN
#define __pyx_n_s_USABLE_WITH_DARK_BACKGROUND __pyx_mstate_global->__pyx_n_s_USABLE_WITH_DARK_BACKGROUND
#define __pyx_n_s_USABLE_WITH_LIGHT_BACKGROUND __pyx_mstate_global->__pyx_n_s_USABLE_WITH_LIGHT_BACKGROUND
#define __pyx_n_s_Union __pyx_mstate_global->__pyx_n_s_Union
#define __pyx_kp_s_Union_str_Path __pyx_mstate_global->__pyx_kp_s_Union_str_Path
#define __pyx_n_s_VERIFY __pyx_mstate_global->__pyx_n_s_VERIFY
#define __pyx_n_s_ValueError __pyx_mstate_global->__pyx_n_s_ValueError
#define __pyx_n_s_XOR __pyx_mstate_global->__pyx_n_s_XOR
#define __pyx_kp_u__10 __pyx_mstate_global->__pyx_kp_u__10
#define __pyx_kp_b__15 __pyx_mstate_global->__pyx_kp_b__15
#define __pyx_kp_u__15 __pyx_mstate_global->__pyx_kp_u__15
#define __pyx_kp_u__16 __pyx_mstate_global->__pyx_kp_u__16
#define __pyx_kp_u__19 __pyx_mstate_global->__pyx_kp_u__19
#define __pyx_kp_u__21 __pyx_mstate_global->__pyx_kp_u__21
#define __pyx_n_s__22 __pyx_mstate_global->__pyx_n_s__22
#define __pyx_kp_u__23 __pyx_mstate_global->__pyx_kp_u__23
#define __pyx_n_s__287 __pyx_mstate_global->__pyx_n_s__287
#define __pyx_n_s_a __pyx_mstate_global->__pyx_n_s_a
#define __pyx_n_s_add __pyx_mstate_global->__pyx_n_s_add
#define __pyx_n_s_add_codepoints __pyx_mstate_global->__pyx_n_s_add_codepoints
#define __pyx_n_s_add_range __pyx_mstate_global->__pyx_n_s_add_range
#define __pyx_n_s_add_str __pyx_mstate_global->__pyx_n_s_add_str
#define __pyx_n_s_add_utf8 __pyx_mstate_global->__pyx_n_s_add_utf8
#define __pyx_n_u_advance __pyx_mstate_global->__pyx_n_u_advance
#define __pyx_n_s_alpha __pyx_mstate_global->__pyx_n_s_alpha
#define __pyx_n_u_alpha __pyx_mstate_global->__pyx_n_u_alpha
#define __pyx_n_s_alternate_count __pyx_mstate_global->__pyx_n_s_alternate_count
#define __pyx_n_s_alternate_glyphs __pyx_mstate_global->__pyx_n_s_alternate_glyphs
#define __pyx_n_s_alternates __pyx_mstate_global->__pyx_n_s_alternates
#define __pyx_n_s_args __pyx_mstate_global->__pyx_n_s_args
#define __pyx_n_s_ascender __pyx_mstate_global->__pyx_n_s_ascender
#define __pyx_n_u_ascender __pyx_mstate_global->__pyx_n_u_ascender
#define __pyx_n_s_assembly __pyx_mstate_global->__pyx_n_s_assembly
#define __pyx_n_s_assembly_array __pyx_mstate_global->__pyx_n_s_assembly_array
#define __pyx_n_s_asyncio_coroutines __pyx_mstate_global->__pyx_n_s_asyncio_coroutines
#define __pyx_n_s_b __pyx_mstate_global->__pyx_n_s_b
#define __pyx_n_s_baseline_tag __pyx_mstate_global->__pyx_n_s_baseline_tag
#define __pyx_n_s_blob __pyx_mstate_global->__pyx_n_s_blob
#define __pyx_n_s_blob_data __pyx_mstate_global->__pyx_n_s_blob_data
#define __pyx_n_s_blob_length __pyx_mstate_global->__pyx_n_s_blob_length
#define __pyx_n_s_blue __pyx_mstate_global->__pyx_n_s_blue
#define __pyx_n_u_blue __pyx_mstate_global->__pyx_n_u_blue
#define __pyx_n_s_bool __pyx_mstate_global->__pyx_n_s_bool
#define __pyx_n_s_buffer __pyx_mstate_global->__pyx_n_s_buffer
#define __pyx_n_s_bytes __pyx_mstate_global->__pyx_n_s_bytes
#define __pyx_n_s_c __pyx_mstate_global->__pyx_n_s_c
#define __pyx_n_s_c_coords __pyx_mstate_global->__pyx_n_s_c_coords
#define __pyx_n_s_c_foreground __pyx_mstate_global->__pyx_n_s_c_foreground
#define __pyx_n_s_c_shapers __pyx_mstate_global->__pyx_n_s_c_shapers
#define __pyx_n_s_callback __pyx_mstate_global->__pyx_n_s_callback
#define __pyx_n_s_clear __pyx_mstate_global->__pyx_n_s_clear
#define __pyx_n_s_clear_contents __pyx_mstate_global->__pyx_n_s_clear_contents
#define __pyx_n_s_cline_in_traceback __pyx_mstate_global->__pyx_n_s_cline_in_traceback
#define __pyx_n_s_close __pyx_mstate_global->__pyx_n_s_close
#define __pyx_n_s_closePath __pyx_mstate_global->__pyx_n_s_closePath
#define __pyx_n_s_cls __pyx_mstate_global->__pyx_n_s_cls
#define __pyx_n_s_codepoints __pyx_mstate_global->__pyx_n_s_codepoints
#define __pyx_n_s_collections __pyx_mstate_global->__pyx_n_s_collections
#define __pyx_n_u_color __pyx_mstate_global->__pyx_n_u_color
#define __pyx_n_s_color_count __pyx_mstate_global->__pyx_n_s_color_count
#define __pyx_n_s_color_index __pyx_mstate_global->__pyx_n_s_color_index
#define __pyx_n_u_color_index __pyx_mstate_global->__pyx_n_u_color_index
#define __pyx_n_s_colors __pyx_mstate_global->__pyx_n_s_colors
#define __pyx_n_s_constant __pyx_mstate_global->__pyx_n_s_constant
#define __pyx_n_s_coords __pyx_mstate_global->__pyx_n_s_coords
#define __pyx_n_s_coords_2dot14 __pyx_mstate_global->__pyx_n_s_coords_2dot14
#define __pyx_n_s_copy __pyx_mstate_global->__pyx_n_s_copy
#define __pyx_n_s_copy_2 __pyx_mstate_global->__pyx_n_s_copy_2
#define __pyx_n_s_correction_height __pyx_mstate_global->__pyx_n_s_correction_height
#define __pyx_n_s_count __pyx_mstate_global->__pyx_n_s_count
#define __pyx_n_s_create __pyx_mstate_global->__pyx_n_s_create
#define __pyx_n_s_create_for_tables __pyx_mstate_global->__pyx_n_s_create_for_tables
#define __pyx_n_s_cstr __pyx_mstate_global->__pyx_n_s_cstr
#define __pyx_n_s_curveTo __pyx_mstate_global->__pyx_n_s_curveTo
#define __pyx_n_s_data __pyx_mstate_global->__pyx_n_s_data
#define __pyx_n_s_del_range __pyx_mstate_global->__pyx_n_s_del_range
#define __pyx_n_s_descender __pyx_mstate_global->__pyx_n_s_descender
#define __pyx_n_u_descender __pyx_mstate_global->__pyx_n_u_descender
#define __pyx_n_s_dict __pyx_mstate_global->__pyx_n_s_dict
#define __pyx_n_s_difference_update __pyx_mstate_global->__pyx_n_s_difference_update
#define __pyx_n_s_direction __pyx_mstate_global->__pyx_n_s_direction
#define __pyx_kp_u_disable __pyx_mstate_global->__pyx_kp_u_disable
#define __pyx_n_s_discard __pyx_mstate_global->__pyx_n_s_discard
#define __pyx_n_s_doc __pyx_mstate_global->__pyx_n_s_doc
#define __pyx_n_s_draw_data __pyx_mstate_global->__pyx_n_s_draw_data
#define __pyx_n_s_draw_funcs __pyx_mstate_global->__pyx_n_s_draw_funcs
#define __pyx_n_s_draw_glyph __pyx_mstate_global->__pyx_n_s_draw_glyph
#define __pyx_kp_u_draw_glyph_is_deprecated_use_Fon __pyx_mstate_global->__pyx_kp_u_draw_glyph_is_deprecated_use_Fon
#define __pyx_n_s_draw_glyph_with_pen __pyx_mstate_global->__pyx_n_s_draw_glyph_with_pen
#define __pyx_n_s_draw_state __pyx_mstate_global->__pyx_n_s_draw_state
#define __pyx_n_s_draw_state_p __pyx_mstate_global->__pyx_n_s_draw_state_p
#define __pyx_kp_u_enable __pyx_mstate_global->__pyx_kp_u_enable
#define __pyx_n_s_encode __pyx_mstate_global->__pyx_n_s_encode
#define __pyx_n_s_end __pyx_mstate_global->__pyx_n_s_end
#define __pyx_n_u_end_connector_length __pyx_mstate_global->__pyx_n_u_end_connector_length
#define __pyx_n_s_enum __pyx_mstate_global->__pyx_n_s_enum
#define __pyx_n_s_enumerate __pyx_mstate_global->__pyx_n_s_enumerate
#define __pyx_n_s_execute __pyx_mstate_global->__pyx_n_s_execute
#define __pyx_n_s_extents __pyx_mstate_global->__pyx_n_s_extents
#define __pyx_n_s_face __pyx_mstate_global->__pyx_n_s_face
#define __pyx_n_s_face_or_font __pyx_mstate_global->__pyx_n_s_face_or_font
#define __pyx_n_s_feat __pyx_mstate_global->__pyx_n_s_feat
#define __pyx_n_s_feature_count __pyx_mstate_global->__pyx_n_s_feature_count
#define __pyx_n_s_feature_tags __pyx_mstate_global->__pyx_n_s_feature_tags
#define __pyx_n_s_features __pyx_mstate_global->__pyx_n_s_features
#define __pyx_n_s_filename __pyx_mstate_global->__pyx_n_s_filename
#define __pyx_n_s_first __pyx_mstate_global->__pyx_n_s_first
#define __pyx_n_u_flags __pyx_mstate_global->__pyx_n_u_flags
#define __pyx_n_s_float __pyx_mstate_global->__pyx_n_s_float
#define __pyx_n_s_font __pyx_mstate_global->__pyx_n_s_font
#define __pyx_n_s_foreground __pyx_mstate_global->__pyx_n_s_foreground
#define __pyx_n_s_from_file_path __pyx_mstate_global->__pyx_n_s_from_file_path
#define __pyx_n_s_from_int __pyx_mstate_global->__pyx_n_s_from_int
#define __pyx_n_s_fsencode __pyx_mstate_global->__pyx_n_s_fsencode
#define __pyx_n_u_full_advance __pyx_mstate_global->__pyx_n_u_full_advance
#define __pyx_n_s_func __pyx_mstate_global->__pyx_n_s_func
#define __pyx_n_s_func_p __pyx_mstate_global->__pyx_n_s_func_p
#define __pyx_n_s_funcs __pyx_mstate_global->__pyx_n_s_funcs
#define __pyx_n_s_g __pyx_mstate_global->__pyx_n_s_g
#define __pyx_kp_u_gc __pyx_mstate_global->__pyx_kp_u_gc
#define __pyx_n_s_genexpr __pyx_mstate_global->__pyx_n_s_genexpr
#define __pyx_n_s_get __pyx_mstate_global->__pyx_n_s_get
#define __pyx_n_s_get_font_extents __pyx_mstate_global->__pyx_n_s_get_font_extents
#define __pyx_n_s_get_glyph_extents __pyx_mstate_global->__pyx_n_s_get_glyph_extents
#define __pyx_n_s_get_glyph_from_name __pyx_mstate_global->__pyx_n_s_get_glyph_from_name
#define __pyx_n_s_get_glyph_h_advance __pyx_mstate_global->__pyx_n_s_get_glyph_h_advance
#define __pyx_n_s_get_glyph_h_origin __pyx_mstate_global->__pyx_n_s_get_glyph_h_origin
#define __pyx_n_s_get_glyph_name __pyx_mstate_global->__pyx_n_s_get_glyph_name
#define __pyx_n_s_get_glyph_shape __pyx_mstate_global->__pyx_n_s_get_glyph_shape
#define __pyx_kp_u_get_glyph_shape_is_deprecated_us __pyx_mstate_global->__pyx_kp_u_get_glyph_shape_is_deprecated_us
#define __pyx_n_s_get_glyph_v_advance __pyx_mstate_global->__pyx_n_s_get_glyph_v_advance
#define __pyx_n_s_get_glyph_v_origin __pyx_mstate_global->__pyx_n_s_get_glyph_v_origin
#define __pyx_n_s_get_nominal_glyph __pyx_mstate_global->__pyx_n_s_get_nominal_glyph
#define __pyx_n_s_get_var_coords_design __pyx_mstate_global->__pyx_n_s_get_var_coords_design
#define __pyx_n_s_get_var_coords_normalized __pyx_mstate_global->__pyx_n_s_get_var_coords_normalized
#define __pyx_n_s_get_variation_glyph __pyx_mstate_global->__pyx_n_s_get_variation_glyph
#define __pyx_n_s_getstate __pyx_mstate_global->__pyx_n_s_getstate
#define __pyx_n_s_gid __pyx_mstate_global->__pyx_n_s_gid
#define __pyx_n_s_glyph __pyx_mstate_global->__pyx_n_s_glyph
#define __pyx_n_u_glyph __pyx_mstate_global->__pyx_n_u_glyph
#define __pyx_n_s_glyph_from_string __pyx_mstate_global->__pyx_n_s_glyph_from_string
#define __pyx_n_s_glyph_to_string __pyx_mstate_global->__pyx_n_s_glyph_to_string
#define __pyx_n_s_graphnodes __pyx_mstate_global->__pyx_n_s_graphnodes
#define __pyx_n_s_green __pyx_mstate_global->__pyx_n_s_green
#define __pyx_n_u_green __pyx_mstate_global->__pyx_n_u_green
#define __pyx_n_s_guess_segment_properties __pyx_mstate_global->__pyx_n_s_guess_segment_properties
#define __pyx_n_u_hang __pyx_mstate_global->__pyx_n_u_hang
#define __pyx_n_s_hb_baseline_tag __pyx_mstate_global->__pyx_n_s_hb_baseline_tag
#define __pyx_n_s_hb_codepoint_t __pyx_mstate_global->__pyx_n_s_hb_codepoint_t
#define __pyx_n_s_hb_codepoints __pyx_mstate_global->__pyx_n_s_hb_codepoints
#define __pyx_n_s_hb_direction __pyx_mstate_global->__pyx_n_s_hb_direction
#define __pyx_n_s_hb_features __pyx_mstate_global->__pyx_n_s_hb_features
#define __pyx_n_s_hb_language __pyx_mstate_global->__pyx_n_s_hb_language
#define __pyx_n_s_hb_language_tag __pyx_mstate_global->__pyx_n_s_hb_language_tag
#define __pyx_n_s_hb_position __pyx_mstate_global->__pyx_n_s_hb_position
#define __pyx_n_s_hb_script __pyx_mstate_global->__pyx_n_s_hb_script
#define __pyx_n_s_hb_script_tag __pyx_mstate_global->__pyx_n_s_hb_script_tag
#define __pyx_n_s_hb_tag __pyx_mstate_global->__pyx_n_s_hb_tag
#define __pyx_n_s_hb_variations __pyx_mstate_global->__pyx_n_s_hb_variations
#define __pyx_n_s_head __pyx_mstate_global->__pyx_n_s_head
#define __pyx_n_u_height __pyx_mstate_global->__pyx_n_u_height
#define __pyx_n_s_i __pyx_mstate_global->__pyx_n_s_i
#define __pyx_n_u_icfb __pyx_mstate_global->__pyx_n_u_icfb
#define __pyx_n_u_icft __pyx_mstate_global->__pyx_n_u_icft
#define __pyx_n_u_ideo __pyx_mstate_global->__pyx_n_u_ideo
#define __pyx_n_u_idtp __pyx_mstate_global->__pyx_n_u_idtp
#define __pyx_n_s_import __pyx_mstate_global->__pyx_n_s_import
#define __pyx_n_s_index __pyx_mstate_global->__pyx_n_s_index
#define __pyx_kp_u_index_out_of_range_for_feature_a __pyx_mstate_global->__pyx_kp_u_index_out_of_range_for_feature_a
#define __pyx_n_s_init __pyx_mstate_global->__pyx_n_s_init
#define __pyx_n_s_init_subclass __pyx_mstate_global->__pyx_n_s_init_subclass
#define __pyx_n_s_initializing __pyx_mstate_global->__pyx_n_s_initializing
#define __pyx_n_s_input __pyx_mstate_global->__pyx_n_s_input
#define __pyx_n_s_inst __pyx_mstate_global->__pyx_n_s_inst
#define __pyx_n_s_int __pyx_mstate_global->__pyx_n_s_int
#define __pyx_kp_s_int_None __pyx_mstate_global->__pyx_kp_s_int_None
#define __pyx_n_s_intersection_update __pyx_mstate_global->__pyx_n_s_intersection_update
#define __pyx_kp_u_invalid_baseline_tag __pyx_mstate_global->__pyx_kp_u_invalid_baseline_tag
#define __pyx_kp_u_invalid_constant __pyx_mstate_global->__pyx_kp_u_invalid_constant
#define __pyx_kp_u_invalid_kern __pyx_mstate_global->__pyx_kp_u_invalid_kern
#define __pyx_n_s_invert __pyx_mstate_global->__pyx_n_s_invert
#define __pyx_n_s_is_coroutine __pyx_mstate_global->__pyx_n_s_is_coroutine
#define __pyx_n_s_is_equal __pyx_mstate_global->__pyx_n_s_is_equal
#define __pyx_n_u_is_foreground __pyx_mstate_global->__pyx_n_u_is_foreground
#define __pyx_n_s_is_inverted __pyx_mstate_global->__pyx_n_s_is_inverted
#define __pyx_kp_u_isenabled __pyx_mstate_global->__pyx_kp_u_isenabled
#define __pyx_n_s_issubset __pyx_mstate_global->__pyx_n_s_issubset
#define __pyx_n_s_issuperset __pyx_mstate_global->__pyx_n_s_issuperset
#define __pyx_n_s_italics_correction __pyx_mstate_global->__pyx_n_s_italics_correction
#define __pyx_n_s_item_length __pyx_mstate_global->__pyx_n_s_item_length
#define __pyx_n_s_item_offset __pyx_mstate_global->__pyx_n_s_item_offset
#define __pyx_n_s_items __pyx_mstate_global->__pyx_n_s_items
#define __pyx_n_s_join __pyx_mstate_global->__pyx_n_s_join
#define __pyx_n_s_k __pyx_mstate_global->__pyx_n_s_k
#define __pyx_n_s_keep_everything __pyx_mstate_global->__pyx_n_s_keep_everything
#define __pyx_n_s_kern __pyx_mstate_global->__pyx_n_s_kern
#define __pyx_n_u_kern_value __pyx_mstate_global->__pyx_n_u_kern_value
#define __pyx_n_s_kerns __pyx_mstate_global->__pyx_n_s_kerns
#define __pyx_n_s_kerns_array __pyx_mstate_global->__pyx_n_s_kerns_array
#define __pyx_n_s_keys __pyx_mstate_global->__pyx_n_s_keys
#define __pyx_n_s_keys_locals_genexpr __pyx_mstate_global->__pyx_n_s_keys_locals_genexpr
#define __pyx_n_s_kind __pyx_mstate_global->__pyx_n_s_kind
#define __pyx_n_s_language_count __pyx_mstate_global->__pyx_n_s_language_count
#define __pyx_n_s_language_index __pyx_mstate_global->__pyx_n_s_language_index
#define __pyx_n_s_language_tag __pyx_mstate_global->__pyx_n_s_language_tag
#define __pyx_n_s_language_tags __pyx_mstate_global->__pyx_n_s_language_tags
#define __pyx_n_s_larger_set __pyx_mstate_global->__pyx_n_s_larger_set
#define __pyx_n_s_last __pyx_mstate_global->__pyx_n_s_last
#define __pyx_n_s_layer_count __pyx_mstate_global->__pyx_n_s_layer_count
#define __pyx_n_s_layers __pyx_mstate_global->__pyx_n_s_layers
#define __pyx_n_s_length __pyx_mstate_global->__pyx_n_s_length
#define __pyx_n_s_lineTo __pyx_mstate_global->__pyx_n_s_lineTo
#define __pyx_n_s_line_gap __pyx_mstate_global->__pyx_n_s_line_gap
#define __pyx_n_u_line_gap __pyx_mstate_global->__pyx_n_u_line_gap
#define __pyx_n_s_lookup_index __pyx_mstate_global->__pyx_n_s_lookup_index
#define __pyx_n_s_m __pyx_mstate_global->__pyx_n_s_m
#define __pyx_n_s_main __pyx_mstate_global->__pyx_n_s_main
#define __pyx_n_u_math __pyx_mstate_global->__pyx_n_u_math
#define __pyx_n_u_max_correction_height __pyx_mstate_global->__pyx_n_u_max_correction_height
#define __pyx_n_s_metaclass __pyx_mstate_global->__pyx_n_s_metaclass
#define __pyx_n_s_methods __pyx_mstate_global->__pyx_n_s_methods
#define __pyx_n_s_module __pyx_mstate_global->__pyx_n_s_module
#define __pyx_n_s_moveTo __pyx_mstate_global->__pyx_n_s_moveTo
#define __pyx_n_s_mro_entries __pyx_mstate_global->__pyx_n_s_mro_entries
#define __pyx_n_s_name __pyx_mstate_global->__pyx_n_s_name
#define __pyx_n_s_name_2 __pyx_mstate_global->__pyx_n_s_name_2
#define __pyx_n_s_name_id __pyx_mstate_global->__pyx_n_s_name_id
#define __pyx_n_s_namedtuple __pyx_mstate_global->__pyx_n_s_namedtuple
#define __pyx_n_s_new_face __pyx_mstate_global->__pyx_n_s_new_face
#define __pyx_kp_s_no_default___reduce___due_to_non __pyx_mstate_global->__pyx_kp_s_no_default___reduce___due_to_non
#define __pyx_n_s_node __pyx_mstate_global->__pyx_n_s_node
#define __pyx_n_s_num_nodes __pyx_mstate_global->__pyx_n_s_num_nodes
#define __pyx_kp_u_num_of_graph_nodes __pyx_mstate_global->__pyx_kp_u_num_of_graph_nodes
#define __pyx_n_s_obj_list __pyx_mstate_global->__pyx_n_s_obj_list
#define __pyx_n_s_object __pyx_mstate_global->__pyx_n_s_object
#define __pyx_n_u_offset __pyx_mstate_global->__pyx_n_u_offset
#define __pyx_n_s_os __pyx_mstate_global->__pyx_n_s_os
#define __pyx_n_s_ot_color_glyph_get_layers __pyx_mstate_global->__pyx_n_s_ot_color_glyph_get_layers
#define __pyx_n_s_ot_color_glyph_get_png __pyx_mstate_global->__pyx_n_s_ot_color_glyph_get_png
#define __pyx_n_s_ot_color_glyph_get_svg __pyx_mstate_global->__pyx_n_s_ot_color_glyph_get_svg
#define __pyx_n_s_ot_color_glyph_has_paint __pyx_mstate_global->__pyx_n_s_ot_color_glyph_has_paint
#define __pyx_n_s_ot_color_has_layers __pyx_mstate_global->__pyx_n_s_ot_color_has_layers
#define __pyx_n_s_ot_color_has_paint __pyx_mstate_global->__pyx_n_s_ot_color_has_paint
#define __pyx_n_s_ot_color_has_palettes __pyx_mstate_global->__pyx_n_s_ot_color_has_palettes
#define __pyx_n_s_ot_color_has_png __pyx_mstate_global->__pyx_n_s_ot_color_has_png
#define __pyx_n_s_ot_color_has_svg __pyx_mstate_global->__pyx_n_s_ot_color_has_svg
#define __pyx_n_s_ot_color_palette_color_get_name __pyx_mstate_global->__pyx_n_s_ot_color_palette_color_get_name
#define __pyx_n_s_ot_color_palette_get_colors __pyx_mstate_global->__pyx_n_s_ot_color_palette_get_colors
#define __pyx_n_s_ot_color_palette_get_count __pyx_mstate_global->__pyx_n_s_ot_color_palette_get_count
#define __pyx_n_s_ot_color_palette_get_flags __pyx_mstate_global->__pyx_n_s_ot_color_palette_get_flags
#define __pyx_n_s_ot_color_palette_get_name_id __pyx_mstate_global->__pyx_n_s_ot_color_palette_get_name_id
#define __pyx_n_s_ot_font_set_funcs __pyx_mstate_global->__pyx_n_s_ot_font_set_funcs
#define __pyx_n_s_ot_layout_get_baseline __pyx_mstate_global->__pyx_n_s_ot_layout_get_baseline
#define __pyx_n_s_ot_layout_get_glyph_class __pyx_mstate_global->__pyx_n_s_ot_layout_get_glyph_class
#define __pyx_n_s_ot_layout_has_glyph_classes __pyx_mstate_global->__pyx_n_s_ot_layout_has_glyph_classes
#define __pyx_n_s_ot_layout_has_positioning __pyx_mstate_global->__pyx_n_s_ot_layout_has_positioning
#define __pyx_n_s_ot_layout_has_substitution __pyx_mstate_global->__pyx_n_s_ot_layout_has_substitution
#define __pyx_n_s_ot_layout_language_get_feature_t __pyx_mstate_global->__pyx_n_s_ot_layout_language_get_feature_t
#define __pyx_n_s_ot_layout_lookup_get_glyph_alter __pyx_mstate_global->__pyx_n_s_ot_layout_lookup_get_glyph_alter
#define __pyx_n_s_ot_layout_script_get_language_ta __pyx_mstate_global->__pyx_n_s_ot_layout_script_get_language_ta
#define __pyx_n_s_ot_layout_table_get_script_tags __pyx_mstate_global->__pyx_n_s_ot_layout_table_get_script_tags
#define __pyx_n_s_ot_math_get_constant __pyx_mstate_global->__pyx_n_s_ot_math_get_constant
#define __pyx_n_s_ot_math_get_glyph_assembly __pyx_mstate_global->__pyx_n_s_ot_math_get_glyph_assembly
#define __pyx_n_s_ot_math_get_glyph_italics_correc __pyx_mstate_global->__pyx_n_s_ot_math_get_glyph_italics_correc
#define __pyx_n_s_ot_math_get_glyph_kerning __pyx_mstate_global->__pyx_n_s_ot_math_get_glyph_kerning
#define __pyx_n_s_ot_math_get_glyph_kernings __pyx_mstate_global->__pyx_n_s_ot_math_get_glyph_kernings
#define __pyx_n_s_ot_math_get_glyph_top_accent_att __pyx_mstate_global->__pyx_n_s_ot_math_get_glyph_top_accent_att
#define __pyx_n_s_ot_math_get_glyph_variants __pyx_mstate_global->__pyx_n_s_ot_math_get_glyph_variants
#define __pyx_n_s_ot_math_get_min_connector_overla __pyx_mstate_global->__pyx_n_s_ot_math_get_min_connector_overla
#define __pyx_n_s_ot_math_has_data __pyx_mstate_global->__pyx_n_s_ot_math_has_data
#define __pyx_n_s_ot_math_is_glyph_extended_shape __pyx_mstate_global->__pyx_n_s_ot_math_is_glyph_extended_shape
#define __pyx_n_s_ot_tag_to_language __pyx_mstate_global->__pyx_n_s_ot_tag_to_language
#define __pyx_n_s_ot_tag_to_script __pyx_mstate_global->__pyx_n_s_ot_tag_to_script
#define __pyx_n_s_other __pyx_mstate_global->__pyx_n_s_other
#define __pyx_n_s_p __pyx_mstate_global->__pyx_n_s_p
#define __pyx_n_s_packed __pyx_mstate_global->__pyx_n_s_packed
#define __pyx_n_s_packed_blob __pyx_mstate_global->__pyx_n_s_packed_blob
#define __pyx_n_s_paint_funcs __pyx_mstate_global->__pyx_n_s_paint_funcs
#define __pyx_n_s_paint_glyph __pyx_mstate_global->__pyx_n_s_paint_glyph
#define __pyx_n_s_paint_state __pyx_mstate_global->__pyx_n_s_paint_state
#define __pyx_n_s_paint_state_p __pyx_mstate_global->__pyx_n_s_paint_state_p
#define __pyx_n_s_palette_index __pyx_mstate_global->__pyx_n_s_palette_index
#define __pyx_n_s_pathlib __pyx_mstate_global->__pyx_n_s_pathlib
#define __pyx_n_s_pen __pyx_mstate_global->__pyx_n_s_pen
#define __pyx_n_s_pin_axis_location __pyx_mstate_global->__pyx_n_s_pin_axis_location
#define __pyx_n_s_pin_axis_to_default __pyx_mstate_global->__pyx_n_s_pin_axis_to_default
#define __pyx_n_s_prepare __pyx_mstate_global->__pyx_n_s_prepare
#define __pyx_n_s_pyx_state __pyx_mstate_global->__pyx_n_s_pyx_state
#define __pyx_n_s_pyx_vtable __pyx_mstate_global->__pyx_n_s_pyx_vtable
#define __pyx_n_s_qCurveTo __pyx_mstate_global->__pyx_n_s_qCurveTo
#define __pyx_n_s_qualname __pyx_mstate_global->__pyx_n_s_qualname
#define __pyx_n_s_r __pyx_mstate_global->__pyx_n_s_r
#define __pyx_n_s_range __pyx_mstate_global->__pyx_n_s_range
#define __pyx_n_s_red __pyx_mstate_global->__pyx_n_s_red
#define __pyx_n_u_red __pyx_mstate_global->__pyx_n_u_red
#define __pyx_n_s_reduce __pyx_mstate_global->__pyx_n_s_reduce
#define __pyx_n_s_reduce_cython __pyx_mstate_global->__pyx_n_s_reduce_cython
#define __pyx_n_s_reduce_ex __pyx_mstate_global->__pyx_n_s_reduce_ex
#define __pyx_n_s_remove __pyx_mstate_global->__pyx_n_s_remove
#define __pyx_n_s_repack __pyx_mstate_global->__pyx_n_s_repack
#define __pyx_n_s_repack_with_tag __pyx_mstate_global->__pyx_n_s_repack_with_tag
#define __pyx_n_s_repr___locals_genexpr __pyx_mstate_global->__pyx_n_s_repr___locals_genexpr
#define __pyx_n_s_reset __pyx_mstate_global->__pyx_n_s_reset
#define __pyx_n_s_ret __pyx_mstate_global->__pyx_n_s_ret
#define __pyx_n_s_return __pyx_mstate_global->__pyx_n_s_return
#define __pyx_n_u_romn __pyx_mstate_global->__pyx_n_u_romn
#define __pyx_n_s_round __pyx_mstate_global->__pyx_n_s_round
#define __pyx_n_s_s __pyx_mstate_global->__pyx_n_s_s
#define __pyx_n_s_script_count __pyx_mstate_global->__pyx_n_s_script_count
#define __pyx_n_s_script_index __pyx_mstate_global->__pyx_n_s_script_index
#define __pyx_n_s_script_tag __pyx_mstate_global->__pyx_n_s_script_tag
#define __pyx_n_s_script_tags __pyx_mstate_global->__pyx_n_s_script_tags
#define __pyx_n_s_self __pyx_mstate_global->__pyx_n_s_self
#define __pyx_n_s_send __pyx_mstate_global->__pyx_n_s_send
#define __pyx_n_s_set __pyx_mstate_global->__pyx_n_s_set
#define __pyx_n_s_set_2 __pyx_mstate_global->__pyx_n_s_set_2
#define __pyx_n_s_set_close_path_func __pyx_mstate_global->__pyx_n_s_set_close_path_func
#define __pyx_n_s_set_color_func __pyx_mstate_global->__pyx_n_s_set_color_func
#define __pyx_n_s_set_color_glyph_func __pyx_mstate_global->__pyx_n_s_set_color_glyph_func
#define __pyx_n_s_set_cubic_to_func __pyx_mstate_global->__pyx_n_s_set_cubic_to_func
#define __pyx_n_s_set_custom_palette_color_func __pyx_mstate_global->__pyx_n_s_set_custom_palette_color_func
#define __pyx_n_s_set_font_h_extents_func __pyx_mstate_global->__pyx_n_s_set_font_h_extents_func
#define __pyx_n_s_set_font_v_extents_func __pyx_mstate_global->__pyx_n_s_set_font_v_extents_func
#define __pyx_n_s_set_glyph_h_advance_func __pyx_mstate_global->__pyx_n_s_set_glyph_h_advance_func
#define __pyx_n_s_set_glyph_name_func __pyx_mstate_global->__pyx_n_s_set_glyph_name_func
#define __pyx_n_s_set_glyph_v_advance_func __pyx_mstate_global->__pyx_n_s_set_glyph_v_advance_func
#define __pyx_n_s_set_glyph_v_origin_func __pyx_mstate_global->__pyx_n_s_set_glyph_v_origin_func
#define __pyx_n_s_set_image_func __pyx_mstate_global->__pyx_n_s_set_image_func
#define __pyx_n_s_set_language_from_ot_tag __pyx_mstate_global->__pyx_n_s_set_language_from_ot_tag
#define __pyx_n_s_set_line_to_func __pyx_mstate_global->__pyx_n_s_set_line_to_func
#define __pyx_n_s_set_linear_gradient_func __pyx_mstate_global->__pyx_n_s_set_linear_gradient_func
#define __pyx_n_s_set_message_func __pyx_mstate_global->__pyx_n_s_set_message_func
#define __pyx_n_s_set_move_to_func __pyx_mstate_global->__pyx_n_s_set_move_to_func
#define __pyx_n_s_set_name __pyx_mstate_global->__pyx_n_s_set_name
#define __pyx_n_s_set_nominal_glyph_func __pyx_mstate_global->__pyx_n_s_set_nominal_glyph_func
#define __pyx_n_s_set_pop_clip_func __pyx_mstate_global->__pyx_n_s_set_pop_clip_func
#define __pyx_n_s_set_pop_group_func __pyx_mstate_global->__pyx_n_s_set_pop_group_func
#define __pyx_n_s_set_pop_transform_func __pyx_mstate_global->__pyx_n_s_set_pop_transform_func
#define __pyx_n_s_set_push_clip_glyph_func __pyx_mstate_global->__pyx_n_s_set_push_clip_glyph_func
#define __pyx_n_s_set_push_clip_rectangle_func __pyx_mstate_global->__pyx_n_s_set_push_clip_rectangle_func
#define __pyx_n_s_set_push_group_func __pyx_mstate_global->__pyx_n_s_set_push_group_func
#define __pyx_n_s_set_push_transform_func __pyx_mstate_global->__pyx_n_s_set_push_transform_func
#define __pyx_n_s_set_quadratic_to_func __pyx_mstate_global->__pyx_n_s_set_quadratic_to_func
#define __pyx_n_s_set_radial_gradient_func __pyx_mstate_global->__pyx_n_s_set_radial_gradient_func
#define __pyx_n_s_set_script_from_ot_tag __pyx_mstate_global->__pyx_n_s_set_script_from_ot_tag
#define __pyx_n_s_set_sweep_gradient_func __pyx_mstate_global->__pyx_n_s_set_sweep_gradient_func
#define __pyx_n_s_set_type __pyx_mstate_global->__pyx_n_s_set_type
#define __pyx_n_s_set_var_coords_design __pyx_mstate_global->__pyx_n_s_set_var_coords_design
#define __pyx_n_s_set_var_coords_normalized __pyx_mstate_global->__pyx_n_s_set_var_coords_normalized
#define __pyx_n_s_set_variation __pyx_mstate_global->__pyx_n_s_set_variation
#define __pyx_n_s_set_variation_glyph_func __pyx_mstate_global->__pyx_n_s_set_variation_glyph_func
#define __pyx_n_s_set_variations __pyx_mstate_global->__pyx_n_s_set_variations
#define __pyx_n_s_sets __pyx_mstate_global->__pyx_n_s_sets
#define __pyx_n_s_setstate __pyx_mstate_global->__pyx_n_s_setstate
#define __pyx_n_s_setstate_cython __pyx_mstate_global->__pyx_n_s_setstate_cython
#define __pyx_n_s_shape __pyx_mstate_global->__pyx_n_s_shape
#define __pyx_n_s_shaper __pyx_mstate_global->__pyx_n_s_shaper
#define __pyx_n_s_shapers __pyx_mstate_global->__pyx_n_s_shapers
#define __pyx_n_s_size __pyx_mstate_global->__pyx_n_s_size
#define __pyx_n_s_smaller_set __pyx_mstate_global->__pyx_n_s_smaller_set
#define __pyx_n_s_source __pyx_mstate_global->__pyx_n_s_source
#define __pyx_n_s_spec __pyx_mstate_global->__pyx_n_s_spec
#define __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx __pyx_mstate_global->__pyx_kp_s_src_uharfbuzz__harfbuzz_pyx
#define __pyx_n_s_start __pyx_mstate_global->__pyx_n_s_start
#define __pyx_n_u_start_connector_length __pyx_mstate_global->__pyx_n_u_start_connector_length
#define __pyx_n_s_start_offset __pyx_mstate_global->__pyx_n_s_start_offset
#define __pyx_n_s_staticmethod __pyx_mstate_global->__pyx_n_s_staticmethod
#define __pyx_n_s_str __pyx_mstate_global->__pyx_n_s_str
#define __pyx_n_s_string __pyx_mstate_global->__pyx_n_s_string
#define __pyx_kp_s_stringsource __pyx_mstate_global->__pyx_kp_s_stringsource
#define __pyx_n_s_subset __pyx_mstate_global->__pyx_n_s_subset
#define __pyx_n_s_subset_preprocess __pyx_mstate_global->__pyx_n_s_subset_preprocess
#define __pyx_n_s_subtables __pyx_mstate_global->__pyx_n_s_subtables
#define __pyx_n_s_success __pyx_mstate_global->__pyx_n_s_success
#define __pyx_n_s_super __pyx_mstate_global->__pyx_n_s_super
#define __pyx_n_s_symmetric_difference_update __pyx_mstate_global->__pyx_n_s_symmetric_difference_update
#define __pyx_n_s_table_bytes __pyx_mstate_global->__pyx_n_s_table_bytes
#define __pyx_n_s_table_data __pyx_mstate_global->__pyx_n_s_table_data
#define __pyx_n_s_tag __pyx_mstate_global->__pyx_n_s_tag
#define __pyx_n_s_tag_packed __pyx_mstate_global->__pyx_n_s_tag_packed
#define __pyx_n_s_tags __pyx_mstate_global->__pyx_n_s_tags
#define __pyx_n_s_tail __pyx_mstate_global->__pyx_n_s_tail
#define __pyx_n_s_test __pyx_mstate_global->__pyx_n_s_test
#define __pyx_n_s_text __pyx_mstate_global->__pyx_n_s_text
#define __pyx_n_s_throw __pyx_mstate_global->__pyx_n_s_throw
#define __pyx_n_s_to_int __pyx_mstate_global->__pyx_n_s_to_int
#define __pyx_n_s_typing __pyx_mstate_global->__pyx_n_s_typing
#define __pyx_n_s_uharfbuzz__harfbuzz __pyx_mstate_global->__pyx_n_s_uharfbuzz__harfbuzz
#define __pyx_n_s_unicode __pyx_mstate_global->__pyx_n_s_unicode
#define __pyx_n_s_update __pyx_mstate_global->__pyx_n_s_update
#define __pyx_n_s_update_2 __pyx_mstate_global->__pyx_n_s_update_2
#define __pyx_n_s_user_data __pyx_mstate_global->__pyx_n_s_user_data
#define __pyx_n_s_user_data_p __pyx_mstate_global->__pyx_n_s_user_data_p
#define __pyx_n_s_v __pyx_mstate_global->__pyx_n_s_v
#define __pyx_n_s_value __pyx_mstate_global->__pyx_n_s_value
#define __pyx_n_s_values __pyx_mstate_global->__pyx_n_s_values
#define __pyx_n_s_values_locals_genexpr __pyx_mstate_global->__pyx_n_s_values_locals_genexpr
#define __pyx_n_s_variants __pyx_mstate_global->__pyx_n_s_variants
#define __pyx_n_s_variants_array __pyx_mstate_global->__pyx_n_s_variants_array
#define __pyx_n_s_variation __pyx_mstate_global->__pyx_n_s_variation
#define __pyx_n_s_variation_selector __pyx_mstate_global->__pyx_n_s_variation_selector
#define __pyx_n_s_variation_unicodes __pyx_mstate_global->__pyx_n_s_variation_unicodes
#define __pyx_n_s_variations __pyx_mstate_global->__pyx_n_s_variations
#define __pyx_n_s_version_string __pyx_mstate_global->__pyx_n_s_version_string
#define __pyx_n_s_warn __pyx_mstate_global->__pyx_n_s_warn
#define __pyx_n_s_warnings __pyx_mstate_global->__pyx_n_s_warnings
#define __pyx_n_u_width __pyx_mstate_global->__pyx_n_u_width
#define __pyx_n_s_x __pyx_mstate_global->__pyx_n_s_x
#define __pyx_n_u_x_bearing __pyx_mstate_global->__pyx_n_u_x_bearing
#define __pyx_n_s_y __pyx_mstate_global->__pyx_n_s_y
#define __pyx_n_u_y_bearing __pyx_mstate_global->__pyx_n_u_y_bearing
#define __pyx_int_0 __pyx_mstate_global->__pyx_int_0
#define __pyx_int_1 __pyx_mstate_global->__pyx_int_1
#define __pyx_int_16384 __pyx_mstate_global->__pyx_int_16384
#define __pyx_int_65535 __pyx_mstate_global->__pyx_int_65535
#define __pyx_int_neg_1 __pyx_mstate_global->__pyx_int_neg_1
#define __pyx_k_ __pyx_mstate_global->__pyx_k_
#define __pyx_k__2 __pyx_mstate_global->__pyx_k__2
#define __pyx_k__3 __pyx_mstate_global->__pyx_k__3
#define __pyx_k__4 __pyx_mstate_global->__pyx_k__4
#define __pyx_k__5 __pyx_mstate_global->__pyx_k__5
#define __pyx_k__6 __pyx_mstate_global->__pyx_k__6
#define __pyx_k__8 __pyx_mstate_global->__pyx_k__8
#define __pyx_k__18 __pyx_mstate_global->__pyx_k__18
#define __pyx_k__20 __pyx_mstate_global->__pyx_k__20
#define __pyx_tuple__7 __pyx_mstate_global->__pyx_tuple__7
#define __pyx_tuple__9 __pyx_mstate_global->__pyx_tuple__9
#define __pyx_tuple__11 __pyx_mstate_global->__pyx_tuple__11
#define __pyx_tuple__12 __pyx_mstate_global->__pyx_tuple__12
#define __pyx_tuple__13 __pyx_mstate_global->__pyx_tuple__13
#define __pyx_tuple__14 __pyx_mstate_global->__pyx_tuple__14
#define __pyx_tuple__17 __pyx_mstate_global->__pyx_tuple__17
#define __pyx_tuple__24 __pyx_mstate_global->__pyx_tuple__24
#define __pyx_tuple__26 __pyx_mstate_global->__pyx_tuple__26
#define __pyx_tuple__28 __pyx_mstate_global->__pyx_tuple__28
#define __pyx_tuple__32 __pyx_mstate_global->__pyx_tuple__32
#define __pyx_tuple__36 __pyx_mstate_global->__pyx_tuple__36
#define __pyx_tuple__39 __pyx_mstate_global->__pyx_tuple__39
#define __pyx_tuple__41 __pyx_mstate_global->__pyx_tuple__41
#define __pyx_tuple__43 __pyx_mstate_global->__pyx_tuple__43
#define __pyx_tuple__46 __pyx_mstate_global->__pyx_tuple__46
#define __pyx_tuple__50 __pyx_mstate_global->__pyx_tuple__50
#define __pyx_tuple__54 __pyx_mstate_global->__pyx_tuple__54
#define __pyx_tuple__56 __pyx_mstate_global->__pyx_tuple__56
#define __pyx_tuple__57 __pyx_mstate_global->__pyx_tuple__57
#define __pyx_tuple__59 __pyx_mstate_global->__pyx_tuple__59
#define __pyx_tuple__63 __pyx_mstate_global->__pyx_tuple__63
#define __pyx_tuple__65 __pyx_mstate_global->__pyx_tuple__65
#define __pyx_tuple__67 __pyx_mstate_global->__pyx_tuple__67
#define __pyx_tuple__69 __pyx_mstate_global->__pyx_tuple__69
#define __pyx_tuple__71 __pyx_mstate_global->__pyx_tuple__71
#define __pyx_tuple__73 __pyx_mstate_global->__pyx_tuple__73
#define __pyx_tuple__75 __pyx_mstate_global->__pyx_tuple__75
#define __pyx_tuple__78 __pyx_mstate_global->__pyx_tuple__78
#define __pyx_tuple__81 __pyx_mstate_global->__pyx_tuple__81
#define __pyx_tuple__83 __pyx_mstate_global->__pyx_tuple__83
#define __pyx_tuple__85 __pyx_mstate_global->__pyx_tuple__85
#define __pyx_tuple__87 __pyx_mstate_global->__pyx_tuple__87
#define __pyx_tuple__89 __pyx_mstate_global->__pyx_tuple__89
#define __pyx_tuple__92 __pyx_mstate_global->__pyx_tuple__92
#define __pyx_tuple__94 __pyx_mstate_global->__pyx_tuple__94
#define __pyx_tuple__96 __pyx_mstate_global->__pyx_tuple__96
#define __pyx_tuple__98 __pyx_mstate_global->__pyx_tuple__98
#define __pyx_tuple__100 __pyx_mstate_global->__pyx_tuple__100
#define __pyx_tuple__102 __pyx_mstate_global->__pyx_tuple__102
#define __pyx_tuple__107 __pyx_mstate_global->__pyx_tuple__107
#define __pyx_tuple__118 __pyx_mstate_global->__pyx_tuple__118
#define __pyx_tuple__120 __pyx_mstate_global->__pyx_tuple__120
#define __pyx_tuple__121 __pyx_mstate_global->__pyx_tuple__121
#define __pyx_tuple__123 __pyx_mstate_global->__pyx_tuple__123
#define __pyx_tuple__125 __pyx_mstate_global->__pyx_tuple__125
#define __pyx_tuple__127 __pyx_mstate_global->__pyx_tuple__127
#define __pyx_tuple__129 __pyx_mstate_global->__pyx_tuple__129
#define __pyx_tuple__131 __pyx_mstate_global->__pyx_tuple__131
#define __pyx_tuple__133 __pyx_mstate_global->__pyx_tuple__133
#define __pyx_tuple__135 __pyx_mstate_global->__pyx_tuple__135
#define __pyx_tuple__139 __pyx_mstate_global->__pyx_tuple__139
#define __pyx_tuple__143 __pyx_mstate_global->__pyx_tuple__143
#define __pyx_tuple__145 __pyx_mstate_global->__pyx_tuple__145
#define __pyx_tuple__147 __pyx_mstate_global->__pyx_tuple__147
#define __pyx_tuple__149 __pyx_mstate_global->__pyx_tuple__149
#define __pyx_tuple__152 __pyx_mstate_global->__pyx_tuple__152
#define __pyx_tuple__157 __pyx_mstate_global->__pyx_tuple__157
#define __pyx_tuple__160 __pyx_mstate_global->__pyx_tuple__160
#define __pyx_tuple__163 __pyx_mstate_global->__pyx_tuple__163
#define __pyx_tuple__165 __pyx_mstate_global->__pyx_tuple__165
#define __pyx_tuple__169 __pyx_mstate_global->__pyx_tuple__169
#define __pyx_tuple__171 __pyx_mstate_global->__pyx_tuple__171
#define __pyx_tuple__173 __pyx_mstate_global->__pyx_tuple__173
#define __pyx_tuple__175 __pyx_mstate_global->__pyx_tuple__175
#define __pyx_tuple__177 __pyx_mstate_global->__pyx_tuple__177
#define __pyx_tuple__179 __pyx_mstate_global->__pyx_tuple__179
#define __pyx_tuple__182 __pyx_mstate_global->__pyx_tuple__182
#define __pyx_tuple__186 __pyx_mstate_global->__pyx_tuple__186
#define __pyx_tuple__203 __pyx_mstate_global->__pyx_tuple__203
#define __pyx_tuple__205 __pyx_mstate_global->__pyx_tuple__205
#define __pyx_tuple__207 __pyx_mstate_global->__pyx_tuple__207
#define __pyx_tuple__217 __pyx_mstate_global->__pyx_tuple__217
#define __pyx_tuple__219 __pyx_mstate_global->__pyx_tuple__219
#define __pyx_tuple__221 __pyx_mstate_global->__pyx_tuple__221
#define __pyx_tuple__223 __pyx_mstate_global->__pyx_tuple__223
#define __pyx_tuple__225 __pyx_mstate_global->__pyx_tuple__225
#define __pyx_tuple__228 __pyx_mstate_global->__pyx_tuple__228
#define __pyx_tuple__230 __pyx_mstate_global->__pyx_tuple__230
#define __pyx_tuple__232 __pyx_mstate_global->__pyx_tuple__232
#define __pyx_tuple__236 __pyx_mstate_global->__pyx_tuple__236
#define __pyx_tuple__240 __pyx_mstate_global->__pyx_tuple__240
#define __pyx_tuple__247 __pyx_mstate_global->__pyx_tuple__247
#define __pyx_tuple__252 __pyx_mstate_global->__pyx_tuple__252
#define __pyx_tuple__254 __pyx_mstate_global->__pyx_tuple__254
#define __pyx_tuple__256 __pyx_mstate_global->__pyx_tuple__256
#define __pyx_tuple__259 __pyx_mstate_global->__pyx_tuple__259
#define __pyx_tuple__273 __pyx_mstate_global->__pyx_tuple__273
#define __pyx_tuple__277 __pyx_mstate_global->__pyx_tuple__277
#define __pyx_tuple__280 __pyx_mstate_global->__pyx_tuple__280
#define __pyx_codeobj__25 __pyx_mstate_global->__pyx_codeobj__25
#define __pyx_codeobj__27 __pyx_mstate_global->__pyx_codeobj__27
#define __pyx_codeobj__29 __pyx_mstate_global->__pyx_codeobj__29
#define __pyx_codeobj__30 __pyx_mstate_global->__pyx_codeobj__30
#define __pyx_codeobj__31 __pyx_mstate_global->__pyx_codeobj__31
#define __pyx_codeobj__33 __pyx_mstate_global->__pyx_codeobj__33
#define __pyx_codeobj__34 __pyx_mstate_global->__pyx_codeobj__34
#define __pyx_codeobj__35 __pyx_mstate_global->__pyx_codeobj__35
#define __pyx_codeobj__37 __pyx_mstate_global->__pyx_codeobj__37
#define __pyx_codeobj__38 __pyx_mstate_global->__pyx_codeobj__38
#define __pyx_codeobj__40 __pyx_mstate_global->__pyx_codeobj__40
#define __pyx_codeobj__42 __pyx_mstate_global->__pyx_codeobj__42
#define __pyx_codeobj__44 __pyx_mstate_global->__pyx_codeobj__44
#define __pyx_codeobj__45 __pyx_mstate_global->__pyx_codeobj__45
#define __pyx_codeobj__47 __pyx_mstate_global->__pyx_codeobj__47
#define __pyx_codeobj__48 __pyx_mstate_global->__pyx_codeobj__48
#define __pyx_codeobj__49 __pyx_mstate_global->__pyx_codeobj__49
#define __pyx_codeobj__51 __pyx_mstate_global->__pyx_codeobj__51
#define __pyx_codeobj__52 __pyx_mstate_global->__pyx_codeobj__52
#define __pyx_codeobj__53 __pyx_mstate_global->__pyx_codeobj__53
#define __pyx_codeobj__55 __pyx_mstate_global->__pyx_codeobj__55
#define __pyx_codeobj__58 __pyx_mstate_global->__pyx_codeobj__58
#define __pyx_codeobj__60 __pyx_mstate_global->__pyx_codeobj__60
#define __pyx_codeobj__61 __pyx_mstate_global->__pyx_codeobj__61
#define __pyx_codeobj__62 __pyx_mstate_global->__pyx_codeobj__62
#define __pyx_codeobj__64 __pyx_mstate_global->__pyx_codeobj__64
#define __pyx_codeobj__66 __pyx_mstate_global->__pyx_codeobj__66
#define __pyx_codeobj__68 __pyx_mstate_global->__pyx_codeobj__68
#define __pyx_codeobj__70 __pyx_mstate_global->__pyx_codeobj__70
#define __pyx_codeobj__72 __pyx_mstate_global->__pyx_codeobj__72
#define __pyx_codeobj__74 __pyx_mstate_global->__pyx_codeobj__74
#define __pyx_codeobj__76 __pyx_mstate_global->__pyx_codeobj__76
#define __pyx_codeobj__77 __pyx_mstate_global->__pyx_codeobj__77
#define __pyx_codeobj__79 __pyx_mstate_global->__pyx_codeobj__79
#define __pyx_codeobj__80 __pyx_mstate_global->__pyx_codeobj__80
#define __pyx_codeobj__82 __pyx_mstate_global->__pyx_codeobj__82
#define __pyx_codeobj__84 __pyx_mstate_global->__pyx_codeobj__84
#define __pyx_codeobj__86 __pyx_mstate_global->__pyx_codeobj__86
#define __pyx_codeobj__88 __pyx_mstate_global->__pyx_codeobj__88
#define __pyx_codeobj__90 __pyx_mstate_global->__pyx_codeobj__90
#define __pyx_codeobj__91 __pyx_mstate_global->__pyx_codeobj__91
#define __pyx_codeobj__93 __pyx_mstate_global->__pyx_codeobj__93
#define __pyx_codeobj__95 __pyx_mstate_global->__pyx_codeobj__95
#define __pyx_codeobj__97 __pyx_mstate_global->__pyx_codeobj__97
#define __pyx_codeobj__99 __pyx_mstate_global->__pyx_codeobj__99
#define __pyx_codeobj__101 __pyx_mstate_global->__pyx_codeobj__101
#define __pyx_codeobj__103 __pyx_mstate_global->__pyx_codeobj__103
#define __pyx_codeobj__104 __pyx_mstate_global->__pyx_codeobj__104
#define __pyx_codeobj__105 __pyx_mstate_global->__pyx_codeobj__105
#define __pyx_codeobj__106 __pyx_mstate_global->__pyx_codeobj__106
#define __pyx_codeobj__108 __pyx_mstate_global->__pyx_codeobj__108
#define __pyx_codeobj__109 __pyx_mstate_global->__pyx_codeobj__109
#define __pyx_codeobj__110 __pyx_mstate_global->__pyx_codeobj__110
#define __pyx_codeobj__111 __pyx_mstate_global->__pyx_codeobj__111
#define __pyx_codeobj__112 __pyx_mstate_global->__pyx_codeobj__112
#define __pyx_codeobj__113 __pyx_mstate_global->__pyx_codeobj__113
#define __pyx_codeobj__114 __pyx_mstate_global->__pyx_codeobj__114
#define __pyx_codeobj__115 __pyx_mstate_global->__pyx_codeobj__115
#define __pyx_codeobj__116 __pyx_mstate_global->__pyx_codeobj__116
#define __pyx_codeobj__117 __pyx_mstate_global->__pyx_codeobj__117
#define __pyx_codeobj__119 __pyx_mstate_global->__pyx_codeobj__119
#define __pyx_codeobj__122 __pyx_mstate_global->__pyx_codeobj__122
#define __pyx_codeobj__124 __pyx_mstate_global->__pyx_codeobj__124
#define __pyx_codeobj__126 __pyx_mstate_global->__pyx_codeobj__126
#define __pyx_codeobj__128 __pyx_mstate_global->__pyx_codeobj__128
#define __pyx_codeobj__130 __pyx_mstate_global->__pyx_codeobj__130
#define __pyx_codeobj__132 __pyx_mstate_global->__pyx_codeobj__132
#define __pyx_codeobj__134 __pyx_mstate_global->__pyx_codeobj__134
#define __pyx_codeobj__136 __pyx_mstate_global->__pyx_codeobj__136
#define __pyx_codeobj__137 __pyx_mstate_global->__pyx_codeobj__137
#define __pyx_codeobj__138 __pyx_mstate_global->__pyx_codeobj__138
#define __pyx_codeobj__140 __pyx_mstate_global->__pyx_codeobj__140
#define __pyx_codeobj__141 __pyx_mstate_global->__pyx_codeobj__141
#define __pyx_codeobj__142 __pyx_mstate_global->__pyx_codeobj__142
#define __pyx_codeobj__144 __pyx_mstate_global->__pyx_codeobj__144
#define __pyx_codeobj__146 __pyx_mstate_global->__pyx_codeobj__146
#define __pyx_codeobj__148 __pyx_mstate_global->__pyx_codeobj__148
#define __pyx_codeobj__150 __pyx_mstate_global->__pyx_codeobj__150
#define __pyx_codeobj__151 __pyx_mstate_global->__pyx_codeobj__151
#define __pyx_codeobj__153 __pyx_mstate_global->__pyx_codeobj__153
#define __pyx_codeobj__154 __pyx_mstate_global->__pyx_codeobj__154
#define __pyx_codeobj__155 __pyx_mstate_global->__pyx_codeobj__155
#define __pyx_codeobj__156 __pyx_mstate_global->__pyx_codeobj__156
#define __pyx_codeobj__158 __pyx_mstate_global->__pyx_codeobj__158
#define __pyx_codeobj__159 __pyx_mstate_global->__pyx_codeobj__159
#define __pyx_codeobj__161 __pyx_mstate_global->__pyx_codeobj__161
#define __pyx_codeobj__162 __pyx_mstate_global->__pyx_codeobj__162
#define __pyx_codeobj__164 __pyx_mstate_global->__pyx_codeobj__164
#define __pyx_codeobj__166 __pyx_mstate_global->__pyx_codeobj__166
#define __pyx_codeobj__167 __pyx_mstate_global->__pyx_codeobj__167
#define __pyx_codeobj__168 __pyx_mstate_global->__pyx_codeobj__168
#define __pyx_codeobj__170 __pyx_mstate_global->__pyx_codeobj__170
#define __pyx_codeobj__172 __pyx_mstate_global->__pyx_codeobj__172
#define __pyx_codeobj__174 __pyx_mstate_global->__pyx_codeobj__174
#define __pyx_codeobj__176 __pyx_mstate_global->__pyx_codeobj__176
#define __pyx_codeobj__178 __pyx_mstate_global->__pyx_codeobj__178
#define __pyx_codeobj__180 __pyx_mstate_global->__pyx_codeobj__180
#define __pyx_codeobj__181 __pyx_mstate_global->__pyx_codeobj__181
#define __pyx_codeobj__183 __pyx_mstate_global->__pyx_codeobj__183
#define __pyx_codeobj__184 __pyx_mstate_global->__pyx_codeobj__184
#define __pyx_codeobj__185 __pyx_mstate_global->__pyx_codeobj__185
#define __pyx_codeobj__187 __pyx_mstate_global->__pyx_codeobj__187
#define __pyx_codeobj__188 __pyx_mstate_global->__pyx_codeobj__188
#define __pyx_codeobj__189 __pyx_mstate_global->__pyx_codeobj__189
#define __pyx_codeobj__190 __pyx_mstate_global->__pyx_codeobj__190
#define __pyx_codeobj__191 __pyx_mstate_global->__pyx_codeobj__191
#define __pyx_codeobj__192 __pyx_mstate_global->__pyx_codeobj__192
#define __pyx_codeobj__193 __pyx_mstate_global->__pyx_codeobj__193
#define __pyx_codeobj__194 __pyx_mstate_global->__pyx_codeobj__194
#define __pyx_codeobj__195 __pyx_mstate_global->__pyx_codeobj__195
#define __pyx_codeobj__196 __pyx_mstate_global->__pyx_codeobj__196
#define __pyx_codeobj__197 __pyx_mstate_global->__pyx_codeobj__197
#define __pyx_codeobj__198 __pyx_mstate_global->__pyx_codeobj__198
#define __pyx_codeobj__199 __pyx_mstate_global->__pyx_codeobj__199
#define __pyx_codeobj__200 __pyx_mstate_global->__pyx_codeobj__200
#define __pyx_codeobj__201 __pyx_mstate_global->__pyx_codeobj__201
#define __pyx_codeobj__202 __pyx_mstate_global->__pyx_codeobj__202
#define __pyx_codeobj__204 __pyx_mstate_global->__pyx_codeobj__204
#define __pyx_codeobj__206 __pyx_mstate_global->__pyx_codeobj__206
#define __pyx_codeobj__208 __pyx_mstate_global->__pyx_codeobj__208
#define __pyx_codeobj__209 __pyx_mstate_global->__pyx_codeobj__209
#define __pyx_codeobj__210 __pyx_mstate_global->__pyx_codeobj__210
#define __pyx_codeobj__211 __pyx_mstate_global->__pyx_codeobj__211
#define __pyx_codeobj__212 __pyx_mstate_global->__pyx_codeobj__212
#define __pyx_codeobj__213 __pyx_mstate_global->__pyx_codeobj__213
#define __pyx_codeobj__214 __pyx_mstate_global->__pyx_codeobj__214
#define __pyx_codeobj__215 __pyx_mstate_global->__pyx_codeobj__215
#define __pyx_codeobj__216 __pyx_mstate_global->__pyx_codeobj__216
#define __pyx_codeobj__218 __pyx_mstate_global->__pyx_codeobj__218
#define __pyx_codeobj__220 __pyx_mstate_global->__pyx_codeobj__220
#define __pyx_codeobj__222 __pyx_mstate_global->__pyx_codeobj__222
#define __pyx_codeobj__224 __pyx_mstate_global->__pyx_codeobj__224
#define __pyx_codeobj__226 __pyx_mstate_global->__pyx_codeobj__226
#define __pyx_codeobj__227 __pyx_mstate_global->__pyx_codeobj__227
#define __pyx_codeobj__229 __pyx_mstate_global->__pyx_codeobj__229
#define __pyx_codeobj__231 __pyx_mstate_global->__pyx_codeobj__231
#define __pyx_codeobj__233 __pyx_mstate_global->__pyx_codeobj__233
#define __pyx_codeobj__234 __pyx_mstate_global->__pyx_codeobj__234
#define __pyx_codeobj__235 __pyx_mstate_global->__pyx_codeobj__235
#define __pyx_codeobj__237 __pyx_mstate_global->__pyx_codeobj__237
#define __pyx_codeobj__238 __pyx_mstate_global->__pyx_codeobj__238
#define __pyx_codeobj__239 __pyx_mstate_global->__pyx_codeobj__239
#define __pyx_codeobj__241 __pyx_mstate_global->__pyx_codeobj__241
#define __pyx_codeobj__242 __pyx_mstate_global->__pyx_codeobj__242
#define __pyx_codeobj__243 __pyx_mstate_global->__pyx_codeobj__243
#define __pyx_codeobj__244 __pyx_mstate_global->__pyx_codeobj__244
#define __pyx_codeobj__245 __pyx_mstate_global->__pyx_codeobj__245
#define __pyx_codeobj__246 __pyx_mstate_global->__pyx_codeobj__246
#define __pyx_codeobj__248 __pyx_mstate_global->__pyx_codeobj__248
#define __pyx_codeobj__249 __pyx_mstate_global->__pyx_codeobj__249
#define __pyx_codeobj__250 __pyx_mstate_global->__pyx_codeobj__250
#define __pyx_codeobj__251 __pyx_mstate_global->__pyx_codeobj__251
#define __pyx_codeobj__253 __pyx_mstate_global->__pyx_codeobj__253
#define __pyx_codeobj__255 __pyx_mstate_global->__pyx_codeobj__255
#define __pyx_codeobj__257 __pyx_mstate_global->__pyx_codeobj__257
#define __pyx_codeobj__258 __pyx_mstate_global->__pyx_codeobj__258
#define __pyx_codeobj__260 __pyx_mstate_global->__pyx_codeobj__260
#define __pyx_codeobj__261 __pyx_mstate_global->__pyx_codeobj__261
#define __pyx_codeobj__262 __pyx_mstate_global->__pyx_codeobj__262
#define __pyx_codeobj__263 __pyx_mstate_global->__pyx_codeobj__263
#define __pyx_codeobj__264 __pyx_mstate_global->__pyx_codeobj__264
#define __pyx_codeobj__265 __pyx_mstate_global->__pyx_codeobj__265
#define __pyx_codeobj__266 __pyx_mstate_global->__pyx_codeobj__266
#define __pyx_codeobj__267 __pyx_mstate_global->__pyx_codeobj__267
#define __pyx_codeobj__268 __pyx_mstate_global->__pyx_codeobj__268
#define __pyx_codeobj__269 __pyx_mstate_global->__pyx_codeobj__269
#define __pyx_codeobj__270 __pyx_mstate_global->__pyx_codeobj__270
#define __pyx_codeobj__271 __pyx_mstate_global->__pyx_codeobj__271
#define __pyx_codeobj__272 __pyx_mstate_global->__pyx_codeobj__272
#define __pyx_codeobj__274 __pyx_mstate_global->__pyx_codeobj__274
#define __pyx_codeobj__275 __pyx_mstate_global->__pyx_codeobj__275
#define __pyx_codeobj__276 __pyx_mstate_global->__pyx_codeobj__276
#define __pyx_codeobj__278 __pyx_mstate_global->__pyx_codeobj__278
#define __pyx_codeobj__279 __pyx_mstate_global->__pyx_codeobj__279
#define __pyx_codeobj__281 __pyx_mstate_global->__pyx_codeobj__281
#define __pyx_codeobj__282 __pyx_mstate_global->__pyx_codeobj__282
#define __pyx_codeobj__283 __pyx_mstate_global->__pyx_codeobj__283
#define __pyx_codeobj__284 __pyx_mstate_global->__pyx_codeobj__284
#define __pyx_codeobj__285 __pyx_mstate_global->__pyx_codeobj__285
#define __pyx_codeobj__286 __pyx_mstate_global->__pyx_codeobj__286
/* #### Code section: module_code ### */

/* "uharfbuzz/_harfbuzz.pyx":34
 * 
 * 
 * cdef int msgcallback(hb_buffer_t *buffer, hb_font_t *font, const char* message, void* userdata) noexcept:             # <<<<<<<<<<<<<<
 *     ret = (<object>userdata)(message.decode('utf-8'))
 *     if ret is None:
 */

static int __pyx_f_9uharfbuzz_9_harfbuzz_msgcallback(CYTHON_UNUSED hb_buffer_t *__pyx_v_buffer, CYTHON_UNUSED hb_font_t *__pyx_v_font, char const *__pyx_v_message, void *__pyx_v_userdata) {
  PyObject *__pyx_v_ret = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("msgcallback", 1);

  /* "uharfbuzz/_harfbuzz.pyx":35
 * 
 * cdef int msgcallback(hb_buffer_t *buffer, hb_font_t *font, const char* message, void* userdata) noexcept:
 *     ret = (<object>userdata)(message.decode('utf-8'))             # <<<<<<<<<<<<<<
 *     if ret is None:
 *         return 1
 */
  __pyx_t_2 = __Pyx_ssize_strlen(__pyx_v_message); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 35, __pyx_L1_error)
  __pyx_t_3 = __Pyx_decode_c_string(__pyx_v_message, 0, __pyx_t_2, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_v_userdata));
  __pyx_t_4 = ((PyObject *)__pyx_v_userdata); __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 35, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_v_ret = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":36
 * cdef int msgcallback(hb_buffer_t *buffer, hb_font_t *font, const char* message, void* userdata) noexcept:
 *     ret = (<object>userdata)(message.decode('utf-8'))
 *     if ret is None:             # <<<<<<<<<<<<<<
 *         return 1
 *     return ret
 */
  __pyx_t_7 = (__pyx_v_ret == Py_None);
  if (__pyx_t_7) {

    /* "uharfbuzz/_harfbuzz.pyx":37
 *     ret = (<object>userdata)(message.decode('utf-8'))
 *     if ret is None:
 *         return 1             # <<<<<<<<<<<<<<
 *     return ret
 * 
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "uharfbuzz/_harfbuzz.pyx":36
 * cdef int msgcallback(hb_buffer_t *buffer, hb_font_t *font, const char* message, void* userdata) noexcept:
 *     ret = (<object>userdata)(message.decode('utf-8'))
 *     if ret is None:             # <<<<<<<<<<<<<<
 *         return 1
 *     return ret
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":38
 *     if ret is None:
 *         return 1
 *     return ret             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_v_ret); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 38, __pyx_L1_error)
  __pyx_r = __pyx_t_8;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":34
 * 
 * 
 * cdef int msgcallback(hb_buffer_t *buffer, hb_font_t *font, const char* message, void* userdata) noexcept:             # <<<<<<<<<<<<<<
 *     ret = (<object>userdata)(message.decode('utf-8'))
 *     if ret is None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz.msgcallback", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":41
 * 
 * 
 * def version_string() -> str:             # <<<<<<<<<<<<<<
 *     cdef const char* cstr = hb_version_string()
 *     cdef bytes packed = cstr
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_1version_string(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_1version_string = {"version_string", (PyCFunction)__pyx_pw_9uharfbuzz_9_harfbuzz_1version_string, METH_NOARGS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_1version_string(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("version_string (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_version_string(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_version_string(CYTHON_UNUSED PyObject *__pyx_self) {
  char const *__pyx_v_cstr;
  PyObject *__pyx_v_packed = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("version_string", 1);

  /* "uharfbuzz/_harfbuzz.pyx":42
 * 
 * def version_string() -> str:
 *     cdef const char* cstr = hb_version_string()             # <<<<<<<<<<<<<<
 *     cdef bytes packed = cstr
 *     return packed.decode()
 */
  __pyx_v_cstr = hb_version_string();

  /* "uharfbuzz/_harfbuzz.pyx":43
 * def version_string() -> str:
 *     cdef const char* cstr = hb_version_string()
 *     cdef bytes packed = cstr             # <<<<<<<<<<<<<<
 *     return packed.decode()
 * 
 */
  __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_cstr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_packed = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":44
 *     cdef const char* cstr = hb_version_string()
 *     cdef bytes packed = cstr
 *     return packed.decode()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_decode_bytes(__pyx_v_packed, 0, PY_SSIZE_T_MAX, NULL, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":41
 * 
 * 
 * def version_string() -> str:             # <<<<<<<<<<<<<<
 *     cdef const char* cstr = hb_version_string()
 *     cdef bytes packed = cstr
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.version_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_packed);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":61
 *     # could maybe store Buffer to prevent GC
 * 
 *     cdef set(self, hb_glyph_info_t info):             # <<<<<<<<<<<<<<
 *         self._hb_glyph_info = info
 * 
 */

static PyObject *__pyx_f_9uharfbuzz_9_harfbuzz_9GlyphInfo_set(struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphInfo *__pyx_v_self, hb_glyph_info_t __pyx_v_info) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set", 1);

  /* "uharfbuzz/_harfbuzz.pyx":62
 * 
 *     cdef set(self, hb_glyph_info_t info):
 *         self._hb_glyph_info = info             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_v_self->_hb_glyph_info = __pyx_v_info;

  /* "uharfbuzz/_harfbuzz.pyx":61
 *     # could maybe store Buffer to prevent GC
 * 
 *     cdef set(self, hb_glyph_info_t info):             # <<<<<<<<<<<<<<
 *         self._hb_glyph_info = info
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":64
 *         self._hb_glyph_info = info
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def codepoint(self) -> int:
 *         return self._hb_glyph_info.codepoint
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9GlyphInfo_9codepoint_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9GlyphInfo_9codepoint_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_9GlyphInfo_9codepoint___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphInfo *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9GlyphInfo_9codepoint___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphInfo *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":66
 *     @property
 *     def codepoint(self) -> int:
 *         return self._hb_glyph_info.codepoint             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_hb_codepoint_t(__pyx_v_self->_hb_glyph_info.codepoint); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":64
 *         self._hb_glyph_info = info
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def codepoint(self) -> int:
 *         return self._hb_glyph_info.codepoint
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.GlyphInfo.codepoint.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":68
 *         return self._hb_glyph_info.codepoint
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def cluster(self) -> int:
 *         return self._hb_glyph_info.cluster
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9GlyphInfo_7cluster_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9GlyphInfo_7cluster_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_9GlyphInfo_7cluster___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphInfo *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9GlyphInfo_7cluster___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphInfo *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":70
 *     @property
 *     def cluster(self) -> int:
 *         return self._hb_glyph_info.cluster             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_long(__pyx_v_self->_hb_glyph_info.cluster); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":68
 *         return self._hb_glyph_info.codepoint
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def cluster(self) -> int:
 *         return self._hb_glyph_info.cluster
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.GlyphInfo.cluster.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":72
 *         return self._hb_glyph_info.cluster
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def flags(self) -> GlyphFlags:
 *         return GlyphFlags(self._hb_glyph_info.mask & HB_GLYPH_FLAG_DEFINED)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9GlyphInfo_5flags_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9GlyphInfo_5flags_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_9GlyphInfo_5flags___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphInfo *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9GlyphInfo_5flags___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphInfo *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":74
 *     @property
 *     def flags(self) -> GlyphFlags:
 *         return GlyphFlags(self._hb_glyph_info.mask & HB_GLYPH_FLAG_DEFINED)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_GlyphFlags); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_hb_mask_t((__pyx_v_self->_hb_glyph_info.mask & HB_GLYPH_FLAG_DEFINED)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":72
 *         return self._hb_glyph_info.cluster
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def flags(self) -> GlyphFlags:
 *         return GlyphFlags(self._hb_glyph_info.mask & HB_GLYPH_FLAG_DEFINED)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.GlyphInfo.flags.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self._hb_glyph_info must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9GlyphInfo_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_9GlyphInfo_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9GlyphInfo_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9GlyphInfo_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_9GlyphInfo___reduce_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphInfo *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9GlyphInfo___reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphInfo *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self._hb_glyph_info must be explicitly requested with @auto_pickle(True)"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "Pickling of struct members such as self._hb_glyph_info must be explicitly requested with @auto_pickle(True)"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_Pickling_of_struct_members_such, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self._hb_glyph_info must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.GlyphInfo.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self._hb_glyph_info must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self._hb_glyph_info must be explicitly requested with @auto_pickle(True)"
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9GlyphInfo_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_9GlyphInfo_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9GlyphInfo_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9GlyphInfo_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.GlyphInfo.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_9GlyphInfo_2__setstate_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphInfo *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9GlyphInfo_2__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphInfo *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "Pickling of struct members such as self._hb_glyph_info must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "Pickling of struct members such as self._hb_glyph_info must be explicitly requested with @auto_pickle(True)"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_Pickling_of_struct_members_such, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self._hb_glyph_info must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self._hb_glyph_info must be explicitly requested with @auto_pickle(True)"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.GlyphInfo.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":81
 *     # could maybe store Buffer to prevent GC
 * 
 *     cdef set(self, hb_glyph_position_t position):             # <<<<<<<<<<<<<<
 *         self._hb_glyph_position = position
 * 
 */

static PyObject *__pyx_f_9uharfbuzz_9_harfbuzz_13GlyphPosition_set(struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition *__pyx_v_self, hb_glyph_position_t __pyx_v_position) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set", 1);

  /* "uharfbuzz/_harfbuzz.pyx":82
 * 
 *     cdef set(self, hb_glyph_position_t position):
 *         self._hb_glyph_position = position             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_v_self->_hb_glyph_position = __pyx_v_position;

  /* "uharfbuzz/_harfbuzz.pyx":81
 *     # could maybe store Buffer to prevent GC
 * 
 *     cdef set(self, hb_glyph_position_t position):             # <<<<<<<<<<<<<<
 *         self._hb_glyph_position = position
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":84
 *         self._hb_glyph_position = position
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def position(self):
 *         return (
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_13GlyphPosition_8position_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_13GlyphPosition_8position_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_13GlyphPosition_8position___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_13GlyphPosition_8position___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":86
 *     @property
 *     def position(self):
 *         return (             # <<<<<<<<<<<<<<
 *             self._hb_glyph_position.x_offset,
 *             self._hb_glyph_position.y_offset,
 */
  __Pyx_XDECREF(__pyx_r);

  /* "uharfbuzz/_harfbuzz.pyx":87
 *     def position(self):
 *         return (
 *             self._hb_glyph_position.x_offset,             # <<<<<<<<<<<<<<
 *             self._hb_glyph_position.y_offset,
 *             self._hb_glyph_position.x_advance,
 */
  __pyx_t_1 = __Pyx_PyInt_From_hb_position_t(__pyx_v_self->_hb_glyph_position.x_offset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "uharfbuzz/_harfbuzz.pyx":88
 *         return (
 *             self._hb_glyph_position.x_offset,
 *             self._hb_glyph_position.y_offset,             # <<<<<<<<<<<<<<
 *             self._hb_glyph_position.x_advance,
 *             self._hb_glyph_position.y_advance
 */
  __pyx_t_2 = __Pyx_PyInt_From_hb_position_t(__pyx_v_self->_hb_glyph_position.y_offset); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "uharfbuzz/_harfbuzz.pyx":89
 *             self._hb_glyph_position.x_offset,
 *             self._hb_glyph_position.y_offset,
 *             self._hb_glyph_position.x_advance,             # <<<<<<<<<<<<<<
 *             self._hb_glyph_position.y_advance
 *         )
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_position_t(__pyx_v_self->_hb_glyph_position.x_advance); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "uharfbuzz/_harfbuzz.pyx":90
 *             self._hb_glyph_position.y_offset,
 *             self._hb_glyph_position.x_advance,
 *             self._hb_glyph_position.y_advance             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __pyx_t_4 = __Pyx_PyInt_From_hb_position_t(__pyx_v_self->_hb_glyph_position.y_advance); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "uharfbuzz/_harfbuzz.pyx":87
 *     def position(self):
 *         return (
 *             self._hb_glyph_position.x_offset,             # <<<<<<<<<<<<<<
 *             self._hb_glyph_position.y_offset,
 *             self._hb_glyph_position.x_advance,
 */
  __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1)) __PYX_ERR(0, 87, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2)) __PYX_ERR(0, 87, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3)) __PYX_ERR(0, 87, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4)) __PYX_ERR(0, 87, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":84
 *         self._hb_glyph_position = position
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def position(self):
 *         return (
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.GlyphPosition.position.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":93
 *         )
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def x_advance(self):
 *         return self._hb_glyph_position.x_advance
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_13GlyphPosition_9x_advance_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_13GlyphPosition_9x_advance_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_13GlyphPosition_9x_advance___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_13GlyphPosition_9x_advance___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":95
 *     @property
 *     def x_advance(self):
 *         return self._hb_glyph_position.x_advance             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_hb_position_t(__pyx_v_self->_hb_glyph_position.x_advance); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":93
 *         )
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def x_advance(self):
 *         return self._hb_glyph_position.x_advance
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.GlyphPosition.x_advance.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":97
 *         return self._hb_glyph_position.x_advance
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def y_advance(self):
 *         return self._hb_glyph_position.y_advance
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_13GlyphPosition_9y_advance_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_13GlyphPosition_9y_advance_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_13GlyphPosition_9y_advance___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_13GlyphPosition_9y_advance___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":99
 *     @property
 *     def y_advance(self):
 *         return self._hb_glyph_position.y_advance             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_hb_position_t(__pyx_v_self->_hb_glyph_position.y_advance); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":97
 *         return self._hb_glyph_position.x_advance
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def y_advance(self):
 *         return self._hb_glyph_position.y_advance
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.GlyphPosition.y_advance.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":101
 *         return self._hb_glyph_position.y_advance
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def x_offset(self):
 *         return self._hb_glyph_position.x_offset
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_13GlyphPosition_8x_offset_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_13GlyphPosition_8x_offset_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_13GlyphPosition_8x_offset___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_13GlyphPosition_8x_offset___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":103
 *     @property
 *     def x_offset(self):
 *         return self._hb_glyph_position.x_offset             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_hb_position_t(__pyx_v_self->_hb_glyph_position.x_offset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":101
 *         return self._hb_glyph_position.y_advance
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def x_offset(self):
 *         return self._hb_glyph_position.x_offset
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.GlyphPosition.x_offset.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":105
 *         return self._hb_glyph_position.x_offset
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def y_offset(self):
 *         return self._hb_glyph_position.y_offset
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_13GlyphPosition_8y_offset_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_13GlyphPosition_8y_offset_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_13GlyphPosition_8y_offset___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_13GlyphPosition_8y_offset___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":107
 *     @property
 *     def y_offset(self):
 *         return self._hb_glyph_position.y_offset             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_hb_position_t(__pyx_v_self->_hb_glyph_position.y_offset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":105
 *         return self._hb_glyph_position.x_offset
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def y_offset(self):
 *         return self._hb_glyph_position.y_offset
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.GlyphPosition.y_offset.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self._hb_glyph_position must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_13GlyphPosition_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_13GlyphPosition_1__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_13GlyphPosition_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_13GlyphPosition_1__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_13GlyphPosition___reduce_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_13GlyphPosition___reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self._hb_glyph_position must be explicitly requested with @auto_pickle(True)"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "Pickling of struct members such as self._hb_glyph_position must be explicitly requested with @auto_pickle(True)"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_Pickling_of_struct_members_such_2, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self._hb_glyph_position must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.GlyphPosition.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self._hb_glyph_position must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self._hb_glyph_position must be explicitly requested with @auto_pickle(True)"
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_13GlyphPosition_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_13GlyphPosition_3__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_13GlyphPosition_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_13GlyphPosition_3__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.GlyphPosition.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_13GlyphPosition_2__setstate_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_13GlyphPosition_2__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "Pickling of struct members such as self._hb_glyph_position must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "Pickling of struct members such as self._hb_glyph_position must be explicitly requested with @auto_pickle(True)"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_Pickling_of_struct_members_such_2, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self._hb_glyph_position must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self._hb_glyph_position must be explicitly requested with @auto_pickle(True)"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.GlyphPosition.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":138
 *     DEFAULT_REPLACEMENT_CODEPOINT = HB_BUFFER_REPLACEMENT_CODEPOINT_DEFAULT
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._hb_buffer = hb_buffer_create()
 *         if not hb_buffer_allocation_successful(self._hb_buffer):
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer___cinit__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer___cinit__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":139
 * 
 *     def __cinit__(self):
 *         self._hb_buffer = hb_buffer_create()             # <<<<<<<<<<<<<<
 *         if not hb_buffer_allocation_successful(self._hb_buffer):
 *             raise MemoryError()
 */
  __pyx_v_self->_hb_buffer = hb_buffer_create();

  /* "uharfbuzz/_harfbuzz.pyx":140
 *     def __cinit__(self):
 *         self._hb_buffer = hb_buffer_create()
 *         if not hb_buffer_allocation_successful(self._hb_buffer):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         self._message_callback = None
 */
  __pyx_t_1 = (!(hb_buffer_allocation_successful(__pyx_v_self->_hb_buffer) != 0));
  if (unlikely(__pyx_t_1)) {

    /* "uharfbuzz/_harfbuzz.pyx":141
 *         self._hb_buffer = hb_buffer_create()
 *         if not hb_buffer_allocation_successful(self._hb_buffer):
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 *         self._message_callback = None
 * 
 */
    PyErr_NoMemory(); __PYX_ERR(0, 141, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":140
 *     def __cinit__(self):
 *         self._hb_buffer = hb_buffer_create()
 *         if not hb_buffer_allocation_successful(self._hb_buffer):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         self._message_callback = None
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":142
 *         if not hb_buffer_allocation_successful(self._hb_buffer):
 *             raise MemoryError()
 *         self._message_callback = None             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_message_callback);
  __Pyx_DECREF(__pyx_v_self->_message_callback);
  __pyx_v_self->_message_callback = Py_None;

  /* "uharfbuzz/_harfbuzz.pyx":138
 *     DEFAULT_REPLACEMENT_CODEPOINT = HB_BUFFER_REPLACEMENT_CODEPOINT_DEFAULT
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._hb_buffer = hb_buffer_create()
 *         if not hb_buffer_allocation_successful(self._hb_buffer):
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":144
 *         self._message_callback = None
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         hb_buffer_destroy(self._hb_buffer)
 * 
 */

/* Python wrapper */
static void __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_2__dealloc__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_2__dealloc__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self) {

  /* "uharfbuzz/_harfbuzz.pyx":145
 * 
 *     def __dealloc__(self):
 *         hb_buffer_destroy(self._hb_buffer)             # <<<<<<<<<<<<<<
 * 
 *     # DEPRECATED: use the normal constructor
 */
  hb_buffer_destroy(__pyx_v_self->_hb_buffer);

  /* "uharfbuzz/_harfbuzz.pyx":144
 *         self._message_callback = None
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         hb_buffer_destroy(self._hb_buffer)
 * 
 */

  /* function exit code */
}

/* "uharfbuzz/_harfbuzz.pyx":148
 * 
 *     # DEPRECATED: use the normal constructor
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def create(cls):
 *         cdef Buffer inst = cls()
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_5create(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_6Buffer_5create = {"create", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_5create, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_5create(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("create", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "create", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_4create(((PyTypeObject*)__pyx_v_cls));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_4create(PyTypeObject *__pyx_v_cls) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_inst = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("create", 1);

  /* "uharfbuzz/_harfbuzz.pyx":150
 *     @classmethod
 *     def create(cls):
 *         cdef Buffer inst = cls()             # <<<<<<<<<<<<<<
 *         return inst
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_v_cls)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer))))) __PYX_ERR(0, 150, __pyx_L1_error)
  __pyx_v_inst = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":151
 *     def create(cls):
 *         cdef Buffer inst = cls()
 *         return inst             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self) -> int:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_inst);
  __pyx_r = ((PyObject *)__pyx_v_inst);
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":148
 * 
 *     # DEPRECATED: use the normal constructor
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def create(cls):
 *         cdef Buffer inst = cls()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_inst);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":153
 *         return inst
 * 
 *     def __len__(self) -> int:             # <<<<<<<<<<<<<<
 *         return hb_buffer_get_length(self._hb_buffer)
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_7__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_7__len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_6__len__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_6__len__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self) {
  Py_ssize_t __pyx_r;

  /* "uharfbuzz/_harfbuzz.pyx":154
 * 
 *     def __len__(self) -> int:
 *         return hb_buffer_get_length(self._hb_buffer)             # <<<<<<<<<<<<<<
 * 
 *     def reset(self):
 */
  __pyx_r = hb_buffer_get_length(__pyx_v_self->_hb_buffer);
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":153
 *         return inst
 * 
 *     def __len__(self) -> int:             # <<<<<<<<<<<<<<
 *         return hb_buffer_get_length(self._hb_buffer)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":156
 *         return hb_buffer_get_length(self._hb_buffer)
 * 
 *     def reset(self):             # <<<<<<<<<<<<<<
 *         hb_buffer_reset (self._hb_buffer)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_9reset(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_6Buffer_9reset = {"reset", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_9reset, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_9reset(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reset (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("reset", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "reset", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_8reset(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_8reset(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reset", 1);

  /* "uharfbuzz/_harfbuzz.pyx":157
 * 
 *     def reset(self):
 *         hb_buffer_reset (self._hb_buffer)             # <<<<<<<<<<<<<<
 * 
 *     def clear_contents(self):
 */
  hb_buffer_reset(__pyx_v_self->_hb_buffer);

  /* "uharfbuzz/_harfbuzz.pyx":156
 *         return hb_buffer_get_length(self._hb_buffer)
 * 
 *     def reset(self):             # <<<<<<<<<<<<<<
 *         hb_buffer_reset (self._hb_buffer)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":159
 *         hb_buffer_reset (self._hb_buffer)
 * 
 *     def clear_contents(self):             # <<<<<<<<<<<<<<
 *         hb_buffer_clear_contents(self._hb_buffer)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_11clear_contents(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_6Buffer_11clear_contents = {"clear_contents", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_11clear_contents, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_11clear_contents(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clear_contents (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("clear_contents", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "clear_contents", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_10clear_contents(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_10clear_contents(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clear_contents", 1);

  /* "uharfbuzz/_harfbuzz.pyx":160
 * 
 *     def clear_contents(self):
 *         hb_buffer_clear_contents(self._hb_buffer)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  hb_buffer_clear_contents(__pyx_v_self->_hb_buffer);

  /* "uharfbuzz/_harfbuzz.pyx":159
 *         hb_buffer_reset (self._hb_buffer)
 * 
 *     def clear_contents(self):             # <<<<<<<<<<<<<<
 *         hb_buffer_clear_contents(self._hb_buffer)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":162
 *         hb_buffer_clear_contents(self._hb_buffer)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def direction(self) -> str:
 *         cdef const_char* cstr = hb_direction_to_string(
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_9direction_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_9direction_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_9direction___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_9direction___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self) {
  const char *__pyx_v_cstr;
  PyObject *__pyx_v_packed = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":164
 *     @property
 *     def direction(self) -> str:
 *         cdef const_char* cstr = hb_direction_to_string(             # <<<<<<<<<<<<<<
 *             hb_buffer_get_direction(self._hb_buffer))
 *         cdef bytes packed = cstr
 */
  __pyx_v_cstr = hb_direction_to_string(hb_buffer_get_direction(__pyx_v_self->_hb_buffer));

  /* "uharfbuzz/_harfbuzz.pyx":166
 *         cdef const_char* cstr = hb_direction_to_string(
 *             hb_buffer_get_direction(self._hb_buffer))
 *         cdef bytes packed = cstr             # <<<<<<<<<<<<<<
 *         return packed.decode()
 * 
 */
  __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_cstr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_packed = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":167
 *             hb_buffer_get_direction(self._hb_buffer))
 *         cdef bytes packed = cstr
 *         return packed.decode()             # <<<<<<<<<<<<<<
 * 
 *     @direction.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_decode_bytes(__pyx_v_packed, 0, PY_SSIZE_T_MAX, NULL, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":162
 *         hb_buffer_clear_contents(self._hb_buffer)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def direction(self) -> str:
 *         cdef const_char* cstr = hb_direction_to_string(
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.direction.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_packed);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":169
 *         return packed.decode()
 * 
 *     @direction.setter             # <<<<<<<<<<<<<<
 *     def direction(self, value: str):
 *         cdef bytes packed = value.encode()
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_9direction_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_9direction_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyUnicode_Type), 0, "value", 1))) __PYX_ERR(0, 170, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_9direction_2__set__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self), ((PyObject*)__pyx_v_value));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_9direction_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_packed = 0;
  char *__pyx_v_cstr;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":171
 *     @direction.setter
 *     def direction(self, value: str):
 *         cdef bytes packed = value.encode()             # <<<<<<<<<<<<<<
 *         cdef char* cstr = packed
 *         hb_buffer_set_direction(
 */
  __pyx_t_1 = PyUnicode_AsEncodedString(__pyx_v_value, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_packed = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":172
 *     def direction(self, value: str):
 *         cdef bytes packed = value.encode()
 *         cdef char* cstr = packed             # <<<<<<<<<<<<<<
 *         hb_buffer_set_direction(
 *             self._hb_buffer, hb_direction_from_string(cstr, -1))
 */
  if (unlikely(__pyx_v_packed == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 172, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBytes_AsWritableString(__pyx_v_packed); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 172, __pyx_L1_error)
  __pyx_v_cstr = __pyx_t_2;

  /* "uharfbuzz/_harfbuzz.pyx":173
 *         cdef bytes packed = value.encode()
 *         cdef char* cstr = packed
 *         hb_buffer_set_direction(             # <<<<<<<<<<<<<<
 *             self._hb_buffer, hb_direction_from_string(cstr, -1))
 * 
 */
  hb_buffer_set_direction(__pyx_v_self->_hb_buffer, hb_direction_from_string(__pyx_v_cstr, -1));

  /* "uharfbuzz/_harfbuzz.pyx":169
 *         return packed.decode()
 * 
 *     @direction.setter             # <<<<<<<<<<<<<<
 *     def direction(self, value: str):
 *         cdef bytes packed = value.encode()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.direction.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_packed);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":176
 *             self._hb_buffer, hb_direction_from_string(cstr, -1))
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def glyph_infos(self) -> List[GlyphInfo]:
 *         cdef unsigned int count
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_11glyph_infos_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_11glyph_infos_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_11glyph_infos___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_11glyph_infos___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self) {
  unsigned int __pyx_v_count;
  hb_glyph_info_t *__pyx_v_glyph_infos;
  PyObject *__pyx_v_infos = 0;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphInfo *__pyx_v_info = 0;
  unsigned int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  unsigned int __pyx_t_2;
  unsigned int __pyx_t_3;
  unsigned int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":179
 *     def glyph_infos(self) -> List[GlyphInfo]:
 *         cdef unsigned int count
 *         cdef hb_glyph_info_t* glyph_infos = hb_buffer_get_glyph_infos(             # <<<<<<<<<<<<<<
 *             self._hb_buffer, &count)
 *         cdef list infos = []
 */
  __pyx_v_glyph_infos = hb_buffer_get_glyph_infos(__pyx_v_self->_hb_buffer, (&__pyx_v_count));

  /* "uharfbuzz/_harfbuzz.pyx":181
 *         cdef hb_glyph_info_t* glyph_infos = hb_buffer_get_glyph_infos(
 *             self._hb_buffer, &count)
 *         cdef list infos = []             # <<<<<<<<<<<<<<
 *         cdef GlyphInfo info
 *         cdef unsigned int i
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_infos = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":184
 *         cdef GlyphInfo info
 *         cdef unsigned int i
 *         for i in range(count):             # <<<<<<<<<<<<<<
 *             info = GlyphInfo()
 *             info.set(glyph_infos[i])
 */
  __pyx_t_2 = __pyx_v_count;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "uharfbuzz/_harfbuzz.pyx":185
 *         cdef unsigned int i
 *         for i in range(count):
 *             info = GlyphInfo()             # <<<<<<<<<<<<<<
 *             info.set(glyph_infos[i])
 *             infos.append(info)
 */
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphInfo)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_info, ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphInfo *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "uharfbuzz/_harfbuzz.pyx":186
 *         for i in range(count):
 *             info = GlyphInfo()
 *             info.set(glyph_infos[i])             # <<<<<<<<<<<<<<
 *             infos.append(info)
 *         return infos
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_GlyphInfo *)__pyx_v_info->__pyx_vtab)->set(__pyx_v_info, (__pyx_v_glyph_infos[__pyx_v_i])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 186, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uharfbuzz/_harfbuzz.pyx":187
 *             info = GlyphInfo()
 *             info.set(glyph_infos[i])
 *             infos.append(info)             # <<<<<<<<<<<<<<
 *         return infos
 * 
 */
    __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_infos, ((PyObject *)__pyx_v_info)); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 187, __pyx_L1_error)
  }

  /* "uharfbuzz/_harfbuzz.pyx":188
 *             info.set(glyph_infos[i])
 *             infos.append(info)
 *         return infos             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_infos);
  __pyx_r = __pyx_v_infos;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":176
 *             self._hb_buffer, hb_direction_from_string(cstr, -1))
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def glyph_infos(self) -> List[GlyphInfo]:
 *         cdef unsigned int count
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.glyph_infos.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_infos);
  __Pyx_XDECREF((PyObject *)__pyx_v_info);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":190
 *         return infos
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def glyph_positions(self) -> List[GlyphPosition]:
 *         cdef unsigned int count
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_15glyph_positions_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_15glyph_positions_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_15glyph_positions___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_15glyph_positions___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self) {
  unsigned int __pyx_v_count;
  hb_glyph_position_t *__pyx_v_glyph_positions;
  PyObject *__pyx_v_positions = 0;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition *__pyx_v_position = 0;
  unsigned int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  unsigned int __pyx_t_3;
  unsigned int __pyx_t_4;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":194
 *         cdef unsigned int count
 *         cdef hb_glyph_position_t* glyph_positions = \
 *             hb_buffer_get_glyph_positions(self._hb_buffer, &count)             # <<<<<<<<<<<<<<
 *         if glyph_positions is NULL:
 *             return None
 */
  __pyx_v_glyph_positions = hb_buffer_get_glyph_positions(__pyx_v_self->_hb_buffer, (&__pyx_v_count));

  /* "uharfbuzz/_harfbuzz.pyx":195
 *         cdef hb_glyph_position_t* glyph_positions = \
 *             hb_buffer_get_glyph_positions(self._hb_buffer, &count)
 *         if glyph_positions is NULL:             # <<<<<<<<<<<<<<
 *             return None
 *         cdef list positions = []
 */
  __pyx_t_1 = (__pyx_v_glyph_positions == NULL);
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":196
 *             hb_buffer_get_glyph_positions(self._hb_buffer, &count)
 *         if glyph_positions is NULL:
 *             return None             # <<<<<<<<<<<<<<
 *         cdef list positions = []
 *         cdef GlyphPosition position
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uharfbuzz/_harfbuzz.pyx":195
 *         cdef hb_glyph_position_t* glyph_positions = \
 *             hb_buffer_get_glyph_positions(self._hb_buffer, &count)
 *         if glyph_positions is NULL:             # <<<<<<<<<<<<<<
 *             return None
 *         cdef list positions = []
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":197
 *         if glyph_positions is NULL:
 *             return None
 *         cdef list positions = []             # <<<<<<<<<<<<<<
 *         cdef GlyphPosition position
 *         cdef unsigned int i
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_positions = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":200
 *         cdef GlyphPosition position
 *         cdef unsigned int i
 *         for i in range(count):             # <<<<<<<<<<<<<<
 *             position = GlyphPosition()
 *             position.set(glyph_positions[i])
 */
  __pyx_t_3 = __pyx_v_count;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "uharfbuzz/_harfbuzz.pyx":201
 *         cdef unsigned int i
 *         for i in range(count):
 *             position = GlyphPosition()             # <<<<<<<<<<<<<<
 *             position.set(glyph_positions[i])
 *             positions.append(position)
 */
    __pyx_t_2 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphPosition)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 201, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_position, ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition *)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "uharfbuzz/_harfbuzz.pyx":202
 *         for i in range(count):
 *             position = GlyphPosition()
 *             position.set(glyph_positions[i])             # <<<<<<<<<<<<<<
 *             positions.append(position)
 *         return positions
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_GlyphPosition *)__pyx_v_position->__pyx_vtab)->set(__pyx_v_position, (__pyx_v_glyph_positions[__pyx_v_i])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uharfbuzz/_harfbuzz.pyx":203
 *             position = GlyphPosition()
 *             position.set(glyph_positions[i])
 *             positions.append(position)             # <<<<<<<<<<<<<<
 *         return positions
 * 
 */
    __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_positions, ((PyObject *)__pyx_v_position)); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 203, __pyx_L1_error)
  }

  /* "uharfbuzz/_harfbuzz.pyx":204
 *             position.set(glyph_positions[i])
 *             positions.append(position)
 *         return positions             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_positions);
  __pyx_r = __pyx_v_positions;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":190
 *         return infos
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def glyph_positions(self) -> List[GlyphPosition]:
 *         cdef unsigned int count
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.glyph_positions.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_positions);
  __Pyx_XDECREF((PyObject *)__pyx_v_position);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":206
 *         return positions
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def language(self) -> str:
 *         cdef const_char* cstr = hb_language_to_string(
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_8language_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_8language_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_8language___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_8language___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self) {
  const char *__pyx_v_cstr;
  PyObject *__pyx_v_packed = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":208
 *     @property
 *     def language(self) -> str:
 *         cdef const_char* cstr = hb_language_to_string(             # <<<<<<<<<<<<<<
 *             hb_buffer_get_language(self._hb_buffer))
 *         if cstr is NULL:
 */
  __pyx_v_cstr = hb_language_to_string(hb_buffer_get_language(__pyx_v_self->_hb_buffer));

  /* "uharfbuzz/_harfbuzz.pyx":210
 *         cdef const_char* cstr = hb_language_to_string(
 *             hb_buffer_get_language(self._hb_buffer))
 *         if cstr is NULL:             # <<<<<<<<<<<<<<
 *             return None
 *         cdef bytes packed = cstr
 */
  __pyx_t_1 = (__pyx_v_cstr == NULL);
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":211
 *             hb_buffer_get_language(self._hb_buffer))
 *         if cstr is NULL:
 *             return None             # <<<<<<<<<<<<<<
 *         cdef bytes packed = cstr
 *         return packed.decode()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uharfbuzz/_harfbuzz.pyx":210
 *         cdef const_char* cstr = hb_language_to_string(
 *             hb_buffer_get_language(self._hb_buffer))
 *         if cstr is NULL:             # <<<<<<<<<<<<<<
 *             return None
 *         cdef bytes packed = cstr
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":212
 *         if cstr is NULL:
 *             return None
 *         cdef bytes packed = cstr             # <<<<<<<<<<<<<<
 *         return packed.decode()
 * 
 */
  __pyx_t_2 = __Pyx_PyBytes_FromString(__pyx_v_cstr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_packed = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":213
 *             return None
 *         cdef bytes packed = cstr
 *         return packed.decode()             # <<<<<<<<<<<<<<
 * 
 *     @language.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_decode_bytes(__pyx_v_packed, 0, PY_SSIZE_T_MAX, NULL, NULL, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":206
 *         return positions
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def language(self) -> str:
 *         cdef const_char* cstr = hb_language_to_string(
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.language.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_packed);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":215
 *         return packed.decode()
 * 
 *     @language.setter             # <<<<<<<<<<<<<<
 *     def language(self, value: str):
 *         cdef bytes packed = value.encode()
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_8language_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_8language_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyUnicode_Type), 0, "value", 1))) __PYX_ERR(0, 216, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_8language_2__set__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self), ((PyObject*)__pyx_v_value));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_8language_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_packed = 0;
  char *__pyx_v_cstr;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":217
 *     @language.setter
 *     def language(self, value: str):
 *         cdef bytes packed = value.encode()             # <<<<<<<<<<<<<<
 *         cdef char* cstr = packed
 *         hb_buffer_set_language(
 */
  __pyx_t_1 = PyUnicode_AsEncodedString(__pyx_v_value, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_packed = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":218
 *     def language(self, value: str):
 *         cdef bytes packed = value.encode()
 *         cdef char* cstr = packed             # <<<<<<<<<<<<<<
 *         hb_buffer_set_language(
 *             self._hb_buffer, hb_language_from_string(cstr, -1))
 */
  if (unlikely(__pyx_v_packed == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 218, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBytes_AsWritableString(__pyx_v_packed); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 218, __pyx_L1_error)
  __pyx_v_cstr = __pyx_t_2;

  /* "uharfbuzz/_harfbuzz.pyx":219
 *         cdef bytes packed = value.encode()
 *         cdef char* cstr = packed
 *         hb_buffer_set_language(             # <<<<<<<<<<<<<<
 *             self._hb_buffer, hb_language_from_string(cstr, -1))
 * 
 */
  hb_buffer_set_language(__pyx_v_self->_hb_buffer, hb_language_from_string(__pyx_v_cstr, -1));

  /* "uharfbuzz/_harfbuzz.pyx":215
 *         return packed.decode()
 * 
 *     @language.setter             # <<<<<<<<<<<<<<
 *     def language(self, value: str):
 *         cdef bytes packed = value.encode()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.language.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_packed);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":222
 *             self._hb_buffer, hb_language_from_string(cstr, -1))
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def script(self) -> str:
 *         cdef char cstr[5]
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_6script_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_6script_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_6script___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_6script___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self) {
  char __pyx_v_cstr[5];
  PyObject *__pyx_v_packed = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":225
 *     def script(self) -> str:
 *         cdef char cstr[5]
 *         hb_tag_to_string(hb_buffer_get_script(self._hb_buffer), cstr)             # <<<<<<<<<<<<<<
 *         cstr[4] = b'\0'
 *         if cstr[0] == b'\0':
 */
  hb_tag_to_string(hb_buffer_get_script(__pyx_v_self->_hb_buffer), __pyx_v_cstr);

  /* "uharfbuzz/_harfbuzz.pyx":226
 *         cdef char cstr[5]
 *         hb_tag_to_string(hb_buffer_get_script(self._hb_buffer), cstr)
 *         cstr[4] = b'\0'             # <<<<<<<<<<<<<<
 *         if cstr[0] == b'\0':
 *             return None
 */
  (__pyx_v_cstr[4]) = '\x00';

  /* "uharfbuzz/_harfbuzz.pyx":227
 *         hb_tag_to_string(hb_buffer_get_script(self._hb_buffer), cstr)
 *         cstr[4] = b'\0'
 *         if cstr[0] == b'\0':             # <<<<<<<<<<<<<<
 *             return None
 *         cdef bytes packed = cstr
 */
  __pyx_t_1 = ((__pyx_v_cstr[0]) == '\x00');
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":228
 *         cstr[4] = b'\0'
 *         if cstr[0] == b'\0':
 *             return None             # <<<<<<<<<<<<<<
 *         cdef bytes packed = cstr
 *         return packed.decode()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uharfbuzz/_harfbuzz.pyx":227
 *         hb_tag_to_string(hb_buffer_get_script(self._hb_buffer), cstr)
 *         cstr[4] = b'\0'
 *         if cstr[0] == b'\0':             # <<<<<<<<<<<<<<
 *             return None
 *         cdef bytes packed = cstr
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":229
 *         if cstr[0] == b'\0':
 *             return None
 *         cdef bytes packed = cstr             # <<<<<<<<<<<<<<
 *         return packed.decode()
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_FromString(__pyx_v_cstr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_packed = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":230
 *             return None
 *         cdef bytes packed = cstr
 *         return packed.decode()             # <<<<<<<<<<<<<<
 * 
 *     @script.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_decode_bytes(__pyx_v_packed, 0, PY_SSIZE_T_MAX, NULL, NULL, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":222
 *             self._hb_buffer, hb_language_from_string(cstr, -1))
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def script(self) -> str:
 *         cdef char cstr[5]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.script.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_packed);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":232
 *         return packed.decode()
 * 
 *     @script.setter             # <<<<<<<<<<<<<<
 *     def script(self, value: str):
 *         cdef bytes packed = value.encode()
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_6script_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_6script_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyUnicode_Type), 0, "value", 1))) __PYX_ERR(0, 233, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_6script_2__set__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self), ((PyObject*)__pyx_v_value));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_6script_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_packed = 0;
  char *__pyx_v_cstr;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":234
 *     @script.setter
 *     def script(self, value: str):
 *         cdef bytes packed = value.encode()             # <<<<<<<<<<<<<<
 *         cdef char* cstr = packed
 *         # all the *_from_string calls should probably be checked and throw an
 */
  __pyx_t_1 = PyUnicode_AsEncodedString(__pyx_v_value, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_packed = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":235
 *     def script(self, value: str):
 *         cdef bytes packed = value.encode()
 *         cdef char* cstr = packed             # <<<<<<<<<<<<<<
 *         # all the *_from_string calls should probably be checked and throw an
 *         # exception if invalid
 */
  if (unlikely(__pyx_v_packed == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 235, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBytes_AsWritableString(__pyx_v_packed); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 235, __pyx_L1_error)
  __pyx_v_cstr = __pyx_t_2;

  /* "uharfbuzz/_harfbuzz.pyx":238
 *         # all the *_from_string calls should probably be checked and throw an
 *         # exception if invalid
 *         hb_buffer_set_script(             # <<<<<<<<<<<<<<
 *             self._hb_buffer, hb_script_from_string(cstr, -1))
 * 
 */
  hb_buffer_set_script(__pyx_v_self->_hb_buffer, hb_script_from_string(__pyx_v_cstr, -1));

  /* "uharfbuzz/_harfbuzz.pyx":232
 *         return packed.decode()
 * 
 *     @script.setter             # <<<<<<<<<<<<<<
 *     def script(self, value: str):
 *         cdef bytes packed = value.encode()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.script.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_packed);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":241
 *             self._hb_buffer, hb_script_from_string(cstr, -1))
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def flags(self) -> BufferFlags:
 *         level = hb_buffer_get_flags(self._hb_buffer)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_5flags_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_5flags_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_5flags___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_5flags___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self) {
  hb_buffer_flags_t __pyx_v_level;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":243
 *     @property
 *     def flags(self) -> BufferFlags:
 *         level = hb_buffer_get_flags(self._hb_buffer)             # <<<<<<<<<<<<<<
 *         return BufferFlags(level)
 * 
 */
  __pyx_v_level = hb_buffer_get_flags(__pyx_v_self->_hb_buffer);

  /* "uharfbuzz/_harfbuzz.pyx":244
 *     def flags(self) -> BufferFlags:
 *         level = hb_buffer_get_flags(self._hb_buffer)
 *         return BufferFlags(level)             # <<<<<<<<<<<<<<
 * 
 *     @flags.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BufferFlags); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_hb_buffer_flags_t(__pyx_v_level); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 244, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":241
 *             self._hb_buffer, hb_script_from_string(cstr, -1))
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def flags(self) -> BufferFlags:
 *         level = hb_buffer_get_flags(self._hb_buffer)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.flags.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":246
 *         return BufferFlags(level)
 * 
 *     @flags.setter             # <<<<<<<<<<<<<<
 *     def flags(self, value: BufferFlags):
 *         level = BufferFlags(value)
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_5flags_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_5flags_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_5flags_2__set__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_5flags_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_level = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  hb_buffer_flags_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":248
 *     @flags.setter
 *     def flags(self, value: BufferFlags):
 *         level = BufferFlags(value)             # <<<<<<<<<<<<<<
 *         hb_buffer_set_flags(self._hb_buffer, level)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BufferFlags); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 248, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_value};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 248, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_level = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":249
 *     def flags(self, value: BufferFlags):
 *         level = BufferFlags(value)
 *         hb_buffer_set_flags(self._hb_buffer, level)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_t_5 = ((hb_buffer_flags_t)__Pyx_PyInt_As_hb_buffer_flags_t(__pyx_v_level)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 249, __pyx_L1_error)
  hb_buffer_set_flags(__pyx_v_self->_hb_buffer, __pyx_t_5);

  /* "uharfbuzz/_harfbuzz.pyx":246
 *         return BufferFlags(level)
 * 
 *     @flags.setter             # <<<<<<<<<<<<<<
 *     def flags(self, value: BufferFlags):
 *         level = BufferFlags(value)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.flags.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_level);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":251
 *         hb_buffer_set_flags(self._hb_buffer, level)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def cluster_level(self) -> BufferClusterLevel:
 *         level = hb_buffer_get_cluster_level(self._hb_buffer)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_13cluster_level_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_13cluster_level_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_13cluster_level___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_13cluster_level___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self) {
  hb_buffer_cluster_level_t __pyx_v_level;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":253
 *     @property
 *     def cluster_level(self) -> BufferClusterLevel:
 *         level = hb_buffer_get_cluster_level(self._hb_buffer)             # <<<<<<<<<<<<<<
 *         return BufferClusterLevel(level)
 * 
 */
  __pyx_v_level = hb_buffer_get_cluster_level(__pyx_v_self->_hb_buffer);

  /* "uharfbuzz/_harfbuzz.pyx":254
 *     def cluster_level(self) -> BufferClusterLevel:
 *         level = hb_buffer_get_cluster_level(self._hb_buffer)
 *         return BufferClusterLevel(level)             # <<<<<<<<<<<<<<
 * 
 *     @cluster_level.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BufferClusterLevel); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_hb_buffer_cluster_level_t(__pyx_v_level); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 254, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":251
 *         hb_buffer_set_flags(self._hb_buffer, level)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def cluster_level(self) -> BufferClusterLevel:
 *         level = hb_buffer_get_cluster_level(self._hb_buffer)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.cluster_level.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":256
 *         return BufferClusterLevel(level)
 * 
 *     @cluster_level.setter             # <<<<<<<<<<<<<<
 *     def cluster_level(self, value: BufferClusterLevel):
 *         level = BufferClusterLevel(value)
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_13cluster_level_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_13cluster_level_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_13cluster_level_2__set__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_13cluster_level_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_level = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  hb_buffer_cluster_level_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":258
 *     @cluster_level.setter
 *     def cluster_level(self, value: BufferClusterLevel):
 *         level = BufferClusterLevel(value)             # <<<<<<<<<<<<<<
 *         hb_buffer_set_cluster_level(self._hb_buffer, level)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BufferClusterLevel); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_value};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_level = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":259
 *     def cluster_level(self, value: BufferClusterLevel):
 *         level = BufferClusterLevel(value)
 *         hb_buffer_set_cluster_level(self._hb_buffer, level)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_t_5 = ((hb_buffer_cluster_level_t)__Pyx_PyInt_As_hb_buffer_cluster_level_t(__pyx_v_level)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 259, __pyx_L1_error)
  hb_buffer_set_cluster_level(__pyx_v_self->_hb_buffer, __pyx_t_5);

  /* "uharfbuzz/_harfbuzz.pyx":256
 *         return BufferClusterLevel(level)
 * 
 *     @cluster_level.setter             # <<<<<<<<<<<<<<
 *     def cluster_level(self, value: BufferClusterLevel):
 *         level = BufferClusterLevel(value)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.cluster_level.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_level);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":261
 *         hb_buffer_set_cluster_level(self._hb_buffer, level)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def content_type(self) -> BufferContentType:
 *         level = hb_buffer_get_content_type(self._hb_buffer)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_12content_type_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_12content_type_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_12content_type___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_12content_type___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self) {
  hb_buffer_content_type_t __pyx_v_level;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":263
 *     @property
 *     def content_type(self) -> BufferContentType:
 *         level = hb_buffer_get_content_type(self._hb_buffer)             # <<<<<<<<<<<<<<
 *         return BufferContentType(level)
 * 
 */
  __pyx_v_level = hb_buffer_get_content_type(__pyx_v_self->_hb_buffer);

  /* "uharfbuzz/_harfbuzz.pyx":264
 *     def content_type(self) -> BufferContentType:
 *         level = hb_buffer_get_content_type(self._hb_buffer)
 *         return BufferContentType(level)             # <<<<<<<<<<<<<<
 * 
 *     @content_type.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BufferContentType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_hb_buffer_content_type_t(__pyx_v_level); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":261
 *         hb_buffer_set_cluster_level(self._hb_buffer, level)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def content_type(self) -> BufferContentType:
 *         level = hb_buffer_get_content_type(self._hb_buffer)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.content_type.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":266
 *         return BufferContentType(level)
 * 
 *     @content_type.setter             # <<<<<<<<<<<<<<
 *     def content_type(self, value: BufferContentType):
 *         level = BufferContentType(value)
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_12content_type_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_12content_type_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_12content_type_2__set__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_12content_type_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_level = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  hb_buffer_content_type_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":268
 *     @content_type.setter
 *     def content_type(self, value: BufferContentType):
 *         level = BufferContentType(value)             # <<<<<<<<<<<<<<
 *         hb_buffer_set_content_type(self._hb_buffer, level)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BufferContentType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_value};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_level = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":269
 *     def content_type(self, value: BufferContentType):
 *         level = BufferContentType(value)
 *         hb_buffer_set_content_type(self._hb_buffer, level)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_t_5 = ((hb_buffer_content_type_t)__Pyx_PyInt_As_hb_buffer_content_type_t(__pyx_v_level)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 269, __pyx_L1_error)
  hb_buffer_set_content_type(__pyx_v_self->_hb_buffer, __pyx_t_5);

  /* "uharfbuzz/_harfbuzz.pyx":266
 *         return BufferContentType(level)
 * 
 *     @content_type.setter             # <<<<<<<<<<<<<<
 *     def content_type(self, value: BufferContentType):
 *         level = BufferContentType(value)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.content_type.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_level);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":271
 *         hb_buffer_set_content_type(self._hb_buffer, level)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def replacement_codepoint(self) -> int:
 *         return hb_buffer_get_replacement_codepoint(self._hb_buffer)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_21replacement_codepoint_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_21replacement_codepoint_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_21replacement_codepoint___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_21replacement_codepoint___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":273
 *     @property
 *     def replacement_codepoint(self) -> int:
 *         return hb_buffer_get_replacement_codepoint(self._hb_buffer)             # <<<<<<<<<<<<<<
 * 
 *     @replacement_codepoint.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_hb_codepoint_t(hb_buffer_get_replacement_codepoint(__pyx_v_self->_hb_buffer)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":271
 *         hb_buffer_set_content_type(self._hb_buffer, level)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def replacement_codepoint(self) -> int:
 *         return hb_buffer_get_replacement_codepoint(self._hb_buffer)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.replacement_codepoint.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":275
 *         return hb_buffer_get_replacement_codepoint(self._hb_buffer)
 * 
 *     @replacement_codepoint.setter             # <<<<<<<<<<<<<<
 *     def replacement_codepoint(self, value: int):
 *         hb_buffer_set_replacement_codepoint(self._hb_buffer, value)
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_21replacement_codepoint_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_21replacement_codepoint_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyInt_Type), 0, "value", 1))) __PYX_ERR(0, 276, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_21replacement_codepoint_2__set__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self), ((PyObject*)__pyx_v_value));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_21replacement_codepoint_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  hb_codepoint_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "uharfbuzz/_harfbuzz.pyx":277
 *     @replacement_codepoint.setter
 *     def replacement_codepoint(self, value: int):
 *         hb_buffer_set_replacement_codepoint(self._hb_buffer, value)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_t_1 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_value); if (unlikely((__pyx_t_1 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 277, __pyx_L1_error)
  hb_buffer_set_replacement_codepoint(__pyx_v_self->_hb_buffer, __pyx_t_1);

  /* "uharfbuzz/_harfbuzz.pyx":275
 *         return hb_buffer_get_replacement_codepoint(self._hb_buffer)
 * 
 *     @replacement_codepoint.setter             # <<<<<<<<<<<<<<
 *     def replacement_codepoint(self, value: int):
 *         hb_buffer_set_replacement_codepoint(self._hb_buffer, value)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.replacement_codepoint.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":279
 *         hb_buffer_set_replacement_codepoint(self._hb_buffer, value)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def invisible_glyph(self) -> int:
 *         return hb_buffer_get_invisible_glyph(self._hb_buffer)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_15invisible_glyph_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_15invisible_glyph_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_15invisible_glyph___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_15invisible_glyph___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":281
 *     @property
 *     def invisible_glyph(self) -> int:
 *         return hb_buffer_get_invisible_glyph(self._hb_buffer)             # <<<<<<<<<<<<<<
 * 
 *     @invisible_glyph.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_hb_codepoint_t(hb_buffer_get_invisible_glyph(__pyx_v_self->_hb_buffer)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":279
 *         hb_buffer_set_replacement_codepoint(self._hb_buffer, value)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def invisible_glyph(self) -> int:
 *         return hb_buffer_get_invisible_glyph(self._hb_buffer)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.invisible_glyph.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":283
 *         return hb_buffer_get_invisible_glyph(self._hb_buffer)
 * 
 *     @invisible_glyph.setter             # <<<<<<<<<<<<<<
 *     def invisible_glyph(self, value: int):
 *         hb_buffer_set_invisible_glyph(self._hb_buffer, value)
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_15invisible_glyph_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_15invisible_glyph_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyInt_Type), 0, "value", 1))) __PYX_ERR(0, 284, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_15invisible_glyph_2__set__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self), ((PyObject*)__pyx_v_value));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_15invisible_glyph_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  hb_codepoint_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "uharfbuzz/_harfbuzz.pyx":285
 *     @invisible_glyph.setter
 *     def invisible_glyph(self, value: int):
 *         hb_buffer_set_invisible_glyph(self._hb_buffer, value)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_t_1 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_value); if (unlikely((__pyx_t_1 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 285, __pyx_L1_error)
  hb_buffer_set_invisible_glyph(__pyx_v_self->_hb_buffer, __pyx_t_1);

  /* "uharfbuzz/_harfbuzz.pyx":283
 *         return hb_buffer_get_invisible_glyph(self._hb_buffer)
 * 
 *     @invisible_glyph.setter             # <<<<<<<<<<<<<<
 *     def invisible_glyph(self, value: int):
 *         hb_buffer_set_invisible_glyph(self._hb_buffer, value)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.invisible_glyph.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":287
 *         hb_buffer_set_invisible_glyph(self._hb_buffer, value)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def not_found_glyph(self) -> int:
 *         return hb_buffer_get_not_found_glyph(self._hb_buffer)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_15not_found_glyph_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_15not_found_glyph_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_15not_found_glyph___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_15not_found_glyph___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":289
 *     @property
 *     def not_found_glyph(self) -> int:
 *         return hb_buffer_get_not_found_glyph(self._hb_buffer)             # <<<<<<<<<<<<<<
 * 
 *     @not_found_glyph.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_hb_codepoint_t(hb_buffer_get_not_found_glyph(__pyx_v_self->_hb_buffer)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":287
 *         hb_buffer_set_invisible_glyph(self._hb_buffer, value)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def not_found_glyph(self) -> int:
 *         return hb_buffer_get_not_found_glyph(self._hb_buffer)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.not_found_glyph.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":291
 *         return hb_buffer_get_not_found_glyph(self._hb_buffer)
 * 
 *     @not_found_glyph.setter             # <<<<<<<<<<<<<<
 *     def not_found_glyph(self, value: int):
 *         hb_buffer_set_not_found_glyph(self._hb_buffer, value)
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_15not_found_glyph_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_15not_found_glyph_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyInt_Type), 0, "value", 1))) __PYX_ERR(0, 292, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_15not_found_glyph_2__set__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self), ((PyObject*)__pyx_v_value));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_15not_found_glyph_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  hb_codepoint_t __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "uharfbuzz/_harfbuzz.pyx":293
 *     @not_found_glyph.setter
 *     def not_found_glyph(self, value: int):
 *         hb_buffer_set_not_found_glyph(self._hb_buffer, value)             # <<<<<<<<<<<<<<
 * 
 *     def set_language_from_ot_tag(self, value: str):
 */
  __pyx_t_1 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_value); if (unlikely((__pyx_t_1 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 293, __pyx_L1_error)
  hb_buffer_set_not_found_glyph(__pyx_v_self->_hb_buffer, __pyx_t_1);

  /* "uharfbuzz/_harfbuzz.pyx":291
 *         return hb_buffer_get_not_found_glyph(self._hb_buffer)
 * 
 *     @not_found_glyph.setter             # <<<<<<<<<<<<<<
 *     def not_found_glyph(self, value: int):
 *         hb_buffer_set_not_found_glyph(self._hb_buffer, value)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.not_found_glyph.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":295
 *         hb_buffer_set_not_found_glyph(self._hb_buffer, value)
 * 
 *     def set_language_from_ot_tag(self, value: str):             # <<<<<<<<<<<<<<
 *         cdef bytes packed = value.encode()
 *         cdef char* cstr = packed
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_13set_language_from_ot_tag(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_6Buffer_13set_language_from_ot_tag = {"set_language_from_ot_tag", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_13set_language_from_ot_tag, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_13set_language_from_ot_tag(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_language_from_ot_tag (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 295, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_language_from_ot_tag") < 0)) __PYX_ERR(0, 295, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_value = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_language_from_ot_tag", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 295, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.set_language_from_ot_tag", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyUnicode_Type), 0, "value", 1))) __PYX_ERR(0, 295, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_12set_language_from_ot_tag(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_12set_language_from_ot_tag(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_packed = 0;
  char *__pyx_v_cstr;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_language_from_ot_tag", 1);

  /* "uharfbuzz/_harfbuzz.pyx":296
 * 
 *     def set_language_from_ot_tag(self, value: str):
 *         cdef bytes packed = value.encode()             # <<<<<<<<<<<<<<
 *         cdef char* cstr = packed
 *         hb_buffer_set_language(
 */
  __pyx_t_1 = PyUnicode_AsEncodedString(__pyx_v_value, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_packed = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":297
 *     def set_language_from_ot_tag(self, value: str):
 *         cdef bytes packed = value.encode()
 *         cdef char* cstr = packed             # <<<<<<<<<<<<<<
 *         hb_buffer_set_language(
 *             self._hb_buffer, hb_ot_tag_to_language(hb_tag_from_string(cstr, -1)))
 */
  if (unlikely(__pyx_v_packed == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 297, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBytes_AsWritableString(__pyx_v_packed); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 297, __pyx_L1_error)
  __pyx_v_cstr = __pyx_t_2;

  /* "uharfbuzz/_harfbuzz.pyx":298
 *         cdef bytes packed = value.encode()
 *         cdef char* cstr = packed
 *         hb_buffer_set_language(             # <<<<<<<<<<<<<<
 *             self._hb_buffer, hb_ot_tag_to_language(hb_tag_from_string(cstr, -1)))
 * 
 */
  hb_buffer_set_language(__pyx_v_self->_hb_buffer, hb_ot_tag_to_language(hb_tag_from_string(__pyx_v_cstr, -1)));

  /* "uharfbuzz/_harfbuzz.pyx":295
 *         hb_buffer_set_not_found_glyph(self._hb_buffer, value)
 * 
 *     def set_language_from_ot_tag(self, value: str):             # <<<<<<<<<<<<<<
 *         cdef bytes packed = value.encode()
 *         cdef char* cstr = packed
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.set_language_from_ot_tag", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_packed);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":301
 *             self._hb_buffer, hb_ot_tag_to_language(hb_tag_from_string(cstr, -1)))
 * 
 *     def set_script_from_ot_tag(self, value: str):             # <<<<<<<<<<<<<<
 *         cdef bytes packed = value.encode()
 *         cdef char* cstr = packed
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_15set_script_from_ot_tag(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_6Buffer_15set_script_from_ot_tag = {"set_script_from_ot_tag", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_15set_script_from_ot_tag, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_15set_script_from_ot_tag(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_script_from_ot_tag (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 301, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_script_from_ot_tag") < 0)) __PYX_ERR(0, 301, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_value = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_script_from_ot_tag", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 301, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.set_script_from_ot_tag", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyUnicode_Type), 0, "value", 1))) __PYX_ERR(0, 301, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_14set_script_from_ot_tag(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_14set_script_from_ot_tag(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_packed = 0;
  char *__pyx_v_cstr;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_script_from_ot_tag", 1);

  /* "uharfbuzz/_harfbuzz.pyx":302
 * 
 *     def set_script_from_ot_tag(self, value: str):
 *         cdef bytes packed = value.encode()             # <<<<<<<<<<<<<<
 *         cdef char* cstr = packed
 *         hb_buffer_set_script(
 */
  __pyx_t_1 = PyUnicode_AsEncodedString(__pyx_v_value, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_packed = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":303
 *     def set_script_from_ot_tag(self, value: str):
 *         cdef bytes packed = value.encode()
 *         cdef char* cstr = packed             # <<<<<<<<<<<<<<
 *         hb_buffer_set_script(
 *             self._hb_buffer, hb_ot_tag_to_script(hb_tag_from_string(cstr, -1)))
 */
  if (unlikely(__pyx_v_packed == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 303, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBytes_AsWritableString(__pyx_v_packed); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 303, __pyx_L1_error)
  __pyx_v_cstr = __pyx_t_2;

  /* "uharfbuzz/_harfbuzz.pyx":304
 *         cdef bytes packed = value.encode()
 *         cdef char* cstr = packed
 *         hb_buffer_set_script(             # <<<<<<<<<<<<<<
 *             self._hb_buffer, hb_ot_tag_to_script(hb_tag_from_string(cstr, -1)))
 * 
 */
  hb_buffer_set_script(__pyx_v_self->_hb_buffer, hb_ot_tag_to_script(hb_tag_from_string(__pyx_v_cstr, -1)));

  /* "uharfbuzz/_harfbuzz.pyx":301
 *             self._hb_buffer, hb_ot_tag_to_language(hb_tag_from_string(cstr, -1)))
 * 
 *     def set_script_from_ot_tag(self, value: str):             # <<<<<<<<<<<<<<
 *         cdef bytes packed = value.encode()
 *         cdef char* cstr = packed
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.set_script_from_ot_tag", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_packed);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":307
 *             self._hb_buffer, hb_ot_tag_to_script(hb_tag_from_string(cstr, -1)))
 * 
 *     def add_codepoints(self, codepoints: List[int],             # <<<<<<<<<<<<<<
 *                        item_offset: int = 0, item_length: int = -1) -> None:
 *         cdef unsigned int size = len(codepoints)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_17add_codepoints(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_6Buffer_17add_codepoints = {"add_codepoints", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_17add_codepoints, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_17add_codepoints(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_codepoints = 0;
  PyObject *__pyx_v_item_offset = 0;
  PyObject *__pyx_v_item_length = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_codepoints (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_codepoints,&__pyx_n_s_item_offset,&__pyx_n_s_item_length,0};
    values[1] = __Pyx_Arg_NewRef_FASTCALL(__pyx_k_);
    values[2] = __Pyx_Arg_NewRef_FASTCALL(__pyx_k__2);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_codepoints)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 307, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_item_offset);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 307, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_item_length);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 307, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "add_codepoints") < 0)) __PYX_ERR(0, 307, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_codepoints = ((PyObject*)values[0]);
    __pyx_v_item_offset = ((PyObject*)values[1]);
    __pyx_v_item_length = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("add_codepoints", 0, 1, 3, __pyx_nargs); __PYX_ERR(0, 307, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.add_codepoints", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_codepoints), (&PyList_Type), 0, "codepoints", 1))) __PYX_ERR(0, 307, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_item_offset), (&PyInt_Type), 0, "item_offset", 1))) __PYX_ERR(0, 308, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_item_length), (&PyInt_Type), 0, "item_length", 1))) __PYX_ERR(0, 308, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_16add_codepoints(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self), __pyx_v_codepoints, __pyx_v_item_offset, __pyx_v_item_length);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_16add_codepoints(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self, PyObject *__pyx_v_codepoints, PyObject *__pyx_v_item_offset, PyObject *__pyx_v_item_length) {
  unsigned int __pyx_v_size;
  hb_codepoint_t *__pyx_v_hb_codepoints;
  unsigned int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  unsigned int __pyx_t_3;
  unsigned int __pyx_t_4;
  unsigned int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  hb_codepoint_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_codepoints", 1);

  /* "uharfbuzz/_harfbuzz.pyx":309
 *     def add_codepoints(self, codepoints: List[int],
 *                        item_offset: int = 0, item_length: int = -1) -> None:
 *         cdef unsigned int size = len(codepoints)             # <<<<<<<<<<<<<<
 *         cdef hb_codepoint_t* hb_codepoints
 *         if not size:
 */
  __pyx_t_1 = __Pyx_PyList_GET_SIZE(__pyx_v_codepoints); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 309, __pyx_L1_error)
  __pyx_v_size = __pyx_t_1;

  /* "uharfbuzz/_harfbuzz.pyx":311
 *         cdef unsigned int size = len(codepoints)
 *         cdef hb_codepoint_t* hb_codepoints
 *         if not size:             # <<<<<<<<<<<<<<
 *             return
 *         hb_codepoints = <hb_codepoint_t*>malloc(
 */
  __pyx_t_2 = (!(__pyx_v_size != 0));
  if (__pyx_t_2) {

    /* "uharfbuzz/_harfbuzz.pyx":312
 *         cdef hb_codepoint_t* hb_codepoints
 *         if not size:
 *             return             # <<<<<<<<<<<<<<
 *         hb_codepoints = <hb_codepoint_t*>malloc(
 *             size * sizeof(hb_codepoint_t))
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uharfbuzz/_harfbuzz.pyx":311
 *         cdef unsigned int size = len(codepoints)
 *         cdef hb_codepoint_t* hb_codepoints
 *         if not size:             # <<<<<<<<<<<<<<
 *             return
 *         hb_codepoints = <hb_codepoint_t*>malloc(
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":313
 *         if not size:
 *             return
 *         hb_codepoints = <hb_codepoint_t*>malloc(             # <<<<<<<<<<<<<<
 *             size * sizeof(hb_codepoint_t))
 *         for i in range(size):
 */
  __pyx_v_hb_codepoints = ((hb_codepoint_t *)malloc((__pyx_v_size * (sizeof(hb_codepoint_t)))));

  /* "uharfbuzz/_harfbuzz.pyx":315
 *         hb_codepoints = <hb_codepoint_t*>malloc(
 *             size * sizeof(hb_codepoint_t))
 *         for i in range(size):             # <<<<<<<<<<<<<<
 *             hb_codepoints[i] = codepoints[i]
 *         hb_buffer_add_codepoints(
 */
  __pyx_t_3 = __pyx_v_size;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "uharfbuzz/_harfbuzz.pyx":316
 *             size * sizeof(hb_codepoint_t))
 *         for i in range(size):
 *             hb_codepoints[i] = codepoints[i]             # <<<<<<<<<<<<<<
 *         hb_buffer_add_codepoints(
 *             self._hb_buffer, hb_codepoints, size, item_offset, item_length)
 */
    __pyx_t_6 = __Pyx_GetItemInt_List(__pyx_v_codepoints, __pyx_v_i, unsigned int, 0, __Pyx_PyInt_From_unsigned_int, 1, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 316, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_t_6); if (unlikely((__pyx_t_7 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 316, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    (__pyx_v_hb_codepoints[__pyx_v_i]) = __pyx_t_7;
  }

  /* "uharfbuzz/_harfbuzz.pyx":318
 *             hb_codepoints[i] = codepoints[i]
 *         hb_buffer_add_codepoints(
 *             self._hb_buffer, hb_codepoints, size, item_offset, item_length)             # <<<<<<<<<<<<<<
 *         free(hb_codepoints)
 *         if not hb_buffer_allocation_successful(self._hb_buffer):
 */
  __pyx_t_3 = __Pyx_PyInt_As_unsigned_int(__pyx_v_item_offset); if (unlikely((__pyx_t_3 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 318, __pyx_L1_error)
  __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_v_item_length); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 318, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":317
 *         for i in range(size):
 *             hb_codepoints[i] = codepoints[i]
 *         hb_buffer_add_codepoints(             # <<<<<<<<<<<<<<
 *             self._hb_buffer, hb_codepoints, size, item_offset, item_length)
 *         free(hb_codepoints)
 */
  hb_buffer_add_codepoints(__pyx_v_self->_hb_buffer, __pyx_v_hb_codepoints, __pyx_v_size, __pyx_t_3, __pyx_t_8);

  /* "uharfbuzz/_harfbuzz.pyx":319
 *         hb_buffer_add_codepoints(
 *             self._hb_buffer, hb_codepoints, size, item_offset, item_length)
 *         free(hb_codepoints)             # <<<<<<<<<<<<<<
 *         if not hb_buffer_allocation_successful(self._hb_buffer):
 *             raise MemoryError()
 */
  free(__pyx_v_hb_codepoints);

  /* "uharfbuzz/_harfbuzz.pyx":320
 *             self._hb_buffer, hb_codepoints, size, item_offset, item_length)
 *         free(hb_codepoints)
 *         if not hb_buffer_allocation_successful(self._hb_buffer):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_2 = (!(hb_buffer_allocation_successful(__pyx_v_self->_hb_buffer) != 0));
  if (unlikely(__pyx_t_2)) {

    /* "uharfbuzz/_harfbuzz.pyx":321
 *         free(hb_codepoints)
 *         if not hb_buffer_allocation_successful(self._hb_buffer):
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *     def add_utf8(self, text: bytes,
 */
    PyErr_NoMemory(); __PYX_ERR(0, 321, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":320
 *             self._hb_buffer, hb_codepoints, size, item_offset, item_length)
 *         free(hb_codepoints)
 *         if not hb_buffer_allocation_successful(self._hb_buffer):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":307
 *             self._hb_buffer, hb_ot_tag_to_script(hb_tag_from_string(cstr, -1)))
 * 
 *     def add_codepoints(self, codepoints: List[int],             # <<<<<<<<<<<<<<
 *                        item_offset: int = 0, item_length: int = -1) -> None:
 *         cdef unsigned int size = len(codepoints)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.add_codepoints", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":323
 *             raise MemoryError()
 * 
 *     def add_utf8(self, text: bytes,             # <<<<<<<<<<<<<<
 *                  item_offset: int = 0, item_length: int = -1) -> None:
 *         hb_buffer_add_utf8(
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_19add_utf8(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_6Buffer_19add_utf8 = {"add_utf8", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_19add_utf8, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_19add_utf8(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_text = 0;
  PyObject *__pyx_v_item_offset = 0;
  PyObject *__pyx_v_item_length = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_utf8 (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_text,&__pyx_n_s_item_offset,&__pyx_n_s_item_length,0};
    values[1] = __Pyx_Arg_NewRef_FASTCALL(__pyx_k__3);
    values[2] = __Pyx_Arg_NewRef_FASTCALL(__pyx_k__4);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_text)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 323, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_item_offset);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 323, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_item_length);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 323, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "add_utf8") < 0)) __PYX_ERR(0, 323, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_text = ((PyObject*)values[0]);
    __pyx_v_item_offset = ((PyObject*)values[1]);
    __pyx_v_item_length = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("add_utf8", 0, 1, 3, __pyx_nargs); __PYX_ERR(0, 323, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.add_utf8", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_text), (&PyBytes_Type), 0, "text", 1))) __PYX_ERR(0, 323, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_item_offset), (&PyInt_Type), 0, "item_offset", 1))) __PYX_ERR(0, 324, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_item_length), (&PyInt_Type), 0, "item_length", 1))) __PYX_ERR(0, 324, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_18add_utf8(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self), __pyx_v_text, __pyx_v_item_offset, __pyx_v_item_length);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_18add_utf8(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self, PyObject *__pyx_v_text, PyObject *__pyx_v_item_offset, PyObject *__pyx_v_item_length) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  char const *__pyx_t_1;
  Py_ssize_t __pyx_t_2;
  unsigned int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_utf8", 1);

  /* "uharfbuzz/_harfbuzz.pyx":326
 *                  item_offset: int = 0, item_length: int = -1) -> None:
 *         hb_buffer_add_utf8(
 *             self._hb_buffer, text, len(text), item_offset, item_length)             # <<<<<<<<<<<<<<
 *         if not hb_buffer_allocation_successful(self._hb_buffer):
 *             raise MemoryError()
 */
  __pyx_t_1 = __Pyx_PyBytes_AsString(__pyx_v_text); if (unlikely((!__pyx_t_1) && PyErr_Occurred())) __PYX_ERR(0, 326, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyBytes_GET_SIZE(__pyx_v_text); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 326, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyInt_As_unsigned_int(__pyx_v_item_offset); if (unlikely((__pyx_t_3 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 326, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_item_length); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 326, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":325
 *     def add_utf8(self, text: bytes,
 *                  item_offset: int = 0, item_length: int = -1) -> None:
 *         hb_buffer_add_utf8(             # <<<<<<<<<<<<<<
 *             self._hb_buffer, text, len(text), item_offset, item_length)
 *         if not hb_buffer_allocation_successful(self._hb_buffer):
 */
  hb_buffer_add_utf8(__pyx_v_self->_hb_buffer, __pyx_t_1, __pyx_t_2, __pyx_t_3, __pyx_t_4);

  /* "uharfbuzz/_harfbuzz.pyx":327
 *         hb_buffer_add_utf8(
 *             self._hb_buffer, text, len(text), item_offset, item_length)
 *         if not hb_buffer_allocation_successful(self._hb_buffer):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_5 = (!(hb_buffer_allocation_successful(__pyx_v_self->_hb_buffer) != 0));
  if (unlikely(__pyx_t_5)) {

    /* "uharfbuzz/_harfbuzz.pyx":328
 *             self._hb_buffer, text, len(text), item_offset, item_length)
 *         if not hb_buffer_allocation_successful(self._hb_buffer):
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *     def add_str(self, text: str,
 */
    PyErr_NoMemory(); __PYX_ERR(0, 328, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":327
 *         hb_buffer_add_utf8(
 *             self._hb_buffer, text, len(text), item_offset, item_length)
 *         if not hb_buffer_allocation_successful(self._hb_buffer):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":323
 *             raise MemoryError()
 * 
 *     def add_utf8(self, text: bytes,             # <<<<<<<<<<<<<<
 *                  item_offset: int = 0, item_length: int = -1) -> None:
 *         hb_buffer_add_utf8(
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.add_utf8", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":330
 *             raise MemoryError()
 * 
 *     def add_str(self, text: str,             # <<<<<<<<<<<<<<
 *                 item_offset: int = 0, item_length: int = -1) -> None:
 *         # ensure unicode string is in the "canonical" representation
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_21add_str(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_6Buffer_21add_str = {"add_str", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_21add_str, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_21add_str(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_text = 0;
  PyObject *__pyx_v_item_offset = 0;
  PyObject *__pyx_v_item_length = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_str (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_text,&__pyx_n_s_item_offset,&__pyx_n_s_item_length,0};
    values[1] = __Pyx_Arg_NewRef_FASTCALL(__pyx_k__5);
    values[2] = __Pyx_Arg_NewRef_FASTCALL(__pyx_k__6);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_text)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 330, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_item_offset);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 330, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_item_length);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 330, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "add_str") < 0)) __PYX_ERR(0, 330, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_text = ((PyObject*)values[0]);
    __pyx_v_item_offset = ((PyObject*)values[1]);
    __pyx_v_item_length = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("add_str", 0, 1, 3, __pyx_nargs); __PYX_ERR(0, 330, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.add_str", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_text), (&PyUnicode_Type), 0, "text", 1))) __PYX_ERR(0, 330, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_item_offset), (&PyInt_Type), 0, "item_offset", 1))) __PYX_ERR(0, 331, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_item_length), (&PyInt_Type), 0, "item_length", 1))) __PYX_ERR(0, 331, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_20add_str(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self), __pyx_v_text, __pyx_v_item_offset, __pyx_v_item_length);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_20add_str(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self, PyObject *__pyx_v_text, PyObject *__pyx_v_item_offset, PyObject *__pyx_v_item_length) {
  Py_ssize_t __pyx_v_length;
  int __pyx_v_kind;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  unsigned int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_str", 1);

  /* "uharfbuzz/_harfbuzz.pyx":333
 *                 item_offset: int = 0, item_length: int = -1) -> None:
 *         # ensure unicode string is in the "canonical" representation
 *         assert PyUnicode_IS_READY(text)             # <<<<<<<<<<<<<<
 * 
 *         cdef Py_ssize_t length = PyUnicode_GET_LENGTH(text)
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(__pyx_assertions_enabled())) {
    __pyx_t_1 = PyUnicode_IS_READY(__pyx_v_text);
    if (unlikely(!__pyx_t_1)) {
      __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
      __PYX_ERR(0, 333, __pyx_L1_error)
    }
  }
  #else
  if ((1)); else __PYX_ERR(0, 333, __pyx_L1_error)
  #endif

  /* "uharfbuzz/_harfbuzz.pyx":335
 *         assert PyUnicode_IS_READY(text)
 * 
 *         cdef Py_ssize_t length = PyUnicode_GET_LENGTH(text)             # <<<<<<<<<<<<<<
 *         cdef int kind = PyUnicode_KIND(text)
 * 
 */
  __pyx_v_length = PyUnicode_GET_LENGTH(__pyx_v_text);

  /* "uharfbuzz/_harfbuzz.pyx":336
 * 
 *         cdef Py_ssize_t length = PyUnicode_GET_LENGTH(text)
 *         cdef int kind = PyUnicode_KIND(text)             # <<<<<<<<<<<<<<
 * 
 *         if kind == PyUnicode_1BYTE_KIND:
 */
  __pyx_v_kind = PyUnicode_KIND(__pyx_v_text);

  /* "uharfbuzz/_harfbuzz.pyx":338
 *         cdef int kind = PyUnicode_KIND(text)
 * 
 *         if kind == PyUnicode_1BYTE_KIND:             # <<<<<<<<<<<<<<
 *             hb_buffer_add_latin1(
 *                 self._hb_buffer,
 */
  __pyx_t_1 = (__pyx_v_kind == PyUnicode_1BYTE_KIND);
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":343
 *                 <uint8_t*>PyUnicode_1BYTE_DATA(text),
 *                 length,
 *                 item_offset,             # <<<<<<<<<<<<<<
 *                 item_length,
 *             )
 */
    __pyx_t_2 = __Pyx_PyInt_As_unsigned_int(__pyx_v_item_offset); if (unlikely((__pyx_t_2 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 343, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":344
 *                 length,
 *                 item_offset,
 *                 item_length,             # <<<<<<<<<<<<<<
 *             )
 *         elif kind == PyUnicode_2BYTE_KIND:
 */
    __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_v_item_length); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 344, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":339
 * 
 *         if kind == PyUnicode_1BYTE_KIND:
 *             hb_buffer_add_latin1(             # <<<<<<<<<<<<<<
 *                 self._hb_buffer,
 *                 <uint8_t*>PyUnicode_1BYTE_DATA(text),
 */
    hb_buffer_add_latin1(__pyx_v_self->_hb_buffer, ((uint8_t *)PyUnicode_1BYTE_DATA(__pyx_v_text)), __pyx_v_length, __pyx_t_2, __pyx_t_3);

    /* "uharfbuzz/_harfbuzz.pyx":338
 *         cdef int kind = PyUnicode_KIND(text)
 * 
 *         if kind == PyUnicode_1BYTE_KIND:             # <<<<<<<<<<<<<<
 *             hb_buffer_add_latin1(
 *                 self._hb_buffer,
 */
    goto __pyx_L3;
  }

  /* "uharfbuzz/_harfbuzz.pyx":346
 *                 item_length,
 *             )
 *         elif kind == PyUnicode_2BYTE_KIND:             # <<<<<<<<<<<<<<
 *             hb_buffer_add_utf16(
 *                 self._hb_buffer,
 */
  __pyx_t_1 = (__pyx_v_kind == PyUnicode_2BYTE_KIND);
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":351
 *                 <uint16_t*>PyUnicode_2BYTE_DATA(text),
 *                 length,
 *                 item_offset,             # <<<<<<<<<<<<<<
 *                 item_length,
 *             )
 */
    __pyx_t_2 = __Pyx_PyInt_As_unsigned_int(__pyx_v_item_offset); if (unlikely((__pyx_t_2 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 351, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":352
 *                 length,
 *                 item_offset,
 *                 item_length,             # <<<<<<<<<<<<<<
 *             )
 *         elif kind == PyUnicode_4BYTE_KIND:
 */
    __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_v_item_length); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 352, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":347
 *             )
 *         elif kind == PyUnicode_2BYTE_KIND:
 *             hb_buffer_add_utf16(             # <<<<<<<<<<<<<<
 *                 self._hb_buffer,
 *                 <uint16_t*>PyUnicode_2BYTE_DATA(text),
 */
    hb_buffer_add_utf16(__pyx_v_self->_hb_buffer, ((uint16_t *)PyUnicode_2BYTE_DATA(__pyx_v_text)), __pyx_v_length, __pyx_t_2, __pyx_t_3);

    /* "uharfbuzz/_harfbuzz.pyx":346
 *                 item_length,
 *             )
 *         elif kind == PyUnicode_2BYTE_KIND:             # <<<<<<<<<<<<<<
 *             hb_buffer_add_utf16(
 *                 self._hb_buffer,
 */
    goto __pyx_L3;
  }

  /* "uharfbuzz/_harfbuzz.pyx":354
 *                 item_length,
 *             )
 *         elif kind == PyUnicode_4BYTE_KIND:             # <<<<<<<<<<<<<<
 *             hb_buffer_add_utf32(
 *                 self._hb_buffer,
 */
  __pyx_t_1 = (__pyx_v_kind == PyUnicode_4BYTE_KIND);
  if (likely(__pyx_t_1)) {

    /* "uharfbuzz/_harfbuzz.pyx":359
 *                 <uint32_t*>PyUnicode_4BYTE_DATA(text),
 *                 length,
 *                 item_offset,             # <<<<<<<<<<<<<<
 *                 item_length,
 *             )
 */
    __pyx_t_2 = __Pyx_PyInt_As_unsigned_int(__pyx_v_item_offset); if (unlikely((__pyx_t_2 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 359, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":360
 *                 length,
 *                 item_offset,
 *                 item_length,             # <<<<<<<<<<<<<<
 *             )
 *         else:
 */
    __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_v_item_length); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 360, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":355
 *             )
 *         elif kind == PyUnicode_4BYTE_KIND:
 *             hb_buffer_add_utf32(             # <<<<<<<<<<<<<<
 *                 self._hb_buffer,
 *                 <uint32_t*>PyUnicode_4BYTE_DATA(text),
 */
    hb_buffer_add_utf32(__pyx_v_self->_hb_buffer, ((uint32_t *)PyUnicode_4BYTE_DATA(__pyx_v_text)), __pyx_v_length, __pyx_t_2, __pyx_t_3);

    /* "uharfbuzz/_harfbuzz.pyx":354
 *                 item_length,
 *             )
 *         elif kind == PyUnicode_4BYTE_KIND:             # <<<<<<<<<<<<<<
 *             hb_buffer_add_utf32(
 *                 self._hb_buffer,
 */
    goto __pyx_L3;
  }

  /* "uharfbuzz/_harfbuzz.pyx":363
 *             )
 *         else:
 *             raise AssertionError(kind)             # <<<<<<<<<<<<<<
 *         if not hb_buffer_allocation_successful(self._hb_buffer):
 *             raise MemoryError()
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_kind); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 363, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_AssertionError, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 363, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 363, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "uharfbuzz/_harfbuzz.pyx":364
 *         else:
 *             raise AssertionError(kind)
 *         if not hb_buffer_allocation_successful(self._hb_buffer):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_1 = (!(hb_buffer_allocation_successful(__pyx_v_self->_hb_buffer) != 0));
  if (unlikely(__pyx_t_1)) {

    /* "uharfbuzz/_harfbuzz.pyx":365
 *             raise AssertionError(kind)
 *         if not hb_buffer_allocation_successful(self._hb_buffer):
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *     def guess_segment_properties(self) -> None:
 */
    PyErr_NoMemory(); __PYX_ERR(0, 365, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":364
 *         else:
 *             raise AssertionError(kind)
 *         if not hb_buffer_allocation_successful(self._hb_buffer):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":330
 *             raise MemoryError()
 * 
 *     def add_str(self, text: str,             # <<<<<<<<<<<<<<
 *                 item_offset: int = 0, item_length: int = -1) -> None:
 *         # ensure unicode string is in the "canonical" representation
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.add_str", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":367
 *             raise MemoryError()
 * 
 *     def guess_segment_properties(self) -> None:             # <<<<<<<<<<<<<<
 *         hb_buffer_guess_segment_properties(self._hb_buffer)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_23guess_segment_properties(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_6Buffer_23guess_segment_properties = {"guess_segment_properties", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_23guess_segment_properties, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_23guess_segment_properties(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("guess_segment_properties (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("guess_segment_properties", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "guess_segment_properties", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_22guess_segment_properties(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_22guess_segment_properties(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("guess_segment_properties", 1);

  /* "uharfbuzz/_harfbuzz.pyx":368
 * 
 *     def guess_segment_properties(self) -> None:
 *         hb_buffer_guess_segment_properties(self._hb_buffer)             # <<<<<<<<<<<<<<
 * 
 *     def set_message_func(self, callback) -> None:
 */
  hb_buffer_guess_segment_properties(__pyx_v_self->_hb_buffer);

  /* "uharfbuzz/_harfbuzz.pyx":367
 *             raise MemoryError()
 * 
 *     def guess_segment_properties(self) -> None:             # <<<<<<<<<<<<<<
 *         hb_buffer_guess_segment_properties(self._hb_buffer)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":370
 *         hb_buffer_guess_segment_properties(self._hb_buffer)
 * 
 *     def set_message_func(self, callback) -> None:             # <<<<<<<<<<<<<<
 *         self._message_callback = callback
 *         hb_buffer_set_message_func(self._hb_buffer, msgcallback, <void*>callback, NULL)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_25set_message_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_6Buffer_25set_message_func = {"set_message_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_25set_message_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_25set_message_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_callback = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_message_func (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_callback,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_callback)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 370, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_message_func") < 0)) __PYX_ERR(0, 370, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_callback = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_message_func", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 370, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.set_message_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_24set_message_func(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self), __pyx_v_callback);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_24set_message_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self, PyObject *__pyx_v_callback) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_message_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":371
 * 
 *     def set_message_func(self, callback) -> None:
 *         self._message_callback = callback             # <<<<<<<<<<<<<<
 *         hb_buffer_set_message_func(self._hb_buffer, msgcallback, <void*>callback, NULL)
 * 
 */
  __Pyx_INCREF(__pyx_v_callback);
  __Pyx_GIVEREF(__pyx_v_callback);
  __Pyx_GOTREF(__pyx_v_self->_message_callback);
  __Pyx_DECREF(__pyx_v_self->_message_callback);
  __pyx_v_self->_message_callback = __pyx_v_callback;

  /* "uharfbuzz/_harfbuzz.pyx":372
 *     def set_message_func(self, callback) -> None:
 *         self._message_callback = callback
 *         hb_buffer_set_message_func(self._hb_buffer, msgcallback, <void*>callback, NULL)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  hb_buffer_set_message_func(__pyx_v_self->_hb_buffer, __pyx_f_9uharfbuzz_9_harfbuzz_msgcallback, ((void *)__pyx_v_callback), NULL);

  /* "uharfbuzz/_harfbuzz.pyx":370
 *         hb_buffer_guess_segment_properties(self._hb_buffer)
 * 
 *     def set_message_func(self, callback) -> None:             # <<<<<<<<<<<<<<
 *         self._message_callback = callback
 *         hb_buffer_set_message_func(self._hb_buffer, msgcallback, <void*>callback, NULL)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_27__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_6Buffer_27__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_27__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_27__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_26__reduce_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_26__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_29__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_6Buffer_29__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_29__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_29__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_28__setstate_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6Buffer_28__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Buffer.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":379
 *     cdef object _data
 * 
 *     def __cinit__(self, bytes data = None):             # <<<<<<<<<<<<<<
 *         if data is not None:
 *             self._data = data
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_4Blob_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_4Blob_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_data = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_data,0};
    values[0] = __Pyx_Arg_NewRef_VARARGS(((PyObject*)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_data);
          if (value) { values[0] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 379, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(0, 379, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_data = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 379, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Blob.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_data), (&PyBytes_Type), 1, "data", 1))) __PYX_ERR(0, 379, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Blob___cinit__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *)__pyx_v_self), __pyx_v_data);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_4Blob___cinit__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *__pyx_v_self, PyObject *__pyx_v_data) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  char const *__pyx_t_2;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":380
 * 
 *     def __cinit__(self, bytes data = None):
 *         if data is not None:             # <<<<<<<<<<<<<<
 *             self._data = data
 *             self._hb_blob = hb_blob_create(
 */
  __pyx_t_1 = (__pyx_v_data != ((PyObject*)Py_None));
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":381
 *     def __cinit__(self, bytes data = None):
 *         if data is not None:
 *             self._data = data             # <<<<<<<<<<<<<<
 *             self._hb_blob = hb_blob_create(
 *                 data, len(data), HB_MEMORY_MODE_READONLY, NULL, NULL)
 */
    __Pyx_INCREF(__pyx_v_data);
    __Pyx_GIVEREF(__pyx_v_data);
    __Pyx_GOTREF(__pyx_v_self->_data);
    __Pyx_DECREF(__pyx_v_self->_data);
    __pyx_v_self->_data = __pyx_v_data;

    /* "uharfbuzz/_harfbuzz.pyx":383
 *             self._data = data
 *             self._hb_blob = hb_blob_create(
 *                 data, len(data), HB_MEMORY_MODE_READONLY, NULL, NULL)             # <<<<<<<<<<<<<<
 *         else:
 *             self._data = bytes()
 */
    if (unlikely(__pyx_v_data == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
      __PYX_ERR(0, 383, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_PyBytes_AsString(__pyx_v_data); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 383, __pyx_L1_error)
    if (unlikely(__pyx_v_data == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 383, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyBytes_GET_SIZE(__pyx_v_data); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 383, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":382
 *         if data is not None:
 *             self._data = data
 *             self._hb_blob = hb_blob_create(             # <<<<<<<<<<<<<<
 *                 data, len(data), HB_MEMORY_MODE_READONLY, NULL, NULL)
 *         else:
 */
    __pyx_v_self->_hb_blob = hb_blob_create(__pyx_t_2, __pyx_t_3, HB_MEMORY_MODE_READONLY, NULL, NULL);

    /* "uharfbuzz/_harfbuzz.pyx":380
 * 
 *     def __cinit__(self, bytes data = None):
 *         if data is not None:             # <<<<<<<<<<<<<<
 *             self._data = data
 *             self._hb_blob = hb_blob_create(
 */
    goto __pyx_L3;
  }

  /* "uharfbuzz/_harfbuzz.pyx":385
 *                 data, len(data), HB_MEMORY_MODE_READONLY, NULL, NULL)
 *         else:
 *             self._data = bytes()             # <<<<<<<<<<<<<<
 *             self._hb_blob = hb_blob_get_empty()
 * 
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyObject_CallNoArg(((PyObject *)(&PyBytes_Type))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __Pyx_GOTREF(__pyx_v_self->_data);
    __Pyx_DECREF(__pyx_v_self->_data);
    __pyx_v_self->_data = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "uharfbuzz/_harfbuzz.pyx":386
 *         else:
 *             self._data = bytes()
 *             self._hb_blob = hb_blob_get_empty()             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
    __pyx_v_self->_hb_blob = hb_blob_get_empty();
  }
  __pyx_L3:;

  /* "uharfbuzz/_harfbuzz.pyx":379
 *     cdef object _data
 * 
 *     def __cinit__(self, bytes data = None):             # <<<<<<<<<<<<<<
 *         if data is not None:
 *             self._data = data
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Blob.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":389
 * 
 *     @staticmethod
 *     cdef Blob from_ptr(hb_blob_t* hb_blob):             # <<<<<<<<<<<<<<
 *         """Create Blob from a pointer taking ownership of a it."""
 * 
 */

static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *__pyx_f_9uharfbuzz_9_harfbuzz_4Blob_from_ptr(hb_blob_t *__pyx_v_hb_blob) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *__pyx_v_wrapper = 0;
  unsigned int __pyx_v_blob_length;
  const char *__pyx_v_blob_data;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("from_ptr", 1);

  /* "uharfbuzz/_harfbuzz.pyx":392
 *         """Create Blob from a pointer taking ownership of a it."""
 * 
 *         cdef Blob wrapper = Blob.__new__(Blob)             # <<<<<<<<<<<<<<
 *         wrapper._hb_blob = hb_blob
 *         cdef unsigned int blob_length
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_9uharfbuzz_9_harfbuzz_Blob(((PyTypeObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Blob), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 392, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":393
 * 
 *         cdef Blob wrapper = Blob.__new__(Blob)
 *         wrapper._hb_blob = hb_blob             # <<<<<<<<<<<<<<
 *         cdef unsigned int blob_length
 *         cdef const_char* blob_data = hb_blob_get_data(hb_blob, &blob_length)
 */
  __pyx_v_wrapper->_hb_blob = __pyx_v_hb_blob;

  /* "uharfbuzz/_harfbuzz.pyx":395
 *         wrapper._hb_blob = hb_blob
 *         cdef unsigned int blob_length
 *         cdef const_char* blob_data = hb_blob_get_data(hb_blob, &blob_length)             # <<<<<<<<<<<<<<
 *         wrapper._data = blob_data[:blob_length]
 *         return wrapper
 */
  __pyx_v_blob_data = hb_blob_get_data(__pyx_v_hb_blob, (&__pyx_v_blob_length));

  /* "uharfbuzz/_harfbuzz.pyx":396
 *         cdef unsigned int blob_length
 *         cdef const_char* blob_data = hb_blob_get_data(hb_blob, &blob_length)
 *         wrapper._data = blob_data[:blob_length]             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_blob_data + 0, __pyx_v_blob_length - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_wrapper->_data);
  __Pyx_DECREF(__pyx_v_wrapper->_data);
  __pyx_v_wrapper->_data = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":397
 *         cdef const_char* blob_data = hb_blob_get_data(hb_blob, &blob_length)
 *         wrapper._data = blob_data[:blob_length]
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = __pyx_v_wrapper;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":389
 * 
 *     @staticmethod
 *     cdef Blob from_ptr(hb_blob_t* hb_blob):             # <<<<<<<<<<<<<<
 *         """Create Blob from a pointer taking ownership of a it."""
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Blob.from_ptr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":399
 *         return wrapper
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def from_file_path(cls, filename: Union[str, Path]):
 *         cdef bytes packed = os.fsencode(filename)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Blob_3from_file_path(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_4Blob_3from_file_path = {"from_file_path", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Blob_3from_file_path, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Blob_3from_file_path(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_filename = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("from_file_path (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_filename,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_filename)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 399, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "from_file_path") < 0)) __PYX_ERR(0, 399, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_filename = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("from_file_path", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 399, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Blob.from_file_path", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Blob_2from_file_path(((PyTypeObject*)__pyx_v_cls), __pyx_v_filename);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Blob_2from_file_path(PyTypeObject *__pyx_v_cls, PyObject *__pyx_v_filename) {
  PyObject *__pyx_v_packed = 0;
  hb_blob_t *__pyx_v_blob;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *__pyx_v_inst = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  char *__pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("from_file_path", 1);

  /* "uharfbuzz/_harfbuzz.pyx":401
 *     @classmethod
 *     def from_file_path(cls, filename: Union[str, Path]):
 *         cdef bytes packed = os.fsencode(filename)             # <<<<<<<<<<<<<<
 *         cdef hb_blob_t* blob = hb_blob_create_from_file_or_fail(<char*>packed)
 *         if blob == NULL:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_os); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_fsencode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_filename};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 401, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_1))) __PYX_ERR(0, 401, __pyx_L1_error)
  __pyx_v_packed = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":402
 *     def from_file_path(cls, filename: Union[str, Path]):
 *         cdef bytes packed = os.fsencode(filename)
 *         cdef hb_blob_t* blob = hb_blob_create_from_file_or_fail(<char*>packed)             # <<<<<<<<<<<<<<
 *         if blob == NULL:
 *             raise HarfBuzzError(f"Failed to open: {filename}")
 */
  if (unlikely(__pyx_v_packed == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 402, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyBytes_AsWritableString(__pyx_v_packed); if (unlikely((!__pyx_t_5) && PyErr_Occurred())) __PYX_ERR(0, 402, __pyx_L1_error)
  __pyx_v_blob = hb_blob_create_from_file_or_fail(((char *)__pyx_t_5));

  /* "uharfbuzz/_harfbuzz.pyx":403
 *         cdef bytes packed = os.fsencode(filename)
 *         cdef hb_blob_t* blob = hb_blob_create_from_file_or_fail(<char*>packed)
 *         if blob == NULL:             # <<<<<<<<<<<<<<
 *             raise HarfBuzzError(f"Failed to open: {filename}")
 *         cdef Blob inst = cls(None)
 */
  __pyx_t_6 = (__pyx_v_blob == NULL);
  if (unlikely(__pyx_t_6)) {

    /* "uharfbuzz/_harfbuzz.pyx":404
 *         cdef hb_blob_t* blob = hb_blob_create_from_file_or_fail(<char*>packed)
 *         if blob == NULL:
 *             raise HarfBuzzError(f"Failed to open: {filename}")             # <<<<<<<<<<<<<<
 *         cdef Blob inst = cls(None)
 *         inst._hb_blob = blob
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_HarfBuzzError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 404, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_v_filename, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 404, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = __Pyx_PyUnicode_Concat(__pyx_kp_u_Failed_to_open, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 404, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_7};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 404, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 404, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":403
 *         cdef bytes packed = os.fsencode(filename)
 *         cdef hb_blob_t* blob = hb_blob_create_from_file_or_fail(<char*>packed)
 *         if blob == NULL:             # <<<<<<<<<<<<<<
 *             raise HarfBuzzError(f"Failed to open: {filename}")
 *         cdef Blob inst = cls(None)
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":405
 *         if blob == NULL:
 *             raise HarfBuzzError(f"Failed to open: {filename}")
 *         cdef Blob inst = cls(None)             # <<<<<<<<<<<<<<
 *         inst._hb_blob = blob
 *         return inst
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_v_cls), __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_9uharfbuzz_9_harfbuzz_Blob))))) __PYX_ERR(0, 405, __pyx_L1_error)
  __pyx_v_inst = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":406
 *             raise HarfBuzzError(f"Failed to open: {filename}")
 *         cdef Blob inst = cls(None)
 *         inst._hb_blob = blob             # <<<<<<<<<<<<<<
 *         return inst
 * 
 */
  __pyx_v_inst->_hb_blob = __pyx_v_blob;

  /* "uharfbuzz/_harfbuzz.pyx":407
 *         cdef Blob inst = cls(None)
 *         inst._hb_blob = blob
 *         return inst             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_inst);
  __pyx_r = ((PyObject *)__pyx_v_inst);
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":399
 *         return wrapper
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def from_file_path(cls, filename: Union[str, Path]):
 *         cdef bytes packed = os.fsencode(filename)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Blob.from_file_path", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_packed);
  __Pyx_XDECREF((PyObject *)__pyx_v_inst);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":409
 *         return inst
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         hb_blob_destroy(self._hb_blob)
 *         self._data = None
 */

/* Python wrapper */
static void __pyx_pw_9uharfbuzz_9_harfbuzz_4Blob_5__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_9uharfbuzz_9_harfbuzz_4Blob_5__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_9uharfbuzz_9_harfbuzz_4Blob_4__dealloc__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_9uharfbuzz_9_harfbuzz_4Blob_4__dealloc__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":410
 * 
 *     def __dealloc__(self):
 *         hb_blob_destroy(self._hb_blob)             # <<<<<<<<<<<<<<
 *         self._data = None
 * 
 */
  hb_blob_destroy(__pyx_v_self->_hb_blob);

  /* "uharfbuzz/_harfbuzz.pyx":411
 *     def __dealloc__(self):
 *         hb_blob_destroy(self._hb_blob)
 *         self._data = None             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self) -> int:
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_data);
  __Pyx_DECREF(__pyx_v_self->_data);
  __pyx_v_self->_data = Py_None;

  /* "uharfbuzz/_harfbuzz.pyx":409
 *         return inst
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         hb_blob_destroy(self._hb_blob)
 *         self._data = None
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "uharfbuzz/_harfbuzz.pyx":413
 *         self._data = None
 * 
 *     def __len__(self) -> int:             # <<<<<<<<<<<<<<
 *         return len(self._data)
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_9uharfbuzz_9_harfbuzz_4Blob_7__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_9uharfbuzz_9_harfbuzz_4Blob_7__len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Blob_6__len__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_9uharfbuzz_9_harfbuzz_4Blob_6__len__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__len__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":414
 * 
 *     def __len__(self) -> int:
 *         return len(self._data)             # <<<<<<<<<<<<<<
 * 
 *     def __bool__(self) -> bool:
 */
  __pyx_t_1 = __pyx_v_self->_data;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 414, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":413
 *         self._data = None
 * 
 *     def __len__(self) -> int:             # <<<<<<<<<<<<<<
 *         return len(self._data)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Blob.__len__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":416
 *         return len(self._data)
 * 
 *     def __bool__(self) -> bool:             # <<<<<<<<<<<<<<
 *         return bool(self._data)
 * 
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_4Blob_9__bool__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_4Blob_9__bool__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__bool__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Blob_8__bool__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_4Blob_8__bool__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *__pyx_v_self) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "uharfbuzz/_harfbuzz.pyx":417
 * 
 *     def __bool__(self) -> bool:
 *         return bool(self._data)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_self->_data); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 417, __pyx_L1_error)
  __pyx_r = (!(!__pyx_t_1));
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":416
 *         return len(self._data)
 * 
 *     def __bool__(self) -> bool:             # <<<<<<<<<<<<<<
 *         return bool(self._data)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Blob.__bool__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":419
 *         return bool(self._data)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def data(self) -> bytes:
 *         return self._data
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Blob_4data_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Blob_4data_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Blob_4data___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Blob_4data___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":421
 *     @property
 *     def data(self) -> bytes:
 *         return self._data             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_data);
  __pyx_r = __pyx_v_self->_data;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":419
 *         return bool(self._data)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def data(self) -> bytes:
 *         return self._data
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Blob_11__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_4Blob_11__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Blob_11__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Blob_11__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Blob_10__reduce_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Blob_10__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Blob.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Blob_13__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_4Blob_13__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Blob_13__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Blob_13__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Blob.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Blob_12__setstate_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Blob_12__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Blob.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":427
 * 
 * 
 * cdef hb_blob_t* _reference_table_func(             # <<<<<<<<<<<<<<
 *         hb_face_t* face, hb_tag_t tag, void* user_data) noexcept:
 *     cdef Face py_face = <object>(hb_face_get_user_data(face, &k))
 */

static hb_blob_t *__pyx_f_9uharfbuzz_9_harfbuzz__reference_table_func(hb_face_t *__pyx_v_face, hb_tag_t __pyx_v_tag, void *__pyx_v_user_data) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_py_face = 0;
  char __pyx_v_cstr[5];
  PyObject *__pyx_v_packed = 0;
  PyObject *__pyx_v_table = 0;
  hb_blob_t *__pyx_r;
  __Pyx_RefNannyDeclarations
  void *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  char const *__pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_reference_table_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":429
 * cdef hb_blob_t* _reference_table_func(
 *         hb_face_t* face, hb_tag_t tag, void* user_data) noexcept:
 *     cdef Face py_face = <object>(hb_face_get_user_data(face, &k))             # <<<<<<<<<<<<<<
 *     #
 *     cdef char cstr[5]
 */
  __pyx_t_1 = hb_face_get_user_data(__pyx_v_face, (&__pyx_v_9uharfbuzz_9_harfbuzz_k));
  if (!(likely(((((PyObject *)__pyx_t_1)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_t_1), __pyx_ptype_9uharfbuzz_9_harfbuzz_Face))))) __PYX_ERR(0, 429, __pyx_L1_error)
  __pyx_t_2 = ((PyObject *)__pyx_t_1);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_py_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":432
 *     #
 *     cdef char cstr[5]
 *     hb_tag_to_string(tag, cstr)             # <<<<<<<<<<<<<<
 *     cstr[4] = b'\0'
 *     cdef bytes packed = cstr
 */
  hb_tag_to_string(__pyx_v_tag, __pyx_v_cstr);

  /* "uharfbuzz/_harfbuzz.pyx":433
 *     cdef char cstr[5]
 *     hb_tag_to_string(tag, cstr)
 *     cstr[4] = b'\0'             # <<<<<<<<<<<<<<
 *     cdef bytes packed = cstr
 *     #
 */
  (__pyx_v_cstr[4]) = '\x00';

  /* "uharfbuzz/_harfbuzz.pyx":434
 *     hb_tag_to_string(tag, cstr)
 *     cstr[4] = b'\0'
 *     cdef bytes packed = cstr             # <<<<<<<<<<<<<<
 *     #
 *     cdef bytes table = py_face._reference_table_func(
 */
  __pyx_t_2 = __Pyx_PyObject_FromString(__pyx_v_cstr); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_packed = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":437
 *     #
 *     cdef bytes table = py_face._reference_table_func(
 *         py_face, packed.decode(), <object>user_data)             # <<<<<<<<<<<<<<
 *     if table is None:
 *         return NULL
 */
  __pyx_t_3 = __Pyx_decode_bytes(__pyx_v_packed, 0, PY_SSIZE_T_MAX, NULL, NULL, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 437, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_py_face->_reference_table_func);
  __pyx_t_4 = __pyx_v_py_face->_reference_table_func; __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_5, ((PyObject *)__pyx_v_py_face), __pyx_t_3, ((PyObject *)__pyx_v_user_data)};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 436, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }

  /* "uharfbuzz/_harfbuzz.pyx":436
 *     cdef bytes packed = cstr
 *     #
 *     cdef bytes table = py_face._reference_table_func(             # <<<<<<<<<<<<<<
 *         py_face, packed.decode(), <object>user_data)
 *     if table is None:
 */
  if (!(likely(PyBytes_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_2))) __PYX_ERR(0, 436, __pyx_L1_error)
  __pyx_v_table = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":438
 *     cdef bytes table = py_face._reference_table_func(
 *         py_face, packed.decode(), <object>user_data)
 *     if table is None:             # <<<<<<<<<<<<<<
 *         return NULL
 *     return hb_blob_create(
 */
  __pyx_t_7 = (__pyx_v_table == ((PyObject*)Py_None));
  if (__pyx_t_7) {

    /* "uharfbuzz/_harfbuzz.pyx":439
 *         py_face, packed.decode(), <object>user_data)
 *     if table is None:
 *         return NULL             # <<<<<<<<<<<<<<
 *     return hb_blob_create(
 *         table, len(table), HB_MEMORY_MODE_READONLY, NULL, NULL)
 */
    __pyx_r = NULL;
    goto __pyx_L0;

    /* "uharfbuzz/_harfbuzz.pyx":438
 *     cdef bytes table = py_face._reference_table_func(
 *         py_face, packed.decode(), <object>user_data)
 *     if table is None:             # <<<<<<<<<<<<<<
 *         return NULL
 *     return hb_blob_create(
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":441
 *         return NULL
 *     return hb_blob_create(
 *         table, len(table), HB_MEMORY_MODE_READONLY, NULL, NULL)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (unlikely(__pyx_v_table == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 441, __pyx_L1_error)
  }
  __pyx_t_8 = __Pyx_PyBytes_AsString(__pyx_v_table); if (unlikely((!__pyx_t_8) && PyErr_Occurred())) __PYX_ERR(0, 441, __pyx_L1_error)
  if (unlikely(__pyx_v_table == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 441, __pyx_L1_error)
  }
  __pyx_t_9 = __Pyx_PyBytes_GET_SIZE(__pyx_v_table); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 441, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":440
 *     if table is None:
 *         return NULL
 *     return hb_blob_create(             # <<<<<<<<<<<<<<
 *         table, len(table), HB_MEMORY_MODE_READONLY, NULL, NULL)
 * 
 */
  __pyx_r = hb_blob_create(__pyx_t_8, __pyx_t_9, HB_MEMORY_MODE_READONLY, NULL, NULL);
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":427
 * 
 * 
 * cdef hb_blob_t* _reference_table_func(             # <<<<<<<<<<<<<<
 *         hb_face_t* face, hb_tag_t tag, void* user_data) noexcept:
 *     cdef Face py_face = <object>(hb_face_get_user_data(face, &k))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz._reference_table_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_py_face);
  __Pyx_XDECREF(__pyx_v_packed);
  __Pyx_XDECREF(__pyx_v_table);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":449
 *     cdef Blob _blob
 * 
 *     def __cinit__(self, blob: Union[Blob, bytes] = None, int index=0):             # <<<<<<<<<<<<<<
 *         if blob is not None:
 *             if not isinstance(blob, Blob):
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_4Face_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_4Face_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_blob = 0;
  int __pyx_v_index;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_blob,&__pyx_n_s_index,0};
    values[0] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_blob);
          if (value) { values[0] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 449, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_index);
          if (value) { values[1] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 449, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(0, 449, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_blob = values[0];
    if (values[1]) {
      __pyx_v_index = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_index == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 449, __pyx_L3_error)
    } else {
      __pyx_v_index = ((int)0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 0, 2, __pyx_nargs); __PYX_ERR(0, 449, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Face.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Face___cinit__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)__pyx_v_self), __pyx_v_blob, __pyx_v_index);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_4Face___cinit__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_self, PyObject *__pyx_v_blob, int __pyx_v_index) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":450
 * 
 *     def __cinit__(self, blob: Union[Blob, bytes] = None, int index=0):
 *         if blob is not None:             # <<<<<<<<<<<<<<
 *             if not isinstance(blob, Blob):
 *                 self._blob = Blob(blob)
 */
  __pyx_t_1 = (__pyx_v_blob != Py_None);
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":451
 *     def __cinit__(self, blob: Union[Blob, bytes] = None, int index=0):
 *         if blob is not None:
 *             if not isinstance(blob, Blob):             # <<<<<<<<<<<<<<
 *                 self._blob = Blob(blob)
 *             else:
 */
    __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_blob, __pyx_ptype_9uharfbuzz_9_harfbuzz_Blob); 
    __pyx_t_2 = (!__pyx_t_1);
    if (__pyx_t_2) {

      /* "uharfbuzz/_harfbuzz.pyx":452
 *         if blob is not None:
 *             if not isinstance(blob, Blob):
 *                 self._blob = Blob(blob)             # <<<<<<<<<<<<<<
 *             else:
 *                 self._blob = blob
 */
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Blob), __pyx_v_blob); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 452, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_3);
      __Pyx_GOTREF((PyObject *)__pyx_v_self->_blob);
      __Pyx_DECREF((PyObject *)__pyx_v_self->_blob);
      __pyx_v_self->_blob = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *)__pyx_t_3);
      __pyx_t_3 = 0;

      /* "uharfbuzz/_harfbuzz.pyx":451
 *     def __cinit__(self, blob: Union[Blob, bytes] = None, int index=0):
 *         if blob is not None:
 *             if not isinstance(blob, Blob):             # <<<<<<<<<<<<<<
 *                 self._blob = Blob(blob)
 *             else:
 */
      goto __pyx_L4;
    }

    /* "uharfbuzz/_harfbuzz.pyx":454
 *                 self._blob = Blob(blob)
 *             else:
 *                 self._blob = blob             # <<<<<<<<<<<<<<
 *             self._hb_face = hb_face_create(self._blob._hb_blob, index)
 *         else:
 */
    /*else*/ {
      if (!(likely(((__pyx_v_blob) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_blob, __pyx_ptype_9uharfbuzz_9_harfbuzz_Blob))))) __PYX_ERR(0, 454, __pyx_L1_error)
      __pyx_t_3 = __pyx_v_blob;
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_3);
      __Pyx_GOTREF((PyObject *)__pyx_v_self->_blob);
      __Pyx_DECREF((PyObject *)__pyx_v_self->_blob);
      __pyx_v_self->_blob = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *)__pyx_t_3);
      __pyx_t_3 = 0;
    }
    __pyx_L4:;

    /* "uharfbuzz/_harfbuzz.pyx":455
 *             else:
 *                 self._blob = blob
 *             self._hb_face = hb_face_create(self._blob._hb_blob, index)             # <<<<<<<<<<<<<<
 *         else:
 *             self._hb_face = hb_face_get_empty()
 */
    __pyx_v_self->_hb_face = hb_face_create(__pyx_v_self->_blob->_hb_blob, __pyx_v_index);

    /* "uharfbuzz/_harfbuzz.pyx":450
 * 
 *     def __cinit__(self, blob: Union[Blob, bytes] = None, int index=0):
 *         if blob is not None:             # <<<<<<<<<<<<<<
 *             if not isinstance(blob, Blob):
 *                 self._blob = Blob(blob)
 */
    goto __pyx_L3;
  }

  /* "uharfbuzz/_harfbuzz.pyx":457
 *             self._hb_face = hb_face_create(self._blob._hb_blob, index)
 *         else:
 *             self._hb_face = hb_face_get_empty()             # <<<<<<<<<<<<<<
 *             self._blob = None
 * 
 */
  /*else*/ {
    __pyx_v_self->_hb_face = hb_face_get_empty();

    /* "uharfbuzz/_harfbuzz.pyx":458
 *         else:
 *             self._hb_face = hb_face_get_empty()
 *             self._blob = None             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF((PyObject *)__pyx_v_self->_blob);
    __Pyx_DECREF((PyObject *)__pyx_v_self->_blob);
    __pyx_v_self->_blob = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *)Py_None);
  }
  __pyx_L3:;

  /* "uharfbuzz/_harfbuzz.pyx":449
 *     cdef Blob _blob
 * 
 *     def __cinit__(self, blob: Union[Blob, bytes] = None, int index=0):             # <<<<<<<<<<<<<<
 *         if blob is not None:
 *             if not isinstance(blob, Blob):
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Face.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":460
 *             self._blob = None
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         hb_face_destroy(self._hb_face)
 *         self._blob = None
 */

/* Python wrapper */
static void __pyx_pw_9uharfbuzz_9_harfbuzz_4Face_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_9uharfbuzz_9_harfbuzz_4Face_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_9uharfbuzz_9_harfbuzz_4Face_2__dealloc__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_9uharfbuzz_9_harfbuzz_4Face_2__dealloc__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":461
 * 
 *     def __dealloc__(self):
 *         hb_face_destroy(self._hb_face)             # <<<<<<<<<<<<<<
 *         self._blob = None
 * 
 */
  hb_face_destroy(__pyx_v_self->_hb_face);

  /* "uharfbuzz/_harfbuzz.pyx":462
 *     def __dealloc__(self):
 *         hb_face_destroy(self._hb_face)
 *         self._blob = None             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_blob);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_blob);
  __pyx_v_self->_blob = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *)Py_None);

  /* "uharfbuzz/_harfbuzz.pyx":460
 *             self._blob = None
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         hb_face_destroy(self._hb_face)
 *         self._blob = None
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "uharfbuzz/_harfbuzz.pyx":465
 * 
 *     @staticmethod
 *     cdef Face from_ptr(hb_face_t* hb_face):             # <<<<<<<<<<<<<<
 *         """Create Face from a pointer taking ownership of a it."""
 * 
 */

static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_f_9uharfbuzz_9_harfbuzz_4Face_from_ptr(hb_face_t *__pyx_v_hb_face) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_wrapper = 0;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("from_ptr", 1);

  /* "uharfbuzz/_harfbuzz.pyx":468
 *         """Create Face from a pointer taking ownership of a it."""
 * 
 *         cdef Face wrapper = Face.__new__(Face)             # <<<<<<<<<<<<<<
 *         wrapper._hb_face = hb_face
 *         return wrapper
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_9uharfbuzz_9_harfbuzz_Face(((PyTypeObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Face), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 468, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":469
 * 
 *         cdef Face wrapper = Face.__new__(Face)
 *         wrapper._hb_face = hb_face             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __pyx_v_wrapper->_hb_face = __pyx_v_hb_face;

  /* "uharfbuzz/_harfbuzz.pyx":470
 *         cdef Face wrapper = Face.__new__(Face)
 *         wrapper._hb_face = hb_face
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     # DEPRECATED: use the normal constructor
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = __pyx_v_wrapper;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":465
 * 
 *     @staticmethod
 *     cdef Face from_ptr(hb_face_t* hb_face):             # <<<<<<<<<<<<<<
 *         """Create Face from a pointer taking ownership of a it."""
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Face.from_ptr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":473
 * 
 *     # DEPRECATED: use the normal constructor
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def create(cls, bytes blob, int index=0):
 *         cdef Face inst = cls(blob, index)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_5create(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_4Face_5create = {"create", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_5create, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_5create(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_blob = 0;
  int __pyx_v_index;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_blob,&__pyx_n_s_index,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_blob)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 473, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_index);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 473, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "create") < 0)) __PYX_ERR(0, 473, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_blob = ((PyObject*)values[0]);
    if (values[1]) {
      __pyx_v_index = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_index == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 474, __pyx_L3_error)
    } else {
      __pyx_v_index = ((int)0);
    }
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 473, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Face.create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_blob), (&PyBytes_Type), 1, "blob", 1))) __PYX_ERR(0, 474, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Face_4create(((PyTypeObject*)__pyx_v_cls), __pyx_v_blob, __pyx_v_index);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Face_4create(PyTypeObject *__pyx_v_cls, PyObject *__pyx_v_blob, int __pyx_v_index) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_inst = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("create", 1);

  /* "uharfbuzz/_harfbuzz.pyx":475
 *     @classmethod
 *     def create(cls, bytes blob, int index=0):
 *         cdef Face inst = cls(blob, index)             # <<<<<<<<<<<<<<
 *         return inst
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_blob);
  __Pyx_GIVEREF(__pyx_v_blob);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_blob)) __PYX_ERR(0, 475, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1)) __PYX_ERR(0, 475, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_v_cls), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_9uharfbuzz_9_harfbuzz_Face))))) __PYX_ERR(0, 475, __pyx_L1_error)
  __pyx_v_inst = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":476
 *     def create(cls, bytes blob, int index=0):
 *         cdef Face inst = cls(blob, index)
 *         return inst             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_inst);
  __pyx_r = ((PyObject *)__pyx_v_inst);
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":473
 * 
 *     # DEPRECATED: use the normal constructor
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def create(cls, bytes blob, int index=0):
 *         cdef Face inst = cls(blob, index)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Face.create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_inst);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":478
 *         return inst
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def create_for_tables(cls,
 *                           func: Callable[[
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_7create_for_tables(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_4Face_7create_for_tables = {"create_for_tables", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_7create_for_tables, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_7create_for_tables(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func = 0;
  PyObject *__pyx_v_user_data = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_for_tables (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func,&__pyx_n_s_user_data,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_func)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 478, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_user_data)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 478, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("create_for_tables", 1, 2, 2, 1); __PYX_ERR(0, 478, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "create_for_tables") < 0)) __PYX_ERR(0, 478, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_func = values[0];
    __pyx_v_user_data = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create_for_tables", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 478, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Face.create_for_tables", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Face_6create_for_tables(((PyTypeObject*)__pyx_v_cls), __pyx_v_func, __pyx_v_user_data);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Face_6create_for_tables(PyTypeObject *__pyx_v_cls, PyObject *__pyx_v_func, PyObject *__pyx_v_user_data) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_inst = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("create_for_tables", 1);

  /* "uharfbuzz/_harfbuzz.pyx":486
 *                           ], bytes],
 *                           user_data: object):
 *         cdef Face inst = cls(None)             # <<<<<<<<<<<<<<
 *         inst._hb_face = hb_face_create_for_tables(
 *             _reference_table_func, <void*>user_data, NULL)
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_v_cls), __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_9uharfbuzz_9_harfbuzz_Face))))) __PYX_ERR(0, 486, __pyx_L1_error)
  __pyx_v_inst = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":487
 *                           user_data: object):
 *         cdef Face inst = cls(None)
 *         inst._hb_face = hb_face_create_for_tables(             # <<<<<<<<<<<<<<
 *             _reference_table_func, <void*>user_data, NULL)
 *         hb_face_set_user_data(inst._hb_face, &k, <void*>inst, NULL, 0)
 */
  __pyx_v_inst->_hb_face = hb_face_create_for_tables(__pyx_f_9uharfbuzz_9_harfbuzz__reference_table_func, ((void *)__pyx_v_user_data), NULL);

  /* "uharfbuzz/_harfbuzz.pyx":489
 *         inst._hb_face = hb_face_create_for_tables(
 *             _reference_table_func, <void*>user_data, NULL)
 *         hb_face_set_user_data(inst._hb_face, &k, <void*>inst, NULL, 0)             # <<<<<<<<<<<<<<
 *         inst._reference_table_func = func
 *         return inst
 */
  (void)(hb_face_set_user_data(__pyx_v_inst->_hb_face, (&__pyx_v_9uharfbuzz_9_harfbuzz_k), ((void *)__pyx_v_inst), NULL, 0));

  /* "uharfbuzz/_harfbuzz.pyx":490
 *             _reference_table_func, <void*>user_data, NULL)
 *         hb_face_set_user_data(inst._hb_face, &k, <void*>inst, NULL, 0)
 *         inst._reference_table_func = func             # <<<<<<<<<<<<<<
 *         return inst
 * 
 */
  __Pyx_INCREF(__pyx_v_func);
  __Pyx_GIVEREF(__pyx_v_func);
  __Pyx_GOTREF(__pyx_v_inst->_reference_table_func);
  __Pyx_DECREF(__pyx_v_inst->_reference_table_func);
  __pyx_v_inst->_reference_table_func = __pyx_v_func;

  /* "uharfbuzz/_harfbuzz.pyx":491
 *         hb_face_set_user_data(inst._hb_face, &k, <void*>inst, NULL, 0)
 *         inst._reference_table_func = func
 *         return inst             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_inst);
  __pyx_r = ((PyObject *)__pyx_v_inst);
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":478
 *         return inst
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def create_for_tables(cls,
 *                           func: Callable[[
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Face.create_for_tables", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_inst);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":493
 *         return inst
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def index(self) -> int:
 *         return hb_face_get_index(self._hb_face)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_5index_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_5index_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Face_5index___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Face_5index___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":495
 *     @property
 *     def index(self) -> int:
 *         return hb_face_get_index(self._hb_face)             # <<<<<<<<<<<<<<
 * 
 *     @index.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(hb_face_get_index(__pyx_v_self->_hb_face)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":493
 *         return inst
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def index(self) -> int:
 *         return hb_face_get_index(self._hb_face)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Face.index.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":497
 *         return hb_face_get_index(self._hb_face)
 * 
 *     @index.setter             # <<<<<<<<<<<<<<
 *     def index(self, value: int):
 *         hb_face_set_index(self._hb_face, value)
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_4Face_5index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_4Face_5index_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyInt_Type), 0, "value", 1))) __PYX_ERR(0, 498, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Face_5index_2__set__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)__pyx_v_self), ((PyObject*)__pyx_v_value));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_4Face_5index_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "uharfbuzz/_harfbuzz.pyx":499
 *     @index.setter
 *     def index(self, value: int):
 *         hb_face_set_index(self._hb_face, value)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_t_1 = __Pyx_PyInt_As_unsigned_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 499, __pyx_L1_error)
  hb_face_set_index(__pyx_v_self->_hb_face, __pyx_t_1);

  /* "uharfbuzz/_harfbuzz.pyx":497
 *         return hb_face_get_index(self._hb_face)
 * 
 *     @index.setter             # <<<<<<<<<<<<<<
 *     def index(self, value: int):
 *         hb_face_set_index(self._hb_face, value)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Face.index.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":501
 *         hb_face_set_index(self._hb_face, value)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def upem(self) -> int:
 *         return hb_face_get_upem(self._hb_face)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_4upem_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_4upem_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Face_4upem___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Face_4upem___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":503
 *     @property
 *     def upem(self) -> int:
 *         return hb_face_get_upem(self._hb_face)             # <<<<<<<<<<<<<<
 * 
 *     @upem.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(hb_face_get_upem(__pyx_v_self->_hb_face)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":501
 *         hb_face_set_index(self._hb_face, value)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def upem(self) -> int:
 *         return hb_face_get_upem(self._hb_face)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Face.upem.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":505
 *         return hb_face_get_upem(self._hb_face)
 * 
 *     @upem.setter             # <<<<<<<<<<<<<<
 *     def upem(self, value: int):
 *         hb_face_set_upem(self._hb_face, value)
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_4Face_4upem_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_4Face_4upem_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyInt_Type), 0, "value", 1))) __PYX_ERR(0, 506, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Face_4upem_2__set__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)__pyx_v_self), ((PyObject*)__pyx_v_value));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_4Face_4upem_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "uharfbuzz/_harfbuzz.pyx":507
 *     @upem.setter
 *     def upem(self, value: int):
 *         hb_face_set_upem(self._hb_face, value)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_t_1 = __Pyx_PyInt_As_unsigned_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 507, __pyx_L1_error)
  hb_face_set_upem(__pyx_v_self->_hb_face, __pyx_t_1);

  /* "uharfbuzz/_harfbuzz.pyx":505
 *         return hb_face_get_upem(self._hb_face)
 * 
 *     @upem.setter             # <<<<<<<<<<<<<<
 *     def upem(self, value: int):
 *         hb_face_set_upem(self._hb_face, value)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Face.upem.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":509
 *         hb_face_set_upem(self._hb_face, value)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def glyph_count(self) -> int:
 *         return hb_face_get_glyph_count(self._hb_face)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_11glyph_count_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_11glyph_count_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Face_11glyph_count___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Face_11glyph_count___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":511
 *     @property
 *     def glyph_count(self) -> int:
 *         return hb_face_get_glyph_count(self._hb_face)             # <<<<<<<<<<<<<<
 * 
 *     @glyph_count.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(hb_face_get_glyph_count(__pyx_v_self->_hb_face)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":509
 *         hb_face_set_upem(self._hb_face, value)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def glyph_count(self) -> int:
 *         return hb_face_get_glyph_count(self._hb_face)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Face.glyph_count.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":513
 *         return hb_face_get_glyph_count(self._hb_face)
 * 
 *     @glyph_count.setter             # <<<<<<<<<<<<<<
 *     def glyph_count(self, value: int):
 *         hb_face_set_glyph_count(self._hb_face, value)
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_4Face_11glyph_count_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_4Face_11glyph_count_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyInt_Type), 0, "value", 1))) __PYX_ERR(0, 514, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Face_11glyph_count_2__set__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)__pyx_v_self), ((PyObject*)__pyx_v_value));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_4Face_11glyph_count_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "uharfbuzz/_harfbuzz.pyx":515
 *     @glyph_count.setter
 *     def glyph_count(self, value: int):
 *         hb_face_set_glyph_count(self._hb_face, value)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_t_1 = __Pyx_PyInt_As_unsigned_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 515, __pyx_L1_error)
  hb_face_set_glyph_count(__pyx_v_self->_hb_face, __pyx_t_1);

  /* "uharfbuzz/_harfbuzz.pyx":513
 *         return hb_face_get_glyph_count(self._hb_face)
 * 
 *     @glyph_count.setter             # <<<<<<<<<<<<<<
 *     def glyph_count(self, value: int):
 *         hb_face_set_glyph_count(self._hb_face, value)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Face.glyph_count.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":517
 *         hb_face_set_glyph_count(self._hb_face, value)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def blob(self) -> Blob:
 *         cdef hb_blob_t* blob = hb_face_reference_blob(self._hb_face)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_4blob_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_4blob_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Face_4blob___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Face_4blob___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_self) {
  hb_blob_t *__pyx_v_blob;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":519
 *     @property
 *     def blob(self) -> Blob:
 *         cdef hb_blob_t* blob = hb_face_reference_blob(self._hb_face)             # <<<<<<<<<<<<<<
 *         if blob is NULL:
 *             raise MemoryError()
 */
  __pyx_v_blob = hb_face_reference_blob(__pyx_v_self->_hb_face);

  /* "uharfbuzz/_harfbuzz.pyx":520
 *     def blob(self) -> Blob:
 *         cdef hb_blob_t* blob = hb_face_reference_blob(self._hb_face)
 *         if blob is NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         return Blob.from_ptr(blob)
 */
  __pyx_t_1 = (__pyx_v_blob == NULL);
  if (unlikely(__pyx_t_1)) {

    /* "uharfbuzz/_harfbuzz.pyx":521
 *         cdef hb_blob_t* blob = hb_face_reference_blob(self._hb_face)
 *         if blob is NULL:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 *         return Blob.from_ptr(blob)
 * 
 */
    PyErr_NoMemory(); __PYX_ERR(0, 521, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":520
 *     def blob(self) -> Blob:
 *         cdef hb_blob_t* blob = hb_face_reference_blob(self._hb_face)
 *         if blob is NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         return Blob.from_ptr(blob)
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":522
 *         if blob is NULL:
 *             raise MemoryError()
 *         return Blob.from_ptr(blob)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = ((PyObject *)__pyx_f_9uharfbuzz_9_harfbuzz_4Blob_from_ptr(__pyx_v_blob)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 522, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":517
 *         hb_face_set_glyph_count(self._hb_face, value)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def blob(self) -> Blob:
 *         cdef hb_blob_t* blob = hb_face_reference_blob(self._hb_face)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Face.blob.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":524
 *         return Blob.from_ptr(blob)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def table_tags(self) -> List[str]:
 *         cdef unsigned int tag_count = STATIC_ARRAY_SIZE
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_10table_tags_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_10table_tags_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Face_10table_tags___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Face_10table_tags___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_self) {
  unsigned int __pyx_v_tag_count;
  hb_tag_t __pyx_v_tags_array[0x80];
  PyObject *__pyx_v_tags = 0;
  char __pyx_v_cstr[5];
  unsigned int __pyx_v_i;
  unsigned int __pyx_v_start_offset;
  char *__pyx_v_packed;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  unsigned int __pyx_t_3;
  unsigned int __pyx_t_4;
  unsigned int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":526
 *     @property
 *     def table_tags(self) -> List[str]:
 *         cdef unsigned int tag_count = STATIC_ARRAY_SIZE             # <<<<<<<<<<<<<<
 *         cdef hb_tag_t tags_array[STATIC_ARRAY_SIZE]
 *         cdef list tags = []
 */
  __pyx_v_tag_count = 0x80;

  /* "uharfbuzz/_harfbuzz.pyx":528
 *         cdef unsigned int tag_count = STATIC_ARRAY_SIZE
 *         cdef hb_tag_t tags_array[STATIC_ARRAY_SIZE]
 *         cdef list tags = []             # <<<<<<<<<<<<<<
 *         cdef char cstr[5]
 *         cdef unsigned int i
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_tags = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":531
 *         cdef char cstr[5]
 *         cdef unsigned int i
 *         cdef unsigned int start_offset = 0             # <<<<<<<<<<<<<<
 *         while tag_count == STATIC_ARRAY_SIZE:
 *             hb_face_get_table_tags(
 */
  __pyx_v_start_offset = 0;

  /* "uharfbuzz/_harfbuzz.pyx":532
 *         cdef unsigned int i
 *         cdef unsigned int start_offset = 0
 *         while tag_count == STATIC_ARRAY_SIZE:             # <<<<<<<<<<<<<<
 *             hb_face_get_table_tags(
 *                 self._hb_face, start_offset, &tag_count, tags_array)
 */
  while (1) {
    __pyx_t_2 = (__pyx_v_tag_count == 0x80);
    if (!__pyx_t_2) break;

    /* "uharfbuzz/_harfbuzz.pyx":533
 *         cdef unsigned int start_offset = 0
 *         while tag_count == STATIC_ARRAY_SIZE:
 *             hb_face_get_table_tags(             # <<<<<<<<<<<<<<
 *                 self._hb_face, start_offset, &tag_count, tags_array)
 *             for i in range(tag_count):
 */
    (void)(hb_face_get_table_tags(__pyx_v_self->_hb_face, __pyx_v_start_offset, (&__pyx_v_tag_count), __pyx_v_tags_array));

    /* "uharfbuzz/_harfbuzz.pyx":535
 *             hb_face_get_table_tags(
 *                 self._hb_face, start_offset, &tag_count, tags_array)
 *             for i in range(tag_count):             # <<<<<<<<<<<<<<
 *                 hb_tag_to_string(tags_array[i], cstr)
 *                 cstr[4] = b'\0'
 */
    __pyx_t_3 = __pyx_v_tag_count;
    __pyx_t_4 = __pyx_t_3;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "uharfbuzz/_harfbuzz.pyx":536
 *                 self._hb_face, start_offset, &tag_count, tags_array)
 *             for i in range(tag_count):
 *                 hb_tag_to_string(tags_array[i], cstr)             # <<<<<<<<<<<<<<
 *                 cstr[4] = b'\0'
 *                 packed = cstr
 */
      hb_tag_to_string((__pyx_v_tags_array[__pyx_v_i]), __pyx_v_cstr);

      /* "uharfbuzz/_harfbuzz.pyx":537
 *             for i in range(tag_count):
 *                 hb_tag_to_string(tags_array[i], cstr)
 *                 cstr[4] = b'\0'             # <<<<<<<<<<<<<<
 *                 packed = cstr
 *                 tags.append(packed.decode())
 */
      (__pyx_v_cstr[4]) = '\x00';

      /* "uharfbuzz/_harfbuzz.pyx":538
 *                 hb_tag_to_string(tags_array[i], cstr)
 *                 cstr[4] = b'\0'
 *                 packed = cstr             # <<<<<<<<<<<<<<
 *                 tags.append(packed.decode())
 *             start_offset += tag_count
 */
      __pyx_v_packed = __pyx_v_cstr;

      /* "uharfbuzz/_harfbuzz.pyx":539
 *                 cstr[4] = b'\0'
 *                 packed = cstr
 *                 tags.append(packed.decode())             # <<<<<<<<<<<<<<
 *             start_offset += tag_count
 *         return tags
 */
      __pyx_t_6 = __Pyx_ssize_strlen(__pyx_v_packed); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 539, __pyx_L1_error)
      __pyx_t_1 = __Pyx_decode_c_string(__pyx_v_packed, 0, __pyx_t_6, NULL, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 539, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_tags, __pyx_t_1); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 539, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }

    /* "uharfbuzz/_harfbuzz.pyx":540
 *                 packed = cstr
 *                 tags.append(packed.decode())
 *             start_offset += tag_count             # <<<<<<<<<<<<<<
 *         return tags
 * 
 */
    __pyx_v_start_offset = (__pyx_v_start_offset + __pyx_v_tag_count);
  }

  /* "uharfbuzz/_harfbuzz.pyx":541
 *                 tags.append(packed.decode())
 *             start_offset += tag_count
 *         return tags             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_tags);
  __pyx_r = __pyx_v_tags;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":524
 *         return Blob.from_ptr(blob)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def table_tags(self) -> List[str]:
 *         cdef unsigned int tag_count = STATIC_ARRAY_SIZE
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Face.table_tags.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tags);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":543
 *         return tags
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def unicodes (self):
 *         s = Set()
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_8unicodes_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_8unicodes_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Face_8unicodes___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Face_8unicodes___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_self) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_s = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":545
 *     @property
 *     def unicodes (self):
 *         s = Set()             # <<<<<<<<<<<<<<
 *         hb_face_collect_unicodes(self._hb_face, s._hb_set)
 *         return s
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Set)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_s = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":546
 *     def unicodes (self):
 *         s = Set()
 *         hb_face_collect_unicodes(self._hb_face, s._hb_set)             # <<<<<<<<<<<<<<
 *         return s
 * 
 */
  hb_face_collect_unicodes(__pyx_v_self->_hb_face, __pyx_v_s->_hb_set);

  /* "uharfbuzz/_harfbuzz.pyx":547
 *         s = Set()
 *         hb_face_collect_unicodes(self._hb_face, s._hb_set)
 *         return s             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_s);
  __pyx_r = ((PyObject *)__pyx_v_s);
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":543
 *         return tags
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def unicodes (self):
 *         s = Set()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Face.unicodes.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_s);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":549
 *         return s
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def variation_selectors(self):
 *         s = Set()
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_19variation_selectors_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_19variation_selectors_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Face_19variation_selectors___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Face_19variation_selectors___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_self) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_s = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":551
 *     @property
 *     def variation_selectors(self):
 *         s = Set()             # <<<<<<<<<<<<<<
 *         hb_face_collect_variation_selectors(self._hb_face, s._hb_set)
 *         return s
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Set)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_s = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":552
 *     def variation_selectors(self):
 *         s = Set()
 *         hb_face_collect_variation_selectors(self._hb_face, s._hb_set)             # <<<<<<<<<<<<<<
 *         return s
 * 
 */
  hb_face_collect_variation_selectors(__pyx_v_self->_hb_face, __pyx_v_s->_hb_set);

  /* "uharfbuzz/_harfbuzz.pyx":553
 *         s = Set()
 *         hb_face_collect_variation_selectors(self._hb_face, s._hb_set)
 *         return s             # <<<<<<<<<<<<<<
 * 
 *     def variation_unicodes(self, variation_selector):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_s);
  __pyx_r = ((PyObject *)__pyx_v_s);
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":549
 *         return s
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def variation_selectors(self):
 *         s = Set()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Face.variation_selectors.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_s);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":555
 *         return s
 * 
 *     def variation_unicodes(self, variation_selector):             # <<<<<<<<<<<<<<
 *         s = Set()
 *         hb_face_collect_variation_unicodes(self._hb_face, variation_selector, s._hb_set)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_9variation_unicodes(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_4Face_9variation_unicodes = {"variation_unicodes", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_9variation_unicodes, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_9variation_unicodes(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_variation_selector = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("variation_unicodes (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_variation_selector,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_variation_selector)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 555, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "variation_unicodes") < 0)) __PYX_ERR(0, 555, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_variation_selector = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("variation_unicodes", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 555, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Face.variation_unicodes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Face_8variation_unicodes(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)__pyx_v_self), __pyx_v_variation_selector);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Face_8variation_unicodes(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_self, PyObject *__pyx_v_variation_selector) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_s = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  hb_codepoint_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("variation_unicodes", 1);

  /* "uharfbuzz/_harfbuzz.pyx":556
 * 
 *     def variation_unicodes(self, variation_selector):
 *         s = Set()             # <<<<<<<<<<<<<<
 *         hb_face_collect_variation_unicodes(self._hb_face, variation_selector, s._hb_set)
 *         return s
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Set)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_s = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":557
 *     def variation_unicodes(self, variation_selector):
 *         s = Set()
 *         hb_face_collect_variation_unicodes(self._hb_face, variation_selector, s._hb_set)             # <<<<<<<<<<<<<<
 *         return s
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_variation_selector); if (unlikely((__pyx_t_2 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 557, __pyx_L1_error)
  hb_face_collect_variation_unicodes(__pyx_v_self->_hb_face, __pyx_t_2, __pyx_v_s->_hb_set);

  /* "uharfbuzz/_harfbuzz.pyx":558
 *         s = Set()
 *         hb_face_collect_variation_unicodes(self._hb_face, variation_selector, s._hb_set)
 *         return s             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_s);
  __pyx_r = ((PyObject *)__pyx_v_s);
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":555
 *         return s
 * 
 *     def variation_unicodes(self, variation_selector):             # <<<<<<<<<<<<<<
 *         s = Set()
 *         hb_face_collect_variation_unicodes(self._hb_face, variation_selector, s._hb_set)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Face.variation_unicodes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_s);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_11__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_4Face_11__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_11__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_11__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Face_10__reduce_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Face_10__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Face.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_13__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_4Face_13__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_13__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_13__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Face.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Face_12__setstate_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Face_12__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Face.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":576
 *     cdef FontFuncs _ffuncs
 * 
 *     def __cinit__(self, face_or_font: Union[Face, Font] = None):             # <<<<<<<<<<<<<<
 *         if face_or_font is not None:
 *             if isinstance(face_or_font, Font):
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_face_or_font = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_face_or_font,0};
    values[0] = __Pyx_Arg_NewRef_VARARGS(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_face_or_font);
          if (value) { values[0] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 576, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(0, 576, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_face_or_font = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 576, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font___cinit__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self), __pyx_v_face_or_font);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_4Font___cinit__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_face_or_font) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":577
 * 
 *     def __cinit__(self, face_or_font: Union[Face, Font] = None):
 *         if face_or_font is not None:             # <<<<<<<<<<<<<<
 *             if isinstance(face_or_font, Font):
 *                 self.__create_sub_font(face_or_font)
 */
  __pyx_t_1 = (__pyx_v_face_or_font != Py_None);
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":578
 *     def __cinit__(self, face_or_font: Union[Face, Font] = None):
 *         if face_or_font is not None:
 *             if isinstance(face_or_font, Font):             # <<<<<<<<<<<<<<
 *                 self.__create_sub_font(face_or_font)
 *                 return
 */
    __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_face_or_font, __pyx_ptype_9uharfbuzz_9_harfbuzz_Font); 
    if (__pyx_t_1) {

      /* "uharfbuzz/_harfbuzz.pyx":579
 *         if face_or_font is not None:
 *             if isinstance(face_or_font, Font):
 *                 self.__create_sub_font(face_or_font)             # <<<<<<<<<<<<<<
 *                 return
 *             self.__create(face_or_font)
 */
      if (!(likely(((__pyx_v_face_or_font) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_face_or_font, __pyx_ptype_9uharfbuzz_9_harfbuzz_Font))))) __PYX_ERR(0, 579, __pyx_L1_error)
      __pyx_t_2 = ((struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self->__pyx_vtab)->_Font__create_sub_font(__pyx_v_self, ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_face_or_font)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 579, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uharfbuzz/_harfbuzz.pyx":580
 *             if isinstance(face_or_font, Font):
 *                 self.__create_sub_font(face_or_font)
 *                 return             # <<<<<<<<<<<<<<
 *             self.__create(face_or_font)
 *         else:
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "uharfbuzz/_harfbuzz.pyx":578
 *     def __cinit__(self, face_or_font: Union[Face, Font] = None):
 *         if face_or_font is not None:
 *             if isinstance(face_or_font, Font):             # <<<<<<<<<<<<<<
 *                 self.__create_sub_font(face_or_font)
 *                 return
 */
    }

    /* "uharfbuzz/_harfbuzz.pyx":581
 *                 self.__create_sub_font(face_or_font)
 *                 return
 *             self.__create(face_or_font)             # <<<<<<<<<<<<<<
 *         else:
 *             self._hb_font = hb_font_get_empty()
 */
    if (!(likely(((__pyx_v_face_or_font) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_face_or_font, __pyx_ptype_9uharfbuzz_9_harfbuzz_Face))))) __PYX_ERR(0, 581, __pyx_L1_error)
    __pyx_t_2 = ((struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self->__pyx_vtab)->_Font__create(__pyx_v_self, ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)__pyx_v_face_or_font)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 581, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uharfbuzz/_harfbuzz.pyx":577
 * 
 *     def __cinit__(self, face_or_font: Union[Face, Font] = None):
 *         if face_or_font is not None:             # <<<<<<<<<<<<<<
 *             if isinstance(face_or_font, Font):
 *                 self.__create_sub_font(face_or_font)
 */
    goto __pyx_L3;
  }

  /* "uharfbuzz/_harfbuzz.pyx":583
 *             self.__create(face_or_font)
 *         else:
 *             self._hb_font = hb_font_get_empty()             # <<<<<<<<<<<<<<
 *             self._face = Face()
 * 
 */
  /*else*/ {
    __pyx_v_self->_hb_font = hb_font_get_empty();

    /* "uharfbuzz/_harfbuzz.pyx":584
 *         else:
 *             self._hb_font = hb_font_get_empty()
 *             self._face = Face()             # <<<<<<<<<<<<<<
 * 
 *     cdef __create(self, Face face):
 */
    __pyx_t_2 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Face)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 584, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF((PyObject *)__pyx_v_self->_face);
    __Pyx_DECREF((PyObject *)__pyx_v_self->_face);
    __pyx_v_self->_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)__pyx_t_2);
    __pyx_t_2 = 0;
  }
  __pyx_L3:;

  /* "uharfbuzz/_harfbuzz.pyx":576
 *     cdef FontFuncs _ffuncs
 * 
 *     def __cinit__(self, face_or_font: Union[Face, Font] = None):             # <<<<<<<<<<<<<<
 *         if face_or_font is not None:
 *             if isinstance(face_or_font, Font):
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":586
 *             self._face = Face()
 * 
 *     cdef __create(self, Face face):             # <<<<<<<<<<<<<<
 *         self._hb_font = hb_font_create(face._hb_face)
 *         self._face = face
 */

static PyObject *__pyx_f_9uharfbuzz_9_harfbuzz_4Font__Font__create(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_Font__create", 1);

  /* "uharfbuzz/_harfbuzz.pyx":587
 * 
 *     cdef __create(self, Face face):
 *         self._hb_font = hb_font_create(face._hb_face)             # <<<<<<<<<<<<<<
 *         self._face = face
 * 
 */
  __pyx_v_self->_hb_font = hb_font_create(__pyx_v_face->_hb_face);

  /* "uharfbuzz/_harfbuzz.pyx":588
 *     cdef __create(self, Face face):
 *         self._hb_font = hb_font_create(face._hb_face)
 *         self._face = face             # <<<<<<<<<<<<<<
 * 
 *     cdef __create_sub_font(self, Font font):
 */
  __Pyx_INCREF((PyObject *)__pyx_v_face);
  __Pyx_GIVEREF((PyObject *)__pyx_v_face);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_face);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_face);
  __pyx_v_self->_face = __pyx_v_face;

  /* "uharfbuzz/_harfbuzz.pyx":586
 *             self._face = Face()
 * 
 *     cdef __create(self, Face face):             # <<<<<<<<<<<<<<
 *         self._hb_font = hb_font_create(face._hb_face)
 *         self._face = face
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":590
 *         self._face = face
 * 
 *     cdef __create_sub_font(self, Font font):             # <<<<<<<<<<<<<<
 *         self._hb_font = hb_font_create_sub_font(font._hb_font)
 *         self._face = font._face
 */

static PyObject *__pyx_f_9uharfbuzz_9_harfbuzz_4Font__Font__create_sub_font(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_Font__create_sub_font", 1);

  /* "uharfbuzz/_harfbuzz.pyx":591
 * 
 *     cdef __create_sub_font(self, Font font):
 *         self._hb_font = hb_font_create_sub_font(font._hb_font)             # <<<<<<<<<<<<<<
 *         self._face = font._face
 * 
 */
  __pyx_v_self->_hb_font = hb_font_create_sub_font(__pyx_v_font->_hb_font);

  /* "uharfbuzz/_harfbuzz.pyx":592
 *     cdef __create_sub_font(self, Font font):
 *         self._hb_font = hb_font_create_sub_font(font._hb_font)
 *         self._face = font._face             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_font->_face);
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_face);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_face);
  __pyx_v_self->_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":590
 *         self._face = face
 * 
 *     cdef __create_sub_font(self, Font font):             # <<<<<<<<<<<<<<
 *         self._hb_font = hb_font_create_sub_font(font._hb_font)
 *         self._face = font._face
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":594
 *         self._face = font._face
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         hb_font_destroy(self._hb_font)
 *         self._face = self._ffuncs = None
 */

/* Python wrapper */
static void __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_2__dealloc__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_2__dealloc__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":595
 * 
 *     def __dealloc__(self):
 *         hb_font_destroy(self._hb_font)             # <<<<<<<<<<<<<<
 *         self._face = self._ffuncs = None
 * 
 */
  hb_font_destroy(__pyx_v_self->_hb_font);

  /* "uharfbuzz/_harfbuzz.pyx":596
 *     def __dealloc__(self):
 *         hb_font_destroy(self._hb_font)
 *         self._face = self._ffuncs = None             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_face);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_face);
  __pyx_v_self->_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)Py_None);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_ffuncs);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_ffuncs);
  __pyx_v_self->_ffuncs = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)Py_None);

  /* "uharfbuzz/_harfbuzz.pyx":594
 *         self._face = font._face
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         hb_font_destroy(self._hb_font)
 *         self._face = self._ffuncs = None
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "uharfbuzz/_harfbuzz.pyx":599
 * 
 *     @staticmethod
 *     cdef Font from_ptr(hb_font_t* hb_font):             # <<<<<<<<<<<<<<
 *         """Create Font from a pointer taking ownership of a it."""
 * 
 */

static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_f_9uharfbuzz_9_harfbuzz_4Font_from_ptr(hb_font_t *__pyx_v_hb_font) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_wrapper = 0;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("from_ptr", 1);

  /* "uharfbuzz/_harfbuzz.pyx":602
 *         """Create Font from a pointer taking ownership of a it."""
 * 
 *         cdef Font wrapper = Font.__new__(Font)             # <<<<<<<<<<<<<<
 *         wrapper._hb_font = hb_font
 *         wrapper._face = Face.from_ptr(hb_face_reference(hb_font_get_face(hb_font)))
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_9uharfbuzz_9_harfbuzz_Font(((PyTypeObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Font), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 602, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":603
 * 
 *         cdef Font wrapper = Font.__new__(Font)
 *         wrapper._hb_font = hb_font             # <<<<<<<<<<<<<<
 *         wrapper._face = Face.from_ptr(hb_face_reference(hb_font_get_face(hb_font)))
 *         return wrapper
 */
  __pyx_v_wrapper->_hb_font = __pyx_v_hb_font;

  /* "uharfbuzz/_harfbuzz.pyx":604
 *         cdef Font wrapper = Font.__new__(Font)
 *         wrapper._hb_font = hb_font
 *         wrapper._face = Face.from_ptr(hb_face_reference(hb_font_get_face(hb_font)))             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_9uharfbuzz_9_harfbuzz_4Face_from_ptr(hb_face_reference(hb_font_get_face(__pyx_v_hb_font)))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 604, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_wrapper->_face);
  __Pyx_DECREF((PyObject *)__pyx_v_wrapper->_face);
  __pyx_v_wrapper->_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":605
 *         wrapper._hb_font = hb_font
 *         wrapper._face = Face.from_ptr(hb_face_reference(hb_font_get_face(hb_font)))
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     # DEPRECATED: use the normal constructor
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = __pyx_v_wrapper;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":599
 * 
 *     @staticmethod
 *     cdef Font from_ptr(hb_font_t* hb_font):             # <<<<<<<<<<<<<<
 *         """Create Font from a pointer taking ownership of a it."""
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.from_ptr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":608
 * 
 *     # DEPRECATED: use the normal constructor
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def create(cls, face: Face):
 *         cdef Font inst = cls(face)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_5create(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_5create = {"create", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_5create, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_5create(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_face,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_face)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 608, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "create") < 0)) __PYX_ERR(0, 608, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("create", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 608, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_face), __pyx_ptype_9uharfbuzz_9_harfbuzz_Face, 0, "face", 0))) __PYX_ERR(0, 609, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_4create(((PyTypeObject*)__pyx_v_cls), __pyx_v_face);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_4create(PyTypeObject *__pyx_v_cls, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_inst = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("create", 1);

  /* "uharfbuzz/_harfbuzz.pyx":610
 *     @classmethod
 *     def create(cls, face: Face):
 *         cdef Font inst = cls(face)             # <<<<<<<<<<<<<<
 *         return inst
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_v_cls), ((PyObject *)__pyx_v_face)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 610, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_9uharfbuzz_9_harfbuzz_Font))))) __PYX_ERR(0, 610, __pyx_L1_error)
  __pyx_v_inst = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":611
 *     def create(cls, face: Face):
 *         cdef Font inst = cls(face)
 *         return inst             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_inst);
  __pyx_r = ((PyObject *)__pyx_v_inst);
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":608
 * 
 *     # DEPRECATED: use the normal constructor
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def create(cls, face: Face):
 *         cdef Font inst = cls(face)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_inst);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":613
 *         return inst
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def face(self):
 *         return self._face
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_4face_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_4face_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_4face___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_4face___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":615
 *     @property
 *     def face(self):
 *         return self._face             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->_face);
  __pyx_r = ((PyObject *)__pyx_v_self->_face);
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":613
 *         return inst
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def face(self):
 *         return self._face
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":617
 *         return self._face
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def funcs(self) -> FontFuncs:
 *         return self._ffuncs
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_5funcs_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_5funcs_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_5funcs___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_5funcs___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":619
 *     @property
 *     def funcs(self) -> FontFuncs:
 *         return self._ffuncs             # <<<<<<<<<<<<<<
 * 
 *     @funcs.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->_ffuncs);
  __pyx_r = ((PyObject *)__pyx_v_self->_ffuncs);
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":617
 *         return self._face
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def funcs(self) -> FontFuncs:
 *         return self._ffuncs
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":621
 *         return self._ffuncs
 * 
 *     @funcs.setter             # <<<<<<<<<<<<<<
 *     def funcs(self, ffuncs: FontFuncs):
 *         hb_font_set_funcs(
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_5funcs_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_ffuncs); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_5funcs_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_ffuncs) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_ffuncs), __pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs, 0, "ffuncs", 0))) __PYX_ERR(0, 622, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_5funcs_2__set__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self), ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)__pyx_v_ffuncs));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_5funcs_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *__pyx_v_ffuncs) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":623
 *     @funcs.setter
 *     def funcs(self, ffuncs: FontFuncs):
 *         hb_font_set_funcs(             # <<<<<<<<<<<<<<
 *             self._hb_font, ffuncs._hb_ffuncs, <void*>self, NULL)
 *         self._ffuncs = ffuncs
 */
  hb_font_set_funcs(__pyx_v_self->_hb_font, __pyx_v_ffuncs->_hb_ffuncs, ((void *)__pyx_v_self), NULL);

  /* "uharfbuzz/_harfbuzz.pyx":625
 *         hb_font_set_funcs(
 *             self._hb_font, ffuncs._hb_ffuncs, <void*>self, NULL)
 *         self._ffuncs = ffuncs             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_INCREF((PyObject *)__pyx_v_ffuncs);
  __Pyx_GIVEREF((PyObject *)__pyx_v_ffuncs);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_ffuncs);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_ffuncs);
  __pyx_v_self->_ffuncs = __pyx_v_ffuncs;

  /* "uharfbuzz/_harfbuzz.pyx":621
 *         return self._ffuncs
 * 
 *     @funcs.setter             # <<<<<<<<<<<<<<
 *     def funcs(self, ffuncs: FontFuncs):
 *         hb_font_set_funcs(
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":627
 *         self._ffuncs = ffuncs
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def scale(self) -> Tuple[int, int]:
 *         cdef int x, y
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_5scale_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_5scale_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_5scale___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_5scale___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self) {
  int __pyx_v_x;
  int __pyx_v_y;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":630
 *     def scale(self) -> Tuple[int, int]:
 *         cdef int x, y
 *         hb_font_get_scale(self._hb_font, &x, &y)             # <<<<<<<<<<<<<<
 *         return (x, y)
 * 
 */
  hb_font_get_scale(__pyx_v_self->_hb_font, (&__pyx_v_x), (&__pyx_v_y));

  /* "uharfbuzz/_harfbuzz.pyx":631
 *         cdef int x, y
 *         hb_font_get_scale(self._hb_font, &x, &y)
 *         return (x, y)             # <<<<<<<<<<<<<<
 * 
 *     @scale.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_x); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_y); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1)) __PYX_ERR(0, 631, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2)) __PYX_ERR(0, 631, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":627
 *         self._ffuncs = ffuncs
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def scale(self) -> Tuple[int, int]:
 *         cdef int x, y
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.scale.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":633
 *         return (x, y)
 * 
 *     @scale.setter             # <<<<<<<<<<<<<<
 *     def scale(self, value: Tuple[int, int]):
 *         x, y = value
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_5scale_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_5scale_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyTuple_Type), 0, "value", 1))) __PYX_ERR(0, 634, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_5scale_2__set__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self), ((PyObject*)__pyx_v_value));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_5scale_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_x = NULL;
  PyObject *__pyx_v_y = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":635
 *     @scale.setter
 *     def scale(self, value: Tuple[int, int]):
 *         x, y = value             # <<<<<<<<<<<<<<
 *         hb_font_set_scale(self._hb_font, x, y)
 * 
 */
  if (1) {
    PyObject* sequence = __pyx_v_value;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 635, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
    __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_2);
    #else
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 635, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 635, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
  }
  __pyx_v_x = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_v_y = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":636
 *     def scale(self, value: Tuple[int, int]):
 *         x, y = value
 *         hb_font_set_scale(self._hb_font, x, y)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_v_x); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 636, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_y); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 636, __pyx_L1_error)
  hb_font_set_scale(__pyx_v_self->_hb_font, __pyx_t_3, __pyx_t_4);

  /* "uharfbuzz/_harfbuzz.pyx":633
 *         return (x, y)
 * 
 *     @scale.setter             # <<<<<<<<<<<<<<
 *     def scale(self, value: Tuple[int, int]):
 *         x, y = value
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.scale.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_x);
  __Pyx_XDECREF(__pyx_v_y);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":638
 *         hb_font_set_scale(self._hb_font, x, y)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ppem(self) -> Tuple[int, int]:
 *         cdef unsigned int x, y
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_4ppem_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_4ppem_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_4ppem___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_4ppem___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self) {
  unsigned int __pyx_v_x;
  unsigned int __pyx_v_y;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":641
 *     def ppem(self) -> Tuple[int, int]:
 *         cdef unsigned int x, y
 *         hb_font_get_ppem(self._hb_font, &x, &y)             # <<<<<<<<<<<<<<
 *         return (x, y)
 * 
 */
  hb_font_get_ppem(__pyx_v_self->_hb_font, (&__pyx_v_x), (&__pyx_v_y));

  /* "uharfbuzz/_harfbuzz.pyx":642
 *         cdef unsigned int x, y
 *         hb_font_get_ppem(self._hb_font, &x, &y)
 *         return (x, y)             # <<<<<<<<<<<<<<
 * 
 *     @ppem.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(__pyx_v_x); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_y); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1)) __PYX_ERR(0, 642, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2)) __PYX_ERR(0, 642, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":638
 *         hb_font_set_scale(self._hb_font, x, y)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ppem(self) -> Tuple[int, int]:
 *         cdef unsigned int x, y
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.ppem.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":644
 *         return (x, y)
 * 
 *     @ppem.setter             # <<<<<<<<<<<<<<
 *     def ppem(self, value: Tuple[int, int]):
 *         x, y = value
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_4ppem_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_4ppem_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyTuple_Type), 0, "value", 1))) __PYX_ERR(0, 645, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_4ppem_2__set__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self), ((PyObject*)__pyx_v_value));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_4ppem_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_x = NULL;
  PyObject *__pyx_v_y = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  unsigned int __pyx_t_3;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":646
 *     @ppem.setter
 *     def ppem(self, value: Tuple[int, int]):
 *         x, y = value             # <<<<<<<<<<<<<<
 *         hb_font_set_ppem(self._hb_font, x, y)
 * 
 */
  if (1) {
    PyObject* sequence = __pyx_v_value;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 646, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
    __pyx_t_2 = PyTuple_GET_ITEM(sequence, 1); 
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_2);
    #else
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 646, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 646, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
  }
  __pyx_v_x = __pyx_t_1;
  __pyx_t_1 = 0;
  __pyx_v_y = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":647
 *     def ppem(self, value: Tuple[int, int]):
 *         x, y = value
 *         hb_font_set_ppem(self._hb_font, x, y)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_t_3 = __Pyx_PyInt_As_unsigned_int(__pyx_v_x); if (unlikely((__pyx_t_3 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 647, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyInt_As_unsigned_int(__pyx_v_y); if (unlikely((__pyx_t_4 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 647, __pyx_L1_error)
  hb_font_set_ppem(__pyx_v_self->_hb_font, __pyx_t_3, __pyx_t_4);

  /* "uharfbuzz/_harfbuzz.pyx":644
 *         return (x, y)
 * 
 *     @ppem.setter             # <<<<<<<<<<<<<<
 *     def ppem(self, value: Tuple[int, int]):
 *         x, y = value
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.ppem.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_x);
  __Pyx_XDECREF(__pyx_v_y);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":649
 *         hb_font_set_ppem(self._hb_font, x, y)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ptem(self) -> float:
 *         return hb_font_get_ptem(self._hb_font)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_4ptem_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_4ptem_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_4ptem___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_4ptem___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":651
 *     @property
 *     def ptem(self) -> float:
 *         return hb_font_get_ptem(self._hb_font)             # <<<<<<<<<<<<<<
 * 
 *     @ptem.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(hb_font_get_ptem(__pyx_v_self->_hb_font)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":649
 *         hb_font_set_ppem(self._hb_font, x, y)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ptem(self) -> float:
 *         return hb_font_get_ptem(self._hb_font)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.ptem.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":653
 *         return hb_font_get_ptem(self._hb_font)
 * 
 *     @ptem.setter             # <<<<<<<<<<<<<<
 *     def ptem(self, value: float):
 *         hb_font_set_ptem(self._hb_font, value)
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_4ptem_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_4ptem_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value) {
  double __pyx_v_value;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  assert(__pyx_arg_value); {
    __pyx_v_value = __pyx_PyFloat_AsDouble(__pyx_arg_value); if (unlikely((__pyx_v_value == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 654, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.ptem.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_4ptem_2__set__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self), ((double)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_4ptem_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, double __pyx_v_value) {
  int __pyx_r;

  /* "uharfbuzz/_harfbuzz.pyx":655
 *     @ptem.setter
 *     def ptem(self, value: float):
 *         hb_font_set_ptem(self._hb_font, value)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  hb_font_set_ptem(__pyx_v_self->_hb_font, __pyx_v_value);

  /* "uharfbuzz/_harfbuzz.pyx":653
 *         return hb_font_get_ptem(self._hb_font)
 * 
 *     @ptem.setter             # <<<<<<<<<<<<<<
 *     def ptem(self, value: float):
 *         hb_font_set_ptem(self._hb_font, value)
 */

  /* function exit code */
  __pyx_r = 0;
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":657
 *         hb_font_set_ptem(self._hb_font, value)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def synthetic_slant(self) -> float:
 *         return hb_font_get_synthetic_slant(self._hb_font)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_15synthetic_slant_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_15synthetic_slant_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_15synthetic_slant___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_15synthetic_slant___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":659
 *     @property
 *     def synthetic_slant(self) -> float:
 *         return hb_font_get_synthetic_slant(self._hb_font)             # <<<<<<<<<<<<<<
 * 
 *     @synthetic_slant.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(hb_font_get_synthetic_slant(__pyx_v_self->_hb_font)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":657
 *         hb_font_set_ptem(self._hb_font, value)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def synthetic_slant(self) -> float:
 *         return hb_font_get_synthetic_slant(self._hb_font)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.synthetic_slant.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":661
 *         return hb_font_get_synthetic_slant(self._hb_font)
 * 
 *     @synthetic_slant.setter             # <<<<<<<<<<<<<<
 *     def synthetic_slant(self, value: float):
 *         hb_font_set_synthetic_slant(self._hb_font, value)
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_15synthetic_slant_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_15synthetic_slant_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_value) {
  double __pyx_v_value;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  assert(__pyx_arg_value); {
    __pyx_v_value = __pyx_PyFloat_AsDouble(__pyx_arg_value); if (unlikely((__pyx_v_value == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 662, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.synthetic_slant.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_15synthetic_slant_2__set__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self), ((double)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_15synthetic_slant_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, double __pyx_v_value) {
  int __pyx_r;

  /* "uharfbuzz/_harfbuzz.pyx":663
 *     @synthetic_slant.setter
 *     def synthetic_slant(self, value: float):
 *         hb_font_set_synthetic_slant(self._hb_font, value)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  hb_font_set_synthetic_slant(__pyx_v_self->_hb_font, __pyx_v_value);

  /* "uharfbuzz/_harfbuzz.pyx":661
 *         return hb_font_get_synthetic_slant(self._hb_font)
 * 
 *     @synthetic_slant.setter             # <<<<<<<<<<<<<<
 *     def synthetic_slant(self, value: float):
 *         hb_font_set_synthetic_slant(self._hb_font, value)
 */

  /* function exit code */
  __pyx_r = 0;
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":665
 *         hb_font_set_synthetic_slant(self._hb_font, value)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def synthetic_bold(self) -> tuple[float, float, bool]:
 *         cdef float x_embolden
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_14synthetic_bold_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_14synthetic_bold_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_14synthetic_bold___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_14synthetic_bold___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self) {
  float __pyx_v_x_embolden;
  float __pyx_v_y_embolden;
  hb_bool_t __pyx_v_in_place;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":670
 *         cdef float y_embolden
 *         cdef hb_bool_t in_place
 *         hb_font_get_synthetic_bold(self._hb_font, &x_embolden, &y_embolden, &in_place)             # <<<<<<<<<<<<<<
 *         return (x_embolden, y_embolden, bool(in_place))
 * 
 */
  hb_font_get_synthetic_bold(__pyx_v_self->_hb_font, (&__pyx_v_x_embolden), (&__pyx_v_y_embolden), (&__pyx_v_in_place));

  /* "uharfbuzz/_harfbuzz.pyx":671
 *         cdef hb_bool_t in_place
 *         hb_font_get_synthetic_bold(self._hb_font, &x_embolden, &y_embolden, &in_place)
 *         return (x_embolden, y_embolden, bool(in_place))             # <<<<<<<<<<<<<<
 * 
 *     @synthetic_bold.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_x_embolden); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_y_embolden); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_hb_bool_t(__pyx_v_in_place); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 671, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyBool_FromLong((!(!__pyx_t_4))); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 671, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1)) __PYX_ERR(0, 671, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2)) __PYX_ERR(0, 671, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3)) __PYX_ERR(0, 671, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":665
 *         hb_font_set_synthetic_slant(self._hb_font, value)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def synthetic_bold(self) -> tuple[float, float, bool]:
 *         cdef float x_embolden
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.synthetic_bold.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":673
 *         return (x_embolden, y_embolden, bool(in_place))
 * 
 *     @synthetic_bold.setter             # <<<<<<<<<<<<<<
 *     def synthetic_bold(self, value: float|tuple[float]|tuple[float,float]|tuple[float,float,bool]):
 *         cdef float x_embolden
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_14synthetic_bold_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_14synthetic_bold_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_14synthetic_bold_2__set__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_14synthetic_bold_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_value) {
  float __pyx_v_x_embolden;
  float __pyx_v_y_embolden;
  hb_bool_t __pyx_v_in_place;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  float __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *(*__pyx_t_7)(PyObject *);
  float __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  hb_bool_t __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":677
 *         cdef float x_embolden
 *         cdef float y_embolden
 *         cdef hb_bool_t in_place = False             # <<<<<<<<<<<<<<
 *         if isinstance(value, tuple):
 *             if len(value) == 1:
 */
  __pyx_v_in_place = 0;

  /* "uharfbuzz/_harfbuzz.pyx":678
 *         cdef float y_embolden
 *         cdef hb_bool_t in_place = False
 *         if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             if len(value) == 1:
 *                 x_embolden = y_embolden = value[0]
 */
  __pyx_t_1 = PyTuple_Check(__pyx_v_value); 
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":679
 *         cdef hb_bool_t in_place = False
 *         if isinstance(value, tuple):
 *             if len(value) == 1:             # <<<<<<<<<<<<<<
 *                 x_embolden = y_embolden = value[0]
 *             elif len(value) == 2:
 */
    __pyx_t_2 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 679, __pyx_L1_error)
    __pyx_t_1 = (__pyx_t_2 == 1);
    if (__pyx_t_1) {

      /* "uharfbuzz/_harfbuzz.pyx":680
 *         if isinstance(value, tuple):
 *             if len(value) == 1:
 *                 x_embolden = y_embolden = value[0]             # <<<<<<<<<<<<<<
 *             elif len(value) == 2:
 *                 x_embolden, y_embolden = value
 */
      __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_value, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 680, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __pyx_PyFloat_AsFloat(__pyx_t_3); if (unlikely((__pyx_t_4 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 680, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_v_x_embolden = __pyx_t_4;
      __pyx_v_y_embolden = __pyx_t_4;

      /* "uharfbuzz/_harfbuzz.pyx":679
 *         cdef hb_bool_t in_place = False
 *         if isinstance(value, tuple):
 *             if len(value) == 1:             # <<<<<<<<<<<<<<
 *                 x_embolden = y_embolden = value[0]
 *             elif len(value) == 2:
 */
      goto __pyx_L4;
    }

    /* "uharfbuzz/_harfbuzz.pyx":681
 *             if len(value) == 1:
 *                 x_embolden = y_embolden = value[0]
 *             elif len(value) == 2:             # <<<<<<<<<<<<<<
 *                 x_embolden, y_embolden = value
 *             else:
 */
    __pyx_t_2 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 681, __pyx_L1_error)
    __pyx_t_1 = (__pyx_t_2 == 2);
    if (__pyx_t_1) {

      /* "uharfbuzz/_harfbuzz.pyx":682
 *                 x_embolden = y_embolden = value[0]
 *             elif len(value) == 2:
 *                 x_embolden, y_embolden = value             # <<<<<<<<<<<<<<
 *             else:
 *                 x_embolden, y_embolden, in_place = value
 */
      if ((likely(PyTuple_CheckExact(__pyx_v_value))) || (PyList_CheckExact(__pyx_v_value))) {
        PyObject* sequence = __pyx_v_value;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 682, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        #else
        __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 682, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 682, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        Py_ssize_t index = -1;
        __pyx_t_6 = PyObject_GetIter(__pyx_v_value); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 682, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_6);
        index = 0; __pyx_t_3 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_3)) goto __pyx_L5_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_3);
        index = 1; __pyx_t_5 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_5);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_6), 2) < 0) __PYX_ERR(0, 682, __pyx_L1_error)
        __pyx_t_7 = NULL;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L6_unpacking_done;
        __pyx_L5_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_7 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 682, __pyx_L1_error)
        __pyx_L6_unpacking_done:;
      }
      __pyx_t_4 = __pyx_PyFloat_AsFloat(__pyx_t_3); if (unlikely((__pyx_t_4 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 682, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_8 = __pyx_PyFloat_AsFloat(__pyx_t_5); if (unlikely((__pyx_t_8 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 682, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_v_x_embolden = __pyx_t_4;
      __pyx_v_y_embolden = __pyx_t_8;

      /* "uharfbuzz/_harfbuzz.pyx":681
 *             if len(value) == 1:
 *                 x_embolden = y_embolden = value[0]
 *             elif len(value) == 2:             # <<<<<<<<<<<<<<
 *                 x_embolden, y_embolden = value
 *             else:
 */
      goto __pyx_L4;
    }

    /* "uharfbuzz/_harfbuzz.pyx":684
 *                 x_embolden, y_embolden = value
 *             else:
 *                 x_embolden, y_embolden, in_place = value             # <<<<<<<<<<<<<<
 *         else:
 *             x_embolden = y_embolden = value
 */
    /*else*/ {
      if ((likely(PyTuple_CheckExact(__pyx_v_value))) || (PyList_CheckExact(__pyx_v_value))) {
        PyObject* sequence = __pyx_v_value;
        Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
        if (unlikely(size != 3)) {
          if (size > 3) __Pyx_RaiseTooManyValuesError(3);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 684, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
          __pyx_t_6 = PyTuple_GET_ITEM(sequence, 2); 
        } else {
          __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
          __pyx_t_6 = PyList_GET_ITEM(sequence, 2); 
        }
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_6);
        #else
        __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 684, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 684, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_6 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 684, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
      } else {
        Py_ssize_t index = -1;
        __pyx_t_9 = PyObject_GetIter(__pyx_v_value); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 684, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_9);
        index = 0; __pyx_t_5 = __pyx_t_7(__pyx_t_9); if (unlikely(!__pyx_t_5)) goto __pyx_L7_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_5);
        index = 1; __pyx_t_3 = __pyx_t_7(__pyx_t_9); if (unlikely(!__pyx_t_3)) goto __pyx_L7_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_3);
        index = 2; __pyx_t_6 = __pyx_t_7(__pyx_t_9); if (unlikely(!__pyx_t_6)) goto __pyx_L7_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_6);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_9), 3) < 0) __PYX_ERR(0, 684, __pyx_L1_error)
        __pyx_t_7 = NULL;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L8_unpacking_done;
        __pyx_L7_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_7 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 684, __pyx_L1_error)
        __pyx_L8_unpacking_done:;
      }
      __pyx_t_8 = __pyx_PyFloat_AsFloat(__pyx_t_5); if (unlikely((__pyx_t_8 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 684, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_4 = __pyx_PyFloat_AsFloat(__pyx_t_3); if (unlikely((__pyx_t_4 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 684, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_10 = __Pyx_PyInt_As_hb_bool_t(__pyx_t_6); if (unlikely((__pyx_t_10 == ((hb_bool_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 684, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_v_x_embolden = __pyx_t_8;
      __pyx_v_y_embolden = __pyx_t_4;
      __pyx_v_in_place = __pyx_t_10;
    }
    __pyx_L4:;

    /* "uharfbuzz/_harfbuzz.pyx":678
 *         cdef float y_embolden
 *         cdef hb_bool_t in_place = False
 *         if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             if len(value) == 1:
 *                 x_embolden = y_embolden = value[0]
 */
    goto __pyx_L3;
  }

  /* "uharfbuzz/_harfbuzz.pyx":686
 *                 x_embolden, y_embolden, in_place = value
 *         else:
 *             x_embolden = y_embolden = value             # <<<<<<<<<<<<<<
 *         hb_font_set_synthetic_bold(self._hb_font, x_embolden, y_embolden, in_place)
 * 
 */
  /*else*/ {
    __pyx_t_4 = __pyx_PyFloat_AsFloat(__pyx_v_value); if (unlikely((__pyx_t_4 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 686, __pyx_L1_error)
    __pyx_v_x_embolden = __pyx_t_4;
    __pyx_v_y_embolden = __pyx_t_4;
  }
  __pyx_L3:;

  /* "uharfbuzz/_harfbuzz.pyx":687
 *         else:
 *             x_embolden = y_embolden = value
 *         hb_font_set_synthetic_bold(self._hb_font, x_embolden, y_embolden, in_place)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  hb_font_set_synthetic_bold(__pyx_v_self->_hb_font, __pyx_v_x_embolden, __pyx_v_y_embolden, __pyx_v_in_place);

  /* "uharfbuzz/_harfbuzz.pyx":673
 *         return (x_embolden, y_embolden, bool(in_place))
 * 
 *     @synthetic_bold.setter             # <<<<<<<<<<<<<<
 *     def synthetic_bold(self, value: float|tuple[float]|tuple[float,float]|tuple[float,float,bool]):
 *         cdef float x_embolden
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.synthetic_bold.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":689
 *         hb_font_set_synthetic_bold(self._hb_font, x_embolden, y_embolden, in_place)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def var_named_instance(self) -> int:
 *         return hb_font_get_var_named_instance(self._hb_font)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_18var_named_instance_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_18var_named_instance_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_18var_named_instance___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_18var_named_instance___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":691
 *     @property
 *     def var_named_instance(self) -> int:
 *         return hb_font_get_var_named_instance(self._hb_font)             # <<<<<<<<<<<<<<
 * 
 *     @var_named_instance.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(hb_font_get_var_named_instance(__pyx_v_self->_hb_font)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":689
 *         hb_font_set_synthetic_bold(self._hb_font, x_embolden, y_embolden, in_place)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def var_named_instance(self) -> int:
 *         return hb_font_get_var_named_instance(self._hb_font)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.var_named_instance.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":693
 *         return hb_font_get_var_named_instance(self._hb_font)
 * 
 *     @var_named_instance.setter             # <<<<<<<<<<<<<<
 *     def var_named_instance(self, value: int):
 *         hb_font_set_var_named_instance(self._hb_font, value)
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_18var_named_instance_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_18var_named_instance_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyInt_Type), 0, "value", 1))) __PYX_ERR(0, 694, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_18var_named_instance_2__set__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self), ((PyObject*)__pyx_v_value));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_18var_named_instance_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  unsigned int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "uharfbuzz/_harfbuzz.pyx":695
 *     @var_named_instance.setter
 *     def var_named_instance(self, value: int):
 *         hb_font_set_var_named_instance(self._hb_font, value)             # <<<<<<<<<<<<<<
 * 
 *     def set_variations(self, variations: Dict[str, float]) -> None:
 */
  __pyx_t_1 = __Pyx_PyInt_As_unsigned_int(__pyx_v_value); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 695, __pyx_L1_error)
  hb_font_set_var_named_instance(__pyx_v_self->_hb_font, __pyx_t_1);

  /* "uharfbuzz/_harfbuzz.pyx":693
 *         return hb_font_get_var_named_instance(self._hb_font)
 * 
 *     @var_named_instance.setter             # <<<<<<<<<<<<<<
 *     def var_named_instance(self, value: int):
 *         hb_font_set_var_named_instance(self._hb_font, value)
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.var_named_instance.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":697
 *         hb_font_set_var_named_instance(self._hb_font, value)
 * 
 *     def set_variations(self, variations: Dict[str, float]) -> None:             # <<<<<<<<<<<<<<
 *         cdef unsigned int size
 *         cdef hb_variation_t* hb_variations
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_7set_variations(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_7set_variations = {"set_variations", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_7set_variations, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_7set_variations(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_variations = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_variations (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_variations,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_variations)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 697, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_variations") < 0)) __PYX_ERR(0, 697, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_variations = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_variations", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 697, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.set_variations", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_variations), (&PyDict_Type), 0, "variations", 1))) __PYX_ERR(0, 697, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_6set_variations(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self), __pyx_v_variations);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_6set_variations(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_variations) {
  unsigned int __pyx_v_size;
  hb_variation_t *__pyx_v_hb_variations;
  PyObject *__pyx_v_packed = 0;
  hb_variation_t __pyx_v_variation;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  unsigned int __pyx_t_11;
  char const *__pyx_t_12;
  float __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_variations", 1);

  /* "uharfbuzz/_harfbuzz.pyx":702
 *         cdef bytes packed
 *         cdef hb_variation_t variation
 *         size = len(variations)             # <<<<<<<<<<<<<<
 *         hb_variations = <hb_variation_t*>malloc(size * sizeof(hb_variation_t))
 *         if not hb_variations:
 */
  __pyx_t_1 = PyDict_Size(__pyx_v_variations); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 702, __pyx_L1_error)
  __pyx_v_size = __pyx_t_1;

  /* "uharfbuzz/_harfbuzz.pyx":703
 *         cdef hb_variation_t variation
 *         size = len(variations)
 *         hb_variations = <hb_variation_t*>malloc(size * sizeof(hb_variation_t))             # <<<<<<<<<<<<<<
 *         if not hb_variations:
 *             raise MemoryError()
 */
  __pyx_v_hb_variations = ((hb_variation_t *)malloc((__pyx_v_size * (sizeof(hb_variation_t)))));

  /* "uharfbuzz/_harfbuzz.pyx":704
 *         size = len(variations)
 *         hb_variations = <hb_variation_t*>malloc(size * sizeof(hb_variation_t))
 *         if not hb_variations:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_2 = (!(__pyx_v_hb_variations != 0));
  if (unlikely(__pyx_t_2)) {

    /* "uharfbuzz/_harfbuzz.pyx":705
 *         hb_variations = <hb_variation_t*>malloc(size * sizeof(hb_variation_t))
 *         if not hb_variations:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
    PyErr_NoMemory(); __PYX_ERR(0, 705, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":704
 *         size = len(variations)
 *         hb_variations = <hb_variation_t*>malloc(size * sizeof(hb_variation_t))
 *         if not hb_variations:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":707
 *             raise MemoryError()
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             for i, (name, value) in enumerate(variations.items()):
 *                 packed = name.encode()
 */
  /*try:*/ {

    /* "uharfbuzz/_harfbuzz.pyx":708
 * 
 *         try:
 *             for i, (name, value) in enumerate(variations.items()):             # <<<<<<<<<<<<<<
 *                 packed = name.encode()
 *                 variation.tag = hb_tag_from_string(packed, -1)
 */
    __Pyx_INCREF(__pyx_int_0);
    __pyx_t_3 = __pyx_int_0;
    __pyx_t_1 = 0;
    __pyx_t_7 = __Pyx_dict_iterator(__pyx_v_variations, 1, __pyx_n_s_items, (&__pyx_t_5), (&__pyx_t_6)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 708, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_4);
    __pyx_t_4 = __pyx_t_7;
    __pyx_t_7 = 0;
    while (1) {
      __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_4, __pyx_t_5, &__pyx_t_1, &__pyx_t_7, &__pyx_t_8, NULL, __pyx_t_6);
      if (unlikely(__pyx_t_9 == 0)) break;
      if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 708, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_7);
      __pyx_t_7 = 0;
      __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_8);
      __pyx_t_8 = 0;
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_3);
      __pyx_t_8 = __Pyx_PyInt_AddObjC(__pyx_t_3, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 708, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_3);
      __pyx_t_3 = __pyx_t_8;
      __pyx_t_8 = 0;

      /* "uharfbuzz/_harfbuzz.pyx":709
 *         try:
 *             for i, (name, value) in enumerate(variations.items()):
 *                 packed = name.encode()             # <<<<<<<<<<<<<<
 *                 variation.tag = hb_tag_from_string(packed, -1)
 *                 variation.value = value
 */
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_name, __pyx_n_s_encode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 709, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_10 = NULL;
      __pyx_t_11 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_11 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_10, NULL};
        __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_11, 0+__pyx_t_11);
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 709, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      if (!(likely(PyBytes_CheckExact(__pyx_t_8))||((__pyx_t_8) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_8))) __PYX_ERR(0, 709, __pyx_L5_error)
      __Pyx_XDECREF_SET(__pyx_v_packed, ((PyObject*)__pyx_t_8));
      __pyx_t_8 = 0;

      /* "uharfbuzz/_harfbuzz.pyx":710
 *             for i, (name, value) in enumerate(variations.items()):
 *                 packed = name.encode()
 *                 variation.tag = hb_tag_from_string(packed, -1)             # <<<<<<<<<<<<<<
 *                 variation.value = value
 *                 hb_variations[i] = variation
 */
      if (unlikely(__pyx_v_packed == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
        __PYX_ERR(0, 710, __pyx_L5_error)
      }
      __pyx_t_12 = __Pyx_PyBytes_AsString(__pyx_v_packed); if (unlikely((!__pyx_t_12) && PyErr_Occurred())) __PYX_ERR(0, 710, __pyx_L5_error)
      __pyx_v_variation.tag = hb_tag_from_string(__pyx_t_12, -1);

      /* "uharfbuzz/_harfbuzz.pyx":711
 *                 packed = name.encode()
 *                 variation.tag = hb_tag_from_string(packed, -1)
 *                 variation.value = value             # <<<<<<<<<<<<<<
 *                 hb_variations[i] = variation
 *             hb_font_set_variations(self._hb_font, hb_variations, size)
 */
      __pyx_t_13 = __pyx_PyFloat_AsFloat(__pyx_v_value); if (unlikely((__pyx_t_13 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 711, __pyx_L5_error)
      __pyx_v_variation.value = __pyx_t_13;

      /* "uharfbuzz/_harfbuzz.pyx":712
 *                 variation.tag = hb_tag_from_string(packed, -1)
 *                 variation.value = value
 *                 hb_variations[i] = variation             # <<<<<<<<<<<<<<
 *             hb_font_set_variations(self._hb_font, hb_variations, size)
 *         finally:
 */
      __pyx_t_14 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_14 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 712, __pyx_L5_error)
      (__pyx_v_hb_variations[__pyx_t_14]) = __pyx_v_variation;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "uharfbuzz/_harfbuzz.pyx":713
 *                 variation.value = value
 *                 hb_variations[i] = variation
 *             hb_font_set_variations(self._hb_font, hb_variations, size)             # <<<<<<<<<<<<<<
 *         finally:
 *             free(hb_variations)
 */
    hb_font_set_variations(__pyx_v_self->_hb_font, __pyx_v_hb_variations, __pyx_v_size);
  }

  /* "uharfbuzz/_harfbuzz.pyx":715
 *             hb_font_set_variations(self._hb_font, hb_variations, size)
 *         finally:
 *             free(hb_variations)             # <<<<<<<<<<<<<<
 * 
 *     def set_variation(self, name: str, value: float) -> None:
 */
  /*finally:*/ {
    /*normal exit:*/{
      free(__pyx_v_hb_variations);
      goto __pyx_L6;
    }
    __pyx_L5_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_16);
      __Pyx_XGOTREF(__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_21);
      __pyx_t_6 = __pyx_lineno; __pyx_t_9 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
      {
        free(__pyx_v_hb_variations);
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_XGIVEREF(__pyx_t_21);
        __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
      }
      __Pyx_XGIVEREF(__pyx_t_16);
      __Pyx_XGIVEREF(__pyx_t_17);
      __Pyx_XGIVEREF(__pyx_t_18);
      __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
      __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
      __pyx_lineno = __pyx_t_6; __pyx_clineno = __pyx_t_9; __pyx_filename = __pyx_t_15;
      goto __pyx_L1_error;
    }
    __pyx_L6:;
  }

  /* "uharfbuzz/_harfbuzz.pyx":697
 *         hb_font_set_var_named_instance(self._hb_font, value)
 * 
 *     def set_variations(self, variations: Dict[str, float]) -> None:             # <<<<<<<<<<<<<<
 *         cdef unsigned int size
 *         cdef hb_variation_t* hb_variations
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.set_variations", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_packed);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":717
 *             free(hb_variations)
 * 
 *     def set_variation(self, name: str, value: float) -> None:             # <<<<<<<<<<<<<<
 *         packed = name.encode()
 *         cdef hb_tag_t tag = hb_tag_from_string(packed, -1)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_9set_variation(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_9set_variation = {"set_variation", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_9set_variation, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_9set_variation(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_name = 0;
  double __pyx_v_value;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_variation (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_name)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 717, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 717, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("set_variation", 1, 2, 2, 1); __PYX_ERR(0, 717, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_variation") < 0)) __PYX_ERR(0, 717, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_name = ((PyObject*)values[0]);
    __pyx_v_value = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_value == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 717, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_variation", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 717, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.set_variation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_name), (&PyUnicode_Type), 0, "name", 1))) __PYX_ERR(0, 717, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_8set_variation(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self), __pyx_v_name, __pyx_v_value);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_8set_variation(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_name, double __pyx_v_value) {
  PyObject *__pyx_v_packed = NULL;
  hb_tag_t __pyx_v_tag;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char const *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_variation", 1);

  /* "uharfbuzz/_harfbuzz.pyx":718
 * 
 *     def set_variation(self, name: str, value: float) -> None:
 *         packed = name.encode()             # <<<<<<<<<<<<<<
 *         cdef hb_tag_t tag = hb_tag_from_string(packed, -1)
 *         hb_font_set_variation(self._hb_font, tag, value)
 */
  __pyx_t_1 = PyUnicode_AsEncodedString(__pyx_v_name, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_packed = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":719
 *     def set_variation(self, name: str, value: float) -> None:
 *         packed = name.encode()
 *         cdef hb_tag_t tag = hb_tag_from_string(packed, -1)             # <<<<<<<<<<<<<<
 *         hb_font_set_variation(self._hb_font, tag, value)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_AsString(__pyx_v_packed); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 719, __pyx_L1_error)
  __pyx_v_tag = hb_tag_from_string(__pyx_t_2, -1);

  /* "uharfbuzz/_harfbuzz.pyx":720
 *         packed = name.encode()
 *         cdef hb_tag_t tag = hb_tag_from_string(packed, -1)
 *         hb_font_set_variation(self._hb_font, tag, value)             # <<<<<<<<<<<<<<
 * 
 *     def get_glyph_name(self, gid: int):
 */
  hb_font_set_variation(__pyx_v_self->_hb_font, __pyx_v_tag, __pyx_v_value);

  /* "uharfbuzz/_harfbuzz.pyx":717
 *             free(hb_variations)
 * 
 *     def set_variation(self, name: str, value: float) -> None:             # <<<<<<<<<<<<<<
 *         packed = name.encode()
 *         cdef hb_tag_t tag = hb_tag_from_string(packed, -1)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.set_variation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_packed);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":722
 *         hb_font_set_variation(self._hb_font, tag, value)
 * 
 *     def get_glyph_name(self, gid: int):             # <<<<<<<<<<<<<<
 *         cdef char name[64]
 *         cdef bytes packed
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_11get_glyph_name(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_11get_glyph_name = {"get_glyph_name", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_11get_glyph_name, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_11get_glyph_name(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_gid = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_glyph_name (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_gid,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_gid)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 722, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_glyph_name") < 0)) __PYX_ERR(0, 722, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_gid = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_glyph_name", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 722, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.get_glyph_name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_gid), (&PyInt_Type), 0, "gid", 1))) __PYX_ERR(0, 722, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_10get_glyph_name(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self), __pyx_v_gid);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_10get_glyph_name(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_gid) {
  char __pyx_v_name[64];
  PyObject *__pyx_v_packed = 0;
  hb_bool_t __pyx_v_success;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  hb_codepoint_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_glyph_name", 1);

  /* "uharfbuzz/_harfbuzz.pyx":725
 *         cdef char name[64]
 *         cdef bytes packed
 *         success = hb_font_get_glyph_name(self._hb_font, gid, name, 64)             # <<<<<<<<<<<<<<
 *         if success:
 *             packed = name
 */
  __pyx_t_1 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_gid); if (unlikely((__pyx_t_1 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 725, __pyx_L1_error)
  __pyx_v_success = hb_font_get_glyph_name(__pyx_v_self->_hb_font, __pyx_t_1, __pyx_v_name, 64);

  /* "uharfbuzz/_harfbuzz.pyx":726
 *         cdef bytes packed
 *         success = hb_font_get_glyph_name(self._hb_font, gid, name, 64)
 *         if success:             # <<<<<<<<<<<<<<
 *             packed = name
 *             return packed.decode()
 */
  __pyx_t_2 = (__pyx_v_success != 0);
  if (__pyx_t_2) {

    /* "uharfbuzz/_harfbuzz.pyx":727
 *         success = hb_font_get_glyph_name(self._hb_font, gid, name, 64)
 *         if success:
 *             packed = name             # <<<<<<<<<<<<<<
 *             return packed.decode()
 *         else:
 */
    __pyx_t_3 = __Pyx_PyObject_FromString(__pyx_v_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 727, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_packed = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "uharfbuzz/_harfbuzz.pyx":728
 *         if success:
 *             packed = name
 *             return packed.decode()             # <<<<<<<<<<<<<<
 *         else:
 *             return None
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_decode_bytes(__pyx_v_packed, 0, PY_SSIZE_T_MAX, NULL, NULL, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 728, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "uharfbuzz/_harfbuzz.pyx":726
 *         cdef bytes packed
 *         success = hb_font_get_glyph_name(self._hb_font, gid, name, 64)
 *         if success:             # <<<<<<<<<<<<<<
 *             packed = name
 *             return packed.decode()
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":730
 *             return packed.decode()
 *         else:
 *             return None             # <<<<<<<<<<<<<<
 * 
 *     def get_glyph_from_name(self, name: str):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;
  }

  /* "uharfbuzz/_harfbuzz.pyx":722
 *         hb_font_set_variation(self._hb_font, tag, value)
 * 
 *     def get_glyph_name(self, gid: int):             # <<<<<<<<<<<<<<
 *         cdef char name[64]
 *         cdef bytes packed
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.get_glyph_name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_packed);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":732
 *             return None
 * 
 *     def get_glyph_from_name(self, name: str):             # <<<<<<<<<<<<<<
 *         cdef hb_codepoint_t gid
 *         cdef bytes packed
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_13get_glyph_from_name(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_13get_glyph_from_name = {"get_glyph_from_name", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_13get_glyph_from_name, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_13get_glyph_from_name(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_name = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_glyph_from_name (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_name)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 732, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_glyph_from_name") < 0)) __PYX_ERR(0, 732, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_name = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_glyph_from_name", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 732, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.get_glyph_from_name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_name), (&PyUnicode_Type), 0, "name", 1))) __PYX_ERR(0, 732, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_12get_glyph_from_name(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self), __pyx_v_name);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_12get_glyph_from_name(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_name) {
  hb_codepoint_t __pyx_v_gid;
  PyObject *__pyx_v_packed = 0;
  hb_bool_t __pyx_v_success;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_glyph_from_name", 1);

  /* "uharfbuzz/_harfbuzz.pyx":735
 *         cdef hb_codepoint_t gid
 *         cdef bytes packed
 *         packed = name.encode()             # <<<<<<<<<<<<<<
 *         success = hb_font_get_glyph_from_name(self._hb_font, <char*>packed, len(packed), &gid)
 *         return gid if success else None
 */
  __pyx_t_1 = PyUnicode_AsEncodedString(__pyx_v_name, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_packed = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":736
 *         cdef bytes packed
 *         packed = name.encode()
 *         success = hb_font_get_glyph_from_name(self._hb_font, <char*>packed, len(packed), &gid)             # <<<<<<<<<<<<<<
 *         return gid if success else None
 * 
 */
  if (unlikely(__pyx_v_packed == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 736, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBytes_AsWritableString(__pyx_v_packed); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 736, __pyx_L1_error)
  if (unlikely(__pyx_v_packed == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 736, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyBytes_GET_SIZE(__pyx_v_packed); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 736, __pyx_L1_error)
  __pyx_v_success = hb_font_get_glyph_from_name(__pyx_v_self->_hb_font, ((char *)__pyx_t_2), __pyx_t_3, (&__pyx_v_gid));

  /* "uharfbuzz/_harfbuzz.pyx":737
 *         packed = name.encode()
 *         success = hb_font_get_glyph_from_name(self._hb_font, <char*>packed, len(packed), &gid)
 *         return gid if success else None             # <<<<<<<<<<<<<<
 * 
 *     def get_glyph_extents(self, gid: int):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = (__pyx_v_success != 0);
  if (__pyx_t_4) {
    __pyx_t_5 = __Pyx_PyInt_From_hb_codepoint_t(__pyx_v_gid); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 737, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __pyx_t_5;
    __pyx_t_5 = 0;
  } else {
    __Pyx_INCREF(Py_None);
    __pyx_t_1 = Py_None;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":732
 *             return None
 * 
 *     def get_glyph_from_name(self, name: str):             # <<<<<<<<<<<<<<
 *         cdef hb_codepoint_t gid
 *         cdef bytes packed
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.get_glyph_from_name", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_packed);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":739
 *         return gid if success else None
 * 
 *     def get_glyph_extents(self, gid: int):             # <<<<<<<<<<<<<<
 *         cdef hb_glyph_extents_t extents
 *         success = hb_font_get_glyph_extents(self._hb_font, gid, &extents)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_15get_glyph_extents(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_15get_glyph_extents = {"get_glyph_extents", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_15get_glyph_extents, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_15get_glyph_extents(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_gid = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_glyph_extents (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_gid,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_gid)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 739, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_glyph_extents") < 0)) __PYX_ERR(0, 739, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_gid = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_glyph_extents", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 739, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.get_glyph_extents", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_gid), (&PyInt_Type), 0, "gid", 1))) __PYX_ERR(0, 739, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_14get_glyph_extents(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self), __pyx_v_gid);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_14get_glyph_extents(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_gid) {
  hb_glyph_extents_t __pyx_v_extents;
  hb_bool_t __pyx_v_success;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  hb_codepoint_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  unsigned int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_glyph_extents", 1);

  /* "uharfbuzz/_harfbuzz.pyx":741
 *     def get_glyph_extents(self, gid: int):
 *         cdef hb_glyph_extents_t extents
 *         success = hb_font_get_glyph_extents(self._hb_font, gid, &extents)             # <<<<<<<<<<<<<<
 *         if success:
 *             return GlyphExtents(
 */
  __pyx_t_1 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_gid); if (unlikely((__pyx_t_1 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 741, __pyx_L1_error)
  __pyx_v_success = hb_font_get_glyph_extents(__pyx_v_self->_hb_font, __pyx_t_1, (&__pyx_v_extents));

  /* "uharfbuzz/_harfbuzz.pyx":742
 *         cdef hb_glyph_extents_t extents
 *         success = hb_font_get_glyph_extents(self._hb_font, gid, &extents)
 *         if success:             # <<<<<<<<<<<<<<
 *             return GlyphExtents(
 *                 extents.x_bearing,
 */
  __pyx_t_2 = (__pyx_v_success != 0);
  if (__pyx_t_2) {

    /* "uharfbuzz/_harfbuzz.pyx":743
 *         success = hb_font_get_glyph_extents(self._hb_font, gid, &extents)
 *         if success:
 *             return GlyphExtents(             # <<<<<<<<<<<<<<
 *                 extents.x_bearing,
 *                 extents.y_bearing,
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_GlyphExtents); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 743, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "uharfbuzz/_harfbuzz.pyx":744
 *         if success:
 *             return GlyphExtents(
 *                 extents.x_bearing,             # <<<<<<<<<<<<<<
 *                 extents.y_bearing,
 *                 extents.width,
 */
    __pyx_t_5 = __Pyx_PyInt_From_hb_position_t(__pyx_v_extents.x_bearing); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 744, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "uharfbuzz/_harfbuzz.pyx":745
 *             return GlyphExtents(
 *                 extents.x_bearing,
 *                 extents.y_bearing,             # <<<<<<<<<<<<<<
 *                 extents.width,
 *                 extents.height,
 */
    __pyx_t_6 = __Pyx_PyInt_From_hb_position_t(__pyx_v_extents.y_bearing); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 745, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);

    /* "uharfbuzz/_harfbuzz.pyx":746
 *                 extents.x_bearing,
 *                 extents.y_bearing,
 *                 extents.width,             # <<<<<<<<<<<<<<
 *                 extents.height,
 *             )
 */
    __pyx_t_7 = __Pyx_PyInt_From_hb_position_t(__pyx_v_extents.width); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 746, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);

    /* "uharfbuzz/_harfbuzz.pyx":747
 *                 extents.y_bearing,
 *                 extents.width,
 *                 extents.height,             # <<<<<<<<<<<<<<
 *             )
 *         else:
 */
    __pyx_t_8 = __Pyx_PyInt_From_hb_position_t(__pyx_v_extents.height); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 747, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = NULL;
    __pyx_t_10 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_10 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[5] = {__pyx_t_9, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_10, 4+__pyx_t_10);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 743, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "uharfbuzz/_harfbuzz.pyx":742
 *         cdef hb_glyph_extents_t extents
 *         success = hb_font_get_glyph_extents(self._hb_font, gid, &extents)
 *         if success:             # <<<<<<<<<<<<<<
 *             return GlyphExtents(
 *                 extents.x_bearing,
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":750
 *             )
 *         else:
 *             return None             # <<<<<<<<<<<<<<
 * 
 *     def get_glyph_h_advance(self, gid: int):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;
  }

  /* "uharfbuzz/_harfbuzz.pyx":739
 *         return gid if success else None
 * 
 *     def get_glyph_extents(self, gid: int):             # <<<<<<<<<<<<<<
 *         cdef hb_glyph_extents_t extents
 *         success = hb_font_get_glyph_extents(self._hb_font, gid, &extents)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.get_glyph_extents", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":752
 *             return None
 * 
 *     def get_glyph_h_advance(self, gid: int):             # <<<<<<<<<<<<<<
 *         return hb_font_get_glyph_h_advance(self._hb_font, gid)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_17get_glyph_h_advance(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_17get_glyph_h_advance = {"get_glyph_h_advance", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_17get_glyph_h_advance, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_17get_glyph_h_advance(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_gid = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_glyph_h_advance (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_gid,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_gid)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 752, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_glyph_h_advance") < 0)) __PYX_ERR(0, 752, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_gid = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_glyph_h_advance", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 752, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.get_glyph_h_advance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_gid), (&PyInt_Type), 0, "gid", 1))) __PYX_ERR(0, 752, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_16get_glyph_h_advance(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self), __pyx_v_gid);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_16get_glyph_h_advance(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_gid) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  hb_codepoint_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_glyph_h_advance", 1);

  /* "uharfbuzz/_harfbuzz.pyx":753
 * 
 *     def get_glyph_h_advance(self, gid: int):
 *         return hb_font_get_glyph_h_advance(self._hb_font, gid)             # <<<<<<<<<<<<<<
 * 
 *     def get_glyph_v_advance(self, gid: int):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_gid); if (unlikely((__pyx_t_1 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 753, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_hb_position_t(hb_font_get_glyph_h_advance(__pyx_v_self->_hb_font, __pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 753, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":752
 *             return None
 * 
 *     def get_glyph_h_advance(self, gid: int):             # <<<<<<<<<<<<<<
 *         return hb_font_get_glyph_h_advance(self._hb_font, gid)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.get_glyph_h_advance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":755
 *         return hb_font_get_glyph_h_advance(self._hb_font, gid)
 * 
 *     def get_glyph_v_advance(self, gid: int):             # <<<<<<<<<<<<<<
 *         return hb_font_get_glyph_v_advance(self._hb_font, gid)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_19get_glyph_v_advance(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_19get_glyph_v_advance = {"get_glyph_v_advance", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_19get_glyph_v_advance, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_19get_glyph_v_advance(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_gid = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_glyph_v_advance (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_gid,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_gid)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 755, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_glyph_v_advance") < 0)) __PYX_ERR(0, 755, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_gid = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_glyph_v_advance", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 755, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.get_glyph_v_advance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_gid), (&PyInt_Type), 0, "gid", 1))) __PYX_ERR(0, 755, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_18get_glyph_v_advance(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self), __pyx_v_gid);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_18get_glyph_v_advance(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_gid) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  hb_codepoint_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_glyph_v_advance", 1);

  /* "uharfbuzz/_harfbuzz.pyx":756
 * 
 *     def get_glyph_v_advance(self, gid: int):
 *         return hb_font_get_glyph_v_advance(self._hb_font, gid)             # <<<<<<<<<<<<<<
 * 
 *     def get_glyph_h_origin(self, gid: int):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_gid); if (unlikely((__pyx_t_1 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 756, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_hb_position_t(hb_font_get_glyph_v_advance(__pyx_v_self->_hb_font, __pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 756, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":755
 *         return hb_font_get_glyph_h_advance(self._hb_font, gid)
 * 
 *     def get_glyph_v_advance(self, gid: int):             # <<<<<<<<<<<<<<
 *         return hb_font_get_glyph_v_advance(self._hb_font, gid)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.get_glyph_v_advance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":758
 *         return hb_font_get_glyph_v_advance(self._hb_font, gid)
 * 
 *     def get_glyph_h_origin(self, gid: int):             # <<<<<<<<<<<<<<
 *         cdef hb_position_t x, y
 *         success = hb_font_get_glyph_h_origin(self._hb_font, gid, &x, &y)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_21get_glyph_h_origin(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_21get_glyph_h_origin = {"get_glyph_h_origin", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_21get_glyph_h_origin, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_21get_glyph_h_origin(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_gid = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_glyph_h_origin (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_gid,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_gid)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 758, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_glyph_h_origin") < 0)) __PYX_ERR(0, 758, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_gid = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_glyph_h_origin", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 758, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.get_glyph_h_origin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_gid), (&PyInt_Type), 0, "gid", 1))) __PYX_ERR(0, 758, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_20get_glyph_h_origin(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self), __pyx_v_gid);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_20get_glyph_h_origin(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_gid) {
  hb_position_t __pyx_v_x;
  hb_position_t __pyx_v_y;
  hb_bool_t __pyx_v_success;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  hb_codepoint_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_glyph_h_origin", 1);

  /* "uharfbuzz/_harfbuzz.pyx":760
 *     def get_glyph_h_origin(self, gid: int):
 *         cdef hb_position_t x, y
 *         success = hb_font_get_glyph_h_origin(self._hb_font, gid, &x, &y)             # <<<<<<<<<<<<<<
 *         return (x, y) if success else None
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_gid); if (unlikely((__pyx_t_1 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 760, __pyx_L1_error)
  __pyx_v_success = hb_font_get_glyph_h_origin(__pyx_v_self->_hb_font, __pyx_t_1, (&__pyx_v_x), (&__pyx_v_y));

  /* "uharfbuzz/_harfbuzz.pyx":761
 *         cdef hb_position_t x, y
 *         success = hb_font_get_glyph_h_origin(self._hb_font, gid, &x, &y)
 *         return (x, y) if success else None             # <<<<<<<<<<<<<<
 * 
 *     def get_glyph_v_origin(self, gid: int):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = (__pyx_v_success != 0);
  if (__pyx_t_3) {
    __pyx_t_4 = __Pyx_PyInt_From_hb_position_t(__pyx_v_x); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 761, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyInt_From_hb_position_t(__pyx_v_y); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 761, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 761, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4)) __PYX_ERR(0, 761, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_5)) __PYX_ERR(0, 761, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_2 = __pyx_t_6;
    __pyx_t_6 = 0;
  } else {
    __Pyx_INCREF(Py_None);
    __pyx_t_2 = Py_None;
  }
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":758
 *         return hb_font_get_glyph_v_advance(self._hb_font, gid)
 * 
 *     def get_glyph_h_origin(self, gid: int):             # <<<<<<<<<<<<<<
 *         cdef hb_position_t x, y
 *         success = hb_font_get_glyph_h_origin(self._hb_font, gid, &x, &y)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.get_glyph_h_origin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":763
 *         return (x, y) if success else None
 * 
 *     def get_glyph_v_origin(self, gid: int):             # <<<<<<<<<<<<<<
 *         cdef hb_position_t x, y
 *         success = hb_font_get_glyph_v_origin(self._hb_font, gid, &x, &y)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_23get_glyph_v_origin(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_23get_glyph_v_origin = {"get_glyph_v_origin", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_23get_glyph_v_origin, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_23get_glyph_v_origin(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_gid = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_glyph_v_origin (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_gid,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_gid)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 763, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_glyph_v_origin") < 0)) __PYX_ERR(0, 763, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_gid = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_glyph_v_origin", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 763, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.get_glyph_v_origin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_gid), (&PyInt_Type), 0, "gid", 1))) __PYX_ERR(0, 763, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_22get_glyph_v_origin(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self), __pyx_v_gid);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_22get_glyph_v_origin(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_gid) {
  hb_position_t __pyx_v_x;
  hb_position_t __pyx_v_y;
  hb_bool_t __pyx_v_success;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  hb_codepoint_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_glyph_v_origin", 1);

  /* "uharfbuzz/_harfbuzz.pyx":765
 *     def get_glyph_v_origin(self, gid: int):
 *         cdef hb_position_t x, y
 *         success = hb_font_get_glyph_v_origin(self._hb_font, gid, &x, &y)             # <<<<<<<<<<<<<<
 *         return (x, y) if success else None
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_gid); if (unlikely((__pyx_t_1 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 765, __pyx_L1_error)
  __pyx_v_success = hb_font_get_glyph_v_origin(__pyx_v_self->_hb_font, __pyx_t_1, (&__pyx_v_x), (&__pyx_v_y));

  /* "uharfbuzz/_harfbuzz.pyx":766
 *         cdef hb_position_t x, y
 *         success = hb_font_get_glyph_v_origin(self._hb_font, gid, &x, &y)
 *         return (x, y) if success else None             # <<<<<<<<<<<<<<
 * 
 *     def get_font_extents(self, direction: str):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = (__pyx_v_success != 0);
  if (__pyx_t_3) {
    __pyx_t_4 = __Pyx_PyInt_From_hb_position_t(__pyx_v_x); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 766, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyInt_From_hb_position_t(__pyx_v_y); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 766, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 766, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4)) __PYX_ERR(0, 766, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_5)) __PYX_ERR(0, 766, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_2 = __pyx_t_6;
    __pyx_t_6 = 0;
  } else {
    __Pyx_INCREF(Py_None);
    __pyx_t_2 = Py_None;
  }
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":763
 *         return (x, y) if success else None
 * 
 *     def get_glyph_v_origin(self, gid: int):             # <<<<<<<<<<<<<<
 *         cdef hb_position_t x, y
 *         success = hb_font_get_glyph_v_origin(self._hb_font, gid, &x, &y)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.get_glyph_v_origin", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":768
 *         return (x, y) if success else None
 * 
 *     def get_font_extents(self, direction: str):             # <<<<<<<<<<<<<<
 *         cdef hb_font_extents_t extents
 *         cdef hb_direction_t hb_direction
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_25get_font_extents(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_25get_font_extents = {"get_font_extents", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_25get_font_extents, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_25get_font_extents(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_direction = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_font_extents (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_direction,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_direction)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 768, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_font_extents") < 0)) __PYX_ERR(0, 768, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_direction = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_font_extents", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 768, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.get_font_extents", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_direction), (&PyUnicode_Type), 0, "direction", 1))) __PYX_ERR(0, 768, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_24get_font_extents(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self), __pyx_v_direction);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_24get_font_extents(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_direction) {
  hb_font_extents_t __pyx_v_extents;
  hb_direction_t __pyx_v_hb_direction;
  PyObject *__pyx_v_packed = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_font_extents", 1);

  /* "uharfbuzz/_harfbuzz.pyx":772
 *         cdef hb_direction_t hb_direction
 *         cdef bytes packed
 *         packed = direction.encode()             # <<<<<<<<<<<<<<
 *         hb_direction = hb_direction_from_string(<char*>packed, -1)
 *         hb_font_get_extents_for_direction(
 */
  __pyx_t_1 = PyUnicode_AsEncodedString(__pyx_v_direction, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_packed = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":773
 *         cdef bytes packed
 *         packed = direction.encode()
 *         hb_direction = hb_direction_from_string(<char*>packed, -1)             # <<<<<<<<<<<<<<
 *         hb_font_get_extents_for_direction(
 *             self._hb_font, hb_direction, &extents
 */
  if (unlikely(__pyx_v_packed == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 773, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBytes_AsWritableString(__pyx_v_packed); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 773, __pyx_L1_error)
  __pyx_v_hb_direction = hb_direction_from_string(((char *)__pyx_t_2), -1);

  /* "uharfbuzz/_harfbuzz.pyx":774
 *         packed = direction.encode()
 *         hb_direction = hb_direction_from_string(<char*>packed, -1)
 *         hb_font_get_extents_for_direction(             # <<<<<<<<<<<<<<
 *             self._hb_font, hb_direction, &extents
 *         )
 */
  hb_font_get_extents_for_direction(__pyx_v_self->_hb_font, __pyx_v_hb_direction, (&__pyx_v_extents));

  /* "uharfbuzz/_harfbuzz.pyx":777
 *             self._hb_font, hb_direction, &extents
 *         )
 *         return FontExtents(             # <<<<<<<<<<<<<<
 *             extents.ascender,
 *             extents.descender,
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_FontExtents); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "uharfbuzz/_harfbuzz.pyx":778
 *         )
 *         return FontExtents(
 *             extents.ascender,             # <<<<<<<<<<<<<<
 *             extents.descender,
 *             extents.line_gap
 */
  __pyx_t_4 = __Pyx_PyInt_From_hb_position_t(__pyx_v_extents.ascender); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 778, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "uharfbuzz/_harfbuzz.pyx":779
 *         return FontExtents(
 *             extents.ascender,
 *             extents.descender,             # <<<<<<<<<<<<<<
 *             extents.line_gap
 *         )
 */
  __pyx_t_5 = __Pyx_PyInt_From_hb_position_t(__pyx_v_extents.descender); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 779, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "uharfbuzz/_harfbuzz.pyx":780
 *             extents.ascender,
 *             extents.descender,
 *             extents.line_gap             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __pyx_t_6 = __Pyx_PyInt_From_hb_position_t(__pyx_v_extents.line_gap); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_7, __pyx_t_4, __pyx_t_5, __pyx_t_6};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_8, 3+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 777, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":768
 *         return (x, y) if success else None
 * 
 *     def get_font_extents(self, direction: str):             # <<<<<<<<<<<<<<
 *         cdef hb_font_extents_t extents
 *         cdef hb_direction_t hb_direction
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.get_font_extents", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_packed);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":783
 *         )
 * 
 *     def get_variation_glyph(self, unicode: int, variation_selector: int):             # <<<<<<<<<<<<<<
 *         cdef hb_codepoint_t gid
 *         success = hb_font_get_variation_glyph(self._hb_font, unicode, variation_selector, &gid)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_27get_variation_glyph(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_27get_variation_glyph = {"get_variation_glyph", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_27get_variation_glyph, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_27get_variation_glyph(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_unicode = 0;
  PyObject *__pyx_v_variation_selector = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_variation_glyph (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_unicode,&__pyx_n_s_variation_selector,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_unicode)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 783, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_variation_selector)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 783, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("get_variation_glyph", 1, 2, 2, 1); __PYX_ERR(0, 783, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_variation_glyph") < 0)) __PYX_ERR(0, 783, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_unicode = ((PyObject*)values[0]);
    __pyx_v_variation_selector = ((PyObject*)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_variation_glyph", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 783, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.get_variation_glyph", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_unicode), (&PyInt_Type), 0, "unicode", 1))) __PYX_ERR(0, 783, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_variation_selector), (&PyInt_Type), 0, "variation_selector", 1))) __PYX_ERR(0, 783, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_26get_variation_glyph(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self), __pyx_v_unicode, __pyx_v_variation_selector);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_26get_variation_glyph(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_unicode, PyObject *__pyx_v_variation_selector) {
  hb_codepoint_t __pyx_v_gid;
  hb_bool_t __pyx_v_success;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  hb_codepoint_t __pyx_t_1;
  hb_codepoint_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_variation_glyph", 1);

  /* "uharfbuzz/_harfbuzz.pyx":785
 *     def get_variation_glyph(self, unicode: int, variation_selector: int):
 *         cdef hb_codepoint_t gid
 *         success = hb_font_get_variation_glyph(self._hb_font, unicode, variation_selector, &gid)             # <<<<<<<<<<<<<<
 *         return gid if success else None
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_unicode); if (unlikely((__pyx_t_1 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 785, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_variation_selector); if (unlikely((__pyx_t_2 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 785, __pyx_L1_error)
  __pyx_v_success = hb_font_get_variation_glyph(__pyx_v_self->_hb_font, __pyx_t_1, __pyx_t_2, (&__pyx_v_gid));

  /* "uharfbuzz/_harfbuzz.pyx":786
 *         cdef hb_codepoint_t gid
 *         success = hb_font_get_variation_glyph(self._hb_font, unicode, variation_selector, &gid)
 *         return gid if success else None             # <<<<<<<<<<<<<<
 * 
 *     def get_nominal_glyph(self, unicode: int):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = (__pyx_v_success != 0);
  if (__pyx_t_4) {
    __pyx_t_5 = __Pyx_PyInt_From_hb_codepoint_t(__pyx_v_gid); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 786, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __pyx_t_5;
    __pyx_t_5 = 0;
  } else {
    __Pyx_INCREF(Py_None);
    __pyx_t_3 = Py_None;
  }
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":783
 *         )
 * 
 *     def get_variation_glyph(self, unicode: int, variation_selector: int):             # <<<<<<<<<<<<<<
 *         cdef hb_codepoint_t gid
 *         success = hb_font_get_variation_glyph(self._hb_font, unicode, variation_selector, &gid)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.get_variation_glyph", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":788
 *         return gid if success else None
 * 
 *     def get_nominal_glyph(self, unicode: int):             # <<<<<<<<<<<<<<
 *         cdef hb_codepoint_t gid
 *         success = hb_font_get_nominal_glyph(self._hb_font, unicode, &gid)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_29get_nominal_glyph(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_29get_nominal_glyph = {"get_nominal_glyph", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_29get_nominal_glyph, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_29get_nominal_glyph(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_unicode = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_nominal_glyph (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_unicode,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_unicode)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 788, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_nominal_glyph") < 0)) __PYX_ERR(0, 788, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_unicode = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_nominal_glyph", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 788, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.get_nominal_glyph", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_unicode), (&PyInt_Type), 0, "unicode", 1))) __PYX_ERR(0, 788, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_28get_nominal_glyph(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self), __pyx_v_unicode);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_28get_nominal_glyph(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_unicode) {
  hb_codepoint_t __pyx_v_gid;
  hb_bool_t __pyx_v_success;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  hb_codepoint_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_nominal_glyph", 1);

  /* "uharfbuzz/_harfbuzz.pyx":790
 *     def get_nominal_glyph(self, unicode: int):
 *         cdef hb_codepoint_t gid
 *         success = hb_font_get_nominal_glyph(self._hb_font, unicode, &gid)             # <<<<<<<<<<<<<<
 *         return gid if success else None
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_unicode); if (unlikely((__pyx_t_1 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 790, __pyx_L1_error)
  __pyx_v_success = hb_font_get_nominal_glyph(__pyx_v_self->_hb_font, __pyx_t_1, (&__pyx_v_gid));

  /* "uharfbuzz/_harfbuzz.pyx":791
 *         cdef hb_codepoint_t gid
 *         success = hb_font_get_nominal_glyph(self._hb_font, unicode, &gid)
 *         return gid if success else None             # <<<<<<<<<<<<<<
 * 
 *     def get_var_coords_normalized(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = (__pyx_v_success != 0);
  if (__pyx_t_3) {
    __pyx_t_4 = __Pyx_PyInt_From_hb_codepoint_t(__pyx_v_gid); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 791, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = __pyx_t_4;
    __pyx_t_4 = 0;
  } else {
    __Pyx_INCREF(Py_None);
    __pyx_t_2 = Py_None;
  }
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":788
 *         return gid if success else None
 * 
 *     def get_nominal_glyph(self, unicode: int):             # <<<<<<<<<<<<<<
 *         cdef hb_codepoint_t gid
 *         success = hb_font_get_nominal_glyph(self._hb_font, unicode, &gid)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.get_nominal_glyph", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":793
 *         return gid if success else None
 * 
 *     def get_var_coords_normalized(self):             # <<<<<<<<<<<<<<
 *         cdef unsigned int length
 *         cdef const int *coords
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_31get_var_coords_normalized(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_31get_var_coords_normalized = {"get_var_coords_normalized", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_31get_var_coords_normalized, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_31get_var_coords_normalized(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_var_coords_normalized (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_var_coords_normalized", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_var_coords_normalized", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_30get_var_coords_normalized(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_30get_var_coords_normalized(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self) {
  unsigned int __pyx_v_length;
  int const *__pyx_v_coords;
  unsigned int __pyx_7genexpr__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  unsigned int __pyx_t_2;
  unsigned int __pyx_t_3;
  unsigned int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_var_coords_normalized", 1);

  /* "uharfbuzz/_harfbuzz.pyx":796
 *         cdef unsigned int length
 *         cdef const int *coords
 *         coords = hb_font_get_var_coords_normalized(self._hb_font, &length)             # <<<<<<<<<<<<<<
 *         # Convert from 2.14 fixed to float: divide by 1 << 14
 *         return [coords[i] / 0x4000 for i in range(length)]
 */
  __pyx_v_coords = hb_font_get_var_coords_normalized(__pyx_v_self->_hb_font, (&__pyx_v_length));

  /* "uharfbuzz/_harfbuzz.pyx":798
 *         coords = hb_font_get_var_coords_normalized(self._hb_font, &length)
 *         # Convert from 2.14 fixed to float: divide by 1 << 14
 *         return [coords[i] / 0x4000 for i in range(length)]             # <<<<<<<<<<<<<<
 * 
 *     def set_var_coords_normalized(self, coords):
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_v_length;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_7genexpr__pyx_v_i = __pyx_t_4;
      __pyx_t_5 = PyFloat_FromDouble((((double)(__pyx_v_coords[__pyx_7genexpr__pyx_v_i])) / 16384.0)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 798, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 798, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
  } /* exit inner scope */
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":793
 *         return gid if success else None
 * 
 *     def get_var_coords_normalized(self):             # <<<<<<<<<<<<<<
 *         cdef unsigned int length
 *         cdef const int *coords
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.get_var_coords_normalized", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":800
 *         return [coords[i] / 0x4000 for i in range(length)]
 * 
 *     def set_var_coords_normalized(self, coords):             # <<<<<<<<<<<<<<
 *         cdef unsigned int length
 *         cdef int *coords_2dot14
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_33set_var_coords_normalized(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_33set_var_coords_normalized = {"set_var_coords_normalized", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_33set_var_coords_normalized, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_33set_var_coords_normalized(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_coords = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_var_coords_normalized (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_coords,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_coords)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 800, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_var_coords_normalized") < 0)) __PYX_ERR(0, 800, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_coords = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_var_coords_normalized", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 800, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.set_var_coords_normalized", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_32set_var_coords_normalized(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self), __pyx_v_coords);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_32set_var_coords_normalized(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_coords) {
  unsigned int __pyx_v_length;
  int *__pyx_v_coords_2dot14;
  unsigned int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  unsigned int __pyx_t_3;
  unsigned int __pyx_t_4;
  unsigned int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  int __pyx_t_9;
  char const *__pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_var_coords_normalized", 1);

  /* "uharfbuzz/_harfbuzz.pyx":803
 *         cdef unsigned int length
 *         cdef int *coords_2dot14
 *         length = len(coords)             # <<<<<<<<<<<<<<
 *         coords_2dot14 = <int *>malloc(length * sizeof(int))
 *         if coords_2dot14 is NULL:
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_coords); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 803, __pyx_L1_error)
  __pyx_v_length = __pyx_t_1;

  /* "uharfbuzz/_harfbuzz.pyx":804
 *         cdef int *coords_2dot14
 *         length = len(coords)
 *         coords_2dot14 = <int *>malloc(length * sizeof(int))             # <<<<<<<<<<<<<<
 *         if coords_2dot14 is NULL:
 *             raise MemoryError()
 */
  __pyx_v_coords_2dot14 = ((int *)malloc((__pyx_v_length * (sizeof(int)))));

  /* "uharfbuzz/_harfbuzz.pyx":805
 *         length = len(coords)
 *         coords_2dot14 = <int *>malloc(length * sizeof(int))
 *         if coords_2dot14 is NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         try:
 */
  __pyx_t_2 = (__pyx_v_coords_2dot14 == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "uharfbuzz/_harfbuzz.pyx":806
 *         coords_2dot14 = <int *>malloc(length * sizeof(int))
 *         if coords_2dot14 is NULL:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 *         try:
 *             for i in range(length):
 */
    PyErr_NoMemory(); __PYX_ERR(0, 806, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":805
 *         length = len(coords)
 *         coords_2dot14 = <int *>malloc(length * sizeof(int))
 *         if coords_2dot14 is NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         try:
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":807
 *         if coords_2dot14 is NULL:
 *             raise MemoryError()
 *         try:             # <<<<<<<<<<<<<<
 *             for i in range(length):
 *                 # Convert from float to 2.14 fixed: multiply by 1 << 14
 */
  /*try:*/ {

    /* "uharfbuzz/_harfbuzz.pyx":808
 *             raise MemoryError()
 *         try:
 *             for i in range(length):             # <<<<<<<<<<<<<<
 *                 # Convert from float to 2.14 fixed: multiply by 1 << 14
 *                 coords_2dot14[i] = round(coords[i] * 0x4000)
 */
    __pyx_t_3 = __pyx_v_length;
    __pyx_t_4 = __pyx_t_3;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "uharfbuzz/_harfbuzz.pyx":810
 *             for i in range(length):
 *                 # Convert from float to 2.14 fixed: multiply by 1 << 14
 *                 coords_2dot14[i] = round(coords[i] * 0x4000)             # <<<<<<<<<<<<<<
 *             hb_font_set_var_coords_normalized(self._hb_font, coords_2dot14, length)
 *         finally:
 */
      __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_coords, __pyx_v_i, unsigned int, 0, __Pyx_PyInt_From_unsigned_int, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 810, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyInt_MultiplyObjC(__pyx_t_6, __pyx_int_16384, 0x4000, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 810, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_round, __pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 810, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_t_6); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 810, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      (__pyx_v_coords_2dot14[__pyx_v_i]) = __pyx_t_8;
    }

    /* "uharfbuzz/_harfbuzz.pyx":811
 *                 # Convert from float to 2.14 fixed: multiply by 1 << 14
 *                 coords_2dot14[i] = round(coords[i] * 0x4000)
 *             hb_font_set_var_coords_normalized(self._hb_font, coords_2dot14, length)             # <<<<<<<<<<<<<<
 *         finally:
 *             free(coords_2dot14)
 */
    hb_font_set_var_coords_normalized(__pyx_v_self->_hb_font, __pyx_v_coords_2dot14, __pyx_v_length);
  }

  /* "uharfbuzz/_harfbuzz.pyx":813
 *             hb_font_set_var_coords_normalized(self._hb_font, coords_2dot14, length)
 *         finally:
 *             free(coords_2dot14)             # <<<<<<<<<<<<<<
 * 
 *     def get_var_coords_design(self):
 */
  /*finally:*/ {
    /*normal exit:*/{
      free(__pyx_v_coords_2dot14);
      goto __pyx_L6;
    }
    __pyx_L5_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13) < 0)) __Pyx_ErrFetch(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_15);
      __Pyx_XGOTREF(__pyx_t_16);
      __pyx_t_8 = __pyx_lineno; __pyx_t_9 = __pyx_clineno; __pyx_t_10 = __pyx_filename;
      {
        free(__pyx_v_coords_2dot14);
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_XGIVEREF(__pyx_t_16);
        __Pyx_ExceptionReset(__pyx_t_14, __pyx_t_15, __pyx_t_16);
      }
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_ErrRestore(__pyx_t_11, __pyx_t_12, __pyx_t_13);
      __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0;
      __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_9; __pyx_filename = __pyx_t_10;
      goto __pyx_L1_error;
    }
    __pyx_L6:;
  }

  /* "uharfbuzz/_harfbuzz.pyx":800
 *         return [coords[i] / 0x4000 for i in range(length)]
 * 
 *     def set_var_coords_normalized(self, coords):             # <<<<<<<<<<<<<<
 *         cdef unsigned int length
 *         cdef int *coords_2dot14
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.set_var_coords_normalized", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":815
 *             free(coords_2dot14)
 * 
 *     def get_var_coords_design(self):             # <<<<<<<<<<<<<<
 *         cdef unsigned int length
 *         cdef const float *coords
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_35get_var_coords_design(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_35get_var_coords_design = {"get_var_coords_design", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_35get_var_coords_design, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_35get_var_coords_design(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_var_coords_design (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_var_coords_design", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_var_coords_design", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_34get_var_coords_design(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_34get_var_coords_design(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self) {
  unsigned int __pyx_v_length;
  float const *__pyx_v_coords;
  unsigned int __pyx_8genexpr1__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  unsigned int __pyx_t_2;
  unsigned int __pyx_t_3;
  unsigned int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_var_coords_design", 1);

  /* "uharfbuzz/_harfbuzz.pyx":818
 *         cdef unsigned int length
 *         cdef const float *coords
 *         coords = hb_font_get_var_coords_design(self._hb_font, &length)             # <<<<<<<<<<<<<<
 *         return [coords[i] for i in range(length)]
 * 
 */
  __pyx_v_coords = hb_font_get_var_coords_design(__pyx_v_self->_hb_font, (&__pyx_v_length));

  /* "uharfbuzz/_harfbuzz.pyx":819
 *         cdef const float *coords
 *         coords = hb_font_get_var_coords_design(self._hb_font, &length)
 *         return [coords[i] for i in range(length)]             # <<<<<<<<<<<<<<
 * 
 *     def set_var_coords_design(self, coords):
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 819, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_v_length;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_8genexpr1__pyx_v_i = __pyx_t_4;
      __pyx_t_5 = PyFloat_FromDouble((__pyx_v_coords[__pyx_8genexpr1__pyx_v_i])); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 819, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 819, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
  } /* exit inner scope */
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":815
 *             free(coords_2dot14)
 * 
 *     def get_var_coords_design(self):             # <<<<<<<<<<<<<<
 *         cdef unsigned int length
 *         cdef const float *coords
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.get_var_coords_design", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":821
 *         return [coords[i] for i in range(length)]
 * 
 *     def set_var_coords_design(self, coords):             # <<<<<<<<<<<<<<
 *         cdef unsigned int length
 *         cdef cython.float *c_coords
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_37set_var_coords_design(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_37set_var_coords_design = {"set_var_coords_design", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_37set_var_coords_design, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_37set_var_coords_design(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_coords = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_var_coords_design (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_coords,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_coords)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 821, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_var_coords_design") < 0)) __PYX_ERR(0, 821, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_coords = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_var_coords_design", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 821, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.set_var_coords_design", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_36set_var_coords_design(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self), __pyx_v_coords);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_36set_var_coords_design(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_coords) {
  unsigned int __pyx_v_length;
  float *__pyx_v_c_coords;
  unsigned int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  unsigned int __pyx_t_3;
  unsigned int __pyx_t_4;
  unsigned int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  float __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  char const *__pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_var_coords_design", 1);

  /* "uharfbuzz/_harfbuzz.pyx":824
 *         cdef unsigned int length
 *         cdef cython.float *c_coords
 *         length = len(coords)             # <<<<<<<<<<<<<<
 *         c_coords = <cython.float*>malloc(length * sizeof(cython.float))
 *         if c_coords is NULL:
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_coords); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 824, __pyx_L1_error)
  __pyx_v_length = __pyx_t_1;

  /* "uharfbuzz/_harfbuzz.pyx":825
 *         cdef cython.float *c_coords
 *         length = len(coords)
 *         c_coords = <cython.float*>malloc(length * sizeof(cython.float))             # <<<<<<<<<<<<<<
 *         if c_coords is NULL:
 *             raise MemoryError()
 */
  __pyx_v_c_coords = ((float *)malloc((__pyx_v_length * (sizeof(float)))));

  /* "uharfbuzz/_harfbuzz.pyx":826
 *         length = len(coords)
 *         c_coords = <cython.float*>malloc(length * sizeof(cython.float))
 *         if c_coords is NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         try:
 */
  __pyx_t_2 = (__pyx_v_c_coords == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "uharfbuzz/_harfbuzz.pyx":827
 *         c_coords = <cython.float*>malloc(length * sizeof(cython.float))
 *         if c_coords is NULL:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 *         try:
 *             for i in range(length):
 */
    PyErr_NoMemory(); __PYX_ERR(0, 827, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":826
 *         length = len(coords)
 *         c_coords = <cython.float*>malloc(length * sizeof(cython.float))
 *         if c_coords is NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         try:
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":828
 *         if c_coords is NULL:
 *             raise MemoryError()
 *         try:             # <<<<<<<<<<<<<<
 *             for i in range(length):
 *                 c_coords[i] = coords[i]
 */
  /*try:*/ {

    /* "uharfbuzz/_harfbuzz.pyx":829
 *             raise MemoryError()
 *         try:
 *             for i in range(length):             # <<<<<<<<<<<<<<
 *                 c_coords[i] = coords[i]
 *             hb_font_set_var_coords_design(self._hb_font, c_coords, length)
 */
    __pyx_t_3 = __pyx_v_length;
    __pyx_t_4 = __pyx_t_3;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "uharfbuzz/_harfbuzz.pyx":830
 *         try:
 *             for i in range(length):
 *                 c_coords[i] = coords[i]             # <<<<<<<<<<<<<<
 *             hb_font_set_var_coords_design(self._hb_font, c_coords, length)
 *         finally:
 */
      __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_coords, __pyx_v_i, unsigned int, 0, __Pyx_PyInt_From_unsigned_int, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 830, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __pyx_PyFloat_AsFloat(__pyx_t_6); if (unlikely((__pyx_t_7 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 830, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      (__pyx_v_c_coords[__pyx_v_i]) = __pyx_t_7;
    }

    /* "uharfbuzz/_harfbuzz.pyx":831
 *             for i in range(length):
 *                 c_coords[i] = coords[i]
 *             hb_font_set_var_coords_design(self._hb_font, c_coords, length)             # <<<<<<<<<<<<<<
 *         finally:
 *             free(c_coords)
 */
    hb_font_set_var_coords_design(__pyx_v_self->_hb_font, __pyx_v_c_coords, __pyx_v_length);
  }

  /* "uharfbuzz/_harfbuzz.pyx":833
 *             hb_font_set_var_coords_design(self._hb_font, c_coords, length)
 *         finally:
 *             free(c_coords)             # <<<<<<<<<<<<<<
 * 
 *     def glyph_to_string(self, gid: int):
 */
  /*finally:*/ {
    /*normal exit:*/{
      free(__pyx_v_c_coords);
      goto __pyx_L6;
    }
    __pyx_L5_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13) < 0)) __Pyx_ErrFetch(&__pyx_t_11, &__pyx_t_12, &__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_13);
      __Pyx_XGOTREF(__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_15);
      __Pyx_XGOTREF(__pyx_t_16);
      __pyx_t_8 = __pyx_lineno; __pyx_t_9 = __pyx_clineno; __pyx_t_10 = __pyx_filename;
      {
        free(__pyx_v_c_coords);
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_XGIVEREF(__pyx_t_16);
        __Pyx_ExceptionReset(__pyx_t_14, __pyx_t_15, __pyx_t_16);
      }
      __Pyx_XGIVEREF(__pyx_t_11);
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_13);
      __Pyx_ErrRestore(__pyx_t_11, __pyx_t_12, __pyx_t_13);
      __pyx_t_11 = 0; __pyx_t_12 = 0; __pyx_t_13 = 0; __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0;
      __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_9; __pyx_filename = __pyx_t_10;
      goto __pyx_L1_error;
    }
    __pyx_L6:;
  }

  /* "uharfbuzz/_harfbuzz.pyx":821
 *         return [coords[i] for i in range(length)]
 * 
 *     def set_var_coords_design(self, coords):             # <<<<<<<<<<<<<<
 *         cdef unsigned int length
 *         cdef cython.float *c_coords
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.set_var_coords_design", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":835
 *             free(c_coords)
 * 
 *     def glyph_to_string(self, gid: int):             # <<<<<<<<<<<<<<
 *         cdef char name[64]
 *         cdef bytes packed
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_39glyph_to_string(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_39glyph_to_string = {"glyph_to_string", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_39glyph_to_string, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_39glyph_to_string(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_gid = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("glyph_to_string (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_gid,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_gid)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 835, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "glyph_to_string") < 0)) __PYX_ERR(0, 835, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_gid = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("glyph_to_string", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 835, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.glyph_to_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_gid), (&PyInt_Type), 0, "gid", 1))) __PYX_ERR(0, 835, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_38glyph_to_string(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self), __pyx_v_gid);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_38glyph_to_string(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_gid) {
  char __pyx_v_name[64];
  PyObject *__pyx_v_packed = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  hb_codepoint_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("glyph_to_string", 1);

  /* "uharfbuzz/_harfbuzz.pyx":838
 *         cdef char name[64]
 *         cdef bytes packed
 *         hb_font_glyph_to_string(self._hb_font, gid, name, 64)             # <<<<<<<<<<<<<<
 *         packed = name
 *         return packed.decode()
 */
  __pyx_t_1 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_gid); if (unlikely((__pyx_t_1 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 838, __pyx_L1_error)
  hb_font_glyph_to_string(__pyx_v_self->_hb_font, __pyx_t_1, __pyx_v_name, 64);

  /* "uharfbuzz/_harfbuzz.pyx":839
 *         cdef bytes packed
 *         hb_font_glyph_to_string(self._hb_font, gid, name, 64)
 *         packed = name             # <<<<<<<<<<<<<<
 *         return packed.decode()
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_FromString(__pyx_v_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_packed = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":840
 *         hb_font_glyph_to_string(self._hb_font, gid, name, 64)
 *         packed = name
 *         return packed.decode()             # <<<<<<<<<<<<<<
 * 
 *     def glyph_from_string(self, string: str):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_decode_bytes(__pyx_v_packed, 0, PY_SSIZE_T_MAX, NULL, NULL, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":835
 *             free(c_coords)
 * 
 *     def glyph_to_string(self, gid: int):             # <<<<<<<<<<<<<<
 *         cdef char name[64]
 *         cdef bytes packed
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.glyph_to_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_packed);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":842
 *         return packed.decode()
 * 
 *     def glyph_from_string(self, string: str):             # <<<<<<<<<<<<<<
 *         cdef hb_codepoint_t gid
 *         cdef bytes packed
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_41glyph_from_string(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_41glyph_from_string = {"glyph_from_string", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_41glyph_from_string, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_41glyph_from_string(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_string = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("glyph_from_string (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_string,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_string)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 842, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "glyph_from_string") < 0)) __PYX_ERR(0, 842, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_string = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("glyph_from_string", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 842, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.glyph_from_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_string), (&PyUnicode_Type), 0, "string", 1))) __PYX_ERR(0, 842, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_40glyph_from_string(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self), __pyx_v_string);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_40glyph_from_string(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_string) {
  hb_codepoint_t __pyx_v_gid;
  PyObject *__pyx_v_packed = 0;
  hb_bool_t __pyx_v_success;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("glyph_from_string", 1);

  /* "uharfbuzz/_harfbuzz.pyx":845
 *         cdef hb_codepoint_t gid
 *         cdef bytes packed
 *         packed = string.encode()             # <<<<<<<<<<<<<<
 *         success = hb_font_glyph_from_string(self._hb_font, <char*>packed, len(packed), &gid)
 *         return gid if success else None
 */
  __pyx_t_1 = PyUnicode_AsEncodedString(__pyx_v_string, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_packed = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":846
 *         cdef bytes packed
 *         packed = string.encode()
 *         success = hb_font_glyph_from_string(self._hb_font, <char*>packed, len(packed), &gid)             # <<<<<<<<<<<<<<
 *         return gid if success else None
 * 
 */
  if (unlikely(__pyx_v_packed == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 846, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBytes_AsWritableString(__pyx_v_packed); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 846, __pyx_L1_error)
  if (unlikely(__pyx_v_packed == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 846, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyBytes_GET_SIZE(__pyx_v_packed); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 846, __pyx_L1_error)
  __pyx_v_success = hb_font_glyph_from_string(__pyx_v_self->_hb_font, ((char *)__pyx_t_2), __pyx_t_3, (&__pyx_v_gid));

  /* "uharfbuzz/_harfbuzz.pyx":847
 *         packed = string.encode()
 *         success = hb_font_glyph_from_string(self._hb_font, <char*>packed, len(packed), &gid)
 *         return gid if success else None             # <<<<<<<<<<<<<<
 * 
 *     def draw_glyph(self, gid: int, draw_funcs: DrawFuncs, draw_state: object = None):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = (__pyx_v_success != 0);
  if (__pyx_t_4) {
    __pyx_t_5 = __Pyx_PyInt_From_hb_codepoint_t(__pyx_v_gid); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 847, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = __pyx_t_5;
    __pyx_t_5 = 0;
  } else {
    __Pyx_INCREF(Py_None);
    __pyx_t_1 = Py_None;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":842
 *         return packed.decode()
 * 
 *     def glyph_from_string(self, string: str):             # <<<<<<<<<<<<<<
 *         cdef hb_codepoint_t gid
 *         cdef bytes packed
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.glyph_from_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_packed);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":849
 *         return gid if success else None
 * 
 *     def draw_glyph(self, gid: int, draw_funcs: DrawFuncs, draw_state: object = None):             # <<<<<<<<<<<<<<
 *         cdef void *draw_state_p = <void *>draw_state
 *         if PyCapsule_IsValid(draw_state, NULL):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_43draw_glyph(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_43draw_glyph = {"draw_glyph", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_43draw_glyph, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_43draw_glyph(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_gid = 0;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *__pyx_v_draw_funcs = 0;
  PyObject *__pyx_v_draw_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("draw_glyph (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_gid,&__pyx_n_s_draw_funcs,&__pyx_n_s_draw_state,0};
    values[2] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_gid)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 849, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_draw_funcs)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 849, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("draw_glyph", 0, 2, 3, 1); __PYX_ERR(0, 849, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_draw_state);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 849, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "draw_glyph") < 0)) __PYX_ERR(0, 849, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_gid = ((PyObject*)values[0]);
    __pyx_v_draw_funcs = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *)values[1]);
    __pyx_v_draw_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("draw_glyph", 0, 2, 3, __pyx_nargs); __PYX_ERR(0, 849, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.draw_glyph", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_gid), (&PyInt_Type), 0, "gid", 1))) __PYX_ERR(0, 849, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_draw_funcs), __pyx_ptype_9uharfbuzz_9_harfbuzz_DrawFuncs, 0, "draw_funcs", 0))) __PYX_ERR(0, 849, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_42draw_glyph(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self), __pyx_v_gid, __pyx_v_draw_funcs, __pyx_v_draw_state);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_42draw_glyph(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_gid, struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *__pyx_v_draw_funcs, PyObject *__pyx_v_draw_state) {
  void *__pyx_v_draw_state_p;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  void *__pyx_t_2;
  hb_codepoint_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("draw_glyph", 1);

  /* "uharfbuzz/_harfbuzz.pyx":850
 * 
 *     def draw_glyph(self, gid: int, draw_funcs: DrawFuncs, draw_state: object = None):
 *         cdef void *draw_state_p = <void *>draw_state             # <<<<<<<<<<<<<<
 *         if PyCapsule_IsValid(draw_state, NULL):
 *             draw_state_p = <void *>PyCapsule_GetPointer(draw_state, NULL)
 */
  __pyx_v_draw_state_p = ((void *)__pyx_v_draw_state);

  /* "uharfbuzz/_harfbuzz.pyx":851
 *     def draw_glyph(self, gid: int, draw_funcs: DrawFuncs, draw_state: object = None):
 *         cdef void *draw_state_p = <void *>draw_state
 *         if PyCapsule_IsValid(draw_state, NULL):             # <<<<<<<<<<<<<<
 *             draw_state_p = <void *>PyCapsule_GetPointer(draw_state, NULL)
 *         hb_font_draw_glyph(self._hb_font, gid, draw_funcs._hb_drawfuncs, draw_state_p);
 */
  __pyx_t_1 = PyCapsule_IsValid(__pyx_v_draw_state, NULL);
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":852
 *         cdef void *draw_state_p = <void *>draw_state
 *         if PyCapsule_IsValid(draw_state, NULL):
 *             draw_state_p = <void *>PyCapsule_GetPointer(draw_state, NULL)             # <<<<<<<<<<<<<<
 *         hb_font_draw_glyph(self._hb_font, gid, draw_funcs._hb_drawfuncs, draw_state_p);
 * 
 */
    __pyx_t_2 = PyCapsule_GetPointer(__pyx_v_draw_state, NULL); if (unlikely(__pyx_t_2 == ((void *)NULL) && PyErr_Occurred())) __PYX_ERR(0, 852, __pyx_L1_error)
    __pyx_v_draw_state_p = ((void *)__pyx_t_2);

    /* "uharfbuzz/_harfbuzz.pyx":851
 *     def draw_glyph(self, gid: int, draw_funcs: DrawFuncs, draw_state: object = None):
 *         cdef void *draw_state_p = <void *>draw_state
 *         if PyCapsule_IsValid(draw_state, NULL):             # <<<<<<<<<<<<<<
 *             draw_state_p = <void *>PyCapsule_GetPointer(draw_state, NULL)
 *         hb_font_draw_glyph(self._hb_font, gid, draw_funcs._hb_drawfuncs, draw_state_p);
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":853
 *         if PyCapsule_IsValid(draw_state, NULL):
 *             draw_state_p = <void *>PyCapsule_GetPointer(draw_state, NULL)
 *         hb_font_draw_glyph(self._hb_font, gid, draw_funcs._hb_drawfuncs, draw_state_p);             # <<<<<<<<<<<<<<
 * 
 *     def paint_glyph(self, gid: int,
 */
  __pyx_t_3 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_gid); if (unlikely((__pyx_t_3 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 853, __pyx_L1_error)
  hb_font_draw_glyph(__pyx_v_self->_hb_font, __pyx_t_3, __pyx_v_draw_funcs->_hb_drawfuncs, __pyx_v_draw_state_p);

  /* "uharfbuzz/_harfbuzz.pyx":849
 *         return gid if success else None
 * 
 *     def draw_glyph(self, gid: int, draw_funcs: DrawFuncs, draw_state: object = None):             # <<<<<<<<<<<<<<
 *         cdef void *draw_state_p = <void *>draw_state
 *         if PyCapsule_IsValid(draw_state, NULL):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.draw_glyph", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":855
 *         hb_font_draw_glyph(self._hb_font, gid, draw_funcs._hb_drawfuncs, draw_state_p);
 * 
 *     def paint_glyph(self, gid: int,             # <<<<<<<<<<<<<<
 *                     paint_funcs: PaintFuncs,
 *                     paint_state: object = None,
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_45paint_glyph(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_45paint_glyph = {"paint_glyph", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_45paint_glyph, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_45paint_glyph(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_gid = 0;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_paint_funcs = 0;
  PyObject *__pyx_v_paint_state = 0;
  PyObject *__pyx_v_palette_index = 0;
  PyObject *__pyx_v_foreground = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("paint_glyph (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_gid,&__pyx_n_s_paint_funcs,&__pyx_n_s_paint_state,&__pyx_n_s_palette_index,&__pyx_n_s_foreground,0};

    /* "uharfbuzz/_harfbuzz.pyx":857
 *     def paint_glyph(self, gid: int,
 *                     paint_funcs: PaintFuncs,
 *                     paint_state: object = None,             # <<<<<<<<<<<<<<
 *                     palette_index: int = 0,
 *                     foreground: Color | None = None) -> None:
 */
    values[2] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    values[3] = __Pyx_Arg_NewRef_FASTCALL(__pyx_k__8);

    /* "uharfbuzz/_harfbuzz.pyx":859
 *                     paint_state: object = None,
 *                     palette_index: int = 0,
 *                     foreground: Color | None = None) -> None:             # <<<<<<<<<<<<<<
 *         cdef void *paint_state_p = <void *>paint_state
 *         cdef hb_color_t c_foreground = 0x000000FF
 */
    values[4] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  5: values[4] = __Pyx_Arg_FASTCALL(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_gid)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 855, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_paint_funcs)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 855, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("paint_glyph", 0, 2, 5, 1); __PYX_ERR(0, 855, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_paint_state);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 855, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_palette_index);
          if (value) { values[3] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 855, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_foreground);
          if (value) { values[4] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 855, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "paint_glyph") < 0)) __PYX_ERR(0, 855, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  5: values[4] = __Pyx_Arg_FASTCALL(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_gid = ((PyObject*)values[0]);
    __pyx_v_paint_funcs = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)values[1]);
    __pyx_v_paint_state = values[2];
    __pyx_v_palette_index = ((PyObject*)values[3]);
    __pyx_v_foreground = values[4];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("paint_glyph", 0, 2, 5, __pyx_nargs); __PYX_ERR(0, 855, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.paint_glyph", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_gid), (&PyInt_Type), 0, "gid", 1))) __PYX_ERR(0, 855, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_paint_funcs), __pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs, 0, "paint_funcs", 0))) __PYX_ERR(0, 856, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_palette_index), (&PyInt_Type), 0, "palette_index", 1))) __PYX_ERR(0, 858, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_44paint_glyph(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self), __pyx_v_gid, __pyx_v_paint_funcs, __pyx_v_paint_state, __pyx_v_palette_index, __pyx_v_foreground);

  /* "uharfbuzz/_harfbuzz.pyx":855
 *         hb_font_draw_glyph(self._hb_font, gid, draw_funcs._hb_drawfuncs, draw_state_p);
 * 
 *     def paint_glyph(self, gid: int,             # <<<<<<<<<<<<<<
 *                     paint_funcs: PaintFuncs,
 *                     paint_state: object = None,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_44paint_glyph(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_gid, struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_paint_funcs, PyObject *__pyx_v_paint_state, PyObject *__pyx_v_palette_index, PyObject *__pyx_v_foreground) {
  void *__pyx_v_paint_state_p;
  hb_color_t __pyx_v_c_foreground;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  hb_color_t __pyx_t_6;
  hb_codepoint_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("paint_glyph", 1);

  /* "uharfbuzz/_harfbuzz.pyx":860
 *                     palette_index: int = 0,
 *                     foreground: Color | None = None) -> None:
 *         cdef void *paint_state_p = <void *>paint_state             # <<<<<<<<<<<<<<
 *         cdef hb_color_t c_foreground = 0x000000FF
 *         if foreground is not None:
 */
  __pyx_v_paint_state_p = ((void *)__pyx_v_paint_state);

  /* "uharfbuzz/_harfbuzz.pyx":861
 *                     foreground: Color | None = None) -> None:
 *         cdef void *paint_state_p = <void *>paint_state
 *         cdef hb_color_t c_foreground = 0x000000FF             # <<<<<<<<<<<<<<
 *         if foreground is not None:
 *             c_foreground = foreground.to_int()
 */
  __pyx_v_c_foreground = 0x000000FF;

  /* "uharfbuzz/_harfbuzz.pyx":862
 *         cdef void *paint_state_p = <void *>paint_state
 *         cdef hb_color_t c_foreground = 0x000000FF
 *         if foreground is not None:             # <<<<<<<<<<<<<<
 *             c_foreground = foreground.to_int()
 *         hb_font_paint_glyph(self._hb_font,
 */
  __pyx_t_1 = (__pyx_v_foreground != Py_None);
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":863
 *         cdef hb_color_t c_foreground = 0x000000FF
 *         if foreground is not None:
 *             c_foreground = foreground.to_int()             # <<<<<<<<<<<<<<
 *         hb_font_paint_glyph(self._hb_font,
 *                             gid,
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_foreground, __pyx_n_s_to_int); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 863, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 863, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_t_6 = __Pyx_PyInt_As_uint32_t(__pyx_t_2); if (unlikely((__pyx_t_6 == ((hb_color_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 863, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_c_foreground = __pyx_t_6;

    /* "uharfbuzz/_harfbuzz.pyx":862
 *         cdef void *paint_state_p = <void *>paint_state
 *         cdef hb_color_t c_foreground = 0x000000FF
 *         if foreground is not None:             # <<<<<<<<<<<<<<
 *             c_foreground = foreground.to_int()
 *         hb_font_paint_glyph(self._hb_font,
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":865
 *             c_foreground = foreground.to_int()
 *         hb_font_paint_glyph(self._hb_font,
 *                             gid,             # <<<<<<<<<<<<<<
 *                             paint_funcs._hb_paintfuncs,
 *                             paint_state_p,
 */
  __pyx_t_7 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_gid); if (unlikely((__pyx_t_7 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 865, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":868
 *                             paint_funcs._hb_paintfuncs,
 *                             paint_state_p,
 *                             palette_index,             # <<<<<<<<<<<<<<
 *                             c_foreground)
 * 
 */
  __pyx_t_5 = __Pyx_PyInt_As_unsigned_int(__pyx_v_palette_index); if (unlikely((__pyx_t_5 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 868, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":864
 *         if foreground is not None:
 *             c_foreground = foreground.to_int()
 *         hb_font_paint_glyph(self._hb_font,             # <<<<<<<<<<<<<<
 *                             gid,
 *                             paint_funcs._hb_paintfuncs,
 */
  hb_font_paint_glyph(__pyx_v_self->_hb_font, __pyx_t_7, __pyx_v_paint_funcs->_hb_paintfuncs, __pyx_v_paint_state_p, __pyx_t_5, __pyx_v_c_foreground);

  /* "uharfbuzz/_harfbuzz.pyx":855
 *         hb_font_draw_glyph(self._hb_font, gid, draw_funcs._hb_drawfuncs, draw_state_p);
 * 
 *     def paint_glyph(self, gid: int,             # <<<<<<<<<<<<<<
 *                     paint_funcs: PaintFuncs,
 *                     paint_state: object = None,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.paint_glyph", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":871
 *                             c_foreground)
 * 
 *     def draw_glyph_with_pen(self, gid: int, pen):             # <<<<<<<<<<<<<<
 *         global drawfuncs
 *         if drawfuncs == NULL:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_47draw_glyph_with_pen(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_47draw_glyph_with_pen = {"draw_glyph_with_pen", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_47draw_glyph_with_pen, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_47draw_glyph_with_pen(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_gid = 0;
  PyObject *__pyx_v_pen = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("draw_glyph_with_pen (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_gid,&__pyx_n_s_pen,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_gid)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 871, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pen)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 871, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("draw_glyph_with_pen", 1, 2, 2, 1); __PYX_ERR(0, 871, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "draw_glyph_with_pen") < 0)) __PYX_ERR(0, 871, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_gid = ((PyObject*)values[0]);
    __pyx_v_pen = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("draw_glyph_with_pen", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 871, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.draw_glyph_with_pen", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_gid), (&PyInt_Type), 0, "gid", 1))) __PYX_ERR(0, 871, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_46draw_glyph_with_pen(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self), __pyx_v_gid, __pyx_v_pen);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_46draw_glyph_with_pen(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, PyObject *__pyx_v_gid, PyObject *__pyx_v_pen) {
  PyObject *__pyx_v_moveTo = NULL;
  PyObject *__pyx_v_lineTo = NULL;
  PyObject *__pyx_v_curveTo = NULL;
  PyObject *__pyx_v_qCurveTo = NULL;
  PyObject *__pyx_v_closePath = NULL;
  struct __pyx_t_9uharfbuzz_9_harfbuzz__pen_methods __pyx_v_methods;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  hb_codepoint_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("draw_glyph_with_pen", 1);

  /* "uharfbuzz/_harfbuzz.pyx":873
 *     def draw_glyph_with_pen(self, gid: int, pen):
 *         global drawfuncs
 *         if drawfuncs == NULL:             # <<<<<<<<<<<<<<
 *             drawfuncs = hb_draw_funcs_create()
 *             hb_draw_funcs_set_move_to_func(drawfuncs, _pen_move_to_func, NULL, NULL)
 */
  __pyx_t_1 = (__pyx_v_9uharfbuzz_9_harfbuzz_drawfuncs == NULL);
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":874
 *         global drawfuncs
 *         if drawfuncs == NULL:
 *             drawfuncs = hb_draw_funcs_create()             # <<<<<<<<<<<<<<
 *             hb_draw_funcs_set_move_to_func(drawfuncs, _pen_move_to_func, NULL, NULL)
 *             hb_draw_funcs_set_line_to_func(drawfuncs, _pen_line_to_func, NULL, NULL)
 */
    __pyx_v_9uharfbuzz_9_harfbuzz_drawfuncs = hb_draw_funcs_create();

    /* "uharfbuzz/_harfbuzz.pyx":875
 *         if drawfuncs == NULL:
 *             drawfuncs = hb_draw_funcs_create()
 *             hb_draw_funcs_set_move_to_func(drawfuncs, _pen_move_to_func, NULL, NULL)             # <<<<<<<<<<<<<<
 *             hb_draw_funcs_set_line_to_func(drawfuncs, _pen_line_to_func, NULL, NULL)
 *             hb_draw_funcs_set_cubic_to_func(drawfuncs, _pen_cubic_to_func, NULL, NULL)
 */
    hb_draw_funcs_set_move_to_func(__pyx_v_9uharfbuzz_9_harfbuzz_drawfuncs, __pyx_f_9uharfbuzz_9_harfbuzz__pen_move_to_func, NULL, NULL);

    /* "uharfbuzz/_harfbuzz.pyx":876
 *             drawfuncs = hb_draw_funcs_create()
 *             hb_draw_funcs_set_move_to_func(drawfuncs, _pen_move_to_func, NULL, NULL)
 *             hb_draw_funcs_set_line_to_func(drawfuncs, _pen_line_to_func, NULL, NULL)             # <<<<<<<<<<<<<<
 *             hb_draw_funcs_set_cubic_to_func(drawfuncs, _pen_cubic_to_func, NULL, NULL)
 *             hb_draw_funcs_set_quadratic_to_func(drawfuncs, _pen_quadratic_to_func, NULL, NULL)
 */
    hb_draw_funcs_set_line_to_func(__pyx_v_9uharfbuzz_9_harfbuzz_drawfuncs, __pyx_f_9uharfbuzz_9_harfbuzz__pen_line_to_func, NULL, NULL);

    /* "uharfbuzz/_harfbuzz.pyx":877
 *             hb_draw_funcs_set_move_to_func(drawfuncs, _pen_move_to_func, NULL, NULL)
 *             hb_draw_funcs_set_line_to_func(drawfuncs, _pen_line_to_func, NULL, NULL)
 *             hb_draw_funcs_set_cubic_to_func(drawfuncs, _pen_cubic_to_func, NULL, NULL)             # <<<<<<<<<<<<<<
 *             hb_draw_funcs_set_quadratic_to_func(drawfuncs, _pen_quadratic_to_func, NULL, NULL)
 *             hb_draw_funcs_set_close_path_func(drawfuncs, _pen_close_path_func, NULL, NULL)
 */
    hb_draw_funcs_set_cubic_to_func(__pyx_v_9uharfbuzz_9_harfbuzz_drawfuncs, __pyx_f_9uharfbuzz_9_harfbuzz__pen_cubic_to_func, NULL, NULL);

    /* "uharfbuzz/_harfbuzz.pyx":878
 *             hb_draw_funcs_set_line_to_func(drawfuncs, _pen_line_to_func, NULL, NULL)
 *             hb_draw_funcs_set_cubic_to_func(drawfuncs, _pen_cubic_to_func, NULL, NULL)
 *             hb_draw_funcs_set_quadratic_to_func(drawfuncs, _pen_quadratic_to_func, NULL, NULL)             # <<<<<<<<<<<<<<
 *             hb_draw_funcs_set_close_path_func(drawfuncs, _pen_close_path_func, NULL, NULL)
 * 
 */
    hb_draw_funcs_set_quadratic_to_func(__pyx_v_9uharfbuzz_9_harfbuzz_drawfuncs, __pyx_f_9uharfbuzz_9_harfbuzz__pen_quadratic_to_func, NULL, NULL);

    /* "uharfbuzz/_harfbuzz.pyx":879
 *             hb_draw_funcs_set_cubic_to_func(drawfuncs, _pen_cubic_to_func, NULL, NULL)
 *             hb_draw_funcs_set_quadratic_to_func(drawfuncs, _pen_quadratic_to_func, NULL, NULL)
 *             hb_draw_funcs_set_close_path_func(drawfuncs, _pen_close_path_func, NULL, NULL)             # <<<<<<<<<<<<<<
 * 
 *         # Keep local copy so they are not GC'ed before the call completes
 */
    hb_draw_funcs_set_close_path_func(__pyx_v_9uharfbuzz_9_harfbuzz_drawfuncs, __pyx_f_9uharfbuzz_9_harfbuzz__pen_close_path_func, NULL, NULL);

    /* "uharfbuzz/_harfbuzz.pyx":873
 *     def draw_glyph_with_pen(self, gid: int, pen):
 *         global drawfuncs
 *         if drawfuncs == NULL:             # <<<<<<<<<<<<<<
 *             drawfuncs = hb_draw_funcs_create()
 *             hb_draw_funcs_set_move_to_func(drawfuncs, _pen_move_to_func, NULL, NULL)
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":882
 * 
 *         # Keep local copy so they are not GC'ed before the call completes
 *         moveTo = pen.moveTo             # <<<<<<<<<<<<<<
 *         lineTo = pen.lineTo
 *         curveTo = pen.curveTo
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_pen, __pyx_n_s_moveTo); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 882, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_moveTo = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":883
 *         # Keep local copy so they are not GC'ed before the call completes
 *         moveTo = pen.moveTo
 *         lineTo = pen.lineTo             # <<<<<<<<<<<<<<
 *         curveTo = pen.curveTo
 *         qCurveTo = pen.qCurveTo
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_pen, __pyx_n_s_lineTo); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 883, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_lineTo = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":884
 *         moveTo = pen.moveTo
 *         lineTo = pen.lineTo
 *         curveTo = pen.curveTo             # <<<<<<<<<<<<<<
 *         qCurveTo = pen.qCurveTo
 *         closePath = pen.closePath
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_pen, __pyx_n_s_curveTo); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_curveTo = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":885
 *         lineTo = pen.lineTo
 *         curveTo = pen.curveTo
 *         qCurveTo = pen.qCurveTo             # <<<<<<<<<<<<<<
 *         closePath = pen.closePath
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_pen, __pyx_n_s_qCurveTo); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_qCurveTo = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":886
 *         curveTo = pen.curveTo
 *         qCurveTo = pen.qCurveTo
 *         closePath = pen.closePath             # <<<<<<<<<<<<<<
 * 
 *         cdef _pen_methods methods
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_pen, __pyx_n_s_closePath); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 886, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_closePath = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":889
 * 
 *         cdef _pen_methods methods
 *         methods.moveTo = <void*>moveTo             # <<<<<<<<<<<<<<
 *         methods.lineTo = <void*>lineTo
 *         methods.curveTo = <void*>curveTo
 */
  __pyx_v_methods.moveTo = ((void *)__pyx_v_moveTo);

  /* "uharfbuzz/_harfbuzz.pyx":890
 *         cdef _pen_methods methods
 *         methods.moveTo = <void*>moveTo
 *         methods.lineTo = <void*>lineTo             # <<<<<<<<<<<<<<
 *         methods.curveTo = <void*>curveTo
 *         methods.qCurveTo = <void*>qCurveTo
 */
  __pyx_v_methods.lineTo = ((void *)__pyx_v_lineTo);

  /* "uharfbuzz/_harfbuzz.pyx":891
 *         methods.moveTo = <void*>moveTo
 *         methods.lineTo = <void*>lineTo
 *         methods.curveTo = <void*>curveTo             # <<<<<<<<<<<<<<
 *         methods.qCurveTo = <void*>qCurveTo
 *         methods.closePath = <void*>closePath
 */
  __pyx_v_methods.curveTo = ((void *)__pyx_v_curveTo);

  /* "uharfbuzz/_harfbuzz.pyx":892
 *         methods.lineTo = <void*>lineTo
 *         methods.curveTo = <void*>curveTo
 *         methods.qCurveTo = <void*>qCurveTo             # <<<<<<<<<<<<<<
 *         methods.closePath = <void*>closePath
 * 
 */
  __pyx_v_methods.qCurveTo = ((void *)__pyx_v_qCurveTo);

  /* "uharfbuzz/_harfbuzz.pyx":893
 *         methods.curveTo = <void*>curveTo
 *         methods.qCurveTo = <void*>qCurveTo
 *         methods.closePath = <void*>closePath             # <<<<<<<<<<<<<<
 * 
 *         hb_font_draw_glyph(self._hb_font, gid, drawfuncs, <void*>&methods)
 */
  __pyx_v_methods.closePath = ((void *)__pyx_v_closePath);

  /* "uharfbuzz/_harfbuzz.pyx":895
 *         methods.closePath = <void*>closePath
 * 
 *         hb_font_draw_glyph(self._hb_font, gid, drawfuncs, <void*>&methods)             # <<<<<<<<<<<<<<
 * 
 * cdef struct _pen_methods:
 */
  __pyx_t_3 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_gid); if (unlikely((__pyx_t_3 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 895, __pyx_L1_error)
  hb_font_draw_glyph(__pyx_v_self->_hb_font, __pyx_t_3, __pyx_v_9uharfbuzz_9_harfbuzz_drawfuncs, ((void *)(&__pyx_v_methods)));

  /* "uharfbuzz/_harfbuzz.pyx":871
 *                             c_foreground)
 * 
 *     def draw_glyph_with_pen(self, gid: int, pen):             # <<<<<<<<<<<<<<
 *         global drawfuncs
 *         if drawfuncs == NULL:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.draw_glyph_with_pen", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_moveTo);
  __Pyx_XDECREF(__pyx_v_lineTo);
  __Pyx_XDECREF(__pyx_v_curveTo);
  __Pyx_XDECREF(__pyx_v_qCurveTo);
  __Pyx_XDECREF(__pyx_v_closePath);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_49__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_49__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_49__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_49__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_48__reduce_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_48__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_51__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_51__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_51__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_51__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4Font_50__setstate_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4Font_50__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Font.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":906
 * cdef hb_draw_funcs_t* drawfuncs = NULL
 * 
 * cdef void _pen_move_to_func(hb_draw_funcs_t *dfuncs,             # <<<<<<<<<<<<<<
 *                             void *draw_data,
 *                             hb_draw_state_t *st,
 */

static void __pyx_f_9uharfbuzz_9_harfbuzz__pen_move_to_func(CYTHON_UNUSED hb_draw_funcs_t *__pyx_v_dfuncs, void *__pyx_v_draw_data, CYTHON_UNUSED hb_draw_state_t *__pyx_v_st, float __pyx_v_to_x, float __pyx_v_to_y, CYTHON_UNUSED void *__pyx_v_user_data) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_pen_move_to_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":912
 *                             float to_y,
 *                             void *user_data) noexcept:
 *     (<object>((<_pen_methods*>draw_data).moveTo))((to_x, to_y))             # <<<<<<<<<<<<<<
 * 
 * cdef void _pen_line_to_func(hb_draw_funcs_t *dfuncs,
 */
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_to_x); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_to_y); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2)) __PYX_ERR(0, 912, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3)) __PYX_ERR(0, 912, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __Pyx_INCREF(((PyObject *)((struct __pyx_t_9uharfbuzz_9_harfbuzz__pen_methods *)__pyx_v_draw_data)->moveTo));
  __pyx_t_3 = ((PyObject *)((struct __pyx_t_9uharfbuzz_9_harfbuzz__pen_methods *)__pyx_v_draw_data)->moveTo); __pyx_t_2 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 912, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":906
 * cdef hb_draw_funcs_t* drawfuncs = NULL
 * 
 * cdef void _pen_move_to_func(hb_draw_funcs_t *dfuncs,             # <<<<<<<<<<<<<<
 *                             void *draw_data,
 *                             hb_draw_state_t *st,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz._pen_move_to_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "uharfbuzz/_harfbuzz.pyx":914
 *     (<object>((<_pen_methods*>draw_data).moveTo))((to_x, to_y))
 * 
 * cdef void _pen_line_to_func(hb_draw_funcs_t *dfuncs,             # <<<<<<<<<<<<<<
 *                             void *draw_data,
 *                             hb_draw_state_t *st,
 */

static void __pyx_f_9uharfbuzz_9_harfbuzz__pen_line_to_func(CYTHON_UNUSED hb_draw_funcs_t *__pyx_v_dfuncs, void *__pyx_v_draw_data, CYTHON_UNUSED hb_draw_state_t *__pyx_v_st, float __pyx_v_to_x, float __pyx_v_to_y, CYTHON_UNUSED void *__pyx_v_user_data) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_pen_line_to_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":920
 *                             float to_y,
 *                             void *user_data) noexcept:
 *     (<object>((<_pen_methods*>draw_data).lineTo))((to_x, to_y))             # <<<<<<<<<<<<<<
 * 
 * cdef void _pen_close_path_func(hb_draw_funcs_t *dfuncs,
 */
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_to_x); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_to_y); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2)) __PYX_ERR(0, 920, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3)) __PYX_ERR(0, 920, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __Pyx_INCREF(((PyObject *)((struct __pyx_t_9uharfbuzz_9_harfbuzz__pen_methods *)__pyx_v_draw_data)->lineTo));
  __pyx_t_3 = ((PyObject *)((struct __pyx_t_9uharfbuzz_9_harfbuzz__pen_methods *)__pyx_v_draw_data)->lineTo); __pyx_t_2 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 920, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":914
 *     (<object>((<_pen_methods*>draw_data).moveTo))((to_x, to_y))
 * 
 * cdef void _pen_line_to_func(hb_draw_funcs_t *dfuncs,             # <<<<<<<<<<<<<<
 *                             void *draw_data,
 *                             hb_draw_state_t *st,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz._pen_line_to_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "uharfbuzz/_harfbuzz.pyx":922
 *     (<object>((<_pen_methods*>draw_data).lineTo))((to_x, to_y))
 * 
 * cdef void _pen_close_path_func(hb_draw_funcs_t *dfuncs,             # <<<<<<<<<<<<<<
 *                                void *draw_data,
 *                                hb_draw_state_t *st,
 */

static void __pyx_f_9uharfbuzz_9_harfbuzz__pen_close_path_func(CYTHON_UNUSED hb_draw_funcs_t *__pyx_v_dfuncs, void *__pyx_v_draw_data, CYTHON_UNUSED hb_draw_state_t *__pyx_v_st, CYTHON_UNUSED void *__pyx_v_user_data) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_pen_close_path_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":926
 *                                hb_draw_state_t *st,
 *                                void *user_data) noexcept:
 *     (<object>((<_pen_methods*>draw_data).closePath))()             # <<<<<<<<<<<<<<
 * 
 * cdef void _pen_quadratic_to_func(hb_draw_funcs_t *dfuncs,
 */
  __Pyx_INCREF(((PyObject *)((struct __pyx_t_9uharfbuzz_9_harfbuzz__pen_methods *)__pyx_v_draw_data)->closePath));
  __pyx_t_2 = ((PyObject *)((struct __pyx_t_9uharfbuzz_9_harfbuzz__pen_methods *)__pyx_v_draw_data)->closePath); __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 926, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":922
 *     (<object>((<_pen_methods*>draw_data).lineTo))((to_x, to_y))
 * 
 * cdef void _pen_close_path_func(hb_draw_funcs_t *dfuncs,             # <<<<<<<<<<<<<<
 *                                void *draw_data,
 *                                hb_draw_state_t *st,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz._pen_close_path_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "uharfbuzz/_harfbuzz.pyx":928
 *     (<object>((<_pen_methods*>draw_data).closePath))()
 * 
 * cdef void _pen_quadratic_to_func(hb_draw_funcs_t *dfuncs,             # <<<<<<<<<<<<<<
 *                                  void *draw_data,
 *                                  hb_draw_state_t *st,
 */

static void __pyx_f_9uharfbuzz_9_harfbuzz__pen_quadratic_to_func(CYTHON_UNUSED hb_draw_funcs_t *__pyx_v_dfuncs, void *__pyx_v_draw_data, CYTHON_UNUSED hb_draw_state_t *__pyx_v_st, float __pyx_v_c1_x, float __pyx_v_c1_y, float __pyx_v_to_x, float __pyx_v_to_y, CYTHON_UNUSED void *__pyx_v_user_data) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_pen_quadratic_to_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":936
 *                                  float to_y,
 *                                  void *user_data) noexcept:
 *     (<object>((<_pen_methods*>draw_data).qCurveTo))((c1_x, c1_y), (to_x, to_y))             # <<<<<<<<<<<<<<
 * 
 * cdef void _pen_cubic_to_func(hb_draw_funcs_t *dfuncs,
 */
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_c1_x); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_c1_y); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2)) __PYX_ERR(0, 936, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3)) __PYX_ERR(0, 936, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_to_x); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_to_y); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(0, 936, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2)) __PYX_ERR(0, 936, __pyx_L1_error);
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __Pyx_INCREF(((PyObject *)((struct __pyx_t_9uharfbuzz_9_harfbuzz__pen_methods *)__pyx_v_draw_data)->qCurveTo));
  __pyx_t_2 = ((PyObject *)((struct __pyx_t_9uharfbuzz_9_harfbuzz__pen_methods *)__pyx_v_draw_data)->qCurveTo); __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_t_4, __pyx_t_5};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 936, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":928
 *     (<object>((<_pen_methods*>draw_data).closePath))()
 * 
 * cdef void _pen_quadratic_to_func(hb_draw_funcs_t *dfuncs,             # <<<<<<<<<<<<<<
 *                                  void *draw_data,
 *                                  hb_draw_state_t *st,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz._pen_quadratic_to_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "uharfbuzz/_harfbuzz.pyx":938
 *     (<object>((<_pen_methods*>draw_data).qCurveTo))((c1_x, c1_y), (to_x, to_y))
 * 
 * cdef void _pen_cubic_to_func(hb_draw_funcs_t *dfuncs,             # <<<<<<<<<<<<<<
 *                              void *draw_data,
 *                              hb_draw_state_t *st,
 */

static void __pyx_f_9uharfbuzz_9_harfbuzz__pen_cubic_to_func(CYTHON_UNUSED hb_draw_funcs_t *__pyx_v_dfuncs, void *__pyx_v_draw_data, CYTHON_UNUSED hb_draw_state_t *__pyx_v_st, float __pyx_v_c1_x, float __pyx_v_c1_y, float __pyx_v_c2_x, float __pyx_v_c2_y, float __pyx_v_to_x, float __pyx_v_to_y, CYTHON_UNUSED void *__pyx_v_user_data) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_pen_cubic_to_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":948
 *                              float to_y,
 *                              void *user_data) noexcept:
 *     (<object>((<_pen_methods*>draw_data).curveTo))((c1_x, c1_y), (c2_x, c2_y), (to_x, to_y))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_c1_x); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 948, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_c1_y); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 948, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 948, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2)) __PYX_ERR(0, 948, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3)) __PYX_ERR(0, 948, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_c2_x); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 948, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_c2_y); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 948, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 948, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(0, 948, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2)) __PYX_ERR(0, 948, __pyx_L1_error);
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_to_x); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 948, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_to_y); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 948, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 948, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2)) __PYX_ERR(0, 948, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_3)) __PYX_ERR(0, 948, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __Pyx_INCREF(((PyObject *)((struct __pyx_t_9uharfbuzz_9_harfbuzz__pen_methods *)__pyx_v_draw_data)->curveTo));
  __pyx_t_3 = ((PyObject *)((struct __pyx_t_9uharfbuzz_9_harfbuzz__pen_methods *)__pyx_v_draw_data)->curveTo); __pyx_t_2 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_2, __pyx_t_4, __pyx_t_5, __pyx_t_6};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_7, 3+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 948, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":938
 *     (<object>((<_pen_methods*>draw_data).qCurveTo))((c1_x, c1_y), (to_x, to_y))
 * 
 * cdef void _pen_cubic_to_func(hb_draw_funcs_t *dfuncs,             # <<<<<<<<<<<<<<
 *                              void *draw_data,
 *                              hb_draw_state_t *st,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz._pen_cubic_to_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "uharfbuzz/_harfbuzz.pyx":951
 * 
 * 
 * cdef hb_position_t _glyph_h_advance_func(hb_font_t* font, void* font_data,             # <<<<<<<<<<<<<<
 *                                          hb_codepoint_t glyph,
 *                                          void* user_data) noexcept:
 */

static hb_position_t __pyx_f_9uharfbuzz_9_harfbuzz__glyph_h_advance_func(CYTHON_UNUSED hb_font_t *__pyx_v_font, void *__pyx_v_font_data, hb_codepoint_t __pyx_v_glyph, void *__pyx_v_user_data) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_py_font = 0;
  hb_position_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  hb_position_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_glyph_h_advance_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":954
 *                                          hb_codepoint_t glyph,
 *                                          void* user_data) noexcept:
 *     cdef Font py_font = <Font>font_data             # <<<<<<<<<<<<<<
 *     return (<FontFuncs>py_font.funcs)._glyph_h_advance_func(
 *         py_font, glyph, <object>user_data)
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_font_data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_py_font = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":955
 *                                          void* user_data) noexcept:
 *     cdef Font py_font = <Font>font_data
 *     return (<FontFuncs>py_font.funcs)._glyph_h_advance_func(             # <<<<<<<<<<<<<<
 *         py_font, glyph, <object>user_data)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_py_font), __pyx_n_s_funcs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "uharfbuzz/_harfbuzz.pyx":956
 *     cdef Font py_font = <Font>font_data
 *     return (<FontFuncs>py_font.funcs)._glyph_h_advance_func(
 *         py_font, glyph, <object>user_data)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_codepoint_t(__pyx_v_glyph); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 956, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)__pyx_t_2)->_glyph_h_advance_func);
  __pyx_t_4 = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)__pyx_t_2)->_glyph_h_advance_func; __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_2, ((PyObject *)__pyx_v_py_font), __pyx_t_3, ((PyObject *)__pyx_v_user_data)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 3+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 955, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }

  /* "uharfbuzz/_harfbuzz.pyx":955
 *                                          void* user_data) noexcept:
 *     cdef Font py_font = <Font>font_data
 *     return (<FontFuncs>py_font.funcs)._glyph_h_advance_func(             # <<<<<<<<<<<<<<
 *         py_font, glyph, <object>user_data)
 * 
 */
  __pyx_t_6 = __Pyx_PyInt_As_hb_position_t(__pyx_t_1); if (unlikely((__pyx_t_6 == ((hb_position_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 955, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_6;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":951
 * 
 * 
 * cdef hb_position_t _glyph_h_advance_func(hb_font_t* font, void* font_data,             # <<<<<<<<<<<<<<
 *                                          hb_codepoint_t glyph,
 *                                          void* user_data) noexcept:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz._glyph_h_advance_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_py_font);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":959
 * 
 * 
 * cdef hb_position_t _glyph_v_advance_func(hb_font_t* font, void* font_data,             # <<<<<<<<<<<<<<
 *                                          hb_codepoint_t glyph,
 *                                          void* user_data) noexcept:
 */

static hb_position_t __pyx_f_9uharfbuzz_9_harfbuzz__glyph_v_advance_func(CYTHON_UNUSED hb_font_t *__pyx_v_font, void *__pyx_v_font_data, hb_codepoint_t __pyx_v_glyph, void *__pyx_v_user_data) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_py_font = 0;
  hb_position_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  hb_position_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_glyph_v_advance_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":962
 *                                          hb_codepoint_t glyph,
 *                                          void* user_data) noexcept:
 *     cdef Font py_font = <Font>font_data             # <<<<<<<<<<<<<<
 *     return (<FontFuncs>py_font.funcs)._glyph_v_advance_func(
 *         py_font, glyph, <object>user_data)
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_font_data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_py_font = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":963
 *                                          void* user_data) noexcept:
 *     cdef Font py_font = <Font>font_data
 *     return (<FontFuncs>py_font.funcs)._glyph_v_advance_func(             # <<<<<<<<<<<<<<
 *         py_font, glyph, <object>user_data)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_py_font), __pyx_n_s_funcs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 963, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "uharfbuzz/_harfbuzz.pyx":964
 *     cdef Font py_font = <Font>font_data
 *     return (<FontFuncs>py_font.funcs)._glyph_v_advance_func(
 *         py_font, glyph, <object>user_data)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_codepoint_t(__pyx_v_glyph); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 964, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)__pyx_t_2)->_glyph_v_advance_func);
  __pyx_t_4 = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)__pyx_t_2)->_glyph_v_advance_func; __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_2, ((PyObject *)__pyx_v_py_font), __pyx_t_3, ((PyObject *)__pyx_v_user_data)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 3+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 963, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }

  /* "uharfbuzz/_harfbuzz.pyx":963
 *                                          void* user_data) noexcept:
 *     cdef Font py_font = <Font>font_data
 *     return (<FontFuncs>py_font.funcs)._glyph_v_advance_func(             # <<<<<<<<<<<<<<
 *         py_font, glyph, <object>user_data)
 * 
 */
  __pyx_t_6 = __Pyx_PyInt_As_hb_position_t(__pyx_t_1); if (unlikely((__pyx_t_6 == ((hb_position_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 963, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_6;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":959
 * 
 * 
 * cdef hb_position_t _glyph_v_advance_func(hb_font_t* font, void* font_data,             # <<<<<<<<<<<<<<
 *                                          hb_codepoint_t glyph,
 *                                          void* user_data) noexcept:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz._glyph_v_advance_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_py_font);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":967
 * 
 * 
 * cdef hb_bool_t _glyph_v_origin_func(hb_font_t* font, void* font_data,             # <<<<<<<<<<<<<<
 *                                     hb_codepoint_t glyph,
 *                                     hb_position_t* x, hb_position_t* y,
 */

static hb_bool_t __pyx_f_9uharfbuzz_9_harfbuzz__glyph_v_origin_func(CYTHON_UNUSED hb_font_t *__pyx_v_font, void *__pyx_v_font_data, hb_codepoint_t __pyx_v_glyph, hb_position_t *__pyx_v_x, hb_position_t *__pyx_v_y, void *__pyx_v_user_data) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_py_font = 0;
  hb_bool_t __pyx_v_success;
  hb_position_t __pyx_v_px;
  hb_position_t __pyx_v_py;
  hb_bool_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *(*__pyx_t_7)(PyObject *);
  hb_bool_t __pyx_t_8;
  hb_position_t __pyx_t_9;
  hb_position_t __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_glyph_v_origin_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":971
 *                                     hb_position_t* x, hb_position_t* y,
 *                                     void* user_data) noexcept:
 *     cdef Font py_font = <Font>font_data             # <<<<<<<<<<<<<<
 *     cdef hb_bool_t success
 *     cdef hb_position_t px
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_font_data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_py_font = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":975
 *     cdef hb_position_t px
 *     cdef hb_position_t py
 *     success, px, py = (<FontFuncs>py_font.funcs)._glyph_v_origin_func(             # <<<<<<<<<<<<<<
 *         py_font, glyph, <object>user_data)
 *     x[0] = px
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_py_font), __pyx_n_s_funcs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 975, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "uharfbuzz/_harfbuzz.pyx":976
 *     cdef hb_position_t py
 *     success, px, py = (<FontFuncs>py_font.funcs)._glyph_v_origin_func(
 *         py_font, glyph, <object>user_data)             # <<<<<<<<<<<<<<
 *     x[0] = px
 *     y[0] = py
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_codepoint_t(__pyx_v_glyph); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 976, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)__pyx_t_2)->_glyph_v_origin_func);
  __pyx_t_4 = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)__pyx_t_2)->_glyph_v_origin_func; __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_2, ((PyObject *)__pyx_v_py_font), __pyx_t_3, ((PyObject *)__pyx_v_user_data)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 3+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 975, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 3)) {
      if (size > 3) __Pyx_RaiseTooManyValuesError(3);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 975, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 2); 
    } else {
      __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_2 = PyList_GET_ITEM(sequence, 2); 
    }
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_2);
    #else
    __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 975, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 975, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 975, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_6 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 975, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_6);
    index = 0; __pyx_t_4 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_4)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_4);
    index = 1; __pyx_t_3 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    index = 2; __pyx_t_2 = __pyx_t_7(__pyx_t_6); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_6), 3) < 0) __PYX_ERR(0, 975, __pyx_L1_error)
    __pyx_t_7 = NULL;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 975, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }

  /* "uharfbuzz/_harfbuzz.pyx":975
 *     cdef hb_position_t px
 *     cdef hb_position_t py
 *     success, px, py = (<FontFuncs>py_font.funcs)._glyph_v_origin_func(             # <<<<<<<<<<<<<<
 *         py_font, glyph, <object>user_data)
 *     x[0] = px
 */
  __pyx_t_8 = __Pyx_PyInt_As_hb_bool_t(__pyx_t_4); if (unlikely((__pyx_t_8 == ((hb_bool_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 975, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_9 = __Pyx_PyInt_As_hb_position_t(__pyx_t_3); if (unlikely((__pyx_t_9 == ((hb_position_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 975, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_10 = __Pyx_PyInt_As_hb_position_t(__pyx_t_2); if (unlikely((__pyx_t_10 == ((hb_position_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 975, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_success = __pyx_t_8;
  __pyx_v_px = __pyx_t_9;
  __pyx_v_py = __pyx_t_10;

  /* "uharfbuzz/_harfbuzz.pyx":977
 *     success, px, py = (<FontFuncs>py_font.funcs)._glyph_v_origin_func(
 *         py_font, glyph, <object>user_data)
 *     x[0] = px             # <<<<<<<<<<<<<<
 *     y[0] = py
 *     return success
 */
  (__pyx_v_x[0]) = __pyx_v_px;

  /* "uharfbuzz/_harfbuzz.pyx":978
 *         py_font, glyph, <object>user_data)
 *     x[0] = px
 *     y[0] = py             # <<<<<<<<<<<<<<
 *     return success
 * 
 */
  (__pyx_v_y[0]) = __pyx_v_py;

  /* "uharfbuzz/_harfbuzz.pyx":979
 *     x[0] = px
 *     y[0] = py
 *     return success             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_success;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":967
 * 
 * 
 * cdef hb_bool_t _glyph_v_origin_func(hb_font_t* font, void* font_data,             # <<<<<<<<<<<<<<
 *                                     hb_codepoint_t glyph,
 *                                     hb_position_t* x, hb_position_t* y,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz._glyph_v_origin_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_py_font);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":982
 * 
 * 
 * cdef hb_bool_t _glyph_name_func(hb_font_t *font, void *font_data,             # <<<<<<<<<<<<<<
 *                                 hb_codepoint_t glyph,
 *                                 char *name, unsigned int size,
 */

static hb_bool_t __pyx_f_9uharfbuzz_9_harfbuzz__glyph_name_func(CYTHON_UNUSED hb_font_t *__pyx_v_font, void *__pyx_v_font_data, hb_codepoint_t __pyx_v_glyph, char *__pyx_v_name, CYTHON_UNUSED unsigned int __pyx_v_size, void *__pyx_v_user_data) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_py_font = 0;
  PyObject *__pyx_v_ret = 0;
  hb_bool_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  char __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_glyph_name_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":986
 *                                 char *name, unsigned int size,
 *                                 void *user_data) noexcept:
 *     cdef Font py_font = <Font>font_data             # <<<<<<<<<<<<<<
 *     cdef bytes ret = (<FontFuncs>py_font.funcs)._glyph_name_func(
 *         py_font, glyph, <object>user_data).encode()
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_font_data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_py_font = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":987
 *                                 void *user_data) noexcept:
 *     cdef Font py_font = <Font>font_data
 *     cdef bytes ret = (<FontFuncs>py_font.funcs)._glyph_name_func(             # <<<<<<<<<<<<<<
 *         py_font, glyph, <object>user_data).encode()
 *     name[0] = ret
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_py_font), __pyx_n_s_funcs); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 987, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "uharfbuzz/_harfbuzz.pyx":988
 *     cdef Font py_font = <Font>font_data
 *     cdef bytes ret = (<FontFuncs>py_font.funcs)._glyph_name_func(
 *         py_font, glyph, <object>user_data).encode()             # <<<<<<<<<<<<<<
 *     name[0] = ret
 *     return 1
 */
  __pyx_t_4 = __Pyx_PyInt_From_hb_codepoint_t(__pyx_v_glyph); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 988, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)__pyx_t_3)->_glyph_name_func);
  __pyx_t_5 = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)__pyx_t_3)->_glyph_name_func; __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, ((PyObject *)__pyx_v_py_font), __pyx_t_4, ((PyObject *)__pyx_v_user_data)};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 987, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_encode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 988, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 988, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_1))) __PYX_ERR(0, 988, __pyx_L1_error)
  __pyx_v_ret = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":989
 *     cdef bytes ret = (<FontFuncs>py_font.funcs)._glyph_name_func(
 *         py_font, glyph, <object>user_data).encode()
 *     name[0] = ret             # <<<<<<<<<<<<<<
 *     return 1
 * 
 */
  __pyx_t_7 = __Pyx_PyInt_As_char(__pyx_v_ret); if (unlikely((__pyx_t_7 == (char)-1) && PyErr_Occurred())) __PYX_ERR(0, 989, __pyx_L1_error)
  (__pyx_v_name[0]) = __pyx_t_7;

  /* "uharfbuzz/_harfbuzz.pyx":990
 *         py_font, glyph, <object>user_data).encode()
 *     name[0] = ret
 *     return 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":982
 * 
 * 
 * cdef hb_bool_t _glyph_name_func(hb_font_t *font, void *font_data,             # <<<<<<<<<<<<<<
 *                                 hb_codepoint_t glyph,
 *                                 char *name, unsigned int size,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz._glyph_name_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_py_font);
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":993
 * 
 * 
 * cdef hb_bool_t _nominal_glyph_func(hb_font_t* font, void* font_data,             # <<<<<<<<<<<<<<
 *                                    hb_codepoint_t unicode,
 *                                    hb_codepoint_t* glyph,
 */

static hb_bool_t __pyx_f_9uharfbuzz_9_harfbuzz__nominal_glyph_func(CYTHON_UNUSED hb_font_t *__pyx_v_font, void *__pyx_v_font_data, hb_codepoint_t __pyx_v_unicode, hb_codepoint_t *__pyx_v_glyph, void *__pyx_v_user_data) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_py_font = 0;
  hb_bool_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  hb_codepoint_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_nominal_glyph_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":997
 *                                    hb_codepoint_t* glyph,
 *                                    void* user_data) noexcept:
 *     cdef Font py_font = <Font>font_data             # <<<<<<<<<<<<<<
 *     glyph[0] = (<FontFuncs>py_font.funcs)._nominal_glyph_func(
 *         py_font, unicode, <object>user_data)
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_font_data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_py_font = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":998
 *                                    void* user_data) noexcept:
 *     cdef Font py_font = <Font>font_data
 *     glyph[0] = (<FontFuncs>py_font.funcs)._nominal_glyph_func(             # <<<<<<<<<<<<<<
 *         py_font, unicode, <object>user_data)
 *     # If the glyph is .notdef, return false, else return true
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_py_font), __pyx_n_s_funcs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 998, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "uharfbuzz/_harfbuzz.pyx":999
 *     cdef Font py_font = <Font>font_data
 *     glyph[0] = (<FontFuncs>py_font.funcs)._nominal_glyph_func(
 *         py_font, unicode, <object>user_data)             # <<<<<<<<<<<<<<
 *     # If the glyph is .notdef, return false, else return true
 *     return int(glyph[0] != 0)
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_codepoint_t(__pyx_v_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 999, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)__pyx_t_2)->_nominal_glyph_func);
  __pyx_t_4 = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)__pyx_t_2)->_nominal_glyph_func; __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_2, ((PyObject *)__pyx_v_py_font), __pyx_t_3, ((PyObject *)__pyx_v_user_data)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 3+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 998, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }

  /* "uharfbuzz/_harfbuzz.pyx":998
 *                                    void* user_data) noexcept:
 *     cdef Font py_font = <Font>font_data
 *     glyph[0] = (<FontFuncs>py_font.funcs)._nominal_glyph_func(             # <<<<<<<<<<<<<<
 *         py_font, unicode, <object>user_data)
 *     # If the glyph is .notdef, return false, else return true
 */
  __pyx_t_6 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_t_1); if (unlikely((__pyx_t_6 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 998, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  (__pyx_v_glyph[0]) = __pyx_t_6;

  /* "uharfbuzz/_harfbuzz.pyx":1001
 *         py_font, unicode, <object>user_data)
 *     # If the glyph is .notdef, return false, else return true
 *     return int(glyph[0] != 0)             # <<<<<<<<<<<<<<
 * 
 * cdef hb_bool_t _variation_glyph_func(hb_font_t* font, void* font_data,
 */
  __pyx_r = ((hb_bool_t)((__pyx_v_glyph[0]) != 0));
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":993
 * 
 * 
 * cdef hb_bool_t _nominal_glyph_func(hb_font_t* font, void* font_data,             # <<<<<<<<<<<<<<
 *                                    hb_codepoint_t unicode,
 *                                    hb_codepoint_t* glyph,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz._nominal_glyph_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_py_font);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1003
 *     return int(glyph[0] != 0)
 * 
 * cdef hb_bool_t _variation_glyph_func(hb_font_t* font, void* font_data,             # <<<<<<<<<<<<<<
 *                                    hb_codepoint_t unicode,
 *                                    hb_codepoint_t variation_selector,
 */

static hb_bool_t __pyx_f_9uharfbuzz_9_harfbuzz__variation_glyph_func(CYTHON_UNUSED hb_font_t *__pyx_v_font, void *__pyx_v_font_data, hb_codepoint_t __pyx_v_unicode, hb_codepoint_t __pyx_v_variation_selector, hb_codepoint_t *__pyx_v_glyph, void *__pyx_v_user_data) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_py_font = 0;
  hb_bool_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  hb_codepoint_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_variation_glyph_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1008
 *                                    hb_codepoint_t* glyph,
 *                                    void* user_data) noexcept:
 *     cdef Font py_font = <Font>font_data             # <<<<<<<<<<<<<<
 *     glyph[0] = (<FontFuncs>py_font.funcs)._variation_glyph_func(
 *         py_font, unicode, variation_selector, <object>user_data)
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_font_data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_py_font = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1009
 *                                    void* user_data) noexcept:
 *     cdef Font py_font = <Font>font_data
 *     glyph[0] = (<FontFuncs>py_font.funcs)._variation_glyph_func(             # <<<<<<<<<<<<<<
 *         py_font, unicode, variation_selector, <object>user_data)
 *     # If the glyph is .notdef, return false, else return true
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_py_font), __pyx_n_s_funcs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1009, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "uharfbuzz/_harfbuzz.pyx":1010
 *     cdef Font py_font = <Font>font_data
 *     glyph[0] = (<FontFuncs>py_font.funcs)._variation_glyph_func(
 *         py_font, unicode, variation_selector, <object>user_data)             # <<<<<<<<<<<<<<
 *     # If the glyph is .notdef, return false, else return true
 *     return int(glyph[0] != 0)
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_codepoint_t(__pyx_v_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1010, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_hb_codepoint_t(__pyx_v_variation_selector); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1010, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)__pyx_t_2)->_variation_glyph_func);
  __pyx_t_5 = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)__pyx_t_2)->_variation_glyph_func; __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[5] = {__pyx_t_2, ((PyObject *)__pyx_v_py_font), __pyx_t_3, __pyx_t_4, ((PyObject *)__pyx_v_user_data)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 4+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1009, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }

  /* "uharfbuzz/_harfbuzz.pyx":1009
 *                                    void* user_data) noexcept:
 *     cdef Font py_font = <Font>font_data
 *     glyph[0] = (<FontFuncs>py_font.funcs)._variation_glyph_func(             # <<<<<<<<<<<<<<
 *         py_font, unicode, variation_selector, <object>user_data)
 *     # If the glyph is .notdef, return false, else return true
 */
  __pyx_t_7 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_t_1); if (unlikely((__pyx_t_7 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1009, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  (__pyx_v_glyph[0]) = __pyx_t_7;

  /* "uharfbuzz/_harfbuzz.pyx":1012
 *         py_font, unicode, variation_selector, <object>user_data)
 *     # If the glyph is .notdef, return false, else return true
 *     return int(glyph[0] != 0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((hb_bool_t)((__pyx_v_glyph[0]) != 0));
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1003
 *     return int(glyph[0] != 0)
 * 
 * cdef hb_bool_t _variation_glyph_func(hb_font_t* font, void* font_data,             # <<<<<<<<<<<<<<
 *                                    hb_codepoint_t unicode,
 *                                    hb_codepoint_t variation_selector,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz._variation_glyph_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_py_font);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1015
 * 
 * 
 * cdef hb_bool_t _font_h_extents_func(hb_font_t* font, void* font_data,             # <<<<<<<<<<<<<<
 *                                     hb_font_extents_t *extents,
 *                                     void* user_data) noexcept:
 */

static hb_bool_t __pyx_f_9uharfbuzz_9_harfbuzz__font_h_extents_func(CYTHON_UNUSED hb_font_t *__pyx_v_font, void *__pyx_v_font_data, hb_font_extents_t *__pyx_v_extents, void *__pyx_v_user_data) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_py_font = 0;
  PyObject *__pyx_v_font_extents = NULL;
  hb_bool_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_t_5;
  hb_position_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_font_h_extents_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1018
 *                                     hb_font_extents_t *extents,
 *                                     void* user_data) noexcept:
 *     cdef Font py_font = <Font>font_data             # <<<<<<<<<<<<<<
 *     font_extents = (<FontFuncs>py_font.funcs)._font_h_extents_func(
 *         py_font, <object>user_data)
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_font_data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_py_font = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1019
 *                                     void* user_data) noexcept:
 *     cdef Font py_font = <Font>font_data
 *     font_extents = (<FontFuncs>py_font.funcs)._font_h_extents_func(             # <<<<<<<<<<<<<<
 *         py_font, <object>user_data)
 *     if font_extents is not None:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_py_font), __pyx_n_s_funcs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1019, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "uharfbuzz/_harfbuzz.pyx":1020
 *     cdef Font py_font = <Font>font_data
 *     font_extents = (<FontFuncs>py_font.funcs)._font_h_extents_func(
 *         py_font, <object>user_data)             # <<<<<<<<<<<<<<
 *     if font_extents is not None:
 *         if font_extents.ascender is not None:
 */
  __Pyx_INCREF(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)__pyx_t_2)->_font_h_extents_func);
  __pyx_t_3 = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)__pyx_t_2)->_font_h_extents_func; __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_2, ((PyObject *)__pyx_v_py_font), ((PyObject *)__pyx_v_user_data)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 2+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1019, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v_font_extents = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1021
 *     font_extents = (<FontFuncs>py_font.funcs)._font_h_extents_func(
 *         py_font, <object>user_data)
 *     if font_extents is not None:             # <<<<<<<<<<<<<<
 *         if font_extents.ascender is not None:
 *             extents.ascender = font_extents.ascender
 */
  __pyx_t_5 = (__pyx_v_font_extents != Py_None);
  if (__pyx_t_5) {

    /* "uharfbuzz/_harfbuzz.pyx":1022
 *         py_font, <object>user_data)
 *     if font_extents is not None:
 *         if font_extents.ascender is not None:             # <<<<<<<<<<<<<<
 *             extents.ascender = font_extents.ascender
 *         if font_extents.descender is not None:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_font_extents, __pyx_n_s_ascender); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1022, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = (__pyx_t_1 != Py_None);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_5) {

      /* "uharfbuzz/_harfbuzz.pyx":1023
 *     if font_extents is not None:
 *         if font_extents.ascender is not None:
 *             extents.ascender = font_extents.ascender             # <<<<<<<<<<<<<<
 *         if font_extents.descender is not None:
 *             extents.descender = font_extents.descender
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_font_extents, __pyx_n_s_ascender); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1023, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyInt_As_hb_position_t(__pyx_t_1); if (unlikely((__pyx_t_6 == ((hb_position_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1023, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_extents->ascender = __pyx_t_6;

      /* "uharfbuzz/_harfbuzz.pyx":1022
 *         py_font, <object>user_data)
 *     if font_extents is not None:
 *         if font_extents.ascender is not None:             # <<<<<<<<<<<<<<
 *             extents.ascender = font_extents.ascender
 *         if font_extents.descender is not None:
 */
    }

    /* "uharfbuzz/_harfbuzz.pyx":1024
 *         if font_extents.ascender is not None:
 *             extents.ascender = font_extents.ascender
 *         if font_extents.descender is not None:             # <<<<<<<<<<<<<<
 *             extents.descender = font_extents.descender
 *         if font_extents.line_gap is not None:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_font_extents, __pyx_n_s_descender); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1024, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = (__pyx_t_1 != Py_None);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_5) {

      /* "uharfbuzz/_harfbuzz.pyx":1025
 *             extents.ascender = font_extents.ascender
 *         if font_extents.descender is not None:
 *             extents.descender = font_extents.descender             # <<<<<<<<<<<<<<
 *         if font_extents.line_gap is not None:
 *             extents.line_gap = font_extents.line_gap
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_font_extents, __pyx_n_s_descender); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1025, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyInt_As_hb_position_t(__pyx_t_1); if (unlikely((__pyx_t_6 == ((hb_position_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1025, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_extents->descender = __pyx_t_6;

      /* "uharfbuzz/_harfbuzz.pyx":1024
 *         if font_extents.ascender is not None:
 *             extents.ascender = font_extents.ascender
 *         if font_extents.descender is not None:             # <<<<<<<<<<<<<<
 *             extents.descender = font_extents.descender
 *         if font_extents.line_gap is not None:
 */
    }

    /* "uharfbuzz/_harfbuzz.pyx":1026
 *         if font_extents.descender is not None:
 *             extents.descender = font_extents.descender
 *         if font_extents.line_gap is not None:             # <<<<<<<<<<<<<<
 *             extents.line_gap = font_extents.line_gap
 *         return 1
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_font_extents, __pyx_n_s_line_gap); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1026, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = (__pyx_t_1 != Py_None);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_5) {

      /* "uharfbuzz/_harfbuzz.pyx":1027
 *             extents.descender = font_extents.descender
 *         if font_extents.line_gap is not None:
 *             extents.line_gap = font_extents.line_gap             # <<<<<<<<<<<<<<
 *         return 1
 *     return 0
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_font_extents, __pyx_n_s_line_gap); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1027, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyInt_As_hb_position_t(__pyx_t_1); if (unlikely((__pyx_t_6 == ((hb_position_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1027, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_extents->line_gap = __pyx_t_6;

      /* "uharfbuzz/_harfbuzz.pyx":1026
 *         if font_extents.descender is not None:
 *             extents.descender = font_extents.descender
 *         if font_extents.line_gap is not None:             # <<<<<<<<<<<<<<
 *             extents.line_gap = font_extents.line_gap
 *         return 1
 */
    }

    /* "uharfbuzz/_harfbuzz.pyx":1028
 *         if font_extents.line_gap is not None:
 *             extents.line_gap = font_extents.line_gap
 *         return 1             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "uharfbuzz/_harfbuzz.pyx":1021
 *     font_extents = (<FontFuncs>py_font.funcs)._font_h_extents_func(
 *         py_font, <object>user_data)
 *     if font_extents is not None:             # <<<<<<<<<<<<<<
 *         if font_extents.ascender is not None:
 *             extents.ascender = font_extents.ascender
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":1029
 *             extents.line_gap = font_extents.line_gap
 *         return 1
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1015
 * 
 * 
 * cdef hb_bool_t _font_h_extents_func(hb_font_t* font, void* font_data,             # <<<<<<<<<<<<<<
 *                                     hb_font_extents_t *extents,
 *                                     void* user_data) noexcept:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz._font_h_extents_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_py_font);
  __Pyx_XDECREF(__pyx_v_font_extents);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1032
 * 
 * 
 * cdef hb_bool_t _font_v_extents_func(hb_font_t* font, void* font_data,             # <<<<<<<<<<<<<<
 *                                     hb_font_extents_t *extents,
 *                                     void* user_data) noexcept:
 */

static hb_bool_t __pyx_f_9uharfbuzz_9_harfbuzz__font_v_extents_func(CYTHON_UNUSED hb_font_t *__pyx_v_font, void *__pyx_v_font_data, hb_font_extents_t *__pyx_v_extents, void *__pyx_v_user_data) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_py_font = 0;
  PyObject *__pyx_v_font_extents = NULL;
  hb_bool_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_t_5;
  hb_position_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_font_v_extents_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1035
 *                                     hb_font_extents_t *extents,
 *                                     void* user_data) noexcept:
 *     cdef Font py_font = <Font>font_data             # <<<<<<<<<<<<<<
 *     font_extents = (<FontFuncs>py_font.funcs)._font_v_extents_func(
 *         py_font, <object>user_data)
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_font_data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_py_font = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1036
 *                                     void* user_data) noexcept:
 *     cdef Font py_font = <Font>font_data
 *     font_extents = (<FontFuncs>py_font.funcs)._font_v_extents_func(             # <<<<<<<<<<<<<<
 *         py_font, <object>user_data)
 *     if font_extents is not None:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_py_font), __pyx_n_s_funcs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1036, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "uharfbuzz/_harfbuzz.pyx":1037
 *     cdef Font py_font = <Font>font_data
 *     font_extents = (<FontFuncs>py_font.funcs)._font_v_extents_func(
 *         py_font, <object>user_data)             # <<<<<<<<<<<<<<
 *     if font_extents is not None:
 *         if font_extents.ascender is not None:
 */
  __Pyx_INCREF(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)__pyx_t_2)->_font_v_extents_func);
  __pyx_t_3 = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)__pyx_t_2)->_font_v_extents_func; __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_2, ((PyObject *)__pyx_v_py_font), ((PyObject *)__pyx_v_user_data)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 2+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1036, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v_font_extents = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1038
 *     font_extents = (<FontFuncs>py_font.funcs)._font_v_extents_func(
 *         py_font, <object>user_data)
 *     if font_extents is not None:             # <<<<<<<<<<<<<<
 *         if font_extents.ascender is not None:
 *             extents.ascender = font_extents.ascender
 */
  __pyx_t_5 = (__pyx_v_font_extents != Py_None);
  if (__pyx_t_5) {

    /* "uharfbuzz/_harfbuzz.pyx":1039
 *         py_font, <object>user_data)
 *     if font_extents is not None:
 *         if font_extents.ascender is not None:             # <<<<<<<<<<<<<<
 *             extents.ascender = font_extents.ascender
 *         if font_extents.descender is not None:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_font_extents, __pyx_n_s_ascender); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1039, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = (__pyx_t_1 != Py_None);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_5) {

      /* "uharfbuzz/_harfbuzz.pyx":1040
 *     if font_extents is not None:
 *         if font_extents.ascender is not None:
 *             extents.ascender = font_extents.ascender             # <<<<<<<<<<<<<<
 *         if font_extents.descender is not None:
 *             extents.descender = font_extents.descender
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_font_extents, __pyx_n_s_ascender); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1040, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyInt_As_hb_position_t(__pyx_t_1); if (unlikely((__pyx_t_6 == ((hb_position_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1040, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_extents->ascender = __pyx_t_6;

      /* "uharfbuzz/_harfbuzz.pyx":1039
 *         py_font, <object>user_data)
 *     if font_extents is not None:
 *         if font_extents.ascender is not None:             # <<<<<<<<<<<<<<
 *             extents.ascender = font_extents.ascender
 *         if font_extents.descender is not None:
 */
    }

    /* "uharfbuzz/_harfbuzz.pyx":1041
 *         if font_extents.ascender is not None:
 *             extents.ascender = font_extents.ascender
 *         if font_extents.descender is not None:             # <<<<<<<<<<<<<<
 *             extents.descender = font_extents.descender
 *         if font_extents.line_gap is not None:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_font_extents, __pyx_n_s_descender); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1041, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = (__pyx_t_1 != Py_None);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_5) {

      /* "uharfbuzz/_harfbuzz.pyx":1042
 *             extents.ascender = font_extents.ascender
 *         if font_extents.descender is not None:
 *             extents.descender = font_extents.descender             # <<<<<<<<<<<<<<
 *         if font_extents.line_gap is not None:
 *             extents.line_gap = font_extents.line_gap
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_font_extents, __pyx_n_s_descender); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1042, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyInt_As_hb_position_t(__pyx_t_1); if (unlikely((__pyx_t_6 == ((hb_position_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1042, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_extents->descender = __pyx_t_6;

      /* "uharfbuzz/_harfbuzz.pyx":1041
 *         if font_extents.ascender is not None:
 *             extents.ascender = font_extents.ascender
 *         if font_extents.descender is not None:             # <<<<<<<<<<<<<<
 *             extents.descender = font_extents.descender
 *         if font_extents.line_gap is not None:
 */
    }

    /* "uharfbuzz/_harfbuzz.pyx":1043
 *         if font_extents.descender is not None:
 *             extents.descender = font_extents.descender
 *         if font_extents.line_gap is not None:             # <<<<<<<<<<<<<<
 *             extents.line_gap = font_extents.line_gap
 *         return 1
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_font_extents, __pyx_n_s_line_gap); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1043, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = (__pyx_t_1 != Py_None);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_5) {

      /* "uharfbuzz/_harfbuzz.pyx":1044
 *             extents.descender = font_extents.descender
 *         if font_extents.line_gap is not None:
 *             extents.line_gap = font_extents.line_gap             # <<<<<<<<<<<<<<
 *         return 1
 *     return 0
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_font_extents, __pyx_n_s_line_gap); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1044, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyInt_As_hb_position_t(__pyx_t_1); if (unlikely((__pyx_t_6 == ((hb_position_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1044, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_extents->line_gap = __pyx_t_6;

      /* "uharfbuzz/_harfbuzz.pyx":1043
 *         if font_extents.descender is not None:
 *             extents.descender = font_extents.descender
 *         if font_extents.line_gap is not None:             # <<<<<<<<<<<<<<
 *             extents.line_gap = font_extents.line_gap
 *         return 1
 */
    }

    /* "uharfbuzz/_harfbuzz.pyx":1045
 *         if font_extents.line_gap is not None:
 *             extents.line_gap = font_extents.line_gap
 *         return 1             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "uharfbuzz/_harfbuzz.pyx":1038
 *     font_extents = (<FontFuncs>py_font.funcs)._font_v_extents_func(
 *         py_font, <object>user_data)
 *     if font_extents is not None:             # <<<<<<<<<<<<<<
 *         if font_extents.ascender is not None:
 *             extents.ascender = font_extents.ascender
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":1046
 *             extents.line_gap = font_extents.line_gap
 *         return 1
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1032
 * 
 * 
 * cdef hb_bool_t _font_v_extents_func(hb_font_t* font, void* font_data,             # <<<<<<<<<<<<<<
 *                                     hb_font_extents_t *extents,
 *                                     void* user_data) noexcept:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz._font_v_extents_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_py_font);
  __Pyx_XDECREF(__pyx_v_font_extents);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1060
 *     cdef object _font_v_extents_func
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._hb_ffuncs = hb_font_funcs_create()
 * 
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs___cinit__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs___cinit__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *__pyx_v_self) {
  int __pyx_r;

  /* "uharfbuzz/_harfbuzz.pyx":1061
 * 
 *     def __cinit__(self):
 *         self._hb_ffuncs = hb_font_funcs_create()             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_v_self->_hb_ffuncs = hb_font_funcs_create();

  /* "uharfbuzz/_harfbuzz.pyx":1060
 *     cdef object _font_v_extents_func
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._hb_ffuncs = hb_font_funcs_create()
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1063
 *         self._hb_ffuncs = hb_font_funcs_create()
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         hb_font_funcs_destroy(self._hb_ffuncs)
 * 
 */

/* Python wrapper */
static void __pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_2__dealloc__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_2__dealloc__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *__pyx_v_self) {

  /* "uharfbuzz/_harfbuzz.pyx":1064
 * 
 *     def __dealloc__(self):
 *         hb_font_funcs_destroy(self._hb_ffuncs)             # <<<<<<<<<<<<<<
 * 
 *     # DEPRECATED: use the normal constructor
 */
  hb_font_funcs_destroy(__pyx_v_self->_hb_ffuncs);

  /* "uharfbuzz/_harfbuzz.pyx":1063
 *         self._hb_ffuncs = hb_font_funcs_create()
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         hb_font_funcs_destroy(self._hb_ffuncs)
 * 
 */

  /* function exit code */
}

/* "uharfbuzz/_harfbuzz.pyx":1067
 * 
 *     # DEPRECATED: use the normal constructor
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def create(cls):
 *         cdef FontFuncs inst = cls()
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_5create(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_9FontFuncs_5create = {"create", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_5create, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_5create(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("create", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "create", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_4create(((PyTypeObject*)__pyx_v_cls));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_4create(PyTypeObject *__pyx_v_cls) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *__pyx_v_inst = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("create", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1069
 *     @classmethod
 *     def create(cls):
 *         cdef FontFuncs inst = cls()             # <<<<<<<<<<<<<<
 *         return inst
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_v_cls)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1069, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs))))) __PYX_ERR(0, 1069, __pyx_L1_error)
  __pyx_v_inst = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1070
 *     def create(cls):
 *         cdef FontFuncs inst = cls()
 *         return inst             # <<<<<<<<<<<<<<
 * 
 *     def set_glyph_h_advance_func(self,
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_inst);
  __pyx_r = ((PyObject *)__pyx_v_inst);
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1067
 * 
 *     # DEPRECATED: use the normal constructor
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def create(cls):
 *         cdef FontFuncs inst = cls()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.FontFuncs.create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_inst);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1072
 *         return inst
 * 
 *     def set_glyph_h_advance_func(self,             # <<<<<<<<<<<<<<
 *                                  func: Callable[[
 *                                      Font,
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_7set_glyph_h_advance_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_9FontFuncs_7set_glyph_h_advance_func = {"set_glyph_h_advance_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_7set_glyph_h_advance_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_7set_glyph_h_advance_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func = 0;
  PyObject *__pyx_v_user_data = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_glyph_h_advance_func (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func,&__pyx_n_s_user_data,0};

    /* "uharfbuzz/_harfbuzz.pyx":1078
 *                                      object,  # user_data
 *                                  ], int],  # h_advance
 *                                  user_data: object = None) -> None:             # <<<<<<<<<<<<<<
 *         hb_font_funcs_set_glyph_h_advance_func(
 *             self._hb_ffuncs, _glyph_h_advance_func, <void*>user_data, NULL)
 */
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_func)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1072, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_user_data);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1072, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_glyph_h_advance_func") < 0)) __PYX_ERR(0, 1072, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_func = values[0];
    __pyx_v_user_data = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_glyph_h_advance_func", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 1072, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.FontFuncs.set_glyph_h_advance_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_6set_glyph_h_advance_func(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)__pyx_v_self), __pyx_v_func, __pyx_v_user_data);

  /* "uharfbuzz/_harfbuzz.pyx":1072
 *         return inst
 * 
 *     def set_glyph_h_advance_func(self,             # <<<<<<<<<<<<<<
 *                                  func: Callable[[
 *                                      Font,
 */

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_6set_glyph_h_advance_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *__pyx_v_self, PyObject *__pyx_v_func, PyObject *__pyx_v_user_data) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_glyph_h_advance_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1079
 *                                  ], int],  # h_advance
 *                                  user_data: object = None) -> None:
 *         hb_font_funcs_set_glyph_h_advance_func(             # <<<<<<<<<<<<<<
 *             self._hb_ffuncs, _glyph_h_advance_func, <void*>user_data, NULL)
 *         self._glyph_h_advance_func = func
 */
  hb_font_funcs_set_glyph_h_advance_func(__pyx_v_self->_hb_ffuncs, __pyx_f_9uharfbuzz_9_harfbuzz__glyph_h_advance_func, ((void *)__pyx_v_user_data), NULL);

  /* "uharfbuzz/_harfbuzz.pyx":1081
 *         hb_font_funcs_set_glyph_h_advance_func(
 *             self._hb_ffuncs, _glyph_h_advance_func, <void*>user_data, NULL)
 *         self._glyph_h_advance_func = func             # <<<<<<<<<<<<<<
 * 
 *     def set_glyph_v_advance_func(self,
 */
  __Pyx_INCREF(__pyx_v_func);
  __Pyx_GIVEREF(__pyx_v_func);
  __Pyx_GOTREF(__pyx_v_self->_glyph_h_advance_func);
  __Pyx_DECREF(__pyx_v_self->_glyph_h_advance_func);
  __pyx_v_self->_glyph_h_advance_func = __pyx_v_func;

  /* "uharfbuzz/_harfbuzz.pyx":1072
 *         return inst
 * 
 *     def set_glyph_h_advance_func(self,             # <<<<<<<<<<<<<<
 *                                  func: Callable[[
 *                                      Font,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1083
 *         self._glyph_h_advance_func = func
 * 
 *     def set_glyph_v_advance_func(self,             # <<<<<<<<<<<<<<
 *                                  func: Callable[[
 *                                      Font,
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_9set_glyph_v_advance_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_9FontFuncs_9set_glyph_v_advance_func = {"set_glyph_v_advance_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_9set_glyph_v_advance_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_9set_glyph_v_advance_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func = 0;
  PyObject *__pyx_v_user_data = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_glyph_v_advance_func (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func,&__pyx_n_s_user_data,0};

    /* "uharfbuzz/_harfbuzz.pyx":1089
 *                                      object,  # user_data
 *                                  ], int],  # v_advance
 *                                  user_data: object = None) -> None:             # <<<<<<<<<<<<<<
 *         hb_font_funcs_set_glyph_v_advance_func(
 *             self._hb_ffuncs, _glyph_v_advance_func, <void*>user_data, NULL)
 */
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_func)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1083, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_user_data);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1083, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_glyph_v_advance_func") < 0)) __PYX_ERR(0, 1083, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_func = values[0];
    __pyx_v_user_data = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_glyph_v_advance_func", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 1083, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.FontFuncs.set_glyph_v_advance_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_8set_glyph_v_advance_func(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)__pyx_v_self), __pyx_v_func, __pyx_v_user_data);

  /* "uharfbuzz/_harfbuzz.pyx":1083
 *         self._glyph_h_advance_func = func
 * 
 *     def set_glyph_v_advance_func(self,             # <<<<<<<<<<<<<<
 *                                  func: Callable[[
 *                                      Font,
 */

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_8set_glyph_v_advance_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *__pyx_v_self, PyObject *__pyx_v_func, PyObject *__pyx_v_user_data) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_glyph_v_advance_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1090
 *                                  ], int],  # v_advance
 *                                  user_data: object = None) -> None:
 *         hb_font_funcs_set_glyph_v_advance_func(             # <<<<<<<<<<<<<<
 *             self._hb_ffuncs, _glyph_v_advance_func, <void*>user_data, NULL)
 *         self._glyph_v_advance_func = func
 */
  hb_font_funcs_set_glyph_v_advance_func(__pyx_v_self->_hb_ffuncs, __pyx_f_9uharfbuzz_9_harfbuzz__glyph_v_advance_func, ((void *)__pyx_v_user_data), NULL);

  /* "uharfbuzz/_harfbuzz.pyx":1092
 *         hb_font_funcs_set_glyph_v_advance_func(
 *             self._hb_ffuncs, _glyph_v_advance_func, <void*>user_data, NULL)
 *         self._glyph_v_advance_func = func             # <<<<<<<<<<<<<<
 * 
 *     def set_glyph_v_origin_func(self,
 */
  __Pyx_INCREF(__pyx_v_func);
  __Pyx_GIVEREF(__pyx_v_func);
  __Pyx_GOTREF(__pyx_v_self->_glyph_v_advance_func);
  __Pyx_DECREF(__pyx_v_self->_glyph_v_advance_func);
  __pyx_v_self->_glyph_v_advance_func = __pyx_v_func;

  /* "uharfbuzz/_harfbuzz.pyx":1083
 *         self._glyph_h_advance_func = func
 * 
 *     def set_glyph_v_advance_func(self,             # <<<<<<<<<<<<<<
 *                                  func: Callable[[
 *                                      Font,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1094
 *         self._glyph_v_advance_func = func
 * 
 *     def set_glyph_v_origin_func(self,             # <<<<<<<<<<<<<<
 *                                 func: Callable[[
 *                                     Font,
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_11set_glyph_v_origin_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_9FontFuncs_11set_glyph_v_origin_func = {"set_glyph_v_origin_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_11set_glyph_v_origin_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_11set_glyph_v_origin_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func = 0;
  PyObject *__pyx_v_user_data = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_glyph_v_origin_func (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func,&__pyx_n_s_user_data,0};

    /* "uharfbuzz/_harfbuzz.pyx":1100
 *                                     object,  # user_data
 *                                 ], (int, int, int)],  # success, v_origin_x, v_origin_y
 *                                 user_data: object = None) -> None:             # <<<<<<<<<<<<<<
 *         hb_font_funcs_set_glyph_v_origin_func(
 *             self._hb_ffuncs, _glyph_v_origin_func, <void*>user_data, NULL)
 */
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_func)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1094, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_user_data);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1094, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_glyph_v_origin_func") < 0)) __PYX_ERR(0, 1094, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_func = values[0];
    __pyx_v_user_data = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_glyph_v_origin_func", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 1094, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.FontFuncs.set_glyph_v_origin_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_10set_glyph_v_origin_func(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)__pyx_v_self), __pyx_v_func, __pyx_v_user_data);

  /* "uharfbuzz/_harfbuzz.pyx":1094
 *         self._glyph_v_advance_func = func
 * 
 *     def set_glyph_v_origin_func(self,             # <<<<<<<<<<<<<<
 *                                 func: Callable[[
 *                                     Font,
 */

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_10set_glyph_v_origin_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *__pyx_v_self, PyObject *__pyx_v_func, PyObject *__pyx_v_user_data) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_glyph_v_origin_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1101
 *                                 ], (int, int, int)],  # success, v_origin_x, v_origin_y
 *                                 user_data: object = None) -> None:
 *         hb_font_funcs_set_glyph_v_origin_func(             # <<<<<<<<<<<<<<
 *             self._hb_ffuncs, _glyph_v_origin_func, <void*>user_data, NULL)
 *         self._glyph_v_origin_func = func
 */
  hb_font_funcs_set_glyph_v_origin_func(__pyx_v_self->_hb_ffuncs, __pyx_f_9uharfbuzz_9_harfbuzz__glyph_v_origin_func, ((void *)__pyx_v_user_data), NULL);

  /* "uharfbuzz/_harfbuzz.pyx":1103
 *         hb_font_funcs_set_glyph_v_origin_func(
 *             self._hb_ffuncs, _glyph_v_origin_func, <void*>user_data, NULL)
 *         self._glyph_v_origin_func = func             # <<<<<<<<<<<<<<
 * 
 *     def set_glyph_name_func(self,
 */
  __Pyx_INCREF(__pyx_v_func);
  __Pyx_GIVEREF(__pyx_v_func);
  __Pyx_GOTREF(__pyx_v_self->_glyph_v_origin_func);
  __Pyx_DECREF(__pyx_v_self->_glyph_v_origin_func);
  __pyx_v_self->_glyph_v_origin_func = __pyx_v_func;

  /* "uharfbuzz/_harfbuzz.pyx":1094
 *         self._glyph_v_advance_func = func
 * 
 *     def set_glyph_v_origin_func(self,             # <<<<<<<<<<<<<<
 *                                 func: Callable[[
 *                                     Font,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1105
 *         self._glyph_v_origin_func = func
 * 
 *     def set_glyph_name_func(self,             # <<<<<<<<<<<<<<
 *                             func: Callable[[
 *                                 Font,
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_13set_glyph_name_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_9FontFuncs_13set_glyph_name_func = {"set_glyph_name_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_13set_glyph_name_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_13set_glyph_name_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func = 0;
  PyObject *__pyx_v_user_data = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_glyph_name_func (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func,&__pyx_n_s_user_data,0};

    /* "uharfbuzz/_harfbuzz.pyx":1111
 *                                 object,  # user_data
 *                             ], str],  # name
 *                             user_data: object = None) -> None:             # <<<<<<<<<<<<<<
 *         hb_font_funcs_set_glyph_name_func(
 *             self._hb_ffuncs, _glyph_name_func, <void*>user_data, NULL)
 */
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_func)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1105, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_user_data);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1105, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_glyph_name_func") < 0)) __PYX_ERR(0, 1105, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_func = values[0];
    __pyx_v_user_data = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_glyph_name_func", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 1105, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.FontFuncs.set_glyph_name_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_12set_glyph_name_func(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)__pyx_v_self), __pyx_v_func, __pyx_v_user_data);

  /* "uharfbuzz/_harfbuzz.pyx":1105
 *         self._glyph_v_origin_func = func
 * 
 *     def set_glyph_name_func(self,             # <<<<<<<<<<<<<<
 *                             func: Callable[[
 *                                 Font,
 */

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_12set_glyph_name_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *__pyx_v_self, PyObject *__pyx_v_func, PyObject *__pyx_v_user_data) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_glyph_name_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1112
 *                             ], str],  # name
 *                             user_data: object = None) -> None:
 *         hb_font_funcs_set_glyph_name_func(             # <<<<<<<<<<<<<<
 *             self._hb_ffuncs, _glyph_name_func, <void*>user_data, NULL)
 *         self._glyph_name_func = func
 */
  hb_font_funcs_set_glyph_name_func(__pyx_v_self->_hb_ffuncs, __pyx_f_9uharfbuzz_9_harfbuzz__glyph_name_func, ((void *)__pyx_v_user_data), NULL);

  /* "uharfbuzz/_harfbuzz.pyx":1114
 *         hb_font_funcs_set_glyph_name_func(
 *             self._hb_ffuncs, _glyph_name_func, <void*>user_data, NULL)
 *         self._glyph_name_func = func             # <<<<<<<<<<<<<<
 * 
 *     def set_nominal_glyph_func(self,
 */
  __Pyx_INCREF(__pyx_v_func);
  __Pyx_GIVEREF(__pyx_v_func);
  __Pyx_GOTREF(__pyx_v_self->_glyph_name_func);
  __Pyx_DECREF(__pyx_v_self->_glyph_name_func);
  __pyx_v_self->_glyph_name_func = __pyx_v_func;

  /* "uharfbuzz/_harfbuzz.pyx":1105
 *         self._glyph_v_origin_func = func
 * 
 *     def set_glyph_name_func(self,             # <<<<<<<<<<<<<<
 *                             func: Callable[[
 *                                 Font,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1116
 *         self._glyph_name_func = func
 * 
 *     def set_nominal_glyph_func(self,             # <<<<<<<<<<<<<<
 *                                func: Callable[[
 *                                    Font,
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_15set_nominal_glyph_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_9FontFuncs_15set_nominal_glyph_func = {"set_nominal_glyph_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_15set_nominal_glyph_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_15set_nominal_glyph_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func = 0;
  PyObject *__pyx_v_user_data = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_nominal_glyph_func (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func,&__pyx_n_s_user_data,0};

    /* "uharfbuzz/_harfbuzz.pyx":1122
 *                                    object,  # user_data
 *                                ], int],  # gid
 *                                user_data: object = None) -> None:             # <<<<<<<<<<<<<<
 *         hb_font_funcs_set_nominal_glyph_func(
 *             self._hb_ffuncs, _nominal_glyph_func, <void*>user_data, NULL)
 */
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_func)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1116, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_user_data);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1116, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_nominal_glyph_func") < 0)) __PYX_ERR(0, 1116, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_func = values[0];
    __pyx_v_user_data = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_nominal_glyph_func", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 1116, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.FontFuncs.set_nominal_glyph_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_14set_nominal_glyph_func(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)__pyx_v_self), __pyx_v_func, __pyx_v_user_data);

  /* "uharfbuzz/_harfbuzz.pyx":1116
 *         self._glyph_name_func = func
 * 
 *     def set_nominal_glyph_func(self,             # <<<<<<<<<<<<<<
 *                                func: Callable[[
 *                                    Font,
 */

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_14set_nominal_glyph_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *__pyx_v_self, PyObject *__pyx_v_func, PyObject *__pyx_v_user_data) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_nominal_glyph_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1123
 *                                ], int],  # gid
 *                                user_data: object = None) -> None:
 *         hb_font_funcs_set_nominal_glyph_func(             # <<<<<<<<<<<<<<
 *             self._hb_ffuncs, _nominal_glyph_func, <void*>user_data, NULL)
 *         self._nominal_glyph_func = func
 */
  hb_font_funcs_set_nominal_glyph_func(__pyx_v_self->_hb_ffuncs, __pyx_f_9uharfbuzz_9_harfbuzz__nominal_glyph_func, ((void *)__pyx_v_user_data), NULL);

  /* "uharfbuzz/_harfbuzz.pyx":1125
 *         hb_font_funcs_set_nominal_glyph_func(
 *             self._hb_ffuncs, _nominal_glyph_func, <void*>user_data, NULL)
 *         self._nominal_glyph_func = func             # <<<<<<<<<<<<<<
 * 
 *     def set_variation_glyph_func(self,
 */
  __Pyx_INCREF(__pyx_v_func);
  __Pyx_GIVEREF(__pyx_v_func);
  __Pyx_GOTREF(__pyx_v_self->_nominal_glyph_func);
  __Pyx_DECREF(__pyx_v_self->_nominal_glyph_func);
  __pyx_v_self->_nominal_glyph_func = __pyx_v_func;

  /* "uharfbuzz/_harfbuzz.pyx":1116
 *         self._glyph_name_func = func
 * 
 *     def set_nominal_glyph_func(self,             # <<<<<<<<<<<<<<
 *                                func: Callable[[
 *                                    Font,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1127
 *         self._nominal_glyph_func = func
 * 
 *     def set_variation_glyph_func(self,             # <<<<<<<<<<<<<<
 *                                func: Callable[[
 *                                    Font,
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_17set_variation_glyph_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_9FontFuncs_17set_variation_glyph_func = {"set_variation_glyph_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_17set_variation_glyph_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_17set_variation_glyph_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func = 0;
  PyObject *__pyx_v_user_data = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_variation_glyph_func (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func,&__pyx_n_s_user_data,0};

    /* "uharfbuzz/_harfbuzz.pyx":1134
 *                                    object,  # user_data
 *                                ], int],  # gid
 *                                user_data: object = None) -> None:             # <<<<<<<<<<<<<<
 *         hb_font_funcs_set_variation_glyph_func(
 *             self._hb_ffuncs, _variation_glyph_func, <void*>user_data, NULL)
 */
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_func)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1127, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_user_data);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1127, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_variation_glyph_func") < 0)) __PYX_ERR(0, 1127, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_func = values[0];
    __pyx_v_user_data = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_variation_glyph_func", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 1127, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.FontFuncs.set_variation_glyph_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_16set_variation_glyph_func(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)__pyx_v_self), __pyx_v_func, __pyx_v_user_data);

  /* "uharfbuzz/_harfbuzz.pyx":1127
 *         self._nominal_glyph_func = func
 * 
 *     def set_variation_glyph_func(self,             # <<<<<<<<<<<<<<
 *                                func: Callable[[
 *                                    Font,
 */

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_16set_variation_glyph_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *__pyx_v_self, PyObject *__pyx_v_func, PyObject *__pyx_v_user_data) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_variation_glyph_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1135
 *                                ], int],  # gid
 *                                user_data: object = None) -> None:
 *         hb_font_funcs_set_variation_glyph_func(             # <<<<<<<<<<<<<<
 *             self._hb_ffuncs, _variation_glyph_func, <void*>user_data, NULL)
 *         self._variation_glyph_func = func
 */
  hb_font_funcs_set_variation_glyph_func(__pyx_v_self->_hb_ffuncs, __pyx_f_9uharfbuzz_9_harfbuzz__variation_glyph_func, ((void *)__pyx_v_user_data), NULL);

  /* "uharfbuzz/_harfbuzz.pyx":1137
 *         hb_font_funcs_set_variation_glyph_func(
 *             self._hb_ffuncs, _variation_glyph_func, <void*>user_data, NULL)
 *         self._variation_glyph_func = func             # <<<<<<<<<<<<<<
 * 
 *     def set_font_h_extents_func(self,
 */
  __Pyx_INCREF(__pyx_v_func);
  __Pyx_GIVEREF(__pyx_v_func);
  __Pyx_GOTREF(__pyx_v_self->_variation_glyph_func);
  __Pyx_DECREF(__pyx_v_self->_variation_glyph_func);
  __pyx_v_self->_variation_glyph_func = __pyx_v_func;

  /* "uharfbuzz/_harfbuzz.pyx":1127
 *         self._nominal_glyph_func = func
 * 
 *     def set_variation_glyph_func(self,             # <<<<<<<<<<<<<<
 *                                func: Callable[[
 *                                    Font,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1139
 *         self._variation_glyph_func = func
 * 
 *     def set_font_h_extents_func(self,             # <<<<<<<<<<<<<<
 *                                 func: Callable[[
 *                                     Font,
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_19set_font_h_extents_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_9FontFuncs_19set_font_h_extents_func = {"set_font_h_extents_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_19set_font_h_extents_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_19set_font_h_extents_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func = 0;
  PyObject *__pyx_v_user_data = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_font_h_extents_func (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func,&__pyx_n_s_user_data,0};

    /* "uharfbuzz/_harfbuzz.pyx":1144
 *                                     object,  # user_data
 *                                 ], FontExtents],  # extents
 *                                 user_data: object = None) -> None:             # <<<<<<<<<<<<<<
 *         hb_font_funcs_set_font_h_extents_func(
 *             self._hb_ffuncs, _font_h_extents_func, <void*>user_data, NULL)
 */
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_func)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1139, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_user_data);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1139, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_font_h_extents_func") < 0)) __PYX_ERR(0, 1139, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_func = values[0];
    __pyx_v_user_data = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_font_h_extents_func", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 1139, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.FontFuncs.set_font_h_extents_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_18set_font_h_extents_func(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)__pyx_v_self), __pyx_v_func, __pyx_v_user_data);

  /* "uharfbuzz/_harfbuzz.pyx":1139
 *         self._variation_glyph_func = func
 * 
 *     def set_font_h_extents_func(self,             # <<<<<<<<<<<<<<
 *                                 func: Callable[[
 *                                     Font,
 */

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_18set_font_h_extents_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *__pyx_v_self, PyObject *__pyx_v_func, PyObject *__pyx_v_user_data) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_font_h_extents_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1145
 *                                 ], FontExtents],  # extents
 *                                 user_data: object = None) -> None:
 *         hb_font_funcs_set_font_h_extents_func(             # <<<<<<<<<<<<<<
 *             self._hb_ffuncs, _font_h_extents_func, <void*>user_data, NULL)
 *         self._font_h_extents_func = func
 */
  hb_font_funcs_set_font_h_extents_func(__pyx_v_self->_hb_ffuncs, __pyx_f_9uharfbuzz_9_harfbuzz__font_h_extents_func, ((void *)__pyx_v_user_data), NULL);

  /* "uharfbuzz/_harfbuzz.pyx":1147
 *         hb_font_funcs_set_font_h_extents_func(
 *             self._hb_ffuncs, _font_h_extents_func, <void*>user_data, NULL)
 *         self._font_h_extents_func = func             # <<<<<<<<<<<<<<
 * 
 *     def set_font_v_extents_func(self,
 */
  __Pyx_INCREF(__pyx_v_func);
  __Pyx_GIVEREF(__pyx_v_func);
  __Pyx_GOTREF(__pyx_v_self->_font_h_extents_func);
  __Pyx_DECREF(__pyx_v_self->_font_h_extents_func);
  __pyx_v_self->_font_h_extents_func = __pyx_v_func;

  /* "uharfbuzz/_harfbuzz.pyx":1139
 *         self._variation_glyph_func = func
 * 
 *     def set_font_h_extents_func(self,             # <<<<<<<<<<<<<<
 *                                 func: Callable[[
 *                                     Font,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1149
 *         self._font_h_extents_func = func
 * 
 *     def set_font_v_extents_func(self,             # <<<<<<<<<<<<<<
 *                                 func: Callable[[
 *                                     Font,
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_21set_font_v_extents_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_9FontFuncs_21set_font_v_extents_func = {"set_font_v_extents_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_21set_font_v_extents_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_21set_font_v_extents_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func = 0;
  PyObject *__pyx_v_user_data = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_font_v_extents_func (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func,&__pyx_n_s_user_data,0};

    /* "uharfbuzz/_harfbuzz.pyx":1154
 *                                     object,  # user_data
 *                                 ], FontExtents],  # extents
 *                                 user_data: object = None) -> None:             # <<<<<<<<<<<<<<
 *         hb_font_funcs_set_font_v_extents_func(
 *             self._hb_ffuncs, _font_v_extents_func, <void*>user_data, NULL)
 */
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_func)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1149, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_user_data);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1149, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_font_v_extents_func") < 0)) __PYX_ERR(0, 1149, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_func = values[0];
    __pyx_v_user_data = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_font_v_extents_func", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 1149, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.FontFuncs.set_font_v_extents_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_20set_font_v_extents_func(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)__pyx_v_self), __pyx_v_func, __pyx_v_user_data);

  /* "uharfbuzz/_harfbuzz.pyx":1149
 *         self._font_h_extents_func = func
 * 
 *     def set_font_v_extents_func(self,             # <<<<<<<<<<<<<<
 *                                 func: Callable[[
 *                                     Font,
 */

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_20set_font_v_extents_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *__pyx_v_self, PyObject *__pyx_v_func, PyObject *__pyx_v_user_data) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_font_v_extents_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1155
 *                                 ], FontExtents],  # extents
 *                                 user_data: object = None) -> None:
 *         hb_font_funcs_set_font_v_extents_func(             # <<<<<<<<<<<<<<
 *             self._hb_ffuncs, _font_v_extents_func, <void*>user_data, NULL)
 *         self._font_v_extents_func = func
 */
  hb_font_funcs_set_font_v_extents_func(__pyx_v_self->_hb_ffuncs, __pyx_f_9uharfbuzz_9_harfbuzz__font_v_extents_func, ((void *)__pyx_v_user_data), NULL);

  /* "uharfbuzz/_harfbuzz.pyx":1157
 *         hb_font_funcs_set_font_v_extents_func(
 *             self._hb_ffuncs, _font_v_extents_func, <void*>user_data, NULL)
 *         self._font_v_extents_func = func             # <<<<<<<<<<<<<<
 * 
 * def shape(font: Font, buffer: Buffer,
 */
  __Pyx_INCREF(__pyx_v_func);
  __Pyx_GIVEREF(__pyx_v_func);
  __Pyx_GOTREF(__pyx_v_self->_font_v_extents_func);
  __Pyx_DECREF(__pyx_v_self->_font_v_extents_func);
  __pyx_v_self->_font_v_extents_func = __pyx_v_func;

  /* "uharfbuzz/_harfbuzz.pyx":1149
 *         self._font_h_extents_func = func
 * 
 *     def set_font_v_extents_func(self,             # <<<<<<<<<<<<<<
 *                                 func: Callable[[
 *                                     Font,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_23__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_9FontFuncs_23__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_23__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_23__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_22__reduce_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_22__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.FontFuncs.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_25__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_9FontFuncs_25__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_25__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_25__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.FontFuncs.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_24__setstate_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9FontFuncs_24__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.FontFuncs.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1159
 *         self._font_v_extents_func = func
 * 
 * def shape(font: Font, buffer: Buffer,             # <<<<<<<<<<<<<<
 *         features: Dict[str,Union[int,bool,Sequence[Tuple[int,int,Union[int,bool]]]]] = None,
 *         shapers: List[str] = None) -> None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3shape(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3shape = {"shape", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3shape, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3shape(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font = 0;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_buffer = 0;
  PyObject *__pyx_v_features = 0;
  PyObject *__pyx_v_shapers = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("shape (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_font,&__pyx_n_s_buffer,&__pyx_n_s_features,&__pyx_n_s_shapers,0};

    /* "uharfbuzz/_harfbuzz.pyx":1160
 * 
 * def shape(font: Font, buffer: Buffer,
 *         features: Dict[str,Union[int,bool,Sequence[Tuple[int,int,Union[int,bool]]]]] = None,             # <<<<<<<<<<<<<<
 *         shapers: List[str] = None) -> None:
 *     cdef unsigned int size
 */
    values[2] = __Pyx_Arg_NewRef_FASTCALL(((PyObject*)Py_None));

    /* "uharfbuzz/_harfbuzz.pyx":1161
 * def shape(font: Font, buffer: Buffer,
 *         features: Dict[str,Union[int,bool,Sequence[Tuple[int,int,Union[int,bool]]]]] = None,
 *         shapers: List[str] = None) -> None:             # <<<<<<<<<<<<<<
 *     cdef unsigned int size
 *     cdef hb_feature_t* hb_features
 */
    values[3] = __Pyx_Arg_NewRef_FASTCALL(((PyObject*)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_font)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1159, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_buffer)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1159, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("shape", 0, 2, 4, 1); __PYX_ERR(0, 1159, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_features);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1159, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_shapers);
          if (value) { values[3] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1159, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "shape") < 0)) __PYX_ERR(0, 1159, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_font = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)values[0]);
    __pyx_v_buffer = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)values[1]);
    __pyx_v_features = ((PyObject*)values[2]);
    __pyx_v_shapers = ((PyObject*)values[3]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("shape", 0, 2, 4, __pyx_nargs); __PYX_ERR(0, 1159, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.shape", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_font), __pyx_ptype_9uharfbuzz_9_harfbuzz_Font, 0, "font", 0))) __PYX_ERR(0, 1159, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_buffer), __pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer, 0, "buffer", 0))) __PYX_ERR(0, 1159, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_features), (&PyDict_Type), 1, "features", 1))) __PYX_ERR(0, 1160, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_shapers), (&PyList_Type), 1, "shapers", 1))) __PYX_ERR(0, 1161, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_2shape(__pyx_self, __pyx_v_font, __pyx_v_buffer, __pyx_v_features, __pyx_v_shapers);

  /* "uharfbuzz/_harfbuzz.pyx":1159
 *         self._font_v_extents_func = func
 * 
 * def shape(font: Font, buffer: Buffer,             # <<<<<<<<<<<<<<
 *         features: Dict[str,Union[int,bool,Sequence[Tuple[int,int,Union[int,bool]]]]] = None,
 *         shapers: List[str] = None) -> None:
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_2shape(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *__pyx_v_buffer, PyObject *__pyx_v_features, PyObject *__pyx_v_shapers) {
  unsigned int __pyx_v_size;
  hb_feature_t *__pyx_v_hb_features;
  PyObject *__pyx_v_packed = 0;
  hb_feature_t __pyx_v_feat;
  char const *__pyx_v_c_shapers[10];
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_v_start = NULL;
  PyObject *__pyx_v_end = NULL;
  PyObject *__pyx_v_shaper = NULL;
  hb_bool_t __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  unsigned int __pyx_t_11;
  char const *__pyx_t_12;
  unsigned long __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *(*__pyx_t_15)(PyObject *);
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *(*__pyx_t_19)(PyObject *);
  Py_ssize_t __pyx_t_20;
  char const *__pyx_t_21;
  char const *__pyx_t_22;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  PyObject *__pyx_t_26 = NULL;
  PyObject *__pyx_t_27 = NULL;
  PyObject *__pyx_t_28 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("shape", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1167
 *     cdef hb_feature_t feat
 *     cdef const char* c_shapers[10]
 *     size = 0             # <<<<<<<<<<<<<<
 *     hb_features = NULL
 *     try:
 */
  __pyx_v_size = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1168
 *     cdef const char* c_shapers[10]
 *     size = 0
 *     hb_features = NULL             # <<<<<<<<<<<<<<
 *     try:
 *         if features:
 */
  __pyx_v_hb_features = NULL;

  /* "uharfbuzz/_harfbuzz.pyx":1169
 *     size = 0
 *     hb_features = NULL
 *     try:             # <<<<<<<<<<<<<<
 *         if features:
 *             for value in features.values():
 */
  /*try:*/ {

    /* "uharfbuzz/_harfbuzz.pyx":1170
 *     hb_features = NULL
 *     try:
 *         if features:             # <<<<<<<<<<<<<<
 *             for value in features.values():
 *                 if isinstance(value, int):
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_features); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1170, __pyx_L4_error)
    if (__pyx_t_1) {

      /* "uharfbuzz/_harfbuzz.pyx":1171
 *     try:
 *         if features:
 *             for value in features.values():             # <<<<<<<<<<<<<<
 *                 if isinstance(value, int):
 *                     size += 1
 */
      __pyx_t_3 = 0;
      if (unlikely(__pyx_v_features == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "values");
        __PYX_ERR(0, 1171, __pyx_L4_error)
      }
      __pyx_t_6 = __Pyx_dict_iterator(__pyx_v_features, 1, __pyx_n_s_values, (&__pyx_t_4), (&__pyx_t_5)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1171, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_2);
      __pyx_t_2 = __pyx_t_6;
      __pyx_t_6 = 0;
      while (1) {
        __pyx_t_7 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_4, &__pyx_t_3, NULL, &__pyx_t_6, NULL, __pyx_t_5);
        if (unlikely(__pyx_t_7 == 0)) break;
        if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 1171, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "uharfbuzz/_harfbuzz.pyx":1172
 *         if features:
 *             for value in features.values():
 *                 if isinstance(value, int):             # <<<<<<<<<<<<<<
 *                     size += 1
 *                 else:
 */
        __pyx_t_1 = PyInt_Check(__pyx_v_value); 
        if (__pyx_t_1) {

          /* "uharfbuzz/_harfbuzz.pyx":1173
 *             for value in features.values():
 *                 if isinstance(value, int):
 *                     size += 1             # <<<<<<<<<<<<<<
 *                 else:
 *                     size += len(value)
 */
          __pyx_v_size = (__pyx_v_size + 1);

          /* "uharfbuzz/_harfbuzz.pyx":1172
 *         if features:
 *             for value in features.values():
 *                 if isinstance(value, int):             # <<<<<<<<<<<<<<
 *                     size += 1
 *                 else:
 */
          goto __pyx_L9;
        }

        /* "uharfbuzz/_harfbuzz.pyx":1175
 *                     size += 1
 *                 else:
 *                     size += len(value)             # <<<<<<<<<<<<<<
 *             hb_features = <hb_feature_t*>malloc(size * sizeof(hb_feature_t))
 *             i = 0
 */
        /*else*/ {
          __pyx_t_8 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1175, __pyx_L4_error)
          __pyx_v_size = (__pyx_v_size + __pyx_t_8);
        }
        __pyx_L9:;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uharfbuzz/_harfbuzz.pyx":1176
 *                 else:
 *                     size += len(value)
 *             hb_features = <hb_feature_t*>malloc(size * sizeof(hb_feature_t))             # <<<<<<<<<<<<<<
 *             i = 0
 *             for name, value in features.items():
 */
      __pyx_v_hb_features = ((hb_feature_t *)malloc((__pyx_v_size * (sizeof(hb_feature_t)))));

      /* "uharfbuzz/_harfbuzz.pyx":1177
 *                     size += len(value)
 *             hb_features = <hb_feature_t*>malloc(size * sizeof(hb_feature_t))
 *             i = 0             # <<<<<<<<<<<<<<
 *             for name, value in features.items():
 *                 assert i < size, "index out of range for feature array capacity"
 */
      __Pyx_INCREF(__pyx_int_0);
      __pyx_v_i = __pyx_int_0;

      /* "uharfbuzz/_harfbuzz.pyx":1178
 *             hb_features = <hb_feature_t*>malloc(size * sizeof(hb_feature_t))
 *             i = 0
 *             for name, value in features.items():             # <<<<<<<<<<<<<<
 *                 assert i < size, "index out of range for feature array capacity"
 *                 packed = name.encode()
 */
      __pyx_t_4 = 0;
      if (unlikely(__pyx_v_features == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
        __PYX_ERR(0, 1178, __pyx_L4_error)
      }
      __pyx_t_6 = __Pyx_dict_iterator(__pyx_v_features, 1, __pyx_n_s_items, (&__pyx_t_3), (&__pyx_t_5)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1178, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_2);
      __pyx_t_2 = __pyx_t_6;
      __pyx_t_6 = 0;
      while (1) {
        __pyx_t_7 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_3, &__pyx_t_4, &__pyx_t_6, &__pyx_t_9, NULL, __pyx_t_5);
        if (unlikely(__pyx_t_7 == 0)) break;
        if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 1178, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_6);
        __pyx_t_6 = 0;
        __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_9);
        __pyx_t_9 = 0;

        /* "uharfbuzz/_harfbuzz.pyx":1179
 *             i = 0
 *             for name, value in features.items():
 *                 assert i < size, "index out of range for feature array capacity"             # <<<<<<<<<<<<<<
 *                 packed = name.encode()
 *                 if isinstance(value, int):
 */
        #ifndef CYTHON_WITHOUT_ASSERTIONS
        if (unlikely(__pyx_assertions_enabled())) {
          __pyx_t_9 = __Pyx_PyInt_From_unsigned_int(__pyx_v_size); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1179, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_6 = PyObject_RichCompare(__pyx_v_i, __pyx_t_9, Py_LT); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1179, __pyx_L4_error)
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1179, __pyx_L4_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (unlikely(!__pyx_t_1)) {
            __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_kp_u_index_out_of_range_for_feature_a, 0, 0);
            __PYX_ERR(0, 1179, __pyx_L4_error)
          }
        }
        #else
        if ((1)); else __PYX_ERR(0, 1179, __pyx_L4_error)
        #endif

        /* "uharfbuzz/_harfbuzz.pyx":1180
 *             for name, value in features.items():
 *                 assert i < size, "index out of range for feature array capacity"
 *                 packed = name.encode()             # <<<<<<<<<<<<<<
 *                 if isinstance(value, int):
 *                     hb_feature_from_string(packed, len(packed), &feat)
 */
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_name, __pyx_n_s_encode); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1180, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = NULL;
        __pyx_t_11 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_9))) {
          __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
          if (likely(__pyx_t_10)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
            __Pyx_INCREF(__pyx_t_10);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_9, function);
            __pyx_t_11 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_10, NULL};
          __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_11, 0+__pyx_t_11);
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1180, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        }
        if (!(likely(PyBytes_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_6))) __PYX_ERR(0, 1180, __pyx_L4_error)
        __Pyx_XDECREF_SET(__pyx_v_packed, ((PyObject*)__pyx_t_6));
        __pyx_t_6 = 0;

        /* "uharfbuzz/_harfbuzz.pyx":1181
 *                 assert i < size, "index out of range for feature array capacity"
 *                 packed = name.encode()
 *                 if isinstance(value, int):             # <<<<<<<<<<<<<<
 *                     hb_feature_from_string(packed, len(packed), &feat)
 *                     feat.value = value
 */
        __pyx_t_1 = PyInt_Check(__pyx_v_value); 
        if (__pyx_t_1) {

          /* "uharfbuzz/_harfbuzz.pyx":1182
 *                 packed = name.encode()
 *                 if isinstance(value, int):
 *                     hb_feature_from_string(packed, len(packed), &feat)             # <<<<<<<<<<<<<<
 *                     feat.value = value
 *                     hb_features[i] = feat
 */
          if (unlikely(__pyx_v_packed == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
            __PYX_ERR(0, 1182, __pyx_L4_error)
          }
          __pyx_t_12 = __Pyx_PyBytes_AsString(__pyx_v_packed); if (unlikely((!__pyx_t_12) && PyErr_Occurred())) __PYX_ERR(0, 1182, __pyx_L4_error)
          if (unlikely(__pyx_v_packed == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
            __PYX_ERR(0, 1182, __pyx_L4_error)
          }
          __pyx_t_8 = __Pyx_PyBytes_GET_SIZE(__pyx_v_packed); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1182, __pyx_L4_error)
          (void)(hb_feature_from_string(__pyx_t_12, __pyx_t_8, (&__pyx_v_feat)));

          /* "uharfbuzz/_harfbuzz.pyx":1183
 *                 if isinstance(value, int):
 *                     hb_feature_from_string(packed, len(packed), &feat)
 *                     feat.value = value             # <<<<<<<<<<<<<<
 *                     hb_features[i] = feat
 *                     i += 1
 */
          __pyx_t_13 = __Pyx_PyInt_As_unsigned_long(__pyx_v_value); if (unlikely((__pyx_t_13 == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1183, __pyx_L4_error)
          __pyx_v_feat.value = __pyx_t_13;

          /* "uharfbuzz/_harfbuzz.pyx":1184
 *                     hb_feature_from_string(packed, len(packed), &feat)
 *                     feat.value = value
 *                     hb_features[i] = feat             # <<<<<<<<<<<<<<
 *                     i += 1
 *                 else:
 */
          __pyx_t_8 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_8 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1184, __pyx_L4_error)
          (__pyx_v_hb_features[__pyx_t_8]) = __pyx_v_feat;

          /* "uharfbuzz/_harfbuzz.pyx":1185
 *                     feat.value = value
 *                     hb_features[i] = feat
 *                     i += 1             # <<<<<<<<<<<<<<
 *                 else:
 *                     feat.tag = hb_tag_from_string(packed, -1)
 */
          __pyx_t_6 = __Pyx_PyInt_AddObjC(__pyx_v_i, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1185, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF_SET(__pyx_v_i, __pyx_t_6);
          __pyx_t_6 = 0;

          /* "uharfbuzz/_harfbuzz.pyx":1181
 *                 assert i < size, "index out of range for feature array capacity"
 *                 packed = name.encode()
 *                 if isinstance(value, int):             # <<<<<<<<<<<<<<
 *                     hb_feature_from_string(packed, len(packed), &feat)
 *                     feat.value = value
 */
          goto __pyx_L12;
        }

        /* "uharfbuzz/_harfbuzz.pyx":1187
 *                     i += 1
 *                 else:
 *                     feat.tag = hb_tag_from_string(packed, -1)             # <<<<<<<<<<<<<<
 *                     for start, end, value in value:
 *                         feat.value = value
 */
        /*else*/ {
          if (unlikely(__pyx_v_packed == Py_None)) {
            PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
            __PYX_ERR(0, 1187, __pyx_L4_error)
          }
          __pyx_t_14 = __Pyx_PyBytes_AsString(__pyx_v_packed); if (unlikely((!__pyx_t_14) && PyErr_Occurred())) __PYX_ERR(0, 1187, __pyx_L4_error)
          __pyx_v_feat.tag = hb_tag_from_string(__pyx_t_14, -1);

          /* "uharfbuzz/_harfbuzz.pyx":1188
 *                 else:
 *                     feat.tag = hb_tag_from_string(packed, -1)
 *                     for start, end, value in value:             # <<<<<<<<<<<<<<
 *                         feat.value = value
 *                         feat.start = start
 */
          if (likely(PyList_CheckExact(__pyx_v_value)) || PyTuple_CheckExact(__pyx_v_value)) {
            __pyx_t_6 = __pyx_v_value; __Pyx_INCREF(__pyx_t_6);
            __pyx_t_8 = 0;
            __pyx_t_15 = NULL;
          } else {
            __pyx_t_8 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_v_value); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1188, __pyx_L4_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_15 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_6); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1188, __pyx_L4_error)
          }
          for (;;) {
            if (likely(!__pyx_t_15)) {
              if (likely(PyList_CheckExact(__pyx_t_6))) {
                {
                  Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_6);
                  #if !CYTHON_ASSUME_SAFE_MACROS
                  if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1188, __pyx_L4_error)
                  #endif
                  if (__pyx_t_8 >= __pyx_temp) break;
                }
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_9 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_8); __Pyx_INCREF(__pyx_t_9); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(0, 1188, __pyx_L4_error)
                #else
                __pyx_t_9 = __Pyx_PySequence_ITEM(__pyx_t_6, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1188, __pyx_L4_error)
                __Pyx_GOTREF(__pyx_t_9);
                #endif
              } else {
                {
                  Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_6);
                  #if !CYTHON_ASSUME_SAFE_MACROS
                  if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1188, __pyx_L4_error)
                  #endif
                  if (__pyx_t_8 >= __pyx_temp) break;
                }
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_8); __Pyx_INCREF(__pyx_t_9); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(0, 1188, __pyx_L4_error)
                #else
                __pyx_t_9 = __Pyx_PySequence_ITEM(__pyx_t_6, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1188, __pyx_L4_error)
                __Pyx_GOTREF(__pyx_t_9);
                #endif
              }
            } else {
              __pyx_t_9 = __pyx_t_15(__pyx_t_6);
              if (unlikely(!__pyx_t_9)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 1188, __pyx_L4_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_9);
            }
            if ((likely(PyTuple_CheckExact(__pyx_t_9))) || (PyList_CheckExact(__pyx_t_9))) {
              PyObject* sequence = __pyx_t_9;
              Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
              if (unlikely(size != 3)) {
                if (size > 3) __Pyx_RaiseTooManyValuesError(3);
                else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                __PYX_ERR(0, 1188, __pyx_L4_error)
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              if (likely(PyTuple_CheckExact(sequence))) {
                __pyx_t_10 = PyTuple_GET_ITEM(sequence, 0); 
                __pyx_t_16 = PyTuple_GET_ITEM(sequence, 1); 
                __pyx_t_17 = PyTuple_GET_ITEM(sequence, 2); 
              } else {
                __pyx_t_10 = PyList_GET_ITEM(sequence, 0); 
                __pyx_t_16 = PyList_GET_ITEM(sequence, 1); 
                __pyx_t_17 = PyList_GET_ITEM(sequence, 2); 
              }
              __Pyx_INCREF(__pyx_t_10);
              __Pyx_INCREF(__pyx_t_16);
              __Pyx_INCREF(__pyx_t_17);
              #else
              __pyx_t_10 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1188, __pyx_L4_error)
              __Pyx_GOTREF(__pyx_t_10);
              __pyx_t_16 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1188, __pyx_L4_error)
              __Pyx_GOTREF(__pyx_t_16);
              __pyx_t_17 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1188, __pyx_L4_error)
              __Pyx_GOTREF(__pyx_t_17);
              #endif
              __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
            } else {
              Py_ssize_t index = -1;
              __pyx_t_18 = PyObject_GetIter(__pyx_t_9); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 1188, __pyx_L4_error)
              __Pyx_GOTREF(__pyx_t_18);
              __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
              __pyx_t_19 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_18);
              index = 0; __pyx_t_10 = __pyx_t_19(__pyx_t_18); if (unlikely(!__pyx_t_10)) goto __pyx_L15_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_10);
              index = 1; __pyx_t_16 = __pyx_t_19(__pyx_t_18); if (unlikely(!__pyx_t_16)) goto __pyx_L15_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_16);
              index = 2; __pyx_t_17 = __pyx_t_19(__pyx_t_18); if (unlikely(!__pyx_t_17)) goto __pyx_L15_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_17);
              if (__Pyx_IternextUnpackEndCheck(__pyx_t_19(__pyx_t_18), 3) < 0) __PYX_ERR(0, 1188, __pyx_L4_error)
              __pyx_t_19 = NULL;
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
              goto __pyx_L16_unpacking_done;
              __pyx_L15_unpacking_failed:;
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
              __pyx_t_19 = NULL;
              if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
              __PYX_ERR(0, 1188, __pyx_L4_error)
              __pyx_L16_unpacking_done:;
            }
            __Pyx_XDECREF_SET(__pyx_v_start, __pyx_t_10);
            __pyx_t_10 = 0;
            __Pyx_XDECREF_SET(__pyx_v_end, __pyx_t_16);
            __pyx_t_16 = 0;
            __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_17);
            __pyx_t_17 = 0;

            /* "uharfbuzz/_harfbuzz.pyx":1189
 *                     feat.tag = hb_tag_from_string(packed, -1)
 *                     for start, end, value in value:
 *                         feat.value = value             # <<<<<<<<<<<<<<
 *                         feat.start = start
 *                         feat.end = end
 */
            __pyx_t_13 = __Pyx_PyInt_As_unsigned_long(__pyx_v_value); if (unlikely((__pyx_t_13 == (unsigned long)-1) && PyErr_Occurred())) __PYX_ERR(0, 1189, __pyx_L4_error)
            __pyx_v_feat.value = __pyx_t_13;

            /* "uharfbuzz/_harfbuzz.pyx":1190
 *                     for start, end, value in value:
 *                         feat.value = value
 *                         feat.start = start             # <<<<<<<<<<<<<<
 *                         feat.end = end
 *                         hb_features[i] = feat
 */
            __pyx_t_11 = __Pyx_PyInt_As_unsigned_int(__pyx_v_start); if (unlikely((__pyx_t_11 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1190, __pyx_L4_error)
            __pyx_v_feat.start = __pyx_t_11;

            /* "uharfbuzz/_harfbuzz.pyx":1191
 *                         feat.value = value
 *                         feat.start = start
 *                         feat.end = end             # <<<<<<<<<<<<<<
 *                         hb_features[i] = feat
 *                         i += 1
 */
            __pyx_t_11 = __Pyx_PyInt_As_unsigned_int(__pyx_v_end); if (unlikely((__pyx_t_11 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1191, __pyx_L4_error)
            __pyx_v_feat.end = __pyx_t_11;

            /* "uharfbuzz/_harfbuzz.pyx":1192
 *                         feat.start = start
 *                         feat.end = end
 *                         hb_features[i] = feat             # <<<<<<<<<<<<<<
 *                         i += 1
 *         if shapers:
 */
            __pyx_t_20 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_20 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1192, __pyx_L4_error)
            (__pyx_v_hb_features[__pyx_t_20]) = __pyx_v_feat;

            /* "uharfbuzz/_harfbuzz.pyx":1193
 *                         feat.end = end
 *                         hb_features[i] = feat
 *                         i += 1             # <<<<<<<<<<<<<<
 *         if shapers:
 *             for i, shaper in enumerate(shapers[:9]):
 */
            __pyx_t_9 = __Pyx_PyInt_AddObjC(__pyx_v_i, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1193, __pyx_L4_error)
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_DECREF_SET(__pyx_v_i, __pyx_t_9);
            __pyx_t_9 = 0;

            /* "uharfbuzz/_harfbuzz.pyx":1188
 *                 else:
 *                     feat.tag = hb_tag_from_string(packed, -1)
 *                     for start, end, value in value:             # <<<<<<<<<<<<<<
 *                         feat.value = value
 *                         feat.start = start
 */
          }
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        }
        __pyx_L12:;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uharfbuzz/_harfbuzz.pyx":1170
 *     hb_features = NULL
 *     try:
 *         if features:             # <<<<<<<<<<<<<<
 *             for value in features.values():
 *                 if isinstance(value, int):
 */
    }

    /* "uharfbuzz/_harfbuzz.pyx":1194
 *                         hb_features[i] = feat
 *                         i += 1
 *         if shapers:             # <<<<<<<<<<<<<<
 *             for i, shaper in enumerate(shapers[:9]):
 *                 packed = shaper.encode()
 */
    __pyx_t_1 = (__pyx_v_shapers != Py_None)&&(PyList_GET_SIZE(__pyx_v_shapers) != 0);
    if (__pyx_t_1) {

      /* "uharfbuzz/_harfbuzz.pyx":1195
 *                         i += 1
 *         if shapers:
 *             for i, shaper in enumerate(shapers[:9]):             # <<<<<<<<<<<<<<
 *                 packed = shaper.encode()
 *                 c_shapers[i] = packed
 */
      __Pyx_INCREF(__pyx_int_0);
      __pyx_t_2 = __pyx_int_0;
      if (unlikely(__pyx_v_shapers == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 1195, __pyx_L4_error)
      }
      __pyx_t_6 = __Pyx_PyList_GetSlice(__pyx_v_shapers, 0, 9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1195, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_9 = __pyx_t_6; __Pyx_INCREF(__pyx_t_9);
      __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      for (;;) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_9);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1195, __pyx_L4_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_6 = PyList_GET_ITEM(__pyx_t_9, __pyx_t_3); __Pyx_INCREF(__pyx_t_6); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 1195, __pyx_L4_error)
        #else
        __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_9, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1195, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_6);
        #endif
        __Pyx_XDECREF_SET(__pyx_v_shaper, __pyx_t_6);
        __pyx_t_6 = 0;
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_2);
        __pyx_t_6 = __Pyx_PyInt_AddObjC(__pyx_t_2, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1195, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_2);
        __pyx_t_2 = __pyx_t_6;
        __pyx_t_6 = 0;

        /* "uharfbuzz/_harfbuzz.pyx":1196
 *         if shapers:
 *             for i, shaper in enumerate(shapers[:9]):
 *                 packed = shaper.encode()             # <<<<<<<<<<<<<<
 *                 c_shapers[i] = packed
 *             c_shapers[i + 1] = NULL
 */
        __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_v_shaper, __pyx_n_s_encode); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1196, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_17);
        __pyx_t_16 = NULL;
        __pyx_t_11 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_17))) {
          __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_17);
          if (likely(__pyx_t_16)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_17);
            __Pyx_INCREF(__pyx_t_16);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_17, function);
            __pyx_t_11 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_16, NULL};
          __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_17, __pyx_callargs+1-__pyx_t_11, 0+__pyx_t_11);
          __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
          if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1196, __pyx_L4_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        }
        if (!(likely(PyBytes_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_6))) __PYX_ERR(0, 1196, __pyx_L4_error)
        __Pyx_XDECREF_SET(__pyx_v_packed, ((PyObject*)__pyx_t_6));
        __pyx_t_6 = 0;

        /* "uharfbuzz/_harfbuzz.pyx":1197
 *             for i, shaper in enumerate(shapers[:9]):
 *                 packed = shaper.encode()
 *                 c_shapers[i] = packed             # <<<<<<<<<<<<<<
 *             c_shapers[i + 1] = NULL
 *             ret = hb_shape_full(font._hb_font, buffer._hb_buffer, hb_features, size, c_shapers)
 */
        if (unlikely(__pyx_v_packed == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
          __PYX_ERR(0, 1197, __pyx_L4_error)
        }
        __pyx_t_21 = __Pyx_PyBytes_AsString(__pyx_v_packed); if (unlikely((!__pyx_t_21) && PyErr_Occurred())) __PYX_ERR(0, 1197, __pyx_L4_error)
        __pyx_t_4 = __Pyx_PyIndex_AsSsize_t(__pyx_v_i); if (unlikely((__pyx_t_4 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1197, __pyx_L4_error)
        (__pyx_v_c_shapers[__pyx_t_4]) = __pyx_t_21;

        /* "uharfbuzz/_harfbuzz.pyx":1195
 *                         i += 1
 *         if shapers:
 *             for i, shaper in enumerate(shapers[:9]):             # <<<<<<<<<<<<<<
 *                 packed = shaper.encode()
 *                 c_shapers[i] = packed
 */
      }
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "uharfbuzz/_harfbuzz.pyx":1198
 *                 packed = shaper.encode()
 *                 c_shapers[i] = packed
 *             c_shapers[i + 1] = NULL             # <<<<<<<<<<<<<<
 *             ret = hb_shape_full(font._hb_font, buffer._hb_buffer, hb_features, size, c_shapers)
 *             if not ret:
 */
      if (unlikely(!__pyx_v_i)) { __Pyx_RaiseUnboundLocalError("i"); __PYX_ERR(0, 1198, __pyx_L4_error) }
      __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_v_i, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1198, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyIndex_AsSsize_t(__pyx_t_2); if (unlikely((__pyx_t_3 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1198, __pyx_L4_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      (__pyx_v_c_shapers[__pyx_t_3]) = NULL;

      /* "uharfbuzz/_harfbuzz.pyx":1199
 *                 c_shapers[i] = packed
 *             c_shapers[i + 1] = NULL
 *             ret = hb_shape_full(font._hb_font, buffer._hb_buffer, hb_features, size, c_shapers)             # <<<<<<<<<<<<<<
 *             if not ret:
 *                 raise RuntimeError("All shapers failed")
 */
      __pyx_v_ret = hb_shape_full(__pyx_v_font->_hb_font, __pyx_v_buffer->_hb_buffer, __pyx_v_hb_features, __pyx_v_size, __pyx_v_c_shapers);

      /* "uharfbuzz/_harfbuzz.pyx":1200
 *             c_shapers[i + 1] = NULL
 *             ret = hb_shape_full(font._hb_font, buffer._hb_buffer, hb_features, size, c_shapers)
 *             if not ret:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError("All shapers failed")
 *         else:
 */
      __pyx_t_1 = (!(__pyx_v_ret != 0));
      if (unlikely(__pyx_t_1)) {

        /* "uharfbuzz/_harfbuzz.pyx":1201
 *             ret = hb_shape_full(font._hb_font, buffer._hb_buffer, hb_features, size, c_shapers)
 *             if not ret:
 *                 raise RuntimeError("All shapers failed")             # <<<<<<<<<<<<<<
 *         else:
 *             hb_shape(font._hb_font, buffer._hb_buffer, hb_features, size)
 */
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1201, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_Raise(__pyx_t_2, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __PYX_ERR(0, 1201, __pyx_L4_error)

        /* "uharfbuzz/_harfbuzz.pyx":1200
 *             c_shapers[i + 1] = NULL
 *             ret = hb_shape_full(font._hb_font, buffer._hb_buffer, hb_features, size, c_shapers)
 *             if not ret:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError("All shapers failed")
 *         else:
 */
      }

      /* "uharfbuzz/_harfbuzz.pyx":1194
 *                         hb_features[i] = feat
 *                         i += 1
 *         if shapers:             # <<<<<<<<<<<<<<
 *             for i, shaper in enumerate(shapers[:9]):
 *                 packed = shaper.encode()
 */
      goto __pyx_L18;
    }

    /* "uharfbuzz/_harfbuzz.pyx":1203
 *                 raise RuntimeError("All shapers failed")
 *         else:
 *             hb_shape(font._hb_font, buffer._hb_buffer, hb_features, size)             # <<<<<<<<<<<<<<
 *         if not hb_buffer_allocation_successful(buffer._hb_buffer):
 *             raise MemoryError()
 */
    /*else*/ {
      hb_shape(__pyx_v_font->_hb_font, __pyx_v_buffer->_hb_buffer, __pyx_v_hb_features, __pyx_v_size);
    }
    __pyx_L18:;

    /* "uharfbuzz/_harfbuzz.pyx":1204
 *         else:
 *             hb_shape(font._hb_font, buffer._hb_buffer, hb_features, size)
 *         if not hb_buffer_allocation_successful(buffer._hb_buffer):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *     finally:
 */
    __pyx_t_1 = (!(hb_buffer_allocation_successful(__pyx_v_buffer->_hb_buffer) != 0));
    if (unlikely(__pyx_t_1)) {

      /* "uharfbuzz/_harfbuzz.pyx":1205
 *             hb_shape(font._hb_font, buffer._hb_buffer, hb_features, size)
 *         if not hb_buffer_allocation_successful(buffer._hb_buffer):
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 *     finally:
 *         if hb_features is not NULL:
 */
      PyErr_NoMemory(); __PYX_ERR(0, 1205, __pyx_L4_error)

      /* "uharfbuzz/_harfbuzz.pyx":1204
 *         else:
 *             hb_shape(font._hb_font, buffer._hb_buffer, hb_features, size)
 *         if not hb_buffer_allocation_successful(buffer._hb_buffer):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *     finally:
 */
    }
  }

  /* "uharfbuzz/_harfbuzz.pyx":1207
 *             raise MemoryError()
 *     finally:
 *         if hb_features is not NULL:             # <<<<<<<<<<<<<<
 *             free(hb_features)
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_1 = (__pyx_v_hb_features != NULL);
      if (__pyx_t_1) {

        /* "uharfbuzz/_harfbuzz.pyx":1208
 *     finally:
 *         if hb_features is not NULL:
 *             free(hb_features)             # <<<<<<<<<<<<<<
 * 
 * 
 */
        free(__pyx_v_hb_features);

        /* "uharfbuzz/_harfbuzz.pyx":1207
 *             raise MemoryError()
 *     finally:
 *         if hb_features is not NULL:             # <<<<<<<<<<<<<<
 *             free(hb_features)
 * 
 */
      }
      goto __pyx_L5;
    }
    __pyx_L4_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_23 = 0; __pyx_t_24 = 0; __pyx_t_25 = 0; __pyx_t_26 = 0; __pyx_t_27 = 0; __pyx_t_28 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
      __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
      __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_26, &__pyx_t_27, &__pyx_t_28);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_23, &__pyx_t_24, &__pyx_t_25) < 0)) __Pyx_ErrFetch(&__pyx_t_23, &__pyx_t_24, &__pyx_t_25);
      __Pyx_XGOTREF(__pyx_t_23);
      __Pyx_XGOTREF(__pyx_t_24);
      __Pyx_XGOTREF(__pyx_t_25);
      __Pyx_XGOTREF(__pyx_t_26);
      __Pyx_XGOTREF(__pyx_t_27);
      __Pyx_XGOTREF(__pyx_t_28);
      __pyx_t_5 = __pyx_lineno; __pyx_t_7 = __pyx_clineno; __pyx_t_22 = __pyx_filename;
      {
        __pyx_t_1 = (__pyx_v_hb_features != NULL);
        if (__pyx_t_1) {

          /* "uharfbuzz/_harfbuzz.pyx":1208
 *     finally:
 *         if hb_features is not NULL:
 *             free(hb_features)             # <<<<<<<<<<<<<<
 * 
 * 
 */
          free(__pyx_v_hb_features);

          /* "uharfbuzz/_harfbuzz.pyx":1207
 *             raise MemoryError()
 *     finally:
 *         if hb_features is not NULL:             # <<<<<<<<<<<<<<
 *             free(hb_features)
 * 
 */
        }
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_26);
        __Pyx_XGIVEREF(__pyx_t_27);
        __Pyx_XGIVEREF(__pyx_t_28);
        __Pyx_ExceptionReset(__pyx_t_26, __pyx_t_27, __pyx_t_28);
      }
      __Pyx_XGIVEREF(__pyx_t_23);
      __Pyx_XGIVEREF(__pyx_t_24);
      __Pyx_XGIVEREF(__pyx_t_25);
      __Pyx_ErrRestore(__pyx_t_23, __pyx_t_24, __pyx_t_25);
      __pyx_t_23 = 0; __pyx_t_24 = 0; __pyx_t_25 = 0; __pyx_t_26 = 0; __pyx_t_27 = 0; __pyx_t_28 = 0;
      __pyx_lineno = __pyx_t_5; __pyx_clineno = __pyx_t_7; __pyx_filename = __pyx_t_22;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "uharfbuzz/_harfbuzz.pyx":1159
 *         self._font_v_extents_func = func
 * 
 * def shape(font: Font, buffer: Buffer,             # <<<<<<<<<<<<<<
 *         features: Dict[str,Union[int,bool,Sequence[Tuple[int,int,Union[int,bool]]]]] = None,
 *         shapers: List[str] = None) -> None:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.shape", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_packed);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XDECREF(__pyx_v_start);
  __Pyx_XDECREF(__pyx_v_end);
  __Pyx_XDECREF(__pyx_v_shaper);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1211
 * 
 * 
 * def ot_tag_to_script(tag: str) -> str:             # <<<<<<<<<<<<<<
 *     cdef bytes packed = tag.encode()
 *     cdef hb_tag_t hb_tag = hb_tag_from_string(<char*>packed, -1)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_5ot_tag_to_script(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_5ot_tag_to_script = {"ot_tag_to_script", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_5ot_tag_to_script, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_5ot_tag_to_script(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_tag = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_tag_to_script (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_tag,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_tag)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1211, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_tag_to_script") < 0)) __PYX_ERR(0, 1211, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_tag = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_tag_to_script", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1211, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_tag_to_script", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_tag), (&PyUnicode_Type), 0, "tag", 1))) __PYX_ERR(0, 1211, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_4ot_tag_to_script(__pyx_self, __pyx_v_tag);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_4ot_tag_to_script(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_tag) {
  PyObject *__pyx_v_packed = 0;
  hb_tag_t __pyx_v_hb_tag;
  hb_script_t __pyx_v_hb_script;
  char __pyx_v_cstr[5];
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_tag_to_script", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1212
 * 
 * def ot_tag_to_script(tag: str) -> str:
 *     cdef bytes packed = tag.encode()             # <<<<<<<<<<<<<<
 *     cdef hb_tag_t hb_tag = hb_tag_from_string(<char*>packed, -1)
 *     cdef hb_script_t hb_script = hb_ot_tag_to_script(hb_tag)
 */
  __pyx_t_1 = PyUnicode_AsEncodedString(__pyx_v_tag, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_packed = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1213
 * def ot_tag_to_script(tag: str) -> str:
 *     cdef bytes packed = tag.encode()
 *     cdef hb_tag_t hb_tag = hb_tag_from_string(<char*>packed, -1)             # <<<<<<<<<<<<<<
 *     cdef hb_script_t hb_script = hb_ot_tag_to_script(hb_tag)
 *     cdef char cstr[5]
 */
  if (unlikely(__pyx_v_packed == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1213, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBytes_AsWritableString(__pyx_v_packed); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 1213, __pyx_L1_error)
  __pyx_v_hb_tag = hb_tag_from_string(((char *)__pyx_t_2), -1);

  /* "uharfbuzz/_harfbuzz.pyx":1214
 *     cdef bytes packed = tag.encode()
 *     cdef hb_tag_t hb_tag = hb_tag_from_string(<char*>packed, -1)
 *     cdef hb_script_t hb_script = hb_ot_tag_to_script(hb_tag)             # <<<<<<<<<<<<<<
 *     cdef char cstr[5]
 *     hb_tag_to_string(hb_script, cstr)
 */
  __pyx_v_hb_script = hb_ot_tag_to_script(__pyx_v_hb_tag);

  /* "uharfbuzz/_harfbuzz.pyx":1216
 *     cdef hb_script_t hb_script = hb_ot_tag_to_script(hb_tag)
 *     cdef char cstr[5]
 *     hb_tag_to_string(hb_script, cstr)             # <<<<<<<<<<<<<<
 *     cstr[4] = b'\0'
 *     packed = cstr
 */
  hb_tag_to_string(__pyx_v_hb_script, __pyx_v_cstr);

  /* "uharfbuzz/_harfbuzz.pyx":1217
 *     cdef char cstr[5]
 *     hb_tag_to_string(hb_script, cstr)
 *     cstr[4] = b'\0'             # <<<<<<<<<<<<<<
 *     packed = cstr
 *     return packed.decode()
 */
  (__pyx_v_cstr[4]) = '\x00';

  /* "uharfbuzz/_harfbuzz.pyx":1218
 *     hb_tag_to_string(hb_script, cstr)
 *     cstr[4] = b'\0'
 *     packed = cstr             # <<<<<<<<<<<<<<
 *     return packed.decode()
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_FromString(__pyx_v_cstr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_packed, ((PyObject*)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1219
 *     cstr[4] = b'\0'
 *     packed = cstr
 *     return packed.decode()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_decode_bytes(__pyx_v_packed, 0, PY_SSIZE_T_MAX, NULL, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1211
 * 
 * 
 * def ot_tag_to_script(tag: str) -> str:             # <<<<<<<<<<<<<<
 *     cdef bytes packed = tag.encode()
 *     cdef hb_tag_t hb_tag = hb_tag_from_string(<char*>packed, -1)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_tag_to_script", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_packed);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1222
 * 
 * 
 * def ot_tag_to_language(tag: str) -> str:             # <<<<<<<<<<<<<<
 *     cdef bytes packed = tag.encode()
 *     cdef hb_tag_t hb_tag = hb_tag_from_string(<char*>packed, -1)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_7ot_tag_to_language(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_7ot_tag_to_language = {"ot_tag_to_language", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_7ot_tag_to_language, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_7ot_tag_to_language(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_tag = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_tag_to_language (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_tag,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_tag)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1222, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_tag_to_language") < 0)) __PYX_ERR(0, 1222, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_tag = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_tag_to_language", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1222, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_tag_to_language", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_tag), (&PyUnicode_Type), 0, "tag", 1))) __PYX_ERR(0, 1222, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_6ot_tag_to_language(__pyx_self, __pyx_v_tag);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_6ot_tag_to_language(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_tag) {
  PyObject *__pyx_v_packed = 0;
  hb_tag_t __pyx_v_hb_tag;
  hb_language_t __pyx_v_hb_language;
  const char *__pyx_v_cstr;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_tag_to_language", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1223
 * 
 * def ot_tag_to_language(tag: str) -> str:
 *     cdef bytes packed = tag.encode()             # <<<<<<<<<<<<<<
 *     cdef hb_tag_t hb_tag = hb_tag_from_string(<char*>packed, -1)
 *     cdef hb_language_t hb_language = hb_ot_tag_to_language(hb_tag)
 */
  __pyx_t_1 = PyUnicode_AsEncodedString(__pyx_v_tag, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_packed = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1224
 * def ot_tag_to_language(tag: str) -> str:
 *     cdef bytes packed = tag.encode()
 *     cdef hb_tag_t hb_tag = hb_tag_from_string(<char*>packed, -1)             # <<<<<<<<<<<<<<
 *     cdef hb_language_t hb_language = hb_ot_tag_to_language(hb_tag)
 *     cdef const_char* cstr = hb_language_to_string(hb_language)
 */
  if (unlikely(__pyx_v_packed == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1224, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBytes_AsWritableString(__pyx_v_packed); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 1224, __pyx_L1_error)
  __pyx_v_hb_tag = hb_tag_from_string(((char *)__pyx_t_2), -1);

  /* "uharfbuzz/_harfbuzz.pyx":1225
 *     cdef bytes packed = tag.encode()
 *     cdef hb_tag_t hb_tag = hb_tag_from_string(<char*>packed, -1)
 *     cdef hb_language_t hb_language = hb_ot_tag_to_language(hb_tag)             # <<<<<<<<<<<<<<
 *     cdef const_char* cstr = hb_language_to_string(hb_language)
 *     if cstr is NULL:
 */
  __pyx_v_hb_language = hb_ot_tag_to_language(__pyx_v_hb_tag);

  /* "uharfbuzz/_harfbuzz.pyx":1226
 *     cdef hb_tag_t hb_tag = hb_tag_from_string(<char*>packed, -1)
 *     cdef hb_language_t hb_language = hb_ot_tag_to_language(hb_tag)
 *     cdef const_char* cstr = hb_language_to_string(hb_language)             # <<<<<<<<<<<<<<
 *     if cstr is NULL:
 *         return None
 */
  __pyx_v_cstr = hb_language_to_string(__pyx_v_hb_language);

  /* "uharfbuzz/_harfbuzz.pyx":1227
 *     cdef hb_language_t hb_language = hb_ot_tag_to_language(hb_tag)
 *     cdef const_char* cstr = hb_language_to_string(hb_language)
 *     if cstr is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     packed = cstr
 */
  __pyx_t_3 = (__pyx_v_cstr == NULL);
  if (__pyx_t_3) {

    /* "uharfbuzz/_harfbuzz.pyx":1228
 *     cdef const_char* cstr = hb_language_to_string(hb_language)
 *     if cstr is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     packed = cstr
 *     return packed.decode()
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = ((PyObject*)Py_None); __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uharfbuzz/_harfbuzz.pyx":1227
 *     cdef hb_language_t hb_language = hb_ot_tag_to_language(hb_tag)
 *     cdef const_char* cstr = hb_language_to_string(hb_language)
 *     if cstr is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     packed = cstr
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":1229
 *     if cstr is NULL:
 *         return None
 *     packed = cstr             # <<<<<<<<<<<<<<
 *     return packed.decode()
 * 
 */
  __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_cstr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1229, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_packed, ((PyObject*)__pyx_t_1));
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1230
 *         return None
 *     packed = cstr
 *     return packed.decode()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_decode_bytes(__pyx_v_packed, 0, PY_SSIZE_T_MAX, NULL, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1222
 * 
 * 
 * def ot_tag_to_language(tag: str) -> str:             # <<<<<<<<<<<<<<
 *     cdef bytes packed = tag.encode()
 *     cdef hb_tag_t hb_tag = hb_tag_from_string(<char*>packed, -1)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_tag_to_language", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_packed);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1233
 * 
 * 
 * def ot_layout_lookup_get_glyph_alternates(             # <<<<<<<<<<<<<<
 *         face: Face, lookup_index : int, glyph : hb_codepoint_t) -> List[int]:
 *     cdef list alternates = []
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9ot_layout_lookup_get_glyph_alternates(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_9ot_layout_lookup_get_glyph_alternates = {"ot_layout_lookup_get_glyph_alternates", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9ot_layout_lookup_get_glyph_alternates, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9ot_layout_lookup_get_glyph_alternates(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face = 0;
  PyObject *__pyx_v_lookup_index = 0;
  hb_codepoint_t __pyx_v_glyph;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_layout_lookup_get_glyph_alternates (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_face,&__pyx_n_s_lookup_index,&__pyx_n_s_glyph,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_face)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1233, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_lookup_index)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1233, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("ot_layout_lookup_get_glyph_alternates", 1, 3, 3, 1); __PYX_ERR(0, 1233, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_glyph)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1233, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("ot_layout_lookup_get_glyph_alternates", 1, 3, 3, 2); __PYX_ERR(0, 1233, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_layout_lookup_get_glyph_alternates") < 0)) __PYX_ERR(0, 1233, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)values[0]);
    __pyx_v_lookup_index = ((PyObject*)values[1]);
    __pyx_v_glyph = __Pyx_PyInt_As_hb_codepoint_t(values[2]); if (unlikely((__pyx_v_glyph == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1234, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_layout_lookup_get_glyph_alternates", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1233, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_layout_lookup_get_glyph_alternates", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_face), __pyx_ptype_9uharfbuzz_9_harfbuzz_Face, 0, "face", 0))) __PYX_ERR(0, 1234, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_lookup_index), (&PyInt_Type), 0, "lookup_index", 1))) __PYX_ERR(0, 1234, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_8ot_layout_lookup_get_glyph_alternates(__pyx_self, __pyx_v_face, __pyx_v_lookup_index, __pyx_v_glyph);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_8ot_layout_lookup_get_glyph_alternates(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, PyObject *__pyx_v_lookup_index, hb_codepoint_t __pyx_v_glyph) {
  PyObject *__pyx_v_alternates = 0;
  unsigned int __pyx_v_i;
  unsigned int __pyx_v_start_offset;
  unsigned int __pyx_v_alternate_count;
  hb_codepoint_t __pyx_v_alternate_glyphs[0x80];
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  unsigned int __pyx_t_3;
  unsigned int __pyx_t_4;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_layout_lookup_get_glyph_alternates", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1235
 * def ot_layout_lookup_get_glyph_alternates(
 *         face: Face, lookup_index : int, glyph : hb_codepoint_t) -> List[int]:
 *     cdef list alternates = []             # <<<<<<<<<<<<<<
 *     cdef unsigned int i
 *     cdef unsigned int start_offset = 0
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_alternates = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1237
 *     cdef list alternates = []
 *     cdef unsigned int i
 *     cdef unsigned int start_offset = 0             # <<<<<<<<<<<<<<
 *     cdef unsigned int alternate_count = STATIC_ARRAY_SIZE
 *     cdef hb_codepoint_t alternate_glyphs[STATIC_ARRAY_SIZE]
 */
  __pyx_v_start_offset = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1238
 *     cdef unsigned int i
 *     cdef unsigned int start_offset = 0
 *     cdef unsigned int alternate_count = STATIC_ARRAY_SIZE             # <<<<<<<<<<<<<<
 *     cdef hb_codepoint_t alternate_glyphs[STATIC_ARRAY_SIZE]
 *     while alternate_count == STATIC_ARRAY_SIZE:
 */
  __pyx_v_alternate_count = 0x80;

  /* "uharfbuzz/_harfbuzz.pyx":1240
 *     cdef unsigned int alternate_count = STATIC_ARRAY_SIZE
 *     cdef hb_codepoint_t alternate_glyphs[STATIC_ARRAY_SIZE]
 *     while alternate_count == STATIC_ARRAY_SIZE:             # <<<<<<<<<<<<<<
 *         hb_ot_layout_lookup_get_glyph_alternates(face._hb_face, lookup_index, glyph, start_offset,
 *             &alternate_count, alternate_glyphs)
 */
  while (1) {
    __pyx_t_2 = (__pyx_v_alternate_count == 0x80);
    if (!__pyx_t_2) break;

    /* "uharfbuzz/_harfbuzz.pyx":1241
 *     cdef hb_codepoint_t alternate_glyphs[STATIC_ARRAY_SIZE]
 *     while alternate_count == STATIC_ARRAY_SIZE:
 *         hb_ot_layout_lookup_get_glyph_alternates(face._hb_face, lookup_index, glyph, start_offset,             # <<<<<<<<<<<<<<
 *             &alternate_count, alternate_glyphs)
 *         for i in range(alternate_count):
 */
    __pyx_t_3 = __Pyx_PyInt_As_unsigned_int(__pyx_v_lookup_index); if (unlikely((__pyx_t_3 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1241, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":1242
 *     while alternate_count == STATIC_ARRAY_SIZE:
 *         hb_ot_layout_lookup_get_glyph_alternates(face._hb_face, lookup_index, glyph, start_offset,
 *             &alternate_count, alternate_glyphs)             # <<<<<<<<<<<<<<
 *         for i in range(alternate_count):
 *             alternates.append(alternate_glyphs[i])
 */
    (void)(hb_ot_layout_lookup_get_glyph_alternates(__pyx_v_face->_hb_face, __pyx_t_3, __pyx_v_glyph, __pyx_v_start_offset, (&__pyx_v_alternate_count), __pyx_v_alternate_glyphs));

    /* "uharfbuzz/_harfbuzz.pyx":1243
 *         hb_ot_layout_lookup_get_glyph_alternates(face._hb_face, lookup_index, glyph, start_offset,
 *             &alternate_count, alternate_glyphs)
 *         for i in range(alternate_count):             # <<<<<<<<<<<<<<
 *             alternates.append(alternate_glyphs[i])
 *         start_offset += alternate_count
 */
    __pyx_t_3 = __pyx_v_alternate_count;
    __pyx_t_4 = __pyx_t_3;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "uharfbuzz/_harfbuzz.pyx":1244
 *             &alternate_count, alternate_glyphs)
 *         for i in range(alternate_count):
 *             alternates.append(alternate_glyphs[i])             # <<<<<<<<<<<<<<
 *         start_offset += alternate_count
 *     return alternates
 */
      __pyx_t_1 = __Pyx_PyInt_From_hb_codepoint_t((__pyx_v_alternate_glyphs[__pyx_v_i])); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1244, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_alternates, __pyx_t_1); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1244, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }

    /* "uharfbuzz/_harfbuzz.pyx":1245
 *         for i in range(alternate_count):
 *             alternates.append(alternate_glyphs[i])
 *         start_offset += alternate_count             # <<<<<<<<<<<<<<
 *     return alternates
 * 
 */
    __pyx_v_start_offset = (__pyx_v_start_offset + __pyx_v_alternate_count);
  }

  /* "uharfbuzz/_harfbuzz.pyx":1246
 *             alternates.append(alternate_glyphs[i])
 *         start_offset += alternate_count
 *     return alternates             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_alternates);
  __pyx_r = __pyx_v_alternates;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1233
 * 
 * 
 * def ot_layout_lookup_get_glyph_alternates(             # <<<<<<<<<<<<<<
 *         face: Face, lookup_index : int, glyph : hb_codepoint_t) -> List[int]:
 *     cdef list alternates = []
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_layout_lookup_get_glyph_alternates", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_alternates);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1249
 * 
 * 
 * def ot_layout_language_get_feature_tags(             # <<<<<<<<<<<<<<
 *         face: Face, tag: str, script_index: int = 0,
 *         language_index: int = 0xFFFF) -> List[str]:
 */

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_84__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_script_index);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_script_index);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_script_index)) __PYX_ERR(0, 1249, __pyx_L1_error);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_language_index);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_language_index);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self)->__pyx_arg_language_index)) __PYX_ERR(0, 1249, __pyx_L1_error);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(0, 1249, __pyx_L1_error);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None)) __PYX_ERR(0, 1249, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_11ot_layout_language_get_feature_tags(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_11ot_layout_language_get_feature_tags = {"ot_layout_language_get_feature_tags", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_11ot_layout_language_get_feature_tags, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_11ot_layout_language_get_feature_tags(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face = 0;
  PyObject *__pyx_v_tag = 0;
  PyObject *__pyx_v_script_index = 0;
  PyObject *__pyx_v_language_index = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_layout_language_get_feature_tags (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_face,&__pyx_n_s_tag,&__pyx_n_s_script_index,&__pyx_n_s_language_index,0};
    __pyx_defaults *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_self);
    values[2] = __Pyx_Arg_NewRef_FASTCALL(__pyx_dynamic_args->__pyx_arg_script_index);
    values[3] = __Pyx_Arg_NewRef_FASTCALL(__pyx_dynamic_args->__pyx_arg_language_index);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_face)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1249, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_tag)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1249, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("ot_layout_language_get_feature_tags", 0, 2, 4, 1); __PYX_ERR(0, 1249, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_script_index);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1249, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_language_index);
          if (value) { values[3] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1249, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_layout_language_get_feature_tags") < 0)) __PYX_ERR(0, 1249, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)values[0]);
    __pyx_v_tag = ((PyObject*)values[1]);
    __pyx_v_script_index = ((PyObject*)values[2]);
    __pyx_v_language_index = ((PyObject*)values[3]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_layout_language_get_feature_tags", 0, 2, 4, __pyx_nargs); __PYX_ERR(0, 1249, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_layout_language_get_feature_tags", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_face), __pyx_ptype_9uharfbuzz_9_harfbuzz_Face, 0, "face", 0))) __PYX_ERR(0, 1250, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_tag), (&PyUnicode_Type), 0, "tag", 1))) __PYX_ERR(0, 1250, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_script_index), (&PyInt_Type), 0, "script_index", 1))) __PYX_ERR(0, 1250, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_language_index), (&PyInt_Type), 0, "language_index", 1))) __PYX_ERR(0, 1251, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_10ot_layout_language_get_feature_tags(__pyx_self, __pyx_v_face, __pyx_v_tag, __pyx_v_script_index, __pyx_v_language_index);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10ot_layout_language_get_feature_tags(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, PyObject *__pyx_v_tag, PyObject *__pyx_v_script_index, PyObject *__pyx_v_language_index) {
  PyObject *__pyx_v_packed = 0;
  hb_tag_t __pyx_v_hb_tag;
  unsigned int __pyx_v_feature_count;
  hb_tag_t __pyx_v_feature_tags[0x80];
  PyObject *__pyx_v_tags = 0;
  char __pyx_v_cstr[5];
  unsigned int __pyx_v_i;
  unsigned int __pyx_v_start_offset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  int __pyx_t_3;
  unsigned int __pyx_t_4;
  unsigned int __pyx_t_5;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_layout_language_get_feature_tags", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1252
 *         face: Face, tag: str, script_index: int = 0,
 *         language_index: int = 0xFFFF) -> List[str]:
 *     cdef bytes packed = tag.encode()             # <<<<<<<<<<<<<<
 *     cdef hb_tag_t hb_tag = hb_tag_from_string(<char*>packed, -1)
 *     cdef unsigned int feature_count = STATIC_ARRAY_SIZE
 */
  __pyx_t_1 = PyUnicode_AsEncodedString(__pyx_v_tag, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_packed = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1253
 *         language_index: int = 0xFFFF) -> List[str]:
 *     cdef bytes packed = tag.encode()
 *     cdef hb_tag_t hb_tag = hb_tag_from_string(<char*>packed, -1)             # <<<<<<<<<<<<<<
 *     cdef unsigned int feature_count = STATIC_ARRAY_SIZE
 *     cdef hb_tag_t feature_tags[STATIC_ARRAY_SIZE]
 */
  if (unlikely(__pyx_v_packed == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1253, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBytes_AsWritableString(__pyx_v_packed); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 1253, __pyx_L1_error)
  __pyx_v_hb_tag = hb_tag_from_string(((char *)__pyx_t_2), -1);

  /* "uharfbuzz/_harfbuzz.pyx":1254
 *     cdef bytes packed = tag.encode()
 *     cdef hb_tag_t hb_tag = hb_tag_from_string(<char*>packed, -1)
 *     cdef unsigned int feature_count = STATIC_ARRAY_SIZE             # <<<<<<<<<<<<<<
 *     cdef hb_tag_t feature_tags[STATIC_ARRAY_SIZE]
 *     cdef list tags = []
 */
  __pyx_v_feature_count = 0x80;

  /* "uharfbuzz/_harfbuzz.pyx":1256
 *     cdef unsigned int feature_count = STATIC_ARRAY_SIZE
 *     cdef hb_tag_t feature_tags[STATIC_ARRAY_SIZE]
 *     cdef list tags = []             # <<<<<<<<<<<<<<
 *     cdef char cstr[5]
 *     cdef unsigned int i
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_tags = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1259
 *     cdef char cstr[5]
 *     cdef unsigned int i
 *     cdef unsigned int start_offset = 0             # <<<<<<<<<<<<<<
 *     while feature_count == STATIC_ARRAY_SIZE:
 *         hb_ot_layout_language_get_feature_tags(
 */
  __pyx_v_start_offset = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1260
 *     cdef unsigned int i
 *     cdef unsigned int start_offset = 0
 *     while feature_count == STATIC_ARRAY_SIZE:             # <<<<<<<<<<<<<<
 *         hb_ot_layout_language_get_feature_tags(
 *             face._hb_face, hb_tag, script_index, language_index, start_offset, &feature_count,
 */
  while (1) {
    __pyx_t_3 = (__pyx_v_feature_count == 0x80);
    if (!__pyx_t_3) break;

    /* "uharfbuzz/_harfbuzz.pyx":1262
 *     while feature_count == STATIC_ARRAY_SIZE:
 *         hb_ot_layout_language_get_feature_tags(
 *             face._hb_face, hb_tag, script_index, language_index, start_offset, &feature_count,             # <<<<<<<<<<<<<<
 *             feature_tags)
 *         for i in range(feature_count):
 */
    __pyx_t_4 = __Pyx_PyInt_As_unsigned_int(__pyx_v_script_index); if (unlikely((__pyx_t_4 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1262, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyInt_As_unsigned_int(__pyx_v_language_index); if (unlikely((__pyx_t_5 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1262, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":1261
 *     cdef unsigned int start_offset = 0
 *     while feature_count == STATIC_ARRAY_SIZE:
 *         hb_ot_layout_language_get_feature_tags(             # <<<<<<<<<<<<<<
 *             face._hb_face, hb_tag, script_index, language_index, start_offset, &feature_count,
 *             feature_tags)
 */
    (void)(hb_ot_layout_language_get_feature_tags(__pyx_v_face->_hb_face, __pyx_v_hb_tag, __pyx_t_4, __pyx_t_5, __pyx_v_start_offset, (&__pyx_v_feature_count), __pyx_v_feature_tags));

    /* "uharfbuzz/_harfbuzz.pyx":1264
 *             face._hb_face, hb_tag, script_index, language_index, start_offset, &feature_count,
 *             feature_tags)
 *         for i in range(feature_count):             # <<<<<<<<<<<<<<
 *             hb_tag_to_string(feature_tags[i], cstr)
 *             cstr[4] = b'\0'
 */
    __pyx_t_5 = __pyx_v_feature_count;
    __pyx_t_4 = __pyx_t_5;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_4; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "uharfbuzz/_harfbuzz.pyx":1265
 *             feature_tags)
 *         for i in range(feature_count):
 *             hb_tag_to_string(feature_tags[i], cstr)             # <<<<<<<<<<<<<<
 *             cstr[4] = b'\0'
 *             packed = cstr
 */
      hb_tag_to_string((__pyx_v_feature_tags[__pyx_v_i]), __pyx_v_cstr);

      /* "uharfbuzz/_harfbuzz.pyx":1266
 *         for i in range(feature_count):
 *             hb_tag_to_string(feature_tags[i], cstr)
 *             cstr[4] = b'\0'             # <<<<<<<<<<<<<<
 *             packed = cstr
 *             tags.append(packed.decode())
 */
      (__pyx_v_cstr[4]) = '\x00';

      /* "uharfbuzz/_harfbuzz.pyx":1267
 *             hb_tag_to_string(feature_tags[i], cstr)
 *             cstr[4] = b'\0'
 *             packed = cstr             # <<<<<<<<<<<<<<
 *             tags.append(packed.decode())
 *         start_offset += feature_count
 */
      __pyx_t_1 = __Pyx_PyObject_FromString(__pyx_v_cstr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1267, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_packed, ((PyObject*)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "uharfbuzz/_harfbuzz.pyx":1268
 *             cstr[4] = b'\0'
 *             packed = cstr
 *             tags.append(packed.decode())             # <<<<<<<<<<<<<<
 *         start_offset += feature_count
 *     return tags
 */
      __pyx_t_1 = __Pyx_decode_bytes(__pyx_v_packed, 0, PY_SSIZE_T_MAX, NULL, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1268, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_tags, __pyx_t_1); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 1268, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }

    /* "uharfbuzz/_harfbuzz.pyx":1269
 *             packed = cstr
 *             tags.append(packed.decode())
 *         start_offset += feature_count             # <<<<<<<<<<<<<<
 *     return tags
 * 
 */
    __pyx_v_start_offset = (__pyx_v_start_offset + __pyx_v_feature_count);
  }

  /* "uharfbuzz/_harfbuzz.pyx":1270
 *             tags.append(packed.decode())
 *         start_offset += feature_count
 *     return tags             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_tags);
  __pyx_r = __pyx_v_tags;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1249
 * 
 * 
 * def ot_layout_language_get_feature_tags(             # <<<<<<<<<<<<<<
 *         face: Face, tag: str, script_index: int = 0,
 *         language_index: int = 0xFFFF) -> List[str]:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_layout_language_get_feature_tags", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_packed);
  __Pyx_XDECREF(__pyx_v_tags);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1273
 * 
 * 
 * def ot_layout_script_get_language_tags(             # <<<<<<<<<<<<<<
 *         face: Face, tag: str, script_index: int = 0) -> List[str]:
 *     cdef bytes packed = tag.encode()
 */

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_86__defaults__(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__defaults__", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_self)->__pyx_arg_script_index);
  __Pyx_GIVEREF(__Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_self)->__pyx_arg_script_index);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_self)->__pyx_arg_script_index)) __PYX_ERR(0, 1273, __pyx_L1_error);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(0, 1273, __pyx_L1_error);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, Py_None)) __PYX_ERR(0, 1273, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.__defaults__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_13ot_layout_script_get_language_tags(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_13ot_layout_script_get_language_tags = {"ot_layout_script_get_language_tags", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_13ot_layout_script_get_language_tags, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_13ot_layout_script_get_language_tags(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face = 0;
  PyObject *__pyx_v_tag = 0;
  PyObject *__pyx_v_script_index = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_layout_script_get_language_tags (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_face,&__pyx_n_s_tag,&__pyx_n_s_script_index,0};
    __pyx_defaults1 *__pyx_dynamic_args = __Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_self);
    values[2] = __Pyx_Arg_NewRef_FASTCALL(__pyx_dynamic_args->__pyx_arg_script_index);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_face)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1273, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_tag)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1273, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("ot_layout_script_get_language_tags", 0, 2, 3, 1); __PYX_ERR(0, 1273, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_script_index);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1273, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_layout_script_get_language_tags") < 0)) __PYX_ERR(0, 1273, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)values[0]);
    __pyx_v_tag = ((PyObject*)values[1]);
    __pyx_v_script_index = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_layout_script_get_language_tags", 0, 2, 3, __pyx_nargs); __PYX_ERR(0, 1273, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_layout_script_get_language_tags", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_face), __pyx_ptype_9uharfbuzz_9_harfbuzz_Face, 0, "face", 0))) __PYX_ERR(0, 1274, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_tag), (&PyUnicode_Type), 0, "tag", 1))) __PYX_ERR(0, 1274, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_script_index), (&PyInt_Type), 0, "script_index", 1))) __PYX_ERR(0, 1274, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_12ot_layout_script_get_language_tags(__pyx_self, __pyx_v_face, __pyx_v_tag, __pyx_v_script_index);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_12ot_layout_script_get_language_tags(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, PyObject *__pyx_v_tag, PyObject *__pyx_v_script_index) {
  PyObject *__pyx_v_packed = 0;
  hb_tag_t __pyx_v_hb_tag;
  unsigned int __pyx_v_language_count;
  hb_tag_t __pyx_v_language_tags[0x80];
  PyObject *__pyx_v_tags = 0;
  char __pyx_v_cstr[5];
  unsigned int __pyx_v_i;
  unsigned int __pyx_v_start_offset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  int __pyx_t_3;
  unsigned int __pyx_t_4;
  unsigned int __pyx_t_5;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_layout_script_get_language_tags", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1275
 * def ot_layout_script_get_language_tags(
 *         face: Face, tag: str, script_index: int = 0) -> List[str]:
 *     cdef bytes packed = tag.encode()             # <<<<<<<<<<<<<<
 *     cdef hb_tag_t hb_tag = hb_tag_from_string(<char*>packed, -1)
 *     cdef unsigned int language_count = STATIC_ARRAY_SIZE
 */
  __pyx_t_1 = PyUnicode_AsEncodedString(__pyx_v_tag, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1275, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_packed = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1276
 *         face: Face, tag: str, script_index: int = 0) -> List[str]:
 *     cdef bytes packed = tag.encode()
 *     cdef hb_tag_t hb_tag = hb_tag_from_string(<char*>packed, -1)             # <<<<<<<<<<<<<<
 *     cdef unsigned int language_count = STATIC_ARRAY_SIZE
 *     cdef hb_tag_t language_tags[STATIC_ARRAY_SIZE]
 */
  if (unlikely(__pyx_v_packed == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1276, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBytes_AsWritableString(__pyx_v_packed); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 1276, __pyx_L1_error)
  __pyx_v_hb_tag = hb_tag_from_string(((char *)__pyx_t_2), -1);

  /* "uharfbuzz/_harfbuzz.pyx":1277
 *     cdef bytes packed = tag.encode()
 *     cdef hb_tag_t hb_tag = hb_tag_from_string(<char*>packed, -1)
 *     cdef unsigned int language_count = STATIC_ARRAY_SIZE             # <<<<<<<<<<<<<<
 *     cdef hb_tag_t language_tags[STATIC_ARRAY_SIZE]
 *     cdef list tags = []
 */
  __pyx_v_language_count = 0x80;

  /* "uharfbuzz/_harfbuzz.pyx":1279
 *     cdef unsigned int language_count = STATIC_ARRAY_SIZE
 *     cdef hb_tag_t language_tags[STATIC_ARRAY_SIZE]
 *     cdef list tags = []             # <<<<<<<<<<<<<<
 *     cdef char cstr[5]
 *     cdef unsigned int i
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_tags = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1282
 *     cdef char cstr[5]
 *     cdef unsigned int i
 *     cdef unsigned int start_offset = 0             # <<<<<<<<<<<<<<
 *     while language_count == STATIC_ARRAY_SIZE:
 *         hb_ot_layout_script_get_language_tags(
 */
  __pyx_v_start_offset = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1283
 *     cdef unsigned int i
 *     cdef unsigned int start_offset = 0
 *     while language_count == STATIC_ARRAY_SIZE:             # <<<<<<<<<<<<<<
 *         hb_ot_layout_script_get_language_tags(
 *             face._hb_face, hb_tag, script_index, start_offset, &language_count, language_tags)
 */
  while (1) {
    __pyx_t_3 = (__pyx_v_language_count == 0x80);
    if (!__pyx_t_3) break;

    /* "uharfbuzz/_harfbuzz.pyx":1285
 *     while language_count == STATIC_ARRAY_SIZE:
 *         hb_ot_layout_script_get_language_tags(
 *             face._hb_face, hb_tag, script_index, start_offset, &language_count, language_tags)             # <<<<<<<<<<<<<<
 *         for i in range(language_count):
 *             hb_tag_to_string(language_tags[i], cstr)
 */
    __pyx_t_4 = __Pyx_PyInt_As_unsigned_int(__pyx_v_script_index); if (unlikely((__pyx_t_4 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1285, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":1284
 *     cdef unsigned int start_offset = 0
 *     while language_count == STATIC_ARRAY_SIZE:
 *         hb_ot_layout_script_get_language_tags(             # <<<<<<<<<<<<<<
 *             face._hb_face, hb_tag, script_index, start_offset, &language_count, language_tags)
 *         for i in range(language_count):
 */
    (void)(hb_ot_layout_script_get_language_tags(__pyx_v_face->_hb_face, __pyx_v_hb_tag, __pyx_t_4, __pyx_v_start_offset, (&__pyx_v_language_count), __pyx_v_language_tags));

    /* "uharfbuzz/_harfbuzz.pyx":1286
 *         hb_ot_layout_script_get_language_tags(
 *             face._hb_face, hb_tag, script_index, start_offset, &language_count, language_tags)
 *         for i in range(language_count):             # <<<<<<<<<<<<<<
 *             hb_tag_to_string(language_tags[i], cstr)
 *             cstr[4] = b'\0'
 */
    __pyx_t_4 = __pyx_v_language_count;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "uharfbuzz/_harfbuzz.pyx":1287
 *             face._hb_face, hb_tag, script_index, start_offset, &language_count, language_tags)
 *         for i in range(language_count):
 *             hb_tag_to_string(language_tags[i], cstr)             # <<<<<<<<<<<<<<
 *             cstr[4] = b'\0'
 *             packed = cstr
 */
      hb_tag_to_string((__pyx_v_language_tags[__pyx_v_i]), __pyx_v_cstr);

      /* "uharfbuzz/_harfbuzz.pyx":1288
 *         for i in range(language_count):
 *             hb_tag_to_string(language_tags[i], cstr)
 *             cstr[4] = b'\0'             # <<<<<<<<<<<<<<
 *             packed = cstr
 *             tags.append(packed.decode())
 */
      (__pyx_v_cstr[4]) = '\x00';

      /* "uharfbuzz/_harfbuzz.pyx":1289
 *             hb_tag_to_string(language_tags[i], cstr)
 *             cstr[4] = b'\0'
 *             packed = cstr             # <<<<<<<<<<<<<<
 *             tags.append(packed.decode())
 *         start_offset += language_count
 */
      __pyx_t_1 = __Pyx_PyObject_FromString(__pyx_v_cstr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1289, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_packed, ((PyObject*)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "uharfbuzz/_harfbuzz.pyx":1290
 *             cstr[4] = b'\0'
 *             packed = cstr
 *             tags.append(packed.decode())             # <<<<<<<<<<<<<<
 *         start_offset += language_count
 *     return tags
 */
      __pyx_t_1 = __Pyx_decode_bytes(__pyx_v_packed, 0, PY_SSIZE_T_MAX, NULL, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1290, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_tags, __pyx_t_1); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 1290, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }

    /* "uharfbuzz/_harfbuzz.pyx":1291
 *             packed = cstr
 *             tags.append(packed.decode())
 *         start_offset += language_count             # <<<<<<<<<<<<<<
 *     return tags
 * 
 */
    __pyx_v_start_offset = (__pyx_v_start_offset + __pyx_v_language_count);
  }

  /* "uharfbuzz/_harfbuzz.pyx":1292
 *             tags.append(packed.decode())
 *         start_offset += language_count
 *     return tags             # <<<<<<<<<<<<<<
 * 
 * def ot_layout_table_get_script_tags(face: Face, tag: str) -> List[str]:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_tags);
  __pyx_r = __pyx_v_tags;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1273
 * 
 * 
 * def ot_layout_script_get_language_tags(             # <<<<<<<<<<<<<<
 *         face: Face, tag: str, script_index: int = 0) -> List[str]:
 *     cdef bytes packed = tag.encode()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_layout_script_get_language_tags", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_packed);
  __Pyx_XDECREF(__pyx_v_tags);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1294
 *     return tags
 * 
 * def ot_layout_table_get_script_tags(face: Face, tag: str) -> List[str]:             # <<<<<<<<<<<<<<
 *     cdef bytes packed = tag.encode()
 *     cdef hb_tag_t hb_tag = hb_tag_from_string(<char*>packed, -1)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_15ot_layout_table_get_script_tags(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_15ot_layout_table_get_script_tags = {"ot_layout_table_get_script_tags", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_15ot_layout_table_get_script_tags, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_15ot_layout_table_get_script_tags(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face = 0;
  PyObject *__pyx_v_tag = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_layout_table_get_script_tags (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_face,&__pyx_n_s_tag,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_face)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1294, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_tag)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1294, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("ot_layout_table_get_script_tags", 1, 2, 2, 1); __PYX_ERR(0, 1294, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_layout_table_get_script_tags") < 0)) __PYX_ERR(0, 1294, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)values[0]);
    __pyx_v_tag = ((PyObject*)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_layout_table_get_script_tags", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1294, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_layout_table_get_script_tags", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_face), __pyx_ptype_9uharfbuzz_9_harfbuzz_Face, 0, "face", 0))) __PYX_ERR(0, 1294, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_tag), (&PyUnicode_Type), 0, "tag", 1))) __PYX_ERR(0, 1294, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_14ot_layout_table_get_script_tags(__pyx_self, __pyx_v_face, __pyx_v_tag);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_14ot_layout_table_get_script_tags(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, PyObject *__pyx_v_tag) {
  PyObject *__pyx_v_packed = 0;
  hb_tag_t __pyx_v_hb_tag;
  unsigned int __pyx_v_script_count;
  hb_tag_t __pyx_v_script_tags[0x80];
  PyObject *__pyx_v_tags = 0;
  char __pyx_v_cstr[5];
  unsigned int __pyx_v_i;
  unsigned int __pyx_v_start_offset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  int __pyx_t_3;
  unsigned int __pyx_t_4;
  unsigned int __pyx_t_5;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_layout_table_get_script_tags", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1295
 * 
 * def ot_layout_table_get_script_tags(face: Face, tag: str) -> List[str]:
 *     cdef bytes packed = tag.encode()             # <<<<<<<<<<<<<<
 *     cdef hb_tag_t hb_tag = hb_tag_from_string(<char*>packed, -1)
 *     cdef unsigned int script_count = STATIC_ARRAY_SIZE
 */
  __pyx_t_1 = PyUnicode_AsEncodedString(__pyx_v_tag, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_packed = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1296
 * def ot_layout_table_get_script_tags(face: Face, tag: str) -> List[str]:
 *     cdef bytes packed = tag.encode()
 *     cdef hb_tag_t hb_tag = hb_tag_from_string(<char*>packed, -1)             # <<<<<<<<<<<<<<
 *     cdef unsigned int script_count = STATIC_ARRAY_SIZE
 *     cdef hb_tag_t script_tags[STATIC_ARRAY_SIZE]
 */
  if (unlikely(__pyx_v_packed == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1296, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBytes_AsWritableString(__pyx_v_packed); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 1296, __pyx_L1_error)
  __pyx_v_hb_tag = hb_tag_from_string(((char *)__pyx_t_2), -1);

  /* "uharfbuzz/_harfbuzz.pyx":1297
 *     cdef bytes packed = tag.encode()
 *     cdef hb_tag_t hb_tag = hb_tag_from_string(<char*>packed, -1)
 *     cdef unsigned int script_count = STATIC_ARRAY_SIZE             # <<<<<<<<<<<<<<
 *     cdef hb_tag_t script_tags[STATIC_ARRAY_SIZE]
 *     cdef list tags = []
 */
  __pyx_v_script_count = 0x80;

  /* "uharfbuzz/_harfbuzz.pyx":1299
 *     cdef unsigned int script_count = STATIC_ARRAY_SIZE
 *     cdef hb_tag_t script_tags[STATIC_ARRAY_SIZE]
 *     cdef list tags = []             # <<<<<<<<<<<<<<
 *     cdef char cstr[5]
 *     cdef unsigned int i
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_tags = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1302
 *     cdef char cstr[5]
 *     cdef unsigned int i
 *     cdef unsigned int start_offset = 0             # <<<<<<<<<<<<<<
 *     while script_count == STATIC_ARRAY_SIZE:
 *         hb_ot_layout_table_get_script_tags(
 */
  __pyx_v_start_offset = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1303
 *     cdef unsigned int i
 *     cdef unsigned int start_offset = 0
 *     while script_count == STATIC_ARRAY_SIZE:             # <<<<<<<<<<<<<<
 *         hb_ot_layout_table_get_script_tags(
 *             face._hb_face, hb_tag, start_offset, &script_count, script_tags)
 */
  while (1) {
    __pyx_t_3 = (__pyx_v_script_count == 0x80);
    if (!__pyx_t_3) break;

    /* "uharfbuzz/_harfbuzz.pyx":1304
 *     cdef unsigned int start_offset = 0
 *     while script_count == STATIC_ARRAY_SIZE:
 *         hb_ot_layout_table_get_script_tags(             # <<<<<<<<<<<<<<
 *             face._hb_face, hb_tag, start_offset, &script_count, script_tags)
 *         for i in range(script_count):
 */
    (void)(hb_ot_layout_table_get_script_tags(__pyx_v_face->_hb_face, __pyx_v_hb_tag, __pyx_v_start_offset, (&__pyx_v_script_count), __pyx_v_script_tags));

    /* "uharfbuzz/_harfbuzz.pyx":1306
 *         hb_ot_layout_table_get_script_tags(
 *             face._hb_face, hb_tag, start_offset, &script_count, script_tags)
 *         for i in range(script_count):             # <<<<<<<<<<<<<<
 *             hb_tag_to_string(script_tags[i], cstr)
 *             cstr[4] = b'\0'
 */
    __pyx_t_4 = __pyx_v_script_count;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "uharfbuzz/_harfbuzz.pyx":1307
 *             face._hb_face, hb_tag, start_offset, &script_count, script_tags)
 *         for i in range(script_count):
 *             hb_tag_to_string(script_tags[i], cstr)             # <<<<<<<<<<<<<<
 *             cstr[4] = b'\0'
 *             packed = cstr
 */
      hb_tag_to_string((__pyx_v_script_tags[__pyx_v_i]), __pyx_v_cstr);

      /* "uharfbuzz/_harfbuzz.pyx":1308
 *         for i in range(script_count):
 *             hb_tag_to_string(script_tags[i], cstr)
 *             cstr[4] = b'\0'             # <<<<<<<<<<<<<<
 *             packed = cstr
 *             tags.append(packed.decode())
 */
      (__pyx_v_cstr[4]) = '\x00';

      /* "uharfbuzz/_harfbuzz.pyx":1309
 *             hb_tag_to_string(script_tags[i], cstr)
 *             cstr[4] = b'\0'
 *             packed = cstr             # <<<<<<<<<<<<<<
 *             tags.append(packed.decode())
 *         start_offset += script_count
 */
      __pyx_t_1 = __Pyx_PyObject_FromString(__pyx_v_cstr); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1309, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_packed, ((PyObject*)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "uharfbuzz/_harfbuzz.pyx":1310
 *             cstr[4] = b'\0'
 *             packed = cstr
 *             tags.append(packed.decode())             # <<<<<<<<<<<<<<
 *         start_offset += script_count
 *     return tags
 */
      __pyx_t_1 = __Pyx_decode_bytes(__pyx_v_packed, 0, PY_SSIZE_T_MAX, NULL, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1310, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_tags, __pyx_t_1); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 1310, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }

    /* "uharfbuzz/_harfbuzz.pyx":1311
 *             packed = cstr
 *             tags.append(packed.decode())
 *         start_offset += script_count             # <<<<<<<<<<<<<<
 *     return tags
 * 
 */
    __pyx_v_start_offset = (__pyx_v_start_offset + __pyx_v_script_count);
  }

  /* "uharfbuzz/_harfbuzz.pyx":1312
 *             tags.append(packed.decode())
 *         start_offset += script_count
 *     return tags             # <<<<<<<<<<<<<<
 * 
 * def ot_layout_get_baseline(font: Font,
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_tags);
  __pyx_r = __pyx_v_tags;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1294
 *     return tags
 * 
 * def ot_layout_table_get_script_tags(face: Face, tag: str) -> List[str]:             # <<<<<<<<<<<<<<
 *     cdef bytes packed = tag.encode()
 *     cdef hb_tag_t hb_tag = hb_tag_from_string(<char*>packed, -1)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_layout_table_get_script_tags", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_packed);
  __Pyx_XDECREF(__pyx_v_tags);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1314
 *     return tags
 * 
 * def ot_layout_get_baseline(font: Font,             # <<<<<<<<<<<<<<
 *                            baseline_tag: str,
 *                            direction: str,
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_17ot_layout_get_baseline(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_17ot_layout_get_baseline = {"ot_layout_get_baseline", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_17ot_layout_get_baseline, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_17ot_layout_get_baseline(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font = 0;
  PyObject *__pyx_v_baseline_tag = 0;
  PyObject *__pyx_v_direction = 0;
  PyObject *__pyx_v_script_tag = 0;
  PyObject *__pyx_v_language_tag = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_layout_get_baseline (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_font,&__pyx_n_s_baseline_tag,&__pyx_n_s_direction,&__pyx_n_s_script_tag,&__pyx_n_s_language_tag,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  5: values[4] = __Pyx_Arg_FASTCALL(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_font)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1314, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_baseline_tag)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1314, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("ot_layout_get_baseline", 1, 5, 5, 1); __PYX_ERR(0, 1314, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_direction)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1314, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("ot_layout_get_baseline", 1, 5, 5, 2); __PYX_ERR(0, 1314, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_script_tag)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[3]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1314, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("ot_layout_get_baseline", 1, 5, 5, 3); __PYX_ERR(0, 1314, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_language_tag)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[4]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1314, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("ot_layout_get_baseline", 1, 5, 5, 4); __PYX_ERR(0, 1314, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_layout_get_baseline") < 0)) __PYX_ERR(0, 1314, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
      values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
      values[4] = __Pyx_Arg_FASTCALL(__pyx_args, 4);
    }
    __pyx_v_font = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)values[0]);
    __pyx_v_baseline_tag = ((PyObject*)values[1]);
    __pyx_v_direction = ((PyObject*)values[2]);
    __pyx_v_script_tag = ((PyObject*)values[3]);
    __pyx_v_language_tag = ((PyObject*)values[4]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_layout_get_baseline", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 1314, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_layout_get_baseline", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_font), __pyx_ptype_9uharfbuzz_9_harfbuzz_Font, 0, "font", 0))) __PYX_ERR(0, 1314, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_baseline_tag), (&PyUnicode_Type), 0, "baseline_tag", 1))) __PYX_ERR(0, 1315, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_direction), (&PyUnicode_Type), 0, "direction", 1))) __PYX_ERR(0, 1316, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_script_tag), (&PyUnicode_Type), 0, "script_tag", 1))) __PYX_ERR(0, 1317, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_language_tag), (&PyUnicode_Type), 0, "language_tag", 1))) __PYX_ERR(0, 1318, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_16ot_layout_get_baseline(__pyx_self, __pyx_v_font, __pyx_v_baseline_tag, __pyx_v_direction, __pyx_v_script_tag, __pyx_v_language_tag);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_16ot_layout_get_baseline(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font, PyObject *__pyx_v_baseline_tag, PyObject *__pyx_v_direction, PyObject *__pyx_v_script_tag, PyObject *__pyx_v_language_tag) {
  hb_ot_layout_baseline_tag_t __pyx_v_hb_baseline_tag;
  hb_direction_t __pyx_v_hb_direction;
  hb_tag_t __pyx_v_hb_script_tag;
  hb_tag_t __pyx_v_hb_language_tag;
  hb_position_t __pyx_v_hb_position;
  hb_bool_t __pyx_v_success;
  PyObject *__pyx_v_packed = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_UCS4 __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  char *__pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_layout_get_baseline", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1327
 *     cdef bytes packed
 * 
 *     if baseline_tag == "romn":             # <<<<<<<<<<<<<<
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_ROMAN
 *     elif baseline_tag == "hang":
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_baseline_tag, __pyx_n_u_romn, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1327, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":1328
 * 
 *     if baseline_tag == "romn":
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_ROMAN             # <<<<<<<<<<<<<<
 *     elif baseline_tag == "hang":
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_HANGING
 */
    __pyx_v_hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_ROMAN;

    /* "uharfbuzz/_harfbuzz.pyx":1327
 *     cdef bytes packed
 * 
 *     if baseline_tag == "romn":             # <<<<<<<<<<<<<<
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_ROMAN
 *     elif baseline_tag == "hang":
 */
    goto __pyx_L3;
  }

  /* "uharfbuzz/_harfbuzz.pyx":1329
 *     if baseline_tag == "romn":
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_ROMAN
 *     elif baseline_tag == "hang":             # <<<<<<<<<<<<<<
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_HANGING
 *     elif baseline_tag == "icfb":
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_baseline_tag, __pyx_n_u_hang, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1329, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":1330
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_ROMAN
 *     elif baseline_tag == "hang":
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_HANGING             # <<<<<<<<<<<<<<
 *     elif baseline_tag == "icfb":
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_IDEO_FACE_BOTTOM_OR_LEFT
 */
    __pyx_v_hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_HANGING;

    /* "uharfbuzz/_harfbuzz.pyx":1329
 *     if baseline_tag == "romn":
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_ROMAN
 *     elif baseline_tag == "hang":             # <<<<<<<<<<<<<<
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_HANGING
 *     elif baseline_tag == "icfb":
 */
    goto __pyx_L3;
  }

  /* "uharfbuzz/_harfbuzz.pyx":1331
 *     elif baseline_tag == "hang":
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_HANGING
 *     elif baseline_tag == "icfb":             # <<<<<<<<<<<<<<
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_IDEO_FACE_BOTTOM_OR_LEFT
 *     elif baseline_tag == "icft":
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_baseline_tag, __pyx_n_u_icfb, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1331, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":1332
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_HANGING
 *     elif baseline_tag == "icfb":
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_IDEO_FACE_BOTTOM_OR_LEFT             # <<<<<<<<<<<<<<
 *     elif baseline_tag == "icft":
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_IDEO_FACE_TOP_OR_RIGHT
 */
    __pyx_v_hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_IDEO_FACE_BOTTOM_OR_LEFT;

    /* "uharfbuzz/_harfbuzz.pyx":1331
 *     elif baseline_tag == "hang":
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_HANGING
 *     elif baseline_tag == "icfb":             # <<<<<<<<<<<<<<
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_IDEO_FACE_BOTTOM_OR_LEFT
 *     elif baseline_tag == "icft":
 */
    goto __pyx_L3;
  }

  /* "uharfbuzz/_harfbuzz.pyx":1333
 *     elif baseline_tag == "icfb":
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_IDEO_FACE_BOTTOM_OR_LEFT
 *     elif baseline_tag == "icft":             # <<<<<<<<<<<<<<
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_IDEO_FACE_TOP_OR_RIGHT
 *     elif baseline_tag == "ideo":
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_baseline_tag, __pyx_n_u_icft, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1333, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":1334
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_IDEO_FACE_BOTTOM_OR_LEFT
 *     elif baseline_tag == "icft":
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_IDEO_FACE_TOP_OR_RIGHT             # <<<<<<<<<<<<<<
 *     elif baseline_tag == "ideo":
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_IDEO_EMBOX_BOTTOM_OR_LEFT
 */
    __pyx_v_hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_IDEO_FACE_TOP_OR_RIGHT;

    /* "uharfbuzz/_harfbuzz.pyx":1333
 *     elif baseline_tag == "icfb":
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_IDEO_FACE_BOTTOM_OR_LEFT
 *     elif baseline_tag == "icft":             # <<<<<<<<<<<<<<
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_IDEO_FACE_TOP_OR_RIGHT
 *     elif baseline_tag == "ideo":
 */
    goto __pyx_L3;
  }

  /* "uharfbuzz/_harfbuzz.pyx":1335
 *     elif baseline_tag == "icft":
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_IDEO_FACE_TOP_OR_RIGHT
 *     elif baseline_tag == "ideo":             # <<<<<<<<<<<<<<
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_IDEO_EMBOX_BOTTOM_OR_LEFT
 *     elif baseline_tag == "idtp":
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_baseline_tag, __pyx_n_u_ideo, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1335, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":1336
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_IDEO_FACE_TOP_OR_RIGHT
 *     elif baseline_tag == "ideo":
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_IDEO_EMBOX_BOTTOM_OR_LEFT             # <<<<<<<<<<<<<<
 *     elif baseline_tag == "idtp":
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_IDEO_EMBOX_TOP_OR_RIGHT
 */
    __pyx_v_hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_IDEO_EMBOX_BOTTOM_OR_LEFT;

    /* "uharfbuzz/_harfbuzz.pyx":1335
 *     elif baseline_tag == "icft":
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_IDEO_FACE_TOP_OR_RIGHT
 *     elif baseline_tag == "ideo":             # <<<<<<<<<<<<<<
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_IDEO_EMBOX_BOTTOM_OR_LEFT
 *     elif baseline_tag == "idtp":
 */
    goto __pyx_L3;
  }

  /* "uharfbuzz/_harfbuzz.pyx":1337
 *     elif baseline_tag == "ideo":
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_IDEO_EMBOX_BOTTOM_OR_LEFT
 *     elif baseline_tag == "idtp":             # <<<<<<<<<<<<<<
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_IDEO_EMBOX_TOP_OR_RIGHT
 *     elif baseline_tag == "math":
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_baseline_tag, __pyx_n_u_idtp, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1337, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":1338
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_IDEO_EMBOX_BOTTOM_OR_LEFT
 *     elif baseline_tag == "idtp":
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_IDEO_EMBOX_TOP_OR_RIGHT             # <<<<<<<<<<<<<<
 *     elif baseline_tag == "math":
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_MATH
 */
    __pyx_v_hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_IDEO_EMBOX_TOP_OR_RIGHT;

    /* "uharfbuzz/_harfbuzz.pyx":1337
 *     elif baseline_tag == "ideo":
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_IDEO_EMBOX_BOTTOM_OR_LEFT
 *     elif baseline_tag == "idtp":             # <<<<<<<<<<<<<<
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_IDEO_EMBOX_TOP_OR_RIGHT
 *     elif baseline_tag == "math":
 */
    goto __pyx_L3;
  }

  /* "uharfbuzz/_harfbuzz.pyx":1339
 *     elif baseline_tag == "idtp":
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_IDEO_EMBOX_TOP_OR_RIGHT
 *     elif baseline_tag == "math":             # <<<<<<<<<<<<<<
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_MATH
 *     else:
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_baseline_tag, __pyx_n_u_math, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1339, __pyx_L1_error)
  if (likely(__pyx_t_1)) {

    /* "uharfbuzz/_harfbuzz.pyx":1340
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_IDEO_EMBOX_TOP_OR_RIGHT
 *     elif baseline_tag == "math":
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_MATH             # <<<<<<<<<<<<<<
 *     else:
 *         raise ValueError(f"invalid baseline tag '{baseline_tag}'")
 */
    __pyx_v_hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_MATH;

    /* "uharfbuzz/_harfbuzz.pyx":1339
 *     elif baseline_tag == "idtp":
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_IDEO_EMBOX_TOP_OR_RIGHT
 *     elif baseline_tag == "math":             # <<<<<<<<<<<<<<
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_MATH
 *     else:
 */
    goto __pyx_L3;
  }

  /* "uharfbuzz/_harfbuzz.pyx":1342
 *         hb_baseline_tag = HB_OT_LAYOUT_BASELINE_TAG_MATH
 *     else:
 *         raise ValueError(f"invalid baseline tag '{baseline_tag}'")             # <<<<<<<<<<<<<<
 *     packed = direction.encode()
 *     hb_direction = hb_direction_from_string(<char*>packed, -1)
 */
  /*else*/ {
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1342, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = 0;
    __pyx_t_4 = 127;
    __Pyx_INCREF(__pyx_kp_u_invalid_baseline_tag);
    __pyx_t_3 += 22;
    __Pyx_GIVEREF(__pyx_kp_u_invalid_baseline_tag);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u_invalid_baseline_tag);
    __Pyx_INCREF(__pyx_v_baseline_tag);
    __pyx_t_4 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_v_baseline_tag) > __pyx_t_4) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_v_baseline_tag) : __pyx_t_4;
    __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_v_baseline_tag);
    __Pyx_GIVEREF(__pyx_v_baseline_tag);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_baseline_tag);
    __Pyx_INCREF(__pyx_kp_u__10);
    __pyx_t_3 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__10);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_kp_u__10);
    __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_2, 3, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1342, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1342, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1342, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "uharfbuzz/_harfbuzz.pyx":1343
 *     else:
 *         raise ValueError(f"invalid baseline tag '{baseline_tag}'")
 *     packed = direction.encode()             # <<<<<<<<<<<<<<
 *     hb_direction = hb_direction_from_string(<char*>packed, -1)
 *     packed = script_tag.encode()
 */
  __pyx_t_2 = PyUnicode_AsEncodedString(__pyx_v_direction, NULL, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_packed = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1344
 *         raise ValueError(f"invalid baseline tag '{baseline_tag}'")
 *     packed = direction.encode()
 *     hb_direction = hb_direction_from_string(<char*>packed, -1)             # <<<<<<<<<<<<<<
 *     packed = script_tag.encode()
 *     hb_script_tag = hb_tag_from_string(<char*>packed, -1)
 */
  if (unlikely(__pyx_v_packed == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1344, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyBytes_AsWritableString(__pyx_v_packed); if (unlikely((!__pyx_t_6) && PyErr_Occurred())) __PYX_ERR(0, 1344, __pyx_L1_error)
  __pyx_v_hb_direction = hb_direction_from_string(((char *)__pyx_t_6), -1);

  /* "uharfbuzz/_harfbuzz.pyx":1345
 *     packed = direction.encode()
 *     hb_direction = hb_direction_from_string(<char*>packed, -1)
 *     packed = script_tag.encode()             # <<<<<<<<<<<<<<
 *     hb_script_tag = hb_tag_from_string(<char*>packed, -1)
 *     packed = language_tag.encode()
 */
  __pyx_t_2 = PyUnicode_AsEncodedString(__pyx_v_script_tag, NULL, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF_SET(__pyx_v_packed, ((PyObject*)__pyx_t_2));
  __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1346
 *     hb_direction = hb_direction_from_string(<char*>packed, -1)
 *     packed = script_tag.encode()
 *     hb_script_tag = hb_tag_from_string(<char*>packed, -1)             # <<<<<<<<<<<<<<
 *     packed = language_tag.encode()
 *     hb_language_tag = hb_tag_from_string(<char*>packed, -1)
 */
  if (unlikely(__pyx_v_packed == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1346, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyBytes_AsWritableString(__pyx_v_packed); if (unlikely((!__pyx_t_6) && PyErr_Occurred())) __PYX_ERR(0, 1346, __pyx_L1_error)
  __pyx_v_hb_script_tag = hb_tag_from_string(((char *)__pyx_t_6), -1);

  /* "uharfbuzz/_harfbuzz.pyx":1347
 *     packed = script_tag.encode()
 *     hb_script_tag = hb_tag_from_string(<char*>packed, -1)
 *     packed = language_tag.encode()             # <<<<<<<<<<<<<<
 *     hb_language_tag = hb_tag_from_string(<char*>packed, -1)
 *     success = hb_ot_layout_get_baseline(font._hb_font,
 */
  __pyx_t_2 = PyUnicode_AsEncodedString(__pyx_v_language_tag, NULL, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF_SET(__pyx_v_packed, ((PyObject*)__pyx_t_2));
  __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1348
 *     hb_script_tag = hb_tag_from_string(<char*>packed, -1)
 *     packed = language_tag.encode()
 *     hb_language_tag = hb_tag_from_string(<char*>packed, -1)             # <<<<<<<<<<<<<<
 *     success = hb_ot_layout_get_baseline(font._hb_font,
 *                                         hb_baseline_tag,
 */
  if (unlikely(__pyx_v_packed == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1348, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyBytes_AsWritableString(__pyx_v_packed); if (unlikely((!__pyx_t_6) && PyErr_Occurred())) __PYX_ERR(0, 1348, __pyx_L1_error)
  __pyx_v_hb_language_tag = hb_tag_from_string(((char *)__pyx_t_6), -1);

  /* "uharfbuzz/_harfbuzz.pyx":1349
 *     packed = language_tag.encode()
 *     hb_language_tag = hb_tag_from_string(<char*>packed, -1)
 *     success = hb_ot_layout_get_baseline(font._hb_font,             # <<<<<<<<<<<<<<
 *                                         hb_baseline_tag,
 *                                         hb_direction,
 */
  __pyx_v_success = hb_ot_layout_get_baseline(__pyx_v_font->_hb_font, __pyx_v_hb_baseline_tag, __pyx_v_hb_direction, __pyx_v_hb_script_tag, __pyx_v_hb_language_tag, (&__pyx_v_hb_position));

  /* "uharfbuzz/_harfbuzz.pyx":1355
 *                                         hb_language_tag,
 *                                         &hb_position)
 *     if success:             # <<<<<<<<<<<<<<
 *         return hb_position
 *     else:
 */
  __pyx_t_1 = (__pyx_v_success != 0);
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":1356
 *                                         &hb_position)
 *     if success:
 *         return hb_position             # <<<<<<<<<<<<<<
 *     else:
 *         return None
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_hb_position_t(__pyx_v_hb_position); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1356, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (!(likely(__Pyx_Py3Int_CheckExact(__pyx_t_2)) || __Pyx_RaiseUnexpectedTypeError("int", __pyx_t_2))) __PYX_ERR(0, 1356, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "uharfbuzz/_harfbuzz.pyx":1355
 *                                         hb_language_tag,
 *                                         &hb_position)
 *     if success:             # <<<<<<<<<<<<<<
 *         return hb_position
 *     else:
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":1358
 *         return hb_position
 *     else:
 *         return None             # <<<<<<<<<<<<<<
 * 
 * def ot_layout_has_glyph_classes(face: Face) -> bool:
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = ((PyObject*)Py_None); __Pyx_INCREF(Py_None);
    goto __pyx_L0;
  }

  /* "uharfbuzz/_harfbuzz.pyx":1314
 *     return tags
 * 
 * def ot_layout_get_baseline(font: Font,             # <<<<<<<<<<<<<<
 *                            baseline_tag: str,
 *                            direction: str,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_layout_get_baseline", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_packed);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1360
 *         return None
 * 
 * def ot_layout_has_glyph_classes(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_layout_has_glyph_classes(face._hb_face)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_19ot_layout_has_glyph_classes(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_19ot_layout_has_glyph_classes = {"ot_layout_has_glyph_classes", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_19ot_layout_has_glyph_classes, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_19ot_layout_has_glyph_classes(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_layout_has_glyph_classes (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_face,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_face)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1360, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_layout_has_glyph_classes") < 0)) __PYX_ERR(0, 1360, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_layout_has_glyph_classes", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1360, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_layout_has_glyph_classes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_face), __pyx_ptype_9uharfbuzz_9_harfbuzz_Face, 0, "face", 0))) __PYX_ERR(0, 1360, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_18ot_layout_has_glyph_classes(__pyx_self, __pyx_v_face);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_18ot_layout_has_glyph_classes(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_layout_has_glyph_classes", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1361
 * 
 * def ot_layout_has_glyph_classes(face: Face) -> bool:
 *     return hb_ot_layout_has_glyph_classes(face._hb_face)             # <<<<<<<<<<<<<<
 * 
 * def ot_layout_has_positioning(face: Face) -> bool:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_hb_bool_t(hb_ot_layout_has_glyph_classes(__pyx_v_face->_hb_face)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1361, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1360
 *         return None
 * 
 * def ot_layout_has_glyph_classes(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_layout_has_glyph_classes(face._hb_face)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_layout_has_glyph_classes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1363
 *     return hb_ot_layout_has_glyph_classes(face._hb_face)
 * 
 * def ot_layout_has_positioning(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_layout_has_positioning(face._hb_face)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_21ot_layout_has_positioning(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_21ot_layout_has_positioning = {"ot_layout_has_positioning", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_21ot_layout_has_positioning, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_21ot_layout_has_positioning(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_layout_has_positioning (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_face,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_face)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1363, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_layout_has_positioning") < 0)) __PYX_ERR(0, 1363, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_layout_has_positioning", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1363, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_layout_has_positioning", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_face), __pyx_ptype_9uharfbuzz_9_harfbuzz_Face, 0, "face", 0))) __PYX_ERR(0, 1363, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_20ot_layout_has_positioning(__pyx_self, __pyx_v_face);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_20ot_layout_has_positioning(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_layout_has_positioning", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1364
 * 
 * def ot_layout_has_positioning(face: Face) -> bool:
 *     return hb_ot_layout_has_positioning(face._hb_face)             # <<<<<<<<<<<<<<
 * 
 * def ot_layout_has_substitution(face: Face) -> bool:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_hb_bool_t(hb_ot_layout_has_positioning(__pyx_v_face->_hb_face)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1364, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1363
 *     return hb_ot_layout_has_glyph_classes(face._hb_face)
 * 
 * def ot_layout_has_positioning(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_layout_has_positioning(face._hb_face)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_layout_has_positioning", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1366
 *     return hb_ot_layout_has_positioning(face._hb_face)
 * 
 * def ot_layout_has_substitution(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_layout_has_substitution(face._hb_face)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_23ot_layout_has_substitution(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_23ot_layout_has_substitution = {"ot_layout_has_substitution", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_23ot_layout_has_substitution, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_23ot_layout_has_substitution(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_layout_has_substitution (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_face,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_face)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1366, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_layout_has_substitution") < 0)) __PYX_ERR(0, 1366, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_layout_has_substitution", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1366, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_layout_has_substitution", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_face), __pyx_ptype_9uharfbuzz_9_harfbuzz_Face, 0, "face", 0))) __PYX_ERR(0, 1366, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_22ot_layout_has_substitution(__pyx_self, __pyx_v_face);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_22ot_layout_has_substitution(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_layout_has_substitution", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1367
 * 
 * def ot_layout_has_substitution(face: Face) -> bool:
 *     return hb_ot_layout_has_substitution(face._hb_face)             # <<<<<<<<<<<<<<
 * 
 * class OTLayoutGlyphClass(IntEnum):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_hb_bool_t(hb_ot_layout_has_substitution(__pyx_v_face->_hb_face)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1366
 *     return hb_ot_layout_has_positioning(face._hb_face)
 * 
 * def ot_layout_has_substitution(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_layout_has_substitution(face._hb_face)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_layout_has_substitution", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1376
 *     COMPONENT = HB_OT_LAYOUT_GLYPH_CLASS_COMPONENT
 * 
 * def ot_layout_get_glyph_class(face: Face, glyph: int) -> OTLayoutGlyphClass:             # <<<<<<<<<<<<<<
 *     return OTLayoutGlyphClass(hb_ot_layout_get_glyph_class(face._hb_face, glyph))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_25ot_layout_get_glyph_class(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_25ot_layout_get_glyph_class = {"ot_layout_get_glyph_class", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_25ot_layout_get_glyph_class, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_25ot_layout_get_glyph_class(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face = 0;
  PyObject *__pyx_v_glyph = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_layout_get_glyph_class (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_face,&__pyx_n_s_glyph,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_face)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1376, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_glyph)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1376, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("ot_layout_get_glyph_class", 1, 2, 2, 1); __PYX_ERR(0, 1376, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_layout_get_glyph_class") < 0)) __PYX_ERR(0, 1376, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)values[0]);
    __pyx_v_glyph = ((PyObject*)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_layout_get_glyph_class", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1376, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_layout_get_glyph_class", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_face), __pyx_ptype_9uharfbuzz_9_harfbuzz_Face, 0, "face", 0))) __PYX_ERR(0, 1376, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_glyph), (&PyInt_Type), 0, "glyph", 1))) __PYX_ERR(0, 1376, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_24ot_layout_get_glyph_class(__pyx_self, __pyx_v_face, __pyx_v_glyph);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_24ot_layout_get_glyph_class(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, PyObject *__pyx_v_glyph) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  hb_codepoint_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_layout_get_glyph_class", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1377
 * 
 * def ot_layout_get_glyph_class(face: Face, glyph: int) -> OTLayoutGlyphClass:
 *     return OTLayoutGlyphClass(hb_ot_layout_get_glyph_class(face._hb_face, glyph))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_OTLayoutGlyphClass); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_glyph); if (unlikely((__pyx_t_3 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1377, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyInt_From_hb_ot_layout_glyph_class_t(hb_ot_layout_get_glyph_class(__pyx_v_face->_hb_face, __pyx_t_3)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1377, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1376
 *     COMPONENT = HB_OT_LAYOUT_GLYPH_CLASS_COMPONENT
 * 
 * def ot_layout_get_glyph_class(face: Face, glyph: int) -> OTLayoutGlyphClass:             # <<<<<<<<<<<<<<
 *     return OTLayoutGlyphClass(hb_ot_layout_get_glyph_class(face._hb_face, glyph))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_layout_get_glyph_class", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1380
 * 
 * 
 * def ot_color_has_palettes(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_color_has_palettes(face._hb_face)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_27ot_color_has_palettes(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_27ot_color_has_palettes = {"ot_color_has_palettes", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_27ot_color_has_palettes, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_27ot_color_has_palettes(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_color_has_palettes (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_face,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_face)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1380, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_color_has_palettes") < 0)) __PYX_ERR(0, 1380, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_color_has_palettes", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1380, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_color_has_palettes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_face), __pyx_ptype_9uharfbuzz_9_harfbuzz_Face, 0, "face", 0))) __PYX_ERR(0, 1380, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_26ot_color_has_palettes(__pyx_self, __pyx_v_face);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_26ot_color_has_palettes(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_color_has_palettes", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1381
 * 
 * def ot_color_has_palettes(face: Face) -> bool:
 *     return hb_ot_color_has_palettes(face._hb_face)             # <<<<<<<<<<<<<<
 * 
 * def ot_color_palette_get_count(face: Face) -> int:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_hb_bool_t(hb_ot_color_has_palettes(__pyx_v_face->_hb_face)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1380
 * 
 * 
 * def ot_color_has_palettes(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_color_has_palettes(face._hb_face)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_color_has_palettes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1383
 *     return hb_ot_color_has_palettes(face._hb_face)
 * 
 * def ot_color_palette_get_count(face: Face) -> int:             # <<<<<<<<<<<<<<
 *     return hb_ot_color_palette_get_count(face._hb_face)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_29ot_color_palette_get_count(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_29ot_color_palette_get_count = {"ot_color_palette_get_count", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_29ot_color_palette_get_count, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_29ot_color_palette_get_count(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_color_palette_get_count (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_face,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_face)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1383, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_color_palette_get_count") < 0)) __PYX_ERR(0, 1383, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_color_palette_get_count", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1383, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_color_palette_get_count", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_face), __pyx_ptype_9uharfbuzz_9_harfbuzz_Face, 0, "face", 0))) __PYX_ERR(0, 1383, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_28ot_color_palette_get_count(__pyx_self, __pyx_v_face);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_28ot_color_palette_get_count(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_color_palette_get_count", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1384
 * 
 * def ot_color_palette_get_count(face: Face) -> int:
 *     return hb_ot_color_palette_get_count(face._hb_face)             # <<<<<<<<<<<<<<
 * 
 * class OTColorPaletteFlags(IntFlag):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_int(hb_ot_color_palette_get_count(__pyx_v_face->_hb_face)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1384, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(__Pyx_Py3Int_CheckExact(__pyx_t_1)) || __Pyx_RaiseUnexpectedTypeError("int", __pyx_t_1))) __PYX_ERR(0, 1384, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1383
 *     return hb_ot_color_has_palettes(face._hb_face)
 * 
 * def ot_color_palette_get_count(face: Face) -> int:             # <<<<<<<<<<<<<<
 *     return hb_ot_color_palette_get_count(face._hb_face)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_color_palette_get_count", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1391
 *     USABLE_WITH_DARK_BACKGROUND = HB_OT_COLOR_PALETTE_FLAG_USABLE_WITH_DARK_BACKGROUND
 * 
 * def ot_color_palette_get_flags(face: Face, palette_index: int) -> OTColorPaletteFlags:             # <<<<<<<<<<<<<<
 *     return OTColorPaletteFlags(hb_ot_color_palette_get_flags(face._hb_face, palette_index))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_31ot_color_palette_get_flags(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_31ot_color_palette_get_flags = {"ot_color_palette_get_flags", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_31ot_color_palette_get_flags, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_31ot_color_palette_get_flags(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face = 0;
  PyObject *__pyx_v_palette_index = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_color_palette_get_flags (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_face,&__pyx_n_s_palette_index,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_face)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1391, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_palette_index)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1391, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("ot_color_palette_get_flags", 1, 2, 2, 1); __PYX_ERR(0, 1391, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_color_palette_get_flags") < 0)) __PYX_ERR(0, 1391, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)values[0]);
    __pyx_v_palette_index = ((PyObject*)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_color_palette_get_flags", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1391, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_color_palette_get_flags", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_face), __pyx_ptype_9uharfbuzz_9_harfbuzz_Face, 0, "face", 0))) __PYX_ERR(0, 1391, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_palette_index), (&PyInt_Type), 0, "palette_index", 1))) __PYX_ERR(0, 1391, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_30ot_color_palette_get_flags(__pyx_self, __pyx_v_face, __pyx_v_palette_index);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_30ot_color_palette_get_flags(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, PyObject *__pyx_v_palette_index) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  unsigned int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_color_palette_get_flags", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1392
 * 
 * def ot_color_palette_get_flags(face: Face, palette_index: int) -> OTColorPaletteFlags:
 *     return OTColorPaletteFlags(hb_ot_color_palette_get_flags(face._hb_face, palette_index))             # <<<<<<<<<<<<<<
 * 
 * def ot_color_palette_get_colors(face: Face, palette_index: int) -> List[Color]:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_OTColorPaletteFlags); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_As_unsigned_int(__pyx_v_palette_index); if (unlikely((__pyx_t_3 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1392, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyInt_From_hb_ot_color_palette_flags_t(hb_ot_color_palette_get_flags(__pyx_v_face->_hb_face, __pyx_t_3)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1392, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_3 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_3 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_3, 1+__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1391
 *     USABLE_WITH_DARK_BACKGROUND = HB_OT_COLOR_PALETTE_FLAG_USABLE_WITH_DARK_BACKGROUND
 * 
 * def ot_color_palette_get_flags(face: Face, palette_index: int) -> OTColorPaletteFlags:             # <<<<<<<<<<<<<<
 *     return OTColorPaletteFlags(hb_ot_color_palette_get_flags(face._hb_face, palette_index))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_color_palette_get_flags", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1394
 *     return OTColorPaletteFlags(hb_ot_color_palette_get_flags(face._hb_face, palette_index))
 * 
 * def ot_color_palette_get_colors(face: Face, palette_index: int) -> List[Color]:             # <<<<<<<<<<<<<<
 *     cdef list ret = []
 *     cdef unsigned int i
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_33ot_color_palette_get_colors(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_33ot_color_palette_get_colors = {"ot_color_palette_get_colors", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_33ot_color_palette_get_colors, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_33ot_color_palette_get_colors(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face = 0;
  PyObject *__pyx_v_palette_index = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_color_palette_get_colors (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_face,&__pyx_n_s_palette_index,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_face)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1394, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_palette_index)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1394, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("ot_color_palette_get_colors", 1, 2, 2, 1); __PYX_ERR(0, 1394, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_color_palette_get_colors") < 0)) __PYX_ERR(0, 1394, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)values[0]);
    __pyx_v_palette_index = ((PyObject*)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_color_palette_get_colors", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1394, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_color_palette_get_colors", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_face), __pyx_ptype_9uharfbuzz_9_harfbuzz_Face, 0, "face", 0))) __PYX_ERR(0, 1394, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_palette_index), (&PyInt_Type), 0, "palette_index", 1))) __PYX_ERR(0, 1394, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_32ot_color_palette_get_colors(__pyx_self, __pyx_v_face, __pyx_v_palette_index);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_32ot_color_palette_get_colors(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, PyObject *__pyx_v_palette_index) {
  PyObject *__pyx_v_ret = 0;
  unsigned int __pyx_v_i;
  unsigned int __pyx_v_start_offset;
  unsigned int __pyx_v_color_count;
  hb_color_t __pyx_v_colors[0x80];
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  unsigned int __pyx_t_3;
  unsigned int __pyx_t_4;
  unsigned int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_color_palette_get_colors", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1395
 * 
 * def ot_color_palette_get_colors(face: Face, palette_index: int) -> List[Color]:
 *     cdef list ret = []             # <<<<<<<<<<<<<<
 *     cdef unsigned int i
 *     cdef unsigned int start_offset = 0
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ret = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1397
 *     cdef list ret = []
 *     cdef unsigned int i
 *     cdef unsigned int start_offset = 0             # <<<<<<<<<<<<<<
 *     cdef unsigned int color_count = STATIC_ARRAY_SIZE
 *     cdef hb_color_t colors[STATIC_ARRAY_SIZE]
 */
  __pyx_v_start_offset = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1398
 *     cdef unsigned int i
 *     cdef unsigned int start_offset = 0
 *     cdef unsigned int color_count = STATIC_ARRAY_SIZE             # <<<<<<<<<<<<<<
 *     cdef hb_color_t colors[STATIC_ARRAY_SIZE]
 *     while color_count == STATIC_ARRAY_SIZE:
 */
  __pyx_v_color_count = 0x80;

  /* "uharfbuzz/_harfbuzz.pyx":1400
 *     cdef unsigned int color_count = STATIC_ARRAY_SIZE
 *     cdef hb_color_t colors[STATIC_ARRAY_SIZE]
 *     while color_count == STATIC_ARRAY_SIZE:             # <<<<<<<<<<<<<<
 *         hb_ot_color_palette_get_colors(face._hb_face, palette_index, start_offset, &color_count, colors)
 *         for i in range(color_count):
 */
  while (1) {
    __pyx_t_2 = (__pyx_v_color_count == 0x80);
    if (!__pyx_t_2) break;

    /* "uharfbuzz/_harfbuzz.pyx":1401
 *     cdef hb_color_t colors[STATIC_ARRAY_SIZE]
 *     while color_count == STATIC_ARRAY_SIZE:
 *         hb_ot_color_palette_get_colors(face._hb_face, palette_index, start_offset, &color_count, colors)             # <<<<<<<<<<<<<<
 *         for i in range(color_count):
 *             ret.append(Color.from_int(colors[i]))
 */
    __pyx_t_3 = __Pyx_PyInt_As_unsigned_int(__pyx_v_palette_index); if (unlikely((__pyx_t_3 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1401, __pyx_L1_error)
    (void)(hb_ot_color_palette_get_colors(__pyx_v_face->_hb_face, __pyx_t_3, __pyx_v_start_offset, (&__pyx_v_color_count), __pyx_v_colors));

    /* "uharfbuzz/_harfbuzz.pyx":1402
 *     while color_count == STATIC_ARRAY_SIZE:
 *         hb_ot_color_palette_get_colors(face._hb_face, palette_index, start_offset, &color_count, colors)
 *         for i in range(color_count):             # <<<<<<<<<<<<<<
 *             ret.append(Color.from_int(colors[i]))
 *     return ret
 */
    __pyx_t_3 = __pyx_v_color_count;
    __pyx_t_4 = __pyx_t_3;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "uharfbuzz/_harfbuzz.pyx":1403
 *         hb_ot_color_palette_get_colors(face._hb_face, palette_index, start_offset, &color_count, colors)
 *         for i in range(color_count):
 *             ret.append(Color.from_int(colors[i]))             # <<<<<<<<<<<<<<
 *     return ret
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_Color); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1403, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_from_int); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1403, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyInt_From_uint32_t((__pyx_v_colors[__pyx_v_i])); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1403, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_8 = NULL;
      __pyx_t_9 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_9 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_6};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1403, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_ret, __pyx_t_1); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 1403, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
  }

  /* "uharfbuzz/_harfbuzz.pyx":1404
 *         for i in range(color_count):
 *             ret.append(Color.from_int(colors[i]))
 *     return ret             # <<<<<<<<<<<<<<
 * 
 * def ot_color_palette_get_name_id(face: Face, palette_index: int) -> int | None:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_ret);
  __pyx_r = __pyx_v_ret;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1394
 *     return OTColorPaletteFlags(hb_ot_color_palette_get_flags(face._hb_face, palette_index))
 * 
 * def ot_color_palette_get_colors(face: Face, palette_index: int) -> List[Color]:             # <<<<<<<<<<<<<<
 *     cdef list ret = []
 *     cdef unsigned int i
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_color_palette_get_colors", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1406
 *     return ret
 * 
 * def ot_color_palette_get_name_id(face: Face, palette_index: int) -> int | None:             # <<<<<<<<<<<<<<
 *     cdef hb_ot_name_id_t name_id
 *     name_id = hb_ot_color_palette_get_name_id(face._hb_face, palette_index)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_35ot_color_palette_get_name_id(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_35ot_color_palette_get_name_id = {"ot_color_palette_get_name_id", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_35ot_color_palette_get_name_id, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_35ot_color_palette_get_name_id(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face = 0;
  PyObject *__pyx_v_palette_index = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_color_palette_get_name_id (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_face,&__pyx_n_s_palette_index,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_face)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1406, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_palette_index)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1406, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("ot_color_palette_get_name_id", 1, 2, 2, 1); __PYX_ERR(0, 1406, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_color_palette_get_name_id") < 0)) __PYX_ERR(0, 1406, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)values[0]);
    __pyx_v_palette_index = ((PyObject*)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_color_palette_get_name_id", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1406, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_color_palette_get_name_id", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_face), __pyx_ptype_9uharfbuzz_9_harfbuzz_Face, 0, "face", 0))) __PYX_ERR(0, 1406, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_palette_index), (&PyInt_Type), 0, "palette_index", 1))) __PYX_ERR(0, 1406, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_34ot_color_palette_get_name_id(__pyx_self, __pyx_v_face, __pyx_v_palette_index);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_34ot_color_palette_get_name_id(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, PyObject *__pyx_v_palette_index) {
  hb_ot_name_id_t __pyx_v_name_id;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  unsigned int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_color_palette_get_name_id", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1408
 * def ot_color_palette_get_name_id(face: Face, palette_index: int) -> int | None:
 *     cdef hb_ot_name_id_t name_id
 *     name_id = hb_ot_color_palette_get_name_id(face._hb_face, palette_index)             # <<<<<<<<<<<<<<
 *     if name_id == HB_OT_NAME_ID_INVALID:
 *         return None
 */
  __pyx_t_1 = __Pyx_PyInt_As_unsigned_int(__pyx_v_palette_index); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1408, __pyx_L1_error)
  __pyx_v_name_id = hb_ot_color_palette_get_name_id(__pyx_v_face->_hb_face, __pyx_t_1);

  /* "uharfbuzz/_harfbuzz.pyx":1409
 *     cdef hb_ot_name_id_t name_id
 *     name_id = hb_ot_color_palette_get_name_id(face._hb_face, palette_index)
 *     if name_id == HB_OT_NAME_ID_INVALID:             # <<<<<<<<<<<<<<
 *         return None
 *     return name_id
 */
  __pyx_t_2 = (__pyx_v_name_id == HB_OT_NAME_ID_INVALID);
  if (__pyx_t_2) {

    /* "uharfbuzz/_harfbuzz.pyx":1410
 *     name_id = hb_ot_color_palette_get_name_id(face._hb_face, palette_index)
 *     if name_id == HB_OT_NAME_ID_INVALID:
 *         return None             # <<<<<<<<<<<<<<
 *     return name_id
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uharfbuzz/_harfbuzz.pyx":1409
 *     cdef hb_ot_name_id_t name_id
 *     name_id = hb_ot_color_palette_get_name_id(face._hb_face, palette_index)
 *     if name_id == HB_OT_NAME_ID_INVALID:             # <<<<<<<<<<<<<<
 *         return None
 *     return name_id
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":1411
 *     if name_id == HB_OT_NAME_ID_INVALID:
 *         return None
 *     return name_id             # <<<<<<<<<<<<<<
 * 
 * def ot_color_palette_color_get_name_id(face: Face, color_index: int) -> int | None:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_name_id_t(__pyx_v_name_id); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1406
 *     return ret
 * 
 * def ot_color_palette_get_name_id(face: Face, palette_index: int) -> int | None:             # <<<<<<<<<<<<<<
 *     cdef hb_ot_name_id_t name_id
 *     name_id = hb_ot_color_palette_get_name_id(face._hb_face, palette_index)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_color_palette_get_name_id", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1413
 *     return name_id
 * 
 * def ot_color_palette_color_get_name_id(face: Face, color_index: int) -> int | None:             # <<<<<<<<<<<<<<
 *     cdef hb_ot_name_id_t name_id
 *     name_id =  hb_ot_color_palette_color_get_name_id(face._hb_face, color_index)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_37ot_color_palette_color_get_name_id(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_37ot_color_palette_color_get_name_id = {"ot_color_palette_color_get_name_id", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_37ot_color_palette_color_get_name_id, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_37ot_color_palette_color_get_name_id(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face = 0;
  PyObject *__pyx_v_color_index = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_color_palette_color_get_name_id (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_face,&__pyx_n_s_color_index,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_face)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1413, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_color_index)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1413, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("ot_color_palette_color_get_name_id", 1, 2, 2, 1); __PYX_ERR(0, 1413, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_color_palette_color_get_name_id") < 0)) __PYX_ERR(0, 1413, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)values[0]);
    __pyx_v_color_index = ((PyObject*)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_color_palette_color_get_name_id", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1413, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_color_palette_color_get_name_id", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_face), __pyx_ptype_9uharfbuzz_9_harfbuzz_Face, 0, "face", 0))) __PYX_ERR(0, 1413, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_color_index), (&PyInt_Type), 0, "color_index", 1))) __PYX_ERR(0, 1413, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_36ot_color_palette_color_get_name_id(__pyx_self, __pyx_v_face, __pyx_v_color_index);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_36ot_color_palette_color_get_name_id(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, PyObject *__pyx_v_color_index) {
  hb_ot_name_id_t __pyx_v_name_id;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  unsigned int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_color_palette_color_get_name_id", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1415
 * def ot_color_palette_color_get_name_id(face: Face, color_index: int) -> int | None:
 *     cdef hb_ot_name_id_t name_id
 *     name_id =  hb_ot_color_palette_color_get_name_id(face._hb_face, color_index)             # <<<<<<<<<<<<<<
 *     if name_id == HB_OT_NAME_ID_INVALID:
 *         return None
 */
  __pyx_t_1 = __Pyx_PyInt_As_unsigned_int(__pyx_v_color_index); if (unlikely((__pyx_t_1 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1415, __pyx_L1_error)
  __pyx_v_name_id = hb_ot_color_palette_color_get_name_id(__pyx_v_face->_hb_face, __pyx_t_1);

  /* "uharfbuzz/_harfbuzz.pyx":1416
 *     cdef hb_ot_name_id_t name_id
 *     name_id =  hb_ot_color_palette_color_get_name_id(face._hb_face, color_index)
 *     if name_id == HB_OT_NAME_ID_INVALID:             # <<<<<<<<<<<<<<
 *         return None
 *     return name_id
 */
  __pyx_t_2 = (__pyx_v_name_id == HB_OT_NAME_ID_INVALID);
  if (__pyx_t_2) {

    /* "uharfbuzz/_harfbuzz.pyx":1417
 *     name_id =  hb_ot_color_palette_color_get_name_id(face._hb_face, color_index)
 *     if name_id == HB_OT_NAME_ID_INVALID:
 *         return None             # <<<<<<<<<<<<<<
 *     return name_id
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uharfbuzz/_harfbuzz.pyx":1416
 *     cdef hb_ot_name_id_t name_id
 *     name_id =  hb_ot_color_palette_color_get_name_id(face._hb_face, color_index)
 *     if name_id == HB_OT_NAME_ID_INVALID:             # <<<<<<<<<<<<<<
 *         return None
 *     return name_id
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":1418
 *     if name_id == HB_OT_NAME_ID_INVALID:
 *         return None
 *     return name_id             # <<<<<<<<<<<<<<
 * 
 * def ot_color_has_layers(face: Face) -> bool:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_name_id_t(__pyx_v_name_id); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1418, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1413
 *     return name_id
 * 
 * def ot_color_palette_color_get_name_id(face: Face, color_index: int) -> int | None:             # <<<<<<<<<<<<<<
 *     cdef hb_ot_name_id_t name_id
 *     name_id =  hb_ot_color_palette_color_get_name_id(face._hb_face, color_index)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_color_palette_color_get_name_id", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1420
 *     return name_id
 * 
 * def ot_color_has_layers(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_color_has_layers(face._hb_face)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_39ot_color_has_layers(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_39ot_color_has_layers = {"ot_color_has_layers", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_39ot_color_has_layers, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_39ot_color_has_layers(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_color_has_layers (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_face,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_face)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1420, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_color_has_layers") < 0)) __PYX_ERR(0, 1420, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_color_has_layers", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1420, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_color_has_layers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_face), __pyx_ptype_9uharfbuzz_9_harfbuzz_Face, 0, "face", 0))) __PYX_ERR(0, 1420, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_38ot_color_has_layers(__pyx_self, __pyx_v_face);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_38ot_color_has_layers(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_color_has_layers", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1421
 * 
 * def ot_color_has_layers(face: Face) -> bool:
 *     return hb_ot_color_has_layers(face._hb_face)             # <<<<<<<<<<<<<<
 * 
 * OTColorLayer = namedtuple("OTColorLayer", ["glyph", "color_index"])
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_hb_bool_t(hb_ot_color_has_layers(__pyx_v_face->_hb_face)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1420
 *     return name_id
 * 
 * def ot_color_has_layers(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_color_has_layers(face._hb_face)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_color_has_layers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1425
 * OTColorLayer = namedtuple("OTColorLayer", ["glyph", "color_index"])
 * 
 * def ot_color_glyph_get_layers(face: Face, glyph: int) -> List[OTColorLayer]:             # <<<<<<<<<<<<<<
 *     cdef list ret = []
 *     cdef unsigned int i
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_41ot_color_glyph_get_layers(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_41ot_color_glyph_get_layers = {"ot_color_glyph_get_layers", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_41ot_color_glyph_get_layers, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_41ot_color_glyph_get_layers(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face = 0;
  PyObject *__pyx_v_glyph = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_color_glyph_get_layers (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_face,&__pyx_n_s_glyph,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_face)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1425, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_glyph)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1425, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("ot_color_glyph_get_layers", 1, 2, 2, 1); __PYX_ERR(0, 1425, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_color_glyph_get_layers") < 0)) __PYX_ERR(0, 1425, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)values[0]);
    __pyx_v_glyph = ((PyObject*)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_color_glyph_get_layers", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1425, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_color_glyph_get_layers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_face), __pyx_ptype_9uharfbuzz_9_harfbuzz_Face, 0, "face", 0))) __PYX_ERR(0, 1425, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_glyph), (&PyInt_Type), 0, "glyph", 1))) __PYX_ERR(0, 1425, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_40ot_color_glyph_get_layers(__pyx_self, __pyx_v_face, __pyx_v_glyph);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_40ot_color_glyph_get_layers(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, PyObject *__pyx_v_glyph) {
  PyObject *__pyx_v_ret = 0;
  unsigned int __pyx_v_i;
  unsigned int __pyx_v_start_offset;
  unsigned int __pyx_v_layer_count;
  hb_ot_color_layer_t __pyx_v_layers[0x80];
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  hb_codepoint_t __pyx_t_3;
  unsigned int __pyx_t_4;
  unsigned int __pyx_t_5;
  unsigned int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  unsigned int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_color_glyph_get_layers", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1426
 * 
 * def ot_color_glyph_get_layers(face: Face, glyph: int) -> List[OTColorLayer]:
 *     cdef list ret = []             # <<<<<<<<<<<<<<
 *     cdef unsigned int i
 *     cdef unsigned int start_offset = 0
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1426, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ret = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1428
 *     cdef list ret = []
 *     cdef unsigned int i
 *     cdef unsigned int start_offset = 0             # <<<<<<<<<<<<<<
 *     cdef unsigned int layer_count = STATIC_ARRAY_SIZE
 *     cdef hb_ot_color_layer_t layers[STATIC_ARRAY_SIZE]
 */
  __pyx_v_start_offset = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1429
 *     cdef unsigned int i
 *     cdef unsigned int start_offset = 0
 *     cdef unsigned int layer_count = STATIC_ARRAY_SIZE             # <<<<<<<<<<<<<<
 *     cdef hb_ot_color_layer_t layers[STATIC_ARRAY_SIZE]
 *     while layer_count == STATIC_ARRAY_SIZE:
 */
  __pyx_v_layer_count = 0x80;

  /* "uharfbuzz/_harfbuzz.pyx":1431
 *     cdef unsigned int layer_count = STATIC_ARRAY_SIZE
 *     cdef hb_ot_color_layer_t layers[STATIC_ARRAY_SIZE]
 *     while layer_count == STATIC_ARRAY_SIZE:             # <<<<<<<<<<<<<<
 *         hb_ot_color_glyph_get_layers(face._hb_face, glyph, start_offset, &layer_count, layers)
 *         for i in range(layer_count):
 */
  while (1) {
    __pyx_t_2 = (__pyx_v_layer_count == 0x80);
    if (!__pyx_t_2) break;

    /* "uharfbuzz/_harfbuzz.pyx":1432
 *     cdef hb_ot_color_layer_t layers[STATIC_ARRAY_SIZE]
 *     while layer_count == STATIC_ARRAY_SIZE:
 *         hb_ot_color_glyph_get_layers(face._hb_face, glyph, start_offset, &layer_count, layers)             # <<<<<<<<<<<<<<
 *         for i in range(layer_count):
 *             ret.append(OTColorLayer(layers[i].glyph, layers[i].color_index))
 */
    __pyx_t_3 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_glyph); if (unlikely((__pyx_t_3 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1432, __pyx_L1_error)
    (void)(hb_ot_color_glyph_get_layers(__pyx_v_face->_hb_face, __pyx_t_3, __pyx_v_start_offset, (&__pyx_v_layer_count), __pyx_v_layers));

    /* "uharfbuzz/_harfbuzz.pyx":1433
 *     while layer_count == STATIC_ARRAY_SIZE:
 *         hb_ot_color_glyph_get_layers(face._hb_face, glyph, start_offset, &layer_count, layers)
 *         for i in range(layer_count):             # <<<<<<<<<<<<<<
 *             ret.append(OTColorLayer(layers[i].glyph, layers[i].color_index))
 *         start_offset += layer_count
 */
    __pyx_t_4 = __pyx_v_layer_count;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "uharfbuzz/_harfbuzz.pyx":1434
 *         hb_ot_color_glyph_get_layers(face._hb_face, glyph, start_offset, &layer_count, layers)
 *         for i in range(layer_count):
 *             ret.append(OTColorLayer(layers[i].glyph, layers[i].color_index))             # <<<<<<<<<<<<<<
 *         start_offset += layer_count
 *     return ret
 */
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_OTColorLayer); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1434, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = __Pyx_PyInt_From_hb_codepoint_t((__pyx_v_layers[__pyx_v_i]).glyph); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1434, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = __Pyx_PyInt_From_unsigned_int((__pyx_v_layers[__pyx_v_i]).color_index); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1434, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = NULL;
      __pyx_t_11 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_11 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_9};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_11, 2+__pyx_t_11);
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1434, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_ret, __pyx_t_1); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 1434, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }

    /* "uharfbuzz/_harfbuzz.pyx":1435
 *         for i in range(layer_count):
 *             ret.append(OTColorLayer(layers[i].glyph, layers[i].color_index))
 *         start_offset += layer_count             # <<<<<<<<<<<<<<
 *     return ret
 * 
 */
    __pyx_v_start_offset = (__pyx_v_start_offset + __pyx_v_layer_count);
  }

  /* "uharfbuzz/_harfbuzz.pyx":1436
 *             ret.append(OTColorLayer(layers[i].glyph, layers[i].color_index))
 *         start_offset += layer_count
 *     return ret             # <<<<<<<<<<<<<<
 * 
 * def ot_color_has_paint(face: Face) -> bool:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_ret);
  __pyx_r = __pyx_v_ret;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1425
 * OTColorLayer = namedtuple("OTColorLayer", ["glyph", "color_index"])
 * 
 * def ot_color_glyph_get_layers(face: Face, glyph: int) -> List[OTColorLayer]:             # <<<<<<<<<<<<<<
 *     cdef list ret = []
 *     cdef unsigned int i
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_color_glyph_get_layers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1438
 *     return ret
 * 
 * def ot_color_has_paint(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_color_has_paint(face._hb_face)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_43ot_color_has_paint(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_43ot_color_has_paint = {"ot_color_has_paint", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_43ot_color_has_paint, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_43ot_color_has_paint(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_color_has_paint (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_face,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_face)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1438, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_color_has_paint") < 0)) __PYX_ERR(0, 1438, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_color_has_paint", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1438, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_color_has_paint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_face), __pyx_ptype_9uharfbuzz_9_harfbuzz_Face, 0, "face", 0))) __PYX_ERR(0, 1438, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_42ot_color_has_paint(__pyx_self, __pyx_v_face);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_42ot_color_has_paint(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_color_has_paint", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1439
 * 
 * def ot_color_has_paint(face: Face) -> bool:
 *     return hb_ot_color_has_paint(face._hb_face)             # <<<<<<<<<<<<<<
 * 
 * def ot_color_glyph_has_paint(face: Face, glyph: int) -> bool:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_hb_bool_t(hb_ot_color_has_paint(__pyx_v_face->_hb_face)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1438
 *     return ret
 * 
 * def ot_color_has_paint(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_color_has_paint(face._hb_face)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_color_has_paint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1441
 *     return hb_ot_color_has_paint(face._hb_face)
 * 
 * def ot_color_glyph_has_paint(face: Face, glyph: int) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_color_glyph_has_paint(face._hb_face, glyph)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_45ot_color_glyph_has_paint(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_45ot_color_glyph_has_paint = {"ot_color_glyph_has_paint", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_45ot_color_glyph_has_paint, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_45ot_color_glyph_has_paint(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face = 0;
  PyObject *__pyx_v_glyph = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_color_glyph_has_paint (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_face,&__pyx_n_s_glyph,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_face)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1441, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_glyph)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1441, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("ot_color_glyph_has_paint", 1, 2, 2, 1); __PYX_ERR(0, 1441, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_color_glyph_has_paint") < 0)) __PYX_ERR(0, 1441, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)values[0]);
    __pyx_v_glyph = ((PyObject*)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_color_glyph_has_paint", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1441, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_color_glyph_has_paint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_face), __pyx_ptype_9uharfbuzz_9_harfbuzz_Face, 0, "face", 0))) __PYX_ERR(0, 1441, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_glyph), (&PyInt_Type), 0, "glyph", 1))) __PYX_ERR(0, 1441, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_44ot_color_glyph_has_paint(__pyx_self, __pyx_v_face, __pyx_v_glyph);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_44ot_color_glyph_has_paint(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, PyObject *__pyx_v_glyph) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  hb_codepoint_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_color_glyph_has_paint", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1442
 * 
 * def ot_color_glyph_has_paint(face: Face, glyph: int) -> bool:
 *     return hb_ot_color_glyph_has_paint(face._hb_face, glyph)             # <<<<<<<<<<<<<<
 * 
 * def ot_color_has_svg(face: Face) -> bool:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_glyph); if (unlikely((__pyx_t_1 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1442, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_hb_bool_t(hb_ot_color_glyph_has_paint(__pyx_v_face->_hb_face, __pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1441
 *     return hb_ot_color_has_paint(face._hb_face)
 * 
 * def ot_color_glyph_has_paint(face: Face, glyph: int) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_color_glyph_has_paint(face._hb_face, glyph)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_color_glyph_has_paint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1444
 *     return hb_ot_color_glyph_has_paint(face._hb_face, glyph)
 * 
 * def ot_color_has_svg(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_color_has_svg(face._hb_face)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_47ot_color_has_svg(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_47ot_color_has_svg = {"ot_color_has_svg", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_47ot_color_has_svg, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_47ot_color_has_svg(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_color_has_svg (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_face,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_face)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1444, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_color_has_svg") < 0)) __PYX_ERR(0, 1444, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_color_has_svg", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1444, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_color_has_svg", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_face), __pyx_ptype_9uharfbuzz_9_harfbuzz_Face, 0, "face", 0))) __PYX_ERR(0, 1444, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_46ot_color_has_svg(__pyx_self, __pyx_v_face);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_46ot_color_has_svg(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_color_has_svg", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1445
 * 
 * def ot_color_has_svg(face: Face) -> bool:
 *     return hb_ot_color_has_svg(face._hb_face)             # <<<<<<<<<<<<<<
 * 
 * def ot_color_glyph_get_svg(face: Face, glyph: int) -> Blob:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_hb_bool_t(hb_ot_color_has_svg(__pyx_v_face->_hb_face)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1445, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1444
 *     return hb_ot_color_glyph_has_paint(face._hb_face, glyph)
 * 
 * def ot_color_has_svg(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_color_has_svg(face._hb_face)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_color_has_svg", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1447
 *     return hb_ot_color_has_svg(face._hb_face)
 * 
 * def ot_color_glyph_get_svg(face: Face, glyph: int) -> Blob:             # <<<<<<<<<<<<<<
 *     cdef hb_blob_t* blob
 *     blob = hb_ot_color_glyph_reference_svg(face._hb_face, glyph)
 */

/* Python wrapper */
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *__pyx_pw_9uharfbuzz_9_harfbuzz_49ot_color_glyph_get_svg(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_49ot_color_glyph_get_svg = {"ot_color_glyph_get_svg", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_49ot_color_glyph_get_svg, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *__pyx_pw_9uharfbuzz_9_harfbuzz_49ot_color_glyph_get_svg(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face = 0;
  PyObject *__pyx_v_glyph = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_color_glyph_get_svg (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_face,&__pyx_n_s_glyph,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_face)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1447, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_glyph)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1447, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("ot_color_glyph_get_svg", 1, 2, 2, 1); __PYX_ERR(0, 1447, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_color_glyph_get_svg") < 0)) __PYX_ERR(0, 1447, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)values[0]);
    __pyx_v_glyph = ((PyObject*)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_color_glyph_get_svg", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1447, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_color_glyph_get_svg", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_face), __pyx_ptype_9uharfbuzz_9_harfbuzz_Face, 0, "face", 0))) __PYX_ERR(0, 1447, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_glyph), (&PyInt_Type), 0, "glyph", 1))) __PYX_ERR(0, 1447, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_48ot_color_glyph_get_svg(__pyx_self, __pyx_v_face, __pyx_v_glyph);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *__pyx_pf_9uharfbuzz_9_harfbuzz_48ot_color_glyph_get_svg(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, PyObject *__pyx_v_glyph) {
  hb_blob_t *__pyx_v_blob;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  hb_codepoint_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_color_glyph_get_svg", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1449
 * def ot_color_glyph_get_svg(face: Face, glyph: int) -> Blob:
 *     cdef hb_blob_t* blob
 *     blob = hb_ot_color_glyph_reference_svg(face._hb_face, glyph)             # <<<<<<<<<<<<<<
 *     return Blob.from_ptr(blob)
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_glyph); if (unlikely((__pyx_t_1 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1449, __pyx_L1_error)
  __pyx_v_blob = hb_ot_color_glyph_reference_svg(__pyx_v_face->_hb_face, __pyx_t_1);

  /* "uharfbuzz/_harfbuzz.pyx":1450
 *     cdef hb_blob_t* blob
 *     blob = hb_ot_color_glyph_reference_svg(face._hb_face, glyph)
 *     return Blob.from_ptr(blob)             # <<<<<<<<<<<<<<
 * 
 * def ot_color_has_png(face: Face) -> bool:
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_2 = ((PyObject *)__pyx_f_9uharfbuzz_9_harfbuzz_4Blob_from_ptr(__pyx_v_blob)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1447
 *     return hb_ot_color_has_svg(face._hb_face)
 * 
 * def ot_color_glyph_get_svg(face: Face, glyph: int) -> Blob:             # <<<<<<<<<<<<<<
 *     cdef hb_blob_t* blob
 *     blob = hb_ot_color_glyph_reference_svg(face._hb_face, glyph)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_color_glyph_get_svg", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1452
 *     return Blob.from_ptr(blob)
 * 
 * def ot_color_has_png(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_color_has_png(face._hb_face)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_51ot_color_has_png(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_51ot_color_has_png = {"ot_color_has_png", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_51ot_color_has_png, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_51ot_color_has_png(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_color_has_png (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_face,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_face)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1452, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_color_has_png") < 0)) __PYX_ERR(0, 1452, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_color_has_png", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1452, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_color_has_png", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_face), __pyx_ptype_9uharfbuzz_9_harfbuzz_Face, 0, "face", 0))) __PYX_ERR(0, 1452, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_50ot_color_has_png(__pyx_self, __pyx_v_face);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_50ot_color_has_png(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_color_has_png", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1453
 * 
 * def ot_color_has_png(face: Face) -> bool:
 *     return hb_ot_color_has_png(face._hb_face)             # <<<<<<<<<<<<<<
 * 
 * def ot_color_glyph_get_png(font: Font, glyph: int) -> Blob:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_hb_bool_t(hb_ot_color_has_png(__pyx_v_face->_hb_face)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1452
 *     return Blob.from_ptr(blob)
 * 
 * def ot_color_has_png(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_color_has_png(face._hb_face)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_color_has_png", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1455
 *     return hb_ot_color_has_png(face._hb_face)
 * 
 * def ot_color_glyph_get_png(font: Font, glyph: int) -> Blob:             # <<<<<<<<<<<<<<
 *     cdef hb_blob_t* blob
 *     blob = hb_ot_color_glyph_reference_png(font._hb_font, glyph)
 */

/* Python wrapper */
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *__pyx_pw_9uharfbuzz_9_harfbuzz_53ot_color_glyph_get_png(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_53ot_color_glyph_get_png = {"ot_color_glyph_get_png", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_53ot_color_glyph_get_png, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *__pyx_pw_9uharfbuzz_9_harfbuzz_53ot_color_glyph_get_png(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font = 0;
  PyObject *__pyx_v_glyph = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_color_glyph_get_png (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_font,&__pyx_n_s_glyph,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_font)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1455, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_glyph)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1455, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("ot_color_glyph_get_png", 1, 2, 2, 1); __PYX_ERR(0, 1455, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_color_glyph_get_png") < 0)) __PYX_ERR(0, 1455, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_font = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)values[0]);
    __pyx_v_glyph = ((PyObject*)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_color_glyph_get_png", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1455, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_color_glyph_get_png", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_font), __pyx_ptype_9uharfbuzz_9_harfbuzz_Font, 0, "font", 0))) __PYX_ERR(0, 1455, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_glyph), (&PyInt_Type), 0, "glyph", 1))) __PYX_ERR(0, 1455, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_52ot_color_glyph_get_png(__pyx_self, __pyx_v_font, __pyx_v_glyph);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *__pyx_pf_9uharfbuzz_9_harfbuzz_52ot_color_glyph_get_png(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font, PyObject *__pyx_v_glyph) {
  hb_blob_t *__pyx_v_blob;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  hb_codepoint_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_color_glyph_get_png", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1457
 * def ot_color_glyph_get_png(font: Font, glyph: int) -> Blob:
 *     cdef hb_blob_t* blob
 *     blob = hb_ot_color_glyph_reference_png(font._hb_font, glyph)             # <<<<<<<<<<<<<<
 *     return Blob.from_ptr(blob)
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_glyph); if (unlikely((__pyx_t_1 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1457, __pyx_L1_error)
  __pyx_v_blob = hb_ot_color_glyph_reference_png(__pyx_v_font->_hb_font, __pyx_t_1);

  /* "uharfbuzz/_harfbuzz.pyx":1458
 *     cdef hb_blob_t* blob
 *     blob = hb_ot_color_glyph_reference_png(font._hb_font, glyph)
 *     return Blob.from_ptr(blob)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_2 = ((PyObject *)__pyx_f_9uharfbuzz_9_harfbuzz_4Blob_from_ptr(__pyx_v_blob)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1458, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1455
 *     return hb_ot_color_has_png(face._hb_face)
 * 
 * def ot_color_glyph_get_png(font: Font, glyph: int) -> Blob:             # <<<<<<<<<<<<<<
 *     cdef hb_blob_t* blob
 *     blob = hb_ot_color_glyph_reference_png(font._hb_font, glyph)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_color_glyph_get_png", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1461
 * 
 * 
 * def ot_math_has_data(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_math_has_data(face._hb_face)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_55ot_math_has_data(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_55ot_math_has_data = {"ot_math_has_data", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_55ot_math_has_data, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_55ot_math_has_data(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_math_has_data (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_face,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_face)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1461, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_math_has_data") < 0)) __PYX_ERR(0, 1461, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_math_has_data", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1461, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_math_has_data", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_face), __pyx_ptype_9uharfbuzz_9_harfbuzz_Face, 0, "face", 0))) __PYX_ERR(0, 1461, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_54ot_math_has_data(__pyx_self, __pyx_v_face);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_54ot_math_has_data(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_math_has_data", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1462
 * 
 * def ot_math_has_data(face: Face) -> bool:
 *     return hb_ot_math_has_data(face._hb_face)             # <<<<<<<<<<<<<<
 * 
 * class OTMathConstant(IntEnum):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_hb_bool_t(hb_ot_math_has_data(__pyx_v_face->_hb_face)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1461
 * 
 * 
 * def ot_math_has_data(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_math_has_data(face._hb_face)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_math_has_data", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1522
 *     RADICAL_DEGREE_BOTTOM_RAISE_PERCENT = HB_OT_MATH_CONSTANT_RADICAL_DEGREE_BOTTOM_RAISE_PERCENT
 * 
 * def ot_math_get_constant(font: Font, constant: OTMathConstant) -> int:             # <<<<<<<<<<<<<<
 *     if constant >= len(OTMathConstant):
 *         raise ValueError("invalid constant")
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_57ot_math_get_constant(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_57ot_math_get_constant = {"ot_math_get_constant", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_57ot_math_get_constant, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_57ot_math_get_constant(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font = 0;
  PyObject *__pyx_v_constant = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_math_get_constant (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_font,&__pyx_n_s_constant,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_font)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1522, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_constant)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1522, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("ot_math_get_constant", 1, 2, 2, 1); __PYX_ERR(0, 1522, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_math_get_constant") < 0)) __PYX_ERR(0, 1522, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_font = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)values[0]);
    __pyx_v_constant = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_math_get_constant", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1522, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_math_get_constant", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_font), __pyx_ptype_9uharfbuzz_9_harfbuzz_Font, 0, "font", 0))) __PYX_ERR(0, 1522, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_56ot_math_get_constant(__pyx_self, __pyx_v_font, __pyx_v_constant);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_56ot_math_get_constant(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font, PyObject *__pyx_v_constant) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  hb_ot_math_constant_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_math_get_constant", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1523
 * 
 * def ot_math_get_constant(font: Font, constant: OTMathConstant) -> int:
 *     if constant >= len(OTMathConstant):             # <<<<<<<<<<<<<<
 *         raise ValueError("invalid constant")
 *     return hb_ot_math_get_constant(font._hb_font, constant)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_OTMathConstant); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1523, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1523, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1523, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_constant, __pyx_t_1, Py_GE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1523, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 1523, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(__pyx_t_4)) {

    /* "uharfbuzz/_harfbuzz.pyx":1524
 * def ot_math_get_constant(font: Font, constant: OTMathConstant) -> int:
 *     if constant >= len(OTMathConstant):
 *         raise ValueError("invalid constant")             # <<<<<<<<<<<<<<
 *     return hb_ot_math_get_constant(font._hb_font, constant)
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1524, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1524, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":1523
 * 
 * def ot_math_get_constant(font: Font, constant: OTMathConstant) -> int:
 *     if constant >= len(OTMathConstant):             # <<<<<<<<<<<<<<
 *         raise ValueError("invalid constant")
 *     return hb_ot_math_get_constant(font._hb_font, constant)
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":1525
 *     if constant >= len(OTMathConstant):
 *         raise ValueError("invalid constant")
 *     return hb_ot_math_get_constant(font._hb_font, constant)             # <<<<<<<<<<<<<<
 * 
 * def ot_math_get_glyph_italics_correction(font: Font, glyph: int) -> int:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = ((hb_ot_math_constant_t)__Pyx_PyInt_As_hb_ot_math_constant_t(__pyx_v_constant)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1525, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyInt_From_hb_position_t(hb_ot_math_get_constant(__pyx_v_font->_hb_font, __pyx_t_5)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!(likely(__Pyx_Py3Int_CheckExact(__pyx_t_3)) || __Pyx_RaiseUnexpectedTypeError("int", __pyx_t_3))) __PYX_ERR(0, 1525, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1522
 *     RADICAL_DEGREE_BOTTOM_RAISE_PERCENT = HB_OT_MATH_CONSTANT_RADICAL_DEGREE_BOTTOM_RAISE_PERCENT
 * 
 * def ot_math_get_constant(font: Font, constant: OTMathConstant) -> int:             # <<<<<<<<<<<<<<
 *     if constant >= len(OTMathConstant):
 *         raise ValueError("invalid constant")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_math_get_constant", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1527
 *     return hb_ot_math_get_constant(font._hb_font, constant)
 * 
 * def ot_math_get_glyph_italics_correction(font: Font, glyph: int) -> int:             # <<<<<<<<<<<<<<
 *     return hb_ot_math_get_glyph_italics_correction(font._hb_font, glyph)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_59ot_math_get_glyph_italics_correction(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_59ot_math_get_glyph_italics_correction = {"ot_math_get_glyph_italics_correction", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_59ot_math_get_glyph_italics_correction, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_59ot_math_get_glyph_italics_correction(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font = 0;
  PyObject *__pyx_v_glyph = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_math_get_glyph_italics_correction (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_font,&__pyx_n_s_glyph,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_font)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1527, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_glyph)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1527, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("ot_math_get_glyph_italics_correction", 1, 2, 2, 1); __PYX_ERR(0, 1527, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_math_get_glyph_italics_correction") < 0)) __PYX_ERR(0, 1527, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_font = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)values[0]);
    __pyx_v_glyph = ((PyObject*)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_math_get_glyph_italics_correction", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1527, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_math_get_glyph_italics_correction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_font), __pyx_ptype_9uharfbuzz_9_harfbuzz_Font, 0, "font", 0))) __PYX_ERR(0, 1527, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_glyph), (&PyInt_Type), 0, "glyph", 1))) __PYX_ERR(0, 1527, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_58ot_math_get_glyph_italics_correction(__pyx_self, __pyx_v_font, __pyx_v_glyph);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_58ot_math_get_glyph_italics_correction(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font, PyObject *__pyx_v_glyph) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  hb_codepoint_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_math_get_glyph_italics_correction", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1528
 * 
 * def ot_math_get_glyph_italics_correction(font: Font, glyph: int) -> int:
 *     return hb_ot_math_get_glyph_italics_correction(font._hb_font, glyph)             # <<<<<<<<<<<<<<
 * 
 * def ot_math_get_glyph_top_accent_attachment(font: Font, glyph: int) -> int:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_glyph); if (unlikely((__pyx_t_1 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1528, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_hb_position_t(hb_ot_math_get_glyph_italics_correction(__pyx_v_font->_hb_font, __pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(__Pyx_Py3Int_CheckExact(__pyx_t_2)) || __Pyx_RaiseUnexpectedTypeError("int", __pyx_t_2))) __PYX_ERR(0, 1528, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1527
 *     return hb_ot_math_get_constant(font._hb_font, constant)
 * 
 * def ot_math_get_glyph_italics_correction(font: Font, glyph: int) -> int:             # <<<<<<<<<<<<<<
 *     return hb_ot_math_get_glyph_italics_correction(font._hb_font, glyph)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_math_get_glyph_italics_correction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1530
 *     return hb_ot_math_get_glyph_italics_correction(font._hb_font, glyph)
 * 
 * def ot_math_get_glyph_top_accent_attachment(font: Font, glyph: int) -> int:             # <<<<<<<<<<<<<<
 *     return hb_ot_math_get_glyph_top_accent_attachment(font._hb_font, glyph)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_61ot_math_get_glyph_top_accent_attachment(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_61ot_math_get_glyph_top_accent_attachment = {"ot_math_get_glyph_top_accent_attachment", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_61ot_math_get_glyph_top_accent_attachment, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_61ot_math_get_glyph_top_accent_attachment(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font = 0;
  PyObject *__pyx_v_glyph = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_math_get_glyph_top_accent_attachment (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_font,&__pyx_n_s_glyph,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_font)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1530, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_glyph)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1530, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("ot_math_get_glyph_top_accent_attachment", 1, 2, 2, 1); __PYX_ERR(0, 1530, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_math_get_glyph_top_accent_attachment") < 0)) __PYX_ERR(0, 1530, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_font = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)values[0]);
    __pyx_v_glyph = ((PyObject*)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_math_get_glyph_top_accent_attachment", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1530, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_math_get_glyph_top_accent_attachment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_font), __pyx_ptype_9uharfbuzz_9_harfbuzz_Font, 0, "font", 0))) __PYX_ERR(0, 1530, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_glyph), (&PyInt_Type), 0, "glyph", 1))) __PYX_ERR(0, 1530, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_60ot_math_get_glyph_top_accent_attachment(__pyx_self, __pyx_v_font, __pyx_v_glyph);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_60ot_math_get_glyph_top_accent_attachment(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font, PyObject *__pyx_v_glyph) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  hb_codepoint_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_math_get_glyph_top_accent_attachment", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1531
 * 
 * def ot_math_get_glyph_top_accent_attachment(font: Font, glyph: int) -> int:
 *     return hb_ot_math_get_glyph_top_accent_attachment(font._hb_font, glyph)             # <<<<<<<<<<<<<<
 * 
 * def ot_math_is_glyph_extended_shape(face: Face, glyph: int) -> bool:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_glyph); if (unlikely((__pyx_t_1 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1531, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_hb_position_t(hb_ot_math_get_glyph_top_accent_attachment(__pyx_v_font->_hb_font, __pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(__Pyx_Py3Int_CheckExact(__pyx_t_2)) || __Pyx_RaiseUnexpectedTypeError("int", __pyx_t_2))) __PYX_ERR(0, 1531, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1530
 *     return hb_ot_math_get_glyph_italics_correction(font._hb_font, glyph)
 * 
 * def ot_math_get_glyph_top_accent_attachment(font: Font, glyph: int) -> int:             # <<<<<<<<<<<<<<
 *     return hb_ot_math_get_glyph_top_accent_attachment(font._hb_font, glyph)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_math_get_glyph_top_accent_attachment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1533
 *     return hb_ot_math_get_glyph_top_accent_attachment(font._hb_font, glyph)
 * 
 * def ot_math_is_glyph_extended_shape(face: Face, glyph: int) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_math_is_glyph_extended_shape(face._hb_face, glyph)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_63ot_math_is_glyph_extended_shape(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_63ot_math_is_glyph_extended_shape = {"ot_math_is_glyph_extended_shape", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_63ot_math_is_glyph_extended_shape, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_63ot_math_is_glyph_extended_shape(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face = 0;
  PyObject *__pyx_v_glyph = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_math_is_glyph_extended_shape (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_face,&__pyx_n_s_glyph,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_face)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1533, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_glyph)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1533, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("ot_math_is_glyph_extended_shape", 1, 2, 2, 1); __PYX_ERR(0, 1533, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_math_is_glyph_extended_shape") < 0)) __PYX_ERR(0, 1533, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)values[0]);
    __pyx_v_glyph = ((PyObject*)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_math_is_glyph_extended_shape", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1533, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_math_is_glyph_extended_shape", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_face), __pyx_ptype_9uharfbuzz_9_harfbuzz_Face, 0, "face", 0))) __PYX_ERR(0, 1533, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_glyph), (&PyInt_Type), 0, "glyph", 1))) __PYX_ERR(0, 1533, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_62ot_math_is_glyph_extended_shape(__pyx_self, __pyx_v_face, __pyx_v_glyph);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_62ot_math_is_glyph_extended_shape(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, PyObject *__pyx_v_glyph) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  hb_codepoint_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_math_is_glyph_extended_shape", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1534
 * 
 * def ot_math_is_glyph_extended_shape(face: Face, glyph: int) -> bool:
 *     return hb_ot_math_is_glyph_extended_shape(face._hb_face, glyph)             # <<<<<<<<<<<<<<
 * 
 * def ot_math_get_min_connector_overlap(font: Font, direction: str) -> int:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_glyph); if (unlikely((__pyx_t_1 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1534, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_hb_bool_t(hb_ot_math_is_glyph_extended_shape(__pyx_v_face->_hb_face, __pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1533
 *     return hb_ot_math_get_glyph_top_accent_attachment(font._hb_font, glyph)
 * 
 * def ot_math_is_glyph_extended_shape(face: Face, glyph: int) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_math_is_glyph_extended_shape(face._hb_face, glyph)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_math_is_glyph_extended_shape", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1536
 *     return hb_ot_math_is_glyph_extended_shape(face._hb_face, glyph)
 * 
 * def ot_math_get_min_connector_overlap(font: Font, direction: str) -> int:             # <<<<<<<<<<<<<<
 *     cdef bytes packed = direction.encode()
 *     cdef char* cstr = packed
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_65ot_math_get_min_connector_overlap(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_65ot_math_get_min_connector_overlap = {"ot_math_get_min_connector_overlap", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_65ot_math_get_min_connector_overlap, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_65ot_math_get_min_connector_overlap(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font = 0;
  PyObject *__pyx_v_direction = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_math_get_min_connector_overlap (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_font,&__pyx_n_s_direction,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_font)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1536, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_direction)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1536, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("ot_math_get_min_connector_overlap", 1, 2, 2, 1); __PYX_ERR(0, 1536, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_math_get_min_connector_overlap") < 0)) __PYX_ERR(0, 1536, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_font = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)values[0]);
    __pyx_v_direction = ((PyObject*)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_math_get_min_connector_overlap", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1536, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_math_get_min_connector_overlap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_font), __pyx_ptype_9uharfbuzz_9_harfbuzz_Font, 0, "font", 0))) __PYX_ERR(0, 1536, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_direction), (&PyUnicode_Type), 0, "direction", 1))) __PYX_ERR(0, 1536, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_64ot_math_get_min_connector_overlap(__pyx_self, __pyx_v_font, __pyx_v_direction);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_64ot_math_get_min_connector_overlap(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font, PyObject *__pyx_v_direction) {
  PyObject *__pyx_v_packed = 0;
  char *__pyx_v_cstr;
  hb_direction_t __pyx_v_hb_direction;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_math_get_min_connector_overlap", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1537
 * 
 * def ot_math_get_min_connector_overlap(font: Font, direction: str) -> int:
 *     cdef bytes packed = direction.encode()             # <<<<<<<<<<<<<<
 *     cdef char* cstr = packed
 *     cdef hb_direction_t hb_direction = hb_direction_from_string(cstr, -1)
 */
  __pyx_t_1 = PyUnicode_AsEncodedString(__pyx_v_direction, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_packed = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1538
 * def ot_math_get_min_connector_overlap(font: Font, direction: str) -> int:
 *     cdef bytes packed = direction.encode()
 *     cdef char* cstr = packed             # <<<<<<<<<<<<<<
 *     cdef hb_direction_t hb_direction = hb_direction_from_string(cstr, -1)
 *     return hb_ot_math_get_min_connector_overlap(font._hb_font, hb_direction)
 */
  if (unlikely(__pyx_v_packed == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1538, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBytes_AsWritableString(__pyx_v_packed); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 1538, __pyx_L1_error)
  __pyx_v_cstr = __pyx_t_2;

  /* "uharfbuzz/_harfbuzz.pyx":1539
 *     cdef bytes packed = direction.encode()
 *     cdef char* cstr = packed
 *     cdef hb_direction_t hb_direction = hb_direction_from_string(cstr, -1)             # <<<<<<<<<<<<<<
 *     return hb_ot_math_get_min_connector_overlap(font._hb_font, hb_direction)
 * 
 */
  __pyx_v_hb_direction = hb_direction_from_string(__pyx_v_cstr, -1);

  /* "uharfbuzz/_harfbuzz.pyx":1540
 *     cdef char* cstr = packed
 *     cdef hb_direction_t hb_direction = hb_direction_from_string(cstr, -1)
 *     return hb_ot_math_get_min_connector_overlap(font._hb_font, hb_direction)             # <<<<<<<<<<<<<<
 * 
 * class OTMathKern(IntEnum):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_hb_position_t(hb_ot_math_get_min_connector_overlap(__pyx_v_font->_hb_font, __pyx_v_hb_direction)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(__Pyx_Py3Int_CheckExact(__pyx_t_1)) || __Pyx_RaiseUnexpectedTypeError("int", __pyx_t_1))) __PYX_ERR(0, 1540, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1536
 *     return hb_ot_math_is_glyph_extended_shape(face._hb_face, glyph)
 * 
 * def ot_math_get_min_connector_overlap(font: Font, direction: str) -> int:             # <<<<<<<<<<<<<<
 *     cdef bytes packed = direction.encode()
 *     cdef char* cstr = packed
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_math_get_min_connector_overlap", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_packed);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1548
 *     BOTTOM_LEFT = HB_OT_MATH_KERN_BOTTOM_LEFT
 * 
 * def ot_math_get_glyph_kerning(font: Font,             # <<<<<<<<<<<<<<
 *                               glyph: int,
 *                               kern: OTMathKern,
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_67ot_math_get_glyph_kerning(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_67ot_math_get_glyph_kerning = {"ot_math_get_glyph_kerning", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_67ot_math_get_glyph_kerning, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_67ot_math_get_glyph_kerning(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font = 0;
  PyObject *__pyx_v_glyph = 0;
  PyObject *__pyx_v_kern = 0;
  int __pyx_v_correction_height;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_math_get_glyph_kerning (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_font,&__pyx_n_s_glyph,&__pyx_n_s_kern,&__pyx_n_s_correction_height,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_font)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1548, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_glyph)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1548, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("ot_math_get_glyph_kerning", 1, 4, 4, 1); __PYX_ERR(0, 1548, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_kern)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1548, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("ot_math_get_glyph_kerning", 1, 4, 4, 2); __PYX_ERR(0, 1548, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_correction_height)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[3]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1548, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("ot_math_get_glyph_kerning", 1, 4, 4, 3); __PYX_ERR(0, 1548, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_math_get_glyph_kerning") < 0)) __PYX_ERR(0, 1548, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 4)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
      values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
    }
    __pyx_v_font = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)values[0]);
    __pyx_v_glyph = ((PyObject*)values[1]);
    __pyx_v_kern = values[2];
    __pyx_v_correction_height = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_correction_height == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1551, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_math_get_glyph_kerning", 1, 4, 4, __pyx_nargs); __PYX_ERR(0, 1548, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_math_get_glyph_kerning", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_font), __pyx_ptype_9uharfbuzz_9_harfbuzz_Font, 0, "font", 0))) __PYX_ERR(0, 1548, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_glyph), (&PyInt_Type), 0, "glyph", 1))) __PYX_ERR(0, 1549, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_66ot_math_get_glyph_kerning(__pyx_self, __pyx_v_font, __pyx_v_glyph, __pyx_v_kern, __pyx_v_correction_height);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_66ot_math_get_glyph_kerning(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font, PyObject *__pyx_v_glyph, PyObject *__pyx_v_kern, int __pyx_v_correction_height) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  hb_codepoint_t __pyx_t_5;
  hb_ot_math_kern_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_math_get_glyph_kerning", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1552
 *                               kern: OTMathKern,
 *                               int correction_height) -> int:
 *     if kern >= len(OTMathKern):             # <<<<<<<<<<<<<<
 *         raise ValueError("invalid kern")
 *     return hb_ot_math_get_glyph_kerning(font._hb_font, glyph, kern, correction_height)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_OTMathKern); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1552, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_kern, __pyx_t_1, Py_GE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1552, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 1552, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(__pyx_t_4)) {

    /* "uharfbuzz/_harfbuzz.pyx":1553
 *                               int correction_height) -> int:
 *     if kern >= len(OTMathKern):
 *         raise ValueError("invalid kern")             # <<<<<<<<<<<<<<
 *     return hb_ot_math_get_glyph_kerning(font._hb_font, glyph, kern, correction_height)
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1553, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1553, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":1552
 *                               kern: OTMathKern,
 *                               int correction_height) -> int:
 *     if kern >= len(OTMathKern):             # <<<<<<<<<<<<<<
 *         raise ValueError("invalid kern")
 *     return hb_ot_math_get_glyph_kerning(font._hb_font, glyph, kern, correction_height)
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":1554
 *     if kern >= len(OTMathKern):
 *         raise ValueError("invalid kern")
 *     return hb_ot_math_get_glyph_kerning(font._hb_font, glyph, kern, correction_height)             # <<<<<<<<<<<<<<
 * 
 * OTMathKernEntry = namedtuple("OTMathKernEntry", ["max_correction_height", "kern_value"])
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_glyph); if (unlikely((__pyx_t_5 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1554, __pyx_L1_error)
  __pyx_t_6 = ((hb_ot_math_kern_t)__Pyx_PyInt_As_hb_ot_math_kern_t(__pyx_v_kern)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1554, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyInt_From_hb_position_t(hb_ot_math_get_glyph_kerning(__pyx_v_font->_hb_font, __pyx_t_5, __pyx_t_6, __pyx_v_correction_height)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!(likely(__Pyx_Py3Int_CheckExact(__pyx_t_3)) || __Pyx_RaiseUnexpectedTypeError("int", __pyx_t_3))) __PYX_ERR(0, 1554, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1548
 *     BOTTOM_LEFT = HB_OT_MATH_KERN_BOTTOM_LEFT
 * 
 * def ot_math_get_glyph_kerning(font: Font,             # <<<<<<<<<<<<<<
 *                               glyph: int,
 *                               kern: OTMathKern,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_math_get_glyph_kerning", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1558
 * OTMathKernEntry = namedtuple("OTMathKernEntry", ["max_correction_height", "kern_value"])
 * 
 * def ot_math_get_glyph_kernings(font: Font,             # <<<<<<<<<<<<<<
 *                                glyph: int,
 *                                kern: OTMathKern) -> List[OTMathKernEntry]:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_69ot_math_get_glyph_kernings(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_69ot_math_get_glyph_kernings = {"ot_math_get_glyph_kernings", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_69ot_math_get_glyph_kernings, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_69ot_math_get_glyph_kernings(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font = 0;
  PyObject *__pyx_v_glyph = 0;
  PyObject *__pyx_v_kern = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_math_get_glyph_kernings (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_font,&__pyx_n_s_glyph,&__pyx_n_s_kern,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_font)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1558, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_glyph)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1558, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("ot_math_get_glyph_kernings", 1, 3, 3, 1); __PYX_ERR(0, 1558, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_kern)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1558, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("ot_math_get_glyph_kernings", 1, 3, 3, 2); __PYX_ERR(0, 1558, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_math_get_glyph_kernings") < 0)) __PYX_ERR(0, 1558, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v_font = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)values[0]);
    __pyx_v_glyph = ((PyObject*)values[1]);
    __pyx_v_kern = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_math_get_glyph_kernings", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1558, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_math_get_glyph_kernings", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_font), __pyx_ptype_9uharfbuzz_9_harfbuzz_Font, 0, "font", 0))) __PYX_ERR(0, 1558, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_glyph), (&PyInt_Type), 0, "glyph", 1))) __PYX_ERR(0, 1559, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_68ot_math_get_glyph_kernings(__pyx_self, __pyx_v_font, __pyx_v_glyph, __pyx_v_kern);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_68ot_math_get_glyph_kernings(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font, PyObject *__pyx_v_glyph, PyObject *__pyx_v_kern) {
  unsigned int __pyx_v_count;
  hb_ot_math_kern_entry_t __pyx_v_kerns_array[0x80];
  PyObject *__pyx_v_kerns = 0;
  unsigned int __pyx_v_i;
  unsigned int __pyx_v_start_offset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  hb_codepoint_t __pyx_t_5;
  hb_ot_math_kern_t __pyx_t_6;
  unsigned int __pyx_t_7;
  unsigned int __pyx_t_8;
  unsigned int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  unsigned int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_math_get_glyph_kernings", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1561
 *                                glyph: int,
 *                                kern: OTMathKern) -> List[OTMathKernEntry]:
 *     if kern >= len(OTMathKern):             # <<<<<<<<<<<<<<
 *         raise ValueError("invalid kern")
 *     cdef unsigned int count = STATIC_ARRAY_SIZE
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_OTMathKern); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1561, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_kern, __pyx_t_1, Py_GE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1561, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 1561, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(__pyx_t_4)) {

    /* "uharfbuzz/_harfbuzz.pyx":1562
 *                                kern: OTMathKern) -> List[OTMathKernEntry]:
 *     if kern >= len(OTMathKern):
 *         raise ValueError("invalid kern")             # <<<<<<<<<<<<<<
 *     cdef unsigned int count = STATIC_ARRAY_SIZE
 *     cdef hb_ot_math_kern_entry_t kerns_array[STATIC_ARRAY_SIZE]
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1562, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1562, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":1561
 *                                glyph: int,
 *                                kern: OTMathKern) -> List[OTMathKernEntry]:
 *     if kern >= len(OTMathKern):             # <<<<<<<<<<<<<<
 *         raise ValueError("invalid kern")
 *     cdef unsigned int count = STATIC_ARRAY_SIZE
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":1563
 *     if kern >= len(OTMathKern):
 *         raise ValueError("invalid kern")
 *     cdef unsigned int count = STATIC_ARRAY_SIZE             # <<<<<<<<<<<<<<
 *     cdef hb_ot_math_kern_entry_t kerns_array[STATIC_ARRAY_SIZE]
 *     cdef list kerns = []
 */
  __pyx_v_count = 0x80;

  /* "uharfbuzz/_harfbuzz.pyx":1565
 *     cdef unsigned int count = STATIC_ARRAY_SIZE
 *     cdef hb_ot_math_kern_entry_t kerns_array[STATIC_ARRAY_SIZE]
 *     cdef list kerns = []             # <<<<<<<<<<<<<<
 *     cdef unsigned int i
 *     cdef unsigned int start_offset = 0
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_kerns = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1567
 *     cdef list kerns = []
 *     cdef unsigned int i
 *     cdef unsigned int start_offset = 0             # <<<<<<<<<<<<<<
 *     while count == STATIC_ARRAY_SIZE:
 *         hb_ot_math_get_glyph_kernings(font._hb_font, glyph, kern, start_offset,
 */
  __pyx_v_start_offset = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1568
 *     cdef unsigned int i
 *     cdef unsigned int start_offset = 0
 *     while count == STATIC_ARRAY_SIZE:             # <<<<<<<<<<<<<<
 *         hb_ot_math_get_glyph_kernings(font._hb_font, glyph, kern, start_offset,
 *             &count, kerns_array)
 */
  while (1) {
    __pyx_t_4 = (__pyx_v_count == 0x80);
    if (!__pyx_t_4) break;

    /* "uharfbuzz/_harfbuzz.pyx":1569
 *     cdef unsigned int start_offset = 0
 *     while count == STATIC_ARRAY_SIZE:
 *         hb_ot_math_get_glyph_kernings(font._hb_font, glyph, kern, start_offset,             # <<<<<<<<<<<<<<
 *             &count, kerns_array)
 *         for i in range(count):
 */
    __pyx_t_5 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_glyph); if (unlikely((__pyx_t_5 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1569, __pyx_L1_error)
    __pyx_t_6 = ((hb_ot_math_kern_t)__Pyx_PyInt_As_hb_ot_math_kern_t(__pyx_v_kern)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1569, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":1570
 *     while count == STATIC_ARRAY_SIZE:
 *         hb_ot_math_get_glyph_kernings(font._hb_font, glyph, kern, start_offset,
 *             &count, kerns_array)             # <<<<<<<<<<<<<<
 *         for i in range(count):
 *             kerns.append(OTMathKernEntry(kerns_array[i].max_correction_height, kerns_array[i].kern_value))
 */
    (void)(hb_ot_math_get_glyph_kernings(__pyx_v_font->_hb_font, __pyx_t_5, __pyx_t_6, __pyx_v_start_offset, (&__pyx_v_count), __pyx_v_kerns_array));

    /* "uharfbuzz/_harfbuzz.pyx":1571
 *         hb_ot_math_get_glyph_kernings(font._hb_font, glyph, kern, start_offset,
 *             &count, kerns_array)
 *         for i in range(count):             # <<<<<<<<<<<<<<
 *             kerns.append(OTMathKernEntry(kerns_array[i].max_correction_height, kerns_array[i].kern_value))
 *         start_offset += count
 */
    __pyx_t_7 = __pyx_v_count;
    __pyx_t_8 = __pyx_t_7;
    for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
      __pyx_v_i = __pyx_t_9;

      /* "uharfbuzz/_harfbuzz.pyx":1572
 *             &count, kerns_array)
 *         for i in range(count):
 *             kerns.append(OTMathKernEntry(kerns_array[i].max_correction_height, kerns_array[i].kern_value))             # <<<<<<<<<<<<<<
 *         start_offset += count
 *     return kerns
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_OTMathKernEntry); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1572, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_10 = __Pyx_PyInt_From_hb_position_t((__pyx_v_kerns_array[__pyx_v_i]).max_correction_height); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1572, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_11 = __Pyx_PyInt_From_hb_position_t((__pyx_v_kerns_array[__pyx_v_i]).kern_value); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1572, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_12 = NULL;
      __pyx_t_13 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_12)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_12);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_13 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_12, __pyx_t_10, __pyx_t_11};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_13, 2+__pyx_t_13);
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1572, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __pyx_t_14 = __Pyx_PyList_Append(__pyx_v_kerns, __pyx_t_3); if (unlikely(__pyx_t_14 == ((int)-1))) __PYX_ERR(0, 1572, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }

    /* "uharfbuzz/_harfbuzz.pyx":1573
 *         for i in range(count):
 *             kerns.append(OTMathKernEntry(kerns_array[i].max_correction_height, kerns_array[i].kern_value))
 *         start_offset += count             # <<<<<<<<<<<<<<
 *     return kerns
 * 
 */
    __pyx_v_start_offset = (__pyx_v_start_offset + __pyx_v_count);
  }

  /* "uharfbuzz/_harfbuzz.pyx":1574
 *             kerns.append(OTMathKernEntry(kerns_array[i].max_correction_height, kerns_array[i].kern_value))
 *         start_offset += count
 *     return kerns             # <<<<<<<<<<<<<<
 * 
 * OTMathGlyphVariant = namedtuple("OTMathGlyphVariant", ["glyph", "advance"])
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_kerns);
  __pyx_r = __pyx_v_kerns;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1558
 * OTMathKernEntry = namedtuple("OTMathKernEntry", ["max_correction_height", "kern_value"])
 * 
 * def ot_math_get_glyph_kernings(font: Font,             # <<<<<<<<<<<<<<
 *                                glyph: int,
 *                                kern: OTMathKern) -> List[OTMathKernEntry]:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_math_get_glyph_kernings", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_kerns);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1578
 * OTMathGlyphVariant = namedtuple("OTMathGlyphVariant", ["glyph", "advance"])
 * 
 * def ot_math_get_glyph_variants(font: Font, glyph: int, direction: str) -> List[OTMathGlyphVariant]:             # <<<<<<<<<<<<<<
 *     cdef bytes packed = direction.encode()
 *     cdef char* cstr = packed
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_71ot_math_get_glyph_variants(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_71ot_math_get_glyph_variants = {"ot_math_get_glyph_variants", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_71ot_math_get_glyph_variants, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_71ot_math_get_glyph_variants(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font = 0;
  PyObject *__pyx_v_glyph = 0;
  PyObject *__pyx_v_direction = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_math_get_glyph_variants (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_font,&__pyx_n_s_glyph,&__pyx_n_s_direction,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_font)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1578, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_glyph)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1578, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("ot_math_get_glyph_variants", 1, 3, 3, 1); __PYX_ERR(0, 1578, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_direction)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1578, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("ot_math_get_glyph_variants", 1, 3, 3, 2); __PYX_ERR(0, 1578, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_math_get_glyph_variants") < 0)) __PYX_ERR(0, 1578, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v_font = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)values[0]);
    __pyx_v_glyph = ((PyObject*)values[1]);
    __pyx_v_direction = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_math_get_glyph_variants", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1578, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_math_get_glyph_variants", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_font), __pyx_ptype_9uharfbuzz_9_harfbuzz_Font, 0, "font", 0))) __PYX_ERR(0, 1578, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_glyph), (&PyInt_Type), 0, "glyph", 1))) __PYX_ERR(0, 1578, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_direction), (&PyUnicode_Type), 0, "direction", 1))) __PYX_ERR(0, 1578, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_70ot_math_get_glyph_variants(__pyx_self, __pyx_v_font, __pyx_v_glyph, __pyx_v_direction);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_70ot_math_get_glyph_variants(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font, PyObject *__pyx_v_glyph, PyObject *__pyx_v_direction) {
  PyObject *__pyx_v_packed = 0;
  char *__pyx_v_cstr;
  hb_direction_t __pyx_v_hb_direction;
  unsigned int __pyx_v_count;
  hb_ot_math_glyph_variant_t __pyx_v_variants_array[0x80];
  PyObject *__pyx_v_variants = 0;
  unsigned int __pyx_v_i;
  unsigned int __pyx_v_start_offset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  int __pyx_t_3;
  hb_codepoint_t __pyx_t_4;
  unsigned int __pyx_t_5;
  unsigned int __pyx_t_6;
  unsigned int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  unsigned int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_math_get_glyph_variants", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1579
 * 
 * def ot_math_get_glyph_variants(font: Font, glyph: int, direction: str) -> List[OTMathGlyphVariant]:
 *     cdef bytes packed = direction.encode()             # <<<<<<<<<<<<<<
 *     cdef char* cstr = packed
 *     cdef hb_direction_t hb_direction = hb_direction_from_string(cstr, -1)
 */
  __pyx_t_1 = PyUnicode_AsEncodedString(__pyx_v_direction, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_packed = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1580
 * def ot_math_get_glyph_variants(font: Font, glyph: int, direction: str) -> List[OTMathGlyphVariant]:
 *     cdef bytes packed = direction.encode()
 *     cdef char* cstr = packed             # <<<<<<<<<<<<<<
 *     cdef hb_direction_t hb_direction = hb_direction_from_string(cstr, -1)
 *     cdef unsigned int count = STATIC_ARRAY_SIZE
 */
  if (unlikely(__pyx_v_packed == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1580, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBytes_AsWritableString(__pyx_v_packed); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 1580, __pyx_L1_error)
  __pyx_v_cstr = __pyx_t_2;

  /* "uharfbuzz/_harfbuzz.pyx":1581
 *     cdef bytes packed = direction.encode()
 *     cdef char* cstr = packed
 *     cdef hb_direction_t hb_direction = hb_direction_from_string(cstr, -1)             # <<<<<<<<<<<<<<
 *     cdef unsigned int count = STATIC_ARRAY_SIZE
 *     cdef hb_ot_math_glyph_variant_t variants_array[STATIC_ARRAY_SIZE]
 */
  __pyx_v_hb_direction = hb_direction_from_string(__pyx_v_cstr, -1);

  /* "uharfbuzz/_harfbuzz.pyx":1582
 *     cdef char* cstr = packed
 *     cdef hb_direction_t hb_direction = hb_direction_from_string(cstr, -1)
 *     cdef unsigned int count = STATIC_ARRAY_SIZE             # <<<<<<<<<<<<<<
 *     cdef hb_ot_math_glyph_variant_t variants_array[STATIC_ARRAY_SIZE]
 *     cdef list variants = []
 */
  __pyx_v_count = 0x80;

  /* "uharfbuzz/_harfbuzz.pyx":1584
 *     cdef unsigned int count = STATIC_ARRAY_SIZE
 *     cdef hb_ot_math_glyph_variant_t variants_array[STATIC_ARRAY_SIZE]
 *     cdef list variants = []             # <<<<<<<<<<<<<<
 *     cdef unsigned int i
 *     cdef unsigned int start_offset = 0
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_variants = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1586
 *     cdef list variants = []
 *     cdef unsigned int i
 *     cdef unsigned int start_offset = 0             # <<<<<<<<<<<<<<
 *     while count == STATIC_ARRAY_SIZE:
 *         hb_ot_math_get_glyph_variants(font._hb_font, glyph, hb_direction, start_offset,
 */
  __pyx_v_start_offset = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1587
 *     cdef unsigned int i
 *     cdef unsigned int start_offset = 0
 *     while count == STATIC_ARRAY_SIZE:             # <<<<<<<<<<<<<<
 *         hb_ot_math_get_glyph_variants(font._hb_font, glyph, hb_direction, start_offset,
 *             &count, variants_array)
 */
  while (1) {
    __pyx_t_3 = (__pyx_v_count == 0x80);
    if (!__pyx_t_3) break;

    /* "uharfbuzz/_harfbuzz.pyx":1588
 *     cdef unsigned int start_offset = 0
 *     while count == STATIC_ARRAY_SIZE:
 *         hb_ot_math_get_glyph_variants(font._hb_font, glyph, hb_direction, start_offset,             # <<<<<<<<<<<<<<
 *             &count, variants_array)
 *         for i in range(count):
 */
    __pyx_t_4 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_glyph); if (unlikely((__pyx_t_4 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1588, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":1589
 *     while count == STATIC_ARRAY_SIZE:
 *         hb_ot_math_get_glyph_variants(font._hb_font, glyph, hb_direction, start_offset,
 *             &count, variants_array)             # <<<<<<<<<<<<<<
 *         for i in range(count):
 *             variants.append(OTMathGlyphVariant(variants_array[i].glyph, variants_array[i].advance))
 */
    (void)(hb_ot_math_get_glyph_variants(__pyx_v_font->_hb_font, __pyx_t_4, __pyx_v_hb_direction, __pyx_v_start_offset, (&__pyx_v_count), __pyx_v_variants_array));

    /* "uharfbuzz/_harfbuzz.pyx":1590
 *         hb_ot_math_get_glyph_variants(font._hb_font, glyph, hb_direction, start_offset,
 *             &count, variants_array)
 *         for i in range(count):             # <<<<<<<<<<<<<<
 *             variants.append(OTMathGlyphVariant(variants_array[i].glyph, variants_array[i].advance))
 *         start_offset += count
 */
    __pyx_t_5 = __pyx_v_count;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_i = __pyx_t_7;

      /* "uharfbuzz/_harfbuzz.pyx":1591
 *             &count, variants_array)
 *         for i in range(count):
 *             variants.append(OTMathGlyphVariant(variants_array[i].glyph, variants_array[i].advance))             # <<<<<<<<<<<<<<
 *         start_offset += count
 *     return variants
 */
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_OTMathGlyphVariant); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1591, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = __Pyx_PyInt_From_hb_codepoint_t((__pyx_v_variants_array[__pyx_v_i]).glyph); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1591, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = __Pyx_PyInt_From_hb_position_t((__pyx_v_variants_array[__pyx_v_i]).advance); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1591, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_11 = NULL;
      __pyx_t_12 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_11)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
          __pyx_t_12 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_11, __pyx_t_9, __pyx_t_10};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_12, 2+__pyx_t_12);
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1591, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_variants, __pyx_t_1); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 1591, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }

    /* "uharfbuzz/_harfbuzz.pyx":1592
 *         for i in range(count):
 *             variants.append(OTMathGlyphVariant(variants_array[i].glyph, variants_array[i].advance))
 *         start_offset += count             # <<<<<<<<<<<<<<
 *     return variants
 * 
 */
    __pyx_v_start_offset = (__pyx_v_start_offset + __pyx_v_count);
  }

  /* "uharfbuzz/_harfbuzz.pyx":1593
 *             variants.append(OTMathGlyphVariant(variants_array[i].glyph, variants_array[i].advance))
 *         start_offset += count
 *     return variants             # <<<<<<<<<<<<<<
 * 
 * class OTMathGlyphPartFlags(IntFlag):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_variants);
  __pyx_r = __pyx_v_variants;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1578
 * OTMathGlyphVariant = namedtuple("OTMathGlyphVariant", ["glyph", "advance"])
 * 
 * def ot_math_get_glyph_variants(font: Font, glyph: int, direction: str) -> List[OTMathGlyphVariant]:             # <<<<<<<<<<<<<<
 *     cdef bytes packed = direction.encode()
 *     cdef char* cstr = packed
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_math_get_glyph_variants", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_packed);
  __Pyx_XDECREF(__pyx_v_variants);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1603
 * )
 * 
 * def ot_math_get_glyph_assembly(font: Font,             # <<<<<<<<<<<<<<
 *                                glyph: int,
 *                                direction: str) -> Tuple[List[OTMathGlyphPart], int]:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_73ot_math_get_glyph_assembly(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_73ot_math_get_glyph_assembly = {"ot_math_get_glyph_assembly", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_73ot_math_get_glyph_assembly, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_73ot_math_get_glyph_assembly(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font = 0;
  PyObject *__pyx_v_glyph = 0;
  PyObject *__pyx_v_direction = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_math_get_glyph_assembly (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_font,&__pyx_n_s_glyph,&__pyx_n_s_direction,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_font)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1603, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_glyph)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1603, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("ot_math_get_glyph_assembly", 1, 3, 3, 1); __PYX_ERR(0, 1603, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_direction)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1603, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("ot_math_get_glyph_assembly", 1, 3, 3, 2); __PYX_ERR(0, 1603, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_math_get_glyph_assembly") < 0)) __PYX_ERR(0, 1603, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v_font = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)values[0]);
    __pyx_v_glyph = ((PyObject*)values[1]);
    __pyx_v_direction = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_math_get_glyph_assembly", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 1603, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_math_get_glyph_assembly", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_font), __pyx_ptype_9uharfbuzz_9_harfbuzz_Font, 0, "font", 0))) __PYX_ERR(0, 1603, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_glyph), (&PyInt_Type), 0, "glyph", 1))) __PYX_ERR(0, 1604, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_direction), (&PyUnicode_Type), 0, "direction", 1))) __PYX_ERR(0, 1605, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_72ot_math_get_glyph_assembly(__pyx_self, __pyx_v_font, __pyx_v_glyph, __pyx_v_direction);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_72ot_math_get_glyph_assembly(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font, PyObject *__pyx_v_glyph, PyObject *__pyx_v_direction) {
  PyObject *__pyx_v_packed = 0;
  char *__pyx_v_cstr;
  hb_direction_t __pyx_v_hb_direction;
  unsigned int __pyx_v_count;
  hb_ot_math_glyph_part_t __pyx_v_assembly_array[0x80];
  PyObject *__pyx_v_assembly = 0;
  unsigned int __pyx_v_i;
  unsigned int __pyx_v_start_offset;
  hb_position_t __pyx_v_italics_correction;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char *__pyx_t_2;
  int __pyx_t_3;
  hb_codepoint_t __pyx_t_4;
  unsigned int __pyx_t_5;
  unsigned int __pyx_t_6;
  unsigned int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  unsigned int __pyx_t_17;
  int __pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("ot_math_get_glyph_assembly", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1606
 *                                glyph: int,
 *                                direction: str) -> Tuple[List[OTMathGlyphPart], int]:
 *     cdef bytes packed = direction.encode()             # <<<<<<<<<<<<<<
 *     cdef char* cstr = packed
 *     cdef hb_direction_t hb_direction = hb_direction_from_string(cstr, -1)
 */
  __pyx_t_1 = PyUnicode_AsEncodedString(__pyx_v_direction, NULL, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_packed = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1607
 *                                direction: str) -> Tuple[List[OTMathGlyphPart], int]:
 *     cdef bytes packed = direction.encode()
 *     cdef char* cstr = packed             # <<<<<<<<<<<<<<
 *     cdef hb_direction_t hb_direction = hb_direction_from_string(cstr, -1)
 *     cdef unsigned int count = STATIC_ARRAY_SIZE
 */
  if (unlikely(__pyx_v_packed == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1607, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyBytes_AsWritableString(__pyx_v_packed); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 1607, __pyx_L1_error)
  __pyx_v_cstr = __pyx_t_2;

  /* "uharfbuzz/_harfbuzz.pyx":1608
 *     cdef bytes packed = direction.encode()
 *     cdef char* cstr = packed
 *     cdef hb_direction_t hb_direction = hb_direction_from_string(cstr, -1)             # <<<<<<<<<<<<<<
 *     cdef unsigned int count = STATIC_ARRAY_SIZE
 *     cdef hb_ot_math_glyph_part_t assembly_array[STATIC_ARRAY_SIZE]
 */
  __pyx_v_hb_direction = hb_direction_from_string(__pyx_v_cstr, -1);

  /* "uharfbuzz/_harfbuzz.pyx":1609
 *     cdef char* cstr = packed
 *     cdef hb_direction_t hb_direction = hb_direction_from_string(cstr, -1)
 *     cdef unsigned int count = STATIC_ARRAY_SIZE             # <<<<<<<<<<<<<<
 *     cdef hb_ot_math_glyph_part_t assembly_array[STATIC_ARRAY_SIZE]
 *     cdef list assembly = []
 */
  __pyx_v_count = 0x80;

  /* "uharfbuzz/_harfbuzz.pyx":1611
 *     cdef unsigned int count = STATIC_ARRAY_SIZE
 *     cdef hb_ot_math_glyph_part_t assembly_array[STATIC_ARRAY_SIZE]
 *     cdef list assembly = []             # <<<<<<<<<<<<<<
 *     cdef unsigned int i
 *     cdef unsigned int start_offset = 0
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1611, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_assembly = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1613
 *     cdef list assembly = []
 *     cdef unsigned int i
 *     cdef unsigned int start_offset = 0             # <<<<<<<<<<<<<<
 *     cdef hb_position_t italics_correction = 0
 *     while count == STATIC_ARRAY_SIZE:
 */
  __pyx_v_start_offset = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1614
 *     cdef unsigned int i
 *     cdef unsigned int start_offset = 0
 *     cdef hb_position_t italics_correction = 0             # <<<<<<<<<<<<<<
 *     while count == STATIC_ARRAY_SIZE:
 *         hb_ot_math_get_glyph_assembly(font._hb_font,
 */
  __pyx_v_italics_correction = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1615
 *     cdef unsigned int start_offset = 0
 *     cdef hb_position_t italics_correction = 0
 *     while count == STATIC_ARRAY_SIZE:             # <<<<<<<<<<<<<<
 *         hb_ot_math_get_glyph_assembly(font._hb_font,
 *             glyph, hb_direction, start_offset,
 */
  while (1) {
    __pyx_t_3 = (__pyx_v_count == 0x80);
    if (!__pyx_t_3) break;

    /* "uharfbuzz/_harfbuzz.pyx":1617
 *     while count == STATIC_ARRAY_SIZE:
 *         hb_ot_math_get_glyph_assembly(font._hb_font,
 *             glyph, hb_direction, start_offset,             # <<<<<<<<<<<<<<
 *             &count, assembly_array, &italics_correction)
 *         for i in range(count):
 */
    __pyx_t_4 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_glyph); if (unlikely((__pyx_t_4 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1617, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":1616
 *     cdef hb_position_t italics_correction = 0
 *     while count == STATIC_ARRAY_SIZE:
 *         hb_ot_math_get_glyph_assembly(font._hb_font,             # <<<<<<<<<<<<<<
 *             glyph, hb_direction, start_offset,
 *             &count, assembly_array, &italics_correction)
 */
    (void)(hb_ot_math_get_glyph_assembly(__pyx_v_font->_hb_font, __pyx_t_4, __pyx_v_hb_direction, __pyx_v_start_offset, (&__pyx_v_count), __pyx_v_assembly_array, (&__pyx_v_italics_correction)));

    /* "uharfbuzz/_harfbuzz.pyx":1619
 *             glyph, hb_direction, start_offset,
 *             &count, assembly_array, &italics_correction)
 *         for i in range(count):             # <<<<<<<<<<<<<<
 *             assembly.append(
 *                 OTMathGlyphPart(assembly_array[i].glyph, assembly_array[i].start_connector_length,
 */
    __pyx_t_5 = __pyx_v_count;
    __pyx_t_6 = __pyx_t_5;
    for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
      __pyx_v_i = __pyx_t_7;

      /* "uharfbuzz/_harfbuzz.pyx":1621
 *         for i in range(count):
 *             assembly.append(
 *                 OTMathGlyphPart(assembly_array[i].glyph, assembly_array[i].start_connector_length,             # <<<<<<<<<<<<<<
 *                     assembly_array[i].end_connector_length, assembly_array[i].full_advance,
 *                     OTMathGlyphPartFlags(assembly_array[i].flags)))
 */
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_OTMathGlyphPart); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1621, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = __Pyx_PyInt_From_hb_codepoint_t((__pyx_v_assembly_array[__pyx_v_i]).glyph); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1621, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = __Pyx_PyInt_From_hb_position_t((__pyx_v_assembly_array[__pyx_v_i]).start_connector_length); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1621, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);

      /* "uharfbuzz/_harfbuzz.pyx":1622
 *             assembly.append(
 *                 OTMathGlyphPart(assembly_array[i].glyph, assembly_array[i].start_connector_length,
 *                     assembly_array[i].end_connector_length, assembly_array[i].full_advance,             # <<<<<<<<<<<<<<
 *                     OTMathGlyphPartFlags(assembly_array[i].flags)))
 *         start_offset += count
 */
      __pyx_t_11 = __Pyx_PyInt_From_hb_position_t((__pyx_v_assembly_array[__pyx_v_i]).end_connector_length); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1622, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_12 = __Pyx_PyInt_From_hb_position_t((__pyx_v_assembly_array[__pyx_v_i]).full_advance); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1622, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);

      /* "uharfbuzz/_harfbuzz.pyx":1623
 *                 OTMathGlyphPart(assembly_array[i].glyph, assembly_array[i].start_connector_length,
 *                     assembly_array[i].end_connector_length, assembly_array[i].full_advance,
 *                     OTMathGlyphPartFlags(assembly_array[i].flags)))             # <<<<<<<<<<<<<<
 *         start_offset += count
 *     return assembly, italics_correction
 */
      __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_OTMathGlyphPartFlags); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1623, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_15 = __Pyx_PyInt_From_hb_ot_math_glyph_part_flags_t((__pyx_v_assembly_array[__pyx_v_i]).flags); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1623, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_15);
      __pyx_t_16 = NULL;
      __pyx_t_17 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_14))) {
        __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_14);
        if (likely(__pyx_t_16)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
          __Pyx_INCREF(__pyx_t_16);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_14, function);
          __pyx_t_17 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_16, __pyx_t_15};
        __pyx_t_13 = __Pyx_PyObject_FastCall(__pyx_t_14, __pyx_callargs+1-__pyx_t_17, 1+__pyx_t_17);
        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1623, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      }
      __pyx_t_14 = NULL;
      __pyx_t_17 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_14)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_14);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
          __pyx_t_17 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[6] = {__pyx_t_14, __pyx_t_9, __pyx_t_10, __pyx_t_11, __pyx_t_12, __pyx_t_13};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_17, 5+__pyx_t_17);
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1621, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }

      /* "uharfbuzz/_harfbuzz.pyx":1620
 *             &count, assembly_array, &italics_correction)
 *         for i in range(count):
 *             assembly.append(             # <<<<<<<<<<<<<<
 *                 OTMathGlyphPart(assembly_array[i].glyph, assembly_array[i].start_connector_length,
 *                     assembly_array[i].end_connector_length, assembly_array[i].full_advance,
 */
      __pyx_t_18 = __Pyx_PyList_Append(__pyx_v_assembly, __pyx_t_1); if (unlikely(__pyx_t_18 == ((int)-1))) __PYX_ERR(0, 1620, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }

    /* "uharfbuzz/_harfbuzz.pyx":1624
 *                     assembly_array[i].end_connector_length, assembly_array[i].full_advance,
 *                     OTMathGlyphPartFlags(assembly_array[i].flags)))
 *         start_offset += count             # <<<<<<<<<<<<<<
 *     return assembly, italics_correction
 * 
 */
    __pyx_v_start_offset = (__pyx_v_start_offset + __pyx_v_count);
  }

  /* "uharfbuzz/_harfbuzz.pyx":1625
 *                     OTMathGlyphPartFlags(assembly_array[i].flags)))
 *         start_offset += count
 *     return assembly, italics_correction             # <<<<<<<<<<<<<<
 * 
 * def ot_font_set_funcs(Font font):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_hb_position_t(__pyx_v_italics_correction); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1625, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(__pyx_v_assembly);
  __Pyx_GIVEREF(__pyx_v_assembly);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_assembly)) __PYX_ERR(0, 1625, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_1)) __PYX_ERR(0, 1625, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_r = ((PyObject*)__pyx_t_8);
  __pyx_t_8 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1603
 * )
 * 
 * def ot_math_get_glyph_assembly(font: Font,             # <<<<<<<<<<<<<<
 *                                glyph: int,
 *                                direction: str) -> Tuple[List[OTMathGlyphPart], int]:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_math_get_glyph_assembly", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_packed);
  __Pyx_XDECREF(__pyx_v_assembly);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1627
 *     return assembly, italics_correction
 * 
 * def ot_font_set_funcs(Font font):             # <<<<<<<<<<<<<<
 *     hb_ot_font_set_funcs(font._hb_font)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_75ot_font_set_funcs(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_75ot_font_set_funcs = {"ot_font_set_funcs", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_75ot_font_set_funcs, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_75ot_font_set_funcs(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_font_set_funcs (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_font,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_font)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1627, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "ot_font_set_funcs") < 0)) __PYX_ERR(0, 1627, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_font = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ot_font_set_funcs", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1627, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ot_font_set_funcs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_font), __pyx_ptype_9uharfbuzz_9_harfbuzz_Font, 1, "font", 0))) __PYX_ERR(0, 1627, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_74ot_font_set_funcs(__pyx_self, __pyx_v_font);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_74ot_font_set_funcs(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ot_font_set_funcs", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1628
 * 
 * def ot_font_set_funcs(Font font):
 *     hb_ot_font_set_funcs(font._hb_font)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  hb_ot_font_set_funcs(__pyx_v_font->_hb_font);

  /* "uharfbuzz/_harfbuzz.pyx":1627
 *     return assembly, italics_correction
 * 
 * def ot_font_set_funcs(Font font):             # <<<<<<<<<<<<<<
 *     hb_ot_font_set_funcs(font._hb_font)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1663
 * 
 * class Color(namedtuple("Color", ["red", "green", "blue", "alpha"])):
 *     def to_int(self) -> int:             # <<<<<<<<<<<<<<
 *         return HB_COLOR(self.blue, self.green, self.red, self.alpha)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_5Color_1to_int(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_5Color_1to_int = {"to_int", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_5Color_1to_int, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_5Color_1to_int(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_self = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("to_int (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_self)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1663, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "to_int") < 0)) __PYX_ERR(0, 1663, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_self = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("to_int", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1663, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Color.to_int", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_5Color_to_int(__pyx_self, __pyx_v_self);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_5Color_to_int(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  uint8_t __pyx_t_2;
  uint8_t __pyx_t_3;
  uint8_t __pyx_t_4;
  uint8_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("to_int", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1664
 * class Color(namedtuple("Color", ["red", "green", "blue", "alpha"])):
 *     def to_int(self) -> int:
 *         return HB_COLOR(self.blue, self.green, self.red, self.alpha)             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_blue); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_uint8_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1664, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_green); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_As_uint8_t(__pyx_t_1); if (unlikely((__pyx_t_3 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1664, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_red); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyInt_As_uint8_t(__pyx_t_1); if (unlikely((__pyx_t_4 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1664, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_alpha); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyInt_As_uint8_t(__pyx_t_1); if (unlikely((__pyx_t_5 == ((uint8_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1664, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_uint32_t(HB_COLOR(__pyx_t_2, __pyx_t_3, __pyx_t_4, __pyx_t_5)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(__Pyx_Py3Int_CheckExact(__pyx_t_1)) || __Pyx_RaiseUnexpectedTypeError("int", __pyx_t_1))) __PYX_ERR(0, 1664, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1663
 * 
 * class Color(namedtuple("Color", ["red", "green", "blue", "alpha"])):
 *     def to_int(self) -> int:             # <<<<<<<<<<<<<<
 *         return HB_COLOR(self.blue, self.green, self.red, self.alpha)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Color.to_int", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1666
 *         return HB_COLOR(self.blue, self.green, self.red, self.alpha)
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def from_int(value: int) -> Color:
 *         r = hb_color_get_red(value)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_5Color_3from_int(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_5Color_3from_int = {"from_int", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_5Color_3from_int, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_5Color_3from_int(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_value = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("from_int (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1666, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "from_int") < 0)) __PYX_ERR(0, 1666, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_value = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("from_int", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1666, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Color.from_int", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyInt_Type), 0, "value", 1))) __PYX_ERR(0, 1667, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_5Color_2from_int(__pyx_self, __pyx_v_value);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_5Color_2from_int(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_value) {
  uint8_t __pyx_v_r;
  uint8_t __pyx_v_g;
  uint8_t __pyx_v_b;
  uint8_t __pyx_v_a;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  hb_color_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("from_int", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1668
 *     @staticmethod
 *     def from_int(value: int) -> Color:
 *         r = hb_color_get_red(value)             # <<<<<<<<<<<<<<
 *         g = hb_color_get_green(value)
 *         b = hb_color_get_blue(value)
 */
  __pyx_t_1 = __Pyx_PyInt_As_uint32_t(__pyx_v_value); if (unlikely((__pyx_t_1 == ((hb_color_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1668, __pyx_L1_error)
  __pyx_v_r = hb_color_get_red(__pyx_t_1);

  /* "uharfbuzz/_harfbuzz.pyx":1669
 *     def from_int(value: int) -> Color:
 *         r = hb_color_get_red(value)
 *         g = hb_color_get_green(value)             # <<<<<<<<<<<<<<
 *         b = hb_color_get_blue(value)
 *         a = hb_color_get_alpha(value)
 */
  __pyx_t_1 = __Pyx_PyInt_As_uint32_t(__pyx_v_value); if (unlikely((__pyx_t_1 == ((hb_color_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1669, __pyx_L1_error)
  __pyx_v_g = hb_color_get_green(__pyx_t_1);

  /* "uharfbuzz/_harfbuzz.pyx":1670
 *         r = hb_color_get_red(value)
 *         g = hb_color_get_green(value)
 *         b = hb_color_get_blue(value)             # <<<<<<<<<<<<<<
 *         a = hb_color_get_alpha(value)
 *         return Color(r, g, b, a)
 */
  __pyx_t_1 = __Pyx_PyInt_As_uint32_t(__pyx_v_value); if (unlikely((__pyx_t_1 == ((hb_color_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1670, __pyx_L1_error)
  __pyx_v_b = hb_color_get_blue(__pyx_t_1);

  /* "uharfbuzz/_harfbuzz.pyx":1671
 *         g = hb_color_get_green(value)
 *         b = hb_color_get_blue(value)
 *         a = hb_color_get_alpha(value)             # <<<<<<<<<<<<<<
 *         return Color(r, g, b, a)
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_As_uint32_t(__pyx_v_value); if (unlikely((__pyx_t_1 == ((hb_color_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1671, __pyx_L1_error)
  __pyx_v_a = hb_color_get_alpha(__pyx_t_1);

  /* "uharfbuzz/_harfbuzz.pyx":1672
 *         b = hb_color_get_blue(value)
 *         a = hb_color_get_alpha(value)
 *         return Color(r, g, b, a)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Color); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_uint8_t(__pyx_v_r); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_uint8_t(__pyx_v_g); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_uint8_t(__pyx_v_b); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_uint8_t(__pyx_v_a); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1672, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = NULL;
  __pyx_t_9 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_9 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[5] = {__pyx_t_8, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_9, 4+__pyx_t_9);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1672, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1666
 *         return HB_COLOR(self.blue, self.green, self.red, self.alpha)
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def from_int(value: int) -> Color:
 *         r = hb_color_get_red(value)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Color.from_int", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1687
 *     cdef hb_color_line_t* _color_line
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._color_line = NULL
 * 
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_9ColorLine_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_9ColorLine_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_9ColorLine___cinit__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_9ColorLine___cinit__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine *__pyx_v_self) {
  int __pyx_r;

  /* "uharfbuzz/_harfbuzz.pyx":1688
 * 
 *     def __cinit__(self):
 *         self._color_line = NULL             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  __pyx_v_self->_color_line = NULL;

  /* "uharfbuzz/_harfbuzz.pyx":1687
 *     cdef hb_color_line_t* _color_line
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._color_line = NULL
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1691
 * 
 *     @staticmethod
 *     cdef ColorLine from_ptr(hb_color_line_t* color_line):             # <<<<<<<<<<<<<<
 *         cdef ColorLine wrapper = ColorLine()
 *         wrapper._color_line = color_line
 */

static struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine *__pyx_f_9uharfbuzz_9_harfbuzz_9ColorLine_from_ptr(hb_color_line_t *__pyx_v_color_line) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine *__pyx_v_wrapper = 0;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("from_ptr", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1692
 *     @staticmethod
 *     cdef ColorLine from_ptr(hb_color_line_t* color_line):
 *         cdef ColorLine wrapper = ColorLine()             # <<<<<<<<<<<<<<
 *         wrapper._color_line = color_line
 *         return wrapper
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_ColorLine)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1693
 *     cdef ColorLine from_ptr(hb_color_line_t* color_line):
 *         cdef ColorLine wrapper = ColorLine()
 *         wrapper._color_line = color_line             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __pyx_v_wrapper->_color_line = __pyx_v_color_line;

  /* "uharfbuzz/_harfbuzz.pyx":1694
 *         cdef ColorLine wrapper = ColorLine()
 *         wrapper._color_line = color_line
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = __pyx_v_wrapper;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1691
 * 
 *     @staticmethod
 *     cdef ColorLine from_ptr(hb_color_line_t* color_line):             # <<<<<<<<<<<<<<
 *         cdef ColorLine wrapper = ColorLine()
 *         wrapper._color_line = color_line
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ColorLine.from_ptr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1697
 * 
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def color_stops(self) -> Sequence[ColorStop]:
 *         if self._color_line is NULL:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9ColorLine_11color_stops_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9ColorLine_11color_stops_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_9ColorLine_11color_stops___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9ColorLine_11color_stops___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine *__pyx_v_self) {
  unsigned int __pyx_v_stop_count;
  hb_color_stop_t __pyx_v_stops_array[0x80];
  PyObject *__pyx_v_stops = 0;
  unsigned int __pyx_v_i;
  unsigned int __pyx_v_start_offset;
  hb_color_stop_t __pyx_v_c_stop;
  PyObject *__pyx_v_py_color = NULL;
  PyObject *__pyx_v_stop = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  unsigned int __pyx_t_3;
  unsigned int __pyx_t_4;
  unsigned int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1699
 *     @property
 *     def color_stops(self) -> Sequence[ColorStop]:
 *         if self._color_line is NULL:             # <<<<<<<<<<<<<<
 *             return []
 *         cdef unsigned int stop_count = STATIC_ARRAY_SIZE
 */
  __pyx_t_1 = (__pyx_v_self->_color_line == NULL);
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":1700
 *     def color_stops(self) -> Sequence[ColorStop]:
 *         if self._color_line is NULL:
 *             return []             # <<<<<<<<<<<<<<
 *         cdef unsigned int stop_count = STATIC_ARRAY_SIZE
 *         cdef hb_color_stop_t stops_array[STATIC_ARRAY_SIZE]
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1700, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "uharfbuzz/_harfbuzz.pyx":1699
 *     @property
 *     def color_stops(self) -> Sequence[ColorStop]:
 *         if self._color_line is NULL:             # <<<<<<<<<<<<<<
 *             return []
 *         cdef unsigned int stop_count = STATIC_ARRAY_SIZE
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":1701
 *         if self._color_line is NULL:
 *             return []
 *         cdef unsigned int stop_count = STATIC_ARRAY_SIZE             # <<<<<<<<<<<<<<
 *         cdef hb_color_stop_t stops_array[STATIC_ARRAY_SIZE]
 *         cdef list stops = []
 */
  __pyx_v_stop_count = 0x80;

  /* "uharfbuzz/_harfbuzz.pyx":1703
 *         cdef unsigned int stop_count = STATIC_ARRAY_SIZE
 *         cdef hb_color_stop_t stops_array[STATIC_ARRAY_SIZE]
 *         cdef list stops = []             # <<<<<<<<<<<<<<
 *         cdef unsigned int i
 *         cdef unsigned int start_offset = 0
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_stops = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1705
 *         cdef list stops = []
 *         cdef unsigned int i
 *         cdef unsigned int start_offset = 0             # <<<<<<<<<<<<<<
 *         while stop_count == STATIC_ARRAY_SIZE:
 *             hb_color_line_get_color_stops(
 */
  __pyx_v_start_offset = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1706
 *         cdef unsigned int i
 *         cdef unsigned int start_offset = 0
 *         while stop_count == STATIC_ARRAY_SIZE:             # <<<<<<<<<<<<<<
 *             hb_color_line_get_color_stops(
 *                 self._color_line, start_offset, &stop_count, stops_array)
 */
  while (1) {
    __pyx_t_1 = (__pyx_v_stop_count == 0x80);
    if (!__pyx_t_1) break;

    /* "uharfbuzz/_harfbuzz.pyx":1707
 *         cdef unsigned int start_offset = 0
 *         while stop_count == STATIC_ARRAY_SIZE:
 *             hb_color_line_get_color_stops(             # <<<<<<<<<<<<<<
 *                 self._color_line, start_offset, &stop_count, stops_array)
 *             for i in range(stop_count):
 */
    (void)(hb_color_line_get_color_stops(__pyx_v_self->_color_line, __pyx_v_start_offset, (&__pyx_v_stop_count), __pyx_v_stops_array));

    /* "uharfbuzz/_harfbuzz.pyx":1709
 *             hb_color_line_get_color_stops(
 *                 self._color_line, start_offset, &stop_count, stops_array)
 *             for i in range(stop_count):             # <<<<<<<<<<<<<<
 *                 c_stop = stops_array[i]
 *                 py_color = Color.from_int(c_stop.color)
 */
    __pyx_t_3 = __pyx_v_stop_count;
    __pyx_t_4 = __pyx_t_3;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "uharfbuzz/_harfbuzz.pyx":1710
 *                 self._color_line, start_offset, &stop_count, stops_array)
 *             for i in range(stop_count):
 *                 c_stop = stops_array[i]             # <<<<<<<<<<<<<<
 *                 py_color = Color.from_int(c_stop.color)
 *                 stop = ColorStop(c_stop.offset, <bint>c_stop.is_foreground, py_color)
 */
      __pyx_v_c_stop = (__pyx_v_stops_array[__pyx_v_i]);

      /* "uharfbuzz/_harfbuzz.pyx":1711
 *             for i in range(stop_count):
 *                 c_stop = stops_array[i]
 *                 py_color = Color.from_int(c_stop.color)             # <<<<<<<<<<<<<<
 *                 stop = ColorStop(c_stop.offset, <bint>c_stop.is_foreground, py_color)
 *                 stops.append(stop)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_Color); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1711, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_from_int); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1711, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyInt_From_uint32_t(__pyx_v_c_stop.color); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1711, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_8 = NULL;
      __pyx_t_9 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_9 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_6};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1711, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_XDECREF_SET(__pyx_v_py_color, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "uharfbuzz/_harfbuzz.pyx":1712
 *                 c_stop = stops_array[i]
 *                 py_color = Color.from_int(c_stop.color)
 *                 stop = ColorStop(c_stop.offset, <bint>c_stop.is_foreground, py_color)             # <<<<<<<<<<<<<<
 *                 stops.append(stop)
 *             start_offset += stop_count
 */
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_ColorStop); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1712, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_6 = PyFloat_FromDouble(__pyx_v_c_stop.offset); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1712, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_8 = __Pyx_PyBool_FromLong((__pyx_v_c_stop.is_foreground != 0)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1712, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_10 = NULL;
      __pyx_t_9 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
          __pyx_t_9 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[4] = {__pyx_t_10, __pyx_t_6, __pyx_t_8, __pyx_v_py_color};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 3+__pyx_t_9);
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1712, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      }
      __Pyx_XDECREF_SET(__pyx_v_stop, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "uharfbuzz/_harfbuzz.pyx":1713
 *                 py_color = Color.from_int(c_stop.color)
 *                 stop = ColorStop(c_stop.offset, <bint>c_stop.is_foreground, py_color)
 *                 stops.append(stop)             # <<<<<<<<<<<<<<
 *             start_offset += stop_count
 *         return stops
 */
      __pyx_t_11 = __Pyx_PyList_Append(__pyx_v_stops, __pyx_v_stop); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 1713, __pyx_L1_error)
    }

    /* "uharfbuzz/_harfbuzz.pyx":1714
 *                 stop = ColorStop(c_stop.offset, <bint>c_stop.is_foreground, py_color)
 *                 stops.append(stop)
 *             start_offset += stop_count             # <<<<<<<<<<<<<<
 *         return stops
 * 
 */
    __pyx_v_start_offset = (__pyx_v_start_offset + __pyx_v_stop_count);
  }

  /* "uharfbuzz/_harfbuzz.pyx":1715
 *                 stops.append(stop)
 *             start_offset += stop_count
 *         return stops             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_stops);
  __pyx_r = __pyx_v_stops;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1697
 * 
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def color_stops(self) -> Sequence[ColorStop]:
 *         if self._color_line is NULL:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ColorLine.color_stops.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_stops);
  __Pyx_XDECREF(__pyx_v_py_color);
  __Pyx_XDECREF(__pyx_v_stop);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1717
 *         return stops
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def extend(self) -> PaintExtend:
 *         if self._color_line is NULL:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9ColorLine_6extend_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9ColorLine_6extend_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_9ColorLine_6extend___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9ColorLine_6extend___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1719
 *     @property
 *     def extend(self) -> PaintExtend:
 *         if self._color_line is NULL:             # <<<<<<<<<<<<<<
 *             return None
 *         return PaintExtend(hb_color_line_get_extend(self._color_line))
 */
  __pyx_t_1 = (__pyx_v_self->_color_line == NULL);
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":1720
 *     def extend(self) -> PaintExtend:
 *         if self._color_line is NULL:
 *             return None             # <<<<<<<<<<<<<<
 *         return PaintExtend(hb_color_line_get_extend(self._color_line))
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uharfbuzz/_harfbuzz.pyx":1719
 *     @property
 *     def extend(self) -> PaintExtend:
 *         if self._color_line is NULL:             # <<<<<<<<<<<<<<
 *             return None
 *         return PaintExtend(hb_color_line_get_extend(self._color_line))
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":1721
 *         if self._color_line is NULL:
 *             return None
 *         return PaintExtend(hb_color_line_get_extend(self._color_line))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_PaintExtend); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1721, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_hb_paint_extend_t(hb_color_line_get_extend(__pyx_v_self->_color_line)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1721, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1721, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1717
 *         return stops
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def extend(self) -> PaintExtend:
 *         if self._color_line is NULL:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ColorLine.extend.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9ColorLine_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_9ColorLine_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9ColorLine_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9ColorLine_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_9ColorLine_2__reduce_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9ColorLine_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ColorLine.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9ColorLine_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_9ColorLine_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9ColorLine_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9ColorLine_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ColorLine.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_9ColorLine_4__setstate_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9ColorLine_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.ColorLine.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1724
 * 
 * 
 * cdef void _paint_push_transform_func(             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_t *funcs,
 *         void *paint_data,
 */

static void __pyx_f_9uharfbuzz_9_harfbuzz__paint_push_transform_func(CYTHON_UNUSED hb_paint_funcs_t *__pyx_v_funcs, void *__pyx_v_paint_data, float __pyx_v_xx, float __pyx_v_yx, float __pyx_v_xy, float __pyx_v_yy, float __pyx_v_dx, float __pyx_v_dy, void *__pyx_v_user_data) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_py_funcs = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  unsigned int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_paint_push_transform_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1734
 *         float dy,
 *         void *user_data) noexcept:
 *     py_funcs = <PaintFuncs>user_data             # <<<<<<<<<<<<<<
 *     py_funcs._push_transform_func(xx, yx, xy, yy, dx, dy, <object>paint_data)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_user_data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_py_funcs = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1735
 *         void *user_data) noexcept:
 *     py_funcs = <PaintFuncs>user_data
 *     py_funcs._push_transform_func(xx, yx, xy, yy, dx, dy, <object>paint_data)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_xx); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_yx); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_xy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_yy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyFloat_FromDouble(__pyx_v_dx); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyFloat_FromDouble(__pyx_v_dy); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_v_py_funcs->_push_transform_func);
  __pyx_t_8 = __pyx_v_py_funcs->_push_transform_func; __pyx_t_9 = NULL;
  __pyx_t_10 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
      __pyx_t_10 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[8] = {__pyx_t_9, __pyx_t_2, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, ((PyObject *)__pyx_v_paint_data)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_10, 7+__pyx_t_10);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1735, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1724
 * 
 * 
 * cdef void _paint_push_transform_func(             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_t *funcs,
 *         void *paint_data,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz._paint_push_transform_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_py_funcs);
  __Pyx_RefNannyFinishContext();
}

/* "uharfbuzz/_harfbuzz.pyx":1738
 * 
 * 
 * cdef void _paint_pop_transform_func(             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_t *funcs,
 *         void *paint_data,
 */

static void __pyx_f_9uharfbuzz_9_harfbuzz__paint_pop_transform_func(CYTHON_UNUSED hb_paint_funcs_t *__pyx_v_funcs, void *__pyx_v_paint_data, void *__pyx_v_user_data) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_py_funcs = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_paint_pop_transform_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1742
 *         void *paint_data,
 *         void *user_data) noexcept:
 *     py_funcs = <PaintFuncs>user_data             # <<<<<<<<<<<<<<
 *     py_funcs._pop_transform_func(<object>paint_data)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_user_data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_py_funcs = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1743
 *         void *user_data) noexcept:
 *     py_funcs = <PaintFuncs>user_data
 *     py_funcs._pop_transform_func(<object>paint_data)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_INCREF(__pyx_v_py_funcs->_pop_transform_func);
  __pyx_t_2 = __pyx_v_py_funcs->_pop_transform_func; __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, ((PyObject *)__pyx_v_paint_data)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1743, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1738
 * 
 * 
 * cdef void _paint_pop_transform_func(             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_t *funcs,
 *         void *paint_data,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz._paint_pop_transform_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_py_funcs);
  __Pyx_RefNannyFinishContext();
}

/* "uharfbuzz/_harfbuzz.pyx":1746
 * 
 * 
 * cdef hb_bool_t _paint_color_glyph_func(             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_t *funcs,
 *         void *paint_data,
 */

static hb_bool_t __pyx_f_9uharfbuzz_9_harfbuzz__paint_color_glyph_func(CYTHON_UNUSED hb_paint_funcs_t *__pyx_v_funcs, void *__pyx_v_paint_data, hb_codepoint_t __pyx_v_glyph, CYTHON_UNUSED hb_font_t *__pyx_v_font, void *__pyx_v_user_data) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_py_funcs = NULL;
  hb_bool_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_paint_color_glyph_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1752
 *         hb_font_t *font,
 *         void *user_data) noexcept:
 *     py_funcs = <PaintFuncs>user_data             # <<<<<<<<<<<<<<
 *     if py_funcs._color_glyph_func(glyph, <object>paint_data):
 *         return 1
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_user_data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_py_funcs = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1753
 *         void *user_data) noexcept:
 *     py_funcs = <PaintFuncs>user_data
 *     if py_funcs._color_glyph_func(glyph, <object>paint_data):             # <<<<<<<<<<<<<<
 *         return 1
 *     return 0
 */
  __pyx_t_2 = __Pyx_PyInt_From_hb_codepoint_t(__pyx_v_glyph); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1753, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_py_funcs->_color_glyph_func);
  __pyx_t_3 = __pyx_v_py_funcs->_color_glyph_func; __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_t_2, ((PyObject *)__pyx_v_paint_data)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1753, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 1753, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_6) {

    /* "uharfbuzz/_harfbuzz.pyx":1754
 *     py_funcs = <PaintFuncs>user_data
 *     if py_funcs._color_glyph_func(glyph, <object>paint_data):
 *         return 1             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "uharfbuzz/_harfbuzz.pyx":1753
 *         void *user_data) noexcept:
 *     py_funcs = <PaintFuncs>user_data
 *     if py_funcs._color_glyph_func(glyph, <object>paint_data):             # <<<<<<<<<<<<<<
 *         return 1
 *     return 0
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":1755
 *     if py_funcs._color_glyph_func(glyph, <object>paint_data):
 *         return 1
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1746
 * 
 * 
 * cdef hb_bool_t _paint_color_glyph_func(             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_t *funcs,
 *         void *paint_data,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz._paint_color_glyph_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_py_funcs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1758
 * 
 * 
 * cdef void _paint_push_clip_glyph_func(             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_t *funcs,
 *         void *paint_data,
 */

static void __pyx_f_9uharfbuzz_9_harfbuzz__paint_push_clip_glyph_func(CYTHON_UNUSED hb_paint_funcs_t *__pyx_v_funcs, void *__pyx_v_paint_data, hb_codepoint_t __pyx_v_glyph, CYTHON_UNUSED hb_font_t *__pyx_v_font, void *__pyx_v_user_data) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_py_funcs = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_paint_push_clip_glyph_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1764
 *         hb_font_t *font,
 *         void *user_data) noexcept:
 *     py_funcs = <PaintFuncs>user_data             # <<<<<<<<<<<<<<
 *     py_funcs._push_clip_glyph_func(glyph, <object>paint_data)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_user_data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_py_funcs = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1765
 *         void *user_data) noexcept:
 *     py_funcs = <PaintFuncs>user_data
 *     py_funcs._push_clip_glyph_func(glyph, <object>paint_data)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_hb_codepoint_t(__pyx_v_glyph); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1765, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_py_funcs->_push_clip_glyph_func);
  __pyx_t_3 = __pyx_v_py_funcs->_push_clip_glyph_func; __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_t_2, ((PyObject *)__pyx_v_paint_data)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1765, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1758
 * 
 * 
 * cdef void _paint_push_clip_glyph_func(             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_t *funcs,
 *         void *paint_data,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz._paint_push_clip_glyph_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_py_funcs);
  __Pyx_RefNannyFinishContext();
}

/* "uharfbuzz/_harfbuzz.pyx":1768
 * 
 * 
 * cdef void _paint_push_clip_rectangle_func(             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_t *funcs,
 *         void *paint_data,
 */

static void __pyx_f_9uharfbuzz_9_harfbuzz__paint_push_clip_rectangle_func(CYTHON_UNUSED hb_paint_funcs_t *__pyx_v_funcs, void *__pyx_v_paint_data, float __pyx_v_xmin, float __pyx_v_ymin, float __pyx_v_xmax, float __pyx_v_ymax, void *__pyx_v_user_data) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_py_funcs = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_paint_push_clip_rectangle_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1776
 *         float ymax,
 *         void *user_data) noexcept:
 *     py_funcs = <PaintFuncs>user_data             # <<<<<<<<<<<<<<
 *     py_funcs._push_clip_rectangle_func(xmin, ymin, xmax, ymax, <object>paint_data)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_user_data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_py_funcs = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1777
 *         void *user_data) noexcept:
 *     py_funcs = <PaintFuncs>user_data
 *     py_funcs._push_clip_rectangle_func(xmin, ymin, xmax, ymax, <object>paint_data)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_xmin); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_ymin); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_xmax); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_ymax); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_py_funcs->_push_clip_rectangle_func);
  __pyx_t_6 = __pyx_v_py_funcs->_push_clip_rectangle_func; __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[6] = {__pyx_t_7, __pyx_t_2, __pyx_t_3, __pyx_t_4, __pyx_t_5, ((PyObject *)__pyx_v_paint_data)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 5+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1777, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1768
 * 
 * 
 * cdef void _paint_push_clip_rectangle_func(             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_t *funcs,
 *         void *paint_data,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz._paint_push_clip_rectangle_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_py_funcs);
  __Pyx_RefNannyFinishContext();
}

/* "uharfbuzz/_harfbuzz.pyx":1780
 * 
 * 
 * cdef void _paint_pop_clip_func(             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_t *funcs,
 *         void *paint_data,
 */

static void __pyx_f_9uharfbuzz_9_harfbuzz__paint_pop_clip_func(CYTHON_UNUSED hb_paint_funcs_t *__pyx_v_funcs, void *__pyx_v_paint_data, void *__pyx_v_user_data) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_py_funcs = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_paint_pop_clip_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1784
 *         void *paint_data,
 *         void *user_data) noexcept:
 *     py_funcs = <PaintFuncs>user_data             # <<<<<<<<<<<<<<
 *     py_funcs._pop_clip_func(<object>paint_data)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_user_data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_py_funcs = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1785
 *         void *user_data) noexcept:
 *     py_funcs = <PaintFuncs>user_data
 *     py_funcs._pop_clip_func(<object>paint_data)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_INCREF(__pyx_v_py_funcs->_pop_clip_func);
  __pyx_t_2 = __pyx_v_py_funcs->_pop_clip_func; __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, ((PyObject *)__pyx_v_paint_data)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1785, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1780
 * 
 * 
 * cdef void _paint_pop_clip_func(             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_t *funcs,
 *         void *paint_data,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz._paint_pop_clip_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_py_funcs);
  __Pyx_RefNannyFinishContext();
}

/* "uharfbuzz/_harfbuzz.pyx":1788
 * 
 * 
 * cdef void _paint_color_func(             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_t *funcs,
 *         void *paint_data,
 */

static void __pyx_f_9uharfbuzz_9_harfbuzz__paint_color_func(CYTHON_UNUSED hb_paint_funcs_t *__pyx_v_funcs, void *__pyx_v_paint_data, hb_bool_t __pyx_v_is_foreground, hb_color_t __pyx_v_color, void *__pyx_v_user_data) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_py_funcs = NULL;
  PyObject *__pyx_v_py_color = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_paint_color_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1794
 *         hb_color_t color,
 *         void *user_data) noexcept:
 *     py_funcs = <PaintFuncs>user_data             # <<<<<<<<<<<<<<
 *     py_color: Color = Color.from_int(color)
 *     py_funcs._color_func(py_color, <bint>is_foreground, <object>paint_data)
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_user_data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_py_funcs = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1795
 *         void *user_data) noexcept:
 *     py_funcs = <PaintFuncs>user_data
 *     py_color: Color = Color.from_int(color)             # <<<<<<<<<<<<<<
 *     py_funcs._color_func(py_color, <bint>is_foreground, <object>paint_data)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Color); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_from_int); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyInt_From_uint32_t(__pyx_v_color); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1795, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v_py_color = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1796
 *     py_funcs = <PaintFuncs>user_data
 *     py_color: Color = Color.from_int(color)
 *     py_funcs._color_func(py_color, <bint>is_foreground, <object>paint_data)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_PyBool_FromLong((__pyx_v_is_foreground != 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1796, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_py_funcs->_color_func);
  __pyx_t_2 = __pyx_v_py_funcs->_color_func; __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_4, __pyx_v_py_color, __pyx_t_3, ((PyObject *)__pyx_v_paint_data)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 3+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1796, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1788
 * 
 * 
 * cdef void _paint_color_func(             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_t *funcs,
 *         void *paint_data,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz._paint_color_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_py_funcs);
  __Pyx_XDECREF(__pyx_v_py_color);
  __Pyx_RefNannyFinishContext();
}

/* "uharfbuzz/_harfbuzz.pyx":1799
 * 
 * 
 * cdef hb_bool_t _paint_image_func(             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_t *funcs,
 *         void *paint_data,
 */

static hb_bool_t __pyx_f_9uharfbuzz_9_harfbuzz__paint_image_func(CYTHON_UNUSED hb_paint_funcs_t *__pyx_v_funcs, void *__pyx_v_paint_data, hb_blob_t *__pyx_v_image, unsigned int __pyx_v_width, unsigned int __pyx_v_height, hb_tag_t __pyx_v_format, float __pyx_v_slant, hb_glyph_extents_t *__pyx_v_extents, void *__pyx_v_user_data) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_py_funcs = NULL;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *__pyx_v_py_image = NULL;
  PyObject *__pyx_v_py_format = NULL;
  PyObject *__pyx_v_py_extents = NULL;
  hb_bool_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_paint_image_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1809
 *         hb_glyph_extents_t *extents,
 *         void *user_data) noexcept:
 *     py_funcs = <PaintFuncs>user_data             # <<<<<<<<<<<<<<
 *     py_image = Blob.from_ptr(hb_blob_reference(image))
 *     py_format = hb_tag_to_string(format, NULL)
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_user_data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_py_funcs = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1810
 *         void *user_data) noexcept:
 *     py_funcs = <PaintFuncs>user_data
 *     py_image = Blob.from_ptr(hb_blob_reference(image))             # <<<<<<<<<<<<<<
 *     py_format = hb_tag_to_string(format, NULL)
 *     py_extents = GlyphExtents(extents.x_bearing, extents.y_bearing, extents.width, extents.height)
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_9uharfbuzz_9_harfbuzz_4Blob_from_ptr(hb_blob_reference(__pyx_v_image))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_py_image = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1811
 *     py_funcs = <PaintFuncs>user_data
 *     py_image = Blob.from_ptr(hb_blob_reference(image))
 *     py_format = hb_tag_to_string(format, NULL)             # <<<<<<<<<<<<<<
 *     py_extents = GlyphExtents(extents.x_bearing, extents.y_bearing, extents.width, extents.height)
 *     if py_funcs._image_func(py_image, width, height, py_format, slant, py_extents, <object>paint_data):
 */
  __pyx_t_1 = __Pyx_void_to_None(hb_tag_to_string(__pyx_v_format, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_py_format = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1812
 *     py_image = Blob.from_ptr(hb_blob_reference(image))
 *     py_format = hb_tag_to_string(format, NULL)
 *     py_extents = GlyphExtents(extents.x_bearing, extents.y_bearing, extents.width, extents.height)             # <<<<<<<<<<<<<<
 *     if py_funcs._image_func(py_image, width, height, py_format, slant, py_extents, <object>paint_data):
 *         return 1
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_GlyphExtents); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_hb_position_t(__pyx_v_extents->x_bearing); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_hb_position_t(__pyx_v_extents->y_bearing); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_hb_position_t(__pyx_v_extents->width); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_hb_position_t(__pyx_v_extents->height); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[5] = {__pyx_t_7, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_8, 4+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1812, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_py_extents = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1813
 *     py_format = hb_tag_to_string(format, NULL)
 *     py_extents = GlyphExtents(extents.x_bearing, extents.y_bearing, extents.width, extents.height)
 *     if py_funcs._image_func(py_image, width, height, py_format, slant, py_extents, <object>paint_data):             # <<<<<<<<<<<<<<
 *         return 1
 *     return 0
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_width); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_int(__pyx_v_height); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_slant); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_py_funcs->_image_func);
  __pyx_t_4 = __pyx_v_py_funcs->_image_func; __pyx_t_3 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[8] = {__pyx_t_3, ((PyObject *)__pyx_v_py_image), __pyx_t_2, __pyx_t_6, __pyx_v_py_format, __pyx_t_5, __pyx_v_py_extents, ((PyObject *)__pyx_v_paint_data)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_8, 7+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1813, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 1813, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_9) {

    /* "uharfbuzz/_harfbuzz.pyx":1814
 *     py_extents = GlyphExtents(extents.x_bearing, extents.y_bearing, extents.width, extents.height)
 *     if py_funcs._image_func(py_image, width, height, py_format, slant, py_extents, <object>paint_data):
 *         return 1             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "uharfbuzz/_harfbuzz.pyx":1813
 *     py_format = hb_tag_to_string(format, NULL)
 *     py_extents = GlyphExtents(extents.x_bearing, extents.y_bearing, extents.width, extents.height)
 *     if py_funcs._image_func(py_image, width, height, py_format, slant, py_extents, <object>paint_data):             # <<<<<<<<<<<<<<
 *         return 1
 *     return 0
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":1815
 *     if py_funcs._image_func(py_image, width, height, py_format, slant, py_extents, <object>paint_data):
 *         return 1
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1799
 * 
 * 
 * cdef hb_bool_t _paint_image_func(             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_t *funcs,
 *         void *paint_data,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz._paint_image_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_py_funcs);
  __Pyx_XDECREF((PyObject *)__pyx_v_py_image);
  __Pyx_XDECREF(__pyx_v_py_format);
  __Pyx_XDECREF(__pyx_v_py_extents);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1818
 * 
 * 
 * cdef void _paint_linear_gradient_func(             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_t *funcs,
 *         void *paint_data,
 */

static void __pyx_f_9uharfbuzz_9_harfbuzz__paint_linear_gradient_func(CYTHON_UNUSED hb_paint_funcs_t *__pyx_v_funcs, void *__pyx_v_paint_data, hb_color_line_t *__pyx_v_color_line, float __pyx_v_x0, float __pyx_v_y0, float __pyx_v_x1, float __pyx_v_y1, float __pyx_v_x2, float __pyx_v_y2, void *__pyx_v_user_data) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_py_funcs = NULL;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine *__pyx_v_py_color_line = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  unsigned int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_paint_linear_gradient_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1829
 *         float y2,
 *         void *user_data) noexcept:
 *     py_funcs = <PaintFuncs>user_data             # <<<<<<<<<<<<<<
 *     py_color_line = ColorLine.from_ptr(color_line)
 *     py_funcs._linear_gradient_func(py_color_line, x0, y0, x1, y1, x2, y2, <object>paint_data)
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_user_data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_py_funcs = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1830
 *         void *user_data) noexcept:
 *     py_funcs = <PaintFuncs>user_data
 *     py_color_line = ColorLine.from_ptr(color_line)             # <<<<<<<<<<<<<<
 *     py_funcs._linear_gradient_func(py_color_line, x0, y0, x1, y1, x2, y2, <object>paint_data)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_9uharfbuzz_9_harfbuzz_9ColorLine_from_ptr(__pyx_v_color_line)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_py_color_line = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1831
 *     py_funcs = <PaintFuncs>user_data
 *     py_color_line = ColorLine.from_ptr(color_line)
 *     py_funcs._linear_gradient_func(py_color_line, x0, y0, x1, y1, x2, y2, <object>paint_data)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_x0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_y0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_x1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_y1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyFloat_FromDouble(__pyx_v_x2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyFloat_FromDouble(__pyx_v_y2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_v_py_funcs->_linear_gradient_func);
  __pyx_t_8 = __pyx_v_py_funcs->_linear_gradient_func; __pyx_t_9 = NULL;
  __pyx_t_10 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
      __pyx_t_10 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[9] = {__pyx_t_9, ((PyObject *)__pyx_v_py_color_line), __pyx_t_2, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, ((PyObject *)__pyx_v_paint_data)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_10, 8+__pyx_t_10);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1831, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1818
 * 
 * 
 * cdef void _paint_linear_gradient_func(             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_t *funcs,
 *         void *paint_data,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz._paint_linear_gradient_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_py_funcs);
  __Pyx_XDECREF((PyObject *)__pyx_v_py_color_line);
  __Pyx_RefNannyFinishContext();
}

/* "uharfbuzz/_harfbuzz.pyx":1834
 * 
 * 
 * cdef void _paint_radial_gradient_func(             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_t *funcs,
 *         void *paint_data,
 */

static void __pyx_f_9uharfbuzz_9_harfbuzz__paint_radial_gradient_func(CYTHON_UNUSED hb_paint_funcs_t *__pyx_v_funcs, void *__pyx_v_paint_data, hb_color_line_t *__pyx_v_color_line, float __pyx_v_x0, float __pyx_v_y0, float __pyx_v_r0, float __pyx_v_x1, float __pyx_v_y1, float __pyx_v_r1, void *__pyx_v_user_data) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_py_funcs = NULL;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine *__pyx_v_py_color_line = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  unsigned int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_paint_radial_gradient_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1845
 *         float r1,
 *         void *user_data) noexcept:
 *     py_funcs = <PaintFuncs>user_data             # <<<<<<<<<<<<<<
 *     py_color_line = ColorLine.from_ptr(color_line)
 *     py_funcs._radial_gradient_func(py_color_line, x0, y0, r0, x1, y1, r1, <object>paint_data)
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_user_data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_py_funcs = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1846
 *         void *user_data) noexcept:
 *     py_funcs = <PaintFuncs>user_data
 *     py_color_line = ColorLine.from_ptr(color_line)             # <<<<<<<<<<<<<<
 *     py_funcs._radial_gradient_func(py_color_line, x0, y0, r0, x1, y1, r1, <object>paint_data)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_9uharfbuzz_9_harfbuzz_9ColorLine_from_ptr(__pyx_v_color_line)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1846, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_py_color_line = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1847
 *     py_funcs = <PaintFuncs>user_data
 *     py_color_line = ColorLine.from_ptr(color_line)
 *     py_funcs._radial_gradient_func(py_color_line, x0, y0, r0, x1, y1, r1, <object>paint_data)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_x0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1847, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_y0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1847, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_r0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1847, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_x1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1847, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyFloat_FromDouble(__pyx_v_y1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1847, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyFloat_FromDouble(__pyx_v_r1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1847, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_v_py_funcs->_radial_gradient_func);
  __pyx_t_8 = __pyx_v_py_funcs->_radial_gradient_func; __pyx_t_9 = NULL;
  __pyx_t_10 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
      __pyx_t_10 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[9] = {__pyx_t_9, ((PyObject *)__pyx_v_py_color_line), __pyx_t_2, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, ((PyObject *)__pyx_v_paint_data)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_10, 8+__pyx_t_10);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1847, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1834
 * 
 * 
 * cdef void _paint_radial_gradient_func(             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_t *funcs,
 *         void *paint_data,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz._paint_radial_gradient_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_py_funcs);
  __Pyx_XDECREF((PyObject *)__pyx_v_py_color_line);
  __Pyx_RefNannyFinishContext();
}

/* "uharfbuzz/_harfbuzz.pyx":1850
 * 
 * 
 * cdef void _paint_sweep_gradient_func(             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_t *funcs,
 *         void *paint_data,
 */

static void __pyx_f_9uharfbuzz_9_harfbuzz__paint_sweep_gradient_func(CYTHON_UNUSED hb_paint_funcs_t *__pyx_v_funcs, void *__pyx_v_paint_data, hb_color_line_t *__pyx_v_color_line, float __pyx_v_x0, float __pyx_v_y0, float __pyx_v_start_angle, float __pyx_v_end_angle, void *__pyx_v_user_data) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_py_funcs = NULL;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine *__pyx_v_py_color_line = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_paint_sweep_gradient_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1859
 *         float end_angle,
 *         void *user_data) noexcept:
 *     py_funcs = <PaintFuncs>user_data             # <<<<<<<<<<<<<<
 *     py_color_line = ColorLine.from_ptr(color_line)
 *     py_funcs._sweep_gradient_func(py_color_line, x0, y0, start_angle, end_angle, <object>paint_data)
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_user_data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_py_funcs = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1860
 *         void *user_data) noexcept:
 *     py_funcs = <PaintFuncs>user_data
 *     py_color_line = ColorLine.from_ptr(color_line)             # <<<<<<<<<<<<<<
 *     py_funcs._sweep_gradient_func(py_color_line, x0, y0, start_angle, end_angle, <object>paint_data)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_f_9uharfbuzz_9_harfbuzz_9ColorLine_from_ptr(__pyx_v_color_line)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1860, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_py_color_line = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1861
 *     py_funcs = <PaintFuncs>user_data
 *     py_color_line = ColorLine.from_ptr(color_line)
 *     py_funcs._sweep_gradient_func(py_color_line, x0, y0, start_angle, end_angle, <object>paint_data)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_x0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_y0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_start_angle); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_end_angle); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1861, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_py_funcs->_sweep_gradient_func);
  __pyx_t_6 = __pyx_v_py_funcs->_sweep_gradient_func; __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[7] = {__pyx_t_7, ((PyObject *)__pyx_v_py_color_line), __pyx_t_2, __pyx_t_3, __pyx_t_4, __pyx_t_5, ((PyObject *)__pyx_v_paint_data)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 6+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1861, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1850
 * 
 * 
 * cdef void _paint_sweep_gradient_func(             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_t *funcs,
 *         void *paint_data,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz._paint_sweep_gradient_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_py_funcs);
  __Pyx_XDECREF((PyObject *)__pyx_v_py_color_line);
  __Pyx_RefNannyFinishContext();
}

/* "uharfbuzz/_harfbuzz.pyx":1864
 * 
 * 
 * cdef void _paint_push_group_func(             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_t *funcs,
 *         void *paint_data,
 */

static void __pyx_f_9uharfbuzz_9_harfbuzz__paint_push_group_func(CYTHON_UNUSED hb_paint_funcs_t *__pyx_v_funcs, void *__pyx_v_paint_data, void *__pyx_v_user_data) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_py_funcs = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_paint_push_group_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1868
 *         void *paint_data,
 *         void *user_data) noexcept:
 *     py_funcs = <PaintFuncs>user_data             # <<<<<<<<<<<<<<
 *     py_funcs._push_group_func(<object>paint_data)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_user_data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_py_funcs = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1869
 *         void *user_data) noexcept:
 *     py_funcs = <PaintFuncs>user_data
 *     py_funcs._push_group_func(<object>paint_data)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_INCREF(__pyx_v_py_funcs->_push_group_func);
  __pyx_t_2 = __pyx_v_py_funcs->_push_group_func; __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, ((PyObject *)__pyx_v_paint_data)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1869, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1864
 * 
 * 
 * cdef void _paint_push_group_func(             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_t *funcs,
 *         void *paint_data,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz._paint_push_group_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_py_funcs);
  __Pyx_RefNannyFinishContext();
}

/* "uharfbuzz/_harfbuzz.pyx":1872
 * 
 * 
 * cdef void _paint_pop_group_func(             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_t *funcs,
 *         void *paint_data,
 */

static void __pyx_f_9uharfbuzz_9_harfbuzz__paint_pop_group_func(CYTHON_UNUSED hb_paint_funcs_t *__pyx_v_funcs, void *__pyx_v_paint_data, hb_paint_composite_mode_t __pyx_v_mode, void *__pyx_v_user_data) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_py_funcs = NULL;
  PyObject *__pyx_v_py_mode = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_paint_pop_group_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1877
 *         hb_paint_composite_mode_t mode,
 *         void *user_data) noexcept:
 *     py_funcs = <PaintFuncs>user_data             # <<<<<<<<<<<<<<
 *     py_mode = PaintCompositeMode(mode)
 *     py_funcs._pop_group_func(py_mode, <object>paint_data)
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_user_data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_py_funcs = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1878
 *         void *user_data) noexcept:
 *     py_funcs = <PaintFuncs>user_data
 *     py_mode = PaintCompositeMode(mode)             # <<<<<<<<<<<<<<
 *     py_funcs._pop_group_func(py_mode, <object>paint_data)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_PaintCompositeMode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_hb_paint_composite_mode_t(__pyx_v_mode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1878, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1878, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_py_mode = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1879
 *     py_funcs = <PaintFuncs>user_data
 *     py_mode = PaintCompositeMode(mode)
 *     py_funcs._pop_group_func(py_mode, <object>paint_data)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_INCREF(__pyx_v_py_funcs->_pop_group_func);
  __pyx_t_2 = __pyx_v_py_funcs->_pop_group_func; __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_v_py_mode, ((PyObject *)__pyx_v_paint_data)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1879, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1872
 * 
 * 
 * cdef void _paint_pop_group_func(             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_t *funcs,
 *         void *paint_data,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz._paint_pop_group_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_py_funcs);
  __Pyx_XDECREF(__pyx_v_py_mode);
  __Pyx_RefNannyFinishContext();
}

/* "uharfbuzz/_harfbuzz.pyx":1882
 * 
 * 
 * cdef hb_bool_t _paint_custom_palette_color_func(             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_t *funcs,
 *         void *paint_data,
 */

static hb_bool_t __pyx_f_9uharfbuzz_9_harfbuzz__paint_custom_palette_color_func(CYTHON_UNUSED hb_paint_funcs_t *__pyx_v_funcs, void *__pyx_v_paint_data, unsigned int __pyx_v_color_index, hb_color_t *__pyx_v_color, void *__pyx_v_user_data) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_py_funcs = NULL;
  PyObject *__pyx_v_py_color = NULL;
  hb_bool_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  hb_color_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_paint_custom_palette_color_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1888
 *         hb_color_t *color,
 *         void *user_data) noexcept:
 *     py_funcs = <PaintFuncs>user_data             # <<<<<<<<<<<<<<
 *     py_color: Color = py_funcs._custom_palette_color_func(color_index, <object>paint_data)
 *     if py_color is not None:
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_user_data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_py_funcs = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1889
 *         void *user_data) noexcept:
 *     py_funcs = <PaintFuncs>user_data
 *     py_color: Color = py_funcs._custom_palette_color_func(color_index, <object>paint_data)             # <<<<<<<<<<<<<<
 *     if py_color is not None:
 *         color[0] = py_color.to_int()
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_color_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1889, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_py_funcs->_custom_palette_color_func);
  __pyx_t_3 = __pyx_v_py_funcs->_custom_palette_color_func; __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_t_2, ((PyObject *)__pyx_v_paint_data)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1889, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v_py_color = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1890
 *     py_funcs = <PaintFuncs>user_data
 *     py_color: Color = py_funcs._custom_palette_color_func(color_index, <object>paint_data)
 *     if py_color is not None:             # <<<<<<<<<<<<<<
 *         color[0] = py_color.to_int()
 *         return 1
 */
  __pyx_t_6 = (__pyx_v_py_color != Py_None);
  if (__pyx_t_6) {

    /* "uharfbuzz/_harfbuzz.pyx":1891
 *     py_color: Color = py_funcs._custom_palette_color_func(color_index, <object>paint_data)
 *     if py_color is not None:
 *         color[0] = py_color.to_int()             # <<<<<<<<<<<<<<
 *         return 1
 *     return 0
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_py_color, __pyx_n_s_to_int); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1891, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1891, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_t_7 = __Pyx_PyInt_As_uint32_t(__pyx_t_1); if (unlikely((__pyx_t_7 == ((hb_color_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1891, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    (__pyx_v_color[0]) = __pyx_t_7;

    /* "uharfbuzz/_harfbuzz.pyx":1892
 *     if py_color is not None:
 *         color[0] = py_color.to_int()
 *         return 1             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
    __pyx_r = 1;
    goto __pyx_L0;

    /* "uharfbuzz/_harfbuzz.pyx":1890
 *     py_funcs = <PaintFuncs>user_data
 *     py_color: Color = py_funcs._custom_palette_color_func(color_index, <object>paint_data)
 *     if py_color is not None:             # <<<<<<<<<<<<<<
 *         color[0] = py_color.to_int()
 *         return 1
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":1893
 *         color[0] = py_color.to_int()
 *         return 1
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":1882
 * 
 * 
 * cdef hb_bool_t _paint_custom_palette_color_func(             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_t *funcs,
 *         void *paint_data,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz._paint_custom_palette_color_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_py_funcs);
  __Pyx_XDECREF(__pyx_v_py_color);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1913
 *     cdef object _custom_palette_color_func
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._hb_paintfuncs = hb_paint_funcs_create()
 * 
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs___cinit__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs___cinit__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self) {
  int __pyx_r;

  /* "uharfbuzz/_harfbuzz.pyx":1914
 * 
 *     def __cinit__(self):
 *         self._hb_paintfuncs = hb_paint_funcs_create()             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_v_self->_hb_paintfuncs = hb_paint_funcs_create();

  /* "uharfbuzz/_harfbuzz.pyx":1913
 *     cdef object _custom_palette_color_func
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._hb_paintfuncs = hb_paint_funcs_create()
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1916
 *         self._hb_paintfuncs = hb_paint_funcs_create()
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_destroy(self._hb_paintfuncs)
 * 
 */

/* Python wrapper */
static void __pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_2__dealloc__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_2__dealloc__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self) {

  /* "uharfbuzz/_harfbuzz.pyx":1917
 * 
 *     def __dealloc__(self):
 *         hb_paint_funcs_destroy(self._hb_paintfuncs)             # <<<<<<<<<<<<<<
 * 
 *     def set_push_transform_func(self,
 */
  hb_paint_funcs_destroy(__pyx_v_self->_hb_paintfuncs);

  /* "uharfbuzz/_harfbuzz.pyx":1916
 *         self._hb_paintfuncs = hb_paint_funcs_create()
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_destroy(self._hb_paintfuncs)
 * 
 */

  /* function exit code */
}

/* "uharfbuzz/_harfbuzz.pyx":1919
 *         hb_paint_funcs_destroy(self._hb_paintfuncs)
 * 
 *     def set_push_transform_func(self,             # <<<<<<<<<<<<<<
 *                                 func: Callable[[
 *                                     float,  # xx
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_5set_push_transform_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_10PaintFuncs_5set_push_transform_func = {"set_push_transform_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_5set_push_transform_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_5set_push_transform_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_push_transform_func (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_func)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1919, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_push_transform_func") < 0)) __PYX_ERR(0, 1919, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_func = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_push_transform_func", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1919, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.PaintFuncs.set_push_transform_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_4set_push_transform_func(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)__pyx_v_self), __pyx_v_func);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_4set_push_transform_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self, PyObject *__pyx_v_func) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_push_transform_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1929
 *                                     object,  # paint_data
 *                                 ], None]) -> None:
 *         self._push_transform_func = func             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_set_push_transform_func(
 *             self._hb_paintfuncs, _paint_push_transform_func, <void*>self, NULL)
 */
  __Pyx_INCREF(__pyx_v_func);
  __Pyx_GIVEREF(__pyx_v_func);
  __Pyx_GOTREF(__pyx_v_self->_push_transform_func);
  __Pyx_DECREF(__pyx_v_self->_push_transform_func);
  __pyx_v_self->_push_transform_func = __pyx_v_func;

  /* "uharfbuzz/_harfbuzz.pyx":1930
 *                                 ], None]) -> None:
 *         self._push_transform_func = func
 *         hb_paint_funcs_set_push_transform_func(             # <<<<<<<<<<<<<<
 *             self._hb_paintfuncs, _paint_push_transform_func, <void*>self, NULL)
 * 
 */
  hb_paint_funcs_set_push_transform_func(__pyx_v_self->_hb_paintfuncs, __pyx_f_9uharfbuzz_9_harfbuzz__paint_push_transform_func, ((void *)__pyx_v_self), NULL);

  /* "uharfbuzz/_harfbuzz.pyx":1919
 *         hb_paint_funcs_destroy(self._hb_paintfuncs)
 * 
 *     def set_push_transform_func(self,             # <<<<<<<<<<<<<<
 *                                 func: Callable[[
 *                                     float,  # xx
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1933
 *             self._hb_paintfuncs, _paint_push_transform_func, <void*>self, NULL)
 * 
 *     def set_pop_transform_func(self,             # <<<<<<<<<<<<<<
 *                                func: Callable[[
 *                                    object,  # paint_data
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_7set_pop_transform_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_10PaintFuncs_7set_pop_transform_func = {"set_pop_transform_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_7set_pop_transform_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_7set_pop_transform_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_pop_transform_func (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_func)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1933, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_pop_transform_func") < 0)) __PYX_ERR(0, 1933, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_func = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_pop_transform_func", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1933, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.PaintFuncs.set_pop_transform_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_6set_pop_transform_func(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)__pyx_v_self), __pyx_v_func);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_6set_pop_transform_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self, PyObject *__pyx_v_func) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_pop_transform_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1937
 *                                    object,  # paint_data
 *                                ], None]) -> None:
 *         self._pop_transform_func = func             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_set_pop_transform_func(
 *             self._hb_paintfuncs, _paint_pop_transform_func, <void*>self, NULL)
 */
  __Pyx_INCREF(__pyx_v_func);
  __Pyx_GIVEREF(__pyx_v_func);
  __Pyx_GOTREF(__pyx_v_self->_pop_transform_func);
  __Pyx_DECREF(__pyx_v_self->_pop_transform_func);
  __pyx_v_self->_pop_transform_func = __pyx_v_func;

  /* "uharfbuzz/_harfbuzz.pyx":1938
 *                                ], None]) -> None:
 *         self._pop_transform_func = func
 *         hb_paint_funcs_set_pop_transform_func(             # <<<<<<<<<<<<<<
 *             self._hb_paintfuncs, _paint_pop_transform_func, <void*>self, NULL)
 * 
 */
  hb_paint_funcs_set_pop_transform_func(__pyx_v_self->_hb_paintfuncs, __pyx_f_9uharfbuzz_9_harfbuzz__paint_pop_transform_func, ((void *)__pyx_v_self), NULL);

  /* "uharfbuzz/_harfbuzz.pyx":1933
 *             self._hb_paintfuncs, _paint_push_transform_func, <void*>self, NULL)
 * 
 *     def set_pop_transform_func(self,             # <<<<<<<<<<<<<<
 *                                func: Callable[[
 *                                    object,  # paint_data
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1941
 *             self._hb_paintfuncs, _paint_pop_transform_func, <void*>self, NULL)
 * 
 *     def set_color_glyph_func(self,             # <<<<<<<<<<<<<<
 *                              func: Callable[[
 *                                  int,  # gid
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_9set_color_glyph_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_10PaintFuncs_9set_color_glyph_func = {"set_color_glyph_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_9set_color_glyph_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_9set_color_glyph_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_color_glyph_func (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_func)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1941, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_color_glyph_func") < 0)) __PYX_ERR(0, 1941, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_func = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_color_glyph_func", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1941, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.PaintFuncs.set_color_glyph_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_8set_color_glyph_func(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)__pyx_v_self), __pyx_v_func);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_8set_color_glyph_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self, PyObject *__pyx_v_func) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_color_glyph_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1946
 *                                  object,  # paint_data
 *                              ], bool]) -> None:
 *         self._color_glyph_func = func             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_set_color_glyph_func(
 *             self._hb_paintfuncs, _paint_color_glyph_func, <void*>self, NULL)
 */
  __Pyx_INCREF(__pyx_v_func);
  __Pyx_GIVEREF(__pyx_v_func);
  __Pyx_GOTREF(__pyx_v_self->_color_glyph_func);
  __Pyx_DECREF(__pyx_v_self->_color_glyph_func);
  __pyx_v_self->_color_glyph_func = __pyx_v_func;

  /* "uharfbuzz/_harfbuzz.pyx":1947
 *                              ], bool]) -> None:
 *         self._color_glyph_func = func
 *         hb_paint_funcs_set_color_glyph_func(             # <<<<<<<<<<<<<<
 *             self._hb_paintfuncs, _paint_color_glyph_func, <void*>self, NULL)
 * 
 */
  hb_paint_funcs_set_color_glyph_func(__pyx_v_self->_hb_paintfuncs, __pyx_f_9uharfbuzz_9_harfbuzz__paint_color_glyph_func, ((void *)__pyx_v_self), NULL);

  /* "uharfbuzz/_harfbuzz.pyx":1941
 *             self._hb_paintfuncs, _paint_pop_transform_func, <void*>self, NULL)
 * 
 *     def set_color_glyph_func(self,             # <<<<<<<<<<<<<<
 *                              func: Callable[[
 *                                  int,  # gid
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1950
 *             self._hb_paintfuncs, _paint_color_glyph_func, <void*>self, NULL)
 * 
 *     def set_push_clip_glyph_func(self,             # <<<<<<<<<<<<<<
 *                                  func: Callable[[
 *                                      int,  # gid
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_11set_push_clip_glyph_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_10PaintFuncs_11set_push_clip_glyph_func = {"set_push_clip_glyph_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_11set_push_clip_glyph_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_11set_push_clip_glyph_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_push_clip_glyph_func (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_func)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1950, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_push_clip_glyph_func") < 0)) __PYX_ERR(0, 1950, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_func = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_push_clip_glyph_func", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1950, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.PaintFuncs.set_push_clip_glyph_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_10set_push_clip_glyph_func(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)__pyx_v_self), __pyx_v_func);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_10set_push_clip_glyph_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self, PyObject *__pyx_v_func) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_push_clip_glyph_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1955
 *                                      object,  # paint_data
 *                                  ], None]) -> None:
 *         self._push_clip_glyph_func = func             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_set_push_clip_glyph_func(
 *             self._hb_paintfuncs, _paint_push_clip_glyph_func, <void*>self, NULL)
 */
  __Pyx_INCREF(__pyx_v_func);
  __Pyx_GIVEREF(__pyx_v_func);
  __Pyx_GOTREF(__pyx_v_self->_push_clip_glyph_func);
  __Pyx_DECREF(__pyx_v_self->_push_clip_glyph_func);
  __pyx_v_self->_push_clip_glyph_func = __pyx_v_func;

  /* "uharfbuzz/_harfbuzz.pyx":1956
 *                                  ], None]) -> None:
 *         self._push_clip_glyph_func = func
 *         hb_paint_funcs_set_push_clip_glyph_func(             # <<<<<<<<<<<<<<
 *             self._hb_paintfuncs, _paint_push_clip_glyph_func, <void*>self, NULL)
 * 
 */
  hb_paint_funcs_set_push_clip_glyph_func(__pyx_v_self->_hb_paintfuncs, __pyx_f_9uharfbuzz_9_harfbuzz__paint_push_clip_glyph_func, ((void *)__pyx_v_self), NULL);

  /* "uharfbuzz/_harfbuzz.pyx":1950
 *             self._hb_paintfuncs, _paint_color_glyph_func, <void*>self, NULL)
 * 
 *     def set_push_clip_glyph_func(self,             # <<<<<<<<<<<<<<
 *                                  func: Callable[[
 *                                      int,  # gid
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1959
 *             self._hb_paintfuncs, _paint_push_clip_glyph_func, <void*>self, NULL)
 * 
 *     def set_push_clip_rectangle_func(self,             # <<<<<<<<<<<<<<
 *                                      func: Callable[[
 *                                          float,  # xmin
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_13set_push_clip_rectangle_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_10PaintFuncs_13set_push_clip_rectangle_func = {"set_push_clip_rectangle_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_13set_push_clip_rectangle_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_13set_push_clip_rectangle_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_push_clip_rectangle_func (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_func)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1959, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_push_clip_rectangle_func") < 0)) __PYX_ERR(0, 1959, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_func = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_push_clip_rectangle_func", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1959, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.PaintFuncs.set_push_clip_rectangle_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_12set_push_clip_rectangle_func(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)__pyx_v_self), __pyx_v_func);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_12set_push_clip_rectangle_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self, PyObject *__pyx_v_func) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_push_clip_rectangle_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1967
 *                                          object,  # paint_data
 *                                      ], None]) -> None:
 *         self._push_clip_rectangle_func = func             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_set_push_clip_rectangle_func(
 *             self._hb_paintfuncs, _paint_push_clip_rectangle_func, <void*>self, NULL)
 */
  __Pyx_INCREF(__pyx_v_func);
  __Pyx_GIVEREF(__pyx_v_func);
  __Pyx_GOTREF(__pyx_v_self->_push_clip_rectangle_func);
  __Pyx_DECREF(__pyx_v_self->_push_clip_rectangle_func);
  __pyx_v_self->_push_clip_rectangle_func = __pyx_v_func;

  /* "uharfbuzz/_harfbuzz.pyx":1968
 *                                      ], None]) -> None:
 *         self._push_clip_rectangle_func = func
 *         hb_paint_funcs_set_push_clip_rectangle_func(             # <<<<<<<<<<<<<<
 *             self._hb_paintfuncs, _paint_push_clip_rectangle_func, <void*>self, NULL)
 * 
 */
  hb_paint_funcs_set_push_clip_rectangle_func(__pyx_v_self->_hb_paintfuncs, __pyx_f_9uharfbuzz_9_harfbuzz__paint_push_clip_rectangle_func, ((void *)__pyx_v_self), NULL);

  /* "uharfbuzz/_harfbuzz.pyx":1959
 *             self._hb_paintfuncs, _paint_push_clip_glyph_func, <void*>self, NULL)
 * 
 *     def set_push_clip_rectangle_func(self,             # <<<<<<<<<<<<<<
 *                                      func: Callable[[
 *                                          float,  # xmin
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1971
 *             self._hb_paintfuncs, _paint_push_clip_rectangle_func, <void*>self, NULL)
 * 
 *     def set_pop_clip_func(self,             # <<<<<<<<<<<<<<
 *                           func: Callable[[
 *                               object,  # paint_data
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_15set_pop_clip_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_10PaintFuncs_15set_pop_clip_func = {"set_pop_clip_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_15set_pop_clip_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_15set_pop_clip_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_pop_clip_func (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_func)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1971, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_pop_clip_func") < 0)) __PYX_ERR(0, 1971, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_func = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_pop_clip_func", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1971, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.PaintFuncs.set_pop_clip_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_14set_pop_clip_func(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)__pyx_v_self), __pyx_v_func);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_14set_pop_clip_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self, PyObject *__pyx_v_func) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_pop_clip_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1975
 *                               object,  # paint_data
 *                           ], None]) -> None:
 *         self._pop_clip_func = func             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_set_pop_clip_func(
 *             self._hb_paintfuncs, _paint_pop_clip_func, <void*>self, NULL)
 */
  __Pyx_INCREF(__pyx_v_func);
  __Pyx_GIVEREF(__pyx_v_func);
  __Pyx_GOTREF(__pyx_v_self->_pop_clip_func);
  __Pyx_DECREF(__pyx_v_self->_pop_clip_func);
  __pyx_v_self->_pop_clip_func = __pyx_v_func;

  /* "uharfbuzz/_harfbuzz.pyx":1976
 *                           ], None]) -> None:
 *         self._pop_clip_func = func
 *         hb_paint_funcs_set_pop_clip_func(             # <<<<<<<<<<<<<<
 *             self._hb_paintfuncs, _paint_pop_clip_func, <void*>self, NULL)
 * 
 */
  hb_paint_funcs_set_pop_clip_func(__pyx_v_self->_hb_paintfuncs, __pyx_f_9uharfbuzz_9_harfbuzz__paint_pop_clip_func, ((void *)__pyx_v_self), NULL);

  /* "uharfbuzz/_harfbuzz.pyx":1971
 *             self._hb_paintfuncs, _paint_push_clip_rectangle_func, <void*>self, NULL)
 * 
 *     def set_pop_clip_func(self,             # <<<<<<<<<<<<<<
 *                           func: Callable[[
 *                               object,  # paint_data
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1979
 *             self._hb_paintfuncs, _paint_pop_clip_func, <void*>self, NULL)
 * 
 *     def set_color_func(self,             # <<<<<<<<<<<<<<
 *                        func: Callable[[
 *                            Color,  # color
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_17set_color_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_10PaintFuncs_17set_color_func = {"set_color_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_17set_color_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_17set_color_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_color_func (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_func)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1979, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_color_func") < 0)) __PYX_ERR(0, 1979, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_func = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_color_func", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1979, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.PaintFuncs.set_color_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_16set_color_func(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)__pyx_v_self), __pyx_v_func);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_16set_color_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self, PyObject *__pyx_v_func) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_color_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1985
 *                            object,  # paint_data
 *                        ], None]) -> None:
 *         self._color_func = func             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_set_color_func(
 *             self._hb_paintfuncs, _paint_color_func, <void*>self, NULL)
 */
  __Pyx_INCREF(__pyx_v_func);
  __Pyx_GIVEREF(__pyx_v_func);
  __Pyx_GOTREF(__pyx_v_self->_color_func);
  __Pyx_DECREF(__pyx_v_self->_color_func);
  __pyx_v_self->_color_func = __pyx_v_func;

  /* "uharfbuzz/_harfbuzz.pyx":1986
 *                        ], None]) -> None:
 *         self._color_func = func
 *         hb_paint_funcs_set_color_func(             # <<<<<<<<<<<<<<
 *             self._hb_paintfuncs, _paint_color_func, <void*>self, NULL)
 * 
 */
  hb_paint_funcs_set_color_func(__pyx_v_self->_hb_paintfuncs, __pyx_f_9uharfbuzz_9_harfbuzz__paint_color_func, ((void *)__pyx_v_self), NULL);

  /* "uharfbuzz/_harfbuzz.pyx":1979
 *             self._hb_paintfuncs, _paint_pop_clip_func, <void*>self, NULL)
 * 
 *     def set_color_func(self,             # <<<<<<<<<<<<<<
 *                        func: Callable[[
 *                            Color,  # color
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":1989
 *             self._hb_paintfuncs, _paint_color_func, <void*>self, NULL)
 * 
 *     def set_image_func(self,             # <<<<<<<<<<<<<<
 *                        func: Callable[[
 *                            Blob,  # image
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_19set_image_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_10PaintFuncs_19set_image_func = {"set_image_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_19set_image_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_19set_image_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_image_func (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_func)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1989, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_image_func") < 0)) __PYX_ERR(0, 1989, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_func = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_image_func", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1989, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.PaintFuncs.set_image_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_18set_image_func(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)__pyx_v_self), __pyx_v_func);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_18set_image_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self, PyObject *__pyx_v_func) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_image_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":1999
 *                            object,  # paint_data
 *                        ], bool]) -> None:
 *         self._image_func = func             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_set_image_func(
 *             self._hb_paintfuncs, _paint_image_func, <void*>self, NULL)
 */
  __Pyx_INCREF(__pyx_v_func);
  __Pyx_GIVEREF(__pyx_v_func);
  __Pyx_GOTREF(__pyx_v_self->_image_func);
  __Pyx_DECREF(__pyx_v_self->_image_func);
  __pyx_v_self->_image_func = __pyx_v_func;

  /* "uharfbuzz/_harfbuzz.pyx":2000
 *                        ], bool]) -> None:
 *         self._image_func = func
 *         hb_paint_funcs_set_image_func(             # <<<<<<<<<<<<<<
 *             self._hb_paintfuncs, _paint_image_func, <void*>self, NULL)
 * 
 */
  hb_paint_funcs_set_image_func(__pyx_v_self->_hb_paintfuncs, __pyx_f_9uharfbuzz_9_harfbuzz__paint_image_func, ((void *)__pyx_v_self), NULL);

  /* "uharfbuzz/_harfbuzz.pyx":1989
 *             self._hb_paintfuncs, _paint_color_func, <void*>self, NULL)
 * 
 *     def set_image_func(self,             # <<<<<<<<<<<<<<
 *                        func: Callable[[
 *                            Blob,  # image
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2003
 *             self._hb_paintfuncs, _paint_image_func, <void*>self, NULL)
 * 
 *     def set_linear_gradient_func(self,             # <<<<<<<<<<<<<<
 *                                  func: Callable[[
 *                                     ColorLine,  # color_line
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_21set_linear_gradient_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_10PaintFuncs_21set_linear_gradient_func = {"set_linear_gradient_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_21set_linear_gradient_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_21set_linear_gradient_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_linear_gradient_func (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_func)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2003, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_linear_gradient_func") < 0)) __PYX_ERR(0, 2003, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_func = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_linear_gradient_func", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2003, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.PaintFuncs.set_linear_gradient_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_20set_linear_gradient_func(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)__pyx_v_self), __pyx_v_func);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_20set_linear_gradient_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self, PyObject *__pyx_v_func) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_linear_gradient_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2014
 *                                     object,  # paint_data
 *                                  ], None]) -> None:
 *         self._linear_gradient_func = func             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_set_linear_gradient_func(
 *             self._hb_paintfuncs, _paint_linear_gradient_func, <void*>self, NULL)
 */
  __Pyx_INCREF(__pyx_v_func);
  __Pyx_GIVEREF(__pyx_v_func);
  __Pyx_GOTREF(__pyx_v_self->_linear_gradient_func);
  __Pyx_DECREF(__pyx_v_self->_linear_gradient_func);
  __pyx_v_self->_linear_gradient_func = __pyx_v_func;

  /* "uharfbuzz/_harfbuzz.pyx":2015
 *                                  ], None]) -> None:
 *         self._linear_gradient_func = func
 *         hb_paint_funcs_set_linear_gradient_func(             # <<<<<<<<<<<<<<
 *             self._hb_paintfuncs, _paint_linear_gradient_func, <void*>self, NULL)
 * 
 */
  hb_paint_funcs_set_linear_gradient_func(__pyx_v_self->_hb_paintfuncs, __pyx_f_9uharfbuzz_9_harfbuzz__paint_linear_gradient_func, ((void *)__pyx_v_self), NULL);

  /* "uharfbuzz/_harfbuzz.pyx":2003
 *             self._hb_paintfuncs, _paint_image_func, <void*>self, NULL)
 * 
 *     def set_linear_gradient_func(self,             # <<<<<<<<<<<<<<
 *                                  func: Callable[[
 *                                     ColorLine,  # color_line
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2018
 *             self._hb_paintfuncs, _paint_linear_gradient_func, <void*>self, NULL)
 * 
 *     def set_radial_gradient_func(self,             # <<<<<<<<<<<<<<
 *                                  func: Callable[[
 *                                     ColorLine,  # color_line
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_23set_radial_gradient_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_10PaintFuncs_23set_radial_gradient_func = {"set_radial_gradient_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_23set_radial_gradient_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_23set_radial_gradient_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_radial_gradient_func (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_func)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2018, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_radial_gradient_func") < 0)) __PYX_ERR(0, 2018, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_func = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_radial_gradient_func", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2018, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.PaintFuncs.set_radial_gradient_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_22set_radial_gradient_func(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)__pyx_v_self), __pyx_v_func);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_22set_radial_gradient_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self, PyObject *__pyx_v_func) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_radial_gradient_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2029
 *                                     object,  # paint_data
 *                                  ], None]) -> None:
 *         self._radial_gradient_func = func             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_set_radial_gradient_func(
 *             self._hb_paintfuncs, _paint_radial_gradient_func, <void*>self, NULL)
 */
  __Pyx_INCREF(__pyx_v_func);
  __Pyx_GIVEREF(__pyx_v_func);
  __Pyx_GOTREF(__pyx_v_self->_radial_gradient_func);
  __Pyx_DECREF(__pyx_v_self->_radial_gradient_func);
  __pyx_v_self->_radial_gradient_func = __pyx_v_func;

  /* "uharfbuzz/_harfbuzz.pyx":2030
 *                                  ], None]) -> None:
 *         self._radial_gradient_func = func
 *         hb_paint_funcs_set_radial_gradient_func(             # <<<<<<<<<<<<<<
 *             self._hb_paintfuncs, _paint_radial_gradient_func, <void*>self, NULL)
 * 
 */
  hb_paint_funcs_set_radial_gradient_func(__pyx_v_self->_hb_paintfuncs, __pyx_f_9uharfbuzz_9_harfbuzz__paint_radial_gradient_func, ((void *)__pyx_v_self), NULL);

  /* "uharfbuzz/_harfbuzz.pyx":2018
 *             self._hb_paintfuncs, _paint_linear_gradient_func, <void*>self, NULL)
 * 
 *     def set_radial_gradient_func(self,             # <<<<<<<<<<<<<<
 *                                  func: Callable[[
 *                                     ColorLine,  # color_line
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2033
 *             self._hb_paintfuncs, _paint_radial_gradient_func, <void*>self, NULL)
 * 
 *     def set_sweep_gradient_func(self,             # <<<<<<<<<<<<<<
 *                                 func: Callable[[
 *                                     ColorLine,  # color_line
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_25set_sweep_gradient_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_10PaintFuncs_25set_sweep_gradient_func = {"set_sweep_gradient_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_25set_sweep_gradient_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_25set_sweep_gradient_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_sweep_gradient_func (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_func)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2033, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_sweep_gradient_func") < 0)) __PYX_ERR(0, 2033, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_func = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_sweep_gradient_func", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2033, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.PaintFuncs.set_sweep_gradient_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_24set_sweep_gradient_func(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)__pyx_v_self), __pyx_v_func);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_24set_sweep_gradient_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self, PyObject *__pyx_v_func) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_sweep_gradient_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2042
 *                                     object,  # paint_data
 *                                 ], None]) -> None:
 *         self._sweep_gradient_func = func             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_set_sweep_gradient_func(
 *             self._hb_paintfuncs, _paint_sweep_gradient_func, <void*>self, NULL)
 */
  __Pyx_INCREF(__pyx_v_func);
  __Pyx_GIVEREF(__pyx_v_func);
  __Pyx_GOTREF(__pyx_v_self->_sweep_gradient_func);
  __Pyx_DECREF(__pyx_v_self->_sweep_gradient_func);
  __pyx_v_self->_sweep_gradient_func = __pyx_v_func;

  /* "uharfbuzz/_harfbuzz.pyx":2043
 *                                 ], None]) -> None:
 *         self._sweep_gradient_func = func
 *         hb_paint_funcs_set_sweep_gradient_func(             # <<<<<<<<<<<<<<
 *             self._hb_paintfuncs, _paint_sweep_gradient_func, <void*>self, NULL)
 * 
 */
  hb_paint_funcs_set_sweep_gradient_func(__pyx_v_self->_hb_paintfuncs, __pyx_f_9uharfbuzz_9_harfbuzz__paint_sweep_gradient_func, ((void *)__pyx_v_self), NULL);

  /* "uharfbuzz/_harfbuzz.pyx":2033
 *             self._hb_paintfuncs, _paint_radial_gradient_func, <void*>self, NULL)
 * 
 *     def set_sweep_gradient_func(self,             # <<<<<<<<<<<<<<
 *                                 func: Callable[[
 *                                     ColorLine,  # color_line
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2046
 *             self._hb_paintfuncs, _paint_sweep_gradient_func, <void*>self, NULL)
 * 
 *     def set_push_group_func(self,             # <<<<<<<<<<<<<<
 *                             func: Callable[[
 *                                 object,  # paint_data
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_27set_push_group_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_10PaintFuncs_27set_push_group_func = {"set_push_group_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_27set_push_group_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_27set_push_group_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_push_group_func (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_func)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2046, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_push_group_func") < 0)) __PYX_ERR(0, 2046, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_func = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_push_group_func", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2046, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.PaintFuncs.set_push_group_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_26set_push_group_func(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)__pyx_v_self), __pyx_v_func);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_26set_push_group_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self, PyObject *__pyx_v_func) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_push_group_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2050
 *                                 object,  # paint_data
 *                             ], None]) -> None:
 *         self._push_group_func = func             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_set_push_group_func(
 *             self._hb_paintfuncs, _paint_push_group_func, <void*>self, NULL)
 */
  __Pyx_INCREF(__pyx_v_func);
  __Pyx_GIVEREF(__pyx_v_func);
  __Pyx_GOTREF(__pyx_v_self->_push_group_func);
  __Pyx_DECREF(__pyx_v_self->_push_group_func);
  __pyx_v_self->_push_group_func = __pyx_v_func;

  /* "uharfbuzz/_harfbuzz.pyx":2051
 *                             ], None]) -> None:
 *         self._push_group_func = func
 *         hb_paint_funcs_set_push_group_func(             # <<<<<<<<<<<<<<
 *             self._hb_paintfuncs, _paint_push_group_func, <void*>self, NULL)
 * 
 */
  hb_paint_funcs_set_push_group_func(__pyx_v_self->_hb_paintfuncs, __pyx_f_9uharfbuzz_9_harfbuzz__paint_push_group_func, ((void *)__pyx_v_self), NULL);

  /* "uharfbuzz/_harfbuzz.pyx":2046
 *             self._hb_paintfuncs, _paint_sweep_gradient_func, <void*>self, NULL)
 * 
 *     def set_push_group_func(self,             # <<<<<<<<<<<<<<
 *                             func: Callable[[
 *                                 object,  # paint_data
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2054
 *             self._hb_paintfuncs, _paint_push_group_func, <void*>self, NULL)
 * 
 *     def set_pop_group_func(self,             # <<<<<<<<<<<<<<
 *                            func: Callable[[
 *                                PaintCompositeMode,  # mode
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_29set_pop_group_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_10PaintFuncs_29set_pop_group_func = {"set_pop_group_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_29set_pop_group_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_29set_pop_group_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_pop_group_func (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_func)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2054, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_pop_group_func") < 0)) __PYX_ERR(0, 2054, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_func = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_pop_group_func", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2054, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.PaintFuncs.set_pop_group_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_28set_pop_group_func(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)__pyx_v_self), __pyx_v_func);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_28set_pop_group_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self, PyObject *__pyx_v_func) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_pop_group_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2059
 *                                object,  # paint_data
 *                            ], None]) -> None:
 *         self._pop_group_func = func             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_set_pop_group_func(
 *             self._hb_paintfuncs, _paint_pop_group_func, <void*>self, NULL)
 */
  __Pyx_INCREF(__pyx_v_func);
  __Pyx_GIVEREF(__pyx_v_func);
  __Pyx_GOTREF(__pyx_v_self->_pop_group_func);
  __Pyx_DECREF(__pyx_v_self->_pop_group_func);
  __pyx_v_self->_pop_group_func = __pyx_v_func;

  /* "uharfbuzz/_harfbuzz.pyx":2060
 *                            ], None]) -> None:
 *         self._pop_group_func = func
 *         hb_paint_funcs_set_pop_group_func(             # <<<<<<<<<<<<<<
 *             self._hb_paintfuncs, _paint_pop_group_func, <void*>self, NULL)
 * 
 */
  hb_paint_funcs_set_pop_group_func(__pyx_v_self->_hb_paintfuncs, __pyx_f_9uharfbuzz_9_harfbuzz__paint_pop_group_func, ((void *)__pyx_v_self), NULL);

  /* "uharfbuzz/_harfbuzz.pyx":2054
 *             self._hb_paintfuncs, _paint_push_group_func, <void*>self, NULL)
 * 
 *     def set_pop_group_func(self,             # <<<<<<<<<<<<<<
 *                            func: Callable[[
 *                                PaintCompositeMode,  # mode
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2063
 *             self._hb_paintfuncs, _paint_pop_group_func, <void*>self, NULL)
 * 
 *     def set_custom_palette_color_func(self,             # <<<<<<<<<<<<<<
 *                                       func: Callable[[
 *                                           int,  # color_index
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_31set_custom_palette_color_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_10PaintFuncs_31set_custom_palette_color_func = {"set_custom_palette_color_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_31set_custom_palette_color_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_31set_custom_palette_color_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_custom_palette_color_func (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_func)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2063, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_custom_palette_color_func") < 0)) __PYX_ERR(0, 2063, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_func = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_custom_palette_color_func", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2063, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.PaintFuncs.set_custom_palette_color_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_30set_custom_palette_color_func(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)__pyx_v_self), __pyx_v_func);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_30set_custom_palette_color_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self, PyObject *__pyx_v_func) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_custom_palette_color_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2068
 *                                           object,  # paint_data
 *                                       ], Color]) -> None:
 *         self._custom_palette_color_func = func             # <<<<<<<<<<<<<<
 *         hb_paint_funcs_set_custom_palette_color_func(
 *             self._hb_paintfuncs, _paint_custom_palette_color_func, <void*>self, NULL)
 */
  __Pyx_INCREF(__pyx_v_func);
  __Pyx_GIVEREF(__pyx_v_func);
  __Pyx_GOTREF(__pyx_v_self->_custom_palette_color_func);
  __Pyx_DECREF(__pyx_v_self->_custom_palette_color_func);
  __pyx_v_self->_custom_palette_color_func = __pyx_v_func;

  /* "uharfbuzz/_harfbuzz.pyx":2069
 *                                       ], Color]) -> None:
 *         self._custom_palette_color_func = func
 *         hb_paint_funcs_set_custom_palette_color_func(             # <<<<<<<<<<<<<<
 *             self._hb_paintfuncs, _paint_custom_palette_color_func, <void*>self, NULL)
 * 
 */
  hb_paint_funcs_set_custom_palette_color_func(__pyx_v_self->_hb_paintfuncs, __pyx_f_9uharfbuzz_9_harfbuzz__paint_custom_palette_color_func, ((void *)__pyx_v_self), NULL);

  /* "uharfbuzz/_harfbuzz.pyx":2063
 *             self._hb_paintfuncs, _paint_pop_group_func, <void*>self, NULL)
 * 
 *     def set_custom_palette_color_func(self,             # <<<<<<<<<<<<<<
 *                                       func: Callable[[
 *                                           int,  # color_index
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_33__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_10PaintFuncs_33__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_33__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_33__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_32__reduce_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_32__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.PaintFuncs.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_35__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_10PaintFuncs_35__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_35__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_35__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.PaintFuncs.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_34__setstate_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10PaintFuncs_34__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.PaintFuncs.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2073
 * 
 * 
 * cdef void _move_to_func(hb_draw_funcs_t *dfuncs,             # <<<<<<<<<<<<<<
 *                         void *draw_data,
 *                         hb_draw_state_t *st,
 */

static void __pyx_f_9uharfbuzz_9_harfbuzz__move_to_func(CYTHON_UNUSED hb_draw_funcs_t *__pyx_v_dfuncs, void *__pyx_v_draw_data, CYTHON_UNUSED hb_draw_state_t *__pyx_v_st, float __pyx_v_to_x, float __pyx_v_to_y, void *__pyx_v_user_data) {
  PyObject *__pyx_v_m = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_move_to_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2079
 *                         float to_y,
 *                         void *user_data) noexcept:
 *     m = <object>user_data             # <<<<<<<<<<<<<<
 *     m(to_x, to_y, <object>draw_data)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_user_data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_m = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2080
 *                         void *user_data) noexcept:
 *     m = <object>user_data
 *     m(to_x, to_y, <object>draw_data)             # <<<<<<<<<<<<<<
 * 
 * cdef void _line_to_func(hb_draw_funcs_t *dfuncs,
 */
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_to_x); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2080, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_to_y); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2080, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_m);
  __pyx_t_4 = __pyx_v_m; __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_5, __pyx_t_2, __pyx_t_3, ((PyObject *)__pyx_v_draw_data)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2080, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2073
 * 
 * 
 * cdef void _move_to_func(hb_draw_funcs_t *dfuncs,             # <<<<<<<<<<<<<<
 *                         void *draw_data,
 *                         hb_draw_state_t *st,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz._move_to_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_m);
  __Pyx_RefNannyFinishContext();
}

/* "uharfbuzz/_harfbuzz.pyx":2082
 *     m(to_x, to_y, <object>draw_data)
 * 
 * cdef void _line_to_func(hb_draw_funcs_t *dfuncs,             # <<<<<<<<<<<<<<
 *                         void *draw_data,
 *                         hb_draw_state_t *st,
 */

static void __pyx_f_9uharfbuzz_9_harfbuzz__line_to_func(CYTHON_UNUSED hb_draw_funcs_t *__pyx_v_dfuncs, void *__pyx_v_draw_data, CYTHON_UNUSED hb_draw_state_t *__pyx_v_st, float __pyx_v_to_x, float __pyx_v_to_y, void *__pyx_v_user_data) {
  PyObject *__pyx_v_l = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_line_to_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2088
 *                         float to_y,
 *                         void *user_data) noexcept:
 *     l = <object>user_data             # <<<<<<<<<<<<<<
 *     l(to_x, to_y, <object>draw_data)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_user_data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_l = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2089
 *                         void *user_data) noexcept:
 *     l = <object>user_data
 *     l(to_x, to_y, <object>draw_data)             # <<<<<<<<<<<<<<
 * 
 * cdef void _close_path_func(hb_draw_funcs_t *dfuncs,
 */
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_to_x); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2089, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_to_y); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2089, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_l);
  __pyx_t_4 = __pyx_v_l; __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_5, __pyx_t_2, __pyx_t_3, ((PyObject *)__pyx_v_draw_data)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 3+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2089, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2082
 *     m(to_x, to_y, <object>draw_data)
 * 
 * cdef void _line_to_func(hb_draw_funcs_t *dfuncs,             # <<<<<<<<<<<<<<
 *                         void *draw_data,
 *                         hb_draw_state_t *st,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz._line_to_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_l);
  __Pyx_RefNannyFinishContext();
}

/* "uharfbuzz/_harfbuzz.pyx":2091
 *     l(to_x, to_y, <object>draw_data)
 * 
 * cdef void _close_path_func(hb_draw_funcs_t *dfuncs,             # <<<<<<<<<<<<<<
 *                            void *draw_data,
 *                            hb_draw_state_t *st,
 */

static void __pyx_f_9uharfbuzz_9_harfbuzz__close_path_func(CYTHON_UNUSED hb_draw_funcs_t *__pyx_v_dfuncs, void *__pyx_v_draw_data, CYTHON_UNUSED hb_draw_state_t *__pyx_v_st, void *__pyx_v_user_data) {
  PyObject *__pyx_v_cl = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_close_path_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2095
 *                            hb_draw_state_t *st,
 *                            void *user_data) noexcept:
 *     cl = <object>user_data             # <<<<<<<<<<<<<<
 *     cl(<object>draw_data)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_user_data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_cl = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2096
 *                            void *user_data) noexcept:
 *     cl = <object>user_data
 *     cl(<object>draw_data)             # <<<<<<<<<<<<<<
 * 
 * cdef void _quadratic_to_func(hb_draw_funcs_t *dfuncs,
 */
  __Pyx_INCREF(__pyx_v_cl);
  __pyx_t_2 = __pyx_v_cl; __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, ((PyObject *)__pyx_v_draw_data)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2096, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2091
 *     l(to_x, to_y, <object>draw_data)
 * 
 * cdef void _close_path_func(hb_draw_funcs_t *dfuncs,             # <<<<<<<<<<<<<<
 *                            void *draw_data,
 *                            hb_draw_state_t *st,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz._close_path_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_cl);
  __Pyx_RefNannyFinishContext();
}

/* "uharfbuzz/_harfbuzz.pyx":2098
 *     cl(<object>draw_data)
 * 
 * cdef void _quadratic_to_func(hb_draw_funcs_t *dfuncs,             # <<<<<<<<<<<<<<
 *                              void *draw_data,
 *                              hb_draw_state_t *st,
 */

static void __pyx_f_9uharfbuzz_9_harfbuzz__quadratic_to_func(CYTHON_UNUSED hb_draw_funcs_t *__pyx_v_dfuncs, void *__pyx_v_draw_data, CYTHON_UNUSED hb_draw_state_t *__pyx_v_st, float __pyx_v_c1_x, float __pyx_v_c1_y, float __pyx_v_to_x, float __pyx_v_to_y, void *__pyx_v_user_data) {
  PyObject *__pyx_v_q = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_quadratic_to_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2106
 *                              float to_y,
 *                              void *user_data) noexcept:
 *     q = <object>user_data             # <<<<<<<<<<<<<<
 *     q(c1_x, c1_y, to_x, to_y, <object>draw_data)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_user_data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_q = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2107
 *                              void *user_data) noexcept:
 *     q = <object>user_data
 *     q(c1_x, c1_y, to_x, to_y, <object>draw_data)             # <<<<<<<<<<<<<<
 * 
 * cdef void _cubic_to_func(hb_draw_funcs_t *dfuncs,
 */
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_c1_x); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_c1_y); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_to_x); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_to_y); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_v_q);
  __pyx_t_6 = __pyx_v_q; __pyx_t_7 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_7)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_7);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[6] = {__pyx_t_7, __pyx_t_2, __pyx_t_3, __pyx_t_4, __pyx_t_5, ((PyObject *)__pyx_v_draw_data)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 5+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2107, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2098
 *     cl(<object>draw_data)
 * 
 * cdef void _quadratic_to_func(hb_draw_funcs_t *dfuncs,             # <<<<<<<<<<<<<<
 *                              void *draw_data,
 *                              hb_draw_state_t *st,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz._quadratic_to_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_q);
  __Pyx_RefNannyFinishContext();
}

/* "uharfbuzz/_harfbuzz.pyx":2109
 *     q(c1_x, c1_y, to_x, to_y, <object>draw_data)
 * 
 * cdef void _cubic_to_func(hb_draw_funcs_t *dfuncs,             # <<<<<<<<<<<<<<
 *                          void *draw_data,
 *                          hb_draw_state_t *st,
 */

static void __pyx_f_9uharfbuzz_9_harfbuzz__cubic_to_func(CYTHON_UNUSED hb_draw_funcs_t *__pyx_v_dfuncs, void *__pyx_v_draw_data, CYTHON_UNUSED hb_draw_state_t *__pyx_v_st, float __pyx_v_c1_x, float __pyx_v_c1_y, float __pyx_v_c2_x, float __pyx_v_c2_y, float __pyx_v_to_x, float __pyx_v_to_y, void *__pyx_v_user_data) {
  PyObject *__pyx_v_c = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  unsigned int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_cubic_to_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2119
 *                          float to_y,
 *                          void *user_data) noexcept:
 *     c = <object>user_data             # <<<<<<<<<<<<<<
 *     c(c1_x, c1_y, c2_x, c2_y, to_x, to_y, <object>draw_data)
 * 
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_user_data);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_c = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2120
 *                          void *user_data) noexcept:
 *     c = <object>user_data
 *     c(c1_x, c1_y, c2_x, c2_y, to_x, to_y, <object>draw_data)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_c1_x); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_c1_y); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_c2_x); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_c2_y); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyFloat_FromDouble(__pyx_v_to_x); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = PyFloat_FromDouble(__pyx_v_to_y); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_v_c);
  __pyx_t_8 = __pyx_v_c; __pyx_t_9 = NULL;
  __pyx_t_10 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
      __pyx_t_10 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[8] = {__pyx_t_9, __pyx_t_2, __pyx_t_3, __pyx_t_4, __pyx_t_5, __pyx_t_6, __pyx_t_7, ((PyObject *)__pyx_v_draw_data)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_10, 7+__pyx_t_10);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2120, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2109
 *     q(c1_x, c1_y, to_x, to_y, <object>draw_data)
 * 
 * cdef void _cubic_to_func(hb_draw_funcs_t *dfuncs,             # <<<<<<<<<<<<<<
 *                          void *draw_data,
 *                          hb_draw_state_t *st,
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_WriteUnraisable("uharfbuzz._harfbuzz._cubic_to_func", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_c);
  __Pyx_RefNannyFinishContext();
}

/* "uharfbuzz/_harfbuzz.pyx":2132
 *     cdef int _warned
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._hb_drawfuncs = hb_draw_funcs_create()
 *         self._warned = False
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_9DrawFuncs___cinit__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_9DrawFuncs___cinit__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *__pyx_v_self) {
  int __pyx_r;

  /* "uharfbuzz/_harfbuzz.pyx":2133
 * 
 *     def __cinit__(self):
 *         self._hb_drawfuncs = hb_draw_funcs_create()             # <<<<<<<<<<<<<<
 *         self._warned = False
 * 
 */
  __pyx_v_self->_hb_drawfuncs = hb_draw_funcs_create();

  /* "uharfbuzz/_harfbuzz.pyx":2134
 *     def __cinit__(self):
 *         self._hb_drawfuncs = hb_draw_funcs_create()
 *         self._warned = False             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_v_self->_warned = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2132
 *     cdef int _warned
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._hb_drawfuncs = hb_draw_funcs_create()
 *         self._warned = False
 */

  /* function exit code */
  __pyx_r = 0;
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2136
 *         self._warned = False
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         hb_draw_funcs_destroy(self._hb_drawfuncs)
 * 
 */

/* Python wrapper */
static void __pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_9uharfbuzz_9_harfbuzz_9DrawFuncs_2__dealloc__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_9uharfbuzz_9_harfbuzz_9DrawFuncs_2__dealloc__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *__pyx_v_self) {

  /* "uharfbuzz/_harfbuzz.pyx":2137
 * 
 *     def __dealloc__(self):
 *         hb_draw_funcs_destroy(self._hb_drawfuncs)             # <<<<<<<<<<<<<<
 * 
 *     def get_glyph_shape(self, font: Font, gid: int):
 */
  hb_draw_funcs_destroy(__pyx_v_self->_hb_drawfuncs);

  /* "uharfbuzz/_harfbuzz.pyx":2136
 *         self._warned = False
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         hb_draw_funcs_destroy(self._hb_drawfuncs)
 * 
 */

  /* function exit code */
}

/* "uharfbuzz/_harfbuzz.pyx":2139
 *         hb_draw_funcs_destroy(self._hb_drawfuncs)
 * 
 *     def get_glyph_shape(self, font: Font, gid: int):             # <<<<<<<<<<<<<<
 *         if not self._warned:
 *             warnings.warn(
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_5get_glyph_shape(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_9DrawFuncs_5get_glyph_shape = {"get_glyph_shape", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_5get_glyph_shape, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_5get_glyph_shape(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font = 0;
  PyObject *__pyx_v_gid = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_glyph_shape (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_font,&__pyx_n_s_gid,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_font)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2139, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_gid)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2139, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("get_glyph_shape", 1, 2, 2, 1); __PYX_ERR(0, 2139, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_glyph_shape") < 0)) __PYX_ERR(0, 2139, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_font = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)values[0]);
    __pyx_v_gid = ((PyObject*)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_glyph_shape", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 2139, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.DrawFuncs.get_glyph_shape", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_font), __pyx_ptype_9uharfbuzz_9_harfbuzz_Font, 0, "font", 0))) __PYX_ERR(0, 2139, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_gid), (&PyInt_Type), 0, "gid", 1))) __PYX_ERR(0, 2139, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_9DrawFuncs_4get_glyph_shape(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *)__pyx_v_self), __pyx_v_font, __pyx_v_gid);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9DrawFuncs_4get_glyph_shape(struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font, PyObject *__pyx_v_gid) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_glyph_shape", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2140
 * 
 *     def get_glyph_shape(self, font: Font, gid: int):
 *         if not self._warned:             # <<<<<<<<<<<<<<
 *             warnings.warn(
 *                 "get_glyph_shape() is deprecated, use Font.draw_glyph() instead",
 */
  __pyx_t_1 = (!(__pyx_v_self->_warned != 0));
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":2141
 *     def get_glyph_shape(self, font: Font, gid: int):
 *         if not self._warned:
 *             warnings.warn(             # <<<<<<<<<<<<<<
 *                 "get_glyph_shape() is deprecated, use Font.draw_glyph() instead",
 *                 DeprecationWarning,
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_warnings); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2141, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_warn); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2141, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uharfbuzz/_harfbuzz.pyx":2143
 *             warnings.warn(
 *                 "get_glyph_shape() is deprecated, use Font.draw_glyph() instead",
 *                 DeprecationWarning,             # <<<<<<<<<<<<<<
 *             )
 *             self._warned = True
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2141, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uharfbuzz/_harfbuzz.pyx":2145
 *                 DeprecationWarning,
 *             )
 *             self._warned = True             # <<<<<<<<<<<<<<
 *         font.draw_glyph(gid, self)
 * 
 */
    __pyx_v_self->_warned = 1;

    /* "uharfbuzz/_harfbuzz.pyx":2140
 * 
 *     def get_glyph_shape(self, font: Font, gid: int):
 *         if not self._warned:             # <<<<<<<<<<<<<<
 *             warnings.warn(
 *                 "get_glyph_shape() is deprecated, use Font.draw_glyph() instead",
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2146
 *             )
 *             self._warned = True
 *         font.draw_glyph(gid, self)             # <<<<<<<<<<<<<<
 * 
 *     def draw_glyph(self, font: Font, gid: int, draw_data: object = None):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_font), __pyx_n_s_draw_glyph); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_v_gid, ((PyObject *)__pyx_v_self)};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2146, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2139
 *         hb_draw_funcs_destroy(self._hb_drawfuncs)
 * 
 *     def get_glyph_shape(self, font: Font, gid: int):             # <<<<<<<<<<<<<<
 *         if not self._warned:
 *             warnings.warn(
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.DrawFuncs.get_glyph_shape", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2148
 *         font.draw_glyph(gid, self)
 * 
 *     def draw_glyph(self, font: Font, gid: int, draw_data: object = None):             # <<<<<<<<<<<<<<
 *         if not self._warned:
 *             warnings.warn(
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_7draw_glyph(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_9DrawFuncs_7draw_glyph = {"draw_glyph", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_7draw_glyph, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_7draw_glyph(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font = 0;
  PyObject *__pyx_v_gid = 0;
  PyObject *__pyx_v_draw_data = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("draw_glyph (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_font,&__pyx_n_s_gid,&__pyx_n_s_draw_data,0};
    values[2] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_font)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2148, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_gid)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2148, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("draw_glyph", 0, 2, 3, 1); __PYX_ERR(0, 2148, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_draw_data);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2148, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "draw_glyph") < 0)) __PYX_ERR(0, 2148, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_font = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)values[0]);
    __pyx_v_gid = ((PyObject*)values[1]);
    __pyx_v_draw_data = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("draw_glyph", 0, 2, 3, __pyx_nargs); __PYX_ERR(0, 2148, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.DrawFuncs.draw_glyph", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_font), __pyx_ptype_9uharfbuzz_9_harfbuzz_Font, 0, "font", 0))) __PYX_ERR(0, 2148, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_gid), (&PyInt_Type), 0, "gid", 1))) __PYX_ERR(0, 2148, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_9DrawFuncs_6draw_glyph(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *)__pyx_v_self), __pyx_v_font, __pyx_v_gid, __pyx_v_draw_data);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9DrawFuncs_6draw_glyph(struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *__pyx_v_font, PyObject *__pyx_v_gid, PyObject *__pyx_v_draw_data) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("draw_glyph", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2149
 * 
 *     def draw_glyph(self, font: Font, gid: int, draw_data: object = None):
 *         if not self._warned:             # <<<<<<<<<<<<<<
 *             warnings.warn(
 *                 "draw_glyph() is deprecated, use Font.draw_glyph() instead",
 */
  __pyx_t_1 = (!(__pyx_v_self->_warned != 0));
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":2150
 *     def draw_glyph(self, font: Font, gid: int, draw_data: object = None):
 *         if not self._warned:
 *             warnings.warn(             # <<<<<<<<<<<<<<
 *                 "draw_glyph() is deprecated, use Font.draw_glyph() instead",
 *                 DeprecationWarning,
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_warnings); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_warn); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uharfbuzz/_harfbuzz.pyx":2152
 *             warnings.warn(
 *                 "draw_glyph() is deprecated, use Font.draw_glyph() instead",
 *                 DeprecationWarning,             # <<<<<<<<<<<<<<
 *             )
 *             self._warned = True
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2150, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "uharfbuzz/_harfbuzz.pyx":2154
 *                 DeprecationWarning,
 *             )
 *             self._warned = True             # <<<<<<<<<<<<<<
 *         font.draw_glyph(gid, self, draw_data)
 * 
 */
    __pyx_v_self->_warned = 1;

    /* "uharfbuzz/_harfbuzz.pyx":2149
 * 
 *     def draw_glyph(self, font: Font, gid: int, draw_data: object = None):
 *         if not self._warned:             # <<<<<<<<<<<<<<
 *             warnings.warn(
 *                 "draw_glyph() is deprecated, use Font.draw_glyph() instead",
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2155
 *             )
 *             self._warned = True
 *         font.draw_glyph(gid, self, draw_data)             # <<<<<<<<<<<<<<
 * 
 *     def set_move_to_func(self,
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_font), __pyx_n_s_draw_glyph); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_4, __pyx_v_gid, ((PyObject *)__pyx_v_self), __pyx_v_draw_data};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 3+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2155, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2148
 *         font.draw_glyph(gid, self)
 * 
 *     def draw_glyph(self, font: Font, gid: int, draw_data: object = None):             # <<<<<<<<<<<<<<
 *         if not self._warned:
 *             warnings.warn(
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.DrawFuncs.draw_glyph", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2157
 *         font.draw_glyph(gid, self, draw_data)
 * 
 *     def set_move_to_func(self,             # <<<<<<<<<<<<<<
 *                          func: Callable[[
 *                              float,
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_9set_move_to_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_9DrawFuncs_9set_move_to_func = {"set_move_to_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_9set_move_to_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_9set_move_to_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func = 0;
  PyObject *__pyx_v_user_data = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_move_to_func (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func,&__pyx_n_s_user_data,0};

    /* "uharfbuzz/_harfbuzz.pyx":2163
 *                              object,  # draw_data
 *                          ], None],
 *                          user_data: object = None) -> None:             # <<<<<<<<<<<<<<
 *         cdef hb_draw_move_to_func_t func_p
 *         cdef void *user_data_p
 */
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_func)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2157, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_user_data);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2157, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_move_to_func") < 0)) __PYX_ERR(0, 2157, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_func = values[0];
    __pyx_v_user_data = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_move_to_func", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 2157, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.DrawFuncs.set_move_to_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_9DrawFuncs_8set_move_to_func(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *)__pyx_v_self), __pyx_v_func, __pyx_v_user_data);

  /* "uharfbuzz/_harfbuzz.pyx":2157
 *         font.draw_glyph(gid, self, draw_data)
 * 
 *     def set_move_to_func(self,             # <<<<<<<<<<<<<<
 *                          func: Callable[[
 *                              float,
 */

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9DrawFuncs_8set_move_to_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *__pyx_v_self, PyObject *__pyx_v_func, PyObject *__pyx_v_user_data) {
  hb_draw_move_to_func_t __pyx_v_func_p;
  void *__pyx_v_user_data_p;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  void *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_move_to_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2166
 *         cdef hb_draw_move_to_func_t func_p
 *         cdef void *user_data_p
 *         if PyCapsule_IsValid(func, NULL):             # <<<<<<<<<<<<<<
 *             self._move_to_func = None
 *             func_p = <hb_draw_move_to_func_t>PyCapsule_GetPointer(func, NULL)
 */
  __pyx_t_1 = PyCapsule_IsValid(__pyx_v_func, NULL);
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":2167
 *         cdef void *user_data_p
 *         if PyCapsule_IsValid(func, NULL):
 *             self._move_to_func = None             # <<<<<<<<<<<<<<
 *             func_p = <hb_draw_move_to_func_t>PyCapsule_GetPointer(func, NULL)
 *             if PyCapsule_IsValid(user_data, NULL):
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->_move_to_func);
    __Pyx_DECREF(__pyx_v_self->_move_to_func);
    __pyx_v_self->_move_to_func = Py_None;

    /* "uharfbuzz/_harfbuzz.pyx":2168
 *         if PyCapsule_IsValid(func, NULL):
 *             self._move_to_func = None
 *             func_p = <hb_draw_move_to_func_t>PyCapsule_GetPointer(func, NULL)             # <<<<<<<<<<<<<<
 *             if PyCapsule_IsValid(user_data, NULL):
 *                 user_data_p = <void*>PyCapsule_GetPointer(user_data, NULL)
 */
    __pyx_t_2 = PyCapsule_GetPointer(__pyx_v_func, NULL); if (unlikely(__pyx_t_2 == ((void *)NULL) && PyErr_Occurred())) __PYX_ERR(0, 2168, __pyx_L1_error)
    __pyx_v_func_p = ((hb_draw_move_to_func_t)__pyx_t_2);

    /* "uharfbuzz/_harfbuzz.pyx":2169
 *             self._move_to_func = None
 *             func_p = <hb_draw_move_to_func_t>PyCapsule_GetPointer(func, NULL)
 *             if PyCapsule_IsValid(user_data, NULL):             # <<<<<<<<<<<<<<
 *                 user_data_p = <void*>PyCapsule_GetPointer(user_data, NULL)
 *             else:
 */
    __pyx_t_1 = PyCapsule_IsValid(__pyx_v_user_data, NULL);
    if (__pyx_t_1) {

      /* "uharfbuzz/_harfbuzz.pyx":2170
 *             func_p = <hb_draw_move_to_func_t>PyCapsule_GetPointer(func, NULL)
 *             if PyCapsule_IsValid(user_data, NULL):
 *                 user_data_p = <void*>PyCapsule_GetPointer(user_data, NULL)             # <<<<<<<<<<<<<<
 *             else:
 *                 user_data_p = <void*>user_data
 */
      __pyx_t_2 = PyCapsule_GetPointer(__pyx_v_user_data, NULL); if (unlikely(__pyx_t_2 == ((void *)NULL) && PyErr_Occurred())) __PYX_ERR(0, 2170, __pyx_L1_error)
      __pyx_v_user_data_p = ((void *)__pyx_t_2);

      /* "uharfbuzz/_harfbuzz.pyx":2169
 *             self._move_to_func = None
 *             func_p = <hb_draw_move_to_func_t>PyCapsule_GetPointer(func, NULL)
 *             if PyCapsule_IsValid(user_data, NULL):             # <<<<<<<<<<<<<<
 *                 user_data_p = <void*>PyCapsule_GetPointer(user_data, NULL)
 *             else:
 */
      goto __pyx_L4;
    }

    /* "uharfbuzz/_harfbuzz.pyx":2172
 *                 user_data_p = <void*>PyCapsule_GetPointer(user_data, NULL)
 *             else:
 *                 user_data_p = <void*>user_data             # <<<<<<<<<<<<<<
 *         else:
 *             self._move_to_func = func
 */
    /*else*/ {
      __pyx_v_user_data_p = ((void *)__pyx_v_user_data);
    }
    __pyx_L4:;

    /* "uharfbuzz/_harfbuzz.pyx":2166
 *         cdef hb_draw_move_to_func_t func_p
 *         cdef void *user_data_p
 *         if PyCapsule_IsValid(func, NULL):             # <<<<<<<<<<<<<<
 *             self._move_to_func = None
 *             func_p = <hb_draw_move_to_func_t>PyCapsule_GetPointer(func, NULL)
 */
    goto __pyx_L3;
  }

  /* "uharfbuzz/_harfbuzz.pyx":2174
 *                 user_data_p = <void*>user_data
 *         else:
 *             self._move_to_func = func             # <<<<<<<<<<<<<<
 *             func_p = _move_to_func
 *             assert user_data is None, "Pass draw_state to Font.draw_glyph"
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_v_func);
    __Pyx_GIVEREF(__pyx_v_func);
    __Pyx_GOTREF(__pyx_v_self->_move_to_func);
    __Pyx_DECREF(__pyx_v_self->_move_to_func);
    __pyx_v_self->_move_to_func = __pyx_v_func;

    /* "uharfbuzz/_harfbuzz.pyx":2175
 *         else:
 *             self._move_to_func = func
 *             func_p = _move_to_func             # <<<<<<<<<<<<<<
 *             assert user_data is None, "Pass draw_state to Font.draw_glyph"
 *             user_data_p = <void*>func
 */
    __pyx_v_func_p = __pyx_f_9uharfbuzz_9_harfbuzz__move_to_func;

    /* "uharfbuzz/_harfbuzz.pyx":2176
 *             self._move_to_func = func
 *             func_p = _move_to_func
 *             assert user_data is None, "Pass draw_state to Font.draw_glyph"             # <<<<<<<<<<<<<<
 *             user_data_p = <void*>func
 *         hb_draw_funcs_set_move_to_func(
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_1 = (__pyx_v_user_data == Py_None);
      if (unlikely(!__pyx_t_1)) {
        __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_kp_u_Pass_draw_state_to_Font_draw_gly, 0, 0);
        __PYX_ERR(0, 2176, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 2176, __pyx_L1_error)
    #endif

    /* "uharfbuzz/_harfbuzz.pyx":2177
 *             func_p = _move_to_func
 *             assert user_data is None, "Pass draw_state to Font.draw_glyph"
 *             user_data_p = <void*>func             # <<<<<<<<<<<<<<
 *         hb_draw_funcs_set_move_to_func(
 *             self._hb_drawfuncs, func_p, user_data_p, NULL)
 */
    __pyx_v_user_data_p = ((void *)__pyx_v_func);
  }
  __pyx_L3:;

  /* "uharfbuzz/_harfbuzz.pyx":2178
 *             assert user_data is None, "Pass draw_state to Font.draw_glyph"
 *             user_data_p = <void*>func
 *         hb_draw_funcs_set_move_to_func(             # <<<<<<<<<<<<<<
 *             self._hb_drawfuncs, func_p, user_data_p, NULL)
 * 
 */
  hb_draw_funcs_set_move_to_func(__pyx_v_self->_hb_drawfuncs, __pyx_v_func_p, __pyx_v_user_data_p, NULL);

  /* "uharfbuzz/_harfbuzz.pyx":2157
 *         font.draw_glyph(gid, self, draw_data)
 * 
 *     def set_move_to_func(self,             # <<<<<<<<<<<<<<
 *                          func: Callable[[
 *                              float,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.DrawFuncs.set_move_to_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2181
 *             self._hb_drawfuncs, func_p, user_data_p, NULL)
 * 
 *     def set_line_to_func(self,             # <<<<<<<<<<<<<<
 *                          func: Callable[[
 *                              float,
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_11set_line_to_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_9DrawFuncs_11set_line_to_func = {"set_line_to_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_11set_line_to_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_11set_line_to_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func = 0;
  PyObject *__pyx_v_user_data = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_line_to_func (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func,&__pyx_n_s_user_data,0};

    /* "uharfbuzz/_harfbuzz.pyx":2187
 *                              object,  # draw_data
 *                          ], None],
 *                          user_data: object = None) -> None:             # <<<<<<<<<<<<<<
 *         cdef hb_draw_line_to_func_t func_p
 *         cdef void *user_data_p
 */
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_func)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2181, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_user_data);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2181, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_line_to_func") < 0)) __PYX_ERR(0, 2181, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_func = values[0];
    __pyx_v_user_data = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_line_to_func", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 2181, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.DrawFuncs.set_line_to_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_9DrawFuncs_10set_line_to_func(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *)__pyx_v_self), __pyx_v_func, __pyx_v_user_data);

  /* "uharfbuzz/_harfbuzz.pyx":2181
 *             self._hb_drawfuncs, func_p, user_data_p, NULL)
 * 
 *     def set_line_to_func(self,             # <<<<<<<<<<<<<<
 *                          func: Callable[[
 *                              float,
 */

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9DrawFuncs_10set_line_to_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *__pyx_v_self, PyObject *__pyx_v_func, PyObject *__pyx_v_user_data) {
  hb_draw_line_to_func_t __pyx_v_func_p;
  void *__pyx_v_user_data_p;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  void *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_line_to_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2190
 *         cdef hb_draw_line_to_func_t func_p
 *         cdef void *user_data_p
 *         if PyCapsule_IsValid(func, NULL):             # <<<<<<<<<<<<<<
 *             self._line_to_func = None
 *             func_p = <hb_draw_line_to_func_t>PyCapsule_GetPointer(func, NULL)
 */
  __pyx_t_1 = PyCapsule_IsValid(__pyx_v_func, NULL);
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":2191
 *         cdef void *user_data_p
 *         if PyCapsule_IsValid(func, NULL):
 *             self._line_to_func = None             # <<<<<<<<<<<<<<
 *             func_p = <hb_draw_line_to_func_t>PyCapsule_GetPointer(func, NULL)
 *             if PyCapsule_IsValid(user_data, NULL):
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->_line_to_func);
    __Pyx_DECREF(__pyx_v_self->_line_to_func);
    __pyx_v_self->_line_to_func = Py_None;

    /* "uharfbuzz/_harfbuzz.pyx":2192
 *         if PyCapsule_IsValid(func, NULL):
 *             self._line_to_func = None
 *             func_p = <hb_draw_line_to_func_t>PyCapsule_GetPointer(func, NULL)             # <<<<<<<<<<<<<<
 *             if PyCapsule_IsValid(user_data, NULL):
 *                 user_data_p = <void*>PyCapsule_GetPointer(user_data, NULL)
 */
    __pyx_t_2 = PyCapsule_GetPointer(__pyx_v_func, NULL); if (unlikely(__pyx_t_2 == ((void *)NULL) && PyErr_Occurred())) __PYX_ERR(0, 2192, __pyx_L1_error)
    __pyx_v_func_p = ((hb_draw_line_to_func_t)__pyx_t_2);

    /* "uharfbuzz/_harfbuzz.pyx":2193
 *             self._line_to_func = None
 *             func_p = <hb_draw_line_to_func_t>PyCapsule_GetPointer(func, NULL)
 *             if PyCapsule_IsValid(user_data, NULL):             # <<<<<<<<<<<<<<
 *                 user_data_p = <void*>PyCapsule_GetPointer(user_data, NULL)
 *             else:
 */
    __pyx_t_1 = PyCapsule_IsValid(__pyx_v_user_data, NULL);
    if (__pyx_t_1) {

      /* "uharfbuzz/_harfbuzz.pyx":2194
 *             func_p = <hb_draw_line_to_func_t>PyCapsule_GetPointer(func, NULL)
 *             if PyCapsule_IsValid(user_data, NULL):
 *                 user_data_p = <void*>PyCapsule_GetPointer(user_data, NULL)             # <<<<<<<<<<<<<<
 *             else:
 *                 user_data_p = <void*>user_data
 */
      __pyx_t_2 = PyCapsule_GetPointer(__pyx_v_user_data, NULL); if (unlikely(__pyx_t_2 == ((void *)NULL) && PyErr_Occurred())) __PYX_ERR(0, 2194, __pyx_L1_error)
      __pyx_v_user_data_p = ((void *)__pyx_t_2);

      /* "uharfbuzz/_harfbuzz.pyx":2193
 *             self._line_to_func = None
 *             func_p = <hb_draw_line_to_func_t>PyCapsule_GetPointer(func, NULL)
 *             if PyCapsule_IsValid(user_data, NULL):             # <<<<<<<<<<<<<<
 *                 user_data_p = <void*>PyCapsule_GetPointer(user_data, NULL)
 *             else:
 */
      goto __pyx_L4;
    }

    /* "uharfbuzz/_harfbuzz.pyx":2196
 *                 user_data_p = <void*>PyCapsule_GetPointer(user_data, NULL)
 *             else:
 *                 user_data_p = <void*>user_data             # <<<<<<<<<<<<<<
 *         else:
 *             self._line_to_func = func
 */
    /*else*/ {
      __pyx_v_user_data_p = ((void *)__pyx_v_user_data);
    }
    __pyx_L4:;

    /* "uharfbuzz/_harfbuzz.pyx":2190
 *         cdef hb_draw_line_to_func_t func_p
 *         cdef void *user_data_p
 *         if PyCapsule_IsValid(func, NULL):             # <<<<<<<<<<<<<<
 *             self._line_to_func = None
 *             func_p = <hb_draw_line_to_func_t>PyCapsule_GetPointer(func, NULL)
 */
    goto __pyx_L3;
  }

  /* "uharfbuzz/_harfbuzz.pyx":2198
 *                 user_data_p = <void*>user_data
 *         else:
 *             self._line_to_func = func             # <<<<<<<<<<<<<<
 *             func_p = _line_to_func
 *             assert user_data is None, "Pass draw_state to Font.draw_glyph"
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_v_func);
    __Pyx_GIVEREF(__pyx_v_func);
    __Pyx_GOTREF(__pyx_v_self->_line_to_func);
    __Pyx_DECREF(__pyx_v_self->_line_to_func);
    __pyx_v_self->_line_to_func = __pyx_v_func;

    /* "uharfbuzz/_harfbuzz.pyx":2199
 *         else:
 *             self._line_to_func = func
 *             func_p = _line_to_func             # <<<<<<<<<<<<<<
 *             assert user_data is None, "Pass draw_state to Font.draw_glyph"
 *             user_data_p = <void*>func
 */
    __pyx_v_func_p = __pyx_f_9uharfbuzz_9_harfbuzz__line_to_func;

    /* "uharfbuzz/_harfbuzz.pyx":2200
 *             self._line_to_func = func
 *             func_p = _line_to_func
 *             assert user_data is None, "Pass draw_state to Font.draw_glyph"             # <<<<<<<<<<<<<<
 *             user_data_p = <void*>func
 *         hb_draw_funcs_set_line_to_func(
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_1 = (__pyx_v_user_data == Py_None);
      if (unlikely(!__pyx_t_1)) {
        __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_kp_u_Pass_draw_state_to_Font_draw_gly, 0, 0);
        __PYX_ERR(0, 2200, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 2200, __pyx_L1_error)
    #endif

    /* "uharfbuzz/_harfbuzz.pyx":2201
 *             func_p = _line_to_func
 *             assert user_data is None, "Pass draw_state to Font.draw_glyph"
 *             user_data_p = <void*>func             # <<<<<<<<<<<<<<
 *         hb_draw_funcs_set_line_to_func(
 *             self._hb_drawfuncs, func_p, user_data_p, NULL)
 */
    __pyx_v_user_data_p = ((void *)__pyx_v_func);
  }
  __pyx_L3:;

  /* "uharfbuzz/_harfbuzz.pyx":2202
 *             assert user_data is None, "Pass draw_state to Font.draw_glyph"
 *             user_data_p = <void*>func
 *         hb_draw_funcs_set_line_to_func(             # <<<<<<<<<<<<<<
 *             self._hb_drawfuncs, func_p, user_data_p, NULL)
 * 
 */
  hb_draw_funcs_set_line_to_func(__pyx_v_self->_hb_drawfuncs, __pyx_v_func_p, __pyx_v_user_data_p, NULL);

  /* "uharfbuzz/_harfbuzz.pyx":2181
 *             self._hb_drawfuncs, func_p, user_data_p, NULL)
 * 
 *     def set_line_to_func(self,             # <<<<<<<<<<<<<<
 *                          func: Callable[[
 *                              float,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.DrawFuncs.set_line_to_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2205
 *             self._hb_drawfuncs, func_p, user_data_p, NULL)
 * 
 *     def set_cubic_to_func(self,             # <<<<<<<<<<<<<<
 *                           func: Callable[[
 *                              float,
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_13set_cubic_to_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_9DrawFuncs_13set_cubic_to_func = {"set_cubic_to_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_13set_cubic_to_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_13set_cubic_to_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func = 0;
  PyObject *__pyx_v_user_data = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_cubic_to_func (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func,&__pyx_n_s_user_data,0};

    /* "uharfbuzz/_harfbuzz.pyx":2215
 *                              object,  # draw_data
 *                           ], None],
 *                           user_data: object = None) -> None:             # <<<<<<<<<<<<<<
 *         cdef hb_draw_cubic_to_func_t func_p
 *         cdef void *user_data_p
 */
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_func)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2205, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_user_data);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2205, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_cubic_to_func") < 0)) __PYX_ERR(0, 2205, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_func = values[0];
    __pyx_v_user_data = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_cubic_to_func", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 2205, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.DrawFuncs.set_cubic_to_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_9DrawFuncs_12set_cubic_to_func(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *)__pyx_v_self), __pyx_v_func, __pyx_v_user_data);

  /* "uharfbuzz/_harfbuzz.pyx":2205
 *             self._hb_drawfuncs, func_p, user_data_p, NULL)
 * 
 *     def set_cubic_to_func(self,             # <<<<<<<<<<<<<<
 *                           func: Callable[[
 *                              float,
 */

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9DrawFuncs_12set_cubic_to_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *__pyx_v_self, PyObject *__pyx_v_func, PyObject *__pyx_v_user_data) {
  hb_draw_cubic_to_func_t __pyx_v_func_p;
  void *__pyx_v_user_data_p;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  void *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_cubic_to_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2218
 *         cdef hb_draw_cubic_to_func_t func_p
 *         cdef void *user_data_p
 *         if PyCapsule_IsValid(func, NULL):             # <<<<<<<<<<<<<<
 *             self._cubic_to_func = None
 *             func_p = <hb_draw_cubic_to_func_t>PyCapsule_GetPointer(func, NULL)
 */
  __pyx_t_1 = PyCapsule_IsValid(__pyx_v_func, NULL);
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":2219
 *         cdef void *user_data_p
 *         if PyCapsule_IsValid(func, NULL):
 *             self._cubic_to_func = None             # <<<<<<<<<<<<<<
 *             func_p = <hb_draw_cubic_to_func_t>PyCapsule_GetPointer(func, NULL)
 *             if PyCapsule_IsValid(user_data, NULL):
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->_cubic_to_func);
    __Pyx_DECREF(__pyx_v_self->_cubic_to_func);
    __pyx_v_self->_cubic_to_func = Py_None;

    /* "uharfbuzz/_harfbuzz.pyx":2220
 *         if PyCapsule_IsValid(func, NULL):
 *             self._cubic_to_func = None
 *             func_p = <hb_draw_cubic_to_func_t>PyCapsule_GetPointer(func, NULL)             # <<<<<<<<<<<<<<
 *             if PyCapsule_IsValid(user_data, NULL):
 *                 user_data_p = <void*>PyCapsule_GetPointer(user_data, NULL)
 */
    __pyx_t_2 = PyCapsule_GetPointer(__pyx_v_func, NULL); if (unlikely(__pyx_t_2 == ((void *)NULL) && PyErr_Occurred())) __PYX_ERR(0, 2220, __pyx_L1_error)
    __pyx_v_func_p = ((hb_draw_cubic_to_func_t)__pyx_t_2);

    /* "uharfbuzz/_harfbuzz.pyx":2221
 *             self._cubic_to_func = None
 *             func_p = <hb_draw_cubic_to_func_t>PyCapsule_GetPointer(func, NULL)
 *             if PyCapsule_IsValid(user_data, NULL):             # <<<<<<<<<<<<<<
 *                 user_data_p = <void*>PyCapsule_GetPointer(user_data, NULL)
 *             else:
 */
    __pyx_t_1 = PyCapsule_IsValid(__pyx_v_user_data, NULL);
    if (__pyx_t_1) {

      /* "uharfbuzz/_harfbuzz.pyx":2222
 *             func_p = <hb_draw_cubic_to_func_t>PyCapsule_GetPointer(func, NULL)
 *             if PyCapsule_IsValid(user_data, NULL):
 *                 user_data_p = <void*>PyCapsule_GetPointer(user_data, NULL)             # <<<<<<<<<<<<<<
 *             else:
 *                 user_data_p = <void*>user_data
 */
      __pyx_t_2 = PyCapsule_GetPointer(__pyx_v_user_data, NULL); if (unlikely(__pyx_t_2 == ((void *)NULL) && PyErr_Occurred())) __PYX_ERR(0, 2222, __pyx_L1_error)
      __pyx_v_user_data_p = ((void *)__pyx_t_2);

      /* "uharfbuzz/_harfbuzz.pyx":2221
 *             self._cubic_to_func = None
 *             func_p = <hb_draw_cubic_to_func_t>PyCapsule_GetPointer(func, NULL)
 *             if PyCapsule_IsValid(user_data, NULL):             # <<<<<<<<<<<<<<
 *                 user_data_p = <void*>PyCapsule_GetPointer(user_data, NULL)
 *             else:
 */
      goto __pyx_L4;
    }

    /* "uharfbuzz/_harfbuzz.pyx":2224
 *                 user_data_p = <void*>PyCapsule_GetPointer(user_data, NULL)
 *             else:
 *                 user_data_p = <void*>user_data             # <<<<<<<<<<<<<<
 *         else:
 *             self._cubic_to_func = func
 */
    /*else*/ {
      __pyx_v_user_data_p = ((void *)__pyx_v_user_data);
    }
    __pyx_L4:;

    /* "uharfbuzz/_harfbuzz.pyx":2218
 *         cdef hb_draw_cubic_to_func_t func_p
 *         cdef void *user_data_p
 *         if PyCapsule_IsValid(func, NULL):             # <<<<<<<<<<<<<<
 *             self._cubic_to_func = None
 *             func_p = <hb_draw_cubic_to_func_t>PyCapsule_GetPointer(func, NULL)
 */
    goto __pyx_L3;
  }

  /* "uharfbuzz/_harfbuzz.pyx":2226
 *                 user_data_p = <void*>user_data
 *         else:
 *             self._cubic_to_func = func             # <<<<<<<<<<<<<<
 *             func_p = _cubic_to_func
 *             assert user_data is None, "Pass draw_state to Font.draw_glyph"
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_v_func);
    __Pyx_GIVEREF(__pyx_v_func);
    __Pyx_GOTREF(__pyx_v_self->_cubic_to_func);
    __Pyx_DECREF(__pyx_v_self->_cubic_to_func);
    __pyx_v_self->_cubic_to_func = __pyx_v_func;

    /* "uharfbuzz/_harfbuzz.pyx":2227
 *         else:
 *             self._cubic_to_func = func
 *             func_p = _cubic_to_func             # <<<<<<<<<<<<<<
 *             assert user_data is None, "Pass draw_state to Font.draw_glyph"
 *             user_data_p = <void*>func
 */
    __pyx_v_func_p = __pyx_f_9uharfbuzz_9_harfbuzz__cubic_to_func;

    /* "uharfbuzz/_harfbuzz.pyx":2228
 *             self._cubic_to_func = func
 *             func_p = _cubic_to_func
 *             assert user_data is None, "Pass draw_state to Font.draw_glyph"             # <<<<<<<<<<<<<<
 *             user_data_p = <void*>func
 *         hb_draw_funcs_set_cubic_to_func(
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_1 = (__pyx_v_user_data == Py_None);
      if (unlikely(!__pyx_t_1)) {
        __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_kp_u_Pass_draw_state_to_Font_draw_gly, 0, 0);
        __PYX_ERR(0, 2228, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 2228, __pyx_L1_error)
    #endif

    /* "uharfbuzz/_harfbuzz.pyx":2229
 *             func_p = _cubic_to_func
 *             assert user_data is None, "Pass draw_state to Font.draw_glyph"
 *             user_data_p = <void*>func             # <<<<<<<<<<<<<<
 *         hb_draw_funcs_set_cubic_to_func(
 *             self._hb_drawfuncs, func_p, user_data_p, NULL)
 */
    __pyx_v_user_data_p = ((void *)__pyx_v_func);
  }
  __pyx_L3:;

  /* "uharfbuzz/_harfbuzz.pyx":2230
 *             assert user_data is None, "Pass draw_state to Font.draw_glyph"
 *             user_data_p = <void*>func
 *         hb_draw_funcs_set_cubic_to_func(             # <<<<<<<<<<<<<<
 *             self._hb_drawfuncs, func_p, user_data_p, NULL)
 * 
 */
  hb_draw_funcs_set_cubic_to_func(__pyx_v_self->_hb_drawfuncs, __pyx_v_func_p, __pyx_v_user_data_p, NULL);

  /* "uharfbuzz/_harfbuzz.pyx":2205
 *             self._hb_drawfuncs, func_p, user_data_p, NULL)
 * 
 *     def set_cubic_to_func(self,             # <<<<<<<<<<<<<<
 *                           func: Callable[[
 *                              float,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.DrawFuncs.set_cubic_to_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2233
 *             self._hb_drawfuncs, func_p, user_data_p, NULL)
 * 
 *     def set_quadratic_to_func(self,             # <<<<<<<<<<<<<<
 *                               func: Callable[[
 *                                  float,
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_15set_quadratic_to_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_9DrawFuncs_15set_quadratic_to_func = {"set_quadratic_to_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_15set_quadratic_to_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_15set_quadratic_to_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func = 0;
  PyObject *__pyx_v_user_data = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_quadratic_to_func (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func,&__pyx_n_s_user_data,0};

    /* "uharfbuzz/_harfbuzz.pyx":2241
 *                                  object,  # draw_data
 *                               ], None],
 *                               user_data: object = None) -> None:             # <<<<<<<<<<<<<<
 *         cdef hb_draw_quadratic_to_func_t func_p
 *         cdef void *user_data_p
 */
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_func)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2233, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_user_data);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2233, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_quadratic_to_func") < 0)) __PYX_ERR(0, 2233, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_func = values[0];
    __pyx_v_user_data = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_quadratic_to_func", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 2233, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.DrawFuncs.set_quadratic_to_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_9DrawFuncs_14set_quadratic_to_func(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *)__pyx_v_self), __pyx_v_func, __pyx_v_user_data);

  /* "uharfbuzz/_harfbuzz.pyx":2233
 *             self._hb_drawfuncs, func_p, user_data_p, NULL)
 * 
 *     def set_quadratic_to_func(self,             # <<<<<<<<<<<<<<
 *                               func: Callable[[
 *                                  float,
 */

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9DrawFuncs_14set_quadratic_to_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *__pyx_v_self, PyObject *__pyx_v_func, PyObject *__pyx_v_user_data) {
  hb_draw_quadratic_to_func_t __pyx_v_func_p;
  void *__pyx_v_user_data_p;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  void *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_quadratic_to_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2244
 *         cdef hb_draw_quadratic_to_func_t func_p
 *         cdef void *user_data_p
 *         if PyCapsule_IsValid(func, NULL):             # <<<<<<<<<<<<<<
 *             self._quadratic_to_func = None
 *             func_p = <hb_draw_quadratic_to_func_t>PyCapsule_GetPointer(func, NULL)
 */
  __pyx_t_1 = PyCapsule_IsValid(__pyx_v_func, NULL);
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":2245
 *         cdef void *user_data_p
 *         if PyCapsule_IsValid(func, NULL):
 *             self._quadratic_to_func = None             # <<<<<<<<<<<<<<
 *             func_p = <hb_draw_quadratic_to_func_t>PyCapsule_GetPointer(func, NULL)
 *             if PyCapsule_IsValid(user_data, NULL):
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->_quadratic_to_func);
    __Pyx_DECREF(__pyx_v_self->_quadratic_to_func);
    __pyx_v_self->_quadratic_to_func = Py_None;

    /* "uharfbuzz/_harfbuzz.pyx":2246
 *         if PyCapsule_IsValid(func, NULL):
 *             self._quadratic_to_func = None
 *             func_p = <hb_draw_quadratic_to_func_t>PyCapsule_GetPointer(func, NULL)             # <<<<<<<<<<<<<<
 *             if PyCapsule_IsValid(user_data, NULL):
 *                 user_data_p = <void*>PyCapsule_GetPointer(user_data, NULL)
 */
    __pyx_t_2 = PyCapsule_GetPointer(__pyx_v_func, NULL); if (unlikely(__pyx_t_2 == ((void *)NULL) && PyErr_Occurred())) __PYX_ERR(0, 2246, __pyx_L1_error)
    __pyx_v_func_p = ((hb_draw_quadratic_to_func_t)__pyx_t_2);

    /* "uharfbuzz/_harfbuzz.pyx":2247
 *             self._quadratic_to_func = None
 *             func_p = <hb_draw_quadratic_to_func_t>PyCapsule_GetPointer(func, NULL)
 *             if PyCapsule_IsValid(user_data, NULL):             # <<<<<<<<<<<<<<
 *                 user_data_p = <void*>PyCapsule_GetPointer(user_data, NULL)
 *             else:
 */
    __pyx_t_1 = PyCapsule_IsValid(__pyx_v_user_data, NULL);
    if (__pyx_t_1) {

      /* "uharfbuzz/_harfbuzz.pyx":2248
 *             func_p = <hb_draw_quadratic_to_func_t>PyCapsule_GetPointer(func, NULL)
 *             if PyCapsule_IsValid(user_data, NULL):
 *                 user_data_p = <void*>PyCapsule_GetPointer(user_data, NULL)             # <<<<<<<<<<<<<<
 *             else:
 *                 user_data_p = <void*>user_data
 */
      __pyx_t_2 = PyCapsule_GetPointer(__pyx_v_user_data, NULL); if (unlikely(__pyx_t_2 == ((void *)NULL) && PyErr_Occurred())) __PYX_ERR(0, 2248, __pyx_L1_error)
      __pyx_v_user_data_p = ((void *)__pyx_t_2);

      /* "uharfbuzz/_harfbuzz.pyx":2247
 *             self._quadratic_to_func = None
 *             func_p = <hb_draw_quadratic_to_func_t>PyCapsule_GetPointer(func, NULL)
 *             if PyCapsule_IsValid(user_data, NULL):             # <<<<<<<<<<<<<<
 *                 user_data_p = <void*>PyCapsule_GetPointer(user_data, NULL)
 *             else:
 */
      goto __pyx_L4;
    }

    /* "uharfbuzz/_harfbuzz.pyx":2250
 *                 user_data_p = <void*>PyCapsule_GetPointer(user_data, NULL)
 *             else:
 *                 user_data_p = <void*>user_data             # <<<<<<<<<<<<<<
 *         else:
 *             self._quadratic_to_func = func
 */
    /*else*/ {
      __pyx_v_user_data_p = ((void *)__pyx_v_user_data);
    }
    __pyx_L4:;

    /* "uharfbuzz/_harfbuzz.pyx":2244
 *         cdef hb_draw_quadratic_to_func_t func_p
 *         cdef void *user_data_p
 *         if PyCapsule_IsValid(func, NULL):             # <<<<<<<<<<<<<<
 *             self._quadratic_to_func = None
 *             func_p = <hb_draw_quadratic_to_func_t>PyCapsule_GetPointer(func, NULL)
 */
    goto __pyx_L3;
  }

  /* "uharfbuzz/_harfbuzz.pyx":2252
 *                 user_data_p = <void*>user_data
 *         else:
 *             self._quadratic_to_func = func             # <<<<<<<<<<<<<<
 *             func_p = _quadratic_to_func
 *             assert user_data is None, "Pass draw_state to Font.draw_glyph"
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_v_func);
    __Pyx_GIVEREF(__pyx_v_func);
    __Pyx_GOTREF(__pyx_v_self->_quadratic_to_func);
    __Pyx_DECREF(__pyx_v_self->_quadratic_to_func);
    __pyx_v_self->_quadratic_to_func = __pyx_v_func;

    /* "uharfbuzz/_harfbuzz.pyx":2253
 *         else:
 *             self._quadratic_to_func = func
 *             func_p = _quadratic_to_func             # <<<<<<<<<<<<<<
 *             assert user_data is None, "Pass draw_state to Font.draw_glyph"
 *             user_data_p = <void*>func
 */
    __pyx_v_func_p = __pyx_f_9uharfbuzz_9_harfbuzz__quadratic_to_func;

    /* "uharfbuzz/_harfbuzz.pyx":2254
 *             self._quadratic_to_func = func
 *             func_p = _quadratic_to_func
 *             assert user_data is None, "Pass draw_state to Font.draw_glyph"             # <<<<<<<<<<<<<<
 *             user_data_p = <void*>func
 *         hb_draw_funcs_set_quadratic_to_func(
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_1 = (__pyx_v_user_data == Py_None);
      if (unlikely(!__pyx_t_1)) {
        __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_kp_u_Pass_draw_state_to_Font_draw_gly, 0, 0);
        __PYX_ERR(0, 2254, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 2254, __pyx_L1_error)
    #endif

    /* "uharfbuzz/_harfbuzz.pyx":2255
 *             func_p = _quadratic_to_func
 *             assert user_data is None, "Pass draw_state to Font.draw_glyph"
 *             user_data_p = <void*>func             # <<<<<<<<<<<<<<
 *         hb_draw_funcs_set_quadratic_to_func(
 *             self._hb_drawfuncs, func_p, user_data_p, NULL)
 */
    __pyx_v_user_data_p = ((void *)__pyx_v_func);
  }
  __pyx_L3:;

  /* "uharfbuzz/_harfbuzz.pyx":2256
 *             assert user_data is None, "Pass draw_state to Font.draw_glyph"
 *             user_data_p = <void*>func
 *         hb_draw_funcs_set_quadratic_to_func(             # <<<<<<<<<<<<<<
 *             self._hb_drawfuncs, func_p, user_data_p, NULL)
 * 
 */
  hb_draw_funcs_set_quadratic_to_func(__pyx_v_self->_hb_drawfuncs, __pyx_v_func_p, __pyx_v_user_data_p, NULL);

  /* "uharfbuzz/_harfbuzz.pyx":2233
 *             self._hb_drawfuncs, func_p, user_data_p, NULL)
 * 
 *     def set_quadratic_to_func(self,             # <<<<<<<<<<<<<<
 *                               func: Callable[[
 *                                  float,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.DrawFuncs.set_quadratic_to_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2259
 *             self._hb_drawfuncs, func_p, user_data_p, NULL)
 * 
 *     def set_close_path_func(self,             # <<<<<<<<<<<<<<
 *                             func: Callable[[
 *                                 object
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_17set_close_path_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_9DrawFuncs_17set_close_path_func = {"set_close_path_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_17set_close_path_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_17set_close_path_func(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_func = 0;
  PyObject *__pyx_v_user_data = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_close_path_func (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_func,&__pyx_n_s_user_data,0};

    /* "uharfbuzz/_harfbuzz.pyx":2263
 *                                 object
 *                             ], None],
 *                             user_data: object = None) -> None:             # <<<<<<<<<<<<<<
 *         cdef hb_draw_close_path_func_t func_p
 *         cdef void *user_data_p
 */
    values[1] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_func)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2259, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_user_data);
          if (value) { values[1] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2259, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_close_path_func") < 0)) __PYX_ERR(0, 2259, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_func = values[0];
    __pyx_v_user_data = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_close_path_func", 0, 1, 2, __pyx_nargs); __PYX_ERR(0, 2259, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.DrawFuncs.set_close_path_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_9DrawFuncs_16set_close_path_func(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *)__pyx_v_self), __pyx_v_func, __pyx_v_user_data);

  /* "uharfbuzz/_harfbuzz.pyx":2259
 *             self._hb_drawfuncs, func_p, user_data_p, NULL)
 * 
 *     def set_close_path_func(self,             # <<<<<<<<<<<<<<
 *                             func: Callable[[
 *                                 object
 */

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9DrawFuncs_16set_close_path_func(struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *__pyx_v_self, PyObject *__pyx_v_func, PyObject *__pyx_v_user_data) {
  hb_draw_close_path_func_t __pyx_v_func_p;
  void *__pyx_v_user_data_p;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  void *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_close_path_func", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2266
 *         cdef hb_draw_close_path_func_t func_p
 *         cdef void *user_data_p
 *         if PyCapsule_IsValid(func, NULL):             # <<<<<<<<<<<<<<
 *             self._close_path_func = None
 *             func_p = <hb_draw_close_path_func_t>PyCapsule_GetPointer(func, NULL)
 */
  __pyx_t_1 = PyCapsule_IsValid(__pyx_v_func, NULL);
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":2267
 *         cdef void *user_data_p
 *         if PyCapsule_IsValid(func, NULL):
 *             self._close_path_func = None             # <<<<<<<<<<<<<<
 *             func_p = <hb_draw_close_path_func_t>PyCapsule_GetPointer(func, NULL)
 *             if PyCapsule_IsValid(user_data, NULL):
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->_close_path_func);
    __Pyx_DECREF(__pyx_v_self->_close_path_func);
    __pyx_v_self->_close_path_func = Py_None;

    /* "uharfbuzz/_harfbuzz.pyx":2268
 *         if PyCapsule_IsValid(func, NULL):
 *             self._close_path_func = None
 *             func_p = <hb_draw_close_path_func_t>PyCapsule_GetPointer(func, NULL)             # <<<<<<<<<<<<<<
 *             if PyCapsule_IsValid(user_data, NULL):
 *                 user_data_p = <void*>PyCapsule_GetPointer(user_data, NULL)
 */
    __pyx_t_2 = PyCapsule_GetPointer(__pyx_v_func, NULL); if (unlikely(__pyx_t_2 == ((void *)NULL) && PyErr_Occurred())) __PYX_ERR(0, 2268, __pyx_L1_error)
    __pyx_v_func_p = ((hb_draw_close_path_func_t)__pyx_t_2);

    /* "uharfbuzz/_harfbuzz.pyx":2269
 *             self._close_path_func = None
 *             func_p = <hb_draw_close_path_func_t>PyCapsule_GetPointer(func, NULL)
 *             if PyCapsule_IsValid(user_data, NULL):             # <<<<<<<<<<<<<<
 *                 user_data_p = <void*>PyCapsule_GetPointer(user_data, NULL)
 *             else:
 */
    __pyx_t_1 = PyCapsule_IsValid(__pyx_v_user_data, NULL);
    if (__pyx_t_1) {

      /* "uharfbuzz/_harfbuzz.pyx":2270
 *             func_p = <hb_draw_close_path_func_t>PyCapsule_GetPointer(func, NULL)
 *             if PyCapsule_IsValid(user_data, NULL):
 *                 user_data_p = <void*>PyCapsule_GetPointer(user_data, NULL)             # <<<<<<<<<<<<<<
 *             else:
 *                 user_data_p = <void*>user_data
 */
      __pyx_t_2 = PyCapsule_GetPointer(__pyx_v_user_data, NULL); if (unlikely(__pyx_t_2 == ((void *)NULL) && PyErr_Occurred())) __PYX_ERR(0, 2270, __pyx_L1_error)
      __pyx_v_user_data_p = ((void *)__pyx_t_2);

      /* "uharfbuzz/_harfbuzz.pyx":2269
 *             self._close_path_func = None
 *             func_p = <hb_draw_close_path_func_t>PyCapsule_GetPointer(func, NULL)
 *             if PyCapsule_IsValid(user_data, NULL):             # <<<<<<<<<<<<<<
 *                 user_data_p = <void*>PyCapsule_GetPointer(user_data, NULL)
 *             else:
 */
      goto __pyx_L4;
    }

    /* "uharfbuzz/_harfbuzz.pyx":2272
 *                 user_data_p = <void*>PyCapsule_GetPointer(user_data, NULL)
 *             else:
 *                 user_data_p = <void*>user_data             # <<<<<<<<<<<<<<
 *         else:
 *             self._close_path_func = func
 */
    /*else*/ {
      __pyx_v_user_data_p = ((void *)__pyx_v_user_data);
    }
    __pyx_L4:;

    /* "uharfbuzz/_harfbuzz.pyx":2266
 *         cdef hb_draw_close_path_func_t func_p
 *         cdef void *user_data_p
 *         if PyCapsule_IsValid(func, NULL):             # <<<<<<<<<<<<<<
 *             self._close_path_func = None
 *             func_p = <hb_draw_close_path_func_t>PyCapsule_GetPointer(func, NULL)
 */
    goto __pyx_L3;
  }

  /* "uharfbuzz/_harfbuzz.pyx":2274
 *                 user_data_p = <void*>user_data
 *         else:
 *             self._close_path_func = func             # <<<<<<<<<<<<<<
 *             func_p = _close_path_func
 *             assert user_data is None, "Pass draw_state to Font.draw_glyph"
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_v_func);
    __Pyx_GIVEREF(__pyx_v_func);
    __Pyx_GOTREF(__pyx_v_self->_close_path_func);
    __Pyx_DECREF(__pyx_v_self->_close_path_func);
    __pyx_v_self->_close_path_func = __pyx_v_func;

    /* "uharfbuzz/_harfbuzz.pyx":2275
 *         else:
 *             self._close_path_func = func
 *             func_p = _close_path_func             # <<<<<<<<<<<<<<
 *             assert user_data is None, "Pass draw_state to Font.draw_glyph"
 *             user_data_p = <void*>func
 */
    __pyx_v_func_p = __pyx_f_9uharfbuzz_9_harfbuzz__close_path_func;

    /* "uharfbuzz/_harfbuzz.pyx":2276
 *             self._close_path_func = func
 *             func_p = _close_path_func
 *             assert user_data is None, "Pass draw_state to Font.draw_glyph"             # <<<<<<<<<<<<<<
 *             user_data_p = <void*>func
 *         hb_draw_funcs_set_close_path_func(
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_1 = (__pyx_v_user_data == Py_None);
      if (unlikely(!__pyx_t_1)) {
        __Pyx_Raise(__pyx_builtin_AssertionError, __pyx_kp_u_Pass_draw_state_to_Font_draw_gly, 0, 0);
        __PYX_ERR(0, 2276, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 2276, __pyx_L1_error)
    #endif

    /* "uharfbuzz/_harfbuzz.pyx":2277
 *             func_p = _close_path_func
 *             assert user_data is None, "Pass draw_state to Font.draw_glyph"
 *             user_data_p = <void*>func             # <<<<<<<<<<<<<<
 *         hb_draw_funcs_set_close_path_func(
 *             self._hb_drawfuncs, func_p, user_data_p, NULL)
 */
    __pyx_v_user_data_p = ((void *)__pyx_v_func);
  }
  __pyx_L3:;

  /* "uharfbuzz/_harfbuzz.pyx":2278
 *             assert user_data is None, "Pass draw_state to Font.draw_glyph"
 *             user_data_p = <void*>func
 *         hb_draw_funcs_set_close_path_func(             # <<<<<<<<<<<<<<
 *             self._hb_drawfuncs, func_p, user_data_p, NULL)
 * 
 */
  hb_draw_funcs_set_close_path_func(__pyx_v_self->_hb_drawfuncs, __pyx_v_func_p, __pyx_v_user_data_p, NULL);

  /* "uharfbuzz/_harfbuzz.pyx":2259
 *             self._hb_drawfuncs, func_p, user_data_p, NULL)
 * 
 *     def set_close_path_func(self,             # <<<<<<<<<<<<<<
 *                             func: Callable[[
 *                                 object
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.DrawFuncs.set_close_path_func", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_19__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_9DrawFuncs_19__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_19__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_19__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_9DrawFuncs_18__reduce_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9DrawFuncs_18__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.DrawFuncs.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_21__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_9DrawFuncs_21__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_21__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_21__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.DrawFuncs.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_9DrawFuncs_20__setstate_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_9DrawFuncs_20__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.DrawFuncs.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2285
 *     cdef unsigned int _num
 * 
 *     def __cinit__(self, num_nodes):             # <<<<<<<<<<<<<<
 *         self._hb_obj_list = <hb_object_t*>calloc(num_nodes, sizeof(hb_object_t))
 *         if self._hb_obj_list == NULL:
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_8HBObject_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_8HBObject_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_num_nodes = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_num_nodes,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_num_nodes)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2285, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(0, 2285, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_num_nodes = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2285, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.HBObject.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_8HBObject___cinit__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_HBObject *)__pyx_v_self), __pyx_v_num_nodes);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_8HBObject___cinit__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_HBObject *__pyx_v_self, PyObject *__pyx_v_num_nodes) {
  int __pyx_r;
  size_t __pyx_t_1;
  int __pyx_t_2;
  unsigned int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2286
 * 
 *     def __cinit__(self, num_nodes):
 *         self._hb_obj_list = <hb_object_t*>calloc(num_nodes, sizeof(hb_object_t))             # <<<<<<<<<<<<<<
 *         if self._hb_obj_list == NULL:
 *             raise MemoryError()
 */
  __pyx_t_1 = __Pyx_PyInt_As_size_t(__pyx_v_num_nodes); if (unlikely((__pyx_t_1 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2286, __pyx_L1_error)
  __pyx_v_self->_hb_obj_list = ((hb_object_t *)calloc(__pyx_t_1, (sizeof(hb_object_t))));

  /* "uharfbuzz/_harfbuzz.pyx":2287
 *     def __cinit__(self, num_nodes):
 *         self._hb_obj_list = <hb_object_t*>calloc(num_nodes, sizeof(hb_object_t))
 *         if self._hb_obj_list == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         self._num = num_nodes
 */
  __pyx_t_2 = (__pyx_v_self->_hb_obj_list == NULL);
  if (unlikely(__pyx_t_2)) {

    /* "uharfbuzz/_harfbuzz.pyx":2288
 *         self._hb_obj_list = <hb_object_t*>calloc(num_nodes, sizeof(hb_object_t))
 *         if self._hb_obj_list == NULL:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 *         self._num = num_nodes
 * 
 */
    PyErr_NoMemory(); __PYX_ERR(0, 2288, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":2287
 *     def __cinit__(self, num_nodes):
 *         self._hb_obj_list = <hb_object_t*>calloc(num_nodes, sizeof(hb_object_t))
 *         if self._hb_obj_list == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 *         self._num = num_nodes
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2289
 *         if self._hb_obj_list == NULL:
 *             raise MemoryError()
 *         self._num = num_nodes             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_t_3 = __Pyx_PyInt_As_unsigned_int(__pyx_v_num_nodes); if (unlikely((__pyx_t_3 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2289, __pyx_L1_error)
  __pyx_v_self->_num = __pyx_t_3;

  /* "uharfbuzz/_harfbuzz.pyx":2285
 *     cdef unsigned int _num
 * 
 *     def __cinit__(self, num_nodes):             # <<<<<<<<<<<<<<
 *         self._hb_obj_list = <hb_object_t*>calloc(num_nodes, sizeof(hb_object_t))
 *         if self._hb_obj_list == NULL:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.HBObject.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2291
 *         self._num = num_nodes
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self._hb_obj_list != NULL:
 *             for i in range(self._num):
 */

/* Python wrapper */
static void __pyx_pw_9uharfbuzz_9_harfbuzz_8HBObject_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_9uharfbuzz_9_harfbuzz_8HBObject_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_9uharfbuzz_9_harfbuzz_8HBObject_2__dealloc__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_HBObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_9uharfbuzz_9_harfbuzz_8HBObject_2__dealloc__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_HBObject *__pyx_v_self) {
  unsigned int __pyx_v_i;
  int __pyx_t_1;
  unsigned int __pyx_t_2;
  unsigned int __pyx_t_3;
  unsigned int __pyx_t_4;

  /* "uharfbuzz/_harfbuzz.pyx":2292
 * 
 *     def __dealloc__(self):
 *         if self._hb_obj_list != NULL:             # <<<<<<<<<<<<<<
 *             for i in range(self._num):
 *                 if self._hb_obj_list[i].real_links != NULL:
 */
  __pyx_t_1 = (__pyx_v_self->_hb_obj_list != NULL);
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":2293
 *     def __dealloc__(self):
 *         if self._hb_obj_list != NULL:
 *             for i in range(self._num):             # <<<<<<<<<<<<<<
 *                 if self._hb_obj_list[i].real_links != NULL:
 *                     free(self._hb_obj_list[i].real_links)
 */
    __pyx_t_2 = __pyx_v_self->_num;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "uharfbuzz/_harfbuzz.pyx":2294
 *         if self._hb_obj_list != NULL:
 *             for i in range(self._num):
 *                 if self._hb_obj_list[i].real_links != NULL:             # <<<<<<<<<<<<<<
 *                     free(self._hb_obj_list[i].real_links)
 *                 if self._hb_obj_list[i].virtual_links != NULL:
 */
      __pyx_t_1 = ((__pyx_v_self->_hb_obj_list[__pyx_v_i]).real_links != NULL);
      if (__pyx_t_1) {

        /* "uharfbuzz/_harfbuzz.pyx":2295
 *             for i in range(self._num):
 *                 if self._hb_obj_list[i].real_links != NULL:
 *                     free(self._hb_obj_list[i].real_links)             # <<<<<<<<<<<<<<
 *                 if self._hb_obj_list[i].virtual_links != NULL:
 *                     free(self._hb_obj_list[i].virtual_links)
 */
        free((__pyx_v_self->_hb_obj_list[__pyx_v_i]).real_links);

        /* "uharfbuzz/_harfbuzz.pyx":2294
 *         if self._hb_obj_list != NULL:
 *             for i in range(self._num):
 *                 if self._hb_obj_list[i].real_links != NULL:             # <<<<<<<<<<<<<<
 *                     free(self._hb_obj_list[i].real_links)
 *                 if self._hb_obj_list[i].virtual_links != NULL:
 */
      }

      /* "uharfbuzz/_harfbuzz.pyx":2296
 *                 if self._hb_obj_list[i].real_links != NULL:
 *                     free(self._hb_obj_list[i].real_links)
 *                 if self._hb_obj_list[i].virtual_links != NULL:             # <<<<<<<<<<<<<<
 *                     free(self._hb_obj_list[i].virtual_links)
 *             free(self._hb_obj_list)
 */
      __pyx_t_1 = ((__pyx_v_self->_hb_obj_list[__pyx_v_i]).virtual_links != NULL);
      if (__pyx_t_1) {

        /* "uharfbuzz/_harfbuzz.pyx":2297
 *                     free(self._hb_obj_list[i].real_links)
 *                 if self._hb_obj_list[i].virtual_links != NULL:
 *                     free(self._hb_obj_list[i].virtual_links)             # <<<<<<<<<<<<<<
 *             free(self._hb_obj_list)
 * 
 */
        free((__pyx_v_self->_hb_obj_list[__pyx_v_i]).virtual_links);

        /* "uharfbuzz/_harfbuzz.pyx":2296
 *                 if self._hb_obj_list[i].real_links != NULL:
 *                     free(self._hb_obj_list[i].real_links)
 *                 if self._hb_obj_list[i].virtual_links != NULL:             # <<<<<<<<<<<<<<
 *                     free(self._hb_obj_list[i].virtual_links)
 *             free(self._hb_obj_list)
 */
      }
    }

    /* "uharfbuzz/_harfbuzz.pyx":2298
 *                 if self._hb_obj_list[i].virtual_links != NULL:
 *                     free(self._hb_obj_list[i].virtual_links)
 *             free(self._hb_obj_list)             # <<<<<<<<<<<<<<
 * 
 *     cdef update_obj_length(self,
 */
    free(__pyx_v_self->_hb_obj_list);

    /* "uharfbuzz/_harfbuzz.pyx":2292
 * 
 *     def __dealloc__(self):
 *         if self._hb_obj_list != NULL:             # <<<<<<<<<<<<<<
 *             for i in range(self._num):
 *                 if self._hb_obj_list[i].real_links != NULL:
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2291
 *         self._num = num_nodes
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self._hb_obj_list != NULL:
 *             for i in range(self._num):
 */

  /* function exit code */
}

/* "uharfbuzz/_harfbuzz.pyx":2300
 *             free(self._hb_obj_list)
 * 
 *     cdef update_obj_length(self,             # <<<<<<<<<<<<<<
 *                            unsigned int idx,
 *                            char* head,
 */

static PyObject *__pyx_f_9uharfbuzz_9_harfbuzz_8HBObject_update_obj_length(struct __pyx_obj_9uharfbuzz_9_harfbuzz_HBObject *__pyx_v_self, unsigned int __pyx_v_idx, char *__pyx_v_head, char *__pyx_v_tail) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("update_obj_length", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2304
 *                            char* head,
 *                            char* tail):
 *         self._hb_obj_list[idx].head = head             # <<<<<<<<<<<<<<
 *         self._hb_obj_list[idx].tail = tail
 * 
 */
  (__pyx_v_self->_hb_obj_list[__pyx_v_idx]).head = __pyx_v_head;

  /* "uharfbuzz/_harfbuzz.pyx":2305
 *                            char* tail):
 *         self._hb_obj_list[idx].head = head
 *         self._hb_obj_list[idx].tail = tail             # <<<<<<<<<<<<<<
 * 
 *     cdef hb_link_t* create_links(self, unsigned int idx,
 */
  (__pyx_v_self->_hb_obj_list[__pyx_v_idx]).tail = __pyx_v_tail;

  /* "uharfbuzz/_harfbuzz.pyx":2300
 *             free(self._hb_obj_list)
 * 
 *     cdef update_obj_length(self,             # <<<<<<<<<<<<<<
 *                            unsigned int idx,
 *                            char* head,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2307
 *         self._hb_obj_list[idx].tail = tail
 * 
 *     cdef hb_link_t* create_links(self, unsigned int idx,             # <<<<<<<<<<<<<<
 *                                  unsigned int link_num,
 *                                  bint is_real_link):
 */

static hb_link_t *__pyx_f_9uharfbuzz_9_harfbuzz_8HBObject_create_links(struct __pyx_obj_9uharfbuzz_9_harfbuzz_HBObject *__pyx_v_self, unsigned int __pyx_v_idx, unsigned int __pyx_v_link_num, int __pyx_v_is_real_link) {
  hb_link_t *__pyx_v_p;
  hb_link_t *__pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2310
 *                                  unsigned int link_num,
 *                                  bint is_real_link):
 *         if link_num == 0:             # <<<<<<<<<<<<<<
 *             return NULL
 * 
 */
  __pyx_t_1 = (__pyx_v_link_num == 0);
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":2311
 *                                  bint is_real_link):
 *         if link_num == 0:
 *             return NULL             # <<<<<<<<<<<<<<
 * 
 *         cdef hb_link_t* p = <hb_link_t*>calloc(link_num, sizeof(hb_link_t))
 */
    __pyx_r = NULL;
    goto __pyx_L0;

    /* "uharfbuzz/_harfbuzz.pyx":2310
 *                                  unsigned int link_num,
 *                                  bint is_real_link):
 *         if link_num == 0:             # <<<<<<<<<<<<<<
 *             return NULL
 * 
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2313
 *             return NULL
 * 
 *         cdef hb_link_t* p = <hb_link_t*>calloc(link_num, sizeof(hb_link_t))             # <<<<<<<<<<<<<<
 *         if p == NULL:
 *             raise MemoryError()
 */
  __pyx_v_p = ((hb_link_t *)calloc(__pyx_v_link_num, (sizeof(hb_link_t))));

  /* "uharfbuzz/_harfbuzz.pyx":2314
 * 
 *         cdef hb_link_t* p = <hb_link_t*>calloc(link_num, sizeof(hb_link_t))
 *         if p == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_1 = (__pyx_v_p == NULL);
  if (unlikely(__pyx_t_1)) {

    /* "uharfbuzz/_harfbuzz.pyx":2315
 *         cdef hb_link_t* p = <hb_link_t*>calloc(link_num, sizeof(hb_link_t))
 *         if p == NULL:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *         if is_real_link:
 */
    PyErr_NoMemory(); __PYX_ERR(0, 2315, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":2314
 * 
 *         cdef hb_link_t* p = <hb_link_t*>calloc(link_num, sizeof(hb_link_t))
 *         if p == NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2317
 *             raise MemoryError()
 * 
 *         if is_real_link:             # <<<<<<<<<<<<<<
 *             self._hb_obj_list[idx].num_real_links = link_num
 *             self._hb_obj_list[idx].real_links = p
 */
  if (__pyx_v_is_real_link) {

    /* "uharfbuzz/_harfbuzz.pyx":2318
 * 
 *         if is_real_link:
 *             self._hb_obj_list[idx].num_real_links = link_num             # <<<<<<<<<<<<<<
 *             self._hb_obj_list[idx].real_links = p
 *         else:
 */
    (__pyx_v_self->_hb_obj_list[__pyx_v_idx]).num_real_links = __pyx_v_link_num;

    /* "uharfbuzz/_harfbuzz.pyx":2319
 *         if is_real_link:
 *             self._hb_obj_list[idx].num_real_links = link_num
 *             self._hb_obj_list[idx].real_links = p             # <<<<<<<<<<<<<<
 *         else:
 *             self._hb_obj_list[idx].num_virtual_links = link_num
 */
    (__pyx_v_self->_hb_obj_list[__pyx_v_idx]).real_links = __pyx_v_p;

    /* "uharfbuzz/_harfbuzz.pyx":2317
 *             raise MemoryError()
 * 
 *         if is_real_link:             # <<<<<<<<<<<<<<
 *             self._hb_obj_list[idx].num_real_links = link_num
 *             self._hb_obj_list[idx].real_links = p
 */
    goto __pyx_L5;
  }

  /* "uharfbuzz/_harfbuzz.pyx":2321
 *             self._hb_obj_list[idx].real_links = p
 *         else:
 *             self._hb_obj_list[idx].num_virtual_links = link_num             # <<<<<<<<<<<<<<
 *             self._hb_obj_list[idx].virtual_links = p
 *         return p
 */
  /*else*/ {
    (__pyx_v_self->_hb_obj_list[__pyx_v_idx]).num_virtual_links = __pyx_v_link_num;

    /* "uharfbuzz/_harfbuzz.pyx":2322
 *         else:
 *             self._hb_obj_list[idx].num_virtual_links = link_num
 *             self._hb_obj_list[idx].virtual_links = p             # <<<<<<<<<<<<<<
 *         return p
 * 
 */
    (__pyx_v_self->_hb_obj_list[__pyx_v_idx]).virtual_links = __pyx_v_p;
  }
  __pyx_L5:;

  /* "uharfbuzz/_harfbuzz.pyx":2323
 *             self._hb_obj_list[idx].num_virtual_links = link_num
 *             self._hb_obj_list[idx].virtual_links = p
 *         return p             # <<<<<<<<<<<<<<
 * 
 *     cdef update_links(self, unsigned int idx, bint is_real_link,
 */
  __pyx_r = __pyx_v_p;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2307
 *         self._hb_obj_list[idx].tail = tail
 * 
 *     cdef hb_link_t* create_links(self, unsigned int idx,             # <<<<<<<<<<<<<<
 *                                  unsigned int link_num,
 *                                  bint is_real_link):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.HBObject.create_links", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2325
 *         return p
 * 
 *     cdef update_links(self, unsigned int idx, bint is_real_link,             # <<<<<<<<<<<<<<
 *                       links: List[Tuple[int, int, int]]):
 *         cdef unsigned int num_links = len(links)
 */

static PyObject *__pyx_f_9uharfbuzz_9_harfbuzz_8HBObject_update_links(struct __pyx_obj_9uharfbuzz_9_harfbuzz_HBObject *__pyx_v_self, unsigned int __pyx_v_idx, int __pyx_v_is_real_link, PyObject *__pyx_v_links) {
  unsigned int __pyx_v_num_links;
  hb_link_t *__pyx_v_l;
  unsigned int __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  hb_link_t *__pyx_t_2;
  unsigned int __pyx_t_3;
  unsigned int __pyx_t_4;
  unsigned int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("update_links", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2327
 *     cdef update_links(self, unsigned int idx, bint is_real_link,
 *                       links: List[Tuple[int, int, int]]):
 *         cdef unsigned int num_links = len(links)             # <<<<<<<<<<<<<<
 *         cdef hb_link_t* l = NULL
 *         if is_real_link:
 */
  __pyx_t_1 = __Pyx_PyList_GET_SIZE(__pyx_v_links); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2327, __pyx_L1_error)
  __pyx_v_num_links = __pyx_t_1;

  /* "uharfbuzz/_harfbuzz.pyx":2328
 *                       links: List[Tuple[int, int, int]]):
 *         cdef unsigned int num_links = len(links)
 *         cdef hb_link_t* l = NULL             # <<<<<<<<<<<<<<
 *         if is_real_link:
 *             l = self._hb_obj_list[idx].real_links
 */
  __pyx_v_l = NULL;

  /* "uharfbuzz/_harfbuzz.pyx":2329
 *         cdef unsigned int num_links = len(links)
 *         cdef hb_link_t* l = NULL
 *         if is_real_link:             # <<<<<<<<<<<<<<
 *             l = self._hb_obj_list[idx].real_links
 *         else:
 */
  if (__pyx_v_is_real_link) {

    /* "uharfbuzz/_harfbuzz.pyx":2330
 *         cdef hb_link_t* l = NULL
 *         if is_real_link:
 *             l = self._hb_obj_list[idx].real_links             # <<<<<<<<<<<<<<
 *         else:
 *             l = self._hb_obj_list[idx].virtual_links
 */
    __pyx_t_2 = (__pyx_v_self->_hb_obj_list[__pyx_v_idx]).real_links;
    __pyx_v_l = __pyx_t_2;

    /* "uharfbuzz/_harfbuzz.pyx":2329
 *         cdef unsigned int num_links = len(links)
 *         cdef hb_link_t* l = NULL
 *         if is_real_link:             # <<<<<<<<<<<<<<
 *             l = self._hb_obj_list[idx].real_links
 *         else:
 */
    goto __pyx_L3;
  }

  /* "uharfbuzz/_harfbuzz.pyx":2332
 *             l = self._hb_obj_list[idx].real_links
 *         else:
 *             l = self._hb_obj_list[idx].virtual_links             # <<<<<<<<<<<<<<
 * 
 *         for i in range(num_links):
 */
  /*else*/ {
    __pyx_t_2 = (__pyx_v_self->_hb_obj_list[__pyx_v_idx]).virtual_links;
    __pyx_v_l = __pyx_t_2;
  }
  __pyx_L3:;

  /* "uharfbuzz/_harfbuzz.pyx":2334
 *             l = self._hb_obj_list[idx].virtual_links
 * 
 *         for i in range(num_links):             # <<<<<<<<<<<<<<
 *             l[i].position = links[i][0]
 *             l[i].width = links[i][1]
 */
  __pyx_t_3 = __pyx_v_num_links;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "uharfbuzz/_harfbuzz.pyx":2335
 * 
 *         for i in range(num_links):
 *             l[i].position = links[i][0]             # <<<<<<<<<<<<<<
 *             l[i].width = links[i][1]
 *             l[i].objidx = links[i][2]
 */
    __pyx_t_6 = __Pyx_GetItemInt_List(__pyx_v_links, __pyx_v_i, unsigned int, 0, __Pyx_PyInt_From_unsigned_int, 1, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2335, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_6, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2335, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_8 = __Pyx_PyInt_As_unsigned_int(__pyx_t_7); if (unlikely((__pyx_t_8 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2335, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    (__pyx_v_l[__pyx_v_i]).position = __pyx_t_8;

    /* "uharfbuzz/_harfbuzz.pyx":2336
 *         for i in range(num_links):
 *             l[i].position = links[i][0]
 *             l[i].width = links[i][1]             # <<<<<<<<<<<<<<
 *             l[i].objidx = links[i][2]
 * 
 */
    __pyx_t_7 = __Pyx_GetItemInt_List(__pyx_v_links, __pyx_v_i, unsigned int, 0, __Pyx_PyInt_From_unsigned_int, 1, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2336, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_7, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2336, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_8 = __Pyx_PyInt_As_unsigned_int(__pyx_t_6); if (unlikely((__pyx_t_8 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2336, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    (__pyx_v_l[__pyx_v_i]).width = __pyx_t_8;

    /* "uharfbuzz/_harfbuzz.pyx":2337
 *             l[i].position = links[i][0]
 *             l[i].width = links[i][1]
 *             l[i].objidx = links[i][2]             # <<<<<<<<<<<<<<
 * 
 * class RepackerError(Exception):
 */
    __pyx_t_6 = __Pyx_GetItemInt_List(__pyx_v_links, __pyx_v_i, unsigned int, 0, __Pyx_PyInt_From_unsigned_int, 1, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_GetItemInt(__pyx_t_6, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_8 = __Pyx_PyInt_As_unsigned_int(__pyx_t_7); if (unlikely((__pyx_t_8 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2337, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    (__pyx_v_l[__pyx_v_i]).objidx = __pyx_t_8;
  }

  /* "uharfbuzz/_harfbuzz.pyx":2325
 *         return p
 * 
 *     cdef update_links(self, unsigned int idx, bint is_real_link,             # <<<<<<<<<<<<<<
 *                       links: List[Tuple[int, int, int]]):
 *         cdef unsigned int num_links = len(links)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.HBObject.update_links", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_8HBObject_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_8HBObject_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_8HBObject_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_8HBObject_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_8HBObject_4__reduce_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_HBObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_8HBObject_4__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_HBObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.HBObject.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_8HBObject_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_8HBObject_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_8HBObject_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_8HBObject_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.HBObject.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_8HBObject_6__setstate_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_HBObject *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_8HBObject_6__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_HBObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.HBObject.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2343
 * 
 * 
 * def repack(subtables: List[bytes],             # <<<<<<<<<<<<<<
 *            graphnodes: List[Tuple[List[Tuple[int, int, int]],
 *                             List[Tuple[int, int, int]]
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_77repack(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_77repack = {"repack", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_77repack, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_77repack(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_subtables = 0;
  PyObject *__pyx_v_graphnodes = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("repack (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_subtables,&__pyx_n_s_graphnodes,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_subtables)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2343, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_graphnodes)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2343, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("repack", 1, 2, 2, 1); __PYX_ERR(0, 2343, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "repack") < 0)) __PYX_ERR(0, 2343, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_subtables = ((PyObject*)values[0]);
    __pyx_v_graphnodes = ((PyObject*)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("repack", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 2343, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.repack", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_subtables), (&PyList_Type), 0, "subtables", 1))) __PYX_ERR(0, 2343, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_graphnodes), (&PyList_Type), 0, "graphnodes", 1))) __PYX_ERR(0, 2344, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_76repack(__pyx_self, __pyx_v_subtables, __pyx_v_graphnodes);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_76repack(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_subtables, PyObject *__pyx_v_graphnodes) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("repack", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2347
 *                             List[Tuple[int, int, int]]
 *                             ]]) -> bytes:
 *     return repack_with_tag("", subtables, graphnodes)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_repack_with_tag); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_kp_u__15, __pyx_v_subtables, __pyx_v_graphnodes};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 3+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2347, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_1))) __PYX_ERR(0, 2347, __pyx_L1_error)
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2343
 * 
 * 
 * def repack(subtables: List[bytes],             # <<<<<<<<<<<<<<
 *            graphnodes: List[Tuple[List[Tuple[int, int, int]],
 *                             List[Tuple[int, int, int]]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.repack", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2350
 * 
 * 
 * def repack_with_tag(tag: str,             # <<<<<<<<<<<<<<
 *                     subtables: List[bytes],
 *                     graphnodes: List[Tuple[List[Tuple[int, int, int]],
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_79repack_with_tag(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_9uharfbuzz_9_harfbuzz_78repack_with_tag, "The whole table is represented as a Graph\n       and the input graphnodes is a flat list of subtables\n       with each node(subtable) represented by a tuple of 2\n       list: real_link list and virtual_link list.\n\n       A link(egde) is an offset link between parent table and\n       child table. It's represented in the format of a tuple:\n       (posiiton: int, width: int, objidx: int):\n\n       - position: means relative position of the offset field\n         in bytes from the beginning of the subtable's C struct.\n         e.g:\n         a GSUB header struct in C looks like below:\n\n         uint16 majorVersion\n         uint16 minorVersion\n         offset16 scriptListOffset\n         offset16 featureListOffset\n         offset16 lookupListOffset\n         And the position for scriptListOffset is 4\n         which is calculated from (16+16)/8\n       - width: size of the offset:\n         e.g 2 for offset16 and 4 for offset32\n       - objidx: objidx is the index of the subtable\n         in graph/tree generated by postorder traversal\n\n       There're 2 types of links:\n       - real_link represents an real offset field in the parent table\n       - virtual_link is not real offset link, it specifies an ordering\n         constraint that harfbuzz packing must follow.\n         A virtual link would have 0 width, 0 position, and a real objidx.\n         e.g:\n         if Node A has a virtual link with objidx b(corresponding Node is B),\n         then that means Node B is always packed after Node A in the final\n         serialized order\n         ");
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_79repack_with_tag = {"repack_with_tag", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_79repack_with_tag, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_9uharfbuzz_9_harfbuzz_78repack_with_tag};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_79repack_with_tag(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_tag = 0;
  PyObject *__pyx_v_subtables = 0;
  PyObject *__pyx_v_graphnodes = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("repack_with_tag (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_tag,&__pyx_n_s_subtables,&__pyx_n_s_graphnodes,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_tag)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2350, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_subtables)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2350, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("repack_with_tag", 1, 3, 3, 1); __PYX_ERR(0, 2350, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_graphnodes)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2350, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("repack_with_tag", 1, 3, 3, 2); __PYX_ERR(0, 2350, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "repack_with_tag") < 0)) __PYX_ERR(0, 2350, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v_tag = ((PyObject*)values[0]);
    __pyx_v_subtables = ((PyObject*)values[1]);
    __pyx_v_graphnodes = ((PyObject*)values[2]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("repack_with_tag", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 2350, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.repack_with_tag", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_tag), (&PyUnicode_Type), 0, "tag", 1))) __PYX_ERR(0, 2350, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_subtables), (&PyList_Type), 0, "subtables", 1))) __PYX_ERR(0, 2351, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_graphnodes), (&PyList_Type), 0, "graphnodes", 1))) __PYX_ERR(0, 2352, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_78repack_with_tag(__pyx_self, __pyx_v_tag, __pyx_v_subtables, __pyx_v_graphnodes);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_78repack_with_tag(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_tag, PyObject *__pyx_v_subtables, PyObject *__pyx_v_graphnodes) {
  unsigned int __pyx_v_num_nodes;
  PyObject *__pyx_v_table_bytes = 0;
  char *__pyx_v_table_data;
  unsigned int __pyx_v_head;
  unsigned int __pyx_v_tail;
  hb_link_t *__pyx_v_p;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_HBObject *__pyx_v_obj_list = 0;
  unsigned int __pyx_v_i;
  PyObject *__pyx_v_node = NULL;
  PyObject *__pyx_v_tag_packed = 0;
  char *__pyx_v_cstr;
  hb_blob_t *__pyx_v_packed_blob;
  unsigned int __pyx_v_blob_length;
  const char *__pyx_v_blob_data;
  PyObject *__pyx_v_packed = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  Py_UCS4 __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  char *__pyx_t_7;
  unsigned int __pyx_t_8;
  unsigned int __pyx_t_9;
  unsigned int __pyx_t_10;
  hb_link_t *__pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("repack_with_tag", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2393
 *          """
 * 
 *     if len(subtables) != len(graphnodes):             # <<<<<<<<<<<<<<
 *         raise ValueError(
 *             f"Input num of subtables({len(subtables)}) != num of graph nodes({len(graphnodes)})"
 */
  __pyx_t_1 = __Pyx_PyList_GET_SIZE(__pyx_v_subtables); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2393, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyList_GET_SIZE(__pyx_v_graphnodes); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2393, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_1 != __pyx_t_2);
  if (unlikely(__pyx_t_3)) {

    /* "uharfbuzz/_harfbuzz.pyx":2395
 *     if len(subtables) != len(graphnodes):
 *         raise ValueError(
 *             f"Input num of subtables({len(subtables)}) != num of graph nodes({len(graphnodes)})"             # <<<<<<<<<<<<<<
 *         )
 *     cdef:
 */
    __pyx_t_4 = PyTuple_New(5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2395, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_5 = 127;
    __Pyx_INCREF(__pyx_kp_u_Input_num_of_subtables);
    __pyx_t_2 += 23;
    __Pyx_GIVEREF(__pyx_kp_u_Input_num_of_subtables);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_kp_u_Input_num_of_subtables);
    __pyx_t_1 = __Pyx_PyList_GET_SIZE(__pyx_v_subtables); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2395, __pyx_L1_error)
    __pyx_t_6 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_t_1, 0, ' ', 'd'); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2395, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u_num_of_graph_nodes);
    __pyx_t_2 += 24;
    __Pyx_GIVEREF(__pyx_kp_u_num_of_graph_nodes);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_kp_u_num_of_graph_nodes);
    __pyx_t_1 = __Pyx_PyList_GET_SIZE(__pyx_v_graphnodes); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2395, __pyx_L1_error)
    __pyx_t_6 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_t_1, 0, ' ', 'd'); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2395, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_kp_u__16);
    __pyx_t_2 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__16);
    PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_kp_u__16);
    __pyx_t_6 = __Pyx_PyUnicode_Join(__pyx_t_4, 5, __pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2395, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "uharfbuzz/_harfbuzz.pyx":2394
 * 
 *     if len(subtables) != len(graphnodes):
 *         raise ValueError(             # <<<<<<<<<<<<<<
 *             f"Input num of subtables({len(subtables)}) != num of graph nodes({len(graphnodes)})"
 *         )
 */
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 2394, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":2393
 *          """
 * 
 *     if len(subtables) != len(graphnodes):             # <<<<<<<<<<<<<<
 *         raise ValueError(
 *             f"Input num of subtables({len(subtables)}) != num of graph nodes({len(graphnodes)})"
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2398
 *         )
 *     cdef:
 *         unsigned int num_nodes = len(subtables)             # <<<<<<<<<<<<<<
 *         bytes table_bytes = b''.join(subtables)
 *         char* table_data = table_bytes
 */
  __pyx_t_2 = __Pyx_PyList_GET_SIZE(__pyx_v_subtables); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2398, __pyx_L1_error)
  __pyx_v_num_nodes = __pyx_t_2;

  /* "uharfbuzz/_harfbuzz.pyx":2399
 *     cdef:
 *         unsigned int num_nodes = len(subtables)
 *         bytes table_bytes = b''.join(subtables)             # <<<<<<<<<<<<<<
 *         char* table_data = table_bytes
 *         unsigned int head = 0, tail = 0
 */
  __pyx_t_4 = __Pyx_PyBytes_Join(__pyx_kp_b__15, __pyx_v_subtables); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("bytes", __pyx_t_4))) __PYX_ERR(0, 2399, __pyx_L1_error)
  __pyx_v_table_bytes = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2400
 *         unsigned int num_nodes = len(subtables)
 *         bytes table_bytes = b''.join(subtables)
 *         char* table_data = table_bytes             # <<<<<<<<<<<<<<
 *         unsigned int head = 0, tail = 0
 *         hb_link_t* p = NULL
 */
  if (unlikely(__pyx_v_table_bytes == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 2400, __pyx_L1_error)
  }
  __pyx_t_7 = __Pyx_PyBytes_AsWritableString(__pyx_v_table_bytes); if (unlikely((!__pyx_t_7) && PyErr_Occurred())) __PYX_ERR(0, 2400, __pyx_L1_error)
  __pyx_v_table_data = __pyx_t_7;

  /* "uharfbuzz/_harfbuzz.pyx":2401
 *         bytes table_bytes = b''.join(subtables)
 *         char* table_data = table_bytes
 *         unsigned int head = 0, tail = 0             # <<<<<<<<<<<<<<
 *         hb_link_t* p = NULL
 *         HBObject obj_list = HBObject(num_nodes)
 */
  __pyx_v_head = 0;
  __pyx_v_tail = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2402
 *         char* table_data = table_bytes
 *         unsigned int head = 0, tail = 0
 *         hb_link_t* p = NULL             # <<<<<<<<<<<<<<
 *         HBObject obj_list = HBObject(num_nodes)
 *     for i in range(num_nodes):
 */
  __pyx_v_p = NULL;

  /* "uharfbuzz/_harfbuzz.pyx":2403
 *         unsigned int head = 0, tail = 0
 *         hb_link_t* p = NULL
 *         HBObject obj_list = HBObject(num_nodes)             # <<<<<<<<<<<<<<
 *     for i in range(num_nodes):
 *         tail += len(subtables[i])
 */
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_int(__pyx_v_num_nodes); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_HBObject), __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_obj_list = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_HBObject *)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2404
 *         hb_link_t* p = NULL
 *         HBObject obj_list = HBObject(num_nodes)
 *     for i in range(num_nodes):             # <<<<<<<<<<<<<<
 *         tail += len(subtables[i])
 *         obj_list.update_obj_length(i, table_data + head, table_data + tail)
 */
  __pyx_t_8 = __pyx_v_num_nodes;
  __pyx_t_9 = __pyx_t_8;
  for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
    __pyx_v_i = __pyx_t_10;

    /* "uharfbuzz/_harfbuzz.pyx":2405
 *         HBObject obj_list = HBObject(num_nodes)
 *     for i in range(num_nodes):
 *         tail += len(subtables[i])             # <<<<<<<<<<<<<<
 *         obj_list.update_obj_length(i, table_data + head, table_data + tail)
 *         head = tail
 */
    __pyx_t_6 = __Pyx_GetItemInt_List(__pyx_v_subtables, __pyx_v_i, unsigned int, 0, __Pyx_PyInt_From_unsigned_int, 1, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = PyObject_Length(__pyx_t_6); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2405, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_tail = (__pyx_v_tail + __pyx_t_2);

    /* "uharfbuzz/_harfbuzz.pyx":2406
 *     for i in range(num_nodes):
 *         tail += len(subtables[i])
 *         obj_list.update_obj_length(i, table_data + head, table_data + tail)             # <<<<<<<<<<<<<<
 *         head = tail
 * 
 */
    __pyx_t_6 = ((struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_HBObject *)__pyx_v_obj_list->__pyx_vtab)->update_obj_length(__pyx_v_obj_list, __pyx_v_i, (__pyx_v_table_data + __pyx_v_head), (__pyx_v_table_data + __pyx_v_tail)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "uharfbuzz/_harfbuzz.pyx":2407
 *         tail += len(subtables[i])
 *         obj_list.update_obj_length(i, table_data + head, table_data + tail)
 *         head = tail             # <<<<<<<<<<<<<<
 * 
 *         node = graphnodes[i]
 */
    __pyx_v_head = __pyx_v_tail;

    /* "uharfbuzz/_harfbuzz.pyx":2409
 *         head = tail
 * 
 *         node = graphnodes[i]             # <<<<<<<<<<<<<<
 *         # real_links
 *         p = obj_list.create_links(i, len(node[0]), True)
 */
    __pyx_t_6 = __Pyx_GetItemInt_List(__pyx_v_graphnodes, __pyx_v_i, unsigned int, 0, __Pyx_PyInt_From_unsigned_int, 1, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2409, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF_SET(__pyx_v_node, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "uharfbuzz/_harfbuzz.pyx":2411
 *         node = graphnodes[i]
 *         # real_links
 *         p = obj_list.create_links(i, len(node[0]), True)             # <<<<<<<<<<<<<<
 *         if p != NULL:
 *             obj_list.update_links(i, True, node[0])
 */
    __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_node, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2411, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = PyObject_Length(__pyx_t_6); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2411, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_11 = ((struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_HBObject *)__pyx_v_obj_list->__pyx_vtab)->create_links(__pyx_v_obj_list, __pyx_v_i, __pyx_t_2, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2411, __pyx_L1_error)
    __pyx_v_p = __pyx_t_11;

    /* "uharfbuzz/_harfbuzz.pyx":2412
 *         # real_links
 *         p = obj_list.create_links(i, len(node[0]), True)
 *         if p != NULL:             # <<<<<<<<<<<<<<
 *             obj_list.update_links(i, True, node[0])
 *         # virtual_links
 */
    __pyx_t_3 = (__pyx_v_p != NULL);
    if (__pyx_t_3) {

      /* "uharfbuzz/_harfbuzz.pyx":2413
 *         p = obj_list.create_links(i, len(node[0]), True)
 *         if p != NULL:
 *             obj_list.update_links(i, True, node[0])             # <<<<<<<<<<<<<<
 *         # virtual_links
 *         p = obj_list.create_links(i, len(node[1]), False)
 */
      __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_node, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2413, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (!(likely(PyList_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_6))) __PYX_ERR(0, 2413, __pyx_L1_error)
      __pyx_t_4 = ((struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_HBObject *)__pyx_v_obj_list->__pyx_vtab)->update_links(__pyx_v_obj_list, __pyx_v_i, 1, ((PyObject*)__pyx_t_6)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2413, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "uharfbuzz/_harfbuzz.pyx":2412
 *         # real_links
 *         p = obj_list.create_links(i, len(node[0]), True)
 *         if p != NULL:             # <<<<<<<<<<<<<<
 *             obj_list.update_links(i, True, node[0])
 *         # virtual_links
 */
    }

    /* "uharfbuzz/_harfbuzz.pyx":2415
 *             obj_list.update_links(i, True, node[0])
 *         # virtual_links
 *         p = obj_list.create_links(i, len(node[1]), False)             # <<<<<<<<<<<<<<
 *         if p != NULL:
 *             obj_list.update_links(i, False, node[1])
 */
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_node, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2415, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = PyObject_Length(__pyx_t_4); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 2415, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_11 = ((struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_HBObject *)__pyx_v_obj_list->__pyx_vtab)->create_links(__pyx_v_obj_list, __pyx_v_i, __pyx_t_2, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2415, __pyx_L1_error)
    __pyx_v_p = __pyx_t_11;

    /* "uharfbuzz/_harfbuzz.pyx":2416
 *         # virtual_links
 *         p = obj_list.create_links(i, len(node[1]), False)
 *         if p != NULL:             # <<<<<<<<<<<<<<
 *             obj_list.update_links(i, False, node[1])
 * 
 */
    __pyx_t_3 = (__pyx_v_p != NULL);
    if (__pyx_t_3) {

      /* "uharfbuzz/_harfbuzz.pyx":2417
 *         p = obj_list.create_links(i, len(node[1]), False)
 *         if p != NULL:
 *             obj_list.update_links(i, False, node[1])             # <<<<<<<<<<<<<<
 * 
 *     cdef bytes tag_packed = tag.encode()
 */
      __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_node, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2417, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (!(likely(PyList_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_4))) __PYX_ERR(0, 2417, __pyx_L1_error)
      __pyx_t_6 = ((struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_HBObject *)__pyx_v_obj_list->__pyx_vtab)->update_links(__pyx_v_obj_list, __pyx_v_i, 0, ((PyObject*)__pyx_t_4)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2417, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "uharfbuzz/_harfbuzz.pyx":2416
 *         # virtual_links
 *         p = obj_list.create_links(i, len(node[1]), False)
 *         if p != NULL:             # <<<<<<<<<<<<<<
 *             obj_list.update_links(i, False, node[1])
 * 
 */
    }
  }

  /* "uharfbuzz/_harfbuzz.pyx":2419
 *             obj_list.update_links(i, False, node[1])
 * 
 *     cdef bytes tag_packed = tag.encode()             # <<<<<<<<<<<<<<
 *     cdef char* cstr = tag_packed
 *     cdef hb_blob_t* packed_blob = hb_subset_repack_or_fail(hb_tag_from_string(cstr, -1),
 */
  __pyx_t_6 = PyUnicode_AsEncodedString(__pyx_v_tag, NULL, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2419, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_tag_packed = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2420
 * 
 *     cdef bytes tag_packed = tag.encode()
 *     cdef char* cstr = tag_packed             # <<<<<<<<<<<<<<
 *     cdef hb_blob_t* packed_blob = hb_subset_repack_or_fail(hb_tag_from_string(cstr, -1),
 * 							   obj_list._hb_obj_list,
 */
  if (unlikely(__pyx_v_tag_packed == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 2420, __pyx_L1_error)
  }
  __pyx_t_7 = __Pyx_PyBytes_AsWritableString(__pyx_v_tag_packed); if (unlikely((!__pyx_t_7) && PyErr_Occurred())) __PYX_ERR(0, 2420, __pyx_L1_error)
  __pyx_v_cstr = __pyx_t_7;

  /* "uharfbuzz/_harfbuzz.pyx":2421
 *     cdef bytes tag_packed = tag.encode()
 *     cdef char* cstr = tag_packed
 *     cdef hb_blob_t* packed_blob = hb_subset_repack_or_fail(hb_tag_from_string(cstr, -1),             # <<<<<<<<<<<<<<
 * 							   obj_list._hb_obj_list,
 * 							   num_nodes)
 */
  __pyx_v_packed_blob = hb_subset_repack_or_fail(hb_tag_from_string(__pyx_v_cstr, -1), __pyx_v_obj_list->_hb_obj_list, __pyx_v_num_nodes);

  /* "uharfbuzz/_harfbuzz.pyx":2424
 * 							   obj_list._hb_obj_list,
 * 							   num_nodes)
 *     if packed_blob == NULL:             # <<<<<<<<<<<<<<
 *         raise RepackerError()
 * 
 */
  __pyx_t_3 = (__pyx_v_packed_blob == NULL);
  if (unlikely(__pyx_t_3)) {

    /* "uharfbuzz/_harfbuzz.pyx":2425
 * 							   num_nodes)
 *     if packed_blob == NULL:
 *         raise RepackerError()             # <<<<<<<<<<<<<<
 * 
 *     cdef unsigned int blob_length
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_RepackerError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2425, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_12 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_12, NULL};
      __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2425, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 2425, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":2424
 * 							   obj_list._hb_obj_list,
 * 							   num_nodes)
 *     if packed_blob == NULL:             # <<<<<<<<<<<<<<
 *         raise RepackerError()
 * 
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2428
 * 
 *     cdef unsigned int blob_length
 *     cdef const_char* blob_data = hb_blob_get_data(packed_blob, &blob_length)             # <<<<<<<<<<<<<<
 *     cdef bytes packed = blob_data[:blob_length]
 * 
 */
  __pyx_v_blob_data = hb_blob_get_data(__pyx_v_packed_blob, (&__pyx_v_blob_length));

  /* "uharfbuzz/_harfbuzz.pyx":2429
 *     cdef unsigned int blob_length
 *     cdef const_char* blob_data = hb_blob_get_data(packed_blob, &blob_length)
 *     cdef bytes packed = blob_data[:blob_length]             # <<<<<<<<<<<<<<
 * 
 *     hb_blob_destroy(packed_blob)
 */
  __pyx_t_6 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_blob_data + 0, __pyx_v_blob_length - 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_packed = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2431
 *     cdef bytes packed = blob_data[:blob_length]
 * 
 *     hb_blob_destroy(packed_blob)             # <<<<<<<<<<<<<<
 *     return packed
 * 
 */
  hb_blob_destroy(__pyx_v_packed_blob);

  /* "uharfbuzz/_harfbuzz.pyx":2432
 * 
 *     hb_blob_destroy(packed_blob)
 *     return packed             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_packed);
  __pyx_r = __pyx_v_packed;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2350
 * 
 * 
 * def repack_with_tag(tag: str,             # <<<<<<<<<<<<<<
 *                     subtables: List[bytes],
 *                     graphnodes: List[Tuple[List[Tuple[int, int, int]],
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.repack_with_tag", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_table_bytes);
  __Pyx_XDECREF((PyObject *)__pyx_v_obj_list);
  __Pyx_XDECREF(__pyx_v_node);
  __Pyx_XDECREF(__pyx_v_tag_packed);
  __Pyx_XDECREF(__pyx_v_packed);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2435
 * 
 * 
 * def subset_preprocess(face: Face) -> Face:             # <<<<<<<<<<<<<<
 *     new_face = hb_subset_preprocess(face._hb_face)
 *     return Face.from_ptr(new_face)
 */

/* Python wrapper */
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_pw_9uharfbuzz_9_harfbuzz_81subset_preprocess(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_81subset_preprocess = {"subset_preprocess", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_81subset_preprocess, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_pw_9uharfbuzz_9_harfbuzz_81subset_preprocess(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("subset_preprocess (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_face,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_face)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2435, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "subset_preprocess") < 0)) __PYX_ERR(0, 2435, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("subset_preprocess", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2435, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.subset_preprocess", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_face), __pyx_ptype_9uharfbuzz_9_harfbuzz_Face, 0, "face", 0))) __PYX_ERR(0, 2435, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_80subset_preprocess(__pyx_self, __pyx_v_face);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_pf_9uharfbuzz_9_harfbuzz_80subset_preprocess(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face) {
  hb_face_t *__pyx_v_new_face;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("subset_preprocess", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2436
 * 
 * def subset_preprocess(face: Face) -> Face:
 *     new_face = hb_subset_preprocess(face._hb_face)             # <<<<<<<<<<<<<<
 *     return Face.from_ptr(new_face)
 * 
 */
  __pyx_v_new_face = hb_subset_preprocess(__pyx_v_face->_hb_face);

  /* "uharfbuzz/_harfbuzz.pyx":2437
 * def subset_preprocess(face: Face) -> Face:
 *     new_face = hb_subset_preprocess(face._hb_face)
 *     return Face.from_ptr(new_face)             # <<<<<<<<<<<<<<
 * 
 * def subset(face: Face, input: SubsetInput) -> Face:
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_9uharfbuzz_9_harfbuzz_4Face_from_ptr(__pyx_v_new_face)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2437, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2435
 * 
 * 
 * def subset_preprocess(face: Face) -> Face:             # <<<<<<<<<<<<<<
 *     new_face = hb_subset_preprocess(face._hb_face)
 *     return Face.from_ptr(new_face)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.subset_preprocess", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2439
 *     return Face.from_ptr(new_face)
 * 
 * def subset(face: Face, input: SubsetInput) -> Face:             # <<<<<<<<<<<<<<
 *     new_face = hb_subset_or_fail(face._hb_face, input._hb_input)
 *     if new_face == NULL:
 */

/* Python wrapper */
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_pw_9uharfbuzz_9_harfbuzz_83subset(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_83subset = {"subset", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_83subset, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_pw_9uharfbuzz_9_harfbuzz_83subset(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face = 0;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_input = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("subset (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_face,&__pyx_n_s_input,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_face)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2439, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_input)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2439, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("subset", 1, 2, 2, 1); __PYX_ERR(0, 2439, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "subset") < 0)) __PYX_ERR(0, 2439, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)values[0]);
    __pyx_v_input = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("subset", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 2439, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.subset", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_face), __pyx_ptype_9uharfbuzz_9_harfbuzz_Face, 0, "face", 0))) __PYX_ERR(0, 2439, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_input), __pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetInput, 0, "input", 0))) __PYX_ERR(0, 2439, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_82subset(__pyx_self, __pyx_v_face, __pyx_v_input);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_pf_9uharfbuzz_9_harfbuzz_82subset(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_input) {
  hb_face_t *__pyx_v_new_face;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("subset", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2440
 * 
 * def subset(face: Face, input: SubsetInput) -> Face:
 *     new_face = hb_subset_or_fail(face._hb_face, input._hb_input)             # <<<<<<<<<<<<<<
 *     if new_face == NULL:
 *         raise RuntimeError("Subsetting failed")
 */
  __pyx_v_new_face = hb_subset_or_fail(__pyx_v_face->_hb_face, __pyx_v_input->_hb_input);

  /* "uharfbuzz/_harfbuzz.pyx":2441
 * def subset(face: Face, input: SubsetInput) -> Face:
 *     new_face = hb_subset_or_fail(face._hb_face, input._hb_input)
 *     if new_face == NULL:             # <<<<<<<<<<<<<<
 *         raise RuntimeError("Subsetting failed")
 *     return Face.from_ptr(new_face)
 */
  __pyx_t_1 = (__pyx_v_new_face == NULL);
  if (unlikely(__pyx_t_1)) {

    /* "uharfbuzz/_harfbuzz.pyx":2442
 *     new_face = hb_subset_or_fail(face._hb_face, input._hb_input)
 *     if new_face == NULL:
 *         raise RuntimeError("Subsetting failed")             # <<<<<<<<<<<<<<
 *     return Face.from_ptr(new_face)
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2442, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 2442, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":2441
 * def subset(face: Face, input: SubsetInput) -> Face:
 *     new_face = hb_subset_or_fail(face._hb_face, input._hb_input)
 *     if new_face == NULL:             # <<<<<<<<<<<<<<
 *         raise RuntimeError("Subsetting failed")
 *     return Face.from_ptr(new_face)
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2443
 *     if new_face == NULL:
 *         raise RuntimeError("Subsetting failed")
 *     return Face.from_ptr(new_face)             # <<<<<<<<<<<<<<
 * 
 * class SubsetInputSets(IntEnum):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_2 = ((PyObject *)__pyx_f_9uharfbuzz_9_harfbuzz_4Face_from_ptr(__pyx_v_new_face)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2443, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2439
 *     return Face.from_ptr(new_face)
 * 
 * def subset(face: Face, input: SubsetInput) -> Face:             # <<<<<<<<<<<<<<
 *     new_face = hb_subset_or_fail(face._hb_face, input._hb_input)
 *     if new_face == NULL:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.subset", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2473
 *     cdef hb_subset_input_t* _hb_input
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._hb_input = hb_subset_input_create_or_fail()
 *         if self._hb_input is NULL:
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 0, 0, __pyx_nargs); return -1;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_VARARGS(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__cinit__", 0))) return -1;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput___cinit__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput___cinit__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2474
 * 
 *     def __cinit__(self):
 *         self._hb_input = hb_subset_input_create_or_fail()             # <<<<<<<<<<<<<<
 *         if self._hb_input is NULL:
 *             raise MemoryError()
 */
  __pyx_v_self->_hb_input = hb_subset_input_create_or_fail();

  /* "uharfbuzz/_harfbuzz.pyx":2475
 *     def __cinit__(self):
 *         self._hb_input = hb_subset_input_create_or_fail()
 *         if self._hb_input is NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_hb_input == NULL);
  if (unlikely(__pyx_t_1)) {

    /* "uharfbuzz/_harfbuzz.pyx":2476
 *         self._hb_input = hb_subset_input_create_or_fail()
 *         if self._hb_input is NULL:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
    PyErr_NoMemory(); __PYX_ERR(0, 2476, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":2475
 *     def __cinit__(self):
 *         self._hb_input = hb_subset_input_create_or_fail()
 *         if self._hb_input is NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2473
 *     cdef hb_subset_input_t* _hb_input
 * 
 *     def __cinit__(self):             # <<<<<<<<<<<<<<
 *         self._hb_input = hb_subset_input_create_or_fail()
 *         if self._hb_input is NULL:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SubsetInput.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2478
 *             raise MemoryError()
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self._hb_input is not NULL:
 *             hb_subset_input_destroy(self._hb_input)
 */

/* Python wrapper */
static void __pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_2__dealloc__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_2__dealloc__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self) {
  int __pyx_t_1;

  /* "uharfbuzz/_harfbuzz.pyx":2479
 * 
 *     def __dealloc__(self):
 *         if self._hb_input is not NULL:             # <<<<<<<<<<<<<<
 *             hb_subset_input_destroy(self._hb_input)
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_hb_input != NULL);
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":2480
 *     def __dealloc__(self):
 *         if self._hb_input is not NULL:
 *             hb_subset_input_destroy(self._hb_input)             # <<<<<<<<<<<<<<
 * 
 *     def subset(self, source: Face) -> Face:
 */
    hb_subset_input_destroy(__pyx_v_self->_hb_input);

    /* "uharfbuzz/_harfbuzz.pyx":2479
 * 
 *     def __dealloc__(self):
 *         if self._hb_input is not NULL:             # <<<<<<<<<<<<<<
 *             hb_subset_input_destroy(self._hb_input)
 * 
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2478
 *             raise MemoryError()
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self._hb_input is not NULL:
 *             hb_subset_input_destroy(self._hb_input)
 */

  /* function exit code */
}

/* "uharfbuzz/_harfbuzz.pyx":2482
 *             hb_subset_input_destroy(self._hb_input)
 * 
 *     def subset(self, source: Face) -> Face:             # <<<<<<<<<<<<<<
 *         return subset(source, self)
 * 
 */

/* Python wrapper */
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_5subset(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_11SubsetInput_5subset = {"subset", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_5subset, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_5subset(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_source = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("subset (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_source,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_source)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2482, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "subset") < 0)) __PYX_ERR(0, 2482, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_source = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("subset", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2482, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SubsetInput.subset", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_source), __pyx_ptype_9uharfbuzz_9_harfbuzz_Face, 0, "source", 0))) __PYX_ERR(0, 2482, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_4subset(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *)__pyx_v_self), __pyx_v_source);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_4subset(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_source) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("subset", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2483
 * 
 *     def subset(self, source: Face) -> Face:
 *         return subset(source, self)             # <<<<<<<<<<<<<<
 * 
 *     def keep_everything(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_subset); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_3, ((PyObject *)__pyx_v_source), ((PyObject *)__pyx_v_self)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 2+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2483, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_9uharfbuzz_9_harfbuzz_Face))))) __PYX_ERR(0, 2483, __pyx_L1_error)
  __pyx_r = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2482
 *             hb_subset_input_destroy(self._hb_input)
 * 
 *     def subset(self, source: Face) -> Face:             # <<<<<<<<<<<<<<
 *         return subset(source, self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SubsetInput.subset", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2485
 *         return subset(source, self)
 * 
 *     def keep_everything(self):             # <<<<<<<<<<<<<<
 *         hb_subset_input_keep_everything(self._hb_input)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_7keep_everything(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_11SubsetInput_7keep_everything = {"keep_everything", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_7keep_everything, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_7keep_everything(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("keep_everything (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("keep_everything", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "keep_everything", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_6keep_everything(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_6keep_everything(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("keep_everything", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2486
 * 
 *     def keep_everything(self):
 *         hb_subset_input_keep_everything(self._hb_input)             # <<<<<<<<<<<<<<
 * 
 *     def pin_axis_to_default(self, face: Face, tag: str) -> bool:
 */
  hb_subset_input_keep_everything(__pyx_v_self->_hb_input);

  /* "uharfbuzz/_harfbuzz.pyx":2485
 *         return subset(source, self)
 * 
 *     def keep_everything(self):             # <<<<<<<<<<<<<<
 *         hb_subset_input_keep_everything(self._hb_input)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2488
 *         hb_subset_input_keep_everything(self._hb_input)
 * 
 *     def pin_axis_to_default(self, face: Face, tag: str) -> bool:             # <<<<<<<<<<<<<<
 *         hb_tag = hb_tag_from_string(tag.encode("ascii"), -1)
 *         return hb_subset_input_pin_axis_to_default(
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_9pin_axis_to_default(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_11SubsetInput_9pin_axis_to_default = {"pin_axis_to_default", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_9pin_axis_to_default, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_9pin_axis_to_default(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face = 0;
  PyObject *__pyx_v_tag = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pin_axis_to_default (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_face,&__pyx_n_s_tag,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_face)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2488, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_tag)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2488, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("pin_axis_to_default", 1, 2, 2, 1); __PYX_ERR(0, 2488, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "pin_axis_to_default") < 0)) __PYX_ERR(0, 2488, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)values[0]);
    __pyx_v_tag = ((PyObject*)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("pin_axis_to_default", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 2488, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SubsetInput.pin_axis_to_default", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_face), __pyx_ptype_9uharfbuzz_9_harfbuzz_Face, 0, "face", 0))) __PYX_ERR(0, 2488, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_tag), (&PyUnicode_Type), 0, "tag", 1))) __PYX_ERR(0, 2488, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_8pin_axis_to_default(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *)__pyx_v_self), __pyx_v_face, __pyx_v_tag);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_8pin_axis_to_default(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, PyObject *__pyx_v_tag) {
  hb_tag_t __pyx_v_hb_tag;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char const *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pin_axis_to_default", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2489
 * 
 *     def pin_axis_to_default(self, face: Face, tag: str) -> bool:
 *         hb_tag = hb_tag_from_string(tag.encode("ascii"), -1)             # <<<<<<<<<<<<<<
 *         return hb_subset_input_pin_axis_to_default(
 *             self._hb_input, face._hb_face, hb_tag
 */
  __pyx_t_1 = PyUnicode_AsASCIIString(__pyx_v_tag); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBytes_AsString(__pyx_t_1); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 2489, __pyx_L1_error)
  __pyx_v_hb_tag = hb_tag_from_string(__pyx_t_2, -1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2490
 *     def pin_axis_to_default(self, face: Face, tag: str) -> bool:
 *         hb_tag = hb_tag_from_string(tag.encode("ascii"), -1)
 *         return hb_subset_input_pin_axis_to_default(             # <<<<<<<<<<<<<<
 *             self._hb_input, face._hb_face, hb_tag
 *         )
 */
  __Pyx_XDECREF(__pyx_r);

  /* "uharfbuzz/_harfbuzz.pyx":2491
 *         hb_tag = hb_tag_from_string(tag.encode("ascii"), -1)
 *         return hb_subset_input_pin_axis_to_default(
 *             self._hb_input, face._hb_face, hb_tag             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_From_hb_bool_t(hb_subset_input_pin_axis_to_default(__pyx_v_self->_hb_input, __pyx_v_face->_hb_face, __pyx_v_hb_tag)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2488
 *         hb_subset_input_keep_everything(self._hb_input)
 * 
 *     def pin_axis_to_default(self, face: Face, tag: str) -> bool:             # <<<<<<<<<<<<<<
 *         hb_tag = hb_tag_from_string(tag.encode("ascii"), -1)
 *         return hb_subset_input_pin_axis_to_default(
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SubsetInput.pin_axis_to_default", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2494
 *         )
 * 
 *     def pin_axis_location(self, face: Face, tag: str, value: float) -> bool:             # <<<<<<<<<<<<<<
 *         hb_tag = hb_tag_from_string(tag.encode("ascii"), -1)
 *         return hb_subset_input_pin_axis_location(
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_11pin_axis_location(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_11SubsetInput_11pin_axis_location = {"pin_axis_location", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_11pin_axis_location, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_11pin_axis_location(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face = 0;
  PyObject *__pyx_v_tag = 0;
  double __pyx_v_value;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pin_axis_location (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_face,&__pyx_n_s_tag,&__pyx_n_s_value,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_face)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2494, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_tag)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2494, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("pin_axis_location", 1, 3, 3, 1); __PYX_ERR(0, 2494, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_value)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2494, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("pin_axis_location", 1, 3, 3, 2); __PYX_ERR(0, 2494, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "pin_axis_location") < 0)) __PYX_ERR(0, 2494, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)values[0]);
    __pyx_v_tag = ((PyObject*)values[1]);
    __pyx_v_value = __pyx_PyFloat_AsDouble(values[2]); if (unlikely((__pyx_v_value == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2494, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("pin_axis_location", 1, 3, 3, __pyx_nargs); __PYX_ERR(0, 2494, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SubsetInput.pin_axis_location", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_face), __pyx_ptype_9uharfbuzz_9_harfbuzz_Face, 0, "face", 0))) __PYX_ERR(0, 2494, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_tag), (&PyUnicode_Type), 0, "tag", 1))) __PYX_ERR(0, 2494, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_10pin_axis_location(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *)__pyx_v_self), __pyx_v_face, __pyx_v_tag, __pyx_v_value);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_10pin_axis_location(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, PyObject *__pyx_v_tag, double __pyx_v_value) {
  hb_tag_t __pyx_v_hb_tag;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char const *__pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pin_axis_location", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2495
 * 
 *     def pin_axis_location(self, face: Face, tag: str, value: float) -> bool:
 *         hb_tag = hb_tag_from_string(tag.encode("ascii"), -1)             # <<<<<<<<<<<<<<
 *         return hb_subset_input_pin_axis_location(
 *             self._hb_input, face._hb_face, hb_tag, value
 */
  __pyx_t_1 = PyUnicode_AsASCIIString(__pyx_v_tag); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBytes_AsString(__pyx_t_1); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 2495, __pyx_L1_error)
  __pyx_v_hb_tag = hb_tag_from_string(__pyx_t_2, -1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2496
 *     def pin_axis_location(self, face: Face, tag: str, value: float) -> bool:
 *         hb_tag = hb_tag_from_string(tag.encode("ascii"), -1)
 *         return hb_subset_input_pin_axis_location(             # <<<<<<<<<<<<<<
 *             self._hb_input, face._hb_face, hb_tag, value
 *         )
 */
  __Pyx_XDECREF(__pyx_r);

  /* "uharfbuzz/_harfbuzz.pyx":2497
 *         hb_tag = hb_tag_from_string(tag.encode("ascii"), -1)
 *         return hb_subset_input_pin_axis_location(
 *             self._hb_input, face._hb_face, hb_tag, value             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_From_hb_bool_t(hb_subset_input_pin_axis_location(__pyx_v_self->_hb_input, __pyx_v_face->_hb_face, __pyx_v_hb_tag, __pyx_v_value)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2494
 *         )
 * 
 *     def pin_axis_location(self, face: Face, tag: str, value: float) -> bool:             # <<<<<<<<<<<<<<
 *         hb_tag = hb_tag_from_string(tag.encode("ascii"), -1)
 *         return hb_subset_input_pin_axis_location(
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SubsetInput.pin_axis_location", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2500
 *         )
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def unicode_set(self) -> Set:
 *         return Set.from_ptr(hb_set_reference (hb_subset_input_unicode_set(self._hb_input)))
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_11unicode_set_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_11unicode_set_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_11unicode_set___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_11unicode_set___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2502
 *     @property
 *     def unicode_set(self) -> Set:
 *         return Set.from_ptr(hb_set_reference (hb_subset_input_unicode_set(self._hb_input)))             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_9uharfbuzz_9_harfbuzz_3Set_from_ptr(hb_set_reference(hb_subset_input_unicode_set(__pyx_v_self->_hb_input)))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2500
 *         )
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def unicode_set(self) -> Set:
 *         return Set.from_ptr(hb_set_reference (hb_subset_input_unicode_set(self._hb_input)))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SubsetInput.unicode_set.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2504
 *         return Set.from_ptr(hb_set_reference (hb_subset_input_unicode_set(self._hb_input)))
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def glyph_set(self) -> Set:
 *         return Set.from_ptr(hb_set_reference (hb_subset_input_glyph_set(self._hb_input)))
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_9glyph_set_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_9glyph_set_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_9glyph_set___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_9glyph_set___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2506
 *     @property
 *     def glyph_set(self) -> Set:
 *         return Set.from_ptr(hb_set_reference (hb_subset_input_glyph_set(self._hb_input)))             # <<<<<<<<<<<<<<
 * 
 *     def sets(self, set_type : SubsetInputSets) -> Set:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_9uharfbuzz_9_harfbuzz_3Set_from_ptr(hb_set_reference(hb_subset_input_glyph_set(__pyx_v_self->_hb_input)))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2504
 *         return Set.from_ptr(hb_set_reference (hb_subset_input_unicode_set(self._hb_input)))
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def glyph_set(self) -> Set:
 *         return Set.from_ptr(hb_set_reference (hb_subset_input_glyph_set(self._hb_input)))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SubsetInput.glyph_set.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2508
 *         return Set.from_ptr(hb_set_reference (hb_subset_input_glyph_set(self._hb_input)))
 * 
 *     def sets(self, set_type : SubsetInputSets) -> Set:             # <<<<<<<<<<<<<<
 *         return Set.from_ptr(hb_set_reference (hb_subset_input_set(self._hb_input, set_type)))
 * 
 */

/* Python wrapper */
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_13sets(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_11SubsetInput_13sets = {"sets", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_13sets, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_13sets(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_set_type = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sets (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_set_type,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_set_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2508, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "sets") < 0)) __PYX_ERR(0, 2508, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_set_type = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sets", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2508, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SubsetInput.sets", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_12sets(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *)__pyx_v_self), __pyx_v_set_type);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_12sets(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self, PyObject *__pyx_v_set_type) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  hb_subset_sets_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sets", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2509
 * 
 *     def sets(self, set_type : SubsetInputSets) -> Set:
 *         return Set.from_ptr(hb_set_reference (hb_subset_input_set(self._hb_input, set_type)))             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = ((hb_subset_sets_t)__Pyx_PyInt_As_hb_subset_sets_t(__pyx_v_set_type)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2509, __pyx_L1_error)
  __pyx_t_2 = ((PyObject *)__pyx_f_9uharfbuzz_9_harfbuzz_3Set_from_ptr(hb_set_reference(hb_subset_input_set(__pyx_v_self->_hb_input, __pyx_t_1)))); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2508
 *         return Set.from_ptr(hb_set_reference (hb_subset_input_glyph_set(self._hb_input)))
 * 
 *     def sets(self, set_type : SubsetInputSets) -> Set:             # <<<<<<<<<<<<<<
 *         return Set.from_ptr(hb_set_reference (hb_subset_input_set(self._hb_input, set_type)))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SubsetInput.sets", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2511
 *         return Set.from_ptr(hb_set_reference (hb_subset_input_set(self._hb_input, set_type)))
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def no_subset_table_tag_set(self) -> Set:
 *         return self.sets(SubsetInputSets.NO_SUBSET_TABLE_TAG)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_23no_subset_table_tag_set_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_23no_subset_table_tag_set_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_23no_subset_table_tag_set___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_23no_subset_table_tag_set___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2513
 *     @property
 *     def no_subset_table_tag_set(self) -> Set:
 *         return self.sets(SubsetInputSets.NO_SUBSET_TABLE_TAG)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_sets); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_SubsetInputSets); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_NO_SUBSET_TABLE_TAG); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2513, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2511
 *         return Set.from_ptr(hb_set_reference (hb_subset_input_set(self._hb_input, set_type)))
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def no_subset_table_tag_set(self) -> Set:
 *         return self.sets(SubsetInputSets.NO_SUBSET_TABLE_TAG)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SubsetInput.no_subset_table_tag_set.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2515
 *         return self.sets(SubsetInputSets.NO_SUBSET_TABLE_TAG)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def drop_table_tag_set(self) -> Set:
 *         return self.sets(SubsetInputSets.DROP_TABLE_TAG)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_18drop_table_tag_set_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_18drop_table_tag_set_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_18drop_table_tag_set___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_18drop_table_tag_set___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2517
 *     @property
 *     def drop_table_tag_set(self) -> Set:
 *         return self.sets(SubsetInputSets.DROP_TABLE_TAG)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_sets); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_SubsetInputSets); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_DROP_TABLE_TAG); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2517, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2515
 *         return self.sets(SubsetInputSets.NO_SUBSET_TABLE_TAG)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def drop_table_tag_set(self) -> Set:
 *         return self.sets(SubsetInputSets.DROP_TABLE_TAG)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SubsetInput.drop_table_tag_set.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2519
 *         return self.sets(SubsetInputSets.DROP_TABLE_TAG)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def name_id_set(self) -> Set:
 *         return self.sets(SubsetInputSets.NAME_ID)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_11name_id_set_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_11name_id_set_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_11name_id_set___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_11name_id_set___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2521
 *     @property
 *     def name_id_set(self) -> Set:
 *         return self.sets(SubsetInputSets.NAME_ID)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_sets); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_SubsetInputSets); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_NAME_ID); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2521, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2519
 *         return self.sets(SubsetInputSets.DROP_TABLE_TAG)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def name_id_set(self) -> Set:
 *         return self.sets(SubsetInputSets.NAME_ID)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SubsetInput.name_id_set.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2523
 *         return self.sets(SubsetInputSets.NAME_ID)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def name_lang_id_set(self) -> Set:
 *         return self.sets(SubsetInputSets.NAME_LANG_ID)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_16name_lang_id_set_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_16name_lang_id_set_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_16name_lang_id_set___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_16name_lang_id_set___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2525
 *     @property
 *     def name_lang_id_set(self) -> Set:
 *         return self.sets(SubsetInputSets.NAME_LANG_ID)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_sets); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_SubsetInputSets); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_NAME_LANG_ID); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2525, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2523
 *         return self.sets(SubsetInputSets.NAME_ID)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def name_lang_id_set(self) -> Set:
 *         return self.sets(SubsetInputSets.NAME_LANG_ID)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SubsetInput.name_lang_id_set.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2527
 *         return self.sets(SubsetInputSets.NAME_LANG_ID)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def layout_feature_tag_set(self) -> Set:
 *         return self.sets(SubsetInputSets.LAYOUT_FEATURE_TAG)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_22layout_feature_tag_set_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_22layout_feature_tag_set_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_22layout_feature_tag_set___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_22layout_feature_tag_set___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2529
 *     @property
 *     def layout_feature_tag_set(self) -> Set:
 *         return self.sets(SubsetInputSets.LAYOUT_FEATURE_TAG)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_sets); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_SubsetInputSets); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_LAYOUT_FEATURE_TAG); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2529, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2527
 *         return self.sets(SubsetInputSets.NAME_LANG_ID)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def layout_feature_tag_set(self) -> Set:
 *         return self.sets(SubsetInputSets.LAYOUT_FEATURE_TAG)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SubsetInput.layout_feature_tag_set.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2531
 *         return self.sets(SubsetInputSets.LAYOUT_FEATURE_TAG)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def layout_script_tag_set(self) -> Set:
 *         return self.sets(SubsetInputSets.LAYOUT_SCRIPT_TAG)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_21layout_script_tag_set_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_21layout_script_tag_set_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_21layout_script_tag_set___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_21layout_script_tag_set___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2533
 *     @property
 *     def layout_script_tag_set(self) -> Set:
 *         return self.sets(SubsetInputSets.LAYOUT_SCRIPT_TAG)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_sets); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_SubsetInputSets); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_LAYOUT_SCRIPT_TAG); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2533, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2531
 *         return self.sets(SubsetInputSets.LAYOUT_FEATURE_TAG)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def layout_script_tag_set(self) -> Set:
 *         return self.sets(SubsetInputSets.LAYOUT_SCRIPT_TAG)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SubsetInput.layout_script_tag_set.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2535
 *         return self.sets(SubsetInputSets.LAYOUT_SCRIPT_TAG)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def flags(self) -> SubsetFlags:
 *         cdef unsigned subset_flags = hb_subset_input_get_flags(self._hb_input)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_5flags_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_5flags_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_5flags___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_5flags___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self) {
  unsigned int __pyx_v_subset_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2537
 *     @property
 *     def flags(self) -> SubsetFlags:
 *         cdef unsigned subset_flags = hb_subset_input_get_flags(self._hb_input)             # <<<<<<<<<<<<<<
 *         return SubsetFlags(subset_flags)
 * 
 */
  __pyx_v_subset_flags = hb_subset_input_get_flags(__pyx_v_self->_hb_input);

  /* "uharfbuzz/_harfbuzz.pyx":2538
 *     def flags(self) -> SubsetFlags:
 *         cdef unsigned subset_flags = hb_subset_input_get_flags(self._hb_input)
 *         return SubsetFlags(subset_flags)             # <<<<<<<<<<<<<<
 * 
 *     @flags.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SubsetFlags); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_unsigned_int(__pyx_v_subset_flags); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2538, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2535
 *         return self.sets(SubsetInputSets.LAYOUT_SCRIPT_TAG)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def flags(self) -> SubsetFlags:
 *         cdef unsigned subset_flags = hb_subset_input_get_flags(self._hb_input)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SubsetInput.flags.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2540
 *         return SubsetFlags(subset_flags)
 * 
 *     @flags.setter             # <<<<<<<<<<<<<<
 *     def flags(self, flags: SubsetFlags) -> None:
 *         hb_subset_input_set_flags(self._hb_input, int(flags))
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_5flags_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_flags); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_5flags_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_flags) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_5flags_2__set__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *)__pyx_v_self), ((PyObject *)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_5flags_2__set__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self, PyObject *__pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  unsigned int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2542
 *     @flags.setter
 *     def flags(self, flags: SubsetFlags) -> None:
 *         hb_subset_input_set_flags(self._hb_input, int(flags))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_unsigned_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2542, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  hb_subset_input_set_flags(__pyx_v_self->_hb_input, __pyx_t_2);

  /* "uharfbuzz/_harfbuzz.pyx":2540
 *         return SubsetFlags(subset_flags)
 * 
 *     @flags.setter             # <<<<<<<<<<<<<<
 *     def flags(self, flags: SubsetFlags) -> None:
 *         hb_subset_input_set_flags(self._hb_input, int(flags))
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SubsetInput.flags.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_15__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_11SubsetInput_15__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_15__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_15__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_14__reduce_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_14__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SubsetInput.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_17__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_11SubsetInput_17__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_17__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_17__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SubsetInput.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_16__setstate_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_11SubsetInput_16__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SubsetInput.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2548
 *     cdef hb_subset_plan_t* _hb_plan
 * 
 *     def __cinit__(self, face: Face, input: SubsetInput):             # <<<<<<<<<<<<<<
 *         self._hb_plan = hb_subset_plan_create_or_fail(face._hb_face, input._hb_input)
 *         if self._hb_plan is NULL:
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_10SubsetPlan_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_10SubsetPlan_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face = 0;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_input = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_face,&__pyx_n_s_input,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_face)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2548, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_input)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2548, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, 1); __PYX_ERR(0, 2548, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(0, 2548, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
    }
    __pyx_v_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)values[0]);
    __pyx_v_input = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 2548, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SubsetPlan.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_face), __pyx_ptype_9uharfbuzz_9_harfbuzz_Face, 0, "face", 0))) __PYX_ERR(0, 2548, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_input), __pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetInput, 0, "input", 0))) __PYX_ERR(0, 2548, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_10SubsetPlan___cinit__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetPlan *)__pyx_v_self), __pyx_v_face, __pyx_v_input);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_10SubsetPlan___cinit__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetPlan *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_v_face, struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput *__pyx_v_input) {
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2549
 * 
 *     def __cinit__(self, face: Face, input: SubsetInput):
 *         self._hb_plan = hb_subset_plan_create_or_fail(face._hb_face, input._hb_input)             # <<<<<<<<<<<<<<
 *         if self._hb_plan is NULL:
 *             raise MemoryError()
 */
  __pyx_v_self->_hb_plan = hb_subset_plan_create_or_fail(__pyx_v_face->_hb_face, __pyx_v_input->_hb_input);

  /* "uharfbuzz/_harfbuzz.pyx":2550
 *     def __cinit__(self, face: Face, input: SubsetInput):
 *         self._hb_plan = hb_subset_plan_create_or_fail(face._hb_face, input._hb_input)
 *         if self._hb_plan is NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_hb_plan == NULL);
  if (unlikely(__pyx_t_1)) {

    /* "uharfbuzz/_harfbuzz.pyx":2551
 *         self._hb_plan = hb_subset_plan_create_or_fail(face._hb_face, input._hb_input)
 *         if self._hb_plan is NULL:
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
    PyErr_NoMemory(); __PYX_ERR(0, 2551, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":2550
 *     def __cinit__(self, face: Face, input: SubsetInput):
 *         self._hb_plan = hb_subset_plan_create_or_fail(face._hb_face, input._hb_input)
 *         if self._hb_plan is NULL:             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2548
 *     cdef hb_subset_plan_t* _hb_plan
 * 
 *     def __cinit__(self, face: Face, input: SubsetInput):             # <<<<<<<<<<<<<<
 *         self._hb_plan = hb_subset_plan_create_or_fail(face._hb_face, input._hb_input)
 *         if self._hb_plan is NULL:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SubsetPlan.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2553
 *             raise MemoryError()
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self._hb_plan is not NULL:
 *             hb_subset_plan_destroy(self._hb_plan)
 */

/* Python wrapper */
static void __pyx_pw_9uharfbuzz_9_harfbuzz_10SubsetPlan_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_9uharfbuzz_9_harfbuzz_10SubsetPlan_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_9uharfbuzz_9_harfbuzz_10SubsetPlan_2__dealloc__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetPlan *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_9uharfbuzz_9_harfbuzz_10SubsetPlan_2__dealloc__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetPlan *__pyx_v_self) {
  int __pyx_t_1;

  /* "uharfbuzz/_harfbuzz.pyx":2554
 * 
 *     def __dealloc__(self):
 *         if self._hb_plan is not NULL:             # <<<<<<<<<<<<<<
 *             hb_subset_plan_destroy(self._hb_plan)
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_hb_plan != NULL);
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":2555
 *     def __dealloc__(self):
 *         if self._hb_plan is not NULL:
 *             hb_subset_plan_destroy(self._hb_plan)             # <<<<<<<<<<<<<<
 * 
 *     def execute(self) -> Face:
 */
    hb_subset_plan_destroy(__pyx_v_self->_hb_plan);

    /* "uharfbuzz/_harfbuzz.pyx":2554
 * 
 *     def __dealloc__(self):
 *         if self._hb_plan is not NULL:             # <<<<<<<<<<<<<<
 *             hb_subset_plan_destroy(self._hb_plan)
 * 
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2553
 *             raise MemoryError()
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self._hb_plan is not NULL:
 *             hb_subset_plan_destroy(self._hb_plan)
 */

  /* function exit code */
}

/* "uharfbuzz/_harfbuzz.pyx":2557
 *             hb_subset_plan_destroy(self._hb_plan)
 * 
 *     def execute(self) -> Face:             # <<<<<<<<<<<<<<
 *         new_face = hb_subset_plan_execute_or_fail(self._hb_plan)
 *         if new_face == NULL:
 */

/* Python wrapper */
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_pw_9uharfbuzz_9_harfbuzz_10SubsetPlan_5execute(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_10SubsetPlan_5execute = {"execute", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10SubsetPlan_5execute, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_pw_9uharfbuzz_9_harfbuzz_10SubsetPlan_5execute(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("execute (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("execute", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "execute", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_10SubsetPlan_4execute(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetPlan *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_pf_9uharfbuzz_9_harfbuzz_10SubsetPlan_4execute(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetPlan *__pyx_v_self) {
  hb_face_t *__pyx_v_new_face;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("execute", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2558
 * 
 *     def execute(self) -> Face:
 *         new_face = hb_subset_plan_execute_or_fail(self._hb_plan)             # <<<<<<<<<<<<<<
 *         if new_face == NULL:
 *             raise RuntimeError("Subsetting failed")
 */
  __pyx_v_new_face = hb_subset_plan_execute_or_fail(__pyx_v_self->_hb_plan);

  /* "uharfbuzz/_harfbuzz.pyx":2559
 *     def execute(self) -> Face:
 *         new_face = hb_subset_plan_execute_or_fail(self._hb_plan)
 *         if new_face == NULL:             # <<<<<<<<<<<<<<
 *             raise RuntimeError("Subsetting failed")
 *         return Face.from_ptr(new_face)
 */
  __pyx_t_1 = (__pyx_v_new_face == NULL);
  if (unlikely(__pyx_t_1)) {

    /* "uharfbuzz/_harfbuzz.pyx":2560
 *         new_face = hb_subset_plan_execute_or_fail(self._hb_plan)
 *         if new_face == NULL:
 *             raise RuntimeError("Subsetting failed")             # <<<<<<<<<<<<<<
 *         return Face.from_ptr(new_face)
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2560, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 2560, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":2559
 *     def execute(self) -> Face:
 *         new_face = hb_subset_plan_execute_or_fail(self._hb_plan)
 *         if new_face == NULL:             # <<<<<<<<<<<<<<
 *             raise RuntimeError("Subsetting failed")
 *         return Face.from_ptr(new_face)
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2561
 *         if new_face == NULL:
 *             raise RuntimeError("Subsetting failed")
 *         return Face.from_ptr(new_face)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_2 = ((PyObject *)__pyx_f_9uharfbuzz_9_harfbuzz_4Face_from_ptr(__pyx_v_new_face)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)__pyx_t_2);
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2557
 *             hb_subset_plan_destroy(self._hb_plan)
 * 
 *     def execute(self) -> Face:             # <<<<<<<<<<<<<<
 *         new_face = hb_subset_plan_execute_or_fail(self._hb_plan)
 *         if new_face == NULL:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SubsetPlan.execute", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2563
 *         return Face.from_ptr(new_face)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def old_to_new_glyph_mapping(self) -> Map:
 *         return Map.from_ptr(hb_map_reference (<hb_map_t*>hb_subset_plan_old_to_new_glyph_mapping(self._hb_plan)))
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10SubsetPlan_24old_to_new_glyph_mapping_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10SubsetPlan_24old_to_new_glyph_mapping_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_10SubsetPlan_24old_to_new_glyph_mapping___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetPlan *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10SubsetPlan_24old_to_new_glyph_mapping___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetPlan *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2565
 *     @property
 *     def old_to_new_glyph_mapping(self) -> Map:
 *         return Map.from_ptr(hb_map_reference (<hb_map_t*>hb_subset_plan_old_to_new_glyph_mapping(self._hb_plan)))             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_9uharfbuzz_9_harfbuzz_3Map_from_ptr(hb_map_reference(((hb_map_t *)hb_subset_plan_old_to_new_glyph_mapping(__pyx_v_self->_hb_plan))))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2563
 *         return Face.from_ptr(new_face)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def old_to_new_glyph_mapping(self) -> Map:
 *         return Map.from_ptr(hb_map_reference (<hb_map_t*>hb_subset_plan_old_to_new_glyph_mapping(self._hb_plan)))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SubsetPlan.old_to_new_glyph_mapping.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2567
 *         return Map.from_ptr(hb_map_reference (<hb_map_t*>hb_subset_plan_old_to_new_glyph_mapping(self._hb_plan)))
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def new_to_old_glyph_mapping(self) -> Map:
 *         return Map.from_ptr(hb_map_reference (<hb_map_t*>hb_subset_plan_new_to_old_glyph_mapping(self._hb_plan)))
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10SubsetPlan_24new_to_old_glyph_mapping_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10SubsetPlan_24new_to_old_glyph_mapping_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_10SubsetPlan_24new_to_old_glyph_mapping___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetPlan *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10SubsetPlan_24new_to_old_glyph_mapping___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetPlan *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2569
 *     @property
 *     def new_to_old_glyph_mapping(self) -> Map:
 *         return Map.from_ptr(hb_map_reference (<hb_map_t*>hb_subset_plan_new_to_old_glyph_mapping(self._hb_plan)))             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_9uharfbuzz_9_harfbuzz_3Map_from_ptr(hb_map_reference(((hb_map_t *)hb_subset_plan_new_to_old_glyph_mapping(__pyx_v_self->_hb_plan))))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2569, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2567
 *         return Map.from_ptr(hb_map_reference (<hb_map_t*>hb_subset_plan_old_to_new_glyph_mapping(self._hb_plan)))
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def new_to_old_glyph_mapping(self) -> Map:
 *         return Map.from_ptr(hb_map_reference (<hb_map_t*>hb_subset_plan_new_to_old_glyph_mapping(self._hb_plan)))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SubsetPlan.new_to_old_glyph_mapping.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2571
 *         return Map.from_ptr(hb_map_reference (<hb_map_t*>hb_subset_plan_new_to_old_glyph_mapping(self._hb_plan)))
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def unicode_to_old_glyph_mapping(self) -> Map:
 *         return Map.from_ptr(hb_map_reference (<hb_map_t*>hb_subset_plan_unicode_to_old_glyph_mapping(self._hb_plan)))
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10SubsetPlan_28unicode_to_old_glyph_mapping_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10SubsetPlan_28unicode_to_old_glyph_mapping_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_10SubsetPlan_28unicode_to_old_glyph_mapping___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetPlan *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10SubsetPlan_28unicode_to_old_glyph_mapping___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetPlan *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2573
 *     @property
 *     def unicode_to_old_glyph_mapping(self) -> Map:
 *         return Map.from_ptr(hb_map_reference (<hb_map_t*>hb_subset_plan_unicode_to_old_glyph_mapping(self._hb_plan)))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((PyObject *)__pyx_f_9uharfbuzz_9_harfbuzz_3Map_from_ptr(hb_map_reference(((hb_map_t *)hb_subset_plan_unicode_to_old_glyph_mapping(__pyx_v_self->_hb_plan))))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2571
 *         return Map.from_ptr(hb_map_reference (<hb_map_t*>hb_subset_plan_new_to_old_glyph_mapping(self._hb_plan)))
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def unicode_to_old_glyph_mapping(self) -> Map:
 *         return Map.from_ptr(hb_map_reference (<hb_map_t*>hb_subset_plan_unicode_to_old_glyph_mapping(self._hb_plan)))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SubsetPlan.unicode_to_old_glyph_mapping.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10SubsetPlan_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_10SubsetPlan_7__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10SubsetPlan_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10SubsetPlan_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_10SubsetPlan_6__reduce_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetPlan *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10SubsetPlan_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetPlan *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SubsetPlan.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10SubsetPlan_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_10SubsetPlan_9__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10SubsetPlan_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_10SubsetPlan_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SubsetPlan.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_10SubsetPlan_8__setstate_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetPlan *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_10SubsetPlan_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetPlan *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SubsetPlan.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2581
 *     INVALID_VALUE = HB_SET_VALUE_INVALID
 * 
 *     def __cinit__(self, init = set()):             # <<<<<<<<<<<<<<
 *         self._hb_set = hb_set_create()
 *         if not hb_set_allocation_successful(self._hb_set):
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_3Set_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_3Set_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_init = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_init,0};
    values[0] = __Pyx_Arg_NewRef_VARARGS(__pyx_k__18);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_init);
          if (value) { values[0] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2581, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(0, 2581, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_init = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 2581, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set___cinit__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self), __pyx_v_init);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_3Set___cinit__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, PyObject *__pyx_v_init) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2582
 * 
 *     def __cinit__(self, init = set()):
 *         self._hb_set = hb_set_create()             # <<<<<<<<<<<<<<
 *         if not hb_set_allocation_successful(self._hb_set):
 *             raise MemoryError()
 */
  __pyx_v_self->_hb_set = hb_set_create();

  /* "uharfbuzz/_harfbuzz.pyx":2583
 *     def __cinit__(self, init = set()):
 *         self._hb_set = hb_set_create()
 *         if not hb_set_allocation_successful(self._hb_set):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_1 = (!(hb_set_allocation_successful(__pyx_v_self->_hb_set) != 0));
  if (unlikely(__pyx_t_1)) {

    /* "uharfbuzz/_harfbuzz.pyx":2584
 *         self._hb_set = hb_set_create()
 *         if not hb_set_allocation_successful(self._hb_set):
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *         self.set(init)
 */
    PyErr_NoMemory(); __PYX_ERR(0, 2584, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":2583
 *     def __cinit__(self, init = set()):
 *         self._hb_set = hb_set_create()
 *         if not hb_set_allocation_successful(self._hb_set):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2586
 *             raise MemoryError()
 * 
 *         self.set(init)             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_init};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2581
 *     INVALID_VALUE = HB_SET_VALUE_INVALID
 * 
 *     def __cinit__(self, init = set()):             # <<<<<<<<<<<<<<
 *         self._hb_set = hb_set_create()
 *         if not hb_set_allocation_successful(self._hb_set):
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2588
 *         self.set(init)
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         hb_set_destroy(self._hb_set)
 * 
 */

/* Python wrapper */
static void __pyx_pw_9uharfbuzz_9_harfbuzz_3Set_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_9uharfbuzz_9_harfbuzz_3Set_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_2__dealloc__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_2__dealloc__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self) {

  /* "uharfbuzz/_harfbuzz.pyx":2589
 * 
 *     def __dealloc__(self):
 *         hb_set_destroy(self._hb_set)             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  hb_set_destroy(__pyx_v_self->_hb_set);

  /* "uharfbuzz/_harfbuzz.pyx":2588
 *         self.set(init)
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         hb_set_destroy(self._hb_set)
 * 
 */

  /* function exit code */
}

/* "uharfbuzz/_harfbuzz.pyx":2592
 * 
 *     @staticmethod
 *     cdef Set from_ptr(hb_set_t* hb_set):             # <<<<<<<<<<<<<<
 *         """Create Set from a pointer taking ownership of a it."""
 * 
 */

static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_f_9uharfbuzz_9_harfbuzz_3Set_from_ptr(hb_set_t *__pyx_v_hb_set) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_wrapper = 0;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("from_ptr", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2595
 *         """Create Set from a pointer taking ownership of a it."""
 * 
 *         cdef Set wrapper = Set.__new__(Set)             # <<<<<<<<<<<<<<
 *         wrapper._hb_set = hb_set
 *         return wrapper
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_9uharfbuzz_9_harfbuzz_Set(((PyTypeObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Set), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2595, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2596
 * 
 *         cdef Set wrapper = Set.__new__(Set)
 *         wrapper._hb_set = hb_set             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __pyx_v_wrapper->_hb_set = __pyx_v_hb_set;

  /* "uharfbuzz/_harfbuzz.pyx":2597
 *         cdef Set wrapper = Set.__new__(Set)
 *         wrapper._hb_set = hb_set
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     def copy(self) -> Set:
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = __pyx_v_wrapper;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2592
 * 
 *     @staticmethod
 *     cdef Set from_ptr(hb_set_t* hb_set):             # <<<<<<<<<<<<<<
 *         """Create Set from a pointer taking ownership of a it."""
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.from_ptr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2599
 *         return wrapper
 * 
 *     def copy(self) -> Set:             # <<<<<<<<<<<<<<
 *         c = Set()
 *         c._hb_set = hb_set_copy(self._hb_set)
 */

/* Python wrapper */
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_5copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_5copy = {"copy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_5copy, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_5copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("copy", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "copy", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_4copy(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_4copy(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_c = NULL;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2600
 * 
 *     def copy(self) -> Set:
 *         c = Set()             # <<<<<<<<<<<<<<
 *         c._hb_set = hb_set_copy(self._hb_set)
 *         return c
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Set)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_c = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2601
 *     def copy(self) -> Set:
 *         c = Set()
 *         c._hb_set = hb_set_copy(self._hb_set)             # <<<<<<<<<<<<<<
 *         return c
 * 
 */
  __pyx_v_c->_hb_set = hb_set_copy(__pyx_v_self->_hb_set);

  /* "uharfbuzz/_harfbuzz.pyx":2602
 *         c = Set()
 *         c._hb_set = hb_set_copy(self._hb_set)
 *         return c             # <<<<<<<<<<<<<<
 * 
 *     def __copy__(self) -> Set:
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_c);
  __pyx_r = __pyx_v_c;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2599
 *         return wrapper
 * 
 *     def copy(self) -> Set:             # <<<<<<<<<<<<<<
 *         c = Set()
 *         c._hb_set = hb_set_copy(self._hb_set)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_c);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2604
 *         return c
 * 
 *     def __copy__(self) -> Set:             # <<<<<<<<<<<<<<
 *         return self.copy()
 * 
 */

/* Python wrapper */
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_7__copy__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_7__copy__ = {"__copy__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_7__copy__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_7__copy__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__copy__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__copy__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__copy__", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_6__copy__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_6__copy__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__copy__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2605
 * 
 *     def __copy__(self) -> Set:
 *         return self.copy()             # <<<<<<<<<<<<<<
 * 
 *     def clear(self):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_copy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2605, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2605, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_9uharfbuzz_9_harfbuzz_Set))))) __PYX_ERR(0, 2605, __pyx_L1_error)
  __pyx_r = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2604
 *         return c
 * 
 *     def __copy__(self) -> Set:             # <<<<<<<<<<<<<<
 *         return self.copy()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.__copy__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2607
 *         return self.copy()
 * 
 *     def clear(self):             # <<<<<<<<<<<<<<
 *         hb_set_clear(self._hb_set)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_9clear(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_9clear = {"clear", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_9clear, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_9clear(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clear (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("clear", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "clear", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_8clear(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_8clear(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clear", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2608
 * 
 *     def clear(self):
 *         hb_set_clear(self._hb_set)             # <<<<<<<<<<<<<<
 * 
 *     def __bool__(self) -> bool:
 */
  hb_set_clear(__pyx_v_self->_hb_set);

  /* "uharfbuzz/_harfbuzz.pyx":2607
 *         return self.copy()
 * 
 *     def clear(self):             # <<<<<<<<<<<<<<
 *         hb_set_clear(self._hb_set)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2610
 *         hb_set_clear(self._hb_set)
 * 
 *     def __bool__(self) -> bool:             # <<<<<<<<<<<<<<
 *         return not hb_set_is_empty(self._hb_set)
 * 
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_3Set_11__bool__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_3Set_11__bool__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__bool__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_10__bool__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_10__bool__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self) {
  int __pyx_r;

  /* "uharfbuzz/_harfbuzz.pyx":2611
 * 
 *     def __bool__(self) -> bool:
 *         return not hb_set_is_empty(self._hb_set)             # <<<<<<<<<<<<<<
 * 
 *     def invert(self):
 */
  __pyx_r = (!(hb_set_is_empty(__pyx_v_self->_hb_set) != 0));
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2610
 *         hb_set_clear(self._hb_set)
 * 
 *     def __bool__(self) -> bool:             # <<<<<<<<<<<<<<
 *         return not hb_set_is_empty(self._hb_set)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2613
 *         return not hb_set_is_empty(self._hb_set)
 * 
 *     def invert(self):             # <<<<<<<<<<<<<<
 *         hb_set_invert(self._hb_set)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_13invert(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_13invert = {"invert", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_13invert, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_13invert(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("invert (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("invert", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "invert", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_12invert(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_12invert(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("invert", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2614
 * 
 *     def invert(self):
 *         hb_set_invert(self._hb_set)             # <<<<<<<<<<<<<<
 * 
 *     def is_inverted(self) -> bool:
 */
  hb_set_invert(__pyx_v_self->_hb_set);

  /* "uharfbuzz/_harfbuzz.pyx":2613
 *         return not hb_set_is_empty(self._hb_set)
 * 
 *     def invert(self):             # <<<<<<<<<<<<<<
 *         hb_set_invert(self._hb_set)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2616
 *         hb_set_invert(self._hb_set)
 * 
 *     def is_inverted(self) -> bool:             # <<<<<<<<<<<<<<
 *         return hb_set_is_inverted(self._hb_set)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_15is_inverted(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_15is_inverted = {"is_inverted", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_15is_inverted, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_15is_inverted(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_inverted (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("is_inverted", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "is_inverted", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_14is_inverted(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_14is_inverted(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_inverted", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2617
 * 
 *     def is_inverted(self) -> bool:
 *         return hb_set_is_inverted(self._hb_set)             # <<<<<<<<<<<<<<
 * 
 *     def __contains__(self, c) -> bool:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_hb_bool_t(hb_set_is_inverted(__pyx_v_self->_hb_set)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2617, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2616
 *         hb_set_invert(self._hb_set)
 * 
 *     def is_inverted(self) -> bool:             # <<<<<<<<<<<<<<
 *         return hb_set_is_inverted(self._hb_set)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.is_inverted", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2619
 *         return hb_set_is_inverted(self._hb_set)
 * 
 *     def __contains__(self, c) -> bool:             # <<<<<<<<<<<<<<
 *         if type(c) != int:
 *             return False
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_3Set_17__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_c); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_3Set_17__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_c) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__contains__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_16__contains__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self), ((PyObject *)__pyx_v_c));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_16__contains__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, PyObject *__pyx_v_c) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  hb_codepoint_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__contains__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2620
 * 
 *     def __contains__(self, c) -> bool:
 *         if type(c) != int:             # <<<<<<<<<<<<<<
 *             return False
 *         if c < 0 or c >= self.INVALID_VALUE:
 */
  __pyx_t_1 = PyObject_RichCompare(((PyObject *)Py_TYPE(__pyx_v_c)), ((PyObject *)(&PyInt_Type)), Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2620, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 2620, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "uharfbuzz/_harfbuzz.pyx":2621
 *     def __contains__(self, c) -> bool:
 *         if type(c) != int:
 *             return False             # <<<<<<<<<<<<<<
 *         if c < 0 or c >= self.INVALID_VALUE:
 *             return False
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "uharfbuzz/_harfbuzz.pyx":2620
 * 
 *     def __contains__(self, c) -> bool:
 *         if type(c) != int:             # <<<<<<<<<<<<<<
 *             return False
 *         if c < 0 or c >= self.INVALID_VALUE:
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2622
 *         if type(c) != int:
 *             return False
 *         if c < 0 or c >= self.INVALID_VALUE:             # <<<<<<<<<<<<<<
 *             return False
 *         return hb_set_has(self._hb_set, c)
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_c, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2622, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 2622, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_INVALID_VALUE); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2622, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_c, __pyx_t_1, Py_GE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2622, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 2622, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = __pyx_t_3;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_2) {

    /* "uharfbuzz/_harfbuzz.pyx":2623
 *             return False
 *         if c < 0 or c >= self.INVALID_VALUE:
 *             return False             # <<<<<<<<<<<<<<
 *         return hb_set_has(self._hb_set, c)
 * 
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "uharfbuzz/_harfbuzz.pyx":2622
 *         if type(c) != int:
 *             return False
 *         if c < 0 or c >= self.INVALID_VALUE:             # <<<<<<<<<<<<<<
 *             return False
 *         return hb_set_has(self._hb_set, c)
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2624
 *         if c < 0 or c >= self.INVALID_VALUE:
 *             return False
 *         return hb_set_has(self._hb_set, c)             # <<<<<<<<<<<<<<
 * 
 *     def add(self, c: int):
 */
  __pyx_t_5 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_c); if (unlikely((__pyx_t_5 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 2624, __pyx_L1_error)
  __pyx_r = hb_set_has(__pyx_v_self->_hb_set, __pyx_t_5);
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2619
 *         return hb_set_is_inverted(self._hb_set)
 * 
 *     def __contains__(self, c) -> bool:             # <<<<<<<<<<<<<<
 *         if type(c) != int:
 *             return False
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2626
 *         return hb_set_has(self._hb_set, c)
 * 
 *     def add(self, c: int):             # <<<<<<<<<<<<<<
 *         hb_set_add(self._hb_set, c)
 *         if not hb_set_allocation_successful(self._hb_set):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_19add(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_19add = {"add", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_19add, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_19add(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_c = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_c,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_c)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2626, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "add") < 0)) __PYX_ERR(0, 2626, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_c = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("add", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2626, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.add", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_c), (&PyInt_Type), 0, "c", 1))) __PYX_ERR(0, 2626, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_18add(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self), __pyx_v_c);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_18add(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  hb_codepoint_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2627
 * 
 *     def add(self, c: int):
 *         hb_set_add(self._hb_set, c)             # <<<<<<<<<<<<<<
 *         if not hb_set_allocation_successful(self._hb_set):
 *             raise MemoryError()
 */
  __pyx_t_1 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_c); if (unlikely((__pyx_t_1 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 2627, __pyx_L1_error)
  hb_set_add(__pyx_v_self->_hb_set, __pyx_t_1);

  /* "uharfbuzz/_harfbuzz.pyx":2628
 *     def add(self, c: int):
 *         hb_set_add(self._hb_set, c)
 *         if not hb_set_allocation_successful(self._hb_set):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_2 = (!(hb_set_allocation_successful(__pyx_v_self->_hb_set) != 0));
  if (unlikely(__pyx_t_2)) {

    /* "uharfbuzz/_harfbuzz.pyx":2629
 *         hb_set_add(self._hb_set, c)
 *         if not hb_set_allocation_successful(self._hb_set):
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *     def add_range(self, first: int, last: int):
 */
    PyErr_NoMemory(); __PYX_ERR(0, 2629, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":2628
 *     def add(self, c: int):
 *         hb_set_add(self._hb_set, c)
 *         if not hb_set_allocation_successful(self._hb_set):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2626
 *         return hb_set_has(self._hb_set, c)
 * 
 *     def add(self, c: int):             # <<<<<<<<<<<<<<
 *         hb_set_add(self._hb_set, c)
 *         if not hb_set_allocation_successful(self._hb_set):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.add", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2631
 *             raise MemoryError()
 * 
 *     def add_range(self, first: int, last: int):             # <<<<<<<<<<<<<<
 *         hb_set_add_range(self._hb_set, first, last)
 *         if not hb_set_allocation_successful(self._hb_set):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_21add_range(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_21add_range = {"add_range", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_21add_range, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_21add_range(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_first = 0;
  PyObject *__pyx_v_last = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("add_range (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_first,&__pyx_n_s_last,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_first)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2631, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_last)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2631, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("add_range", 1, 2, 2, 1); __PYX_ERR(0, 2631, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "add_range") < 0)) __PYX_ERR(0, 2631, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_first = ((PyObject*)values[0]);
    __pyx_v_last = ((PyObject*)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("add_range", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 2631, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.add_range", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_first), (&PyInt_Type), 0, "first", 1))) __PYX_ERR(0, 2631, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_last), (&PyInt_Type), 0, "last", 1))) __PYX_ERR(0, 2631, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_20add_range(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self), __pyx_v_first, __pyx_v_last);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_20add_range(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, PyObject *__pyx_v_first, PyObject *__pyx_v_last) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  hb_codepoint_t __pyx_t_1;
  hb_codepoint_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("add_range", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2632
 * 
 *     def add_range(self, first: int, last: int):
 *         hb_set_add_range(self._hb_set, first, last)             # <<<<<<<<<<<<<<
 *         if not hb_set_allocation_successful(self._hb_set):
 *             raise MemoryError()
 */
  __pyx_t_1 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_first); if (unlikely((__pyx_t_1 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 2632, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_last); if (unlikely((__pyx_t_2 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 2632, __pyx_L1_error)
  hb_set_add_range(__pyx_v_self->_hb_set, __pyx_t_1, __pyx_t_2);

  /* "uharfbuzz/_harfbuzz.pyx":2633
 *     def add_range(self, first: int, last: int):
 *         hb_set_add_range(self._hb_set, first, last)
 *         if not hb_set_allocation_successful(self._hb_set):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_3 = (!(hb_set_allocation_successful(__pyx_v_self->_hb_set) != 0));
  if (unlikely(__pyx_t_3)) {

    /* "uharfbuzz/_harfbuzz.pyx":2634
 *         hb_set_add_range(self._hb_set, first, last)
 *         if not hb_set_allocation_successful(self._hb_set):
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *     def remove(self, c: int):
 */
    PyErr_NoMemory(); __PYX_ERR(0, 2634, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":2633
 *     def add_range(self, first: int, last: int):
 *         hb_set_add_range(self._hb_set, first, last)
 *         if not hb_set_allocation_successful(self._hb_set):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2631
 *             raise MemoryError()
 * 
 *     def add_range(self, first: int, last: int):             # <<<<<<<<<<<<<<
 *         hb_set_add_range(self._hb_set, first, last)
 *         if not hb_set_allocation_successful(self._hb_set):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.add_range", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2636
 *             raise MemoryError()
 * 
 *     def remove(self, c: int):             # <<<<<<<<<<<<<<
 *         if not c in self:
 *             raise KeyError, c
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_23remove(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_23remove = {"remove", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_23remove, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_23remove(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_c = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("remove (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_c,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_c)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2636, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "remove") < 0)) __PYX_ERR(0, 2636, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_c = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("remove", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2636, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_c), (&PyInt_Type), 0, "c", 1))) __PYX_ERR(0, 2636, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_22remove(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self), __pyx_v_c);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_22remove(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  hb_codepoint_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("remove", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2637
 * 
 *     def remove(self, c: int):
 *         if not c in self:             # <<<<<<<<<<<<<<
 *             raise KeyError, c
 *         hb_set_del(self._hb_set, c)
 */
  __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_c, ((PyObject *)__pyx_v_self), Py_NE)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 2637, __pyx_L1_error)
  if (unlikely(__pyx_t_1)) {

    /* "uharfbuzz/_harfbuzz.pyx":2638
 *     def remove(self, c: int):
 *         if not c in self:
 *             raise KeyError, c             # <<<<<<<<<<<<<<
 *         hb_set_del(self._hb_set, c)
 *         if not hb_set_allocation_successful(self._hb_set):
 */
    __Pyx_Raise(__pyx_builtin_KeyError, __pyx_v_c, 0, 0);
    __PYX_ERR(0, 2638, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":2637
 * 
 *     def remove(self, c: int):
 *         if not c in self:             # <<<<<<<<<<<<<<
 *             raise KeyError, c
 *         hb_set_del(self._hb_set, c)
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2639
 *         if not c in self:
 *             raise KeyError, c
 *         hb_set_del(self._hb_set, c)             # <<<<<<<<<<<<<<
 *         if not hb_set_allocation_successful(self._hb_set):
 *             raise MemoryError()
 */
  __pyx_t_2 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_c); if (unlikely((__pyx_t_2 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 2639, __pyx_L1_error)
  hb_set_del(__pyx_v_self->_hb_set, __pyx_t_2);

  /* "uharfbuzz/_harfbuzz.pyx":2640
 *             raise KeyError, c
 *         hb_set_del(self._hb_set, c)
 *         if not hb_set_allocation_successful(self._hb_set):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_1 = (!(hb_set_allocation_successful(__pyx_v_self->_hb_set) != 0));
  if (unlikely(__pyx_t_1)) {

    /* "uharfbuzz/_harfbuzz.pyx":2641
 *         hb_set_del(self._hb_set, c)
 *         if not hb_set_allocation_successful(self._hb_set):
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *     def discard(self, c: int):
 */
    PyErr_NoMemory(); __PYX_ERR(0, 2641, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":2640
 *             raise KeyError, c
 *         hb_set_del(self._hb_set, c)
 *         if not hb_set_allocation_successful(self._hb_set):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2636
 *             raise MemoryError()
 * 
 *     def remove(self, c: int):             # <<<<<<<<<<<<<<
 *         if not c in self:
 *             raise KeyError, c
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2643
 *             raise MemoryError()
 * 
 *     def discard(self, c: int):             # <<<<<<<<<<<<<<
 *         hb_set_del(self._hb_set, c)
 *         if not hb_set_allocation_successful(self._hb_set):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_25discard(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_25discard = {"discard", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_25discard, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_25discard(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_c = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("discard (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_c,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_c)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2643, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "discard") < 0)) __PYX_ERR(0, 2643, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_c = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("discard", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2643, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.discard", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_c), (&PyInt_Type), 0, "c", 1))) __PYX_ERR(0, 2643, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_24discard(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self), __pyx_v_c);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_24discard(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  hb_codepoint_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("discard", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2644
 * 
 *     def discard(self, c: int):
 *         hb_set_del(self._hb_set, c)             # <<<<<<<<<<<<<<
 *         if not hb_set_allocation_successful(self._hb_set):
 *             raise MemoryError()
 */
  __pyx_t_1 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_c); if (unlikely((__pyx_t_1 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 2644, __pyx_L1_error)
  hb_set_del(__pyx_v_self->_hb_set, __pyx_t_1);

  /* "uharfbuzz/_harfbuzz.pyx":2645
 *     def discard(self, c: int):
 *         hb_set_del(self._hb_set, c)
 *         if not hb_set_allocation_successful(self._hb_set):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_2 = (!(hb_set_allocation_successful(__pyx_v_self->_hb_set) != 0));
  if (unlikely(__pyx_t_2)) {

    /* "uharfbuzz/_harfbuzz.pyx":2646
 *         hb_set_del(self._hb_set, c)
 *         if not hb_set_allocation_successful(self._hb_set):
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *     def del_range(self, first: int, last: int):
 */
    PyErr_NoMemory(); __PYX_ERR(0, 2646, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":2645
 *     def discard(self, c: int):
 *         hb_set_del(self._hb_set, c)
 *         if not hb_set_allocation_successful(self._hb_set):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2643
 *             raise MemoryError()
 * 
 *     def discard(self, c: int):             # <<<<<<<<<<<<<<
 *         hb_set_del(self._hb_set, c)
 *         if not hb_set_allocation_successful(self._hb_set):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.discard", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2648
 *             raise MemoryError()
 * 
 *     def del_range(self, first: int, last: int):             # <<<<<<<<<<<<<<
 *         hb_set_del_range(self._hb_set, first, last)
 *         if not hb_set_allocation_successful(self._hb_set):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_27del_range(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_27del_range = {"del_range", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_27del_range, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_27del_range(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_first = 0;
  PyObject *__pyx_v_last = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("del_range (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_first,&__pyx_n_s_last,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_first)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2648, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_last)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2648, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("del_range", 1, 2, 2, 1); __PYX_ERR(0, 2648, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "del_range") < 0)) __PYX_ERR(0, 2648, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_first = ((PyObject*)values[0]);
    __pyx_v_last = ((PyObject*)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("del_range", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 2648, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.del_range", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_first), (&PyInt_Type), 0, "first", 1))) __PYX_ERR(0, 2648, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_last), (&PyInt_Type), 0, "last", 1))) __PYX_ERR(0, 2648, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_26del_range(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self), __pyx_v_first, __pyx_v_last);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_26del_range(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, PyObject *__pyx_v_first, PyObject *__pyx_v_last) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  hb_codepoint_t __pyx_t_1;
  hb_codepoint_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("del_range", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2649
 * 
 *     def del_range(self, first: int, last: int):
 *         hb_set_del_range(self._hb_set, first, last)             # <<<<<<<<<<<<<<
 *         if not hb_set_allocation_successful(self._hb_set):
 *             raise MemoryError()
 */
  __pyx_t_1 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_first); if (unlikely((__pyx_t_1 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 2649, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_last); if (unlikely((__pyx_t_2 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 2649, __pyx_L1_error)
  hb_set_del_range(__pyx_v_self->_hb_set, __pyx_t_1, __pyx_t_2);

  /* "uharfbuzz/_harfbuzz.pyx":2650
 *     def del_range(self, first: int, last: int):
 *         hb_set_del_range(self._hb_set, first, last)
 *         if not hb_set_allocation_successful(self._hb_set):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_3 = (!(hb_set_allocation_successful(__pyx_v_self->_hb_set) != 0));
  if (unlikely(__pyx_t_3)) {

    /* "uharfbuzz/_harfbuzz.pyx":2651
 *         hb_set_del_range(self._hb_set, first, last)
 *         if not hb_set_allocation_successful(self._hb_set):
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *     def _is_equal(self, other: Set) -> bool:
 */
    PyErr_NoMemory(); __PYX_ERR(0, 2651, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":2650
 *     def del_range(self, first: int, last: int):
 *         hb_set_del_range(self._hb_set, first, last)
 *         if not hb_set_allocation_successful(self._hb_set):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2648
 *             raise MemoryError()
 * 
 *     def del_range(self, first: int, last: int):             # <<<<<<<<<<<<<<
 *         hb_set_del_range(self._hb_set, first, last)
 *         if not hb_set_allocation_successful(self._hb_set):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.del_range", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2653
 *             raise MemoryError()
 * 
 *     def _is_equal(self, other: Set) -> bool:             # <<<<<<<<<<<<<<
 *         return hb_set_is_equal(self._hb_set, other._hb_set)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_29_is_equal(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_29_is_equal = {"_is_equal", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_29_is_equal, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_29_is_equal(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_other = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_is_equal (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_other,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_other)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2653, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_is_equal") < 0)) __PYX_ERR(0, 2653, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_other = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_is_equal", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2653, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set._is_equal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_9uharfbuzz_9_harfbuzz_Set, 0, "other", 0))) __PYX_ERR(0, 2653, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_28_is_equal(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self), __pyx_v_other);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_28_is_equal(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_is_equal", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2654
 * 
 *     def _is_equal(self, other: Set) -> bool:
 *         return hb_set_is_equal(self._hb_set, other._hb_set)             # <<<<<<<<<<<<<<
 * 
 *     def __eq__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_hb_bool_t(hb_set_is_equal(__pyx_v_self->_hb_set, __pyx_v_other->_hb_set)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2653
 *             raise MemoryError()
 * 
 *     def _is_equal(self, other: Set) -> bool:             # <<<<<<<<<<<<<<
 *         return hb_set_is_equal(self._hb_set, other._hb_set)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set._is_equal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2656
 *         return hb_set_is_equal(self._hb_set, other._hb_set)
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         if type(other) != Set:
 *             return NotImplemented
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_31__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_31__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__eq__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_30__eq__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_30__eq__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__eq__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2657
 * 
 *     def __eq__(self, other):
 *         if type(other) != Set:             # <<<<<<<<<<<<<<
 *             return NotImplemented
 *         return self._is_equal(other)
 */
  __pyx_t_1 = PyObject_RichCompare(((PyObject *)Py_TYPE(__pyx_v_other)), ((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Set), Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2657, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 2657, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "uharfbuzz/_harfbuzz.pyx":2658
 *     def __eq__(self, other):
 *         if type(other) != Set:
 *             return NotImplemented             # <<<<<<<<<<<<<<
 *         return self._is_equal(other)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_builtin_NotImplemented);
    __pyx_r = __pyx_builtin_NotImplemented;
    goto __pyx_L0;

    /* "uharfbuzz/_harfbuzz.pyx":2657
 * 
 *     def __eq__(self, other):
 *         if type(other) != Set:             # <<<<<<<<<<<<<<
 *             return NotImplemented
 *         return self._is_equal(other)
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2659
 *         if type(other) != Set:
 *             return NotImplemented
 *         return self._is_equal(other)             # <<<<<<<<<<<<<<
 * 
 *     def issubset(self, larger_set: Set) -> bool:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_is_equal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_other};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2659, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2656
 *         return hb_set_is_equal(self._hb_set, other._hb_set)
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         if type(other) != Set:
 *             return NotImplemented
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.__eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2661
 *         return self._is_equal(other)
 * 
 *     def issubset(self, larger_set: Set) -> bool:             # <<<<<<<<<<<<<<
 *         return hb_set_is_subset(self._hb_set, larger_set._hb_set)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_33issubset(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_33issubset = {"issubset", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_33issubset, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_33issubset(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_larger_set = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("issubset (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_larger_set,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_larger_set)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2661, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "issubset") < 0)) __PYX_ERR(0, 2661, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_larger_set = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("issubset", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2661, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.issubset", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_larger_set), __pyx_ptype_9uharfbuzz_9_harfbuzz_Set, 0, "larger_set", 0))) __PYX_ERR(0, 2661, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_32issubset(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self), __pyx_v_larger_set);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_32issubset(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_larger_set) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("issubset", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2662
 * 
 *     def issubset(self, larger_set: Set) -> bool:
 *         return hb_set_is_subset(self._hb_set, larger_set._hb_set)             # <<<<<<<<<<<<<<
 * 
 *     def issuperset(self, smaller_set: Set) -> bool:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_hb_bool_t(hb_set_is_subset(__pyx_v_self->_hb_set, __pyx_v_larger_set->_hb_set)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2661
 *         return self._is_equal(other)
 * 
 *     def issubset(self, larger_set: Set) -> bool:             # <<<<<<<<<<<<<<
 *         return hb_set_is_subset(self._hb_set, larger_set._hb_set)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.issubset", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2664
 *         return hb_set_is_subset(self._hb_set, larger_set._hb_set)
 * 
 *     def issuperset(self, smaller_set: Set) -> bool:             # <<<<<<<<<<<<<<
 *         return hb_set_is_subset(smaller_set._hb_set, self._hb_set)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_35issuperset(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_35issuperset = {"issuperset", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_35issuperset, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_35issuperset(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_smaller_set = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("issuperset (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_smaller_set,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_smaller_set)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2664, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "issuperset") < 0)) __PYX_ERR(0, 2664, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_smaller_set = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("issuperset", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2664, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.issuperset", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_smaller_set), __pyx_ptype_9uharfbuzz_9_harfbuzz_Set, 0, "smaller_set", 0))) __PYX_ERR(0, 2664, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_34issuperset(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self), __pyx_v_smaller_set);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_34issuperset(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_smaller_set) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("issuperset", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2665
 * 
 *     def issuperset(self, smaller_set: Set) -> bool:
 *         return hb_set_is_subset(smaller_set._hb_set, self._hb_set)             # <<<<<<<<<<<<<<
 * 
 *     def _set(self, other: Set):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_hb_bool_t(hb_set_is_subset(__pyx_v_smaller_set->_hb_set, __pyx_v_self->_hb_set)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2665, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2664
 *         return hb_set_is_subset(self._hb_set, larger_set._hb_set)
 * 
 *     def issuperset(self, smaller_set: Set) -> bool:             # <<<<<<<<<<<<<<
 *         return hb_set_is_subset(smaller_set._hb_set, self._hb_set)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.issuperset", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2667
 *         return hb_set_is_subset(smaller_set._hb_set, self._hb_set)
 * 
 *     def _set(self, other: Set):             # <<<<<<<<<<<<<<
 *         hb_set_set(self._hb_set, other._hb_set)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_37_set(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_37_set = {"_set", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_37_set, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_37_set(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_other = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_set (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_other,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_other)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2667, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_set") < 0)) __PYX_ERR(0, 2667, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_other = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_set", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2667, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set._set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_9uharfbuzz_9_harfbuzz_Set, 0, "other", 0))) __PYX_ERR(0, 2667, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_36_set(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self), __pyx_v_other);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_36_set(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_set", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2668
 * 
 *     def _set(self, other: Set):
 *         hb_set_set(self._hb_set, other._hb_set)             # <<<<<<<<<<<<<<
 * 
 *     def set(self, other):
 */
  hb_set_set(__pyx_v_self->_hb_set, __pyx_v_other->_hb_set);

  /* "uharfbuzz/_harfbuzz.pyx":2667
 *         return hb_set_is_subset(smaller_set._hb_set, self._hb_set)
 * 
 *     def _set(self, other: Set):             # <<<<<<<<<<<<<<
 *         hb_set_set(self._hb_set, other._hb_set)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2670
 *         hb_set_set(self._hb_set, other._hb_set)
 * 
 *     def set(self, other):             # <<<<<<<<<<<<<<
 *         if type(other) == Set:
 *             self._set(other)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_39set(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_39set = {"set", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_39set, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_39set(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_other = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_other,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_other)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2670, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set") < 0)) __PYX_ERR(0, 2670, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_other = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2670, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_38set(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self), __pyx_v_other);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_38set(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_v_c = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  hb_codepoint_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2671
 * 
 *     def set(self, other):
 *         if type(other) == Set:             # <<<<<<<<<<<<<<
 *             self._set(other)
 *         else:
 */
  __pyx_t_1 = PyObject_RichCompare(((PyObject *)Py_TYPE(__pyx_v_other)), ((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Set), Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2671, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 2671, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "uharfbuzz/_harfbuzz.pyx":2672
 *     def set(self, other):
 *         if type(other) == Set:
 *             self._set(other)             # <<<<<<<<<<<<<<
 *         else:
 *             for c in other:
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2672, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_other};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2672, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uharfbuzz/_harfbuzz.pyx":2671
 * 
 *     def set(self, other):
 *         if type(other) == Set:             # <<<<<<<<<<<<<<
 *             self._set(other)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "uharfbuzz/_harfbuzz.pyx":2674
 *             self._set(other)
 *         else:
 *             for c in other:             # <<<<<<<<<<<<<<
 *                 hb_set_add(self._hb_set, c)
 * 
 */
  /*else*/ {
    if (likely(PyList_CheckExact(__pyx_v_other)) || PyTuple_CheckExact(__pyx_v_other)) {
      __pyx_t_1 = __pyx_v_other; __Pyx_INCREF(__pyx_t_1);
      __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_other); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2674, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2674, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 2674, __pyx_L1_error)
            #endif
            if (__pyx_t_6 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 2674, __pyx_L1_error)
          #else
          __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2674, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 2674, __pyx_L1_error)
            #endif
            if (__pyx_t_6 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 2674, __pyx_L1_error)
          #else
          __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2674, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_7(__pyx_t_1);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 2674, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_XDECREF_SET(__pyx_v_c, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "uharfbuzz/_harfbuzz.pyx":2675
 *         else:
 *             for c in other:
 *                 hb_set_add(self._hb_set, c)             # <<<<<<<<<<<<<<
 * 
 *         if not hb_set_allocation_successful(self._hb_set):
 */
      __pyx_t_8 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_c); if (unlikely((__pyx_t_8 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 2675, __pyx_L1_error)
      hb_set_add(__pyx_v_self->_hb_set, __pyx_t_8);

      /* "uharfbuzz/_harfbuzz.pyx":2674
 *             self._set(other)
 *         else:
 *             for c in other:             # <<<<<<<<<<<<<<
 *                 hb_set_add(self._hb_set, c)
 * 
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "uharfbuzz/_harfbuzz.pyx":2677
 *                 hb_set_add(self._hb_set, c)
 * 
 *         if not hb_set_allocation_successful(self._hb_set):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_2 = (!(hb_set_allocation_successful(__pyx_v_self->_hb_set) != 0));
  if (unlikely(__pyx_t_2)) {

    /* "uharfbuzz/_harfbuzz.pyx":2678
 * 
 *         if not hb_set_allocation_successful(self._hb_set):
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *     def _update(self, other: Set):
 */
    PyErr_NoMemory(); __PYX_ERR(0, 2678, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":2677
 *                 hb_set_add(self._hb_set, c)
 * 
 *         if not hb_set_allocation_successful(self._hb_set):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2670
 *         hb_set_set(self._hb_set, other._hb_set)
 * 
 *     def set(self, other):             # <<<<<<<<<<<<<<
 *         if type(other) == Set:
 *             self._set(other)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.set", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_c);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2680
 *             raise MemoryError()
 * 
 *     def _update(self, other: Set):             # <<<<<<<<<<<<<<
 *         hb_set_union(self._hb_set, other._hb_set)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_41_update(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_41_update = {"_update", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_41_update, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_41_update(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_other = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_update (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_other,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_other)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2680, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_update") < 0)) __PYX_ERR(0, 2680, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_other = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_update", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2680, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set._update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_9uharfbuzz_9_harfbuzz_Set, 0, "other", 0))) __PYX_ERR(0, 2680, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_40_update(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self), __pyx_v_other);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_40_update(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_update", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2681
 * 
 *     def _update(self, other: Set):
 *         hb_set_union(self._hb_set, other._hb_set)             # <<<<<<<<<<<<<<
 * 
 *     def update(self, other):
 */
  hb_set_union(__pyx_v_self->_hb_set, __pyx_v_other->_hb_set);

  /* "uharfbuzz/_harfbuzz.pyx":2680
 *             raise MemoryError()
 * 
 *     def _update(self, other: Set):             # <<<<<<<<<<<<<<
 *         hb_set_union(self._hb_set, other._hb_set)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2683
 *         hb_set_union(self._hb_set, other._hb_set)
 * 
 *     def update(self, other):             # <<<<<<<<<<<<<<
 *         if type(other) == Set:
 *             self._update(other)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_43update(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_43update = {"update", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_43update, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_43update(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_other = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("update (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_other,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_other)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2683, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "update") < 0)) __PYX_ERR(0, 2683, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_other = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("update", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2683, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_42update(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self), __pyx_v_other);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_42update(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_v_c = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  hb_codepoint_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("update", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2684
 * 
 *     def update(self, other):
 *         if type(other) == Set:             # <<<<<<<<<<<<<<
 *             self._update(other)
 *         else:
 */
  __pyx_t_1 = PyObject_RichCompare(((PyObject *)Py_TYPE(__pyx_v_other)), ((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Set), Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2684, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 2684, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "uharfbuzz/_harfbuzz.pyx":2685
 *     def update(self, other):
 *         if type(other) == Set:
 *             self._update(other)             # <<<<<<<<<<<<<<
 *         else:
 *             for c in other:
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_update); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2685, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_other};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2685, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uharfbuzz/_harfbuzz.pyx":2684
 * 
 *     def update(self, other):
 *         if type(other) == Set:             # <<<<<<<<<<<<<<
 *             self._update(other)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "uharfbuzz/_harfbuzz.pyx":2687
 *             self._update(other)
 *         else:
 *             for c in other:             # <<<<<<<<<<<<<<
 *                 hb_set_add(self._hb_set, c)
 * 
 */
  /*else*/ {
    if (likely(PyList_CheckExact(__pyx_v_other)) || PyTuple_CheckExact(__pyx_v_other)) {
      __pyx_t_1 = __pyx_v_other; __Pyx_INCREF(__pyx_t_1);
      __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_other); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2687, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2687, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 2687, __pyx_L1_error)
            #endif
            if (__pyx_t_6 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 2687, __pyx_L1_error)
          #else
          __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2687, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 2687, __pyx_L1_error)
            #endif
            if (__pyx_t_6 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_6); __Pyx_INCREF(__pyx_t_3); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 2687, __pyx_L1_error)
          #else
          __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2687, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_7(__pyx_t_1);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 2687, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_XDECREF_SET(__pyx_v_c, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "uharfbuzz/_harfbuzz.pyx":2688
 *         else:
 *             for c in other:
 *                 hb_set_add(self._hb_set, c)             # <<<<<<<<<<<<<<
 * 
 *         if not hb_set_allocation_successful(self._hb_set):
 */
      __pyx_t_8 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_c); if (unlikely((__pyx_t_8 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 2688, __pyx_L1_error)
      hb_set_add(__pyx_v_self->_hb_set, __pyx_t_8);

      /* "uharfbuzz/_harfbuzz.pyx":2687
 *             self._update(other)
 *         else:
 *             for c in other:             # <<<<<<<<<<<<<<
 *                 hb_set_add(self._hb_set, c)
 * 
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "uharfbuzz/_harfbuzz.pyx":2690
 *                 hb_set_add(self._hb_set, c)
 * 
 *         if not hb_set_allocation_successful(self._hb_set):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_2 = (!(hb_set_allocation_successful(__pyx_v_self->_hb_set) != 0));
  if (unlikely(__pyx_t_2)) {

    /* "uharfbuzz/_harfbuzz.pyx":2691
 * 
 *         if not hb_set_allocation_successful(self._hb_set):
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *     def __ior__(self, other):
 */
    PyErr_NoMemory(); __PYX_ERR(0, 2691, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":2690
 *                 hb_set_add(self._hb_set, c)
 * 
 *         if not hb_set_allocation_successful(self._hb_set):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2683
 *         hb_set_union(self._hb_set, other._hb_set)
 * 
 *     def update(self, other):             # <<<<<<<<<<<<<<
 *         if type(other) == Set:
 *             self._update(other)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_c);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2693
 *             raise MemoryError()
 * 
 *     def __ior__(self, other):             # <<<<<<<<<<<<<<
 *         self.update(other)
 *         return self
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_45__ior__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_45__ior__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__ior__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_44__ior__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_44__ior__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__ior__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2694
 * 
 *     def __ior__(self, other):
 *         self.update(other)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_update_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2694, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_other};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2694, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2695
 *     def __ior__(self, other):
 *         self.update(other)
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def intersection_update(self, other: Set):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2693
 *             raise MemoryError()
 * 
 *     def __ior__(self, other):             # <<<<<<<<<<<<<<
 *         self.update(other)
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.__ior__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2697
 *         return self
 * 
 *     def intersection_update(self, other: Set):             # <<<<<<<<<<<<<<
 *         hb_set_intersect(self._hb_set, other._hb_set)
 *         if not hb_set_allocation_successful(self._hb_set):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_47intersection_update(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_47intersection_update = {"intersection_update", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_47intersection_update, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_47intersection_update(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_other = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("intersection_update (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_other,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_other)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2697, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "intersection_update") < 0)) __PYX_ERR(0, 2697, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_other = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("intersection_update", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2697, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.intersection_update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_9uharfbuzz_9_harfbuzz_Set, 0, "other", 0))) __PYX_ERR(0, 2697, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_46intersection_update(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self), __pyx_v_other);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_46intersection_update(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("intersection_update", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2698
 * 
 *     def intersection_update(self, other: Set):
 *         hb_set_intersect(self._hb_set, other._hb_set)             # <<<<<<<<<<<<<<
 *         if not hb_set_allocation_successful(self._hb_set):
 *             raise MemoryError()
 */
  hb_set_intersect(__pyx_v_self->_hb_set, __pyx_v_other->_hb_set);

  /* "uharfbuzz/_harfbuzz.pyx":2699
 *     def intersection_update(self, other: Set):
 *         hb_set_intersect(self._hb_set, other._hb_set)
 *         if not hb_set_allocation_successful(self._hb_set):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_1 = (!(hb_set_allocation_successful(__pyx_v_self->_hb_set) != 0));
  if (unlikely(__pyx_t_1)) {

    /* "uharfbuzz/_harfbuzz.pyx":2700
 *         hb_set_intersect(self._hb_set, other._hb_set)
 *         if not hb_set_allocation_successful(self._hb_set):
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *     def __iand__(self, other: Set):
 */
    PyErr_NoMemory(); __PYX_ERR(0, 2700, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":2699
 *     def intersection_update(self, other: Set):
 *         hb_set_intersect(self._hb_set, other._hb_set)
 *         if not hb_set_allocation_successful(self._hb_set):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2697
 *         return self
 * 
 *     def intersection_update(self, other: Set):             # <<<<<<<<<<<<<<
 *         hb_set_intersect(self._hb_set, other._hb_set)
 *         if not hb_set_allocation_successful(self._hb_set):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.intersection_update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2702
 *             raise MemoryError()
 * 
 *     def __iand__(self, other: Set):             # <<<<<<<<<<<<<<
 *         self.intersection_update(other)
 *         return self
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_49__iand__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_49__iand__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iand__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_9uharfbuzz_9_harfbuzz_Set, 0, "other", 0))) __PYX_ERR(0, 2702, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_48__iand__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self), ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_48__iand__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iand__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2703
 * 
 *     def __iand__(self, other: Set):
 *         self.intersection_update(other)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_intersection_update); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, ((PyObject *)__pyx_v_other)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2703, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2704
 *     def __iand__(self, other: Set):
 *         self.intersection_update(other)
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def difference_update(self, other: Set):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2702
 *             raise MemoryError()
 * 
 *     def __iand__(self, other: Set):             # <<<<<<<<<<<<<<
 *         self.intersection_update(other)
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.__iand__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2706
 *         return self
 * 
 *     def difference_update(self, other: Set):             # <<<<<<<<<<<<<<
 *         hb_set_subtract(self._hb_set, other._hb_set)
 *         if not hb_set_allocation_successful(self._hb_set):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_51difference_update(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_51difference_update = {"difference_update", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_51difference_update, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_51difference_update(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_other = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("difference_update (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_other,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_other)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2706, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "difference_update") < 0)) __PYX_ERR(0, 2706, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_other = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("difference_update", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2706, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.difference_update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_9uharfbuzz_9_harfbuzz_Set, 0, "other", 0))) __PYX_ERR(0, 2706, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_50difference_update(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self), __pyx_v_other);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_50difference_update(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("difference_update", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2707
 * 
 *     def difference_update(self, other: Set):
 *         hb_set_subtract(self._hb_set, other._hb_set)             # <<<<<<<<<<<<<<
 *         if not hb_set_allocation_successful(self._hb_set):
 *             raise MemoryError()
 */
  hb_set_subtract(__pyx_v_self->_hb_set, __pyx_v_other->_hb_set);

  /* "uharfbuzz/_harfbuzz.pyx":2708
 *     def difference_update(self, other: Set):
 *         hb_set_subtract(self._hb_set, other._hb_set)
 *         if not hb_set_allocation_successful(self._hb_set):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_1 = (!(hb_set_allocation_successful(__pyx_v_self->_hb_set) != 0));
  if (unlikely(__pyx_t_1)) {

    /* "uharfbuzz/_harfbuzz.pyx":2709
 *         hb_set_subtract(self._hb_set, other._hb_set)
 *         if not hb_set_allocation_successful(self._hb_set):
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *     def __isub__(self, other: Set):
 */
    PyErr_NoMemory(); __PYX_ERR(0, 2709, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":2708
 *     def difference_update(self, other: Set):
 *         hb_set_subtract(self._hb_set, other._hb_set)
 *         if not hb_set_allocation_successful(self._hb_set):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2706
 *         return self
 * 
 *     def difference_update(self, other: Set):             # <<<<<<<<<<<<<<
 *         hb_set_subtract(self._hb_set, other._hb_set)
 *         if not hb_set_allocation_successful(self._hb_set):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.difference_update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2711
 *             raise MemoryError()
 * 
 *     def __isub__(self, other: Set):             # <<<<<<<<<<<<<<
 *         self.difference_update(other)
 *         return self
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_53__isub__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_53__isub__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__isub__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_9uharfbuzz_9_harfbuzz_Set, 0, "other", 0))) __PYX_ERR(0, 2711, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_52__isub__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self), ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_52__isub__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__isub__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2712
 * 
 *     def __isub__(self, other: Set):
 *         self.difference_update(other)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_difference_update); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2712, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, ((PyObject *)__pyx_v_other)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2712, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2713
 *     def __isub__(self, other: Set):
 *         self.difference_update(other)
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def symmetric_difference_update(self, other: Set):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2711
 *             raise MemoryError()
 * 
 *     def __isub__(self, other: Set):             # <<<<<<<<<<<<<<
 *         self.difference_update(other)
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.__isub__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2715
 *         return self
 * 
 *     def symmetric_difference_update(self, other: Set):             # <<<<<<<<<<<<<<
 *         hb_set_symmetric_difference(self._hb_set, other._hb_set)
 *         if not hb_set_allocation_successful(self._hb_set):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_55symmetric_difference_update(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_55symmetric_difference_update = {"symmetric_difference_update", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_55symmetric_difference_update, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_55symmetric_difference_update(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_other = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("symmetric_difference_update (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_other,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_other)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2715, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "symmetric_difference_update") < 0)) __PYX_ERR(0, 2715, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_other = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("symmetric_difference_update", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2715, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.symmetric_difference_update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_9uharfbuzz_9_harfbuzz_Set, 0, "other", 0))) __PYX_ERR(0, 2715, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_54symmetric_difference_update(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self), __pyx_v_other);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_54symmetric_difference_update(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("symmetric_difference_update", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2716
 * 
 *     def symmetric_difference_update(self, other: Set):
 *         hb_set_symmetric_difference(self._hb_set, other._hb_set)             # <<<<<<<<<<<<<<
 *         if not hb_set_allocation_successful(self._hb_set):
 *             raise MemoryError()
 */
  hb_set_symmetric_difference(__pyx_v_self->_hb_set, __pyx_v_other->_hb_set);

  /* "uharfbuzz/_harfbuzz.pyx":2717
 *     def symmetric_difference_update(self, other: Set):
 *         hb_set_symmetric_difference(self._hb_set, other._hb_set)
 *         if not hb_set_allocation_successful(self._hb_set):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_1 = (!(hb_set_allocation_successful(__pyx_v_self->_hb_set) != 0));
  if (unlikely(__pyx_t_1)) {

    /* "uharfbuzz/_harfbuzz.pyx":2718
 *         hb_set_symmetric_difference(self._hb_set, other._hb_set)
 *         if not hb_set_allocation_successful(self._hb_set):
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *     def __ixor__(self, other: Set):
 */
    PyErr_NoMemory(); __PYX_ERR(0, 2718, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":2717
 *     def symmetric_difference_update(self, other: Set):
 *         hb_set_symmetric_difference(self._hb_set, other._hb_set)
 *         if not hb_set_allocation_successful(self._hb_set):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2715
 *         return self
 * 
 *     def symmetric_difference_update(self, other: Set):             # <<<<<<<<<<<<<<
 *         hb_set_symmetric_difference(self._hb_set, other._hb_set)
 *         if not hb_set_allocation_successful(self._hb_set):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.symmetric_difference_update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2720
 *             raise MemoryError()
 * 
 *     def __ixor__(self, other: Set):             # <<<<<<<<<<<<<<
 *         self.symmetric_difference_update(other)
 *         return self
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_57__ixor__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_57__ixor__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__ixor__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_9uharfbuzz_9_harfbuzz_Set, 0, "other", 0))) __PYX_ERR(0, 2720, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_56__ixor__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self), ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_other));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_56__ixor__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__ixor__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2721
 * 
 *     def __ixor__(self, other: Set):
 *         self.symmetric_difference_update(other)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_symmetric_difference_update); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2721, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, ((PyObject *)__pyx_v_other)};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2721, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2722
 *     def __ixor__(self, other: Set):
 *         self.symmetric_difference_update(other)
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self) -> int:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2720
 *             raise MemoryError()
 * 
 *     def __ixor__(self, other: Set):             # <<<<<<<<<<<<<<
 *         self.symmetric_difference_update(other)
 *         return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.__ixor__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2724
 *         return self
 * 
 *     def __len__(self) -> int:             # <<<<<<<<<<<<<<
 *         return hb_set_get_population(self._hb_set)
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_9uharfbuzz_9_harfbuzz_3Set_59__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_9uharfbuzz_9_harfbuzz_3Set_59__len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_58__len__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_58__len__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self) {
  Py_ssize_t __pyx_r;

  /* "uharfbuzz/_harfbuzz.pyx":2725
 * 
 *     def __len__(self) -> int:
 *         return hb_set_get_population(self._hb_set)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_r = hb_set_get_population(__pyx_v_self->_hb_set);
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2724
 *         return self
 * 
 *     def __len__(self) -> int:             # <<<<<<<<<<<<<<
 *         return hb_set_get_population(self._hb_set)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2727
 *         return hb_set_get_population(self._hb_set)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def min(self) -> int:
 *         return hb_set_get_min(self._hb_set)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_3min_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_3min_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_3min___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_3min___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2729
 *     @property
 *     def min(self) -> int:
 *         return hb_set_get_min(self._hb_set)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_hb_codepoint_t(hb_set_get_min(__pyx_v_self->_hb_set)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2729, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2727
 *         return hb_set_get_population(self._hb_set)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def min(self) -> int:
 *         return hb_set_get_min(self._hb_set)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.min.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2731
 *         return hb_set_get_min(self._hb_set)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max(self) -> int:
 *         return hb_set_get_max(self._hb_set)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_3max_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_3max_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_3max___get__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_3max___get__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2733
 *     @property
 *     def max(self) -> int:
 *         return hb_set_get_max(self._hb_set)             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_hb_codepoint_t(hb_set_get_max(__pyx_v_self->_hb_set)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2733, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2731
 *         return hb_set_get_min(self._hb_set)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max(self) -> int:
 *         return hb_set_get_max(self._hb_set)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.max.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2735
 *         return hb_set_get_max(self._hb_set)
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return SetIter(self)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_61__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_61__iter__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_60__iter__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_60__iter__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iter__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2736
 * 
 *     def __iter__(self):
 *         return SetIter(self)             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_SetIter), ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2735
 *         return hb_set_get_max(self._hb_set)
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return SetIter(self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2738
 *         return SetIter(self)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_inverted():
 *             return "Set({...})"
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_63__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_63__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_62__repr__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9uharfbuzz_9_harfbuzz_3Set_8__repr___2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uharfbuzz/_harfbuzz.pyx":2741
 *         if self.is_inverted():
 *             return "Set({...})"
 *         s = ', '.join(repr(v) for v in self)             # <<<<<<<<<<<<<<
 *         return ("Set({%s})" % s)
 * 
 */

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_8__repr___genexpr(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_genexpr_arg_0) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr *)__pyx_tp_new_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr(__pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 2741, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9uharfbuzz_9_harfbuzz_3Set_8__repr___2generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_repr___locals_genexpr, __pyx_n_s_uharfbuzz__harfbuzz); if (unlikely(!gen)) __PYX_ERR(0, 2741, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.__repr__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9uharfbuzz_9_harfbuzz_3Set_8__repr___2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr *__pyx_cur_scope = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2741, __pyx_L1_error)
  __pyx_r = PyList_New(0); if (unlikely(!__pyx_r)) __PYX_ERR(0, 2741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(0, 2741, __pyx_L1_error) }
  if (likely(PyList_CheckExact(((PyObject *)__pyx_cur_scope->__pyx_genexpr_arg_0))) || PyTuple_CheckExact(((PyObject *)__pyx_cur_scope->__pyx_genexpr_arg_0))) {
    __pyx_t_1 = ((PyObject *)__pyx_cur_scope->__pyx_genexpr_arg_0); __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(((PyObject *)__pyx_cur_scope->__pyx_genexpr_arg_0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2741, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2741, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 2741, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 2741, __pyx_L1_error)
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2741, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 2741, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 2741, __pyx_L1_error)
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2741, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 2741, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_v);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_v, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = PyObject_Repr(__pyx_cur_scope->__pyx_v_v); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2741, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely(__Pyx_ListComp_Append(__pyx_r, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 2741, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2738
 *         return SetIter(self)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_inverted():
 *             return "Set({...})"
 */

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_62__repr__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self) {
  PyObject *__pyx_v_s = NULL;
  PyObject *__pyx_gb_9uharfbuzz_9_harfbuzz_3Set_8__repr___2generator = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2739
 * 
 *     def __repr__(self):
 *         if self.is_inverted():             # <<<<<<<<<<<<<<
 *             return "Set({...})"
 *         s = ', '.join(repr(v) for v in self)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_is_inverted); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2739, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 2739, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "uharfbuzz/_harfbuzz.pyx":2740
 *     def __repr__(self):
 *         if self.is_inverted():
 *             return "Set({...})"             # <<<<<<<<<<<<<<
 *         s = ', '.join(repr(v) for v in self)
 *         return ("Set({%s})" % s)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_kp_u_Set);
    __pyx_r = __pyx_kp_u_Set;
    goto __pyx_L0;

    /* "uharfbuzz/_harfbuzz.pyx":2739
 * 
 *     def __repr__(self):
 *         if self.is_inverted():             # <<<<<<<<<<<<<<
 *             return "Set({...})"
 *         s = ', '.join(repr(v) for v in self)
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2741
 *         if self.is_inverted():
 *             return "Set({...})"
 *         s = ', '.join(repr(v) for v in self)             # <<<<<<<<<<<<<<
 *         return ("Set({%s})" % s)
 * 
 */
  __pyx_t_1 = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_8__repr___genexpr(NULL, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_Generator_Next(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyUnicode_Join(__pyx_kp_u__19, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_s = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2742
 *             return "Set({...})"
 *         s = ', '.join(repr(v) for v in self)
 *         return ("Set({%s})" % s)             # <<<<<<<<<<<<<<
 * 
 * cdef class SetIter:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyUnicode_Format(__pyx_kp_u_Set_s, __pyx_v_s); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2742, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2738
 *         return SetIter(self)
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         if self.is_inverted():
 *             return "Set({...})"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XDECREF(__pyx_gb_9uharfbuzz_9_harfbuzz_3Set_8__repr___2generator);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_65__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_65__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_65__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_65__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_64__reduce_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_64__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_67__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_67__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_67__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_67__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Set_66__setstate_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Set_66__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Set.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2749
 *     cdef hb_codepoint_t _c
 * 
 *     def __cinit__(self, s: Set):             # <<<<<<<<<<<<<<
 *         self.s = s
 *         self._hb_set = s._hb_set
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_7SetIter_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_7SetIter_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_s = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_s,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_s)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2749, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(0, 2749, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_s = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2749, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SetIter.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_s), __pyx_ptype_9uharfbuzz_9_harfbuzz_Set, 0, "s", 0))) __PYX_ERR(0, 2749, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_7SetIter___cinit__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SetIter *)__pyx_v_self), __pyx_v_s);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_7SetIter___cinit__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SetIter *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *__pyx_v_s) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  hb_set_t *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  hb_codepoint_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2750
 * 
 *     def __cinit__(self, s: Set):
 *         self.s = s             # <<<<<<<<<<<<<<
 *         self._hb_set = s._hb_set
 *         self._c = s.INVALID_VALUE
 */
  __Pyx_INCREF((PyObject *)__pyx_v_s);
  __Pyx_GIVEREF((PyObject *)__pyx_v_s);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->s);
  __Pyx_DECREF((PyObject *)__pyx_v_self->s);
  __pyx_v_self->s = __pyx_v_s;

  /* "uharfbuzz/_harfbuzz.pyx":2751
 *     def __cinit__(self, s: Set):
 *         self.s = s
 *         self._hb_set = s._hb_set             # <<<<<<<<<<<<<<
 *         self._c = s.INVALID_VALUE
 * 
 */
  __pyx_t_1 = __pyx_v_s->_hb_set;
  __pyx_v_self->_hb_set = __pyx_t_1;

  /* "uharfbuzz/_harfbuzz.pyx":2752
 *         self.s = s
 *         self._hb_set = s._hb_set
 *         self._c = s.INVALID_VALUE             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_s), __pyx_n_s_INVALID_VALUE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_t_2); if (unlikely((__pyx_t_3 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 2752, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_self->_c = __pyx_t_3;

  /* "uharfbuzz/_harfbuzz.pyx":2749
 *     cdef hb_codepoint_t _c
 * 
 *     def __cinit__(self, s: Set):             # <<<<<<<<<<<<<<
 *         self.s = s
 *         self._hb_set = s._hb_set
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SetIter.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2754
 *         self._c = s.INVALID_VALUE
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_7SetIter_3__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_7SetIter_3__iter__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_7SetIter_2__iter__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SetIter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_7SetIter_2__iter__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SetIter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2755
 * 
 *     def __iter__(self):
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __next__(self) -> int:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2754
 *         self._c = s.INVALID_VALUE
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2757
 *         return self
 * 
 *     def __next__(self) -> int:             # <<<<<<<<<<<<<<
 *         ret = hb_set_next(self._hb_set, &self._c)
 *         if not ret:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_7SetIter_5__next__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_7SetIter_5__next__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__next__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_7SetIter_4__next__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SetIter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_7SetIter_4__next__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SetIter *__pyx_v_self) {
  hb_bool_t __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_error_without_exception = 0; /* StopIteration */
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__next__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2758
 * 
 *     def __next__(self) -> int:
 *         ret = hb_set_next(self._hb_set, &self._c)             # <<<<<<<<<<<<<<
 *         if not ret:
 *             raise StopIteration
 */
  __pyx_v_ret = hb_set_next(__pyx_v_self->_hb_set, (&__pyx_v_self->_c));

  /* "uharfbuzz/_harfbuzz.pyx":2759
 *     def __next__(self) -> int:
 *         ret = hb_set_next(self._hb_set, &self._c)
 *         if not ret:             # <<<<<<<<<<<<<<
 *             raise StopIteration
 *         return self._c
 */
  __pyx_t_1 = (!(__pyx_v_ret != 0));
  if (unlikely(__pyx_t_1)) {

    /* "uharfbuzz/_harfbuzz.pyx":2760
 *         ret = hb_set_next(self._hb_set, &self._c)
 *         if not ret:
 *             raise StopIteration             # <<<<<<<<<<<<<<
 *         return self._c
 * 
 */
    __pyx_error_without_exception = 1;
    goto __pyx_L1_error;;

    /* "uharfbuzz/_harfbuzz.pyx":2759
 *     def __next__(self) -> int:
 *         ret = hb_set_next(self._hb_set, &self._c)
 *         if not ret:             # <<<<<<<<<<<<<<
 *             raise StopIteration
 *         return self._c
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2761
 *         if not ret:
 *             raise StopIteration
 *         return self._c             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_hb_codepoint_t(__pyx_v_self->_c); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2761, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2757
 *         return self
 * 
 *     def __next__(self) -> int:             # <<<<<<<<<<<<<<
 *         ret = hb_set_next(self._hb_set, &self._c)
 *         if not ret:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  if (!__pyx_error_without_exception) {
    __Pyx_AddTraceback("uharfbuzz._harfbuzz.SetIter.__next__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  }
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_7SetIter_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_7SetIter_7__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_7SetIter_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_7SetIter_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_7SetIter_6__reduce_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SetIter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_7SetIter_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_SetIter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SetIter.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_7SetIter_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_7SetIter_9__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_7SetIter_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_7SetIter_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SetIter.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_7SetIter_8__setstate_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SetIter *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_7SetIter_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_SetIter *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.SetIter.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2769
 *     INVALID_VALUE = HB_MAP_VALUE_INVALID
 * 
 *     def __cinit__(self, init = dict()):             # <<<<<<<<<<<<<<
 *         self._hb_map = hb_map_create()
 *         if not hb_map_allocation_successful(self._hb_map):
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_3Map_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_3Map_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_init = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_init,0};
    values[0] = __Pyx_Arg_NewRef_VARARGS(__pyx_k__20);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_init);
          if (value) { values[0] = __Pyx_Arg_NewRef_VARARGS(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2769, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(0, 2769, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_init = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 2769, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Map.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Map___cinit__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *)__pyx_v_self), __pyx_v_init);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_3Map___cinit__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self, PyObject *__pyx_v_init) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2770
 * 
 *     def __cinit__(self, init = dict()):
 *         self._hb_map = hb_map_create()             # <<<<<<<<<<<<<<
 *         if not hb_map_allocation_successful(self._hb_map):
 *             raise MemoryError()
 */
  __pyx_v_self->_hb_map = hb_map_create();

  /* "uharfbuzz/_harfbuzz.pyx":2771
 *     def __cinit__(self, init = dict()):
 *         self._hb_map = hb_map_create()
 *         if not hb_map_allocation_successful(self._hb_map):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_1 = (!(hb_map_allocation_successful(__pyx_v_self->_hb_map) != 0));
  if (unlikely(__pyx_t_1)) {

    /* "uharfbuzz/_harfbuzz.pyx":2772
 *         self._hb_map = hb_map_create()
 *         if not hb_map_allocation_successful(self._hb_map):
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *         self.update(init)
 */
    PyErr_NoMemory(); __PYX_ERR(0, 2772, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":2771
 *     def __cinit__(self, init = dict()):
 *         self._hb_map = hb_map_create()
 *         if not hb_map_allocation_successful(self._hb_map):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2774
 *             raise MemoryError()
 * 
 *         self.update(init)             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_update_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_init};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2774, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2769
 *     INVALID_VALUE = HB_MAP_VALUE_INVALID
 * 
 *     def __cinit__(self, init = dict()):             # <<<<<<<<<<<<<<
 *         self._hb_map = hb_map_create()
 *         if not hb_map_allocation_successful(self._hb_map):
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Map.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2776
 *         self.update(init)
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         hb_map_destroy(self._hb_map)
 * 
 */

/* Python wrapper */
static void __pyx_pw_9uharfbuzz_9_harfbuzz_3Map_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_9uharfbuzz_9_harfbuzz_3Map_3__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_2__dealloc__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_2__dealloc__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self) {

  /* "uharfbuzz/_harfbuzz.pyx":2777
 * 
 *     def __dealloc__(self):
 *         hb_map_destroy(self._hb_map)             # <<<<<<<<<<<<<<
 * 
 *     @staticmethod
 */
  hb_map_destroy(__pyx_v_self->_hb_map);

  /* "uharfbuzz/_harfbuzz.pyx":2776
 *         self.update(init)
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         hb_map_destroy(self._hb_map)
 * 
 */

  /* function exit code */
}

/* "uharfbuzz/_harfbuzz.pyx":2780
 * 
 *     @staticmethod
 *     cdef Map from_ptr(hb_map_t* hb_map):             # <<<<<<<<<<<<<<
 *         """Create Map from a pointer taking ownership of a it."""
 * 
 */

static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_f_9uharfbuzz_9_harfbuzz_3Map_from_ptr(hb_map_t *__pyx_v_hb_map) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_wrapper = 0;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("from_ptr", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2783
 *         """Create Map from a pointer taking ownership of a it."""
 * 
 *         cdef Map wrapper = Map.__new__(Map)             # <<<<<<<<<<<<<<
 *         wrapper._hb_map = hb_map
 *         return wrapper
 */
  __pyx_t_1 = ((PyObject *)__pyx_tp_new_9uharfbuzz_9_harfbuzz_Map(((PyTypeObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Map), __pyx_empty_tuple, NULL)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2783, __pyx_L1_error)
  __Pyx_GOTREF((PyObject *)__pyx_t_1);
  __pyx_v_wrapper = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2784
 * 
 *         cdef Map wrapper = Map.__new__(Map)
 *         wrapper._hb_map = hb_map             # <<<<<<<<<<<<<<
 *         return wrapper
 * 
 */
  __pyx_v_wrapper->_hb_map = __pyx_v_hb_map;

  /* "uharfbuzz/_harfbuzz.pyx":2785
 *         cdef Map wrapper = Map.__new__(Map)
 *         wrapper._hb_map = hb_map
 *         return wrapper             # <<<<<<<<<<<<<<
 * 
 *     def copy(self) -> Map:
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_wrapper);
  __pyx_r = __pyx_v_wrapper;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2780
 * 
 *     @staticmethod
 *     cdef Map from_ptr(hb_map_t* hb_map):             # <<<<<<<<<<<<<<
 *         """Create Map from a pointer taking ownership of a it."""
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Map.from_ptr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_wrapper);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2787
 *         return wrapper
 * 
 *     def copy(self) -> Map:             # <<<<<<<<<<<<<<
 *         c = Map()
 *         c._hb_map = hb_map_copy(self._hb_map)
 */

/* Python wrapper */
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_5copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Map_5copy = {"copy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_5copy, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_5copy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("copy", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "copy", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_4copy(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_4copy(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_c = NULL;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2788
 * 
 *     def copy(self) -> Map:
 *         c = Map()             # <<<<<<<<<<<<<<
 *         c._hb_map = hb_map_copy(self._hb_map)
 *         return c
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Map)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2788, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_c = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2789
 *     def copy(self) -> Map:
 *         c = Map()
 *         c._hb_map = hb_map_copy(self._hb_map)             # <<<<<<<<<<<<<<
 *         return c
 * 
 */
  __pyx_v_c->_hb_map = hb_map_copy(__pyx_v_self->_hb_map);

  /* "uharfbuzz/_harfbuzz.pyx":2790
 *         c = Map()
 *         c._hb_map = hb_map_copy(self._hb_map)
 *         return c             # <<<<<<<<<<<<<<
 * 
 *     def __copy__(self) -> Map:
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_c);
  __pyx_r = __pyx_v_c;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2787
 *         return wrapper
 * 
 *     def copy(self) -> Map:             # <<<<<<<<<<<<<<
 *         c = Map()
 *         c._hb_map = hb_map_copy(self._hb_map)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Map.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_c);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2792
 *         return c
 * 
 *     def __copy__(self) -> Map:             # <<<<<<<<<<<<<<
 *         return self.copy()
 * 
 */

/* Python wrapper */
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_7__copy__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Map_7__copy__ = {"__copy__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_7__copy__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_7__copy__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__copy__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__copy__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__copy__", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_6__copy__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_6__copy__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__copy__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2793
 * 
 *     def __copy__(self) -> Map:
 *         return self.copy()             # <<<<<<<<<<<<<<
 * 
 *     def _update(self, other : Map):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_copy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2793, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_9uharfbuzz_9_harfbuzz_Map))))) __PYX_ERR(0, 2793, __pyx_L1_error)
  __pyx_r = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2792
 *         return c
 * 
 *     def __copy__(self) -> Map:             # <<<<<<<<<<<<<<
 *         return self.copy()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Map.__copy__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2795
 *         return self.copy()
 * 
 *     def _update(self, other : Map):             # <<<<<<<<<<<<<<
 *         hb_map_update(self._hb_map, other._hb_map)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_9_update(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Map_9_update = {"_update", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_9_update, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_9_update(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_other = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_update (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_other,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_other)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2795, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_update") < 0)) __PYX_ERR(0, 2795, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_other = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_update", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2795, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Map._update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_9uharfbuzz_9_harfbuzz_Map, 0, "other", 0))) __PYX_ERR(0, 2795, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_8_update(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *)__pyx_v_self), __pyx_v_other);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_8_update(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_update", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2796
 * 
 *     def _update(self, other : Map):
 *         hb_map_update(self._hb_map, other._hb_map)             # <<<<<<<<<<<<<<
 * 
 *     def update(self, other):
 */
  hb_map_update(__pyx_v_self->_hb_map, __pyx_v_other->_hb_map);

  /* "uharfbuzz/_harfbuzz.pyx":2795
 *         return self.copy()
 * 
 *     def _update(self, other : Map):             # <<<<<<<<<<<<<<
 *         hb_map_update(self._hb_map, other._hb_map)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2798
 *         hb_map_update(self._hb_map, other._hb_map)
 * 
 *     def update(self, other):             # <<<<<<<<<<<<<<
 *         if type(other) == Map:
 *             self._update(other)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_11update(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Map_11update = {"update", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_11update, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_11update(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_other = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("update (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_other,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_other)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2798, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "update") < 0)) __PYX_ERR(0, 2798, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_other = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("update", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2798, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Map.update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_10update(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *)__pyx_v_self), __pyx_v_other);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_10update(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  hb_codepoint_t __pyx_t_10;
  hb_codepoint_t __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("update", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2799
 * 
 *     def update(self, other):
 *         if type(other) == Map:             # <<<<<<<<<<<<<<
 *             self._update(other)
 *         else:
 */
  __pyx_t_1 = PyObject_RichCompare(((PyObject *)Py_TYPE(__pyx_v_other)), ((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Map), Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2799, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 2799, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "uharfbuzz/_harfbuzz.pyx":2800
 *     def update(self, other):
 *         if type(other) == Map:
 *             self._update(other)             # <<<<<<<<<<<<<<
 *         else:
 *             for k,v in other.items():
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_update); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2800, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_other};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2800, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "uharfbuzz/_harfbuzz.pyx":2799
 * 
 *     def update(self, other):
 *         if type(other) == Map:             # <<<<<<<<<<<<<<
 *             self._update(other)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "uharfbuzz/_harfbuzz.pyx":2802
 *             self._update(other)
 *         else:
 *             for k,v in other.items():             # <<<<<<<<<<<<<<
 *                 hb_map_set(self._hb_map, k, v)
 * 
 */
  /*else*/ {
    __pyx_t_6 = 0;
    if (unlikely(__pyx_v_other == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(0, 2802, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_dict_iterator(__pyx_v_other, 0, __pyx_n_s_items, (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2802, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_1);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
    while (1) {
      __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_7, &__pyx_t_6, &__pyx_t_3, &__pyx_t_4, NULL, __pyx_t_8);
      if (unlikely(__pyx_t_9 == 0)) break;
      if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 2802, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_3);
      __pyx_t_3 = 0;
      __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "uharfbuzz/_harfbuzz.pyx":2803
 *         else:
 *             for k,v in other.items():
 *                 hb_map_set(self._hb_map, k, v)             # <<<<<<<<<<<<<<
 * 
 *         if not hb_map_allocation_successful(self._hb_map):
 */
      __pyx_t_10 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_k); if (unlikely((__pyx_t_10 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 2803, __pyx_L1_error)
      __pyx_t_11 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_v); if (unlikely((__pyx_t_11 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 2803, __pyx_L1_error)
      hb_map_set(__pyx_v_self->_hb_map, __pyx_t_10, __pyx_t_11);
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "uharfbuzz/_harfbuzz.pyx":2805
 *                 hb_map_set(self._hb_map, k, v)
 * 
 *         if not hb_map_allocation_successful(self._hb_map):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_2 = (!(hb_map_allocation_successful(__pyx_v_self->_hb_map) != 0));
  if (unlikely(__pyx_t_2)) {

    /* "uharfbuzz/_harfbuzz.pyx":2806
 * 
 *         if not hb_map_allocation_successful(self._hb_map):
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *     def clear(self):
 */
    PyErr_NoMemory(); __PYX_ERR(0, 2806, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":2805
 *                 hb_map_set(self._hb_map, k, v)
 * 
 *         if not hb_map_allocation_successful(self._hb_map):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2798
 *         hb_map_update(self._hb_map, other._hb_map)
 * 
 *     def update(self, other):             # <<<<<<<<<<<<<<
 *         if type(other) == Map:
 *             self._update(other)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Map.update", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2808
 *             raise MemoryError()
 * 
 *     def clear(self):             # <<<<<<<<<<<<<<
 *         hb_map_clear(self._hb_map)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_13clear(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Map_13clear = {"clear", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_13clear, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_13clear(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clear (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("clear", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "clear", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_12clear(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_12clear(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("clear", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2809
 * 
 *     def clear(self):
 *         hb_map_clear(self._hb_map)             # <<<<<<<<<<<<<<
 * 
 *     def __bool__(self) -> bool:
 */
  hb_map_clear(__pyx_v_self->_hb_map);

  /* "uharfbuzz/_harfbuzz.pyx":2808
 *             raise MemoryError()
 * 
 *     def clear(self):             # <<<<<<<<<<<<<<
 *         hb_map_clear(self._hb_map)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2811
 *         hb_map_clear(self._hb_map)
 * 
 *     def __bool__(self) -> bool:             # <<<<<<<<<<<<<<
 *         return not hb_map_is_empty(self._hb_map)
 * 
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_3Map_15__bool__(PyObject *__pyx_v_self); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_3Map_15__bool__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__bool__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_14__bool__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_14__bool__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self) {
  int __pyx_r;

  /* "uharfbuzz/_harfbuzz.pyx":2812
 * 
 *     def __bool__(self) -> bool:
 *         return not hb_map_is_empty(self._hb_map)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self) -> int:
 */
  __pyx_r = (!(hb_map_is_empty(__pyx_v_self->_hb_map) != 0));
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2811
 *         hb_map_clear(self._hb_map)
 * 
 *     def __bool__(self) -> bool:             # <<<<<<<<<<<<<<
 *         return not hb_map_is_empty(self._hb_map)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2814
 *         return not hb_map_is_empty(self._hb_map)
 * 
 *     def __len__(self) -> int:             # <<<<<<<<<<<<<<
 *         return hb_map_get_population(self._hb_map)
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_9uharfbuzz_9_harfbuzz_3Map_17__len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_pw_9uharfbuzz_9_harfbuzz_3Map_17__len__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_16__len__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_16__len__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self) {
  Py_ssize_t __pyx_r;

  /* "uharfbuzz/_harfbuzz.pyx":2815
 * 
 *     def __len__(self) -> int:
 *         return hb_map_get_population(self._hb_map)             # <<<<<<<<<<<<<<
 * 
 *     def _is_equal(self, other: Map) -> bool:
 */
  __pyx_r = hb_map_get_population(__pyx_v_self->_hb_map);
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2814
 *         return not hb_map_is_empty(self._hb_map)
 * 
 *     def __len__(self) -> int:             # <<<<<<<<<<<<<<
 *         return hb_map_get_population(self._hb_map)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2817
 *         return hb_map_get_population(self._hb_map)
 * 
 *     def _is_equal(self, other: Map) -> bool:             # <<<<<<<<<<<<<<
 *         return hb_map_is_equal(self._hb_map, other._hb_map)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_19_is_equal(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Map_19_is_equal = {"_is_equal", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_19_is_equal, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_19_is_equal(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_other = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_is_equal (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_other,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_other)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2817, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "_is_equal") < 0)) __PYX_ERR(0, 2817, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_other = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_is_equal", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2817, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Map._is_equal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_other), __pyx_ptype_9uharfbuzz_9_harfbuzz_Map, 0, "other", 0))) __PYX_ERR(0, 2817, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_18_is_equal(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *)__pyx_v_self), __pyx_v_other);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_18_is_equal(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_is_equal", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2818
 * 
 *     def _is_equal(self, other: Map) -> bool:
 *         return hb_map_is_equal(self._hb_map, other._hb_map)             # <<<<<<<<<<<<<<
 * 
 *     def __eq__(self, other):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_hb_bool_t(hb_map_is_equal(__pyx_v_self->_hb_map, __pyx_v_other->_hb_map)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2818, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2817
 *         return hb_map_get_population(self._hb_map)
 * 
 *     def _is_equal(self, other: Map) -> bool:             # <<<<<<<<<<<<<<
 *         return hb_map_is_equal(self._hb_map, other._hb_map)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Map._is_equal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2820
 *         return hb_map_is_equal(self._hb_map, other._hb_map)
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         if type(other) != Map:
 *             return NotImplemented
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_21__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_21__eq__(PyObject *__pyx_v_self, PyObject *__pyx_v_other) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__eq__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_20__eq__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *)__pyx_v_self), ((PyObject *)__pyx_v_other));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_20__eq__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self, PyObject *__pyx_v_other) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__eq__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2821
 * 
 *     def __eq__(self, other):
 *         if type(other) != Map:             # <<<<<<<<<<<<<<
 *             return NotImplemented
 *         return self._is_equal(other)
 */
  __pyx_t_1 = PyObject_RichCompare(((PyObject *)Py_TYPE(__pyx_v_other)), ((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Map), Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2821, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 2821, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "uharfbuzz/_harfbuzz.pyx":2822
 *     def __eq__(self, other):
 *         if type(other) != Map:
 *             return NotImplemented             # <<<<<<<<<<<<<<
 *         return self._is_equal(other)
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_builtin_NotImplemented);
    __pyx_r = __pyx_builtin_NotImplemented;
    goto __pyx_L0;

    /* "uharfbuzz/_harfbuzz.pyx":2821
 * 
 *     def __eq__(self, other):
 *         if type(other) != Map:             # <<<<<<<<<<<<<<
 *             return NotImplemented
 *         return self._is_equal(other)
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2823
 *         if type(other) != Map:
 *             return NotImplemented
 *         return self._is_equal(other)             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, k: int, v: int):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_is_equal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_other};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2823, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2820
 *         return hb_map_is_equal(self._hb_map, other._hb_map)
 * 
 *     def __eq__(self, other):             # <<<<<<<<<<<<<<
 *         if type(other) != Map:
 *             return NotImplemented
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Map.__eq__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2825
 *         return self._is_equal(other)
 * 
 *     def __setitem__(self, k: int, v: int):             # <<<<<<<<<<<<<<
 *         hb_map_set(self._hb_map, k, v)
 *         if not hb_map_allocation_successful(self._hb_map):
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_3Map_23__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_k, PyObject *__pyx_v_v); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_3Map_23__setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_k, PyObject *__pyx_v_v) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_k), (&PyInt_Type), 0, "k", 1))) __PYX_ERR(0, 2825, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_v), (&PyInt_Type), 0, "v", 1))) __PYX_ERR(0, 2825, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_22__setitem__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *)__pyx_v_self), ((PyObject*)__pyx_v_k), ((PyObject*)__pyx_v_v));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_22__setitem__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self, PyObject *__pyx_v_k, PyObject *__pyx_v_v) {
  int __pyx_r;
  hb_codepoint_t __pyx_t_1;
  hb_codepoint_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2826
 * 
 *     def __setitem__(self, k: int, v: int):
 *         hb_map_set(self._hb_map, k, v)             # <<<<<<<<<<<<<<
 *         if not hb_map_allocation_successful(self._hb_map):
 *             raise MemoryError()
 */
  __pyx_t_1 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_k); if (unlikely((__pyx_t_1 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 2826, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_v); if (unlikely((__pyx_t_2 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 2826, __pyx_L1_error)
  hb_map_set(__pyx_v_self->_hb_map, __pyx_t_1, __pyx_t_2);

  /* "uharfbuzz/_harfbuzz.pyx":2827
 *     def __setitem__(self, k: int, v: int):
 *         hb_map_set(self._hb_map, k, v)
 *         if not hb_map_allocation_successful(self._hb_map):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  __pyx_t_3 = (!(hb_map_allocation_successful(__pyx_v_self->_hb_map) != 0));
  if (unlikely(__pyx_t_3)) {

    /* "uharfbuzz/_harfbuzz.pyx":2828
 *         hb_map_set(self._hb_map, k, v)
 *         if not hb_map_allocation_successful(self._hb_map):
 *             raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *     def get(self, k: int):
 */
    PyErr_NoMemory(); __PYX_ERR(0, 2828, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":2827
 *     def __setitem__(self, k: int, v: int):
 *         hb_map_set(self._hb_map, k, v)
 *         if not hb_map_allocation_successful(self._hb_map):             # <<<<<<<<<<<<<<
 *             raise MemoryError()
 * 
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2825
 *         return self._is_equal(other)
 * 
 *     def __setitem__(self, k: int, v: int):             # <<<<<<<<<<<<<<
 *         hb_map_set(self._hb_map, k, v)
 *         if not hb_map_allocation_successful(self._hb_map):
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Map.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2830
 *             raise MemoryError()
 * 
 *     def get(self, k: int):             # <<<<<<<<<<<<<<
 *         if k < 0 or k >= self.INVALID_VALUE:
 *             return None
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_25get(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Map_25get = {"get", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_25get, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_25get(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_k = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_k,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_k)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2830, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get") < 0)) __PYX_ERR(0, 2830, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_k = ((PyObject*)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2830, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Map.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_k), (&PyInt_Type), 0, "k", 1))) __PYX_ERR(0, 2830, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_24get(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *)__pyx_v_self), __pyx_v_k);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_24get(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self, PyObject *__pyx_v_k) {
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  hb_codepoint_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2831
 * 
 *     def get(self, k: int):
 *         if k < 0 or k >= self.INVALID_VALUE:             # <<<<<<<<<<<<<<
 *             return None
 *         v = hb_map_get(self._hb_map, k)
 */
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_k, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2831, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 2831, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_INVALID_VALUE); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_k, __pyx_t_2, Py_GE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2831, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 2831, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":2832
 *     def get(self, k: int):
 *         if k < 0 or k >= self.INVALID_VALUE:
 *             return None             # <<<<<<<<<<<<<<
 *         v = hb_map_get(self._hb_map, k)
 *         if v == self.INVALID_VALUE:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "uharfbuzz/_harfbuzz.pyx":2831
 * 
 *     def get(self, k: int):
 *         if k < 0 or k >= self.INVALID_VALUE:             # <<<<<<<<<<<<<<
 *             return None
 *         v = hb_map_get(self._hb_map, k)
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2833
 *         if k < 0 or k >= self.INVALID_VALUE:
 *             return None
 *         v = hb_map_get(self._hb_map, k)             # <<<<<<<<<<<<<<
 *         if v == self.INVALID_VALUE:
 *             v = None
 */
  __pyx_t_5 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_k); if (unlikely((__pyx_t_5 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 2833, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyInt_From_hb_codepoint_t(hb_map_get(__pyx_v_self->_hb_map, __pyx_t_5)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_v = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2834
 *             return None
 *         v = hb_map_get(self._hb_map, k)
 *         if v == self.INVALID_VALUE:             # <<<<<<<<<<<<<<
 *             v = None
 *         return v
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_INVALID_VALUE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2834, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_v, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2834, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 2834, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_1) {

    /* "uharfbuzz/_harfbuzz.pyx":2835
 *         v = hb_map_get(self._hb_map, k)
 *         if v == self.INVALID_VALUE:
 *             v = None             # <<<<<<<<<<<<<<
 *         return v
 * 
 */
    __Pyx_INCREF(Py_None);
    __Pyx_DECREF_SET(__pyx_v_v, Py_None);

    /* "uharfbuzz/_harfbuzz.pyx":2834
 *             return None
 *         v = hb_map_get(self._hb_map, k)
 *         if v == self.INVALID_VALUE:             # <<<<<<<<<<<<<<
 *             v = None
 *         return v
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2836
 *         if v == self.INVALID_VALUE:
 *             v = None
 *         return v             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, k: int) -> int:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_v);
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2830
 *             raise MemoryError()
 * 
 *     def get(self, k: int):             # <<<<<<<<<<<<<<
 *         if k < 0 or k >= self.INVALID_VALUE:
 *             return None
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Map.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2838
 *         return v
 * 
 *     def __getitem__(self, k: int) -> int:             # <<<<<<<<<<<<<<
 *         v = self.get(k)
 *         if v is None:
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_27__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_k); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_27__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_k) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_k), (&PyInt_Type), 0, "k", 1))) __PYX_ERR(0, 2838, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_26__getitem__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *)__pyx_v_self), ((PyObject*)__pyx_v_k));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_26__getitem__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self, PyObject *__pyx_v_k) {
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2839
 * 
 *     def __getitem__(self, k: int) -> int:
 *         v = self.get(k)             # <<<<<<<<<<<<<<
 *         if v is None:
 *             raise KeyError, v
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2839, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_k};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2839, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_v = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2840
 *     def __getitem__(self, k: int) -> int:
 *         v = self.get(k)
 *         if v is None:             # <<<<<<<<<<<<<<
 *             raise KeyError, v
 *         return v
 */
  __pyx_t_5 = (__pyx_v_v == Py_None);
  if (unlikely(__pyx_t_5)) {

    /* "uharfbuzz/_harfbuzz.pyx":2841
 *         v = self.get(k)
 *         if v is None:
 *             raise KeyError, v             # <<<<<<<<<<<<<<
 *         return v
 * 
 */
    __Pyx_Raise(__pyx_builtin_KeyError, __pyx_v_v, 0, 0);
    __PYX_ERR(0, 2841, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":2840
 *     def __getitem__(self, k: int) -> int:
 *         v = self.get(k)
 *         if v is None:             # <<<<<<<<<<<<<<
 *             raise KeyError, v
 *         return v
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2842
 *         if v is None:
 *             raise KeyError, v
 *         return v             # <<<<<<<<<<<<<<
 * 
 *     def __contains__(self, k) -> bool:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_v);
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2838
 *         return v
 * 
 *     def __getitem__(self, k: int) -> int:             # <<<<<<<<<<<<<<
 *         v = self.get(k)
 *         if v is None:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Map.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2844
 *         return v
 * 
 *     def __contains__(self, k) -> bool:             # <<<<<<<<<<<<<<
 *         if type(k) != int:
 *             return False
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_3Map_29__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_k); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_3Map_29__contains__(PyObject *__pyx_v_self, PyObject *__pyx_v_k) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__contains__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_28__contains__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *)__pyx_v_self), ((PyObject *)__pyx_v_k));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_28__contains__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self, PyObject *__pyx_v_k) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  hb_codepoint_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__contains__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2845
 * 
 *     def __contains__(self, k) -> bool:
 *         if type(k) != int:             # <<<<<<<<<<<<<<
 *             return False
 *         if k < 0 or k >= self.INVALID_VALUE:
 */
  __pyx_t_1 = PyObject_RichCompare(((PyObject *)Py_TYPE(__pyx_v_k)), ((PyObject *)(&PyInt_Type)), Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2845, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 2845, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "uharfbuzz/_harfbuzz.pyx":2846
 *     def __contains__(self, k) -> bool:
 *         if type(k) != int:
 *             return False             # <<<<<<<<<<<<<<
 *         if k < 0 or k >= self.INVALID_VALUE:
 *             return False
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "uharfbuzz/_harfbuzz.pyx":2845
 * 
 *     def __contains__(self, k) -> bool:
 *         if type(k) != int:             # <<<<<<<<<<<<<<
 *             return False
 *         if k < 0 or k >= self.INVALID_VALUE:
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2847
 *         if type(k) != int:
 *             return False
 *         if k < 0 or k >= self.INVALID_VALUE:             # <<<<<<<<<<<<<<
 *             return False
 *         return hb_map_has(self._hb_map, k)
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_k, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2847, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 2847, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_INVALID_VALUE); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2847, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_k, __pyx_t_1, Py_GE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2847, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 2847, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_2 = __pyx_t_3;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_2) {

    /* "uharfbuzz/_harfbuzz.pyx":2848
 *             return False
 *         if k < 0 or k >= self.INVALID_VALUE:
 *             return False             # <<<<<<<<<<<<<<
 *         return hb_map_has(self._hb_map, k)
 * 
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "uharfbuzz/_harfbuzz.pyx":2847
 *         if type(k) != int:
 *             return False
 *         if k < 0 or k >= self.INVALID_VALUE:             # <<<<<<<<<<<<<<
 *             return False
 *         return hb_map_has(self._hb_map, k)
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2849
 *         if k < 0 or k >= self.INVALID_VALUE:
 *             return False
 *         return hb_map_has(self._hb_map, k)             # <<<<<<<<<<<<<<
 * 
 *     def __delitem__(self, c: int):
 */
  __pyx_t_5 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_k); if (unlikely((__pyx_t_5 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 2849, __pyx_L1_error)
  __pyx_r = hb_map_has(__pyx_v_self->_hb_map, __pyx_t_5);
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2844
 *         return v
 * 
 *     def __contains__(self, k) -> bool:             # <<<<<<<<<<<<<<
 *         if type(k) != int:
 *             return False
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Map.__contains__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2851
 *         return hb_map_has(self._hb_map, k)
 * 
 *     def __delitem__(self, c: int):             # <<<<<<<<<<<<<<
 *         if not c in self:
 *             raise KeyError, c
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_3Map_31__delitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_c); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_3Map_31__delitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_c) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__delitem__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_c), (&PyInt_Type), 0, "c", 1))) __PYX_ERR(0, 2851, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_30__delitem__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *)__pyx_v_self), ((PyObject*)__pyx_v_c));

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_30__delitem__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self, PyObject *__pyx_v_c) {
  int __pyx_r;
  int __pyx_t_1;
  hb_codepoint_t __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2852
 * 
 *     def __delitem__(self, c: int):
 *         if not c in self:             # <<<<<<<<<<<<<<
 *             raise KeyError, c
 *         hb_map_del(self._hb_map, c)
 */
  __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_v_c, ((PyObject *)__pyx_v_self), Py_NE)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 2852, __pyx_L1_error)
  if (unlikely(__pyx_t_1)) {

    /* "uharfbuzz/_harfbuzz.pyx":2853
 *     def __delitem__(self, c: int):
 *         if not c in self:
 *             raise KeyError, c             # <<<<<<<<<<<<<<
 *         hb_map_del(self._hb_map, c)
 * 
 */
    __Pyx_Raise(__pyx_builtin_KeyError, __pyx_v_c, 0, 0);
    __PYX_ERR(0, 2853, __pyx_L1_error)

    /* "uharfbuzz/_harfbuzz.pyx":2852
 * 
 *     def __delitem__(self, c: int):
 *         if not c in self:             # <<<<<<<<<<<<<<
 *             raise KeyError, c
 *         hb_map_del(self._hb_map, c)
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2854
 *         if not c in self:
 *             raise KeyError, c
 *         hb_map_del(self._hb_map, c)             # <<<<<<<<<<<<<<
 * 
 *     def items(self):
 */
  __pyx_t_2 = __Pyx_PyInt_As_hb_codepoint_t(__pyx_v_c); if (unlikely((__pyx_t_2 == ((hb_codepoint_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 2854, __pyx_L1_error)
  hb_map_del(__pyx_v_self->_hb_map, __pyx_t_2);

  /* "uharfbuzz/_harfbuzz.pyx":2851
 *         return hb_map_has(self._hb_map, k)
 * 
 *     def __delitem__(self, c: int):             # <<<<<<<<<<<<<<
 *         if not c in self:
 *             raise KeyError, c
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Map.__delitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2856
 *         hb_map_del(self._hb_map, c)
 * 
 *     def items(self):             # <<<<<<<<<<<<<<
 *         return MapIter(self)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_33items(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Map_33items = {"items", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_33items, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_33items(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("items (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("items", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "items", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_32items(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_32items(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("items", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2857
 * 
 *     def items(self):
 *         return MapIter(self)             # <<<<<<<<<<<<<<
 * 
 *     def keys(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_MapIter), ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2856
 *         hb_map_del(self._hb_map, c)
 * 
 *     def items(self):             # <<<<<<<<<<<<<<
 *         return MapIter(self)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Map.items", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2859
 *         return MapIter(self)
 * 
 *     def keys(self):             # <<<<<<<<<<<<<<
 *         return (k for k,v in self.items())
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_35keys(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Map_35keys = {"keys", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_35keys, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_35keys(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("keys (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("keys", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "keys", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_34keys(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9uharfbuzz_9_harfbuzz_3Map_4keys_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uharfbuzz/_harfbuzz.pyx":2860
 * 
 *     def keys(self):
 *         return (k for k,v in self.items())             # <<<<<<<<<<<<<<
 * 
 *     def values(self):
 */

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_4keys_genexpr(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_genexpr_arg_0) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr *)__pyx_tp_new_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr(__pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 2860, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9uharfbuzz_9_harfbuzz_3Map_4keys_2generator1, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_keys_locals_genexpr, __pyx_n_s_uharfbuzz__harfbuzz); if (unlikely(!gen)) __PYX_ERR(0, 2860, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Map.keys.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9uharfbuzz_9_harfbuzz_3Map_4keys_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2860, __pyx_L1_error)
  __pyx_t_2 = 0;
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(0, 2860, __pyx_L1_error) }
  if (unlikely(((PyObject *)__pyx_cur_scope->__pyx_genexpr_arg_0) == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
    __PYX_ERR(0, 2860, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_dict_iterator(((PyObject *)__pyx_cur_scope->__pyx_genexpr_arg_0), 0, __pyx_n_s_items, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2860, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_7 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_5, &__pyx_t_6, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_7 == 0)) break;
    if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 2860, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_k);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_k, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_v);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_v, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_k);
    __pyx_r = __pyx_cur_scope->__pyx_v_k;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_3 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_3;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2860, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2859
 *         return MapIter(self)
 * 
 *     def keys(self):             # <<<<<<<<<<<<<<
 *         return (k for k,v in self.items())
 * 
 */

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_34keys(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self) {
  PyObject *__pyx_gb_9uharfbuzz_9_harfbuzz_3Map_4keys_2generator1 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("keys", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2860
 * 
 *     def keys(self):
 *         return (k for k,v in self.items())             # <<<<<<<<<<<<<<
 * 
 *     def values(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_4keys_genexpr(NULL, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2860, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2859
 *         return MapIter(self)
 * 
 *     def keys(self):             # <<<<<<<<<<<<<<
 *         return (k for k,v in self.items())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Map.keys", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_gb_9uharfbuzz_9_harfbuzz_3Map_4keys_2generator1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2862
 *         return (k for k,v in self.items())
 * 
 *     def values(self):             # <<<<<<<<<<<<<<
 *         return (v for k,v in self.items())
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_37values(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Map_37values = {"values", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_37values, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_37values(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("values (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("values", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "values", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_36values(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9uharfbuzz_9_harfbuzz_3Map_6values_2generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uharfbuzz/_harfbuzz.pyx":2863
 * 
 *     def values(self):
 *         return (v for k,v in self.items())             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_6values_genexpr(CYTHON_UNUSED PyObject *__pyx_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_genexpr_arg_0) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr *)__pyx_tp_new_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr(__pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 2863, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9uharfbuzz_9_harfbuzz_3Map_6values_2generator2, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_values_locals_genexpr, __pyx_n_s_uharfbuzz__harfbuzz); if (unlikely(!gen)) __PYX_ERR(0, 2863, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Map.values.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9uharfbuzz_9_harfbuzz_3Map_6values_2generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2863, __pyx_L1_error)
  __pyx_t_2 = 0;
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(0, 2863, __pyx_L1_error) }
  if (unlikely(((PyObject *)__pyx_cur_scope->__pyx_genexpr_arg_0) == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
    __PYX_ERR(0, 2863, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_dict_iterator(((PyObject *)__pyx_cur_scope->__pyx_genexpr_arg_0), 0, __pyx_n_s_items, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2863, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_7 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_5, &__pyx_t_6, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_7 == 0)) break;
    if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 2863, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_k);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_k, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_v);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_v, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_v);
    __pyx_r = __pyx_cur_scope->__pyx_v_v;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_3 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_3;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2863, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2862
 *         return (k for k,v in self.items())
 * 
 *     def values(self):             # <<<<<<<<<<<<<<
 *         return (v for k,v in self.items())
 * 
 */

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_36values(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self) {
  PyObject *__pyx_gb_9uharfbuzz_9_harfbuzz_3Map_6values_2generator2 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("values", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2863
 * 
 *     def values(self):
 *         return (v for k,v in self.items())             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_6values_genexpr(NULL, __pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2863, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2862
 *         return (k for k,v in self.items())
 * 
 *     def values(self):             # <<<<<<<<<<<<<<
 *         return (v for k,v in self.items())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Map.values", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_gb_9uharfbuzz_9_harfbuzz_3Map_6values_2generator2);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2865
 *         return (v for k,v in self.items())
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return self.keys()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_39__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_39__iter__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_38__iter__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_38__iter__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__iter__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2866
 * 
 *     def __iter__(self):
 *         return self.keys()             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_keys); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2866, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2866, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2865
 *         return (v for k,v in self.items())
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return self.keys()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Map.__iter__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2868
 *         return self.keys()
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         s = ', '.join("%s: %s" % (repr(k), repr(v)) for k,v in sorted(self.items()))
 *         return ("Map({%s})" % s)
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_41__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_41__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_40__repr__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_9uharfbuzz_9_harfbuzz_3Map_8__repr___2generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "uharfbuzz/_harfbuzz.pyx":2869
 * 
 *     def __repr__(self):
 *         s = ', '.join("%s: %s" % (repr(k), repr(v)) for k,v in sorted(self.items()))             # <<<<<<<<<<<<<<
 *         return ("Map({%s})" % s)
 * 
 */

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_8__repr___genexpr(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr *)__pyx_tp_new_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr(__pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 2869, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_9uharfbuzz_9_harfbuzz_3Map_8__repr___2generator3, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_repr___locals_genexpr, __pyx_n_s_uharfbuzz__harfbuzz); if (unlikely(!gen)) __PYX_ERR(0, 2869, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Map.__repr__.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_9uharfbuzz_9_harfbuzz_3Map_8__repr___2generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr *__pyx_cur_scope = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  Py_ssize_t __pyx_t_9;
  Py_UCS4 __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 2869, __pyx_L1_error)
  __pyx_r = PyList_New(0); if (unlikely(!__pyx_r)) __PYX_ERR(0, 2869, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(0, 2869, __pyx_L1_error) }
  __pyx_t_2 = PySequence_List(__pyx_cur_scope->__pyx_genexpr_arg_0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2869, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_3 = PyList_Sort(__pyx_t_1); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 2869, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2);
  __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 2869, __pyx_L1_error)
      #endif
      if (__pyx_t_4 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_4); __Pyx_INCREF(__pyx_t_1); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 2869, __pyx_L1_error)
    #else
    __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2869, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    #endif
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 2869, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2869, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2869, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_7 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2869, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_8 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_7);
      index = 0; __pyx_t_5 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_5)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(0, 2869, __pyx_L1_error)
      __pyx_t_8 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L7_unpacking_done;
      __pyx_L6_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 2869, __pyx_L1_error)
      __pyx_L7_unpacking_done:;
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_k);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_k, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_v);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_v, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    __pyx_t_6 = 0;
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2869, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = 0;
    __pyx_t_10 = 127;
    __pyx_t_6 = PyObject_Repr(__pyx_cur_scope->__pyx_v_k); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2869, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_6), __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2869, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_10;
    __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_kp_u__21);
    __pyx_t_9 += 2;
    __Pyx_GIVEREF(__pyx_kp_u__21);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_kp_u__21);
    __pyx_t_5 = PyObject_Repr(__pyx_cur_scope->__pyx_v_v); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2869, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_t_5), __pyx_empty_unicode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2869, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_6) : __pyx_t_10;
    __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_t_6);
    __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyUnicode_Join(__pyx_t_1, 3, __pyx_t_9, __pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2869, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(__Pyx_ListComp_Append(__pyx_r, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 2869, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2868
 *         return self.keys()
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         s = ', '.join("%s: %s" % (repr(k), repr(v)) for k,v in sorted(self.items()))
 *         return ("Map({%s})" % s)
 */

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_40__repr__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self) {
  PyObject *__pyx_v_s = NULL;
  PyObject *__pyx_gb_9uharfbuzz_9_harfbuzz_3Map_8__repr___2generator3 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2869
 * 
 *     def __repr__(self):
 *         s = ', '.join("%s: %s" % (repr(k), repr(v)) for k,v in sorted(self.items()))             # <<<<<<<<<<<<<<
 *         return ("Map({%s})" % s)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_items); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2869, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2869, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_2 = __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_8__repr___genexpr(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2869, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_Generator_Next(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2869, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyUnicode_Join(__pyx_kp_u__19, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2869, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_s = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2870
 *     def __repr__(self):
 *         s = ', '.join("%s: %s" % (repr(k), repr(v)) for k,v in sorted(self.items()))
 *         return ("Map({%s})" % s)             # <<<<<<<<<<<<<<
 * 
 * cdef class MapIter:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyUnicode_Format(__pyx_kp_u_Map_s, __pyx_v_s); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2870, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2868
 *         return self.keys()
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         s = ', '.join("%s: %s" % (repr(k), repr(v)) for k,v in sorted(self.items()))
 *         return ("Map({%s})" % s)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Map.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_s);
  __Pyx_XDECREF(__pyx_gb_9uharfbuzz_9_harfbuzz_3Map_8__repr___2generator3);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_43__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Map_43__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_43__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_43__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_42__reduce_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_42__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Map.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_45__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_3Map_45__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_45__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_45__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Map.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_3Map_44__setstate_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_3Map_44__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.Map.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2877
 *     cdef int _i
 * 
 *     def __cinit__(self, m: Map):             # <<<<<<<<<<<<<<
 *         self.m = m
 *         self._hb_map = m._hb_map
 */

/* Python wrapper */
static int __pyx_pw_9uharfbuzz_9_harfbuzz_7MapIter_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_9uharfbuzz_9_harfbuzz_7MapIter_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_m = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_m,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_m)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 2877, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__cinit__") < 0)) __PYX_ERR(0, 2877, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_m = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 2877, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.MapIter.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_m), __pyx_ptype_9uharfbuzz_9_harfbuzz_Map, 0, "m", 0))) __PYX_ERR(0, 2877, __pyx_L1_error)
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_7MapIter___cinit__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_MapIter *)__pyx_v_self), __pyx_v_m);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_9uharfbuzz_9_harfbuzz_7MapIter___cinit__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_MapIter *__pyx_v_self, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *__pyx_v_m) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  hb_map_t *__pyx_t_1;
  __Pyx_RefNannySetupContext("__cinit__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2878
 * 
 *     def __cinit__(self, m: Map):
 *         self.m = m             # <<<<<<<<<<<<<<
 *         self._hb_map = m._hb_map
 *         self._i = -1
 */
  __Pyx_INCREF((PyObject *)__pyx_v_m);
  __Pyx_GIVEREF((PyObject *)__pyx_v_m);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->m);
  __Pyx_DECREF((PyObject *)__pyx_v_self->m);
  __pyx_v_self->m = __pyx_v_m;

  /* "uharfbuzz/_harfbuzz.pyx":2879
 *     def __cinit__(self, m: Map):
 *         self.m = m
 *         self._hb_map = m._hb_map             # <<<<<<<<<<<<<<
 *         self._i = -1
 * 
 */
  __pyx_t_1 = __pyx_v_m->_hb_map;
  __pyx_v_self->_hb_map = __pyx_t_1;

  /* "uharfbuzz/_harfbuzz.pyx":2880
 *         self.m = m
 *         self._hb_map = m._hb_map
 *         self._i = -1             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  __pyx_v_self->_i = -1;

  /* "uharfbuzz/_harfbuzz.pyx":2877
 *     cdef int _i
 * 
 *     def __cinit__(self, m: Map):             # <<<<<<<<<<<<<<
 *         self.m = m
 *         self._hb_map = m._hb_map
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2882
 *         self._i = -1
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_7MapIter_3__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_7MapIter_3__iter__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_7MapIter_2__iter__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_MapIter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_7MapIter_2__iter__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_MapIter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2883
 * 
 *     def __iter__(self):
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __next__(self) -> Tuple[int, int]:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2882
 *         self._i = -1
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "uharfbuzz/_harfbuzz.pyx":2885
 *         return self
 * 
 *     def __next__(self) -> Tuple[int, int]:             # <<<<<<<<<<<<<<
 *         cdef hb_codepoint_t k
 *         cdef hb_codepoint_t v
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_7MapIter_5__next__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_7MapIter_5__next__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__next__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_7MapIter_4__next__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_MapIter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_7MapIter_4__next__(struct __pyx_obj_9uharfbuzz_9_harfbuzz_MapIter *__pyx_v_self) {
  hb_codepoint_t __pyx_v_k;
  hb_codepoint_t __pyx_v_v;
  hb_bool_t __pyx_v_ret;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_error_without_exception = 0; /* StopIteration */
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__next__", 1);

  /* "uharfbuzz/_harfbuzz.pyx":2888
 *         cdef hb_codepoint_t k
 *         cdef hb_codepoint_t v
 *         ret = hb_map_next(self._hb_map, &self._i, &k, &v)             # <<<<<<<<<<<<<<
 *         if not ret:
 *             raise StopIteration
 */
  __pyx_v_ret = hb_map_next(__pyx_v_self->_hb_map, (&__pyx_v_self->_i), (&__pyx_v_k), (&__pyx_v_v));

  /* "uharfbuzz/_harfbuzz.pyx":2889
 *         cdef hb_codepoint_t v
 *         ret = hb_map_next(self._hb_map, &self._i, &k, &v)
 *         if not ret:             # <<<<<<<<<<<<<<
 *             raise StopIteration
 *         return (k, v)
 */
  __pyx_t_1 = (!(__pyx_v_ret != 0));
  if (unlikely(__pyx_t_1)) {

    /* "uharfbuzz/_harfbuzz.pyx":2890
 *         ret = hb_map_next(self._hb_map, &self._i, &k, &v)
 *         if not ret:
 *             raise StopIteration             # <<<<<<<<<<<<<<
 *         return (k, v)
 */
    __pyx_error_without_exception = 1;
    goto __pyx_L1_error;;

    /* "uharfbuzz/_harfbuzz.pyx":2889
 *         cdef hb_codepoint_t v
 *         ret = hb_map_next(self._hb_map, &self._i, &k, &v)
 *         if not ret:             # <<<<<<<<<<<<<<
 *             raise StopIteration
 *         return (k, v)
 */
  }

  /* "uharfbuzz/_harfbuzz.pyx":2891
 *         if not ret:
 *             raise StopIteration
 *         return (k, v)             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_hb_codepoint_t(__pyx_v_k); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_hb_codepoint_t(__pyx_v_v); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2891, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2)) __PYX_ERR(0, 2891, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3)) __PYX_ERR(0, 2891, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "uharfbuzz/_harfbuzz.pyx":2885
 *         return self
 * 
 *     def __next__(self) -> Tuple[int, int]:             # <<<<<<<<<<<<<<
 *         cdef hb_codepoint_t k
 *         cdef hb_codepoint_t v
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  if (!__pyx_error_without_exception) {
    __Pyx_AddTraceback("uharfbuzz._harfbuzz.MapIter.__next__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  }
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_7MapIter_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_7MapIter_7__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_7MapIter_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_7MapIter_7__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_7MapIter_6__reduce_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_MapIter *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_7MapIter_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_MapIter *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.MapIter.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

/* Python wrapper */
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_7MapIter_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_9uharfbuzz_9_harfbuzz_7MapIter_9__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_7MapIter_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_9uharfbuzz_9_harfbuzz_7MapIter_9__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.MapIter.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_9uharfbuzz_9_harfbuzz_7MapIter_8__setstate_cython__(((struct __pyx_obj_9uharfbuzz_9_harfbuzz_MapIter *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_9uharfbuzz_9_harfbuzz_7MapIter_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_9uharfbuzz_9_harfbuzz_MapIter *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_no_default___reduce___due_to_non, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("uharfbuzz._harfbuzz.MapIter.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_GlyphInfo __pyx_vtable_9uharfbuzz_9_harfbuzz_GlyphInfo;

static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_GlyphInfo(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphInfo *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphInfo *)o);
  p->__pyx_vtab = __pyx_vtabptr_9uharfbuzz_9_harfbuzz_GlyphInfo;
  new((void*)&(p->_hb_glyph_info)) hb_glyph_info_t();
  return o;
}

static void __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_GlyphInfo(PyObject *o) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphInfo *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphInfo *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_GlyphInfo) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  __Pyx_call_destructor(p->_hb_glyph_info);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_9GlyphInfo_codepoint(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_9GlyphInfo_9codepoint_1__get__(o);
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_9GlyphInfo_cluster(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_9GlyphInfo_7cluster_1__get__(o);
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_9GlyphInfo_flags(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_9GlyphInfo_5flags_1__get__(o);
}

static PyMethodDef __pyx_methods_9uharfbuzz_9_harfbuzz_GlyphInfo[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9GlyphInfo_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9GlyphInfo_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_9uharfbuzz_9_harfbuzz_GlyphInfo[] = {
  {(char *)"codepoint", __pyx_getprop_9uharfbuzz_9_harfbuzz_9GlyphInfo_codepoint, 0, (char *)0, 0},
  {(char *)"cluster", __pyx_getprop_9uharfbuzz_9_harfbuzz_9GlyphInfo_cluster, 0, (char *)0, 0},
  {(char *)"flags", __pyx_getprop_9uharfbuzz_9_harfbuzz_9GlyphInfo_flags, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9uharfbuzz_9_harfbuzz_GlyphInfo_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_GlyphInfo},
  {Py_tp_methods, (void *)__pyx_methods_9uharfbuzz_9_harfbuzz_GlyphInfo},
  {Py_tp_getset, (void *)__pyx_getsets_9uharfbuzz_9_harfbuzz_GlyphInfo},
  {Py_tp_new, (void *)__pyx_tp_new_9uharfbuzz_9_harfbuzz_GlyphInfo},
  {0, 0},
};
static PyType_Spec __pyx_type_9uharfbuzz_9_harfbuzz_GlyphInfo_spec = {
  "uharfbuzz._harfbuzz.GlyphInfo",
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphInfo),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_9uharfbuzz_9_harfbuzz_GlyphInfo_slots,
};
#else

static PyTypeObject __pyx_type_9uharfbuzz_9_harfbuzz_GlyphInfo = {
  PyVarObject_HEAD_INIT(0, 0)
  "uharfbuzz._harfbuzz.""GlyphInfo", /*tp_name*/
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphInfo), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_GlyphInfo, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_9uharfbuzz_9_harfbuzz_GlyphInfo, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_9uharfbuzz_9_harfbuzz_GlyphInfo, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9uharfbuzz_9_harfbuzz_GlyphInfo, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_GlyphPosition __pyx_vtable_9uharfbuzz_9_harfbuzz_GlyphPosition;

static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_GlyphPosition(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition *)o);
  p->__pyx_vtab = __pyx_vtabptr_9uharfbuzz_9_harfbuzz_GlyphPosition;
  new((void*)&(p->_hb_glyph_position)) hb_glyph_position_t();
  return o;
}

static void __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_GlyphPosition(PyObject *o) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_GlyphPosition) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  __Pyx_call_destructor(p->_hb_glyph_position);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_13GlyphPosition_position(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_13GlyphPosition_8position_1__get__(o);
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_13GlyphPosition_x_advance(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_13GlyphPosition_9x_advance_1__get__(o);
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_13GlyphPosition_y_advance(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_13GlyphPosition_9y_advance_1__get__(o);
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_13GlyphPosition_x_offset(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_13GlyphPosition_8x_offset_1__get__(o);
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_13GlyphPosition_y_offset(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_13GlyphPosition_8y_offset_1__get__(o);
}

static PyMethodDef __pyx_methods_9uharfbuzz_9_harfbuzz_GlyphPosition[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_13GlyphPosition_1__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_13GlyphPosition_3__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_9uharfbuzz_9_harfbuzz_GlyphPosition[] = {
  {(char *)"position", __pyx_getprop_9uharfbuzz_9_harfbuzz_13GlyphPosition_position, 0, (char *)0, 0},
  {(char *)"x_advance", __pyx_getprop_9uharfbuzz_9_harfbuzz_13GlyphPosition_x_advance, 0, (char *)0, 0},
  {(char *)"y_advance", __pyx_getprop_9uharfbuzz_9_harfbuzz_13GlyphPosition_y_advance, 0, (char *)0, 0},
  {(char *)"x_offset", __pyx_getprop_9uharfbuzz_9_harfbuzz_13GlyphPosition_x_offset, 0, (char *)0, 0},
  {(char *)"y_offset", __pyx_getprop_9uharfbuzz_9_harfbuzz_13GlyphPosition_y_offset, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9uharfbuzz_9_harfbuzz_GlyphPosition_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_GlyphPosition},
  {Py_tp_methods, (void *)__pyx_methods_9uharfbuzz_9_harfbuzz_GlyphPosition},
  {Py_tp_getset, (void *)__pyx_getsets_9uharfbuzz_9_harfbuzz_GlyphPosition},
  {Py_tp_new, (void *)__pyx_tp_new_9uharfbuzz_9_harfbuzz_GlyphPosition},
  {0, 0},
};
static PyType_Spec __pyx_type_9uharfbuzz_9_harfbuzz_GlyphPosition_spec = {
  "uharfbuzz._harfbuzz.GlyphPosition",
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_9uharfbuzz_9_harfbuzz_GlyphPosition_slots,
};
#else

static PyTypeObject __pyx_type_9uharfbuzz_9_harfbuzz_GlyphPosition = {
  PyVarObject_HEAD_INIT(0, 0)
  "uharfbuzz._harfbuzz.""GlyphPosition", /*tp_name*/
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_GlyphPosition, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_9uharfbuzz_9_harfbuzz_GlyphPosition, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_9uharfbuzz_9_harfbuzz_GlyphPosition, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9uharfbuzz_9_harfbuzz_GlyphPosition, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_Buffer(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)o);
  p->_message_callback = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_Buffer(PyObject *o) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_Buffer) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->_message_callback);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_9uharfbuzz_9_harfbuzz_Buffer(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)o;
  if (p->_message_callback) {
    e = (*v)(p->_message_callback, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9uharfbuzz_9_harfbuzz_Buffer(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer *)o;
  tmp = ((PyObject*)p->_message_callback);
  p->_message_callback = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_6Buffer_direction(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_9direction_1__get__(o);
}

static int __pyx_setprop_9uharfbuzz_9_harfbuzz_6Buffer_direction(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_9direction_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_6Buffer_glyph_infos(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_11glyph_infos_1__get__(o);
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_6Buffer_glyph_positions(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_15glyph_positions_1__get__(o);
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_6Buffer_language(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_8language_1__get__(o);
}

static int __pyx_setprop_9uharfbuzz_9_harfbuzz_6Buffer_language(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_8language_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_6Buffer_script(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_6script_1__get__(o);
}

static int __pyx_setprop_9uharfbuzz_9_harfbuzz_6Buffer_script(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_6script_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_6Buffer_flags(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_5flags_1__get__(o);
}

static int __pyx_setprop_9uharfbuzz_9_harfbuzz_6Buffer_flags(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_5flags_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_6Buffer_cluster_level(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_13cluster_level_1__get__(o);
}

static int __pyx_setprop_9uharfbuzz_9_harfbuzz_6Buffer_cluster_level(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_13cluster_level_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_6Buffer_content_type(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_12content_type_1__get__(o);
}

static int __pyx_setprop_9uharfbuzz_9_harfbuzz_6Buffer_content_type(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_12content_type_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_6Buffer_replacement_codepoint(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_21replacement_codepoint_1__get__(o);
}

static int __pyx_setprop_9uharfbuzz_9_harfbuzz_6Buffer_replacement_codepoint(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_21replacement_codepoint_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_6Buffer_invisible_glyph(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_15invisible_glyph_1__get__(o);
}

static int __pyx_setprop_9uharfbuzz_9_harfbuzz_6Buffer_invisible_glyph(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_15invisible_glyph_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_6Buffer_not_found_glyph(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_15not_found_glyph_1__get__(o);
}

static int __pyx_setprop_9uharfbuzz_9_harfbuzz_6Buffer_not_found_glyph(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_15not_found_glyph_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_9uharfbuzz_9_harfbuzz_Buffer[] = {
  {"create", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_5create, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"reset", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_9reset, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"clear_contents", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_11clear_contents, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_language_from_ot_tag", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_13set_language_from_ot_tag, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_script_from_ot_tag", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_15set_script_from_ot_tag, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"add_codepoints", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_17add_codepoints, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"add_utf8", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_19add_utf8, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"add_str", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_21add_str, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"guess_segment_properties", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_23guess_segment_properties, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_message_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_25set_message_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_27__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_29__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_9uharfbuzz_9_harfbuzz_Buffer[] = {
  {(char *)"direction", __pyx_getprop_9uharfbuzz_9_harfbuzz_6Buffer_direction, __pyx_setprop_9uharfbuzz_9_harfbuzz_6Buffer_direction, (char *)0, 0},
  {(char *)"glyph_infos", __pyx_getprop_9uharfbuzz_9_harfbuzz_6Buffer_glyph_infos, 0, (char *)0, 0},
  {(char *)"glyph_positions", __pyx_getprop_9uharfbuzz_9_harfbuzz_6Buffer_glyph_positions, 0, (char *)0, 0},
  {(char *)"language", __pyx_getprop_9uharfbuzz_9_harfbuzz_6Buffer_language, __pyx_setprop_9uharfbuzz_9_harfbuzz_6Buffer_language, (char *)0, 0},
  {(char *)"script", __pyx_getprop_9uharfbuzz_9_harfbuzz_6Buffer_script, __pyx_setprop_9uharfbuzz_9_harfbuzz_6Buffer_script, (char *)0, 0},
  {(char *)"flags", __pyx_getprop_9uharfbuzz_9_harfbuzz_6Buffer_flags, __pyx_setprop_9uharfbuzz_9_harfbuzz_6Buffer_flags, (char *)0, 0},
  {(char *)"cluster_level", __pyx_getprop_9uharfbuzz_9_harfbuzz_6Buffer_cluster_level, __pyx_setprop_9uharfbuzz_9_harfbuzz_6Buffer_cluster_level, (char *)0, 0},
  {(char *)"content_type", __pyx_getprop_9uharfbuzz_9_harfbuzz_6Buffer_content_type, __pyx_setprop_9uharfbuzz_9_harfbuzz_6Buffer_content_type, (char *)0, 0},
  {(char *)"replacement_codepoint", __pyx_getprop_9uharfbuzz_9_harfbuzz_6Buffer_replacement_codepoint, __pyx_setprop_9uharfbuzz_9_harfbuzz_6Buffer_replacement_codepoint, (char *)0, 0},
  {(char *)"invisible_glyph", __pyx_getprop_9uharfbuzz_9_harfbuzz_6Buffer_invisible_glyph, __pyx_setprop_9uharfbuzz_9_harfbuzz_6Buffer_invisible_glyph, (char *)0, 0},
  {(char *)"not_found_glyph", __pyx_getprop_9uharfbuzz_9_harfbuzz_6Buffer_not_found_glyph, __pyx_setprop_9uharfbuzz_9_harfbuzz_6Buffer_not_found_glyph, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9uharfbuzz_9_harfbuzz_Buffer_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_Buffer},
  {Py_sq_length, (void *)__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_7__len__},
  {Py_mp_length, (void *)__pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_7__len__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_9uharfbuzz_9_harfbuzz_Buffer},
  {Py_tp_clear, (void *)__pyx_tp_clear_9uharfbuzz_9_harfbuzz_Buffer},
  {Py_tp_methods, (void *)__pyx_methods_9uharfbuzz_9_harfbuzz_Buffer},
  {Py_tp_getset, (void *)__pyx_getsets_9uharfbuzz_9_harfbuzz_Buffer},
  {Py_tp_new, (void *)__pyx_tp_new_9uharfbuzz_9_harfbuzz_Buffer},
  {0, 0},
};
static PyType_Spec __pyx_type_9uharfbuzz_9_harfbuzz_Buffer_spec = {
  "uharfbuzz._harfbuzz.Buffer",
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_9uharfbuzz_9_harfbuzz_Buffer_slots,
};
#else

static PySequenceMethods __pyx_tp_as_sequence_Buffer = {
  __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_7__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  0, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_Buffer = {
  __pyx_pw_9uharfbuzz_9_harfbuzz_6Buffer_7__len__, /*mp_length*/
  0, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_9uharfbuzz_9_harfbuzz_Buffer = {
  PyVarObject_HEAD_INIT(0, 0)
  "uharfbuzz._harfbuzz.""Buffer", /*tp_name*/
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Buffer), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_Buffer, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_Buffer, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_Buffer, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9uharfbuzz_9_harfbuzz_Buffer, /*tp_traverse*/
  __pyx_tp_clear_9uharfbuzz_9_harfbuzz_Buffer, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_9uharfbuzz_9_harfbuzz_Buffer, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_9uharfbuzz_9_harfbuzz_Buffer, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9uharfbuzz_9_harfbuzz_Buffer, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_Blob __pyx_vtable_9uharfbuzz_9_harfbuzz_Blob;

static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_Blob(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *)o);
  p->__pyx_vtab = __pyx_vtabptr_9uharfbuzz_9_harfbuzz_Blob;
  p->_data = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_9uharfbuzz_9_harfbuzz_4Blob_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_Blob(PyObject *o) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_Blob) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_9uharfbuzz_9_harfbuzz_4Blob_5__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->_data);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_9uharfbuzz_9_harfbuzz_Blob(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *)o;
  if (p->_data) {
    e = (*v)(p->_data, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9uharfbuzz_9_harfbuzz_Blob(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *)o;
  tmp = ((PyObject*)p->_data);
  p->_data = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_4Blob_data(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_4Blob_4data_1__get__(o);
}

static PyMethodDef __pyx_methods_9uharfbuzz_9_harfbuzz_Blob[] = {
  {"from_file_path", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Blob_3from_file_path, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Blob_11__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Blob_13__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_9uharfbuzz_9_harfbuzz_Blob[] = {
  {(char *)"data", __pyx_getprop_9uharfbuzz_9_harfbuzz_4Blob_data, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9uharfbuzz_9_harfbuzz_Blob_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_Blob},
  {Py_nb_bool, (void *)__pyx_pw_9uharfbuzz_9_harfbuzz_4Blob_9__bool__},
  {Py_sq_length, (void *)__pyx_pw_9uharfbuzz_9_harfbuzz_4Blob_7__len__},
  {Py_mp_length, (void *)__pyx_pw_9uharfbuzz_9_harfbuzz_4Blob_7__len__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_9uharfbuzz_9_harfbuzz_Blob},
  {Py_tp_clear, (void *)__pyx_tp_clear_9uharfbuzz_9_harfbuzz_Blob},
  {Py_tp_methods, (void *)__pyx_methods_9uharfbuzz_9_harfbuzz_Blob},
  {Py_tp_getset, (void *)__pyx_getsets_9uharfbuzz_9_harfbuzz_Blob},
  {Py_tp_new, (void *)__pyx_tp_new_9uharfbuzz_9_harfbuzz_Blob},
  {0, 0},
};
static PyType_Spec __pyx_type_9uharfbuzz_9_harfbuzz_Blob_spec = {
  "uharfbuzz._harfbuzz.Blob",
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_9uharfbuzz_9_harfbuzz_Blob_slots,
};
#else

static PyNumberMethods __pyx_tp_as_number_Blob = {
  0, /*nb_add*/
  0, /*nb_subtract*/
  0, /*nb_multiply*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_divide*/
  #endif
  0, /*nb_remainder*/
  0, /*nb_divmod*/
  0, /*nb_power*/
  0, /*nb_negative*/
  0, /*nb_positive*/
  0, /*nb_absolute*/
  __pyx_pw_9uharfbuzz_9_harfbuzz_4Blob_9__bool__, /*nb_bool*/
  0, /*nb_invert*/
  0, /*nb_lshift*/
  0, /*nb_rshift*/
  0, /*nb_and*/
  0, /*nb_xor*/
  0, /*nb_or*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_coerce*/
  #endif
  0, /*nb_int*/
  #if PY_MAJOR_VERSION < 3
  0, /*nb_long*/
  #else
  0, /*reserved*/
  #endif
  0, /*nb_float*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_oct*/
  #endif
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_hex*/
  #endif
  0, /*nb_inplace_add*/
  0, /*nb_inplace_subtract*/
  0, /*nb_inplace_multiply*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_inplace_divide*/
  #endif
  0, /*nb_inplace_remainder*/
  0, /*nb_inplace_power*/
  0, /*nb_inplace_lshift*/
  0, /*nb_inplace_rshift*/
  0, /*nb_inplace_and*/
  0, /*nb_inplace_xor*/
  0, /*nb_inplace_or*/
  0, /*nb_floor_divide*/
  0, /*nb_true_divide*/
  0, /*nb_inplace_floor_divide*/
  0, /*nb_inplace_true_divide*/
  0, /*nb_index*/
  #if PY_VERSION_HEX >= 0x03050000
  0, /*nb_matrix_multiply*/
  #endif
  #if PY_VERSION_HEX >= 0x03050000
  0, /*nb_inplace_matrix_multiply*/
  #endif
};

static PySequenceMethods __pyx_tp_as_sequence_Blob = {
  __pyx_pw_9uharfbuzz_9_harfbuzz_4Blob_7__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  0, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_Blob = {
  __pyx_pw_9uharfbuzz_9_harfbuzz_4Blob_7__len__, /*mp_length*/
  0, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_9uharfbuzz_9_harfbuzz_Blob = {
  PyVarObject_HEAD_INIT(0, 0)
  "uharfbuzz._harfbuzz.""Blob", /*tp_name*/
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_Blob, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  &__pyx_tp_as_number_Blob, /*tp_as_number*/
  &__pyx_tp_as_sequence_Blob, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_Blob, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9uharfbuzz_9_harfbuzz_Blob, /*tp_traverse*/
  __pyx_tp_clear_9uharfbuzz_9_harfbuzz_Blob, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_9uharfbuzz_9_harfbuzz_Blob, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_9uharfbuzz_9_harfbuzz_Blob, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9uharfbuzz_9_harfbuzz_Blob, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_Face __pyx_vtable_9uharfbuzz_9_harfbuzz_Face;

static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_Face(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)o);
  p->__pyx_vtab = __pyx_vtabptr_9uharfbuzz_9_harfbuzz_Face;
  p->_reference_table_func = Py_None; Py_INCREF(Py_None);
  p->_blob = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_Face(PyObject *o) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_Face) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_9uharfbuzz_9_harfbuzz_4Face_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->_reference_table_func);
  Py_CLEAR(p->_blob);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_9uharfbuzz_9_harfbuzz_Face(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)o;
  if (p->_reference_table_func) {
    e = (*v)(p->_reference_table_func, a); if (e) return e;
  }
  if (p->_blob) {
    e = (*v)(((PyObject *)p->_blob), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9uharfbuzz_9_harfbuzz_Face(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)o;
  tmp = ((PyObject*)p->_reference_table_func);
  p->_reference_table_func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_blob);
  p->_blob = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_4Face_index(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_4Face_5index_1__get__(o);
}

static int __pyx_setprop_9uharfbuzz_9_harfbuzz_4Face_index(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9uharfbuzz_9_harfbuzz_4Face_5index_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_4Face_upem(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_4Face_4upem_1__get__(o);
}

static int __pyx_setprop_9uharfbuzz_9_harfbuzz_4Face_upem(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9uharfbuzz_9_harfbuzz_4Face_4upem_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_4Face_glyph_count(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_4Face_11glyph_count_1__get__(o);
}

static int __pyx_setprop_9uharfbuzz_9_harfbuzz_4Face_glyph_count(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9uharfbuzz_9_harfbuzz_4Face_11glyph_count_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_4Face_blob(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_4Face_4blob_1__get__(o);
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_4Face_table_tags(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_4Face_10table_tags_1__get__(o);
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_4Face_unicodes(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_4Face_8unicodes_1__get__(o);
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_4Face_variation_selectors(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_4Face_19variation_selectors_1__get__(o);
}

static PyMethodDef __pyx_methods_9uharfbuzz_9_harfbuzz_Face[] = {
  {"create", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_5create, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"create_for_tables", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_7create_for_tables, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"variation_unicodes", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_9variation_unicodes, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_11__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Face_13__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_9uharfbuzz_9_harfbuzz_Face[] = {
  {(char *)"index", __pyx_getprop_9uharfbuzz_9_harfbuzz_4Face_index, __pyx_setprop_9uharfbuzz_9_harfbuzz_4Face_index, (char *)0, 0},
  {(char *)"upem", __pyx_getprop_9uharfbuzz_9_harfbuzz_4Face_upem, __pyx_setprop_9uharfbuzz_9_harfbuzz_4Face_upem, (char *)0, 0},
  {(char *)"glyph_count", __pyx_getprop_9uharfbuzz_9_harfbuzz_4Face_glyph_count, __pyx_setprop_9uharfbuzz_9_harfbuzz_4Face_glyph_count, (char *)0, 0},
  {(char *)"blob", __pyx_getprop_9uharfbuzz_9_harfbuzz_4Face_blob, 0, (char *)0, 0},
  {(char *)"table_tags", __pyx_getprop_9uharfbuzz_9_harfbuzz_4Face_table_tags, 0, (char *)0, 0},
  {(char *)"unicodes", __pyx_getprop_9uharfbuzz_9_harfbuzz_4Face_unicodes, 0, (char *)0, 0},
  {(char *)"variation_selectors", __pyx_getprop_9uharfbuzz_9_harfbuzz_4Face_variation_selectors, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9uharfbuzz_9_harfbuzz_Face_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_Face},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_9uharfbuzz_9_harfbuzz_Face},
  {Py_tp_clear, (void *)__pyx_tp_clear_9uharfbuzz_9_harfbuzz_Face},
  {Py_tp_methods, (void *)__pyx_methods_9uharfbuzz_9_harfbuzz_Face},
  {Py_tp_getset, (void *)__pyx_getsets_9uharfbuzz_9_harfbuzz_Face},
  {Py_tp_new, (void *)__pyx_tp_new_9uharfbuzz_9_harfbuzz_Face},
  {0, 0},
};
static PyType_Spec __pyx_type_9uharfbuzz_9_harfbuzz_Face_spec = {
  "uharfbuzz._harfbuzz.Face",
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_9uharfbuzz_9_harfbuzz_Face_slots,
};
#else

static PyTypeObject __pyx_type_9uharfbuzz_9_harfbuzz_Face = {
  PyVarObject_HEAD_INIT(0, 0)
  "uharfbuzz._harfbuzz.""Face", /*tp_name*/
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_Face, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9uharfbuzz_9_harfbuzz_Face, /*tp_traverse*/
  __pyx_tp_clear_9uharfbuzz_9_harfbuzz_Face, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_9uharfbuzz_9_harfbuzz_Face, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_9uharfbuzz_9_harfbuzz_Face, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9uharfbuzz_9_harfbuzz_Face, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_Font __pyx_vtable_9uharfbuzz_9_harfbuzz_Font;

static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_Font(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)o);
  p->__pyx_vtab = __pyx_vtabptr_9uharfbuzz_9_harfbuzz_Font;
  p->_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)Py_None); Py_INCREF(Py_None);
  p->_ffuncs = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_Font(PyObject *o) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_Font) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->_face);
  Py_CLEAR(p->_ffuncs);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_9uharfbuzz_9_harfbuzz_Font(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)o;
  if (p->_face) {
    e = (*v)(((PyObject *)p->_face), a); if (e) return e;
  }
  if (p->_ffuncs) {
    e = (*v)(((PyObject *)p->_ffuncs), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9uharfbuzz_9_harfbuzz_Font(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *)o;
  tmp = ((PyObject*)p->_face);
  p->_face = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_ffuncs);
  p->_ffuncs = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_4Font_face(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_4face_1__get__(o);
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_4Font_funcs(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_5funcs_1__get__(o);
}

static int __pyx_setprop_9uharfbuzz_9_harfbuzz_4Font_funcs(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_5funcs_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_4Font_scale(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_5scale_1__get__(o);
}

static int __pyx_setprop_9uharfbuzz_9_harfbuzz_4Font_scale(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_5scale_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_4Font_ppem(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_4ppem_1__get__(o);
}

static int __pyx_setprop_9uharfbuzz_9_harfbuzz_4Font_ppem(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_4ppem_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_4Font_ptem(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_4ptem_1__get__(o);
}

static int __pyx_setprop_9uharfbuzz_9_harfbuzz_4Font_ptem(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_4ptem_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_4Font_synthetic_slant(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_15synthetic_slant_1__get__(o);
}

static int __pyx_setprop_9uharfbuzz_9_harfbuzz_4Font_synthetic_slant(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_15synthetic_slant_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_4Font_synthetic_bold(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_14synthetic_bold_1__get__(o);
}

static int __pyx_setprop_9uharfbuzz_9_harfbuzz_4Font_synthetic_bold(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_14synthetic_bold_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_4Font_var_named_instance(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_18var_named_instance_1__get__(o);
}

static int __pyx_setprop_9uharfbuzz_9_harfbuzz_4Font_var_named_instance(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9uharfbuzz_9_harfbuzz_4Font_18var_named_instance_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_9uharfbuzz_9_harfbuzz_Font[] = {
  {"create", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_5create, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_variations", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_7set_variations, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_variation", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_9set_variation, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_glyph_name", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_11get_glyph_name, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_glyph_from_name", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_13get_glyph_from_name, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_glyph_extents", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_15get_glyph_extents, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_glyph_h_advance", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_17get_glyph_h_advance, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_glyph_v_advance", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_19get_glyph_v_advance, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_glyph_h_origin", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_21get_glyph_h_origin, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_glyph_v_origin", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_23get_glyph_v_origin, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_font_extents", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_25get_font_extents, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_variation_glyph", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_27get_variation_glyph, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_nominal_glyph", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_29get_nominal_glyph, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_var_coords_normalized", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_31get_var_coords_normalized, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_var_coords_normalized", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_33set_var_coords_normalized, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_var_coords_design", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_35get_var_coords_design, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_var_coords_design", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_37set_var_coords_design, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"glyph_to_string", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_39glyph_to_string, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"glyph_from_string", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_41glyph_from_string, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"draw_glyph", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_43draw_glyph, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"paint_glyph", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_45paint_glyph, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"draw_glyph_with_pen", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_47draw_glyph_with_pen, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_49__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_4Font_51__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_9uharfbuzz_9_harfbuzz_Font[] = {
  {(char *)"face", __pyx_getprop_9uharfbuzz_9_harfbuzz_4Font_face, 0, (char *)0, 0},
  {(char *)"funcs", __pyx_getprop_9uharfbuzz_9_harfbuzz_4Font_funcs, __pyx_setprop_9uharfbuzz_9_harfbuzz_4Font_funcs, (char *)0, 0},
  {(char *)"scale", __pyx_getprop_9uharfbuzz_9_harfbuzz_4Font_scale, __pyx_setprop_9uharfbuzz_9_harfbuzz_4Font_scale, (char *)0, 0},
  {(char *)"ppem", __pyx_getprop_9uharfbuzz_9_harfbuzz_4Font_ppem, __pyx_setprop_9uharfbuzz_9_harfbuzz_4Font_ppem, (char *)0, 0},
  {(char *)"ptem", __pyx_getprop_9uharfbuzz_9_harfbuzz_4Font_ptem, __pyx_setprop_9uharfbuzz_9_harfbuzz_4Font_ptem, (char *)0, 0},
  {(char *)"synthetic_slant", __pyx_getprop_9uharfbuzz_9_harfbuzz_4Font_synthetic_slant, __pyx_setprop_9uharfbuzz_9_harfbuzz_4Font_synthetic_slant, (char *)0, 0},
  {(char *)"synthetic_bold", __pyx_getprop_9uharfbuzz_9_harfbuzz_4Font_synthetic_bold, __pyx_setprop_9uharfbuzz_9_harfbuzz_4Font_synthetic_bold, (char *)0, 0},
  {(char *)"var_named_instance", __pyx_getprop_9uharfbuzz_9_harfbuzz_4Font_var_named_instance, __pyx_setprop_9uharfbuzz_9_harfbuzz_4Font_var_named_instance, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9uharfbuzz_9_harfbuzz_Font_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_Font},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_9uharfbuzz_9_harfbuzz_Font},
  {Py_tp_clear, (void *)__pyx_tp_clear_9uharfbuzz_9_harfbuzz_Font},
  {Py_tp_methods, (void *)__pyx_methods_9uharfbuzz_9_harfbuzz_Font},
  {Py_tp_getset, (void *)__pyx_getsets_9uharfbuzz_9_harfbuzz_Font},
  {Py_tp_new, (void *)__pyx_tp_new_9uharfbuzz_9_harfbuzz_Font},
  {0, 0},
};
static PyType_Spec __pyx_type_9uharfbuzz_9_harfbuzz_Font_spec = {
  "uharfbuzz._harfbuzz.Font",
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_9uharfbuzz_9_harfbuzz_Font_slots,
};
#else

static PyTypeObject __pyx_type_9uharfbuzz_9_harfbuzz_Font = {
  PyVarObject_HEAD_INIT(0, 0)
  "uharfbuzz._harfbuzz.""Font", /*tp_name*/
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_Font, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9uharfbuzz_9_harfbuzz_Font, /*tp_traverse*/
  __pyx_tp_clear_9uharfbuzz_9_harfbuzz_Font, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_9uharfbuzz_9_harfbuzz_Font, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_9uharfbuzz_9_harfbuzz_Font, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9uharfbuzz_9_harfbuzz_Font, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_FontFuncs(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)o);
  p->_glyph_h_advance_func = Py_None; Py_INCREF(Py_None);
  p->_glyph_v_advance_func = Py_None; Py_INCREF(Py_None);
  p->_glyph_v_origin_func = Py_None; Py_INCREF(Py_None);
  p->_glyph_name_func = Py_None; Py_INCREF(Py_None);
  p->_nominal_glyph_func = Py_None; Py_INCREF(Py_None);
  p->_variation_glyph_func = Py_None; Py_INCREF(Py_None);
  p->_font_h_extents_func = Py_None; Py_INCREF(Py_None);
  p->_font_v_extents_func = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_FontFuncs(PyObject *o) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_FontFuncs) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->_glyph_h_advance_func);
  Py_CLEAR(p->_glyph_v_advance_func);
  Py_CLEAR(p->_glyph_v_origin_func);
  Py_CLEAR(p->_glyph_name_func);
  Py_CLEAR(p->_nominal_glyph_func);
  Py_CLEAR(p->_variation_glyph_func);
  Py_CLEAR(p->_font_h_extents_func);
  Py_CLEAR(p->_font_v_extents_func);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_9uharfbuzz_9_harfbuzz_FontFuncs(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)o;
  if (p->_glyph_h_advance_func) {
    e = (*v)(p->_glyph_h_advance_func, a); if (e) return e;
  }
  if (p->_glyph_v_advance_func) {
    e = (*v)(p->_glyph_v_advance_func, a); if (e) return e;
  }
  if (p->_glyph_v_origin_func) {
    e = (*v)(p->_glyph_v_origin_func, a); if (e) return e;
  }
  if (p->_glyph_name_func) {
    e = (*v)(p->_glyph_name_func, a); if (e) return e;
  }
  if (p->_nominal_glyph_func) {
    e = (*v)(p->_nominal_glyph_func, a); if (e) return e;
  }
  if (p->_variation_glyph_func) {
    e = (*v)(p->_variation_glyph_func, a); if (e) return e;
  }
  if (p->_font_h_extents_func) {
    e = (*v)(p->_font_h_extents_func, a); if (e) return e;
  }
  if (p->_font_v_extents_func) {
    e = (*v)(p->_font_v_extents_func, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9uharfbuzz_9_harfbuzz_FontFuncs(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs *)o;
  tmp = ((PyObject*)p->_glyph_h_advance_func);
  p->_glyph_h_advance_func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_glyph_v_advance_func);
  p->_glyph_v_advance_func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_glyph_v_origin_func);
  p->_glyph_v_origin_func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_glyph_name_func);
  p->_glyph_name_func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_nominal_glyph_func);
  p->_nominal_glyph_func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_variation_glyph_func);
  p->_variation_glyph_func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_font_h_extents_func);
  p->_font_h_extents_func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_font_v_extents_func);
  p->_font_v_extents_func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_9uharfbuzz_9_harfbuzz_FontFuncs[] = {
  {"create", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_5create, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_glyph_h_advance_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_7set_glyph_h_advance_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_glyph_v_advance_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_9set_glyph_v_advance_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_glyph_v_origin_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_11set_glyph_v_origin_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_glyph_name_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_13set_glyph_name_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_nominal_glyph_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_15set_nominal_glyph_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_variation_glyph_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_17set_variation_glyph_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_font_h_extents_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_19set_font_h_extents_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_font_v_extents_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_21set_font_v_extents_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_23__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9FontFuncs_25__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9uharfbuzz_9_harfbuzz_FontFuncs_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_FontFuncs},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_9uharfbuzz_9_harfbuzz_FontFuncs},
  {Py_tp_clear, (void *)__pyx_tp_clear_9uharfbuzz_9_harfbuzz_FontFuncs},
  {Py_tp_methods, (void *)__pyx_methods_9uharfbuzz_9_harfbuzz_FontFuncs},
  {Py_tp_new, (void *)__pyx_tp_new_9uharfbuzz_9_harfbuzz_FontFuncs},
  {0, 0},
};
static PyType_Spec __pyx_type_9uharfbuzz_9_harfbuzz_FontFuncs_spec = {
  "uharfbuzz._harfbuzz.FontFuncs",
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_9uharfbuzz_9_harfbuzz_FontFuncs_slots,
};
#else

static PyTypeObject __pyx_type_9uharfbuzz_9_harfbuzz_FontFuncs = {
  PyVarObject_HEAD_INIT(0, 0)
  "uharfbuzz._harfbuzz.""FontFuncs", /*tp_name*/
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_FontFuncs), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_FontFuncs, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9uharfbuzz_9_harfbuzz_FontFuncs, /*tp_traverse*/
  __pyx_tp_clear_9uharfbuzz_9_harfbuzz_FontFuncs, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_9uharfbuzz_9_harfbuzz_FontFuncs, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9uharfbuzz_9_harfbuzz_FontFuncs, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_ColorLine __pyx_vtable_9uharfbuzz_9_harfbuzz_ColorLine;

static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_ColorLine(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine *)o);
  p->__pyx_vtab = __pyx_vtabptr_9uharfbuzz_9_harfbuzz_ColorLine;
  if (unlikely(__pyx_pw_9uharfbuzz_9_harfbuzz_9ColorLine_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_ColorLine(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_ColorLine) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_9ColorLine_color_stops(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_9ColorLine_11color_stops_1__get__(o);
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_9ColorLine_extend(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_9ColorLine_6extend_1__get__(o);
}

static PyMethodDef __pyx_methods_9uharfbuzz_9_harfbuzz_ColorLine[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9ColorLine_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9ColorLine_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_9uharfbuzz_9_harfbuzz_ColorLine[] = {
  {(char *)"color_stops", __pyx_getprop_9uharfbuzz_9_harfbuzz_9ColorLine_color_stops, 0, (char *)0, 0},
  {(char *)"extend", __pyx_getprop_9uharfbuzz_9_harfbuzz_9ColorLine_extend, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9uharfbuzz_9_harfbuzz_ColorLine_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_ColorLine},
  {Py_tp_methods, (void *)__pyx_methods_9uharfbuzz_9_harfbuzz_ColorLine},
  {Py_tp_getset, (void *)__pyx_getsets_9uharfbuzz_9_harfbuzz_ColorLine},
  {Py_tp_new, (void *)__pyx_tp_new_9uharfbuzz_9_harfbuzz_ColorLine},
  {0, 0},
};
static PyType_Spec __pyx_type_9uharfbuzz_9_harfbuzz_ColorLine_spec = {
  "uharfbuzz._harfbuzz.ColorLine",
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_9uharfbuzz_9_harfbuzz_ColorLine_slots,
};
#else

static PyTypeObject __pyx_type_9uharfbuzz_9_harfbuzz_ColorLine = {
  PyVarObject_HEAD_INIT(0, 0)
  "uharfbuzz._harfbuzz.""ColorLine", /*tp_name*/
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_ColorLine, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_9uharfbuzz_9_harfbuzz_ColorLine, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_9uharfbuzz_9_harfbuzz_ColorLine, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9uharfbuzz_9_harfbuzz_ColorLine, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_PaintFuncs(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)o);
  p->_push_transform_func = Py_None; Py_INCREF(Py_None);
  p->_pop_transform_func = Py_None; Py_INCREF(Py_None);
  p->_color_glyph_func = Py_None; Py_INCREF(Py_None);
  p->_push_clip_glyph_func = Py_None; Py_INCREF(Py_None);
  p->_push_clip_rectangle_func = Py_None; Py_INCREF(Py_None);
  p->_pop_clip_func = Py_None; Py_INCREF(Py_None);
  p->_color_func = Py_None; Py_INCREF(Py_None);
  p->_image_func = Py_None; Py_INCREF(Py_None);
  p->_linear_gradient_func = Py_None; Py_INCREF(Py_None);
  p->_radial_gradient_func = Py_None; Py_INCREF(Py_None);
  p->_sweep_gradient_func = Py_None; Py_INCREF(Py_None);
  p->_push_group_func = Py_None; Py_INCREF(Py_None);
  p->_pop_group_func = Py_None; Py_INCREF(Py_None);
  p->_custom_palette_color_func = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_PaintFuncs(PyObject *o) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_PaintFuncs) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->_push_transform_func);
  Py_CLEAR(p->_pop_transform_func);
  Py_CLEAR(p->_color_glyph_func);
  Py_CLEAR(p->_push_clip_glyph_func);
  Py_CLEAR(p->_push_clip_rectangle_func);
  Py_CLEAR(p->_pop_clip_func);
  Py_CLEAR(p->_color_func);
  Py_CLEAR(p->_image_func);
  Py_CLEAR(p->_linear_gradient_func);
  Py_CLEAR(p->_radial_gradient_func);
  Py_CLEAR(p->_sweep_gradient_func);
  Py_CLEAR(p->_push_group_func);
  Py_CLEAR(p->_pop_group_func);
  Py_CLEAR(p->_custom_palette_color_func);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_9uharfbuzz_9_harfbuzz_PaintFuncs(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)o;
  if (p->_push_transform_func) {
    e = (*v)(p->_push_transform_func, a); if (e) return e;
  }
  if (p->_pop_transform_func) {
    e = (*v)(p->_pop_transform_func, a); if (e) return e;
  }
  if (p->_color_glyph_func) {
    e = (*v)(p->_color_glyph_func, a); if (e) return e;
  }
  if (p->_push_clip_glyph_func) {
    e = (*v)(p->_push_clip_glyph_func, a); if (e) return e;
  }
  if (p->_push_clip_rectangle_func) {
    e = (*v)(p->_push_clip_rectangle_func, a); if (e) return e;
  }
  if (p->_pop_clip_func) {
    e = (*v)(p->_pop_clip_func, a); if (e) return e;
  }
  if (p->_color_func) {
    e = (*v)(p->_color_func, a); if (e) return e;
  }
  if (p->_image_func) {
    e = (*v)(p->_image_func, a); if (e) return e;
  }
  if (p->_linear_gradient_func) {
    e = (*v)(p->_linear_gradient_func, a); if (e) return e;
  }
  if (p->_radial_gradient_func) {
    e = (*v)(p->_radial_gradient_func, a); if (e) return e;
  }
  if (p->_sweep_gradient_func) {
    e = (*v)(p->_sweep_gradient_func, a); if (e) return e;
  }
  if (p->_push_group_func) {
    e = (*v)(p->_push_group_func, a); if (e) return e;
  }
  if (p->_pop_group_func) {
    e = (*v)(p->_pop_group_func, a); if (e) return e;
  }
  if (p->_custom_palette_color_func) {
    e = (*v)(p->_custom_palette_color_func, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9uharfbuzz_9_harfbuzz_PaintFuncs(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs *)o;
  tmp = ((PyObject*)p->_push_transform_func);
  p->_push_transform_func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_pop_transform_func);
  p->_pop_transform_func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_color_glyph_func);
  p->_color_glyph_func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_push_clip_glyph_func);
  p->_push_clip_glyph_func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_push_clip_rectangle_func);
  p->_push_clip_rectangle_func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_pop_clip_func);
  p->_pop_clip_func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_color_func);
  p->_color_func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_image_func);
  p->_image_func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_linear_gradient_func);
  p->_linear_gradient_func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_radial_gradient_func);
  p->_radial_gradient_func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_sweep_gradient_func);
  p->_sweep_gradient_func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_push_group_func);
  p->_push_group_func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_pop_group_func);
  p->_pop_group_func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_custom_palette_color_func);
  p->_custom_palette_color_func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_9uharfbuzz_9_harfbuzz_PaintFuncs[] = {
  {"set_push_transform_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_5set_push_transform_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_pop_transform_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_7set_pop_transform_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_color_glyph_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_9set_color_glyph_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_push_clip_glyph_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_11set_push_clip_glyph_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_push_clip_rectangle_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_13set_push_clip_rectangle_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_pop_clip_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_15set_pop_clip_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_color_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_17set_color_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_image_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_19set_image_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_linear_gradient_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_21set_linear_gradient_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_radial_gradient_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_23set_radial_gradient_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_sweep_gradient_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_25set_sweep_gradient_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_push_group_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_27set_push_group_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_pop_group_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_29set_pop_group_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_custom_palette_color_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_31set_custom_palette_color_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_33__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10PaintFuncs_35__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9uharfbuzz_9_harfbuzz_PaintFuncs_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_PaintFuncs},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_9uharfbuzz_9_harfbuzz_PaintFuncs},
  {Py_tp_clear, (void *)__pyx_tp_clear_9uharfbuzz_9_harfbuzz_PaintFuncs},
  {Py_tp_methods, (void *)__pyx_methods_9uharfbuzz_9_harfbuzz_PaintFuncs},
  {Py_tp_new, (void *)__pyx_tp_new_9uharfbuzz_9_harfbuzz_PaintFuncs},
  {0, 0},
};
static PyType_Spec __pyx_type_9uharfbuzz_9_harfbuzz_PaintFuncs_spec = {
  "uharfbuzz._harfbuzz.PaintFuncs",
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_9uharfbuzz_9_harfbuzz_PaintFuncs_slots,
};
#else

static PyTypeObject __pyx_type_9uharfbuzz_9_harfbuzz_PaintFuncs = {
  PyVarObject_HEAD_INIT(0, 0)
  "uharfbuzz._harfbuzz.""PaintFuncs", /*tp_name*/
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_PaintFuncs), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_PaintFuncs, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9uharfbuzz_9_harfbuzz_PaintFuncs, /*tp_traverse*/
  __pyx_tp_clear_9uharfbuzz_9_harfbuzz_PaintFuncs, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_9uharfbuzz_9_harfbuzz_PaintFuncs, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9uharfbuzz_9_harfbuzz_PaintFuncs, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_DrawFuncs(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *)o);
  p->_move_to_func = Py_None; Py_INCREF(Py_None);
  p->_line_to_func = Py_None; Py_INCREF(Py_None);
  p->_cubic_to_func = Py_None; Py_INCREF(Py_None);
  p->_quadratic_to_func = Py_None; Py_INCREF(Py_None);
  p->_close_path_func = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_DrawFuncs(PyObject *o) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_DrawFuncs) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->_move_to_func);
  Py_CLEAR(p->_line_to_func);
  Py_CLEAR(p->_cubic_to_func);
  Py_CLEAR(p->_quadratic_to_func);
  Py_CLEAR(p->_close_path_func);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_9uharfbuzz_9_harfbuzz_DrawFuncs(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *)o;
  if (p->_move_to_func) {
    e = (*v)(p->_move_to_func, a); if (e) return e;
  }
  if (p->_line_to_func) {
    e = (*v)(p->_line_to_func, a); if (e) return e;
  }
  if (p->_cubic_to_func) {
    e = (*v)(p->_cubic_to_func, a); if (e) return e;
  }
  if (p->_quadratic_to_func) {
    e = (*v)(p->_quadratic_to_func, a); if (e) return e;
  }
  if (p->_close_path_func) {
    e = (*v)(p->_close_path_func, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9uharfbuzz_9_harfbuzz_DrawFuncs(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs *)o;
  tmp = ((PyObject*)p->_move_to_func);
  p->_move_to_func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_line_to_func);
  p->_line_to_func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_cubic_to_func);
  p->_cubic_to_func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_quadratic_to_func);
  p->_quadratic_to_func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_close_path_func);
  p->_close_path_func = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_9uharfbuzz_9_harfbuzz_DrawFuncs[] = {
  {"get_glyph_shape", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_5get_glyph_shape, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"draw_glyph", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_7draw_glyph, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_move_to_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_9set_move_to_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_line_to_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_11set_line_to_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_cubic_to_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_13set_cubic_to_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_quadratic_to_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_15set_quadratic_to_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_close_path_func", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_17set_close_path_func, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_19__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_9DrawFuncs_21__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9uharfbuzz_9_harfbuzz_DrawFuncs_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_DrawFuncs},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_9uharfbuzz_9_harfbuzz_DrawFuncs},
  {Py_tp_clear, (void *)__pyx_tp_clear_9uharfbuzz_9_harfbuzz_DrawFuncs},
  {Py_tp_methods, (void *)__pyx_methods_9uharfbuzz_9_harfbuzz_DrawFuncs},
  {Py_tp_new, (void *)__pyx_tp_new_9uharfbuzz_9_harfbuzz_DrawFuncs},
  {0, 0},
};
static PyType_Spec __pyx_type_9uharfbuzz_9_harfbuzz_DrawFuncs_spec = {
  "uharfbuzz._harfbuzz.DrawFuncs",
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_9uharfbuzz_9_harfbuzz_DrawFuncs_slots,
};
#else

static PyTypeObject __pyx_type_9uharfbuzz_9_harfbuzz_DrawFuncs = {
  PyVarObject_HEAD_INIT(0, 0)
  "uharfbuzz._harfbuzz.""DrawFuncs", /*tp_name*/
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_DrawFuncs), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_DrawFuncs, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9uharfbuzz_9_harfbuzz_DrawFuncs, /*tp_traverse*/
  __pyx_tp_clear_9uharfbuzz_9_harfbuzz_DrawFuncs, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_9uharfbuzz_9_harfbuzz_DrawFuncs, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9uharfbuzz_9_harfbuzz_DrawFuncs, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_HBObject __pyx_vtable_9uharfbuzz_9_harfbuzz_HBObject;

static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_HBObject(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_HBObject *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_HBObject *)o);
  p->__pyx_vtab = __pyx_vtabptr_9uharfbuzz_9_harfbuzz_HBObject;
  if (unlikely(__pyx_pw_9uharfbuzz_9_harfbuzz_8HBObject_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_HBObject(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_HBObject) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_9uharfbuzz_9_harfbuzz_8HBObject_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static PyMethodDef __pyx_methods_9uharfbuzz_9_harfbuzz_HBObject[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_8HBObject_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_8HBObject_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9uharfbuzz_9_harfbuzz_HBObject_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_HBObject},
  {Py_tp_methods, (void *)__pyx_methods_9uharfbuzz_9_harfbuzz_HBObject},
  {Py_tp_new, (void *)__pyx_tp_new_9uharfbuzz_9_harfbuzz_HBObject},
  {0, 0},
};
static PyType_Spec __pyx_type_9uharfbuzz_9_harfbuzz_HBObject_spec = {
  "uharfbuzz._harfbuzz.HBObject",
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_HBObject),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_9uharfbuzz_9_harfbuzz_HBObject_slots,
};
#else

static PyTypeObject __pyx_type_9uharfbuzz_9_harfbuzz_HBObject = {
  PyVarObject_HEAD_INIT(0, 0)
  "uharfbuzz._harfbuzz.""HBObject", /*tp_name*/
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_HBObject), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_HBObject, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_9uharfbuzz_9_harfbuzz_HBObject, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9uharfbuzz_9_harfbuzz_HBObject, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_SubsetInput(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  if (unlikely(__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_1__cinit__(o, __pyx_empty_tuple, NULL) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_SubsetInput(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_SubsetInput) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_11SubsetInput_unicode_set(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_11unicode_set_1__get__(o);
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_11SubsetInput_glyph_set(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_9glyph_set_1__get__(o);
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_11SubsetInput_no_subset_table_tag_set(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_23no_subset_table_tag_set_1__get__(o);
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_11SubsetInput_drop_table_tag_set(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_18drop_table_tag_set_1__get__(o);
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_11SubsetInput_name_id_set(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_11name_id_set_1__get__(o);
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_11SubsetInput_name_lang_id_set(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_16name_lang_id_set_1__get__(o);
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_11SubsetInput_layout_feature_tag_set(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_22layout_feature_tag_set_1__get__(o);
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_11SubsetInput_layout_script_tag_set(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_21layout_script_tag_set_1__get__(o);
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_11SubsetInput_flags(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_5flags_1__get__(o);
}

static int __pyx_setprop_9uharfbuzz_9_harfbuzz_11SubsetInput_flags(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_5flags_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyMethodDef __pyx_methods_9uharfbuzz_9_harfbuzz_SubsetInput[] = {
  {"subset", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_5subset, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"keep_everything", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_7keep_everything, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"pin_axis_to_default", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_9pin_axis_to_default, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"pin_axis_location", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_11pin_axis_location, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"sets", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_13sets, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_15__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_11SubsetInput_17__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_9uharfbuzz_9_harfbuzz_SubsetInput[] = {
  {(char *)"unicode_set", __pyx_getprop_9uharfbuzz_9_harfbuzz_11SubsetInput_unicode_set, 0, (char *)0, 0},
  {(char *)"glyph_set", __pyx_getprop_9uharfbuzz_9_harfbuzz_11SubsetInput_glyph_set, 0, (char *)0, 0},
  {(char *)"no_subset_table_tag_set", __pyx_getprop_9uharfbuzz_9_harfbuzz_11SubsetInput_no_subset_table_tag_set, 0, (char *)0, 0},
  {(char *)"drop_table_tag_set", __pyx_getprop_9uharfbuzz_9_harfbuzz_11SubsetInput_drop_table_tag_set, 0, (char *)0, 0},
  {(char *)"name_id_set", __pyx_getprop_9uharfbuzz_9_harfbuzz_11SubsetInput_name_id_set, 0, (char *)0, 0},
  {(char *)"name_lang_id_set", __pyx_getprop_9uharfbuzz_9_harfbuzz_11SubsetInput_name_lang_id_set, 0, (char *)0, 0},
  {(char *)"layout_feature_tag_set", __pyx_getprop_9uharfbuzz_9_harfbuzz_11SubsetInput_layout_feature_tag_set, 0, (char *)0, 0},
  {(char *)"layout_script_tag_set", __pyx_getprop_9uharfbuzz_9_harfbuzz_11SubsetInput_layout_script_tag_set, 0, (char *)0, 0},
  {(char *)"flags", __pyx_getprop_9uharfbuzz_9_harfbuzz_11SubsetInput_flags, __pyx_setprop_9uharfbuzz_9_harfbuzz_11SubsetInput_flags, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9uharfbuzz_9_harfbuzz_SubsetInput_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_SubsetInput},
  {Py_tp_methods, (void *)__pyx_methods_9uharfbuzz_9_harfbuzz_SubsetInput},
  {Py_tp_getset, (void *)__pyx_getsets_9uharfbuzz_9_harfbuzz_SubsetInput},
  {Py_tp_new, (void *)__pyx_tp_new_9uharfbuzz_9_harfbuzz_SubsetInput},
  {0, 0},
};
static PyType_Spec __pyx_type_9uharfbuzz_9_harfbuzz_SubsetInput_spec = {
  "uharfbuzz._harfbuzz.SubsetInput",
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_9uharfbuzz_9_harfbuzz_SubsetInput_slots,
};
#else

static PyTypeObject __pyx_type_9uharfbuzz_9_harfbuzz_SubsetInput = {
  PyVarObject_HEAD_INIT(0, 0)
  "uharfbuzz._harfbuzz.""SubsetInput", /*tp_name*/
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetInput), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_SubsetInput, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_9uharfbuzz_9_harfbuzz_SubsetInput, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_9uharfbuzz_9_harfbuzz_SubsetInput, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9uharfbuzz_9_harfbuzz_SubsetInput, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_SubsetPlan(PyTypeObject *t, PyObject *a, PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  if (unlikely(__pyx_pw_9uharfbuzz_9_harfbuzz_10SubsetPlan_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_SubsetPlan(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_SubsetPlan) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_9uharfbuzz_9_harfbuzz_10SubsetPlan_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_10SubsetPlan_old_to_new_glyph_mapping(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_10SubsetPlan_24old_to_new_glyph_mapping_1__get__(o);
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_10SubsetPlan_new_to_old_glyph_mapping(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_10SubsetPlan_24new_to_old_glyph_mapping_1__get__(o);
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_10SubsetPlan_unicode_to_old_glyph_mapping(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_10SubsetPlan_28unicode_to_old_glyph_mapping_1__get__(o);
}

static PyMethodDef __pyx_methods_9uharfbuzz_9_harfbuzz_SubsetPlan[] = {
  {"execute", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10SubsetPlan_5execute, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10SubsetPlan_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_10SubsetPlan_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_9uharfbuzz_9_harfbuzz_SubsetPlan[] = {
  {(char *)"old_to_new_glyph_mapping", __pyx_getprop_9uharfbuzz_9_harfbuzz_10SubsetPlan_old_to_new_glyph_mapping, 0, (char *)0, 0},
  {(char *)"new_to_old_glyph_mapping", __pyx_getprop_9uharfbuzz_9_harfbuzz_10SubsetPlan_new_to_old_glyph_mapping, 0, (char *)0, 0},
  {(char *)"unicode_to_old_glyph_mapping", __pyx_getprop_9uharfbuzz_9_harfbuzz_10SubsetPlan_unicode_to_old_glyph_mapping, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9uharfbuzz_9_harfbuzz_SubsetPlan_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_SubsetPlan},
  {Py_tp_methods, (void *)__pyx_methods_9uharfbuzz_9_harfbuzz_SubsetPlan},
  {Py_tp_getset, (void *)__pyx_getsets_9uharfbuzz_9_harfbuzz_SubsetPlan},
  {Py_tp_new, (void *)__pyx_tp_new_9uharfbuzz_9_harfbuzz_SubsetPlan},
  {0, 0},
};
static PyType_Spec __pyx_type_9uharfbuzz_9_harfbuzz_SubsetPlan_spec = {
  "uharfbuzz._harfbuzz.SubsetPlan",
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetPlan),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_9uharfbuzz_9_harfbuzz_SubsetPlan_slots,
};
#else

static PyTypeObject __pyx_type_9uharfbuzz_9_harfbuzz_SubsetPlan = {
  PyVarObject_HEAD_INIT(0, 0)
  "uharfbuzz._harfbuzz.""SubsetPlan", /*tp_name*/
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SubsetPlan), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_SubsetPlan, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_9uharfbuzz_9_harfbuzz_SubsetPlan, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_9uharfbuzz_9_harfbuzz_SubsetPlan, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9uharfbuzz_9_harfbuzz_SubsetPlan, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_Set __pyx_vtable_9uharfbuzz_9_harfbuzz_Set;

static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_Set(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)o);
  p->__pyx_vtab = __pyx_vtabptr_9uharfbuzz_9_harfbuzz_Set;
  if (unlikely(__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_Set(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_Set) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_9uharfbuzz_9_harfbuzz_3Set_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static PyObject *__pyx_tp_richcompare_9uharfbuzz_9_harfbuzz_Set(PyObject *o1, PyObject *o2, int op) {
  switch (op) {
    case Py_EQ: {
      return __pyx_pw_9uharfbuzz_9_harfbuzz_3Set_31__eq__(o1, o2);
    }
    case Py_NE: {
      PyObject *ret;
      ret = __pyx_pw_9uharfbuzz_9_harfbuzz_3Set_31__eq__(o1, o2);
      if (likely(ret && ret != Py_NotImplemented)) {
        int b = __Pyx_PyObject_IsTrue(ret);
        Py_DECREF(ret);
        if (unlikely(b < 0)) return NULL;
        ret = (b) ? Py_False : Py_True;
        Py_INCREF(ret);
      }
      return ret;
    }
    default: {
      return __Pyx_NewRef(Py_NotImplemented);
    }
  }
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_3Set_min(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_3Set_3min_1__get__(o);
}

static PyObject *__pyx_getprop_9uharfbuzz_9_harfbuzz_3Set_max(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_3Set_3max_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_9uharfbuzz_9_harfbuzz_3Set_63__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_3Set_63__repr__(self);
}

static PyMethodDef __pyx_methods_9uharfbuzz_9_harfbuzz_Set[] = {
  {"copy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_5copy, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__copy__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_7__copy__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"clear", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_9clear, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"invert", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_13invert, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"is_inverted", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_15is_inverted, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"add", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_19add, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"add_range", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_21add_range, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"remove", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_23remove, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"discard", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_25discard, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"del_range", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_27del_range, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_is_equal", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_29_is_equal, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"issubset", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_33issubset, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"issuperset", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_35issuperset, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_set", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_37_set, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_39set, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_update", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_41_update, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"update", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_43update, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"intersection_update", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_47intersection_update, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"difference_update", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_51difference_update, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"symmetric_difference_update", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_55symmetric_difference_update, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_9uharfbuzz_9_harfbuzz_3Set_63__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_65__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_67__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_9uharfbuzz_9_harfbuzz_Set[] = {
  {(char *)"min", __pyx_getprop_9uharfbuzz_9_harfbuzz_3Set_min, 0, (char *)0, 0},
  {(char *)"max", __pyx_getprop_9uharfbuzz_9_harfbuzz_3Set_max, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9uharfbuzz_9_harfbuzz_Set_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_Set},
  {Py_tp_repr, (void *)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_63__repr__},
  {Py_nb_bool, (void *)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_11__bool__},
  {Py_nb_inplace_subtract, (void *)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_53__isub__},
  {Py_nb_inplace_and, (void *)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_49__iand__},
  {Py_nb_inplace_xor, (void *)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_57__ixor__},
  {Py_nb_inplace_or, (void *)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_45__ior__},
  {Py_sq_length, (void *)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_59__len__},
  {Py_sq_contains, (void *)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_17__contains__},
  {Py_mp_length, (void *)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_59__len__},
  {Py_tp_richcompare, (void *)__pyx_tp_richcompare_9uharfbuzz_9_harfbuzz_Set},
  {Py_tp_iter, (void *)__pyx_pw_9uharfbuzz_9_harfbuzz_3Set_61__iter__},
  {Py_tp_methods, (void *)__pyx_methods_9uharfbuzz_9_harfbuzz_Set},
  {Py_tp_getset, (void *)__pyx_getsets_9uharfbuzz_9_harfbuzz_Set},
  {Py_tp_new, (void *)__pyx_tp_new_9uharfbuzz_9_harfbuzz_Set},
  {0, 0},
};
static PyType_Spec __pyx_type_9uharfbuzz_9_harfbuzz_Set_spec = {
  "uharfbuzz._harfbuzz.Set",
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_9uharfbuzz_9_harfbuzz_Set_slots,
};
#else

static PyNumberMethods __pyx_tp_as_number_Set = {
  0, /*nb_add*/
  0, /*nb_subtract*/
  0, /*nb_multiply*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_divide*/
  #endif
  0, /*nb_remainder*/
  0, /*nb_divmod*/
  0, /*nb_power*/
  0, /*nb_negative*/
  0, /*nb_positive*/
  0, /*nb_absolute*/
  __pyx_pw_9uharfbuzz_9_harfbuzz_3Set_11__bool__, /*nb_bool*/
  0, /*nb_invert*/
  0, /*nb_lshift*/
  0, /*nb_rshift*/
  0, /*nb_and*/
  0, /*nb_xor*/
  0, /*nb_or*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_coerce*/
  #endif
  0, /*nb_int*/
  #if PY_MAJOR_VERSION < 3
  0, /*nb_long*/
  #else
  0, /*reserved*/
  #endif
  0, /*nb_float*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_oct*/
  #endif
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_hex*/
  #endif
  0, /*nb_inplace_add*/
  __pyx_pw_9uharfbuzz_9_harfbuzz_3Set_53__isub__, /*nb_inplace_subtract*/
  0, /*nb_inplace_multiply*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_inplace_divide*/
  #endif
  0, /*nb_inplace_remainder*/
  0, /*nb_inplace_power*/
  0, /*nb_inplace_lshift*/
  0, /*nb_inplace_rshift*/
  __pyx_pw_9uharfbuzz_9_harfbuzz_3Set_49__iand__, /*nb_inplace_and*/
  __pyx_pw_9uharfbuzz_9_harfbuzz_3Set_57__ixor__, /*nb_inplace_xor*/
  __pyx_pw_9uharfbuzz_9_harfbuzz_3Set_45__ior__, /*nb_inplace_or*/
  0, /*nb_floor_divide*/
  0, /*nb_true_divide*/
  0, /*nb_inplace_floor_divide*/
  0, /*nb_inplace_true_divide*/
  0, /*nb_index*/
  #if PY_VERSION_HEX >= 0x03050000
  0, /*nb_matrix_multiply*/
  #endif
  #if PY_VERSION_HEX >= 0x03050000
  0, /*nb_inplace_matrix_multiply*/
  #endif
};

static PySequenceMethods __pyx_tp_as_sequence_Set = {
  __pyx_pw_9uharfbuzz_9_harfbuzz_3Set_59__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  0, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  __pyx_pw_9uharfbuzz_9_harfbuzz_3Set_17__contains__, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_Set = {
  __pyx_pw_9uharfbuzz_9_harfbuzz_3Set_59__len__, /*mp_length*/
  0, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_9uharfbuzz_9_harfbuzz_Set = {
  PyVarObject_HEAD_INIT(0, 0)
  "uharfbuzz._harfbuzz.""Set", /*tp_name*/
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_Set, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_9uharfbuzz_9_harfbuzz_3Set_63__repr__, /*tp_repr*/
  &__pyx_tp_as_number_Set, /*tp_as_number*/
  &__pyx_tp_as_sequence_Set, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_Set, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  __pyx_tp_richcompare_9uharfbuzz_9_harfbuzz_Set, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_9uharfbuzz_9_harfbuzz_3Set_61__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_9uharfbuzz_9_harfbuzz_Set, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_9uharfbuzz_9_harfbuzz_Set, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9uharfbuzz_9_harfbuzz_Set, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_SetIter(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_SetIter *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_SetIter *)o);
  p->s = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_9uharfbuzz_9_harfbuzz_7SetIter_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_SetIter(PyObject *o) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_SetIter *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_SetIter *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_SetIter) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->s);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_9uharfbuzz_9_harfbuzz_SetIter(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_SetIter *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_SetIter *)o;
  if (p->s) {
    e = (*v)(((PyObject *)p->s), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9uharfbuzz_9_harfbuzz_SetIter(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_SetIter *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_SetIter *)o;
  tmp = ((PyObject*)p->s);
  p->s = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_specialmethod___pyx_pw_9uharfbuzz_9_harfbuzz_7SetIter_5__next__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  PyObject *res = __pyx_pw_9uharfbuzz_9_harfbuzz_7SetIter_5__next__(self);
  if (!res && !PyErr_Occurred()) { PyErr_SetNone(PyExc_StopIteration); }
  return res;
}

static PyMethodDef __pyx_methods_9uharfbuzz_9_harfbuzz_SetIter[] = {
  {"__next__", (PyCFunction)__pyx_specialmethod___pyx_pw_9uharfbuzz_9_harfbuzz_7SetIter_5__next__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_7SetIter_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_7SetIter_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9uharfbuzz_9_harfbuzz_SetIter_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_SetIter},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_9uharfbuzz_9_harfbuzz_SetIter},
  {Py_tp_clear, (void *)__pyx_tp_clear_9uharfbuzz_9_harfbuzz_SetIter},
  {Py_tp_iter, (void *)__pyx_pw_9uharfbuzz_9_harfbuzz_7SetIter_3__iter__},
  {Py_tp_iternext, (void *)__pyx_pw_9uharfbuzz_9_harfbuzz_7SetIter_5__next__},
  {Py_tp_methods, (void *)__pyx_methods_9uharfbuzz_9_harfbuzz_SetIter},
  {Py_tp_new, (void *)__pyx_tp_new_9uharfbuzz_9_harfbuzz_SetIter},
  {0, 0},
};
static PyType_Spec __pyx_type_9uharfbuzz_9_harfbuzz_SetIter_spec = {
  "uharfbuzz._harfbuzz.SetIter",
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SetIter),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_9uharfbuzz_9_harfbuzz_SetIter_slots,
};
#else

static PyTypeObject __pyx_type_9uharfbuzz_9_harfbuzz_SetIter = {
  PyVarObject_HEAD_INIT(0, 0)
  "uharfbuzz._harfbuzz.""SetIter", /*tp_name*/
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_SetIter), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_SetIter, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9uharfbuzz_9_harfbuzz_SetIter, /*tp_traverse*/
  __pyx_tp_clear_9uharfbuzz_9_harfbuzz_SetIter, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_9uharfbuzz_9_harfbuzz_7SetIter_3__iter__, /*tp_iter*/
  __pyx_pw_9uharfbuzz_9_harfbuzz_7SetIter_5__next__, /*tp_iternext*/
  __pyx_methods_9uharfbuzz_9_harfbuzz_SetIter, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9uharfbuzz_9_harfbuzz_SetIter, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_9uharfbuzz_9_harfbuzz_Map __pyx_vtable_9uharfbuzz_9_harfbuzz_Map;

static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_Map(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *)o);
  p->__pyx_vtab = __pyx_vtabptr_9uharfbuzz_9_harfbuzz_Map;
  if (unlikely(__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_Map(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_Map) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_9uharfbuzz_9_harfbuzz_3Map_3__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}
static PyObject *__pyx_sq_item_9uharfbuzz_9_harfbuzz_Map(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_9uharfbuzz_9_harfbuzz_Map(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_pw_9uharfbuzz_9_harfbuzz_3Map_23__setitem__(o, i, v);
  }
  else {
    return __pyx_pw_9uharfbuzz_9_harfbuzz_3Map_31__delitem__(o, i);
  }
}

static PyObject *__pyx_tp_richcompare_9uharfbuzz_9_harfbuzz_Map(PyObject *o1, PyObject *o2, int op) {
  switch (op) {
    case Py_EQ: {
      return __pyx_pw_9uharfbuzz_9_harfbuzz_3Map_21__eq__(o1, o2);
    }
    case Py_NE: {
      PyObject *ret;
      ret = __pyx_pw_9uharfbuzz_9_harfbuzz_3Map_21__eq__(o1, o2);
      if (likely(ret && ret != Py_NotImplemented)) {
        int b = __Pyx_PyObject_IsTrue(ret);
        Py_DECREF(ret);
        if (unlikely(b < 0)) return NULL;
        ret = (b) ? Py_False : Py_True;
        Py_INCREF(ret);
      }
      return ret;
    }
    default: {
      return __Pyx_NewRef(Py_NotImplemented);
    }
  }
}

static PyObject *__pyx_specialmethod___pyx_pw_9uharfbuzz_9_harfbuzz_3Map_41__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_9uharfbuzz_9_harfbuzz_3Map_41__repr__(self);
}

static PyMethodDef __pyx_methods_9uharfbuzz_9_harfbuzz_Map[] = {
  {"copy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_5copy, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__copy__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_7__copy__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_update", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_9_update, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"update", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_11update, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"clear", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_13clear, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_is_equal", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_19_is_equal, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_25get, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"items", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_33items, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"keys", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_35keys, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"values", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_37values, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_9uharfbuzz_9_harfbuzz_3Map_41__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_43__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_45__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9uharfbuzz_9_harfbuzz_Map_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_Map},
  {Py_tp_repr, (void *)__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_41__repr__},
  {Py_nb_bool, (void *)__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_15__bool__},
  {Py_sq_length, (void *)__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_17__len__},
  {Py_sq_item, (void *)__pyx_sq_item_9uharfbuzz_9_harfbuzz_Map},
  {Py_sq_contains, (void *)__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_29__contains__},
  {Py_mp_length, (void *)__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_17__len__},
  {Py_mp_subscript, (void *)__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_27__getitem__},
  {Py_mp_ass_subscript, (void *)__pyx_mp_ass_subscript_9uharfbuzz_9_harfbuzz_Map},
  {Py_tp_richcompare, (void *)__pyx_tp_richcompare_9uharfbuzz_9_harfbuzz_Map},
  {Py_tp_iter, (void *)__pyx_pw_9uharfbuzz_9_harfbuzz_3Map_39__iter__},
  {Py_tp_methods, (void *)__pyx_methods_9uharfbuzz_9_harfbuzz_Map},
  {Py_tp_new, (void *)__pyx_tp_new_9uharfbuzz_9_harfbuzz_Map},
  {0, 0},
};
static PyType_Spec __pyx_type_9uharfbuzz_9_harfbuzz_Map_spec = {
  "uharfbuzz._harfbuzz.Map",
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_9uharfbuzz_9_harfbuzz_Map_slots,
};
#else

static PyNumberMethods __pyx_tp_as_number_Map = {
  0, /*nb_add*/
  0, /*nb_subtract*/
  0, /*nb_multiply*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_divide*/
  #endif
  0, /*nb_remainder*/
  0, /*nb_divmod*/
  0, /*nb_power*/
  0, /*nb_negative*/
  0, /*nb_positive*/
  0, /*nb_absolute*/
  __pyx_pw_9uharfbuzz_9_harfbuzz_3Map_15__bool__, /*nb_bool*/
  0, /*nb_invert*/
  0, /*nb_lshift*/
  0, /*nb_rshift*/
  0, /*nb_and*/
  0, /*nb_xor*/
  0, /*nb_or*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_coerce*/
  #endif
  0, /*nb_int*/
  #if PY_MAJOR_VERSION < 3
  0, /*nb_long*/
  #else
  0, /*reserved*/
  #endif
  0, /*nb_float*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_oct*/
  #endif
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_hex*/
  #endif
  0, /*nb_inplace_add*/
  0, /*nb_inplace_subtract*/
  0, /*nb_inplace_multiply*/
  #if PY_MAJOR_VERSION < 3 || (CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03050000)
  0, /*nb_inplace_divide*/
  #endif
  0, /*nb_inplace_remainder*/
  0, /*nb_inplace_power*/
  0, /*nb_inplace_lshift*/
  0, /*nb_inplace_rshift*/
  0, /*nb_inplace_and*/
  0, /*nb_inplace_xor*/
  0, /*nb_inplace_or*/
  0, /*nb_floor_divide*/
  0, /*nb_true_divide*/
  0, /*nb_inplace_floor_divide*/
  0, /*nb_inplace_true_divide*/
  0, /*nb_index*/
  #if PY_VERSION_HEX >= 0x03050000
  0, /*nb_matrix_multiply*/
  #endif
  #if PY_VERSION_HEX >= 0x03050000
  0, /*nb_inplace_matrix_multiply*/
  #endif
};

static PySequenceMethods __pyx_tp_as_sequence_Map = {
  __pyx_pw_9uharfbuzz_9_harfbuzz_3Map_17__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_9uharfbuzz_9_harfbuzz_Map, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  __pyx_pw_9uharfbuzz_9_harfbuzz_3Map_29__contains__, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_Map = {
  __pyx_pw_9uharfbuzz_9_harfbuzz_3Map_17__len__, /*mp_length*/
  __pyx_pw_9uharfbuzz_9_harfbuzz_3Map_27__getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_9uharfbuzz_9_harfbuzz_Map, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_9uharfbuzz_9_harfbuzz_Map = {
  PyVarObject_HEAD_INIT(0, 0)
  "uharfbuzz._harfbuzz.""Map", /*tp_name*/
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_Map, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_9uharfbuzz_9_harfbuzz_3Map_41__repr__, /*tp_repr*/
  &__pyx_tp_as_number_Map, /*tp_as_number*/
  &__pyx_tp_as_sequence_Map, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_Map, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  __pyx_tp_richcompare_9uharfbuzz_9_harfbuzz_Map, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_9uharfbuzz_9_harfbuzz_3Map_39__iter__, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_9uharfbuzz_9_harfbuzz_Map, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9uharfbuzz_9_harfbuzz_Map, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz_MapIter(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_MapIter *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_MapIter *)o);
  p->m = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_9uharfbuzz_9_harfbuzz_7MapIter_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_MapIter(PyObject *o) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_MapIter *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_MapIter *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_MapIter) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->m);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_9uharfbuzz_9_harfbuzz_MapIter(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_MapIter *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_MapIter *)o;
  if (p->m) {
    e = (*v)(((PyObject *)p->m), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_9uharfbuzz_9_harfbuzz_MapIter(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz_MapIter *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_MapIter *)o;
  tmp = ((PyObject*)p->m);
  p->m = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_specialmethod___pyx_pw_9uharfbuzz_9_harfbuzz_7MapIter_5__next__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  PyObject *res = __pyx_pw_9uharfbuzz_9_harfbuzz_7MapIter_5__next__(self);
  if (!res && !PyErr_Occurred()) { PyErr_SetNone(PyExc_StopIteration); }
  return res;
}

static PyMethodDef __pyx_methods_9uharfbuzz_9_harfbuzz_MapIter[] = {
  {"__next__", (PyCFunction)__pyx_specialmethod___pyx_pw_9uharfbuzz_9_harfbuzz_7MapIter_5__next__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_7MapIter_7__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_9uharfbuzz_9_harfbuzz_7MapIter_9__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9uharfbuzz_9_harfbuzz_MapIter_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_MapIter},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_9uharfbuzz_9_harfbuzz_MapIter},
  {Py_tp_clear, (void *)__pyx_tp_clear_9uharfbuzz_9_harfbuzz_MapIter},
  {Py_tp_iter, (void *)__pyx_pw_9uharfbuzz_9_harfbuzz_7MapIter_3__iter__},
  {Py_tp_iternext, (void *)__pyx_pw_9uharfbuzz_9_harfbuzz_7MapIter_5__next__},
  {Py_tp_methods, (void *)__pyx_methods_9uharfbuzz_9_harfbuzz_MapIter},
  {Py_tp_new, (void *)__pyx_tp_new_9uharfbuzz_9_harfbuzz_MapIter},
  {0, 0},
};
static PyType_Spec __pyx_type_9uharfbuzz_9_harfbuzz_MapIter_spec = {
  "uharfbuzz._harfbuzz.MapIter",
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_MapIter),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_9uharfbuzz_9_harfbuzz_MapIter_slots,
};
#else

static PyTypeObject __pyx_type_9uharfbuzz_9_harfbuzz_MapIter = {
  PyVarObject_HEAD_INIT(0, 0)
  "uharfbuzz._harfbuzz.""MapIter", /*tp_name*/
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz_MapIter), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz_MapIter, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9uharfbuzz_9_harfbuzz_MapIter, /*tp_traverse*/
  __pyx_tp_clear_9uharfbuzz_9_harfbuzz_MapIter, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_9uharfbuzz_9_harfbuzz_7MapIter_3__iter__, /*tp_iter*/
  __pyx_pw_9uharfbuzz_9_harfbuzz_7MapIter_5__next__, /*tp_iternext*/
  __pyx_methods_9uharfbuzz_9_harfbuzz_MapIter, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9uharfbuzz_9_harfbuzz_MapIter, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr *__pyx_freelist_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr[8];
static int __pyx_freecount_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr = 0;
#endif

static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr)))) {
    o = (PyObject*)__pyx_freelist_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr[--__pyx_freecount_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr(PyObject *o) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v_v);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr)))) {
    __pyx_freelist_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr[__pyx_freecount_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr++] = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr *)o;
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(((PyObject *)p->__pyx_genexpr_arg_0), a); if (e) return e;
  }
  if (p->__pyx_v_v) {
    e = (*v)(p->__pyx_v_v, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr_spec = {
  "uharfbuzz._harfbuzz.__pyx_scope_struct__genexpr",
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr_slots,
};
#else

static PyTypeObject __pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "uharfbuzz._harfbuzz.""__pyx_scope_struct__genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr *__pyx_freelist_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr[8];
static int __pyx_freecount_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr = 0;
#endif

static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr[--__pyx_freecount_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr(PyObject *o) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v_k);
  Py_CLEAR(p->__pyx_v_v);
  Py_CLEAR(p->__pyx_t_0);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr)))) {
    __pyx_freelist_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr[__pyx_freecount_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr++] = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr *)o;
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(((PyObject *)p->__pyx_genexpr_arg_0), a); if (e) return e;
  }
  if (p->__pyx_v_k) {
    e = (*v)(p->__pyx_v_k, a); if (e) return e;
  }
  if (p->__pyx_v_v) {
    e = (*v)(p->__pyx_v_v, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr_spec = {
  "uharfbuzz._harfbuzz.__pyx_scope_struct_1_genexpr",
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr_slots,
};
#else

static PyTypeObject __pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "uharfbuzz._harfbuzz.""__pyx_scope_struct_1_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr *__pyx_freelist_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr[8];
static int __pyx_freecount_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr = 0;
#endif

static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr[--__pyx_freecount_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr(PyObject *o) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v_k);
  Py_CLEAR(p->__pyx_v_v);
  Py_CLEAR(p->__pyx_t_0);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr)))) {
    __pyx_freelist_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr[__pyx_freecount_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr++] = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr *)o;
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(((PyObject *)p->__pyx_genexpr_arg_0), a); if (e) return e;
  }
  if (p->__pyx_v_k) {
    e = (*v)(p->__pyx_v_k, a); if (e) return e;
  }
  if (p->__pyx_v_v) {
    e = (*v)(p->__pyx_v_v, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr_spec = {
  "uharfbuzz._harfbuzz.__pyx_scope_struct_2_genexpr",
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr_slots,
};
#else

static PyTypeObject __pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "uharfbuzz._harfbuzz.""__pyx_scope_struct_2_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr *__pyx_freelist_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr[8];
static int __pyx_freecount_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr = 0;
#endif

static PyObject *__pyx_tp_new_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr)))) {
    o = (PyObject*)__pyx_freelist_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr[--__pyx_freecount_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr(PyObject *o) {
  struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v_k);
  Py_CLEAR(p->__pyx_v_v);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr)))) {
    __pyx_freelist_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr[__pyx_freecount_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr++] = ((struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr *p = (struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr *)o;
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(p->__pyx_genexpr_arg_0, a); if (e) return e;
  }
  if (p->__pyx_v_k) {
    e = (*v)(p->__pyx_v_k, a); if (e) return e;
  }
  if (p->__pyx_v_v) {
    e = (*v)(p->__pyx_v_v, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr_spec = {
  "uharfbuzz._harfbuzz.__pyx_scope_struct_3_genexpr",
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr_slots,
};
#else

static PyTypeObject __pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "uharfbuzz._harfbuzz.""__pyx_scope_struct_3_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif
/* #### Code section: pystring_table ### */

static int __Pyx_CreateStringTabAndInitStrings(void) {
  __Pyx_StringTabEntry __pyx_string_tab[] = {
    {&__pyx_n_s_ACCENT_BASE_HEIGHT, __pyx_k_ACCENT_BASE_HEIGHT, sizeof(__pyx_k_ACCENT_BASE_HEIGHT), 0, 0, 1, 1},
    {&__pyx_n_s_AXIS_HEIGHT, __pyx_k_AXIS_HEIGHT, sizeof(__pyx_k_AXIS_HEIGHT), 0, 0, 1, 1},
    {&__pyx_kp_u_All_shapers_failed, __pyx_k_All_shapers_failed, sizeof(__pyx_k_All_shapers_failed), 0, 1, 0, 0},
    {&__pyx_n_s_AssertionError, __pyx_k_AssertionError, sizeof(__pyx_k_AssertionError), 0, 0, 1, 1},
    {&__pyx_n_s_BASE_GLYPH, __pyx_k_BASE_GLYPH, sizeof(__pyx_k_BASE_GLYPH), 0, 0, 1, 1},
    {&__pyx_n_s_BOT, __pyx_k_BOT, sizeof(__pyx_k_BOT), 0, 0, 1, 1},
    {&__pyx_n_s_BOTTOM_LEFT, __pyx_k_BOTTOM_LEFT, sizeof(__pyx_k_BOTTOM_LEFT), 0, 0, 1, 1},
    {&__pyx_n_s_BOTTOM_RIGHT, __pyx_k_BOTTOM_RIGHT, sizeof(__pyx_k_BOTTOM_RIGHT), 0, 0, 1, 1},
    {&__pyx_n_s_Blob, __pyx_k_Blob, sizeof(__pyx_k_Blob), 0, 0, 1, 1},
    {&__pyx_n_s_Blob___reduce_cython, __pyx_k_Blob___reduce_cython, sizeof(__pyx_k_Blob___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Blob___setstate_cython, __pyx_k_Blob___setstate_cython, sizeof(__pyx_k_Blob___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Blob_from_file_path, __pyx_k_Blob_from_file_path, sizeof(__pyx_k_Blob_from_file_path), 0, 0, 1, 1},
    {&__pyx_n_s_Buffer, __pyx_k_Buffer, sizeof(__pyx_k_Buffer), 0, 0, 1, 1},
    {&__pyx_n_s_BufferClusterLevel, __pyx_k_BufferClusterLevel, sizeof(__pyx_k_BufferClusterLevel), 0, 0, 1, 1},
    {&__pyx_n_s_BufferContentType, __pyx_k_BufferContentType, sizeof(__pyx_k_BufferContentType), 0, 0, 1, 1},
    {&__pyx_n_s_BufferFlags, __pyx_k_BufferFlags, sizeof(__pyx_k_BufferFlags), 0, 0, 1, 1},
    {&__pyx_n_s_Buffer___reduce_cython, __pyx_k_Buffer___reduce_cython, sizeof(__pyx_k_Buffer___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Buffer___setstate_cython, __pyx_k_Buffer___setstate_cython, sizeof(__pyx_k_Buffer___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Buffer_add_codepoints, __pyx_k_Buffer_add_codepoints, sizeof(__pyx_k_Buffer_add_codepoints), 0, 0, 1, 1},
    {&__pyx_n_s_Buffer_add_str, __pyx_k_Buffer_add_str, sizeof(__pyx_k_Buffer_add_str), 0, 0, 1, 1},
    {&__pyx_n_s_Buffer_add_utf8, __pyx_k_Buffer_add_utf8, sizeof(__pyx_k_Buffer_add_utf8), 0, 0, 1, 1},
    {&__pyx_n_s_Buffer_clear_contents, __pyx_k_Buffer_clear_contents, sizeof(__pyx_k_Buffer_clear_contents), 0, 0, 1, 1},
    {&__pyx_n_s_Buffer_create, __pyx_k_Buffer_create, sizeof(__pyx_k_Buffer_create), 0, 0, 1, 1},
    {&__pyx_n_s_Buffer_guess_segment_properties, __pyx_k_Buffer_guess_segment_properties, sizeof(__pyx_k_Buffer_guess_segment_properties), 0, 0, 1, 1},
    {&__pyx_n_s_Buffer_reset, __pyx_k_Buffer_reset, sizeof(__pyx_k_Buffer_reset), 0, 0, 1, 1},
    {&__pyx_n_s_Buffer_set_language_from_ot_tag, __pyx_k_Buffer_set_language_from_ot_tag, sizeof(__pyx_k_Buffer_set_language_from_ot_tag), 0, 0, 1, 1},
    {&__pyx_n_s_Buffer_set_message_func, __pyx_k_Buffer_set_message_func, sizeof(__pyx_k_Buffer_set_message_func), 0, 0, 1, 1},
    {&__pyx_n_s_Buffer_set_script_from_ot_tag, __pyx_k_Buffer_set_script_from_ot_tag, sizeof(__pyx_k_Buffer_set_script_from_ot_tag), 0, 0, 1, 1},
    {&__pyx_n_s_CHARACTERS, __pyx_k_CHARACTERS, sizeof(__pyx_k_CHARACTERS), 0, 0, 1, 1},
    {&__pyx_n_s_CLEAR, __pyx_k_CLEAR, sizeof(__pyx_k_CLEAR), 0, 0, 1, 1},
    {&__pyx_n_s_COLOR_BURN, __pyx_k_COLOR_BURN, sizeof(__pyx_k_COLOR_BURN), 0, 0, 1, 1},
    {&__pyx_n_s_COLOR_DODGE, __pyx_k_COLOR_DODGE, sizeof(__pyx_k_COLOR_DODGE), 0, 0, 1, 1},
    {&__pyx_n_s_COMPONENT, __pyx_k_COMPONENT, sizeof(__pyx_k_COMPONENT), 0, 0, 1, 1},
    {&__pyx_n_s_Callable, __pyx_k_Callable, sizeof(__pyx_k_Callable), 0, 0, 1, 1},
    {&__pyx_kp_s_Callable_Blob_int_int_str_float, __pyx_k_Callable_Blob_int_int_str_float, sizeof(__pyx_k_Callable_Blob_int_int_str_float), 0, 0, 1, 0},
    {&__pyx_kp_s_Callable_ColorLine_float_float_f, __pyx_k_Callable_ColorLine_float_float_f, sizeof(__pyx_k_Callable_ColorLine_float_float_f), 0, 0, 1, 0},
    {&__pyx_kp_s_Callable_ColorLine_float_float_f_2, __pyx_k_Callable_ColorLine_float_float_f_2, sizeof(__pyx_k_Callable_ColorLine_float_float_f_2), 0, 0, 1, 0},
    {&__pyx_kp_s_Callable_Color_bool_object_None, __pyx_k_Callable_Color_bool_object_None, sizeof(__pyx_k_Callable_Color_bool_object_None), 0, 0, 1, 0},
    {&__pyx_kp_s_Callable_Face_str_object_bytes, __pyx_k_Callable_Face_str_object_bytes, sizeof(__pyx_k_Callable_Face_str_object_bytes), 0, 0, 1, 0},
    {&__pyx_kp_s_Callable_Font_int_int_object_int, __pyx_k_Callable_Font_int_int_object_int, sizeof(__pyx_k_Callable_Font_int_int_object_int), 0, 0, 1, 0},
    {&__pyx_kp_s_Callable_Font_int_object_int, __pyx_k_Callable_Font_int_object_int, sizeof(__pyx_k_Callable_Font_int_object_int), 0, 0, 1, 0},
    {&__pyx_kp_s_Callable_Font_int_object_int_int, __pyx_k_Callable_Font_int_object_int_int, sizeof(__pyx_k_Callable_Font_int_object_int_int), 0, 0, 1, 0},
    {&__pyx_kp_s_Callable_Font_int_object_str, __pyx_k_Callable_Font_int_object_str, sizeof(__pyx_k_Callable_Font_int_object_str), 0, 0, 1, 0},
    {&__pyx_kp_s_Callable_Font_object_FontExtents, __pyx_k_Callable_Font_object_FontExtents, sizeof(__pyx_k_Callable_Font_object_FontExtents), 0, 0, 1, 0},
    {&__pyx_kp_s_Callable_PaintCompositeMode_obje, __pyx_k_Callable_PaintCompositeMode_obje, sizeof(__pyx_k_Callable_PaintCompositeMode_obje), 0, 0, 1, 0},
    {&__pyx_kp_s_Callable_float_float_float_float, __pyx_k_Callable_float_float_float_float, sizeof(__pyx_k_Callable_float_float_float_float), 0, 0, 1, 0},
    {&__pyx_kp_s_Callable_float_float_float_float_2, __pyx_k_Callable_float_float_float_float_2, sizeof(__pyx_k_Callable_float_float_float_float_2), 0, 0, 1, 0},
    {&__pyx_kp_s_Callable_float_float_object_None, __pyx_k_Callable_float_float_object_None, sizeof(__pyx_k_Callable_float_float_object_None), 0, 0, 1, 0},
    {&__pyx_kp_s_Callable_int_object_Color, __pyx_k_Callable_int_object_Color, sizeof(__pyx_k_Callable_int_object_Color), 0, 0, 1, 0},
    {&__pyx_kp_s_Callable_int_object_None, __pyx_k_Callable_int_object_None, sizeof(__pyx_k_Callable_int_object_None), 0, 0, 1, 0},
    {&__pyx_kp_s_Callable_int_object_bool, __pyx_k_Callable_int_object_bool, sizeof(__pyx_k_Callable_int_object_bool), 0, 0, 1, 0},
    {&__pyx_kp_s_Callable_object_None, __pyx_k_Callable_object_None, sizeof(__pyx_k_Callable_object_None), 0, 0, 1, 0},
    {&__pyx_n_s_Color, __pyx_k_Color, sizeof(__pyx_k_Color), 0, 0, 1, 1},
    {&__pyx_n_u_Color, __pyx_k_Color, sizeof(__pyx_k_Color), 0, 1, 0, 1},
    {&__pyx_n_s_ColorLine, __pyx_k_ColorLine, sizeof(__pyx_k_ColorLine), 0, 0, 1, 1},
    {&__pyx_n_s_ColorLine___reduce_cython, __pyx_k_ColorLine___reduce_cython, sizeof(__pyx_k_ColorLine___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_ColorLine___setstate_cython, __pyx_k_ColorLine___setstate_cython, sizeof(__pyx_k_ColorLine___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_ColorStop, __pyx_k_ColorStop, sizeof(__pyx_k_ColorStop), 0, 0, 1, 1},
    {&__pyx_n_u_ColorStop, __pyx_k_ColorStop, sizeof(__pyx_k_ColorStop), 0, 1, 0, 1},
    {&__pyx_kp_s_Color_None, __pyx_k_Color_None, sizeof(__pyx_k_Color_None), 0, 0, 1, 0},
    {&__pyx_n_s_Color_from_int, __pyx_k_Color_from_int, sizeof(__pyx_k_Color_from_int), 0, 0, 1, 1},
    {&__pyx_n_s_Color_to_int, __pyx_k_Color_to_int, sizeof(__pyx_k_Color_to_int), 0, 0, 1, 1},
    {&__pyx_n_s_DARKEN, __pyx_k_DARKEN, sizeof(__pyx_k_DARKEN), 0, 0, 1, 1},
    {&__pyx_n_s_DEFAULT, __pyx_k_DEFAULT, sizeof(__pyx_k_DEFAULT), 0, 0, 1, 1},
    {&__pyx_n_s_DEFAULT_REPLACEMENT_CODEPOINT, __pyx_k_DEFAULT_REPLACEMENT_CODEPOINT, sizeof(__pyx_k_DEFAULT_REPLACEMENT_CODEPOINT), 0, 0, 1, 1},
    {&__pyx_n_s_DELIMITED_SUB_FORMULA_MIN_HEIGHT, __pyx_k_DELIMITED_SUB_FORMULA_MIN_HEIGHT, sizeof(__pyx_k_DELIMITED_SUB_FORMULA_MIN_HEIGHT), 0, 0, 1, 1},
    {&__pyx_n_s_DEST, __pyx_k_DEST, sizeof(__pyx_k_DEST), 0, 0, 1, 1},
    {&__pyx_n_s_DEST_ATOP, __pyx_k_DEST_ATOP, sizeof(__pyx_k_DEST_ATOP), 0, 0, 1, 1},
    {&__pyx_n_s_DEST_IN, __pyx_k_DEST_IN, sizeof(__pyx_k_DEST_IN), 0, 0, 1, 1},
    {&__pyx_n_s_DEST_OUT, __pyx_k_DEST_OUT, sizeof(__pyx_k_DEST_OUT), 0, 0, 1, 1},
    {&__pyx_n_s_DEST_OVER, __pyx_k_DEST_OVER, sizeof(__pyx_k_DEST_OVER), 0, 0, 1, 1},
    {&__pyx_n_s_DESUBROUTINIZE, __pyx_k_DESUBROUTINIZE, sizeof(__pyx_k_DESUBROUTINIZE), 0, 0, 1, 1},
    {&__pyx_n_s_DIFFERENCE, __pyx_k_DIFFERENCE, sizeof(__pyx_k_DIFFERENCE), 0, 0, 1, 1},
    {&__pyx_n_s_DISPLAY_OPERATOR_MIN_HEIGHT, __pyx_k_DISPLAY_OPERATOR_MIN_HEIGHT, sizeof(__pyx_k_DISPLAY_OPERATOR_MIN_HEIGHT), 0, 0, 1, 1},
    {&__pyx_n_s_DO_NOT_INSERT_DOTTED_CIRCLE, __pyx_k_DO_NOT_INSERT_DOTTED_CIRCLE, sizeof(__pyx_k_DO_NOT_INSERT_DOTTED_CIRCLE), 0, 0, 1, 1},
    {&__pyx_n_s_DROP_TABLE_TAG, __pyx_k_DROP_TABLE_TAG, sizeof(__pyx_k_DROP_TABLE_TAG), 0, 0, 1, 1},
    {&__pyx_n_s_DeprecationWarning, __pyx_k_DeprecationWarning, sizeof(__pyx_k_DeprecationWarning), 0, 0, 1, 1},
    {&__pyx_n_s_Dict, __pyx_k_Dict, sizeof(__pyx_k_Dict), 0, 0, 1, 1},
    {&__pyx_kp_s_Dict_str_Union_int_bool_Sequence, __pyx_k_Dict_str_Union_int_bool_Sequence, sizeof(__pyx_k_Dict_str_Union_int_bool_Sequence), 0, 0, 1, 0},
    {&__pyx_kp_s_Dict_str_float, __pyx_k_Dict_str_float, sizeof(__pyx_k_Dict_str_float), 0, 0, 1, 0},
    {&__pyx_n_s_DrawFuncs, __pyx_k_DrawFuncs, sizeof(__pyx_k_DrawFuncs), 0, 0, 1, 1},
    {&__pyx_n_s_DrawFuncs___reduce_cython, __pyx_k_DrawFuncs___reduce_cython, sizeof(__pyx_k_DrawFuncs___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_DrawFuncs___setstate_cython, __pyx_k_DrawFuncs___setstate_cython, sizeof(__pyx_k_DrawFuncs___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_DrawFuncs_draw_glyph, __pyx_k_DrawFuncs_draw_glyph, sizeof(__pyx_k_DrawFuncs_draw_glyph), 0, 0, 1, 1},
    {&__pyx_n_s_DrawFuncs_get_glyph_shape, __pyx_k_DrawFuncs_get_glyph_shape, sizeof(__pyx_k_DrawFuncs_get_glyph_shape), 0, 0, 1, 1},
    {&__pyx_n_s_DrawFuncs_set_close_path_func, __pyx_k_DrawFuncs_set_close_path_func, sizeof(__pyx_k_DrawFuncs_set_close_path_func), 0, 0, 1, 1},
    {&__pyx_n_s_DrawFuncs_set_cubic_to_func, __pyx_k_DrawFuncs_set_cubic_to_func, sizeof(__pyx_k_DrawFuncs_set_cubic_to_func), 0, 0, 1, 1},
    {&__pyx_n_s_DrawFuncs_set_line_to_func, __pyx_k_DrawFuncs_set_line_to_func, sizeof(__pyx_k_DrawFuncs_set_line_to_func), 0, 0, 1, 1},
    {&__pyx_n_s_DrawFuncs_set_move_to_func, __pyx_k_DrawFuncs_set_move_to_func, sizeof(__pyx_k_DrawFuncs_set_move_to_func), 0, 0, 1, 1},
    {&__pyx_n_s_DrawFuncs_set_quadratic_to_func, __pyx_k_DrawFuncs_set_quadratic_to_func, sizeof(__pyx_k_DrawFuncs_set_quadratic_to_func), 0, 0, 1, 1},
    {&__pyx_n_s_EOT, __pyx_k_EOT, sizeof(__pyx_k_EOT), 0, 0, 1, 1},
    {&__pyx_n_s_EXCLUSION, __pyx_k_EXCLUSION, sizeof(__pyx_k_EXCLUSION), 0, 0, 1, 1},
    {&__pyx_n_s_EXTENDER, __pyx_k_EXTENDER, sizeof(__pyx_k_EXTENDER), 0, 0, 1, 1},
    {&__pyx_n_s_FLATTENED_ACCENT_BASE_HEIGHT, __pyx_k_FLATTENED_ACCENT_BASE_HEIGHT, sizeof(__pyx_k_FLATTENED_ACCENT_BASE_HEIGHT), 0, 0, 1, 1},
    {&__pyx_n_s_FRACTION_DENOMINATOR_DISPLAY_STY, __pyx_k_FRACTION_DENOMINATOR_DISPLAY_STY, sizeof(__pyx_k_FRACTION_DENOMINATOR_DISPLAY_STY), 0, 0, 1, 1},
    {&__pyx_n_s_FRACTION_DENOMINATOR_GAP_MIN, __pyx_k_FRACTION_DENOMINATOR_GAP_MIN, sizeof(__pyx_k_FRACTION_DENOMINATOR_GAP_MIN), 0, 0, 1, 1},
    {&__pyx_n_s_FRACTION_DENOMINATOR_SHIFT_DOWN, __pyx_k_FRACTION_DENOMINATOR_SHIFT_DOWN, sizeof(__pyx_k_FRACTION_DENOMINATOR_SHIFT_DOWN), 0, 0, 1, 1},
    {&__pyx_n_s_FRACTION_DENOM_DISPLAY_STYLE_GAP, __pyx_k_FRACTION_DENOM_DISPLAY_STYLE_GAP, sizeof(__pyx_k_FRACTION_DENOM_DISPLAY_STYLE_GAP), 0, 0, 1, 1},
    {&__pyx_n_s_FRACTION_NUMERATOR_DISPLAY_STYLE, __pyx_k_FRACTION_NUMERATOR_DISPLAY_STYLE, sizeof(__pyx_k_FRACTION_NUMERATOR_DISPLAY_STYLE), 0, 0, 1, 1},
    {&__pyx_n_s_FRACTION_NUMERATOR_GAP_MIN, __pyx_k_FRACTION_NUMERATOR_GAP_MIN, sizeof(__pyx_k_FRACTION_NUMERATOR_GAP_MIN), 0, 0, 1, 1},
    {&__pyx_n_s_FRACTION_NUMERATOR_SHIFT_UP, __pyx_k_FRACTION_NUMERATOR_SHIFT_UP, sizeof(__pyx_k_FRACTION_NUMERATOR_SHIFT_UP), 0, 0, 1, 1},
    {&__pyx_n_s_FRACTION_NUM_DISPLAY_STYLE_GAP_M, __pyx_k_FRACTION_NUM_DISPLAY_STYLE_GAP_M, sizeof(__pyx_k_FRACTION_NUM_DISPLAY_STYLE_GAP_M), 0, 0, 1, 1},
    {&__pyx_n_s_FRACTION_RULE_THICKNESS, __pyx_k_FRACTION_RULE_THICKNESS, sizeof(__pyx_k_FRACTION_RULE_THICKNESS), 0, 0, 1, 1},
    {&__pyx_n_s_Face, __pyx_k_Face, sizeof(__pyx_k_Face), 0, 0, 1, 1},
    {&__pyx_n_s_Face___reduce_cython, __pyx_k_Face___reduce_cython, sizeof(__pyx_k_Face___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Face___setstate_cython, __pyx_k_Face___setstate_cython, sizeof(__pyx_k_Face___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Face_create, __pyx_k_Face_create, sizeof(__pyx_k_Face_create), 0, 0, 1, 1},
    {&__pyx_n_s_Face_create_for_tables, __pyx_k_Face_create_for_tables, sizeof(__pyx_k_Face_create_for_tables), 0, 0, 1, 1},
    {&__pyx_n_s_Face_variation_unicodes, __pyx_k_Face_variation_unicodes, sizeof(__pyx_k_Face_variation_unicodes), 0, 0, 1, 1},
    {&__pyx_kp_u_Failed_to_open, __pyx_k_Failed_to_open, sizeof(__pyx_k_Failed_to_open), 0, 1, 0, 0},
    {&__pyx_n_s_Font, __pyx_k_Font, sizeof(__pyx_k_Font), 0, 0, 1, 1},
    {&__pyx_n_s_FontExtents, __pyx_k_FontExtents, sizeof(__pyx_k_FontExtents), 0, 0, 1, 1},
    {&__pyx_n_u_FontExtents, __pyx_k_FontExtents, sizeof(__pyx_k_FontExtents), 0, 1, 0, 1},
    {&__pyx_n_s_FontFuncs, __pyx_k_FontFuncs, sizeof(__pyx_k_FontFuncs), 0, 0, 1, 1},
    {&__pyx_n_s_FontFuncs___reduce_cython, __pyx_k_FontFuncs___reduce_cython, sizeof(__pyx_k_FontFuncs___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_FontFuncs___setstate_cython, __pyx_k_FontFuncs___setstate_cython, sizeof(__pyx_k_FontFuncs___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_FontFuncs_create, __pyx_k_FontFuncs_create, sizeof(__pyx_k_FontFuncs_create), 0, 0, 1, 1},
    {&__pyx_n_s_FontFuncs_set_font_h_extents_fun, __pyx_k_FontFuncs_set_font_h_extents_fun, sizeof(__pyx_k_FontFuncs_set_font_h_extents_fun), 0, 0, 1, 1},
    {&__pyx_n_s_FontFuncs_set_font_v_extents_fun, __pyx_k_FontFuncs_set_font_v_extents_fun, sizeof(__pyx_k_FontFuncs_set_font_v_extents_fun), 0, 0, 1, 1},
    {&__pyx_n_s_FontFuncs_set_glyph_h_advance_fu, __pyx_k_FontFuncs_set_glyph_h_advance_fu, sizeof(__pyx_k_FontFuncs_set_glyph_h_advance_fu), 0, 0, 1, 1},
    {&__pyx_n_s_FontFuncs_set_glyph_name_func, __pyx_k_FontFuncs_set_glyph_name_func, sizeof(__pyx_k_FontFuncs_set_glyph_name_func), 0, 0, 1, 1},
    {&__pyx_n_s_FontFuncs_set_glyph_v_advance_fu, __pyx_k_FontFuncs_set_glyph_v_advance_fu, sizeof(__pyx_k_FontFuncs_set_glyph_v_advance_fu), 0, 0, 1, 1},
    {&__pyx_n_s_FontFuncs_set_glyph_v_origin_fun, __pyx_k_FontFuncs_set_glyph_v_origin_fun, sizeof(__pyx_k_FontFuncs_set_glyph_v_origin_fun), 0, 0, 1, 1},
    {&__pyx_n_s_FontFuncs_set_nominal_glyph_func, __pyx_k_FontFuncs_set_nominal_glyph_func, sizeof(__pyx_k_FontFuncs_set_nominal_glyph_func), 0, 0, 1, 1},
    {&__pyx_n_s_FontFuncs_set_variation_glyph_fu, __pyx_k_FontFuncs_set_variation_glyph_fu, sizeof(__pyx_k_FontFuncs_set_variation_glyph_fu), 0, 0, 1, 1},
    {&__pyx_n_s_Font___reduce_cython, __pyx_k_Font___reduce_cython, sizeof(__pyx_k_Font___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Font___setstate_cython, __pyx_k_Font___setstate_cython, sizeof(__pyx_k_Font___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Font_create, __pyx_k_Font_create, sizeof(__pyx_k_Font_create), 0, 0, 1, 1},
    {&__pyx_n_s_Font_draw_glyph, __pyx_k_Font_draw_glyph, sizeof(__pyx_k_Font_draw_glyph), 0, 0, 1, 1},
    {&__pyx_n_s_Font_draw_glyph_with_pen, __pyx_k_Font_draw_glyph_with_pen, sizeof(__pyx_k_Font_draw_glyph_with_pen), 0, 0, 1, 1},
    {&__pyx_n_s_Font_get_font_extents, __pyx_k_Font_get_font_extents, sizeof(__pyx_k_Font_get_font_extents), 0, 0, 1, 1},
    {&__pyx_n_s_Font_get_glyph_extents, __pyx_k_Font_get_glyph_extents, sizeof(__pyx_k_Font_get_glyph_extents), 0, 0, 1, 1},
    {&__pyx_n_s_Font_get_glyph_from_name, __pyx_k_Font_get_glyph_from_name, sizeof(__pyx_k_Font_get_glyph_from_name), 0, 0, 1, 1},
    {&__pyx_n_s_Font_get_glyph_h_advance, __pyx_k_Font_get_glyph_h_advance, sizeof(__pyx_k_Font_get_glyph_h_advance), 0, 0, 1, 1},
    {&__pyx_n_s_Font_get_glyph_h_origin, __pyx_k_Font_get_glyph_h_origin, sizeof(__pyx_k_Font_get_glyph_h_origin), 0, 0, 1, 1},
    {&__pyx_n_s_Font_get_glyph_name, __pyx_k_Font_get_glyph_name, sizeof(__pyx_k_Font_get_glyph_name), 0, 0, 1, 1},
    {&__pyx_n_s_Font_get_glyph_v_advance, __pyx_k_Font_get_glyph_v_advance, sizeof(__pyx_k_Font_get_glyph_v_advance), 0, 0, 1, 1},
    {&__pyx_n_s_Font_get_glyph_v_origin, __pyx_k_Font_get_glyph_v_origin, sizeof(__pyx_k_Font_get_glyph_v_origin), 0, 0, 1, 1},
    {&__pyx_n_s_Font_get_nominal_glyph, __pyx_k_Font_get_nominal_glyph, sizeof(__pyx_k_Font_get_nominal_glyph), 0, 0, 1, 1},
    {&__pyx_n_s_Font_get_var_coords_design, __pyx_k_Font_get_var_coords_design, sizeof(__pyx_k_Font_get_var_coords_design), 0, 0, 1, 1},
    {&__pyx_n_s_Font_get_var_coords_normalized, __pyx_k_Font_get_var_coords_normalized, sizeof(__pyx_k_Font_get_var_coords_normalized), 0, 0, 1, 1},
    {&__pyx_n_s_Font_get_variation_glyph, __pyx_k_Font_get_variation_glyph, sizeof(__pyx_k_Font_get_variation_glyph), 0, 0, 1, 1},
    {&__pyx_n_s_Font_glyph_from_string, __pyx_k_Font_glyph_from_string, sizeof(__pyx_k_Font_glyph_from_string), 0, 0, 1, 1},
    {&__pyx_n_s_Font_glyph_to_string, __pyx_k_Font_glyph_to_string, sizeof(__pyx_k_Font_glyph_to_string), 0, 0, 1, 1},
    {&__pyx_n_s_Font_paint_glyph, __pyx_k_Font_paint_glyph, sizeof(__pyx_k_Font_paint_glyph), 0, 0, 1, 1},
    {&__pyx_n_s_Font_set_var_coords_design, __pyx_k_Font_set_var_coords_design, sizeof(__pyx_k_Font_set_var_coords_design), 0, 0, 1, 1},
    {&__pyx_n_s_Font_set_var_coords_normalized, __pyx_k_Font_set_var_coords_normalized, sizeof(__pyx_k_Font_set_var_coords_normalized), 0, 0, 1, 1},
    {&__pyx_n_s_Font_set_variation, __pyx_k_Font_set_variation, sizeof(__pyx_k_Font_set_variation), 0, 0, 1, 1},
    {&__pyx_n_s_Font_set_variations, __pyx_k_Font_set_variations, sizeof(__pyx_k_Font_set_variations), 0, 0, 1, 1},
    {&__pyx_n_s_GLYPHS, __pyx_k_GLYPHS, sizeof(__pyx_k_GLYPHS), 0, 0, 1, 1},
    {&__pyx_n_s_GLYPH_INDEX, __pyx_k_GLYPH_INDEX, sizeof(__pyx_k_GLYPH_INDEX), 0, 0, 1, 1},
    {&__pyx_n_s_GLYPH_NAMES, __pyx_k_GLYPH_NAMES, sizeof(__pyx_k_GLYPH_NAMES), 0, 0, 1, 1},
    {&__pyx_n_s_GlyphExtents, __pyx_k_GlyphExtents, sizeof(__pyx_k_GlyphExtents), 0, 0, 1, 1},
    {&__pyx_n_u_GlyphExtents, __pyx_k_GlyphExtents, sizeof(__pyx_k_GlyphExtents), 0, 1, 0, 1},
    {&__pyx_n_s_GlyphFlags, __pyx_k_GlyphFlags, sizeof(__pyx_k_GlyphFlags), 0, 0, 1, 1},
    {&__pyx_n_s_GlyphInfo, __pyx_k_GlyphInfo, sizeof(__pyx_k_GlyphInfo), 0, 0, 1, 1},
    {&__pyx_n_s_GlyphInfo___reduce_cython, __pyx_k_GlyphInfo___reduce_cython, sizeof(__pyx_k_GlyphInfo___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_GlyphInfo___setstate_cython, __pyx_k_GlyphInfo___setstate_cython, sizeof(__pyx_k_GlyphInfo___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_GlyphPosition, __pyx_k_GlyphPosition, sizeof(__pyx_k_GlyphPosition), 0, 0, 1, 1},
    {&__pyx_n_s_GlyphPosition___reduce_cython, __pyx_k_GlyphPosition___reduce_cython, sizeof(__pyx_k_GlyphPosition___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_GlyphPosition___setstate_cython, __pyx_k_GlyphPosition___setstate_cython, sizeof(__pyx_k_GlyphPosition___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_HARD_LIGHT, __pyx_k_HARD_LIGHT, sizeof(__pyx_k_HARD_LIGHT), 0, 0, 1, 1},
    {&__pyx_n_s_HBObject, __pyx_k_HBObject, sizeof(__pyx_k_HBObject), 0, 0, 1, 1},
    {&__pyx_n_s_HBObject___reduce_cython, __pyx_k_HBObject___reduce_cython, sizeof(__pyx_k_HBObject___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_HBObject___setstate_cython, __pyx_k_HBObject___setstate_cython, sizeof(__pyx_k_HBObject___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_HSL_COLOR, __pyx_k_HSL_COLOR, sizeof(__pyx_k_HSL_COLOR), 0, 0, 1, 1},
    {&__pyx_n_s_HSL_HUE, __pyx_k_HSL_HUE, sizeof(__pyx_k_HSL_HUE), 0, 0, 1, 1},
    {&__pyx_n_s_HSL_LUMINOSITY, __pyx_k_HSL_LUMINOSITY, sizeof(__pyx_k_HSL_LUMINOSITY), 0, 0, 1, 1},
    {&__pyx_n_s_HSL_SATURATION, __pyx_k_HSL_SATURATION, sizeof(__pyx_k_HSL_SATURATION), 0, 0, 1, 1},
    {&__pyx_n_s_HarfBuzzError, __pyx_k_HarfBuzzError, sizeof(__pyx_k_HarfBuzzError), 0, 0, 1, 1},
    {&__pyx_n_s_INVALID, __pyx_k_INVALID, sizeof(__pyx_k_INVALID), 0, 0, 1, 1},
    {&__pyx_n_s_INVALID_VALUE, __pyx_k_INVALID_VALUE, sizeof(__pyx_k_INVALID_VALUE), 0, 0, 1, 1},
    {&__pyx_kp_u_Input_num_of_subtables, __pyx_k_Input_num_of_subtables, sizeof(__pyx_k_Input_num_of_subtables), 0, 1, 0, 0},
    {&__pyx_n_s_IntEnum, __pyx_k_IntEnum, sizeof(__pyx_k_IntEnum), 0, 0, 1, 1},
    {&__pyx_n_s_IntFlag, __pyx_k_IntFlag, sizeof(__pyx_k_IntFlag), 0, 0, 1, 1},
    {&__pyx_n_s_KeyError, __pyx_k_KeyError, sizeof(__pyx_k_KeyError), 0, 0, 1, 1},
    {&__pyx_n_s_LAYOUT_FEATURE_TAG, __pyx_k_LAYOUT_FEATURE_TAG, sizeof(__pyx_k_LAYOUT_FEATURE_TAG), 0, 0, 1, 1},
    {&__pyx_n_s_LAYOUT_SCRIPT_TAG, __pyx_k_LAYOUT_SCRIPT_TAG, sizeof(__pyx_k_LAYOUT_SCRIPT_TAG), 0, 0, 1, 1},
    {&__pyx_n_s_LIGATURE, __pyx_k_LIGATURE, sizeof(__pyx_k_LIGATURE), 0, 0, 1, 1},
    {&__pyx_n_s_LIGHTEN, __pyx_k_LIGHTEN, sizeof(__pyx_k_LIGHTEN), 0, 0, 1, 1},
    {&__pyx_n_s_LOWER_LIMIT_BASELINE_DROP_MIN, __pyx_k_LOWER_LIMIT_BASELINE_DROP_MIN, sizeof(__pyx_k_LOWER_LIMIT_BASELINE_DROP_MIN), 0, 0, 1, 1},
    {&__pyx_n_s_LOWER_LIMIT_GAP_MIN, __pyx_k_LOWER_LIMIT_GAP_MIN, sizeof(__pyx_k_LOWER_LIMIT_GAP_MIN), 0, 0, 1, 1},
    {&__pyx_n_s_List, __pyx_k_List, sizeof(__pyx_k_List), 0, 0, 1, 1},
    {&__pyx_kp_s_List_Color, __pyx_k_List_Color, sizeof(__pyx_k_List_Color), 0, 0, 1, 0},
    {&__pyx_kp_s_List_OTColorLayer, __pyx_k_List_OTColorLayer, sizeof(__pyx_k_List_OTColorLayer), 0, 0, 1, 0},
    {&__pyx_kp_s_List_OTMathGlyphVariant, __pyx_k_List_OTMathGlyphVariant, sizeof(__pyx_k_List_OTMathGlyphVariant), 0, 0, 1, 0},
    {&__pyx_kp_s_List_OTMathKernEntry, __pyx_k_List_OTMathKernEntry, sizeof(__pyx_k_List_OTMathKernEntry), 0, 0, 1, 0},
    {&__pyx_kp_s_List_Tuple_List_Tuple_int_int_in, __pyx_k_List_Tuple_List_Tuple_int_int_in, sizeof(__pyx_k_List_Tuple_List_Tuple_int_int_in), 0, 0, 1, 0},
    {&__pyx_kp_s_List_bytes, __pyx_k_List_bytes, sizeof(__pyx_k_List_bytes), 0, 0, 1, 0},
    {&__pyx_kp_s_List_int, __pyx_k_List_int, sizeof(__pyx_k_List_int), 0, 0, 1, 0},
    {&__pyx_kp_s_List_str, __pyx_k_List_str, sizeof(__pyx_k_List_str), 0, 0, 1, 0},
    {&__pyx_n_s_MARK, __pyx_k_MARK, sizeof(__pyx_k_MARK), 0, 0, 1, 1},
    {&__pyx_n_s_MATH_LEADING, __pyx_k_MATH_LEADING, sizeof(__pyx_k_MATH_LEADING), 0, 0, 1, 1},
    {&__pyx_n_s_MONOTONE_CHARACTERS, __pyx_k_MONOTONE_CHARACTERS, sizeof(__pyx_k_MONOTONE_CHARACTERS), 0, 0, 1, 1},
    {&__pyx_n_s_MONOTONE_GRAPHEMES, __pyx_k_MONOTONE_GRAPHEMES, sizeof(__pyx_k_MONOTONE_GRAPHEMES), 0, 0, 1, 1},
    {&__pyx_n_s_MULTIPLY, __pyx_k_MULTIPLY, sizeof(__pyx_k_MULTIPLY), 0, 0, 1, 1},
    {&__pyx_n_s_Map, __pyx_k_Map, sizeof(__pyx_k_Map), 0, 0, 1, 1},
    {&__pyx_n_s_MapIter, __pyx_k_MapIter, sizeof(__pyx_k_MapIter), 0, 0, 1, 1},
    {&__pyx_n_s_MapIter___reduce_cython, __pyx_k_MapIter___reduce_cython, sizeof(__pyx_k_MapIter___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_MapIter___setstate_cython, __pyx_k_MapIter___setstate_cython, sizeof(__pyx_k_MapIter___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Map___copy, __pyx_k_Map___copy, sizeof(__pyx_k_Map___copy), 0, 0, 1, 1},
    {&__pyx_n_s_Map___reduce_cython, __pyx_k_Map___reduce_cython, sizeof(__pyx_k_Map___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Map___setstate_cython, __pyx_k_Map___setstate_cython, sizeof(__pyx_k_Map___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Map__is_equal, __pyx_k_Map__is_equal, sizeof(__pyx_k_Map__is_equal), 0, 0, 1, 1},
    {&__pyx_n_s_Map__update, __pyx_k_Map__update, sizeof(__pyx_k_Map__update), 0, 0, 1, 1},
    {&__pyx_n_s_Map_clear, __pyx_k_Map_clear, sizeof(__pyx_k_Map_clear), 0, 0, 1, 1},
    {&__pyx_n_s_Map_copy, __pyx_k_Map_copy, sizeof(__pyx_k_Map_copy), 0, 0, 1, 1},
    {&__pyx_n_s_Map_get, __pyx_k_Map_get, sizeof(__pyx_k_Map_get), 0, 0, 1, 1},
    {&__pyx_n_s_Map_items, __pyx_k_Map_items, sizeof(__pyx_k_Map_items), 0, 0, 1, 1},
    {&__pyx_n_s_Map_keys, __pyx_k_Map_keys, sizeof(__pyx_k_Map_keys), 0, 0, 1, 1},
    {&__pyx_kp_u_Map_s, __pyx_k_Map_s, sizeof(__pyx_k_Map_s), 0, 1, 0, 0},
    {&__pyx_n_s_Map_update, __pyx_k_Map_update, sizeof(__pyx_k_Map_update), 0, 0, 1, 1},
    {&__pyx_n_s_Map_values, __pyx_k_Map_values, sizeof(__pyx_k_Map_values), 0, 0, 1, 1},
    {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
    {&__pyx_n_s_NAME_ID, __pyx_k_NAME_ID, sizeof(__pyx_k_NAME_ID), 0, 0, 1, 1},
    {&__pyx_n_s_NAME_LANG_ID, __pyx_k_NAME_LANG_ID, sizeof(__pyx_k_NAME_LANG_ID), 0, 0, 1, 1},
    {&__pyx_n_s_NAME_LEGACY, __pyx_k_NAME_LEGACY, sizeof(__pyx_k_NAME_LEGACY), 0, 0, 1, 1},
    {&__pyx_n_s_NOTDEF_OUTLINE, __pyx_k_NOTDEF_OUTLINE, sizeof(__pyx_k_NOTDEF_OUTLINE), 0, 0, 1, 1},
    {&__pyx_n_s_NO_HINTING, __pyx_k_NO_HINTING, sizeof(__pyx_k_NO_HINTING), 0, 0, 1, 1},
    {&__pyx_n_s_NO_LAYOUT_CLOSURE, __pyx_k_NO_LAYOUT_CLOSURE, sizeof(__pyx_k_NO_LAYOUT_CLOSURE), 0, 0, 1, 1},
    {&__pyx_n_s_NO_PRUNE_UNICODE_RANGES, __pyx_k_NO_PRUNE_UNICODE_RANGES, sizeof(__pyx_k_NO_PRUNE_UNICODE_RANGES), 0, 0, 1, 1},
    {&__pyx_n_s_NO_SUBSET_TABLE_TAG, __pyx_k_NO_SUBSET_TABLE_TAG, sizeof(__pyx_k_NO_SUBSET_TABLE_TAG), 0, 0, 1, 1},
    {&__pyx_n_s_None, __pyx_k_None, sizeof(__pyx_k_None), 0, 0, 1, 1},
    {&__pyx_n_s_NotImplemented, __pyx_k_NotImplemented, sizeof(__pyx_k_NotImplemented), 0, 0, 1, 1},
    {&__pyx_n_s_OTColorLayer, __pyx_k_OTColorLayer, sizeof(__pyx_k_OTColorLayer), 0, 0, 1, 1},
    {&__pyx_n_u_OTColorLayer, __pyx_k_OTColorLayer, sizeof(__pyx_k_OTColorLayer), 0, 1, 0, 1},
    {&__pyx_n_s_OTColorPaletteFlags, __pyx_k_OTColorPaletteFlags, sizeof(__pyx_k_OTColorPaletteFlags), 0, 0, 1, 1},
    {&__pyx_n_s_OTLayoutGlyphClass, __pyx_k_OTLayoutGlyphClass, sizeof(__pyx_k_OTLayoutGlyphClass), 0, 0, 1, 1},
    {&__pyx_n_s_OTMathConstant, __pyx_k_OTMathConstant, sizeof(__pyx_k_OTMathConstant), 0, 0, 1, 1},
    {&__pyx_n_s_OTMathGlyphPart, __pyx_k_OTMathGlyphPart, sizeof(__pyx_k_OTMathGlyphPart), 0, 0, 1, 1},
    {&__pyx_n_u_OTMathGlyphPart, __pyx_k_OTMathGlyphPart, sizeof(__pyx_k_OTMathGlyphPart), 0, 1, 0, 1},
    {&__pyx_n_s_OTMathGlyphPartFlags, __pyx_k_OTMathGlyphPartFlags, sizeof(__pyx_k_OTMathGlyphPartFlags), 0, 0, 1, 1},
    {&__pyx_n_s_OTMathGlyphVariant, __pyx_k_OTMathGlyphVariant, sizeof(__pyx_k_OTMathGlyphVariant), 0, 0, 1, 1},
    {&__pyx_n_u_OTMathGlyphVariant, __pyx_k_OTMathGlyphVariant, sizeof(__pyx_k_OTMathGlyphVariant), 0, 1, 0, 1},
    {&__pyx_n_s_OTMathKern, __pyx_k_OTMathKern, sizeof(__pyx_k_OTMathKern), 0, 0, 1, 1},
    {&__pyx_n_s_OTMathKernEntry, __pyx_k_OTMathKernEntry, sizeof(__pyx_k_OTMathKernEntry), 0, 0, 1, 1},
    {&__pyx_n_u_OTMathKernEntry, __pyx_k_OTMathKernEntry, sizeof(__pyx_k_OTMathKernEntry), 0, 1, 0, 1},
    {&__pyx_n_s_OVERBAR_EXTRA_ASCENDER, __pyx_k_OVERBAR_EXTRA_ASCENDER, sizeof(__pyx_k_OVERBAR_EXTRA_ASCENDER), 0, 0, 1, 1},
    {&__pyx_n_s_OVERBAR_RULE_THICKNESS, __pyx_k_OVERBAR_RULE_THICKNESS, sizeof(__pyx_k_OVERBAR_RULE_THICKNESS), 0, 0, 1, 1},
    {&__pyx_n_s_OVERBAR_VERTICAL_GAP, __pyx_k_OVERBAR_VERTICAL_GAP, sizeof(__pyx_k_OVERBAR_VERTICAL_GAP), 0, 0, 1, 1},
    {&__pyx_n_s_OVERLAY, __pyx_k_OVERLAY, sizeof(__pyx_k_OVERLAY), 0, 0, 1, 1},
    {&__pyx_n_s_PAD, __pyx_k_PAD, sizeof(__pyx_k_PAD), 0, 0, 1, 1},
    {&__pyx_n_s_PASSTHROUGH_UNRECOGNIZED, __pyx_k_PASSTHROUGH_UNRECOGNIZED, sizeof(__pyx_k_PASSTHROUGH_UNRECOGNIZED), 0, 0, 1, 1},
    {&__pyx_n_s_PLUS, __pyx_k_PLUS, sizeof(__pyx_k_PLUS), 0, 0, 1, 1},
    {&__pyx_n_s_PRESERVE_DEFAULT_IGNORABLES, __pyx_k_PRESERVE_DEFAULT_IGNORABLES, sizeof(__pyx_k_PRESERVE_DEFAULT_IGNORABLES), 0, 0, 1, 1},
    {&__pyx_n_s_PRODUCE_SAFE_TO_INSERT_TATWEEL, __pyx_k_PRODUCE_SAFE_TO_INSERT_TATWEEL, sizeof(__pyx_k_PRODUCE_SAFE_TO_INSERT_TATWEEL), 0, 0, 1, 1},
    {&__pyx_n_s_PRODUCE_UNSAFE_TO_CONCAT, __pyx_k_PRODUCE_UNSAFE_TO_CONCAT, sizeof(__pyx_k_PRODUCE_UNSAFE_TO_CONCAT), 0, 0, 1, 1},
    {&__pyx_n_s_PaintCompositeMode, __pyx_k_PaintCompositeMode, sizeof(__pyx_k_PaintCompositeMode), 0, 0, 1, 1},
    {&__pyx_n_s_PaintExtend, __pyx_k_PaintExtend, sizeof(__pyx_k_PaintExtend), 0, 0, 1, 1},
    {&__pyx_n_s_PaintFuncs, __pyx_k_PaintFuncs, sizeof(__pyx_k_PaintFuncs), 0, 0, 1, 1},
    {&__pyx_n_s_PaintFuncs___reduce_cython, __pyx_k_PaintFuncs___reduce_cython, sizeof(__pyx_k_PaintFuncs___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_PaintFuncs___setstate_cython, __pyx_k_PaintFuncs___setstate_cython, sizeof(__pyx_k_PaintFuncs___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_PaintFuncs_set_color_func, __pyx_k_PaintFuncs_set_color_func, sizeof(__pyx_k_PaintFuncs_set_color_func), 0, 0, 1, 1},
    {&__pyx_n_s_PaintFuncs_set_color_glyph_func, __pyx_k_PaintFuncs_set_color_glyph_func, sizeof(__pyx_k_PaintFuncs_set_color_glyph_func), 0, 0, 1, 1},
    {&__pyx_n_s_PaintFuncs_set_custom_palette_co, __pyx_k_PaintFuncs_set_custom_palette_co, sizeof(__pyx_k_PaintFuncs_set_custom_palette_co), 0, 0, 1, 1},
    {&__pyx_n_s_PaintFuncs_set_image_func, __pyx_k_PaintFuncs_set_image_func, sizeof(__pyx_k_PaintFuncs_set_image_func), 0, 0, 1, 1},
    {&__pyx_n_s_PaintFuncs_set_linear_gradient_f, __pyx_k_PaintFuncs_set_linear_gradient_f, sizeof(__pyx_k_PaintFuncs_set_linear_gradient_f), 0, 0, 1, 1},
    {&__pyx_n_s_PaintFuncs_set_pop_clip_func, __pyx_k_PaintFuncs_set_pop_clip_func, sizeof(__pyx_k_PaintFuncs_set_pop_clip_func), 0, 0, 1, 1},
    {&__pyx_n_s_PaintFuncs_set_pop_group_func, __pyx_k_PaintFuncs_set_pop_group_func, sizeof(__pyx_k_PaintFuncs_set_pop_group_func), 0, 0, 1, 1},
    {&__pyx_n_s_PaintFuncs_set_pop_transform_fun, __pyx_k_PaintFuncs_set_pop_transform_fun, sizeof(__pyx_k_PaintFuncs_set_pop_transform_fun), 0, 0, 1, 1},
    {&__pyx_n_s_PaintFuncs_set_push_clip_glyph_f, __pyx_k_PaintFuncs_set_push_clip_glyph_f, sizeof(__pyx_k_PaintFuncs_set_push_clip_glyph_f), 0, 0, 1, 1},
    {&__pyx_n_s_PaintFuncs_set_push_clip_rectang, __pyx_k_PaintFuncs_set_push_clip_rectang, sizeof(__pyx_k_PaintFuncs_set_push_clip_rectang), 0, 0, 1, 1},
    {&__pyx_n_s_PaintFuncs_set_push_group_func, __pyx_k_PaintFuncs_set_push_group_func, sizeof(__pyx_k_PaintFuncs_set_push_group_func), 0, 0, 1, 1},
    {&__pyx_n_s_PaintFuncs_set_push_transform_fu, __pyx_k_PaintFuncs_set_push_transform_fu, sizeof(__pyx_k_PaintFuncs_set_push_transform_fu), 0, 0, 1, 1},
    {&__pyx_n_s_PaintFuncs_set_radial_gradient_f, __pyx_k_PaintFuncs_set_radial_gradient_f, sizeof(__pyx_k_PaintFuncs_set_radial_gradient_f), 0, 0, 1, 1},
    {&__pyx_n_s_PaintFuncs_set_sweep_gradient_fu, __pyx_k_PaintFuncs_set_sweep_gradient_fu, sizeof(__pyx_k_PaintFuncs_set_sweep_gradient_fu), 0, 0, 1, 1},
    {&__pyx_kp_u_Pass_draw_state_to_Font_draw_gly, __pyx_k_Pass_draw_state_to_Font_draw_gly, sizeof(__pyx_k_Pass_draw_state_to_Font_draw_gly), 0, 1, 0, 0},
    {&__pyx_n_s_Path, __pyx_k_Path, sizeof(__pyx_k_Path), 0, 0, 1, 1},
    {&__pyx_kp_s_Pickling_of_struct_members_such, __pyx_k_Pickling_of_struct_members_such, sizeof(__pyx_k_Pickling_of_struct_members_such), 0, 0, 1, 0},
    {&__pyx_kp_s_Pickling_of_struct_members_such_2, __pyx_k_Pickling_of_struct_members_such_2, sizeof(__pyx_k_Pickling_of_struct_members_such_2), 0, 0, 1, 0},
    {&__pyx_n_s_RADICAL_DEGREE_BOTTOM_RAISE_PERC, __pyx_k_RADICAL_DEGREE_BOTTOM_RAISE_PERC, sizeof(__pyx_k_RADICAL_DEGREE_BOTTOM_RAISE_PERC), 0, 0, 1, 1},
    {&__pyx_n_s_RADICAL_DISPLAY_STYLE_VERTICAL_G, __pyx_k_RADICAL_DISPLAY_STYLE_VERTICAL_G, sizeof(__pyx_k_RADICAL_DISPLAY_STYLE_VERTICAL_G), 0, 0, 1, 1},
    {&__pyx_n_s_RADICAL_EXTRA_ASCENDER, __pyx_k_RADICAL_EXTRA_ASCENDER, sizeof(__pyx_k_RADICAL_EXTRA_ASCENDER), 0, 0, 1, 1},
    {&__pyx_n_s_RADICAL_KERN_AFTER_DEGREE, __pyx_k_RADICAL_KERN_AFTER_DEGREE, sizeof(__pyx_k_RADICAL_KERN_AFTER_DEGREE), 0, 0, 1, 1},
    {&__pyx_n_s_RADICAL_KERN_BEFORE_DEGREE, __pyx_k_RADICAL_KERN_BEFORE_DEGREE, sizeof(__pyx_k_RADICAL_KERN_BEFORE_DEGREE), 0, 0, 1, 1},
    {&__pyx_n_s_RADICAL_RULE_THICKNESS, __pyx_k_RADICAL_RULE_THICKNESS, sizeof(__pyx_k_RADICAL_RULE_THICKNESS), 0, 0, 1, 1},
    {&__pyx_n_s_RADICAL_VERTICAL_GAP, __pyx_k_RADICAL_VERTICAL_GAP, sizeof(__pyx_k_RADICAL_VERTICAL_GAP), 0, 0, 1, 1},
    {&__pyx_n_s_REFLECT, __pyx_k_REFLECT, sizeof(__pyx_k_REFLECT), 0, 0, 1, 1},
    {&__pyx_n_s_REMOVE_DEFAULT_IGNORABLES, __pyx_k_REMOVE_DEFAULT_IGNORABLES, sizeof(__pyx_k_REMOVE_DEFAULT_IGNORABLES), 0, 0, 1, 1},
    {&__pyx_n_s_REPEAT, __pyx_k_REPEAT, sizeof(__pyx_k_REPEAT), 0, 0, 1, 1},
    {&__pyx_n_s_RETAIN_GIDS, __pyx_k_RETAIN_GIDS, sizeof(__pyx_k_RETAIN_GIDS), 0, 0, 1, 1},
    {&__pyx_n_s_RepackerError, __pyx_k_RepackerError, sizeof(__pyx_k_RepackerError), 0, 0, 1, 1},
    {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
    {&__pyx_n_s_SAFE_TO_INSERT_TATWEEL, __pyx_k_SAFE_TO_INSERT_TATWEEL, sizeof(__pyx_k_SAFE_TO_INSERT_TATWEEL), 0, 0, 1, 1},
    {&__pyx_n_s_SCREEN, __pyx_k_SCREEN, sizeof(__pyx_k_SCREEN), 0, 0, 1, 1},
    {&__pyx_n_s_SCRIPT_PERCENT_SCALE_DOWN, __pyx_k_SCRIPT_PERCENT_SCALE_DOWN, sizeof(__pyx_k_SCRIPT_PERCENT_SCALE_DOWN), 0, 0, 1, 1},
    {&__pyx_n_s_SCRIPT_SCRIPT_PERCENT_SCALE_DOWN, __pyx_k_SCRIPT_SCRIPT_PERCENT_SCALE_DOWN, sizeof(__pyx_k_SCRIPT_SCRIPT_PERCENT_SCALE_DOWN), 0, 0, 1, 1},
    {&__pyx_n_s_SET_OVERLAPS_FLAG, __pyx_k_SET_OVERLAPS_FLAG, sizeof(__pyx_k_SET_OVERLAPS_FLAG), 0, 0, 1, 1},
    {&__pyx_n_s_SKEWED_FRACTION_HORIZONTAL_GAP, __pyx_k_SKEWED_FRACTION_HORIZONTAL_GAP, sizeof(__pyx_k_SKEWED_FRACTION_HORIZONTAL_GAP), 0, 0, 1, 1},
    {&__pyx_n_s_SKEWED_FRACTION_VERTICAL_GAP, __pyx_k_SKEWED_FRACTION_VERTICAL_GAP, sizeof(__pyx_k_SKEWED_FRACTION_VERTICAL_GAP), 0, 0, 1, 1},
    {&__pyx_n_s_SOFT_LIGHT, __pyx_k_SOFT_LIGHT, sizeof(__pyx_k_SOFT_LIGHT), 0, 0, 1, 1},
    {&__pyx_n_s_SPACE_AFTER_SCRIPT, __pyx_k_SPACE_AFTER_SCRIPT, sizeof(__pyx_k_SPACE_AFTER_SCRIPT), 0, 0, 1, 1},
    {&__pyx_n_s_SRC, __pyx_k_SRC, sizeof(__pyx_k_SRC), 0, 0, 1, 1},
    {&__pyx_n_s_SRC_ATOP, __pyx_k_SRC_ATOP, sizeof(__pyx_k_SRC_ATOP), 0, 0, 1, 1},
    {&__pyx_n_s_SRC_IN, __pyx_k_SRC_IN, sizeof(__pyx_k_SRC_IN), 0, 0, 1, 1},
    {&__pyx_n_s_SRC_OUT, __pyx_k_SRC_OUT, sizeof(__pyx_k_SRC_OUT), 0, 0, 1, 1},
    {&__pyx_n_s_SRC_OVER, __pyx_k_SRC_OVER, sizeof(__pyx_k_SRC_OVER), 0, 0, 1, 1},
    {&__pyx_n_s_STACK_BOTTOM_DISPLAY_STYLE_SHIFT, __pyx_k_STACK_BOTTOM_DISPLAY_STYLE_SHIFT, sizeof(__pyx_k_STACK_BOTTOM_DISPLAY_STYLE_SHIFT), 0, 0, 1, 1},
    {&__pyx_n_s_STACK_BOTTOM_SHIFT_DOWN, __pyx_k_STACK_BOTTOM_SHIFT_DOWN, sizeof(__pyx_k_STACK_BOTTOM_SHIFT_DOWN), 0, 0, 1, 1},
    {&__pyx_n_s_STACK_DISPLAY_STYLE_GAP_MIN, __pyx_k_STACK_DISPLAY_STYLE_GAP_MIN, sizeof(__pyx_k_STACK_DISPLAY_STYLE_GAP_MIN), 0, 0, 1, 1},
    {&__pyx_n_s_STACK_GAP_MIN, __pyx_k_STACK_GAP_MIN, sizeof(__pyx_k_STACK_GAP_MIN), 0, 0, 1, 1},
    {&__pyx_n_s_STACK_TOP_DISPLAY_STYLE_SHIFT_UP, __pyx_k_STACK_TOP_DISPLAY_STYLE_SHIFT_UP, sizeof(__pyx_k_STACK_TOP_DISPLAY_STYLE_SHIFT_UP), 0, 0, 1, 1},
    {&__pyx_n_s_STACK_TOP_SHIFT_UP, __pyx_k_STACK_TOP_SHIFT_UP, sizeof(__pyx_k_STACK_TOP_SHIFT_UP), 0, 0, 1, 1},
    {&__pyx_n_s_STRETCH_STACK_BOTTOM_SHIFT_DOWN, __pyx_k_STRETCH_STACK_BOTTOM_SHIFT_DOWN, sizeof(__pyx_k_STRETCH_STACK_BOTTOM_SHIFT_DOWN), 0, 0, 1, 1},
    {&__pyx_n_s_STRETCH_STACK_GAP_ABOVE_MIN, __pyx_k_STRETCH_STACK_GAP_ABOVE_MIN, sizeof(__pyx_k_STRETCH_STACK_GAP_ABOVE_MIN), 0, 0, 1, 1},
    {&__pyx_n_s_STRETCH_STACK_GAP_BELOW_MIN, __pyx_k_STRETCH_STACK_GAP_BELOW_MIN, sizeof(__pyx_k_STRETCH_STACK_GAP_BELOW_MIN), 0, 0, 1, 1},
    {&__pyx_n_s_STRETCH_STACK_TOP_SHIFT_UP, __pyx_k_STRETCH_STACK_TOP_SHIFT_UP, sizeof(__pyx_k_STRETCH_STACK_TOP_SHIFT_UP), 0, 0, 1, 1},
    {&__pyx_n_s_SUBSCRIPT_BASELINE_DROP_MIN, __pyx_k_SUBSCRIPT_BASELINE_DROP_MIN, sizeof(__pyx_k_SUBSCRIPT_BASELINE_DROP_MIN), 0, 0, 1, 1},
    {&__pyx_n_s_SUBSCRIPT_SHIFT_DOWN, __pyx_k_SUBSCRIPT_SHIFT_DOWN, sizeof(__pyx_k_SUBSCRIPT_SHIFT_DOWN), 0, 0, 1, 1},
    {&__pyx_n_s_SUBSCRIPT_TOP_MAX, __pyx_k_SUBSCRIPT_TOP_MAX, sizeof(__pyx_k_SUBSCRIPT_TOP_MAX), 0, 0, 1, 1},
    {&__pyx_n_s_SUB_SUPERSCRIPT_GAP_MIN, __pyx_k_SUB_SUPERSCRIPT_GAP_MIN, sizeof(__pyx_k_SUB_SUPERSCRIPT_GAP_MIN), 0, 0, 1, 1},
    {&__pyx_n_s_SUPERSCRIPT_BASELINE_DROP_MAX, __pyx_k_SUPERSCRIPT_BASELINE_DROP_MAX, sizeof(__pyx_k_SUPERSCRIPT_BASELINE_DROP_MAX), 0, 0, 1, 1},
    {&__pyx_n_s_SUPERSCRIPT_BOTTOM_MAX_WITH_SUBS, __pyx_k_SUPERSCRIPT_BOTTOM_MAX_WITH_SUBS, sizeof(__pyx_k_SUPERSCRIPT_BOTTOM_MAX_WITH_SUBS), 0, 0, 1, 1},
    {&__pyx_n_s_SUPERSCRIPT_BOTTOM_MIN, __pyx_k_SUPERSCRIPT_BOTTOM_MIN, sizeof(__pyx_k_SUPERSCRIPT_BOTTOM_MIN), 0, 0, 1, 1},
    {&__pyx_n_s_SUPERSCRIPT_SHIFT_UP, __pyx_k_SUPERSCRIPT_SHIFT_UP, sizeof(__pyx_k_SUPERSCRIPT_SHIFT_UP), 0, 0, 1, 1},
    {&__pyx_n_s_SUPERSCRIPT_SHIFT_UP_CRAMPED, __pyx_k_SUPERSCRIPT_SHIFT_UP_CRAMPED, sizeof(__pyx_k_SUPERSCRIPT_SHIFT_UP_CRAMPED), 0, 0, 1, 1},
    {&__pyx_n_s_Sequence, __pyx_k_Sequence, sizeof(__pyx_k_Sequence), 0, 0, 1, 1},
    {&__pyx_kp_u_Set, __pyx_k_Set, sizeof(__pyx_k_Set), 0, 1, 0, 0},
    {&__pyx_n_s_SetIter, __pyx_k_SetIter, sizeof(__pyx_k_SetIter), 0, 0, 1, 1},
    {&__pyx_n_s_SetIter___reduce_cython, __pyx_k_SetIter___reduce_cython, sizeof(__pyx_k_SetIter___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_SetIter___setstate_cython, __pyx_k_SetIter___setstate_cython, sizeof(__pyx_k_SetIter___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Set_2, __pyx_k_Set_2, sizeof(__pyx_k_Set_2), 0, 0, 1, 1},
    {&__pyx_n_s_Set___copy, __pyx_k_Set___copy, sizeof(__pyx_k_Set___copy), 0, 0, 1, 1},
    {&__pyx_n_s_Set___reduce_cython, __pyx_k_Set___reduce_cython, sizeof(__pyx_k_Set___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Set___setstate_cython, __pyx_k_Set___setstate_cython, sizeof(__pyx_k_Set___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_Set__is_equal, __pyx_k_Set__is_equal, sizeof(__pyx_k_Set__is_equal), 0, 0, 1, 1},
    {&__pyx_n_s_Set__set, __pyx_k_Set__set, sizeof(__pyx_k_Set__set), 0, 0, 1, 1},
    {&__pyx_n_s_Set__update, __pyx_k_Set__update, sizeof(__pyx_k_Set__update), 0, 0, 1, 1},
    {&__pyx_n_s_Set_add, __pyx_k_Set_add, sizeof(__pyx_k_Set_add), 0, 0, 1, 1},
    {&__pyx_n_s_Set_add_range, __pyx_k_Set_add_range, sizeof(__pyx_k_Set_add_range), 0, 0, 1, 1},
    {&__pyx_n_s_Set_clear, __pyx_k_Set_clear, sizeof(__pyx_k_Set_clear), 0, 0, 1, 1},
    {&__pyx_n_s_Set_copy, __pyx_k_Set_copy, sizeof(__pyx_k_Set_copy), 0, 0, 1, 1},
    {&__pyx_n_s_Set_del_range, __pyx_k_Set_del_range, sizeof(__pyx_k_Set_del_range), 0, 0, 1, 1},
    {&__pyx_n_s_Set_difference_update, __pyx_k_Set_difference_update, sizeof(__pyx_k_Set_difference_update), 0, 0, 1, 1},
    {&__pyx_n_s_Set_discard, __pyx_k_Set_discard, sizeof(__pyx_k_Set_discard), 0, 0, 1, 1},
    {&__pyx_n_s_Set_intersection_update, __pyx_k_Set_intersection_update, sizeof(__pyx_k_Set_intersection_update), 0, 0, 1, 1},
    {&__pyx_n_s_Set_invert, __pyx_k_Set_invert, sizeof(__pyx_k_Set_invert), 0, 0, 1, 1},
    {&__pyx_n_s_Set_is_inverted, __pyx_k_Set_is_inverted, sizeof(__pyx_k_Set_is_inverted), 0, 0, 1, 1},
    {&__pyx_n_s_Set_issubset, __pyx_k_Set_issubset, sizeof(__pyx_k_Set_issubset), 0, 0, 1, 1},
    {&__pyx_n_s_Set_issuperset, __pyx_k_Set_issuperset, sizeof(__pyx_k_Set_issuperset), 0, 0, 1, 1},
    {&__pyx_n_s_Set_remove, __pyx_k_Set_remove, sizeof(__pyx_k_Set_remove), 0, 0, 1, 1},
    {&__pyx_kp_u_Set_s, __pyx_k_Set_s, sizeof(__pyx_k_Set_s), 0, 1, 0, 0},
    {&__pyx_n_s_Set_set, __pyx_k_Set_set, sizeof(__pyx_k_Set_set), 0, 0, 1, 1},
    {&__pyx_n_s_Set_symmetric_difference_update, __pyx_k_Set_symmetric_difference_update, sizeof(__pyx_k_Set_symmetric_difference_update), 0, 0, 1, 1},
    {&__pyx_n_s_Set_update, __pyx_k_Set_update, sizeof(__pyx_k_Set_update), 0, 0, 1, 1},
    {&__pyx_n_s_StopIteration, __pyx_k_StopIteration, sizeof(__pyx_k_StopIteration), 0, 0, 1, 1},
    {&__pyx_n_s_SubsetFlags, __pyx_k_SubsetFlags, sizeof(__pyx_k_SubsetFlags), 0, 0, 1, 1},
    {&__pyx_n_s_SubsetInput, __pyx_k_SubsetInput, sizeof(__pyx_k_SubsetInput), 0, 0, 1, 1},
    {&__pyx_n_s_SubsetInputSets, __pyx_k_SubsetInputSets, sizeof(__pyx_k_SubsetInputSets), 0, 0, 1, 1},
    {&__pyx_n_s_SubsetInput___reduce_cython, __pyx_k_SubsetInput___reduce_cython, sizeof(__pyx_k_SubsetInput___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_SubsetInput___setstate_cython, __pyx_k_SubsetInput___setstate_cython, sizeof(__pyx_k_SubsetInput___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_SubsetInput_keep_everything, __pyx_k_SubsetInput_keep_everything, sizeof(__pyx_k_SubsetInput_keep_everything), 0, 0, 1, 1},
    {&__pyx_n_s_SubsetInput_pin_axis_location, __pyx_k_SubsetInput_pin_axis_location, sizeof(__pyx_k_SubsetInput_pin_axis_location), 0, 0, 1, 1},
    {&__pyx_n_s_SubsetInput_pin_axis_to_default, __pyx_k_SubsetInput_pin_axis_to_default, sizeof(__pyx_k_SubsetInput_pin_axis_to_default), 0, 0, 1, 1},
    {&__pyx_n_s_SubsetInput_sets, __pyx_k_SubsetInput_sets, sizeof(__pyx_k_SubsetInput_sets), 0, 0, 1, 1},
    {&__pyx_n_s_SubsetInput_subset, __pyx_k_SubsetInput_subset, sizeof(__pyx_k_SubsetInput_subset), 0, 0, 1, 1},
    {&__pyx_n_s_SubsetPlan, __pyx_k_SubsetPlan, sizeof(__pyx_k_SubsetPlan), 0, 0, 1, 1},
    {&__pyx_n_s_SubsetPlan___reduce_cython, __pyx_k_SubsetPlan___reduce_cython, sizeof(__pyx_k_SubsetPlan___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_SubsetPlan___setstate_cython, __pyx_k_SubsetPlan___setstate_cython, sizeof(__pyx_k_SubsetPlan___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_SubsetPlan_execute, __pyx_k_SubsetPlan_execute, sizeof(__pyx_k_SubsetPlan_execute), 0, 0, 1, 1},
    {&__pyx_kp_u_Subsetting_failed, __pyx_k_Subsetting_failed, sizeof(__pyx_k_Subsetting_failed), 0, 1, 0, 0},
    {&__pyx_n_s_TOP_LEFT, __pyx_k_TOP_LEFT, sizeof(__pyx_k_TOP_LEFT), 0, 0, 1, 1},
    {&__pyx_n_s_TOP_RIGHT, __pyx_k_TOP_RIGHT, sizeof(__pyx_k_TOP_RIGHT), 0, 0, 1, 1},
    {&__pyx_n_s_Tuple, __pyx_k_Tuple, sizeof(__pyx_k_Tuple), 0, 0, 1, 1},
    {&__pyx_kp_s_Tuple_List_OTMathGlyphPart_int, __pyx_k_Tuple_List_OTMathGlyphPart_int, sizeof(__pyx_k_Tuple_List_OTMathGlyphPart_int), 0, 0, 1, 0},
    {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
    {&__pyx_n_s_UNCLASSIFIED, __pyx_k_UNCLASSIFIED, sizeof(__pyx_k_UNCLASSIFIED), 0, 0, 1, 1},
    {&__pyx_n_s_UNDERBAR_EXTRA_DESCENDER, __pyx_k_UNDERBAR_EXTRA_DESCENDER, sizeof(__pyx_k_UNDERBAR_EXTRA_DESCENDER), 0, 0, 1, 1},
    {&__pyx_n_s_UNDERBAR_RULE_THICKNESS, __pyx_k_UNDERBAR_RULE_THICKNESS, sizeof(__pyx_k_UNDERBAR_RULE_THICKNESS), 0, 0, 1, 1},
    {&__pyx_n_s_UNDERBAR_VERTICAL_GAP, __pyx_k_UNDERBAR_VERTICAL_GAP, sizeof(__pyx_k_UNDERBAR_VERTICAL_GAP), 0, 0, 1, 1},
    {&__pyx_n_s_UNICODE, __pyx_k_UNICODE, sizeof(__pyx_k_UNICODE), 0, 0, 1, 1},
    {&__pyx_n_s_UNSAFE_TO_BREAK, __pyx_k_UNSAFE_TO_BREAK, sizeof(__pyx_k_UNSAFE_TO_BREAK), 0, 0, 1, 1},
    {&__pyx_n_s_UNSAFE_TO_CONCAT, __pyx_k_UNSAFE_TO_CONCAT, sizeof(__pyx_k_UNSAFE_TO_CONCAT), 0, 0, 1, 1},
    {&__pyx_n_s_UPPER_LIMIT_BASELINE_RISE_MIN, __pyx_k_UPPER_LIMIT_BASELINE_RISE_MIN, sizeof(__pyx_k_UPPER_LIMIT_BASELINE_RISE_MIN), 0, 0, 1, 1},
    {&__pyx_n_s_UPPER_LIMIT_GAP_MIN, __pyx_k_UPPER_LIMIT_GAP_MIN, sizeof(__pyx_k_UPPER_LIMIT_GAP_MIN), 0, 0, 1, 1},
    {&__pyx_n_s_USABLE_WITH_DARK_BACKGROUND, __pyx_k_USABLE_WITH_DARK_BACKGROUND, sizeof(__pyx_k_USABLE_WITH_DARK_BACKGROUND), 0, 0, 1, 1},
    {&__pyx_n_s_USABLE_WITH_LIGHT_BACKGROUND, __pyx_k_USABLE_WITH_LIGHT_BACKGROUND, sizeof(__pyx_k_USABLE_WITH_LIGHT_BACKGROUND), 0, 0, 1, 1},
    {&__pyx_n_s_Union, __pyx_k_Union, sizeof(__pyx_k_Union), 0, 0, 1, 1},
    {&__pyx_kp_s_Union_str_Path, __pyx_k_Union_str_Path, sizeof(__pyx_k_Union_str_Path), 0, 0, 1, 0},
    {&__pyx_n_s_VERIFY, __pyx_k_VERIFY, sizeof(__pyx_k_VERIFY), 0, 0, 1, 1},
    {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
    {&__pyx_n_s_XOR, __pyx_k_XOR, sizeof(__pyx_k_XOR), 0, 0, 1, 1},
    {&__pyx_kp_u__10, __pyx_k__10, sizeof(__pyx_k__10), 0, 1, 0, 0},
    {&__pyx_kp_b__15, __pyx_k__15, sizeof(__pyx_k__15), 0, 0, 0, 0},
    {&__pyx_kp_u__15, __pyx_k__15, sizeof(__pyx_k__15), 0, 1, 0, 0},
    {&__pyx_kp_u__16, __pyx_k__16, sizeof(__pyx_k__16), 0, 1, 0, 0},
    {&__pyx_kp_u__19, __pyx_k__19, sizeof(__pyx_k__19), 0, 1, 0, 0},
    {&__pyx_kp_u__21, __pyx_k__21, sizeof(__pyx_k__21), 0, 1, 0, 0},
    {&__pyx_n_s__22, __pyx_k__22, sizeof(__pyx_k__22), 0, 0, 1, 1},
    {&__pyx_kp_u__23, __pyx_k__23, sizeof(__pyx_k__23), 0, 1, 0, 0},
    {&__pyx_n_s__287, __pyx_k__287, sizeof(__pyx_k__287), 0, 0, 1, 1},
    {&__pyx_n_s_a, __pyx_k_a, sizeof(__pyx_k_a), 0, 0, 1, 1},
    {&__pyx_n_s_add, __pyx_k_add, sizeof(__pyx_k_add), 0, 0, 1, 1},
    {&__pyx_n_s_add_codepoints, __pyx_k_add_codepoints, sizeof(__pyx_k_add_codepoints), 0, 0, 1, 1},
    {&__pyx_n_s_add_range, __pyx_k_add_range, sizeof(__pyx_k_add_range), 0, 0, 1, 1},
    {&__pyx_n_s_add_str, __pyx_k_add_str, sizeof(__pyx_k_add_str), 0, 0, 1, 1},
    {&__pyx_n_s_add_utf8, __pyx_k_add_utf8, sizeof(__pyx_k_add_utf8), 0, 0, 1, 1},
    {&__pyx_n_u_advance, __pyx_k_advance, sizeof(__pyx_k_advance), 0, 1, 0, 1},
    {&__pyx_n_s_alpha, __pyx_k_alpha, sizeof(__pyx_k_alpha), 0, 0, 1, 1},
    {&__pyx_n_u_alpha, __pyx_k_alpha, sizeof(__pyx_k_alpha), 0, 1, 0, 1},
    {&__pyx_n_s_alternate_count, __pyx_k_alternate_count, sizeof(__pyx_k_alternate_count), 0, 0, 1, 1},
    {&__pyx_n_s_alternate_glyphs, __pyx_k_alternate_glyphs, sizeof(__pyx_k_alternate_glyphs), 0, 0, 1, 1},
    {&__pyx_n_s_alternates, __pyx_k_alternates, sizeof(__pyx_k_alternates), 0, 0, 1, 1},
    {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
    {&__pyx_n_s_ascender, __pyx_k_ascender, sizeof(__pyx_k_ascender), 0, 0, 1, 1},
    {&__pyx_n_u_ascender, __pyx_k_ascender, sizeof(__pyx_k_ascender), 0, 1, 0, 1},
    {&__pyx_n_s_assembly, __pyx_k_assembly, sizeof(__pyx_k_assembly), 0, 0, 1, 1},
    {&__pyx_n_s_assembly_array, __pyx_k_assembly_array, sizeof(__pyx_k_assembly_array), 0, 0, 1, 1},
    {&__pyx_n_s_asyncio_coroutines, __pyx_k_asyncio_coroutines, sizeof(__pyx_k_asyncio_coroutines), 0, 0, 1, 1},
    {&__pyx_n_s_b, __pyx_k_b, sizeof(__pyx_k_b), 0, 0, 1, 1},
    {&__pyx_n_s_baseline_tag, __pyx_k_baseline_tag, sizeof(__pyx_k_baseline_tag), 0, 0, 1, 1},
    {&__pyx_n_s_blob, __pyx_k_blob, sizeof(__pyx_k_blob), 0, 0, 1, 1},
    {&__pyx_n_s_blob_data, __pyx_k_blob_data, sizeof(__pyx_k_blob_data), 0, 0, 1, 1},
    {&__pyx_n_s_blob_length, __pyx_k_blob_length, sizeof(__pyx_k_blob_length), 0, 0, 1, 1},
    {&__pyx_n_s_blue, __pyx_k_blue, sizeof(__pyx_k_blue), 0, 0, 1, 1},
    {&__pyx_n_u_blue, __pyx_k_blue, sizeof(__pyx_k_blue), 0, 1, 0, 1},
    {&__pyx_n_s_bool, __pyx_k_bool, sizeof(__pyx_k_bool), 0, 0, 1, 1},
    {&__pyx_n_s_buffer, __pyx_k_buffer, sizeof(__pyx_k_buffer), 0, 0, 1, 1},
    {&__pyx_n_s_bytes, __pyx_k_bytes, sizeof(__pyx_k_bytes), 0, 0, 1, 1},
    {&__pyx_n_s_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
    {&__pyx_n_s_c_coords, __pyx_k_c_coords, sizeof(__pyx_k_c_coords), 0, 0, 1, 1},
    {&__pyx_n_s_c_foreground, __pyx_k_c_foreground, sizeof(__pyx_k_c_foreground), 0, 0, 1, 1},
    {&__pyx_n_s_c_shapers, __pyx_k_c_shapers, sizeof(__pyx_k_c_shapers), 0, 0, 1, 1},
    {&__pyx_n_s_callback, __pyx_k_callback, sizeof(__pyx_k_callback), 0, 0, 1, 1},
    {&__pyx_n_s_clear, __pyx_k_clear, sizeof(__pyx_k_clear), 0, 0, 1, 1},
    {&__pyx_n_s_clear_contents, __pyx_k_clear_contents, sizeof(__pyx_k_clear_contents), 0, 0, 1, 1},
    {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
    {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
    {&__pyx_n_s_closePath, __pyx_k_closePath, sizeof(__pyx_k_closePath), 0, 0, 1, 1},
    {&__pyx_n_s_cls, __pyx_k_cls, sizeof(__pyx_k_cls), 0, 0, 1, 1},
    {&__pyx_n_s_codepoints, __pyx_k_codepoints, sizeof(__pyx_k_codepoints), 0, 0, 1, 1},
    {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
    {&__pyx_n_u_color, __pyx_k_color, sizeof(__pyx_k_color), 0, 1, 0, 1},
    {&__pyx_n_s_color_count, __pyx_k_color_count, sizeof(__pyx_k_color_count), 0, 0, 1, 1},
    {&__pyx_n_s_color_index, __pyx_k_color_index, sizeof(__pyx_k_color_index), 0, 0, 1, 1},
    {&__pyx_n_u_color_index, __pyx_k_color_index, sizeof(__pyx_k_color_index), 0, 1, 0, 1},
    {&__pyx_n_s_colors, __pyx_k_colors, sizeof(__pyx_k_colors), 0, 0, 1, 1},
    {&__pyx_n_s_constant, __pyx_k_constant, sizeof(__pyx_k_constant), 0, 0, 1, 1},
    {&__pyx_n_s_coords, __pyx_k_coords, sizeof(__pyx_k_coords), 0, 0, 1, 1},
    {&__pyx_n_s_coords_2dot14, __pyx_k_coords_2dot14, sizeof(__pyx_k_coords_2dot14), 0, 0, 1, 1},
    {&__pyx_n_s_copy, __pyx_k_copy, sizeof(__pyx_k_copy), 0, 0, 1, 1},
    {&__pyx_n_s_copy_2, __pyx_k_copy_2, sizeof(__pyx_k_copy_2), 0, 0, 1, 1},
    {&__pyx_n_s_correction_height, __pyx_k_correction_height, sizeof(__pyx_k_correction_height), 0, 0, 1, 1},
    {&__pyx_n_s_count, __pyx_k_count, sizeof(__pyx_k_count), 0, 0, 1, 1},
    {&__pyx_n_s_create, __pyx_k_create, sizeof(__pyx_k_create), 0, 0, 1, 1},
    {&__pyx_n_s_create_for_tables, __pyx_k_create_for_tables, sizeof(__pyx_k_create_for_tables), 0, 0, 1, 1},
    {&__pyx_n_s_cstr, __pyx_k_cstr, sizeof(__pyx_k_cstr), 0, 0, 1, 1},
    {&__pyx_n_s_curveTo, __pyx_k_curveTo, sizeof(__pyx_k_curveTo), 0, 0, 1, 1},
    {&__pyx_n_s_data, __pyx_k_data, sizeof(__pyx_k_data), 0, 0, 1, 1},
    {&__pyx_n_s_del_range, __pyx_k_del_range, sizeof(__pyx_k_del_range), 0, 0, 1, 1},
    {&__pyx_n_s_descender, __pyx_k_descender, sizeof(__pyx_k_descender), 0, 0, 1, 1},
    {&__pyx_n_u_descender, __pyx_k_descender, sizeof(__pyx_k_descender), 0, 1, 0, 1},
    {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
    {&__pyx_n_s_difference_update, __pyx_k_difference_update, sizeof(__pyx_k_difference_update), 0, 0, 1, 1},
    {&__pyx_n_s_direction, __pyx_k_direction, sizeof(__pyx_k_direction), 0, 0, 1, 1},
    {&__pyx_kp_u_disable, __pyx_k_disable, sizeof(__pyx_k_disable), 0, 1, 0, 0},
    {&__pyx_n_s_discard, __pyx_k_discard, sizeof(__pyx_k_discard), 0, 0, 1, 1},
    {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
    {&__pyx_n_s_draw_data, __pyx_k_draw_data, sizeof(__pyx_k_draw_data), 0, 0, 1, 1},
    {&__pyx_n_s_draw_funcs, __pyx_k_draw_funcs, sizeof(__pyx_k_draw_funcs), 0, 0, 1, 1},
    {&__pyx_n_s_draw_glyph, __pyx_k_draw_glyph, sizeof(__pyx_k_draw_glyph), 0, 0, 1, 1},
    {&__pyx_kp_u_draw_glyph_is_deprecated_use_Fon, __pyx_k_draw_glyph_is_deprecated_use_Fon, sizeof(__pyx_k_draw_glyph_is_deprecated_use_Fon), 0, 1, 0, 0},
    {&__pyx_n_s_draw_glyph_with_pen, __pyx_k_draw_glyph_with_pen, sizeof(__pyx_k_draw_glyph_with_pen), 0, 0, 1, 1},
    {&__pyx_n_s_draw_state, __pyx_k_draw_state, sizeof(__pyx_k_draw_state), 0, 0, 1, 1},
    {&__pyx_n_s_draw_state_p, __pyx_k_draw_state_p, sizeof(__pyx_k_draw_state_p), 0, 0, 1, 1},
    {&__pyx_kp_u_enable, __pyx_k_enable, sizeof(__pyx_k_enable), 0, 1, 0, 0},
    {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
    {&__pyx_n_s_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
    {&__pyx_n_u_end_connector_length, __pyx_k_end_connector_length, sizeof(__pyx_k_end_connector_length), 0, 1, 0, 1},
    {&__pyx_n_s_enum, __pyx_k_enum, sizeof(__pyx_k_enum), 0, 0, 1, 1},
    {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
    {&__pyx_n_s_execute, __pyx_k_execute, sizeof(__pyx_k_execute), 0, 0, 1, 1},
    {&__pyx_n_s_extents, __pyx_k_extents, sizeof(__pyx_k_extents), 0, 0, 1, 1},
    {&__pyx_n_s_face, __pyx_k_face, sizeof(__pyx_k_face), 0, 0, 1, 1},
    {&__pyx_n_s_face_or_font, __pyx_k_face_or_font, sizeof(__pyx_k_face_or_font), 0, 0, 1, 1},
    {&__pyx_n_s_feat, __pyx_k_feat, sizeof(__pyx_k_feat), 0, 0, 1, 1},
    {&__pyx_n_s_feature_count, __pyx_k_feature_count, sizeof(__pyx_k_feature_count), 0, 0, 1, 1},
    {&__pyx_n_s_feature_tags, __pyx_k_feature_tags, sizeof(__pyx_k_feature_tags), 0, 0, 1, 1},
    {&__pyx_n_s_features, __pyx_k_features, sizeof(__pyx_k_features), 0, 0, 1, 1},
    {&__pyx_n_s_filename, __pyx_k_filename, sizeof(__pyx_k_filename), 0, 0, 1, 1},
    {&__pyx_n_s_first, __pyx_k_first, sizeof(__pyx_k_first), 0, 0, 1, 1},
    {&__pyx_n_u_flags, __pyx_k_flags, sizeof(__pyx_k_flags), 0, 1, 0, 1},
    {&__pyx_n_s_float, __pyx_k_float, sizeof(__pyx_k_float), 0, 0, 1, 1},
    {&__pyx_n_s_font, __pyx_k_font, sizeof(__pyx_k_font), 0, 0, 1, 1},
    {&__pyx_n_s_foreground, __pyx_k_foreground, sizeof(__pyx_k_foreground), 0, 0, 1, 1},
    {&__pyx_n_s_from_file_path, __pyx_k_from_file_path, sizeof(__pyx_k_from_file_path), 0, 0, 1, 1},
    {&__pyx_n_s_from_int, __pyx_k_from_int, sizeof(__pyx_k_from_int), 0, 0, 1, 1},
    {&__pyx_n_s_fsencode, __pyx_k_fsencode, sizeof(__pyx_k_fsencode), 0, 0, 1, 1},
    {&__pyx_n_u_full_advance, __pyx_k_full_advance, sizeof(__pyx_k_full_advance), 0, 1, 0, 1},
    {&__pyx_n_s_func, __pyx_k_func, sizeof(__pyx_k_func), 0, 0, 1, 1},
    {&__pyx_n_s_func_p, __pyx_k_func_p, sizeof(__pyx_k_func_p), 0, 0, 1, 1},
    {&__pyx_n_s_funcs, __pyx_k_funcs, sizeof(__pyx_k_funcs), 0, 0, 1, 1},
    {&__pyx_n_s_g, __pyx_k_g, sizeof(__pyx_k_g), 0, 0, 1, 1},
    {&__pyx_kp_u_gc, __pyx_k_gc, sizeof(__pyx_k_gc), 0, 1, 0, 0},
    {&__pyx_n_s_genexpr, __pyx_k_genexpr, sizeof(__pyx_k_genexpr), 0, 0, 1, 1},
    {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
    {&__pyx_n_s_get_font_extents, __pyx_k_get_font_extents, sizeof(__pyx_k_get_font_extents), 0, 0, 1, 1},
    {&__pyx_n_s_get_glyph_extents, __pyx_k_get_glyph_extents, sizeof(__pyx_k_get_glyph_extents), 0, 0, 1, 1},
    {&__pyx_n_s_get_glyph_from_name, __pyx_k_get_glyph_from_name, sizeof(__pyx_k_get_glyph_from_name), 0, 0, 1, 1},
    {&__pyx_n_s_get_glyph_h_advance, __pyx_k_get_glyph_h_advance, sizeof(__pyx_k_get_glyph_h_advance), 0, 0, 1, 1},
    {&__pyx_n_s_get_glyph_h_origin, __pyx_k_get_glyph_h_origin, sizeof(__pyx_k_get_glyph_h_origin), 0, 0, 1, 1},
    {&__pyx_n_s_get_glyph_name, __pyx_k_get_glyph_name, sizeof(__pyx_k_get_glyph_name), 0, 0, 1, 1},
    {&__pyx_n_s_get_glyph_shape, __pyx_k_get_glyph_shape, sizeof(__pyx_k_get_glyph_shape), 0, 0, 1, 1},
    {&__pyx_kp_u_get_glyph_shape_is_deprecated_us, __pyx_k_get_glyph_shape_is_deprecated_us, sizeof(__pyx_k_get_glyph_shape_is_deprecated_us), 0, 1, 0, 0},
    {&__pyx_n_s_get_glyph_v_advance, __pyx_k_get_glyph_v_advance, sizeof(__pyx_k_get_glyph_v_advance), 0, 0, 1, 1},
    {&__pyx_n_s_get_glyph_v_origin, __pyx_k_get_glyph_v_origin, sizeof(__pyx_k_get_glyph_v_origin), 0, 0, 1, 1},
    {&__pyx_n_s_get_nominal_glyph, __pyx_k_get_nominal_glyph, sizeof(__pyx_k_get_nominal_glyph), 0, 0, 1, 1},
    {&__pyx_n_s_get_var_coords_design, __pyx_k_get_var_coords_design, sizeof(__pyx_k_get_var_coords_design), 0, 0, 1, 1},
    {&__pyx_n_s_get_var_coords_normalized, __pyx_k_get_var_coords_normalized, sizeof(__pyx_k_get_var_coords_normalized), 0, 0, 1, 1},
    {&__pyx_n_s_get_variation_glyph, __pyx_k_get_variation_glyph, sizeof(__pyx_k_get_variation_glyph), 0, 0, 1, 1},
    {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
    {&__pyx_n_s_gid, __pyx_k_gid, sizeof(__pyx_k_gid), 0, 0, 1, 1},
    {&__pyx_n_s_glyph, __pyx_k_glyph, sizeof(__pyx_k_glyph), 0, 0, 1, 1},
    {&__pyx_n_u_glyph, __pyx_k_glyph, sizeof(__pyx_k_glyph), 0, 1, 0, 1},
    {&__pyx_n_s_glyph_from_string, __pyx_k_glyph_from_string, sizeof(__pyx_k_glyph_from_string), 0, 0, 1, 1},
    {&__pyx_n_s_glyph_to_string, __pyx_k_glyph_to_string, sizeof(__pyx_k_glyph_to_string), 0, 0, 1, 1},
    {&__pyx_n_s_graphnodes, __pyx_k_graphnodes, sizeof(__pyx_k_graphnodes), 0, 0, 1, 1},
    {&__pyx_n_s_green, __pyx_k_green, sizeof(__pyx_k_green), 0, 0, 1, 1},
    {&__pyx_n_u_green, __pyx_k_green, sizeof(__pyx_k_green), 0, 1, 0, 1},
    {&__pyx_n_s_guess_segment_properties, __pyx_k_guess_segment_properties, sizeof(__pyx_k_guess_segment_properties), 0, 0, 1, 1},
    {&__pyx_n_u_hang, __pyx_k_hang, sizeof(__pyx_k_hang), 0, 1, 0, 1},
    {&__pyx_n_s_hb_baseline_tag, __pyx_k_hb_baseline_tag, sizeof(__pyx_k_hb_baseline_tag), 0, 0, 1, 1},
    {&__pyx_n_s_hb_codepoint_t, __pyx_k_hb_codepoint_t, sizeof(__pyx_k_hb_codepoint_t), 0, 0, 1, 1},
    {&__pyx_n_s_hb_codepoints, __pyx_k_hb_codepoints, sizeof(__pyx_k_hb_codepoints), 0, 0, 1, 1},
    {&__pyx_n_s_hb_direction, __pyx_k_hb_direction, sizeof(__pyx_k_hb_direction), 0, 0, 1, 1},
    {&__pyx_n_s_hb_features, __pyx_k_hb_features, sizeof(__pyx_k_hb_features), 0, 0, 1, 1},
    {&__pyx_n_s_hb_language, __pyx_k_hb_language, sizeof(__pyx_k_hb_language), 0, 0, 1, 1},
    {&__pyx_n_s_hb_language_tag, __pyx_k_hb_language_tag, sizeof(__pyx_k_hb_language_tag), 0, 0, 1, 1},
    {&__pyx_n_s_hb_position, __pyx_k_hb_position, sizeof(__pyx_k_hb_position), 0, 0, 1, 1},
    {&__pyx_n_s_hb_script, __pyx_k_hb_script, sizeof(__pyx_k_hb_script), 0, 0, 1, 1},
    {&__pyx_n_s_hb_script_tag, __pyx_k_hb_script_tag, sizeof(__pyx_k_hb_script_tag), 0, 0, 1, 1},
    {&__pyx_n_s_hb_tag, __pyx_k_hb_tag, sizeof(__pyx_k_hb_tag), 0, 0, 1, 1},
    {&__pyx_n_s_hb_variations, __pyx_k_hb_variations, sizeof(__pyx_k_hb_variations), 0, 0, 1, 1},
    {&__pyx_n_s_head, __pyx_k_head, sizeof(__pyx_k_head), 0, 0, 1, 1},
    {&__pyx_n_u_height, __pyx_k_height, sizeof(__pyx_k_height), 0, 1, 0, 1},
    {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
    {&__pyx_n_u_icfb, __pyx_k_icfb, sizeof(__pyx_k_icfb), 0, 1, 0, 1},
    {&__pyx_n_u_icft, __pyx_k_icft, sizeof(__pyx_k_icft), 0, 1, 0, 1},
    {&__pyx_n_u_ideo, __pyx_k_ideo, sizeof(__pyx_k_ideo), 0, 1, 0, 1},
    {&__pyx_n_u_idtp, __pyx_k_idtp, sizeof(__pyx_k_idtp), 0, 1, 0, 1},
    {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
    {&__pyx_n_s_index, __pyx_k_index, sizeof(__pyx_k_index), 0, 0, 1, 1},
    {&__pyx_kp_u_index_out_of_range_for_feature_a, __pyx_k_index_out_of_range_for_feature_a, sizeof(__pyx_k_index_out_of_range_for_feature_a), 0, 1, 0, 0},
    {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
    {&__pyx_n_s_init_subclass, __pyx_k_init_subclass, sizeof(__pyx_k_init_subclass), 0, 0, 1, 1},
    {&__pyx_n_s_initializing, __pyx_k_initializing, sizeof(__pyx_k_initializing), 0, 0, 1, 1},
    {&__pyx_n_s_input, __pyx_k_input, sizeof(__pyx_k_input), 0, 0, 1, 1},
    {&__pyx_n_s_inst, __pyx_k_inst, sizeof(__pyx_k_inst), 0, 0, 1, 1},
    {&__pyx_n_s_int, __pyx_k_int, sizeof(__pyx_k_int), 0, 0, 1, 1},
    {&__pyx_kp_s_int_None, __pyx_k_int_None, sizeof(__pyx_k_int_None), 0, 0, 1, 0},
    {&__pyx_n_s_intersection_update, __pyx_k_intersection_update, sizeof(__pyx_k_intersection_update), 0, 0, 1, 1},
    {&__pyx_kp_u_invalid_baseline_tag, __pyx_k_invalid_baseline_tag, sizeof(__pyx_k_invalid_baseline_tag), 0, 1, 0, 0},
    {&__pyx_kp_u_invalid_constant, __pyx_k_invalid_constant, sizeof(__pyx_k_invalid_constant), 0, 1, 0, 0},
    {&__pyx_kp_u_invalid_kern, __pyx_k_invalid_kern, sizeof(__pyx_k_invalid_kern), 0, 1, 0, 0},
    {&__pyx_n_s_invert, __pyx_k_invert, sizeof(__pyx_k_invert), 0, 0, 1, 1},
    {&__pyx_n_s_is_coroutine, __pyx_k_is_coroutine, sizeof(__pyx_k_is_coroutine), 0, 0, 1, 1},
    {&__pyx_n_s_is_equal, __pyx_k_is_equal, sizeof(__pyx_k_is_equal), 0, 0, 1, 1},
    {&__pyx_n_u_is_foreground, __pyx_k_is_foreground, sizeof(__pyx_k_is_foreground), 0, 1, 0, 1},
    {&__pyx_n_s_is_inverted, __pyx_k_is_inverted, sizeof(__pyx_k_is_inverted), 0, 0, 1, 1},
    {&__pyx_kp_u_isenabled, __pyx_k_isenabled, sizeof(__pyx_k_isenabled), 0, 1, 0, 0},
    {&__pyx_n_s_issubset, __pyx_k_issubset, sizeof(__pyx_k_issubset), 0, 0, 1, 1},
    {&__pyx_n_s_issuperset, __pyx_k_issuperset, sizeof(__pyx_k_issuperset), 0, 0, 1, 1},
    {&__pyx_n_s_italics_correction, __pyx_k_italics_correction, sizeof(__pyx_k_italics_correction), 0, 0, 1, 1},
    {&__pyx_n_s_item_length, __pyx_k_item_length, sizeof(__pyx_k_item_length), 0, 0, 1, 1},
    {&__pyx_n_s_item_offset, __pyx_k_item_offset, sizeof(__pyx_k_item_offset), 0, 0, 1, 1},
    {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
    {&__pyx_n_s_join, __pyx_k_join, sizeof(__pyx_k_join), 0, 0, 1, 1},
    {&__pyx_n_s_k, __pyx_k_k, sizeof(__pyx_k_k), 0, 0, 1, 1},
    {&__pyx_n_s_keep_everything, __pyx_k_keep_everything, sizeof(__pyx_k_keep_everything), 0, 0, 1, 1},
    {&__pyx_n_s_kern, __pyx_k_kern, sizeof(__pyx_k_kern), 0, 0, 1, 1},
    {&__pyx_n_u_kern_value, __pyx_k_kern_value, sizeof(__pyx_k_kern_value), 0, 1, 0, 1},
    {&__pyx_n_s_kerns, __pyx_k_kerns, sizeof(__pyx_k_kerns), 0, 0, 1, 1},
    {&__pyx_n_s_kerns_array, __pyx_k_kerns_array, sizeof(__pyx_k_kerns_array), 0, 0, 1, 1},
    {&__pyx_n_s_keys, __pyx_k_keys, sizeof(__pyx_k_keys), 0, 0, 1, 1},
    {&__pyx_n_s_keys_locals_genexpr, __pyx_k_keys_locals_genexpr, sizeof(__pyx_k_keys_locals_genexpr), 0, 0, 1, 1},
    {&__pyx_n_s_kind, __pyx_k_kind, sizeof(__pyx_k_kind), 0, 0, 1, 1},
    {&__pyx_n_s_language_count, __pyx_k_language_count, sizeof(__pyx_k_language_count), 0, 0, 1, 1},
    {&__pyx_n_s_language_index, __pyx_k_language_index, sizeof(__pyx_k_language_index), 0, 0, 1, 1},
    {&__pyx_n_s_language_tag, __pyx_k_language_tag, sizeof(__pyx_k_language_tag), 0, 0, 1, 1},
    {&__pyx_n_s_language_tags, __pyx_k_language_tags, sizeof(__pyx_k_language_tags), 0, 0, 1, 1},
    {&__pyx_n_s_larger_set, __pyx_k_larger_set, sizeof(__pyx_k_larger_set), 0, 0, 1, 1},
    {&__pyx_n_s_last, __pyx_k_last, sizeof(__pyx_k_last), 0, 0, 1, 1},
    {&__pyx_n_s_layer_count, __pyx_k_layer_count, sizeof(__pyx_k_layer_count), 0, 0, 1, 1},
    {&__pyx_n_s_layers, __pyx_k_layers, sizeof(__pyx_k_layers), 0, 0, 1, 1},
    {&__pyx_n_s_length, __pyx_k_length, sizeof(__pyx_k_length), 0, 0, 1, 1},
    {&__pyx_n_s_lineTo, __pyx_k_lineTo, sizeof(__pyx_k_lineTo), 0, 0, 1, 1},
    {&__pyx_n_s_line_gap, __pyx_k_line_gap, sizeof(__pyx_k_line_gap), 0, 0, 1, 1},
    {&__pyx_n_u_line_gap, __pyx_k_line_gap, sizeof(__pyx_k_line_gap), 0, 1, 0, 1},
    {&__pyx_n_s_lookup_index, __pyx_k_lookup_index, sizeof(__pyx_k_lookup_index), 0, 0, 1, 1},
    {&__pyx_n_s_m, __pyx_k_m, sizeof(__pyx_k_m), 0, 0, 1, 1},
    {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
    {&__pyx_n_u_math, __pyx_k_math, sizeof(__pyx_k_math), 0, 1, 0, 1},
    {&__pyx_n_u_max_correction_height, __pyx_k_max_correction_height, sizeof(__pyx_k_max_correction_height), 0, 1, 0, 1},
    {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
    {&__pyx_n_s_methods, __pyx_k_methods, sizeof(__pyx_k_methods), 0, 0, 1, 1},
    {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
    {&__pyx_n_s_moveTo, __pyx_k_moveTo, sizeof(__pyx_k_moveTo), 0, 0, 1, 1},
    {&__pyx_n_s_mro_entries, __pyx_k_mro_entries, sizeof(__pyx_k_mro_entries), 0, 0, 1, 1},
    {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
    {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
    {&__pyx_n_s_name_id, __pyx_k_name_id, sizeof(__pyx_k_name_id), 0, 0, 1, 1},
    {&__pyx_n_s_namedtuple, __pyx_k_namedtuple, sizeof(__pyx_k_namedtuple), 0, 0, 1, 1},
    {&__pyx_n_s_new_face, __pyx_k_new_face, sizeof(__pyx_k_new_face), 0, 0, 1, 1},
    {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
    {&__pyx_n_s_node, __pyx_k_node, sizeof(__pyx_k_node), 0, 0, 1, 1},
    {&__pyx_n_s_num_nodes, __pyx_k_num_nodes, sizeof(__pyx_k_num_nodes), 0, 0, 1, 1},
    {&__pyx_kp_u_num_of_graph_nodes, __pyx_k_num_of_graph_nodes, sizeof(__pyx_k_num_of_graph_nodes), 0, 1, 0, 0},
    {&__pyx_n_s_obj_list, __pyx_k_obj_list, sizeof(__pyx_k_obj_list), 0, 0, 1, 1},
    {&__pyx_n_s_object, __pyx_k_object, sizeof(__pyx_k_object), 0, 0, 1, 1},
    {&__pyx_n_u_offset, __pyx_k_offset, sizeof(__pyx_k_offset), 0, 1, 0, 1},
    {&__pyx_n_s_os, __pyx_k_os, sizeof(__pyx_k_os), 0, 0, 1, 1},
    {&__pyx_n_s_ot_color_glyph_get_layers, __pyx_k_ot_color_glyph_get_layers, sizeof(__pyx_k_ot_color_glyph_get_layers), 0, 0, 1, 1},
    {&__pyx_n_s_ot_color_glyph_get_png, __pyx_k_ot_color_glyph_get_png, sizeof(__pyx_k_ot_color_glyph_get_png), 0, 0, 1, 1},
    {&__pyx_n_s_ot_color_glyph_get_svg, __pyx_k_ot_color_glyph_get_svg, sizeof(__pyx_k_ot_color_glyph_get_svg), 0, 0, 1, 1},
    {&__pyx_n_s_ot_color_glyph_has_paint, __pyx_k_ot_color_glyph_has_paint, sizeof(__pyx_k_ot_color_glyph_has_paint), 0, 0, 1, 1},
    {&__pyx_n_s_ot_color_has_layers, __pyx_k_ot_color_has_layers, sizeof(__pyx_k_ot_color_has_layers), 0, 0, 1, 1},
    {&__pyx_n_s_ot_color_has_paint, __pyx_k_ot_color_has_paint, sizeof(__pyx_k_ot_color_has_paint), 0, 0, 1, 1},
    {&__pyx_n_s_ot_color_has_palettes, __pyx_k_ot_color_has_palettes, sizeof(__pyx_k_ot_color_has_palettes), 0, 0, 1, 1},
    {&__pyx_n_s_ot_color_has_png, __pyx_k_ot_color_has_png, sizeof(__pyx_k_ot_color_has_png), 0, 0, 1, 1},
    {&__pyx_n_s_ot_color_has_svg, __pyx_k_ot_color_has_svg, sizeof(__pyx_k_ot_color_has_svg), 0, 0, 1, 1},
    {&__pyx_n_s_ot_color_palette_color_get_name, __pyx_k_ot_color_palette_color_get_name, sizeof(__pyx_k_ot_color_palette_color_get_name), 0, 0, 1, 1},
    {&__pyx_n_s_ot_color_palette_get_colors, __pyx_k_ot_color_palette_get_colors, sizeof(__pyx_k_ot_color_palette_get_colors), 0, 0, 1, 1},
    {&__pyx_n_s_ot_color_palette_get_count, __pyx_k_ot_color_palette_get_count, sizeof(__pyx_k_ot_color_palette_get_count), 0, 0, 1, 1},
    {&__pyx_n_s_ot_color_palette_get_flags, __pyx_k_ot_color_palette_get_flags, sizeof(__pyx_k_ot_color_palette_get_flags), 0, 0, 1, 1},
    {&__pyx_n_s_ot_color_palette_get_name_id, __pyx_k_ot_color_palette_get_name_id, sizeof(__pyx_k_ot_color_palette_get_name_id), 0, 0, 1, 1},
    {&__pyx_n_s_ot_font_set_funcs, __pyx_k_ot_font_set_funcs, sizeof(__pyx_k_ot_font_set_funcs), 0, 0, 1, 1},
    {&__pyx_n_s_ot_layout_get_baseline, __pyx_k_ot_layout_get_baseline, sizeof(__pyx_k_ot_layout_get_baseline), 0, 0, 1, 1},
    {&__pyx_n_s_ot_layout_get_glyph_class, __pyx_k_ot_layout_get_glyph_class, sizeof(__pyx_k_ot_layout_get_glyph_class), 0, 0, 1, 1},
    {&__pyx_n_s_ot_layout_has_glyph_classes, __pyx_k_ot_layout_has_glyph_classes, sizeof(__pyx_k_ot_layout_has_glyph_classes), 0, 0, 1, 1},
    {&__pyx_n_s_ot_layout_has_positioning, __pyx_k_ot_layout_has_positioning, sizeof(__pyx_k_ot_layout_has_positioning), 0, 0, 1, 1},
    {&__pyx_n_s_ot_layout_has_substitution, __pyx_k_ot_layout_has_substitution, sizeof(__pyx_k_ot_layout_has_substitution), 0, 0, 1, 1},
    {&__pyx_n_s_ot_layout_language_get_feature_t, __pyx_k_ot_layout_language_get_feature_t, sizeof(__pyx_k_ot_layout_language_get_feature_t), 0, 0, 1, 1},
    {&__pyx_n_s_ot_layout_lookup_get_glyph_alter, __pyx_k_ot_layout_lookup_get_glyph_alter, sizeof(__pyx_k_ot_layout_lookup_get_glyph_alter), 0, 0, 1, 1},
    {&__pyx_n_s_ot_layout_script_get_language_ta, __pyx_k_ot_layout_script_get_language_ta, sizeof(__pyx_k_ot_layout_script_get_language_ta), 0, 0, 1, 1},
    {&__pyx_n_s_ot_layout_table_get_script_tags, __pyx_k_ot_layout_table_get_script_tags, sizeof(__pyx_k_ot_layout_table_get_script_tags), 0, 0, 1, 1},
    {&__pyx_n_s_ot_math_get_constant, __pyx_k_ot_math_get_constant, sizeof(__pyx_k_ot_math_get_constant), 0, 0, 1, 1},
    {&__pyx_n_s_ot_math_get_glyph_assembly, __pyx_k_ot_math_get_glyph_assembly, sizeof(__pyx_k_ot_math_get_glyph_assembly), 0, 0, 1, 1},
    {&__pyx_n_s_ot_math_get_glyph_italics_correc, __pyx_k_ot_math_get_glyph_italics_correc, sizeof(__pyx_k_ot_math_get_glyph_italics_correc), 0, 0, 1, 1},
    {&__pyx_n_s_ot_math_get_glyph_kerning, __pyx_k_ot_math_get_glyph_kerning, sizeof(__pyx_k_ot_math_get_glyph_kerning), 0, 0, 1, 1},
    {&__pyx_n_s_ot_math_get_glyph_kernings, __pyx_k_ot_math_get_glyph_kernings, sizeof(__pyx_k_ot_math_get_glyph_kernings), 0, 0, 1, 1},
    {&__pyx_n_s_ot_math_get_glyph_top_accent_att, __pyx_k_ot_math_get_glyph_top_accent_att, sizeof(__pyx_k_ot_math_get_glyph_top_accent_att), 0, 0, 1, 1},
    {&__pyx_n_s_ot_math_get_glyph_variants, __pyx_k_ot_math_get_glyph_variants, sizeof(__pyx_k_ot_math_get_glyph_variants), 0, 0, 1, 1},
    {&__pyx_n_s_ot_math_get_min_connector_overla, __pyx_k_ot_math_get_min_connector_overla, sizeof(__pyx_k_ot_math_get_min_connector_overla), 0, 0, 1, 1},
    {&__pyx_n_s_ot_math_has_data, __pyx_k_ot_math_has_data, sizeof(__pyx_k_ot_math_has_data), 0, 0, 1, 1},
    {&__pyx_n_s_ot_math_is_glyph_extended_shape, __pyx_k_ot_math_is_glyph_extended_shape, sizeof(__pyx_k_ot_math_is_glyph_extended_shape), 0, 0, 1, 1},
    {&__pyx_n_s_ot_tag_to_language, __pyx_k_ot_tag_to_language, sizeof(__pyx_k_ot_tag_to_language), 0, 0, 1, 1},
    {&__pyx_n_s_ot_tag_to_script, __pyx_k_ot_tag_to_script, sizeof(__pyx_k_ot_tag_to_script), 0, 0, 1, 1},
    {&__pyx_n_s_other, __pyx_k_other, sizeof(__pyx_k_other), 0, 0, 1, 1},
    {&__pyx_n_s_p, __pyx_k_p, sizeof(__pyx_k_p), 0, 0, 1, 1},
    {&__pyx_n_s_packed, __pyx_k_packed, sizeof(__pyx_k_packed), 0, 0, 1, 1},
    {&__pyx_n_s_packed_blob, __pyx_k_packed_blob, sizeof(__pyx_k_packed_blob), 0, 0, 1, 1},
    {&__pyx_n_s_paint_funcs, __pyx_k_paint_funcs, sizeof(__pyx_k_paint_funcs), 0, 0, 1, 1},
    {&__pyx_n_s_paint_glyph, __pyx_k_paint_glyph, sizeof(__pyx_k_paint_glyph), 0, 0, 1, 1},
    {&__pyx_n_s_paint_state, __pyx_k_paint_state, sizeof(__pyx_k_paint_state), 0, 0, 1, 1},
    {&__pyx_n_s_paint_state_p, __pyx_k_paint_state_p, sizeof(__pyx_k_paint_state_p), 0, 0, 1, 1},
    {&__pyx_n_s_palette_index, __pyx_k_palette_index, sizeof(__pyx_k_palette_index), 0, 0, 1, 1},
    {&__pyx_n_s_pathlib, __pyx_k_pathlib, sizeof(__pyx_k_pathlib), 0, 0, 1, 1},
    {&__pyx_n_s_pen, __pyx_k_pen, sizeof(__pyx_k_pen), 0, 0, 1, 1},
    {&__pyx_n_s_pin_axis_location, __pyx_k_pin_axis_location, sizeof(__pyx_k_pin_axis_location), 0, 0, 1, 1},
    {&__pyx_n_s_pin_axis_to_default, __pyx_k_pin_axis_to_default, sizeof(__pyx_k_pin_axis_to_default), 0, 0, 1, 1},
    {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
    {&__pyx_n_s_qCurveTo, __pyx_k_qCurveTo, sizeof(__pyx_k_qCurveTo), 0, 0, 1, 1},
    {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
    {&__pyx_n_s_r, __pyx_k_r, sizeof(__pyx_k_r), 0, 0, 1, 1},
    {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
    {&__pyx_n_s_red, __pyx_k_red, sizeof(__pyx_k_red), 0, 0, 1, 1},
    {&__pyx_n_u_red, __pyx_k_red, sizeof(__pyx_k_red), 0, 1, 0, 1},
    {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
    {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
    {&__pyx_n_s_remove, __pyx_k_remove, sizeof(__pyx_k_remove), 0, 0, 1, 1},
    {&__pyx_n_s_repack, __pyx_k_repack, sizeof(__pyx_k_repack), 0, 0, 1, 1},
    {&__pyx_n_s_repack_with_tag, __pyx_k_repack_with_tag, sizeof(__pyx_k_repack_with_tag), 0, 0, 1, 1},
    {&__pyx_n_s_repr___locals_genexpr, __pyx_k_repr___locals_genexpr, sizeof(__pyx_k_repr___locals_genexpr), 0, 0, 1, 1},
    {&__pyx_n_s_reset, __pyx_k_reset, sizeof(__pyx_k_reset), 0, 0, 1, 1},
    {&__pyx_n_s_ret, __pyx_k_ret, sizeof(__pyx_k_ret), 0, 0, 1, 1},
    {&__pyx_n_s_return, __pyx_k_return, sizeof(__pyx_k_return), 0, 0, 1, 1},
    {&__pyx_n_u_romn, __pyx_k_romn, sizeof(__pyx_k_romn), 0, 1, 0, 1},
    {&__pyx_n_s_round, __pyx_k_round, sizeof(__pyx_k_round), 0, 0, 1, 1},
    {&__pyx_n_s_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 0, 1, 1},
    {&__pyx_n_s_script_count, __pyx_k_script_count, sizeof(__pyx_k_script_count), 0, 0, 1, 1},
    {&__pyx_n_s_script_index, __pyx_k_script_index, sizeof(__pyx_k_script_index), 0, 0, 1, 1},
    {&__pyx_n_s_script_tag, __pyx_k_script_tag, sizeof(__pyx_k_script_tag), 0, 0, 1, 1},
    {&__pyx_n_s_script_tags, __pyx_k_script_tags, sizeof(__pyx_k_script_tags), 0, 0, 1, 1},
    {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
    {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
    {&__pyx_n_s_set, __pyx_k_set, sizeof(__pyx_k_set), 0, 0, 1, 1},
    {&__pyx_n_s_set_2, __pyx_k_set_2, sizeof(__pyx_k_set_2), 0, 0, 1, 1},
    {&__pyx_n_s_set_close_path_func, __pyx_k_set_close_path_func, sizeof(__pyx_k_set_close_path_func), 0, 0, 1, 1},
    {&__pyx_n_s_set_color_func, __pyx_k_set_color_func, sizeof(__pyx_k_set_color_func), 0, 0, 1, 1},
    {&__pyx_n_s_set_color_glyph_func, __pyx_k_set_color_glyph_func, sizeof(__pyx_k_set_color_glyph_func), 0, 0, 1, 1},
    {&__pyx_n_s_set_cubic_to_func, __pyx_k_set_cubic_to_func, sizeof(__pyx_k_set_cubic_to_func), 0, 0, 1, 1},
    {&__pyx_n_s_set_custom_palette_color_func, __pyx_k_set_custom_palette_color_func, sizeof(__pyx_k_set_custom_palette_color_func), 0, 0, 1, 1},
    {&__pyx_n_s_set_font_h_extents_func, __pyx_k_set_font_h_extents_func, sizeof(__pyx_k_set_font_h_extents_func), 0, 0, 1, 1},
    {&__pyx_n_s_set_font_v_extents_func, __pyx_k_set_font_v_extents_func, sizeof(__pyx_k_set_font_v_extents_func), 0, 0, 1, 1},
    {&__pyx_n_s_set_glyph_h_advance_func, __pyx_k_set_glyph_h_advance_func, sizeof(__pyx_k_set_glyph_h_advance_func), 0, 0, 1, 1},
    {&__pyx_n_s_set_glyph_name_func, __pyx_k_set_glyph_name_func, sizeof(__pyx_k_set_glyph_name_func), 0, 0, 1, 1},
    {&__pyx_n_s_set_glyph_v_advance_func, __pyx_k_set_glyph_v_advance_func, sizeof(__pyx_k_set_glyph_v_advance_func), 0, 0, 1, 1},
    {&__pyx_n_s_set_glyph_v_origin_func, __pyx_k_set_glyph_v_origin_func, sizeof(__pyx_k_set_glyph_v_origin_func), 0, 0, 1, 1},
    {&__pyx_n_s_set_image_func, __pyx_k_set_image_func, sizeof(__pyx_k_set_image_func), 0, 0, 1, 1},
    {&__pyx_n_s_set_language_from_ot_tag, __pyx_k_set_language_from_ot_tag, sizeof(__pyx_k_set_language_from_ot_tag), 0, 0, 1, 1},
    {&__pyx_n_s_set_line_to_func, __pyx_k_set_line_to_func, sizeof(__pyx_k_set_line_to_func), 0, 0, 1, 1},
    {&__pyx_n_s_set_linear_gradient_func, __pyx_k_set_linear_gradient_func, sizeof(__pyx_k_set_linear_gradient_func), 0, 0, 1, 1},
    {&__pyx_n_s_set_message_func, __pyx_k_set_message_func, sizeof(__pyx_k_set_message_func), 0, 0, 1, 1},
    {&__pyx_n_s_set_move_to_func, __pyx_k_set_move_to_func, sizeof(__pyx_k_set_move_to_func), 0, 0, 1, 1},
    {&__pyx_n_s_set_name, __pyx_k_set_name, sizeof(__pyx_k_set_name), 0, 0, 1, 1},
    {&__pyx_n_s_set_nominal_glyph_func, __pyx_k_set_nominal_glyph_func, sizeof(__pyx_k_set_nominal_glyph_func), 0, 0, 1, 1},
    {&__pyx_n_s_set_pop_clip_func, __pyx_k_set_pop_clip_func, sizeof(__pyx_k_set_pop_clip_func), 0, 0, 1, 1},
    {&__pyx_n_s_set_pop_group_func, __pyx_k_set_pop_group_func, sizeof(__pyx_k_set_pop_group_func), 0, 0, 1, 1},
    {&__pyx_n_s_set_pop_transform_func, __pyx_k_set_pop_transform_func, sizeof(__pyx_k_set_pop_transform_func), 0, 0, 1, 1},
    {&__pyx_n_s_set_push_clip_glyph_func, __pyx_k_set_push_clip_glyph_func, sizeof(__pyx_k_set_push_clip_glyph_func), 0, 0, 1, 1},
    {&__pyx_n_s_set_push_clip_rectangle_func, __pyx_k_set_push_clip_rectangle_func, sizeof(__pyx_k_set_push_clip_rectangle_func), 0, 0, 1, 1},
    {&__pyx_n_s_set_push_group_func, __pyx_k_set_push_group_func, sizeof(__pyx_k_set_push_group_func), 0, 0, 1, 1},
    {&__pyx_n_s_set_push_transform_func, __pyx_k_set_push_transform_func, sizeof(__pyx_k_set_push_transform_func), 0, 0, 1, 1},
    {&__pyx_n_s_set_quadratic_to_func, __pyx_k_set_quadratic_to_func, sizeof(__pyx_k_set_quadratic_to_func), 0, 0, 1, 1},
    {&__pyx_n_s_set_radial_gradient_func, __pyx_k_set_radial_gradient_func, sizeof(__pyx_k_set_radial_gradient_func), 0, 0, 1, 1},
    {&__pyx_n_s_set_script_from_ot_tag, __pyx_k_set_script_from_ot_tag, sizeof(__pyx_k_set_script_from_ot_tag), 0, 0, 1, 1},
    {&__pyx_n_s_set_sweep_gradient_func, __pyx_k_set_sweep_gradient_func, sizeof(__pyx_k_set_sweep_gradient_func), 0, 0, 1, 1},
    {&__pyx_n_s_set_type, __pyx_k_set_type, sizeof(__pyx_k_set_type), 0, 0, 1, 1},
    {&__pyx_n_s_set_var_coords_design, __pyx_k_set_var_coords_design, sizeof(__pyx_k_set_var_coords_design), 0, 0, 1, 1},
    {&__pyx_n_s_set_var_coords_normalized, __pyx_k_set_var_coords_normalized, sizeof(__pyx_k_set_var_coords_normalized), 0, 0, 1, 1},
    {&__pyx_n_s_set_variation, __pyx_k_set_variation, sizeof(__pyx_k_set_variation), 0, 0, 1, 1},
    {&__pyx_n_s_set_variation_glyph_func, __pyx_k_set_variation_glyph_func, sizeof(__pyx_k_set_variation_glyph_func), 0, 0, 1, 1},
    {&__pyx_n_s_set_variations, __pyx_k_set_variations, sizeof(__pyx_k_set_variations), 0, 0, 1, 1},
    {&__pyx_n_s_sets, __pyx_k_sets, sizeof(__pyx_k_sets), 0, 0, 1, 1},
    {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
    {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_shape, __pyx_k_shape, sizeof(__pyx_k_shape), 0, 0, 1, 1},
    {&__pyx_n_s_shaper, __pyx_k_shaper, sizeof(__pyx_k_shaper), 0, 0, 1, 1},
    {&__pyx_n_s_shapers, __pyx_k_shapers, sizeof(__pyx_k_shapers), 0, 0, 1, 1},
    {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
    {&__pyx_n_s_smaller_set, __pyx_k_smaller_set, sizeof(__pyx_k_smaller_set), 0, 0, 1, 1},
    {&__pyx_n_s_source, __pyx_k_source, sizeof(__pyx_k_source), 0, 0, 1, 1},
    {&__pyx_n_s_spec, __pyx_k_spec, sizeof(__pyx_k_spec), 0, 0, 1, 1},
    {&__pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_k_src_uharfbuzz__harfbuzz_pyx, sizeof(__pyx_k_src_uharfbuzz__harfbuzz_pyx), 0, 0, 1, 0},
    {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
    {&__pyx_n_u_start_connector_length, __pyx_k_start_connector_length, sizeof(__pyx_k_start_connector_length), 0, 1, 0, 1},
    {&__pyx_n_s_start_offset, __pyx_k_start_offset, sizeof(__pyx_k_start_offset), 0, 0, 1, 1},
    {&__pyx_n_s_staticmethod, __pyx_k_staticmethod, sizeof(__pyx_k_staticmethod), 0, 0, 1, 1},
    {&__pyx_n_s_str, __pyx_k_str, sizeof(__pyx_k_str), 0, 0, 1, 1},
    {&__pyx_n_s_string, __pyx_k_string, sizeof(__pyx_k_string), 0, 0, 1, 1},
    {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
    {&__pyx_n_s_subset, __pyx_k_subset, sizeof(__pyx_k_subset), 0, 0, 1, 1},
    {&__pyx_n_s_subset_preprocess, __pyx_k_subset_preprocess, sizeof(__pyx_k_subset_preprocess), 0, 0, 1, 1},
    {&__pyx_n_s_subtables, __pyx_k_subtables, sizeof(__pyx_k_subtables), 0, 0, 1, 1},
    {&__pyx_n_s_success, __pyx_k_success, sizeof(__pyx_k_success), 0, 0, 1, 1},
    {&__pyx_n_s_super, __pyx_k_super, sizeof(__pyx_k_super), 0, 0, 1, 1},
    {&__pyx_n_s_symmetric_difference_update, __pyx_k_symmetric_difference_update, sizeof(__pyx_k_symmetric_difference_update), 0, 0, 1, 1},
    {&__pyx_n_s_table_bytes, __pyx_k_table_bytes, sizeof(__pyx_k_table_bytes), 0, 0, 1, 1},
    {&__pyx_n_s_table_data, __pyx_k_table_data, sizeof(__pyx_k_table_data), 0, 0, 1, 1},
    {&__pyx_n_s_tag, __pyx_k_tag, sizeof(__pyx_k_tag), 0, 0, 1, 1},
    {&__pyx_n_s_tag_packed, __pyx_k_tag_packed, sizeof(__pyx_k_tag_packed), 0, 0, 1, 1},
    {&__pyx_n_s_tags, __pyx_k_tags, sizeof(__pyx_k_tags), 0, 0, 1, 1},
    {&__pyx_n_s_tail, __pyx_k_tail, sizeof(__pyx_k_tail), 0, 0, 1, 1},
    {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
    {&__pyx_n_s_text, __pyx_k_text, sizeof(__pyx_k_text), 0, 0, 1, 1},
    {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
    {&__pyx_n_s_to_int, __pyx_k_to_int, sizeof(__pyx_k_to_int), 0, 0, 1, 1},
    {&__pyx_n_s_typing, __pyx_k_typing, sizeof(__pyx_k_typing), 0, 0, 1, 1},
    {&__pyx_n_s_uharfbuzz__harfbuzz, __pyx_k_uharfbuzz__harfbuzz, sizeof(__pyx_k_uharfbuzz__harfbuzz), 0, 0, 1, 1},
    {&__pyx_n_s_unicode, __pyx_k_unicode, sizeof(__pyx_k_unicode), 0, 0, 1, 1},
    {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
    {&__pyx_n_s_update_2, __pyx_k_update_2, sizeof(__pyx_k_update_2), 0, 0, 1, 1},
    {&__pyx_n_s_user_data, __pyx_k_user_data, sizeof(__pyx_k_user_data), 0, 0, 1, 1},
    {&__pyx_n_s_user_data_p, __pyx_k_user_data_p, sizeof(__pyx_k_user_data_p), 0, 0, 1, 1},
    {&__pyx_n_s_v, __pyx_k_v, sizeof(__pyx_k_v), 0, 0, 1, 1},
    {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
    {&__pyx_n_s_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
    {&__pyx_n_s_values_locals_genexpr, __pyx_k_values_locals_genexpr, sizeof(__pyx_k_values_locals_genexpr), 0, 0, 1, 1},
    {&__pyx_n_s_variants, __pyx_k_variants, sizeof(__pyx_k_variants), 0, 0, 1, 1},
    {&__pyx_n_s_variants_array, __pyx_k_variants_array, sizeof(__pyx_k_variants_array), 0, 0, 1, 1},
    {&__pyx_n_s_variation, __pyx_k_variation, sizeof(__pyx_k_variation), 0, 0, 1, 1},
    {&__pyx_n_s_variation_selector, __pyx_k_variation_selector, sizeof(__pyx_k_variation_selector), 0, 0, 1, 1},
    {&__pyx_n_s_variation_unicodes, __pyx_k_variation_unicodes, sizeof(__pyx_k_variation_unicodes), 0, 0, 1, 1},
    {&__pyx_n_s_variations, __pyx_k_variations, sizeof(__pyx_k_variations), 0, 0, 1, 1},
    {&__pyx_n_s_version_string, __pyx_k_version_string, sizeof(__pyx_k_version_string), 0, 0, 1, 1},
    {&__pyx_n_s_warn, __pyx_k_warn, sizeof(__pyx_k_warn), 0, 0, 1, 1},
    {&__pyx_n_s_warnings, __pyx_k_warnings, sizeof(__pyx_k_warnings), 0, 0, 1, 1},
    {&__pyx_n_u_width, __pyx_k_width, sizeof(__pyx_k_width), 0, 1, 0, 1},
    {&__pyx_n_s_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 0, 1, 1},
    {&__pyx_n_u_x_bearing, __pyx_k_x_bearing, sizeof(__pyx_k_x_bearing), 0, 1, 0, 1},
    {&__pyx_n_s_y, __pyx_k_y, sizeof(__pyx_k_y), 0, 0, 1, 1},
    {&__pyx_n_u_y_bearing, __pyx_k_y_bearing, sizeof(__pyx_k_y_bearing), 0, 1, 0, 1},
    {0, 0, 0, 0, 0, 0, 0}
  };
  return __Pyx_InitStrings(__pyx_string_tab);
}
/* #### Code section: cached_builtins ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_staticmethod = __Pyx_GetBuiltinName(__pyx_n_s_staticmethod); if (!__pyx_builtin_staticmethod) __PYX_ERR(0, 1666, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(0, 141, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 184, __pyx_L1_error)
  __pyx_builtin_AssertionError = __Pyx_GetBuiltinName(__pyx_n_s_AssertionError); if (!__pyx_builtin_AssertionError) __PYX_ERR(0, 333, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 708, __pyx_L1_error)
  __pyx_builtin_round = __Pyx_GetBuiltinName(__pyx_n_s_round); if (!__pyx_builtin_round) __PYX_ERR(0, 810, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(0, 1201, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 1342, __pyx_L1_error)
  __pyx_builtin_DeprecationWarning = __Pyx_GetBuiltinName(__pyx_n_s_DeprecationWarning); if (!__pyx_builtin_DeprecationWarning) __PYX_ERR(0, 2143, __pyx_L1_error)
  __pyx_builtin_KeyError = __Pyx_GetBuiltinName(__pyx_n_s_KeyError); if (!__pyx_builtin_KeyError) __PYX_ERR(0, 2638, __pyx_L1_error)
  __pyx_builtin_NotImplemented = __Pyx_GetBuiltinName(__pyx_n_s_NotImplemented); if (!__pyx_builtin_NotImplemented) __PYX_ERR(0, 2658, __pyx_L1_error)
  __pyx_builtin_StopIteration = __Pyx_GetBuiltinName(__pyx_n_s_StopIteration); if (!__pyx_builtin_StopIteration) __PYX_ERR(0, 2760, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cached_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "uharfbuzz/_harfbuzz.pyx":405
 *         if blob == NULL:
 *             raise HarfBuzzError(f"Failed to open: {filename}")
 *         cdef Blob inst = cls(None)             # <<<<<<<<<<<<<<
 *         inst._hb_blob = blob
 *         return inst
 */
  __pyx_tuple__7 = PyTuple_Pack(1, Py_None); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "uharfbuzz/_harfbuzz.pyx":1201
 *             ret = hb_shape_full(font._hb_font, buffer._hb_buffer, hb_features, size, c_shapers)
 *             if not ret:
 *                 raise RuntimeError("All shapers failed")             # <<<<<<<<<<<<<<
 *         else:
 *             hb_shape(font._hb_font, buffer._hb_buffer, hb_features, size)
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_u_All_shapers_failed); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 1201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "uharfbuzz/_harfbuzz.pyx":1524
 * def ot_math_get_constant(font: Font, constant: OTMathConstant) -> int:
 *     if constant >= len(OTMathConstant):
 *         raise ValueError("invalid constant")             # <<<<<<<<<<<<<<
 *     return hb_ot_math_get_constant(font._hb_font, constant)
 * 
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_u_invalid_constant); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(0, 1524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "uharfbuzz/_harfbuzz.pyx":1553
 *                               int correction_height) -> int:
 *     if kern >= len(OTMathKern):
 *         raise ValueError("invalid kern")             # <<<<<<<<<<<<<<
 *     return hb_ot_math_get_glyph_kerning(font._hb_font, glyph, kern, correction_height)
 * 
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_u_invalid_kern); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 1553, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "uharfbuzz/_harfbuzz.pyx":2141
 *     def get_glyph_shape(self, font: Font, gid: int):
 *         if not self._warned:
 *             warnings.warn(             # <<<<<<<<<<<<<<
 *                 "get_glyph_shape() is deprecated, use Font.draw_glyph() instead",
 *                 DeprecationWarning,
 */
  __pyx_tuple__13 = PyTuple_Pack(2, __pyx_kp_u_get_glyph_shape_is_deprecated_us, __pyx_builtin_DeprecationWarning); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(0, 2141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "uharfbuzz/_harfbuzz.pyx":2150
 *     def draw_glyph(self, font: Font, gid: int, draw_data: object = None):
 *         if not self._warned:
 *             warnings.warn(             # <<<<<<<<<<<<<<
 *                 "draw_glyph() is deprecated, use Font.draw_glyph() instead",
 *                 DeprecationWarning,
 */
  __pyx_tuple__14 = PyTuple_Pack(2, __pyx_kp_u_draw_glyph_is_deprecated_use_Fon, __pyx_builtin_DeprecationWarning); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 2150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "uharfbuzz/_harfbuzz.pyx":2442
 *     new_face = hb_subset_or_fail(face._hb_face, input._hb_input)
 *     if new_face == NULL:
 *         raise RuntimeError("Subsetting failed")             # <<<<<<<<<<<<<<
 *     return Face.from_ptr(new_face)
 * 
 */
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_kp_u_Subsetting_failed); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 2442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "uharfbuzz/_harfbuzz.pyx":41
 * 
 * 
 * def version_string() -> str:             # <<<<<<<<<<<<<<
 *     cdef const char* cstr = hb_version_string()
 *     cdef bytes packed = cstr
 */
  __pyx_tuple__24 = PyTuple_Pack(2, __pyx_n_s_cstr, __pyx_n_s_packed); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);
  __pyx_codeobj__25 = (PyObject*)__Pyx_PyCode_New(0, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_version_string, 41, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__25)) __PYX_ERR(0, 41, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self._hb_glyph_info must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_tuple__26 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);
  __pyx_codeobj__27 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__27)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self._hb_glyph_info must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self._hb_glyph_info must be explicitly requested with @auto_pickle(True)"
 */
  __pyx_tuple__28 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_pyx_state); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);
  __pyx_codeobj__29 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__29)) __PYX_ERR(1, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self._hb_glyph_position must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__30 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__30)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self._hb_glyph_position must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self._hb_glyph_position must be explicitly requested with @auto_pickle(True)"
 */
  __pyx_codeobj__31 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__31)) __PYX_ERR(1, 3, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":148
 * 
 *     # DEPRECATED: use the normal constructor
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def create(cls):
 *         cdef Buffer inst = cls()
 */
  __pyx_tuple__32 = PyTuple_Pack(2, __pyx_n_s_cls, __pyx_n_s_inst); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);
  __pyx_codeobj__33 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__32, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_create, 148, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__33)) __PYX_ERR(0, 148, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":156
 *         return hb_buffer_get_length(self._hb_buffer)
 * 
 *     def reset(self):             # <<<<<<<<<<<<<<
 *         hb_buffer_reset (self._hb_buffer)
 * 
 */
  __pyx_codeobj__34 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_reset, 156, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__34)) __PYX_ERR(0, 156, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":159
 *         hb_buffer_reset (self._hb_buffer)
 * 
 *     def clear_contents(self):             # <<<<<<<<<<<<<<
 *         hb_buffer_clear_contents(self._hb_buffer)
 * 
 */
  __pyx_codeobj__35 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_clear_contents, 159, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__35)) __PYX_ERR(0, 159, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":295
 *         hb_buffer_set_not_found_glyph(self._hb_buffer, value)
 * 
 *     def set_language_from_ot_tag(self, value: str):             # <<<<<<<<<<<<<<
 *         cdef bytes packed = value.encode()
 *         cdef char* cstr = packed
 */
  __pyx_tuple__36 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_value, __pyx_n_s_packed, __pyx_n_s_cstr); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);
  __pyx_codeobj__37 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__36, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_language_from_ot_tag, 295, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__37)) __PYX_ERR(0, 295, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":301
 *             self._hb_buffer, hb_ot_tag_to_language(hb_tag_from_string(cstr, -1)))
 * 
 *     def set_script_from_ot_tag(self, value: str):             # <<<<<<<<<<<<<<
 *         cdef bytes packed = value.encode()
 *         cdef char* cstr = packed
 */
  __pyx_codeobj__38 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__36, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_script_from_ot_tag, 301, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__38)) __PYX_ERR(0, 301, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":307
 *             self._hb_buffer, hb_ot_tag_to_script(hb_tag_from_string(cstr, -1)))
 * 
 *     def add_codepoints(self, codepoints: List[int],             # <<<<<<<<<<<<<<
 *                        item_offset: int = 0, item_length: int = -1) -> None:
 *         cdef unsigned int size = len(codepoints)
 */
  __pyx_tuple__39 = PyTuple_Pack(7, __pyx_n_s_self, __pyx_n_s_codepoints, __pyx_n_s_item_offset, __pyx_n_s_item_length, __pyx_n_s_size, __pyx_n_s_hb_codepoints, __pyx_n_s_i); if (unlikely(!__pyx_tuple__39)) __PYX_ERR(0, 307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__39);
  __Pyx_GIVEREF(__pyx_tuple__39);
  __pyx_codeobj__40 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_add_codepoints, 307, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__40)) __PYX_ERR(0, 307, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":323
 *             raise MemoryError()
 * 
 *     def add_utf8(self, text: bytes,             # <<<<<<<<<<<<<<
 *                  item_offset: int = 0, item_length: int = -1) -> None:
 *         hb_buffer_add_utf8(
 */
  __pyx_tuple__41 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_text, __pyx_n_s_item_offset, __pyx_n_s_item_length); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(0, 323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__41);
  __Pyx_GIVEREF(__pyx_tuple__41);
  __pyx_codeobj__42 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_add_utf8, 323, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__42)) __PYX_ERR(0, 323, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":330
 *             raise MemoryError()
 * 
 *     def add_str(self, text: str,             # <<<<<<<<<<<<<<
 *                 item_offset: int = 0, item_length: int = -1) -> None:
 *         # ensure unicode string is in the "canonical" representation
 */
  __pyx_tuple__43 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_text, __pyx_n_s_item_offset, __pyx_n_s_item_length, __pyx_n_s_length, __pyx_n_s_kind); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(0, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__43);
  __Pyx_GIVEREF(__pyx_tuple__43);
  __pyx_codeobj__44 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__43, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_add_str, 330, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__44)) __PYX_ERR(0, 330, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":367
 *             raise MemoryError()
 * 
 *     def guess_segment_properties(self) -> None:             # <<<<<<<<<<<<<<
 *         hb_buffer_guess_segment_properties(self._hb_buffer)
 * 
 */
  __pyx_codeobj__45 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_guess_segment_properties, 367, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__45)) __PYX_ERR(0, 367, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":370
 *         hb_buffer_guess_segment_properties(self._hb_buffer)
 * 
 *     def set_message_func(self, callback) -> None:             # <<<<<<<<<<<<<<
 *         self._message_callback = callback
 *         hb_buffer_set_message_func(self._hb_buffer, msgcallback, <void*>callback, NULL)
 */
  __pyx_tuple__46 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_callback); if (unlikely(!__pyx_tuple__46)) __PYX_ERR(0, 370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__46);
  __Pyx_GIVEREF(__pyx_tuple__46);
  __pyx_codeobj__47 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_message_func, 370, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__47)) __PYX_ERR(0, 370, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__48 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__48)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__49 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__49)) __PYX_ERR(1, 3, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":399
 *         return wrapper
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def from_file_path(cls, filename: Union[str, Path]):
 *         cdef bytes packed = os.fsencode(filename)
 */
  __pyx_tuple__50 = PyTuple_Pack(5, __pyx_n_s_cls, __pyx_n_s_filename, __pyx_n_s_packed, __pyx_n_s_blob, __pyx_n_s_inst); if (unlikely(!__pyx_tuple__50)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__50);
  __Pyx_GIVEREF(__pyx_tuple__50);
  __pyx_codeobj__51 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__50, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_from_file_path, 399, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__51)) __PYX_ERR(0, 399, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__52 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__52)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__53 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__53)) __PYX_ERR(1, 3, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":473
 * 
 *     # DEPRECATED: use the normal constructor
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def create(cls, bytes blob, int index=0):
 *         cdef Face inst = cls(blob, index)
 */
  __pyx_tuple__54 = PyTuple_Pack(4, __pyx_n_s_cls, __pyx_n_s_blob, __pyx_n_s_index, __pyx_n_s_inst); if (unlikely(!__pyx_tuple__54)) __PYX_ERR(0, 473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__54);
  __Pyx_GIVEREF(__pyx_tuple__54);
  __pyx_codeobj__55 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__54, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_create, 473, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__55)) __PYX_ERR(0, 473, __pyx_L1_error)
  __pyx_tuple__56 = PyTuple_Pack(1, __pyx_int_0); if (unlikely(!__pyx_tuple__56)) __PYX_ERR(0, 473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__56);
  __Pyx_GIVEREF(__pyx_tuple__56);

  /* "uharfbuzz/_harfbuzz.pyx":478
 *         return inst
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def create_for_tables(cls,
 *                           func: Callable[[
 */
  __pyx_tuple__57 = PyTuple_Pack(4, __pyx_n_s_cls, __pyx_n_s_func, __pyx_n_s_user_data, __pyx_n_s_inst); if (unlikely(!__pyx_tuple__57)) __PYX_ERR(0, 478, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__57);
  __Pyx_GIVEREF(__pyx_tuple__57);
  __pyx_codeobj__58 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__57, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_create_for_tables, 478, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__58)) __PYX_ERR(0, 478, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":555
 *         return s
 * 
 *     def variation_unicodes(self, variation_selector):             # <<<<<<<<<<<<<<
 *         s = Set()
 *         hb_face_collect_variation_unicodes(self._hb_face, variation_selector, s._hb_set)
 */
  __pyx_tuple__59 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_variation_selector, __pyx_n_s_s); if (unlikely(!__pyx_tuple__59)) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__59);
  __Pyx_GIVEREF(__pyx_tuple__59);
  __pyx_codeobj__60 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__59, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_variation_unicodes, 555, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__60)) __PYX_ERR(0, 555, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__61 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__61)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__62 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__62)) __PYX_ERR(1, 3, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":608
 * 
 *     # DEPRECATED: use the normal constructor
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def create(cls, face: Face):
 *         cdef Font inst = cls(face)
 */
  __pyx_tuple__63 = PyTuple_Pack(3, __pyx_n_s_cls, __pyx_n_s_face, __pyx_n_s_inst); if (unlikely(!__pyx_tuple__63)) __PYX_ERR(0, 608, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__63);
  __Pyx_GIVEREF(__pyx_tuple__63);
  __pyx_codeobj__64 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__63, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_create, 608, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__64)) __PYX_ERR(0, 608, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":697
 *         hb_font_set_var_named_instance(self._hb_font, value)
 * 
 *     def set_variations(self, variations: Dict[str, float]) -> None:             # <<<<<<<<<<<<<<
 *         cdef unsigned int size
 *         cdef hb_variation_t* hb_variations
 */
  __pyx_tuple__65 = PyTuple_Pack(9, __pyx_n_s_self, __pyx_n_s_variations, __pyx_n_s_size, __pyx_n_s_hb_variations, __pyx_n_s_packed, __pyx_n_s_variation, __pyx_n_s_i, __pyx_n_s_name, __pyx_n_s_value); if (unlikely(!__pyx_tuple__65)) __PYX_ERR(0, 697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__65);
  __Pyx_GIVEREF(__pyx_tuple__65);
  __pyx_codeobj__66 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__65, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_variations, 697, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__66)) __PYX_ERR(0, 697, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":717
 *             free(hb_variations)
 * 
 *     def set_variation(self, name: str, value: float) -> None:             # <<<<<<<<<<<<<<
 *         packed = name.encode()
 *         cdef hb_tag_t tag = hb_tag_from_string(packed, -1)
 */
  __pyx_tuple__67 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_name, __pyx_n_s_value, __pyx_n_s_packed, __pyx_n_s_tag); if (unlikely(!__pyx_tuple__67)) __PYX_ERR(0, 717, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__67);
  __Pyx_GIVEREF(__pyx_tuple__67);
  __pyx_codeobj__68 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__67, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_variation, 717, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__68)) __PYX_ERR(0, 717, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":722
 *         hb_font_set_variation(self._hb_font, tag, value)
 * 
 *     def get_glyph_name(self, gid: int):             # <<<<<<<<<<<<<<
 *         cdef char name[64]
 *         cdef bytes packed
 */
  __pyx_tuple__69 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_gid, __pyx_n_s_name, __pyx_n_s_packed, __pyx_n_s_success); if (unlikely(!__pyx_tuple__69)) __PYX_ERR(0, 722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__69);
  __Pyx_GIVEREF(__pyx_tuple__69);
  __pyx_codeobj__70 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__69, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_get_glyph_name, 722, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__70)) __PYX_ERR(0, 722, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":732
 *             return None
 * 
 *     def get_glyph_from_name(self, name: str):             # <<<<<<<<<<<<<<
 *         cdef hb_codepoint_t gid
 *         cdef bytes packed
 */
  __pyx_tuple__71 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_name, __pyx_n_s_gid, __pyx_n_s_packed, __pyx_n_s_success); if (unlikely(!__pyx_tuple__71)) __PYX_ERR(0, 732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__71);
  __Pyx_GIVEREF(__pyx_tuple__71);
  __pyx_codeobj__72 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__71, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_get_glyph_from_name, 732, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__72)) __PYX_ERR(0, 732, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":739
 *         return gid if success else None
 * 
 *     def get_glyph_extents(self, gid: int):             # <<<<<<<<<<<<<<
 *         cdef hb_glyph_extents_t extents
 *         success = hb_font_get_glyph_extents(self._hb_font, gid, &extents)
 */
  __pyx_tuple__73 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_gid, __pyx_n_s_extents, __pyx_n_s_success); if (unlikely(!__pyx_tuple__73)) __PYX_ERR(0, 739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__73);
  __Pyx_GIVEREF(__pyx_tuple__73);
  __pyx_codeobj__74 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__73, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_get_glyph_extents, 739, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__74)) __PYX_ERR(0, 739, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":752
 *             return None
 * 
 *     def get_glyph_h_advance(self, gid: int):             # <<<<<<<<<<<<<<
 *         return hb_font_get_glyph_h_advance(self._hb_font, gid)
 * 
 */
  __pyx_tuple__75 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_gid); if (unlikely(!__pyx_tuple__75)) __PYX_ERR(0, 752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__75);
  __Pyx_GIVEREF(__pyx_tuple__75);
  __pyx_codeobj__76 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__75, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_get_glyph_h_advance, 752, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__76)) __PYX_ERR(0, 752, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":755
 *         return hb_font_get_glyph_h_advance(self._hb_font, gid)
 * 
 *     def get_glyph_v_advance(self, gid: int):             # <<<<<<<<<<<<<<
 *         return hb_font_get_glyph_v_advance(self._hb_font, gid)
 * 
 */
  __pyx_codeobj__77 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__75, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_get_glyph_v_advance, 755, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__77)) __PYX_ERR(0, 755, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":758
 *         return hb_font_get_glyph_v_advance(self._hb_font, gid)
 * 
 *     def get_glyph_h_origin(self, gid: int):             # <<<<<<<<<<<<<<
 *         cdef hb_position_t x, y
 *         success = hb_font_get_glyph_h_origin(self._hb_font, gid, &x, &y)
 */
  __pyx_tuple__78 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_gid, __pyx_n_s_x, __pyx_n_s_y, __pyx_n_s_success); if (unlikely(!__pyx_tuple__78)) __PYX_ERR(0, 758, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__78);
  __Pyx_GIVEREF(__pyx_tuple__78);
  __pyx_codeobj__79 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__78, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_get_glyph_h_origin, 758, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__79)) __PYX_ERR(0, 758, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":763
 *         return (x, y) if success else None
 * 
 *     def get_glyph_v_origin(self, gid: int):             # <<<<<<<<<<<<<<
 *         cdef hb_position_t x, y
 *         success = hb_font_get_glyph_v_origin(self._hb_font, gid, &x, &y)
 */
  __pyx_codeobj__80 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__78, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_get_glyph_v_origin, 763, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__80)) __PYX_ERR(0, 763, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":768
 *         return (x, y) if success else None
 * 
 *     def get_font_extents(self, direction: str):             # <<<<<<<<<<<<<<
 *         cdef hb_font_extents_t extents
 *         cdef hb_direction_t hb_direction
 */
  __pyx_tuple__81 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_direction, __pyx_n_s_extents, __pyx_n_s_hb_direction, __pyx_n_s_packed); if (unlikely(!__pyx_tuple__81)) __PYX_ERR(0, 768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__81);
  __Pyx_GIVEREF(__pyx_tuple__81);
  __pyx_codeobj__82 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__81, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_get_font_extents, 768, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__82)) __PYX_ERR(0, 768, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":783
 *         )
 * 
 *     def get_variation_glyph(self, unicode: int, variation_selector: int):             # <<<<<<<<<<<<<<
 *         cdef hb_codepoint_t gid
 *         success = hb_font_get_variation_glyph(self._hb_font, unicode, variation_selector, &gid)
 */
  __pyx_tuple__83 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_unicode, __pyx_n_s_variation_selector, __pyx_n_s_gid, __pyx_n_s_success); if (unlikely(!__pyx_tuple__83)) __PYX_ERR(0, 783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__83);
  __Pyx_GIVEREF(__pyx_tuple__83);
  __pyx_codeobj__84 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__83, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_get_variation_glyph, 783, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__84)) __PYX_ERR(0, 783, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":788
 *         return gid if success else None
 * 
 *     def get_nominal_glyph(self, unicode: int):             # <<<<<<<<<<<<<<
 *         cdef hb_codepoint_t gid
 *         success = hb_font_get_nominal_glyph(self._hb_font, unicode, &gid)
 */
  __pyx_tuple__85 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_unicode, __pyx_n_s_gid, __pyx_n_s_success); if (unlikely(!__pyx_tuple__85)) __PYX_ERR(0, 788, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__85);
  __Pyx_GIVEREF(__pyx_tuple__85);
  __pyx_codeobj__86 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__85, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_get_nominal_glyph, 788, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__86)) __PYX_ERR(0, 788, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":793
 *         return gid if success else None
 * 
 *     def get_var_coords_normalized(self):             # <<<<<<<<<<<<<<
 *         cdef unsigned int length
 *         cdef const int *coords
 */
  __pyx_tuple__87 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_length, __pyx_n_s_coords, __pyx_n_s_i); if (unlikely(!__pyx_tuple__87)) __PYX_ERR(0, 793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__87);
  __Pyx_GIVEREF(__pyx_tuple__87);
  __pyx_codeobj__88 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__87, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_get_var_coords_normalized, 793, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__88)) __PYX_ERR(0, 793, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":800
 *         return [coords[i] / 0x4000 for i in range(length)]
 * 
 *     def set_var_coords_normalized(self, coords):             # <<<<<<<<<<<<<<
 *         cdef unsigned int length
 *         cdef int *coords_2dot14
 */
  __pyx_tuple__89 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_coords, __pyx_n_s_length, __pyx_n_s_coords_2dot14, __pyx_n_s_i); if (unlikely(!__pyx_tuple__89)) __PYX_ERR(0, 800, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__89);
  __Pyx_GIVEREF(__pyx_tuple__89);
  __pyx_codeobj__90 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__89, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_var_coords_normalized, 800, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__90)) __PYX_ERR(0, 800, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":815
 *             free(coords_2dot14)
 * 
 *     def get_var_coords_design(self):             # <<<<<<<<<<<<<<
 *         cdef unsigned int length
 *         cdef const float *coords
 */
  __pyx_codeobj__91 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__87, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_get_var_coords_design, 815, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__91)) __PYX_ERR(0, 815, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":821
 *         return [coords[i] for i in range(length)]
 * 
 *     def set_var_coords_design(self, coords):             # <<<<<<<<<<<<<<
 *         cdef unsigned int length
 *         cdef cython.float *c_coords
 */
  __pyx_tuple__92 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_coords, __pyx_n_s_length, __pyx_n_s_c_coords, __pyx_n_s_i); if (unlikely(!__pyx_tuple__92)) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__92);
  __Pyx_GIVEREF(__pyx_tuple__92);
  __pyx_codeobj__93 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__92, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_var_coords_design, 821, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__93)) __PYX_ERR(0, 821, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":835
 *             free(c_coords)
 * 
 *     def glyph_to_string(self, gid: int):             # <<<<<<<<<<<<<<
 *         cdef char name[64]
 *         cdef bytes packed
 */
  __pyx_tuple__94 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_gid, __pyx_n_s_name, __pyx_n_s_packed); if (unlikely(!__pyx_tuple__94)) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__94);
  __Pyx_GIVEREF(__pyx_tuple__94);
  __pyx_codeobj__95 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__94, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_glyph_to_string, 835, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__95)) __PYX_ERR(0, 835, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":842
 *         return packed.decode()
 * 
 *     def glyph_from_string(self, string: str):             # <<<<<<<<<<<<<<
 *         cdef hb_codepoint_t gid
 *         cdef bytes packed
 */
  __pyx_tuple__96 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_string, __pyx_n_s_gid, __pyx_n_s_packed, __pyx_n_s_success); if (unlikely(!__pyx_tuple__96)) __PYX_ERR(0, 842, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__96);
  __Pyx_GIVEREF(__pyx_tuple__96);
  __pyx_codeobj__97 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__96, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_glyph_from_string, 842, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__97)) __PYX_ERR(0, 842, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":849
 *         return gid if success else None
 * 
 *     def draw_glyph(self, gid: int, draw_funcs: DrawFuncs, draw_state: object = None):             # <<<<<<<<<<<<<<
 *         cdef void *draw_state_p = <void *>draw_state
 *         if PyCapsule_IsValid(draw_state, NULL):
 */
  __pyx_tuple__98 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_gid, __pyx_n_s_draw_funcs, __pyx_n_s_draw_state, __pyx_n_s_draw_state_p); if (unlikely(!__pyx_tuple__98)) __PYX_ERR(0, 849, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__98);
  __Pyx_GIVEREF(__pyx_tuple__98);
  __pyx_codeobj__99 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__98, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_draw_glyph, 849, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__99)) __PYX_ERR(0, 849, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":855
 *         hb_font_draw_glyph(self._hb_font, gid, draw_funcs._hb_drawfuncs, draw_state_p);
 * 
 *     def paint_glyph(self, gid: int,             # <<<<<<<<<<<<<<
 *                     paint_funcs: PaintFuncs,
 *                     paint_state: object = None,
 */
  __pyx_tuple__100 = PyTuple_Pack(8, __pyx_n_s_self, __pyx_n_s_gid, __pyx_n_s_paint_funcs, __pyx_n_s_paint_state, __pyx_n_s_palette_index, __pyx_n_s_foreground, __pyx_n_s_paint_state_p, __pyx_n_s_c_foreground); if (unlikely(!__pyx_tuple__100)) __PYX_ERR(0, 855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__100);
  __Pyx_GIVEREF(__pyx_tuple__100);
  __pyx_codeobj__101 = (PyObject*)__Pyx_PyCode_New(6, 0, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__100, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_paint_glyph, 855, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__101)) __PYX_ERR(0, 855, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":871
 *                             c_foreground)
 * 
 *     def draw_glyph_with_pen(self, gid: int, pen):             # <<<<<<<<<<<<<<
 *         global drawfuncs
 *         if drawfuncs == NULL:
 */
  __pyx_tuple__102 = PyTuple_Pack(9, __pyx_n_s_self, __pyx_n_s_gid, __pyx_n_s_pen, __pyx_n_s_moveTo, __pyx_n_s_lineTo, __pyx_n_s_curveTo, __pyx_n_s_qCurveTo, __pyx_n_s_closePath, __pyx_n_s_methods); if (unlikely(!__pyx_tuple__102)) __PYX_ERR(0, 871, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__102);
  __Pyx_GIVEREF(__pyx_tuple__102);
  __pyx_codeobj__103 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 9, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__102, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_draw_glyph_with_pen, 871, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__103)) __PYX_ERR(0, 871, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__104 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__104)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__105 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__105)) __PYX_ERR(1, 3, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1067
 * 
 *     # DEPRECATED: use the normal constructor
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def create(cls):
 *         cdef FontFuncs inst = cls()
 */
  __pyx_codeobj__106 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__32, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_create, 1067, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__106)) __PYX_ERR(0, 1067, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1072
 *         return inst
 * 
 *     def set_glyph_h_advance_func(self,             # <<<<<<<<<<<<<<
 *                                  func: Callable[[
 *                                      Font,
 */
  __pyx_tuple__107 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_func, __pyx_n_s_user_data); if (unlikely(!__pyx_tuple__107)) __PYX_ERR(0, 1072, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__107);
  __Pyx_GIVEREF(__pyx_tuple__107);
  __pyx_codeobj__108 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__107, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_glyph_h_advance_func, 1072, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__108)) __PYX_ERR(0, 1072, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1083
 *         self._glyph_h_advance_func = func
 * 
 *     def set_glyph_v_advance_func(self,             # <<<<<<<<<<<<<<
 *                                  func: Callable[[
 *                                      Font,
 */
  __pyx_codeobj__109 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__107, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_glyph_v_advance_func, 1083, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__109)) __PYX_ERR(0, 1083, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1094
 *         self._glyph_v_advance_func = func
 * 
 *     def set_glyph_v_origin_func(self,             # <<<<<<<<<<<<<<
 *                                 func: Callable[[
 *                                     Font,
 */
  __pyx_codeobj__110 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__107, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_glyph_v_origin_func, 1094, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__110)) __PYX_ERR(0, 1094, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1105
 *         self._glyph_v_origin_func = func
 * 
 *     def set_glyph_name_func(self,             # <<<<<<<<<<<<<<
 *                             func: Callable[[
 *                                 Font,
 */
  __pyx_codeobj__111 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__107, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_glyph_name_func, 1105, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__111)) __PYX_ERR(0, 1105, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1116
 *         self._glyph_name_func = func
 * 
 *     def set_nominal_glyph_func(self,             # <<<<<<<<<<<<<<
 *                                func: Callable[[
 *                                    Font,
 */
  __pyx_codeobj__112 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__107, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_nominal_glyph_func, 1116, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__112)) __PYX_ERR(0, 1116, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1127
 *         self._nominal_glyph_func = func
 * 
 *     def set_variation_glyph_func(self,             # <<<<<<<<<<<<<<
 *                                func: Callable[[
 *                                    Font,
 */
  __pyx_codeobj__113 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__107, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_variation_glyph_func, 1127, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__113)) __PYX_ERR(0, 1127, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1139
 *         self._variation_glyph_func = func
 * 
 *     def set_font_h_extents_func(self,             # <<<<<<<<<<<<<<
 *                                 func: Callable[[
 *                                     Font,
 */
  __pyx_codeobj__114 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__107, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_font_h_extents_func, 1139, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__114)) __PYX_ERR(0, 1139, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1149
 *         self._font_h_extents_func = func
 * 
 *     def set_font_v_extents_func(self,             # <<<<<<<<<<<<<<
 *                                 func: Callable[[
 *                                     Font,
 */
  __pyx_codeobj__115 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__107, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_font_v_extents_func, 1149, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__115)) __PYX_ERR(0, 1149, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__116 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__116)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__117 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__117)) __PYX_ERR(1, 3, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1159
 *         self._font_v_extents_func = func
 * 
 * def shape(font: Font, buffer: Buffer,             # <<<<<<<<<<<<<<
 *         features: Dict[str,Union[int,bool,Sequence[Tuple[int,int,Union[int,bool]]]]] = None,
 *         shapers: List[str] = None) -> None:
 */
  __pyx_tuple__118 = PyTuple_Pack(16, __pyx_n_s_font, __pyx_n_s_buffer, __pyx_n_s_features, __pyx_n_s_shapers, __pyx_n_s_size, __pyx_n_s_hb_features, __pyx_n_s_packed, __pyx_n_s_feat, __pyx_n_s_c_shapers, __pyx_n_s_value, __pyx_n_s_i, __pyx_n_s_name, __pyx_n_s_start, __pyx_n_s_end, __pyx_n_s_shaper, __pyx_n_s_ret); if (unlikely(!__pyx_tuple__118)) __PYX_ERR(0, 1159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__118);
  __Pyx_GIVEREF(__pyx_tuple__118);
  __pyx_codeobj__119 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 16, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__118, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_shape, 1159, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__119)) __PYX_ERR(0, 1159, __pyx_L1_error)
  __pyx_tuple__120 = PyTuple_Pack(2, Py_None, Py_None); if (unlikely(!__pyx_tuple__120)) __PYX_ERR(0, 1159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__120);
  __Pyx_GIVEREF(__pyx_tuple__120);

  /* "uharfbuzz/_harfbuzz.pyx":1211
 * 
 * 
 * def ot_tag_to_script(tag: str) -> str:             # <<<<<<<<<<<<<<
 *     cdef bytes packed = tag.encode()
 *     cdef hb_tag_t hb_tag = hb_tag_from_string(<char*>packed, -1)
 */
  __pyx_tuple__121 = PyTuple_Pack(5, __pyx_n_s_tag, __pyx_n_s_packed, __pyx_n_s_hb_tag, __pyx_n_s_hb_script, __pyx_n_s_cstr); if (unlikely(!__pyx_tuple__121)) __PYX_ERR(0, 1211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__121);
  __Pyx_GIVEREF(__pyx_tuple__121);
  __pyx_codeobj__122 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__121, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_tag_to_script, 1211, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__122)) __PYX_ERR(0, 1211, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1222
 * 
 * 
 * def ot_tag_to_language(tag: str) -> str:             # <<<<<<<<<<<<<<
 *     cdef bytes packed = tag.encode()
 *     cdef hb_tag_t hb_tag = hb_tag_from_string(<char*>packed, -1)
 */
  __pyx_tuple__123 = PyTuple_Pack(5, __pyx_n_s_tag, __pyx_n_s_packed, __pyx_n_s_hb_tag, __pyx_n_s_hb_language, __pyx_n_s_cstr); if (unlikely(!__pyx_tuple__123)) __PYX_ERR(0, 1222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__123);
  __Pyx_GIVEREF(__pyx_tuple__123);
  __pyx_codeobj__124 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__123, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_tag_to_language, 1222, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__124)) __PYX_ERR(0, 1222, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1233
 * 
 * 
 * def ot_layout_lookup_get_glyph_alternates(             # <<<<<<<<<<<<<<
 *         face: Face, lookup_index : int, glyph : hb_codepoint_t) -> List[int]:
 *     cdef list alternates = []
 */
  __pyx_tuple__125 = PyTuple_Pack(8, __pyx_n_s_face, __pyx_n_s_lookup_index, __pyx_n_s_glyph, __pyx_n_s_alternates, __pyx_n_s_i, __pyx_n_s_start_offset, __pyx_n_s_alternate_count, __pyx_n_s_alternate_glyphs); if (unlikely(!__pyx_tuple__125)) __PYX_ERR(0, 1233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__125);
  __Pyx_GIVEREF(__pyx_tuple__125);
  __pyx_codeobj__126 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__125, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_layout_lookup_get_glyph_alter, 1233, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__126)) __PYX_ERR(0, 1233, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1249
 * 
 * 
 * def ot_layout_language_get_feature_tags(             # <<<<<<<<<<<<<<
 *         face: Face, tag: str, script_index: int = 0,
 *         language_index: int = 0xFFFF) -> List[str]:
 */
  __pyx_tuple__127 = PyTuple_Pack(12, __pyx_n_s_face, __pyx_n_s_tag, __pyx_n_s_script_index, __pyx_n_s_language_index, __pyx_n_s_packed, __pyx_n_s_hb_tag, __pyx_n_s_feature_count, __pyx_n_s_feature_tags, __pyx_n_s_tags, __pyx_n_s_cstr, __pyx_n_s_i, __pyx_n_s_start_offset); if (unlikely(!__pyx_tuple__127)) __PYX_ERR(0, 1249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__127);
  __Pyx_GIVEREF(__pyx_tuple__127);
  __pyx_codeobj__128 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 12, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__127, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_layout_language_get_feature_t, 1249, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__128)) __PYX_ERR(0, 1249, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1273
 * 
 * 
 * def ot_layout_script_get_language_tags(             # <<<<<<<<<<<<<<
 *         face: Face, tag: str, script_index: int = 0) -> List[str]:
 *     cdef bytes packed = tag.encode()
 */
  __pyx_tuple__129 = PyTuple_Pack(11, __pyx_n_s_face, __pyx_n_s_tag, __pyx_n_s_script_index, __pyx_n_s_packed, __pyx_n_s_hb_tag, __pyx_n_s_language_count, __pyx_n_s_language_tags, __pyx_n_s_tags, __pyx_n_s_cstr, __pyx_n_s_i, __pyx_n_s_start_offset); if (unlikely(!__pyx_tuple__129)) __PYX_ERR(0, 1273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__129);
  __Pyx_GIVEREF(__pyx_tuple__129);
  __pyx_codeobj__130 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 11, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__129, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_layout_script_get_language_ta, 1273, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__130)) __PYX_ERR(0, 1273, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1294
 *     return tags
 * 
 * def ot_layout_table_get_script_tags(face: Face, tag: str) -> List[str]:             # <<<<<<<<<<<<<<
 *     cdef bytes packed = tag.encode()
 *     cdef hb_tag_t hb_tag = hb_tag_from_string(<char*>packed, -1)
 */
  __pyx_tuple__131 = PyTuple_Pack(10, __pyx_n_s_face, __pyx_n_s_tag, __pyx_n_s_packed, __pyx_n_s_hb_tag, __pyx_n_s_script_count, __pyx_n_s_script_tags, __pyx_n_s_tags, __pyx_n_s_cstr, __pyx_n_s_i, __pyx_n_s_start_offset); if (unlikely(!__pyx_tuple__131)) __PYX_ERR(0, 1294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__131);
  __Pyx_GIVEREF(__pyx_tuple__131);
  __pyx_codeobj__132 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 10, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__131, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_layout_table_get_script_tags, 1294, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__132)) __PYX_ERR(0, 1294, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1314
 *     return tags
 * 
 * def ot_layout_get_baseline(font: Font,             # <<<<<<<<<<<<<<
 *                            baseline_tag: str,
 *                            direction: str,
 */
  __pyx_tuple__133 = PyTuple_Pack(12, __pyx_n_s_font, __pyx_n_s_baseline_tag, __pyx_n_s_direction, __pyx_n_s_script_tag, __pyx_n_s_language_tag, __pyx_n_s_hb_baseline_tag, __pyx_n_s_hb_direction, __pyx_n_s_hb_script_tag, __pyx_n_s_hb_language_tag, __pyx_n_s_hb_position, __pyx_n_s_success, __pyx_n_s_packed); if (unlikely(!__pyx_tuple__133)) __PYX_ERR(0, 1314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__133);
  __Pyx_GIVEREF(__pyx_tuple__133);
  __pyx_codeobj__134 = (PyObject*)__Pyx_PyCode_New(5, 0, 0, 12, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__133, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_layout_get_baseline, 1314, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__134)) __PYX_ERR(0, 1314, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1360
 *         return None
 * 
 * def ot_layout_has_glyph_classes(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_layout_has_glyph_classes(face._hb_face)
 * 
 */
  __pyx_tuple__135 = PyTuple_Pack(1, __pyx_n_s_face); if (unlikely(!__pyx_tuple__135)) __PYX_ERR(0, 1360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__135);
  __Pyx_GIVEREF(__pyx_tuple__135);
  __pyx_codeobj__136 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__135, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_layout_has_glyph_classes, 1360, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__136)) __PYX_ERR(0, 1360, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1363
 *     return hb_ot_layout_has_glyph_classes(face._hb_face)
 * 
 * def ot_layout_has_positioning(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_layout_has_positioning(face._hb_face)
 * 
 */
  __pyx_codeobj__137 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__135, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_layout_has_positioning, 1363, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__137)) __PYX_ERR(0, 1363, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1366
 *     return hb_ot_layout_has_positioning(face._hb_face)
 * 
 * def ot_layout_has_substitution(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_layout_has_substitution(face._hb_face)
 * 
 */
  __pyx_codeobj__138 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__135, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_layout_has_substitution, 1366, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__138)) __PYX_ERR(0, 1366, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1376
 *     COMPONENT = HB_OT_LAYOUT_GLYPH_CLASS_COMPONENT
 * 
 * def ot_layout_get_glyph_class(face: Face, glyph: int) -> OTLayoutGlyphClass:             # <<<<<<<<<<<<<<
 *     return OTLayoutGlyphClass(hb_ot_layout_get_glyph_class(face._hb_face, glyph))
 * 
 */
  __pyx_tuple__139 = PyTuple_Pack(2, __pyx_n_s_face, __pyx_n_s_glyph); if (unlikely(!__pyx_tuple__139)) __PYX_ERR(0, 1376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__139);
  __Pyx_GIVEREF(__pyx_tuple__139);
  __pyx_codeobj__140 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__139, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_layout_get_glyph_class, 1376, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__140)) __PYX_ERR(0, 1376, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1380
 * 
 * 
 * def ot_color_has_palettes(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_color_has_palettes(face._hb_face)
 * 
 */
  __pyx_codeobj__141 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__135, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_color_has_palettes, 1380, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__141)) __PYX_ERR(0, 1380, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1383
 *     return hb_ot_color_has_palettes(face._hb_face)
 * 
 * def ot_color_palette_get_count(face: Face) -> int:             # <<<<<<<<<<<<<<
 *     return hb_ot_color_palette_get_count(face._hb_face)
 * 
 */
  __pyx_codeobj__142 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__135, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_color_palette_get_count, 1383, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__142)) __PYX_ERR(0, 1383, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1391
 *     USABLE_WITH_DARK_BACKGROUND = HB_OT_COLOR_PALETTE_FLAG_USABLE_WITH_DARK_BACKGROUND
 * 
 * def ot_color_palette_get_flags(face: Face, palette_index: int) -> OTColorPaletteFlags:             # <<<<<<<<<<<<<<
 *     return OTColorPaletteFlags(hb_ot_color_palette_get_flags(face._hb_face, palette_index))
 * 
 */
  __pyx_tuple__143 = PyTuple_Pack(2, __pyx_n_s_face, __pyx_n_s_palette_index); if (unlikely(!__pyx_tuple__143)) __PYX_ERR(0, 1391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__143);
  __Pyx_GIVEREF(__pyx_tuple__143);
  __pyx_codeobj__144 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__143, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_color_palette_get_flags, 1391, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__144)) __PYX_ERR(0, 1391, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1394
 *     return OTColorPaletteFlags(hb_ot_color_palette_get_flags(face._hb_face, palette_index))
 * 
 * def ot_color_palette_get_colors(face: Face, palette_index: int) -> List[Color]:             # <<<<<<<<<<<<<<
 *     cdef list ret = []
 *     cdef unsigned int i
 */
  __pyx_tuple__145 = PyTuple_Pack(7, __pyx_n_s_face, __pyx_n_s_palette_index, __pyx_n_s_ret, __pyx_n_s_i, __pyx_n_s_start_offset, __pyx_n_s_color_count, __pyx_n_s_colors); if (unlikely(!__pyx_tuple__145)) __PYX_ERR(0, 1394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__145);
  __Pyx_GIVEREF(__pyx_tuple__145);
  __pyx_codeobj__146 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__145, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_color_palette_get_colors, 1394, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__146)) __PYX_ERR(0, 1394, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1406
 *     return ret
 * 
 * def ot_color_palette_get_name_id(face: Face, palette_index: int) -> int | None:             # <<<<<<<<<<<<<<
 *     cdef hb_ot_name_id_t name_id
 *     name_id = hb_ot_color_palette_get_name_id(face._hb_face, palette_index)
 */
  __pyx_tuple__147 = PyTuple_Pack(3, __pyx_n_s_face, __pyx_n_s_palette_index, __pyx_n_s_name_id); if (unlikely(!__pyx_tuple__147)) __PYX_ERR(0, 1406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__147);
  __Pyx_GIVEREF(__pyx_tuple__147);
  __pyx_codeobj__148 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__147, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_color_palette_get_name_id, 1406, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__148)) __PYX_ERR(0, 1406, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1413
 *     return name_id
 * 
 * def ot_color_palette_color_get_name_id(face: Face, color_index: int) -> int | None:             # <<<<<<<<<<<<<<
 *     cdef hb_ot_name_id_t name_id
 *     name_id =  hb_ot_color_palette_color_get_name_id(face._hb_face, color_index)
 */
  __pyx_tuple__149 = PyTuple_Pack(3, __pyx_n_s_face, __pyx_n_s_color_index, __pyx_n_s_name_id); if (unlikely(!__pyx_tuple__149)) __PYX_ERR(0, 1413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__149);
  __Pyx_GIVEREF(__pyx_tuple__149);
  __pyx_codeobj__150 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__149, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_color_palette_color_get_name, 1413, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__150)) __PYX_ERR(0, 1413, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1420
 *     return name_id
 * 
 * def ot_color_has_layers(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_color_has_layers(face._hb_face)
 * 
 */
  __pyx_codeobj__151 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__135, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_color_has_layers, 1420, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__151)) __PYX_ERR(0, 1420, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1425
 * OTColorLayer = namedtuple("OTColorLayer", ["glyph", "color_index"])
 * 
 * def ot_color_glyph_get_layers(face: Face, glyph: int) -> List[OTColorLayer]:             # <<<<<<<<<<<<<<
 *     cdef list ret = []
 *     cdef unsigned int i
 */
  __pyx_tuple__152 = PyTuple_Pack(7, __pyx_n_s_face, __pyx_n_s_glyph, __pyx_n_s_ret, __pyx_n_s_i, __pyx_n_s_start_offset, __pyx_n_s_layer_count, __pyx_n_s_layers); if (unlikely(!__pyx_tuple__152)) __PYX_ERR(0, 1425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__152);
  __Pyx_GIVEREF(__pyx_tuple__152);
  __pyx_codeobj__153 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 7, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__152, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_color_glyph_get_layers, 1425, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__153)) __PYX_ERR(0, 1425, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1438
 *     return ret
 * 
 * def ot_color_has_paint(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_color_has_paint(face._hb_face)
 * 
 */
  __pyx_codeobj__154 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__135, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_color_has_paint, 1438, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__154)) __PYX_ERR(0, 1438, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1441
 *     return hb_ot_color_has_paint(face._hb_face)
 * 
 * def ot_color_glyph_has_paint(face: Face, glyph: int) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_color_glyph_has_paint(face._hb_face, glyph)
 * 
 */
  __pyx_codeobj__155 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__139, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_color_glyph_has_paint, 1441, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__155)) __PYX_ERR(0, 1441, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1444
 *     return hb_ot_color_glyph_has_paint(face._hb_face, glyph)
 * 
 * def ot_color_has_svg(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_color_has_svg(face._hb_face)
 * 
 */
  __pyx_codeobj__156 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__135, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_color_has_svg, 1444, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__156)) __PYX_ERR(0, 1444, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1447
 *     return hb_ot_color_has_svg(face._hb_face)
 * 
 * def ot_color_glyph_get_svg(face: Face, glyph: int) -> Blob:             # <<<<<<<<<<<<<<
 *     cdef hb_blob_t* blob
 *     blob = hb_ot_color_glyph_reference_svg(face._hb_face, glyph)
 */
  __pyx_tuple__157 = PyTuple_Pack(3, __pyx_n_s_face, __pyx_n_s_glyph, __pyx_n_s_blob); if (unlikely(!__pyx_tuple__157)) __PYX_ERR(0, 1447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__157);
  __Pyx_GIVEREF(__pyx_tuple__157);
  __pyx_codeobj__158 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__157, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_color_glyph_get_svg, 1447, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__158)) __PYX_ERR(0, 1447, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1452
 *     return Blob.from_ptr(blob)
 * 
 * def ot_color_has_png(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_color_has_png(face._hb_face)
 * 
 */
  __pyx_codeobj__159 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__135, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_color_has_png, 1452, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__159)) __PYX_ERR(0, 1452, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1455
 *     return hb_ot_color_has_png(face._hb_face)
 * 
 * def ot_color_glyph_get_png(font: Font, glyph: int) -> Blob:             # <<<<<<<<<<<<<<
 *     cdef hb_blob_t* blob
 *     blob = hb_ot_color_glyph_reference_png(font._hb_font, glyph)
 */
  __pyx_tuple__160 = PyTuple_Pack(3, __pyx_n_s_font, __pyx_n_s_glyph, __pyx_n_s_blob); if (unlikely(!__pyx_tuple__160)) __PYX_ERR(0, 1455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__160);
  __Pyx_GIVEREF(__pyx_tuple__160);
  __pyx_codeobj__161 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__160, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_color_glyph_get_png, 1455, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__161)) __PYX_ERR(0, 1455, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1461
 * 
 * 
 * def ot_math_has_data(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_math_has_data(face._hb_face)
 * 
 */
  __pyx_codeobj__162 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__135, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_math_has_data, 1461, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__162)) __PYX_ERR(0, 1461, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1522
 *     RADICAL_DEGREE_BOTTOM_RAISE_PERCENT = HB_OT_MATH_CONSTANT_RADICAL_DEGREE_BOTTOM_RAISE_PERCENT
 * 
 * def ot_math_get_constant(font: Font, constant: OTMathConstant) -> int:             # <<<<<<<<<<<<<<
 *     if constant >= len(OTMathConstant):
 *         raise ValueError("invalid constant")
 */
  __pyx_tuple__163 = PyTuple_Pack(2, __pyx_n_s_font, __pyx_n_s_constant); if (unlikely(!__pyx_tuple__163)) __PYX_ERR(0, 1522, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__163);
  __Pyx_GIVEREF(__pyx_tuple__163);
  __pyx_codeobj__164 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__163, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_math_get_constant, 1522, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__164)) __PYX_ERR(0, 1522, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1527
 *     return hb_ot_math_get_constant(font._hb_font, constant)
 * 
 * def ot_math_get_glyph_italics_correction(font: Font, glyph: int) -> int:             # <<<<<<<<<<<<<<
 *     return hb_ot_math_get_glyph_italics_correction(font._hb_font, glyph)
 * 
 */
  __pyx_tuple__165 = PyTuple_Pack(2, __pyx_n_s_font, __pyx_n_s_glyph); if (unlikely(!__pyx_tuple__165)) __PYX_ERR(0, 1527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__165);
  __Pyx_GIVEREF(__pyx_tuple__165);
  __pyx_codeobj__166 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__165, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_math_get_glyph_italics_correc, 1527, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__166)) __PYX_ERR(0, 1527, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1530
 *     return hb_ot_math_get_glyph_italics_correction(font._hb_font, glyph)
 * 
 * def ot_math_get_glyph_top_accent_attachment(font: Font, glyph: int) -> int:             # <<<<<<<<<<<<<<
 *     return hb_ot_math_get_glyph_top_accent_attachment(font._hb_font, glyph)
 * 
 */
  __pyx_codeobj__167 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__165, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_math_get_glyph_top_accent_att, 1530, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__167)) __PYX_ERR(0, 1530, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1533
 *     return hb_ot_math_get_glyph_top_accent_attachment(font._hb_font, glyph)
 * 
 * def ot_math_is_glyph_extended_shape(face: Face, glyph: int) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_math_is_glyph_extended_shape(face._hb_face, glyph)
 * 
 */
  __pyx_codeobj__168 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__139, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_math_is_glyph_extended_shape, 1533, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__168)) __PYX_ERR(0, 1533, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1536
 *     return hb_ot_math_is_glyph_extended_shape(face._hb_face, glyph)
 * 
 * def ot_math_get_min_connector_overlap(font: Font, direction: str) -> int:             # <<<<<<<<<<<<<<
 *     cdef bytes packed = direction.encode()
 *     cdef char* cstr = packed
 */
  __pyx_tuple__169 = PyTuple_Pack(5, __pyx_n_s_font, __pyx_n_s_direction, __pyx_n_s_packed, __pyx_n_s_cstr, __pyx_n_s_hb_direction); if (unlikely(!__pyx_tuple__169)) __PYX_ERR(0, 1536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__169);
  __Pyx_GIVEREF(__pyx_tuple__169);
  __pyx_codeobj__170 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_math_get_min_connector_overla, 1536, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__170)) __PYX_ERR(0, 1536, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1548
 *     BOTTOM_LEFT = HB_OT_MATH_KERN_BOTTOM_LEFT
 * 
 * def ot_math_get_glyph_kerning(font: Font,             # <<<<<<<<<<<<<<
 *                               glyph: int,
 *                               kern: OTMathKern,
 */
  __pyx_tuple__171 = PyTuple_Pack(4, __pyx_n_s_font, __pyx_n_s_glyph, __pyx_n_s_kern, __pyx_n_s_correction_height); if (unlikely(!__pyx_tuple__171)) __PYX_ERR(0, 1548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__171);
  __Pyx_GIVEREF(__pyx_tuple__171);
  __pyx_codeobj__172 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__171, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_math_get_glyph_kerning, 1548, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__172)) __PYX_ERR(0, 1548, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1558
 * OTMathKernEntry = namedtuple("OTMathKernEntry", ["max_correction_height", "kern_value"])
 * 
 * def ot_math_get_glyph_kernings(font: Font,             # <<<<<<<<<<<<<<
 *                                glyph: int,
 *                                kern: OTMathKern) -> List[OTMathKernEntry]:
 */
  __pyx_tuple__173 = PyTuple_Pack(8, __pyx_n_s_font, __pyx_n_s_glyph, __pyx_n_s_kern, __pyx_n_s_count, __pyx_n_s_kerns_array, __pyx_n_s_kerns, __pyx_n_s_i, __pyx_n_s_start_offset); if (unlikely(!__pyx_tuple__173)) __PYX_ERR(0, 1558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__173);
  __Pyx_GIVEREF(__pyx_tuple__173);
  __pyx_codeobj__174 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__173, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_math_get_glyph_kernings, 1558, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__174)) __PYX_ERR(0, 1558, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1578
 * OTMathGlyphVariant = namedtuple("OTMathGlyphVariant", ["glyph", "advance"])
 * 
 * def ot_math_get_glyph_variants(font: Font, glyph: int, direction: str) -> List[OTMathGlyphVariant]:             # <<<<<<<<<<<<<<
 *     cdef bytes packed = direction.encode()
 *     cdef char* cstr = packed
 */
  __pyx_tuple__175 = PyTuple_Pack(11, __pyx_n_s_font, __pyx_n_s_glyph, __pyx_n_s_direction, __pyx_n_s_packed, __pyx_n_s_cstr, __pyx_n_s_hb_direction, __pyx_n_s_count, __pyx_n_s_variants_array, __pyx_n_s_variants, __pyx_n_s_i, __pyx_n_s_start_offset); if (unlikely(!__pyx_tuple__175)) __PYX_ERR(0, 1578, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__175);
  __Pyx_GIVEREF(__pyx_tuple__175);
  __pyx_codeobj__176 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 11, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__175, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_math_get_glyph_variants, 1578, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__176)) __PYX_ERR(0, 1578, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1603
 * )
 * 
 * def ot_math_get_glyph_assembly(font: Font,             # <<<<<<<<<<<<<<
 *                                glyph: int,
 *                                direction: str) -> Tuple[List[OTMathGlyphPart], int]:
 */
  __pyx_tuple__177 = PyTuple_Pack(12, __pyx_n_s_font, __pyx_n_s_glyph, __pyx_n_s_direction, __pyx_n_s_packed, __pyx_n_s_cstr, __pyx_n_s_hb_direction, __pyx_n_s_count, __pyx_n_s_assembly_array, __pyx_n_s_assembly, __pyx_n_s_i, __pyx_n_s_start_offset, __pyx_n_s_italics_correction); if (unlikely(!__pyx_tuple__177)) __PYX_ERR(0, 1603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__177);
  __Pyx_GIVEREF(__pyx_tuple__177);
  __pyx_codeobj__178 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 12, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__177, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_math_get_glyph_assembly, 1603, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__178)) __PYX_ERR(0, 1603, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1627
 *     return assembly, italics_correction
 * 
 * def ot_font_set_funcs(Font font):             # <<<<<<<<<<<<<<
 *     hb_ot_font_set_funcs(font._hb_font)
 * 
 */
  __pyx_tuple__179 = PyTuple_Pack(1, __pyx_n_s_font); if (unlikely(!__pyx_tuple__179)) __PYX_ERR(0, 1627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__179);
  __Pyx_GIVEREF(__pyx_tuple__179);
  __pyx_codeobj__180 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__179, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_ot_font_set_funcs, 1627, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__180)) __PYX_ERR(0, 1627, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1663
 * 
 * class Color(namedtuple("Color", ["red", "green", "blue", "alpha"])):
 *     def to_int(self) -> int:             # <<<<<<<<<<<<<<
 *         return HB_COLOR(self.blue, self.green, self.red, self.alpha)
 * 
 */
  __pyx_codeobj__181 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_to_int, 1663, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__181)) __PYX_ERR(0, 1663, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1666
 *         return HB_COLOR(self.blue, self.green, self.red, self.alpha)
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def from_int(value: int) -> Color:
 *         r = hb_color_get_red(value)
 */
  __pyx_tuple__182 = PyTuple_Pack(5, __pyx_n_s_value, __pyx_n_s_r, __pyx_n_s_g, __pyx_n_s_b, __pyx_n_s_a); if (unlikely(!__pyx_tuple__182)) __PYX_ERR(0, 1666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__182);
  __Pyx_GIVEREF(__pyx_tuple__182);
  __pyx_codeobj__183 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__182, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_from_int, 1666, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__183)) __PYX_ERR(0, 1666, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__184 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__184)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__185 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__185)) __PYX_ERR(1, 3, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1919
 *         hb_paint_funcs_destroy(self._hb_paintfuncs)
 * 
 *     def set_push_transform_func(self,             # <<<<<<<<<<<<<<
 *                                 func: Callable[[
 *                                     float,  # xx
 */
  __pyx_tuple__186 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_func); if (unlikely(!__pyx_tuple__186)) __PYX_ERR(0, 1919, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__186);
  __Pyx_GIVEREF(__pyx_tuple__186);
  __pyx_codeobj__187 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__186, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_push_transform_func, 1919, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__187)) __PYX_ERR(0, 1919, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1933
 *             self._hb_paintfuncs, _paint_push_transform_func, <void*>self, NULL)
 * 
 *     def set_pop_transform_func(self,             # <<<<<<<<<<<<<<
 *                                func: Callable[[
 *                                    object,  # paint_data
 */
  __pyx_codeobj__188 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__186, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_pop_transform_func, 1933, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__188)) __PYX_ERR(0, 1933, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1941
 *             self._hb_paintfuncs, _paint_pop_transform_func, <void*>self, NULL)
 * 
 *     def set_color_glyph_func(self,             # <<<<<<<<<<<<<<
 *                              func: Callable[[
 *                                  int,  # gid
 */
  __pyx_codeobj__189 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__186, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_color_glyph_func, 1941, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__189)) __PYX_ERR(0, 1941, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1950
 *             self._hb_paintfuncs, _paint_color_glyph_func, <void*>self, NULL)
 * 
 *     def set_push_clip_glyph_func(self,             # <<<<<<<<<<<<<<
 *                                  func: Callable[[
 *                                      int,  # gid
 */
  __pyx_codeobj__190 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__186, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_push_clip_glyph_func, 1950, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__190)) __PYX_ERR(0, 1950, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1959
 *             self._hb_paintfuncs, _paint_push_clip_glyph_func, <void*>self, NULL)
 * 
 *     def set_push_clip_rectangle_func(self,             # <<<<<<<<<<<<<<
 *                                      func: Callable[[
 *                                          float,  # xmin
 */
  __pyx_codeobj__191 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__186, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_push_clip_rectangle_func, 1959, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__191)) __PYX_ERR(0, 1959, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1971
 *             self._hb_paintfuncs, _paint_push_clip_rectangle_func, <void*>self, NULL)
 * 
 *     def set_pop_clip_func(self,             # <<<<<<<<<<<<<<
 *                           func: Callable[[
 *                               object,  # paint_data
 */
  __pyx_codeobj__192 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__186, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_pop_clip_func, 1971, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__192)) __PYX_ERR(0, 1971, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1979
 *             self._hb_paintfuncs, _paint_pop_clip_func, <void*>self, NULL)
 * 
 *     def set_color_func(self,             # <<<<<<<<<<<<<<
 *                        func: Callable[[
 *                            Color,  # color
 */
  __pyx_codeobj__193 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__186, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_color_func, 1979, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__193)) __PYX_ERR(0, 1979, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1989
 *             self._hb_paintfuncs, _paint_color_func, <void*>self, NULL)
 * 
 *     def set_image_func(self,             # <<<<<<<<<<<<<<
 *                        func: Callable[[
 *                            Blob,  # image
 */
  __pyx_codeobj__194 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__186, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_image_func, 1989, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__194)) __PYX_ERR(0, 1989, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2003
 *             self._hb_paintfuncs, _paint_image_func, <void*>self, NULL)
 * 
 *     def set_linear_gradient_func(self,             # <<<<<<<<<<<<<<
 *                                  func: Callable[[
 *                                     ColorLine,  # color_line
 */
  __pyx_codeobj__195 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__186, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_linear_gradient_func, 2003, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__195)) __PYX_ERR(0, 2003, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2018
 *             self._hb_paintfuncs, _paint_linear_gradient_func, <void*>self, NULL)
 * 
 *     def set_radial_gradient_func(self,             # <<<<<<<<<<<<<<
 *                                  func: Callable[[
 *                                     ColorLine,  # color_line
 */
  __pyx_codeobj__196 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__186, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_radial_gradient_func, 2018, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__196)) __PYX_ERR(0, 2018, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2033
 *             self._hb_paintfuncs, _paint_radial_gradient_func, <void*>self, NULL)
 * 
 *     def set_sweep_gradient_func(self,             # <<<<<<<<<<<<<<
 *                                 func: Callable[[
 *                                     ColorLine,  # color_line
 */
  __pyx_codeobj__197 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__186, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_sweep_gradient_func, 2033, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__197)) __PYX_ERR(0, 2033, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2046
 *             self._hb_paintfuncs, _paint_sweep_gradient_func, <void*>self, NULL)
 * 
 *     def set_push_group_func(self,             # <<<<<<<<<<<<<<
 *                             func: Callable[[
 *                                 object,  # paint_data
 */
  __pyx_codeobj__198 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__186, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_push_group_func, 2046, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__198)) __PYX_ERR(0, 2046, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2054
 *             self._hb_paintfuncs, _paint_push_group_func, <void*>self, NULL)
 * 
 *     def set_pop_group_func(self,             # <<<<<<<<<<<<<<
 *                            func: Callable[[
 *                                PaintCompositeMode,  # mode
 */
  __pyx_codeobj__199 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__186, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_pop_group_func, 2054, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__199)) __PYX_ERR(0, 2054, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2063
 *             self._hb_paintfuncs, _paint_pop_group_func, <void*>self, NULL)
 * 
 *     def set_custom_palette_color_func(self,             # <<<<<<<<<<<<<<
 *                                       func: Callable[[
 *                                           int,  # color_index
 */
  __pyx_codeobj__200 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__186, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_custom_palette_color_func, 2063, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__200)) __PYX_ERR(0, 2063, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__201 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__201)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__202 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__202)) __PYX_ERR(1, 3, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2139
 *         hb_draw_funcs_destroy(self._hb_drawfuncs)
 * 
 *     def get_glyph_shape(self, font: Font, gid: int):             # <<<<<<<<<<<<<<
 *         if not self._warned:
 *             warnings.warn(
 */
  __pyx_tuple__203 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_font, __pyx_n_s_gid); if (unlikely(!__pyx_tuple__203)) __PYX_ERR(0, 2139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__203);
  __Pyx_GIVEREF(__pyx_tuple__203);
  __pyx_codeobj__204 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__203, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_get_glyph_shape, 2139, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__204)) __PYX_ERR(0, 2139, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2148
 *         font.draw_glyph(gid, self)
 * 
 *     def draw_glyph(self, font: Font, gid: int, draw_data: object = None):             # <<<<<<<<<<<<<<
 *         if not self._warned:
 *             warnings.warn(
 */
  __pyx_tuple__205 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_font, __pyx_n_s_gid, __pyx_n_s_draw_data); if (unlikely(!__pyx_tuple__205)) __PYX_ERR(0, 2148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__205);
  __Pyx_GIVEREF(__pyx_tuple__205);
  __pyx_codeobj__206 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__205, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_draw_glyph, 2148, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__206)) __PYX_ERR(0, 2148, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2157
 *         font.draw_glyph(gid, self, draw_data)
 * 
 *     def set_move_to_func(self,             # <<<<<<<<<<<<<<
 *                          func: Callable[[
 *                              float,
 */
  __pyx_tuple__207 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_func, __pyx_n_s_user_data, __pyx_n_s_func_p, __pyx_n_s_user_data_p); if (unlikely(!__pyx_tuple__207)) __PYX_ERR(0, 2157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__207);
  __Pyx_GIVEREF(__pyx_tuple__207);
  __pyx_codeobj__208 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__207, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_move_to_func, 2157, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__208)) __PYX_ERR(0, 2157, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2181
 *             self._hb_drawfuncs, func_p, user_data_p, NULL)
 * 
 *     def set_line_to_func(self,             # <<<<<<<<<<<<<<
 *                          func: Callable[[
 *                              float,
 */
  __pyx_codeobj__209 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__207, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_line_to_func, 2181, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__209)) __PYX_ERR(0, 2181, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2205
 *             self._hb_drawfuncs, func_p, user_data_p, NULL)
 * 
 *     def set_cubic_to_func(self,             # <<<<<<<<<<<<<<
 *                           func: Callable[[
 *                              float,
 */
  __pyx_codeobj__210 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__207, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_cubic_to_func, 2205, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__210)) __PYX_ERR(0, 2205, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2233
 *             self._hb_drawfuncs, func_p, user_data_p, NULL)
 * 
 *     def set_quadratic_to_func(self,             # <<<<<<<<<<<<<<
 *                               func: Callable[[
 *                                  float,
 */
  __pyx_codeobj__211 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__207, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_quadratic_to_func, 2233, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__211)) __PYX_ERR(0, 2233, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2259
 *             self._hb_drawfuncs, func_p, user_data_p, NULL)
 * 
 *     def set_close_path_func(self,             # <<<<<<<<<<<<<<
 *                             func: Callable[[
 *                                 object
 */
  __pyx_codeobj__212 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__207, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_close_path_func, 2259, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__212)) __PYX_ERR(0, 2259, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__213 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__213)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__214 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__214)) __PYX_ERR(1, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__215 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__215)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__216 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__216)) __PYX_ERR(1, 3, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2343
 * 
 * 
 * def repack(subtables: List[bytes],             # <<<<<<<<<<<<<<
 *            graphnodes: List[Tuple[List[Tuple[int, int, int]],
 *                             List[Tuple[int, int, int]]
 */
  __pyx_tuple__217 = PyTuple_Pack(2, __pyx_n_s_subtables, __pyx_n_s_graphnodes); if (unlikely(!__pyx_tuple__217)) __PYX_ERR(0, 2343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__217);
  __Pyx_GIVEREF(__pyx_tuple__217);
  __pyx_codeobj__218 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__217, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_repack, 2343, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__218)) __PYX_ERR(0, 2343, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2350
 * 
 * 
 * def repack_with_tag(tag: str,             # <<<<<<<<<<<<<<
 *                     subtables: List[bytes],
 *                     graphnodes: List[Tuple[List[Tuple[int, int, int]],
 */
  __pyx_tuple__219 = PyTuple_Pack(18, __pyx_n_s_tag, __pyx_n_s_subtables, __pyx_n_s_graphnodes, __pyx_n_s_num_nodes, __pyx_n_s_table_bytes, __pyx_n_s_table_data, __pyx_n_s_head, __pyx_n_s_tail, __pyx_n_s_p, __pyx_n_s_obj_list, __pyx_n_s_i, __pyx_n_s_node, __pyx_n_s_tag_packed, __pyx_n_s_cstr, __pyx_n_s_packed_blob, __pyx_n_s_blob_length, __pyx_n_s_blob_data, __pyx_n_s_packed); if (unlikely(!__pyx_tuple__219)) __PYX_ERR(0, 2350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__219);
  __Pyx_GIVEREF(__pyx_tuple__219);
  __pyx_codeobj__220 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 18, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__219, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_repack_with_tag, 2350, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__220)) __PYX_ERR(0, 2350, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2435
 * 
 * 
 * def subset_preprocess(face: Face) -> Face:             # <<<<<<<<<<<<<<
 *     new_face = hb_subset_preprocess(face._hb_face)
 *     return Face.from_ptr(new_face)
 */
  __pyx_tuple__221 = PyTuple_Pack(2, __pyx_n_s_face, __pyx_n_s_new_face); if (unlikely(!__pyx_tuple__221)) __PYX_ERR(0, 2435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__221);
  __Pyx_GIVEREF(__pyx_tuple__221);
  __pyx_codeobj__222 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__221, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_subset_preprocess, 2435, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__222)) __PYX_ERR(0, 2435, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2439
 *     return Face.from_ptr(new_face)
 * 
 * def subset(face: Face, input: SubsetInput) -> Face:             # <<<<<<<<<<<<<<
 *     new_face = hb_subset_or_fail(face._hb_face, input._hb_input)
 *     if new_face == NULL:
 */
  __pyx_tuple__223 = PyTuple_Pack(3, __pyx_n_s_face, __pyx_n_s_input, __pyx_n_s_new_face); if (unlikely(!__pyx_tuple__223)) __PYX_ERR(0, 2439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__223);
  __Pyx_GIVEREF(__pyx_tuple__223);
  __pyx_codeobj__224 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__223, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_subset, 2439, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__224)) __PYX_ERR(0, 2439, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2482
 *             hb_subset_input_destroy(self._hb_input)
 * 
 *     def subset(self, source: Face) -> Face:             # <<<<<<<<<<<<<<
 *         return subset(source, self)
 * 
 */
  __pyx_tuple__225 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_source); if (unlikely(!__pyx_tuple__225)) __PYX_ERR(0, 2482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__225);
  __Pyx_GIVEREF(__pyx_tuple__225);
  __pyx_codeobj__226 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__225, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_subset, 2482, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__226)) __PYX_ERR(0, 2482, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2485
 *         return subset(source, self)
 * 
 *     def keep_everything(self):             # <<<<<<<<<<<<<<
 *         hb_subset_input_keep_everything(self._hb_input)
 * 
 */
  __pyx_codeobj__227 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_keep_everything, 2485, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__227)) __PYX_ERR(0, 2485, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2488
 *         hb_subset_input_keep_everything(self._hb_input)
 * 
 *     def pin_axis_to_default(self, face: Face, tag: str) -> bool:             # <<<<<<<<<<<<<<
 *         hb_tag = hb_tag_from_string(tag.encode("ascii"), -1)
 *         return hb_subset_input_pin_axis_to_default(
 */
  __pyx_tuple__228 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_face, __pyx_n_s_tag, __pyx_n_s_hb_tag); if (unlikely(!__pyx_tuple__228)) __PYX_ERR(0, 2488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__228);
  __Pyx_GIVEREF(__pyx_tuple__228);
  __pyx_codeobj__229 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__228, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_pin_axis_to_default, 2488, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__229)) __PYX_ERR(0, 2488, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2494
 *         )
 * 
 *     def pin_axis_location(self, face: Face, tag: str, value: float) -> bool:             # <<<<<<<<<<<<<<
 *         hb_tag = hb_tag_from_string(tag.encode("ascii"), -1)
 *         return hb_subset_input_pin_axis_location(
 */
  __pyx_tuple__230 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_face, __pyx_n_s_tag, __pyx_n_s_value, __pyx_n_s_hb_tag); if (unlikely(!__pyx_tuple__230)) __PYX_ERR(0, 2494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__230);
  __Pyx_GIVEREF(__pyx_tuple__230);
  __pyx_codeobj__231 = (PyObject*)__Pyx_PyCode_New(4, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__230, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_pin_axis_location, 2494, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__231)) __PYX_ERR(0, 2494, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2508
 *         return Set.from_ptr(hb_set_reference (hb_subset_input_glyph_set(self._hb_input)))
 * 
 *     def sets(self, set_type : SubsetInputSets) -> Set:             # <<<<<<<<<<<<<<
 *         return Set.from_ptr(hb_set_reference (hb_subset_input_set(self._hb_input, set_type)))
 * 
 */
  __pyx_tuple__232 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_set_type); if (unlikely(!__pyx_tuple__232)) __PYX_ERR(0, 2508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__232);
  __Pyx_GIVEREF(__pyx_tuple__232);
  __pyx_codeobj__233 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__232, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_sets, 2508, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__233)) __PYX_ERR(0, 2508, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__234 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__234)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__235 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__235)) __PYX_ERR(1, 3, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2557
 *             hb_subset_plan_destroy(self._hb_plan)
 * 
 *     def execute(self) -> Face:             # <<<<<<<<<<<<<<
 *         new_face = hb_subset_plan_execute_or_fail(self._hb_plan)
 *         if new_face == NULL:
 */
  __pyx_tuple__236 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_new_face); if (unlikely(!__pyx_tuple__236)) __PYX_ERR(0, 2557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__236);
  __Pyx_GIVEREF(__pyx_tuple__236);
  __pyx_codeobj__237 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__236, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_execute, 2557, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__237)) __PYX_ERR(0, 2557, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__238 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__238)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__239 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__239)) __PYX_ERR(1, 3, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2599
 *         return wrapper
 * 
 *     def copy(self) -> Set:             # <<<<<<<<<<<<<<
 *         c = Set()
 *         c._hb_set = hb_set_copy(self._hb_set)
 */
  __pyx_tuple__240 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_c); if (unlikely(!__pyx_tuple__240)) __PYX_ERR(0, 2599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__240);
  __Pyx_GIVEREF(__pyx_tuple__240);
  __pyx_codeobj__241 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__240, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_copy, 2599, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__241)) __PYX_ERR(0, 2599, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2604
 *         return c
 * 
 *     def __copy__(self) -> Set:             # <<<<<<<<<<<<<<
 *         return self.copy()
 * 
 */
  __pyx_codeobj__242 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_copy_2, 2604, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__242)) __PYX_ERR(0, 2604, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2607
 *         return self.copy()
 * 
 *     def clear(self):             # <<<<<<<<<<<<<<
 *         hb_set_clear(self._hb_set)
 * 
 */
  __pyx_codeobj__243 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_clear, 2607, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__243)) __PYX_ERR(0, 2607, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2613
 *         return not hb_set_is_empty(self._hb_set)
 * 
 *     def invert(self):             # <<<<<<<<<<<<<<
 *         hb_set_invert(self._hb_set)
 * 
 */
  __pyx_codeobj__244 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_invert, 2613, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__244)) __PYX_ERR(0, 2613, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2616
 *         hb_set_invert(self._hb_set)
 * 
 *     def is_inverted(self) -> bool:             # <<<<<<<<<<<<<<
 *         return hb_set_is_inverted(self._hb_set)
 * 
 */
  __pyx_codeobj__245 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_is_inverted, 2616, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__245)) __PYX_ERR(0, 2616, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2626
 *         return hb_set_has(self._hb_set, c)
 * 
 *     def add(self, c: int):             # <<<<<<<<<<<<<<
 *         hb_set_add(self._hb_set, c)
 *         if not hb_set_allocation_successful(self._hb_set):
 */
  __pyx_codeobj__246 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__240, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_add, 2626, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__246)) __PYX_ERR(0, 2626, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2631
 *             raise MemoryError()
 * 
 *     def add_range(self, first: int, last: int):             # <<<<<<<<<<<<<<
 *         hb_set_add_range(self._hb_set, first, last)
 *         if not hb_set_allocation_successful(self._hb_set):
 */
  __pyx_tuple__247 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_first, __pyx_n_s_last); if (unlikely(!__pyx_tuple__247)) __PYX_ERR(0, 2631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__247);
  __Pyx_GIVEREF(__pyx_tuple__247);
  __pyx_codeobj__248 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__247, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_add_range, 2631, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__248)) __PYX_ERR(0, 2631, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2636
 *             raise MemoryError()
 * 
 *     def remove(self, c: int):             # <<<<<<<<<<<<<<
 *         if not c in self:
 *             raise KeyError, c
 */
  __pyx_codeobj__249 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__240, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_remove, 2636, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__249)) __PYX_ERR(0, 2636, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2643
 *             raise MemoryError()
 * 
 *     def discard(self, c: int):             # <<<<<<<<<<<<<<
 *         hb_set_del(self._hb_set, c)
 *         if not hb_set_allocation_successful(self._hb_set):
 */
  __pyx_codeobj__250 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__240, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_discard, 2643, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__250)) __PYX_ERR(0, 2643, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2648
 *             raise MemoryError()
 * 
 *     def del_range(self, first: int, last: int):             # <<<<<<<<<<<<<<
 *         hb_set_del_range(self._hb_set, first, last)
 *         if not hb_set_allocation_successful(self._hb_set):
 */
  __pyx_codeobj__251 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__247, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_del_range, 2648, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__251)) __PYX_ERR(0, 2648, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2653
 *             raise MemoryError()
 * 
 *     def _is_equal(self, other: Set) -> bool:             # <<<<<<<<<<<<<<
 *         return hb_set_is_equal(self._hb_set, other._hb_set)
 * 
 */
  __pyx_tuple__252 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_other); if (unlikely(!__pyx_tuple__252)) __PYX_ERR(0, 2653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__252);
  __Pyx_GIVEREF(__pyx_tuple__252);
  __pyx_codeobj__253 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__252, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_is_equal, 2653, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__253)) __PYX_ERR(0, 2653, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2661
 *         return self._is_equal(other)
 * 
 *     def issubset(self, larger_set: Set) -> bool:             # <<<<<<<<<<<<<<
 *         return hb_set_is_subset(self._hb_set, larger_set._hb_set)
 * 
 */
  __pyx_tuple__254 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_larger_set); if (unlikely(!__pyx_tuple__254)) __PYX_ERR(0, 2661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__254);
  __Pyx_GIVEREF(__pyx_tuple__254);
  __pyx_codeobj__255 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__254, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_issubset, 2661, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__255)) __PYX_ERR(0, 2661, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2664
 *         return hb_set_is_subset(self._hb_set, larger_set._hb_set)
 * 
 *     def issuperset(self, smaller_set: Set) -> bool:             # <<<<<<<<<<<<<<
 *         return hb_set_is_subset(smaller_set._hb_set, self._hb_set)
 * 
 */
  __pyx_tuple__256 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_smaller_set); if (unlikely(!__pyx_tuple__256)) __PYX_ERR(0, 2664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__256);
  __Pyx_GIVEREF(__pyx_tuple__256);
  __pyx_codeobj__257 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__256, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_issuperset, 2664, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__257)) __PYX_ERR(0, 2664, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2667
 *         return hb_set_is_subset(smaller_set._hb_set, self._hb_set)
 * 
 *     def _set(self, other: Set):             # <<<<<<<<<<<<<<
 *         hb_set_set(self._hb_set, other._hb_set)
 * 
 */
  __pyx_codeobj__258 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__252, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set_2, 2667, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__258)) __PYX_ERR(0, 2667, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2670
 *         hb_set_set(self._hb_set, other._hb_set)
 * 
 *     def set(self, other):             # <<<<<<<<<<<<<<
 *         if type(other) == Set:
 *             self._set(other)
 */
  __pyx_tuple__259 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_other, __pyx_n_s_c); if (unlikely(!__pyx_tuple__259)) __PYX_ERR(0, 2670, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__259);
  __Pyx_GIVEREF(__pyx_tuple__259);
  __pyx_codeobj__260 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__259, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_set, 2670, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__260)) __PYX_ERR(0, 2670, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2680
 *             raise MemoryError()
 * 
 *     def _update(self, other: Set):             # <<<<<<<<<<<<<<
 *         hb_set_union(self._hb_set, other._hb_set)
 * 
 */
  __pyx_codeobj__261 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__252, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_update, 2680, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__261)) __PYX_ERR(0, 2680, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2683
 *         hb_set_union(self._hb_set, other._hb_set)
 * 
 *     def update(self, other):             # <<<<<<<<<<<<<<
 *         if type(other) == Set:
 *             self._update(other)
 */
  __pyx_codeobj__262 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__259, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_update_2, 2683, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__262)) __PYX_ERR(0, 2683, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2697
 *         return self
 * 
 *     def intersection_update(self, other: Set):             # <<<<<<<<<<<<<<
 *         hb_set_intersect(self._hb_set, other._hb_set)
 *         if not hb_set_allocation_successful(self._hb_set):
 */
  __pyx_codeobj__263 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__252, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_intersection_update, 2697, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__263)) __PYX_ERR(0, 2697, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2706
 *         return self
 * 
 *     def difference_update(self, other: Set):             # <<<<<<<<<<<<<<
 *         hb_set_subtract(self._hb_set, other._hb_set)
 *         if not hb_set_allocation_successful(self._hb_set):
 */
  __pyx_codeobj__264 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__252, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_difference_update, 2706, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__264)) __PYX_ERR(0, 2706, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2715
 *         return self
 * 
 *     def symmetric_difference_update(self, other: Set):             # <<<<<<<<<<<<<<
 *         hb_set_symmetric_difference(self._hb_set, other._hb_set)
 *         if not hb_set_allocation_successful(self._hb_set):
 */
  __pyx_codeobj__265 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__252, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_symmetric_difference_update, 2715, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__265)) __PYX_ERR(0, 2715, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__266 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__266)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__267 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__267)) __PYX_ERR(1, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__268 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__268)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__269 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__269)) __PYX_ERR(1, 3, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2787
 *         return wrapper
 * 
 *     def copy(self) -> Map:             # <<<<<<<<<<<<<<
 *         c = Map()
 *         c._hb_map = hb_map_copy(self._hb_map)
 */
  __pyx_codeobj__270 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__240, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_copy, 2787, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__270)) __PYX_ERR(0, 2787, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2792
 *         return c
 * 
 *     def __copy__(self) -> Map:             # <<<<<<<<<<<<<<
 *         return self.copy()
 * 
 */
  __pyx_codeobj__271 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_copy_2, 2792, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__271)) __PYX_ERR(0, 2792, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2795
 *         return self.copy()
 * 
 *     def _update(self, other : Map):             # <<<<<<<<<<<<<<
 *         hb_map_update(self._hb_map, other._hb_map)
 * 
 */
  __pyx_codeobj__272 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__252, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_update, 2795, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__272)) __PYX_ERR(0, 2795, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2798
 *         hb_map_update(self._hb_map, other._hb_map)
 * 
 *     def update(self, other):             # <<<<<<<<<<<<<<
 *         if type(other) == Map:
 *             self._update(other)
 */
  __pyx_tuple__273 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_other, __pyx_n_s_k, __pyx_n_s_v); if (unlikely(!__pyx_tuple__273)) __PYX_ERR(0, 2798, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__273);
  __Pyx_GIVEREF(__pyx_tuple__273);
  __pyx_codeobj__274 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__273, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_update_2, 2798, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__274)) __PYX_ERR(0, 2798, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2808
 *             raise MemoryError()
 * 
 *     def clear(self):             # <<<<<<<<<<<<<<
 *         hb_map_clear(self._hb_map)
 * 
 */
  __pyx_codeobj__275 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_clear, 2808, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__275)) __PYX_ERR(0, 2808, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2817
 *         return hb_map_get_population(self._hb_map)
 * 
 *     def _is_equal(self, other: Map) -> bool:             # <<<<<<<<<<<<<<
 *         return hb_map_is_equal(self._hb_map, other._hb_map)
 * 
 */
  __pyx_codeobj__276 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__252, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_is_equal, 2817, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__276)) __PYX_ERR(0, 2817, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2830
 *             raise MemoryError()
 * 
 *     def get(self, k: int):             # <<<<<<<<<<<<<<
 *         if k < 0 or k >= self.INVALID_VALUE:
 *             return None
 */
  __pyx_tuple__277 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_k, __pyx_n_s_v); if (unlikely(!__pyx_tuple__277)) __PYX_ERR(0, 2830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__277);
  __Pyx_GIVEREF(__pyx_tuple__277);
  __pyx_codeobj__278 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__277, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_get, 2830, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__278)) __PYX_ERR(0, 2830, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2856
 *         hb_map_del(self._hb_map, c)
 * 
 *     def items(self):             # <<<<<<<<<<<<<<
 *         return MapIter(self)
 * 
 */
  __pyx_codeobj__279 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_items, 2856, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__279)) __PYX_ERR(0, 2856, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2859
 *         return MapIter(self)
 * 
 *     def keys(self):             # <<<<<<<<<<<<<<
 *         return (k for k,v in self.items())
 * 
 */
  __pyx_tuple__280 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__280)) __PYX_ERR(0, 2859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__280);
  __Pyx_GIVEREF(__pyx_tuple__280);
  __pyx_codeobj__281 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__280, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_keys, 2859, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__281)) __PYX_ERR(0, 2859, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":2862
 *         return (k for k,v in self.items())
 * 
 *     def values(self):             # <<<<<<<<<<<<<<
 *         return (v for k,v in self.items())
 * 
 */
  __pyx_codeobj__282 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__280, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_src_uharfbuzz__harfbuzz_pyx, __pyx_n_s_values, 2862, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__282)) __PYX_ERR(0, 2862, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__283 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__283)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__284 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__284)) __PYX_ERR(1, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__285 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__285)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_codeobj__286 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__286)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitConstants(void) {
  if (__Pyx_CreateStringTabAndInitStrings() < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_16384 = PyInt_FromLong(16384L); if (unlikely(!__pyx_int_16384)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_65535 = PyInt_FromLong(65535L); if (unlikely(!__pyx_int_65535)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_globals ### */

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  /* AssertionsEnabled.init */
  if (likely(__Pyx_init_assertions_enabled() == 0)); else

if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_module ### */

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_9uharfbuzz_9_harfbuzz_GlyphInfo = &__pyx_vtable_9uharfbuzz_9_harfbuzz_GlyphInfo;
  __pyx_vtable_9uharfbuzz_9_harfbuzz_GlyphInfo.set = (PyObject *(*)(struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphInfo *, hb_glyph_info_t))__pyx_f_9uharfbuzz_9_harfbuzz_9GlyphInfo_set;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphInfo = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9uharfbuzz_9_harfbuzz_GlyphInfo_spec, NULL); if (unlikely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphInfo)) __PYX_ERR(0, 56, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9uharfbuzz_9_harfbuzz_GlyphInfo_spec, __pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphInfo) < 0) __PYX_ERR(0, 56, __pyx_L1_error)
  #else
  __pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphInfo = &__pyx_type_9uharfbuzz_9_harfbuzz_GlyphInfo;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphInfo) < 0) __PYX_ERR(0, 56, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphInfo->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphInfo->tp_dictoffset && __pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphInfo->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphInfo->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphInfo, __pyx_vtabptr_9uharfbuzz_9_harfbuzz_GlyphInfo) < 0) __PYX_ERR(0, 56, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphInfo) < 0) __PYX_ERR(0, 56, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_GlyphInfo, (PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphInfo) < 0) __PYX_ERR(0, 56, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphInfo) < 0) __PYX_ERR(0, 56, __pyx_L1_error)
  #endif
  __pyx_vtabptr_9uharfbuzz_9_harfbuzz_GlyphPosition = &__pyx_vtable_9uharfbuzz_9_harfbuzz_GlyphPosition;
  __pyx_vtable_9uharfbuzz_9_harfbuzz_GlyphPosition.set = (PyObject *(*)(struct __pyx_obj_9uharfbuzz_9_harfbuzz_GlyphPosition *, hb_glyph_position_t))__pyx_f_9uharfbuzz_9_harfbuzz_13GlyphPosition_set;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphPosition = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9uharfbuzz_9_harfbuzz_GlyphPosition_spec, NULL); if (unlikely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphPosition)) __PYX_ERR(0, 77, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9uharfbuzz_9_harfbuzz_GlyphPosition_spec, __pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphPosition) < 0) __PYX_ERR(0, 77, __pyx_L1_error)
  #else
  __pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphPosition = &__pyx_type_9uharfbuzz_9_harfbuzz_GlyphPosition;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphPosition) < 0) __PYX_ERR(0, 77, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphPosition->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphPosition->tp_dictoffset && __pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphPosition->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphPosition->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphPosition, __pyx_vtabptr_9uharfbuzz_9_harfbuzz_GlyphPosition) < 0) __PYX_ERR(0, 77, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphPosition) < 0) __PYX_ERR(0, 77, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_GlyphPosition, (PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphPosition) < 0) __PYX_ERR(0, 77, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_GlyphPosition) < 0) __PYX_ERR(0, 77, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9uharfbuzz_9_harfbuzz_Buffer_spec, NULL); if (unlikely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer)) __PYX_ERR(0, 132, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9uharfbuzz_9_harfbuzz_Buffer_spec, __pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer) < 0) __PYX_ERR(0, 132, __pyx_L1_error)
  #else
  __pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer = &__pyx_type_9uharfbuzz_9_harfbuzz_Buffer;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer) < 0) __PYX_ERR(0, 132, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer->tp_dictoffset && __pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Buffer, (PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer) < 0) __PYX_ERR(0, 132, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer) < 0) __PYX_ERR(0, 132, __pyx_L1_error)
  #endif
  __pyx_vtabptr_9uharfbuzz_9_harfbuzz_Blob = &__pyx_vtable_9uharfbuzz_9_harfbuzz_Blob;
  __pyx_vtable_9uharfbuzz_9_harfbuzz_Blob.from_ptr = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_Blob *(*)(hb_blob_t *))__pyx_f_9uharfbuzz_9_harfbuzz_4Blob_from_ptr;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_9uharfbuzz_9_harfbuzz_Blob = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9uharfbuzz_9_harfbuzz_Blob_spec, NULL); if (unlikely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_Blob)) __PYX_ERR(0, 375, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9uharfbuzz_9_harfbuzz_Blob_spec, __pyx_ptype_9uharfbuzz_9_harfbuzz_Blob) < 0) __PYX_ERR(0, 375, __pyx_L1_error)
  #else
  __pyx_ptype_9uharfbuzz_9_harfbuzz_Blob = &__pyx_type_9uharfbuzz_9_harfbuzz_Blob;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9uharfbuzz_9_harfbuzz_Blob) < 0) __PYX_ERR(0, 375, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9uharfbuzz_9_harfbuzz_Blob->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_Blob->tp_dictoffset && __pyx_ptype_9uharfbuzz_9_harfbuzz_Blob->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9uharfbuzz_9_harfbuzz_Blob->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_9uharfbuzz_9_harfbuzz_Blob, __pyx_vtabptr_9uharfbuzz_9_harfbuzz_Blob) < 0) __PYX_ERR(0, 375, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_9uharfbuzz_9_harfbuzz_Blob) < 0) __PYX_ERR(0, 375, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Blob, (PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_Blob) < 0) __PYX_ERR(0, 375, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_Blob) < 0) __PYX_ERR(0, 375, __pyx_L1_error)
  #endif
  __pyx_vtabptr_9uharfbuzz_9_harfbuzz_Face = &__pyx_vtable_9uharfbuzz_9_harfbuzz_Face;
  __pyx_vtable_9uharfbuzz_9_harfbuzz_Face.from_ptr = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *(*)(hb_face_t *))__pyx_f_9uharfbuzz_9_harfbuzz_4Face_from_ptr;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_9uharfbuzz_9_harfbuzz_Face = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9uharfbuzz_9_harfbuzz_Face_spec, NULL); if (unlikely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_Face)) __PYX_ERR(0, 444, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9uharfbuzz_9_harfbuzz_Face_spec, __pyx_ptype_9uharfbuzz_9_harfbuzz_Face) < 0) __PYX_ERR(0, 444, __pyx_L1_error)
  #else
  __pyx_ptype_9uharfbuzz_9_harfbuzz_Face = &__pyx_type_9uharfbuzz_9_harfbuzz_Face;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9uharfbuzz_9_harfbuzz_Face) < 0) __PYX_ERR(0, 444, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9uharfbuzz_9_harfbuzz_Face->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_Face->tp_dictoffset && __pyx_ptype_9uharfbuzz_9_harfbuzz_Face->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9uharfbuzz_9_harfbuzz_Face->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_9uharfbuzz_9_harfbuzz_Face, __pyx_vtabptr_9uharfbuzz_9_harfbuzz_Face) < 0) __PYX_ERR(0, 444, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_9uharfbuzz_9_harfbuzz_Face) < 0) __PYX_ERR(0, 444, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Face, (PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_Face) < 0) __PYX_ERR(0, 444, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_Face) < 0) __PYX_ERR(0, 444, __pyx_L1_error)
  #endif
  __pyx_vtabptr_9uharfbuzz_9_harfbuzz_Font = &__pyx_vtable_9uharfbuzz_9_harfbuzz_Font;
  __pyx_vtable_9uharfbuzz_9_harfbuzz_Font._Font__create = (PyObject *(*)(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Face *))__pyx_f_9uharfbuzz_9_harfbuzz_4Font__Font__create;
  __pyx_vtable_9uharfbuzz_9_harfbuzz_Font._Font__create_sub_font = (PyObject *(*)(struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *, struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *))__pyx_f_9uharfbuzz_9_harfbuzz_4Font__Font__create_sub_font;
  __pyx_vtable_9uharfbuzz_9_harfbuzz_Font.from_ptr = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_Font *(*)(hb_font_t *))__pyx_f_9uharfbuzz_9_harfbuzz_4Font_from_ptr;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_9uharfbuzz_9_harfbuzz_Font = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9uharfbuzz_9_harfbuzz_Font_spec, NULL); if (unlikely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_Font)) __PYX_ERR(0, 570, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9uharfbuzz_9_harfbuzz_Font_spec, __pyx_ptype_9uharfbuzz_9_harfbuzz_Font) < 0) __PYX_ERR(0, 570, __pyx_L1_error)
  #else
  __pyx_ptype_9uharfbuzz_9_harfbuzz_Font = &__pyx_type_9uharfbuzz_9_harfbuzz_Font;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9uharfbuzz_9_harfbuzz_Font) < 0) __PYX_ERR(0, 570, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9uharfbuzz_9_harfbuzz_Font->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_Font->tp_dictoffset && __pyx_ptype_9uharfbuzz_9_harfbuzz_Font->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9uharfbuzz_9_harfbuzz_Font->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_9uharfbuzz_9_harfbuzz_Font, __pyx_vtabptr_9uharfbuzz_9_harfbuzz_Font) < 0) __PYX_ERR(0, 570, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_9uharfbuzz_9_harfbuzz_Font) < 0) __PYX_ERR(0, 570, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Font, (PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_Font) < 0) __PYX_ERR(0, 570, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_Font) < 0) __PYX_ERR(0, 570, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9uharfbuzz_9_harfbuzz_FontFuncs_spec, NULL); if (unlikely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs)) __PYX_ERR(0, 1049, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9uharfbuzz_9_harfbuzz_FontFuncs_spec, __pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs) < 0) __PYX_ERR(0, 1049, __pyx_L1_error)
  #else
  __pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs = &__pyx_type_9uharfbuzz_9_harfbuzz_FontFuncs;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs) < 0) __PYX_ERR(0, 1049, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs->tp_dictoffset && __pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_FontFuncs, (PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs) < 0) __PYX_ERR(0, 1049, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs) < 0) __PYX_ERR(0, 1049, __pyx_L1_error)
  #endif
  __pyx_vtabptr_9uharfbuzz_9_harfbuzz_ColorLine = &__pyx_vtable_9uharfbuzz_9_harfbuzz_ColorLine;
  __pyx_vtable_9uharfbuzz_9_harfbuzz_ColorLine.from_ptr = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_ColorLine *(*)(hb_color_line_t *))__pyx_f_9uharfbuzz_9_harfbuzz_9ColorLine_from_ptr;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_9uharfbuzz_9_harfbuzz_ColorLine = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9uharfbuzz_9_harfbuzz_ColorLine_spec, NULL); if (unlikely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_ColorLine)) __PYX_ERR(0, 1684, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9uharfbuzz_9_harfbuzz_ColorLine_spec, __pyx_ptype_9uharfbuzz_9_harfbuzz_ColorLine) < 0) __PYX_ERR(0, 1684, __pyx_L1_error)
  #else
  __pyx_ptype_9uharfbuzz_9_harfbuzz_ColorLine = &__pyx_type_9uharfbuzz_9_harfbuzz_ColorLine;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9uharfbuzz_9_harfbuzz_ColorLine) < 0) __PYX_ERR(0, 1684, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9uharfbuzz_9_harfbuzz_ColorLine->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_ColorLine->tp_dictoffset && __pyx_ptype_9uharfbuzz_9_harfbuzz_ColorLine->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9uharfbuzz_9_harfbuzz_ColorLine->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_9uharfbuzz_9_harfbuzz_ColorLine, __pyx_vtabptr_9uharfbuzz_9_harfbuzz_ColorLine) < 0) __PYX_ERR(0, 1684, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_9uharfbuzz_9_harfbuzz_ColorLine) < 0) __PYX_ERR(0, 1684, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ColorLine, (PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_ColorLine) < 0) __PYX_ERR(0, 1684, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_ColorLine) < 0) __PYX_ERR(0, 1684, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9uharfbuzz_9_harfbuzz_PaintFuncs_spec, NULL); if (unlikely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs)) __PYX_ERR(0, 1896, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9uharfbuzz_9_harfbuzz_PaintFuncs_spec, __pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs) < 0) __PYX_ERR(0, 1896, __pyx_L1_error)
  #else
  __pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs = &__pyx_type_9uharfbuzz_9_harfbuzz_PaintFuncs;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs) < 0) __PYX_ERR(0, 1896, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs->tp_dictoffset && __pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PaintFuncs, (PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs) < 0) __PYX_ERR(0, 1896, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs) < 0) __PYX_ERR(0, 1896, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_9uharfbuzz_9_harfbuzz_DrawFuncs = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9uharfbuzz_9_harfbuzz_DrawFuncs_spec, NULL); if (unlikely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_DrawFuncs)) __PYX_ERR(0, 2123, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9uharfbuzz_9_harfbuzz_DrawFuncs_spec, __pyx_ptype_9uharfbuzz_9_harfbuzz_DrawFuncs) < 0) __PYX_ERR(0, 2123, __pyx_L1_error)
  #else
  __pyx_ptype_9uharfbuzz_9_harfbuzz_DrawFuncs = &__pyx_type_9uharfbuzz_9_harfbuzz_DrawFuncs;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9uharfbuzz_9_harfbuzz_DrawFuncs) < 0) __PYX_ERR(0, 2123, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9uharfbuzz_9_harfbuzz_DrawFuncs->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_DrawFuncs->tp_dictoffset && __pyx_ptype_9uharfbuzz_9_harfbuzz_DrawFuncs->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9uharfbuzz_9_harfbuzz_DrawFuncs->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_DrawFuncs, (PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_DrawFuncs) < 0) __PYX_ERR(0, 2123, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_DrawFuncs) < 0) __PYX_ERR(0, 2123, __pyx_L1_error)
  #endif
  __pyx_vtabptr_9uharfbuzz_9_harfbuzz_HBObject = &__pyx_vtable_9uharfbuzz_9_harfbuzz_HBObject;
  __pyx_vtable_9uharfbuzz_9_harfbuzz_HBObject.update_obj_length = (PyObject *(*)(struct __pyx_obj_9uharfbuzz_9_harfbuzz_HBObject *, unsigned int, char *, char *))__pyx_f_9uharfbuzz_9_harfbuzz_8HBObject_update_obj_length;
  __pyx_vtable_9uharfbuzz_9_harfbuzz_HBObject.create_links = (hb_link_t *(*)(struct __pyx_obj_9uharfbuzz_9_harfbuzz_HBObject *, unsigned int, unsigned int, int))__pyx_f_9uharfbuzz_9_harfbuzz_8HBObject_create_links;
  __pyx_vtable_9uharfbuzz_9_harfbuzz_HBObject.update_links = (PyObject *(*)(struct __pyx_obj_9uharfbuzz_9_harfbuzz_HBObject *, unsigned int, int, PyObject *))__pyx_f_9uharfbuzz_9_harfbuzz_8HBObject_update_links;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_9uharfbuzz_9_harfbuzz_HBObject = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9uharfbuzz_9_harfbuzz_HBObject_spec, NULL); if (unlikely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_HBObject)) __PYX_ERR(0, 2281, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9uharfbuzz_9_harfbuzz_HBObject_spec, __pyx_ptype_9uharfbuzz_9_harfbuzz_HBObject) < 0) __PYX_ERR(0, 2281, __pyx_L1_error)
  #else
  __pyx_ptype_9uharfbuzz_9_harfbuzz_HBObject = &__pyx_type_9uharfbuzz_9_harfbuzz_HBObject;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9uharfbuzz_9_harfbuzz_HBObject) < 0) __PYX_ERR(0, 2281, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9uharfbuzz_9_harfbuzz_HBObject->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_HBObject->tp_dictoffset && __pyx_ptype_9uharfbuzz_9_harfbuzz_HBObject->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9uharfbuzz_9_harfbuzz_HBObject->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_9uharfbuzz_9_harfbuzz_HBObject, __pyx_vtabptr_9uharfbuzz_9_harfbuzz_HBObject) < 0) __PYX_ERR(0, 2281, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_9uharfbuzz_9_harfbuzz_HBObject) < 0) __PYX_ERR(0, 2281, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_HBObject, (PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_HBObject) < 0) __PYX_ERR(0, 2281, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_HBObject) < 0) __PYX_ERR(0, 2281, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetInput = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9uharfbuzz_9_harfbuzz_SubsetInput_spec, NULL); if (unlikely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetInput)) __PYX_ERR(0, 2470, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9uharfbuzz_9_harfbuzz_SubsetInput_spec, __pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetInput) < 0) __PYX_ERR(0, 2470, __pyx_L1_error)
  #else
  __pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetInput = &__pyx_type_9uharfbuzz_9_harfbuzz_SubsetInput;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetInput) < 0) __PYX_ERR(0, 2470, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetInput->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetInput->tp_dictoffset && __pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetInput->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetInput->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SubsetInput, (PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetInput) < 0) __PYX_ERR(0, 2470, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetInput) < 0) __PYX_ERR(0, 2470, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetPlan = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9uharfbuzz_9_harfbuzz_SubsetPlan_spec, NULL); if (unlikely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetPlan)) __PYX_ERR(0, 2545, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9uharfbuzz_9_harfbuzz_SubsetPlan_spec, __pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetPlan) < 0) __PYX_ERR(0, 2545, __pyx_L1_error)
  #else
  __pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetPlan = &__pyx_type_9uharfbuzz_9_harfbuzz_SubsetPlan;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetPlan) < 0) __PYX_ERR(0, 2545, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetPlan->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetPlan->tp_dictoffset && __pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetPlan->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetPlan->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SubsetPlan, (PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetPlan) < 0) __PYX_ERR(0, 2545, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetPlan) < 0) __PYX_ERR(0, 2545, __pyx_L1_error)
  #endif
  __pyx_vtabptr_9uharfbuzz_9_harfbuzz_Set = &__pyx_vtable_9uharfbuzz_9_harfbuzz_Set;
  __pyx_vtable_9uharfbuzz_9_harfbuzz_Set.from_ptr = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_Set *(*)(hb_set_t *))__pyx_f_9uharfbuzz_9_harfbuzz_3Set_from_ptr;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_9uharfbuzz_9_harfbuzz_Set = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9uharfbuzz_9_harfbuzz_Set_spec, NULL); if (unlikely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_Set)) __PYX_ERR(0, 2576, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9uharfbuzz_9_harfbuzz_Set_spec, __pyx_ptype_9uharfbuzz_9_harfbuzz_Set) < 0) __PYX_ERR(0, 2576, __pyx_L1_error)
  #else
  __pyx_ptype_9uharfbuzz_9_harfbuzz_Set = &__pyx_type_9uharfbuzz_9_harfbuzz_Set;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9uharfbuzz_9_harfbuzz_Set) < 0) __PYX_ERR(0, 2576, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9uharfbuzz_9_harfbuzz_Set->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_Set->tp_dictoffset && __pyx_ptype_9uharfbuzz_9_harfbuzz_Set->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9uharfbuzz_9_harfbuzz_Set->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_9uharfbuzz_9_harfbuzz_Set, __pyx_vtabptr_9uharfbuzz_9_harfbuzz_Set) < 0) __PYX_ERR(0, 2576, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_9uharfbuzz_9_harfbuzz_Set) < 0) __PYX_ERR(0, 2576, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Set_2, (PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_Set) < 0) __PYX_ERR(0, 2576, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_Set) < 0) __PYX_ERR(0, 2576, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_9uharfbuzz_9_harfbuzz_SetIter = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9uharfbuzz_9_harfbuzz_SetIter_spec, NULL); if (unlikely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_SetIter)) __PYX_ERR(0, 2744, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9uharfbuzz_9_harfbuzz_SetIter_spec, __pyx_ptype_9uharfbuzz_9_harfbuzz_SetIter) < 0) __PYX_ERR(0, 2744, __pyx_L1_error)
  #else
  __pyx_ptype_9uharfbuzz_9_harfbuzz_SetIter = &__pyx_type_9uharfbuzz_9_harfbuzz_SetIter;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9uharfbuzz_9_harfbuzz_SetIter) < 0) __PYX_ERR(0, 2744, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9uharfbuzz_9_harfbuzz_SetIter->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_SetIter->tp_dictoffset && __pyx_ptype_9uharfbuzz_9_harfbuzz_SetIter->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9uharfbuzz_9_harfbuzz_SetIter->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_SetIter, (PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_SetIter) < 0) __PYX_ERR(0, 2744, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_SetIter) < 0) __PYX_ERR(0, 2744, __pyx_L1_error)
  #endif
  __pyx_vtabptr_9uharfbuzz_9_harfbuzz_Map = &__pyx_vtable_9uharfbuzz_9_harfbuzz_Map;
  __pyx_vtable_9uharfbuzz_9_harfbuzz_Map.from_ptr = (struct __pyx_obj_9uharfbuzz_9_harfbuzz_Map *(*)(hb_map_t *))__pyx_f_9uharfbuzz_9_harfbuzz_3Map_from_ptr;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_9uharfbuzz_9_harfbuzz_Map = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9uharfbuzz_9_harfbuzz_Map_spec, NULL); if (unlikely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_Map)) __PYX_ERR(0, 2764, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9uharfbuzz_9_harfbuzz_Map_spec, __pyx_ptype_9uharfbuzz_9_harfbuzz_Map) < 0) __PYX_ERR(0, 2764, __pyx_L1_error)
  #else
  __pyx_ptype_9uharfbuzz_9_harfbuzz_Map = &__pyx_type_9uharfbuzz_9_harfbuzz_Map;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9uharfbuzz_9_harfbuzz_Map) < 0) __PYX_ERR(0, 2764, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9uharfbuzz_9_harfbuzz_Map->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_Map->tp_dictoffset && __pyx_ptype_9uharfbuzz_9_harfbuzz_Map->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9uharfbuzz_9_harfbuzz_Map->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_9uharfbuzz_9_harfbuzz_Map, __pyx_vtabptr_9uharfbuzz_9_harfbuzz_Map) < 0) __PYX_ERR(0, 2764, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_9uharfbuzz_9_harfbuzz_Map) < 0) __PYX_ERR(0, 2764, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Map, (PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_Map) < 0) __PYX_ERR(0, 2764, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_Map) < 0) __PYX_ERR(0, 2764, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_9uharfbuzz_9_harfbuzz_MapIter = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9uharfbuzz_9_harfbuzz_MapIter_spec, NULL); if (unlikely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_MapIter)) __PYX_ERR(0, 2872, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9uharfbuzz_9_harfbuzz_MapIter_spec, __pyx_ptype_9uharfbuzz_9_harfbuzz_MapIter) < 0) __PYX_ERR(0, 2872, __pyx_L1_error)
  #else
  __pyx_ptype_9uharfbuzz_9_harfbuzz_MapIter = &__pyx_type_9uharfbuzz_9_harfbuzz_MapIter;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9uharfbuzz_9_harfbuzz_MapIter) < 0) __PYX_ERR(0, 2872, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9uharfbuzz_9_harfbuzz_MapIter->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9uharfbuzz_9_harfbuzz_MapIter->tp_dictoffset && __pyx_ptype_9uharfbuzz_9_harfbuzz_MapIter->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9uharfbuzz_9_harfbuzz_MapIter->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_MapIter, (PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_MapIter) < 0) __PYX_ERR(0, 2872, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_9uharfbuzz_9_harfbuzz_MapIter) < 0) __PYX_ERR(0, 2872, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr_spec, NULL); if (unlikely(!__pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr)) __PYX_ERR(0, 2741, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr_spec, __pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr) < 0) __PYX_ERR(0, 2741, __pyx_L1_error)
  #else
  __pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr = &__pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr) < 0) __PYX_ERR(0, 2741, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr->tp_dictoffset && __pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct__genexpr->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr_spec, NULL); if (unlikely(!__pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr)) __PYX_ERR(0, 2860, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr_spec, __pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr) < 0) __PYX_ERR(0, 2860, __pyx_L1_error)
  #else
  __pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr = &__pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr) < 0) __PYX_ERR(0, 2860, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr->tp_dictoffset && __pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_1_genexpr->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr_spec, NULL); if (unlikely(!__pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr)) __PYX_ERR(0, 2863, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr_spec, __pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr) < 0) __PYX_ERR(0, 2863, __pyx_L1_error)
  #else
  __pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr = &__pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr) < 0) __PYX_ERR(0, 2863, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr->tp_dictoffset && __pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_2_genexpr->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr_spec, NULL); if (unlikely(!__pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr)) __PYX_ERR(0, 2869, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr_spec, __pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr) < 0) __PYX_ERR(0, 2869, __pyx_L1_error)
  #else
  __pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr = &__pyx_type_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr) < 0) __PYX_ERR(0, 2869, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr->tp_dictoffset && __pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_9uharfbuzz_9_harfbuzz___pyx_scope_struct_3_genexpr->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec__harfbuzz(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec__harfbuzz},
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "_harfbuzz",
      0, /* m_doc */
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      0, /* m_size */
    #elif CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstate), /* m_size */
    #else
      -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif
#endif

#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC init_harfbuzz(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC init_harfbuzz(void)
#else
__Pyx_PyMODINIT_FUNC PyInit__harfbuzz(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit__harfbuzz(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *module, const char* from_name, const char* to_name, int allow_none)
#else
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
#endif
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
#if CYTHON_COMPILING_IN_LIMITED_API
            result = PyModule_AddObject(module, to_name, value);
#else
            result = PyDict_SetItemString(moddict, to_name, value);
#endif
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    moddict = module;
#else
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
#endif
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec__harfbuzz(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module '_harfbuzz' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("_harfbuzz", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #elif CYTHON_USE_MODULE_STATE
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  {
    int add_module_result = PyState_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "_harfbuzz" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #endif
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_cython_runtime = __Pyx_PyImport_AddModuleRef((const char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit__harfbuzz(void)", 0);
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_uharfbuzz___harfbuzz) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "uharfbuzz._harfbuzz")) {
      if (unlikely((PyDict_SetItemString(modules, "uharfbuzz._harfbuzz", __pyx_m) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely((__Pyx_modinit_type_init_code() < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_type_import_code();
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "uharfbuzz/_harfbuzz.pyx":3
 * #cython: language_level=3
 * cimport cython
 * import os             # <<<<<<<<<<<<<<
 * import warnings
 * from enum import IntEnum, IntFlag
 */
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_n_s_os, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_os, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":4
 * cimport cython
 * import os
 * import warnings             # <<<<<<<<<<<<<<
 * from enum import IntEnum, IntFlag
 * from .charfbuzz cimport *
 */
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_n_s_warnings, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_warnings, __pyx_t_2) < 0) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":5
 * import os
 * import warnings
 * from enum import IntEnum, IntFlag             # <<<<<<<<<<<<<<
 * from .charfbuzz cimport *
 * from libc.stdlib cimport free, malloc, calloc
 */
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_IntEnum);
  __Pyx_GIVEREF(__pyx_n_s_IntEnum);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_IntEnum)) __PYX_ERR(0, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_IntFlag);
  __Pyx_GIVEREF(__pyx_n_s_IntFlag);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_IntFlag)) __PYX_ERR(0, 5, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_enum, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_IntEnum); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_IntEnum, __pyx_t_2) < 0) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_IntFlag); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_IntFlag, __pyx_t_2) < 0) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":9
 * from libc.stdlib cimport free, malloc, calloc
 * from libc.string cimport const_char
 * from collections import namedtuple             # <<<<<<<<<<<<<<
 * from cpython.pycapsule cimport PyCapsule_GetPointer, PyCapsule_IsValid
 * from typing import Callable, Dict, List, Sequence, Tuple, Union
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_namedtuple);
  __Pyx_GIVEREF(__pyx_n_s_namedtuple);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_namedtuple)) __PYX_ERR(0, 9, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_collections, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_namedtuple); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_namedtuple, __pyx_t_3) < 0) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":11
 * from collections import namedtuple
 * from cpython.pycapsule cimport PyCapsule_GetPointer, PyCapsule_IsValid
 * from typing import Callable, Dict, List, Sequence, Tuple, Union             # <<<<<<<<<<<<<<
 * from pathlib import Path
 * 
 */
  __pyx_t_2 = PyList_New(6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_Callable);
  __Pyx_GIVEREF(__pyx_n_s_Callable);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_Callable)) __PYX_ERR(0, 11, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_Dict);
  __Pyx_GIVEREF(__pyx_n_s_Dict);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_Dict)) __PYX_ERR(0, 11, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_List);
  __Pyx_GIVEREF(__pyx_n_s_List);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 2, __pyx_n_s_List)) __PYX_ERR(0, 11, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_Sequence);
  __Pyx_GIVEREF(__pyx_n_s_Sequence);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 3, __pyx_n_s_Sequence)) __PYX_ERR(0, 11, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_Tuple);
  __Pyx_GIVEREF(__pyx_n_s_Tuple);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 4, __pyx_n_s_Tuple)) __PYX_ERR(0, 11, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_Union);
  __Pyx_GIVEREF(__pyx_n_s_Union);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 5, __pyx_n_s_Union)) __PYX_ERR(0, 11, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_typing, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_Callable); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Callable, __pyx_t_2) < 0) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_Dict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Dict, __pyx_t_2) < 0) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_List); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_List, __pyx_t_2) < 0) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_Sequence); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Sequence, __pyx_t_2) < 0) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_Tuple); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Tuple, __pyx_t_2) < 0) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_Union); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Union, __pyx_t_2) < 0) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":12
 * from cpython.pycapsule cimport PyCapsule_GetPointer, PyCapsule_IsValid
 * from typing import Callable, Dict, List, Sequence, Tuple, Union
 * from pathlib import Path             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_Path);
  __Pyx_GIVEREF(__pyx_n_s_Path);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_Path)) __PYX_ERR(0, 12, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_pathlib, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_Path); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Path, __pyx_t_3) < 0) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":41
 * 
 * 
 * def version_string() -> str:             # <<<<<<<<<<<<<<
 *     cdef const char* cstr = hb_version_string()
 *     cdef bytes packed = cstr
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_str) < 0) __PYX_ERR(0, 41, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_1version_string, 0, __pyx_n_s_version_string, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__25)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_version_string, __pyx_t_3) < 0) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":47
 * 
 * 
 * class HarfBuzzError(Exception):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0]));
  __Pyx_GIVEREF((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0]));
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])))) __PYX_ERR(0, 47, __pyx_L1_error);
  __pyx_t_2 = __Pyx_PEP560_update_bases(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_Py3MetaclassPrepare(__pyx_t_4, __pyx_t_2, __pyx_n_s_HarfBuzzError, __pyx_n_s_HarfBuzzError, (PyObject *) NULL, __pyx_n_s_uharfbuzz__harfbuzz, (PyObject *) NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__pyx_t_2 != __pyx_t_3) {
    if (unlikely((PyDict_SetItemString(__pyx_t_5, "__orig_bases__", __pyx_t_3) < 0))) __PYX_ERR(0, 47, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_Py3ClassCreate(__pyx_t_4, __pyx_n_s_HarfBuzzError, __pyx_t_2, __pyx_t_5, NULL, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_HarfBuzzError, __pyx_t_3) < 0) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":51
 * 
 * 
 * class GlyphFlags(IntFlag):             # <<<<<<<<<<<<<<
 *     UNSAFE_TO_BREAK = HB_GLYPH_FLAG_UNSAFE_TO_BREAK
 *     UNSAFE_TO_CONCAT = HB_GLYPH_FLAG_UNSAFE_TO_CONCAT
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_IntFlag); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2)) __PYX_ERR(0, 51, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PEP560_update_bases(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_2, __pyx_n_s_GlyphFlags, __pyx_n_s_GlyphFlags, (PyObject *) NULL, __pyx_n_s_uharfbuzz__harfbuzz, (PyObject *) NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__pyx_t_2 != __pyx_t_4) {
    if (unlikely((PyDict_SetItemString(__pyx_t_3, "__orig_bases__", __pyx_t_4) < 0))) __PYX_ERR(0, 51, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":52
 * 
 * class GlyphFlags(IntFlag):
 *     UNSAFE_TO_BREAK = HB_GLYPH_FLAG_UNSAFE_TO_BREAK             # <<<<<<<<<<<<<<
 *     UNSAFE_TO_CONCAT = HB_GLYPH_FLAG_UNSAFE_TO_CONCAT
 *     SAFE_TO_INSERT_TATWEEL = HB_GLYPH_FLAG_SAFE_TO_INSERT_TATWEEL
 */
  __pyx_t_4 = __Pyx_PyInt_From_hb_glyph_flags_t(HB_GLYPH_FLAG_UNSAFE_TO_BREAK); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_UNSAFE_TO_BREAK, __pyx_t_4) < 0) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":53
 * class GlyphFlags(IntFlag):
 *     UNSAFE_TO_BREAK = HB_GLYPH_FLAG_UNSAFE_TO_BREAK
 *     UNSAFE_TO_CONCAT = HB_GLYPH_FLAG_UNSAFE_TO_CONCAT             # <<<<<<<<<<<<<<
 *     SAFE_TO_INSERT_TATWEEL = HB_GLYPH_FLAG_SAFE_TO_INSERT_TATWEEL
 * 
 */
  __pyx_t_4 = __Pyx_PyInt_From_hb_glyph_flags_t(HB_GLYPH_FLAG_UNSAFE_TO_CONCAT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_UNSAFE_TO_CONCAT, __pyx_t_4) < 0) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":54
 *     UNSAFE_TO_BREAK = HB_GLYPH_FLAG_UNSAFE_TO_BREAK
 *     UNSAFE_TO_CONCAT = HB_GLYPH_FLAG_UNSAFE_TO_CONCAT
 *     SAFE_TO_INSERT_TATWEEL = HB_GLYPH_FLAG_SAFE_TO_INSERT_TATWEEL             # <<<<<<<<<<<<<<
 * 
 * cdef class GlyphInfo:
 */
  __pyx_t_4 = __Pyx_PyInt_From_hb_glyph_flags_t(HB_GLYPH_FLAG_SAFE_TO_INSERT_TATWEEL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_SAFE_TO_INSERT_TATWEEL, __pyx_t_4) < 0) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":51
 * 
 * 
 * class GlyphFlags(IntFlag):             # <<<<<<<<<<<<<<
 *     UNSAFE_TO_BREAK = HB_GLYPH_FLAG_UNSAFE_TO_BREAK
 *     UNSAFE_TO_CONCAT = HB_GLYPH_FLAG_UNSAFE_TO_CONCAT
 */
  __pyx_t_4 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_GlyphFlags, __pyx_t_2, __pyx_t_3, NULL, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_GlyphFlags, __pyx_t_4) < 0) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self._hb_glyph_info must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_9GlyphInfo_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlyphInfo___reduce_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__27)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self._hb_glyph_info must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self._hb_glyph_info must be explicitly requested with @auto_pickle(True)"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_9GlyphInfo_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlyphInfo___setstate_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__29)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self._hb_glyph_position must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_13GlyphPosition_1__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlyphPosition___reduce_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__30)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "Pickling of struct members such as self._hb_glyph_position must be explicitly requested with @auto_pickle(True)"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "Pickling of struct members such as self._hb_glyph_position must be explicitly requested with @auto_pickle(True)"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_13GlyphPosition_3__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_GlyphPosition___setstate_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__31)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":110
 * 
 * 
 * class BufferFlags(IntFlag):             # <<<<<<<<<<<<<<
 *     DEFAULT = HB_BUFFER_FLAG_DEFAULT
 *     BOT = HB_BUFFER_FLAG_BOT
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_IntFlag); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2)) __PYX_ERR(0, 110, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PEP560_update_bases(__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_Py3MetaclassPrepare(__pyx_t_3, __pyx_t_2, __pyx_n_s_BufferFlags, __pyx_n_s_BufferFlags, (PyObject *) NULL, __pyx_n_s_uharfbuzz__harfbuzz, (PyObject *) NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__pyx_t_2 != __pyx_t_5) {
    if (unlikely((PyDict_SetItemString(__pyx_t_4, "__orig_bases__", __pyx_t_5) < 0))) __PYX_ERR(0, 110, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":111
 * 
 * class BufferFlags(IntFlag):
 *     DEFAULT = HB_BUFFER_FLAG_DEFAULT             # <<<<<<<<<<<<<<
 *     BOT = HB_BUFFER_FLAG_BOT
 *     EOT = HB_BUFFER_FLAG_EOT
 */
  __pyx_t_5 = __Pyx_PyInt_From_hb_buffer_flags_t(HB_BUFFER_FLAG_DEFAULT); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_DEFAULT, __pyx_t_5) < 0) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":112
 * class BufferFlags(IntFlag):
 *     DEFAULT = HB_BUFFER_FLAG_DEFAULT
 *     BOT = HB_BUFFER_FLAG_BOT             # <<<<<<<<<<<<<<
 *     EOT = HB_BUFFER_FLAG_EOT
 *     PRESERVE_DEFAULT_IGNORABLES = HB_BUFFER_FLAG_PRESERVE_DEFAULT_IGNORABLES
 */
  __pyx_t_5 = __Pyx_PyInt_From_hb_buffer_flags_t(HB_BUFFER_FLAG_BOT); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_BOT, __pyx_t_5) < 0) __PYX_ERR(0, 112, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":113
 *     DEFAULT = HB_BUFFER_FLAG_DEFAULT
 *     BOT = HB_BUFFER_FLAG_BOT
 *     EOT = HB_BUFFER_FLAG_EOT             # <<<<<<<<<<<<<<
 *     PRESERVE_DEFAULT_IGNORABLES = HB_BUFFER_FLAG_PRESERVE_DEFAULT_IGNORABLES
 *     REMOVE_DEFAULT_IGNORABLES = HB_BUFFER_FLAG_REMOVE_DEFAULT_IGNORABLES
 */
  __pyx_t_5 = __Pyx_PyInt_From_hb_buffer_flags_t(HB_BUFFER_FLAG_EOT); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_EOT, __pyx_t_5) < 0) __PYX_ERR(0, 113, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":114
 *     BOT = HB_BUFFER_FLAG_BOT
 *     EOT = HB_BUFFER_FLAG_EOT
 *     PRESERVE_DEFAULT_IGNORABLES = HB_BUFFER_FLAG_PRESERVE_DEFAULT_IGNORABLES             # <<<<<<<<<<<<<<
 *     REMOVE_DEFAULT_IGNORABLES = HB_BUFFER_FLAG_REMOVE_DEFAULT_IGNORABLES
 *     DO_NOT_INSERT_DOTTED_CIRCLE = HB_BUFFER_FLAG_DO_NOT_INSERT_DOTTED_CIRCLE
 */
  __pyx_t_5 = __Pyx_PyInt_From_hb_buffer_flags_t(HB_BUFFER_FLAG_PRESERVE_DEFAULT_IGNORABLES); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_PRESERVE_DEFAULT_IGNORABLES, __pyx_t_5) < 0) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":115
 *     EOT = HB_BUFFER_FLAG_EOT
 *     PRESERVE_DEFAULT_IGNORABLES = HB_BUFFER_FLAG_PRESERVE_DEFAULT_IGNORABLES
 *     REMOVE_DEFAULT_IGNORABLES = HB_BUFFER_FLAG_REMOVE_DEFAULT_IGNORABLES             # <<<<<<<<<<<<<<
 *     DO_NOT_INSERT_DOTTED_CIRCLE = HB_BUFFER_FLAG_DO_NOT_INSERT_DOTTED_CIRCLE
 *     VERIFY = HB_BUFFER_FLAG_VERIFY
 */
  __pyx_t_5 = __Pyx_PyInt_From_hb_buffer_flags_t(HB_BUFFER_FLAG_REMOVE_DEFAULT_IGNORABLES); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_REMOVE_DEFAULT_IGNORABLES, __pyx_t_5) < 0) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":116
 *     PRESERVE_DEFAULT_IGNORABLES = HB_BUFFER_FLAG_PRESERVE_DEFAULT_IGNORABLES
 *     REMOVE_DEFAULT_IGNORABLES = HB_BUFFER_FLAG_REMOVE_DEFAULT_IGNORABLES
 *     DO_NOT_INSERT_DOTTED_CIRCLE = HB_BUFFER_FLAG_DO_NOT_INSERT_DOTTED_CIRCLE             # <<<<<<<<<<<<<<
 *     VERIFY = HB_BUFFER_FLAG_VERIFY
 *     PRODUCE_UNSAFE_TO_CONCAT = HB_BUFFER_FLAG_PRODUCE_UNSAFE_TO_CONCAT
 */
  __pyx_t_5 = __Pyx_PyInt_From_hb_buffer_flags_t(HB_BUFFER_FLAG_DO_NOT_INSERT_DOTTED_CIRCLE); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_DO_NOT_INSERT_DOTTED_CIRCLE, __pyx_t_5) < 0) __PYX_ERR(0, 116, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":117
 *     REMOVE_DEFAULT_IGNORABLES = HB_BUFFER_FLAG_REMOVE_DEFAULT_IGNORABLES
 *     DO_NOT_INSERT_DOTTED_CIRCLE = HB_BUFFER_FLAG_DO_NOT_INSERT_DOTTED_CIRCLE
 *     VERIFY = HB_BUFFER_FLAG_VERIFY             # <<<<<<<<<<<<<<
 *     PRODUCE_UNSAFE_TO_CONCAT = HB_BUFFER_FLAG_PRODUCE_UNSAFE_TO_CONCAT
 *     PRODUCE_SAFE_TO_INSERT_TATWEEL = HB_BUFFER_FLAG_PRODUCE_SAFE_TO_INSERT_TATWEEL
 */
  __pyx_t_5 = __Pyx_PyInt_From_hb_buffer_flags_t(HB_BUFFER_FLAG_VERIFY); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_VERIFY, __pyx_t_5) < 0) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":118
 *     DO_NOT_INSERT_DOTTED_CIRCLE = HB_BUFFER_FLAG_DO_NOT_INSERT_DOTTED_CIRCLE
 *     VERIFY = HB_BUFFER_FLAG_VERIFY
 *     PRODUCE_UNSAFE_TO_CONCAT = HB_BUFFER_FLAG_PRODUCE_UNSAFE_TO_CONCAT             # <<<<<<<<<<<<<<
 *     PRODUCE_SAFE_TO_INSERT_TATWEEL = HB_BUFFER_FLAG_PRODUCE_SAFE_TO_INSERT_TATWEEL
 * 
 */
  __pyx_t_5 = __Pyx_PyInt_From_hb_buffer_flags_t(HB_BUFFER_FLAG_PRODUCE_UNSAFE_TO_CONCAT); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_PRODUCE_UNSAFE_TO_CONCAT, __pyx_t_5) < 0) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":119
 *     VERIFY = HB_BUFFER_FLAG_VERIFY
 *     PRODUCE_UNSAFE_TO_CONCAT = HB_BUFFER_FLAG_PRODUCE_UNSAFE_TO_CONCAT
 *     PRODUCE_SAFE_TO_INSERT_TATWEEL = HB_BUFFER_FLAG_PRODUCE_SAFE_TO_INSERT_TATWEEL             # <<<<<<<<<<<<<<
 * 
 * class BufferClusterLevel(IntEnum):
 */
  __pyx_t_5 = __Pyx_PyInt_From_hb_buffer_flags_t(HB_BUFFER_FLAG_PRODUCE_SAFE_TO_INSERT_TATWEEL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_PRODUCE_SAFE_TO_INSERT_TATWEEL, __pyx_t_5) < 0) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":110
 * 
 * 
 * class BufferFlags(IntFlag):             # <<<<<<<<<<<<<<
 *     DEFAULT = HB_BUFFER_FLAG_DEFAULT
 *     BOT = HB_BUFFER_FLAG_BOT
 */
  __pyx_t_5 = __Pyx_Py3ClassCreate(__pyx_t_3, __pyx_n_s_BufferFlags, __pyx_t_2, __pyx_t_4, NULL, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_BufferFlags, __pyx_t_5) < 0) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":121
 *     PRODUCE_SAFE_TO_INSERT_TATWEEL = HB_BUFFER_FLAG_PRODUCE_SAFE_TO_INSERT_TATWEEL
 * 
 * class BufferClusterLevel(IntEnum):             # <<<<<<<<<<<<<<
 *     MONOTONE_GRAPHEMES = HB_BUFFER_CLUSTER_LEVEL_MONOTONE_GRAPHEMES
 *     MONOTONE_CHARACTERS = HB_BUFFER_CLUSTER_LEVEL_MONOTONE_CHARACTERS
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_IntEnum); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2)) __PYX_ERR(0, 121, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PEP560_update_bases(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_Py3MetaclassPrepare(__pyx_t_4, __pyx_t_2, __pyx_n_s_BufferClusterLevel, __pyx_n_s_BufferClusterLevel, (PyObject *) NULL, __pyx_n_s_uharfbuzz__harfbuzz, (PyObject *) NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__pyx_t_2 != __pyx_t_3) {
    if (unlikely((PyDict_SetItemString(__pyx_t_5, "__orig_bases__", __pyx_t_3) < 0))) __PYX_ERR(0, 121, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":122
 * 
 * class BufferClusterLevel(IntEnum):
 *     MONOTONE_GRAPHEMES = HB_BUFFER_CLUSTER_LEVEL_MONOTONE_GRAPHEMES             # <<<<<<<<<<<<<<
 *     MONOTONE_CHARACTERS = HB_BUFFER_CLUSTER_LEVEL_MONOTONE_CHARACTERS
 *     CHARACTERS = HB_BUFFER_CLUSTER_LEVEL_CHARACTERS
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_buffer_cluster_level_t(HB_BUFFER_CLUSTER_LEVEL_MONOTONE_GRAPHEMES); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_MONOTONE_GRAPHEMES, __pyx_t_3) < 0) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":123
 * class BufferClusterLevel(IntEnum):
 *     MONOTONE_GRAPHEMES = HB_BUFFER_CLUSTER_LEVEL_MONOTONE_GRAPHEMES
 *     MONOTONE_CHARACTERS = HB_BUFFER_CLUSTER_LEVEL_MONOTONE_CHARACTERS             # <<<<<<<<<<<<<<
 *     CHARACTERS = HB_BUFFER_CLUSTER_LEVEL_CHARACTERS
 *     DEFAULT = HB_BUFFER_CLUSTER_LEVEL_DEFAULT
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_buffer_cluster_level_t(HB_BUFFER_CLUSTER_LEVEL_MONOTONE_CHARACTERS); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_MONOTONE_CHARACTERS, __pyx_t_3) < 0) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":124
 *     MONOTONE_GRAPHEMES = HB_BUFFER_CLUSTER_LEVEL_MONOTONE_GRAPHEMES
 *     MONOTONE_CHARACTERS = HB_BUFFER_CLUSTER_LEVEL_MONOTONE_CHARACTERS
 *     CHARACTERS = HB_BUFFER_CLUSTER_LEVEL_CHARACTERS             # <<<<<<<<<<<<<<
 *     DEFAULT = HB_BUFFER_CLUSTER_LEVEL_DEFAULT
 * 
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_buffer_cluster_level_t(HB_BUFFER_CLUSTER_LEVEL_CHARACTERS); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_CHARACTERS, __pyx_t_3) < 0) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":125
 *     MONOTONE_CHARACTERS = HB_BUFFER_CLUSTER_LEVEL_MONOTONE_CHARACTERS
 *     CHARACTERS = HB_BUFFER_CLUSTER_LEVEL_CHARACTERS
 *     DEFAULT = HB_BUFFER_CLUSTER_LEVEL_DEFAULT             # <<<<<<<<<<<<<<
 * 
 * class BufferContentType(IntEnum):
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_buffer_cluster_level_t(HB_BUFFER_CLUSTER_LEVEL_DEFAULT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_DEFAULT, __pyx_t_3) < 0) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":121
 *     PRODUCE_SAFE_TO_INSERT_TATWEEL = HB_BUFFER_FLAG_PRODUCE_SAFE_TO_INSERT_TATWEEL
 * 
 * class BufferClusterLevel(IntEnum):             # <<<<<<<<<<<<<<
 *     MONOTONE_GRAPHEMES = HB_BUFFER_CLUSTER_LEVEL_MONOTONE_GRAPHEMES
 *     MONOTONE_CHARACTERS = HB_BUFFER_CLUSTER_LEVEL_MONOTONE_CHARACTERS
 */
  __pyx_t_3 = __Pyx_Py3ClassCreate(__pyx_t_4, __pyx_n_s_BufferClusterLevel, __pyx_t_2, __pyx_t_5, NULL, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_BufferClusterLevel, __pyx_t_3) < 0) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":127
 *     DEFAULT = HB_BUFFER_CLUSTER_LEVEL_DEFAULT
 * 
 * class BufferContentType(IntEnum):             # <<<<<<<<<<<<<<
 *     INVALID = HB_BUFFER_CONTENT_TYPE_INVALID
 *     UNICODE = HB_BUFFER_CONTENT_TYPE_UNICODE
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_IntEnum); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_2)) __PYX_ERR(0, 127, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PEP560_update_bases(__pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_2, __pyx_n_s_BufferContentType, __pyx_n_s_BufferContentType, (PyObject *) NULL, __pyx_n_s_uharfbuzz__harfbuzz, (PyObject *) NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__pyx_t_2 != __pyx_t_4) {
    if (unlikely((PyDict_SetItemString(__pyx_t_3, "__orig_bases__", __pyx_t_4) < 0))) __PYX_ERR(0, 127, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":128
 * 
 * class BufferContentType(IntEnum):
 *     INVALID = HB_BUFFER_CONTENT_TYPE_INVALID             # <<<<<<<<<<<<<<
 *     UNICODE = HB_BUFFER_CONTENT_TYPE_UNICODE
 *     GLYPHS = HB_BUFFER_CONTENT_TYPE_GLYPHS
 */
  __pyx_t_4 = __Pyx_PyInt_From_hb_buffer_content_type_t(HB_BUFFER_CONTENT_TYPE_INVALID); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_INVALID, __pyx_t_4) < 0) __PYX_ERR(0, 128, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":129
 * class BufferContentType(IntEnum):
 *     INVALID = HB_BUFFER_CONTENT_TYPE_INVALID
 *     UNICODE = HB_BUFFER_CONTENT_TYPE_UNICODE             # <<<<<<<<<<<<<<
 *     GLYPHS = HB_BUFFER_CONTENT_TYPE_GLYPHS
 * 
 */
  __pyx_t_4 = __Pyx_PyInt_From_hb_buffer_content_type_t(HB_BUFFER_CONTENT_TYPE_UNICODE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_UNICODE, __pyx_t_4) < 0) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":130
 *     INVALID = HB_BUFFER_CONTENT_TYPE_INVALID
 *     UNICODE = HB_BUFFER_CONTENT_TYPE_UNICODE
 *     GLYPHS = HB_BUFFER_CONTENT_TYPE_GLYPHS             # <<<<<<<<<<<<<<
 * 
 * cdef class Buffer:
 */
  __pyx_t_4 = __Pyx_PyInt_From_hb_buffer_content_type_t(HB_BUFFER_CONTENT_TYPE_GLYPHS); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_GLYPHS, __pyx_t_4) < 0) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":127
 *     DEFAULT = HB_BUFFER_CLUSTER_LEVEL_DEFAULT
 * 
 * class BufferContentType(IntEnum):             # <<<<<<<<<<<<<<
 *     INVALID = HB_BUFFER_CONTENT_TYPE_INVALID
 *     UNICODE = HB_BUFFER_CONTENT_TYPE_UNICODE
 */
  __pyx_t_4 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_BufferContentType, __pyx_t_2, __pyx_t_3, NULL, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_BufferContentType, __pyx_t_4) < 0) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":136
 *     cdef object _message_callback
 * 
 *     DEFAULT_REPLACEMENT_CODEPOINT = HB_BUFFER_REPLACEMENT_CODEPOINT_DEFAULT             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self):
 */
  __pyx_t_2 = __Pyx_PyInt_From_hb_codepoint_t(HB_BUFFER_REPLACEMENT_CODEPOINT_DEFAULT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer, __pyx_n_s_DEFAULT_REPLACEMENT_CODEPOINT, __pyx_t_2) < 0) __PYX_ERR(0, 136, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer);

  /* "uharfbuzz/_harfbuzz.pyx":148
 * 
 *     # DEPRECATED: use the normal constructor
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def create(cls):
 *         cdef Buffer inst = cls()
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_6Buffer_5create, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Buffer_create, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__33)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer, __pyx_n_s_create, __pyx_t_2) < 0) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer);
  __Pyx_GetNameInClass(__pyx_t_2, (PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer, __pyx_n_s_create); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_Method_ClassMethod(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer, __pyx_n_s_create, __pyx_t_5) < 0) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer);

  /* "uharfbuzz/_harfbuzz.pyx":156
 *         return hb_buffer_get_length(self._hb_buffer)
 * 
 *     def reset(self):             # <<<<<<<<<<<<<<
 *         hb_buffer_reset (self._hb_buffer)
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_6Buffer_9reset, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Buffer_reset, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__34)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer, __pyx_n_s_reset, __pyx_t_5) < 0) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer);

  /* "uharfbuzz/_harfbuzz.pyx":159
 *         hb_buffer_reset (self._hb_buffer)
 * 
 *     def clear_contents(self):             # <<<<<<<<<<<<<<
 *         hb_buffer_clear_contents(self._hb_buffer)
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_6Buffer_11clear_contents, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Buffer_clear_contents, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__35)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer, __pyx_n_s_clear_contents, __pyx_t_5) < 0) __PYX_ERR(0, 159, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer);

  /* "uharfbuzz/_harfbuzz.pyx":295
 *         hb_buffer_set_not_found_glyph(self._hb_buffer, value)
 * 
 *     def set_language_from_ot_tag(self, value: str):             # <<<<<<<<<<<<<<
 *         cdef bytes packed = value.encode()
 *         cdef char* cstr = packed
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_value, __pyx_n_s_str) < 0) __PYX_ERR(0, 295, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_6Buffer_13set_language_from_ot_tag, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Buffer_set_language_from_ot_tag, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__37)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer, __pyx_n_s_set_language_from_ot_tag, __pyx_t_2) < 0) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer);

  /* "uharfbuzz/_harfbuzz.pyx":301
 *             self._hb_buffer, hb_ot_tag_to_language(hb_tag_from_string(cstr, -1)))
 * 
 *     def set_script_from_ot_tag(self, value: str):             # <<<<<<<<<<<<<<
 *         cdef bytes packed = value.encode()
 *         cdef char* cstr = packed
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_value, __pyx_n_s_str) < 0) __PYX_ERR(0, 301, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_6Buffer_15set_script_from_ot_tag, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Buffer_set_script_from_ot_tag, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__38)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer, __pyx_n_s_set_script_from_ot_tag, __pyx_t_5) < 0) __PYX_ERR(0, 301, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer);

  /* "uharfbuzz/_harfbuzz.pyx":308
 * 
 *     def add_codepoints(self, codepoints: List[int],
 *                        item_offset: int = 0, item_length: int = -1) -> None:             # <<<<<<<<<<<<<<
 *         cdef unsigned int size = len(codepoints)
 *         cdef hb_codepoint_t* hb_codepoints
 */
  if (!(likely(__Pyx_Py3Int_CheckExact(__pyx_int_0)) || __Pyx_RaiseUnexpectedTypeError("int", __pyx_int_0))) __PYX_ERR(0, 308, __pyx_L1_error)
  __Pyx_INCREF(__pyx_int_0);
  __pyx_k_ = ((PyObject*)__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  if (!(likely(__Pyx_Py3Int_CheckExact(__pyx_int_neg_1)) || __Pyx_RaiseUnexpectedTypeError("int", __pyx_int_neg_1))) __PYX_ERR(0, 308, __pyx_L1_error)
  __Pyx_INCREF(__pyx_int_neg_1);
  __pyx_k__2 = ((PyObject*)__pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);

  /* "uharfbuzz/_harfbuzz.pyx":307
 *             self._hb_buffer, hb_ot_tag_to_script(hb_tag_from_string(cstr, -1)))
 * 
 *     def add_codepoints(self, codepoints: List[int],             # <<<<<<<<<<<<<<
 *                        item_offset: int = 0, item_length: int = -1) -> None:
 *         cdef unsigned int size = len(codepoints)
 */
  if (!(likely(__Pyx_Py3Int_CheckExact(__pyx_int_0)) || __Pyx_RaiseUnexpectedTypeError("int", __pyx_int_0))) __PYX_ERR(0, 308, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":308
 * 
 *     def add_codepoints(self, codepoints: List[int],
 *                        item_offset: int = 0, item_length: int = -1) -> None:             # <<<<<<<<<<<<<<
 *         cdef unsigned int size = len(codepoints)
 *         cdef hb_codepoint_t* hb_codepoints
 */
  if (!(likely(__Pyx_Py3Int_CheckExact(__pyx_int_neg_1)) || __Pyx_RaiseUnexpectedTypeError("int", __pyx_int_neg_1))) __PYX_ERR(0, 308, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":307
 *             self._hb_buffer, hb_ot_tag_to_script(hb_tag_from_string(cstr, -1)))
 * 
 *     def add_codepoints(self, codepoints: List[int],             # <<<<<<<<<<<<<<
 *                        item_offset: int = 0, item_length: int = -1) -> None:
 *         cdef unsigned int size = len(codepoints)
 */
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_int_0)) __PYX_ERR(0, 307, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_neg_1)) __PYX_ERR(0, 307, __pyx_L1_error);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_codepoints, __pyx_kp_s_List_int) < 0) __PYX_ERR(0, 307, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_item_offset, __pyx_n_s_int) < 0) __PYX_ERR(0, 307, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_item_length, __pyx_n_s_int) < 0) __PYX_ERR(0, 307, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 307, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_6Buffer_17add_codepoints, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Buffer_add_codepoints, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__40)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer, __pyx_n_s_add_codepoints, __pyx_t_3) < 0) __PYX_ERR(0, 307, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer);

  /* "uharfbuzz/_harfbuzz.pyx":324
 * 
 *     def add_utf8(self, text: bytes,
 *                  item_offset: int = 0, item_length: int = -1) -> None:             # <<<<<<<<<<<<<<
 *         hb_buffer_add_utf8(
 *             self._hb_buffer, text, len(text), item_offset, item_length)
 */
  if (!(likely(__Pyx_Py3Int_CheckExact(__pyx_int_0)) || __Pyx_RaiseUnexpectedTypeError("int", __pyx_int_0))) __PYX_ERR(0, 324, __pyx_L1_error)
  __Pyx_INCREF(__pyx_int_0);
  __pyx_k__3 = ((PyObject*)__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  if (!(likely(__Pyx_Py3Int_CheckExact(__pyx_int_neg_1)) || __Pyx_RaiseUnexpectedTypeError("int", __pyx_int_neg_1))) __PYX_ERR(0, 324, __pyx_L1_error)
  __Pyx_INCREF(__pyx_int_neg_1);
  __pyx_k__4 = ((PyObject*)__pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);

  /* "uharfbuzz/_harfbuzz.pyx":323
 *             raise MemoryError()
 * 
 *     def add_utf8(self, text: bytes,             # <<<<<<<<<<<<<<
 *                  item_offset: int = 0, item_length: int = -1) -> None:
 *         hb_buffer_add_utf8(
 */
  if (!(likely(__Pyx_Py3Int_CheckExact(__pyx_int_0)) || __Pyx_RaiseUnexpectedTypeError("int", __pyx_int_0))) __PYX_ERR(0, 324, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":324
 * 
 *     def add_utf8(self, text: bytes,
 *                  item_offset: int = 0, item_length: int = -1) -> None:             # <<<<<<<<<<<<<<
 *         hb_buffer_add_utf8(
 *             self._hb_buffer, text, len(text), item_offset, item_length)
 */
  if (!(likely(__Pyx_Py3Int_CheckExact(__pyx_int_neg_1)) || __Pyx_RaiseUnexpectedTypeError("int", __pyx_int_neg_1))) __PYX_ERR(0, 324, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":323
 *             raise MemoryError()
 * 
 *     def add_utf8(self, text: bytes,             # <<<<<<<<<<<<<<
 *                  item_offset: int = 0, item_length: int = -1) -> None:
 *         hb_buffer_add_utf8(
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_int_0)) __PYX_ERR(0, 323, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_neg_1)) __PYX_ERR(0, 323, __pyx_L1_error);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_text, __pyx_n_s_bytes) < 0) __PYX_ERR(0, 323, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_item_offset, __pyx_n_s_int) < 0) __PYX_ERR(0, 323, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_item_length, __pyx_n_s_int) < 0) __PYX_ERR(0, 323, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 323, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_6Buffer_19add_utf8, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Buffer_add_utf8, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__42)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer, __pyx_n_s_add_utf8, __pyx_t_5) < 0) __PYX_ERR(0, 323, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer);

  /* "uharfbuzz/_harfbuzz.pyx":331
 * 
 *     def add_str(self, text: str,
 *                 item_offset: int = 0, item_length: int = -1) -> None:             # <<<<<<<<<<<<<<
 *         # ensure unicode string is in the "canonical" representation
 *         assert PyUnicode_IS_READY(text)
 */
  if (!(likely(__Pyx_Py3Int_CheckExact(__pyx_int_0)) || __Pyx_RaiseUnexpectedTypeError("int", __pyx_int_0))) __PYX_ERR(0, 331, __pyx_L1_error)
  __Pyx_INCREF(__pyx_int_0);
  __pyx_k__5 = ((PyObject*)__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  if (!(likely(__Pyx_Py3Int_CheckExact(__pyx_int_neg_1)) || __Pyx_RaiseUnexpectedTypeError("int", __pyx_int_neg_1))) __PYX_ERR(0, 331, __pyx_L1_error)
  __Pyx_INCREF(__pyx_int_neg_1);
  __pyx_k__6 = ((PyObject*)__pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);

  /* "uharfbuzz/_harfbuzz.pyx":330
 *             raise MemoryError()
 * 
 *     def add_str(self, text: str,             # <<<<<<<<<<<<<<
 *                 item_offset: int = 0, item_length: int = -1) -> None:
 *         # ensure unicode string is in the "canonical" representation
 */
  if (!(likely(__Pyx_Py3Int_CheckExact(__pyx_int_0)) || __Pyx_RaiseUnexpectedTypeError("int", __pyx_int_0))) __PYX_ERR(0, 331, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":331
 * 
 *     def add_str(self, text: str,
 *                 item_offset: int = 0, item_length: int = -1) -> None:             # <<<<<<<<<<<<<<
 *         # ensure unicode string is in the "canonical" representation
 *         assert PyUnicode_IS_READY(text)
 */
  if (!(likely(__Pyx_Py3Int_CheckExact(__pyx_int_neg_1)) || __Pyx_RaiseUnexpectedTypeError("int", __pyx_int_neg_1))) __PYX_ERR(0, 331, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":330
 *             raise MemoryError()
 * 
 *     def add_str(self, text: str,             # <<<<<<<<<<<<<<
 *                 item_offset: int = 0, item_length: int = -1) -> None:
 *         # ensure unicode string is in the "canonical" representation
 */
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_int_0)) __PYX_ERR(0, 330, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_neg_1)) __PYX_ERR(0, 330, __pyx_L1_error);
  __pyx_t_2 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_text, __pyx_n_s_str) < 0) __PYX_ERR(0, 330, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_item_offset, __pyx_n_s_int) < 0) __PYX_ERR(0, 330, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_item_length, __pyx_n_s_int) < 0) __PYX_ERR(0, 330, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 330, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_6Buffer_21add_str, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Buffer_add_str, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__44)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 330, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer, __pyx_n_s_add_str, __pyx_t_3) < 0) __PYX_ERR(0, 330, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer);

  /* "uharfbuzz/_harfbuzz.pyx":367
 *             raise MemoryError()
 * 
 *     def guess_segment_properties(self) -> None:             # <<<<<<<<<<<<<<
 *         hb_buffer_guess_segment_properties(self._hb_buffer)
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 367, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_6Buffer_23guess_segment_properties, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Buffer_guess_segment_properties, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__45)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer, __pyx_n_s_guess_segment_properties, __pyx_t_2) < 0) __PYX_ERR(0, 367, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer);

  /* "uharfbuzz/_harfbuzz.pyx":370
 *         hb_buffer_guess_segment_properties(self._hb_buffer)
 * 
 *     def set_message_func(self, callback) -> None:             # <<<<<<<<<<<<<<
 *         self._message_callback = callback
 *         hb_buffer_set_message_func(self._hb_buffer, msgcallback, <void*>callback, NULL)
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 370, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_6Buffer_25set_message_func, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Buffer_set_message_func, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__47)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer, __pyx_n_s_set_message_func, __pyx_t_3) < 0) __PYX_ERR(0, 370, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Buffer);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_6Buffer_27__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Buffer___reduce_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__48)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_6Buffer_29__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Buffer___setstate_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__49)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":399
 *         return wrapper
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def from_file_path(cls, filename: Union[str, Path]):
 *         cdef bytes packed = os.fsencode(filename)
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_filename, __pyx_kp_s_Union_str_Path) < 0) __PYX_ERR(0, 399, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_4Blob_3from_file_path, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Blob_from_file_path, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__51)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Blob, __pyx_n_s_from_file_path, __pyx_t_2) < 0) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Blob);
  __Pyx_GetNameInClass(__pyx_t_2, (PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Blob, __pyx_n_s_from_file_path); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_Method_ClassMethod(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Blob, __pyx_n_s_from_file_path, __pyx_t_3) < 0) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Blob);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_4Blob_11__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Blob___reduce_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__52)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_4Blob_13__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Blob___setstate_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__53)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":473
 * 
 *     # DEPRECATED: use the normal constructor
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def create(cls, bytes blob, int index=0):
 *         cdef Face inst = cls(blob, index)
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_4Face_5create, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Face_create, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__55)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__56);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Face, __pyx_n_s_create, __pyx_t_3) < 0) __PYX_ERR(0, 473, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Face);
  __Pyx_GetNameInClass(__pyx_t_3, (PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Face, __pyx_n_s_create); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_Method_ClassMethod(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Face, __pyx_n_s_create, __pyx_t_2) < 0) __PYX_ERR(0, 473, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Face);

  /* "uharfbuzz/_harfbuzz.pyx":478
 *         return inst
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def create_for_tables(cls,
 *                           func: Callable[[
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 478, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_func, __pyx_kp_s_Callable_Face_str_object_bytes) < 0) __PYX_ERR(0, 478, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_user_data, __pyx_n_s_object) < 0) __PYX_ERR(0, 478, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_4Face_7create_for_tables, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Face_create_for_tables, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__58)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 478, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Face, __pyx_n_s_create_for_tables, __pyx_t_3) < 0) __PYX_ERR(0, 478, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Face);
  __Pyx_GetNameInClass(__pyx_t_3, (PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Face, __pyx_n_s_create_for_tables); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 478, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_Method_ClassMethod(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 478, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Face, __pyx_n_s_create_for_tables, __pyx_t_2) < 0) __PYX_ERR(0, 478, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Face);

  /* "uharfbuzz/_harfbuzz.pyx":555
 *         return s
 * 
 *     def variation_unicodes(self, variation_selector):             # <<<<<<<<<<<<<<
 *         s = Set()
 *         hb_face_collect_variation_unicodes(self._hb_face, variation_selector, s._hb_set)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_4Face_9variation_unicodes, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Face_variation_unicodes, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__60)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Face, __pyx_n_s_variation_unicodes, __pyx_t_2) < 0) __PYX_ERR(0, 555, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Face);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_4Face_11__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Face___reduce_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__61)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_4Face_13__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Face___setstate_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__62)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":562
 * 
 * # typing.NamedTuple doesn't seem to work with cython
 * GlyphExtents = namedtuple(             # <<<<<<<<<<<<<<
 *     "GlyphExtents", ["x_bearing", "y_bearing", "width", "height"]
 * )
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_namedtuple); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 562, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "uharfbuzz/_harfbuzz.pyx":563
 * # typing.NamedTuple doesn't seem to work with cython
 * GlyphExtents = namedtuple(
 *     "GlyphExtents", ["x_bearing", "y_bearing", "width", "height"]             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_t_3 = PyList_New(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_u_x_bearing);
  __Pyx_GIVEREF(__pyx_n_u_x_bearing);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_u_x_bearing)) __PYX_ERR(0, 563, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_y_bearing);
  __Pyx_GIVEREF(__pyx_n_u_y_bearing);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 1, __pyx_n_u_y_bearing)) __PYX_ERR(0, 563, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_width);
  __Pyx_GIVEREF(__pyx_n_u_width);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 2, __pyx_n_u_width)) __PYX_ERR(0, 563, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_height);
  __Pyx_GIVEREF(__pyx_n_u_height);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 3, __pyx_n_u_height)) __PYX_ERR(0, 563, __pyx_L1_error);

  /* "uharfbuzz/_harfbuzz.pyx":562
 * 
 * # typing.NamedTuple doesn't seem to work with cython
 * GlyphExtents = namedtuple(             # <<<<<<<<<<<<<<
 *     "GlyphExtents", ["x_bearing", "y_bearing", "width", "height"]
 * )
 */
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 562, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_n_u_GlyphExtents);
  __Pyx_GIVEREF(__pyx_n_u_GlyphExtents);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_u_GlyphExtents)) __PYX_ERR(0, 562, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_3)) __PYX_ERR(0, 562, __pyx_L1_error);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 562, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_GlyphExtents, __pyx_t_3) < 0) __PYX_ERR(0, 562, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":566
 * )
 * 
 * FontExtents = namedtuple(             # <<<<<<<<<<<<<<
 *     "FontExtents", ["ascender", "descender", "line_gap"]
 * )
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_namedtuple); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "uharfbuzz/_harfbuzz.pyx":567
 * 
 * FontExtents = namedtuple(
 *     "FontExtents", ["ascender", "descender", "line_gap"]             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_t_5 = PyList_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_n_u_ascender);
  __Pyx_GIVEREF(__pyx_n_u_ascender);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 0, __pyx_n_u_ascender)) __PYX_ERR(0, 567, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_descender);
  __Pyx_GIVEREF(__pyx_n_u_descender);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 1, __pyx_n_u_descender)) __PYX_ERR(0, 567, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_line_gap);
  __Pyx_GIVEREF(__pyx_n_u_line_gap);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_5, 2, __pyx_n_u_line_gap)) __PYX_ERR(0, 567, __pyx_L1_error);

  /* "uharfbuzz/_harfbuzz.pyx":566
 * )
 * 
 * FontExtents = namedtuple(             # <<<<<<<<<<<<<<
 *     "FontExtents", ["ascender", "descender", "line_gap"]
 * )
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_u_FontExtents);
  __Pyx_GIVEREF(__pyx_n_u_FontExtents);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_n_u_FontExtents)) __PYX_ERR(0, 566, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_5)) __PYX_ERR(0, 566, __pyx_L1_error);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_2, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_FontExtents, __pyx_t_5) < 0) __PYX_ERR(0, 566, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":608
 * 
 *     # DEPRECATED: use the normal constructor
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def create(cls, face: Face):
 *         cdef Font inst = cls(face)
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 608, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_face, __pyx_n_s_Face) < 0) __PYX_ERR(0, 608, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_5create, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Font_create, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__64)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 608, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Font, __pyx_n_s_create, __pyx_t_2) < 0) __PYX_ERR(0, 608, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Font);
  __Pyx_GetNameInClass(__pyx_t_2, (PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Font, __pyx_n_s_create); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 608, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_Method_ClassMethod(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 608, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Font, __pyx_n_s_create, __pyx_t_5) < 0) __PYX_ERR(0, 608, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Font);

  /* "uharfbuzz/_harfbuzz.pyx":697
 *         hb_font_set_var_named_instance(self._hb_font, value)
 * 
 *     def set_variations(self, variations: Dict[str, float]) -> None:             # <<<<<<<<<<<<<<
 *         cdef unsigned int size
 *         cdef hb_variation_t* hb_variations
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_variations, __pyx_kp_s_Dict_str_float) < 0) __PYX_ERR(0, 697, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 697, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_7set_variations, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Font_set_variations, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__66)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Font, __pyx_n_s_set_variations, __pyx_t_2) < 0) __PYX_ERR(0, 697, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Font);

  /* "uharfbuzz/_harfbuzz.pyx":717
 *             free(hb_variations)
 * 
 *     def set_variation(self, name: str, value: float) -> None:             # <<<<<<<<<<<<<<
 *         packed = name.encode()
 *         cdef hb_tag_t tag = hb_tag_from_string(packed, -1)
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 717, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_name, __pyx_n_s_str) < 0) __PYX_ERR(0, 717, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_value, __pyx_n_s_float) < 0) __PYX_ERR(0, 717, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 717, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_9set_variation, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Font_set_variation, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__68)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 717, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Font, __pyx_n_s_set_variation, __pyx_t_5) < 0) __PYX_ERR(0, 717, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Font);

  /* "uharfbuzz/_harfbuzz.pyx":722
 *         hb_font_set_variation(self._hb_font, tag, value)
 * 
 *     def get_glyph_name(self, gid: int):             # <<<<<<<<<<<<<<
 *         cdef char name[64]
 *         cdef bytes packed
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_gid, __pyx_n_s_int) < 0) __PYX_ERR(0, 722, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_11get_glyph_name, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Font_get_glyph_name, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__70)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Font, __pyx_n_s_get_glyph_name, __pyx_t_2) < 0) __PYX_ERR(0, 722, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Font);

  /* "uharfbuzz/_harfbuzz.pyx":732
 *             return None
 * 
 *     def get_glyph_from_name(self, name: str):             # <<<<<<<<<<<<<<
 *         cdef hb_codepoint_t gid
 *         cdef bytes packed
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_name, __pyx_n_s_str) < 0) __PYX_ERR(0, 732, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_13get_glyph_from_name, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Font_get_glyph_from_name, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__72)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Font, __pyx_n_s_get_glyph_from_name, __pyx_t_5) < 0) __PYX_ERR(0, 732, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Font);

  /* "uharfbuzz/_harfbuzz.pyx":739
 *         return gid if success else None
 * 
 *     def get_glyph_extents(self, gid: int):             # <<<<<<<<<<<<<<
 *         cdef hb_glyph_extents_t extents
 *         success = hb_font_get_glyph_extents(self._hb_font, gid, &extents)
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_gid, __pyx_n_s_int) < 0) __PYX_ERR(0, 739, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_15get_glyph_extents, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Font_get_glyph_extents, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__74)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Font, __pyx_n_s_get_glyph_extents, __pyx_t_2) < 0) __PYX_ERR(0, 739, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Font);

  /* "uharfbuzz/_harfbuzz.pyx":752
 *             return None
 * 
 *     def get_glyph_h_advance(self, gid: int):             # <<<<<<<<<<<<<<
 *         return hb_font_get_glyph_h_advance(self._hb_font, gid)
 * 
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_gid, __pyx_n_s_int) < 0) __PYX_ERR(0, 752, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_17get_glyph_h_advance, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Font_get_glyph_h_advance, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__76)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Font, __pyx_n_s_get_glyph_h_advance, __pyx_t_5) < 0) __PYX_ERR(0, 752, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Font);

  /* "uharfbuzz/_harfbuzz.pyx":755
 *         return hb_font_get_glyph_h_advance(self._hb_font, gid)
 * 
 *     def get_glyph_v_advance(self, gid: int):             # <<<<<<<<<<<<<<
 *         return hb_font_get_glyph_v_advance(self._hb_font, gid)
 * 
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 755, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_gid, __pyx_n_s_int) < 0) __PYX_ERR(0, 755, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_19get_glyph_v_advance, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Font_get_glyph_v_advance, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__77)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 755, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Font, __pyx_n_s_get_glyph_v_advance, __pyx_t_2) < 0) __PYX_ERR(0, 755, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Font);

  /* "uharfbuzz/_harfbuzz.pyx":758
 *         return hb_font_get_glyph_v_advance(self._hb_font, gid)
 * 
 *     def get_glyph_h_origin(self, gid: int):             # <<<<<<<<<<<<<<
 *         cdef hb_position_t x, y
 *         success = hb_font_get_glyph_h_origin(self._hb_font, gid, &x, &y)
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 758, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_gid, __pyx_n_s_int) < 0) __PYX_ERR(0, 758, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_21get_glyph_h_origin, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Font_get_glyph_h_origin, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__79)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 758, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Font, __pyx_n_s_get_glyph_h_origin, __pyx_t_5) < 0) __PYX_ERR(0, 758, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Font);

  /* "uharfbuzz/_harfbuzz.pyx":763
 *         return (x, y) if success else None
 * 
 *     def get_glyph_v_origin(self, gid: int):             # <<<<<<<<<<<<<<
 *         cdef hb_position_t x, y
 *         success = hb_font_get_glyph_v_origin(self._hb_font, gid, &x, &y)
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 763, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_gid, __pyx_n_s_int) < 0) __PYX_ERR(0, 763, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_23get_glyph_v_origin, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Font_get_glyph_v_origin, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__80)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 763, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Font, __pyx_n_s_get_glyph_v_origin, __pyx_t_2) < 0) __PYX_ERR(0, 763, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Font);

  /* "uharfbuzz/_harfbuzz.pyx":768
 *         return (x, y) if success else None
 * 
 *     def get_font_extents(self, direction: str):             # <<<<<<<<<<<<<<
 *         cdef hb_font_extents_t extents
 *         cdef hb_direction_t hb_direction
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_direction, __pyx_n_s_str) < 0) __PYX_ERR(0, 768, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_25get_font_extents, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Font_get_font_extents, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__82)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Font, __pyx_n_s_get_font_extents, __pyx_t_5) < 0) __PYX_ERR(0, 768, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Font);

  /* "uharfbuzz/_harfbuzz.pyx":783
 *         )
 * 
 *     def get_variation_glyph(self, unicode: int, variation_selector: int):             # <<<<<<<<<<<<<<
 *         cdef hb_codepoint_t gid
 *         success = hb_font_get_variation_glyph(self._hb_font, unicode, variation_selector, &gid)
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_unicode, __pyx_n_s_int) < 0) __PYX_ERR(0, 783, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_variation_selector, __pyx_n_s_int) < 0) __PYX_ERR(0, 783, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_27get_variation_glyph, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Font_get_variation_glyph, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__84)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Font, __pyx_n_s_get_variation_glyph, __pyx_t_2) < 0) __PYX_ERR(0, 783, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Font);

  /* "uharfbuzz/_harfbuzz.pyx":788
 *         return gid if success else None
 * 
 *     def get_nominal_glyph(self, unicode: int):             # <<<<<<<<<<<<<<
 *         cdef hb_codepoint_t gid
 *         success = hb_font_get_nominal_glyph(self._hb_font, unicode, &gid)
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 788, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_unicode, __pyx_n_s_int) < 0) __PYX_ERR(0, 788, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_29get_nominal_glyph, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Font_get_nominal_glyph, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__86)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 788, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Font, __pyx_n_s_get_nominal_glyph, __pyx_t_5) < 0) __PYX_ERR(0, 788, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Font);

  /* "uharfbuzz/_harfbuzz.pyx":793
 *         return gid if success else None
 * 
 *     def get_var_coords_normalized(self):             # <<<<<<<<<<<<<<
 *         cdef unsigned int length
 *         cdef const int *coords
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_31get_var_coords_normalized, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Font_get_var_coords_normalized, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__88)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Font, __pyx_n_s_get_var_coords_normalized, __pyx_t_5) < 0) __PYX_ERR(0, 793, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Font);

  /* "uharfbuzz/_harfbuzz.pyx":800
 *         return [coords[i] / 0x4000 for i in range(length)]
 * 
 *     def set_var_coords_normalized(self, coords):             # <<<<<<<<<<<<<<
 *         cdef unsigned int length
 *         cdef int *coords_2dot14
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_33set_var_coords_normalized, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Font_set_var_coords_normalized, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__90)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 800, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Font, __pyx_n_s_set_var_coords_normalized, __pyx_t_5) < 0) __PYX_ERR(0, 800, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Font);

  /* "uharfbuzz/_harfbuzz.pyx":815
 *             free(coords_2dot14)
 * 
 *     def get_var_coords_design(self):             # <<<<<<<<<<<<<<
 *         cdef unsigned int length
 *         cdef const float *coords
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_35get_var_coords_design, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Font_get_var_coords_design, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__91)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Font, __pyx_n_s_get_var_coords_design, __pyx_t_5) < 0) __PYX_ERR(0, 815, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Font);

  /* "uharfbuzz/_harfbuzz.pyx":821
 *         return [coords[i] for i in range(length)]
 * 
 *     def set_var_coords_design(self, coords):             # <<<<<<<<<<<<<<
 *         cdef unsigned int length
 *         cdef cython.float *c_coords
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_37set_var_coords_design, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Font_set_var_coords_design, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__93)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Font, __pyx_n_s_set_var_coords_design, __pyx_t_5) < 0) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Font);

  /* "uharfbuzz/_harfbuzz.pyx":835
 *             free(c_coords)
 * 
 *     def glyph_to_string(self, gid: int):             # <<<<<<<<<<<<<<
 *         cdef char name[64]
 *         cdef bytes packed
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_gid, __pyx_n_s_int) < 0) __PYX_ERR(0, 835, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_39glyph_to_string, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Font_glyph_to_string, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__95)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Font, __pyx_n_s_glyph_to_string, __pyx_t_2) < 0) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Font);

  /* "uharfbuzz/_harfbuzz.pyx":842
 *         return packed.decode()
 * 
 *     def glyph_from_string(self, string: str):             # <<<<<<<<<<<<<<
 *         cdef hb_codepoint_t gid
 *         cdef bytes packed
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 842, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_string, __pyx_n_s_str) < 0) __PYX_ERR(0, 842, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_41glyph_from_string, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Font_glyph_from_string, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__97)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 842, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Font, __pyx_n_s_glyph_from_string, __pyx_t_5) < 0) __PYX_ERR(0, 842, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Font);

  /* "uharfbuzz/_harfbuzz.pyx":849
 *         return gid if success else None
 * 
 *     def draw_glyph(self, gid: int, draw_funcs: DrawFuncs, draw_state: object = None):             # <<<<<<<<<<<<<<
 *         cdef void *draw_state_p = <void *>draw_state
 *         if PyCapsule_IsValid(draw_state, NULL):
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 849, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_gid, __pyx_n_s_int) < 0) __PYX_ERR(0, 849, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_draw_funcs, __pyx_n_s_DrawFuncs) < 0) __PYX_ERR(0, 849, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_draw_state, __pyx_n_s_object) < 0) __PYX_ERR(0, 849, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_43draw_glyph, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Font_draw_glyph, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__99)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 849, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_tuple__7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Font, __pyx_n_s_draw_glyph, __pyx_t_2) < 0) __PYX_ERR(0, 849, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Font);

  /* "uharfbuzz/_harfbuzz.pyx":858
 *                     paint_funcs: PaintFuncs,
 *                     paint_state: object = None,
 *                     palette_index: int = 0,             # <<<<<<<<<<<<<<
 *                     foreground: Color | None = None) -> None:
 *         cdef void *paint_state_p = <void *>paint_state
 */
  if (!(likely(__Pyx_Py3Int_CheckExact(__pyx_int_0)) || __Pyx_RaiseUnexpectedTypeError("int", __pyx_int_0))) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_INCREF(__pyx_int_0);
  __pyx_k__8 = ((PyObject*)__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);

  /* "uharfbuzz/_harfbuzz.pyx":857
 *     def paint_glyph(self, gid: int,
 *                     paint_funcs: PaintFuncs,
 *                     paint_state: object = None,             # <<<<<<<<<<<<<<
 *                     palette_index: int = 0,
 *                     foreground: Color | None = None) -> None:
 */
  if (!(likely(__Pyx_Py3Int_CheckExact(__pyx_int_0)) || __Pyx_RaiseUnexpectedTypeError("int", __pyx_int_0))) __PYX_ERR(0, 858, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":855
 *         hb_font_draw_glyph(self._hb_font, gid, draw_funcs._hb_drawfuncs, draw_state_p);
 * 
 *     def paint_glyph(self, gid: int,             # <<<<<<<<<<<<<<
 *                     paint_funcs: PaintFuncs,
 *                     paint_state: object = None,
 */
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, Py_None)) __PYX_ERR(0, 855, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_0)) __PYX_ERR(0, 855, __pyx_L1_error);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None)) __PYX_ERR(0, 855, __pyx_L1_error);
  __pyx_t_5 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_gid, __pyx_n_s_int) < 0) __PYX_ERR(0, 855, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_paint_funcs, __pyx_n_s_PaintFuncs) < 0) __PYX_ERR(0, 855, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_paint_state, __pyx_n_s_object) < 0) __PYX_ERR(0, 855, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_palette_index, __pyx_n_s_int) < 0) __PYX_ERR(0, 855, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_foreground, __pyx_kp_s_Color_None) < 0) __PYX_ERR(0, 855, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 855, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_45paint_glyph, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Font_paint_glyph, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__101)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 855, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Font, __pyx_n_s_paint_glyph, __pyx_t_3) < 0) __PYX_ERR(0, 855, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Font);

  /* "uharfbuzz/_harfbuzz.pyx":871
 *                             c_foreground)
 * 
 *     def draw_glyph_with_pen(self, gid: int, pen):             # <<<<<<<<<<<<<<
 *         global drawfuncs
 *         if drawfuncs == NULL:
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 871, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_gid, __pyx_n_s_int) < 0) __PYX_ERR(0, 871, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_47draw_glyph_with_pen, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Font_draw_glyph_with_pen, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__103)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 871, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Font, __pyx_n_s_draw_glyph_with_pen, __pyx_t_5) < 0) __PYX_ERR(0, 871, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Font);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_49__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Font___reduce_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__104)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_5) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_4Font_51__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Font___setstate_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__105)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_5) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":904
 *     void *closePath
 * 
 * cdef hb_draw_funcs_t* drawfuncs = NULL             # <<<<<<<<<<<<<<
 * 
 * cdef void _pen_move_to_func(hb_draw_funcs_t *dfuncs,
 */
  __pyx_v_9uharfbuzz_9_harfbuzz_drawfuncs = NULL;

  /* "uharfbuzz/_harfbuzz.pyx":1067
 * 
 *     # DEPRECATED: use the normal constructor
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def create(cls):
 *         cdef FontFuncs inst = cls()
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_9FontFuncs_5create, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FontFuncs_create, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__106)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1067, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs, __pyx_n_s_create, __pyx_t_5) < 0) __PYX_ERR(0, 1067, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs);
  __Pyx_GetNameInClass(__pyx_t_5, (PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs, __pyx_n_s_create); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1067, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_Method_ClassMethod(__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1067, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs, __pyx_n_s_create, __pyx_t_3) < 0) __PYX_ERR(0, 1067, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs);

  /* "uharfbuzz/_harfbuzz.pyx":1072
 *         return inst
 * 
 *     def set_glyph_h_advance_func(self,             # <<<<<<<<<<<<<<
 *                                  func: Callable[[
 *                                      Font,
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1072, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_func, __pyx_kp_s_Callable_Font_int_object_int) < 0) __PYX_ERR(0, 1072, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_user_data, __pyx_n_s_object) < 0) __PYX_ERR(0, 1072, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 1072, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_9FontFuncs_7set_glyph_h_advance_func, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FontFuncs_set_glyph_h_advance_fu, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__108)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1072, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_tuple__7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs, __pyx_n_s_set_glyph_h_advance_func, __pyx_t_5) < 0) __PYX_ERR(0, 1072, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs);

  /* "uharfbuzz/_harfbuzz.pyx":1083
 *         self._glyph_h_advance_func = func
 * 
 *     def set_glyph_v_advance_func(self,             # <<<<<<<<<<<<<<
 *                                  func: Callable[[
 *                                      Font,
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1083, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_func, __pyx_kp_s_Callable_Font_int_object_int) < 0) __PYX_ERR(0, 1083, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_user_data, __pyx_n_s_object) < 0) __PYX_ERR(0, 1083, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 1083, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_9FontFuncs_9set_glyph_v_advance_func, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FontFuncs_set_glyph_v_advance_fu, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__109)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1083, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs, __pyx_n_s_set_glyph_v_advance_func, __pyx_t_3) < 0) __PYX_ERR(0, 1083, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs);

  /* "uharfbuzz/_harfbuzz.pyx":1094
 *         self._glyph_v_advance_func = func
 * 
 *     def set_glyph_v_origin_func(self,             # <<<<<<<<<<<<<<
 *                                 func: Callable[[
 *                                     Font,
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1094, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_func, __pyx_kp_s_Callable_Font_int_object_int_int) < 0) __PYX_ERR(0, 1094, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_user_data, __pyx_n_s_object) < 0) __PYX_ERR(0, 1094, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 1094, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_9FontFuncs_11set_glyph_v_origin_func, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FontFuncs_set_glyph_v_origin_fun, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__110)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1094, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_tuple__7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs, __pyx_n_s_set_glyph_v_origin_func, __pyx_t_5) < 0) __PYX_ERR(0, 1094, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs);

  /* "uharfbuzz/_harfbuzz.pyx":1105
 *         self._glyph_v_origin_func = func
 * 
 *     def set_glyph_name_func(self,             # <<<<<<<<<<<<<<
 *                             func: Callable[[
 *                                 Font,
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_func, __pyx_kp_s_Callable_Font_int_object_str) < 0) __PYX_ERR(0, 1105, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_user_data, __pyx_n_s_object) < 0) __PYX_ERR(0, 1105, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 1105, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_9FontFuncs_13set_glyph_name_func, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FontFuncs_set_glyph_name_func, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__111)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs, __pyx_n_s_set_glyph_name_func, __pyx_t_3) < 0) __PYX_ERR(0, 1105, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs);

  /* "uharfbuzz/_harfbuzz.pyx":1116
 *         self._glyph_name_func = func
 * 
 *     def set_nominal_glyph_func(self,             # <<<<<<<<<<<<<<
 *                                func: Callable[[
 *                                    Font,
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_func, __pyx_kp_s_Callable_Font_int_object_int) < 0) __PYX_ERR(0, 1116, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_user_data, __pyx_n_s_object) < 0) __PYX_ERR(0, 1116, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 1116, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_9FontFuncs_15set_nominal_glyph_func, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FontFuncs_set_nominal_glyph_func, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__112)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_tuple__7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs, __pyx_n_s_set_nominal_glyph_func, __pyx_t_5) < 0) __PYX_ERR(0, 1116, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs);

  /* "uharfbuzz/_harfbuzz.pyx":1127
 *         self._nominal_glyph_func = func
 * 
 *     def set_variation_glyph_func(self,             # <<<<<<<<<<<<<<
 *                                func: Callable[[
 *                                    Font,
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_func, __pyx_kp_s_Callable_Font_int_int_object_int) < 0) __PYX_ERR(0, 1127, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_user_data, __pyx_n_s_object) < 0) __PYX_ERR(0, 1127, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 1127, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_9FontFuncs_17set_variation_glyph_func, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FontFuncs_set_variation_glyph_fu, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__113)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs, __pyx_n_s_set_variation_glyph_func, __pyx_t_3) < 0) __PYX_ERR(0, 1127, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs);

  /* "uharfbuzz/_harfbuzz.pyx":1139
 *         self._variation_glyph_func = func
 * 
 *     def set_font_h_extents_func(self,             # <<<<<<<<<<<<<<
 *                                 func: Callable[[
 *                                     Font,
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_func, __pyx_kp_s_Callable_Font_object_FontExtents) < 0) __PYX_ERR(0, 1139, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_user_data, __pyx_n_s_object) < 0) __PYX_ERR(0, 1139, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 1139, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_9FontFuncs_19set_font_h_extents_func, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FontFuncs_set_font_h_extents_fun, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__114)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_tuple__7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs, __pyx_n_s_set_font_h_extents_func, __pyx_t_5) < 0) __PYX_ERR(0, 1139, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs);

  /* "uharfbuzz/_harfbuzz.pyx":1149
 *         self._font_h_extents_func = func
 * 
 *     def set_font_v_extents_func(self,             # <<<<<<<<<<<<<<
 *                                 func: Callable[[
 *                                     Font,
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_func, __pyx_kp_s_Callable_Font_object_FontExtents) < 0) __PYX_ERR(0, 1149, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_user_data, __pyx_n_s_object) < 0) __PYX_ERR(0, 1149, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 1149, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_9FontFuncs_21set_font_v_extents_func, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FontFuncs_set_font_v_extents_fun, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__115)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_tuple__7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs, __pyx_n_s_set_font_v_extents_func, __pyx_t_3) < 0) __PYX_ERR(0, 1149, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_FontFuncs);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_9FontFuncs_23__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FontFuncs___reduce_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__116)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_9FontFuncs_25__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_FontFuncs___setstate_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__117)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1159
 *         self._font_v_extents_func = func
 * 
 * def shape(font: Font, buffer: Buffer,             # <<<<<<<<<<<<<<
 *         features: Dict[str,Union[int,bool,Sequence[Tuple[int,int,Union[int,bool]]]]] = None,
 *         shapers: List[str] = None) -> None:
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_font, __pyx_n_s_Font) < 0) __PYX_ERR(0, 1159, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_buffer, __pyx_n_s_Buffer) < 0) __PYX_ERR(0, 1159, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_features, __pyx_kp_s_Dict_str_Union_int_bool_Sequence) < 0) __PYX_ERR(0, 1159, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_shapers, __pyx_kp_s_List_str) < 0) __PYX_ERR(0, 1159, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 1159, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3shape, 0, __pyx_n_s_shape, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__119)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_tuple__120);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_shape, __pyx_t_5) < 0) __PYX_ERR(0, 1159, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1211
 * 
 * 
 * def ot_tag_to_script(tag: str) -> str:             # <<<<<<<<<<<<<<
 *     cdef bytes packed = tag.encode()
 *     cdef hb_tag_t hb_tag = hb_tag_from_string(<char*>packed, -1)
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_tag, __pyx_n_s_str) < 0) __PYX_ERR(0, 1211, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return, __pyx_n_s_str) < 0) __PYX_ERR(0, 1211, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_5ot_tag_to_script, 0, __pyx_n_s_ot_tag_to_script, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__122)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_tag_to_script, __pyx_t_3) < 0) __PYX_ERR(0, 1211, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1222
 * 
 * 
 * def ot_tag_to_language(tag: str) -> str:             # <<<<<<<<<<<<<<
 *     cdef bytes packed = tag.encode()
 *     cdef hb_tag_t hb_tag = hb_tag_from_string(<char*>packed, -1)
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_tag, __pyx_n_s_str) < 0) __PYX_ERR(0, 1222, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_str) < 0) __PYX_ERR(0, 1222, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_7ot_tag_to_language, 0, __pyx_n_s_ot_tag_to_language, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__124)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_tag_to_language, __pyx_t_5) < 0) __PYX_ERR(0, 1222, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1233
 * 
 * 
 * def ot_layout_lookup_get_glyph_alternates(             # <<<<<<<<<<<<<<
 *         face: Face, lookup_index : int, glyph : hb_codepoint_t) -> List[int]:
 *     cdef list alternates = []
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_face, __pyx_n_s_Face) < 0) __PYX_ERR(0, 1233, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_lookup_index, __pyx_n_s_int) < 0) __PYX_ERR(0, 1233, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_glyph, __pyx_n_s_hb_codepoint_t) < 0) __PYX_ERR(0, 1233, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return, __pyx_kp_s_List_int) < 0) __PYX_ERR(0, 1233, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_9ot_layout_lookup_get_glyph_alternates, 0, __pyx_n_s_ot_layout_lookup_get_glyph_alter, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__126)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_layout_lookup_get_glyph_alter, __pyx_t_3) < 0) __PYX_ERR(0, 1233, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1249
 * 
 * 
 * def ot_layout_language_get_feature_tags(             # <<<<<<<<<<<<<<
 *         face: Face, tag: str, script_index: int = 0,
 *         language_index: int = 0xFFFF) -> List[str]:
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_face, __pyx_n_s_Face) < 0) __PYX_ERR(0, 1249, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_tag, __pyx_n_s_str) < 0) __PYX_ERR(0, 1249, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_script_index, __pyx_n_s_int) < 0) __PYX_ERR(0, 1249, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_language_index, __pyx_n_s_int) < 0) __PYX_ERR(0, 1249, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_kp_s_List_str) < 0) __PYX_ERR(0, 1249, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_11ot_layout_language_get_feature_tags, 0, __pyx_n_s_ot_layout_language_get_feature_t, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__128)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_5, sizeof(__pyx_defaults), 2)) __PYX_ERR(0, 1249, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1250
 * 
 * def ot_layout_language_get_feature_tags(
 *         face: Face, tag: str, script_index: int = 0,             # <<<<<<<<<<<<<<
 *         language_index: int = 0xFFFF) -> List[str]:
 *     cdef bytes packed = tag.encode()
 */
  if (!(likely(__Pyx_Py3Int_CheckExact(__pyx_int_0)) || __Pyx_RaiseUnexpectedTypeError("int", __pyx_int_0))) __PYX_ERR(0, 1250, __pyx_L1_error)
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_t_5)->__pyx_arg_script_index = ((PyObject*)__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);

  /* "uharfbuzz/_harfbuzz.pyx":1251
 * def ot_layout_language_get_feature_tags(
 *         face: Face, tag: str, script_index: int = 0,
 *         language_index: int = 0xFFFF) -> List[str]:             # <<<<<<<<<<<<<<
 *     cdef bytes packed = tag.encode()
 *     cdef hb_tag_t hb_tag = hb_tag_from_string(<char*>packed, -1)
 */
  if (!(likely(__Pyx_Py3Int_CheckExact(__pyx_int_65535)) || __Pyx_RaiseUnexpectedTypeError("int", __pyx_int_65535))) __PYX_ERR(0, 1251, __pyx_L1_error)
  __Pyx_INCREF(__pyx_int_65535);
  __Pyx_CyFunction_Defaults(__pyx_defaults, __pyx_t_5)->__pyx_arg_language_index = ((PyObject*)__pyx_int_65535);
  __Pyx_GIVEREF(__pyx_int_65535);
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_5, __pyx_pf_9uharfbuzz_9_harfbuzz_84__defaults__);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_layout_language_get_feature_t, __pyx_t_5) < 0) __PYX_ERR(0, 1249, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1273
 * 
 * 
 * def ot_layout_script_get_language_tags(             # <<<<<<<<<<<<<<
 *         face: Face, tag: str, script_index: int = 0) -> List[str]:
 *     cdef bytes packed = tag.encode()
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_face, __pyx_n_s_Face) < 0) __PYX_ERR(0, 1273, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_tag, __pyx_n_s_str) < 0) __PYX_ERR(0, 1273, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_script_index, __pyx_n_s_int) < 0) __PYX_ERR(0, 1273, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return, __pyx_kp_s_List_str) < 0) __PYX_ERR(0, 1273, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_13ot_layout_script_get_language_tags, 0, __pyx_n_s_ot_layout_script_get_language_ta, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__130)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!__Pyx_CyFunction_InitDefaults(__pyx_t_3, sizeof(__pyx_defaults1), 1)) __PYX_ERR(0, 1273, __pyx_L1_error)

  /* "uharfbuzz/_harfbuzz.pyx":1274
 * 
 * def ot_layout_script_get_language_tags(
 *         face: Face, tag: str, script_index: int = 0) -> List[str]:             # <<<<<<<<<<<<<<
 *     cdef bytes packed = tag.encode()
 *     cdef hb_tag_t hb_tag = hb_tag_from_string(<char*>packed, -1)
 */
  if (!(likely(__Pyx_Py3Int_CheckExact(__pyx_int_0)) || __Pyx_RaiseUnexpectedTypeError("int", __pyx_int_0))) __PYX_ERR(0, 1274, __pyx_L1_error)
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_CyFunction_Defaults(__pyx_defaults1, __pyx_t_3)->__pyx_arg_script_index = ((PyObject*)__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __Pyx_CyFunction_SetDefaultsGetter(__pyx_t_3, __pyx_pf_9uharfbuzz_9_harfbuzz_86__defaults__);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_layout_script_get_language_ta, __pyx_t_3) < 0) __PYX_ERR(0, 1273, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1294
 *     return tags
 * 
 * def ot_layout_table_get_script_tags(face: Face, tag: str) -> List[str]:             # <<<<<<<<<<<<<<
 *     cdef bytes packed = tag.encode()
 *     cdef hb_tag_t hb_tag = hb_tag_from_string(<char*>packed, -1)
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_face, __pyx_n_s_Face) < 0) __PYX_ERR(0, 1294, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_tag, __pyx_n_s_str) < 0) __PYX_ERR(0, 1294, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_kp_s_List_str) < 0) __PYX_ERR(0, 1294, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_15ot_layout_table_get_script_tags, 0, __pyx_n_s_ot_layout_table_get_script_tags, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__132)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_layout_table_get_script_tags, __pyx_t_5) < 0) __PYX_ERR(0, 1294, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1314
 *     return tags
 * 
 * def ot_layout_get_baseline(font: Font,             # <<<<<<<<<<<<<<
 *                            baseline_tag: str,
 *                            direction: str,
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_font, __pyx_n_s_Font) < 0) __PYX_ERR(0, 1314, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_baseline_tag, __pyx_n_s_str) < 0) __PYX_ERR(0, 1314, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_direction, __pyx_n_s_str) < 0) __PYX_ERR(0, 1314, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_script_tag, __pyx_n_s_str) < 0) __PYX_ERR(0, 1314, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_language_tag, __pyx_n_s_str) < 0) __PYX_ERR(0, 1314, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return, __pyx_n_s_int) < 0) __PYX_ERR(0, 1314, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_17ot_layout_get_baseline, 0, __pyx_n_s_ot_layout_get_baseline, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__134)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1314, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_layout_get_baseline, __pyx_t_3) < 0) __PYX_ERR(0, 1314, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1360
 *         return None
 * 
 * def ot_layout_has_glyph_classes(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_layout_has_glyph_classes(face._hb_face)
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_face, __pyx_n_s_Face) < 0) __PYX_ERR(0, 1360, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 1360, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_19ot_layout_has_glyph_classes, 0, __pyx_n_s_ot_layout_has_glyph_classes, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__136)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_layout_has_glyph_classes, __pyx_t_5) < 0) __PYX_ERR(0, 1360, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1363
 *     return hb_ot_layout_has_glyph_classes(face._hb_face)
 * 
 * def ot_layout_has_positioning(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_layout_has_positioning(face._hb_face)
 * 
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_face, __pyx_n_s_Face) < 0) __PYX_ERR(0, 1363, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 1363, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_21ot_layout_has_positioning, 0, __pyx_n_s_ot_layout_has_positioning, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__137)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_layout_has_positioning, __pyx_t_3) < 0) __PYX_ERR(0, 1363, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1366
 *     return hb_ot_layout_has_positioning(face._hb_face)
 * 
 * def ot_layout_has_substitution(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_layout_has_substitution(face._hb_face)
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_face, __pyx_n_s_Face) < 0) __PYX_ERR(0, 1366, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 1366, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_23ot_layout_has_substitution, 0, __pyx_n_s_ot_layout_has_substitution, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__138)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_layout_has_substitution, __pyx_t_5) < 0) __PYX_ERR(0, 1366, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1369
 *     return hb_ot_layout_has_substitution(face._hb_face)
 * 
 * class OTLayoutGlyphClass(IntEnum):             # <<<<<<<<<<<<<<
 *     UNCLASSIFIED = HB_OT_LAYOUT_GLYPH_CLASS_UNCLASSIFIED
 *     BASE_GLYPH = HB_OT_LAYOUT_GLYPH_CLASS_BASE_GLYPH
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_IntEnum); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(0, 1369, __pyx_L1_error);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PEP560_update_bases(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_5, __pyx_n_s_OTLayoutGlyphClass, __pyx_n_s_OTLayoutGlyphClass, (PyObject *) NULL, __pyx_n_s_uharfbuzz__harfbuzz, (PyObject *) NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__pyx_t_5 != __pyx_t_3) {
    if (unlikely((PyDict_SetItemString(__pyx_t_4, "__orig_bases__", __pyx_t_3) < 0))) __PYX_ERR(0, 1369, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1370
 * 
 * class OTLayoutGlyphClass(IntEnum):
 *     UNCLASSIFIED = HB_OT_LAYOUT_GLYPH_CLASS_UNCLASSIFIED             # <<<<<<<<<<<<<<
 *     BASE_GLYPH = HB_OT_LAYOUT_GLYPH_CLASS_BASE_GLYPH
 *     LIGATURE = HB_OT_LAYOUT_GLYPH_CLASS_LIGATURE
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_layout_glyph_class_t(HB_OT_LAYOUT_GLYPH_CLASS_UNCLASSIFIED); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1370, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_UNCLASSIFIED, __pyx_t_3) < 0) __PYX_ERR(0, 1370, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1371
 * class OTLayoutGlyphClass(IntEnum):
 *     UNCLASSIFIED = HB_OT_LAYOUT_GLYPH_CLASS_UNCLASSIFIED
 *     BASE_GLYPH = HB_OT_LAYOUT_GLYPH_CLASS_BASE_GLYPH             # <<<<<<<<<<<<<<
 *     LIGATURE = HB_OT_LAYOUT_GLYPH_CLASS_LIGATURE
 *     MARK = HB_OT_LAYOUT_GLYPH_CLASS_MARK
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_layout_glyph_class_t(HB_OT_LAYOUT_GLYPH_CLASS_BASE_GLYPH); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_BASE_GLYPH, __pyx_t_3) < 0) __PYX_ERR(0, 1371, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1372
 *     UNCLASSIFIED = HB_OT_LAYOUT_GLYPH_CLASS_UNCLASSIFIED
 *     BASE_GLYPH = HB_OT_LAYOUT_GLYPH_CLASS_BASE_GLYPH
 *     LIGATURE = HB_OT_LAYOUT_GLYPH_CLASS_LIGATURE             # <<<<<<<<<<<<<<
 *     MARK = HB_OT_LAYOUT_GLYPH_CLASS_MARK
 *     COMPONENT = HB_OT_LAYOUT_GLYPH_CLASS_COMPONENT
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_layout_glyph_class_t(HB_OT_LAYOUT_GLYPH_CLASS_LIGATURE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_LIGATURE, __pyx_t_3) < 0) __PYX_ERR(0, 1372, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1373
 *     BASE_GLYPH = HB_OT_LAYOUT_GLYPH_CLASS_BASE_GLYPH
 *     LIGATURE = HB_OT_LAYOUT_GLYPH_CLASS_LIGATURE
 *     MARK = HB_OT_LAYOUT_GLYPH_CLASS_MARK             # <<<<<<<<<<<<<<
 *     COMPONENT = HB_OT_LAYOUT_GLYPH_CLASS_COMPONENT
 * 
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_layout_glyph_class_t(HB_OT_LAYOUT_GLYPH_CLASS_MARK); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_MARK, __pyx_t_3) < 0) __PYX_ERR(0, 1373, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1374
 *     LIGATURE = HB_OT_LAYOUT_GLYPH_CLASS_LIGATURE
 *     MARK = HB_OT_LAYOUT_GLYPH_CLASS_MARK
 *     COMPONENT = HB_OT_LAYOUT_GLYPH_CLASS_COMPONENT             # <<<<<<<<<<<<<<
 * 
 * def ot_layout_get_glyph_class(face: Face, glyph: int) -> OTLayoutGlyphClass:
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_layout_glyph_class_t(HB_OT_LAYOUT_GLYPH_CLASS_COMPONENT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_COMPONENT, __pyx_t_3) < 0) __PYX_ERR(0, 1374, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1369
 *     return hb_ot_layout_has_substitution(face._hb_face)
 * 
 * class OTLayoutGlyphClass(IntEnum):             # <<<<<<<<<<<<<<
 *     UNCLASSIFIED = HB_OT_LAYOUT_GLYPH_CLASS_UNCLASSIFIED
 *     BASE_GLYPH = HB_OT_LAYOUT_GLYPH_CLASS_BASE_GLYPH
 */
  __pyx_t_3 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_OTLayoutGlyphClass, __pyx_t_5, __pyx_t_4, NULL, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OTLayoutGlyphClass, __pyx_t_3) < 0) __PYX_ERR(0, 1369, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1376
 *     COMPONENT = HB_OT_LAYOUT_GLYPH_CLASS_COMPONENT
 * 
 * def ot_layout_get_glyph_class(face: Face, glyph: int) -> OTLayoutGlyphClass:             # <<<<<<<<<<<<<<
 *     return OTLayoutGlyphClass(hb_ot_layout_get_glyph_class(face._hb_face, glyph))
 * 
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_face, __pyx_n_s_Face) < 0) __PYX_ERR(0, 1376, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_glyph, __pyx_n_s_int) < 0) __PYX_ERR(0, 1376, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return, __pyx_n_s_OTLayoutGlyphClass) < 0) __PYX_ERR(0, 1376, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_25ot_layout_get_glyph_class, 0, __pyx_n_s_ot_layout_get_glyph_class, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__140)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_layout_get_glyph_class, __pyx_t_2) < 0) __PYX_ERR(0, 1376, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1380
 * 
 * 
 * def ot_color_has_palettes(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_color_has_palettes(face._hb_face)
 * 
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_face, __pyx_n_s_Face) < 0) __PYX_ERR(0, 1380, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 1380, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_27ot_color_has_palettes, 0, __pyx_n_s_ot_color_has_palettes, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__141)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_color_has_palettes, __pyx_t_5) < 0) __PYX_ERR(0, 1380, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1383
 *     return hb_ot_color_has_palettes(face._hb_face)
 * 
 * def ot_color_palette_get_count(face: Face) -> int:             # <<<<<<<<<<<<<<
 *     return hb_ot_color_palette_get_count(face._hb_face)
 * 
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_face, __pyx_n_s_Face) < 0) __PYX_ERR(0, 1383, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return, __pyx_n_s_int) < 0) __PYX_ERR(0, 1383, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_29ot_color_palette_get_count, 0, __pyx_n_s_ot_color_palette_get_count, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__142)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_color_palette_get_count, __pyx_t_2) < 0) __PYX_ERR(0, 1383, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1386
 *     return hb_ot_color_palette_get_count(face._hb_face)
 * 
 * class OTColorPaletteFlags(IntFlag):             # <<<<<<<<<<<<<<
 *     DEFAULT = HB_OT_COLOR_PALETTE_FLAG_DEFAULT
 *     USABLE_WITH_LIGHT_BACKGROUND = HB_OT_COLOR_PALETTE_FLAG_USABLE_WITH_LIGHT_BACKGROUND
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_IntFlag); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2)) __PYX_ERR(0, 1386, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PEP560_update_bases(__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_Py3MetaclassPrepare(__pyx_t_4, __pyx_t_2, __pyx_n_s_OTColorPaletteFlags, __pyx_n_s_OTColorPaletteFlags, (PyObject *) NULL, __pyx_n_s_uharfbuzz__harfbuzz, (PyObject *) NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__pyx_t_2 != __pyx_t_5) {
    if (unlikely((PyDict_SetItemString(__pyx_t_3, "__orig_bases__", __pyx_t_5) < 0))) __PYX_ERR(0, 1386, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1387
 * 
 * class OTColorPaletteFlags(IntFlag):
 *     DEFAULT = HB_OT_COLOR_PALETTE_FLAG_DEFAULT             # <<<<<<<<<<<<<<
 *     USABLE_WITH_LIGHT_BACKGROUND = HB_OT_COLOR_PALETTE_FLAG_USABLE_WITH_LIGHT_BACKGROUND
 *     USABLE_WITH_DARK_BACKGROUND = HB_OT_COLOR_PALETTE_FLAG_USABLE_WITH_DARK_BACKGROUND
 */
  __pyx_t_5 = __Pyx_PyInt_From_hb_ot_color_palette_flags_t(HB_OT_COLOR_PALETTE_FLAG_DEFAULT); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_DEFAULT, __pyx_t_5) < 0) __PYX_ERR(0, 1387, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1388
 * class OTColorPaletteFlags(IntFlag):
 *     DEFAULT = HB_OT_COLOR_PALETTE_FLAG_DEFAULT
 *     USABLE_WITH_LIGHT_BACKGROUND = HB_OT_COLOR_PALETTE_FLAG_USABLE_WITH_LIGHT_BACKGROUND             # <<<<<<<<<<<<<<
 *     USABLE_WITH_DARK_BACKGROUND = HB_OT_COLOR_PALETTE_FLAG_USABLE_WITH_DARK_BACKGROUND
 * 
 */
  __pyx_t_5 = __Pyx_PyInt_From_hb_ot_color_palette_flags_t(HB_OT_COLOR_PALETTE_FLAG_USABLE_WITH_LIGHT_BACKGROUND); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_USABLE_WITH_LIGHT_BACKGROUND, __pyx_t_5) < 0) __PYX_ERR(0, 1388, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1389
 *     DEFAULT = HB_OT_COLOR_PALETTE_FLAG_DEFAULT
 *     USABLE_WITH_LIGHT_BACKGROUND = HB_OT_COLOR_PALETTE_FLAG_USABLE_WITH_LIGHT_BACKGROUND
 *     USABLE_WITH_DARK_BACKGROUND = HB_OT_COLOR_PALETTE_FLAG_USABLE_WITH_DARK_BACKGROUND             # <<<<<<<<<<<<<<
 * 
 * def ot_color_palette_get_flags(face: Face, palette_index: int) -> OTColorPaletteFlags:
 */
  __pyx_t_5 = __Pyx_PyInt_From_hb_ot_color_palette_flags_t(HB_OT_COLOR_PALETTE_FLAG_USABLE_WITH_DARK_BACKGROUND); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1389, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_USABLE_WITH_DARK_BACKGROUND, __pyx_t_5) < 0) __PYX_ERR(0, 1389, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1386
 *     return hb_ot_color_palette_get_count(face._hb_face)
 * 
 * class OTColorPaletteFlags(IntFlag):             # <<<<<<<<<<<<<<
 *     DEFAULT = HB_OT_COLOR_PALETTE_FLAG_DEFAULT
 *     USABLE_WITH_LIGHT_BACKGROUND = HB_OT_COLOR_PALETTE_FLAG_USABLE_WITH_LIGHT_BACKGROUND
 */
  __pyx_t_5 = __Pyx_Py3ClassCreate(__pyx_t_4, __pyx_n_s_OTColorPaletteFlags, __pyx_t_2, __pyx_t_3, NULL, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OTColorPaletteFlags, __pyx_t_5) < 0) __PYX_ERR(0, 1386, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1391
 *     USABLE_WITH_DARK_BACKGROUND = HB_OT_COLOR_PALETTE_FLAG_USABLE_WITH_DARK_BACKGROUND
 * 
 * def ot_color_palette_get_flags(face: Face, palette_index: int) -> OTColorPaletteFlags:             # <<<<<<<<<<<<<<
 *     return OTColorPaletteFlags(hb_ot_color_palette_get_flags(face._hb_face, palette_index))
 * 
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_face, __pyx_n_s_Face) < 0) __PYX_ERR(0, 1391, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_palette_index, __pyx_n_s_int) < 0) __PYX_ERR(0, 1391, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_OTColorPaletteFlags) < 0) __PYX_ERR(0, 1391, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_31ot_color_palette_get_flags, 0, __pyx_n_s_ot_color_palette_get_flags, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__144)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_color_palette_get_flags, __pyx_t_4) < 0) __PYX_ERR(0, 1391, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1394
 *     return OTColorPaletteFlags(hb_ot_color_palette_get_flags(face._hb_face, palette_index))
 * 
 * def ot_color_palette_get_colors(face: Face, palette_index: int) -> List[Color]:             # <<<<<<<<<<<<<<
 *     cdef list ret = []
 *     cdef unsigned int i
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_face, __pyx_n_s_Face) < 0) __PYX_ERR(0, 1394, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_palette_index, __pyx_n_s_int) < 0) __PYX_ERR(0, 1394, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_return, __pyx_kp_s_List_Color) < 0) __PYX_ERR(0, 1394, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_33ot_color_palette_get_colors, 0, __pyx_n_s_ot_color_palette_get_colors, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__146)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_color_palette_get_colors, __pyx_t_2) < 0) __PYX_ERR(0, 1394, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1406
 *     return ret
 * 
 * def ot_color_palette_get_name_id(face: Face, palette_index: int) -> int | None:             # <<<<<<<<<<<<<<
 *     cdef hb_ot_name_id_t name_id
 *     name_id = hb_ot_color_palette_get_name_id(face._hb_face, palette_index)
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_face, __pyx_n_s_Face) < 0) __PYX_ERR(0, 1406, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_palette_index, __pyx_n_s_int) < 0) __PYX_ERR(0, 1406, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_kp_s_int_None) < 0) __PYX_ERR(0, 1406, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_35ot_color_palette_get_name_id, 0, __pyx_n_s_ot_color_palette_get_name_id, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__148)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_color_palette_get_name_id, __pyx_t_4) < 0) __PYX_ERR(0, 1406, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1413
 *     return name_id
 * 
 * def ot_color_palette_color_get_name_id(face: Face, color_index: int) -> int | None:             # <<<<<<<<<<<<<<
 *     cdef hb_ot_name_id_t name_id
 *     name_id =  hb_ot_color_palette_color_get_name_id(face._hb_face, color_index)
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_face, __pyx_n_s_Face) < 0) __PYX_ERR(0, 1413, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_color_index, __pyx_n_s_int) < 0) __PYX_ERR(0, 1413, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_return, __pyx_kp_s_int_None) < 0) __PYX_ERR(0, 1413, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_37ot_color_palette_color_get_name_id, 0, __pyx_n_s_ot_color_palette_color_get_name, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__150)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_color_palette_color_get_name, __pyx_t_2) < 0) __PYX_ERR(0, 1413, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1420
 *     return name_id
 * 
 * def ot_color_has_layers(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_color_has_layers(face._hb_face)
 * 
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_face, __pyx_n_s_Face) < 0) __PYX_ERR(0, 1420, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 1420, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_39ot_color_has_layers, 0, __pyx_n_s_ot_color_has_layers, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__151)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_color_has_layers, __pyx_t_4) < 0) __PYX_ERR(0, 1420, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1423
 *     return hb_ot_color_has_layers(face._hb_face)
 * 
 * OTColorLayer = namedtuple("OTColorLayer", ["glyph", "color_index"])             # <<<<<<<<<<<<<<
 * 
 * def ot_color_glyph_get_layers(face: Face, glyph: int) -> List[OTColorLayer]:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_namedtuple); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_u_glyph);
  __Pyx_GIVEREF(__pyx_n_u_glyph);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_u_glyph)) __PYX_ERR(0, 1423, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_color_index);
  __Pyx_GIVEREF(__pyx_n_u_color_index);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_u_color_index)) __PYX_ERR(0, 1423, __pyx_L1_error);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_u_OTColorLayer);
  __Pyx_GIVEREF(__pyx_n_u_OTColorLayer);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_n_u_OTColorLayer)) __PYX_ERR(0, 1423, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2)) __PYX_ERR(0, 1423, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OTColorLayer, __pyx_t_2) < 0) __PYX_ERR(0, 1423, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1425
 * OTColorLayer = namedtuple("OTColorLayer", ["glyph", "color_index"])
 * 
 * def ot_color_glyph_get_layers(face: Face, glyph: int) -> List[OTColorLayer]:             # <<<<<<<<<<<<<<
 *     cdef list ret = []
 *     cdef unsigned int i
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_face, __pyx_n_s_Face) < 0) __PYX_ERR(0, 1425, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_glyph, __pyx_n_s_int) < 0) __PYX_ERR(0, 1425, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_kp_s_List_OTColorLayer) < 0) __PYX_ERR(0, 1425, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_41ot_color_glyph_get_layers, 0, __pyx_n_s_ot_color_glyph_get_layers, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__153)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1425, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_color_glyph_get_layers, __pyx_t_3) < 0) __PYX_ERR(0, 1425, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1438
 *     return ret
 * 
 * def ot_color_has_paint(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_color_has_paint(face._hb_face)
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_face, __pyx_n_s_Face) < 0) __PYX_ERR(0, 1438, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 1438, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_43ot_color_has_paint, 0, __pyx_n_s_ot_color_has_paint, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__154)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_color_has_paint, __pyx_t_2) < 0) __PYX_ERR(0, 1438, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1441
 *     return hb_ot_color_has_paint(face._hb_face)
 * 
 * def ot_color_glyph_has_paint(face: Face, glyph: int) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_color_glyph_has_paint(face._hb_face, glyph)
 * 
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_face, __pyx_n_s_Face) < 0) __PYX_ERR(0, 1441, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_glyph, __pyx_n_s_int) < 0) __PYX_ERR(0, 1441, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 1441, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_45ot_color_glyph_has_paint, 0, __pyx_n_s_ot_color_glyph_has_paint, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__155)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_color_glyph_has_paint, __pyx_t_3) < 0) __PYX_ERR(0, 1441, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1444
 *     return hb_ot_color_glyph_has_paint(face._hb_face, glyph)
 * 
 * def ot_color_has_svg(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_color_has_svg(face._hb_face)
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_face, __pyx_n_s_Face) < 0) __PYX_ERR(0, 1444, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 1444, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_47ot_color_has_svg, 0, __pyx_n_s_ot_color_has_svg, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__156)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_color_has_svg, __pyx_t_2) < 0) __PYX_ERR(0, 1444, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1447
 *     return hb_ot_color_has_svg(face._hb_face)
 * 
 * def ot_color_glyph_get_svg(face: Face, glyph: int) -> Blob:             # <<<<<<<<<<<<<<
 *     cdef hb_blob_t* blob
 *     blob = hb_ot_color_glyph_reference_svg(face._hb_face, glyph)
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_face, __pyx_n_s_Face) < 0) __PYX_ERR(0, 1447, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_glyph, __pyx_n_s_int) < 0) __PYX_ERR(0, 1447, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_Blob) < 0) __PYX_ERR(0, 1447, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_49ot_color_glyph_get_svg, 0, __pyx_n_s_ot_color_glyph_get_svg, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__158)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_color_glyph_get_svg, __pyx_t_3) < 0) __PYX_ERR(0, 1447, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1452
 *     return Blob.from_ptr(blob)
 * 
 * def ot_color_has_png(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_color_has_png(face._hb_face)
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_face, __pyx_n_s_Face) < 0) __PYX_ERR(0, 1452, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 1452, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_51ot_color_has_png, 0, __pyx_n_s_ot_color_has_png, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__159)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_color_has_png, __pyx_t_2) < 0) __PYX_ERR(0, 1452, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1455
 *     return hb_ot_color_has_png(face._hb_face)
 * 
 * def ot_color_glyph_get_png(font: Font, glyph: int) -> Blob:             # <<<<<<<<<<<<<<
 *     cdef hb_blob_t* blob
 *     blob = hb_ot_color_glyph_reference_png(font._hb_font, glyph)
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_font, __pyx_n_s_Font) < 0) __PYX_ERR(0, 1455, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_glyph, __pyx_n_s_int) < 0) __PYX_ERR(0, 1455, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_Blob) < 0) __PYX_ERR(0, 1455, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_53ot_color_glyph_get_png, 0, __pyx_n_s_ot_color_glyph_get_png, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__161)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_color_glyph_get_png, __pyx_t_3) < 0) __PYX_ERR(0, 1455, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1461
 * 
 * 
 * def ot_math_has_data(face: Face) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_math_has_data(face._hb_face)
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_face, __pyx_n_s_Face) < 0) __PYX_ERR(0, 1461, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 1461, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_55ot_math_has_data, 0, __pyx_n_s_ot_math_has_data, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__162)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_math_has_data, __pyx_t_2) < 0) __PYX_ERR(0, 1461, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1464
 *     return hb_ot_math_has_data(face._hb_face)
 * 
 * class OTMathConstant(IntEnum):             # <<<<<<<<<<<<<<
 *     SCRIPT_PERCENT_SCALE_DOWN = HB_OT_MATH_CONSTANT_SCRIPT_PERCENT_SCALE_DOWN
 *     SCRIPT_SCRIPT_PERCENT_SCALE_DOWN = HB_OT_MATH_CONSTANT_SCRIPT_SCRIPT_PERCENT_SCALE_DOWN
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_IntEnum); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2)) __PYX_ERR(0, 1464, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PEP560_update_bases(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_Py3MetaclassPrepare(__pyx_t_4, __pyx_t_2, __pyx_n_s_OTMathConstant, __pyx_n_s_OTMathConstant, (PyObject *) NULL, __pyx_n_s_uharfbuzz__harfbuzz, (PyObject *) NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__pyx_t_2 != __pyx_t_3) {
    if (unlikely((PyDict_SetItemString(__pyx_t_5, "__orig_bases__", __pyx_t_3) < 0))) __PYX_ERR(0, 1464, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1465
 * 
 * class OTMathConstant(IntEnum):
 *     SCRIPT_PERCENT_SCALE_DOWN = HB_OT_MATH_CONSTANT_SCRIPT_PERCENT_SCALE_DOWN             # <<<<<<<<<<<<<<
 *     SCRIPT_SCRIPT_PERCENT_SCALE_DOWN = HB_OT_MATH_CONSTANT_SCRIPT_SCRIPT_PERCENT_SCALE_DOWN
 *     DELIMITED_SUB_FORMULA_MIN_HEIGHT = HB_OT_MATH_CONSTANT_DELIMITED_SUB_FORMULA_MIN_HEIGHT
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_SCRIPT_PERCENT_SCALE_DOWN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_SCRIPT_PERCENT_SCALE_DOWN, __pyx_t_3) < 0) __PYX_ERR(0, 1465, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1466
 * class OTMathConstant(IntEnum):
 *     SCRIPT_PERCENT_SCALE_DOWN = HB_OT_MATH_CONSTANT_SCRIPT_PERCENT_SCALE_DOWN
 *     SCRIPT_SCRIPT_PERCENT_SCALE_DOWN = HB_OT_MATH_CONSTANT_SCRIPT_SCRIPT_PERCENT_SCALE_DOWN             # <<<<<<<<<<<<<<
 *     DELIMITED_SUB_FORMULA_MIN_HEIGHT = HB_OT_MATH_CONSTANT_DELIMITED_SUB_FORMULA_MIN_HEIGHT
 *     DISPLAY_OPERATOR_MIN_HEIGHT = HB_OT_MATH_CONSTANT_DISPLAY_OPERATOR_MIN_HEIGHT
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_SCRIPT_SCRIPT_PERCENT_SCALE_DOWN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_SCRIPT_SCRIPT_PERCENT_SCALE_DOWN, __pyx_t_3) < 0) __PYX_ERR(0, 1466, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1467
 *     SCRIPT_PERCENT_SCALE_DOWN = HB_OT_MATH_CONSTANT_SCRIPT_PERCENT_SCALE_DOWN
 *     SCRIPT_SCRIPT_PERCENT_SCALE_DOWN = HB_OT_MATH_CONSTANT_SCRIPT_SCRIPT_PERCENT_SCALE_DOWN
 *     DELIMITED_SUB_FORMULA_MIN_HEIGHT = HB_OT_MATH_CONSTANT_DELIMITED_SUB_FORMULA_MIN_HEIGHT             # <<<<<<<<<<<<<<
 *     DISPLAY_OPERATOR_MIN_HEIGHT = HB_OT_MATH_CONSTANT_DISPLAY_OPERATOR_MIN_HEIGHT
 *     MATH_LEADING = HB_OT_MATH_CONSTANT_MATH_LEADING
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_DELIMITED_SUB_FORMULA_MIN_HEIGHT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_DELIMITED_SUB_FORMULA_MIN_HEIGHT, __pyx_t_3) < 0) __PYX_ERR(0, 1467, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1468
 *     SCRIPT_SCRIPT_PERCENT_SCALE_DOWN = HB_OT_MATH_CONSTANT_SCRIPT_SCRIPT_PERCENT_SCALE_DOWN
 *     DELIMITED_SUB_FORMULA_MIN_HEIGHT = HB_OT_MATH_CONSTANT_DELIMITED_SUB_FORMULA_MIN_HEIGHT
 *     DISPLAY_OPERATOR_MIN_HEIGHT = HB_OT_MATH_CONSTANT_DISPLAY_OPERATOR_MIN_HEIGHT             # <<<<<<<<<<<<<<
 *     MATH_LEADING = HB_OT_MATH_CONSTANT_MATH_LEADING
 *     AXIS_HEIGHT = HB_OT_MATH_CONSTANT_AXIS_HEIGHT
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_DISPLAY_OPERATOR_MIN_HEIGHT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_DISPLAY_OPERATOR_MIN_HEIGHT, __pyx_t_3) < 0) __PYX_ERR(0, 1468, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1469
 *     DELIMITED_SUB_FORMULA_MIN_HEIGHT = HB_OT_MATH_CONSTANT_DELIMITED_SUB_FORMULA_MIN_HEIGHT
 *     DISPLAY_OPERATOR_MIN_HEIGHT = HB_OT_MATH_CONSTANT_DISPLAY_OPERATOR_MIN_HEIGHT
 *     MATH_LEADING = HB_OT_MATH_CONSTANT_MATH_LEADING             # <<<<<<<<<<<<<<
 *     AXIS_HEIGHT = HB_OT_MATH_CONSTANT_AXIS_HEIGHT
 *     ACCENT_BASE_HEIGHT = HB_OT_MATH_CONSTANT_ACCENT_BASE_HEIGHT
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_MATH_LEADING); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_MATH_LEADING, __pyx_t_3) < 0) __PYX_ERR(0, 1469, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1470
 *     DISPLAY_OPERATOR_MIN_HEIGHT = HB_OT_MATH_CONSTANT_DISPLAY_OPERATOR_MIN_HEIGHT
 *     MATH_LEADING = HB_OT_MATH_CONSTANT_MATH_LEADING
 *     AXIS_HEIGHT = HB_OT_MATH_CONSTANT_AXIS_HEIGHT             # <<<<<<<<<<<<<<
 *     ACCENT_BASE_HEIGHT = HB_OT_MATH_CONSTANT_ACCENT_BASE_HEIGHT
 *     FLATTENED_ACCENT_BASE_HEIGHT = HB_OT_MATH_CONSTANT_FLATTENED_ACCENT_BASE_HEIGHT
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_AXIS_HEIGHT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_AXIS_HEIGHT, __pyx_t_3) < 0) __PYX_ERR(0, 1470, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1471
 *     MATH_LEADING = HB_OT_MATH_CONSTANT_MATH_LEADING
 *     AXIS_HEIGHT = HB_OT_MATH_CONSTANT_AXIS_HEIGHT
 *     ACCENT_BASE_HEIGHT = HB_OT_MATH_CONSTANT_ACCENT_BASE_HEIGHT             # <<<<<<<<<<<<<<
 *     FLATTENED_ACCENT_BASE_HEIGHT = HB_OT_MATH_CONSTANT_FLATTENED_ACCENT_BASE_HEIGHT
 *     SUBSCRIPT_SHIFT_DOWN = HB_OT_MATH_CONSTANT_SUBSCRIPT_SHIFT_DOWN
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_ACCENT_BASE_HEIGHT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_ACCENT_BASE_HEIGHT, __pyx_t_3) < 0) __PYX_ERR(0, 1471, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1472
 *     AXIS_HEIGHT = HB_OT_MATH_CONSTANT_AXIS_HEIGHT
 *     ACCENT_BASE_HEIGHT = HB_OT_MATH_CONSTANT_ACCENT_BASE_HEIGHT
 *     FLATTENED_ACCENT_BASE_HEIGHT = HB_OT_MATH_CONSTANT_FLATTENED_ACCENT_BASE_HEIGHT             # <<<<<<<<<<<<<<
 *     SUBSCRIPT_SHIFT_DOWN = HB_OT_MATH_CONSTANT_SUBSCRIPT_SHIFT_DOWN
 *     SUBSCRIPT_TOP_MAX = HB_OT_MATH_CONSTANT_SUBSCRIPT_TOP_MAX
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_FLATTENED_ACCENT_BASE_HEIGHT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1472, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_FLATTENED_ACCENT_BASE_HEIGHT, __pyx_t_3) < 0) __PYX_ERR(0, 1472, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1473
 *     ACCENT_BASE_HEIGHT = HB_OT_MATH_CONSTANT_ACCENT_BASE_HEIGHT
 *     FLATTENED_ACCENT_BASE_HEIGHT = HB_OT_MATH_CONSTANT_FLATTENED_ACCENT_BASE_HEIGHT
 *     SUBSCRIPT_SHIFT_DOWN = HB_OT_MATH_CONSTANT_SUBSCRIPT_SHIFT_DOWN             # <<<<<<<<<<<<<<
 *     SUBSCRIPT_TOP_MAX = HB_OT_MATH_CONSTANT_SUBSCRIPT_TOP_MAX
 *     SUBSCRIPT_BASELINE_DROP_MIN = HB_OT_MATH_CONSTANT_SUBSCRIPT_BASELINE_DROP_MIN
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_SUBSCRIPT_SHIFT_DOWN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_SUBSCRIPT_SHIFT_DOWN, __pyx_t_3) < 0) __PYX_ERR(0, 1473, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1474
 *     FLATTENED_ACCENT_BASE_HEIGHT = HB_OT_MATH_CONSTANT_FLATTENED_ACCENT_BASE_HEIGHT
 *     SUBSCRIPT_SHIFT_DOWN = HB_OT_MATH_CONSTANT_SUBSCRIPT_SHIFT_DOWN
 *     SUBSCRIPT_TOP_MAX = HB_OT_MATH_CONSTANT_SUBSCRIPT_TOP_MAX             # <<<<<<<<<<<<<<
 *     SUBSCRIPT_BASELINE_DROP_MIN = HB_OT_MATH_CONSTANT_SUBSCRIPT_BASELINE_DROP_MIN
 *     SUPERSCRIPT_SHIFT_UP = HB_OT_MATH_CONSTANT_SUPERSCRIPT_SHIFT_UP
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_SUBSCRIPT_TOP_MAX); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_SUBSCRIPT_TOP_MAX, __pyx_t_3) < 0) __PYX_ERR(0, 1474, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1475
 *     SUBSCRIPT_SHIFT_DOWN = HB_OT_MATH_CONSTANT_SUBSCRIPT_SHIFT_DOWN
 *     SUBSCRIPT_TOP_MAX = HB_OT_MATH_CONSTANT_SUBSCRIPT_TOP_MAX
 *     SUBSCRIPT_BASELINE_DROP_MIN = HB_OT_MATH_CONSTANT_SUBSCRIPT_BASELINE_DROP_MIN             # <<<<<<<<<<<<<<
 *     SUPERSCRIPT_SHIFT_UP = HB_OT_MATH_CONSTANT_SUPERSCRIPT_SHIFT_UP
 *     SUPERSCRIPT_SHIFT_UP_CRAMPED = HB_OT_MATH_CONSTANT_SUPERSCRIPT_SHIFT_UP_CRAMPED
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_SUBSCRIPT_BASELINE_DROP_MIN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_SUBSCRIPT_BASELINE_DROP_MIN, __pyx_t_3) < 0) __PYX_ERR(0, 1475, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1476
 *     SUBSCRIPT_TOP_MAX = HB_OT_MATH_CONSTANT_SUBSCRIPT_TOP_MAX
 *     SUBSCRIPT_BASELINE_DROP_MIN = HB_OT_MATH_CONSTANT_SUBSCRIPT_BASELINE_DROP_MIN
 *     SUPERSCRIPT_SHIFT_UP = HB_OT_MATH_CONSTANT_SUPERSCRIPT_SHIFT_UP             # <<<<<<<<<<<<<<
 *     SUPERSCRIPT_SHIFT_UP_CRAMPED = HB_OT_MATH_CONSTANT_SUPERSCRIPT_SHIFT_UP_CRAMPED
 *     SUPERSCRIPT_BOTTOM_MIN = HB_OT_MATH_CONSTANT_SUPERSCRIPT_BOTTOM_MIN
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_SUPERSCRIPT_SHIFT_UP); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1476, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_SUPERSCRIPT_SHIFT_UP, __pyx_t_3) < 0) __PYX_ERR(0, 1476, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1477
 *     SUBSCRIPT_BASELINE_DROP_MIN = HB_OT_MATH_CONSTANT_SUBSCRIPT_BASELINE_DROP_MIN
 *     SUPERSCRIPT_SHIFT_UP = HB_OT_MATH_CONSTANT_SUPERSCRIPT_SHIFT_UP
 *     SUPERSCRIPT_SHIFT_UP_CRAMPED = HB_OT_MATH_CONSTANT_SUPERSCRIPT_SHIFT_UP_CRAMPED             # <<<<<<<<<<<<<<
 *     SUPERSCRIPT_BOTTOM_MIN = HB_OT_MATH_CONSTANT_SUPERSCRIPT_BOTTOM_MIN
 *     SUPERSCRIPT_BASELINE_DROP_MAX = HB_OT_MATH_CONSTANT_SUPERSCRIPT_BASELINE_DROP_MAX
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_SUPERSCRIPT_SHIFT_UP_CRAMPED); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1477, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_SUPERSCRIPT_SHIFT_UP_CRAMPED, __pyx_t_3) < 0) __PYX_ERR(0, 1477, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1478
 *     SUPERSCRIPT_SHIFT_UP = HB_OT_MATH_CONSTANT_SUPERSCRIPT_SHIFT_UP
 *     SUPERSCRIPT_SHIFT_UP_CRAMPED = HB_OT_MATH_CONSTANT_SUPERSCRIPT_SHIFT_UP_CRAMPED
 *     SUPERSCRIPT_BOTTOM_MIN = HB_OT_MATH_CONSTANT_SUPERSCRIPT_BOTTOM_MIN             # <<<<<<<<<<<<<<
 *     SUPERSCRIPT_BASELINE_DROP_MAX = HB_OT_MATH_CONSTANT_SUPERSCRIPT_BASELINE_DROP_MAX
 *     SUB_SUPERSCRIPT_GAP_MIN = HB_OT_MATH_CONSTANT_SUB_SUPERSCRIPT_GAP_MIN
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_SUPERSCRIPT_BOTTOM_MIN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1478, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_SUPERSCRIPT_BOTTOM_MIN, __pyx_t_3) < 0) __PYX_ERR(0, 1478, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1479
 *     SUPERSCRIPT_SHIFT_UP_CRAMPED = HB_OT_MATH_CONSTANT_SUPERSCRIPT_SHIFT_UP_CRAMPED
 *     SUPERSCRIPT_BOTTOM_MIN = HB_OT_MATH_CONSTANT_SUPERSCRIPT_BOTTOM_MIN
 *     SUPERSCRIPT_BASELINE_DROP_MAX = HB_OT_MATH_CONSTANT_SUPERSCRIPT_BASELINE_DROP_MAX             # <<<<<<<<<<<<<<
 *     SUB_SUPERSCRIPT_GAP_MIN = HB_OT_MATH_CONSTANT_SUB_SUPERSCRIPT_GAP_MIN
 *     SUPERSCRIPT_BOTTOM_MAX_WITH_SUBSCRIPT = HB_OT_MATH_CONSTANT_SUPERSCRIPT_BOTTOM_MAX_WITH_SUBSCRIPT
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_SUPERSCRIPT_BASELINE_DROP_MAX); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1479, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_SUPERSCRIPT_BASELINE_DROP_MAX, __pyx_t_3) < 0) __PYX_ERR(0, 1479, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1480
 *     SUPERSCRIPT_BOTTOM_MIN = HB_OT_MATH_CONSTANT_SUPERSCRIPT_BOTTOM_MIN
 *     SUPERSCRIPT_BASELINE_DROP_MAX = HB_OT_MATH_CONSTANT_SUPERSCRIPT_BASELINE_DROP_MAX
 *     SUB_SUPERSCRIPT_GAP_MIN = HB_OT_MATH_CONSTANT_SUB_SUPERSCRIPT_GAP_MIN             # <<<<<<<<<<<<<<
 *     SUPERSCRIPT_BOTTOM_MAX_WITH_SUBSCRIPT = HB_OT_MATH_CONSTANT_SUPERSCRIPT_BOTTOM_MAX_WITH_SUBSCRIPT
 *     SPACE_AFTER_SCRIPT = HB_OT_MATH_CONSTANT_SPACE_AFTER_SCRIPT
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_SUB_SUPERSCRIPT_GAP_MIN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_SUB_SUPERSCRIPT_GAP_MIN, __pyx_t_3) < 0) __PYX_ERR(0, 1480, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1481
 *     SUPERSCRIPT_BASELINE_DROP_MAX = HB_OT_MATH_CONSTANT_SUPERSCRIPT_BASELINE_DROP_MAX
 *     SUB_SUPERSCRIPT_GAP_MIN = HB_OT_MATH_CONSTANT_SUB_SUPERSCRIPT_GAP_MIN
 *     SUPERSCRIPT_BOTTOM_MAX_WITH_SUBSCRIPT = HB_OT_MATH_CONSTANT_SUPERSCRIPT_BOTTOM_MAX_WITH_SUBSCRIPT             # <<<<<<<<<<<<<<
 *     SPACE_AFTER_SCRIPT = HB_OT_MATH_CONSTANT_SPACE_AFTER_SCRIPT
 *     UPPER_LIMIT_GAP_MIN = HB_OT_MATH_CONSTANT_UPPER_LIMIT_GAP_MIN
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_SUPERSCRIPT_BOTTOM_MAX_WITH_SUBSCRIPT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_SUPERSCRIPT_BOTTOM_MAX_WITH_SUBS, __pyx_t_3) < 0) __PYX_ERR(0, 1481, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1482
 *     SUB_SUPERSCRIPT_GAP_MIN = HB_OT_MATH_CONSTANT_SUB_SUPERSCRIPT_GAP_MIN
 *     SUPERSCRIPT_BOTTOM_MAX_WITH_SUBSCRIPT = HB_OT_MATH_CONSTANT_SUPERSCRIPT_BOTTOM_MAX_WITH_SUBSCRIPT
 *     SPACE_AFTER_SCRIPT = HB_OT_MATH_CONSTANT_SPACE_AFTER_SCRIPT             # <<<<<<<<<<<<<<
 *     UPPER_LIMIT_GAP_MIN = HB_OT_MATH_CONSTANT_UPPER_LIMIT_GAP_MIN
 *     UPPER_LIMIT_BASELINE_RISE_MIN = HB_OT_MATH_CONSTANT_UPPER_LIMIT_BASELINE_RISE_MIN
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_SPACE_AFTER_SCRIPT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_SPACE_AFTER_SCRIPT, __pyx_t_3) < 0) __PYX_ERR(0, 1482, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1483
 *     SUPERSCRIPT_BOTTOM_MAX_WITH_SUBSCRIPT = HB_OT_MATH_CONSTANT_SUPERSCRIPT_BOTTOM_MAX_WITH_SUBSCRIPT
 *     SPACE_AFTER_SCRIPT = HB_OT_MATH_CONSTANT_SPACE_AFTER_SCRIPT
 *     UPPER_LIMIT_GAP_MIN = HB_OT_MATH_CONSTANT_UPPER_LIMIT_GAP_MIN             # <<<<<<<<<<<<<<
 *     UPPER_LIMIT_BASELINE_RISE_MIN = HB_OT_MATH_CONSTANT_UPPER_LIMIT_BASELINE_RISE_MIN
 *     LOWER_LIMIT_GAP_MIN = HB_OT_MATH_CONSTANT_LOWER_LIMIT_GAP_MIN
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_UPPER_LIMIT_GAP_MIN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_UPPER_LIMIT_GAP_MIN, __pyx_t_3) < 0) __PYX_ERR(0, 1483, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1484
 *     SPACE_AFTER_SCRIPT = HB_OT_MATH_CONSTANT_SPACE_AFTER_SCRIPT
 *     UPPER_LIMIT_GAP_MIN = HB_OT_MATH_CONSTANT_UPPER_LIMIT_GAP_MIN
 *     UPPER_LIMIT_BASELINE_RISE_MIN = HB_OT_MATH_CONSTANT_UPPER_LIMIT_BASELINE_RISE_MIN             # <<<<<<<<<<<<<<
 *     LOWER_LIMIT_GAP_MIN = HB_OT_MATH_CONSTANT_LOWER_LIMIT_GAP_MIN
 *     LOWER_LIMIT_BASELINE_DROP_MIN = HB_OT_MATH_CONSTANT_LOWER_LIMIT_BASELINE_DROP_MIN
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_UPPER_LIMIT_BASELINE_RISE_MIN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_UPPER_LIMIT_BASELINE_RISE_MIN, __pyx_t_3) < 0) __PYX_ERR(0, 1484, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1485
 *     UPPER_LIMIT_GAP_MIN = HB_OT_MATH_CONSTANT_UPPER_LIMIT_GAP_MIN
 *     UPPER_LIMIT_BASELINE_RISE_MIN = HB_OT_MATH_CONSTANT_UPPER_LIMIT_BASELINE_RISE_MIN
 *     LOWER_LIMIT_GAP_MIN = HB_OT_MATH_CONSTANT_LOWER_LIMIT_GAP_MIN             # <<<<<<<<<<<<<<
 *     LOWER_LIMIT_BASELINE_DROP_MIN = HB_OT_MATH_CONSTANT_LOWER_LIMIT_BASELINE_DROP_MIN
 *     STACK_TOP_SHIFT_UP = HB_OT_MATH_CONSTANT_STACK_TOP_SHIFT_UP
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_LOWER_LIMIT_GAP_MIN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_LOWER_LIMIT_GAP_MIN, __pyx_t_3) < 0) __PYX_ERR(0, 1485, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1486
 *     UPPER_LIMIT_BASELINE_RISE_MIN = HB_OT_MATH_CONSTANT_UPPER_LIMIT_BASELINE_RISE_MIN
 *     LOWER_LIMIT_GAP_MIN = HB_OT_MATH_CONSTANT_LOWER_LIMIT_GAP_MIN
 *     LOWER_LIMIT_BASELINE_DROP_MIN = HB_OT_MATH_CONSTANT_LOWER_LIMIT_BASELINE_DROP_MIN             # <<<<<<<<<<<<<<
 *     STACK_TOP_SHIFT_UP = HB_OT_MATH_CONSTANT_STACK_TOP_SHIFT_UP
 *     STACK_TOP_DISPLAY_STYLE_SHIFT_UP = HB_OT_MATH_CONSTANT_STACK_TOP_DISPLAY_STYLE_SHIFT_UP
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_LOWER_LIMIT_BASELINE_DROP_MIN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_LOWER_LIMIT_BASELINE_DROP_MIN, __pyx_t_3) < 0) __PYX_ERR(0, 1486, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1487
 *     LOWER_LIMIT_GAP_MIN = HB_OT_MATH_CONSTANT_LOWER_LIMIT_GAP_MIN
 *     LOWER_LIMIT_BASELINE_DROP_MIN = HB_OT_MATH_CONSTANT_LOWER_LIMIT_BASELINE_DROP_MIN
 *     STACK_TOP_SHIFT_UP = HB_OT_MATH_CONSTANT_STACK_TOP_SHIFT_UP             # <<<<<<<<<<<<<<
 *     STACK_TOP_DISPLAY_STYLE_SHIFT_UP = HB_OT_MATH_CONSTANT_STACK_TOP_DISPLAY_STYLE_SHIFT_UP
 *     STACK_BOTTOM_SHIFT_DOWN = HB_OT_MATH_CONSTANT_STACK_BOTTOM_SHIFT_DOWN
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_STACK_TOP_SHIFT_UP); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1487, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_STACK_TOP_SHIFT_UP, __pyx_t_3) < 0) __PYX_ERR(0, 1487, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1488
 *     LOWER_LIMIT_BASELINE_DROP_MIN = HB_OT_MATH_CONSTANT_LOWER_LIMIT_BASELINE_DROP_MIN
 *     STACK_TOP_SHIFT_UP = HB_OT_MATH_CONSTANT_STACK_TOP_SHIFT_UP
 *     STACK_TOP_DISPLAY_STYLE_SHIFT_UP = HB_OT_MATH_CONSTANT_STACK_TOP_DISPLAY_STYLE_SHIFT_UP             # <<<<<<<<<<<<<<
 *     STACK_BOTTOM_SHIFT_DOWN = HB_OT_MATH_CONSTANT_STACK_BOTTOM_SHIFT_DOWN
 *     STACK_BOTTOM_DISPLAY_STYLE_SHIFT_DOWN = HB_OT_MATH_CONSTANT_STACK_BOTTOM_DISPLAY_STYLE_SHIFT_DOWN
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_STACK_TOP_DISPLAY_STYLE_SHIFT_UP); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_STACK_TOP_DISPLAY_STYLE_SHIFT_UP, __pyx_t_3) < 0) __PYX_ERR(0, 1488, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1489
 *     STACK_TOP_SHIFT_UP = HB_OT_MATH_CONSTANT_STACK_TOP_SHIFT_UP
 *     STACK_TOP_DISPLAY_STYLE_SHIFT_UP = HB_OT_MATH_CONSTANT_STACK_TOP_DISPLAY_STYLE_SHIFT_UP
 *     STACK_BOTTOM_SHIFT_DOWN = HB_OT_MATH_CONSTANT_STACK_BOTTOM_SHIFT_DOWN             # <<<<<<<<<<<<<<
 *     STACK_BOTTOM_DISPLAY_STYLE_SHIFT_DOWN = HB_OT_MATH_CONSTANT_STACK_BOTTOM_DISPLAY_STYLE_SHIFT_DOWN
 *     STACK_GAP_MIN = HB_OT_MATH_CONSTANT_STACK_GAP_MIN
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_STACK_BOTTOM_SHIFT_DOWN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_STACK_BOTTOM_SHIFT_DOWN, __pyx_t_3) < 0) __PYX_ERR(0, 1489, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1490
 *     STACK_TOP_DISPLAY_STYLE_SHIFT_UP = HB_OT_MATH_CONSTANT_STACK_TOP_DISPLAY_STYLE_SHIFT_UP
 *     STACK_BOTTOM_SHIFT_DOWN = HB_OT_MATH_CONSTANT_STACK_BOTTOM_SHIFT_DOWN
 *     STACK_BOTTOM_DISPLAY_STYLE_SHIFT_DOWN = HB_OT_MATH_CONSTANT_STACK_BOTTOM_DISPLAY_STYLE_SHIFT_DOWN             # <<<<<<<<<<<<<<
 *     STACK_GAP_MIN = HB_OT_MATH_CONSTANT_STACK_GAP_MIN
 *     STACK_DISPLAY_STYLE_GAP_MIN = HB_OT_MATH_CONSTANT_STACK_DISPLAY_STYLE_GAP_MIN
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_STACK_BOTTOM_DISPLAY_STYLE_SHIFT_DOWN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_STACK_BOTTOM_DISPLAY_STYLE_SHIFT, __pyx_t_3) < 0) __PYX_ERR(0, 1490, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1491
 *     STACK_BOTTOM_SHIFT_DOWN = HB_OT_MATH_CONSTANT_STACK_BOTTOM_SHIFT_DOWN
 *     STACK_BOTTOM_DISPLAY_STYLE_SHIFT_DOWN = HB_OT_MATH_CONSTANT_STACK_BOTTOM_DISPLAY_STYLE_SHIFT_DOWN
 *     STACK_GAP_MIN = HB_OT_MATH_CONSTANT_STACK_GAP_MIN             # <<<<<<<<<<<<<<
 *     STACK_DISPLAY_STYLE_GAP_MIN = HB_OT_MATH_CONSTANT_STACK_DISPLAY_STYLE_GAP_MIN
 *     STRETCH_STACK_TOP_SHIFT_UP = HB_OT_MATH_CONSTANT_STRETCH_STACK_TOP_SHIFT_UP
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_STACK_GAP_MIN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_STACK_GAP_MIN, __pyx_t_3) < 0) __PYX_ERR(0, 1491, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1492
 *     STACK_BOTTOM_DISPLAY_STYLE_SHIFT_DOWN = HB_OT_MATH_CONSTANT_STACK_BOTTOM_DISPLAY_STYLE_SHIFT_DOWN
 *     STACK_GAP_MIN = HB_OT_MATH_CONSTANT_STACK_GAP_MIN
 *     STACK_DISPLAY_STYLE_GAP_MIN = HB_OT_MATH_CONSTANT_STACK_DISPLAY_STYLE_GAP_MIN             # <<<<<<<<<<<<<<
 *     STRETCH_STACK_TOP_SHIFT_UP = HB_OT_MATH_CONSTANT_STRETCH_STACK_TOP_SHIFT_UP
 *     STRETCH_STACK_BOTTOM_SHIFT_DOWN = HB_OT_MATH_CONSTANT_STRETCH_STACK_BOTTOM_SHIFT_DOWN
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_STACK_DISPLAY_STYLE_GAP_MIN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_STACK_DISPLAY_STYLE_GAP_MIN, __pyx_t_3) < 0) __PYX_ERR(0, 1492, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1493
 *     STACK_GAP_MIN = HB_OT_MATH_CONSTANT_STACK_GAP_MIN
 *     STACK_DISPLAY_STYLE_GAP_MIN = HB_OT_MATH_CONSTANT_STACK_DISPLAY_STYLE_GAP_MIN
 *     STRETCH_STACK_TOP_SHIFT_UP = HB_OT_MATH_CONSTANT_STRETCH_STACK_TOP_SHIFT_UP             # <<<<<<<<<<<<<<
 *     STRETCH_STACK_BOTTOM_SHIFT_DOWN = HB_OT_MATH_CONSTANT_STRETCH_STACK_BOTTOM_SHIFT_DOWN
 *     STRETCH_STACK_GAP_ABOVE_MIN = HB_OT_MATH_CONSTANT_STRETCH_STACK_GAP_ABOVE_MIN
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_STRETCH_STACK_TOP_SHIFT_UP); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_STRETCH_STACK_TOP_SHIFT_UP, __pyx_t_3) < 0) __PYX_ERR(0, 1493, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1494
 *     STACK_DISPLAY_STYLE_GAP_MIN = HB_OT_MATH_CONSTANT_STACK_DISPLAY_STYLE_GAP_MIN
 *     STRETCH_STACK_TOP_SHIFT_UP = HB_OT_MATH_CONSTANT_STRETCH_STACK_TOP_SHIFT_UP
 *     STRETCH_STACK_BOTTOM_SHIFT_DOWN = HB_OT_MATH_CONSTANT_STRETCH_STACK_BOTTOM_SHIFT_DOWN             # <<<<<<<<<<<<<<
 *     STRETCH_STACK_GAP_ABOVE_MIN = HB_OT_MATH_CONSTANT_STRETCH_STACK_GAP_ABOVE_MIN
 *     STRETCH_STACK_GAP_BELOW_MIN = HB_OT_MATH_CONSTANT_STRETCH_STACK_GAP_BELOW_MIN
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_STRETCH_STACK_BOTTOM_SHIFT_DOWN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_STRETCH_STACK_BOTTOM_SHIFT_DOWN, __pyx_t_3) < 0) __PYX_ERR(0, 1494, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1495
 *     STRETCH_STACK_TOP_SHIFT_UP = HB_OT_MATH_CONSTANT_STRETCH_STACK_TOP_SHIFT_UP
 *     STRETCH_STACK_BOTTOM_SHIFT_DOWN = HB_OT_MATH_CONSTANT_STRETCH_STACK_BOTTOM_SHIFT_DOWN
 *     STRETCH_STACK_GAP_ABOVE_MIN = HB_OT_MATH_CONSTANT_STRETCH_STACK_GAP_ABOVE_MIN             # <<<<<<<<<<<<<<
 *     STRETCH_STACK_GAP_BELOW_MIN = HB_OT_MATH_CONSTANT_STRETCH_STACK_GAP_BELOW_MIN
 *     FRACTION_NUMERATOR_SHIFT_UP = HB_OT_MATH_CONSTANT_FRACTION_NUMERATOR_SHIFT_UP
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_STRETCH_STACK_GAP_ABOVE_MIN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_STRETCH_STACK_GAP_ABOVE_MIN, __pyx_t_3) < 0) __PYX_ERR(0, 1495, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1496
 *     STRETCH_STACK_BOTTOM_SHIFT_DOWN = HB_OT_MATH_CONSTANT_STRETCH_STACK_BOTTOM_SHIFT_DOWN
 *     STRETCH_STACK_GAP_ABOVE_MIN = HB_OT_MATH_CONSTANT_STRETCH_STACK_GAP_ABOVE_MIN
 *     STRETCH_STACK_GAP_BELOW_MIN = HB_OT_MATH_CONSTANT_STRETCH_STACK_GAP_BELOW_MIN             # <<<<<<<<<<<<<<
 *     FRACTION_NUMERATOR_SHIFT_UP = HB_OT_MATH_CONSTANT_FRACTION_NUMERATOR_SHIFT_UP
 *     FRACTION_NUMERATOR_DISPLAY_STYLE_SHIFT_UP = HB_OT_MATH_CONSTANT_FRACTION_NUMERATOR_DISPLAY_STYLE_SHIFT_UP
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_STRETCH_STACK_GAP_BELOW_MIN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_STRETCH_STACK_GAP_BELOW_MIN, __pyx_t_3) < 0) __PYX_ERR(0, 1496, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1497
 *     STRETCH_STACK_GAP_ABOVE_MIN = HB_OT_MATH_CONSTANT_STRETCH_STACK_GAP_ABOVE_MIN
 *     STRETCH_STACK_GAP_BELOW_MIN = HB_OT_MATH_CONSTANT_STRETCH_STACK_GAP_BELOW_MIN
 *     FRACTION_NUMERATOR_SHIFT_UP = HB_OT_MATH_CONSTANT_FRACTION_NUMERATOR_SHIFT_UP             # <<<<<<<<<<<<<<
 *     FRACTION_NUMERATOR_DISPLAY_STYLE_SHIFT_UP = HB_OT_MATH_CONSTANT_FRACTION_NUMERATOR_DISPLAY_STYLE_SHIFT_UP
 *     FRACTION_DENOMINATOR_SHIFT_DOWN = HB_OT_MATH_CONSTANT_FRACTION_DENOMINATOR_SHIFT_DOWN
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_FRACTION_NUMERATOR_SHIFT_UP); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_FRACTION_NUMERATOR_SHIFT_UP, __pyx_t_3) < 0) __PYX_ERR(0, 1497, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1498
 *     STRETCH_STACK_GAP_BELOW_MIN = HB_OT_MATH_CONSTANT_STRETCH_STACK_GAP_BELOW_MIN
 *     FRACTION_NUMERATOR_SHIFT_UP = HB_OT_MATH_CONSTANT_FRACTION_NUMERATOR_SHIFT_UP
 *     FRACTION_NUMERATOR_DISPLAY_STYLE_SHIFT_UP = HB_OT_MATH_CONSTANT_FRACTION_NUMERATOR_DISPLAY_STYLE_SHIFT_UP             # <<<<<<<<<<<<<<
 *     FRACTION_DENOMINATOR_SHIFT_DOWN = HB_OT_MATH_CONSTANT_FRACTION_DENOMINATOR_SHIFT_DOWN
 *     FRACTION_DENOMINATOR_DISPLAY_STYLE_SHIFT_DOWN = HB_OT_MATH_CONSTANT_FRACTION_DENOMINATOR_DISPLAY_STYLE_SHIFT_DOWN
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_FRACTION_NUMERATOR_DISPLAY_STYLE_SHIFT_UP); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_FRACTION_NUMERATOR_DISPLAY_STYLE, __pyx_t_3) < 0) __PYX_ERR(0, 1498, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1499
 *     FRACTION_NUMERATOR_SHIFT_UP = HB_OT_MATH_CONSTANT_FRACTION_NUMERATOR_SHIFT_UP
 *     FRACTION_NUMERATOR_DISPLAY_STYLE_SHIFT_UP = HB_OT_MATH_CONSTANT_FRACTION_NUMERATOR_DISPLAY_STYLE_SHIFT_UP
 *     FRACTION_DENOMINATOR_SHIFT_DOWN = HB_OT_MATH_CONSTANT_FRACTION_DENOMINATOR_SHIFT_DOWN             # <<<<<<<<<<<<<<
 *     FRACTION_DENOMINATOR_DISPLAY_STYLE_SHIFT_DOWN = HB_OT_MATH_CONSTANT_FRACTION_DENOMINATOR_DISPLAY_STYLE_SHIFT_DOWN
 *     FRACTION_NUMERATOR_GAP_MIN = HB_OT_MATH_CONSTANT_FRACTION_NUMERATOR_GAP_MIN
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_FRACTION_DENOMINATOR_SHIFT_DOWN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_FRACTION_DENOMINATOR_SHIFT_DOWN, __pyx_t_3) < 0) __PYX_ERR(0, 1499, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1500
 *     FRACTION_NUMERATOR_DISPLAY_STYLE_SHIFT_UP = HB_OT_MATH_CONSTANT_FRACTION_NUMERATOR_DISPLAY_STYLE_SHIFT_UP
 *     FRACTION_DENOMINATOR_SHIFT_DOWN = HB_OT_MATH_CONSTANT_FRACTION_DENOMINATOR_SHIFT_DOWN
 *     FRACTION_DENOMINATOR_DISPLAY_STYLE_SHIFT_DOWN = HB_OT_MATH_CONSTANT_FRACTION_DENOMINATOR_DISPLAY_STYLE_SHIFT_DOWN             # <<<<<<<<<<<<<<
 *     FRACTION_NUMERATOR_GAP_MIN = HB_OT_MATH_CONSTANT_FRACTION_NUMERATOR_GAP_MIN
 *     FRACTION_NUM_DISPLAY_STYLE_GAP_MIN = HB_OT_MATH_CONSTANT_FRACTION_NUM_DISPLAY_STYLE_GAP_MIN
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_FRACTION_DENOMINATOR_DISPLAY_STYLE_SHIFT_DOWN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_FRACTION_DENOMINATOR_DISPLAY_STY, __pyx_t_3) < 0) __PYX_ERR(0, 1500, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1501
 *     FRACTION_DENOMINATOR_SHIFT_DOWN = HB_OT_MATH_CONSTANT_FRACTION_DENOMINATOR_SHIFT_DOWN
 *     FRACTION_DENOMINATOR_DISPLAY_STYLE_SHIFT_DOWN = HB_OT_MATH_CONSTANT_FRACTION_DENOMINATOR_DISPLAY_STYLE_SHIFT_DOWN
 *     FRACTION_NUMERATOR_GAP_MIN = HB_OT_MATH_CONSTANT_FRACTION_NUMERATOR_GAP_MIN             # <<<<<<<<<<<<<<
 *     FRACTION_NUM_DISPLAY_STYLE_GAP_MIN = HB_OT_MATH_CONSTANT_FRACTION_NUM_DISPLAY_STYLE_GAP_MIN
 *     FRACTION_RULE_THICKNESS = HB_OT_MATH_CONSTANT_FRACTION_RULE_THICKNESS
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_FRACTION_NUMERATOR_GAP_MIN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_FRACTION_NUMERATOR_GAP_MIN, __pyx_t_3) < 0) __PYX_ERR(0, 1501, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1502
 *     FRACTION_DENOMINATOR_DISPLAY_STYLE_SHIFT_DOWN = HB_OT_MATH_CONSTANT_FRACTION_DENOMINATOR_DISPLAY_STYLE_SHIFT_DOWN
 *     FRACTION_NUMERATOR_GAP_MIN = HB_OT_MATH_CONSTANT_FRACTION_NUMERATOR_GAP_MIN
 *     FRACTION_NUM_DISPLAY_STYLE_GAP_MIN = HB_OT_MATH_CONSTANT_FRACTION_NUM_DISPLAY_STYLE_GAP_MIN             # <<<<<<<<<<<<<<
 *     FRACTION_RULE_THICKNESS = HB_OT_MATH_CONSTANT_FRACTION_RULE_THICKNESS
 *     FRACTION_DENOMINATOR_GAP_MIN = HB_OT_MATH_CONSTANT_FRACTION_DENOMINATOR_GAP_MIN
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_FRACTION_NUM_DISPLAY_STYLE_GAP_MIN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_FRACTION_NUM_DISPLAY_STYLE_GAP_M, __pyx_t_3) < 0) __PYX_ERR(0, 1502, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1503
 *     FRACTION_NUMERATOR_GAP_MIN = HB_OT_MATH_CONSTANT_FRACTION_NUMERATOR_GAP_MIN
 *     FRACTION_NUM_DISPLAY_STYLE_GAP_MIN = HB_OT_MATH_CONSTANT_FRACTION_NUM_DISPLAY_STYLE_GAP_MIN
 *     FRACTION_RULE_THICKNESS = HB_OT_MATH_CONSTANT_FRACTION_RULE_THICKNESS             # <<<<<<<<<<<<<<
 *     FRACTION_DENOMINATOR_GAP_MIN = HB_OT_MATH_CONSTANT_FRACTION_DENOMINATOR_GAP_MIN
 *     FRACTION_DENOM_DISPLAY_STYLE_GAP_MIN = HB_OT_MATH_CONSTANT_FRACTION_DENOM_DISPLAY_STYLE_GAP_MIN
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_FRACTION_RULE_THICKNESS); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_FRACTION_RULE_THICKNESS, __pyx_t_3) < 0) __PYX_ERR(0, 1503, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1504
 *     FRACTION_NUM_DISPLAY_STYLE_GAP_MIN = HB_OT_MATH_CONSTANT_FRACTION_NUM_DISPLAY_STYLE_GAP_MIN
 *     FRACTION_RULE_THICKNESS = HB_OT_MATH_CONSTANT_FRACTION_RULE_THICKNESS
 *     FRACTION_DENOMINATOR_GAP_MIN = HB_OT_MATH_CONSTANT_FRACTION_DENOMINATOR_GAP_MIN             # <<<<<<<<<<<<<<
 *     FRACTION_DENOM_DISPLAY_STYLE_GAP_MIN = HB_OT_MATH_CONSTANT_FRACTION_DENOM_DISPLAY_STYLE_GAP_MIN
 *     SKEWED_FRACTION_HORIZONTAL_GAP = HB_OT_MATH_CONSTANT_SKEWED_FRACTION_HORIZONTAL_GAP
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_FRACTION_DENOMINATOR_GAP_MIN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_FRACTION_DENOMINATOR_GAP_MIN, __pyx_t_3) < 0) __PYX_ERR(0, 1504, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1505
 *     FRACTION_RULE_THICKNESS = HB_OT_MATH_CONSTANT_FRACTION_RULE_THICKNESS
 *     FRACTION_DENOMINATOR_GAP_MIN = HB_OT_MATH_CONSTANT_FRACTION_DENOMINATOR_GAP_MIN
 *     FRACTION_DENOM_DISPLAY_STYLE_GAP_MIN = HB_OT_MATH_CONSTANT_FRACTION_DENOM_DISPLAY_STYLE_GAP_MIN             # <<<<<<<<<<<<<<
 *     SKEWED_FRACTION_HORIZONTAL_GAP = HB_OT_MATH_CONSTANT_SKEWED_FRACTION_HORIZONTAL_GAP
 *     SKEWED_FRACTION_VERTICAL_GAP = HB_OT_MATH_CONSTANT_SKEWED_FRACTION_VERTICAL_GAP
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_FRACTION_DENOM_DISPLAY_STYLE_GAP_MIN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1505, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_FRACTION_DENOM_DISPLAY_STYLE_GAP, __pyx_t_3) < 0) __PYX_ERR(0, 1505, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1506
 *     FRACTION_DENOMINATOR_GAP_MIN = HB_OT_MATH_CONSTANT_FRACTION_DENOMINATOR_GAP_MIN
 *     FRACTION_DENOM_DISPLAY_STYLE_GAP_MIN = HB_OT_MATH_CONSTANT_FRACTION_DENOM_DISPLAY_STYLE_GAP_MIN
 *     SKEWED_FRACTION_HORIZONTAL_GAP = HB_OT_MATH_CONSTANT_SKEWED_FRACTION_HORIZONTAL_GAP             # <<<<<<<<<<<<<<
 *     SKEWED_FRACTION_VERTICAL_GAP = HB_OT_MATH_CONSTANT_SKEWED_FRACTION_VERTICAL_GAP
 *     OVERBAR_VERTICAL_GAP = HB_OT_MATH_CONSTANT_OVERBAR_VERTICAL_GAP
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_SKEWED_FRACTION_HORIZONTAL_GAP); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_SKEWED_FRACTION_HORIZONTAL_GAP, __pyx_t_3) < 0) __PYX_ERR(0, 1506, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1507
 *     FRACTION_DENOM_DISPLAY_STYLE_GAP_MIN = HB_OT_MATH_CONSTANT_FRACTION_DENOM_DISPLAY_STYLE_GAP_MIN
 *     SKEWED_FRACTION_HORIZONTAL_GAP = HB_OT_MATH_CONSTANT_SKEWED_FRACTION_HORIZONTAL_GAP
 *     SKEWED_FRACTION_VERTICAL_GAP = HB_OT_MATH_CONSTANT_SKEWED_FRACTION_VERTICAL_GAP             # <<<<<<<<<<<<<<
 *     OVERBAR_VERTICAL_GAP = HB_OT_MATH_CONSTANT_OVERBAR_VERTICAL_GAP
 *     OVERBAR_RULE_THICKNESS = HB_OT_MATH_CONSTANT_OVERBAR_RULE_THICKNESS
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_SKEWED_FRACTION_VERTICAL_GAP); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1507, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_SKEWED_FRACTION_VERTICAL_GAP, __pyx_t_3) < 0) __PYX_ERR(0, 1507, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1508
 *     SKEWED_FRACTION_HORIZONTAL_GAP = HB_OT_MATH_CONSTANT_SKEWED_FRACTION_HORIZONTAL_GAP
 *     SKEWED_FRACTION_VERTICAL_GAP = HB_OT_MATH_CONSTANT_SKEWED_FRACTION_VERTICAL_GAP
 *     OVERBAR_VERTICAL_GAP = HB_OT_MATH_CONSTANT_OVERBAR_VERTICAL_GAP             # <<<<<<<<<<<<<<
 *     OVERBAR_RULE_THICKNESS = HB_OT_MATH_CONSTANT_OVERBAR_RULE_THICKNESS
 *     OVERBAR_EXTRA_ASCENDER = HB_OT_MATH_CONSTANT_OVERBAR_EXTRA_ASCENDER
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_OVERBAR_VERTICAL_GAP); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_OVERBAR_VERTICAL_GAP, __pyx_t_3) < 0) __PYX_ERR(0, 1508, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1509
 *     SKEWED_FRACTION_VERTICAL_GAP = HB_OT_MATH_CONSTANT_SKEWED_FRACTION_VERTICAL_GAP
 *     OVERBAR_VERTICAL_GAP = HB_OT_MATH_CONSTANT_OVERBAR_VERTICAL_GAP
 *     OVERBAR_RULE_THICKNESS = HB_OT_MATH_CONSTANT_OVERBAR_RULE_THICKNESS             # <<<<<<<<<<<<<<
 *     OVERBAR_EXTRA_ASCENDER = HB_OT_MATH_CONSTANT_OVERBAR_EXTRA_ASCENDER
 *     UNDERBAR_VERTICAL_GAP = HB_OT_MATH_CONSTANT_UNDERBAR_VERTICAL_GAP
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_OVERBAR_RULE_THICKNESS); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_OVERBAR_RULE_THICKNESS, __pyx_t_3) < 0) __PYX_ERR(0, 1509, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1510
 *     OVERBAR_VERTICAL_GAP = HB_OT_MATH_CONSTANT_OVERBAR_VERTICAL_GAP
 *     OVERBAR_RULE_THICKNESS = HB_OT_MATH_CONSTANT_OVERBAR_RULE_THICKNESS
 *     OVERBAR_EXTRA_ASCENDER = HB_OT_MATH_CONSTANT_OVERBAR_EXTRA_ASCENDER             # <<<<<<<<<<<<<<
 *     UNDERBAR_VERTICAL_GAP = HB_OT_MATH_CONSTANT_UNDERBAR_VERTICAL_GAP
 *     UNDERBAR_RULE_THICKNESS = HB_OT_MATH_CONSTANT_UNDERBAR_RULE_THICKNESS
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_OVERBAR_EXTRA_ASCENDER); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1510, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_OVERBAR_EXTRA_ASCENDER, __pyx_t_3) < 0) __PYX_ERR(0, 1510, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1511
 *     OVERBAR_RULE_THICKNESS = HB_OT_MATH_CONSTANT_OVERBAR_RULE_THICKNESS
 *     OVERBAR_EXTRA_ASCENDER = HB_OT_MATH_CONSTANT_OVERBAR_EXTRA_ASCENDER
 *     UNDERBAR_VERTICAL_GAP = HB_OT_MATH_CONSTANT_UNDERBAR_VERTICAL_GAP             # <<<<<<<<<<<<<<
 *     UNDERBAR_RULE_THICKNESS = HB_OT_MATH_CONSTANT_UNDERBAR_RULE_THICKNESS
 *     UNDERBAR_EXTRA_DESCENDER = HB_OT_MATH_CONSTANT_UNDERBAR_EXTRA_DESCENDER
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_UNDERBAR_VERTICAL_GAP); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_UNDERBAR_VERTICAL_GAP, __pyx_t_3) < 0) __PYX_ERR(0, 1511, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1512
 *     OVERBAR_EXTRA_ASCENDER = HB_OT_MATH_CONSTANT_OVERBAR_EXTRA_ASCENDER
 *     UNDERBAR_VERTICAL_GAP = HB_OT_MATH_CONSTANT_UNDERBAR_VERTICAL_GAP
 *     UNDERBAR_RULE_THICKNESS = HB_OT_MATH_CONSTANT_UNDERBAR_RULE_THICKNESS             # <<<<<<<<<<<<<<
 *     UNDERBAR_EXTRA_DESCENDER = HB_OT_MATH_CONSTANT_UNDERBAR_EXTRA_DESCENDER
 *     RADICAL_VERTICAL_GAP = HB_OT_MATH_CONSTANT_RADICAL_VERTICAL_GAP
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_UNDERBAR_RULE_THICKNESS); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1512, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_UNDERBAR_RULE_THICKNESS, __pyx_t_3) < 0) __PYX_ERR(0, 1512, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1513
 *     UNDERBAR_VERTICAL_GAP = HB_OT_MATH_CONSTANT_UNDERBAR_VERTICAL_GAP
 *     UNDERBAR_RULE_THICKNESS = HB_OT_MATH_CONSTANT_UNDERBAR_RULE_THICKNESS
 *     UNDERBAR_EXTRA_DESCENDER = HB_OT_MATH_CONSTANT_UNDERBAR_EXTRA_DESCENDER             # <<<<<<<<<<<<<<
 *     RADICAL_VERTICAL_GAP = HB_OT_MATH_CONSTANT_RADICAL_VERTICAL_GAP
 *     RADICAL_DISPLAY_STYLE_VERTICAL_GAP = HB_OT_MATH_CONSTANT_RADICAL_DISPLAY_STYLE_VERTICAL_GAP
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_UNDERBAR_EXTRA_DESCENDER); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1513, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_UNDERBAR_EXTRA_DESCENDER, __pyx_t_3) < 0) __PYX_ERR(0, 1513, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1514
 *     UNDERBAR_RULE_THICKNESS = HB_OT_MATH_CONSTANT_UNDERBAR_RULE_THICKNESS
 *     UNDERBAR_EXTRA_DESCENDER = HB_OT_MATH_CONSTANT_UNDERBAR_EXTRA_DESCENDER
 *     RADICAL_VERTICAL_GAP = HB_OT_MATH_CONSTANT_RADICAL_VERTICAL_GAP             # <<<<<<<<<<<<<<
 *     RADICAL_DISPLAY_STYLE_VERTICAL_GAP = HB_OT_MATH_CONSTANT_RADICAL_DISPLAY_STYLE_VERTICAL_GAP
 *     RADICAL_RULE_THICKNESS = HB_OT_MATH_CONSTANT_RADICAL_RULE_THICKNESS
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_RADICAL_VERTICAL_GAP); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1514, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_RADICAL_VERTICAL_GAP, __pyx_t_3) < 0) __PYX_ERR(0, 1514, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1515
 *     UNDERBAR_EXTRA_DESCENDER = HB_OT_MATH_CONSTANT_UNDERBAR_EXTRA_DESCENDER
 *     RADICAL_VERTICAL_GAP = HB_OT_MATH_CONSTANT_RADICAL_VERTICAL_GAP
 *     RADICAL_DISPLAY_STYLE_VERTICAL_GAP = HB_OT_MATH_CONSTANT_RADICAL_DISPLAY_STYLE_VERTICAL_GAP             # <<<<<<<<<<<<<<
 *     RADICAL_RULE_THICKNESS = HB_OT_MATH_CONSTANT_RADICAL_RULE_THICKNESS
 *     RADICAL_EXTRA_ASCENDER = HB_OT_MATH_CONSTANT_RADICAL_EXTRA_ASCENDER
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_RADICAL_DISPLAY_STYLE_VERTICAL_GAP); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1515, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_RADICAL_DISPLAY_STYLE_VERTICAL_G, __pyx_t_3) < 0) __PYX_ERR(0, 1515, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1516
 *     RADICAL_VERTICAL_GAP = HB_OT_MATH_CONSTANT_RADICAL_VERTICAL_GAP
 *     RADICAL_DISPLAY_STYLE_VERTICAL_GAP = HB_OT_MATH_CONSTANT_RADICAL_DISPLAY_STYLE_VERTICAL_GAP
 *     RADICAL_RULE_THICKNESS = HB_OT_MATH_CONSTANT_RADICAL_RULE_THICKNESS             # <<<<<<<<<<<<<<
 *     RADICAL_EXTRA_ASCENDER = HB_OT_MATH_CONSTANT_RADICAL_EXTRA_ASCENDER
 *     RADICAL_KERN_BEFORE_DEGREE = HB_OT_MATH_CONSTANT_RADICAL_KERN_BEFORE_DEGREE
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_RADICAL_RULE_THICKNESS); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_RADICAL_RULE_THICKNESS, __pyx_t_3) < 0) __PYX_ERR(0, 1516, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1517
 *     RADICAL_DISPLAY_STYLE_VERTICAL_GAP = HB_OT_MATH_CONSTANT_RADICAL_DISPLAY_STYLE_VERTICAL_GAP
 *     RADICAL_RULE_THICKNESS = HB_OT_MATH_CONSTANT_RADICAL_RULE_THICKNESS
 *     RADICAL_EXTRA_ASCENDER = HB_OT_MATH_CONSTANT_RADICAL_EXTRA_ASCENDER             # <<<<<<<<<<<<<<
 *     RADICAL_KERN_BEFORE_DEGREE = HB_OT_MATH_CONSTANT_RADICAL_KERN_BEFORE_DEGREE
 *     RADICAL_KERN_AFTER_DEGREE = HB_OT_MATH_CONSTANT_RADICAL_KERN_AFTER_DEGREE
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_RADICAL_EXTRA_ASCENDER); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_RADICAL_EXTRA_ASCENDER, __pyx_t_3) < 0) __PYX_ERR(0, 1517, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1518
 *     RADICAL_RULE_THICKNESS = HB_OT_MATH_CONSTANT_RADICAL_RULE_THICKNESS
 *     RADICAL_EXTRA_ASCENDER = HB_OT_MATH_CONSTANT_RADICAL_EXTRA_ASCENDER
 *     RADICAL_KERN_BEFORE_DEGREE = HB_OT_MATH_CONSTANT_RADICAL_KERN_BEFORE_DEGREE             # <<<<<<<<<<<<<<
 *     RADICAL_KERN_AFTER_DEGREE = HB_OT_MATH_CONSTANT_RADICAL_KERN_AFTER_DEGREE
 *     RADICAL_DEGREE_BOTTOM_RAISE_PERCENT = HB_OT_MATH_CONSTANT_RADICAL_DEGREE_BOTTOM_RAISE_PERCENT
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_RADICAL_KERN_BEFORE_DEGREE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_RADICAL_KERN_BEFORE_DEGREE, __pyx_t_3) < 0) __PYX_ERR(0, 1518, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1519
 *     RADICAL_EXTRA_ASCENDER = HB_OT_MATH_CONSTANT_RADICAL_EXTRA_ASCENDER
 *     RADICAL_KERN_BEFORE_DEGREE = HB_OT_MATH_CONSTANT_RADICAL_KERN_BEFORE_DEGREE
 *     RADICAL_KERN_AFTER_DEGREE = HB_OT_MATH_CONSTANT_RADICAL_KERN_AFTER_DEGREE             # <<<<<<<<<<<<<<
 *     RADICAL_DEGREE_BOTTOM_RAISE_PERCENT = HB_OT_MATH_CONSTANT_RADICAL_DEGREE_BOTTOM_RAISE_PERCENT
 * 
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_RADICAL_KERN_AFTER_DEGREE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1519, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_RADICAL_KERN_AFTER_DEGREE, __pyx_t_3) < 0) __PYX_ERR(0, 1519, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1520
 *     RADICAL_KERN_BEFORE_DEGREE = HB_OT_MATH_CONSTANT_RADICAL_KERN_BEFORE_DEGREE
 *     RADICAL_KERN_AFTER_DEGREE = HB_OT_MATH_CONSTANT_RADICAL_KERN_AFTER_DEGREE
 *     RADICAL_DEGREE_BOTTOM_RAISE_PERCENT = HB_OT_MATH_CONSTANT_RADICAL_DEGREE_BOTTOM_RAISE_PERCENT             # <<<<<<<<<<<<<<
 * 
 * def ot_math_get_constant(font: Font, constant: OTMathConstant) -> int:
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_ot_math_constant_t(HB_OT_MATH_CONSTANT_RADICAL_DEGREE_BOTTOM_RAISE_PERCENT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_5, __pyx_n_s_RADICAL_DEGREE_BOTTOM_RAISE_PERC, __pyx_t_3) < 0) __PYX_ERR(0, 1520, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1464
 *     return hb_ot_math_has_data(face._hb_face)
 * 
 * class OTMathConstant(IntEnum):             # <<<<<<<<<<<<<<
 *     SCRIPT_PERCENT_SCALE_DOWN = HB_OT_MATH_CONSTANT_SCRIPT_PERCENT_SCALE_DOWN
 *     SCRIPT_SCRIPT_PERCENT_SCALE_DOWN = HB_OT_MATH_CONSTANT_SCRIPT_SCRIPT_PERCENT_SCALE_DOWN
 */
  __pyx_t_3 = __Pyx_Py3ClassCreate(__pyx_t_4, __pyx_n_s_OTMathConstant, __pyx_t_2, __pyx_t_5, NULL, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OTMathConstant, __pyx_t_3) < 0) __PYX_ERR(0, 1464, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1522
 *     RADICAL_DEGREE_BOTTOM_RAISE_PERCENT = HB_OT_MATH_CONSTANT_RADICAL_DEGREE_BOTTOM_RAISE_PERCENT
 * 
 * def ot_math_get_constant(font: Font, constant: OTMathConstant) -> int:             # <<<<<<<<<<<<<<
 *     if constant >= len(OTMathConstant):
 *         raise ValueError("invalid constant")
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1522, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_font, __pyx_n_s_Font) < 0) __PYX_ERR(0, 1522, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_constant, __pyx_n_s_OTMathConstant) < 0) __PYX_ERR(0, 1522, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_int) < 0) __PYX_ERR(0, 1522, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_57ot_math_get_constant, 0, __pyx_n_s_ot_math_get_constant, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__164)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1522, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_math_get_constant, __pyx_t_4) < 0) __PYX_ERR(0, 1522, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1527
 *     return hb_ot_math_get_constant(font._hb_font, constant)
 * 
 * def ot_math_get_glyph_italics_correction(font: Font, glyph: int) -> int:             # <<<<<<<<<<<<<<
 *     return hb_ot_math_get_glyph_italics_correction(font._hb_font, glyph)
 * 
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_font, __pyx_n_s_Font) < 0) __PYX_ERR(0, 1527, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_glyph, __pyx_n_s_int) < 0) __PYX_ERR(0, 1527, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_return, __pyx_n_s_int) < 0) __PYX_ERR(0, 1527, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_59ot_math_get_glyph_italics_correction, 0, __pyx_n_s_ot_math_get_glyph_italics_correc, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__166)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_math_get_glyph_italics_correc, __pyx_t_2) < 0) __PYX_ERR(0, 1527, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1530
 *     return hb_ot_math_get_glyph_italics_correction(font._hb_font, glyph)
 * 
 * def ot_math_get_glyph_top_accent_attachment(font: Font, glyph: int) -> int:             # <<<<<<<<<<<<<<
 *     return hb_ot_math_get_glyph_top_accent_attachment(font._hb_font, glyph)
 * 
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_font, __pyx_n_s_Font) < 0) __PYX_ERR(0, 1530, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_glyph, __pyx_n_s_int) < 0) __PYX_ERR(0, 1530, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_int) < 0) __PYX_ERR(0, 1530, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_61ot_math_get_glyph_top_accent_attachment, 0, __pyx_n_s_ot_math_get_glyph_top_accent_att, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__167)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_math_get_glyph_top_accent_att, __pyx_t_4) < 0) __PYX_ERR(0, 1530, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1533
 *     return hb_ot_math_get_glyph_top_accent_attachment(font._hb_font, glyph)
 * 
 * def ot_math_is_glyph_extended_shape(face: Face, glyph: int) -> bool:             # <<<<<<<<<<<<<<
 *     return hb_ot_math_is_glyph_extended_shape(face._hb_face, glyph)
 * 
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_face, __pyx_n_s_Face) < 0) __PYX_ERR(0, 1533, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_glyph, __pyx_n_s_int) < 0) __PYX_ERR(0, 1533, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 1533, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_63ot_math_is_glyph_extended_shape, 0, __pyx_n_s_ot_math_is_glyph_extended_shape, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__168)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_math_is_glyph_extended_shape, __pyx_t_2) < 0) __PYX_ERR(0, 1533, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1536
 *     return hb_ot_math_is_glyph_extended_shape(face._hb_face, glyph)
 * 
 * def ot_math_get_min_connector_overlap(font: Font, direction: str) -> int:             # <<<<<<<<<<<<<<
 *     cdef bytes packed = direction.encode()
 *     cdef char* cstr = packed
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_font, __pyx_n_s_Font) < 0) __PYX_ERR(0, 1536, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_direction, __pyx_n_s_str) < 0) __PYX_ERR(0, 1536, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_int) < 0) __PYX_ERR(0, 1536, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_65ot_math_get_min_connector_overlap, 0, __pyx_n_s_ot_math_get_min_connector_overla, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__170)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1536, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_math_get_min_connector_overla, __pyx_t_4) < 0) __PYX_ERR(0, 1536, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1542
 *     return hb_ot_math_get_min_connector_overlap(font._hb_font, hb_direction)
 * 
 * class OTMathKern(IntEnum):             # <<<<<<<<<<<<<<
 *     TOP_RIGHT = HB_OT_MATH_KERN_TOP_RIGHT
 *     TOP_LEFT = HB_OT_MATH_KERN_TOP_LEFT
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_IntEnum); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_4)) __PYX_ERR(0, 1542, __pyx_L1_error);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PEP560_update_bases(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_4, __pyx_n_s_OTMathKern, __pyx_n_s_OTMathKern, (PyObject *) NULL, __pyx_n_s_uharfbuzz__harfbuzz, (PyObject *) NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__pyx_t_4 != __pyx_t_2) {
    if (unlikely((PyDict_SetItemString(__pyx_t_3, "__orig_bases__", __pyx_t_2) < 0))) __PYX_ERR(0, 1542, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1543
 * 
 * class OTMathKern(IntEnum):
 *     TOP_RIGHT = HB_OT_MATH_KERN_TOP_RIGHT             # <<<<<<<<<<<<<<
 *     TOP_LEFT = HB_OT_MATH_KERN_TOP_LEFT
 *     BOTTOM_RIGHT = HB_OT_MATH_KERN_BOTTOM_RIGHT
 */
  __pyx_t_2 = __Pyx_PyInt_From_hb_ot_math_kern_t(HB_OT_MATH_KERN_TOP_RIGHT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_TOP_RIGHT, __pyx_t_2) < 0) __PYX_ERR(0, 1543, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1544
 * class OTMathKern(IntEnum):
 *     TOP_RIGHT = HB_OT_MATH_KERN_TOP_RIGHT
 *     TOP_LEFT = HB_OT_MATH_KERN_TOP_LEFT             # <<<<<<<<<<<<<<
 *     BOTTOM_RIGHT = HB_OT_MATH_KERN_BOTTOM_RIGHT
 *     BOTTOM_LEFT = HB_OT_MATH_KERN_BOTTOM_LEFT
 */
  __pyx_t_2 = __Pyx_PyInt_From_hb_ot_math_kern_t(HB_OT_MATH_KERN_TOP_LEFT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_TOP_LEFT, __pyx_t_2) < 0) __PYX_ERR(0, 1544, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1545
 *     TOP_RIGHT = HB_OT_MATH_KERN_TOP_RIGHT
 *     TOP_LEFT = HB_OT_MATH_KERN_TOP_LEFT
 *     BOTTOM_RIGHT = HB_OT_MATH_KERN_BOTTOM_RIGHT             # <<<<<<<<<<<<<<
 *     BOTTOM_LEFT = HB_OT_MATH_KERN_BOTTOM_LEFT
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_hb_ot_math_kern_t(HB_OT_MATH_KERN_BOTTOM_RIGHT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_BOTTOM_RIGHT, __pyx_t_2) < 0) __PYX_ERR(0, 1545, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1546
 *     TOP_LEFT = HB_OT_MATH_KERN_TOP_LEFT
 *     BOTTOM_RIGHT = HB_OT_MATH_KERN_BOTTOM_RIGHT
 *     BOTTOM_LEFT = HB_OT_MATH_KERN_BOTTOM_LEFT             # <<<<<<<<<<<<<<
 * 
 * def ot_math_get_glyph_kerning(font: Font,
 */
  __pyx_t_2 = __Pyx_PyInt_From_hb_ot_math_kern_t(HB_OT_MATH_KERN_BOTTOM_LEFT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1546, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_BOTTOM_LEFT, __pyx_t_2) < 0) __PYX_ERR(0, 1546, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1542
 *     return hb_ot_math_get_min_connector_overlap(font._hb_font, hb_direction)
 * 
 * class OTMathKern(IntEnum):             # <<<<<<<<<<<<<<
 *     TOP_RIGHT = HB_OT_MATH_KERN_TOP_RIGHT
 *     TOP_LEFT = HB_OT_MATH_KERN_TOP_LEFT
 */
  __pyx_t_2 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_OTMathKern, __pyx_t_4, __pyx_t_3, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1542, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OTMathKern, __pyx_t_2) < 0) __PYX_ERR(0, 1542, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1548
 *     BOTTOM_LEFT = HB_OT_MATH_KERN_BOTTOM_LEFT
 * 
 * def ot_math_get_glyph_kerning(font: Font,             # <<<<<<<<<<<<<<
 *                               glyph: int,
 *                               kern: OTMathKern,
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_font, __pyx_n_s_Font) < 0) __PYX_ERR(0, 1548, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_glyph, __pyx_n_s_int) < 0) __PYX_ERR(0, 1548, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_kern, __pyx_n_s_OTMathKern) < 0) __PYX_ERR(0, 1548, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_return, __pyx_n_s_int) < 0) __PYX_ERR(0, 1548, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_67ot_math_get_glyph_kerning, 0, __pyx_n_s_ot_math_get_glyph_kerning, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__172)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_math_get_glyph_kerning, __pyx_t_5) < 0) __PYX_ERR(0, 1548, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1556
 *     return hb_ot_math_get_glyph_kerning(font._hb_font, glyph, kern, correction_height)
 * 
 * OTMathKernEntry = namedtuple("OTMathKernEntry", ["max_correction_height", "kern_value"])             # <<<<<<<<<<<<<<
 * 
 * def ot_math_get_glyph_kernings(font: Font,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_namedtuple); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyList_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_n_u_max_correction_height);
  __Pyx_GIVEREF(__pyx_n_u_max_correction_height);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 0, __pyx_n_u_max_correction_height)) __PYX_ERR(0, 1556, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_kern_value);
  __Pyx_GIVEREF(__pyx_n_u_kern_value);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 1, __pyx_n_u_kern_value)) __PYX_ERR(0, 1556, __pyx_L1_error);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_u_OTMathKernEntry);
  __Pyx_GIVEREF(__pyx_n_u_OTMathKernEntry);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_n_u_OTMathKernEntry)) __PYX_ERR(0, 1556, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4)) __PYX_ERR(0, 1556, __pyx_L1_error);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OTMathKernEntry, __pyx_t_4) < 0) __PYX_ERR(0, 1556, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1558
 * OTMathKernEntry = namedtuple("OTMathKernEntry", ["max_correction_height", "kern_value"])
 * 
 * def ot_math_get_glyph_kernings(font: Font,             # <<<<<<<<<<<<<<
 *                                glyph: int,
 *                                kern: OTMathKern) -> List[OTMathKernEntry]:
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_font, __pyx_n_s_Font) < 0) __PYX_ERR(0, 1558, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_glyph, __pyx_n_s_int) < 0) __PYX_ERR(0, 1558, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_kern, __pyx_n_s_OTMathKern) < 0) __PYX_ERR(0, 1558, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_return, __pyx_kp_s_List_OTMathKernEntry) < 0) __PYX_ERR(0, 1558, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_69ot_math_get_glyph_kernings, 0, __pyx_n_s_ot_math_get_glyph_kernings, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__174)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_math_get_glyph_kernings, __pyx_t_3) < 0) __PYX_ERR(0, 1558, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1576
 *     return kerns
 * 
 * OTMathGlyphVariant = namedtuple("OTMathGlyphVariant", ["glyph", "advance"])             # <<<<<<<<<<<<<<
 * 
 * def ot_math_get_glyph_variants(font: Font, glyph: int, direction: str) -> List[OTMathGlyphVariant]:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_namedtuple); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyList_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_n_u_glyph);
  __Pyx_GIVEREF(__pyx_n_u_glyph);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 0, __pyx_n_u_glyph)) __PYX_ERR(0, 1576, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_advance);
  __Pyx_GIVEREF(__pyx_n_u_advance);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 1, __pyx_n_u_advance)) __PYX_ERR(0, 1576, __pyx_L1_error);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_n_u_OTMathGlyphVariant);
  __Pyx_GIVEREF(__pyx_n_u_OTMathGlyphVariant);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_n_u_OTMathGlyphVariant)) __PYX_ERR(0, 1576, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_4)) __PYX_ERR(0, 1576, __pyx_L1_error);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OTMathGlyphVariant, __pyx_t_4) < 0) __PYX_ERR(0, 1576, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1578
 * OTMathGlyphVariant = namedtuple("OTMathGlyphVariant", ["glyph", "advance"])
 * 
 * def ot_math_get_glyph_variants(font: Font, glyph: int, direction: str) -> List[OTMathGlyphVariant]:             # <<<<<<<<<<<<<<
 *     cdef bytes packed = direction.encode()
 *     cdef char* cstr = packed
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1578, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_font, __pyx_n_s_Font) < 0) __PYX_ERR(0, 1578, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_glyph, __pyx_n_s_int) < 0) __PYX_ERR(0, 1578, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_direction, __pyx_n_s_str) < 0) __PYX_ERR(0, 1578, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_return, __pyx_kp_s_List_OTMathGlyphVariant) < 0) __PYX_ERR(0, 1578, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_71ot_math_get_glyph_variants, 0, __pyx_n_s_ot_math_get_glyph_variants, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__176)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1578, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_math_get_glyph_variants, __pyx_t_5) < 0) __PYX_ERR(0, 1578, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1595
 *     return variants
 * 
 * class OTMathGlyphPartFlags(IntFlag):             # <<<<<<<<<<<<<<
 *     EXTENDER = HB_OT_MATH_GLYPH_PART_FLAG_EXTENDER
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_IntFlag); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5)) __PYX_ERR(0, 1595, __pyx_L1_error);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PEP560_update_bases(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_CalculateMetaclass(NULL, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_Py3MetaclassPrepare(__pyx_t_3, __pyx_t_5, __pyx_n_s_OTMathGlyphPartFlags, __pyx_n_s_OTMathGlyphPartFlags, (PyObject *) NULL, __pyx_n_s_uharfbuzz__harfbuzz, (PyObject *) NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__pyx_t_5 != __pyx_t_4) {
    if (unlikely((PyDict_SetItemString(__pyx_t_2, "__orig_bases__", __pyx_t_4) < 0))) __PYX_ERR(0, 1595, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1596
 * 
 * class OTMathGlyphPartFlags(IntFlag):
 *     EXTENDER = HB_OT_MATH_GLYPH_PART_FLAG_EXTENDER             # <<<<<<<<<<<<<<
 * 
 * OTMathGlyphPart = namedtuple(
 */
  __pyx_t_4 = __Pyx_PyInt_From_hb_ot_math_glyph_part_flags_t(HB_OT_MATH_GLYPH_PART_FLAG_EXTENDER); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_EXTENDER, __pyx_t_4) < 0) __PYX_ERR(0, 1596, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1595
 *     return variants
 * 
 * class OTMathGlyphPartFlags(IntFlag):             # <<<<<<<<<<<<<<
 *     EXTENDER = HB_OT_MATH_GLYPH_PART_FLAG_EXTENDER
 * 
 */
  __pyx_t_4 = __Pyx_Py3ClassCreate(__pyx_t_3, __pyx_n_s_OTMathGlyphPartFlags, __pyx_t_5, __pyx_t_2, NULL, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OTMathGlyphPartFlags, __pyx_t_4) < 0) __PYX_ERR(0, 1595, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1598
 *     EXTENDER = HB_OT_MATH_GLYPH_PART_FLAG_EXTENDER
 * 
 * OTMathGlyphPart = namedtuple(             # <<<<<<<<<<<<<<
 *     "OTMathGlyphPart",
 *     ["glyph", "start_connector_length", "end_connector_length", "full_advance", "flags"]
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_namedtuple); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "uharfbuzz/_harfbuzz.pyx":1600
 * OTMathGlyphPart = namedtuple(
 *     "OTMathGlyphPart",
 *     ["glyph", "start_connector_length", "end_connector_length", "full_advance", "flags"]             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_t_3 = PyList_New(5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_u_glyph);
  __Pyx_GIVEREF(__pyx_n_u_glyph);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_u_glyph)) __PYX_ERR(0, 1600, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_start_connector_length);
  __Pyx_GIVEREF(__pyx_n_u_start_connector_length);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 1, __pyx_n_u_start_connector_length)) __PYX_ERR(0, 1600, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_end_connector_length);
  __Pyx_GIVEREF(__pyx_n_u_end_connector_length);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 2, __pyx_n_u_end_connector_length)) __PYX_ERR(0, 1600, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_full_advance);
  __Pyx_GIVEREF(__pyx_n_u_full_advance);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 3, __pyx_n_u_full_advance)) __PYX_ERR(0, 1600, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_flags);
  __Pyx_GIVEREF(__pyx_n_u_flags);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 4, __pyx_n_u_flags)) __PYX_ERR(0, 1600, __pyx_L1_error);

  /* "uharfbuzz/_harfbuzz.pyx":1598
 *     EXTENDER = HB_OT_MATH_GLYPH_PART_FLAG_EXTENDER
 * 
 * OTMathGlyphPart = namedtuple(             # <<<<<<<<<<<<<<
 *     "OTMathGlyphPart",
 *     ["glyph", "start_connector_length", "end_connector_length", "full_advance", "flags"]
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_u_OTMathGlyphPart);
  __Pyx_GIVEREF(__pyx_n_u_OTMathGlyphPart);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_n_u_OTMathGlyphPart)) __PYX_ERR(0, 1598, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3)) __PYX_ERR(0, 1598, __pyx_L1_error);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OTMathGlyphPart, __pyx_t_3) < 0) __PYX_ERR(0, 1598, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1603
 * )
 * 
 * def ot_math_get_glyph_assembly(font: Font,             # <<<<<<<<<<<<<<
 *                                glyph: int,
 *                                direction: str) -> Tuple[List[OTMathGlyphPart], int]:
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_font, __pyx_n_s_Font) < 0) __PYX_ERR(0, 1603, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_glyph, __pyx_n_s_int) < 0) __PYX_ERR(0, 1603, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_direction, __pyx_n_s_str) < 0) __PYX_ERR(0, 1603, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_kp_s_Tuple_List_OTMathGlyphPart_int) < 0) __PYX_ERR(0, 1603, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_73ot_math_get_glyph_assembly, 0, __pyx_n_s_ot_math_get_glyph_assembly, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__178)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_math_get_glyph_assembly, __pyx_t_2) < 0) __PYX_ERR(0, 1603, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1627
 *     return assembly, italics_correction
 * 
 * def ot_font_set_funcs(Font font):             # <<<<<<<<<<<<<<
 *     hb_ot_font_set_funcs(font._hb_font)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_75ot_font_set_funcs, 0, __pyx_n_s_ot_font_set_funcs, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__180)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ot_font_set_funcs, __pyx_t_2) < 0) __PYX_ERR(0, 1627, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1631
 * 
 * 
 * class PaintCompositeMode(IntEnum):             # <<<<<<<<<<<<<<
 *     CLEAR = HB_PAINT_COMPOSITE_MODE_CLEAR
 *     SRC = HB_PAINT_COMPOSITE_MODE_SRC
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_IntEnum); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2)) __PYX_ERR(0, 1631, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PEP560_update_bases(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_Py3MetaclassPrepare(__pyx_t_5, __pyx_t_2, __pyx_n_s_PaintCompositeMode, __pyx_n_s_PaintCompositeMode, (PyObject *) NULL, __pyx_n_s_uharfbuzz__harfbuzz, (PyObject *) NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__pyx_t_2 != __pyx_t_3) {
    if (unlikely((PyDict_SetItemString(__pyx_t_4, "__orig_bases__", __pyx_t_3) < 0))) __PYX_ERR(0, 1631, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1632
 * 
 * class PaintCompositeMode(IntEnum):
 *     CLEAR = HB_PAINT_COMPOSITE_MODE_CLEAR             # <<<<<<<<<<<<<<
 *     SRC = HB_PAINT_COMPOSITE_MODE_SRC
 *     DEST = HB_PAINT_COMPOSITE_MODE_DEST
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_paint_composite_mode_t(HB_PAINT_COMPOSITE_MODE_CLEAR); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_CLEAR, __pyx_t_3) < 0) __PYX_ERR(0, 1632, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1633
 * class PaintCompositeMode(IntEnum):
 *     CLEAR = HB_PAINT_COMPOSITE_MODE_CLEAR
 *     SRC = HB_PAINT_COMPOSITE_MODE_SRC             # <<<<<<<<<<<<<<
 *     DEST = HB_PAINT_COMPOSITE_MODE_DEST
 *     SRC_OVER = HB_PAINT_COMPOSITE_MODE_SRC_OVER
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_paint_composite_mode_t(HB_PAINT_COMPOSITE_MODE_SRC); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_SRC, __pyx_t_3) < 0) __PYX_ERR(0, 1633, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1634
 *     CLEAR = HB_PAINT_COMPOSITE_MODE_CLEAR
 *     SRC = HB_PAINT_COMPOSITE_MODE_SRC
 *     DEST = HB_PAINT_COMPOSITE_MODE_DEST             # <<<<<<<<<<<<<<
 *     SRC_OVER = HB_PAINT_COMPOSITE_MODE_SRC_OVER
 *     DEST_OVER = HB_PAINT_COMPOSITE_MODE_DEST_OVER
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_paint_composite_mode_t(HB_PAINT_COMPOSITE_MODE_DEST); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1634, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_DEST, __pyx_t_3) < 0) __PYX_ERR(0, 1634, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1635
 *     SRC = HB_PAINT_COMPOSITE_MODE_SRC
 *     DEST = HB_PAINT_COMPOSITE_MODE_DEST
 *     SRC_OVER = HB_PAINT_COMPOSITE_MODE_SRC_OVER             # <<<<<<<<<<<<<<
 *     DEST_OVER = HB_PAINT_COMPOSITE_MODE_DEST_OVER
 *     SRC_IN = HB_PAINT_COMPOSITE_MODE_SRC_IN
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_paint_composite_mode_t(HB_PAINT_COMPOSITE_MODE_SRC_OVER); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1635, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_SRC_OVER, __pyx_t_3) < 0) __PYX_ERR(0, 1635, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1636
 *     DEST = HB_PAINT_COMPOSITE_MODE_DEST
 *     SRC_OVER = HB_PAINT_COMPOSITE_MODE_SRC_OVER
 *     DEST_OVER = HB_PAINT_COMPOSITE_MODE_DEST_OVER             # <<<<<<<<<<<<<<
 *     SRC_IN = HB_PAINT_COMPOSITE_MODE_SRC_IN
 *     DEST_IN = HB_PAINT_COMPOSITE_MODE_DEST_IN
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_paint_composite_mode_t(HB_PAINT_COMPOSITE_MODE_DEST_OVER); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_DEST_OVER, __pyx_t_3) < 0) __PYX_ERR(0, 1636, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1637
 *     SRC_OVER = HB_PAINT_COMPOSITE_MODE_SRC_OVER
 *     DEST_OVER = HB_PAINT_COMPOSITE_MODE_DEST_OVER
 *     SRC_IN = HB_PAINT_COMPOSITE_MODE_SRC_IN             # <<<<<<<<<<<<<<
 *     DEST_IN = HB_PAINT_COMPOSITE_MODE_DEST_IN
 *     SRC_OUT = HB_PAINT_COMPOSITE_MODE_SRC_OUT
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_paint_composite_mode_t(HB_PAINT_COMPOSITE_MODE_SRC_IN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_SRC_IN, __pyx_t_3) < 0) __PYX_ERR(0, 1637, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1638
 *     DEST_OVER = HB_PAINT_COMPOSITE_MODE_DEST_OVER
 *     SRC_IN = HB_PAINT_COMPOSITE_MODE_SRC_IN
 *     DEST_IN = HB_PAINT_COMPOSITE_MODE_DEST_IN             # <<<<<<<<<<<<<<
 *     SRC_OUT = HB_PAINT_COMPOSITE_MODE_SRC_OUT
 *     DEST_OUT = HB_PAINT_COMPOSITE_MODE_DEST_OUT
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_paint_composite_mode_t(HB_PAINT_COMPOSITE_MODE_DEST_IN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1638, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_DEST_IN, __pyx_t_3) < 0) __PYX_ERR(0, 1638, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1639
 *     SRC_IN = HB_PAINT_COMPOSITE_MODE_SRC_IN
 *     DEST_IN = HB_PAINT_COMPOSITE_MODE_DEST_IN
 *     SRC_OUT = HB_PAINT_COMPOSITE_MODE_SRC_OUT             # <<<<<<<<<<<<<<
 *     DEST_OUT = HB_PAINT_COMPOSITE_MODE_DEST_OUT
 *     SRC_ATOP = HB_PAINT_COMPOSITE_MODE_SRC_ATOP
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_paint_composite_mode_t(HB_PAINT_COMPOSITE_MODE_SRC_OUT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_SRC_OUT, __pyx_t_3) < 0) __PYX_ERR(0, 1639, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1640
 *     DEST_IN = HB_PAINT_COMPOSITE_MODE_DEST_IN
 *     SRC_OUT = HB_PAINT_COMPOSITE_MODE_SRC_OUT
 *     DEST_OUT = HB_PAINT_COMPOSITE_MODE_DEST_OUT             # <<<<<<<<<<<<<<
 *     SRC_ATOP = HB_PAINT_COMPOSITE_MODE_SRC_ATOP
 *     DEST_ATOP = HB_PAINT_COMPOSITE_MODE_DEST_ATOP
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_paint_composite_mode_t(HB_PAINT_COMPOSITE_MODE_DEST_OUT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1640, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_DEST_OUT, __pyx_t_3) < 0) __PYX_ERR(0, 1640, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1641
 *     SRC_OUT = HB_PAINT_COMPOSITE_MODE_SRC_OUT
 *     DEST_OUT = HB_PAINT_COMPOSITE_MODE_DEST_OUT
 *     SRC_ATOP = HB_PAINT_COMPOSITE_MODE_SRC_ATOP             # <<<<<<<<<<<<<<
 *     DEST_ATOP = HB_PAINT_COMPOSITE_MODE_DEST_ATOP
 *     XOR = HB_PAINT_COMPOSITE_MODE_XOR
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_paint_composite_mode_t(HB_PAINT_COMPOSITE_MODE_SRC_ATOP); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_SRC_ATOP, __pyx_t_3) < 0) __PYX_ERR(0, 1641, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1642
 *     DEST_OUT = HB_PAINT_COMPOSITE_MODE_DEST_OUT
 *     SRC_ATOP = HB_PAINT_COMPOSITE_MODE_SRC_ATOP
 *     DEST_ATOP = HB_PAINT_COMPOSITE_MODE_DEST_ATOP             # <<<<<<<<<<<<<<
 *     XOR = HB_PAINT_COMPOSITE_MODE_XOR
 *     PLUS = HB_PAINT_COMPOSITE_MODE_PLUS
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_paint_composite_mode_t(HB_PAINT_COMPOSITE_MODE_DEST_ATOP); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_DEST_ATOP, __pyx_t_3) < 0) __PYX_ERR(0, 1642, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1643
 *     SRC_ATOP = HB_PAINT_COMPOSITE_MODE_SRC_ATOP
 *     DEST_ATOP = HB_PAINT_COMPOSITE_MODE_DEST_ATOP
 *     XOR = HB_PAINT_COMPOSITE_MODE_XOR             # <<<<<<<<<<<<<<
 *     PLUS = HB_PAINT_COMPOSITE_MODE_PLUS
 *     SCREEN = HB_PAINT_COMPOSITE_MODE_SCREEN
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_paint_composite_mode_t(HB_PAINT_COMPOSITE_MODE_XOR); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_XOR, __pyx_t_3) < 0) __PYX_ERR(0, 1643, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1644
 *     DEST_ATOP = HB_PAINT_COMPOSITE_MODE_DEST_ATOP
 *     XOR = HB_PAINT_COMPOSITE_MODE_XOR
 *     PLUS = HB_PAINT_COMPOSITE_MODE_PLUS             # <<<<<<<<<<<<<<
 *     SCREEN = HB_PAINT_COMPOSITE_MODE_SCREEN
 *     OVERLAY = HB_PAINT_COMPOSITE_MODE_OVERLAY
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_paint_composite_mode_t(HB_PAINT_COMPOSITE_MODE_PLUS); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1644, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_PLUS, __pyx_t_3) < 0) __PYX_ERR(0, 1644, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1645
 *     XOR = HB_PAINT_COMPOSITE_MODE_XOR
 *     PLUS = HB_PAINT_COMPOSITE_MODE_PLUS
 *     SCREEN = HB_PAINT_COMPOSITE_MODE_SCREEN             # <<<<<<<<<<<<<<
 *     OVERLAY = HB_PAINT_COMPOSITE_MODE_OVERLAY
 *     DARKEN = HB_PAINT_COMPOSITE_MODE_DARKEN
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_paint_composite_mode_t(HB_PAINT_COMPOSITE_MODE_SCREEN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_SCREEN, __pyx_t_3) < 0) __PYX_ERR(0, 1645, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1646
 *     PLUS = HB_PAINT_COMPOSITE_MODE_PLUS
 *     SCREEN = HB_PAINT_COMPOSITE_MODE_SCREEN
 *     OVERLAY = HB_PAINT_COMPOSITE_MODE_OVERLAY             # <<<<<<<<<<<<<<
 *     DARKEN = HB_PAINT_COMPOSITE_MODE_DARKEN
 *     LIGHTEN = HB_PAINT_COMPOSITE_MODE_LIGHTEN
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_paint_composite_mode_t(HB_PAINT_COMPOSITE_MODE_OVERLAY); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_OVERLAY, __pyx_t_3) < 0) __PYX_ERR(0, 1646, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1647
 *     SCREEN = HB_PAINT_COMPOSITE_MODE_SCREEN
 *     OVERLAY = HB_PAINT_COMPOSITE_MODE_OVERLAY
 *     DARKEN = HB_PAINT_COMPOSITE_MODE_DARKEN             # <<<<<<<<<<<<<<
 *     LIGHTEN = HB_PAINT_COMPOSITE_MODE_LIGHTEN
 *     COLOR_DODGE = HB_PAINT_COMPOSITE_MODE_COLOR_DODGE
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_paint_composite_mode_t(HB_PAINT_COMPOSITE_MODE_DARKEN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_DARKEN, __pyx_t_3) < 0) __PYX_ERR(0, 1647, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1648
 *     OVERLAY = HB_PAINT_COMPOSITE_MODE_OVERLAY
 *     DARKEN = HB_PAINT_COMPOSITE_MODE_DARKEN
 *     LIGHTEN = HB_PAINT_COMPOSITE_MODE_LIGHTEN             # <<<<<<<<<<<<<<
 *     COLOR_DODGE = HB_PAINT_COMPOSITE_MODE_COLOR_DODGE
 *     COLOR_BURN = HB_PAINT_COMPOSITE_MODE_COLOR_BURN
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_paint_composite_mode_t(HB_PAINT_COMPOSITE_MODE_LIGHTEN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_LIGHTEN, __pyx_t_3) < 0) __PYX_ERR(0, 1648, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1649
 *     DARKEN = HB_PAINT_COMPOSITE_MODE_DARKEN
 *     LIGHTEN = HB_PAINT_COMPOSITE_MODE_LIGHTEN
 *     COLOR_DODGE = HB_PAINT_COMPOSITE_MODE_COLOR_DODGE             # <<<<<<<<<<<<<<
 *     COLOR_BURN = HB_PAINT_COMPOSITE_MODE_COLOR_BURN
 *     HARD_LIGHT = HB_PAINT_COMPOSITE_MODE_HARD_LIGHT
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_paint_composite_mode_t(HB_PAINT_COMPOSITE_MODE_COLOR_DODGE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_COLOR_DODGE, __pyx_t_3) < 0) __PYX_ERR(0, 1649, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1650
 *     LIGHTEN = HB_PAINT_COMPOSITE_MODE_LIGHTEN
 *     COLOR_DODGE = HB_PAINT_COMPOSITE_MODE_COLOR_DODGE
 *     COLOR_BURN = HB_PAINT_COMPOSITE_MODE_COLOR_BURN             # <<<<<<<<<<<<<<
 *     HARD_LIGHT = HB_PAINT_COMPOSITE_MODE_HARD_LIGHT
 *     SOFT_LIGHT = HB_PAINT_COMPOSITE_MODE_SOFT_LIGHT
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_paint_composite_mode_t(HB_PAINT_COMPOSITE_MODE_COLOR_BURN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1650, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_COLOR_BURN, __pyx_t_3) < 0) __PYX_ERR(0, 1650, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1651
 *     COLOR_DODGE = HB_PAINT_COMPOSITE_MODE_COLOR_DODGE
 *     COLOR_BURN = HB_PAINT_COMPOSITE_MODE_COLOR_BURN
 *     HARD_LIGHT = HB_PAINT_COMPOSITE_MODE_HARD_LIGHT             # <<<<<<<<<<<<<<
 *     SOFT_LIGHT = HB_PAINT_COMPOSITE_MODE_SOFT_LIGHT
 *     DIFFERENCE = HB_PAINT_COMPOSITE_MODE_DIFFERENCE
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_paint_composite_mode_t(HB_PAINT_COMPOSITE_MODE_HARD_LIGHT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_HARD_LIGHT, __pyx_t_3) < 0) __PYX_ERR(0, 1651, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1652
 *     COLOR_BURN = HB_PAINT_COMPOSITE_MODE_COLOR_BURN
 *     HARD_LIGHT = HB_PAINT_COMPOSITE_MODE_HARD_LIGHT
 *     SOFT_LIGHT = HB_PAINT_COMPOSITE_MODE_SOFT_LIGHT             # <<<<<<<<<<<<<<
 *     DIFFERENCE = HB_PAINT_COMPOSITE_MODE_DIFFERENCE
 *     EXCLUSION = HB_PAINT_COMPOSITE_MODE_EXCLUSION
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_paint_composite_mode_t(HB_PAINT_COMPOSITE_MODE_SOFT_LIGHT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_SOFT_LIGHT, __pyx_t_3) < 0) __PYX_ERR(0, 1652, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1653
 *     HARD_LIGHT = HB_PAINT_COMPOSITE_MODE_HARD_LIGHT
 *     SOFT_LIGHT = HB_PAINT_COMPOSITE_MODE_SOFT_LIGHT
 *     DIFFERENCE = HB_PAINT_COMPOSITE_MODE_DIFFERENCE             # <<<<<<<<<<<<<<
 *     EXCLUSION = HB_PAINT_COMPOSITE_MODE_EXCLUSION
 *     MULTIPLY = HB_PAINT_COMPOSITE_MODE_MULTIPLY
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_paint_composite_mode_t(HB_PAINT_COMPOSITE_MODE_DIFFERENCE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_DIFFERENCE, __pyx_t_3) < 0) __PYX_ERR(0, 1653, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1654
 *     SOFT_LIGHT = HB_PAINT_COMPOSITE_MODE_SOFT_LIGHT
 *     DIFFERENCE = HB_PAINT_COMPOSITE_MODE_DIFFERENCE
 *     EXCLUSION = HB_PAINT_COMPOSITE_MODE_EXCLUSION             # <<<<<<<<<<<<<<
 *     MULTIPLY = HB_PAINT_COMPOSITE_MODE_MULTIPLY
 *     HSL_HUE = HB_PAINT_COMPOSITE_MODE_HSL_HUE
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_paint_composite_mode_t(HB_PAINT_COMPOSITE_MODE_EXCLUSION); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_EXCLUSION, __pyx_t_3) < 0) __PYX_ERR(0, 1654, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1655
 *     DIFFERENCE = HB_PAINT_COMPOSITE_MODE_DIFFERENCE
 *     EXCLUSION = HB_PAINT_COMPOSITE_MODE_EXCLUSION
 *     MULTIPLY = HB_PAINT_COMPOSITE_MODE_MULTIPLY             # <<<<<<<<<<<<<<
 *     HSL_HUE = HB_PAINT_COMPOSITE_MODE_HSL_HUE
 *     HSL_SATURATION = HB_PAINT_COMPOSITE_MODE_HSL_SATURATION
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_paint_composite_mode_t(HB_PAINT_COMPOSITE_MODE_MULTIPLY); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_MULTIPLY, __pyx_t_3) < 0) __PYX_ERR(0, 1655, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1656
 *     EXCLUSION = HB_PAINT_COMPOSITE_MODE_EXCLUSION
 *     MULTIPLY = HB_PAINT_COMPOSITE_MODE_MULTIPLY
 *     HSL_HUE = HB_PAINT_COMPOSITE_MODE_HSL_HUE             # <<<<<<<<<<<<<<
 *     HSL_SATURATION = HB_PAINT_COMPOSITE_MODE_HSL_SATURATION
 *     HSL_COLOR = HB_PAINT_COMPOSITE_MODE_HSL_COLOR
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_paint_composite_mode_t(HB_PAINT_COMPOSITE_MODE_HSL_HUE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_HSL_HUE, __pyx_t_3) < 0) __PYX_ERR(0, 1656, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1657
 *     MULTIPLY = HB_PAINT_COMPOSITE_MODE_MULTIPLY
 *     HSL_HUE = HB_PAINT_COMPOSITE_MODE_HSL_HUE
 *     HSL_SATURATION = HB_PAINT_COMPOSITE_MODE_HSL_SATURATION             # <<<<<<<<<<<<<<
 *     HSL_COLOR = HB_PAINT_COMPOSITE_MODE_HSL_COLOR
 *     HSL_LUMINOSITY = HB_PAINT_COMPOSITE_MODE_HSL_LUMINOSITY
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_paint_composite_mode_t(HB_PAINT_COMPOSITE_MODE_HSL_SATURATION); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_HSL_SATURATION, __pyx_t_3) < 0) __PYX_ERR(0, 1657, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1658
 *     HSL_HUE = HB_PAINT_COMPOSITE_MODE_HSL_HUE
 *     HSL_SATURATION = HB_PAINT_COMPOSITE_MODE_HSL_SATURATION
 *     HSL_COLOR = HB_PAINT_COMPOSITE_MODE_HSL_COLOR             # <<<<<<<<<<<<<<
 *     HSL_LUMINOSITY = HB_PAINT_COMPOSITE_MODE_HSL_LUMINOSITY
 * 
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_paint_composite_mode_t(HB_PAINT_COMPOSITE_MODE_HSL_COLOR); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_HSL_COLOR, __pyx_t_3) < 0) __PYX_ERR(0, 1658, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1659
 *     HSL_SATURATION = HB_PAINT_COMPOSITE_MODE_HSL_SATURATION
 *     HSL_COLOR = HB_PAINT_COMPOSITE_MODE_HSL_COLOR
 *     HSL_LUMINOSITY = HB_PAINT_COMPOSITE_MODE_HSL_LUMINOSITY             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_paint_composite_mode_t(HB_PAINT_COMPOSITE_MODE_HSL_LUMINOSITY); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_4, __pyx_n_s_HSL_LUMINOSITY, __pyx_t_3) < 0) __PYX_ERR(0, 1659, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1631
 * 
 * 
 * class PaintCompositeMode(IntEnum):             # <<<<<<<<<<<<<<
 *     CLEAR = HB_PAINT_COMPOSITE_MODE_CLEAR
 *     SRC = HB_PAINT_COMPOSITE_MODE_SRC
 */
  __pyx_t_3 = __Pyx_Py3ClassCreate(__pyx_t_5, __pyx_n_s_PaintCompositeMode, __pyx_t_2, __pyx_t_4, NULL, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PaintCompositeMode, __pyx_t_3) < 0) __PYX_ERR(0, 1631, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1662
 * 
 * 
 * class Color(namedtuple("Color", ["red", "green", "blue", "alpha"])):             # <<<<<<<<<<<<<<
 *     def to_int(self) -> int:
 *         return HB_COLOR(self.blue, self.green, self.red, self.alpha)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_namedtuple); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyList_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_n_u_red);
  __Pyx_GIVEREF(__pyx_n_u_red);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 0, __pyx_n_u_red)) __PYX_ERR(0, 1662, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_green);
  __Pyx_GIVEREF(__pyx_n_u_green);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 1, __pyx_n_u_green)) __PYX_ERR(0, 1662, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_blue);
  __Pyx_GIVEREF(__pyx_n_u_blue);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 2, __pyx_n_u_blue)) __PYX_ERR(0, 1662, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_alpha);
  __Pyx_GIVEREF(__pyx_n_u_alpha);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 3, __pyx_n_u_alpha)) __PYX_ERR(0, 1662, __pyx_L1_error);
  __pyx_t_3 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_n_u_Color, __pyx_t_4};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1662, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2)) __PYX_ERR(0, 1662, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PEP560_update_bases(__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_CalculateMetaclass(NULL, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_Py3MetaclassPrepare(__pyx_t_4, __pyx_t_2, __pyx_n_s_Color, __pyx_n_s_Color, (PyObject *) NULL, __pyx_n_s_uharfbuzz__harfbuzz, (PyObject *) NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__pyx_t_2 != __pyx_t_5) {
    if (unlikely((PyDict_SetItemString(__pyx_t_3, "__orig_bases__", __pyx_t_5) < 0))) __PYX_ERR(0, 1662, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1663
 * 
 * class Color(namedtuple("Color", ["red", "green", "blue", "alpha"])):
 *     def to_int(self) -> int:             # <<<<<<<<<<<<<<
 *         return HB_COLOR(self.blue, self.green, self.red, self.alpha)
 * 
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return, __pyx_n_s_int) < 0) __PYX_ERR(0, 1663, __pyx_L1_error)
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_5Color_1to_int, 0, __pyx_n_s_Color_to_int, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__181)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_to_int, __pyx_t_7) < 0) __PYX_ERR(0, 1663, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1666
 *         return HB_COLOR(self.blue, self.green, self.red, self.alpha)
 * 
 *     @staticmethod             # <<<<<<<<<<<<<<
 *     def from_int(value: int) -> Color:
 *         r = hb_color_get_red(value)
 */
  __pyx_t_7 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_value, __pyx_n_s_int) < 0) __PYX_ERR(0, 1666, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_return, __pyx_n_s_Color) < 0) __PYX_ERR(0, 1666, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_5Color_3from_int, __Pyx_CYFUNCTION_STATICMETHOD, __pyx_n_s_Color_from_int, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__183)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_builtin_staticmethod, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetNameInClass(__pyx_t_3, __pyx_n_s_from_int, __pyx_t_7) < 0) __PYX_ERR(0, 1666, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1662
 * 
 * 
 * class Color(namedtuple("Color", ["red", "green", "blue", "alpha"])):             # <<<<<<<<<<<<<<
 *     def to_int(self) -> int:
 *         return HB_COLOR(self.blue, self.green, self.red, self.alpha)
 */
  __pyx_t_7 = __Pyx_Py3ClassCreate(__pyx_t_4, __pyx_n_s_Color, __pyx_t_2, __pyx_t_3, NULL, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Color, __pyx_t_7) < 0) __PYX_ERR(0, 1662, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1675
 * 
 * 
 * ColorStop = namedtuple("ColorStop", ["offset", "is_foreground", "color"])             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_namedtuple); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1675, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = PyList_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1675, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_n_u_offset);
  __Pyx_GIVEREF(__pyx_n_u_offset);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 0, __pyx_n_u_offset)) __PYX_ERR(0, 1675, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_is_foreground);
  __Pyx_GIVEREF(__pyx_n_u_is_foreground);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 1, __pyx_n_u_is_foreground)) __PYX_ERR(0, 1675, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_color);
  __Pyx_GIVEREF(__pyx_n_u_color);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 2, __pyx_n_u_color)) __PYX_ERR(0, 1675, __pyx_L1_error);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1675, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_u_ColorStop);
  __Pyx_GIVEREF(__pyx_n_u_ColorStop);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_n_u_ColorStop)) __PYX_ERR(0, 1675, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4)) __PYX_ERR(0, 1675, __pyx_L1_error);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1675, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ColorStop, __pyx_t_4) < 0) __PYX_ERR(0, 1675, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1678
 * 
 * 
 * class PaintExtend(IntEnum):             # <<<<<<<<<<<<<<
 *     PAD = HB_PAINT_EXTEND_PAD
 *     REPEAT = HB_PAINT_EXTEND_REPEAT
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_IntEnum); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(0, 1678, __pyx_L1_error);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PEP560_update_bases(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_4, __pyx_n_s_PaintExtend, __pyx_n_s_PaintExtend, (PyObject *) NULL, __pyx_n_s_uharfbuzz__harfbuzz, (PyObject *) NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__pyx_t_4 != __pyx_t_3) {
    if (unlikely((PyDict_SetItemString(__pyx_t_7, "__orig_bases__", __pyx_t_3) < 0))) __PYX_ERR(0, 1678, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1679
 * 
 * class PaintExtend(IntEnum):
 *     PAD = HB_PAINT_EXTEND_PAD             # <<<<<<<<<<<<<<
 *     REPEAT = HB_PAINT_EXTEND_REPEAT
 *     REFLECT = HB_PAINT_EXTEND_REFLECT
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_paint_extend_t(HB_PAINT_EXTEND_PAD); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_PAD, __pyx_t_3) < 0) __PYX_ERR(0, 1679, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1680
 * class PaintExtend(IntEnum):
 *     PAD = HB_PAINT_EXTEND_PAD
 *     REPEAT = HB_PAINT_EXTEND_REPEAT             # <<<<<<<<<<<<<<
 *     REFLECT = HB_PAINT_EXTEND_REFLECT
 * 
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_paint_extend_t(HB_PAINT_EXTEND_REPEAT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_REPEAT, __pyx_t_3) < 0) __PYX_ERR(0, 1680, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1681
 *     PAD = HB_PAINT_EXTEND_PAD
 *     REPEAT = HB_PAINT_EXTEND_REPEAT
 *     REFLECT = HB_PAINT_EXTEND_REFLECT             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_paint_extend_t(HB_PAINT_EXTEND_REFLECT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1681, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_REFLECT, __pyx_t_3) < 0) __PYX_ERR(0, 1681, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1678
 * 
 * 
 * class PaintExtend(IntEnum):             # <<<<<<<<<<<<<<
 *     PAD = HB_PAINT_EXTEND_PAD
 *     REPEAT = HB_PAINT_EXTEND_REPEAT
 */
  __pyx_t_3 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_PaintExtend, __pyx_t_4, __pyx_t_7, NULL, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PaintExtend, __pyx_t_3) < 0) __PYX_ERR(0, 1678, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_9ColorLine_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ColorLine___reduce_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__184)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_9ColorLine_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_ColorLine___setstate_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__185)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1919
 *         hb_paint_funcs_destroy(self._hb_paintfuncs)
 * 
 *     def set_push_transform_func(self,             # <<<<<<<<<<<<<<
 *                                 func: Callable[[
 *                                     float,  # xx
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1919, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_func, __pyx_kp_s_Callable_float_float_float_float) < 0) __PYX_ERR(0, 1919, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 1919, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_10PaintFuncs_5set_push_transform_func, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaintFuncs_set_push_transform_fu, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__187)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1919, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs, __pyx_n_s_set_push_transform_func, __pyx_t_2) < 0) __PYX_ERR(0, 1919, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs);

  /* "uharfbuzz/_harfbuzz.pyx":1933
 *             self._hb_paintfuncs, _paint_push_transform_func, <void*>self, NULL)
 * 
 *     def set_pop_transform_func(self,             # <<<<<<<<<<<<<<
 *                                func: Callable[[
 *                                    object,  # paint_data
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1933, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_func, __pyx_kp_s_Callable_object_None) < 0) __PYX_ERR(0, 1933, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 1933, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_10PaintFuncs_7set_pop_transform_func, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaintFuncs_set_pop_transform_fun, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__188)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1933, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs, __pyx_n_s_set_pop_transform_func, __pyx_t_4) < 0) __PYX_ERR(0, 1933, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs);

  /* "uharfbuzz/_harfbuzz.pyx":1941
 *             self._hb_paintfuncs, _paint_pop_transform_func, <void*>self, NULL)
 * 
 *     def set_color_glyph_func(self,             # <<<<<<<<<<<<<<
 *                              func: Callable[[
 *                                  int,  # gid
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1941, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_func, __pyx_kp_s_Callable_int_object_bool) < 0) __PYX_ERR(0, 1941, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 1941, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_10PaintFuncs_9set_color_glyph_func, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaintFuncs_set_color_glyph_func, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__189)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1941, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs, __pyx_n_s_set_color_glyph_func, __pyx_t_2) < 0) __PYX_ERR(0, 1941, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs);

  /* "uharfbuzz/_harfbuzz.pyx":1950
 *             self._hb_paintfuncs, _paint_color_glyph_func, <void*>self, NULL)
 * 
 *     def set_push_clip_glyph_func(self,             # <<<<<<<<<<<<<<
 *                                  func: Callable[[
 *                                      int,  # gid
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1950, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_func, __pyx_kp_s_Callable_int_object_None) < 0) __PYX_ERR(0, 1950, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 1950, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_10PaintFuncs_11set_push_clip_glyph_func, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaintFuncs_set_push_clip_glyph_f, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__190)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1950, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs, __pyx_n_s_set_push_clip_glyph_func, __pyx_t_4) < 0) __PYX_ERR(0, 1950, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs);

  /* "uharfbuzz/_harfbuzz.pyx":1959
 *             self._hb_paintfuncs, _paint_push_clip_glyph_func, <void*>self, NULL)
 * 
 *     def set_push_clip_rectangle_func(self,             # <<<<<<<<<<<<<<
 *                                      func: Callable[[
 *                                          float,  # xmin
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1959, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_func, __pyx_kp_s_Callable_float_float_float_float_2) < 0) __PYX_ERR(0, 1959, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 1959, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_10PaintFuncs_13set_push_clip_rectangle_func, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaintFuncs_set_push_clip_rectang, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__191)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1959, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs, __pyx_n_s_set_push_clip_rectangle_func, __pyx_t_2) < 0) __PYX_ERR(0, 1959, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs);

  /* "uharfbuzz/_harfbuzz.pyx":1971
 *             self._hb_paintfuncs, _paint_push_clip_rectangle_func, <void*>self, NULL)
 * 
 *     def set_pop_clip_func(self,             # <<<<<<<<<<<<<<
 *                           func: Callable[[
 *                               object,  # paint_data
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1971, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_func, __pyx_kp_s_Callable_object_None) < 0) __PYX_ERR(0, 1971, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 1971, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_10PaintFuncs_15set_pop_clip_func, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaintFuncs_set_pop_clip_func, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__192)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1971, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs, __pyx_n_s_set_pop_clip_func, __pyx_t_4) < 0) __PYX_ERR(0, 1971, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs);

  /* "uharfbuzz/_harfbuzz.pyx":1979
 *             self._hb_paintfuncs, _paint_pop_clip_func, <void*>self, NULL)
 * 
 *     def set_color_func(self,             # <<<<<<<<<<<<<<
 *                        func: Callable[[
 *                            Color,  # color
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1979, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_func, __pyx_kp_s_Callable_Color_bool_object_None) < 0) __PYX_ERR(0, 1979, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 1979, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_10PaintFuncs_17set_color_func, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaintFuncs_set_color_func, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__193)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1979, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs, __pyx_n_s_set_color_func, __pyx_t_2) < 0) __PYX_ERR(0, 1979, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs);

  /* "uharfbuzz/_harfbuzz.pyx":1989
 *             self._hb_paintfuncs, _paint_color_func, <void*>self, NULL)
 * 
 *     def set_image_func(self,             # <<<<<<<<<<<<<<
 *                        func: Callable[[
 *                            Blob,  # image
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_func, __pyx_kp_s_Callable_Blob_int_int_str_float) < 0) __PYX_ERR(0, 1989, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 1989, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_10PaintFuncs_19set_image_func, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaintFuncs_set_image_func, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__194)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs, __pyx_n_s_set_image_func, __pyx_t_4) < 0) __PYX_ERR(0, 1989, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs);

  /* "uharfbuzz/_harfbuzz.pyx":2003
 *             self._hb_paintfuncs, _paint_image_func, <void*>self, NULL)
 * 
 *     def set_linear_gradient_func(self,             # <<<<<<<<<<<<<<
 *                                  func: Callable[[
 *                                     ColorLine,  # color_line
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_func, __pyx_kp_s_Callable_ColorLine_float_float_f) < 0) __PYX_ERR(0, 2003, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 2003, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_10PaintFuncs_21set_linear_gradient_func, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaintFuncs_set_linear_gradient_f, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__195)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs, __pyx_n_s_set_linear_gradient_func, __pyx_t_2) < 0) __PYX_ERR(0, 2003, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs);

  /* "uharfbuzz/_harfbuzz.pyx":2018
 *             self._hb_paintfuncs, _paint_linear_gradient_func, <void*>self, NULL)
 * 
 *     def set_radial_gradient_func(self,             # <<<<<<<<<<<<<<
 *                                  func: Callable[[
 *                                     ColorLine,  # color_line
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_func, __pyx_kp_s_Callable_ColorLine_float_float_f) < 0) __PYX_ERR(0, 2018, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 2018, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_10PaintFuncs_23set_radial_gradient_func, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaintFuncs_set_radial_gradient_f, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__196)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs, __pyx_n_s_set_radial_gradient_func, __pyx_t_4) < 0) __PYX_ERR(0, 2018, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs);

  /* "uharfbuzz/_harfbuzz.pyx":2033
 *             self._hb_paintfuncs, _paint_radial_gradient_func, <void*>self, NULL)
 * 
 *     def set_sweep_gradient_func(self,             # <<<<<<<<<<<<<<
 *                                 func: Callable[[
 *                                     ColorLine,  # color_line
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2033, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_func, __pyx_kp_s_Callable_ColorLine_float_float_f_2) < 0) __PYX_ERR(0, 2033, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 2033, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_10PaintFuncs_25set_sweep_gradient_func, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaintFuncs_set_sweep_gradient_fu, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__197)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2033, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs, __pyx_n_s_set_sweep_gradient_func, __pyx_t_2) < 0) __PYX_ERR(0, 2033, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs);

  /* "uharfbuzz/_harfbuzz.pyx":2046
 *             self._hb_paintfuncs, _paint_sweep_gradient_func, <void*>self, NULL)
 * 
 *     def set_push_group_func(self,             # <<<<<<<<<<<<<<
 *                             func: Callable[[
 *                                 object,  # paint_data
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2046, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_func, __pyx_kp_s_Callable_object_None) < 0) __PYX_ERR(0, 2046, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 2046, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_10PaintFuncs_27set_push_group_func, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaintFuncs_set_push_group_func, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__198)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2046, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs, __pyx_n_s_set_push_group_func, __pyx_t_4) < 0) __PYX_ERR(0, 2046, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs);

  /* "uharfbuzz/_harfbuzz.pyx":2054
 *             self._hb_paintfuncs, _paint_push_group_func, <void*>self, NULL)
 * 
 *     def set_pop_group_func(self,             # <<<<<<<<<<<<<<
 *                            func: Callable[[
 *                                PaintCompositeMode,  # mode
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2054, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_func, __pyx_kp_s_Callable_PaintCompositeMode_obje) < 0) __PYX_ERR(0, 2054, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 2054, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_10PaintFuncs_29set_pop_group_func, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaintFuncs_set_pop_group_func, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__199)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2054, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs, __pyx_n_s_set_pop_group_func, __pyx_t_2) < 0) __PYX_ERR(0, 2054, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs);

  /* "uharfbuzz/_harfbuzz.pyx":2063
 *             self._hb_paintfuncs, _paint_pop_group_func, <void*>self, NULL)
 * 
 *     def set_custom_palette_color_func(self,             # <<<<<<<<<<<<<<
 *                                       func: Callable[[
 *                                           int,  # color_index
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2063, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_func, __pyx_kp_s_Callable_int_object_Color) < 0) __PYX_ERR(0, 2063, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 2063, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_10PaintFuncs_31set_custom_palette_color_func, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaintFuncs_set_custom_palette_co, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__200)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2063, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs, __pyx_n_s_set_custom_palette_color_func, __pyx_t_4) < 0) __PYX_ERR(0, 2063, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_PaintFuncs);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_10PaintFuncs_33__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaintFuncs___reduce_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__201)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_10PaintFuncs_35__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaintFuncs___setstate_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__202)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2139
 *         hb_draw_funcs_destroy(self._hb_drawfuncs)
 * 
 *     def get_glyph_shape(self, font: Font, gid: int):             # <<<<<<<<<<<<<<
 *         if not self._warned:
 *             warnings.warn(
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_font, __pyx_n_s_Font) < 0) __PYX_ERR(0, 2139, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_gid, __pyx_n_s_int) < 0) __PYX_ERR(0, 2139, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_9DrawFuncs_5get_glyph_shape, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_DrawFuncs_get_glyph_shape, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__204)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_DrawFuncs, __pyx_n_s_get_glyph_shape, __pyx_t_2) < 0) __PYX_ERR(0, 2139, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_DrawFuncs);

  /* "uharfbuzz/_harfbuzz.pyx":2148
 *         font.draw_glyph(gid, self)
 * 
 *     def draw_glyph(self, font: Font, gid: int, draw_data: object = None):             # <<<<<<<<<<<<<<
 *         if not self._warned:
 *             warnings.warn(
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_font, __pyx_n_s_Font) < 0) __PYX_ERR(0, 2148, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_gid, __pyx_n_s_int) < 0) __PYX_ERR(0, 2148, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_draw_data, __pyx_n_s_object) < 0) __PYX_ERR(0, 2148, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_9DrawFuncs_7draw_glyph, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_DrawFuncs_draw_glyph, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__206)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_4, __pyx_tuple__7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_DrawFuncs, __pyx_n_s_draw_glyph, __pyx_t_4) < 0) __PYX_ERR(0, 2148, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_DrawFuncs);

  /* "uharfbuzz/_harfbuzz.pyx":2157
 *         font.draw_glyph(gid, self, draw_data)
 * 
 *     def set_move_to_func(self,             # <<<<<<<<<<<<<<
 *                          func: Callable[[
 *                              float,
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_func, __pyx_kp_s_Callable_float_float_object_None) < 0) __PYX_ERR(0, 2157, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_user_data, __pyx_n_s_object) < 0) __PYX_ERR(0, 2157, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 2157, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_9DrawFuncs_9set_move_to_func, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_DrawFuncs_set_move_to_func, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__208)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_tuple__7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_DrawFuncs, __pyx_n_s_set_move_to_func, __pyx_t_2) < 0) __PYX_ERR(0, 2157, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_DrawFuncs);

  /* "uharfbuzz/_harfbuzz.pyx":2181
 *             self._hb_drawfuncs, func_p, user_data_p, NULL)
 * 
 *     def set_line_to_func(self,             # <<<<<<<<<<<<<<
 *                          func: Callable[[
 *                              float,
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_func, __pyx_kp_s_Callable_float_float_object_None) < 0) __PYX_ERR(0, 2181, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_user_data, __pyx_n_s_object) < 0) __PYX_ERR(0, 2181, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 2181, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_9DrawFuncs_11set_line_to_func, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_DrawFuncs_set_line_to_func, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__209)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_4, __pyx_tuple__7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_DrawFuncs, __pyx_n_s_set_line_to_func, __pyx_t_4) < 0) __PYX_ERR(0, 2181, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_DrawFuncs);

  /* "uharfbuzz/_harfbuzz.pyx":2205
 *             self._hb_drawfuncs, func_p, user_data_p, NULL)
 * 
 *     def set_cubic_to_func(self,             # <<<<<<<<<<<<<<
 *                           func: Callable[[
 *                              float,
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_func, __pyx_kp_s_Callable_float_float_float_float) < 0) __PYX_ERR(0, 2205, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_user_data, __pyx_n_s_object) < 0) __PYX_ERR(0, 2205, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 2205, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_9DrawFuncs_13set_cubic_to_func, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_DrawFuncs_set_cubic_to_func, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__210)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_tuple__7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_DrawFuncs, __pyx_n_s_set_cubic_to_func, __pyx_t_2) < 0) __PYX_ERR(0, 2205, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_DrawFuncs);

  /* "uharfbuzz/_harfbuzz.pyx":2233
 *             self._hb_drawfuncs, func_p, user_data_p, NULL)
 * 
 *     def set_quadratic_to_func(self,             # <<<<<<<<<<<<<<
 *                               func: Callable[[
 *                                  float,
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_func, __pyx_kp_s_Callable_float_float_float_float_2) < 0) __PYX_ERR(0, 2233, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_user_data, __pyx_n_s_object) < 0) __PYX_ERR(0, 2233, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 2233, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_9DrawFuncs_15set_quadratic_to_func, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_DrawFuncs_set_quadratic_to_func, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__211)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_4, __pyx_tuple__7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_DrawFuncs, __pyx_n_s_set_quadratic_to_func, __pyx_t_4) < 0) __PYX_ERR(0, 2233, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_DrawFuncs);

  /* "uharfbuzz/_harfbuzz.pyx":2259
 *             self._hb_drawfuncs, func_p, user_data_p, NULL)
 * 
 *     def set_close_path_func(self,             # <<<<<<<<<<<<<<
 *                             func: Callable[[
 *                                 object
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_func, __pyx_kp_s_Callable_object_None) < 0) __PYX_ERR(0, 2259, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_user_data, __pyx_n_s_object) < 0) __PYX_ERR(0, 2259, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_return, __pyx_n_s_None) < 0) __PYX_ERR(0, 2259, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_9DrawFuncs_17set_close_path_func, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_DrawFuncs_set_close_path_func, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__212)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_2, __pyx_tuple__7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_DrawFuncs, __pyx_n_s_set_close_path_func, __pyx_t_2) < 0) __PYX_ERR(0, 2259, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_DrawFuncs);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_9DrawFuncs_19__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_DrawFuncs___reduce_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__213)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_9DrawFuncs_21__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_DrawFuncs___setstate_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__214)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_8HBObject_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_HBObject___reduce_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__215)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_8HBObject_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_HBObject___setstate_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__216)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2339
 *             l[i].objidx = links[i][2]
 * 
 * class RepackerError(Exception):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0]));
  __Pyx_GIVEREF((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0]));
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])))) __PYX_ERR(0, 2339, __pyx_L1_error);
  __pyx_t_4 = __Pyx_PEP560_update_bases(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = __Pyx_CalculateMetaclass(NULL, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_3 = __Pyx_Py3MetaclassPrepare(__pyx_t_7, __pyx_t_4, __pyx_n_s_RepackerError, __pyx_n_s_RepackerError, (PyObject *) NULL, __pyx_n_s_uharfbuzz__harfbuzz, (PyObject *) NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__pyx_t_4 != __pyx_t_2) {
    if (unlikely((PyDict_SetItemString(__pyx_t_3, "__orig_bases__", __pyx_t_2) < 0))) __PYX_ERR(0, 2339, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_Py3ClassCreate(__pyx_t_7, __pyx_n_s_RepackerError, __pyx_t_4, __pyx_t_3, NULL, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2339, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_RepackerError, __pyx_t_2) < 0) __PYX_ERR(0, 2339, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2343
 * 
 * 
 * def repack(subtables: List[bytes],             # <<<<<<<<<<<<<<
 *            graphnodes: List[Tuple[List[Tuple[int, int, int]],
 *                             List[Tuple[int, int, int]]
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_subtables, __pyx_kp_s_List_bytes) < 0) __PYX_ERR(0, 2343, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_graphnodes, __pyx_kp_s_List_Tuple_List_Tuple_int_int_in) < 0) __PYX_ERR(0, 2343, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_return, __pyx_n_s_bytes) < 0) __PYX_ERR(0, 2343, __pyx_L1_error)
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_77repack, 0, __pyx_n_s_repack, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__218)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2343, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_repack, __pyx_t_7) < 0) __PYX_ERR(0, 2343, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2350
 * 
 * 
 * def repack_with_tag(tag: str,             # <<<<<<<<<<<<<<
 *                     subtables: List[bytes],
 *                     graphnodes: List[Tuple[List[Tuple[int, int, int]],
 */
  __pyx_t_7 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_tag, __pyx_n_s_str) < 0) __PYX_ERR(0, 2350, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_subtables, __pyx_kp_s_List_bytes) < 0) __PYX_ERR(0, 2350, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_graphnodes, __pyx_kp_s_List_Tuple_List_Tuple_int_int_in) < 0) __PYX_ERR(0, 2350, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_return, __pyx_n_s_bytes) < 0) __PYX_ERR(0, 2350, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_79repack_with_tag, 0, __pyx_n_s_repack_with_tag, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__220)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_repack_with_tag, __pyx_t_4) < 0) __PYX_ERR(0, 2350, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2435
 * 
 * 
 * def subset_preprocess(face: Face) -> Face:             # <<<<<<<<<<<<<<
 *     new_face = hb_subset_preprocess(face._hb_face)
 *     return Face.from_ptr(new_face)
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_face, __pyx_n_s_Face) < 0) __PYX_ERR(0, 2435, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_return, __pyx_n_s_Face) < 0) __PYX_ERR(0, 2435, __pyx_L1_error)
  __pyx_t_7 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_81subset_preprocess, 0, __pyx_n_s_subset_preprocess, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__222)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_subset_preprocess, __pyx_t_7) < 0) __PYX_ERR(0, 2435, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2439
 *     return Face.from_ptr(new_face)
 * 
 * def subset(face: Face, input: SubsetInput) -> Face:             # <<<<<<<<<<<<<<
 *     new_face = hb_subset_or_fail(face._hb_face, input._hb_input)
 *     if new_face == NULL:
 */
  __pyx_t_7 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_face, __pyx_n_s_Face) < 0) __PYX_ERR(0, 2439, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_input, __pyx_n_s_SubsetInput) < 0) __PYX_ERR(0, 2439, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_return, __pyx_n_s_Face) < 0) __PYX_ERR(0, 2439, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_83subset, 0, __pyx_n_s_subset, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__224)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_subset, __pyx_t_4) < 0) __PYX_ERR(0, 2439, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2445
 *     return Face.from_ptr(new_face)
 * 
 * class SubsetInputSets(IntEnum):             # <<<<<<<<<<<<<<
 *     GLYPH_INDEX = HB_SUBSET_SETS_GLYPH_INDEX
 *     UNICODE = HB_SUBSET_SETS_UNICODE
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_IntEnum); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2445, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2445, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_4)) __PYX_ERR(0, 2445, __pyx_L1_error);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PEP560_update_bases(__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2445, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_CalculateMetaclass(NULL, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2445, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_Py3MetaclassPrepare(__pyx_t_3, __pyx_t_4, __pyx_n_s_SubsetInputSets, __pyx_n_s_SubsetInputSets, (PyObject *) NULL, __pyx_n_s_uharfbuzz__harfbuzz, (PyObject *) NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2445, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__pyx_t_4 != __pyx_t_7) {
    if (unlikely((PyDict_SetItemString(__pyx_t_2, "__orig_bases__", __pyx_t_7) < 0))) __PYX_ERR(0, 2445, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2446
 * 
 * class SubsetInputSets(IntEnum):
 *     GLYPH_INDEX = HB_SUBSET_SETS_GLYPH_INDEX             # <<<<<<<<<<<<<<
 *     UNICODE = HB_SUBSET_SETS_UNICODE
 *     NO_SUBSET_TABLE_TAG = HB_SUBSET_SETS_NO_SUBSET_TABLE_TAG
 */
  __pyx_t_7 = __Pyx_PyInt_From_hb_subset_sets_t(HB_SUBSET_SETS_GLYPH_INDEX); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_GLYPH_INDEX, __pyx_t_7) < 0) __PYX_ERR(0, 2446, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2447
 * class SubsetInputSets(IntEnum):
 *     GLYPH_INDEX = HB_SUBSET_SETS_GLYPH_INDEX
 *     UNICODE = HB_SUBSET_SETS_UNICODE             # <<<<<<<<<<<<<<
 *     NO_SUBSET_TABLE_TAG = HB_SUBSET_SETS_NO_SUBSET_TABLE_TAG
 *     DROP_TABLE_TAG = HB_SUBSET_SETS_DROP_TABLE_TAG
 */
  __pyx_t_7 = __Pyx_PyInt_From_hb_subset_sets_t(HB_SUBSET_SETS_UNICODE); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_UNICODE, __pyx_t_7) < 0) __PYX_ERR(0, 2447, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2448
 *     GLYPH_INDEX = HB_SUBSET_SETS_GLYPH_INDEX
 *     UNICODE = HB_SUBSET_SETS_UNICODE
 *     NO_SUBSET_TABLE_TAG = HB_SUBSET_SETS_NO_SUBSET_TABLE_TAG             # <<<<<<<<<<<<<<
 *     DROP_TABLE_TAG = HB_SUBSET_SETS_DROP_TABLE_TAG
 *     NAME_ID = HB_SUBSET_SETS_NAME_ID
 */
  __pyx_t_7 = __Pyx_PyInt_From_hb_subset_sets_t(HB_SUBSET_SETS_NO_SUBSET_TABLE_TAG); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_NO_SUBSET_TABLE_TAG, __pyx_t_7) < 0) __PYX_ERR(0, 2448, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2449
 *     UNICODE = HB_SUBSET_SETS_UNICODE
 *     NO_SUBSET_TABLE_TAG = HB_SUBSET_SETS_NO_SUBSET_TABLE_TAG
 *     DROP_TABLE_TAG = HB_SUBSET_SETS_DROP_TABLE_TAG             # <<<<<<<<<<<<<<
 *     NAME_ID = HB_SUBSET_SETS_NAME_ID
 *     NAME_LANG_ID = HB_SUBSET_SETS_NAME_LANG_ID
 */
  __pyx_t_7 = __Pyx_PyInt_From_hb_subset_sets_t(HB_SUBSET_SETS_DROP_TABLE_TAG); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2449, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_DROP_TABLE_TAG, __pyx_t_7) < 0) __PYX_ERR(0, 2449, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2450
 *     NO_SUBSET_TABLE_TAG = HB_SUBSET_SETS_NO_SUBSET_TABLE_TAG
 *     DROP_TABLE_TAG = HB_SUBSET_SETS_DROP_TABLE_TAG
 *     NAME_ID = HB_SUBSET_SETS_NAME_ID             # <<<<<<<<<<<<<<
 *     NAME_LANG_ID = HB_SUBSET_SETS_NAME_LANG_ID
 *     LAYOUT_FEATURE_TAG = HB_SUBSET_SETS_LAYOUT_FEATURE_TAG
 */
  __pyx_t_7 = __Pyx_PyInt_From_hb_subset_sets_t(HB_SUBSET_SETS_NAME_ID); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_NAME_ID, __pyx_t_7) < 0) __PYX_ERR(0, 2450, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2451
 *     DROP_TABLE_TAG = HB_SUBSET_SETS_DROP_TABLE_TAG
 *     NAME_ID = HB_SUBSET_SETS_NAME_ID
 *     NAME_LANG_ID = HB_SUBSET_SETS_NAME_LANG_ID             # <<<<<<<<<<<<<<
 *     LAYOUT_FEATURE_TAG = HB_SUBSET_SETS_LAYOUT_FEATURE_TAG
 *     LAYOUT_SCRIPT_TAG = HB_SUBSET_SETS_LAYOUT_SCRIPT_TAG
 */
  __pyx_t_7 = __Pyx_PyInt_From_hb_subset_sets_t(HB_SUBSET_SETS_NAME_LANG_ID); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_NAME_LANG_ID, __pyx_t_7) < 0) __PYX_ERR(0, 2451, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2452
 *     NAME_ID = HB_SUBSET_SETS_NAME_ID
 *     NAME_LANG_ID = HB_SUBSET_SETS_NAME_LANG_ID
 *     LAYOUT_FEATURE_TAG = HB_SUBSET_SETS_LAYOUT_FEATURE_TAG             # <<<<<<<<<<<<<<
 *     LAYOUT_SCRIPT_TAG = HB_SUBSET_SETS_LAYOUT_SCRIPT_TAG
 * 
 */
  __pyx_t_7 = __Pyx_PyInt_From_hb_subset_sets_t(HB_SUBSET_SETS_LAYOUT_FEATURE_TAG); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_LAYOUT_FEATURE_TAG, __pyx_t_7) < 0) __PYX_ERR(0, 2452, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2453
 *     NAME_LANG_ID = HB_SUBSET_SETS_NAME_LANG_ID
 *     LAYOUT_FEATURE_TAG = HB_SUBSET_SETS_LAYOUT_FEATURE_TAG
 *     LAYOUT_SCRIPT_TAG = HB_SUBSET_SETS_LAYOUT_SCRIPT_TAG             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_7 = __Pyx_PyInt_From_hb_subset_sets_t(HB_SUBSET_SETS_LAYOUT_SCRIPT_TAG); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_LAYOUT_SCRIPT_TAG, __pyx_t_7) < 0) __PYX_ERR(0, 2453, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2445
 *     return Face.from_ptr(new_face)
 * 
 * class SubsetInputSets(IntEnum):             # <<<<<<<<<<<<<<
 *     GLYPH_INDEX = HB_SUBSET_SETS_GLYPH_INDEX
 *     UNICODE = HB_SUBSET_SETS_UNICODE
 */
  __pyx_t_7 = __Pyx_Py3ClassCreate(__pyx_t_3, __pyx_n_s_SubsetInputSets, __pyx_t_4, __pyx_t_2, NULL, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2445, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SubsetInputSets, __pyx_t_7) < 0) __PYX_ERR(0, 2445, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2456
 * 
 * 
 * class SubsetFlags(IntFlag):             # <<<<<<<<<<<<<<
 *     DEFAULT = HB_SUBSET_FLAGS_DEFAULT
 *     NO_HINTING = HB_SUBSET_FLAGS_NO_HINTING
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_IntFlag); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(0, 2456, __pyx_L1_error);
  __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PEP560_update_bases(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_4, __pyx_n_s_SubsetFlags, __pyx_n_s_SubsetFlags, (PyObject *) NULL, __pyx_n_s_uharfbuzz__harfbuzz, (PyObject *) NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (__pyx_t_4 != __pyx_t_3) {
    if (unlikely((PyDict_SetItemString(__pyx_t_7, "__orig_bases__", __pyx_t_3) < 0))) __PYX_ERR(0, 2456, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2457
 * 
 * class SubsetFlags(IntFlag):
 *     DEFAULT = HB_SUBSET_FLAGS_DEFAULT             # <<<<<<<<<<<<<<
 *     NO_HINTING = HB_SUBSET_FLAGS_NO_HINTING
 *     RETAIN_GIDS = HB_SUBSET_FLAGS_RETAIN_GIDS
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_subset_flags_t(HB_SUBSET_FLAGS_DEFAULT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_DEFAULT, __pyx_t_3) < 0) __PYX_ERR(0, 2457, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2458
 * class SubsetFlags(IntFlag):
 *     DEFAULT = HB_SUBSET_FLAGS_DEFAULT
 *     NO_HINTING = HB_SUBSET_FLAGS_NO_HINTING             # <<<<<<<<<<<<<<
 *     RETAIN_GIDS = HB_SUBSET_FLAGS_RETAIN_GIDS
 *     DESUBROUTINIZE = HB_SUBSET_FLAGS_DESUBROUTINIZE
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_subset_flags_t(HB_SUBSET_FLAGS_NO_HINTING); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2458, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_NO_HINTING, __pyx_t_3) < 0) __PYX_ERR(0, 2458, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2459
 *     DEFAULT = HB_SUBSET_FLAGS_DEFAULT
 *     NO_HINTING = HB_SUBSET_FLAGS_NO_HINTING
 *     RETAIN_GIDS = HB_SUBSET_FLAGS_RETAIN_GIDS             # <<<<<<<<<<<<<<
 *     DESUBROUTINIZE = HB_SUBSET_FLAGS_DESUBROUTINIZE
 *     NAME_LEGACY = HB_SUBSET_FLAGS_NAME_LEGACY
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_subset_flags_t(HB_SUBSET_FLAGS_RETAIN_GIDS); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2459, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_RETAIN_GIDS, __pyx_t_3) < 0) __PYX_ERR(0, 2459, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2460
 *     NO_HINTING = HB_SUBSET_FLAGS_NO_HINTING
 *     RETAIN_GIDS = HB_SUBSET_FLAGS_RETAIN_GIDS
 *     DESUBROUTINIZE = HB_SUBSET_FLAGS_DESUBROUTINIZE             # <<<<<<<<<<<<<<
 *     NAME_LEGACY = HB_SUBSET_FLAGS_NAME_LEGACY
 *     SET_OVERLAPS_FLAG = HB_SUBSET_FLAGS_SET_OVERLAPS_FLAG
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_subset_flags_t(HB_SUBSET_FLAGS_DESUBROUTINIZE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_DESUBROUTINIZE, __pyx_t_3) < 0) __PYX_ERR(0, 2460, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2461
 *     RETAIN_GIDS = HB_SUBSET_FLAGS_RETAIN_GIDS
 *     DESUBROUTINIZE = HB_SUBSET_FLAGS_DESUBROUTINIZE
 *     NAME_LEGACY = HB_SUBSET_FLAGS_NAME_LEGACY             # <<<<<<<<<<<<<<
 *     SET_OVERLAPS_FLAG = HB_SUBSET_FLAGS_SET_OVERLAPS_FLAG
 *     PASSTHROUGH_UNRECOGNIZED = HB_SUBSET_FLAGS_PASSTHROUGH_UNRECOGNIZED
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_subset_flags_t(HB_SUBSET_FLAGS_NAME_LEGACY); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_NAME_LEGACY, __pyx_t_3) < 0) __PYX_ERR(0, 2461, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2462
 *     DESUBROUTINIZE = HB_SUBSET_FLAGS_DESUBROUTINIZE
 *     NAME_LEGACY = HB_SUBSET_FLAGS_NAME_LEGACY
 *     SET_OVERLAPS_FLAG = HB_SUBSET_FLAGS_SET_OVERLAPS_FLAG             # <<<<<<<<<<<<<<
 *     PASSTHROUGH_UNRECOGNIZED = HB_SUBSET_FLAGS_PASSTHROUGH_UNRECOGNIZED
 *     NOTDEF_OUTLINE = HB_SUBSET_FLAGS_NOTDEF_OUTLINE
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_subset_flags_t(HB_SUBSET_FLAGS_SET_OVERLAPS_FLAG); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_SET_OVERLAPS_FLAG, __pyx_t_3) < 0) __PYX_ERR(0, 2462, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2463
 *     NAME_LEGACY = HB_SUBSET_FLAGS_NAME_LEGACY
 *     SET_OVERLAPS_FLAG = HB_SUBSET_FLAGS_SET_OVERLAPS_FLAG
 *     PASSTHROUGH_UNRECOGNIZED = HB_SUBSET_FLAGS_PASSTHROUGH_UNRECOGNIZED             # <<<<<<<<<<<<<<
 *     NOTDEF_OUTLINE = HB_SUBSET_FLAGS_NOTDEF_OUTLINE
 *     GLYPH_NAMES = HB_SUBSET_FLAGS_GLYPH_NAMES
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_subset_flags_t(HB_SUBSET_FLAGS_PASSTHROUGH_UNRECOGNIZED); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2463, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_PASSTHROUGH_UNRECOGNIZED, __pyx_t_3) < 0) __PYX_ERR(0, 2463, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2464
 *     SET_OVERLAPS_FLAG = HB_SUBSET_FLAGS_SET_OVERLAPS_FLAG
 *     PASSTHROUGH_UNRECOGNIZED = HB_SUBSET_FLAGS_PASSTHROUGH_UNRECOGNIZED
 *     NOTDEF_OUTLINE = HB_SUBSET_FLAGS_NOTDEF_OUTLINE             # <<<<<<<<<<<<<<
 *     GLYPH_NAMES = HB_SUBSET_FLAGS_GLYPH_NAMES
 *     NO_PRUNE_UNICODE_RANGES = HB_SUBSET_FLAGS_NO_PRUNE_UNICODE_RANGES
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_subset_flags_t(HB_SUBSET_FLAGS_NOTDEF_OUTLINE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_NOTDEF_OUTLINE, __pyx_t_3) < 0) __PYX_ERR(0, 2464, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2465
 *     PASSTHROUGH_UNRECOGNIZED = HB_SUBSET_FLAGS_PASSTHROUGH_UNRECOGNIZED
 *     NOTDEF_OUTLINE = HB_SUBSET_FLAGS_NOTDEF_OUTLINE
 *     GLYPH_NAMES = HB_SUBSET_FLAGS_GLYPH_NAMES             # <<<<<<<<<<<<<<
 *     NO_PRUNE_UNICODE_RANGES = HB_SUBSET_FLAGS_NO_PRUNE_UNICODE_RANGES
 *     NO_LAYOUT_CLOSURE = HB_SUBSET_FLAGS_NO_LAYOUT_CLOSURE
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_subset_flags_t(HB_SUBSET_FLAGS_GLYPH_NAMES); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_GLYPH_NAMES, __pyx_t_3) < 0) __PYX_ERR(0, 2465, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2466
 *     NOTDEF_OUTLINE = HB_SUBSET_FLAGS_NOTDEF_OUTLINE
 *     GLYPH_NAMES = HB_SUBSET_FLAGS_GLYPH_NAMES
 *     NO_PRUNE_UNICODE_RANGES = HB_SUBSET_FLAGS_NO_PRUNE_UNICODE_RANGES             # <<<<<<<<<<<<<<
 *     NO_LAYOUT_CLOSURE = HB_SUBSET_FLAGS_NO_LAYOUT_CLOSURE
 * 
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_subset_flags_t(HB_SUBSET_FLAGS_NO_PRUNE_UNICODE_RANGES); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_NO_PRUNE_UNICODE_RANGES, __pyx_t_3) < 0) __PYX_ERR(0, 2466, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2467
 *     GLYPH_NAMES = HB_SUBSET_FLAGS_GLYPH_NAMES
 *     NO_PRUNE_UNICODE_RANGES = HB_SUBSET_FLAGS_NO_PRUNE_UNICODE_RANGES
 *     NO_LAYOUT_CLOSURE = HB_SUBSET_FLAGS_NO_LAYOUT_CLOSURE             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_PyInt_From_hb_subset_flags_t(HB_SUBSET_FLAGS_NO_LAYOUT_CLOSURE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_7, __pyx_n_s_NO_LAYOUT_CLOSURE, __pyx_t_3) < 0) __PYX_ERR(0, 2467, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2456
 * 
 * 
 * class SubsetFlags(IntFlag):             # <<<<<<<<<<<<<<
 *     DEFAULT = HB_SUBSET_FLAGS_DEFAULT
 *     NO_HINTING = HB_SUBSET_FLAGS_NO_HINTING
 */
  __pyx_t_3 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_SubsetFlags, __pyx_t_4, __pyx_t_7, NULL, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SubsetFlags, __pyx_t_3) < 0) __PYX_ERR(0, 2456, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2482
 *             hb_subset_input_destroy(self._hb_input)
 * 
 *     def subset(self, source: Face) -> Face:             # <<<<<<<<<<<<<<
 *         return subset(source, self)
 * 
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_source, __pyx_n_s_Face) < 0) __PYX_ERR(0, 2482, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_return, __pyx_n_s_Face) < 0) __PYX_ERR(0, 2482, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_11SubsetInput_5subset, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SubsetInput_subset, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__226)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2482, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetInput, __pyx_n_s_subset, __pyx_t_2) < 0) __PYX_ERR(0, 2482, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetInput);

  /* "uharfbuzz/_harfbuzz.pyx":2485
 *         return subset(source, self)
 * 
 *     def keep_everything(self):             # <<<<<<<<<<<<<<
 *         hb_subset_input_keep_everything(self._hb_input)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_11SubsetInput_7keep_everything, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SubsetInput_keep_everything, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__227)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetInput, __pyx_n_s_keep_everything, __pyx_t_2) < 0) __PYX_ERR(0, 2485, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetInput);

  /* "uharfbuzz/_harfbuzz.pyx":2488
 *         hb_subset_input_keep_everything(self._hb_input)
 * 
 *     def pin_axis_to_default(self, face: Face, tag: str) -> bool:             # <<<<<<<<<<<<<<
 *         hb_tag = hb_tag_from_string(tag.encode("ascii"), -1)
 *         return hb_subset_input_pin_axis_to_default(
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_face, __pyx_n_s_Face) < 0) __PYX_ERR(0, 2488, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_tag, __pyx_n_s_str) < 0) __PYX_ERR(0, 2488, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 2488, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_11SubsetInput_9pin_axis_to_default, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SubsetInput_pin_axis_to_default, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__229)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetInput, __pyx_n_s_pin_axis_to_default, __pyx_t_4) < 0) __PYX_ERR(0, 2488, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetInput);

  /* "uharfbuzz/_harfbuzz.pyx":2494
 *         )
 * 
 *     def pin_axis_location(self, face: Face, tag: str, value: float) -> bool:             # <<<<<<<<<<<<<<
 *         hb_tag = hb_tag_from_string(tag.encode("ascii"), -1)
 *         return hb_subset_input_pin_axis_location(
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_face, __pyx_n_s_Face) < 0) __PYX_ERR(0, 2494, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_tag, __pyx_n_s_str) < 0) __PYX_ERR(0, 2494, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_value, __pyx_n_s_float) < 0) __PYX_ERR(0, 2494, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 2494, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_11SubsetInput_11pin_axis_location, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SubsetInput_pin_axis_location, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__231)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetInput, __pyx_n_s_pin_axis_location, __pyx_t_2) < 0) __PYX_ERR(0, 2494, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetInput);

  /* "uharfbuzz/_harfbuzz.pyx":2508
 *         return Set.from_ptr(hb_set_reference (hb_subset_input_glyph_set(self._hb_input)))
 * 
 *     def sets(self, set_type : SubsetInputSets) -> Set:             # <<<<<<<<<<<<<<
 *         return Set.from_ptr(hb_set_reference (hb_subset_input_set(self._hb_input, set_type)))
 * 
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_set_type, __pyx_n_s_SubsetInputSets) < 0) __PYX_ERR(0, 2508, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_Set_2) < 0) __PYX_ERR(0, 2508, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_11SubsetInput_13sets, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SubsetInput_sets, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__233)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetInput, __pyx_n_s_sets, __pyx_t_4) < 0) __PYX_ERR(0, 2508, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetInput);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_11SubsetInput_15__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SubsetInput___reduce_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__234)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_11SubsetInput_17__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SubsetInput___setstate_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__235)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2557
 *             hb_subset_plan_destroy(self._hb_plan)
 * 
 *     def execute(self) -> Face:             # <<<<<<<<<<<<<<
 *         new_face = hb_subset_plan_execute_or_fail(self._hb_plan)
 *         if new_face == NULL:
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_return, __pyx_n_s_Face) < 0) __PYX_ERR(0, 2557, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_10SubsetPlan_5execute, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SubsetPlan_execute, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__237)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetPlan, __pyx_n_s_execute, __pyx_t_2) < 0) __PYX_ERR(0, 2557, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_SubsetPlan);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_10SubsetPlan_7__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SubsetPlan___reduce_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__238)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_10SubsetPlan_9__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SubsetPlan___setstate_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__239)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2579
 *     cdef hb_set_t* _hb_set
 * 
 *     INVALID_VALUE = HB_SET_VALUE_INVALID             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self, init = set()):
 */
  __pyx_t_2 = __Pyx_PyInt_From_hb_codepoint_t(HB_SET_VALUE_INVALID); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Set, __pyx_n_s_INVALID_VALUE, __pyx_t_2) < 0) __PYX_ERR(0, 2579, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Set);

  /* "uharfbuzz/_harfbuzz.pyx":2581
 *     INVALID_VALUE = HB_SET_VALUE_INVALID
 * 
 *     def __cinit__(self, init = set()):             # <<<<<<<<<<<<<<
 *         self._hb_set = hb_set_create()
 *         if not hb_set_allocation_successful(self._hb_set):
 */
  __pyx_t_2 = PySet_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2581, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k__18 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2599
 *         return wrapper
 * 
 *     def copy(self) -> Set:             # <<<<<<<<<<<<<<
 *         c = Set()
 *         c._hb_set = hb_set_copy(self._hb_set)
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_Set_2) < 0) __PYX_ERR(0, 2599, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_5copy, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Set_copy, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__241)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Set, __pyx_n_s_copy, __pyx_t_4) < 0) __PYX_ERR(0, 2599, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Set);

  /* "uharfbuzz/_harfbuzz.pyx":2604
 *         return c
 * 
 *     def __copy__(self) -> Set:             # <<<<<<<<<<<<<<
 *         return self.copy()
 * 
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2604, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_return, __pyx_n_s_Set_2) < 0) __PYX_ERR(0, 2604, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_7__copy__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Set___copy, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__242)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2604, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Set, __pyx_n_s_copy_2, __pyx_t_2) < 0) __PYX_ERR(0, 2604, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Set);

  /* "uharfbuzz/_harfbuzz.pyx":2607
 *         return self.copy()
 * 
 *     def clear(self):             # <<<<<<<<<<<<<<
 *         hb_set_clear(self._hb_set)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_9clear, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Set_clear, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__243)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2607, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Set, __pyx_n_s_clear, __pyx_t_2) < 0) __PYX_ERR(0, 2607, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Set);

  /* "uharfbuzz/_harfbuzz.pyx":2613
 *         return not hb_set_is_empty(self._hb_set)
 * 
 *     def invert(self):             # <<<<<<<<<<<<<<
 *         hb_set_invert(self._hb_set)
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_13invert, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Set_invert, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__244)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Set, __pyx_n_s_invert, __pyx_t_2) < 0) __PYX_ERR(0, 2613, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Set);

  /* "uharfbuzz/_harfbuzz.pyx":2616
 *         hb_set_invert(self._hb_set)
 * 
 *     def is_inverted(self) -> bool:             # <<<<<<<<<<<<<<
 *         return hb_set_is_inverted(self._hb_set)
 * 
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 2616, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_15is_inverted, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Set_is_inverted, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__245)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Set, __pyx_n_s_is_inverted, __pyx_t_4) < 0) __PYX_ERR(0, 2616, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Set);

  /* "uharfbuzz/_harfbuzz.pyx":2626
 *         return hb_set_has(self._hb_set, c)
 * 
 *     def add(self, c: int):             # <<<<<<<<<<<<<<
 *         hb_set_add(self._hb_set, c)
 *         if not hb_set_allocation_successful(self._hb_set):
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_c, __pyx_n_s_int) < 0) __PYX_ERR(0, 2626, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_19add, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Set_add, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__246)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2626, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Set, __pyx_n_s_add, __pyx_t_2) < 0) __PYX_ERR(0, 2626, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Set);

  /* "uharfbuzz/_harfbuzz.pyx":2631
 *             raise MemoryError()
 * 
 *     def add_range(self, first: int, last: int):             # <<<<<<<<<<<<<<
 *         hb_set_add_range(self._hb_set, first, last)
 *         if not hb_set_allocation_successful(self._hb_set):
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_first, __pyx_n_s_int) < 0) __PYX_ERR(0, 2631, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_last, __pyx_n_s_int) < 0) __PYX_ERR(0, 2631, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_21add_range, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Set_add_range, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__248)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2631, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Set, __pyx_n_s_add_range, __pyx_t_4) < 0) __PYX_ERR(0, 2631, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Set);

  /* "uharfbuzz/_harfbuzz.pyx":2636
 *             raise MemoryError()
 * 
 *     def remove(self, c: int):             # <<<<<<<<<<<<<<
 *         if not c in self:
 *             raise KeyError, c
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_c, __pyx_n_s_int) < 0) __PYX_ERR(0, 2636, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_23remove, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Set_remove, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__249)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Set, __pyx_n_s_remove, __pyx_t_2) < 0) __PYX_ERR(0, 2636, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Set);

  /* "uharfbuzz/_harfbuzz.pyx":2643
 *             raise MemoryError()
 * 
 *     def discard(self, c: int):             # <<<<<<<<<<<<<<
 *         hb_set_del(self._hb_set, c)
 *         if not hb_set_allocation_successful(self._hb_set):
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_c, __pyx_n_s_int) < 0) __PYX_ERR(0, 2643, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_25discard, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Set_discard, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__250)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Set, __pyx_n_s_discard, __pyx_t_4) < 0) __PYX_ERR(0, 2643, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Set);

  /* "uharfbuzz/_harfbuzz.pyx":2648
 *             raise MemoryError()
 * 
 *     def del_range(self, first: int, last: int):             # <<<<<<<<<<<<<<
 *         hb_set_del_range(self._hb_set, first, last)
 *         if not hb_set_allocation_successful(self._hb_set):
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_first, __pyx_n_s_int) < 0) __PYX_ERR(0, 2648, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_last, __pyx_n_s_int) < 0) __PYX_ERR(0, 2648, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_27del_range, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Set_del_range, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__251)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Set, __pyx_n_s_del_range, __pyx_t_2) < 0) __PYX_ERR(0, 2648, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Set);

  /* "uharfbuzz/_harfbuzz.pyx":2653
 *             raise MemoryError()
 * 
 *     def _is_equal(self, other: Set) -> bool:             # <<<<<<<<<<<<<<
 *         return hb_set_is_equal(self._hb_set, other._hb_set)
 * 
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_other, __pyx_n_s_Set_2) < 0) __PYX_ERR(0, 2653, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 2653, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_29_is_equal, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Set__is_equal, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__253)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Set, __pyx_n_s_is_equal, __pyx_t_4) < 0) __PYX_ERR(0, 2653, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Set);

  /* "uharfbuzz/_harfbuzz.pyx":2661
 *         return self._is_equal(other)
 * 
 *     def issubset(self, larger_set: Set) -> bool:             # <<<<<<<<<<<<<<
 *         return hb_set_is_subset(self._hb_set, larger_set._hb_set)
 * 
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_larger_set, __pyx_n_s_Set_2) < 0) __PYX_ERR(0, 2661, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 2661, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_33issubset, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Set_issubset, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__255)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Set, __pyx_n_s_issubset, __pyx_t_2) < 0) __PYX_ERR(0, 2661, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Set);

  /* "uharfbuzz/_harfbuzz.pyx":2664
 *         return hb_set_is_subset(self._hb_set, larger_set._hb_set)
 * 
 *     def issuperset(self, smaller_set: Set) -> bool:             # <<<<<<<<<<<<<<
 *         return hb_set_is_subset(smaller_set._hb_set, self._hb_set)
 * 
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_smaller_set, __pyx_n_s_Set_2) < 0) __PYX_ERR(0, 2664, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 2664, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_35issuperset, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Set_issuperset, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__257)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Set, __pyx_n_s_issuperset, __pyx_t_4) < 0) __PYX_ERR(0, 2664, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Set);

  /* "uharfbuzz/_harfbuzz.pyx":2667
 *         return hb_set_is_subset(smaller_set._hb_set, self._hb_set)
 * 
 *     def _set(self, other: Set):             # <<<<<<<<<<<<<<
 *         hb_set_set(self._hb_set, other._hb_set)
 * 
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_other, __pyx_n_s_Set_2) < 0) __PYX_ERR(0, 2667, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_37_set, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Set__set, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__258)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Set, __pyx_n_s_set_2, __pyx_t_2) < 0) __PYX_ERR(0, 2667, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Set);

  /* "uharfbuzz/_harfbuzz.pyx":2670
 *         hb_set_set(self._hb_set, other._hb_set)
 * 
 *     def set(self, other):             # <<<<<<<<<<<<<<
 *         if type(other) == Set:
 *             self._set(other)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_39set, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Set_set, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__260)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2670, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Set, __pyx_n_s_set, __pyx_t_2) < 0) __PYX_ERR(0, 2670, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Set);

  /* "uharfbuzz/_harfbuzz.pyx":2680
 *             raise MemoryError()
 * 
 *     def _update(self, other: Set):             # <<<<<<<<<<<<<<
 *         hb_set_union(self._hb_set, other._hb_set)
 * 
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_other, __pyx_n_s_Set_2) < 0) __PYX_ERR(0, 2680, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_41_update, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Set__update, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__261)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Set, __pyx_n_s_update, __pyx_t_4) < 0) __PYX_ERR(0, 2680, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Set);

  /* "uharfbuzz/_harfbuzz.pyx":2683
 *         hb_set_union(self._hb_set, other._hb_set)
 * 
 *     def update(self, other):             # <<<<<<<<<<<<<<
 *         if type(other) == Set:
 *             self._update(other)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_43update, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Set_update, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__262)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2683, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Set, __pyx_n_s_update_2, __pyx_t_4) < 0) __PYX_ERR(0, 2683, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Set);

  /* "uharfbuzz/_harfbuzz.pyx":2697
 *         return self
 * 
 *     def intersection_update(self, other: Set):             # <<<<<<<<<<<<<<
 *         hb_set_intersect(self._hb_set, other._hb_set)
 *         if not hb_set_allocation_successful(self._hb_set):
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_other, __pyx_n_s_Set_2) < 0) __PYX_ERR(0, 2697, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_47intersection_update, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Set_intersection_update, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__263)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Set, __pyx_n_s_intersection_update, __pyx_t_2) < 0) __PYX_ERR(0, 2697, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Set);

  /* "uharfbuzz/_harfbuzz.pyx":2706
 *         return self
 * 
 *     def difference_update(self, other: Set):             # <<<<<<<<<<<<<<
 *         hb_set_subtract(self._hb_set, other._hb_set)
 *         if not hb_set_allocation_successful(self._hb_set):
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_other, __pyx_n_s_Set_2) < 0) __PYX_ERR(0, 2706, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_51difference_update, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Set_difference_update, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__264)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Set, __pyx_n_s_difference_update, __pyx_t_4) < 0) __PYX_ERR(0, 2706, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Set);

  /* "uharfbuzz/_harfbuzz.pyx":2715
 *         return self
 * 
 *     def symmetric_difference_update(self, other: Set):             # <<<<<<<<<<<<<<
 *         hb_set_symmetric_difference(self._hb_set, other._hb_set)
 *         if not hb_set_allocation_successful(self._hb_set):
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2715, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_other, __pyx_n_s_Set_2) < 0) __PYX_ERR(0, 2715, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_55symmetric_difference_update, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Set_symmetric_difference_update, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__265)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2715, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Set, __pyx_n_s_symmetric_difference_update, __pyx_t_2) < 0) __PYX_ERR(0, 2715, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Set);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_65__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Set___reduce_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__266)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Set_67__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Set___setstate_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__267)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_7SetIter_7__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SetIter___reduce_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__268)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_7SetIter_9__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_SetIter___setstate_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__269)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2767
 *     cdef hb_map_t* _hb_map
 * 
 *     INVALID_VALUE = HB_MAP_VALUE_INVALID             # <<<<<<<<<<<<<<
 * 
 *     def __cinit__(self, init = dict()):
 */
  __pyx_t_2 = __Pyx_PyInt_From_hb_codepoint_t(HB_MAP_VALUE_INVALID); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2767, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Map, __pyx_n_s_INVALID_VALUE, __pyx_t_2) < 0) __PYX_ERR(0, 2767, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Map);

  /* "uharfbuzz/_harfbuzz.pyx":2769
 *     INVALID_VALUE = HB_MAP_VALUE_INVALID
 * 
 *     def __cinit__(self, init = dict()):             # <<<<<<<<<<<<<<
 *         self._hb_map = hb_map_create()
 *         if not hb_map_allocation_successful(self._hb_map):
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2769, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k__20 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":2787
 *         return wrapper
 * 
 *     def copy(self) -> Map:             # <<<<<<<<<<<<<<
 *         c = Map()
 *         c._hb_map = hb_map_copy(self._hb_map)
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_Map) < 0) __PYX_ERR(0, 2787, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Map_5copy, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Map_copy, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__270)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Map, __pyx_n_s_copy, __pyx_t_4) < 0) __PYX_ERR(0, 2787, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Map);

  /* "uharfbuzz/_harfbuzz.pyx":2792
 *         return c
 * 
 *     def __copy__(self) -> Map:             # <<<<<<<<<<<<<<
 *         return self.copy()
 * 
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_return, __pyx_n_s_Map) < 0) __PYX_ERR(0, 2792, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Map_7__copy__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Map___copy, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__271)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2792, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Map, __pyx_n_s_copy_2, __pyx_t_2) < 0) __PYX_ERR(0, 2792, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Map);

  /* "uharfbuzz/_harfbuzz.pyx":2795
 *         return self.copy()
 * 
 *     def _update(self, other : Map):             # <<<<<<<<<<<<<<
 *         hb_map_update(self._hb_map, other._hb_map)
 * 
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_other, __pyx_n_s_Map) < 0) __PYX_ERR(0, 2795, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Map_9_update, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Map__update, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__272)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Map, __pyx_n_s_update, __pyx_t_4) < 0) __PYX_ERR(0, 2795, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Map);

  /* "uharfbuzz/_harfbuzz.pyx":2798
 *         hb_map_update(self._hb_map, other._hb_map)
 * 
 *     def update(self, other):             # <<<<<<<<<<<<<<
 *         if type(other) == Map:
 *             self._update(other)
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Map_11update, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Map_update, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__274)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2798, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Map, __pyx_n_s_update_2, __pyx_t_4) < 0) __PYX_ERR(0, 2798, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Map);

  /* "uharfbuzz/_harfbuzz.pyx":2808
 *             raise MemoryError()
 * 
 *     def clear(self):             # <<<<<<<<<<<<<<
 *         hb_map_clear(self._hb_map)
 * 
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Map_13clear, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Map_clear, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__275)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Map, __pyx_n_s_clear, __pyx_t_4) < 0) __PYX_ERR(0, 2808, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Map);

  /* "uharfbuzz/_harfbuzz.pyx":2817
 *         return hb_map_get_population(self._hb_map)
 * 
 *     def _is_equal(self, other: Map) -> bool:             # <<<<<<<<<<<<<<
 *         return hb_map_is_equal(self._hb_map, other._hb_map)
 * 
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_other, __pyx_n_s_Map) < 0) __PYX_ERR(0, 2817, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 2817, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Map_19_is_equal, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Map__is_equal, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__276)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Map, __pyx_n_s_is_equal, __pyx_t_2) < 0) __PYX_ERR(0, 2817, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Map);

  /* "uharfbuzz/_harfbuzz.pyx":2830
 *             raise MemoryError()
 * 
 *     def get(self, k: int):             # <<<<<<<<<<<<<<
 *         if k < 0 or k >= self.INVALID_VALUE:
 *             return None
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_k, __pyx_n_s_int) < 0) __PYX_ERR(0, 2830, __pyx_L1_error)
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Map_25get, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Map_get, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__278)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2830, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_4, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Map, __pyx_n_s_get, __pyx_t_4) < 0) __PYX_ERR(0, 2830, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Map);

  /* "uharfbuzz/_harfbuzz.pyx":2856
 *         hb_map_del(self._hb_map, c)
 * 
 *     def items(self):             # <<<<<<<<<<<<<<
 *         return MapIter(self)
 * 
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Map_33items, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Map_items, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__279)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Map, __pyx_n_s_items, __pyx_t_4) < 0) __PYX_ERR(0, 2856, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Map);

  /* "uharfbuzz/_harfbuzz.pyx":2859
 *         return MapIter(self)
 * 
 *     def keys(self):             # <<<<<<<<<<<<<<
 *         return (k for k,v in self.items())
 * 
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Map_35keys, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Map_keys, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__281)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Map, __pyx_n_s_keys, __pyx_t_4) < 0) __PYX_ERR(0, 2859, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Map);

  /* "uharfbuzz/_harfbuzz.pyx":2862
 *         return (k for k,v in self.items())
 * 
 *     def values(self):             # <<<<<<<<<<<<<<
 *         return (v for k,v in self.items())
 * 
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Map_37values, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Map_values, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__282)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_9uharfbuzz_9_harfbuzz_Map, __pyx_n_s_values, __pyx_t_4) < 0) __PYX_ERR(0, 2862, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  PyType_Modified(__pyx_ptype_9uharfbuzz_9_harfbuzz_Map);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Map_43__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Map___reduce_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__283)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_3Map_45__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_Map___setstate_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__284)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_7MapIter_7__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_MapIter___reduce_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__285)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_4) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "no default __reduce__ due to non-trivial __cinit__"
 */
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_9uharfbuzz_9_harfbuzz_7MapIter_9__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_MapIter___setstate_cython, NULL, __pyx_n_s_uharfbuzz__harfbuzz, __pyx_d, ((PyObject *)__pyx_codeobj__286)); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_4) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "uharfbuzz/_harfbuzz.pyx":1
 * #cython: language_level=3             # <<<<<<<<<<<<<<
 * cimport cython
 * import os
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_4) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  if (__pyx_m) {
    if (__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init uharfbuzz._harfbuzz", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init uharfbuzz._harfbuzz");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError */
#if __PYX_LIMITED_VERSION_HEX < 0x030d00A1
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* decode_c_string */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    Py_ssize_t length;
    if (unlikely((start < 0) | (stop < 0))) {
        size_t slen = strlen(cstring);
        if (unlikely(slen > (size_t) PY_SSIZE_T_MAX)) {
            PyErr_SetString(PyExc_OverflowError,
                            "c-string too long to convert to Python");
            return NULL;
        }
        length = (Py_ssize_t) slen;
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    if (unlikely(stop <= start))
        return __Pyx_NewRef(__pyx_empty_unicode);
    length = stop - start;
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL && !CYTHON_VECTORCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object"))) {
        return NULL;
    }
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object"))) {
        return NULL;
    }
    #endif
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    #endif
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    #endif
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) < 0) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    #if PY_VERSION_HEX < 0x030800B1
    #if CYTHON_FAST_PYCCALL
    if (PyCFunction_Check(func)) {
        if (kwargs) {
            return _PyCFunction_FastCallDict(func, args, nargs, kwargs);
        } else {
            return _PyCFunction_FastCallKeywords(func, args, nargs, NULL);
        }
    }
    #if PY_VERSION_HEX >= 0x030700A1
    if (!kwargs && __Pyx_IS_TYPE(func, &PyMethodDescr_Type)) {
        return _PyMethodDescr_FastCallKeywords(func, args, nargs, NULL);
    }
    #endif
    #endif
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs);
    }
    #endif
    #endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL
        #if PY_VERSION_HEX < 0x03090000
        vectorcallfunc f = _PyVectorcall_Function(func);
        #else
        vectorcallfunc f = PyVectorcall_Function(func);
        #endif
        if (f) {
            return f(func, args, (size_t)nargs, NULL);
        }
        #elif defined(__Pyx_CyFunction_USED) && CYTHON_BACKPORT_VECTORCALL
        if (__Pyx_CyFunction_CheckExact(func)) {
            __pyx_vectorcallfunc f = __Pyx_CyFunction_func_vectorcall(func);
            if (f) return f(func, args, (size_t)nargs, NULL);
        }
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
    else state = (PyGILState_STATE)0;
#endif
    CYTHON_UNUSED_VAR(clineno);
    CYTHON_UNUSED_VAR(lineno);
    CYTHON_UNUSED_VAR(filename);
    CYTHON_MAYBE_UNUSED_VAR(nogil);
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(0);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* TupleAndListFromArray */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        Py_INCREF(__pyx_empty_tuple);
        return __pyx_empty_tuple;
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = PyTuple_GET_SIZE(kwnames);
    for (i = 0; i < n; i++)
    {
        if (s == PyTuple_GET_ITEM(kwnames, i)) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        int eq = __Pyx_PyUnicode_Equals(s, PyTuple_GET_ITEM(kwnames, i), Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;
            return kwvalues[i];
        }
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues) {
    Py_ssize_t i, nkwargs = PyTuple_GET_SIZE(kwnames);
    PyObject *dict;
    dict = PyDict_New();
    if (unlikely(!dict))
        return NULL;
    for (i=0; i<nkwargs; i++) {
        PyObject *key = PyTuple_GET_ITEM(kwnames, i);
        if (unlikely(PyDict_SetItem(dict, key, kwvalues[i]) < 0))
            goto bad;
    }
    return dict;
bad:
    Py_DECREF(dict);
    return NULL;
}
#endif
#endif

/* decode_c_bytes */
static CYTHON_INLINE PyObject* __Pyx_decode_c_bytes(
         const char* cstring, Py_ssize_t length, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    if (unlikely((start < 0) | (stop < 0))) {
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    if (stop > length)
        stop = length;
    if (unlikely(stop <= start))
        return __Pyx_NewRef(__pyx_empty_unicode);
    length = stop - start;
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && PY_VERSION_HEX < 0x030d0000
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#elif CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kw,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kw, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kw))) {
        Py_ssize_t kwsize;
#if CYTHON_ASSUME_SAFE_MACROS
        kwsize = PyTuple_GET_SIZE(kw);
#else
        kwsize = PyTuple_Size(kw);
        if (kwsize < 0) return 0;
#endif
        if (unlikely(kwsize == 0))
            return 1;
        if (!kw_allowed) {
#if CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kw, 0);
#else
            key = PyTuple_GetItem(kw, pos);
            if (!key) return 0;
#endif
            goto invalid_keyword;
        }
#if PY_VERSION_HEX < 0x03090000
        for (pos = 0; pos < kwsize; pos++) {
#if CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kw, pos);
#else
            key = PyTuple_GetItem(kw, pos);
            if (!key) return 0;
#endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
        }
#endif
        return 1;
    }
    while (PyDict_Next(kw, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if (!kw_allowed && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    __Pyx_PyThreadState_declare
    CYTHON_UNUSED_VAR(cause);
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
      #if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
      #elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    int kwds_is_tuple = CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds));
    while (1) {
        Py_XDECREF(key); key = NULL;
        Py_XDECREF(value); value = NULL;
        if (kwds_is_tuple) {
            Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(kwds);
#else
            size = PyTuple_Size(kwds);
            if (size < 0) goto bad;
#endif
            if (pos >= size) break;
#if CYTHON_AVOID_BORROWED_REFS
            key = __Pyx_PySequence_ITEM(kwds, pos);
            if (!key) goto bad;
#elif CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kwds, pos);
#else
            key = PyTuple_GetItem(kwds, pos);
            if (!key) goto bad;
#endif
            value = kwvalues[pos];
            pos++;
        }
        else
        {
            if (!PyDict_Next(kwds, &pos, &key, &value)) break;
#if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(key);
#endif
        }
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(value);
            Py_DECREF(key);
#endif
            key = NULL;
            value = NULL;
            continue;
        }
#if !CYTHON_AVOID_BORROWED_REFS
        Py_INCREF(key);
#endif
        Py_INCREF(value);
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
                    value = NULL;
#endif
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key)
                );
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
                    value = NULL;
#endif
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    Py_XDECREF(key);
    Py_XDECREF(value);
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    Py_XDECREF(key);
    Py_XDECREF(value);
    return -1;
}

/* PyObjectCallNoArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg[2] = {NULL, NULL};
    return __Pyx_PyObject_FastCall(func, arg + 1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    __Pyx_TypeName obj_type_name;
    __Pyx_TypeName type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    type_name = __Pyx_PyType_GetName(type);
    PyErr_Format(PyExc_TypeError,
                 "Cannot convert " __Pyx_FMT_TYPENAME " to " __Pyx_FMT_TYPENAME,
                 obj_type_name, type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    __Pyx_TypeName type_name;
    __Pyx_TypeName obj_type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    type_name = __Pyx_PyType_GetName(type);
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected " __Pyx_FMT_TYPENAME
        ", got " __Pyx_FMT_TYPENAME ")", name, type_name, obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (mm && mm->mp_subscript) {
            PyObject *r, *key = PyInt_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || !PyMapping_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* PyObjectCallOneArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* RaiseUnexpectedTypeError */
static int
__Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj)
{
    __Pyx_TypeName obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError, "Expected %s, got " __Pyx_FMT_TYPENAME,
                 expected, obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
    PyObject* exc_type;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (unlikely(exc_type)) {
        if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return -1;
        __Pyx_PyErr_Clear();
        return 0;
    }
    return 0;
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#elif PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = __Pyx_PyObject_GetIterNextFunc(iter);
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
#if CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
#include <string.h>
#endif
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
        
            x = (long)((unsigned long)a + (unsigned long)b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        if (unlikely(__Pyx_PyLong_IsZero(op1))) {
            return __Pyx_NewRef(op2);
        }
        if (likely(__Pyx_PyLong_IsCompact(op1))) {
            a = __Pyx_PyLong_CompactValue(op1);
        } else {
            const digit* digits = __Pyx_PyLong_Digits(op1);
            const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op1);
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double a = __pyx_PyFloat_AsDouble(op1);
#else
        double a = PyFloat_AS_DOUBLE(op1);
#endif
            double result;
            
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type = NULL, *local_value, *local_tb = NULL;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if PY_VERSION_HEX >= 0x030C00A6
    local_value = tstate->current_exception;
    tstate->current_exception = 0;
    if (likely(local_value)) {
        local_type = (PyObject*) Py_TYPE(local_value);
        Py_INCREF(local_type);
        local_tb = PyException_GetTraceback(local_value);
    }
  #else
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
  #endif
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE && PY_VERSION_HEX >= 0x030C00A6
    if (unlikely(tstate->current_exception))
#elif CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
      #if PY_VERSION_HEX >= 0x030B00a4
        tmp_value = exc_info->exc_value;
        exc_info->exc_value = local_value;
        tmp_type = NULL;
        tmp_tb = NULL;
        Py_XDECREF(local_type);
        Py_XDECREF(local_tb);
      #else
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
      #endif
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_value = exc_info->exc_value;
    exc_info->exc_value = *value;
    if (tmp_value == NULL || tmp_value == Py_None) {
        Py_XDECREF(tmp_value);
        tmp_value = NULL;
        tmp_type = NULL;
        tmp_tb = NULL;
    } else {
        tmp_type = (PyObject*) Py_TYPE(tmp_value);
        Py_INCREF(tmp_type);
        #if CYTHON_COMPILING_IN_CPYTHON
        tmp_tb = ((PyBaseExceptionObject*) tmp_value)->traceback;
        Py_XINCREF(tmp_tb);
        #else
        tmp_tb = PyException_GetTraceback(tmp_value);
        #endif
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
  #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
  #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_value == NULL || exc_info->exc_value == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    PyObject *exc_value = exc_info->exc_value;
    if (exc_value == NULL || exc_value == Py_None) {
        *value = NULL;
        *type = NULL;
        *tb = NULL;
    } else {
        *value = exc_value;
        Py_INCREF(*value);
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        *tb = PyException_GetTraceback(exc_value);
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #endif
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    PyObject *tmp_value = exc_info->exc_value;
    exc_info->exc_value = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
  #else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
  #endif
}
#endif

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_MultiplyObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        
#ifdef HAVE_LONG_LONG
            if (sizeof(PY_LONG_LONG) > sizeof(long)) {
                PY_LONG_LONG result = (PY_LONG_LONG)a * (PY_LONG_LONG)b;
                return (result >= LONG_MIN && result <= LONG_MAX) ?
                    PyInt_FromLong((long)result) : PyLong_FromLongLong(result);
            }
#endif
#if CYTHON_USE_TYPE_SLOTS
            return PyInt_Type.tp_as_number->nb_multiply(op1, op2);
#else
            return PyNumber_Multiply(op1, op2);
#endif
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        if (unlikely(__Pyx_PyLong_IsZero(op1))) {
            return __Pyx_NewRef(op1);
        }
        if (likely(__Pyx_PyLong_IsCompact(op1))) {
            a = __Pyx_PyLong_CompactValue(op1);
        } else {
            const digit* digits = __Pyx_PyLong_Digits(op1);
            const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op1);
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT+30) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT+30) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT+30) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT+30) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT+30) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT+30) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT+30) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT+30) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT+30) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT+30) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT+30) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT+30) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_multiply(op1, op2);
            }
        }
                CYTHON_UNUSED_VAR(a);
                CYTHON_UNUSED_VAR(b);
                #ifdef HAVE_LONG_LONG
                lla = a;
                goto long_long;
                #else
                return PyLong_Type.tp_as_number->nb_multiply(op1, op2);
                #endif
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla * llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double a = __pyx_PyFloat_AsDouble(op1);
#else
        double a = PyFloat_AS_DOUBLE(op1);
#endif
            double result;
            
            PyFPE_START_PROTECT("multiply", return NULL)
            result = ((double)a) * (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceMultiply : PyNumber_Multiply)(op1, op2);
}
#endif

/* SliceTupleAndList */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_crop_slice(Py_ssize_t* _start, Py_ssize_t* _stop, Py_ssize_t* _length) {
    Py_ssize_t start = *_start, stop = *_stop, length = *_length;
    if (start < 0) {
        start += length;
        if (start < 0)
            start = 0;
    }
    if (stop < 0)
        stop += length;
    else if (stop > length)
        stop = length;
    *_length = stop - start;
    *_start = start;
    *_stop = stop;
}
static CYTHON_INLINE PyObject* __Pyx_PyList_GetSlice(
            PyObject* src, Py_ssize_t start, Py_ssize_t stop) {
    Py_ssize_t length = PyList_GET_SIZE(src);
    __Pyx_crop_slice(&start, &stop, &length);
    if (length <= 0) {
        return PyList_New(0);
    }
    return __Pyx_PyList_FromArray(((PyListObject*)src)->ob_item + start, length);
}
static CYTHON_INLINE PyObject* __Pyx_PyTuple_GetSlice(
            PyObject* src, Py_ssize_t start, Py_ssize_t stop) {
    Py_ssize_t length = PyTuple_GET_SIZE(src);
    __Pyx_crop_slice(&start, &stop, &length);
    return __Pyx_PyTuple_FromArray(((PyTupleObject*)src)->ob_item + start, length);
}
#endif

/* RaiseUnboundLocalError */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind, kind_shift;
    Py_ssize_t i, char_pos;
    void *result_udata;
    CYTHON_MAYBE_UNUSED_VAR(max_char);
#if CYTHON_PEP393_ENABLED
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    kind_shift = (result_ukind == PyUnicode_4BYTE_KIND) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_DATA(result_uval);
#else
    result_uval = PyUnicode_FromUnicode(NULL, result_ulength);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = sizeof(Py_UNICODE);
    kind_shift = (result_ukind == 4) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_AS_UNICODE(result_uval);
#endif
    assert(kind_shift == 2 || kind_shift == 1 || kind_shift == 0);
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = PyTuple_GET_ITEM(value_tuple, i);
        if (unlikely(__Pyx_PyUnicode_READY(uval)))
            goto bad;
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        if (unlikely(!ulength))
            continue;
        if (unlikely((PY_SSIZE_T_MAX >> kind_shift) - ulength < char_pos))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (!CYTHON_PEP393_ENABLED || ukind == result_ukind) {
            memcpy((char *)result_udata + (char_pos << kind_shift), udata, (size_t) (ulength << kind_shift));
        } else {
            #if PY_VERSION_HEX >= 0x030d0000
            if (unlikely(PyUnicode_CopyCharacters(result_uval, char_pos, uval, 0, ulength) < 0)) goto bad;
            #elif CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030300F0 || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    CYTHON_UNUSED_VAR(max_char);
    CYTHON_UNUSED_VAR(result_ulength);
    CYTHON_UNUSED_VAR(value_count);
    return PyUnicode_Join(__pyx_empty_unicode, value_tuple);
#endif
}

/* CIntToDigits */
static const char DIGIT_PAIRS_10[2*10*10+1] = {
    "00010203040506070809"
    "10111213141516171819"
    "20212223242526272829"
    "30313233343536373839"
    "40414243444546474849"
    "50515253545556575859"
    "60616263646566676869"
    "70717273747576777879"
    "80818283848586878889"
    "90919293949596979899"
};
static const char DIGIT_PAIRS_8[2*8*8+1] = {
    "0001020304050607"
    "1011121314151617"
    "2021222324252627"
    "3031323334353637"
    "4041424344454647"
    "5051525354555657"
    "6061626364656667"
    "7071727374757677"
};
static const char DIGITS_HEX[2*16+1] = {
    "0123456789abcdef"
    "0123456789ABCDEF"
};

/* BuildPyUnicode */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char) {
    PyObject *uval;
    Py_ssize_t uoffset = ulength - clength;
#if CYTHON_USE_UNICODE_INTERNALS
    Py_ssize_t i;
#if CYTHON_PEP393_ENABLED
    void *udata;
    uval = PyUnicode_New(ulength, 127);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_DATA(uval);
#else
    Py_UNICODE *udata;
    uval = PyUnicode_FromUnicode(NULL, ulength);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_AS_UNICODE(uval);
#endif
    if (uoffset > 0) {
        i = 0;
        if (prepend_sign) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, 0, '-');
            i++;
        }
        for (; i < uoffset; i++) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, i, padding_char);
        }
    }
    for (i=0; i < clength; i++) {
        __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, uoffset+i, chars[i]);
    }
#else
    {
        PyObject *sign = NULL, *padding = NULL;
        uval = NULL;
        if (uoffset > 0) {
            prepend_sign = !!prepend_sign;
            if (uoffset > prepend_sign) {
                padding = PyUnicode_FromOrdinal(padding_char);
                if (likely(padding) && uoffset > prepend_sign + 1) {
                    PyObject *tmp;
                    PyObject *repeat = PyInt_FromSsize_t(uoffset - prepend_sign);
                    if (unlikely(!repeat)) goto done_or_error;
                    tmp = PyNumber_Multiply(padding, repeat);
                    Py_DECREF(repeat);
                    Py_DECREF(padding);
                    padding = tmp;
                }
                if (unlikely(!padding)) goto done_or_error;
            }
            if (prepend_sign) {
                sign = PyUnicode_FromOrdinal('-');
                if (unlikely(!sign)) goto done_or_error;
            }
        }
        uval = PyUnicode_DecodeASCII(chars, clength, NULL);
        if (likely(uval) && padding) {
            PyObject *tmp = PyNumber_Add(padding, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
        if (likely(uval) && sign) {
            PyObject *tmp = PyNumber_Add(sign, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
done_or_error:
        Py_XDECREF(padding);
        Py_XDECREF(sign);
    }
#endif
    return uval;
}

/* CIntToPyUnicode */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(Py_ssize_t)*3+2];
    char *dpos, *end = digits + sizeof(Py_ssize_t)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    Py_ssize_t remaining;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const Py_ssize_t neg_one = (Py_ssize_t) -1, const_zero = (Py_ssize_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (Py_ssize_t) (remaining / (8*8));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_8 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (Py_ssize_t) (remaining / (10*10));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_10 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (Py_ssize_t) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    assert(!last_one_off || *dpos == '0');
    dpos += last_one_off;
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* PyObjectCall2Args */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args[3] = {NULL, arg1, arg2};
    return __Pyx_PyObject_FastCall(function, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectCallMethod1 */
#if !(CYTHON_VECTORCALL && __PYX_LIMITED_VERSION_HEX >= 0x030C00A2)
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
#endif
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
#if CYTHON_VECTORCALL && __PYX_LIMITED_VERSION_HEX >= 0x030C00A2
    PyObject *args[2] = {obj, arg};
    (void) __Pyx_PyObject_GetMethod;
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_Call2Args;
    return PyObject_VectorcallMethod(method_name, args, 2 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
#endif
}

/* StringJoin */
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values) {
    (void) __Pyx_PyObject_CallMethod1;
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION < 3
    return _PyString_Join(sep, values);
#elif CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
    return _PyBytes_Join(sep, values);
#else
    return __Pyx_PyObject_CallMethod1(sep, __pyx_n_s_join, values);
#endif
}

/* pep479 */
static void __Pyx_Generator_Replace_StopIteration(int in_async_gen) {
    PyObject *exc, *val, *tb, *cur_exc;
    __Pyx_PyThreadState_declare
    #ifdef __Pyx_StopAsyncIteration_USED
    int is_async_stopiteration = 0;
    #endif
    CYTHON_MAYBE_UNUSED_VAR(in_async_gen);
    cur_exc = PyErr_Occurred();
    if (likely(!__Pyx_PyErr_GivenExceptionMatches(cur_exc, PyExc_StopIteration))) {
        #ifdef __Pyx_StopAsyncIteration_USED
        if (in_async_gen && unlikely(__Pyx_PyErr_GivenExceptionMatches(cur_exc, __Pyx_PyExc_StopAsyncIteration))) {
            is_async_stopiteration = 1;
        } else
        #endif
            return;
    }
    __Pyx_PyThreadState_assign
    __Pyx_GetException(&exc, &val, &tb);
    Py_XDECREF(exc);
    Py_XDECREF(val);
    Py_XDECREF(tb);
    PyErr_SetString(PyExc_RuntimeError,
        #ifdef __Pyx_StopAsyncIteration_USED
        is_async_stopiteration ? "async generator raised StopAsyncIteration" :
        in_async_gen ? "async generator raised StopIteration" :
        #endif
        "generator raised StopIteration");
}

/* PyObjectFormatAndDecref */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f) {
    if (unlikely(!s)) return NULL;
    if (likely(PyUnicode_CheckExact(s))) return s;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyString_CheckExact(s))) {
        PyObject *result = PyUnicode_FromEncodedObject(s, NULL, "strict");
        Py_DECREF(s);
        return result;
    }
    #endif
    return __Pyx_PyObject_FormatAndDecref(s, f);
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f) {
    PyObject *result;
    if (unlikely(!s)) return NULL;
    result = PyObject_Format(s, f);
    Py_DECREF(s);
    return result;
}

/* FixUpExtensionType */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if PY_VERSION_HEX > 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
#else
    const PyType_Slot *slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
        int changed = 0;
#if !(PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON)
        const
#endif
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
#if PY_VERSION_HEX < 0x030900b1
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
#if PY_VERSION_HEX >= 0x030800b4
                    type->tp_vectorcall_offset = memb->offset;
#else
                    type->tp_print = (printfunc) memb->offset;
#endif
                    changed = 1;
                }
#endif
#else
                if ((0));
#endif
#if PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    if (unlikely(PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr) < 0)) {
                        Py_DECREF(descr);
                        return -1;
                    }
                    Py_DECREF(descr);
                    changed = 1;
                }
#endif
            }
            memb++;
        }
        if (changed)
            PyType_Modified(type);
    }
#endif
    return 0;
}
#endif

/* ValidateBasesTuple */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases) {
    Py_ssize_t i, n;
#if CYTHON_ASSUME_SAFE_MACROS
    n = PyTuple_GET_SIZE(bases);
#else
    n = PyTuple_Size(bases);
    if (n < 0) return -1;
#endif
    for (i = 1; i < n; i++)
    {
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *b0 = PySequence_GetItem(bases, i);
        if (!b0) return -1;
#elif CYTHON_ASSUME_SAFE_MACROS
        PyObject *b0 = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *b0 = PyTuple_GetItem(bases, i);
        if (!b0) return -1;
#endif
        PyTypeObject *b;
#if PY_MAJOR_VERSION < 3
        if (PyClass_Check(b0))
        {
            PyErr_Format(PyExc_TypeError, "base class '%.200s' is an old-style class",
                         PyString_AS_STRING(((PyClassObject*)b0)->cl_name));
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
#endif
        b = (PyTypeObject*) b0;
        if (!__Pyx_PyType_HasFeature(b, Py_TPFLAGS_HEAPTYPE))
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
            PyErr_Format(PyExc_TypeError,
                "base class '" __Pyx_FMT_TYPENAME "' is not a heap type", b_name);
            __Pyx_DECREF_TypeName(b_name);
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
        if (dictoffset == 0)
        {
            Py_ssize_t b_dictoffset = 0;
#if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
            b_dictoffset = b->tp_dictoffset;
#else
            PyObject *py_b_dictoffset = PyObject_GetAttrString((PyObject*)b, "__dictoffset__");
            if (!py_b_dictoffset) goto dictoffset_return;
            b_dictoffset = PyLong_AsSsize_t(py_b_dictoffset);
            Py_DECREF(py_b_dictoffset);
            if (b_dictoffset == -1 && PyErr_Occurred()) goto dictoffset_return;
#endif
            if (b_dictoffset) {
                {
                    __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
                    PyErr_Format(PyExc_TypeError,
                        "extension type '%.200s' has no __dict__ slot, "
                        "but base type '" __Pyx_FMT_TYPENAME "' has: "
                        "either add 'cdef dict __dict__' to the extension type "
                        "or add '__slots__ = [...]' to the base type",
                        type_name, b_name);
                    __Pyx_DECREF_TypeName(b_name);
                }
#if !(CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY)
              dictoffset_return:
#endif
#if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(b0);
#endif
                return -1;
            }
        }
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(b0);
#endif
    }
    return 0;
}
#endif

/* PyType_Ready */
static int __Pyx_PyType_Ready(PyTypeObject *t) {
#if CYTHON_USE_TYPE_SPECS || !(CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API) || defined(PYSTON_MAJOR_VERSION)
    (void)__Pyx_PyObject_CallMethod0;
#if CYTHON_USE_TYPE_SPECS
    (void)__Pyx_validate_bases_tuple;
#endif
    return PyType_Ready(t);
#else
    int r;
    PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
    if (bases && unlikely(__Pyx_validate_bases_tuple(t->tp_name, t->tp_dictoffset, bases) == -1))
        return -1;
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
    {
        int gc_was_enabled;
    #if PY_VERSION_HEX >= 0x030A00b1
        gc_was_enabled = PyGC_Disable();
        (void)__Pyx_PyObject_CallMethod0;
    #else
        PyObject *ret, *py_status;
        PyObject *gc = NULL;
        #if PY_VERSION_HEX >= 0x030700a1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM+0 >= 0x07030400)
        gc = PyImport_GetModule(__pyx_kp_u_gc);
        #endif
        if (unlikely(!gc)) gc = PyImport_Import(__pyx_kp_u_gc);
        if (unlikely(!gc)) return -1;
        py_status = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_isenabled);
        if (unlikely(!py_status)) {
            Py_DECREF(gc);
            return -1;
        }
        gc_was_enabled = __Pyx_PyObject_IsTrue(py_status);
        Py_DECREF(py_status);
        if (gc_was_enabled > 0) {
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_disable);
            if (unlikely(!ret)) {
                Py_DECREF(gc);
                return -1;
            }
            Py_DECREF(ret);
        } else if (unlikely(gc_was_enabled == -1)) {
            Py_DECREF(gc);
            return -1;
        }
    #endif
        t->tp_flags |= Py_TPFLAGS_HEAPTYPE;
#if PY_VERSION_HEX >= 0x030A0000
        t->tp_flags |= Py_TPFLAGS_IMMUTABLETYPE;
#endif
#else
        (void)__Pyx_PyObject_CallMethod0;
#endif
    r = PyType_Ready(t);
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
        t->tp_flags &= ~Py_TPFLAGS_HEAPTYPE;
    #if PY_VERSION_HEX >= 0x030A00b1
        if (gc_was_enabled)
            PyGC_Enable();
    #else
        if (gc_was_enabled) {
            PyObject *tp, *v, *tb;
            PyErr_Fetch(&tp, &v, &tb);
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_enable);
            if (likely(ret || r == -1)) {
                Py_XDECREF(ret);
                PyErr_Restore(tp, v, tb);
            } else {
                Py_XDECREF(tp);
                Py_XDECREF(v);
                Py_XDECREF(tb);
                r = -1;
            }
        }
        Py_DECREF(gc);
    #endif
    }
#endif
    return r;
#endif
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    __Pyx_TypeName type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, attr_name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(attr_name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyTypeObject *type, void *vtable) {
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
    if (unlikely(!ob))
        goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(PyObject_SetAttr((PyObject *) type, __pyx_n_s_pyx_vtable, ob) < 0))
#else
    if (unlikely(PyDict_SetItem(type->tp_dict, __pyx_n_s_pyx_vtable, ob) < 0))
#endif
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* GetVTable */
static void* __Pyx_GetVtable(PyTypeObject *type) {
    void* ptr;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *ob = PyObject_GetAttr((PyObject *)type, __pyx_n_s_pyx_vtable);
#else
    PyObject *ob = PyObject_GetItem(type->tp_dict, __pyx_n_s_pyx_vtable);
#endif
    if (!ob)
        goto bad;
    ptr = PyCapsule_GetPointer(ob, 0);
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* MergeVTables */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_MergeVtables(PyTypeObject *type) {
    int i;
    void** base_vtables;
    __Pyx_TypeName tp_base_name;
    __Pyx_TypeName base_name;
    void* unknown = (void*)-1;
    PyObject* bases = type->tp_bases;
    int base_depth = 0;
    {
        PyTypeObject* base = type->tp_base;
        while (base) {
            base_depth += 1;
            base = base->tp_base;
        }
    }
    base_vtables = (void**) malloc(sizeof(void*) * (size_t)(base_depth + 1));
    base_vtables[0] = unknown;
    for (i = 1; i < PyTuple_GET_SIZE(bases); i++) {
        void* base_vtable = __Pyx_GetVtable(((PyTypeObject*)PyTuple_GET_ITEM(bases, i)));
        if (base_vtable != NULL) {
            int j;
            PyTypeObject* base = type->tp_base;
            for (j = 0; j < base_depth; j++) {
                if (base_vtables[j] == unknown) {
                    base_vtables[j] = __Pyx_GetVtable(base);
                    base_vtables[j + 1] = unknown;
                }
                if (base_vtables[j] == base_vtable) {
                    break;
                } else if (base_vtables[j] == NULL) {
                    goto bad;
                }
                base = base->tp_base;
            }
        }
    }
    PyErr_Clear();
    free(base_vtables);
    return 0;
bad:
    tp_base_name = __Pyx_PyType_GetName(type->tp_base);
    base_name = __Pyx_PyType_GetName((PyTypeObject*)PyTuple_GET_ITEM(bases, i));
    PyErr_Format(PyExc_TypeError,
        "multiple bases have vtable conflict: '" __Pyx_FMT_TYPENAME "' and '" __Pyx_FMT_TYPENAME "'", tp_base_name, base_name);
    __Pyx_DECREF_TypeName(tp_base_name);
    __Pyx_DECREF_TypeName(base_name);
    free(base_vtables);
    return -1;
}
#endif

/* SetupReduce */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStrNoError(meth, __pyx_n_s_name_2);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred()) {
        __Pyx_TypeName type_obj_name =
            __Pyx_PyType_GetName((PyTypeObject*)type_obj);
        PyErr_Format(PyExc_RuntimeError,
            "Unable to initialize pickling for " __Pyx_FMT_TYPENAME, type_obj_name);
        __Pyx_DECREF_TypeName(type_obj_name);
    }
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *empty_list = 0;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (unlikely(!py_import))
        goto bad;
    if (!from_list) {
        empty_list = PyList_New(0);
        if (unlikely(!empty_list))
            goto bad;
        from_list = empty_list;
    }
    #endif
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.') != NULL) {
                module = PyImport_ImportModuleLevelObject(
                    name, __pyx_d, empty_dict, from_list, 1);
                if (unlikely(!module)) {
                    if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (unlikely(!py_level))
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, __pyx_d, empty_dict, from_list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, __pyx_d, empty_dict, from_list, level);
            #endif
        }
    }
bad:
    Py_XDECREF(empty_dict);
    Py_XDECREF(empty_list);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    return module;
}

/* ImportDottedModule */
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Error(PyObject *name, PyObject *parts_tuple, Py_ssize_t count) {
    PyObject *partial_name = NULL, *slice = NULL, *sep = NULL;
    if (unlikely(PyErr_Occurred())) {
        PyErr_Clear();
    }
    if (likely(PyTuple_GET_SIZE(parts_tuple) == count)) {
        partial_name = name;
    } else {
        slice = PySequence_GetSlice(parts_tuple, 0, count);
        if (unlikely(!slice))
            goto bad;
        sep = PyUnicode_FromStringAndSize(".", 1);
        if (unlikely(!sep))
            goto bad;
        partial_name = PyUnicode_Join(sep, slice);
    }
    PyErr_Format(
#if PY_MAJOR_VERSION < 3
        PyExc_ImportError,
        "No module named '%s'", PyString_AS_STRING(partial_name));
#else
#if PY_VERSION_HEX >= 0x030600B1
        PyExc_ModuleNotFoundError,
#else
        PyExc_ImportError,
#endif
        "No module named '%U'", partial_name);
#endif
bad:
    Py_XDECREF(sep);
    Py_XDECREF(slice);
    Py_XDECREF(partial_name);
    return NULL;
}
#endif
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Lookup(PyObject *name) {
    PyObject *imported_module;
#if PY_VERSION_HEX < 0x030700A1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400)
    PyObject *modules = PyImport_GetModuleDict();
    if (unlikely(!modules))
        return NULL;
    imported_module = __Pyx_PyDict_GetItemStr(modules, name);
    Py_XINCREF(imported_module);
#else
    imported_module = PyImport_GetModule(name);
#endif
    return imported_module;
}
#endif
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_ImportDottedModule_WalkParts(PyObject *module, PyObject *name, PyObject *parts_tuple) {
    Py_ssize_t i, nparts;
    nparts = PyTuple_GET_SIZE(parts_tuple);
    for (i=1; i < nparts && module; i++) {
        PyObject *part, *submodule;
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        part = PyTuple_GET_ITEM(parts_tuple, i);
#else
        part = PySequence_ITEM(parts_tuple, i);
#endif
        submodule = __Pyx_PyObject_GetAttrStrNoError(module, part);
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
        Py_DECREF(part);
#endif
        Py_DECREF(module);
        module = submodule;
    }
    if (unlikely(!module)) {
        return __Pyx__ImportDottedModule_Error(name, parts_tuple, i);
    }
    return module;
}
#endif
static PyObject *__Pyx__ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if PY_MAJOR_VERSION < 3
    PyObject *module, *from_list, *star = __pyx_n_s__22;
    CYTHON_UNUSED_VAR(parts_tuple);
    from_list = PyList_New(1);
    if (unlikely(!from_list))
        return NULL;
    Py_INCREF(star);
    PyList_SET_ITEM(from_list, 0, star);
    module = __Pyx_Import(name, from_list, 0);
    Py_DECREF(from_list);
    return module;
#else
    PyObject *imported_module;
    PyObject *module = __Pyx_Import(name, NULL, 0);
    if (!parts_tuple || unlikely(!module))
        return module;
    imported_module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(imported_module)) {
        Py_DECREF(module);
        return imported_module;
    }
    PyErr_Clear();
    return __Pyx_ImportDottedModule_WalkParts(module, name, parts_tuple);
#endif
}
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030400B1
    PyObject *module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(module)) {
        PyObject *spec = __Pyx_PyObject_GetAttrStrNoError(module, __pyx_n_s_spec);
        if (likely(spec)) {
            PyObject *unsafe = __Pyx_PyObject_GetAttrStrNoError(spec, __pyx_n_s_initializing);
            if (likely(!unsafe || !__Pyx_PyObject_IsTrue(unsafe))) {
                Py_DECREF(spec);
                spec = NULL;
            }
            Py_XDECREF(unsafe);
        }
        if (likely(!spec)) {
            PyErr_Clear();
            return module;
        }
        Py_DECREF(spec);
        Py_DECREF(module);
    } else if (PyErr_Occurred()) {
        PyErr_Clear();
    }
#endif
    return __Pyx__ImportDottedModule(name, parts_tuple);
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        const char* module_name_str = 0;
        PyObject* module_name = 0;
        PyObject* module_dot = 0;
        PyObject* full_name = 0;
        PyErr_Clear();
        module_name_str = PyModule_GetName(module);
        if (unlikely(!module_name_str)) { goto modbad; }
        module_name = PyUnicode_FromString(module_name_str);
        if (unlikely(!module_name)) { goto modbad; }
        module_dot = PyUnicode_Concat(module_name, __pyx_kp_u__23);
        if (unlikely(!module_dot)) { goto modbad; }
        full_name = PyUnicode_Concat(module_dot, name);
        if (unlikely(!full_name)) { goto modbad; }
        #if PY_VERSION_HEX < 0x030700A1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400)
        {
            PyObject *modules = PyImport_GetModuleDict();
            if (unlikely(!modules))
                goto modbad;
            value = PyObject_GetItem(modules, full_name);
        }
        #else
        value = PyImport_GetModule(full_name);
        #endif
      modbad:
        Py_XDECREF(full_name);
        Py_XDECREF(module_dot);
        Py_XDECREF(module_name);
    }
    if (unlikely(!value)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* FetchSharedCythonModule */
static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    return __Pyx_PyImport_AddModuleRef((char*) __PYX_ABI_MODULE_NAME);
}

/* FetchCommonType */
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t basicsize,
                               Py_ssize_t expected_basicsize) {
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* abi_module;
    const char* object_name;
    PyTypeObject *cached_type = NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    object_name = strrchr(type->tp_name, '.');
    object_name = object_name ? object_name+1 : type->tp_name;
    cached_type = (PyTypeObject*) PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        if (__Pyx_VerifyCachedType(
              (PyObject *)cached_type,
              object_name,
              cached_type->tp_basicsize,
              type->tp_basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    if (PyType_Ready(type) < 0) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, (PyObject *)type) < 0)
        goto bad;
    Py_INCREF(type);
    cached_type = type;
done:
    Py_DECREF(abi_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#else
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module, *cached_type = NULL;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    cached_type = PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        Py_ssize_t basicsize;
#if CYTHON_COMPILING_IN_LIMITED_API
        PyObject *py_basicsize;
        py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
        if (unlikely(!py_basicsize)) goto bad;
        basicsize = PyLong_AsSsize_t(py_basicsize);
        Py_DECREF(py_basicsize);
        py_basicsize = 0;
        if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
#else
        basicsize = likely(PyType_Check(cached_type)) ? ((PyTypeObject*) cached_type)->tp_basicsize : -1;
#endif
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              basicsize,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    CYTHON_UNUSED_VAR(module);
    cached_type = __Pyx_PyType_FromModuleAndSpec(abi_module, spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, cached_type) < 0) goto bad;
done:
    Py_DECREF(abi_module);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#endif

/* PyVectorcallFastCallDict */
#if CYTHON_METH_FASTCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i, pos;
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (PyDict_Next(kw, &pos, &key, &value)) {
        keys_are_strings &= Py_TYPE(key)->tp_flags;
        Py_INCREF(key);
        Py_INCREF(value);
        PyTuple_SET_ITEM(kwnames, i, key);
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    if (likely(kw == NULL) || PyDict_GET_SIZE(kw) == 0) {
        return vc(func, args, nargs, NULL);
    }
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared */
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc) {
    if (__Pyx_CyFunction_Check(func)) {
        return PyCFunction_GetFunction(((__pyx_CyFunctionObject*)func)->func) == (PyCFunction) cfunc;
    } else if (PyCFunction_Check(func)) {
        return PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
    }
    return 0;
}
#else
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc) {
    return __Pyx_CyOrPyCFunction_Check(func) && __Pyx_CyOrPyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
}
#endif
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure)
{
    CYTHON_UNUSED_VAR(closure);
    if (unlikely(op->func_doc == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_doc = PyObject_GetAttrString(op->func, "__doc__");
        if (unlikely(!op->func_doc)) return NULL;
#else
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
#endif
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_name == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_name = PyObject_GetAttrString(op->func, "__name__");
#elif PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_dict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = __Pyx_PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = __Pyx_PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_tuple;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_kwdict;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->func_annotations;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    int is_coroutine;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
#if PY_VERSION_HEX >= 0x03050000
    if (is_coroutine) {
        PyObject *module, *fromlist, *marker = __pyx_n_s_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
#if CYTHON_ASSUME_SAFE_MACROS
        PyList_SET_ITEM(fromlist, 0, marker);
#else
        if (unlikely(PyList_SetItem(fromlist, 0, marker) < 0)) {
            Py_DECREF(marker);
            Py_DECREF(fromlist);
            return NULL;
        }
#endif
        module = PyImport_ImportModuleLevelObject(__pyx_n_s_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        op->func_is_coroutine = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(op->func_is_coroutine)) {
            return __Pyx_NewRef(op->func_is_coroutine);
        }
ignore:
        PyErr_Clear();
    }
#endif
    op->func_is_coroutine = __Pyx_PyBool_FromLong(is_coroutine);
    return __Pyx_NewRef(op->func_is_coroutine);
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *
__Pyx_CyFunction_get_module(__pyx_CyFunctionObject *op, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_GetAttrString(op->func, "__module__");
}
static int
__Pyx_CyFunction_set_module(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_SetAttrString(op->func, "__module__", value);
}
#endif
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {(char *) "_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", (getter)__Pyx_CyFunction_get_module, (setter)__Pyx_CyFunction_set_module, 0, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
#if !CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#endif
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#if CYTHON_METH_FASTCALL
#if CYTHON_BACKPORT_VECTORCALL
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
#if !CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#endif
#endif
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    CYTHON_UNUSED_VAR(args);
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(((PyCFunctionObject*)m)->m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
#endif
    if (unlikely(op == NULL))
        return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    op->func = PyCFunction_NewEx(ml, (PyObject*)op, module);
    if (unlikely(!op->func)) return NULL;
#endif
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
#if !CYTHON_COMPILING_IN_LIMITED_API
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
#endif
    Py_XINCREF(closure);
    op->func_closure = closure;
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_XINCREF(module);
    cf->m_module = module;
#endif
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func);
#else
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
#endif
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
#if PY_VERSION_HEX < 0x030900B1
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func);
#else
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
#endif
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
#endif
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *f = ((__pyx_CyFunctionObject*)func)->func;
    PyObject *py_name = NULL;
    PyCFunction meth;
    int flags;
    meth = PyCFunction_GetFunction(f);
    if (unlikely(!meth)) return NULL;
    flags = PyCFunction_GetFlags(f);
    if (unlikely(flags < 0)) return NULL;
#else
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    int flags = f->m_ml->ml_flags;
#endif
    Py_ssize_t size;
    switch (flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 0))
                return (*meth)(self, NULL);
#if CYTHON_COMPILING_IN_LIMITED_API
            py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
            if (!py_name) return NULL;
            PyErr_Format(PyExc_TypeError,
                "%.200S() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                py_name, size);
            Py_DECREF(py_name);
#else
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
#endif
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = __Pyx_PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
#if CYTHON_COMPILING_IN_LIMITED_API
            py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
            if (!py_name) return NULL;
            PyErr_Format(PyExc_TypeError,
                "%.200S() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                py_name, size);
            Py_DECREF(py_name);
#else
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
#endif
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
#if CYTHON_COMPILING_IN_LIMITED_API
    py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
    if (!py_name) return NULL;
    PyErr_Format(PyExc_TypeError, "%.200S() takes no keyword arguments",
                 py_name);
    Py_DECREF(py_name);
#else
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
#endif
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *self, *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)func)->func);
    if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
    self = ((PyCFunctionObject*)func)->m_self;
#endif
    result = __Pyx_CyFunction_CallMethod(func, self, arg, kw);
    return result;
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
#if CYTHON_ASSUME_SAFE_MACROS
        argc = PyTuple_GET_SIZE(args);
#else
        argc = PyTuple_Size(args);
        if (unlikely(!argc) < 0) return NULL;
#endif
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
#if PY_MAJOR_VERSION > 2
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
#else
            PyErr_SetString(PyExc_TypeError,
                            "unbound method needs an argument");
#endif
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            PyErr_Format(PyExc_TypeError, "%.200s() needs an argument",
                         ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(PyTuple_GET_SIZE(kwnames))) {
        PyErr_Format(PyExc_TypeError,
                     "%.200s() takes no keyword arguments", ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))def->ml_meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCMethod)(void(*)(void))def->ml_meth)(self, cls, args, (size_t)nargs, kwnames);
}
#endif
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if (defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
#else
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
#if !CYTHON_METH_FASTCALL
    0,
#elif CYTHON_BACKPORT_VECTORCALL
    (printfunc)offsetof(__pyx_CyFunctionObject, func_vectorcall),
#else
    offsetof(PyCFunctionObject, vectorcall),
#endif
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_PyMethod_New,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if PY_VERSION_HEX >= 0x030d00A4
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
static int __pyx_CyFunction_init(PyObject *module) {
#if CYTHON_USE_TYPE_SPECS
    __pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_CyFunctionType_spec, NULL);
#else
    CYTHON_UNUSED_VAR(module);
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
#endif
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* Py3UpdateBases */
static PyObject*
__Pyx_PEP560_update_bases(PyObject *bases)
{
    Py_ssize_t i, j, size_bases;
    PyObject *base, *meth, *new_base, *result, *new_bases = NULL;
    size_bases = PyTuple_GET_SIZE(bases);
    for (i = 0; i < size_bases; i++) {
        base  = PyTuple_GET_ITEM(bases, i);
        if (PyType_Check(base)) {
            if (new_bases) {
                if (PyList_Append(new_bases, base) < 0) {
                    goto error;
                }
            }
            continue;
        }
        meth = __Pyx_PyObject_GetAttrStrNoError(base, __pyx_n_s_mro_entries);
        if (!meth && PyErr_Occurred()) {
            goto error;
        }
        if (!meth) {
            if (new_bases) {
                if (PyList_Append(new_bases, base) < 0) {
                    goto error;
                }
            }
            continue;
        }
        new_base = __Pyx_PyObject_CallOneArg(meth, bases);
        Py_DECREF(meth);
        if (!new_base) {
            goto error;
        }
        if (!PyTuple_Check(new_base)) {
            PyErr_SetString(PyExc_TypeError,
                            "__mro_entries__ must return a tuple");
            Py_DECREF(new_base);
            goto error;
        }
        if (!new_bases) {
            if (!(new_bases = PyList_New(i))) {
                goto error;
            }
            for (j = 0; j < i; j++) {
                base = PyTuple_GET_ITEM(bases, j);
                PyList_SET_ITEM(new_bases, j, base);
                Py_INCREF(base);
            }
        }
        j = PyList_GET_SIZE(new_bases);
        if (PyList_SetSlice(new_bases, j, j, new_base) < 0) {
            goto error;
        }
        Py_DECREF(new_base);
    }
    if (!new_bases) {
        Py_INCREF(bases);
        return bases;
    }
    result = PyList_AsTuple(new_bases);
    Py_DECREF(new_bases);
    return result;
error:
    Py_XDECREF(new_bases);
    return NULL;
}

/* CalculateMetaclass */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases;
#if CYTHON_ASSUME_SAFE_MACROS
    nbases = PyTuple_GET_SIZE(bases);
#else
    nbases = PyTuple_Size(bases);
    if (nbases < 0) return NULL;
#endif
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
#if CYTHON_ASSUME_SAFE_MACROS
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *tmp = PyTuple_GetItem(bases, i);
        if (!tmp) return NULL;
#endif
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* PyObjectLookupSpecial */
#if CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx__PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name, int with_error) {
    PyObject *res;
    PyTypeObject *tp = Py_TYPE(obj);
#if PY_MAJOR_VERSION < 3
    if (unlikely(PyInstance_Check(obj)))
        return with_error ? __Pyx_PyObject_GetAttrStr(obj, attr_name) : __Pyx_PyObject_GetAttrStrNoError(obj, attr_name);
#endif
    res = _PyType_Lookup(tp, attr_name);
    if (likely(res)) {
        descrgetfunc f = Py_TYPE(res)->tp_descr_get;
        if (!f) {
            Py_INCREF(res);
        } else {
            res = f(res, obj, (PyObject *)tp);
        }
    } else if (with_error) {
        PyErr_SetObject(PyExc_AttributeError, attr_name);
    }
    return res;
}
#endif

/* Py3ClassCreate */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStrNoError(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs[3] = {NULL, name, bases};
            ns = __Pyx_PyObject_FastCallDict(prep, pargs+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, mkw);
            Py_DECREF(prep);
        } else {
            if (unlikely(PyErr_Occurred()))
                return NULL;
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
#if PY_VERSION_HEX >= 0x03030000
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
#else
    CYTHON_MAYBE_UNUSED_VAR(qualname);
#endif
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
#if PY_VERSION_HEX < 0x030600A4 && CYTHON_PEP487_INIT_SUBCLASS
static int __Pyx_SetNamesPEP487(PyObject *type_obj) {
    PyTypeObject *type = (PyTypeObject*) type_obj;
    PyObject *names_to_set, *key, *value, *set_name, *tmp;
    Py_ssize_t i = 0;
#if CYTHON_USE_TYPE_SLOTS
    names_to_set = PyDict_Copy(type->tp_dict);
#else
    {
        PyObject *d = PyObject_GetAttr(type_obj, __pyx_n_s_dict);
        names_to_set = NULL;
        if (likely(d)) {
            PyObject *names_to_set = PyDict_New();
            int ret = likely(names_to_set) ? PyDict_Update(names_to_set, d) : -1;
            Py_DECREF(d);
            if (unlikely(ret < 0))
                Py_CLEAR(names_to_set);
        }
    }
#endif
    if (unlikely(names_to_set == NULL))
        goto bad;
    while (PyDict_Next(names_to_set, &i, &key, &value)) {
        set_name = __Pyx_PyObject_LookupSpecialNoError(value, __pyx_n_s_set_name);
        if (unlikely(set_name != NULL)) {
            tmp = __Pyx_PyObject_Call2Args(set_name, type_obj, key);
            Py_DECREF(set_name);
            if (unlikely(tmp == NULL)) {
                __Pyx_TypeName value_type_name =
                    __Pyx_PyType_GetName(Py_TYPE(value));
                __Pyx_TypeName type_name = __Pyx_PyType_GetName(type);
                PyErr_Format(PyExc_RuntimeError,
#if PY_MAJOR_VERSION >= 3
                    "Error calling __set_name__ on '" __Pyx_FMT_TYPENAME "' instance %R " "in '" __Pyx_FMT_TYPENAME "'",
                    value_type_name, key, type_name);
#else
                    "Error calling __set_name__ on '" __Pyx_FMT_TYPENAME "' instance %.100s in '" __Pyx_FMT_TYPENAME "'",
                    value_type_name,
                    PyString_Check(key) ? PyString_AS_STRING(key) : "?",
                    type_name);
#endif
                goto bad;
            } else {
                Py_DECREF(tmp);
            }
        }
        else if (unlikely(PyErr_Occurred())) {
            goto bad;
        }
    }
    Py_DECREF(names_to_set);
    return 0;
bad:
    Py_XDECREF(names_to_set);
    return -1;
}
static PyObject *__Pyx_InitSubclassPEP487(PyObject *type_obj, PyObject *mkw) {
#if CYTHON_USE_TYPE_SLOTS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyTypeObject *type = (PyTypeObject*) type_obj;
    PyObject *mro = type->tp_mro;
    Py_ssize_t i, nbases;
    if (unlikely(!mro)) goto done;
    (void) &__Pyx_GetBuiltinName;
    Py_INCREF(mro);
    nbases = PyTuple_GET_SIZE(mro);
    assert(PyTuple_GET_ITEM(mro, 0) == type_obj);
    for (i = 1; i < nbases-1; i++) {
        PyObject *base, *dict, *meth;
        base = PyTuple_GET_ITEM(mro, i);
        dict = ((PyTypeObject *)base)->tp_dict;
        meth = __Pyx_PyDict_GetItemStrWithError(dict, __pyx_n_s_init_subclass);
        if (unlikely(meth)) {
            descrgetfunc f = Py_TYPE(meth)->tp_descr_get;
            PyObject *res;
            Py_INCREF(meth);
            if (likely(f)) {
                res = f(meth, NULL, type_obj);
                Py_DECREF(meth);
                if (unlikely(!res)) goto bad;
                meth = res;
            }
            res = __Pyx_PyObject_FastCallDict(meth, NULL, 0, mkw);
            Py_DECREF(meth);
            if (unlikely(!res)) goto bad;
            Py_DECREF(res);
            goto done;
        } else if (unlikely(PyErr_Occurred())) {
            goto bad;
        }
    }
done:
    Py_XDECREF(mro);
    return type_obj;
bad:
    Py_XDECREF(mro);
    Py_DECREF(type_obj);
    return NULL;
#else
    PyObject *super_type, *super, *func, *res;
#if CYTHON_COMPILING_IN_PYPY && !defined(PySuper_Type)
    super_type = __Pyx_GetBuiltinName(__pyx_n_s_super);
#else
    super_type = (PyObject*) &PySuper_Type;
    (void) &__Pyx_GetBuiltinName;
#endif
    super = likely(super_type) ? __Pyx_PyObject_Call2Args(super_type, type_obj, type_obj) : NULL;
#if CYTHON_COMPILING_IN_PYPY && !defined(PySuper_Type)
    Py_XDECREF(super_type);
#endif
    if (unlikely(!super)) {
        Py_CLEAR(type_obj);
        goto done;
    }
    func = __Pyx_PyObject_GetAttrStrNoError(super, __pyx_n_s_init_subclass);
    Py_DECREF(super);
    if (likely(!func)) {
        if (unlikely(PyErr_Occurred()))
            Py_CLEAR(type_obj);
        goto done;
    }
    res = __Pyx_PyObject_FastCallDict(func, NULL, 0, mkw);
    Py_DECREF(func);
    if (unlikely(!res))
        Py_CLEAR(type_obj);
    Py_XDECREF(res);
done:
    return type_obj;
#endif
}
#endif
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result;
    PyObject *owned_metaclass = NULL;
    PyObject *margs[4] = {NULL, name, bases, dict};
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    result = __Pyx_PyObject_FastCallDict(metaclass, margs+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET,
#if PY_VERSION_HEX < 0x030600A4
        (metaclass == (PyObject*)&PyType_Type) ? NULL : mkw
#else
        mkw
#endif
    );
    Py_XDECREF(owned_metaclass);
#if PY_VERSION_HEX < 0x030600A4 && CYTHON_PEP487_INIT_SUBCLASS
    if (likely(result) && likely(PyType_Check(result))) {
        if (unlikely(__Pyx_SetNamesPEP487(result) < 0)) {
            Py_CLEAR(result);
        } else {
            result = __Pyx_InitSubclassPEP487(result, mkw);
        }
    }
#else
    (void) &__Pyx_GetBuiltinName;
#endif
    return result;
}

/* ClassMethod */
static PyObject* __Pyx_Method_ClassMethod(PyObject *method) {
#if CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM <= 0x05080000
    if (PyObject_TypeCheck(method, &PyWrapperDescr_Type)) {
        return PyClassMethod_New(method);
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (PyMethodDescr_Check(method))
#else
    #if PY_MAJOR_VERSION == 2
    static PyTypeObject *methoddescr_type = NULL;
    if (unlikely(methoddescr_type == NULL)) {
       PyObject *meth = PyObject_GetAttrString((PyObject*)&PyList_Type, "append");
       if (unlikely(!meth)) return NULL;
       methoddescr_type = Py_TYPE(meth);
       Py_DECREF(meth);
    }
    #else
    PyTypeObject *methoddescr_type = &PyMethodDescr_Type;
    #endif
    if (__Pyx_TypeCheck(method, methoddescr_type))
#endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject *)method;
        #if PY_VERSION_HEX < 0x03020000
        PyTypeObject *d_type = descr->d_type;
        #else
        PyTypeObject *d_type = descr->d_common.d_type;
        #endif
        return PyDescr_NewClassMethod(d_type, descr->d_method);
    }
#endif
    else if (PyMethod_Check(method)) {
        return PyClassMethod_New(PyMethod_GET_FUNCTION(method));
    }
    else {
        return PyClassMethod_New(method);
    }
}

/* GetNameInClass */
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name) {
    PyObject *result;
    PyObject *dict;
    assert(PyType_Check(nmspace));
#if CYTHON_USE_TYPE_SLOTS
    dict = ((PyTypeObject*)nmspace)->tp_dict;
    Py_XINCREF(dict);
#else
    dict = PyObject_GetAttr(nmspace, __pyx_n_s_dict);
#endif
    if (likely(dict)) {
        result = PyObject_GetItem(dict, name);
        Py_DECREF(dict);
        if (result) {
            return result;
        }
    }
    PyErr_Clear();
    __Pyx_GetModuleGlobalNameUncached(result, name);
    return result;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStrNoError(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}
#endif

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result;
        result = PyObject_Call(replace, __pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    #if __PYX_LIMITED_VERSION_HEX < 0x030780000
    {
        PyObject *compiled = NULL, *result = NULL;
        if (unlikely(PyDict_SetItemString(scratch_dict, "code", code))) return NULL;
        if (unlikely(PyDict_SetItemString(scratch_dict, "type", (PyObject*)(&PyType_Type)))) return NULL;
        compiled = Py_CompileString(
            "out = type(code)(\n"
            "  code.co_argcount, code.co_kwonlyargcount, code.co_nlocals, code.co_stacksize,\n"
            "  code.co_flags, code.co_code, code.co_consts, code.co_names,\n"
            "  code.co_varnames, code.co_filename, co_name, co_firstlineno,\n"
            "  code.co_lnotab)\n", "<dummy>", Py_file_input);
        if (!compiled) return NULL;
        result = PyEval_EvalCode(compiled, scratch_dict, scratch_dict);
        Py_DECREF(compiled);
        if (!result) PyErr_Print();
        Py_DECREF(result);
        result = PyDict_GetItemString(scratch_dict, "out");
        if (result) Py_INCREF(result);
        return result;
    }
    #else
    return NULL;
    #endif
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        (void) __pyx_cfilenm;
        (void) __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
    if (unlikely(!code_object)) goto bad;
    py_py_line = PyLong_FromLong(py_line);
    if (unlikely(!py_py_line)) goto bad;
    py_funcname = PyUnicode_FromString(funcname);
    if (unlikely(!py_funcname)) goto bad;
    dict = PyDict_New();
    if (unlikely(!dict)) goto bad;
    {
        PyObject *old_code_object = code_object;
        code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
        Py_DECREF(old_code_object);
    }
    if (unlikely(!code_object)) goto bad;
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_glyph_flags_t(hb_glyph_flags_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hb_glyph_flags_t neg_one = (hb_glyph_flags_t) -1, const_zero = (hb_glyph_flags_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(hb_glyph_flags_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(hb_glyph_flags_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_glyph_flags_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(hb_glyph_flags_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_glyph_flags_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(hb_glyph_flags_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(hb_glyph_flags_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_buffer_flags_t(hb_buffer_flags_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hb_buffer_flags_t neg_one = (hb_buffer_flags_t) -1, const_zero = (hb_buffer_flags_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(hb_buffer_flags_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(hb_buffer_flags_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_buffer_flags_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(hb_buffer_flags_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_buffer_flags_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(hb_buffer_flags_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(hb_buffer_flags_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_buffer_cluster_level_t(hb_buffer_cluster_level_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hb_buffer_cluster_level_t neg_one = (hb_buffer_cluster_level_t) -1, const_zero = (hb_buffer_cluster_level_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(hb_buffer_cluster_level_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(hb_buffer_cluster_level_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_buffer_cluster_level_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(hb_buffer_cluster_level_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_buffer_cluster_level_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(hb_buffer_cluster_level_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(hb_buffer_cluster_level_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_buffer_content_type_t(hb_buffer_content_type_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hb_buffer_content_type_t neg_one = (hb_buffer_content_type_t) -1, const_zero = (hb_buffer_content_type_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(hb_buffer_content_type_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(hb_buffer_content_type_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_buffer_content_type_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(hb_buffer_content_type_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_buffer_content_type_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(hb_buffer_content_type_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(hb_buffer_content_type_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_codepoint_t(hb_codepoint_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hb_codepoint_t neg_one = (hb_codepoint_t) -1, const_zero = (hb_codepoint_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(hb_codepoint_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(hb_codepoint_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_codepoint_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(hb_codepoint_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_codepoint_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(hb_codepoint_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(hb_codepoint_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(int) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE hb_codepoint_t __Pyx_PyInt_As_hb_codepoint_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hb_codepoint_t neg_one = (hb_codepoint_t) -1, const_zero = (hb_codepoint_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(hb_codepoint_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(hb_codepoint_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (hb_codepoint_t) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        hb_codepoint_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (hb_codepoint_t) -1;
        val = __Pyx_PyInt_As_hb_codepoint_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(hb_codepoint_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(hb_codepoint_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_codepoint_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_codepoint_t) >= 2 * PyLong_SHIFT)) {
                            return (hb_codepoint_t) (((((hb_codepoint_t)digits[1]) << PyLong_SHIFT) | (hb_codepoint_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(hb_codepoint_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_codepoint_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_codepoint_t) >= 3 * PyLong_SHIFT)) {
                            return (hb_codepoint_t) (((((((hb_codepoint_t)digits[2]) << PyLong_SHIFT) | (hb_codepoint_t)digits[1]) << PyLong_SHIFT) | (hb_codepoint_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(hb_codepoint_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_codepoint_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_codepoint_t) >= 4 * PyLong_SHIFT)) {
                            return (hb_codepoint_t) (((((((((hb_codepoint_t)digits[3]) << PyLong_SHIFT) | (hb_codepoint_t)digits[2]) << PyLong_SHIFT) | (hb_codepoint_t)digits[1]) << PyLong_SHIFT) | (hb_codepoint_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (hb_codepoint_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(hb_codepoint_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_codepoint_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(hb_codepoint_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_codepoint_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(hb_codepoint_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(hb_codepoint_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_codepoint_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_codepoint_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (hb_codepoint_t) (((hb_codepoint_t)-1)*(((((hb_codepoint_t)digits[1]) << PyLong_SHIFT) | (hb_codepoint_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(hb_codepoint_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_codepoint_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_codepoint_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (hb_codepoint_t) ((((((hb_codepoint_t)digits[1]) << PyLong_SHIFT) | (hb_codepoint_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(hb_codepoint_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_codepoint_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_codepoint_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (hb_codepoint_t) (((hb_codepoint_t)-1)*(((((((hb_codepoint_t)digits[2]) << PyLong_SHIFT) | (hb_codepoint_t)digits[1]) << PyLong_SHIFT) | (hb_codepoint_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(hb_codepoint_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_codepoint_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_codepoint_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (hb_codepoint_t) ((((((((hb_codepoint_t)digits[2]) << PyLong_SHIFT) | (hb_codepoint_t)digits[1]) << PyLong_SHIFT) | (hb_codepoint_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(hb_codepoint_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_codepoint_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_codepoint_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (hb_codepoint_t) (((hb_codepoint_t)-1)*(((((((((hb_codepoint_t)digits[3]) << PyLong_SHIFT) | (hb_codepoint_t)digits[2]) << PyLong_SHIFT) | (hb_codepoint_t)digits[1]) << PyLong_SHIFT) | (hb_codepoint_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(hb_codepoint_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_codepoint_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_codepoint_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (hb_codepoint_t) ((((((((((hb_codepoint_t)digits[3]) << PyLong_SHIFT) | (hb_codepoint_t)digits[2]) << PyLong_SHIFT) | (hb_codepoint_t)digits[1]) << PyLong_SHIFT) | (hb_codepoint_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(hb_codepoint_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_codepoint_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(hb_codepoint_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_codepoint_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        hb_codepoint_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (hb_codepoint_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (hb_codepoint_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (hb_codepoint_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (hb_codepoint_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(hb_codepoint_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((hb_codepoint_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(hb_codepoint_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((hb_codepoint_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((hb_codepoint_t) 1) << (sizeof(hb_codepoint_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (hb_codepoint_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to hb_codepoint_t");
    return (hb_codepoint_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to hb_codepoint_t");
    return (hb_codepoint_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_ot_layout_glyph_class_t(hb_ot_layout_glyph_class_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hb_ot_layout_glyph_class_t neg_one = (hb_ot_layout_glyph_class_t) -1, const_zero = (hb_ot_layout_glyph_class_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(hb_ot_layout_glyph_class_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(hb_ot_layout_glyph_class_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_ot_layout_glyph_class_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(hb_ot_layout_glyph_class_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_ot_layout_glyph_class_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(hb_ot_layout_glyph_class_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(hb_ot_layout_glyph_class_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_ot_color_palette_flags_t(hb_ot_color_palette_flags_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hb_ot_color_palette_flags_t neg_one = (hb_ot_color_palette_flags_t) -1, const_zero = (hb_ot_color_palette_flags_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(hb_ot_color_palette_flags_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(hb_ot_color_palette_flags_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_ot_color_palette_flags_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(hb_ot_color_palette_flags_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_ot_color_palette_flags_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(hb_ot_color_palette_flags_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(hb_ot_color_palette_flags_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_ot_math_constant_t(hb_ot_math_constant_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hb_ot_math_constant_t neg_one = (hb_ot_math_constant_t) -1, const_zero = (hb_ot_math_constant_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(hb_ot_math_constant_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(hb_ot_math_constant_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_ot_math_constant_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(hb_ot_math_constant_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_ot_math_constant_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(hb_ot_math_constant_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(hb_ot_math_constant_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_ot_math_kern_t(hb_ot_math_kern_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hb_ot_math_kern_t neg_one = (hb_ot_math_kern_t) -1, const_zero = (hb_ot_math_kern_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(hb_ot_math_kern_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(hb_ot_math_kern_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_ot_math_kern_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(hb_ot_math_kern_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_ot_math_kern_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(hb_ot_math_kern_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(hb_ot_math_kern_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_ot_math_glyph_part_flags_t(hb_ot_math_glyph_part_flags_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hb_ot_math_glyph_part_flags_t neg_one = (hb_ot_math_glyph_part_flags_t) -1, const_zero = (hb_ot_math_glyph_part_flags_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(hb_ot_math_glyph_part_flags_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(hb_ot_math_glyph_part_flags_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_ot_math_glyph_part_flags_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(hb_ot_math_glyph_part_flags_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_ot_math_glyph_part_flags_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(hb_ot_math_glyph_part_flags_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(hb_ot_math_glyph_part_flags_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_paint_composite_mode_t(hb_paint_composite_mode_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hb_paint_composite_mode_t neg_one = (hb_paint_composite_mode_t) -1, const_zero = (hb_paint_composite_mode_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(hb_paint_composite_mode_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(hb_paint_composite_mode_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_paint_composite_mode_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(hb_paint_composite_mode_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_paint_composite_mode_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(hb_paint_composite_mode_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(hb_paint_composite_mode_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_paint_extend_t(hb_paint_extend_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hb_paint_extend_t neg_one = (hb_paint_extend_t) -1, const_zero = (hb_paint_extend_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(hb_paint_extend_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(hb_paint_extend_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_paint_extend_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(hb_paint_extend_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_paint_extend_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(hb_paint_extend_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(hb_paint_extend_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_subset_sets_t(hb_subset_sets_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hb_subset_sets_t neg_one = (hb_subset_sets_t) -1, const_zero = (hb_subset_sets_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(hb_subset_sets_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(hb_subset_sets_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_subset_sets_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(hb_subset_sets_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_subset_sets_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(hb_subset_sets_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(hb_subset_sets_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_subset_flags_t(hb_subset_flags_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hb_subset_flags_t neg_one = (hb_subset_flags_t) -1, const_zero = (hb_subset_flags_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(hb_subset_flags_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(hb_subset_flags_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_subset_flags_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(hb_subset_flags_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_subset_flags_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(hb_subset_flags_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(hb_subset_flags_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_long(unsigned long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned long neg_one = (unsigned long) -1, const_zero = (unsigned long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(unsigned long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_mask_t(hb_mask_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hb_mask_t neg_one = (hb_mask_t) -1, const_zero = (hb_mask_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(hb_mask_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(hb_mask_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_mask_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(hb_mask_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_mask_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(hb_mask_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(hb_mask_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_position_t(hb_position_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hb_position_t neg_one = (hb_position_t) -1, const_zero = (hb_position_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(hb_position_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(hb_position_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_position_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(hb_position_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_position_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(hb_position_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(hb_position_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_int(unsigned int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned int neg_one = (unsigned int) -1, const_zero = (unsigned int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(unsigned int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE unsigned int __Pyx_PyInt_As_unsigned_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned int neg_one = (unsigned int) -1, const_zero = (unsigned int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(unsigned int) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(unsigned int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (unsigned int) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        unsigned int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (unsigned int) -1;
        val = __Pyx_PyInt_As_unsigned_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(unsigned int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(unsigned int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) >= 2 * PyLong_SHIFT)) {
                            return (unsigned int) (((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) >= 3 * PyLong_SHIFT)) {
                            return (unsigned int) (((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) >= 4 * PyLong_SHIFT)) {
                            return (unsigned int) (((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (unsigned int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(unsigned int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(unsigned int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(unsigned int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned int) (((unsigned int)-1)*(((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(unsigned int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned int) ((((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned int) (((unsigned int)-1)*(((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned int) ((((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned int) (((unsigned int)-1)*(((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned int) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned int) ((((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(unsigned int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(unsigned int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        unsigned int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (unsigned int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (unsigned int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (unsigned int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (unsigned int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(unsigned int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((unsigned int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(unsigned int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((unsigned int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((unsigned int) 1) << (sizeof(unsigned int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (unsigned int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned int");
    return (unsigned int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned int");
    return (unsigned int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE hb_buffer_flags_t __Pyx_PyInt_As_hb_buffer_flags_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hb_buffer_flags_t neg_one = (hb_buffer_flags_t) -1, const_zero = (hb_buffer_flags_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(hb_buffer_flags_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(hb_buffer_flags_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (hb_buffer_flags_t) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        hb_buffer_flags_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (hb_buffer_flags_t) -1;
        val = __Pyx_PyInt_As_hb_buffer_flags_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(hb_buffer_flags_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(hb_buffer_flags_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_buffer_flags_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_buffer_flags_t) >= 2 * PyLong_SHIFT)) {
                            return (hb_buffer_flags_t) (((((hb_buffer_flags_t)digits[1]) << PyLong_SHIFT) | (hb_buffer_flags_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(hb_buffer_flags_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_buffer_flags_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_buffer_flags_t) >= 3 * PyLong_SHIFT)) {
                            return (hb_buffer_flags_t) (((((((hb_buffer_flags_t)digits[2]) << PyLong_SHIFT) | (hb_buffer_flags_t)digits[1]) << PyLong_SHIFT) | (hb_buffer_flags_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(hb_buffer_flags_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_buffer_flags_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_buffer_flags_t) >= 4 * PyLong_SHIFT)) {
                            return (hb_buffer_flags_t) (((((((((hb_buffer_flags_t)digits[3]) << PyLong_SHIFT) | (hb_buffer_flags_t)digits[2]) << PyLong_SHIFT) | (hb_buffer_flags_t)digits[1]) << PyLong_SHIFT) | (hb_buffer_flags_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (hb_buffer_flags_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(hb_buffer_flags_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_buffer_flags_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(hb_buffer_flags_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_buffer_flags_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(hb_buffer_flags_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(hb_buffer_flags_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_buffer_flags_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_buffer_flags_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (hb_buffer_flags_t) (((hb_buffer_flags_t)-1)*(((((hb_buffer_flags_t)digits[1]) << PyLong_SHIFT) | (hb_buffer_flags_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(hb_buffer_flags_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_buffer_flags_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_buffer_flags_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (hb_buffer_flags_t) ((((((hb_buffer_flags_t)digits[1]) << PyLong_SHIFT) | (hb_buffer_flags_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(hb_buffer_flags_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_buffer_flags_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_buffer_flags_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (hb_buffer_flags_t) (((hb_buffer_flags_t)-1)*(((((((hb_buffer_flags_t)digits[2]) << PyLong_SHIFT) | (hb_buffer_flags_t)digits[1]) << PyLong_SHIFT) | (hb_buffer_flags_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(hb_buffer_flags_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_buffer_flags_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_buffer_flags_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (hb_buffer_flags_t) ((((((((hb_buffer_flags_t)digits[2]) << PyLong_SHIFT) | (hb_buffer_flags_t)digits[1]) << PyLong_SHIFT) | (hb_buffer_flags_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(hb_buffer_flags_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_buffer_flags_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_buffer_flags_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (hb_buffer_flags_t) (((hb_buffer_flags_t)-1)*(((((((((hb_buffer_flags_t)digits[3]) << PyLong_SHIFT) | (hb_buffer_flags_t)digits[2]) << PyLong_SHIFT) | (hb_buffer_flags_t)digits[1]) << PyLong_SHIFT) | (hb_buffer_flags_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(hb_buffer_flags_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_buffer_flags_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_buffer_flags_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (hb_buffer_flags_t) ((((((((((hb_buffer_flags_t)digits[3]) << PyLong_SHIFT) | (hb_buffer_flags_t)digits[2]) << PyLong_SHIFT) | (hb_buffer_flags_t)digits[1]) << PyLong_SHIFT) | (hb_buffer_flags_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(hb_buffer_flags_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_buffer_flags_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(hb_buffer_flags_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_buffer_flags_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        hb_buffer_flags_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyErr_SetString(PyExc_RuntimeError,
                        "_PyLong_AsByteArray() or PyLong_AsNativeBytes() not available, cannot convert large enums");
        val = (hb_buffer_flags_t) -1;
#endif
        if (unlikely(ret))
            return (hb_buffer_flags_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to hb_buffer_flags_t");
    return (hb_buffer_flags_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to hb_buffer_flags_t");
    return (hb_buffer_flags_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE hb_buffer_cluster_level_t __Pyx_PyInt_As_hb_buffer_cluster_level_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hb_buffer_cluster_level_t neg_one = (hb_buffer_cluster_level_t) -1, const_zero = (hb_buffer_cluster_level_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(hb_buffer_cluster_level_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(hb_buffer_cluster_level_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (hb_buffer_cluster_level_t) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        hb_buffer_cluster_level_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (hb_buffer_cluster_level_t) -1;
        val = __Pyx_PyInt_As_hb_buffer_cluster_level_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(hb_buffer_cluster_level_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(hb_buffer_cluster_level_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_buffer_cluster_level_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_buffer_cluster_level_t) >= 2 * PyLong_SHIFT)) {
                            return (hb_buffer_cluster_level_t) (((((hb_buffer_cluster_level_t)digits[1]) << PyLong_SHIFT) | (hb_buffer_cluster_level_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(hb_buffer_cluster_level_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_buffer_cluster_level_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_buffer_cluster_level_t) >= 3 * PyLong_SHIFT)) {
                            return (hb_buffer_cluster_level_t) (((((((hb_buffer_cluster_level_t)digits[2]) << PyLong_SHIFT) | (hb_buffer_cluster_level_t)digits[1]) << PyLong_SHIFT) | (hb_buffer_cluster_level_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(hb_buffer_cluster_level_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_buffer_cluster_level_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_buffer_cluster_level_t) >= 4 * PyLong_SHIFT)) {
                            return (hb_buffer_cluster_level_t) (((((((((hb_buffer_cluster_level_t)digits[3]) << PyLong_SHIFT) | (hb_buffer_cluster_level_t)digits[2]) << PyLong_SHIFT) | (hb_buffer_cluster_level_t)digits[1]) << PyLong_SHIFT) | (hb_buffer_cluster_level_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (hb_buffer_cluster_level_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(hb_buffer_cluster_level_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_buffer_cluster_level_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(hb_buffer_cluster_level_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_buffer_cluster_level_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(hb_buffer_cluster_level_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(hb_buffer_cluster_level_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_buffer_cluster_level_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_buffer_cluster_level_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (hb_buffer_cluster_level_t) (((hb_buffer_cluster_level_t)-1)*(((((hb_buffer_cluster_level_t)digits[1]) << PyLong_SHIFT) | (hb_buffer_cluster_level_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(hb_buffer_cluster_level_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_buffer_cluster_level_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_buffer_cluster_level_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (hb_buffer_cluster_level_t) ((((((hb_buffer_cluster_level_t)digits[1]) << PyLong_SHIFT) | (hb_buffer_cluster_level_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(hb_buffer_cluster_level_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_buffer_cluster_level_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_buffer_cluster_level_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (hb_buffer_cluster_level_t) (((hb_buffer_cluster_level_t)-1)*(((((((hb_buffer_cluster_level_t)digits[2]) << PyLong_SHIFT) | (hb_buffer_cluster_level_t)digits[1]) << PyLong_SHIFT) | (hb_buffer_cluster_level_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(hb_buffer_cluster_level_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_buffer_cluster_level_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_buffer_cluster_level_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (hb_buffer_cluster_level_t) ((((((((hb_buffer_cluster_level_t)digits[2]) << PyLong_SHIFT) | (hb_buffer_cluster_level_t)digits[1]) << PyLong_SHIFT) | (hb_buffer_cluster_level_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(hb_buffer_cluster_level_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_buffer_cluster_level_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_buffer_cluster_level_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (hb_buffer_cluster_level_t) (((hb_buffer_cluster_level_t)-1)*(((((((((hb_buffer_cluster_level_t)digits[3]) << PyLong_SHIFT) | (hb_buffer_cluster_level_t)digits[2]) << PyLong_SHIFT) | (hb_buffer_cluster_level_t)digits[1]) << PyLong_SHIFT) | (hb_buffer_cluster_level_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(hb_buffer_cluster_level_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_buffer_cluster_level_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_buffer_cluster_level_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (hb_buffer_cluster_level_t) ((((((((((hb_buffer_cluster_level_t)digits[3]) << PyLong_SHIFT) | (hb_buffer_cluster_level_t)digits[2]) << PyLong_SHIFT) | (hb_buffer_cluster_level_t)digits[1]) << PyLong_SHIFT) | (hb_buffer_cluster_level_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(hb_buffer_cluster_level_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_buffer_cluster_level_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(hb_buffer_cluster_level_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_buffer_cluster_level_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        hb_buffer_cluster_level_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyErr_SetString(PyExc_RuntimeError,
                        "_PyLong_AsByteArray() or PyLong_AsNativeBytes() not available, cannot convert large enums");
        val = (hb_buffer_cluster_level_t) -1;
#endif
        if (unlikely(ret))
            return (hb_buffer_cluster_level_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to hb_buffer_cluster_level_t");
    return (hb_buffer_cluster_level_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to hb_buffer_cluster_level_t");
    return (hb_buffer_cluster_level_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE hb_buffer_content_type_t __Pyx_PyInt_As_hb_buffer_content_type_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hb_buffer_content_type_t neg_one = (hb_buffer_content_type_t) -1, const_zero = (hb_buffer_content_type_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(hb_buffer_content_type_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(hb_buffer_content_type_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (hb_buffer_content_type_t) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        hb_buffer_content_type_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (hb_buffer_content_type_t) -1;
        val = __Pyx_PyInt_As_hb_buffer_content_type_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(hb_buffer_content_type_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(hb_buffer_content_type_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_buffer_content_type_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_buffer_content_type_t) >= 2 * PyLong_SHIFT)) {
                            return (hb_buffer_content_type_t) (((((hb_buffer_content_type_t)digits[1]) << PyLong_SHIFT) | (hb_buffer_content_type_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(hb_buffer_content_type_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_buffer_content_type_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_buffer_content_type_t) >= 3 * PyLong_SHIFT)) {
                            return (hb_buffer_content_type_t) (((((((hb_buffer_content_type_t)digits[2]) << PyLong_SHIFT) | (hb_buffer_content_type_t)digits[1]) << PyLong_SHIFT) | (hb_buffer_content_type_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(hb_buffer_content_type_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_buffer_content_type_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_buffer_content_type_t) >= 4 * PyLong_SHIFT)) {
                            return (hb_buffer_content_type_t) (((((((((hb_buffer_content_type_t)digits[3]) << PyLong_SHIFT) | (hb_buffer_content_type_t)digits[2]) << PyLong_SHIFT) | (hb_buffer_content_type_t)digits[1]) << PyLong_SHIFT) | (hb_buffer_content_type_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (hb_buffer_content_type_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(hb_buffer_content_type_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_buffer_content_type_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(hb_buffer_content_type_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_buffer_content_type_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(hb_buffer_content_type_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(hb_buffer_content_type_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_buffer_content_type_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_buffer_content_type_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (hb_buffer_content_type_t) (((hb_buffer_content_type_t)-1)*(((((hb_buffer_content_type_t)digits[1]) << PyLong_SHIFT) | (hb_buffer_content_type_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(hb_buffer_content_type_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_buffer_content_type_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_buffer_content_type_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (hb_buffer_content_type_t) ((((((hb_buffer_content_type_t)digits[1]) << PyLong_SHIFT) | (hb_buffer_content_type_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(hb_buffer_content_type_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_buffer_content_type_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_buffer_content_type_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (hb_buffer_content_type_t) (((hb_buffer_content_type_t)-1)*(((((((hb_buffer_content_type_t)digits[2]) << PyLong_SHIFT) | (hb_buffer_content_type_t)digits[1]) << PyLong_SHIFT) | (hb_buffer_content_type_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(hb_buffer_content_type_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_buffer_content_type_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_buffer_content_type_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (hb_buffer_content_type_t) ((((((((hb_buffer_content_type_t)digits[2]) << PyLong_SHIFT) | (hb_buffer_content_type_t)digits[1]) << PyLong_SHIFT) | (hb_buffer_content_type_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(hb_buffer_content_type_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_buffer_content_type_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_buffer_content_type_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (hb_buffer_content_type_t) (((hb_buffer_content_type_t)-1)*(((((((((hb_buffer_content_type_t)digits[3]) << PyLong_SHIFT) | (hb_buffer_content_type_t)digits[2]) << PyLong_SHIFT) | (hb_buffer_content_type_t)digits[1]) << PyLong_SHIFT) | (hb_buffer_content_type_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(hb_buffer_content_type_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_buffer_content_type_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_buffer_content_type_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (hb_buffer_content_type_t) ((((((((((hb_buffer_content_type_t)digits[3]) << PyLong_SHIFT) | (hb_buffer_content_type_t)digits[2]) << PyLong_SHIFT) | (hb_buffer_content_type_t)digits[1]) << PyLong_SHIFT) | (hb_buffer_content_type_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(hb_buffer_content_type_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_buffer_content_type_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(hb_buffer_content_type_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_buffer_content_type_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        hb_buffer_content_type_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyErr_SetString(PyExc_RuntimeError,
                        "_PyLong_AsByteArray() or PyLong_AsNativeBytes() not available, cannot convert large enums");
        val = (hb_buffer_content_type_t) -1;
#endif
        if (unlikely(ret))
            return (hb_buffer_content_type_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to hb_buffer_content_type_t");
    return (hb_buffer_content_type_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to hb_buffer_content_type_t");
    return (hb_buffer_content_type_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_bool_t(hb_bool_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hb_bool_t neg_one = (hb_bool_t) -1, const_zero = (hb_bool_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(hb_bool_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(hb_bool_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_bool_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(hb_bool_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_bool_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(hb_bool_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(hb_bool_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE hb_bool_t __Pyx_PyInt_As_hb_bool_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hb_bool_t neg_one = (hb_bool_t) -1, const_zero = (hb_bool_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(hb_bool_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(hb_bool_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (hb_bool_t) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        hb_bool_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (hb_bool_t) -1;
        val = __Pyx_PyInt_As_hb_bool_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(hb_bool_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(hb_bool_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_bool_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_bool_t) >= 2 * PyLong_SHIFT)) {
                            return (hb_bool_t) (((((hb_bool_t)digits[1]) << PyLong_SHIFT) | (hb_bool_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(hb_bool_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_bool_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_bool_t) >= 3 * PyLong_SHIFT)) {
                            return (hb_bool_t) (((((((hb_bool_t)digits[2]) << PyLong_SHIFT) | (hb_bool_t)digits[1]) << PyLong_SHIFT) | (hb_bool_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(hb_bool_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_bool_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_bool_t) >= 4 * PyLong_SHIFT)) {
                            return (hb_bool_t) (((((((((hb_bool_t)digits[3]) << PyLong_SHIFT) | (hb_bool_t)digits[2]) << PyLong_SHIFT) | (hb_bool_t)digits[1]) << PyLong_SHIFT) | (hb_bool_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (hb_bool_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(hb_bool_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_bool_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(hb_bool_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_bool_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(hb_bool_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(hb_bool_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_bool_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_bool_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (hb_bool_t) (((hb_bool_t)-1)*(((((hb_bool_t)digits[1]) << PyLong_SHIFT) | (hb_bool_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(hb_bool_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_bool_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_bool_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (hb_bool_t) ((((((hb_bool_t)digits[1]) << PyLong_SHIFT) | (hb_bool_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(hb_bool_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_bool_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_bool_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (hb_bool_t) (((hb_bool_t)-1)*(((((((hb_bool_t)digits[2]) << PyLong_SHIFT) | (hb_bool_t)digits[1]) << PyLong_SHIFT) | (hb_bool_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(hb_bool_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_bool_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_bool_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (hb_bool_t) ((((((((hb_bool_t)digits[2]) << PyLong_SHIFT) | (hb_bool_t)digits[1]) << PyLong_SHIFT) | (hb_bool_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(hb_bool_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_bool_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_bool_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (hb_bool_t) (((hb_bool_t)-1)*(((((((((hb_bool_t)digits[3]) << PyLong_SHIFT) | (hb_bool_t)digits[2]) << PyLong_SHIFT) | (hb_bool_t)digits[1]) << PyLong_SHIFT) | (hb_bool_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(hb_bool_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_bool_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_bool_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (hb_bool_t) ((((((((((hb_bool_t)digits[3]) << PyLong_SHIFT) | (hb_bool_t)digits[2]) << PyLong_SHIFT) | (hb_bool_t)digits[1]) << PyLong_SHIFT) | (hb_bool_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(hb_bool_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_bool_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(hb_bool_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_bool_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        hb_bool_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (hb_bool_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (hb_bool_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (hb_bool_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (hb_bool_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(hb_bool_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((hb_bool_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(hb_bool_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((hb_bool_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((hb_bool_t) 1) << (sizeof(hb_bool_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (hb_bool_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to hb_bool_t");
    return (hb_bool_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to hb_bool_t");
    return (hb_bool_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE uint32_t __Pyx_PyInt_As_uint32_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint32_t neg_one = (uint32_t) -1, const_zero = (uint32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(uint32_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(uint32_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint32_t) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        uint32_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint32_t) -1;
        val = __Pyx_PyInt_As_uint32_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint32_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(uint32_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) >= 2 * PyLong_SHIFT)) {
                            return (uint32_t) (((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint32_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) >= 3 * PyLong_SHIFT)) {
                            return (uint32_t) (((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint32_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) >= 4 * PyLong_SHIFT)) {
                            return (uint32_t) (((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (uint32_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(uint32_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint32_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(uint32_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint32_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint32_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(uint32_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint32_t) (((uint32_t)-1)*(((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(uint32_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint32_t) ((((((uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(uint32_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint32_t) (((uint32_t)-1)*(((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint32_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint32_t) ((((((((uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(uint32_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint32_t) (((uint32_t)-1)*(((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint32_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint32_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint32_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint32_t) ((((((((((uint32_t)digits[3]) << PyLong_SHIFT) | (uint32_t)digits[2]) << PyLong_SHIFT) | (uint32_t)digits[1]) << PyLong_SHIFT) | (uint32_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(uint32_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint32_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(uint32_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint32_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        uint32_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (uint32_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (uint32_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (uint32_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (uint32_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(uint32_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((uint32_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(uint32_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((uint32_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((uint32_t) 1) << (sizeof(uint32_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (uint32_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint32_t");
    return (uint32_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint32_t");
    return (uint32_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE hb_position_t __Pyx_PyInt_As_hb_position_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hb_position_t neg_one = (hb_position_t) -1, const_zero = (hb_position_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(hb_position_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(hb_position_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (hb_position_t) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        hb_position_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (hb_position_t) -1;
        val = __Pyx_PyInt_As_hb_position_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(hb_position_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(hb_position_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_position_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_position_t) >= 2 * PyLong_SHIFT)) {
                            return (hb_position_t) (((((hb_position_t)digits[1]) << PyLong_SHIFT) | (hb_position_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(hb_position_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_position_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_position_t) >= 3 * PyLong_SHIFT)) {
                            return (hb_position_t) (((((((hb_position_t)digits[2]) << PyLong_SHIFT) | (hb_position_t)digits[1]) << PyLong_SHIFT) | (hb_position_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(hb_position_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_position_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_position_t) >= 4 * PyLong_SHIFT)) {
                            return (hb_position_t) (((((((((hb_position_t)digits[3]) << PyLong_SHIFT) | (hb_position_t)digits[2]) << PyLong_SHIFT) | (hb_position_t)digits[1]) << PyLong_SHIFT) | (hb_position_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (hb_position_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(hb_position_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_position_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(hb_position_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_position_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(hb_position_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(hb_position_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_position_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_position_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (hb_position_t) (((hb_position_t)-1)*(((((hb_position_t)digits[1]) << PyLong_SHIFT) | (hb_position_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(hb_position_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_position_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_position_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (hb_position_t) ((((((hb_position_t)digits[1]) << PyLong_SHIFT) | (hb_position_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(hb_position_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_position_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_position_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (hb_position_t) (((hb_position_t)-1)*(((((((hb_position_t)digits[2]) << PyLong_SHIFT) | (hb_position_t)digits[1]) << PyLong_SHIFT) | (hb_position_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(hb_position_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_position_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_position_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (hb_position_t) ((((((((hb_position_t)digits[2]) << PyLong_SHIFT) | (hb_position_t)digits[1]) << PyLong_SHIFT) | (hb_position_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(hb_position_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_position_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_position_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (hb_position_t) (((hb_position_t)-1)*(((((((((hb_position_t)digits[3]) << PyLong_SHIFT) | (hb_position_t)digits[2]) << PyLong_SHIFT) | (hb_position_t)digits[1]) << PyLong_SHIFT) | (hb_position_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(hb_position_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_position_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_position_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (hb_position_t) ((((((((((hb_position_t)digits[3]) << PyLong_SHIFT) | (hb_position_t)digits[2]) << PyLong_SHIFT) | (hb_position_t)digits[1]) << PyLong_SHIFT) | (hb_position_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(hb_position_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_position_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(hb_position_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_position_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        hb_position_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (hb_position_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (hb_position_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (hb_position_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (hb_position_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(hb_position_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((hb_position_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(hb_position_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((hb_position_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((hb_position_t) 1) << (sizeof(hb_position_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (hb_position_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to hb_position_t");
    return (hb_position_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to hb_position_t");
    return (hb_position_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const char neg_one = (char) -1, const_zero = (char) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(char) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(char, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (char) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        char val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (char) -1;
        val = __Pyx_PyInt_As_char(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(char, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(char) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(char) >= 2 * PyLong_SHIFT)) {
                            return (char) (((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(char) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(char) >= 3 * PyLong_SHIFT)) {
                            return (char) (((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(char) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(char) >= 4 * PyLong_SHIFT)) {
                            return (char) (((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (char) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(char) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(char, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(char) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(char, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(char, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(char) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(char) - 1 > 2 * PyLong_SHIFT)) {
                            return (char) (((char)-1)*(((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(char) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(char) - 1 > 2 * PyLong_SHIFT)) {
                            return (char) ((((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(char) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(char) - 1 > 3 * PyLong_SHIFT)) {
                            return (char) (((char)-1)*(((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(char) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(char) - 1 > 3 * PyLong_SHIFT)) {
                            return (char) ((((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(char) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(char) - 1 > 4 * PyLong_SHIFT)) {
                            return (char) (((char)-1)*(((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(char) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(char) - 1 > 4 * PyLong_SHIFT)) {
                            return (char) ((((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(char) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(char, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(char) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(char, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        char val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (char) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (char) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (char) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (char) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(char) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((char) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(char) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((char) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((char) 1) << (sizeof(char) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (char) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to char");
    return (char) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to char");
    return (char) -1;
}

/* CIntFromPy */
static CYTHON_INLINE unsigned long __Pyx_PyInt_As_unsigned_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned long neg_one = (unsigned long) -1, const_zero = (unsigned long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(unsigned long) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(unsigned long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (unsigned long) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        unsigned long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (unsigned long) -1;
        val = __Pyx_PyInt_As_unsigned_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(unsigned long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(unsigned long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) >= 2 * PyLong_SHIFT)) {
                            return (unsigned long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) >= 3 * PyLong_SHIFT)) {
                            return (unsigned long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) >= 4 * PyLong_SHIFT)) {
                            return (unsigned long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (unsigned long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(unsigned long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(unsigned long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(unsigned long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(unsigned long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned long) (((unsigned long)-1)*(((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(unsigned long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) - 1 > 2 * PyLong_SHIFT)) {
                            return (unsigned long) ((((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(unsigned long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned long) (((unsigned long)-1)*(((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) - 1 > 3 * PyLong_SHIFT)) {
                            return (unsigned long) ((((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(unsigned long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned long) (((unsigned long)-1)*(((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(unsigned long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(unsigned long) - 1 > 4 * PyLong_SHIFT)) {
                            return (unsigned long) ((((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(unsigned long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(unsigned long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(unsigned long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        unsigned long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (unsigned long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (unsigned long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (unsigned long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (unsigned long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(unsigned long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((unsigned long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(unsigned long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((unsigned long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((unsigned long) 1) << (sizeof(unsigned long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (unsigned long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned long");
    return (unsigned long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned long");
    return (unsigned long) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint32_t(uint32_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint32_t neg_one = (uint32_t) -1, const_zero = (uint32_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint32_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint32_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint32_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint32_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint32_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(uint32_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(uint32_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_hb_ot_name_id_t(hb_ot_name_id_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hb_ot_name_id_t neg_one = (hb_ot_name_id_t) -1, const_zero = (hb_ot_name_id_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(hb_ot_name_id_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(hb_ot_name_id_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_ot_name_id_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(hb_ot_name_id_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(hb_ot_name_id_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(hb_ot_name_id_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(hb_ot_name_id_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE hb_ot_math_constant_t __Pyx_PyInt_As_hb_ot_math_constant_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hb_ot_math_constant_t neg_one = (hb_ot_math_constant_t) -1, const_zero = (hb_ot_math_constant_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(hb_ot_math_constant_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(hb_ot_math_constant_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (hb_ot_math_constant_t) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        hb_ot_math_constant_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (hb_ot_math_constant_t) -1;
        val = __Pyx_PyInt_As_hb_ot_math_constant_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(hb_ot_math_constant_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(hb_ot_math_constant_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_ot_math_constant_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_ot_math_constant_t) >= 2 * PyLong_SHIFT)) {
                            return (hb_ot_math_constant_t) (((((hb_ot_math_constant_t)digits[1]) << PyLong_SHIFT) | (hb_ot_math_constant_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(hb_ot_math_constant_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_ot_math_constant_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_ot_math_constant_t) >= 3 * PyLong_SHIFT)) {
                            return (hb_ot_math_constant_t) (((((((hb_ot_math_constant_t)digits[2]) << PyLong_SHIFT) | (hb_ot_math_constant_t)digits[1]) << PyLong_SHIFT) | (hb_ot_math_constant_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(hb_ot_math_constant_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_ot_math_constant_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_ot_math_constant_t) >= 4 * PyLong_SHIFT)) {
                            return (hb_ot_math_constant_t) (((((((((hb_ot_math_constant_t)digits[3]) << PyLong_SHIFT) | (hb_ot_math_constant_t)digits[2]) << PyLong_SHIFT) | (hb_ot_math_constant_t)digits[1]) << PyLong_SHIFT) | (hb_ot_math_constant_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (hb_ot_math_constant_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(hb_ot_math_constant_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_ot_math_constant_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(hb_ot_math_constant_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_ot_math_constant_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(hb_ot_math_constant_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(hb_ot_math_constant_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_ot_math_constant_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_ot_math_constant_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (hb_ot_math_constant_t) (((hb_ot_math_constant_t)-1)*(((((hb_ot_math_constant_t)digits[1]) << PyLong_SHIFT) | (hb_ot_math_constant_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(hb_ot_math_constant_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_ot_math_constant_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_ot_math_constant_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (hb_ot_math_constant_t) ((((((hb_ot_math_constant_t)digits[1]) << PyLong_SHIFT) | (hb_ot_math_constant_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(hb_ot_math_constant_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_ot_math_constant_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_ot_math_constant_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (hb_ot_math_constant_t) (((hb_ot_math_constant_t)-1)*(((((((hb_ot_math_constant_t)digits[2]) << PyLong_SHIFT) | (hb_ot_math_constant_t)digits[1]) << PyLong_SHIFT) | (hb_ot_math_constant_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(hb_ot_math_constant_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_ot_math_constant_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_ot_math_constant_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (hb_ot_math_constant_t) ((((((((hb_ot_math_constant_t)digits[2]) << PyLong_SHIFT) | (hb_ot_math_constant_t)digits[1]) << PyLong_SHIFT) | (hb_ot_math_constant_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(hb_ot_math_constant_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_ot_math_constant_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_ot_math_constant_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (hb_ot_math_constant_t) (((hb_ot_math_constant_t)-1)*(((((((((hb_ot_math_constant_t)digits[3]) << PyLong_SHIFT) | (hb_ot_math_constant_t)digits[2]) << PyLong_SHIFT) | (hb_ot_math_constant_t)digits[1]) << PyLong_SHIFT) | (hb_ot_math_constant_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(hb_ot_math_constant_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_ot_math_constant_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_ot_math_constant_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (hb_ot_math_constant_t) ((((((((((hb_ot_math_constant_t)digits[3]) << PyLong_SHIFT) | (hb_ot_math_constant_t)digits[2]) << PyLong_SHIFT) | (hb_ot_math_constant_t)digits[1]) << PyLong_SHIFT) | (hb_ot_math_constant_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(hb_ot_math_constant_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_ot_math_constant_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(hb_ot_math_constant_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_ot_math_constant_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        hb_ot_math_constant_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyErr_SetString(PyExc_RuntimeError,
                        "_PyLong_AsByteArray() or PyLong_AsNativeBytes() not available, cannot convert large enums");
        val = (hb_ot_math_constant_t) -1;
#endif
        if (unlikely(ret))
            return (hb_ot_math_constant_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to hb_ot_math_constant_t");
    return (hb_ot_math_constant_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to hb_ot_math_constant_t");
    return (hb_ot_math_constant_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE hb_ot_math_kern_t __Pyx_PyInt_As_hb_ot_math_kern_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hb_ot_math_kern_t neg_one = (hb_ot_math_kern_t) -1, const_zero = (hb_ot_math_kern_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(hb_ot_math_kern_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(hb_ot_math_kern_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (hb_ot_math_kern_t) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        hb_ot_math_kern_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (hb_ot_math_kern_t) -1;
        val = __Pyx_PyInt_As_hb_ot_math_kern_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(hb_ot_math_kern_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(hb_ot_math_kern_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_ot_math_kern_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_ot_math_kern_t) >= 2 * PyLong_SHIFT)) {
                            return (hb_ot_math_kern_t) (((((hb_ot_math_kern_t)digits[1]) << PyLong_SHIFT) | (hb_ot_math_kern_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(hb_ot_math_kern_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_ot_math_kern_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_ot_math_kern_t) >= 3 * PyLong_SHIFT)) {
                            return (hb_ot_math_kern_t) (((((((hb_ot_math_kern_t)digits[2]) << PyLong_SHIFT) | (hb_ot_math_kern_t)digits[1]) << PyLong_SHIFT) | (hb_ot_math_kern_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(hb_ot_math_kern_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_ot_math_kern_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_ot_math_kern_t) >= 4 * PyLong_SHIFT)) {
                            return (hb_ot_math_kern_t) (((((((((hb_ot_math_kern_t)digits[3]) << PyLong_SHIFT) | (hb_ot_math_kern_t)digits[2]) << PyLong_SHIFT) | (hb_ot_math_kern_t)digits[1]) << PyLong_SHIFT) | (hb_ot_math_kern_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (hb_ot_math_kern_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(hb_ot_math_kern_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_ot_math_kern_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(hb_ot_math_kern_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_ot_math_kern_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(hb_ot_math_kern_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(hb_ot_math_kern_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_ot_math_kern_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_ot_math_kern_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (hb_ot_math_kern_t) (((hb_ot_math_kern_t)-1)*(((((hb_ot_math_kern_t)digits[1]) << PyLong_SHIFT) | (hb_ot_math_kern_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(hb_ot_math_kern_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_ot_math_kern_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_ot_math_kern_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (hb_ot_math_kern_t) ((((((hb_ot_math_kern_t)digits[1]) << PyLong_SHIFT) | (hb_ot_math_kern_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(hb_ot_math_kern_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_ot_math_kern_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_ot_math_kern_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (hb_ot_math_kern_t) (((hb_ot_math_kern_t)-1)*(((((((hb_ot_math_kern_t)digits[2]) << PyLong_SHIFT) | (hb_ot_math_kern_t)digits[1]) << PyLong_SHIFT) | (hb_ot_math_kern_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(hb_ot_math_kern_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_ot_math_kern_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_ot_math_kern_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (hb_ot_math_kern_t) ((((((((hb_ot_math_kern_t)digits[2]) << PyLong_SHIFT) | (hb_ot_math_kern_t)digits[1]) << PyLong_SHIFT) | (hb_ot_math_kern_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(hb_ot_math_kern_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_ot_math_kern_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_ot_math_kern_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (hb_ot_math_kern_t) (((hb_ot_math_kern_t)-1)*(((((((((hb_ot_math_kern_t)digits[3]) << PyLong_SHIFT) | (hb_ot_math_kern_t)digits[2]) << PyLong_SHIFT) | (hb_ot_math_kern_t)digits[1]) << PyLong_SHIFT) | (hb_ot_math_kern_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(hb_ot_math_kern_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_ot_math_kern_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_ot_math_kern_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (hb_ot_math_kern_t) ((((((((((hb_ot_math_kern_t)digits[3]) << PyLong_SHIFT) | (hb_ot_math_kern_t)digits[2]) << PyLong_SHIFT) | (hb_ot_math_kern_t)digits[1]) << PyLong_SHIFT) | (hb_ot_math_kern_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(hb_ot_math_kern_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_ot_math_kern_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(hb_ot_math_kern_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_ot_math_kern_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        hb_ot_math_kern_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyErr_SetString(PyExc_RuntimeError,
                        "_PyLong_AsByteArray() or PyLong_AsNativeBytes() not available, cannot convert large enums");
        val = (hb_ot_math_kern_t) -1;
#endif
        if (unlikely(ret))
            return (hb_ot_math_kern_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to hb_ot_math_kern_t");
    return (hb_ot_math_kern_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to hb_ot_math_kern_t");
    return (hb_ot_math_kern_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE uint8_t __Pyx_PyInt_As_uint8_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint8_t neg_one = (uint8_t) -1, const_zero = (uint8_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(uint8_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(uint8_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (uint8_t) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        uint8_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (uint8_t) -1;
        val = __Pyx_PyInt_As_uint8_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint8_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(uint8_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) >= 2 * PyLong_SHIFT)) {
                            return (uint8_t) (((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint8_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) >= 3 * PyLong_SHIFT)) {
                            return (uint8_t) (((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint8_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) >= 4 * PyLong_SHIFT)) {
                            return (uint8_t) (((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (uint8_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(uint8_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint8_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(uint8_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint8_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(uint8_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(uint8_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint8_t) (((uint8_t)-1)*(((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(uint8_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (uint8_t) ((((((uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(uint8_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint8_t) (((uint8_t)-1)*(((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(uint8_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (uint8_t) ((((((((uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(uint8_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint8_t) (((uint8_t)-1)*(((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(uint8_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(uint8_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(uint8_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (uint8_t) ((((((((((uint8_t)digits[3]) << PyLong_SHIFT) | (uint8_t)digits[2]) << PyLong_SHIFT) | (uint8_t)digits[1]) << PyLong_SHIFT) | (uint8_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(uint8_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint8_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(uint8_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(uint8_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        uint8_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (uint8_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (uint8_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (uint8_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (uint8_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(uint8_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((uint8_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(uint8_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((uint8_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((uint8_t) 1) << (sizeof(uint8_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (uint8_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to uint8_t");
    return (uint8_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to uint8_t");
    return (uint8_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_uint8_t(uint8_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const uint8_t neg_one = (uint8_t) -1, const_zero = (uint8_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(uint8_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(uint8_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint8_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(uint8_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(uint8_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(uint8_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(uint8_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(size_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(size_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(size_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 2 * PyLong_SHIFT)) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(size_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 3 * PyLong_SHIFT)) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(size_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 4 * PyLong_SHIFT)) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (size_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(size_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(size_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(size_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(size_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(size_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(size_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(size_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        size_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (size_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (size_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (size_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (size_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(size_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((size_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(size_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((size_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((size_t) 1) << (sizeof(size_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (size_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE hb_subset_sets_t __Pyx_PyInt_As_hb_subset_sets_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const hb_subset_sets_t neg_one = (hb_subset_sets_t) -1, const_zero = (hb_subset_sets_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(hb_subset_sets_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(hb_subset_sets_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (hb_subset_sets_t) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        hb_subset_sets_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (hb_subset_sets_t) -1;
        val = __Pyx_PyInt_As_hb_subset_sets_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(hb_subset_sets_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(hb_subset_sets_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_subset_sets_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_subset_sets_t) >= 2 * PyLong_SHIFT)) {
                            return (hb_subset_sets_t) (((((hb_subset_sets_t)digits[1]) << PyLong_SHIFT) | (hb_subset_sets_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(hb_subset_sets_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_subset_sets_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_subset_sets_t) >= 3 * PyLong_SHIFT)) {
                            return (hb_subset_sets_t) (((((((hb_subset_sets_t)digits[2]) << PyLong_SHIFT) | (hb_subset_sets_t)digits[1]) << PyLong_SHIFT) | (hb_subset_sets_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(hb_subset_sets_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_subset_sets_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_subset_sets_t) >= 4 * PyLong_SHIFT)) {
                            return (hb_subset_sets_t) (((((((((hb_subset_sets_t)digits[3]) << PyLong_SHIFT) | (hb_subset_sets_t)digits[2]) << PyLong_SHIFT) | (hb_subset_sets_t)digits[1]) << PyLong_SHIFT) | (hb_subset_sets_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (hb_subset_sets_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(hb_subset_sets_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_subset_sets_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(hb_subset_sets_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_subset_sets_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(hb_subset_sets_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(hb_subset_sets_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_subset_sets_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_subset_sets_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (hb_subset_sets_t) (((hb_subset_sets_t)-1)*(((((hb_subset_sets_t)digits[1]) << PyLong_SHIFT) | (hb_subset_sets_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(hb_subset_sets_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_subset_sets_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_subset_sets_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (hb_subset_sets_t) ((((((hb_subset_sets_t)digits[1]) << PyLong_SHIFT) | (hb_subset_sets_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(hb_subset_sets_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_subset_sets_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_subset_sets_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (hb_subset_sets_t) (((hb_subset_sets_t)-1)*(((((((hb_subset_sets_t)digits[2]) << PyLong_SHIFT) | (hb_subset_sets_t)digits[1]) << PyLong_SHIFT) | (hb_subset_sets_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(hb_subset_sets_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_subset_sets_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_subset_sets_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (hb_subset_sets_t) ((((((((hb_subset_sets_t)digits[2]) << PyLong_SHIFT) | (hb_subset_sets_t)digits[1]) << PyLong_SHIFT) | (hb_subset_sets_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(hb_subset_sets_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_subset_sets_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_subset_sets_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (hb_subset_sets_t) (((hb_subset_sets_t)-1)*(((((((((hb_subset_sets_t)digits[3]) << PyLong_SHIFT) | (hb_subset_sets_t)digits[2]) << PyLong_SHIFT) | (hb_subset_sets_t)digits[1]) << PyLong_SHIFT) | (hb_subset_sets_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(hb_subset_sets_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(hb_subset_sets_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(hb_subset_sets_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (hb_subset_sets_t) ((((((((((hb_subset_sets_t)digits[3]) << PyLong_SHIFT) | (hb_subset_sets_t)digits[2]) << PyLong_SHIFT) | (hb_subset_sets_t)digits[1]) << PyLong_SHIFT) | (hb_subset_sets_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(hb_subset_sets_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_subset_sets_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(hb_subset_sets_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(hb_subset_sets_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        hb_subset_sets_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyErr_SetString(PyExc_RuntimeError,
                        "_PyLong_AsByteArray() or PyLong_AsNativeBytes() not available, cannot convert large enums");
        val = (hb_subset_sets_t) -1;
#endif
        if (unlikely(ret))
            return (hb_subset_sets_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to hb_subset_sets_t");
    return (hb_subset_sets_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to hb_subset_sets_t");
    return (hb_subset_sets_t) -1;
}

/* FormatTypeName */
#if CYTHON_COMPILING_IN_LIMITED_API
static __Pyx_TypeName
__Pyx_PyType_GetName(PyTypeObject* tp)
{
    PyObject *name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_n_s_name_2);
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) {
        PyErr_Clear();
        Py_XDECREF(name);
        name = __Pyx_NewRef(__pyx_n_s__287);
    }
    return name;
}
#endif

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(long) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CoroutineBase */
#include <frameobject.h>
#if PY_VERSION_HEX >= 0x030b00a6
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    CYTHON_UNUSED_VAR(__pyx_tstate);
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (likely(__Pyx_IS_TYPE(ev, (PyTypeObject*)PyExc_StopIteration))) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_CLEAR(exc_state->exc_value);
#else
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
#endif
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(__pyx_CoroutineObject *gen) {
    const char *msg;
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(PyObject *gen) {
    const char *msg;
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(PyObject *gen, PyObject *value, int closing) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    CYTHON_MAYBE_UNUSED_VAR(closing);
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_value) {
        #if CYTHON_COMPILING_IN_PYPY
        #else
        PyObject *exc_tb;
        #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_CPYTHON
        exc_tb = PyException_GetTraceback(exc_state->exc_value);
        #elif PY_VERSION_HEX >= 0x030B00a4
        exc_tb = ((PyBaseExceptionObject*) exc_state->exc_value)->traceback;
        #else
        exc_tb = exc_state->exc_traceback;
        #endif
        if (exc_tb) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
            PyFrameObject *f = tb->tb_frame;
            assert(f->f_back == NULL);
            #if PY_VERSION_HEX >= 0x030B00A1
            f->f_back = PyThreadState_GetFrame(tstate);
            #else
            Py_XINCREF(tstate->frame);
            f->f_back = tstate->frame;
            #endif
            #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_CPYTHON
            Py_DECREF(exc_tb);
            #endif
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body(self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
#if CYTHON_COMPILING_IN_PYPY
    CYTHON_UNUSED_VAR(exc_state);
#else
    PyObject *exc_tb;
    #if PY_VERSION_HEX >= 0x030B00a4
    if (!exc_state->exc_value) return;
    exc_tb = PyException_GetTraceback(exc_state->exc_value);
    #else
    exc_tb = exc_state->exc_traceback;
    #endif
    if (likely(exc_tb)) {
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
        #if PY_VERSION_HEX >= 0x030B00a4
        Py_DECREF(exc_tb);
        #endif
    }
#endif
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(PyObject* gen, PyObject *retval) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
#if PY_VERSION_HEX < 0x030d00A1
            _PyGen_SetStopIterationValue(result);
#else
            if (!PyTuple_Check(result) && !PyExceptionInstance_Check(result)) {
                PyErr_SetObject(PyExc_StopIteration, result);
            } else {
                PyObject *exc = __Pyx_PyObject_CallOneArg(PyExc_StopIteration, result);
                if (likely(exc != NULL)) {
                    PyErr_SetObject(PyExc_StopIteration, exc);
                    Py_DECREF(exc);
                }
            }
#endif
        }
        Py_DECREF(result);
        result = NULL;
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = __Pyx_PyObject_GetIterNextFunc(yf)(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (unlikely(PyErr_Occurred())) {
                PyErr_WriteUnraisable(yf);
            }
        } else {
            retval = __Pyx_PyObject_CallNoArg(meth);
            Py_DECREF(meth);
            if (unlikely(!retval))
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = __Pyx_PyObject_GetIterNextFunc(yf)(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, PyObject *arg) {
    CYTHON_UNUSED_VAR(arg);
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (unlikely(PyErr_Occurred())) {
                    gen->is_running = 0;
                    return NULL;
                }
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = __Pyx_PyObject_Call(meth, args, NULL);
            } else {
                PyObject *cargs[4] = {NULL, typ, val, tb};
                ret = __Pyx_PyObject_FastCall(meth, cargs+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (unlikely(!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb)))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_VISIT(exc_state->exc_value);
#else
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
#endif
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_frame);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (unlikely(PyObject_CallFinalizerFromDealloc(self)))
#else
        Py_TYPE(gen)->tp_del(self);
        if (unlikely(Py_REFCNT(self) > 0))
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    __Pyx_PyHeapTypeObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    __Pyx_SET_REFCNT(self, 1);
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(Py_REFCNT(self) > 0);
    if (likely(--self->ob_refcnt == 0)) {
        return;
    }
    {
        Py_ssize_t refcnt = Py_REFCNT(self);
        _Py_NewReference(self);
        __Pyx_SET_REFCNT(self, refcnt);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(Py_TYPE(self)) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_name;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_name, value);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_qualname;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_qualname, value);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_frame(__pyx_CoroutineObject *self, void *context)
{
    PyObject *frame = self->gi_frame;
    CYTHON_UNUSED_VAR(context);
    if (!frame) {
        if (unlikely(!self->gi_code)) {
            Py_RETURN_NONE;
        }
        frame = (PyObject *) PyFrame_New(
            PyThreadState_Get(),            /*PyThreadState *tstate,*/
            (PyCodeObject*) self->gi_code,  /*PyCodeObject *code,*/
            __pyx_d,                 /*PyObject *globals,*/
            0                               /*PyObject *locals*/
        );
        if (unlikely(!frame))
            return NULL;
        self->gi_frame = frame;
    }
    Py_INCREF(frame);
    return frame;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    #if PY_VERSION_HEX >= 0x030B00a4
    gen->gi_exc_state.exc_value = NULL;
    #else
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
    #endif
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    gen->gi_frame = NULL;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
#ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (unlikely(!module)) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {(char *) "__module__", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_modulename), 0, 0},
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CoroutineObject, gi_weakreflist), READONLY, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {(char *) "gi_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     (char*) PyDoc_STR("Frame of the generator"), 0},
    {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_GeneratorType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_Coroutine_dealloc},
    {Py_tp_traverse, (void *)__Pyx_Coroutine_traverse},
    {Py_tp_iter, (void *)PyObject_SelfIter},
    {Py_tp_iternext, (void *)__Pyx_Generator_Next},
    {Py_tp_methods, (void *)__pyx_Generator_methods},
    {Py_tp_members, (void *)__pyx_Generator_memberlist},
    {Py_tp_getset, (void *)__pyx_Generator_getsets},
    {Py_tp_getattro, (void *) __Pyx_PyObject_GenericGetAttrNoDict},
#if CYTHON_USE_TP_FINALIZE
    {Py_tp_finalize, (void *)__Pyx_Coroutine_del},
#endif
    {0, 0},
};
static PyType_Spec __pyx_GeneratorType_spec = {
    __PYX_TYPE_MODULE_PREFIX "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    __pyx_GeneratorType_slots
};
#else
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if PY_VERSION_HEX >= 0x030d00A4
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
static int __pyx_Generator_init(PyObject *module) {
#if CYTHON_USE_TYPE_SPECS
    __pyx_GeneratorType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_GeneratorType_spec, NULL);
#else
    CYTHON_UNUSED_VAR(module);
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
#endif
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030B00A4
    return Py_Version & ~0xFFUL;
#else
    const char* rt_version = Py_GetVersion();
    unsigned long version = 0;
    unsigned long factor = 0x01000000UL;
    unsigned int digit = 0;
    int i = 0;
    while (factor) {
        while ('0' <= rt_version[i] && rt_version[i] <= '9') {
            digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
            ++i;
        }
        version += factor * digit;
        if (rt_version[i] != '.')
            break;
        digit = 0;
        factor >>= 8;
        ++i;
    }
    return version;
#endif
}
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* InitStrings */
#if PY_MAJOR_VERSION >= 3
static int __Pyx_InitString(__Pyx_StringTabEntry t, PyObject **str) {
    if (t.is_unicode | t.is_str) {
        if (t.intern) {
            *str = PyUnicode_InternFromString(t.s);
        } else if (t.encoding) {
            *str = PyUnicode_Decode(t.s, t.n - 1, t.encoding, NULL);
        } else {
            *str = PyUnicode_FromStringAndSize(t.s, t.n - 1);
        }
    } else {
        *str = PyBytes_FromStringAndSize(t.s, t.n - 1);
    }
    if (!*str)
        return -1;
    if (PyObject_Hash(*str) == -1)
        return -1;
    return 0;
}
#endif
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION >= 3
        __Pyx_InitString(*t, t->p);
        #else
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        #endif
        ++t;
    }
    return 0;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY && !CYTHON_COMPILING_IN_LIMITED_API) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetName(Py_TYPE(result));
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type " __Pyx_FMT_TYPENAME ")",
                 type_name, type_name, result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
